/***********************************************************
 * Copyright 2001,2002,2003: {{{
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * }}}
 **********************************************************/

#include "arm_opcodes.h"
#include "arm_registers.h"
#include "arm_instructions.h"
#include "arm_utils.h"
#include "../../atom_internal.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_instrument.h"
#include "../../utils/util_error.h"
#include "../../diablo_options.h"
#include <string.h>
#if MAX_NR_REG_SUBSETS <= 4
#define OPTIMIZE
#endif

#if 0
extern t_object * inst_object;
#endif
/* ArmPatchReturnSet {{{ */
void ArmPatchReturnSet(t_object * obj)
{
  t_bbl * bbl;
  t_ins * ins;
  t_cfg_edge * edge;
  t_cfg * cfg=T_CFG(obj->code[0]->data);
  t_reloc * rel;
  t_address null;
  CFG_FOREACH_BBL(cfg,bbl)
  {
    BBL_FOREACH_INS(bbl,ins)
    {
      if (ARMINSOPCODE(ins)==ARM_MOV)
	if (ARMINSREGA(ins)==ARM_REG_R14)
	  if (ARMINSREGC(ins)==ARM_REG_R15)
	  {
	    BBL_FOREACH_SUCC_EDGE(bbl,edge)
	    {
	      if (EDGE_CAT(edge)==ET_FUNLINK) break;
	    }
	    if (!edge) continue;

	    GenericSet_uint32(null,0);
	    rel=RelocTableAddRelocToRelocatable(obj->reloc_table,null,T_RELOCATABLE(ins),null,T_RELOCATABLE(CFG_EDGE_TAIL(edge)),null,FALSE,NULL,NULL,NULL,"S\\$",FALSE);

	    ArmMakeAddressProducer(__FILE__, __LINE__,ins, G_T_UINT32(BBL_OLD_ADDRESS(CFG_EDGE_TAIL(edge))), rel);
	  }
    }
  }
}
/* }}} */
/* ArmPatchSwitches {{{ */
void ArmPatchSwitches(t_object * obj)
{
  t_bbl * bbl;
  t_cfg_edge * edge, *edge2;
  t_cfg * cfg=T_CFG(obj->code[0]->data);
  CFG_FOREACH_BBL(T_CFG(obj->code[0]->data),bbl)
  {
    /* skip blocks that are not in functions: 
     * there's no need to keep them correct because they won't be
     * executed anyway */
    if (!BBL_FUN(bbl)) continue;

    BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      if (EDGE_CAT(edge)==ET_SWITCH)
	break;
    }

    if (edge) 
    {
      BBL_FOREACH_SUCC_EDGE(bbl,edge)
      {
	if (EDGE_CAT(edge)==ET_SWITCH)
	{
	  t_bbl * new= AtomMoveBbl(CFG_EDGE_TAIL(edge));
	  t_ins * ins=InsNewForBbl(new);
	  ArmInstructionMakeCondBranch(ins, ARM_CONDITION_AL);
	  CfgEdgeNew(cfg,CFG_EDGE_TAIL(edge),new,ET_JUMP);
	  InsAppendToBbl(ins,CFG_EDGE_TAIL(edge));

	  do
	  {
	    BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),edge2)
	    {
	      if (EDGE_CAT(edge2)==ET_FALLTHROUGH)
		break;
	    }
	    if (edge2)
	    {
	      CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge2),new,EDGE_CAT(edge2));
	      CfgEdgeKill(edge2);
	    }
	  } while(edge2);
	}
	else if (EDGE_CAT(edge)==ET_FALLTHROUGH)
	{
	  t_bbl * new= AtomMoveBbl(CFG_EDGE_TAIL(edge));
	  t_ins * ins=InsNewForBbl(new);
	  ArmInstructionMakeCondBranch(ins, ARM_CONDITION_AL);
	  CfgEdgeNew(cfg,CFG_EDGE_TAIL(edge),new,ET_JUMP);
	  InsAppendToBbl(ins,CFG_EDGE_TAIL(edge));
	}
      }
    }

  }
}
/* }}} */

/* {{{ ArmAtomAddSave */
#ifdef AVOID_SAVES
#endif
t_uint32 ArmAtomAddSave(t_bbl * in, t_regset liveregs, t_regset changedregs) 
{
  t_ins * save, * save2, * mrs, * reload;
  t_uint32 nsave=0;
#ifdef AVOID_SAVES
  t_reg trans[32];
  t_reg scratch;
  t_reg spill;
  t_regset scratchregs=in->fun->cfg->sec->obj->description->int_registers & (~liveregs) & (in->fun->cfg->sec->obj->description->callee_saved) &(~ARM_REG_R15) & (~ARM_REG_R13);
#endif
#ifdef OPTIMIZE
  t_regset set=liveregs & in->fun->cfg->sec->obj->description->callee_may_change & (~((1<<ARM_REG_R15)|(1<<ARM_REG_R13)));
  t_regset orig_set=set;

#else
  t_regset set=AllRegs();
  t_regset orig_set=set;
#endif
#ifdef AVOID_SAVES
  {
    //DiabloPrint(stdout,"BBL = @B\n",in);
    while ((!RegsetIsEmpty(scratchregs))&&(!RegsetIsEmpty(set&0xffff)))
    {
      REGSET_FOREACH_REG(scratchregs,scratch)
	break;

      REGSET_FOREACH_REG(set&0xffff,spill)
	break;

      RegsetSubReg(scratchregs,scratch);
      RegsetSubReg(set,spill);
      save = InsNewForBbl(in);
      ArmInstructionMakeMov(save,scratch,spill,0,ARM_CONDITION_AL);
      InsAppendToBbl(save,in);
      trans[spill]=scratch;
    }
    //DiabloPrint(stdout,"BBL = @B\n",in);
  }
#endif

#ifdef ANGEL_HACK
    save= InsNewForBbl(in);
    ArmInstructionMakeAdd(save,ARM_REG_R13,ARM_REG_R13,ARM_REG_NONE,0x1000000,ARM_CONDITION_AL);
    InsAppendToBbl(save,in);
#endif

  if(!RegsetIsEmpty(set & (in->fun->cfg->sec->obj->description->int_registers) & (~((1<<ARM_REG_R15)|(1<<ARM_REG_R13)))))
  {
    save= InsNewForBbl(in);
    if(RegsetCountRegs(set & (in->fun->cfg->sec->obj->description->int_registers) & (~((1<<ARM_REG_R15)|(1<<ARM_REG_R13)))) > 1)
      ArmInstructionMakeStm(save,ARM_REG_R13,0xffff&set&(~((1<<ARM_REG_R15)|(1<<ARM_REG_R13))),ARM_CONDITION_AL,TRUE,TRUE,FALSE);
    else
    {
      t_reg iter_reg;
      REGSET_FOREACH_REG(set,iter_reg) break;
      ArmInstructionMakeStr(save,iter_reg,ARM_REG_R13,ARM_REG_NONE,4,ARM_CONDITION_AL,TRUE,FALSE,TRUE);
    }
    nsave+=RegsetCountRegs(0xffff&set&(~((1<<ARM_REG_R15)|(1<<ARM_REG_R13))))*4;
    InsAppendToBbl(save,in);
  }

  if (set & ((1<<ARM_REG_C_CONDITION) | (1<<ARM_REG_V_CONDITION) | (1<<ARM_REG_Z_CONDITION) | (1<<ARM_REG_N_CONDITION)))
  {
    mrs=InsNewForBbl(in);
    ArmInstructionMakeMrs(mrs,ARM_REG_R0,ARM_CONDITION_AL);
    InsAppendToBbl(mrs,in);

    save2=InsNewForBbl(in);
    ArmInstructionMakeStr(save2,ARM_REG_R0,ARM_REG_R13,ARM_REG_NONE,4,ARM_CONDITION_AL,TRUE,FALSE,TRUE);
    InsAppendToBbl(save2,in);
    nsave+=4;

    if (orig_set & ((1<<ARM_REG_R0))) {
      if (set & ((1<<ARM_REG_R0))) {
	reload=InsNewForBbl(in);
	ArmInstructionMakeLdr(reload,ARM_REG_R0,ARM_REG_R13,ARM_REG_NONE,4,ARM_CONDITION_AL,TRUE,TRUE,FALSE);
	InsAppendToBbl(reload,in);
      }
      else
      {
#ifdef AVOID_SAVES
	reload=InsNewForBbl(in);
	ArmInstructionMakeMov(reload,ARM_REG_R0,trans[ARM_REG_R0],0,ARM_CONDITION_AL);
	InsAppendToBbl(reload,in);
#else
	FATAL(("Impossible!"));
#endif
      }
    }
  }
  return nsave;
} /* }}} */

/* ArmAtomAddRestore {{{ */ 
void ArmAtomAddRestore(t_bbl * out,t_regset liveregs,t_regset changedregs,t_uint32 stack_change) 
{
#ifdef OPTIMIZE
  t_regset set=liveregs & out->fun->cfg->sec->obj->description->callee_may_change  & (~((1<<ARM_REG_R15)|(1<<ARM_REG_R13)));
#else
  t_regset set=0xffffffff;
#endif
#ifdef AVOID_SAVES
  t_regset scratchregs=out->fun->cfg->sec->obj->description->int_registers & (~liveregs) & (out->fun->cfg->sec->obj->description->callee_saved) &(~ARM_REG_R15) & (~ARM_REG_R13) ;
  t_regset orig_set=set;
  t_reg scratch;
  t_reg spill;
  t_bool needsave=!RegsetIsEmpty(set & (out->fun->cfg->sec->obj->description->int_registers) & (~((1<<ARM_REG_R15)|(1<<ARM_REG_R13))));
#endif

  t_ins * restore, * msr, *restore2 ;

#ifdef AVOID_SAVES  
  {
    while ((!RegsetIsEmpty(scratchregs))&&(!RegsetIsEmpty(set&0xffff)))
    {
      REGSET_FOREACH_REG(scratchregs,scratch)
	break;

      REGSET_FOREACH_REG(set&0xffff,spill)
	break;

      RegsetSubReg(scratchregs,scratch);
      RegsetSubReg(set,spill);
      restore = InsNewForBbl(out);
      ArmInstructionMakeMov(restore,spill,scratch,0,ARM_CONDITION_AL);
      InsPrependToBbl(restore,out);
    }
  }
#endif
#ifdef ANGEL_HACK
    restore= InsNewForBbl(out);
    ArmInstructionMakeSub(restore,ARM_REG_R13,ARM_REG_R13,ARM_REG_NONE,0x1000000,ARM_CONDITION_AL);
    InsAppendToBbl(restore,out);
#endif


  if(!RegsetIsEmpty(set & (out->fun->cfg->sec->obj->description->int_registers) & (~((1<<ARM_REG_R15)|(1<<ARM_REG_R13)))))
  {
    restore = InsNewForBbl(out);
    if(RegsetCountRegs(set & (out->fun->cfg->sec->obj->description->int_registers) & (~((1<<ARM_REG_R15)|(1<<ARM_REG_R13)))) > 1)
      ArmInstructionMakeLdm(restore,ARM_REG_R13,0xffff&set&(~((1<<ARM_REG_R15)|(1<<ARM_REG_R13))),ARM_CONDITION_AL,TRUE,FALSE,TRUE);
    else
    {
      t_reg iter_reg;
      REGSET_FOREACH_REG(set,iter_reg) break;
      ArmInstructionMakeLdr(restore,iter_reg,ARM_REG_R13,ARM_REG_NONE,4,ARM_CONDITION_AL,FALSE,TRUE,FALSE);
    }
    stack_change-=RegsetCountRegs(0xffff&set&(~((1<<ARM_REG_R15)|(1<<ARM_REG_R13))))*4;
    InsPrependToBbl(restore,out);
  }
  if (set & (/*(1<<ARM_REG_Q_CONDITION) |*/ (1<<ARM_REG_C_CONDITION) | (1<<ARM_REG_V_CONDITION) | (1<<ARM_REG_Z_CONDITION) | (1<<ARM_REG_N_CONDITION)))
  {
    msr=InsNewForBbl(out);
    ArmInstructionMakeMsr(msr,ARM_REG_R14,ARM_CONDITION_AL,TRUE);
    InsPrependToBbl(msr,out);

    restore2 = InsNewForBbl(out);
    ArmInstructionMakeLdr(restore2,ARM_REG_R14,ARM_REG_R13,ARM_REG_NONE,4,ARM_CONDITION_AL,FALSE,TRUE,FALSE);
    InsPrependToBbl(restore2,out);
    stack_change-=4;
  }

  if (stack_change!=0) FATAL(("Stack mismatch!"));
} /* }}} */

void ArmAtomAddParams(t_bbl * in, t_ins * ins, t_bbl * orig, t_atom_instrument_marker * marker,t_uint32 stack_change, t_address pc)
{
  t_ins ** first_inses_for_params=NULL;
  t_ins * parstart=BBL_INS_LAST(in);
  t_atom_proto * proto=marker->proto;
  t_uint32 tel;
  if (proto->nargs) first_inses_for_params=Calloc(proto->nargs+1,sizeof(t_ins *));

  if (proto->nargs>4) FATAL(("Implement more then 4 arguments for the ARM!"));
  
  /* {{{ add parameter construction instructions */
  for (tel=0; tel<proto->nargs; tel++)
  {

    switch(proto->args[tel])
    {
      case ATOM_REG_VALUE:
	if ((marker->values[tel].reg==REG_PC)||(marker->values[tel].reg==REG_R15))
	{
	  t_ins * param;
	  param= InsNewForBbl(in);
	  first_inses_for_params[tel]=param;
	  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R15,0,ARM_CONDITION_AL);
	  /*ArmMakeConstantProducer(param, G_T_UINT32(BBL_OLD_ADDRESS(orig)));*/
	  ArmMakeConstantProducer(param, G_T_UINT32(pc));
	  InsAppendToBbl(param,in);
	}
	else if ((marker->values[tel].reg==REG_SP)||(marker->values[tel].reg==REG_R13))
	{
	  /* Arm implementation {{{*/
	  t_ins * param = InsNewForBbl(in);
	  first_inses_for_params[tel]=param;
	  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R13,ARM_REG_NONE,stack_change,ARM_CONDITION_AL);
	  InsAppendToBbl(param,in);
	  /* }}} */
	}
	/* Other Arm Registers {{{ */
	else if ((marker->values[tel].reg>=REG_R0)||(marker->values[tel].reg<=REG_R15))
	{
	  t_ins * param= InsNewForBbl(in);
	  first_inses_for_params[tel]=param;
	  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R0+marker->values[tel].reg-REG_R0,0,ARM_CONDITION_AL);
	  InsAppendToBbl(param,in);
	}
	/* }}} */
	else
	{
	  FATAL(("Implement other registers (reg=%d)!",marker->values[tel].reg));
	}
	break;
      case ATOM_SPECIAL_VALUE:
	{
	  switch(marker->values[tel].query)
	  {
	    case BrCondValue:
	      if (!ins) FATAL(("Can only query BrCondValue for instructions!"));
	      /* Arm implementation {{{ */
	      {
		t_ins * param;
		param= InsNewForBbl(in);
		first_inses_for_params[tel]=param;
		ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_NONE,0,ARM_CONDITION_AL);
		InsAppendToBbl(param,in);
		param= InsNewForBbl(in);
		ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_NONE,1,ARMINSCONDITION(ins));
		InsAppendToBbl(param,in);
	      }
	      /* }}} */
	      break;
	    case EffStoreAddrValue:
do_EffStoreAddrValue:
	      if (!ins) FATAL(("Can only query EffStoreAddrValue for instructions!"));

	    case EffLoadAddrValue:
do_EffLoadAddrValue:
	      if (!ins) FATAL(("Can only query EffLoadAddrValue for instructions!"));

	      /* Arm implementation {{{*/
	      if (INS_TYPE(ins)==IT_LOAD || INS_TYPE(ins)==IT_STORE)
	      {
		if ((ARMINSOPCODE(ins)==ARM_LDR) || (ARMINSOPCODE(ins)==ARM_LDRB) || (ARMINSOPCODE(ins)==ARM_LDRSB) || (ARMINSOPCODE(ins)==ARM_LDRH) || (ARMINSOPCODE(ins)==ARM_LDRSH) || (ARMINSOPCODE(ins)==ARM_STR) || (ARMINSOPCODE(ins)==ARM_STRB) || (ARMINSOPCODE(ins)==ARM_STRH) )
		{
		  t_bool switched_regs = FALSE;
		  t_ins * param;
#ifdef VERBOSE_INSTRUMENT
		  DiabloPrint(stdout,">>>>> @I\n",ins);
#endif
		  /* Not implemented cases  {{{ */
		  if (ARMINSREGS(ins)!=ARM_REG_NONE) FATAL(("Implement shifted registers!"));
		  if (ARMINSREGC(ins) == ARM_REG_R13 || ARMINSREGC(ins) == ARM_REG_R15)
		  {
		    /* ugly hack to avoid extra boring implementation work:
		     * if regc is r13 or r15, and the addressing mode is preindexed,
		     * swap regb and regc. the calculated address remains the same, but
		     * we don't have to implement them as special cases */
		    if (ARMINSFLAGS(ins) & FL_PREINDEX)
		    {
		      t_reg tmp = ARMINSREGB(ins);
		      ARMINSREGB(ins) = ARMINSREGC(ins);
		      ARMINSREGC(ins) = tmp;
		      switched_regs = TRUE;
		    }
		  }
		  if (ARMINSREGC(ins)==ARM_REG_R13) FATAL(("Implement SP used as second reg in load"));
		  if (ARMINSREGC(ins)==ARM_REG_R15) FATAL(("Implement PC used as second reg in load")); 
		  /* }}} */
		  /* Register + Register loads/stores {{{ */
		  if (ARMINSREGC(ins)!=ARM_REG_NONE) 
		  {
		    /* Preindexing: we have to take the index into account */
		    if (ARMINSFLAGS(ins)&FL_PREINDEX)
		    {
		      /* The index is added to the base register */
		      if (ARMINSFLAGS(ins)&FL_DIRUP)
		      {
			/* Special case 1, ARM_REG_R15 is used as base
			 * register. We know the old value of the pc, so we can
			 * calculate the old value. */
			if (ARMINSREGB(ins)==ARM_REG_R15)
			{
			  param=InsNewForBbl(in);
			  first_inses_for_params[tel]=param;
			  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R15,0,ARM_CONDITION_AL);
			  ArmMakeConstantProducer(param, (8+G_T_UINT32(INS_OLD_ADDRESS(ins))));
			  InsAppendToBbl(param,in);
			  param= InsNewForBbl(in);
			  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARMINSREGC(ins),0,ARM_CONDITION_AL);
			  ARMINSSHIFTTYPE(param)=ARMINSSHIFTTYPE(ins);
			  ARMINSSHIFTLENGTH(param)=ARMINSSHIFTLENGTH(ins);
			  InsAppendToBbl(param,in);

			}
			/* Special case 2, ARM_REG_R13 is used as base
			 * register. We know the old value of sp (we changed it
			 * when adding parameters, but for the rest it should
			 * be unchanged), so we can calculate the original
			 * address for this load */ 
			else if (ARMINSREGB(ins)==ARM_REG_R13)
			{
			  param= InsNewForBbl(in);
			  first_inses_for_params[tel]=param;
			  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARMINSREGC(ins),0,ARM_CONDITION_AL);
			  ARMINSSHIFTTYPE(param)=ARMINSSHIFTTYPE(ins);
			  ARMINSSHIFTLENGTH(param)=ARMINSSHIFTLENGTH(ins);
			  InsAppendToBbl(param,in);
			  param= InsNewForBbl(in);
			  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,stack_change,ARM_CONDITION_AL);
			  InsAppendToBbl(param,in);
			}
			/* The general case: we calculate the new address and
			 * add a call to translate to translate the address
			 * back to it's original value. This is implemented
			 * with a pseudo call that will be changed in a real
			 * call by a post-pass */
			else
			{
			  param= InsNewForBbl(in);
			  first_inses_for_params[tel]=param;
			  ArmInstructionMakeAdd(param,ARM_REG_R0,ARMINSREGB(ins),ARMINSREGC(ins),0,ARM_CONDITION_AL);
			  ARMINSSHIFTTYPE(param)=ARMINSSHIFTTYPE(ins);
			  ARMINSSHIFTLENGTH(param)=ARMINSSHIFTLENGTH(ins);
			  InsAppendToBbl(param,in);
			  param= InsNewForBbl(in);
			  ArmInstructionMakePseudoCall(param,translate);
			  InsAppendToBbl(param,in);
			  if (tel!=0) {
			    param= InsNewForBbl(in);
			    ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R0,0,ARM_CONDITION_AL);
			    InsAppendToBbl(param,in);
			  }
			}
		      }
		      /* The index is substracted from the base register */
		      else
		      {
			/* Special case 1, ARM_REG_R15 is used as base
			 * register. We know the old value of the pc, so we can
			 * calculate the old value. */
			if (ARMINSREGB(ins)==ARM_REG_R15)
			{
			  param=InsNewForBbl(in);
			  first_inses_for_params[tel]=param;
			  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R15,0,ARM_CONDITION_AL);
			  ArmMakeConstantProducer(param, (8+G_T_UINT32(INS_OLD_ADDRESS(ins))));
			  InsAppendToBbl(param,in);
			  param= InsNewForBbl(in);
			  ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARMINSREGC(ins),0,ARM_CONDITION_AL);
			  ARMINSSHIFTTYPE(param)=ARMINSSHIFTTYPE(ins);
			  ARMINSSHIFTLENGTH(param)=ARMINSSHIFTLENGTH(ins);
			  InsAppendToBbl(param,in);
			}
			/* Special case 2, ARM_REG_R13 is used as base
			 * register. We know the old value of sp (we changed it
			 * when adding parameters, but for the rest it should
			 * be unchanged), so we can calculate the original
			 * address for this load */ 
			else if (ARMINSREGB(ins)==ARM_REG_R13)
			{
			  param= InsNewForBbl(in);
			  first_inses_for_params[tel]=param;
			  ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARMINSREGC(ins),0,ARM_CONDITION_AL);
			  ARMINSSHIFTTYPE(param)=ARMINSSHIFTTYPE(ins);
			  ARMINSSHIFTLENGTH(param)=ARMINSSHIFTLENGTH(ins);
			  InsAppendToBbl(param,in);
			  param= InsNewForBbl(in);
			  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,stack_change,ARM_CONDITION_AL);
			  InsAppendToBbl(param,in);

			}
			/* The general case: we calculate the new address and
			 * add a call to translate to translate the address
			 * back to it's original value. This is implemented
			 * with a pseudo call that will be changed in a real
			 * call by a post-pass */
			else
			{
			  param= InsNewForBbl(in);
			  first_inses_for_params[tel]=param;
			  ArmInstructionMakeSub(param,ARM_REG_R0,ARMINSREGB(ins),ARMINSREGC(ins),0,ARM_CONDITION_AL);
			  ARMINSSHIFTTYPE(param)=ARMINSSHIFTTYPE(ins);
			  ARMINSSHIFTLENGTH(param)=ARMINSSHIFTLENGTH(ins);
			  InsAppendToBbl(param,in);
			  param= InsNewForBbl(in);
			  ArmInstructionMakePseudoCall(param,translate);
			  InsAppendToBbl(param,in);
			  if (tel!=0) {
			    param= InsNewForBbl(in);
			    ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R0,0,ARM_CONDITION_AL);
			    InsAppendToBbl(param,in);
			  }
			}
		      }
		    }
		    /* Post indexing - We can safely ignore the index */
		    else
		    {
		      /* Special case 1, ARM_REG_R15 is used as base
		       * register. Let's pray this cannot happen :) */
		      if (ARMINSREGB(ins)==ARM_REG_R15)
		      {
			FATAL(("ARM_REG_R15 used in a post indexing load/store!"));
		      }
		      /* Special case 2, ARM_REG_R13 is used as base
		       * register. We know the old value of sp (we changed it
		       * when adding parameters, but for the rest it should
		       * be unchanged), so we can calculate the original
		       * address for this load */ 
		      else if (ARMINSREGA(ins)==ARM_REG_R13)
		      {
			param= InsNewForBbl(in);
			first_inses_for_params[tel]=param;
			ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARMINSREGB(ins),0,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);
			param= InsNewForBbl(in);
			ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,stack_change,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);
		      }
		      /* The general case: we calculate the new address and
		       * add a call to translate to translate the address
		       * back to it's original value. This is implemented
		       * with a pseudo call that will be changed in a real
		       * call by a post-pass */
		      else
		      {
			param= InsNewForBbl(in);
			first_inses_for_params[tel]=param;
			ArmInstructionMakeMov(param,ARM_REG_R0,ARMINSREGB(ins),0,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);
			param= InsNewForBbl(in);
			ArmInstructionMakePseudoCall(param,translate);
			InsAppendToBbl(param,in);
			if (tel!=0) {
			  param= InsNewForBbl(in);
			  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R0,0,ARM_CONDITION_AL);
			  InsAppendToBbl(param,in);
			}
		      }
		    }
		  }
		  /* End Register Register }}} */
		  /* Register - constant loads {{{ */
		  else
		  {
		    /* Preindexing: need to take the index into account */
		    if (ARMINSFLAGS(ins)&FL_PREINDEX)
		    {
		      /* Mind that loads from REG_R15 are/should be imposible
		       * here: they should have been turned in address
		       * producers! */
		      
		      /* Special case,  ARM_REG_R13 is used as base
		       * register. We know the old value of sp (we changed it
		       * when adding parameters, but for the rest it should
		       * be unchanged), so we can calculate the original
		       * address for this load */ 
		      if (ARMINSREGB(ins)==ARM_REG_R13)
		      {
			param= InsNewForBbl(in);
			first_inses_for_params[tel]=param;

			if (ARMINSFLAGS(ins)&FL_DIRUP)
			  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,ARMINSIMMEDIATE(ins)&0xff,ARM_CONDITION_AL);
			else
			  ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,ARMINSIMMEDIATE(ins)&0xff,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);

			if (ARMINSIMMEDIATE(ins)&(~0xff))
			{
			  param= InsNewForBbl(in);
			  if (ARMINSFLAGS(ins)&FL_DIRUP)
			    ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,ARMINSIMMEDIATE(ins)&(~0xff),ARM_CONDITION_AL);
			  else
			    ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,ARMINSIMMEDIATE(ins)&(~0xff),ARM_CONDITION_AL);

			  InsAppendToBbl(param,in);
			}

			param= InsNewForBbl(in);
			ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,stack_change,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);
		      }
		      /* The general case: we calculate the new address and
		       * add a call to translate to translate the address
		       * back to it's original value. This is implemented
		       * with a pseudo call that will be changed in a real
		       * call by a post-pass */
		      else
		      {
			param= InsNewForBbl(in);
			first_inses_for_params[tel]=param;

			if (ARMINSFLAGS(ins)&FL_DIRUP)
			  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,ARMINSIMMEDIATE(ins)&0xff,ARM_CONDITION_AL);
			else
			  ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,ARMINSIMMEDIATE(ins)&0xff,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);

			if (ARMINSIMMEDIATE(ins)&(~0xff))
			{
			  param= InsNewForBbl(in);
			  if (ARMINSFLAGS(ins)&FL_DIRUP)
			    ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,ARMINSIMMEDIATE(ins)&(~0xff),ARM_CONDITION_AL);
			  else
			    ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,ARMINSIMMEDIATE(ins)&(~0xff),ARM_CONDITION_AL);

			  InsAppendToBbl(param,in);
			}

			param= InsNewForBbl(in);
			ArmInstructionMakePseudoCall(param,translate);
			InsAppendToBbl(param,in);
			if (tel!=0) {
			  param= InsNewForBbl(in);
			  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R0,0,ARM_CONDITION_AL);
			  InsAppendToBbl(param,in);
			}
		      }
		    }
		    /* Post indexing.... we can ignore the index */
		    else
		    {
		      /* Mind that loads from REG_R15 are/should be imposible
		       * here: they should have been turned in address
		       * producers! */
		      
		      /* Special case,  ARM_REG_R13 is used as base
		       * register. We know the old value of sp (we changed it
		       * when adding parameters, but for the rest it should
		       * be unchanged), so we can calculate the original
		       * address for this load */ 
		      if (ARMINSREGB(ins)==ARM_REG_R13)
		      {
			param= InsNewForBbl(in);
			first_inses_for_params[tel]=param;
			ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARMINSREGB(ins),0,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);
			param= InsNewForBbl(in);
			ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,stack_change,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);
		      }
		      /* The general case: we calculate the new address and
		       * add a call to translate to translate the address
		       * back to it's original value. This is implemented
		       * with a pseudo call that will be changed in a real
		       * call by a post-pass */
		      else
		      {
			param= InsNewForBbl(in);
			first_inses_for_params[tel]=param;
			ArmInstructionMakeMov(param,ARM_REG_R0,ARMINSREGB(ins),0,ARM_CONDITION_AL);
			InsAppendToBbl(param,in);
			param= InsNewForBbl(in);
			ArmInstructionMakePseudoCall(param,translate);
			InsAppendToBbl(param,in);
			if (tel!=0) {
			  param= InsNewForBbl(in);
			  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R0,0,ARM_CONDITION_AL);
			  InsAppendToBbl(param,in);
			}
		      }
		    }
		  }
		  /* End Register - constant loads }}} */
		  if (switched_regs)
		  {
		    t_reg tmp = ARMINSREGB(ins);
		    ARMINSREGB(ins) = ARMINSREGC(ins);
		    ARMINSREGC(ins) = tmp;
		  }
		}
		else FATAL(("Strange IT_LOAD!"));
	      }
	      else if ((ARMINSOPCODE(ins)==ARM_LDM)||(ARMINSOPCODE(ins)==ARM_STM))
	      {
		t_ins * param;
                /* To match panalyzer */
		t_uint32 nregs=RegsetCountRegs(0xffff&ARMINSIMMEDIATE(ins))*4-4;

		/* Special case,  ARM_REG_R13 is used as base
		 * register. We know the old value of sp (we changed it
		 * when adding parameters, but for the rest it should
		 * be unchanged), so we can calculate the original
		 * address for this load */ 
		if (ARMINSREGB(ins)==ARM_REG_R13)
		{

		  param= InsNewForBbl(in);
		  first_inses_for_params[tel]=param;
		  if (!(ARMINSFLAGS(ins)&FL_WRITEBACK))
		  {
		    if (ARMINSFLAGS(ins)&FL_PREINDEX)
		    {
		      if (ARMINSFLAGS(ins)&FL_DIRUP)
			ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,4,ARM_CONDITION_AL);
		      else
			ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,4,ARM_CONDITION_AL);
		    }
		    else
		      ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,0,ARM_CONDITION_AL);
		  }
		  else if (ARMINSFLAGS(ins)&FL_DIRUP)
		    ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,nregs,ARM_CONDITION_AL);
		  else
		    ArmInstructionMakeSub(param,ARM_REG_R0+tel,ARMINSREGB(ins),ARM_REG_NONE,nregs,ARM_CONDITION_AL);
		  InsAppendToBbl(param,in);
		  
		  param= InsNewForBbl(in);
		  ArmInstructionMakeAdd(param,ARM_REG_R0+tel,ARM_REG_R0+tel,ARM_REG_NONE,stack_change,ARM_CONDITION_AL);
		  InsAppendToBbl(param,in);
		}
		/* The general case: we calculate the new address and
		 * add a call to translate to translate the address
		 * back to it's original value. This is implemented
		 * with a pseudo call that will be changed in a real
		 * call by a post-pass */
		else
		{
		  param= InsNewForBbl(in);
		  first_inses_for_params[tel]=param;
		  if (!(ARMINSFLAGS(ins)&FL_WRITEBACK))
		  {
		    if (ARMINSFLAGS(ins)&FL_PREINDEX)
		    {
		      if (ARMINSFLAGS(ins)&FL_DIRUP)
			ArmInstructionMakeAdd(param,ARM_REG_R0,ARMINSREGB(ins),ARM_REG_NONE,4,ARM_CONDITION_AL);
		      else
			ArmInstructionMakeSub(param,ARM_REG_R0,ARMINSREGB(ins),ARM_REG_NONE,4,ARM_CONDITION_AL);
		    }
		    else
		      ArmInstructionMakeAdd(param,ARM_REG_R0,ARMINSREGB(ins),ARM_REG_NONE,0,ARM_CONDITION_AL);
		  }
		  else if (ARMINSFLAGS(ins)&FL_DIRUP)
		    ArmInstructionMakeAdd(param,ARM_REG_R0,ARMINSREGB(ins),ARM_REG_NONE,nregs,ARM_CONDITION_AL);
		  else
		    ArmInstructionMakeSub(param,ARM_REG_R0,ARMINSREGB(ins),ARM_REG_NONE,nregs,ARM_CONDITION_AL);
		  InsAppendToBbl(param,in);

		  param= InsNewForBbl(in);
		  ArmInstructionMakePseudoCall(param,translate);
		  InsAppendToBbl(param,in);
		  if (tel!=0) {
		    param= InsNewForBbl(in);
		    ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R0,0,ARM_CONDITION_AL);
		    InsAppendToBbl(param,in);
		  }

		}
#ifdef VERBOSE_INSTRUMENT
		DiabloPrint(stdout,"M >>>>> @I\n",ins);
#endif
	      }
	      else if ((T_ARM_INS(ins)->info)&&(T_ARM_INS(ins)->info->original_type=='L'))
	      {
		t_ins * param= InsNewForBbl(in);
		first_inses_for_params[tel]=param;
		ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R15,0,ARM_CONDITION_AL);
		ArmMakeConstantProducer(param, T_ARM_INS(ins)->info->load_address);
		InsAppendToBbl(param,in);
	      }
	      else  
	      { 
		FATAL (("EffAddrValue only returns the address for loads, not for @I!",ins)); 
	      }
	      /*}}}*/
	      break;

	    case EffLoadAddrValue2:
	      if (!ins) FATAL(("EffLoadAddrValue only works for instructions"));
	      FATAL(("ARM has no need for EffLoadAddrValue2"));
	      break;

	    case EffAddrValue:
	      if (!ins) FATAL(("Can only query EffAddrValue for instructions"));
	      if (IsInstType(ins, InstTypeStore) && IsInstType(ins, InstTypeLoad))
		FATAL(("@I is both store and load. Disambiguate by using EffLoadAddrValue and EffStoreAddrValue",ins));
	      if (IsInstType(ins, InstTypeStore))
		goto do_EffStoreAddrValue;
	      if (IsInstType(ins, InstTypeLoad))
		goto do_EffLoadAddrValue;
	      FATAL(("@I is neither load nor store. There is no effective address value to be calculated", ins));
	      break;

	    case FunctionName:
	      /* FunctionName {{{ */
	      {
		t_section * added_section;
		t_object * obj;
		t_section * sec;
		t_string secname;
		t_ins * param;
		t_address null;
		t_reloc * rel;
		if (!(BBL_FUN(orig))) FATAL(("Oeps!"));
		obj=SECTION_OBJECT(CFG_SECTION(BBL_CFG(orig)));
		sec=SectionGetFromObjectByName(obj,".rodata");
		if (!sec) FATAL(("Implement no rodata!"));

		if (BBL_FUN(orig)->name)
		{
		  secname=Collate("funname_",BBL_FUN(orig)->name);
		  added_section=InstrumentAddExtraSection(obj,sec,RO_DATA_SECTION,secname,sec->name,strlen(BBL_FUN(orig)->name)+1,0x0);
		  strcpy(added_section->data,BBL_FUN(orig)->name);
		  Free(secname);
		}
		else 
		{
		  secname=Collate("funname_","noname");
		  added_section=InstrumentAddExtraSection(obj,sec,RO_DATA_SECTION,secname,sec->name,strlen("noname"),0x0);
		  strcpy(added_section->data,"noname");
		  Free(secname);
		}

		GenericSet_uint32(null,0);
		param= InsNewForBbl(in);
		first_inses_for_params[tel]=param;
		rel=RelocTableAddRelocToRelocatable(obj->reloc_table,null,T_RELOCATABLE(param),null,T_RELOCATABLE(added_section),null,FALSE,NULL,NULL,NULL,"S\\\\$",FALSE);

		ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R15,0,ARM_CONDITION_AL);
		ArmMakeAddressProducer(__FILE__, __LINE__,param, G_T_UINT32(SECTION_GET_OLD_ADDRESS(added_section)), rel);
		InsAppendToBbl(param,in);
	      }
	      /* }}} */
#if 0
	    case ObjectName:
	      /* ObjectName {{{ */
	      {
		t_section * added_section;
		t_string secname;
		t_ins * param;
		t_address null;
		t_reloc * rel;
		t_section * sec;

		sec=SectionGetFromObjectByName(inst_object,".rodata");
		if (!sec) sec=SectionGetFromObjectByName(inst_object,".rconst");
		if ((!sec) && inst_object->nrodatas) sec=inst_object->rodata[0];
		if (!sec) sec=SectionGetFromObjectByName(inst_object,"ER_RW");
		if (!sec) FATAL(("Implement no rodata!"));
		
		secname="objectname";
		added_section=InstrumentAddExtraSection(inst_object,sec,RO_DATA_SECTION,secname,sec->name,strlen(basename(inst_object->name))+1,0x0);
		strcpy(added_section->data,basename(inst_object->name));
		
		GenericSet_uint32(null,0);
		param= InsNewForBbl(in);
		first_inses_for_params[tel]=param;
		rel=RelocTableAddRelocToRelocatable(inst_object->reloc_table,null,T_RELOCATABLE(param),null,T_RELOCATABLE(added_section),null,FALSE,NULL,NULL,NULL,"S\\\\$",FALSE);

		ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_R15,0,ARM_CONDITION_AL);
		ArmMakeAddressProducer(__FILE__, __LINE__,param, G_T_UINT32(SECTION_GET_OLD_ADDRESS(added_section)), rel);
		InsAppendToBbl(param,in);
	      }
#endif
	      /* }}} */

	      break;
	    default:
	      FATAL(("Implement special values!"));
	  }
	  break;
	}

      case ATOM_INT_VALUE:
	{
	  t_ins * param= InsNewForBbl(in);
	  first_inses_for_params[tel]=param;
	  ArmInstructionMakeMov(param,ARM_REG_R0+tel,ARM_REG_NONE,0,ARM_CONDITION_AL);
	  ArmMakeConstantProducer(param, marker->values[tel].int32);
	  InsAppendToBbl(param,in);
	}
	break;

      default:
	FATAL(("Implement argument type %d\n",proto->args[tel]));
	/* TODO: Add the right parameter */
    }
  } /* }}} */

  AtomReorderParameters(proto,first_inses_for_params,parstart,in);
}

/* ArmAtomAddCall {{{ */ 
void ArmAtomAddCall(t_bbl * in, t_bbl * out, t_atom_proto * proto)
{
  t_cfg * cfg = FUN_CFG(BBL_FUN(in));
  t_bbl * instr_entry;
  t_ins * call;

  instr_entry = proto->bbl;

  call = InsNewForBbl(in);
  ArmInstructionMakeCondBranchAndLink(call,ARM_CONDITION_AL);
  InsAppendToBbl(call,in);
  CfgEdgeNewCall(cfg,in,instr_entry,out,FUN_BBL_LAST(BBL_FUN(instr_entry)));
  CfgEdgeNew(cfg,in,out,ET_FUNLINK);
} /* }}} */

t_bool ArmIsInstType(Inst * ins, t_atom_inst_type type)
{
  switch(type)
  {
    case InstTypeLoad:
      if ((INS_TYPE(ins)==IT_LOAD)||(INS_TYPE(ins)==IT_LOAD_MULTIPLE)||((T_ARM_INS(ins)->info)&&(T_ARM_INS(ins)->info->original_type=='L')))
	return TRUE;
      break;
    case InstTypeStore:
      if ((INS_TYPE(ins)==IT_STORE)||(INS_TYPE(ins)==IT_STORE_MULTIPLE))
	return TRUE;
      break;
    case InstTypeCondBr:
      if (ARMINSOPCODE(ins) == ARM_B && ARMINSCONDITION(ins) != ARM_CONDITION_AL)
	return TRUE;
      if (ARMINSOPCODE(ins) == ARM_BL && ARMINSCONDITION(ins) != ARM_CONDITION_AL)
	return TRUE;
      return FALSE;
      break;
    case InstTypeConditional:
      if (ARMINSCONDITION(ins) != ARM_CONDITION_AL)
	return TRUE;
      return FALSE;
      break;
    default:
      FATAL(("Implement IsInstType"));
  }
  return FALSE;
}

/* vim: set shiftwidth=2 foldmethod=marker:*/
