/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/*#define DEBUG_CONST_OPTIMIZATIONS*/

/* Includes {{{ */
#include "arm_utils.h"
#include "arm_emulation.h"
#include "arm_constant_propagation.h"
#include "arm_constantoptimizations.h"
#include "arm_registers.h"
#include "arm_instructions.h"
#ifdef DIETER
#include "arm_description.h"
#endif
#include "../../utils/util_error.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_procstate.h"
#include "../../kernel/diablo_constprop.h"
#include "../../diablo_options.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_profile.h"

/* }}} */
/*! 
 * Counts the number of defined registers for an instruction
 *
 * \todo This is a weird function, check use, name, ... Its should be in
 * arm_utils it also uses 16 as the number of registers ... 
 *
 * \param ins the instruction for which we want to count the defined registers 
 *
 * \return unsigned int the numer of registers that is defined by this
 * instruction 
 */
/* ArmInsDefinedRegCount {{{ */

void MakeConstProducers2(t_ins * ins, t_procstate *procstate);

t_ins * FindCmpThatDeterminesJump(t_ins * ins)
{
  ins = INS_PREV(ins);

  while (ins)
    {
      if (ARMINSOPCODE(ins)==ARM_CMP)
	break;
      ins = INS_PREV(ins);
    }

  return ins;
}

t_bool BblEndsWithConditionalBranchAfterCMP(t_bbl * bbl)
{
  t_ins * last_ins = BBL_INS_LAST(bbl);
  t_ins * ins;
  t_regset regs;
  t_regset def_regs = RegsetNew();

  if (ARMINSOPCODE(last_ins)!=ARM_B) return FALSE;

  if (ARMINSCONDITION(last_ins)!=ARM_CONDITION_EQ && ARMINSCONDITION(last_ins)!=ARM_CONDITION_NE) return FALSE;

  regs = INS_REGS_USE(last_ins);
  ins = INS_PREV(last_ins);

  while (ins)
    {
      if (ARMINSOPCODE(ins)==ARM_CMP)
	break;
      if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),regs)))
	ins = NULL;
      else
	{
	  RegsetSetUnion(def_regs,INS_REGS_DEF(ins));
	  ins = INS_PREV(ins);
	}
    }

  if (ins==NULL)
    return FALSE;

  if (ARMINSREGC(ins)!=ARM_REG_NONE) return FALSE;

  if (RegsetIn(def_regs,ARMINSREGB(ins))) return FALSE;
  if (ArmInstructionIsConditional(ins)) return FALSE;
  
  return TRUE;
}

unsigned int ArmInsDefinedRegCount(t_ins * ins) {
  t_reg i;
  int count;
  t_regset mask;

  mask = ArmDefinedRegisters(ins);
  count = 0;

  REGSET_FOREACH_REG(mask,i)
    count++;
  return count;
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetSingleCondition {{{ */
static void ARMTestAndSetSingleCondition(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false, t_reg flag)
{
  t_bool value;
  if (BOT==ProcStateGetCond(state,flag,&value))
    {
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
      ProcStateSetCond(*state_true,flag,TRUE);
      ProcStateSetCond(*state_false,flag,FALSE);
    }
  else if (value)
    {
      *state_true = state;
      *state_false = NULL;
    }
  else
    {
      *state_false = state;
      *state_true = NULL;
    }
}

static void ARMTestAndSetSingleConditionExt(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false, t_reg flag)
{
  t_bool value;
  /*#define TEST*/
#ifdef TEST
  static int teller = 0;
#endif
  if (BOT==ProcStateGetCond(state,flag,&value))
    {
      t_ins * last_ins = BBL_INS_LAST(CFG_EDGE_HEAD(edge));
      t_ins * ins = FindCmpThatDeterminesJump(last_ins);
      t_uint32 value = ARMINSIMMEDIATE(ins);
      t_reg reg = ARMINSREGB(ins);
      register_content content;
      GenericSet_uint32(content.i,value);

      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
      ProcStateSetCond(*state_true,flag,TRUE);
      ProcStateSetCond(*state_false,flag,FALSE);

#ifdef TEST
      if (teller++<global_options.debugcounter)
#endif
	{
	  ProcStateSetReg(*state_true,reg,content);
	  ProcStateSetTagTop(*state_true,reg);
	  /*	  DiabloPrintArch(stdout,BBL_FUN(CFG_EDGE_HEAD(edge))->cfg->sec->obj->description,"@iB\n @E\n @P\n", CFG_EDGE_HEAD(edge),edge,*state_true); */
	}
    }
  else if (value)
    {
      *state_true = state;
      *state_false = NULL;
    }
  else
    {
      *state_false = state;
      *state_true = NULL;
    }
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_EQ {{{ */
void ARMTestAndSetConditionCN_EQ(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge,state, state_true, state_false, ARM_REG_Z_CONDITION);
}

void ARMTestAndSetConditionCN_EQExt(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleConditionExt(edge,state, state_true, state_false, ARM_REG_Z_CONDITION);
}


/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_NE {{{ */
void ARMTestAndSetConditionCN_NE(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge,state, state_false, state_true, ARM_REG_Z_CONDITION);
}

void ARMTestAndSetConditionCN_NEExt(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleConditionExt(edge,state, state_false, state_true, ARM_REG_Z_CONDITION);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_CS {{{ */
void ARMTestAndSetConditionCN_CS(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge,state, state_true, state_false, ARM_REG_C_CONDITION);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_CC {{{ */
void ARMTestAndSetConditionCN_CC(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge,state, state_false, state_true, ARM_REG_C_CONDITION);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_MI {{{ */
void ARMTestAndSetConditionCN_MI(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge, state, state_true, state_false, ARM_REG_N_CONDITION);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_PL {{{ */
void ARMTestAndSetConditionCN_PL(t_cfg_edge * edge, t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge, state, state_false, state_true, ARM_REG_N_CONDITION);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_VS {{{ */
void ARMTestAndSetConditionCN_VS(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge,state, state_true, state_false, ARM_REG_V_CONDITION);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_VC {{{ */
void ARMTestAndSetConditionCN_VC(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetSingleCondition(edge,state, state_false, state_true, ARM_REG_V_CONDITION);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_HI {{{ */
void ARMTestAndSetConditionCN_HI(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{ 
  t_bool c_flag;
  t_bool z_flag;
  t_lattice_level c_level = ProcStateGetCond(state,ARM_REG_C_CONDITION,&c_flag);
  t_lattice_level z_level = ProcStateGetCond(state,ARM_REG_Z_CONDITION,&z_flag);

  if (BOT==c_level && BOT==z_level)
    {
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
      ProcStateSetCond(*state_true,ARM_REG_C_CONDITION,TRUE);
      ProcStateSetCond(*state_true,ARM_REG_Z_CONDITION,FALSE);
    }
  else if (BOT==c_level)
    {
      if (!z_flag)
	{
	  *state_true  = ProcStateNew();
	  *state_false = state;
	  ProcStateDup(*state_true,state);
	  ProcStateSetCond(*state_true,ARM_REG_C_CONDITION,TRUE);
	  ProcStateSetCond(*state_false,ARM_REG_C_CONDITION,FALSE);
	}
      else
	{
	  *state_true = NULL;
	  *state_false = state;
	}
    }
  else if (BOT==z_level)
    {
      if (c_flag)
	{
	  *state_true  = ProcStateNew();
	  *state_false = state;
	  ProcStateDup(*state_true,state);
	  ProcStateSetCond(*state_true,ARM_REG_Z_CONDITION,FALSE);
	  ProcStateSetCond(*state_false,ARM_REG_Z_CONDITION,TRUE);
	}
      else
	{
	  *state_true = NULL;
	  *state_false = state;
	}
    }
  else if (c_flag && !z_flag)
    {
      *state_true = state;
      *state_false = NULL;
    }
  else
    {
      *state_true  = NULL;
      *state_false = state;
    }
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_LS {{{ */
void ARMTestAndSetConditionCN_LS(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetConditionCN_HI(edge,state, state_false, state_true);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_GE {{{ */
void ARMTestAndSetConditionCN_GE(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  t_bool n_flag;
  t_bool v_flag;
  t_lattice_level n_level = ProcStateGetCond(state,ARM_REG_N_CONDITION,&n_flag);
  t_lattice_level v_level = ProcStateGetCond(state,ARM_REG_V_CONDITION,&v_flag);

  if (BOT==n_level && BOT==v_level)
    {
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
    }
  else if (BOT==n_level)
    {
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
      if (v_flag)
	{
	  ProcStateSetCond(*state_true,ARM_REG_N_CONDITION,TRUE);
	  ProcStateSetCond(*state_false,ARM_REG_N_CONDITION,FALSE);
	}
      else
	{
	  ProcStateSetCond(*state_true,ARM_REG_N_CONDITION,FALSE);
	  ProcStateSetCond(*state_false,ARM_REG_N_CONDITION,TRUE);
	}
    }
  else if (BOT==v_level)
    {
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
      if (n_flag)
	{
	  ProcStateSetCond(*state_true,ARM_REG_V_CONDITION,TRUE);
	  ProcStateSetCond(*state_false,ARM_REG_V_CONDITION,FALSE);
	}
      else
	{
	  ProcStateSetCond(*state_true,ARM_REG_V_CONDITION,FALSE);
	  ProcStateSetCond(*state_false,ARM_REG_V_CONDITION,TRUE);
	}
    }
  else if ((n_flag && v_flag)
	   ||
	   (!n_flag && ! v_flag)
	   )
    {
      *state_true = state;
      *state_false = NULL;
    }
  else
    {
      *state_true  = NULL;
      *state_false = state;
    }
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_LT {{{ */
void ARMTestAndSetConditionCN_LT(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetConditionCN_GE(edge,state, state_false, state_true);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_GT {{{ */
void ARMTestAndSetConditionCN_GT(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  t_bool z_flag;
  t_lattice_level z_level = ProcStateGetCond(state,ARM_REG_Z_CONDITION,&z_flag);

  if (BOT==z_level)
    {
      ARMTestAndSetConditionCN_GE(edge,state, state_true, state_false);

      if (*state_true)
	ProcStateSetCond(*state_true,ARM_REG_Z_CONDITION,FALSE);
    }
  else if (z_flag)
    {
      *state_false = state;
      *state_true = NULL;
    }
  else
    {
      ARMTestAndSetConditionCN_GE(edge,state, state_true, state_false);
    }
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_LE {{{ */
void ARMTestAndSetConditionCN_LE(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  ARMTestAndSetConditionCN_GT(edge,state, state_false, state_true);
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_AL {{{ */
void ARMTestAndSetConditionCN_AL(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true = state;
  *state_false = NULL;
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionCN_NV {{{ */
void ARMTestAndSetConditionCN_NV(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_false = state;
  *state_true = NULL;
}
/* }}} */
/*!
 *
 * \todo Document
 *
 * This is a function that is attached to system call edges 
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionSetAllToBot {{{ */
void ARMTestAndSetConditionSetAllToBot(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
  ProcStateSetAllBot(*state_true,AllRegs());
}
/* }}} */
/*!
 * This is a function for return and compensating edges: information must be
 * propagated over all outgoing edges of an exit block, at least as far as we
 * are concerned. The constant propagator might add different constraints, such
 * as in a context-sensitive constant propagation. This function is also used
 * for hell edges.
 *
 * In general it is used where more than two successor edges are found in the
 * graph, which model anomalous control flow. They are not taken or not-taken
 * edges, but reflect different execution contexts.  
 *
 * \todo Document
 * 
 * \param edge
 * \param state
 * \param state_true
 * \param state_false
 * \param flag
 *
 * \return void
 */
/* ARMTestAndSetConditionAllPaths {{{ */
void ARMTestAndSetConditionAllPaths(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = state;
  *state_false = state;
}
/* }}} */
/*! \todo Document */
TestAndSetConditionFunction ARMTestAndSetConditionFunctionsArray [] = { &ARMTestAndSetConditionCN_EQ,
									&ARMTestAndSetConditionCN_NE,
									&ARMTestAndSetConditionCN_CS,
									&ARMTestAndSetConditionCN_CC,
									&ARMTestAndSetConditionCN_MI,
									&ARMTestAndSetConditionCN_PL,
									&ARMTestAndSetConditionCN_VS,
									&ARMTestAndSetConditionCN_VC,
									&ARMTestAndSetConditionCN_HI,
									&ARMTestAndSetConditionCN_LS,
									&ARMTestAndSetConditionCN_GE,
									&ARMTestAndSetConditionCN_LT,
									&ARMTestAndSetConditionCN_GT,
									&ARMTestAndSetConditionCN_LE,
									&ARMTestAndSetConditionCN_AL,
									&ARMTestAndSetConditionCN_NV
};


TestAndSetConditionFunction ARMTestAndSetConditionFunctionsArrayExt [] = { &ARMTestAndSetConditionCN_EQExt,
									&ARMTestAndSetConditionCN_NEExt,
									&ARMTestAndSetConditionCN_CS,
									&ARMTestAndSetConditionCN_CC,
									&ARMTestAndSetConditionCN_MI,
									&ARMTestAndSetConditionCN_PL,
									&ARMTestAndSetConditionCN_VS,
									&ARMTestAndSetConditionCN_VC,
									&ARMTestAndSetConditionCN_HI,
									&ARMTestAndSetConditionCN_LS,
									&ARMTestAndSetConditionCN_GE,
									&ARMTestAndSetConditionCN_LT,
									&ARMTestAndSetConditionCN_GT,
									&ARMTestAndSetConditionCN_LE,
									&ARMTestAndSetConditionCN_AL,
									&ARMTestAndSetConditionCN_NV
};

/*! 
 * \todo Document
 *
 * \param edge
 * \param last_ins
 *
 * \return void
 */
/* ArmEdgePropagator {{{ */
void ArmEdgePropagator(t_cfg_edge * edge, t_ins * last_ins)
{
  t_uint8  condition = last_ins?ARMINSCONDITION(last_ins):ARM_CONDITION_AL;
  t_uint16 edgetype = EDGE_CAT(edge);

  if (IsHellNode(EDGE_HEAD(edge)))
    {
      edge->testcondition = &ARMTestAndSetConditionAllPaths;
      return;
    }

  switch (edgetype)
  {
    case ET_JUMP:
      /* this is a special case: jump tables jumps are unconditional,
	 but we need to propagate to all the blocks in the jump table */

      if (CFG_EDGE_FLAGS(edge) & EF_FROM_SWITCH_TABLE)
      {
	edge->testcondition = &ARMTestAndSetConditionAllPaths;
	break;
      }
    case ET_CALL:
    case ET_IPJUMP:
      if (BblEndsWithConditionalBranchAfterCMP(INS_BBL(last_ins)))
	edge->testcondition = ARMTestAndSetConditionFunctionsArrayExt[condition];
      else
	edge->testcondition = ARMTestAndSetConditionFunctionsArray[condition];
      break;

    case ET_FALLTHROUGH:
    case ET_IPFALLTHRU:
    case ET_UNKNOWN:
    case ET_IPUNKNOWN:

      /* if there is a fall-through-like path and PropOverEdge is
	 called for the corresponding edge with some state, then this
	 state needs to be propagated 

	 if this edge comes from a block ending with a conditional
	 branch, PropOverEdge() will have been called for the
	 branch-taken edge before it is called for the fall-through
	 edge, and if the propagated state could be evaluated to
	 "always take the branch", PropOverEdge() will not even be
	 called for the fall-through edges.  */

      edge->testcondition = ARMTestAndSetConditionFunctionsArray[ARM_CONDITION_AL];
      break;
    case ET_SWITCH:
    case ET_RETURN:
    case ET_COMPENSATING:
    case ET_IPSWITCH:
      edge->testcondition = &ARMTestAndSetConditionAllPaths;
      break;
    case ET_FUNLINK:
    case ET_IPFUNLINK:
    case ET_SWI:

      /* these case (and especially link edges) are dealt with in
	 PropOverEdge() */

      edge->testcondition = ARMTestAndSetConditionFunctionsArray[ARM_CONDITION_NV];
      break;
    case ET_SWILINK:
    case ET_IPSWILINK:

      /* this is overly conservative, we assume that systems calls
	 define all registers in unknown ways */

      edge->testcondition = ARMTestAndSetConditionFunctionsArray[ARM_CONDITION_AL];/*&ARMTestAndSetConditionSetAllToBot;*/
      break;
    default:
      FATAL(("ARMSetConditionToProp invalid type %lx\n",edgetype));
      break;
  }

}
/* }}} */
/*! 
 * \todo Document
 *
 * \param ins
 *
 * \return t_bool
 */
/* ArmInstructionUnconditionalizer {{{ */
t_bool ArmInstructionUnconditionalizer(t_ins * ins)
{
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL)
    {
      ARMINSCONDITION(ins)=ARM_CONDITION_AL;
      INS_REGS_USE(ins)=ArmUsedRegisters(ins); 
      INS_REGS_DEF(ins)=ArmDefinedRegisters(ins); 
      INS_ATTRIB(ins) &= ~IF_CONDITIONAL;
      return TRUE;
    }
  return FALSE;
}
/* }}} */
/*!
 *
 * \todo Document
 * 
 * \param ins
 * \param prev_state
 * \param next_state
 *
 * \return t_bool
 */
/* ArmInstructionsConstantOptimizer {{{ */
t_bool ArmInstructionsConstantOptimizer(t_ins * ins, t_procstate * prev_state, t_procstate * next_state)
{
  
  static int teller = 0;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  teller++;

  if (teller>global_options.debugcounter)
    return TRUE;
#endif

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"1 %d @I\n",teller,ins);
#endif
  if (!ArmInsOptKillSBit(ins,prev_state,next_state)) return FALSE;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"2 %d @I\n",teller,ins);
#endif

  if (!global_options.rely_on_calling_conventions)
    if (!ArmInsOptKillIdempotent(ins,prev_state,next_state)) return FALSE;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"3 %d @I\n",teller,ins);
#endif
  
  if (!ArmInsOptMakeIndirectJumpsDirect(ins,prev_state,next_state)) return FALSE;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"4 %d @I\n",teller,ins);
#endif
  
  if (!ArmInsOptSwitchOptimizer(ins,prev_state,next_state)) return FALSE;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"5 %d @I\n",teller,ins);
#endif

  if (!ArmInsOptEncodeConstantResult(ins,prev_state,next_state)) return FALSE;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"6 %d @I\n",teller,ins);
#endif
  
  if (!global_options.rely_on_calling_conventions)
    if (!ArmInsOptProduceConstantFromOtherRegister(ins,prev_state,next_state,teller)) return FALSE;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"7 %d @I\n",teller,ins);
#endif
  
  if (!ArmInsOptEncodeConstantOperands(ins,prev_state,next_state)) return FALSE;

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"8 %d @I\n",teller,ins);
#endif

  ArmOptMoveNullIntoPC(ins);

#ifdef DEBUG_CONST_OPTIMIZATIONS  
  if (teller>global_options.debugcounter-10)
    DiabloPrint(stdout,"9 %d @I\n",teller,ins);
#endif

  MakeConstProducers2(ins, next_state);
  
  /* switched off because it makes programs bigger instead of smaller!
   if (!ArmInsOptEncodeConstantResultInTwoInstructions(ins,prev_state,next_state)) return FALSE; */

  /*  ArmInsOptChangeLoadBaseRegister(ins, prev_state, next_state);*/
  return TRUE;
}


void RenameLocalAddressProducers(t_bbl * bbl)
{
  t_ins * i_ins, *j_ins, *k_ins;
  t_regset used_regs = RegsetNew();

  BBL_FOREACH_INS(bbl,i_ins)
    {
      if (ARMINSOPCODE(i_ins)==ARM_ADDRESS_PRODUCER && !ArmInstructionIsConditional(i_ins))
	{
	  t_reg def = ARMINSREGA(i_ins);
	  t_reg new_reg;
	  t_regset free_regs;
	  j_ins = INS_NEXT(i_ins);
	  
	  while (j_ins)
	    {
	      if (RegsetIn(INS_REGS_DEF(j_ins),def) && RegsetIn(INS_REGS_USE(j_ins),def) && ArmInsWriteBackHappens(j_ins))
		{
		  j_ins = NULL;
		  break;
		}
	      
	      if (RegsetIn(INS_REGS_DEF(j_ins),def) && !ArmInstructionIsConditional(j_ins))
		{
		  break;
		}
	      if (INS_TYPE(j_ins)==IT_STORE_MULTIPLE && RegsetIn(ARMINSIMMEDIATE(j_ins),def))
		{
		  break;
		}
	      j_ins = INS_NEXT(j_ins);
	    }
	  
	  if (j_ins==NULL)
	    continue;
	  
	  if (INS_TYPE(j_ins)==IT_STORE_MULTIPLE && RegsetIn(ARMINSIMMEDIATE(j_ins),def))
	    continue;

	  RegsetDup(free_regs,InsRegsLiveAfter(i_ins));
	  RegsetSetInvers(free_regs);
	  
	  RegsetSubReg(free_regs,def);
#if 1
	  k_ins = INS_NEXT(i_ins);
	  while (k_ins)
	    {
	      RegsetSetDiff(free_regs,INS_REGS_DEF(k_ins));
	      k_ins=INS_NEXT(k_ins);
	    }
#else
	  for (k_ins=INS_NEXT(i_ins);k_ins!=j_ins;k_ins=INS_NEXT(k_ins))
	    RegsetSetDiff(free_regs,INS_REGS_DEF(k_ins));
#endif  
	  RegsetSetIntersect(free_regs,BBL_FUN(bbl)->cfg->sec->obj->description->int_registers);
	  RegsetSetDiff(free_regs,BBL_REGS_LIVE_OUT(bbl));
	  RegsetSetDiff(free_regs,used_regs);
	  RegsetSubReg(free_regs,ARM_REG_R14);
	  RegsetSubReg(free_regs,ARM_REG_R15);
	  
	  if (RegsetIsEmpty(free_regs))
	    continue;

	  REGSET_FOREACH_REG_INVERSE(free_regs,new_reg)
	    break;

	  RegsetAddReg(used_regs,new_reg);

	  ARMINSREGA(i_ins)=new_reg;

	  INS_REGS_DEF(i_ins) = ArmDefinedRegisters(i_ins);

	  k_ins = INS_NEXT(i_ins);

	  while (k_ins)
	    {
	      if (ARMINSREGB(k_ins)==def)
		ARMINSREGB(k_ins)=new_reg;
	      if (ARMINSREGC(k_ins)==def)
		ARMINSREGC(k_ins)=new_reg;
	      if (ARMINSREGS(k_ins)==def)
		ARMINSREGS(k_ins)=new_reg;
	      if (INS_TYPE(k_ins)==IT_STORE)
		if (ARMINSREGA(k_ins)==def)
		  ARMINSREGA(k_ins)=new_reg;

	      INS_REGS_DEF(k_ins) = ArmDefinedRegisters(k_ins);
	      INS_REGS_USE(k_ins) = ArmUsedRegisters(k_ins);

	      if (RegsetIn(INS_REGS_DEF(k_ins),def) && !ArmInstructionIsConditional(k_ins))
		break;
	      k_ins = INS_NEXT(k_ins);
	    }
	}
    }
}

#define rotate_left(v, n) (((v) << (n)) | ((t_uint32)(v))>>(((t_uint32)(32-(n)))))

t_bool ConstProdCanGenerateInOne(t_int32 val)
{
  t_int32 val_neg = -val;
  t_int32 val_invert = ~val;
  
  t_uint32 a;
  t_uint32 i;
  
  /* first we try to generate the constant from scratch in one instruction */
  {
    for (i = 0; i < 32; i += 2)
      if ((a = rotate_left (val, i)) <= 0xff)
	{
	  return TRUE;
	}
    
    for (i = 0; i < 32; i += 2)
      if ((a = rotate_left (val_neg-1, i)) <= 0xff)
	{
	  return TRUE; 
	}

    for (i = 0; i < 32; i += 2)
      if ((a = rotate_left (val_invert, i)) <= 0xff)
      {
	return TRUE;
      }
  }
  return FALSE;
}

void MakeConstProducers(t_cfg * cfg)
{
  t_bbl * bbl;
  t_procstate * temp_procstate = ProcStateNew();
  register_content content;
  t_reloc * reloc;
  t_ins * ins;

  t_ConstantPropagationInsEmul propagator = CFG_OBJECT(cfg)->actions->InstructionEmulator;

  if (global_options.blockprofilefile)
    CfgComputeHotBblThreshold(cfg,0.95);

  CFG_FOREACH_BBL(cfg,bbl)
    {
      ProcStateSetAllBot(temp_procstate,cfg->sec->obj->description->all_registers);
      BBL_FOREACH_INS(bbl,ins) 
	{
	  propagator(ins,temp_procstate);
	  if ((
	       ARMINSOPCODE(ins)==ARM_MOV || 
	       ARMINSOPCODE(ins)==ARM_MVN || 
	       ARMINSOPCODE(ins)==ARM_SUB || 
	       ARMINSOPCODE(ins)==ARM_ADD || 
	       ARMINSOPCODE(ins)==ARM_LDR 
	       )
	      && ARMINSREGA(ins)!=ARM_REG_R15 && !ArmInstructionIsConditional(ins) && !ArmInsUpdatesCond(ins) && RegsetCountRegs(INS_REGS_DEF(ins)) == 1)
	    
	    if (VALUE == ProcStateGetReg(temp_procstate,ARMINSREGA(ins),&content))
	      if (TOP == ProcStateGetTag(temp_procstate,ARMINSREGA(ins),&reloc))
		{
		  if (!BblIsHot(bbl) || ConstProdCanGenerateInOne(GENERIC_EXTRACT_UINT32(content.i)))
		    ArmMakeConstantProducer(ins,GENERIC_EXTRACT_UINT32(content.i));
		}
	}
    }
}

void MakeConstProducers2(t_ins * ins, t_procstate *procstate)
{
  register_content content;
  t_reloc * reloc;
  t_uint32 value;

  if (ARMINSOPCODE(ins)!=ARM_CONSTANT_PRODUCER && 
      ARMINSREGA(ins)!=ARM_REG_R15 && 
      ARMINSREGA(ins)!=ARM_REG_NONE && 
      ARMINSOPCODE(ins)!=ARM_LDM &&
      ARMINSOPCODE(ins)!=ARM_LFM &&
      !ArmInstructionIsConditional(ins) && 
      !ArmInsUpdatesCond(ins) && 
      !ArmInsHasSideEffect(ins) &&
      RegsetCountRegs(INS_REGS_DEF(ins)) == 1 &&
      VALUE == ProcStateGetReg(procstate,ARMINSREGA(ins),&content) &&
      TOP == ProcStateGetTag(procstate,ARMINSREGA(ins),&reloc) && 
      !RegsetIsEmpty(INS_REGS_USE(ins))
      )
    {
      value = GENERIC_EXTRACT_UINT32(content.i);
      
      if (ConstProdCanGenerateInOne(value))
	{
	  ArmMakeConstantProducer(ins,value);
	}
    }
}

#ifdef DIETER
void ArmAddressAnalyser(t_ins * ins, t_procstate * procstate1, t_procstate * procstate2)
{
  t_reg i_reg;
  t_reloc * ins_reloc, * i_reloc;

  t_bool flag = FALSE; // Used for debugging

  if(ARMINSOPCODE(ins) == ARM_ADDRESS_PRODUCER)
    {
      ins_reloc = INS_REFERS_TO(ins)->rel;
      
      /* Check pre-conditions */
      if(ins_reloc &&  // relocation is available?
	 REL_IS_TO_RELOCATABLE(ins_reloc) &&  // to a relocatable?
	 RELOCATABLE_TYPE(REL_TO_RELOCATABLE(ins_reloc)) == RT_SUBSECTION  // to a subsection?
	 )
	{
	  REGSET_FOREACH_REG(arm_description.all_registers, i_reg)
	    {
	      if(i_reg != ARMINSREGA(ins) &&  // not the same reg?
		 VALUE == ProcStateGetTag(procstate1, i_reg, &i_reloc) &&  // contains a usable value?
		 i_reloc != ins_reloc &&  // not the same reloc as before?
		 REL_IS_TO_RELOCATABLE(i_reloc) &&  // to a relocatable?
		 RELOCATABLE_TYPE(REL_TO_RELOCATABLE(i_reloc)) == RT_SUBSECTION &&  // to a subsection?
		 REL_TO_RELOCATABLE(ins_reloc) != REL_TO_RELOCATABLE(i_reloc) &&  // not the same subsection as before? // Reconsider
		 SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(ins_reloc)) == SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(i_reloc))  // beloning to the same parent-section?
		 )
		{
#if 1 // Debug
		  if(!flag)
		    {
		      printf("ADDRESS:\n");
		      DiabloPrint(stdout, "  @R\n", ins_reloc);
		      flag = TRUE;
		    }
		  DiabloPrint(stdout, "    @R\n", i_reloc);
#endif

#if 0 // Analyse
		  if(INS_BBL(T_INS(REL_FROM(ins_reloc))) == INS_BBL(T_INS(REL_FROM(i_reloc))))
		    {
		      if(!flag)
			{
			  printf("ADDRESS (2 producers for same subsection in same BBL):\n");
			  DiabloPrint(stdout, "  @R\n", ins_reloc);
			  flag = TRUE;
			}
		      DiabloPrint(stdout, "    @R\n", i_reloc);
		    }
#endif

		  SectionConnUpdate(T_SECTION(REL_TO_RELOCATABLE(ins_reloc)), T_SECTION(REL_TO_RELOCATABLE(i_reloc)), 1);
		}
	    }
	}
    }
  else if(ARMINSOPCODE(ins) == ARM_LDR || ARMINSOPCODE(ins) == ARM_LDRB ||
	  ARMINSOPCODE(ins) == ARM_STR || ARMINSOPCODE(ins) == ARM_STRB)
    {
      if(VALUE == ProcStateGetTag(procstate1, ARMINSREGB(ins), &ins_reloc) &&
	 REL_IS_TO_RELOCATABLE(ins_reloc) &&
	 RELOCATABLE_TYPE(REL_TO_RELOCATABLE(ins_reloc)) == RT_SUBSECTION)
	{
	  /* If the value of an relocation is effectively used for memory access,
	     its hotness will increase faster. */
	  ins_reloc->hotness += (1 << 16);

	  t_regset dead_after = InsRegsLiveAfter(ins);
	  RegsetSetInvers(dead_after);
	  REGSET_FOREACH_REG(arm_description.all_registers, i_reg)
	    {
	      if(i_reg != ARMINSREGB(ins) &&
		 VALUE == ProcStateGetTag(procstate1, i_reg, &i_reloc) &&
		 i_reloc != ins_reloc &&
		 REL_IS_TO_RELOCATABLE(i_reloc) &&
		 RELOCATABLE_TYPE(REL_TO_RELOCATABLE(i_reloc)) == RT_SUBSECTION && 
		 SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(ins_reloc)) == SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(i_reloc))
		 )
		{
#if 1 // Debug
		  if(!flag)
		    {
		      if(INS_TYPE(ins) == IT_LOAD)
			printf("LOAD:\n");
		      else
			printf("STORE:\n");
		      DiabloPrint(stdout, "  @R\n", ins_reloc);
		      flag = TRUE;
		    }
		  DiabloPrint(stdout, "    @R\n", i_reloc);
#endif

#if 0 // Analyse
		  if(INS_BBL(T_INS(REL_FROM(ins_reloc))) == INS_BBL(ins) &&
		     RegsetIn(dead_after, ARMINSREGB(ins)))
		    {
		      if(!flag)
			{
			  if(INS_TYPE(ins) == IT_LOAD)
			    printf("LOAD (producer in same BBL as load and dead reg afer use):\n");
			  else
			    printf("STORE (producer in same BBL as store and dead reg afer use):\n");
			  DiabloPrint(stdout, "  @R\n", ins_reloc);
			  flag = TRUE;
			}
		      DiabloPrint(stdout, "    @R\n", i_reloc);
		    }
#endif

		  /* If the value of an relocation is potentially used for memory access,
		     its hotness will only increase slow. */
		  i_reloc->hotness += 1;

		  /* Update Section-connections (only if ins_reloc != i_reloc) */
		  if(REL_TO_RELOCATABLE(ins_reloc) != REL_TO_RELOCATABLE(i_reloc))
		    {
		      if(RegsetIn(dead_after, ARMINSREGB(ins)))
			{
			  /* If the used reg is dead after this ins, we will do a bigger effort
			     to bring the 2 subsections closer to eachother.*/
			  SectionConnUpdate(T_SECTION(REL_TO_RELOCATABLE(ins_reloc)), T_SECTION(REL_TO_RELOCATABLE(i_reloc)), (1 << 16));
			}
		      else
			{
			  SectionConnUpdate(T_SECTION(REL_TO_RELOCATABLE(ins_reloc)), T_SECTION(REL_TO_RELOCATABLE(i_reloc)), 1);
			}
		    }
		}
	    }
	}
    }
  else
    {
      REGSET_FOREACH_REG(INS_REGS_USE(ins), i_reg)
	{
	  if(VALUE == ProcStateGetTag(procstate1, i_reg, &i_reloc))
	    {
	      i_reloc->hotness += (1 << 8);
	    }
	}
    }

#if 0 // Debug
  if(flag)
    {
      DiabloPrint(stdout, "        @B\n", INS_BBL(ins));
      t_ins * temp_ins = BBL_INS_FIRST(INS_BBL(ins));
      while(temp_ins)
	{
	  if(temp_ins == ins)
	    DiabloPrint(stdout, "        ->@I\n", temp_ins);
	  else
	    DiabloPrint(stdout, "          @I\n", temp_ins);
	  temp_ins = INS_NEXT(temp_ins);
	}
    }
#endif

}

void ArmOptimizeSectionLayoutForAddresses(t_section * sec)
{
  t_section ** subs;
  t_section * temp_sec, * join_norm_a, * join_norm_b, * join_rev_a, * join_rev_b;
  t_section_conn * conn;
  t_address offset_norm, offset_rev;
  t_uint32 nsubs, j;

  subs = SectionGetSubsections(sec, &nsubs);
  printf("  #subsections in section: %d\n", nsubs);

  if(nsubs < 2) return;

  /* Make sure all subsections will be joined */
  for(j = 1; j < nsubs; j++)
    {
      SectionConnUpdate(subs[j-1], subs[j], 0);
    }

#if 0 // Debug
  SectionConnPrint(sec->section_connections, TRUE);
#endif

  /* A. Determin the best order and line them up, no joining yet */

  /* 1. Try to make very hot clusters of subsections that are frequently 
     used together. Keep large sections out of this clusters since they
     will reduce the opportinities to optimize the addressproducers. */
  conn = sec->section_connections;
mainloop:
  while(conn)
    {

      /* Calculate offset if joined as AB */
      offset_norm = NULL_GENERIC;

      temp_sec = conn->section_a->join_next;
      join_norm_a = conn->section_a;
      while(temp_sec)
	{
	  if(temp_sec == conn->section_b)
	    {
	      conn = conn->next;
	      goto mainloop;
	    }
	  join_norm_a = temp_sec;
	  offset_norm = GENERIC_ADD(offset_norm, SECTION_CSIZE(temp_sec));
	  temp_sec = temp_sec->join_next;
	}

      temp_sec = conn->section_b->join_prev;
      join_norm_b = conn->section_b;
      while(temp_sec)
	{
	  join_norm_b = temp_sec;
	  offset_norm = GENERIC_ADD(offset_norm, SECTION_CSIZE(temp_sec));
	  temp_sec = temp_sec->join_prev;
	}
      
      /* Calculate offset if joined as BA */
      offset_rev = NULL_GENERIC;

      temp_sec = conn->section_a->join_prev;
      join_rev_a = conn->section_a;
      while(temp_sec)
	{
	  if(temp_sec == conn->section_b)
	    {
	      conn = conn->next;
	      goto mainloop;
	    }
	  join_rev_a = temp_sec;
	  offset_rev = GENERIC_ADD(offset_rev, SECTION_CSIZE(temp_sec));
	  temp_sec = temp_sec->join_prev;
	}

      temp_sec = conn->section_b->join_next;
      join_rev_b = conn->section_b;
      while(temp_sec)
	{
	  join_rev_b = temp_sec;
	  offset_rev = GENERIC_ADD(offset_rev, SECTION_CSIZE(temp_sec));
	  temp_sec = temp_sec->join_next;
	}

      /* Take the smallest offset and line up de clusters */
      if(GENERIC_LE(offset_norm, offset_rev))
	{
	  join_norm_a->join_next = join_norm_b;
	  join_norm_b->join_prev = join_norm_a;
	  sec->join_first = join_rev_a;
	}
      else
	{
	  join_rev_a->join_prev = join_rev_b;
	  join_rev_b->join_next = join_rev_a;
	  sec->join_first = join_norm_b;
	}

      conn = conn->next;
    }

  /* 2. Try to place the hottest relocations on easy to calculate offsets
     form their hottest remaining connections. */

  /* 3. Try to fill the gaps in a best-fit way. */

  /* B. Join the subsections for real now */

#if 1 // Debug
      temp_sec = sec->join_first;
      while(temp_sec)
	{
	  DiabloPrint(stdout, "    %s\n", temp_sec->obj->name);
	  temp_sec = temp_sec->join_next;
	}
#endif

  temp_sec = sec->join_first->join_next;
  while(temp_sec)
    {
      SectionAppend(sec->join_first, temp_sec);
      temp_sec = temp_sec->join_next;
    }
}

void ArmOptimizeLayoutForAddresses(t_object * obj)
{
  t_uint32 i;
  t_bbl * bbl;
  t_ins * ins;

#if 0
  printf("Optimizing layout for Code-sections. (%d)\n", obj->ncodes);
  for(i = 0 ; i < obj->ncodes; i++)
    ArmOptimizeSectionLayoutForAddresses(obj->code[i]);
#endif

  printf("Optimizing layout for Data-sections. (%d)\n", obj->ndatas);
  for(i = 0 ; i < obj->ndatas; i++)
    ArmOptimizeSectionLayoutForAddresses(obj->data[i]);

  printf("Optimizing layout for Rodata-sections. (%d)\n", obj->nrodatas);
  for(i = 0 ; i < obj->nrodatas; i++)
    ArmOptimizeSectionLayoutForAddresses(obj->rodata[i]);

  printf("Optimizing layout for Bss-sections. (%d)\n", obj->nbsss);
  for(i = 0 ; i < obj->nbsss; i++)
    ArmOptimizeSectionLayoutForAddresses(obj->bss[i]);
  
  printf("Updating address producers.\n");
  CFG_FOREACH_BBL(T_CFG(obj->code[0]->data),bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      if(ARMINSOPCODE(ins) == ARM_ADDRESS_PRODUCER)
	{
	  ARMINSIMMEDIATE(ins) = 
	    INS_CADDRESS(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(ins)->rel)) + 
	    RELOCATABLE_REFERS_TO(ins)->rel->to_offset + 
	    RELOCATABLE_REFERS_TO(ins)->rel->addend;
	}
    }

  printf("Updating data-BLLs.\n");
  CFG_FOREACH_BBL(T_CFG(obj->code[0]->data),bbl)
    {
      if(BBL_ATTRIB(bbl) & BBL_IS_DATA)
	{
	  printf("Data-BBL found!!!\n");
	  BBL_FOREACH_INS(bbl,ins)
	    {
	      if(INS_REFERS_TO(ins))
		{
		  t_reloc * reloc = INS_REFERS_TO(ins)->rel;
		  DiabloPrint(stdout, "  @R\n", reloc);
		}
	    }
	}
    }
}
#endif

/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
