/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_copy_analysis.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "arm_utils.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../diablo_options.h"
/* }}} */
/*!
 * For a copy instruction, this function finds all uses of the copy in the
 * basic block and renames them to the original register whenever possible. 
 *
 * \param j_ins The instruction following the copy instruction
 * \param copy The copy of the register 
 * \param copied The original register
 *
 * \return int (0) if we can keep renaming in the following blocks, else (-1) 
*/
/* ArmOptCopyInner {{{ */
static int ArmOptCopyInner(t_ins * j_ins,t_reg copy, t_reg copied)
{
  static int debug = 0;
   while (j_ins!=NULL)
   {
/*     int old_debug = debug;*/
/*     if(debug >= global_options.debugcounter) return -1;*/
     /* don't propagate copies into multiply instructions: they have additional 
      * requirements about which registers may be used as operands.
      * we might check if these are fulfilled, but it's just easier to skip 
      * them altogether (possibly losing some opportunities that way)
      */
      if ((INS_TYPE(j_ins) != IT_MUL) && (!(ArmInsWriteBackHappens(j_ins)) || ARMINSREGB(j_ins)!=copy))
      {
	 if (ARMINSREGB(j_ins)==copy)
	 {
	   /*DiabloPrint(stdout,"Using %d instead of %d in @I\n",copied,copy,j_ins);*/
	    ARMINSREGB(j_ins) = copied; /* use original */
	    INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
	    debug++;
	 }
	 if (ARMINSREGC(j_ins)==copy)
	 {
	   if(!ArmInstructionIsNOOP(j_ins))
	   {
	   /*DiabloPrint(stdout,"Using %d instead of %d in @I\n",copied,copy,j_ins);*/
	    ARMINSREGC(j_ins) = copied; /* use original */
	    INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
	    debug++;

	    /* Got to be carefull here: if the instruction turns in a no-op badness happens (ask Bruno) */
	    if ((ARMINSOPCODE(j_ins)==ARM_MOV)&&(ARMINSREGA(j_ins)==copied))
	       INS_REGS_DEF(j_ins)= ArmDefinedRegisters(j_ins);
	   }
	 }
	 if (ARMINSREGS(j_ins)==copy)
	 {
	   /*DiabloPrint(stdout,"Using %d instead of %d in @I\n",copied,copy,j_ins);*/
	    ARMINSREGS(j_ins) = copied; /* use original */
	    INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
	    debug++;
	 }

	 if ((INS_TYPE(j_ins)==IT_STORE) && (ARMINSREGA(j_ins)==copy))
	 {
	   /*DiabloPrint(stdout,"Using %d instead of %d in @I\n",copied,copy,j_ins);*/
	    ARMINSREGA(j_ins) = copied; /* use original */
	    INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
	    debug++;
	 }
/*
	 if (ARMINSOPCODE(j_ins)==ARM_STM)
	    if (ARMINSREGB(j_ins)==ARM_REG_R13)
	       if (copy==ARM_REG_R12 && copied == ARM_REG_R13)
		  if (RegsetIn(ARMINSIMMEDIATE(j_ins),ARM_REG_R12))
		     if (!(RegsetIn(ARMINSIMMEDIATE(j_ins),ARM_REG_R13)))
		     {
			BblPrint(j_ins->bbl,PRINT_INSTRUCTIONS);
			printf("converted to \n");
			RegsetSubReg(ARMINSIMMEDIATE(j_ins),copy);
			RegsetAddReg(ARMINSIMMEDIATE(j_ins),copied);
			RegsetSubReg(INS_REGS_USE(j_ins),copy);
			RegsetAddReg(INS_REGS_USE(j_ins),copied);
			INS_ATTRIB(i_ins)|=IF_CHANGED;
			BblPrint(j_ins->bbl,PRINT_INSTRUCTIONS);
		     }
*/

      } 
      if(ARMINSOPCODE(j_ins) == ARM_MUL || ARMINSOPCODE(j_ins) == ARM_MLA)
      {
	if(ARMINSREGB(j_ins) == copy && ARMINSREGA(j_ins) != copied)
	{
/*          DiabloPrint(stdout,"Using %d instead of %d in @I\n",copied,copy,j_ins);*/
	  ARMINSREGB(j_ins) = copied; /* use original */
	  INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
	  debug++;
	}
	else if(ARMINSREGB(j_ins) == copy && ARMINSREGC(j_ins) != copy && ARMINSREGC(j_ins) != ARMINSREGA(j_ins))
	{
/*          DiabloPrint(stdout,"Using %d instead of %d in @I but switching regs!\n",copied,copy,j_ins);*/
	  ARMINSREGB(j_ins) = ARMINSREGC(j_ins);
	  ARMINSREGC(j_ins) = copied;
	  INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
	  debug++;
	}
	if(ARMINSREGC(j_ins) == copy)
	{
/*          DiabloPrint(stdout,"Using %d instead of %d in @I\n",copied,copy,j_ins);*/
	  ARMINSREGC(j_ins) = copied;
	  INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
	  debug++;
	}
      }

      if (RegsetIn(INS_REGS_DEF(j_ins),copy))
      {
	 return -1;
      }
      if (RegsetIn(INS_REGS_DEF(j_ins),copied))
      {
	 return -1;
      }

/*      if(old_debug != debug) VERBOSE(0,("Changed @I",j_ins));*/
      j_ins=INS_NEXT(j_ins);
   }
   return 0;
}
/* }}} */
/*!
 * Tries to follow all paths and applies ArmCopyInner on all basic blocks on the path 
 *
 * \param i_bbl a bbl in which copy propagation was already applied.
 * \param copy The copy of the register 
 * \param copied The original register
 *
 * \return void 
*/
/* ArmOptCopyRecurse {{{ */
static void ArmOptCopyRecurse(t_bbl * i_bbl, t_reg copy, t_reg copied)
{
  t_cfg_edge * edge,* edge2;
  t_uint32 real_preds;
  if (NodeIsMarked(i_bbl)) return;
  BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
   {
      EdgeMark(edge);

      /* don't propagate over function calls: this might change the callee-saved registers
       * of the function we propagate the constant into, and we can't handle that at the moment */
      if (global_options.rely_on_calling_conventions)
	if (EDGE_CAT(edge) == ET_CALL)
	  {
	    continue;
	  }
      
      if ((EDGE_CAT(edge)==ET_FUNLINK) || (EDGE_CAT(edge)==ET_IPFUNLINK)) 
      {
#if 1
	continue;
#else
	t_cfg_edge * edge2;
	BBL_FOREACH_SUCC_EDGE(EDGE_HEAD(edge),edge2)
	  if (EDGE_CAT(edge2)==ET_CALL) break;

	if (!edge2) 
	{
	  BblPrint(i_bbl,0xf);
	  FATAL(("LINK WITHOUT CALL! @E",edge));
	}
	if ((FUN_REGS_SAVED(BBL_FUN(EDGE_TAIL(edge2))) & (1<<copy) ) && (FUN_REGS_SAVED(BBL_FUN(EDGE_TAIL(edge2))) & (1<<copied) ) )
	  continue;
	real_preds=0;
	if(global_options.copy_analysis)// &&BBL_EQS_IN(CFG_EDGE_TAIL(edge)))
	/* We have extra information from extended copy-analysis*/
	{
	  if(EquationsRegsEqual(BBL_EQS_IN(CFG_EDGE_TAIL(edge)),copy,copied,0) == TRUE)
	  {
	    NodeMark(i_bbl);
	    if (!ArmOptCopyInner(BBL_INS_FIRST(CFG_EDGE_TAIL(edge)),copy,copied))
	    {
	      ArmOptCopyRecurse(CFG_EDGE_TAIL(edge),copy,copied);
	    }
	  }
	}
	else
	{
	  BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),edge2)
	  {
	    /* Fun link edges don't count as predecessor edges for copy propagation */
	    if ((EDGE_CAT(edge2)==ET_FUNLINK) || (EDGE_CAT(edge2)==ET_IPFUNLINK)) continue;
	    if (EDGE_CAT(edge2)==ET_SWITCH)  continue;
	    if (EDGE_CAT(edge2)==ET_RETURN)  continue;
	    if (EdgeIsMarked(edge2)) continue;
	    real_preds++;
	  }
	  if (real_preds==0)
	  {
	    NodeMark(i_bbl);
	    if (!ArmOptCopyInner(BBL_INS_FIRST(CFG_EDGE_TAIL(edge)),copy,copied))
	    {
	      ArmOptCopyRecurse(CFG_EDGE_TAIL(edge),copy,copied);
	    }
	  }
	}
#endif
      }
      else
      {
	if(global_options.copy_analysis)/* && BBL_EQS_IN(CFG_EDGE_TAIL(edge)))*/
	/* We have extra information from extended copy-analysis*/
	{
	  if(EquationsRegsEqual(BBL_EQS_IN(CFG_EDGE_TAIL(edge)),copy,copied,0,NULL,NULL) == TRUE)
	  {
	    NodeMark(i_bbl);
	    if (!ArmOptCopyInner(BBL_INS_FIRST(CFG_EDGE_TAIL(edge)),copy,copied))
	    {
	      ArmOptCopyRecurse(CFG_EDGE_TAIL(edge),copy,copied);
	    }
	  }
	}
	else
	{
	  real_preds=0;
	  BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),edge2)
	  {
	    /* Fun link edges don't count as predecessor edges for copy propagation */
	    if ((EDGE_CAT(edge2)==ET_FUNLINK) || (EDGE_CAT(edge2)==ET_IPFUNLINK)) continue;
	    if (EDGE_CAT(edge2)==ET_SWITCH)  continue;
	    if (EdgeIsMarked(edge2)) continue;
	    real_preds++;
	  }
	  if (real_preds==0)
	  {
	    NodeMark(i_bbl);
	    if (!ArmOptCopyInner(BBL_INS_FIRST(CFG_EDGE_TAIL(edge)),copy,copied))
	    {
	      ArmOptCopyRecurse(CFG_EDGE_TAIL(edge),copy,copied);
	    }
	  }
	}
      }
   }
}
/* }}} */
/*!
 * \todo Document and move to arm_instructions 
 *
 * \param i_ins
 * \param copy
 * \param original
 *
 * \return t_bool 
*/
/* ArmInsIsCopy {{{ */
t_bool ArmInsIsCopy(t_ins * i_ins, t_reg * copy, t_reg * original)
{
   if (INS_IS_CONDITIONAL(i_ins)) return FALSE;
   if (ArmInsHasShiftedFlexible(i_ins)) return FALSE; 
   if (ARMINSOPCODE(i_ins)!=ARM_MOV) return FALSE;
   if (ARMINSREGC(i_ins)==ARM_REG_NONE) return FALSE;
   if (ARMINSREGC(i_ins)==ARM_REG_R15) return FALSE;
   if (ARMINSREGA(i_ins)==ARM_REG_R15) return FALSE;
   if (ARMINSREGC(i_ins)==ARMINSREGA(i_ins) && (!(ARMINSFLAGS(i_ins)&FL_S)))
   {
      ARMINSREGA(i_ins)=ARMINSREGC(i_ins)=0;
      INS_REGS_USE(i_ins)=NullRegs;
      INS_REGS_DEF(i_ins)=NullRegs;
      return FALSE;
   }
   *original = ARMINSREGC(i_ins); /* Original value */
   *copy = ARMINSREGA(i_ins);   /* Copy */
   return TRUE;
}
/* }}} */
/*!
 * Copy propagation for the arm. This optimization tries to eliminate
 * unnecessary register copies. Calls ArmOptCopyInner to find a candidate and
 * rename copies in one bbl and calls ArmOptCopyRecurse to rename copies in the
 * following blocks.
 *
 * \param cfg The flowgraph to work on
 *
 * \return void 
*/
/* ArmOptCopyPropagation {{{ */
void ArmOptCopyPropagation(t_cfg * cfg)
{
  t_object * obj=CFG_OBJECT(cfg);
  t_bbl * i_bbl;
  t_ins * i_ins, *j_ins;
  t_reg copied;
  t_reg copy;

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (!BBL_FUN(i_bbl)) continue;

    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      if (!obj->actions->IsCopy(i_ins,&copy,&copied)) continue;
      j_ins=INS_NEXT(i_ins);

      if (ArmOptCopyInner(j_ins,copy,copied)) continue;
      InitNodeMarking();
      InitEdgeMarking();
      ArmOptCopyRecurse(i_bbl,copy,copied);
    }
  }
}
/* }}} */

/*{{{*/
void ArmCopyInstructionPropagator(t_ins * ins, t_equations eqs)
{
  t_reg reg;
/*  DiabloPrint(stdout,"Propping @I\n",ins);*/
  if(!INS_IS_CONDITIONAL(ins))
  {
    if(ARMINSOPCODE(ins) == ARM_ADD)
    {
      if(ARMINSFLAGS(ins) & FL_IMMED)
      {
        EquationsAdd(eqs,ARMINSREGA(ins),ARMINSREGB(ins),0-ARMINSIMMEDIATE(ins),NULL,NULL);
	return;
      }
    }
    else if(ARMINSOPCODE(ins) == ARM_SUB)
    {
      if(ARMINSFLAGS(ins) & FL_IMMED)
      {
        EquationsAdd(eqs,ARMINSREGA(ins),ARMINSREGB(ins),ARMINSIMMEDIATE(ins),NULL,NULL);
	return;
      }
    }
    else if(ARMINSOPCODE(ins) == ARM_MOV)
    {
      if(ARMINSREGA(ins) != ARM_REG_NONE && ARMINSREGC(ins) != ARM_REG_NONE && ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_NONE)
      {
        EquationsAdd(eqs,ARMINSREGA(ins),ARMINSREGC(ins),0,NULL,NULL);
	return;
      }
      else if(ARMINSREGC(ins) == ARM_REG_NONE && ARMINSFLAGS(ins) & FL_IMMED)
      {
	EquationsAdd(eqs,ARMINSREGA(ins),null_register,0-ARMINSIMMEDIATE(ins),NULL,NULL);
	return;
      }
    }
    else if(ARMINSOPCODE(ins) == ARM_MVN)
    {
      if(ARMINSREGC(ins) == ARM_REG_NONE && ARMINSFLAGS(ins) & FL_IMMED)
      {
	t_uint32 imm = ARMINSIMMEDIATE(ins);
	imm = ~imm;
	EquationsAdd(eqs,ARMINSREGA(ins),null_register,0-ARMINSIMMEDIATE(ins),NULL,NULL);
	return;
      }
    }
    else if(ARMINSOPCODE(ins) == ARM_STR && ARMINSFLAGS(ins) & FL_WRITEBACK && ARMINSFLAGS(ins) & FL_IMMED)
    /* Case : str  r2,[r8,#12]!  or str r2,[r8],#12 */
    {
      if(ARMINSFLAGS(ins) & FL_DIRUP)
        EquationsAdd(eqs,ARMINSREGB(ins),ARMINSREGB(ins),0-ARMINSIMMEDIATE(ins),NULL,NULL); 
      else
        EquationsAdd(eqs,ARMINSREGB(ins),ARMINSREGB(ins),ARMINSIMMEDIATE(ins),NULL,NULL); 
      return;
    }
    else if(ARMINSOPCODE(ins) == ARM_LDR && ARMINSFLAGS(ins) & FL_WRITEBACK && ARMINSFLAGS(ins) & FL_IMMED)
    /* Case : ldr  r2,[r8,#12]!  or ldr r2,[r8],#12 */
    {
      if(ARMINSFLAGS(ins) & FL_DIRUP)
        EquationsAdd(eqs,ARMINSREGB(ins),ARMINSREGB(ins),0-ARMINSIMMEDIATE(ins),NULL,NULL); 
      else
        EquationsAdd(eqs,ARMINSREGB(ins),ARMINSREGB(ins),ARMINSIMMEDIATE(ins),NULL,NULL); 
      EquationsInvalidate(eqs,ARMINSREGA(ins)); 
      return;
    }
    else if((ARMINSOPCODE(ins) == ARM_STM || ARMINSOPCODE(ins) == ARM_LDM) && (ARMINSFLAGS(ins) & (FL_WRITEBACK) ))
    /* Case : stm/ldm r13!,(r1,r2,r5,r9) */
    {
#if  MAX_NR_REG_SUBSETS > 4
      FATAL(("Implement me with real regsets!"));
#else
      t_int32 nr_regs = RegsetCountRegs(ARMINSIMMEDIATE(ins));
      nr_regs *= 4;
      if(ARMINSFLAGS(ins) & FL_DIRUP)
        EquationsAdd(eqs,ARMINSREGB(ins),ARMINSREGB(ins),0-nr_regs,NULL,NULL);
      else
        EquationsAdd(eqs,ARMINSREGB(ins),ARMINSREGB(ins),nr_regs,NULL,NULL);
      if(ARMINSOPCODE(ins) == ARM_LDM)
        REGSET_FOREACH_REG(ARMINSIMMEDIATE(ins),reg)
	  EquationsInvalidate(eqs,reg);
      return;
#endif
    }
    else if(ARMINSOPCODE(ins) == ARM_CONSTANT_PRODUCER)
    {
      EquationsAdd(eqs,ARMINSREGA(ins),null_register,0-ARMINSIMMEDIATE(ins),NULL,NULL);
      return;
    }
    else if(ARMINSOPCODE(ins) == ARM_ADDRESS_PRODUCER)
    {
      /*DiabloPrint(stdout,"Adding address prod with reloc @R\n",RELOCATABLE_REFERS_TO(ins)->rel);*/
      EquationsAdd(eqs,ARMINSREGA(ins),null_register,0,RELOCATABLE_REFERS_TO(ins)->rel,NULL);
      return;
    }
  }
  REGSET_FOREACH_REG(INS_REGS_DEF(ins),reg)
  {
    if(RegsetIn(INS_OBJECT(ins)->description->int_registers,reg))
      EquationsInvalidate(eqs,reg);
  }
  return;
}/*}}}*/
/* vim: set shiftwidth=2 foldmethod=marker : */
