/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_deflowgraph.h"
#include "arm_instructions.h"
#include "arm_layout.h"
#include "../../diablo_options.h"
#include "../../utils/util_error.h"
#include "../../diablo_options.h"
#include "../../utils/util_malloc.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"
#include <string.h>
/* }}} */
/*!
 * Layouts a flowgraph, and builds a linear list of instructions that can be executed. 
 *
 * \todo needs only one parameter
 *
 * \param obj Remove this parameter
 * \param code The section to layout (we actually take the flowgraph from this
 * section)
 *
 * \return void 
*/
/* ArmDeflowgraphSection {{{ */
#ifdef INSTRUMENTATION
extern t_section * added_section;
#endif
void ArmDeflowgraphSection(t_object * obj, t_section * code)
{
  t_cfg * cfg = SECTION_CFG(code);
  t_bbl * bbl;
  t_uint32 size=0;
  t_uint32 tel=0;

  size=ArmCfgLayout(cfg,0);
  SECTION_TMP_BUF(code)=Calloc(sizeof(t_arm_ins),(size/4+2));

  VERIFY_FASSERT(SECTION_TMP_BUF(code),("Failed to allocate buffer for instructions!"));
  /* remove all relocations that point to basic blocks
   * this is a temporary measure, the correct thing to do is
   * move the relocations from the bbl to the first instruction of
   * the bbl */
  CFG_FOREACH_BBL(cfg,bbl)
  {
    while (RELOCATABLE_REFED_BY(bbl))
      RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFED_BY(bbl)->rel);

    VERIFY_FASSERT(RELOCATABLE_REFERS_TO(bbl) == NULL,("Relocations corrupt"));
#ifdef INSTRUMENTATION
    
    if ((BBL_INS_FIRST(bbl)) 
	&& (INS_TYPE(BBL_INS_FIRST(bbl))==IT_DATA) 
	&& (GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl))!=0) 
	&& (GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl))!=0xffffffff) 
	&& RELOCATABLE_REFED_BY(bbl))
    {
      TransTableAddElem32(obj,GENERIC_EXTRACT_UINT32(BBL_CADDRESS(bbl)),GENERIC_EXTRACT_UINT32(BBL_CADDRESS(bbl))+GENERIC_EXTRACT_UINT32(BBL_CSIZE(bbl)),GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl)));
    }

#endif
  }
  
  CFG_FOREACH_BBL(cfg,bbl)
  {
    t_uint32 addr;
    t_arm_ins * current=(t_arm_ins *) bbl->ins_first;
    if (!current) continue;

    if(!BBL_NEXT_IN_CHAIN(bbl)) WARNING(("No next in chain! @iB\n",bbl));
    if(!BBL_PREV_IN_CHAIN(bbl)) WARNING(("No prev in chain! @iB\n",bbl));

    addr=G_T_UINT32(GENERIC_SUB(BBL_CADDRESS(bbl),SECTION_GET_CADDRESS(code)))/4;

    while(current)
    {
      void * result;
      t_arm_ins * to;
      /* sanity check */
      to = &((t_arm_ins *) SECTION_TMP_BUF(code))[addr];
      if (ARMINSREGA(to)+ARMINSREGB(to)+ARMINSREGC(to)+ARMINSREGS(to)+ARMINSIMMEDIATE(to))
	FATAL(("Overwriting already written instructions at offset %d: @I in \n@ieB\n@I",addr,current,bbl,to));

      tel++;
      INS_COPY(current)=(t_ins *) to;
      result=memcpy(to,current,sizeof(t_arm_ins));
      INS_BBL(to)=NULL;

      RELOCATABLE_REFERS_TO(INS_COPY(current))=NULL;
      RELOCATABLE_REFED_BY(INS_COPY(current))=NULL;

      while(RELOCATABLE_REFERS_TO(((t_ins *) current))) { RelocSetFrom(RELOCATABLE_REFERS_TO(((t_ins *) current))->rel,T_RELOCATABLE(INS_COPY(current))); }
      while(RELOCATABLE_REFED_BY(((t_ins *) current))) { FATAL(("Should not get here!")); }
      
      addr++;
      current=(t_arm_ins *) INS_NEXT(current);
    }
  }

  printf("Done, wrote %d\n",tel);

  for (tel=0; tel<size/4-1; tel++)
  {
    INS_NEXT(&((t_arm_ins *) SECTION_TMP_BUF(code))[tel])=(t_ins *) (&((t_arm_ins *) SECTION_TMP_BUF(code))[tel+1]);
    INS_BBL(&((t_arm_ins *) SECTION_TMP_BUF(code))[tel]) = NULL;
  }

  INS_NEXT(&((t_arm_ins *) SECTION_TMP_BUF(code))[size/4-1])=NULL;

  CfgFree(SECTION_CFG(code));

  code->data=SECTION_TMP_BUF(code);
  code->type=DISASSEMBLED_CODE_SECTION;

  printf("Counted till %d\n",tel);
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
