/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_registers.h"
#include "arm_disassemble.h"
#include "arm_disassemble_one.h"
#include "arm_instructions.h"
#include "arm_opcodes.h"
#include "arm_utils.h"
#include "../../utils/util_error.h"
#include "../../utils/util_malloc.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_symbol.h"
#include "thumb_opcodes.h"
#include "thumb_disassemble_one.h"
#include "thumb_print.h"
/* }}} */
/*!
 * Disassemble an instruction at a given address. Not yet implemented. 
 *
 * \param obj
 * \param start
 * \param size_ret
 *
 * \return void *
*/
/* ArmDisassembleOneInstruction {{{ */
void * ArmDisassembleOneInstruction(t_object * obj, t_address start, int * size_ret)
{
  FATAL(("Implement"));
  return NULL; /* Keep the compiler happy */
}
/* }}} */
/*!
 * Disassemble an instruction at a given offset in a section. 
 *
 * \todo parameter size_ret seems unused!
 * 
 * \param sec The section in which we want to disassemble something
 * \param offset The offset in the section
 * \param size_ret seems unused!
 *
 * \return void * a void casted t_arm_ins pointer
*/
/* ArmDisassembleOneInstructionForSection {{{ */
void * ArmDisassembleOneInstructionForSection(t_section * sec, t_uint32 offset, int * size_ret)
{
  t_ins * ret=NULL;
  t_uint32 encoded_instruction;
  t_uint16 i;
  t_address goffset;

  GenericSet_uint32(goffset,offset);
  if (G_T_UINT32(SECTION_CSIZE(sec))>offset)
  {
    ret=Malloc(sizeof(t_arm_ins));
    encoded_instruction=SectionGetData32(sec,goffset);
    /* Find the opcode in the opcode table */
    for (i = 0; (encoded_instruction & arm_opcode_table[i].mask) != arm_opcode_table[i].opcode; i++);
    arm_opcode_table[i].Disassemble(ret,encoded_instruction,i);

    /* Initialize the other fields */
    INS_PREV(ret) = INS_NEXT(ret) = NULL; 
    INS_REGS_USE(ret)=ArmUsedRegisters(ret);
    INS_REGS_DEF(ret)=ArmDefinedRegisters(ret);
    GenericSet_uint32(INS_CSIZE(ret),4);
    GenericSet_uint32(INS_OLD_SIZE(ret),0);
    
  }
  return ret;
}
/* }}} */
/*!
 * Disassemble an entire section. Creates an array of t_arm_ins's.
 *
 * \todo Needs only one parameter
 *
 * \param obj Not needed
 * \param code The section we want to disassemble
 *
 * \return void 
*/
/* ArmDisassembleSection {{{ */
void ArmDisassembleSection(t_object * obj, t_section * code)
{
  t_address offset;
  t_address o_offset;
  t_uint16 i;
  t_ins * ins_s=NULL;
  t_ins * prev;
  t_uint32 data;
  t_bool previous_set=FALSE;
  t_string out;
  t_address previous=NULL_GENERIC; /* Just to keep the compiler happy */
  out = (t_string)Malloc(100);

  GenericSet_uint32(offset,0);
  GenericSet_uint32(o_offset,0);

  prev = NULL;
  while(GENERIC_LT(offset, SECTION_CSIZE(code)))
  {
    ins_s = InsNewForSec(code);
    if (!ins_s) FATAL(("No instructions!"));
    RELOCATABLE_REFERS_TO(ins_s)=NULL;  /* TODO: Check if this is necessary */
    RELOCATABLE_REFED_BY(ins_s)=NULL; /* TODO: Check if this is necessary */
    if(SymbolTableGetCodeType(obj, GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) == ADDRESS_IS_SMALL_CODE)
    {
      if (SymbolTableGetDataType(obj, GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) == ADDRESS_IS_DATA) 
      {
	if(GENERIC_EXTRACT_UINT32(GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) & 0x3)
	{
	  data=SectionGetData16(code,offset);
	  ThumbDisassembleData(ins_s,(data),ARM_DATA);
	  /*FATAL(("16 bits data here!")); */
	  GenericSet_uint32(INS_CSIZE(ins_s),2);
	  GenericSet_uint32(INS_OLD_SIZE(ins_s),2);
	}
	else
	{
	  data=SectionGetData32(code,offset);
	  ArmDisassembleData(ins_s,(data),ARM_DATA);
	  GenericSet_uint32(INS_CSIZE(ins_s),4);
	  GenericSet_uint32(INS_OLD_SIZE(ins_s),4);
	}
      }
      else
      {
	data=SectionGetData16(code,offset);
	for (i = 0; ((data) & thumb_opcode_table[i].mask) != thumb_opcode_table[i].opcode; i++);
	fflush(stdout);
	if (((data) & 0xf000) == 0xf000)
	{
	  data=SectionGetData32(code,offset);
	  ThumbDisassembleBranchLink(ins_s,(data),i);
	  INS_REGS_USE(ins_s)=ArmUsedRegisters(ins_s);
	  INS_REGS_DEF(ins_s)=ArmDefinedRegisters(ins_s);
	  GenericSet_uint32(INS_CSIZE(ins_s),4);
	  GenericSet_uint32(INS_OLD_SIZE(ins_s),4);
	}
	else
	{
	  thumb_opcode_table[i].Disassemble(ins_s,(data),i);
	  INS_REGS_USE(ins_s)=ArmUsedRegisters(ins_s);
	  INS_REGS_DEF(ins_s)=ArmDefinedRegisters(ins_s);
	  /*	if (RegsetIn(INS_REGS_DEF(ins_s),ARM_REG_R15)) INS_ATTRIB(ins_s) |= IF_PCDEF; */
	  GenericSet_uint32(INS_CSIZE(ins_s),2);
	  GenericSet_uint32(INS_OLD_SIZE(ins_s),2);
	}
	ARMINSFLAGS(ins_s) |= FL_THUMB;
	DiabloPrint(stdout, "@I\n", ins_s);
	fflush(stdout);
      }
    }
    else
    {
      if(GENERIC_EXTRACT_UINT32(GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) & 0x3) FATAL(("Not aligned in arm modus!"));
      data=SectionGetData32(code,offset);
      if (SymbolTableGetDataType(obj, GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) == ADDRESS_IS_DATA) 
      {
	ArmDisassembleData(ins_s,(data),ARM_DATA);
	GenericSet_uint32(INS_CSIZE(ins_s),4);
	GenericSet_uint32(INS_OLD_SIZE(ins_s),4);
      }
      else
      {
	for (i = 0; ((data) & arm_opcode_table[i].mask) != arm_opcode_table[i].opcode; i++);
	arm_opcode_table[i].Disassemble(ins_s,(data),i);

	INS_REGS_USE(ins_s)=ArmUsedRegisters(ins_s);
	INS_REGS_DEF(ins_s)=ArmDefinedRegisters(ins_s);
	if (ARMINSCONDITION(ins_s)!=ARM_CONDITION_AL)
	  INS_ATTRIB(ins_s) |= IF_CONDITIONAL;
	if (RegsetIn(INS_REGS_DEF(ins_s),ARM_REG_R15)) INS_ATTRIB(ins_s) |= IF_PCDEF;
	GenericSet_uint32(INS_CSIZE(ins_s),4);
	GenericSet_uint32(INS_OLD_SIZE(ins_s),4);
      }

    }
    INS_CADDRESS(ins_s) = GENERIC_ADD(T_RELOCATABLE(code)->caddress,offset);

    /* Bookkeeping to give the old_address a meaningfull value {{{*/
    if (obj->n_orig_codes)
    {
      t_bool found=FALSE;
      t_uint32 tel;
      for (tel=0;tel<obj->n_orig_codes; tel++)
      {
	if ((GENERIC_LE(obj->orig_codes[tel].base,INS_CADDRESS(ins_s)))&&(GENERIC_GT(GENERIC_ADD(obj->orig_codes[tel].base,obj->orig_codes[tel].size),INS_CADDRESS(ins_s))))
	{
	  found = TRUE;

	  if (previous_set && (!GENERIC_EQ(previous,obj->orig_codes[tel].obase)))
	  {
	    GenericSet_uint32(o_offset,0);
	  }
	  previous=obj->orig_codes[tel].obase;
	  previous_set=TRUE;
	  if (G_T_UINT32(obj->orig_codes[tel].obase)==0x0) GenericSet_uint32(INS_OLD_ADDRESS(ins_s),0xffffffff);
	  else  INS_OLD_ADDRESS(ins_s) = GENERIC_ADD(obj->orig_codes[tel].obase,o_offset);
	  break;
	}
      }
      if (!found) 
      {
	if (!previous_set)
	{
	  FATAL(("Could not map merged address (@G -- @I) to original",INS_CADDRESS(ins_s),ins_s));
	}
	else
	{
	  INS_OLD_ADDRESS(ins_s) = GENERIC_ADD(previous,o_offset);
	}
      }
    }
    else
    {
      /*      FATAL(("HIER\n")); Commented out, because i couldn't find out why this is fatal? */
      INS_OLD_ADDRESS(ins_s) = GENERIC_ADD(SECTION_GET_CADDRESS(code),offset);
    }/*}}}*/

    INS_PREV(ins_s) = T_INS(prev);
    if (prev) INS_NEXT(prev) = ins_s;
    prev=ins_s;
    if(GENERIC_EXTRACT_UINT32(INS_CSIZE(ins_s)) == 4)
    {
      offset=GENERIC_ADD_UINT32(offset,4);
      o_offset=GENERIC_ADD_UINT32(o_offset,4);
    }
    else if (GENERIC_EXTRACT_UINT32(INS_CSIZE(ins_s)) == 2)
    {
      offset=GENERIC_ADD_UINT32(offset,2);
      o_offset=GENERIC_ADD_UINT32(o_offset,2);
    }
    else FATAL(("Impossible CSIZE for instructions!"));
  }
  INS_NEXT(ins_s) = NULL;

#if 0  /* Pre-thumb code {{{*/
  for ( prev = NULL; GENERIC_LT(offset, SECTION_CSIZE(code)); offset=GENERIC_ADD_UINT32(offset,4) , o_offset=GENERIC_ADD_UINT32(o_offset,4), prev=ins_s)
  {
    ins_s = InsNewForSec(code);
    data=SectionGetData32(code,offset);
    if (!ins_s) FATAL(("No instructions!"));
    RELOCATABLE_REFERS_TO(ins_s)=NULL;  /* TODO: Check if this is necessary */
    RELOCATABLE_REFED_BY(ins_s)=NULL; /* TODO: Check if this is necessary */


    /* Use the symbol table to find out if this is data or code */
    if (SymbolTableGetDataType(obj, GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) == ADDRESS_IS_DATA) 
    {
      /*DiabloPrint(stdout,"Data at @G\n",INS_CADDRESS(ins)); */
      ArmDisassembleData(ins_s,(data),ARM_DATA);
    }
    else
    {
      for (i = 0; ((data) & arm_opcode_table[i].mask) != arm_opcode_table[i].opcode; i++);
      arm_opcode_table[i].Disassemble(ins_s,(data),i);
    }

    INS_PREV(ins_s) = T_INS(prev);
    INS_REGS_USE(ins_s)=ArmUsedRegisters(ins_s);
    INS_REGS_DEF(ins_s)=ArmDefinedRegisters(ins_s);
    if (ARMINSCONDITION(ins_s)!=ARM_CONDITION_AL)
      INS_ATTRIB(ins_s) |= IF_CONDITIONAL;
    if (RegsetIn(INS_REGS_DEF(ins_s),ARM_REG_R15)) INS_ATTRIB(ins_s) |= IF_PCDEF;
    if (prev) INS_NEXT(prev) = ins_s;
    INS_CADDRESS(ins_s) = GENERIC_ADD(T_RELOCATABLE(code)->caddress,offset);
    GenericSet_uint32(INS_CSIZE(ins_s),4);

    if (obj->n_orig_codes)
    {
      t_bool found=FALSE;
      t_uint32 tel;
      for (tel=0;tel<obj->n_orig_codes; tel++)
      {
	if ((GENERIC_LE(obj->orig_codes[tel].base,INS_CADDRESS(ins_s)))&&(GENERIC_GT(GENERIC_ADD(obj->orig_codes[tel].base,obj->orig_codes[tel].size),INS_CADDRESS(ins_s))))
	{
	  found = TRUE;

	  if (previous_set && (!GENERIC_EQ(previous,obj->orig_codes[tel].obase)))
	  {
	    GenericSet_uint32(o_offset,0);
	  }
	  previous=obj->orig_codes[tel].obase;
	  previous_set=TRUE;
	  if (G_T_UINT32(obj->orig_codes[tel].obase)==0x0) GenericSet_uint32(INS_OLD_ADDRESS(ins_s),0xffffffff);
	  else  INS_OLD_ADDRESS(ins_s) = GENERIC_ADD(obj->orig_codes[tel].obase,o_offset);
	  break;
	}
      }
      if (!found) 
      {
	if (!previous_set)
	{
	  FATAL(("Could not map merged address (@G -- @I) to original",INS_CADDRESS(ins_s),ins_s));
	}
	else
	{
	  INS_OLD_ADDRESS(ins_s) = GENERIC_ADD(previous,o_offset);
	}
      }
    }
    else
    {
      FATAL(("HIER\n"));
      INS_OLD_ADDRESS(ins_s) = GENERIC_ADD(SECTION_GET_CADDRESS(code),offset);
    }
    GenericSet_uint32(INS_OLD_SIZE(ins_s),4);

    
    if (0) { 
      t_symbol * sym=SymbolTableGetSymbolByAddress(obj->sub_symbol_table, INS_CADDRESS(ins_s));


      if (sym) printf("%s:\n",sym->name);
      DiabloPrint(stdout,"@I\n",ins_s);
    }
  }
  INS_NEXT(ins_s) = NULL;
#endif /*}}}*/
  Free(out);
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
