/*****************************************************************************
 * Copyright 2001,2002,2003: {{{
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This file contains all arch-independent functions to handle basic blocks in
 * objects
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002,2003 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (ARM Backend)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_instructions.h"
#include "arm_registers.h"
#include "../../kernel/diablo_bbl.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_factor.h"
#include "../../kernel/diablo_iterators.h"
/* }}} */
#define NOCOPY 33
#define FREE   34

#ifdef FACTOR_SUPPORT
extern t_uint32 killcount;  				/*Defined in diablo_factor.c*/
extern t_uint32 addcount;
extern t_uint32 count;
#else
t_uint32 killcount;
t_uint32 addcount;
t_uint32 count;
#endif

/*ArmBblCmpExtReg:	Compare used external registers{{{*/
t_uint32 ArmBblCmpExtReg(t_bbl * master,t_bbl * slave,t_uint32 * in)
{
  t_ins* ins1=BBL_INS_FIRST(master);
  t_ins* ins2=BBL_INS_FIRST(slave);

  t_uint32 def1[32]={0}; /*Keeps track of internally defined registers*/
  t_uint32 def2[32]={0};	
  
  while(ins1)
  {
    t_reg reg1b=ARMINSREGB(ins1);
    t_reg reg1c=ARMINSREGC(ins1);
    t_reg reg2b=ARMINSREGB(ins2);
    t_reg reg2c=ARMINSREGC(ins2);

    DiabloPrint(stdout,"@I\t\t@I\n",ins1,ins2);
    
    if ((reg1b!=255) && (reg1b!=reg2b) && (def2[reg2b]==0)) /*if reg1b diffs from reg2b and they're not internal*/
    {
      if((in[reg1b]==NOCOPY)||(in[reg1b]==reg2b)) in[reg1b]=reg2b;
      else return 1;			/*two regs to be copied to the same -> abstraction impossible*/
    }
    
    if ((reg1c!=255) && (reg1c!=reg2c) && (def2[reg2c]==0))
    {
      if((in[reg1c]==NOCOPY)||(in[reg1c]==reg2c)) in[reg1c]=reg2c;
      else return 1;			/*two regs to be copied to the same -> abstraction impossible*/
    }

    def1[ARMINSREGA(ins1)]=1;		/*insert defined register into array*/
    def2[ARMINSREGA(ins2)]=1;		/*TODO: what about LDMIA?           */

    ins1=INS_NEXT(ins1);
    ins2=INS_NEXT(ins2);
  }
  return 0;
}/*}}}*/

/*FindFreeReg:	Find a free register to resolve copy ops loops{{{*/
t_uint32 FindFreeReg(t_uint32 * in)
{
  int i;
  for(i=0;i<32;i++) if(in[i]==FREE)
  {
    in[i]=NOCOPY;
    return i;
  }
  return NOCOPY;
}/*}}}*/

/*ArmBblAbstractSlave: Detect Copy Op Loops. Insert Copy Ops in slave block. Insert call instr. Split bbl.{{{*/
t_uint32 ArmBblAbstractSlave(t_bbl * master,t_bbl * slave,t_uint32 * in,t_uint32 * out)
{
 t_uint32 i,j,k=0; /* Check assignment ... */
 t_uint32 inloop[32],outloop[32];
 t_ins * last=BBL_INS_LAST(slave);
 t_ins * it,* it2;
 
 t_uint32 type,linklive;
 t_ins * i_ins,* split;
 t_cfg_edge * i_edge,* i_edgetmp;
 t_bbl * B=slave;
 t_bbl * C=BblNew(slave->cfg);

 for(i=0;i<32;i++) inloop[i]=outloop[i]=NOCOPY;
 split=last;
  
 if(ARMINSOPCODE(split)==ARM_B||ARMINSOPCODE(split)==ARM_BX) type=1;
 else if(ARMINSOPCODE(split)==ARM_BL) type=2;
 else if(ARMINSREGA(split)==ARM_REG_R15 && ARMINSREGC(INS_PREV(split))==ARM_REG_R15)
 {
   type=3;
   split=INS_PREV(split);
 }
 else if(ARMINSREGA(split)==ARM_REG_R15) type=4;
 else if(ARMINSOPCODE(split)==ARM_LDM && (ARMINSIMMEDIATE(split)&0x00008000)) type=5;
 else
 {
   type=6;
   split=NULL;
 }
#if MAX_NR_REG_SUBSETS > 4
 FATAL(("Implement me!"));
#else
 if((slave->regs_defined & 0x00004000)|(slave->regs_used & 0x00004000)) return 2; 

 if(InsRegsLiveBefore(BBL_INS_FIRST(slave))&0x00004000) linklive=1;
 else linklive=0;
#endif
 if(type==1)
 {
   /*split bbl*/
   BBL_INS_LAST(C)=BBL_INS_LAST(B);
   BBL_INS_FIRST(C)=split;
   INS_NEXT(INS_PREV(split))=NULL;
   BBL_INS_LAST(B)=INS_PREV(split);
   INS_PREV(split)=NULL;
   
   BBL_FOREACH_INS(C,i_ins)
   {
     INS_BBL(i_ins)=C;
     BBL_NINS(C)++;
     BBL_NINS(B)--;
   } 
   
   /*add call instruction*/
   ArmInstructionMakeCondBranchAndLink(BBL_INS_FIRST(B),ARM_CONDITION_AL);

   /*remove instructions from B, except BL-instr*/
   it=INS_NEXT(BBL_INS_FIRST(B));
   while(it!=NULL)
   {
   it2=INS_NEXT(it);
   InsKill(it);
   killcount++;
   it=it2;
   } 
   
   /*sanity check*/
   if(BBL_NINS(B)!=1) FATAL(("B zou alleen uit BL mogen bestaan..."));
   
   /*insert copy ops in B and C*/
   if(in!=NULL){			/*No copy ops for completely identical slave*/
 /*Loop Detection{{{*/
  for(i=0;i<32;i++)
  { 
   j=in[i];
   while(j!=NOCOPY && j!=FREE && j!=i) 
   {
     k=j;
     j=in[k];
   } 
   
   if(j==i)
   {
     j=FindFreeReg(in);
     if(j==NOCOPY) return 1;    /*No free register found: abstraction is not possible*/
     inloop[j]=i;        	/*Add extra copy-instruction to inloop[]*/
     inloop[i]=in[i];		/*Add original copy-intruction to inloop[]*/
     in[i]=NOCOPY;		/*Remove original copy-ins from in[] to break loop*/
     in[k]=j;                   /*Modify copy-instruction in in[]*/ 
   } 
  }
  
  
  for(i=0;i<32;i++)
  { 
   j=out[i];
   while(j!=NOCOPY && j!=FREE && j!=i) 
   {
     k=j;
     j=out[k];
   } 
   
   if(j==i)
   {
     j=FindFreeReg(out);
     printf("Free gevonden: %d\n",j);
     if(j==NOCOPY) return 1;    /*No free register found: abstraction is not possible*/
     outloop[j]=i;        	/*Add extra copy-instruction to outloop[]*/
     outloop[i]=out[i];		/*Add original copy-intruction to outloop[]*/
     out[i]=NOCOPY;		/*Remove original copy-ins from out[] to break loop*/
     out[k]=j;                  /*Modify copy-instruction in out[]*/ 
   } 
  }
 
 /*End of loop detection}}}*/
 /*Insert Copy Instructions{{{*/
  
  printf("inloop:\t");
  for(i=0;i<32;i++) if(inloop[i]!=NOCOPY) printf("%s:=%s\t",BBL_OBJECT(master)->actions->RegName(i),BBL_OBJECT(master)->actions->RegName(inloop[i]));
  printf("\n");

  printf("outloop:\t");
  for(i=0;i<32;i++) if(outloop[i]!=NOCOPY) printf("%s:=%s\t",BBL_OBJECT(master)->actions->RegName(i),BBL_OBJECT(master)->actions->RegName(outloop[i]));
  printf("\n");
  
  /*in*/
  for(i=0;i<32;i++)
  {
    t_arm_ins * ins;
   while(in[i]<NOCOPY)
   {
    k=i; 
    j=in[i];
    while(in[j]<NOCOPY)
    {
     k=j; 
     j=in[j];
    }  
    ins=T_ARM_INS(InsNewForCfg(master->cfg));
    ArmInstructionMakeMov(T_INS(ins),k,j,0,ARM_CONDITION_AL);
    InsPrependToBbl(T_INS(ins),B);
    addcount++;
    DiabloPrint(stdout,"in:\t@I\n",ins);
    in[k]=NOCOPY;
   }
  }

  /*inloop*/
  for(i=0;i<32;i++)
  {
    t_arm_ins * ins; 
   while(inloop[i]<NOCOPY)
   {
    k=i;
    j=inloop[i];
    while(inloop[j]<NOCOPY)
    {
     k=j; 
     j=inloop[k];
    }  
    ins=T_ARM_INS(InsNewForCfg(master->cfg));
    ArmInstructionMakeMov(T_INS(ins),k,j,0,ARM_CONDITION_AL);
    InsPrependToBbl(T_INS(ins),B);
    addcount++;
    DiabloPrint(stdout,"inloop:\t@I\n",ins);
    inloop[k]=NOCOPY;
   }
  }

  /*out*/
  for(i=0;i<32;i++)
  {
    t_arm_ins *  ins;
   while(out[i]<NOCOPY)
   {
    k=i;
    j=out[i];
    while(out[j]<NOCOPY)
    {
     k=j; 
     j=out[j];
    }  
    ins=T_ARM_INS(InsNewForCfg(master->cfg));
    ArmInstructionMakeMov(T_INS(ins),k,j,0,ARM_CONDITION_AL);
    split?InsInsertBefore(T_INS(ins),split):InsAppendToBbl(T_INS(ins),C);
    addcount++;
    split=T_INS(ins);
    DiabloPrint(stdout,"out:\t@I\n",ins);
    out[k]=NOCOPY;
    
   }
  }
 
  /*outloop*/
  for(i=0;i<32;i++)
  {
    t_arm_ins * ins; 
   while(outloop[i]<NOCOPY)
   {
    k=i;
    j=outloop[i];
    while(outloop[j]<NOCOPY)
    {
     k=j; 
     j=outloop[k];
    }  
    ins=T_ARM_INS(InsNewForCfg(master->cfg));
    ArmInstructionMakeMov(T_INS(ins),k,j,0,ARM_CONDITION_AL);
    split?InsInsertBefore(T_INS(ins),split):InsAppendToBbl(T_INS(ins),C);
    addcount++;
    split=T_INS(ins);
    DiabloPrint(stdout,"outloop:\t@I\n",ins);
    outloop[k]=NOCOPY;
   }
  }
 
 /*End of Inserts}}}*/
   } /*end of if(in!=NULL)*/		
   
   /*add pop and push r14*/
   if(linklive)
   {
     t_arm_ins * pop;
     t_arm_ins * push=T_ARM_INS(InsNewForCfg(BBL_CFG(B)));
     InsInsertBefore(T_INS(push),BBL_INS_LAST(B));
     addcount++;
     ArmInstructionMakeStm(T_INS(push),ARM_REG_R13,0x00004000,ARM_CONDITION_AL,TRUE,TRUE,FALSE);
     
     pop=T_ARM_INS(InsNewForCfg(BBL_CFG(C)));
     InsPrependToBbl(T_INS(pop),C);
     addcount++;
     ArmInstructionMakeLdm(T_INS(pop),ARM_REG_R13,0x00004000,ARM_CONDITION_AL,TRUE,FALSE,TRUE);
   }   

   /*replace edges*/
   BBL_FOREACH_SUCC_EDGE_SAFE(B,i_edge,i_edgetmp)
   { 
     if(EDGE_CAT(i_edge)==ET_CALL)
     {
      CfgEdgeNewCall(BBL_CFG(C),C,CFG_EDGE_TAIL(i_edge),CFG_EDGE_TAIL(EDGE_CORR(i_edge)),CFG_EDGE_HEAD(EDGE_CORR(i_edge)));
      CfgEdgeKill(EDGE_CORR(i_edge));
     }
     else CfgEdgeNew(BBL_CFG(C),C,CFG_EDGE_TAIL(i_edge),EDGE_CAT(i_edge));
     CfgEdgeKill(i_edge);
   }

   /*add call and return edges*/
   CfgEdgeNewCall(BBL_CFG(B),B,master,C,FUN_BBL_LAST(BBL_FUN(master)));   
   CfgEdgeNew(BBL_CFG(B),B,C,ET_FUNLINK);   
   
   /*add C to function*/ 
   BblInsertInFunction(C,BBL_FUN(B));
 } 
 return 0;
}
/*}}}*/

/*ArmBblAbstractMaster: Inserts Branch and Return instructions in master block and splits it.{{{*/
t_uint32 ArmBblAbstractMaster(t_bbl * current,t_bbl * new)
{
 /*declarations*/ 
 t_uint32 type,linklive;
 extern t_uint32 addcount;
 t_ins * split=BBL_INS_LAST(current),* i_ins;
 t_cfg_edge * i_edge,* i_edgetmp;
 t_bbl * A=BblNew(current->cfg);
 t_bbl * B=current;
 t_bbl * C=BblNew(current->cfg);

 /*determine type of bbl*/
 if(ARMINSOPCODE(split)==ARM_B||ARMINSOPCODE(split)==ARM_BX) type=1;
 else if(ARMINSOPCODE(split)==ARM_BL) type=2;
 else if(ARMINSREGA(split)==ARM_REG_R15 && ARMINSREGC(INS_PREV(split))==ARM_REG_R15)
 {
   type=3;
   split=INS_PREV(split);
 }
 else if(ARMINSREGA(split)==ARM_REG_R15) type=4;
 else if(ARMINSOPCODE(split)==ARM_LDM && (ARMINSIMMEDIATE(split)&0x00008000)) type=5;
 else
 {
   type=6;
   split=NULL;
 }

#if MAX_NR_REG_SUBSETS > 4
 FATAL(("Implement me!"));
#else
 /*look if r14(which is used for the function call)is used or defined in the bbl */
 if((current->regs_defined & 0x00004000)|(current->regs_used & 0x00004000))
   return 2; 

 /*look if r14 is live before the bbl -> if true, it has to be stored*/
 if(InsRegsLiveBefore(BBL_INS_FIRST(current))&0x00004000) linklive=1;
 else linklive=0;
#endif
   

 if(type==1)
 {
   t_arm_ins * bl;
   t_arm_ins * mov;
   /*split bbl*/
   
   BBL_INS_LAST(C)=BBL_INS_LAST(B);
   BBL_INS_FIRST(C)=split;
   INS_NEXT(INS_PREV(split))=NULL;
   BBL_INS_LAST(B)=INS_PREV(split);
   INS_PREV(split)=NULL;
   
   BBL_FOREACH_INS(C,i_ins)
   {
     INS_BBL(i_ins)=C;
     BBL_NINS(C)++;
     BBL_NINS(B)--;
   } 
   
   /*add call and return instructions*/
   bl=T_ARM_INS(InsNewForCfg(BBL_CFG(A)));
   InsAppendToBbl(T_INS(bl),A);
   addcount++;
   ArmInstructionMakeCondBranchAndLink(T_INS(bl),ARM_CONDITION_AL);

   mov=T_ARM_INS(InsNewForCfg(BBL_CFG(B)));
   InsAppendToBbl(T_INS(mov),B);
   addcount++;
   ArmInstructionMakeMov(T_INS(mov),ARM_REG_R15,ARM_REG_R14,0,ARM_CONDITION_AL);

   /*add pop and push r14*/
   if(linklive)
   {
     t_arm_ins * pop;
     t_arm_ins * push=T_ARM_INS(InsNewForCfg(BBL_CFG(A)));
     InsPrependToBbl(T_INS(push),A);
     addcount++;
     ArmInstructionMakeStm(T_INS(push),ARM_REG_R13,0x00004000,ARM_CONDITION_AL,TRUE,TRUE,FALSE);
     
     pop=T_ARM_INS(InsNewForCfg(BBL_CFG(A)));
     InsPrependToBbl(T_INS(pop),C);
     addcount++;
     ArmInstructionMakeLdm(T_INS(pop),ARM_REG_R13,0x00004000,ARM_CONDITION_AL,TRUE,FALSE,TRUE);
   }   

   /*add A and C to function, make a new function for B*/ 
   BblInsertInFunction(A,BBL_FUN(B));
   BblInsertInFunction(C,BBL_FUN(B));

   if(B->prev_in_fun) B->prev_in_fun->next_in_fun=B->next_in_fun;
   if(B->next_in_fun) B->next_in_fun->prev_in_fun=B->prev_in_fun;
   
   if(BBL_FUN(B)->bbl_first==B) BBL_FUN(B)->bbl_first=B->next_in_fun;

   B->next_in_fun=NULL;
   B->prev_in_fun=NULL;

   FunctionNew(B,"noname",FT_NORMAL);

   /*replace edges*/
   BBL_FOREACH_PRED_EDGE_SAFE(B,i_edge,i_edgetmp)
   {  
     if(EDGE_CAT(i_edge)==ET_CALL)
     {
      CfgEdgeNewCall(BBL_CFG(A),CFG_EDGE_HEAD(i_edge),A,CFG_EDGE_TAIL(EDGE_CORR(i_edge)),CFG_EDGE_HEAD(EDGE_CORR(i_edge)));
      CfgEdgeKill(EDGE_CORR(i_edge));
     }
     
     else if(EDGE_CAT(i_edge)==ET_RETURN)
     {
      CfgEdgeNewCall(BBL_CFG(A),CFG_EDGE_HEAD(EDGE_CORR(i_edge)),CFG_EDGE_TAIL(EDGE_CORR(i_edge)),A,CFG_EDGE_HEAD(i_edge));
      CfgEdgeKill(EDGE_CORR(i_edge));
     }
      
     else CfgEdgeNew(BBL_CFG(C),CFG_EDGE_HEAD(i_edge),A,EDGE_CAT(i_edge));
     CfgEdgeKill(i_edge);
   }

   BBL_FOREACH_SUCC_EDGE_SAFE(B,i_edge,i_edgetmp)
   { 
     if(EDGE_CAT(i_edge)==ET_CALL)
     {
      CfgEdgeNewCall(BBL_CFG(C),C,CFG_EDGE_TAIL(i_edge),CFG_EDGE_TAIL(EDGE_CORR(i_edge)),CFG_EDGE_HEAD(EDGE_CORR(i_edge)));
      CfgEdgeKill(EDGE_CORR(i_edge));
     }
     else CfgEdgeNew(BBL_CFG(C),C,CFG_EDGE_TAIL(i_edge),EDGE_CAT(i_edge));
     CfgEdgeKill(i_edge);
   }

       
   /*add call and return edges*/
   CfgEdgeNew(BBL_CFG(B),B,FUN_BBL_LAST(BBL_FUN(B)),ET_JUMP);
   CfgEdgeNew(BBL_CFG(A),A,C,ET_FUNLINK); 
   printf("%s\n",BBL_FUN(C)->name);
   CfgEdgeNewCall(BBL_CFG(A),A,B,C,FUN_BBL_LAST(BBL_FUN(B)));   
   
 
   return 1;
 } 
 return type;
}
/*}}}*/

/*ArmBblAbstract: Wrapper to choose to abstract a master or a slave{{{*/
t_uint32 ArmBblAbstract(t_bbl * master,t_bbl * slave,t_uint32 * in,t_uint32 * out)
{
  if(slave==NULL) return  ArmBblAbstractMaster(master,slave);
  else return  ArmBblAbstractSlave(master,slave,in,out);
}/*}}}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
