/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_instructions.h"
#include "arm_registers.h"
#include "arm_utils.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../diablo_options.h"
#include "../../kernel/diablo_regset.h"
#include <string.h>
/* }}} */


t_arm_addr_info * ArmAddrInfoNew()
{
  return Malloc(sizeof(t_arm_addr_info));
}


/*!
 * \todo Document
 *
 * \param ins
 *
 * \return t_bool 
 */
/* ArmInsHasSideEffect {{{ */ 
t_bool ArmInsHasSideEffect(t_ins * ins)
{
  if ((INS_TYPE(ins) == IT_STORE) || RegsetIn(INS_REGS_DEF(ins),ARM_REG_R15))
  {
    return TRUE;
  }
  if (INS_TYPE(ins) == IT_FLT_STORE)
  {
    return TRUE;
  }
  if (INS_TYPE(ins) == IT_STORE_MULTIPLE) 
  {
    return TRUE;
  }
  if (INS_TYPE(ins) == IT_SWI) 
  {
    return TRUE;
  }

  return FALSE;
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 *
 * \return void 
 */
/* ArmInstructionMakeNoop {{{ */
void ArmInstructionMakeNoop(t_ins * ins)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  /* Make a noop */

  ARMINSCONDITION(ins) = ARM_CONDITION_AL;
  INS_TYPE(ins) = IT_DATAPROC;
  ARMINSOPCODE(ins) = ARM_MOV;
  ARMINSREGA(ins) = ARM_REG_R0;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGC(ins) = ARM_REG_R0;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = 0;
  ARMINSFLAGS(ins) = 0x0;
  INS_REGS_USE(ins) = NullRegs; 
  INS_REGS_DEF(ins) = NullRegs; 
  INS_ATTRIB(ins) = IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 *
 * \return void 
*/
/* ArmInstructionMakeUncondBranch {{{ */
void ArmInstructionMakeUncondBranch(t_ins * ins)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  /* Make a noop */

  ARMINSCONDITION(ins) = ARM_CONDITION_AL;
  INS_TYPE(ins) = IT_BRANCH;
  ARMINSOPCODE(ins) = ARM_B;
  ARMINSREGA(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) - (G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) + BBL_NINS(INS_BBL(ins)) * 4 - 4) - 8;
  ARMINSFLAGS(ins) |= FL_IMMED;
  RegsetSetEmpty(INS_REGS_USE(ins));
  RegsetSetEmpty(INS_REGS_DEF(ins));
  RegsetAddReg(INS_REGS_DEF(ins),ARM_REG_R15);
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_ATTRIB(ins) |= IF_PCDEF;
  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*!
 * 
 * \todo Document
 *
 * \param ins
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeCondBranchAndLink {{{ */
void ArmInstructionMakeCondBranchAndLink(t_ins * ins, t_uint32 cond)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  /* Make a branch and link */

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_BRANCH;
  ARMINSOPCODE(ins) = ARM_BL;
  ARMINSREGA(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) - (G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) + BBL_NINS(INS_BBL(ins)) * 4 - 4) - 8;
  ARMINSFLAGS(ins) |= FL_IMMED;
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_ATTRIB(ins) |= IF_PCDEF;
  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeCondBranch {{{ */
void ArmInstructionMakeCondBranch(t_ins * ins, t_uint32 cond)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  /* Make a noop */

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_BRANCH;
  ARMINSOPCODE(ins) = ARM_B;
  ARMINSREGA(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) - (G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) + BBL_NINS(INS_BBL(ins)) * 4 - 4) - 8;
  ARMINSFLAGS(ins) |= FL_IMMED;
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_ATTRIB(ins) |= IF_PCDEF;
  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}

void ArmInstructionConvertToBranch(t_ins * ins)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  /* Make a noop */

  INS_TYPE(ins) = IT_BRANCH;
  ARMINSOPCODE(ins) = ARM_B;
  ARMINSREGA(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) - (G_T_UINT32(BBL_CADDRESS(INS_BBL(ins))) + BBL_NINS(INS_BBL(ins)) * 4 - 4) - 8;
  ARMINSFLAGS(ins) |= FL_IMMED;
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_ATTRIB(ins) |= IF_PCDEF;
  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}



/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param regA
 * \param regB
 * \param regC
 * \param immed
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeLdr {{{ */
void ArmInstructionMakeLdr(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond, t_bool pre, t_bool up, t_bool wb)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_LOAD;
  ARMINSOPCODE(ins) = ARM_LDR;
  ARMINSREGA(ins) = regA;
  ARMINSREGB(ins) = regB;
  ARMINSREGC(ins) = regC;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = immed;
  RegsetSetEmpty(INS_REGS_USE(ins));
  RegsetAddReg(INS_REGS_USE(ins),regB);

  ARMINSFLAGS(ins) &= ~FL_DIRUP & ~FL_WRITEBACK & ~FL_PREINDEX;
  if (pre) ARMINSFLAGS(ins) |= FL_PREINDEX;
  if (up) ARMINSFLAGS(ins) |= FL_DIRUP;
  if (wb) ARMINSFLAGS(ins) |= FL_WRITEBACK;
  if (regC==ARM_REG_NONE)
    {
      ARMINSFLAGS(ins) |= FL_IMMED;
    }
  else
    {
      RegsetAddReg(INS_REGS_USE(ins),regC);
    }
  RegsetSetEmpty(INS_REGS_DEF(ins));
  RegsetAddReg(INS_REGS_DEF(ins),regA);
  if(!pre || wb)
    RegsetAddReg(INS_REGS_DEF(ins),regB);
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param regA
 * \param regB
 * \param regC
 * \param immed
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeStr {{{ */
void ArmInstructionMakeStr(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond, t_bool pre, t_bool up, t_bool wb)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_STORE;
  ARMINSOPCODE(ins) = ARM_STR;
  ARMINSREGA(ins) = regA;
  ARMINSREGB(ins) = regB;
  ARMINSREGC(ins) = regC;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = immed;
  RegsetSetEmpty(INS_REGS_USE(ins));
  RegsetAddReg(INS_REGS_USE(ins),regB);
  RegsetAddReg(INS_REGS_USE(ins),regA);

  ARMINSFLAGS(ins) &= ~FL_DIRUP & ~FL_WRITEBACK & ~FL_PREINDEX;
  if (pre) ARMINSFLAGS(ins) |= FL_PREINDEX;
  if (up) ARMINSFLAGS(ins) |= FL_DIRUP;
  if (wb) ARMINSFLAGS(ins) |= FL_WRITEBACK;
  if (regC==ARM_REG_NONE)
    {
      ARMINSFLAGS(ins) |= FL_IMMED;
    }
  else
    {
      RegsetAddReg(INS_REGS_USE(ins),regC);
    }
  RegsetSetEmpty(INS_REGS_DEF(ins));
  if(!pre || wb)
    RegsetAddReg(INS_REGS_DEF(ins),regB);
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/*}}}*/
/*!
 * \todo Document
 *
 * \param ins
 * \param regA
 * \param regB
 * \param regC
 * \param immed
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeAdd {{{ */
void ArmInstructionMakeAdd(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_DATAPROC;
  ARMINSOPCODE(ins) = ARM_ADD;
  ARMINSREGA(ins) = regA;
  ARMINSREGB(ins) = regB;
  ARMINSREGC(ins) = regC;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = immed;
  /*RegsetSetEmpty(INS_REGS_USE(ins));
  RegsetAddReg(INS_REGS_USE(ins),regB);*/
  ARMINSFLAGS(ins) &= ~FL_DIRUP & ~FL_WRITEBACK & ~FL_PREINDEX;
  if (regC==ARM_REG_NONE)
    {
      ARMINSFLAGS(ins) |= FL_IMMED;
    }
  /*else
    {
      RegsetAddReg(INS_REGS_USE(ins),regC);
    }
  RegsetSetEmpty(INS_REGS_DEF(ins));
  RegsetAddReg(INS_REGS_DEF(ins),regA);*/
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param regA
 * \param regB
 * \param regC
 * \param immed
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeSub {{{ */
void ArmInstructionMakeSub(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_DATAPROC;
  ARMINSOPCODE(ins) = ARM_SUB;
  ARMINSREGA(ins) = regA;
  ARMINSREGB(ins) = regB;
  ARMINSREGC(ins) = regC;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = immed;
  /*RegsetSetEmpty(INS_REGS_USE(ins));
  RegsetAddReg(INS_REGS_USE(ins),regB);*/
  ARMINSFLAGS(ins) &= ~FL_DIRUP & ~FL_WRITEBACK & ~FL_PREINDEX;
  if (regC==ARM_REG_NONE)
    {
      ARMINSFLAGS(ins) |= FL_IMMED;
    }
  /*else
    {
      RegsetAddReg(INS_REGS_USE(ins),regC);
    }
  RegsetSetEmpty(INS_REGS_DEF(ins));
  RegsetAddReg(INS_REGS_DEF(ins),regA);*/
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param regC
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeMsr {{{ */
void ArmInstructionMakeMsr(t_ins * ins, t_reg regC, t_uint32 cond, t_bool set)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_STATUS;
  ARMINSOPCODE(ins) = ARM_MSR;
  ARMINSREGC(ins) = regC;
  ARMINSREGA(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) =0;
  ARMINSFLAGS(ins) &= ~FL_DIRUP & ~FL_WRITEBACK & ~FL_PREINDEX &~FL_S;

  if (set) ARMINSFLAGS(ins)|=FL_S;

  INS_ATTRIB(ins) |= IF_CHANGED;

  /* Set used and defined */
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param regA
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeMrs {{{ */
void ArmInstructionMakeMrs(t_ins * ins, t_reg regA, t_uint32 cond)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_STATUS;
  ARMINSOPCODE(ins) = ARM_MRS;
  ARMINSREGA(ins) = regA;
  ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) =0;
  ARMINSFLAGS(ins) &= ~FL_DIRUP & ~FL_WRITEBACK & ~FL_PREINDEX &~FL_S;


  /* Set used and defined */
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*ArmInstructionMakeStm {{{ */
void ArmInstructionMakeStm(t_ins * ins, t_reg regB, t_uint32 regs, t_uint32 cond, t_bool writeback, t_bool preindex, t_bool dirup)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_STORE_MULTIPLE;
  ARMINSOPCODE(ins) = ARM_STM;
  ARMINSREGA(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = regB;
  ARMINSREGC(ins) = ARM_REG_NONE; 
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = regs;
  
  /* Set flags */

  if (writeback) ARMINSFLAGS(ins) |= FL_WRITEBACK;
  else ARMINSFLAGS(ins) &=~FL_WRITEBACK;
  if (preindex) ARMINSFLAGS(ins) |= FL_PREINDEX;
  else ARMINSFLAGS(ins) &=~FL_PREINDEX;
  if (dirup) ARMINSFLAGS(ins) |= FL_DIRUP;
  else ARMINSFLAGS(ins) &=~FL_DIRUP;
     
  
  /* Set used and defined */
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/*ArmInstructionMakeLdm {{{ */
void ArmInstructionMakeLdm(t_ins * ins, t_reg regB, t_uint32 regs, t_uint32 cond, t_bool writeback, t_bool preindex, t_bool dirup)
{
  t_reloc_ref * relref;
  t_section * sec=InsGetSection(ins);
  
  t_reloc_table * reloc_table;
 
  VERIFY_FASSERT(sec,("Section not set for instruction"));
  reloc_table=sec->obj->reloc_table;

  ARMINSCONDITION(ins) = cond;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  INS_TYPE(ins) = IT_STORE_MULTIPLE;
  ARMINSOPCODE(ins) = ARM_LDM;
  ARMINSREGA(ins) = ARM_REG_NONE;
  ARMINSREGB(ins) = regB;
  ARMINSREGC(ins) = ARM_REG_NONE; 
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  ARMINSIMMEDIATE(ins) = regs;
  
  /* Set flags */

  if (writeback) ARMINSFLAGS(ins) |= FL_WRITEBACK;
  else ARMINSFLAGS(ins) &=~FL_WRITEBACK;
  if (preindex) ARMINSFLAGS(ins) |= FL_PREINDEX;
  else ARMINSFLAGS(ins) &=~FL_PREINDEX;
  if (dirup) ARMINSFLAGS(ins) |= FL_DIRUP;
  else ARMINSFLAGS(ins) &=~FL_DIRUP;
     
  
  /* Set used and defined */
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  
  INS_ATTRIB(ins) |= IF_CHANGED;

  /* If this instruction pointed to something else (e.g. it produced an
   * address, or jumped, or ....) then it no longer does that. So we can remove
   * the reloc (and all associated structures). */

  relref = RELOCATABLE_REFERS_TO(ins);
  while (relref)
  {
    RelocTableRemoveReloc(reloc_table, relref->rel);
    relref = RELOCATABLE_REFERS_TO(ins);
  }
}
/* }}} */
/* ArmInstructionMakeSwi{{{ */
void ArmInstructionMakeSwi(t_ins * ins, t_uint32 immediate, t_uint32 cond)
{
  ARMINSOPCODE(ins) = ARM_SWI;
  ARMINSIMMEDIATE(ins) = immediate;
  ARMINSCONDITION(ins) = cond;
  INS_TYPE(ins) = IT_SWI;
  if (cond!=ARM_CONDITION_AL)
    INS_ATTRIB(ins) |= IF_CONDITIONAL;
  ARMINSREGA(ins) = ARMINSREGB(ins) = ARMINSREGC(ins) = ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSFLAGS(ins) = FL_IMMED;
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;

  INS_ATTRIB(ins) |= IF_CHANGED;

  /* Set used and defined */
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
}
/* }}} */
/*!
 * \todo Document
 *
 * \param f
 * \param lnno
 * \param ins
 * \param immed
 * \param rel
 *
 * \return void 
 */
/* ArmMakeAddressProducer {{{ */
void ArmMakeAddressProducer(char * f, t_uint32 lnno, t_ins * ins, t_uint32 immed, t_reloc * rel)
{
  ARMINSOPCODE(ins)=ARM_ADDRESS_PRODUCER;
  /* The produced register stays the same, as do the conditionflags */
  ARMINSREGB(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  ARMINSREGC(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_TYPE(ins) = IT_CONSTS;
  ARMINSIMMEDIATE(ins)=immed;
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_BBL(ins)->bbl_attributes |= BBL_ADDR_PROD;
}
/* }}} */
/* ArmMakePseudoCall {{{ */
void ArmInstructionMakePseudoCall(t_ins * ins, t_function * fun)
{
  ARMINSOPCODE(ins)=ARM_PSEUDO_CALL;
  /* The produced register stays the same, as do the conditionflags */
  ARMINSREGA(ins)=ARM_REG_R0; /* The previous operands are useless now */
  ARMINSREGB(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  ARMINSREGC(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_TYPE(ins) = IT_CALL;
  ARMINSIMMEDIATE(ins)=0;
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  T_ARM_INS(ins)->data=(void*)fun;
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param immed
 *
 * \return void 
 */
/* ArmMakeConstantProducer {{{ */
void  ArmMakeConstantProducer(t_ins * ins, t_uint32 immed)
{
  ARMINSOPCODE(ins)=ARM_CONSTANT_PRODUCER;
  /* The produced register stays the same, as do the conditionflags */
  ARMINSREGB(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  ARMINSREGC(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_TYPE(ins) = IT_CONSTS;
  if (RELOCATABLE_REFERS_TO(ins)) Free(RELOCATABLE_REFERS_TO(ins));
  RELOCATABLE_REFERS_TO(ins)=NULL;
  ARMINSIMMEDIATE(ins)=immed;
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins); 
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
}
/* }}} */
/*!
 * \todo Document
 * 
 *
 * \param ins
 * \param data
 *
 * \return void 
 */
/* ArmMakeFloatProducer {{{ */
void  ArmMakeFloatProducer(t_ins * ins, char * data)
{
  ARMINSOPCODE(ins)=ARM_FLOAT_PRODUCER;
  /* The produced register stays the same, as do the conditionflags */
  ARMINSREGB(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  ARMINSREGC(ins)=ARM_REG_NONE; /* The previous operands are useless now */
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_TYPE(ins) = IT_CONSTS;
  if (RELOCATABLE_REFERS_TO(ins)) Free(RELOCATABLE_REFERS_TO(ins));
  RELOCATABLE_REFERS_TO(ins)=NULL;
  ARMINSDATA(ins)=data;
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins); 
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
}
/* }}} */
/*!
 * \todo Document
 *
 * \param ins
 * \param regA
 * \param regC
 * \param immed
 * \param cond
 *
 * \return void 
 */
/* ArmInstructionMakeMov {{{ */
void ArmInstructionMakeMov(t_ins * ins, t_reg regA, t_reg regC, t_uint32 immed, t_uint32 cond)
{
  ARMINSCONDITION(ins) = cond;
  ARMINSOPCODE(ins) = ARM_MOV;
  ARMINSREGA(ins) = regA;
  ARMINSREGB(ins) = ARM_REG_NONE;
  ARMINSREGC(ins) = regC;
  ARMINSREGS(ins) = ARM_REG_NONE;
  ARMINSIMMEDIATE(ins) = 0;
   if (ARMINSREGC(ins) != ARM_REG_NONE)
    ARMINSFLAGS(ins) &= ~FL_IMMED;
   else 
   {
     ARMINSFLAGS(ins)|= FL_IMMED;
     ARMINSIMMEDIATE(ins) = immed;
   }
  ARMINSFLAGS(ins) &= ~(FL_DIRUP | FL_WRITEBACK | FL_PREINDEX);
  ARMINSSHIFTTYPE(ins) = ARM_SHIFT_TYPE_NONE;
  ARMINSSHIFTLENGTH(ins) = 0;
  INS_TYPE(ins) = IT_DATAPROC;
  INS_ATTRIB(ins) |= IF_CHANGED;
  if (ARMINSREGA(ins) == ARM_REG_R15) INS_ATTRIB(ins) |= IF_PCDEF;
  if(ARMINSREGA(ins) != ARMINSREGC(ins) || (ARMINSFLAGS(ins)| FL_IMMED))
  {
    INS_REGS_DEF(ins) = ArmDefinedRegisters(ins); 
    INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  }
}
/* }}} */
/*!
 * \todo Document 
 *
 * \param i1
 * \param i2
 *
 * \return t_bool 
 */
/* ArmCompareInstructions {{{ */
t_bool ArmCompareInstructions(t_ins * i1, t_ins * i2)
{
   if (ARMINSOPCODE(i1)==ARM_B) return FALSE;
   if (ARMINSOPCODE(i1)==ARMINSOPCODE(i2))
   {
      if (ARMINSREGA(i1)==ARMINSREGA(i2))
      if (ARMINSREGB(i1)==ARMINSREGB(i2))
      if (ARMINSREGC(i1)==ARMINSREGC(i2))
      if (ARMINSREGS(i1)==ARMINSREGS(i2))
      if (ARMINSCONDITION(i1)==ARMINSCONDITION(i2))
      if (ARMINSFLAGS(i1)==ARMINSFLAGS(i2))
      if (ARMINSSHIFTTYPE(i1)==ARMINSSHIFTTYPE(i2))
      if (ARMINSSHIFTLENGTH(i1)==ARMINSSHIFTLENGTH(i2))
      if (ARMINSIMMEDIATE(i1)==ARMINSIMMEDIATE(i2))
	 return TRUE;
   }
   return FALSE;
}

t_bool ArmCompareInstructionsOppositeCond(t_ins * i1, t_ins * i2)
{
   if (ARMINSOPCODE(i1)==ARM_B) return FALSE;
   if (ARMINSOPCODE(i1)==ARMINSOPCODE(i2))
   {
      if (ARMINSREGA(i1)==ARMINSREGA(i2))
      if (ARMINSREGB(i1)==ARMINSREGB(i2))
      if (ARMINSREGC(i1)==ARMINSREGC(i2))
      if (ARMINSREGS(i1)==ARMINSREGS(i2))
      if (ARMINSCONDITION(i1)==ArmInvertCondition(ARMINSCONDITION(i2)))
      if (ARMINSFLAGS(i1)==ARMINSFLAGS(i2))
      if (ARMINSSHIFTTYPE(i1)==ARMINSSHIFTTYPE(i2))
      if (ARMINSSHIFTLENGTH(i1)==ARMINSSHIFTLENGTH(i2))
      if (ARMINSIMMEDIATE(i1)==ARMINSIMMEDIATE(i2))
	 return TRUE;
   }
   return FALSE;
}

/* }}} */

t_tristate ArmIsSyscallExit(t_ins * ins)
{
  if(INS_TYPE(ins) != IT_SWI) return YES;
  if(ARMINSIMMEDIATE(ins) == 0x900001) return YES;
  if(ARMINSIMMEDIATE(ins) == 0x123456)
  {
    /* ARM ADS syscalls, we have to look for a MOV r0,0x18 instruction */
    t_ins * prev = INS_PREV(ins);
    while(prev)
    {
      if(RegsetIn(INS_REGS_DEF(prev),ARM_REG_R0))
      {
        if(ARMINSOPCODE(prev) == ARM_MOV && ARMINSREGA(prev) == ARM_REG_R0 && ARMINSFLAGS(prev) & FL_IMMED)
	{
	  if(ARMINSIMMEDIATE(prev) == 0x18) return YES;
	  else return NO;
	}
	else
	{
	  WARNING(("Unusual way of setting the syscall number by @I",prev));
	  return PERHAPS;
	}
      }
      else
	prev = INS_PREV(prev);
    }
    
  }
  return NO;
}

/* Unconditionalizer {{{ */
t_bool ArmInsUnconditionalize(t_ins * ins)
{
  if (ARMINSCONDITION(ins)!=ARM_CONDITION_AL)
    {
      ARMINSCONDITION(ins)=ARM_CONDITION_AL;
      return TRUE;
    }
  INS_ATTRIB(ins) &= ~IF_CONDITIONAL;
  return FALSE;
}
/* }}} */

t_bool ArmIsControlflow(t_ins * ins)
{
  if(INS_TYPE(ins) == IT_BRANCH) return TRUE;
  if(INS_TYPE(ins) == IT_SWI) return TRUE;
  if(RegsetIn(INS_REGS_DEF(ins),ARM_REG_R15)) return TRUE;
  return FALSE;
}

t_bool ArmIsIndirectCall(t_ins * ins)
{
  if(ARMINSOPCODE(ins) == ARM_BL)
    return FALSE;

  if(ARMINSOPCODE(ins) == ARM_MOV && ARMINSREGA(ins) == ARM_REG_R15)
    return TRUE;

  FATAL(("IMPLEMENT ME @I",ins));
  return TRUE;
}

void ArmInstructionMakeData(t_ins * ins, t_uint32 data)
{
  INS_TYPE(ins)=IT_DATA;
  ARMINSOPCODE(ins) = ARM_DATA;
  ARMINSIMMEDIATE(ins)=data;
}

void ArmInsCleanup(t_ins * ins)
{
  if(ARMINSDATA(ins))
  {
    fflush(stdout);
    Free(ARMINSDATA(ins));
    ARMINSDATA(ins) = NULL;
  }
  if(T_ARM_INS(ins)->info) Free(T_ARM_INS(ins)->info);
  T_ARM_INS(ins)->info = NULL;
}

void ArmInsDupDynamic(t_ins * target, t_ins * source)
{
  if(T_ARM_INS(source)->info)
  {
    T_ARM_INS(target)->info = ArmAddrInfoNew();

    memcpy(T_ARM_INS(target)->info,T_ARM_INS(source)->info,sizeof(t_arm_addr_info));
  }
}

EXEC_CONDITION ArmInsCondition(t_ins * ins)
{
  switch (ARMINSCONDITION(ins))
    {
    case ARM_CONDITION_EQ :
      return EQ;
    case ARM_CONDITION_NE :
      return NE;
    case ARM_CONDITION_GE :
      return GE;
    case ARM_CONDITION_LT : 
      return LT;
    case ARM_CONDITION_GT : 
      return GT;
    case ARM_CONDITION_LE : 
      return LE;
    case ARM_CONDITION_AL:
      return ALWAYS;
    case ARM_CONDITION_NV : 
      return NEVER;
    default:
      return UNKNOWN ;

    }
  return UNKNOWN;
}

t_ins * ARMaddJumpInstruction(t_bbl * bbl)
{
  t_ins * ins = InsNewForBbl(bbl);
  ArmInstructionMakeUncondBranch(ins);
  return ins;
}


t_bool ArmIsCallInstruction(t_ins * ins)
{
  return ARMINSOPCODE(ins) == ARM_BL;
}

void ArmcomputeLiveRegsBeforeSwi(t_regset * live_after_swi, t_ins * ins)
{
  switch (ARMINSIMMEDIATE(ins))
    {
    case 0x900005:
    case 0x900004:
    case 0x900003:
      RegsetSubReg(*live_after_swi,ARM_REG_R3);
      break;
    default:
      break;
    }
  return;
}
/* vim: set shiftwidth=2 foldmethod=marker : */
