/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/

/*!\addtogroup ARM_BACKEND */
/*@{ */
#include "../../config.h"
#include "arm_opcodes.h"

#ifndef ARM_INSTRUCTION_TYPEDEFS
#define ARM_INSTRUCTION_TYPEDEFS
typedef struct _t_arm_ins t_arm_ins;
typedef struct _t_arm_addr_info t_arm_addr_info;
#endif

#ifndef ARM_INSTRUCTION_H
#define ARM_INSTRUCTION_H
/* Includes {{{ */
#include "../../kernel/diablo_ins.h"
#include "arm_opcodes.h"
/* }}} */

struct _t_arm_addr_info
{
  char original_type;
  t_uint32 load_address;
  t_uint32 load_value;
};

/*! Arm instructions. Arm specific code uses this representation, generic code
 * uses the generic representation */

struct _t_arm_ins
{
  /*! The generic instruction */
  t_ins generic;
  /*! (translated) opcode */ 
  t_arm_opcode opcode;
  /*! destination register */ 
  t_reg rega;
  /*! operand register 1 */
  t_reg regb;
  /*! operand register 2 */
  t_reg regc;
  /*! shift register */
  t_reg regs;     
  /*! immediate operand */
  t_int32 immediate;
  /*! multiple regs operand */
  t_regset multiple;
  /*! instructiegebonden vlaggen (set CPSR, ...) */
  t_instruction_flags attributes;
  /*! Condition code */
  t_arm_condition_code condition;
  /*! The type of shift */
  t_arm_shift_type shifttype;
  /*! immediate shiftlengte */
  t_uint8 shiftlength;		
  /*! Holds float data for ARM_FLOAT_PRODUCER */
  char * data;
  t_arm_addr_info * info;
};

/* Often used shorthands {{{ */
/*! \todo Document */
#define ARMINSOPCODE(arm_ins)      (((t_arm_ins *) arm_ins)->opcode)
/*! \todo Document */
#define ARMINSREGA(arm_ins)        (((t_arm_ins *) arm_ins)->rega)
/*! \todo Document */
#define ARMINSREGB(arm_ins)        (((t_arm_ins *) arm_ins)->regb)
/*! \todo Document */
#define ARMINSREGC(arm_ins)        (((t_arm_ins *) arm_ins)->regc)
/*! \todo Document */
#define ARMINSREGS(arm_ins)        (((t_arm_ins *) arm_ins)->regs)
/*! \todo Document */
#define ARMINSFLAGS(arm_ins)       (((t_arm_ins *) arm_ins)->attributes)
/*! \todo Document */
#define ARMINSCONDITION(arm_ins)   (((t_arm_ins *) arm_ins)->condition)
/*! \todo Document */
#define ARMINSIMMEDIATE(arm_ins)   (((t_arm_ins *) arm_ins)->immediate)
/*! \todo Document */
#define ARMINSDATA(arm_ins)        (((t_arm_ins *) arm_ins)->data)
/*! \todo Document */
#define ARMINSMULTIPLE(arm_ins)    (((t_arm_ins *) arm_ins)->multiple)
/*! \todo Document */
#define ARMINSSHIFTTYPE(arm_ins)   (((t_arm_ins *) arm_ins)->shifttype)
/*! \todo Document */
#define ARMINSSHIFTLENGTH(arm_ins) (((t_arm_ins *) arm_ins)->shiftlength)
/* }}} */

/*! \todo Document */
#define T_ARM_INS(arm_ins)            ((t_arm_ins *) arm_ins)

/* Instruction flags {{{ */
#define FL_S	        	0x1
#define FL_PREINDEX		0x2
#define FL_DIRUP		0x4
#define FL_WRITEBACK		0x8
#define FL_SPSR			0x10
#define FL_IMMED		0x20	/* instructie heeft immediate tweede operand -> voor gevallen */
				        /* waarin twijfel kan bestaan */
#define FL_CONTROL		0x40
#define FL_FLT_SINGLE       	0x80
#define FL_FLT_DOUBLE       	0x100
#define FL_FLT_DOUBLE_EXTENDED 	0x200
#define FL_FLT_PACKED 	        0x400
#define FL_FLT_IMMED		0x800	/* instructie heeft float immediate ipv REGC */
#define FL_FLT_RPI		0x1000  /* float: rounding mode: towards plus infinity */
#define FL_FLT_RMI		0x2000  /* float: rounding mode: towards minus infinity */
#define FL_FLT_RZ		0x4000  /* float: rounding mode: towards zero */

#define FL_NO_LOAD		0x10000 /* address producer MAY NOT load its value */
#define FL_THUMB		0x20000 /* thumb code */
#define FL_ORIG_CONST_PROD	0x40000
/* }}} */

void ArmInstructionMakeNoop(t_ins * ins);
void ArmInstructionMakeData(t_ins * ins, t_uint32 data);
void ArmInstructionMakeUncondBranch(t_ins * ins);
void ArmInstructionMakeCondBranch(t_ins * ins, t_uint32 cond);
void ArmInstructionMakeCondBranchAndLink(t_ins * ins, t_uint32 cond);
void ArmInstructionMakeAdd(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond);
void ArmInstructionMakeSub(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond);
void ArmInstructionMakeMov(t_ins * ins, t_reg regA, t_reg regC, t_uint32 immed, t_uint32 cond);
void ArmInstructionMakeLdm(t_ins * ins, t_reg regB, t_uint32 regs, t_uint32 cond, t_bool writeback, t_bool preindex, t_bool dirup);
void ArmInstructionMakeStm(t_ins * ins, t_reg regB, t_uint32 regs, t_uint32 cond, t_bool writeback, t_bool preindex, t_bool dirup);
void ArmInstructionMakeMrs(t_ins * ins, t_reg regA, t_uint32 cond);
void ArmMakeAddressProducer(char * f, t_uint32 lnno, t_ins * ins, t_uint32 immed, t_reloc * rel);
void ArmInstructionMakeLdr(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond, t_bool pre, t_bool up, t_bool wb);
void ArmInstructionMakeStr(t_ins * ins, t_reg regA, t_reg regB, t_reg regC, t_uint32 immed, t_uint32 cond, t_bool pre, t_bool up, t_bool wb);
void ArmInstructionMakeMsr(t_ins * ins, t_reg regC, t_uint32 cond, t_bool set);
void ArmMakeConstantProducer(t_ins * ins, t_uint32 immed);
void ArmMakeFloatProducer(t_ins * ins, char * data);
void ArmInstructionMakePseudoCall(t_ins * ins, t_function * fun);
void ArmInstructionMakeSwi(t_ins * ins, t_uint32 immediate, t_uint32 cond);
t_bool ArmCompareInstructions(t_ins *,t_ins *);
t_bool ArmCompareInstructionsOppositeCond(t_ins * i1, t_ins * i2);
t_bool ArmInsHasSideEffect(t_ins * ins);
t_tristate ArmIsSyscallExit(t_ins * ins);
t_bool ArmIsControlflow(t_ins * ins);
t_bool ArmInsUnconditionalize(t_ins * ins);
t_arm_addr_info * ArmAddrInfoNew();
t_bool ArmIsIndirectCall(t_ins * ins);
void ArmInsCleanup(t_ins * ins);
void ArmInsDupDynamic(t_ins * target, t_ins * source);
EXEC_CONDITION ArmInsCondition(t_ins * ins);
void ArmInstructionConvertToBranch(t_ins * ins);
t_ins * ARMaddJumpInstruction(t_bbl * bbl);
t_bool ArmIsCallInstruction(t_ins * ins);
void ArmcomputeLiveRegsBeforeSwi(t_regset *, t_ins * );
#endif
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
