/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../../kernel/diablo_edge.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_profile.h"
#include "arm_opcodes.h"
#include "arm_registers.h"
#include "arm_utils.h"
#include "arm_assemble.h"
#include "arm_disassemble_one.h"
#include "arm_instructions.h"
#include "arm_layout.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_section.h"
#include "../../kernel/diablo_stack_machine.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_layout.h"
#include <string.h>
/* }}} */
#define SCHEDULE

/*#ifdef INSTRUMENTATION*/
/*#  define LOAD_EVERYTHING*/
/*#endif*/

/*#define DEBUG_LAYOUT*/

#define LOAD_OFFSET 0x1000
#ifdef SCHEDULE
#include "../../kernel/diablo_schedule.h"
#endif

#define INS_DATA_POOL(x)	((t_data_pool*)INS_TEMP(x))
#define INS_DATA_POOL_BBL(x)	(((t_data_pool*)INS_TEMP(x))->bbl)

t_bool InsAddToPool(t_ins * ins, t_bbl * bbl, t_int32 slack);
t_ins* InsFindIndexInPool(t_ins * ins);
void InsRemoveFromPool(char * file, int line, t_ins * ins);
void FreePoolData();
void ScheduleCfg(t_cfg * cfg);
void ArmReorderChains(t_cfg * cfg,t_bbl ** chains, int nchains);

/*! \todo Remove, use utils_bits.h */
#define rotate_left(v, n) (((v) << (n)) | ((t_uint32)(v))>>(((t_uint32)(32-(n)))))
/*! Document */
typedef struct 
{
  t_bbl * bbl;
  int links_to_next_chain;
  int links_to_prev_chain; 
  int links_to_all_prev_chains;
  int links_to_all_next_chains;
} layouted_chain;

#define IS_LARGE_DATABBL(x)   (IS_DATABBL(x) && BBL_NINS(x)>200)

/*!
 * \todo Document
 *
 * \param ins
 * \param nr_ins
 *
 * \return t_bool 
*/
/* FindValueNearby {{{ */
t_bool FindValueNearby(t_ins * ins,int nr_ins)
{
  t_bbl * i_bbl;
  t_ins * i_ins;
  int address_ins;
  int address_data;
  int range = 0x1000+0xff*0x1000*(nr_ins-1);
  t_int32 val = ARMINSIMMEDIATE(ins);
  t_bool restarted = FALSE;
  t_ins * extra_ins=NULL;
  
 restart:
  
  address_ins = G_T_UINT32(INS_CADDRESS(ins));
  address_data = address_ins;
  
  i_ins = ins;

  while (address_ins-address_data<range-8)
    {
      if (ARMINSOPCODE(i_ins)==ARM_DATA)
	if (ARMINSIMMEDIATE(i_ins)==val)
	  if (!RELOCATABLE_REFERS_TO(i_ins))
	    {
	      return TRUE;
	    }
      if (INS_PREV(i_ins))
	i_ins = INS_PREV(i_ins);
      else
	{
	  i_bbl = INS_BBL(i_ins);
	  if (!BBL_PREV_IN_CHAIN(i_bbl))
	    break;
	  while (BBL_PREV_IN_CHAIN(i_bbl))
	    {
	      if (BBL_NINS(BBL_PREV_IN_CHAIN(i_bbl))!=0)
		{
		  i_ins = BBL_INS_LAST(BBL_PREV_IN_CHAIN(i_bbl));
		  break;
		}
	      i_bbl = BBL_PREV_IN_CHAIN(i_bbl);
	    }
	  if (BBL_NINS(i_bbl) == 0) break;
	}
      address_data=G_T_UINT32(INS_CADDRESS(i_ins));
    }

  address_data = address_ins;
  
  i_ins = ins;

  while (address_data-address_ins<range+8)
    {
      if (ARMINSOPCODE(i_ins)==ARM_DATA)
	if (ARMINSIMMEDIATE(i_ins)==val)
	  if (!RELOCATABLE_REFERS_TO(i_ins))
	    {
	      return TRUE;
	    }
      if (INS_NEXT(i_ins))
	{
	  i_ins = INS_NEXT(i_ins);
	}
      else
	{
	  i_bbl = INS_BBL(i_ins);
	  if (!BBL_NEXT_IN_CHAIN(i_bbl))
	    break;
	  if (IsHellNode(BBL_NEXT_IN_CHAIN(i_bbl)))
	    break;
	  while (BBL_NEXT_IN_CHAIN(i_bbl))
	    {
	      if (BBL_NINS(BBL_NEXT_IN_CHAIN(i_bbl))!=0)
		{
		  i_ins = BBL_INS_FIRST(BBL_NEXT_IN_CHAIN(i_bbl));
		  break;
		}
	      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
	    }
	  if (!i_bbl || IsHellNode(i_bbl) || (BBL_NINS(i_bbl) == 0)) break;
	}
      address_data=G_T_UINT32(INS_CADDRESS(i_ins));
    }

  if (restarted)
  {
    /* if we reach this point, adding the extra data instruction was of no use.
     * we should remove it again and report that we've failed. */
    InsKill(extra_ins);
    return FALSE;
  }
  else if (nr_ins==2)
    {
      extra_ins = i_ins = InsNewForCfg(INS_BBL(ins)->cfg);
      GenericSet_uint32(INS_CSIZE(i_ins),4);
      GenericSet_uint32(INS_OLD_SIZE(i_ins),0);
      
      ARMINSIMMEDIATE(i_ins)=val;
      ARMINSOPCODE(i_ins)=ARM_DATA;
      ARMINSIMMEDIATE(i_ins)=val;
      INS_TYPE(i_ins)=IT_DATA;
      
      InsAppendToBbl(i_ins,BBL_PREV_IN_CHAIN(BBL_PREV_IN_CHAIN(BBL_LAST_IN_CHAIN(INS_BBL(ins)))));
      if (!restarted)
      {
	restarted = TRUE;
	goto restart;
      }
    }
  
  return FALSE;
}
/*}}}*/
/*!
 * \todo Document
 *
 * \param ins
 * \param nr_ins
 *
 * \return t_bool 
*/
/* LoadValueNearby {{{ */
t_bool LoadValueNearby(t_ins * ins,int nr_ins)
{
  t_bbl * i_bbl;
  t_ins * i_ins;
  int address_ins;
  int address_data;
  int range = 0x1000+0xff*0x1000*(nr_ins-1);
  t_uint32 val = ARMINSIMMEDIATE(ins);

  address_ins = G_T_UINT32(INS_CADDRESS(ins));
  address_data = address_ins;

  i_ins = ins;

  while (address_ins-address_data<range-8)
  {
    if (ARMINSOPCODE(i_ins)==ARM_DATA)
      if (ARMINSIMMEDIATE(i_ins)==val)
	if (!RELOCATABLE_REFERS_TO(i_ins))
	{
	  if (nr_ins==1)
	  {
	    ARMINSOPCODE(ins)=ARM_LDR;
	    ARMINSIMMEDIATE(ins)=address_ins-address_data+8;
	    ARMINSREGB(ins)=15;
	    ARMINSFLAGS(ins) = FL_IMMED|FL_PREINDEX;
	    INS_TYPE(ins)=IT_LOAD;
	    return TRUE;
	  }
	  else
	  {
	    t_ins * extra_ins = INS_NEXT(ins);

	    ARMINSOPCODE(ins) = ARM_SUB;
	    ARMINSREGB(ins) = ARM_REG_R15; 
	    ARMINSREGC(ins) = ARM_REG_NONE;
	    ARMINSIMMEDIATE(ins) = (G_T_UINT32(INS_CADDRESS(ins))-address_data+8-4) & ~0xfff;
	    ARMINSFLAGS(ins) = FL_IMMED;
	    INS_ATTRIB(ins) |= IF_CHANGED;
	    INS_TYPE(ins) = IT_DATAPROC;

	    ARMINSOPCODE(extra_ins)=ARM_LDR;
	    ARMINSIMMEDIATE(extra_ins)=(G_T_UINT32(INS_CADDRESS(extra_ins))-address_data+8-4) & 0xfff;
	    ARMINSREGB(extra_ins)=ARMINSREGA(ins);
	    ARMINSREGA(extra_ins)=ARMINSREGA(ins);
	    ARMINSFLAGS(extra_ins) = FL_IMMED|FL_PREINDEX;
	    INS_TYPE(extra_ins)=IT_LOAD;
	    return TRUE;
	  }
	}


    if (INS_PREV(i_ins))
    {
      i_ins = INS_PREV(i_ins);
    }
    else
    {
      i_bbl = INS_BBL(i_ins);
      if (!BBL_PREV_IN_CHAIN(i_bbl))
	break;

      while (BBL_PREV_IN_CHAIN(i_bbl))
      {
	if (BBL_NINS(BBL_PREV_IN_CHAIN(i_bbl))!=0)
	{
	  i_ins = BBL_INS_LAST(BBL_PREV_IN_CHAIN(i_bbl));
	  break;
	}
	i_bbl = BBL_PREV_IN_CHAIN(i_bbl);
      }
      if (BBL_NINS(i_bbl) == 0) break;
    }
    address_data=G_T_UINT32(INS_CADDRESS(i_ins));
  }

  address_data = address_ins;

  i_ins = ins;

  while (address_data-address_ins<range+8)
  {
    if (ARMINSOPCODE(i_ins)==ARM_DATA)
      if (ARMINSIMMEDIATE(i_ins)==val)
	if (!RELOCATABLE_REFERS_TO(i_ins))
	{
	  if (nr_ins==1)
	  {
	    ARMINSOPCODE(ins)=ARM_LDR;
	    ARMINSIMMEDIATE(ins)=address_data-address_ins-8;
	    ARMINSREGB(ins)=15;
	    ARMINSFLAGS(ins) = FL_IMMED|FL_DIRUP|FL_PREINDEX;
	    INS_TYPE(ins)=IT_LOAD;
	    return TRUE;
	  }
	  else
	  {
	    t_ins * extra_ins = INS_NEXT(ins);

	    ARMINSOPCODE(ins) = ARM_ADD;
	    ARMINSREGB(ins) = ARM_REG_R15; 
	    ARMINSREGC(ins) = ARM_REG_NONE;
	    ARMINSIMMEDIATE(ins) = (address_data-G_T_UINT32(INS_CADDRESS(ins))-8) & ~0xfff;
	    ARMINSFLAGS(ins) = FL_IMMED;
	    INS_ATTRIB(ins) |= IF_CHANGED;
	    INS_TYPE(ins) = IT_DATAPROC;

	    ARMINSOPCODE(extra_ins)=ARM_LDR;
	    ARMINSIMMEDIATE(extra_ins)=(address_data-G_T_UINT32(INS_CADDRESS(ins))-8) & 0xfff;
	    ARMINSREGB(extra_ins)=ARMINSREGA(ins);
	    ARMINSREGA(extra_ins)=ARMINSREGA(ins);
	    ARMINSFLAGS(extra_ins) = FL_IMMED|FL_PREINDEX|FL_DIRUP;
	    INS_TYPE(extra_ins)=IT_LOAD;
	    return TRUE;
	  }
	}
    if (INS_NEXT(i_ins))
      i_ins = INS_NEXT(i_ins);
    else
    {
      i_bbl = INS_BBL(i_ins);
      if (!BBL_NEXT_IN_CHAIN(i_bbl))
	break;
      if (IsHellNode(BBL_NEXT_IN_CHAIN(i_bbl)))
	break;
      while (BBL_NEXT_IN_CHAIN(i_bbl))
      {
	if (BBL_NINS(BBL_NEXT_IN_CHAIN(i_bbl))!=0)
	{
	  i_ins = BBL_INS_FIRST(BBL_NEXT_IN_CHAIN(i_bbl));
	  break;
	} 
	i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
      }
      if (!i_bbl || IsHellNode(i_bbl) || (BBL_NINS(i_bbl) == 0)) break;
    }
    address_data=G_T_UINT32(INS_CADDRESS(i_ins));
  }
  return FALSE;
}
/*}}}*/
/*! Returns the number of address and constant producers that load their value
 * from a constant pool that lays beyond this instruction, but that are layouted 
 * them selves before this instruction */
t_uint32 NumberOfPrevInstructionsWithForwardDatapool(t_ins * ins)
{
  t_bbl* bbl;
  t_ins* i_ins;
  t_uint32 ret = 0;

  if(!ins) return 0;
  bbl = INS_BBL(ins);
  i_ins = INS_PREV(ins);
  while(1)
  {
    if(!i_ins)
    {
      if(!BBL_PREV_IN_CHAIN(bbl)) break;
      else
      {
	do
	{
	  /* skip empty blocks (eg. return blocks) */
	  bbl = BBL_PREV_IN_CHAIN(bbl);
	  if(!bbl) return 0;
	}
	while(!BBL_INS_LAST(bbl));
	i_ins = BBL_INS_LAST(bbl);
      }
    }
    /* If distance gets too big, we cannot address the datapool anymore */
    if(GENERIC_EXTRACT_UINT32(GENERIC_SUB(GENERIC_ADD_UINT32(INS_CADDRESS(ins),8),INS_CADDRESS(i_ins))) > LOAD_OFFSET) break;

    if(INS_DATA_POOL(i_ins))
    {
      if(GENERIC_LE(BBL_CADDRESS(INS_DATA_POOL_BBL(i_ins)),BBL_CADDRESS(INS_BBL(ins))))
	break;
      ret++;
    }
    i_ins = INS_PREV(i_ins);
  }
#ifdef DEBUG_LAYOUT
  VERBOSE(0,("Forward @I returns %d",ins,ret));
#endif

  return ret;
}
/*! Returns the number of address and constant producers that load their value
 * from a constant pool that lays before this instruction, but that are layouted 
 * themselves after this instruction */
t_uint32 NumberOfNextInstructionsWithBackwardDatapool(t_ins * ins)
{
  t_bbl* bbl;
  t_ins* i_ins;
  t_uint32 ret = 0;

  if(!ins) return 0;

  bbl = INS_BBL(ins);
  i_ins = INS_NEXT(ins);

  while(1)
  {
    if(!i_ins)
    {
      if(!BBL_NEXT_IN_CHAIN(bbl)) break;
      else
      {
	do
	{
	  /* skip empty blocks (eg. return blocks) */
	  bbl = BBL_NEXT_IN_CHAIN(bbl);
	  if(!bbl) return 0;
	}
	while(!BBL_INS_FIRST(bbl));
	i_ins = BBL_INS_FIRST(bbl);
      }
    }
    /* If distance gets too big, we cannot address the datapool anymore */
    
    if(((t_int32)GENERIC_EXTRACT_UINT32(GENERIC_SUB(INS_CADDRESS(i_ins),GENERIC_ADD_UINT32(INS_CADDRESS(ins),8)))) > LOAD_OFFSET) break;

    if(INS_DATA_POOL(i_ins))
    {
      if(GENERIC_GT(BBL_CADDRESS(INS_DATA_POOL_BBL(i_ins)),INS_CADDRESS(ins)))
	break;
      ret++;
    }
    i_ins = INS_NEXT(i_ins);
  }
#ifdef DEBUG_LAYOUT
  VERBOSE(0,("Backward @I returns %d",ins,ret));
#endif

  return ret;
}


/*!
 * \todo Document
 *
 * \param ins
 *
 * \return t_uint32 
*/
/* ConstProdGuessNumberOfInstructionsNeeded {{{ */
static t_uint32 ConstProdGuessNumberOfInstructionsNeeded(t_ins * ins, t_uint32 min_addr, t_uint32 max_addr)
{
  t_int32 val = ARMINSIMMEDIATE(ins);
  t_int32 val_neg = -val;
  t_int32 val_invert = ~val;
  t_int32 val_diff, val_abs;
  t_uint32 uval, max_offset = 0;

  t_ins * j_ins;
  t_bbl * bbl = INS_BBL(ins);

  t_uint32 a;
  t_uint32 i;

  if(min_addr)
  {
    uval = ARMINSIMMEDIATE(ins);
    if (uval <= min_addr) 
      max_offset = max_addr - uval;
    else if(uval >= max_addr) 
      max_offset = uval - min_addr;
    else
      max_offset = (uval - min_addr) > (max_addr - uval) ? (uval - min_addr) : (max_addr - uval);
    /*max_offset += LAYOUT_SAFETY_MARGIN;*/
    /*DiabloPrint(stdout,"Max_offset using addresses = %x\n",max_offset);*/

  }
  /* first we try to generate the constant from scratch in one instruction */
  {
    for (i = 0; i < 32; i += 2)
      if ((a = rotate_left (val, i)) <= 0xff)
      {
	return 1;
      }

    for (i = 0; i < 32; i += 2)
      if ((a = rotate_left (val_neg-1, i)) <= 0xff)
      {
	return 1; 
      }

    for (i = 0; i < 32; i += 2)
      if ((a = rotate_left (val_invert, i)) <= 0xff)
      {
	return 1;
      }
  }

  /* Maybe there is another constantproducer preceding this one, that we can use?*/
  BBL_FOREACH_INS(bbl,j_ins)
    if(ARMINSOPCODE(j_ins) == ARM_CONSTANT_PRODUCER && ARMINSCONDITION(j_ins) == ARM_CONDITION_AL && j_ins != ins)
    {
      t_ins * r_ins = INS_NEXT(j_ins);
      while(r_ins != ins)
      {
	if(RegsetIn(INS_REGS_DEF(r_ins),ARMINSREGA(j_ins))) break;
	r_ins = INS_NEXT(r_ins);
      }
      
      if(r_ins == ins)
      {
	val_abs = val_diff = ARMINSIMMEDIATE(j_ins) - val;
	if(val_diff < 0) val_abs = -val_diff; 

	for (i = 0; i < 32; i += 2)
	  if ((a = rotate_left (val_abs, i)) <= 0xff)
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("Using a previous constant producer, @I, to produce @I",j_ins,ins));
#endif
	    /* effectively change the instruction, to remove the dependency from the constprod in
	     * j_ins, which may end up in another block when data pools are created */
	    if(val_diff < 0)
	    {
	      ArmInstructionMakeAdd(ins,ARMINSREGA(ins),ARMINSREGA(j_ins),ARM_REG_NONE,val_abs,ARMINSCONDITION(ins));
	    }
	    else
	    {
	      ArmInstructionMakeSub(ins,ARMINSREGA(ins),ARMINSREGA(j_ins),ARM_REG_NONE,val_abs,ARMINSCONDITION(ins));
	    }
	    return 1;
	  }
      }
    }
    else if (j_ins == ins)
      break;
  

  /* Try to generate the constant from the pc in 1 instruction */
#if 0
  if(min_addr)
  {
    if(max_offset < 256)
    {
      printf("C MIN = %x MAX = %x OFF= %x\n",min_addr,max_addr,max_offset);
      /*DiabloPrint(stdout,"Certainly 1, using addresses! @I\n",ins);*/
      return 1;
    }
  }
#endif

  if (!BblIsHot(INS_BBL(ins))) return 0; /* NO CONST PRODS IN TWO INSTRUCTIONS */
  if(global_options.insprofilefile)
    if ((INS_EXEC_COUNT(ins) <= BBL_EXEC_COUNT(INS_BBL(ins))/2) && INS_ATTRIB(ins) & IF_CONDITIONAL) return 0;

  /* Try to generate the constant from scratch in two instructions */
  {
    for (i = 0; i < 32; i += 2)
      if (((a = rotate_left (val, i)) & 0xff) != 0)
      {
	if (a & 0xff00)
	{
	  if (a & ~ 0xffff)
	    continue;
	}
	else if (a & 0xff0000)
	{
	  if (a & 0xff000000)
	    continue;
	}
	else
	{
	}

	return 2;
      }

    for (i = 0; i < 32; i += 2)
      if (((a = rotate_left (val_neg-1, i)) & 0xff) != 0)
      {
	if (a & 0xff00)
	{
	  if (a & ~ 0xffff)
	    continue;
	}
	else if (a & 0xff0000)
	{
	  if (a & 0xff000000)
	    continue;
	}

	return 2;
      }

    for (i = 0; i < 32; i += 2)
      if (((a = rotate_left (val_invert, i)) & 0xff) != 0)
      {
	if (a & 0xff00)
	{
	  if (a & ~ 0xffff)
	    continue;
	}
	else if (a & 0xff0000)
	{
	  if (a & 0xff000000)
	    continue;
	}
	else
	{
	}

	return 2;
      }
  }

  /* Maybe there is another const prod preceding this one, that we can use?*/
  BBL_FOREACH_INS(bbl,j_ins)
    if(ARMINSOPCODE(j_ins) == ARM_CONSTANT_PRODUCER && ARMINSCONDITION(j_ins) == ARM_CONDITION_AL && j_ins != ins)
    {
      t_ins * r_ins = INS_NEXT(j_ins);
      while(r_ins != ins)
      {
	if(RegsetIn(INS_REGS_DEF(r_ins),ARMINSREGA(j_ins))) break;
	r_ins = INS_NEXT(r_ins);
      }

      if(r_ins == ins)
      {
	val_abs = val_diff = ARMINSIMMEDIATE(j_ins) - val;
	if(val_diff < 0) val_abs = -val_diff; 

	for (i = 0; i < 32; i += 2)
	  if (((a = rotate_left (val_abs, i)) & 0xff) != 0)
	  {
	    if (a & 0xff00)
	    {
	      if (a & ~ 0xffff)
		continue;
	    }
	    else if (a & 0xff0000)
	    {
	      if (a & 0xff000000)
		continue;
	    }
	    else
	    {
	    }

	    return 2;
	  }

      }
    }
    else if(j_ins == ins) break;

  if(min_addr) 
  {
    if(max_offset < 65536)
    {
      /*DiabloPrint(stdout,"Certainly 2, using addresses! @I\n",ins);*/
      return 2;
    }
  }

  return 0;
}
/*}}}*/
/*!
 * \todo Document 
 *
 * \param ins
 *
 * \return void 
*/
/*GenerateInstructionsForAddrProd {{{ */
static t_uint32 unefficient = 0;

static void GenerateInstructionsForAddrProd(t_ins * ins)
{
  t_reloc * reloc = RELOCATABLE_REFERS_TO(ins)->rel; 

  t_uint32 producer_new_address;
  t_uint32 produced_new_address;
  t_int32 val, val_neg,val_invert;
  t_uint32 a;
  t_uint32 i;
  t_bool not_needed = FALSE;
  

  t_int32 range = 0x3ff;
  t_int32 mask  = 0x3fc;

  t_int32 offset;

  producer_new_address = G_T_UINT32(INS_CADDRESS(ins)) + 8;
  produced_new_address = G_T_UINT32(StackExec(reloc->code,reloc,NULL,FALSE,0));
      
  offset = produced_new_address - producer_new_address;
  val = produced_new_address;
  val_neg = -val;
  val_invert = ~val;

  /*VERBOSE(0,("Producing %d\n",produced_new_address));*/
/*{{{ Treat it as a constant, and try to generate it in 1 instruction */
  for (i = 0; i < 32; i += 2)
    if ((a = rotate_left (val, i)) <= 0xff)
      {
	ARMINSOPCODE(ins) = ARM_MOV;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = val;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	if(INS_DATA_POOL(ins))
	{
#ifdef DEBUG_LAYOUT
	  VERBOSE(0,("Estimated 2, doing it in 1 for @I",ins));
#endif
	  not_needed = TRUE;
	}
	return;
      }

  for (i = 0; i < 32; i += 2)
    if ((a = rotate_left (val_neg-1, i)) <= 0xff)
      {
	ARMINSOPCODE(ins) = ARM_MVN;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = val_neg-1;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	if(INS_DATA_POOL(ins))
	{
#ifdef DEBUG_LAYOUT
	  VERBOSE(0,("Estimated 2, doing it in 1 for @I",ins));
#endif
	  not_needed = TRUE;
	}
	return;
      }
  

  for (i = 0; i < 32; i += 2)
    if ((a = rotate_left (val_invert, i)) <= 0xff)
      {
	ARMINSOPCODE(ins) = ARM_MVN;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = val_invert;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	if(INS_DATA_POOL(ins))
	{
#ifdef DEBUG_LAYOUT
	  VERBOSE(0,("Estimated 2, doing it in 1 for @I",ins));
#endif
	  not_needed = TRUE;
	}
	return;
      }
/*}}}*/

  if ((offset & 0xfffffffc)!=offset)
  {
#if 0
    DiabloPrint(stdout,"@I\n",ins);
    printf("Not aligned! %x %x\n",offset & 0xfffffffc,offset);
    printf("PRODUCER = %x\n",producer_new_address);
    printf("PRODUCED= %x\n",produced_new_address);
    DiabloPrint(stdout,"Reloc: @R\n",reloc);
    printf("reloc code %s\n",reloc->code);
#endif
    range = 0xff;
    mask  = 0xff; 
  }

  if (0<=offset && offset<range)
  {
    ArmInstructionMakeAdd(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,offset,ARMINSCONDITION(ins));
    if(INS_DATA_POOL(ins))
    {
#ifdef DEBUG_LAYOUT
      VERBOSE(0,("Estimated 2, doing it in 1 from the PC for @I",ins));
#endif
      not_needed = TRUE;
    }
    return;
  }
  if (-range<=offset && offset<=0)
  {
    ArmInstructionMakeSub(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,-offset,ARMINSCONDITION(ins));
    if(INS_DATA_POOL(ins))
    {
#ifdef DEBUG_LAYOUT
      VERBOSE(0,("Estimated 2, doing it in 1 from the PC for @I",ins));
#endif
      not_needed = TRUE;
    }
    return;
  }
  /*{{{ Treat it as a constant and try to generate it in two instructions */
  {
    for (i = 0; i < 32; i += 2)
      if (((a = rotate_left (val, i)) & 0xff) != 0)
      {
	t_ins * extra_ins = INS_NEXT(ins);

	if (a & 0xff00)
	{
	  if (a & ~ 0xffff)
	    continue;
	}
	else if (a & 0xff0000)
	{
	  if (a & 0xff000000)
	    continue;
	}

	if ((!extra_ins) || !ArmInstructionIsNOOP(extra_ins))
	{
	  if(!INS_DATA_POOL(ins))
	    FATAL(("no NOOP 5 for @I, val = %x, range = %x, offset = %x\n",ins,val,range,offset));
	  else
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("Estimated load, but could do it in 2 ins from scratch: @I, val = %x",ins,val));
#endif
	    not_needed = TRUE;
	    break;
	  }
	}
	else
	{
	  ARMINSOPCODE(ins) = ARM_MOV;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = val & rotate_left(0xff,32-i);
	  ARMINSFLAGS(ins) = FL_IMMED;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_TYPE(ins) = IT_DATAPROC;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

	  ARMINSOPCODE(extra_ins) = ARM_ORR;
	  ARMINSREGB(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGC(extra_ins) = ARM_REG_NONE;
	  ARMINSREGA(extra_ins) = ARMINSREGA(ins);
	  ARMINSIMMEDIATE(extra_ins) = val & ~rotate_left(0xff,32-i);
	  ARMINSFLAGS(extra_ins) = FL_IMMED;
	  INS_ATTRIB(extra_ins) |= IF_CHANGED;
	  INS_TYPE(extra_ins) = IT_DATAPROC;
	  ARMINSCONDITION(extra_ins) = ARMINSCONDITION(ins);
	  INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
	  INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);
	  return;
	}
      }

    for (i = 0; i < 32; i += 2)
      if (((a = rotate_left (val_neg-1, i)) & 0xff) != 0)
      {
	t_ins * extra_ins = INS_NEXT(ins);
	if (a & 0xff00)
	{
	  if (a & ~ 0xffff)
	    continue;
	}
	else if (a & 0xff0000)
	{
	  if (a & 0xff000000)
	    continue;
	}

	if ((!extra_ins) || !ArmInstructionIsNOOP(extra_ins))
	{
	  if(!INS_DATA_POOL(ins))
	    FATAL(("no NOOP 6, val_neg-1 = %x\n",val_neg-1));
	  else
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("Estimated load, but could do it in 2 ins from scratch: @I, val_neg-1 = %x",ins,val_neg-1));
#endif
	    not_needed = TRUE;
	    break;
	  }
	}
	else
	{
	  ARMINSOPCODE(ins) = ARM_MVN;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = (val_neg-1) & rotate_left(0xff,32-i);
	  ARMINSFLAGS(ins) = FL_IMMED;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_TYPE(ins) = IT_DATAPROC;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

	  ARMINSREGA(extra_ins) = ARMINSREGA(ins);
	  ARMINSOPCODE(extra_ins) = ARM_BIC;
	  ARMINSREGB(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGC(extra_ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(extra_ins) = (val_neg-1) & ~rotate_left(0xff,32-i);
	  ARMINSFLAGS(extra_ins) = FL_IMMED;
	  INS_ATTRIB(extra_ins) |= IF_CHANGED;
	  INS_TYPE(extra_ins) = IT_DATAPROC;
	  ARMINSCONDITION(extra_ins) = ARMINSCONDITION(ins);
	  INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
	  INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);
	  return;
	}
      }

    for (i = 0; i < 32; i += 2)
      if (((a = rotate_left (val_invert, i)) & 0xff) != 0)
      {
	t_ins * extra_ins = INS_NEXT(ins);
	if (a & 0xff00)
	{
	  if (a & ~ 0xffff)
	    continue;
	}
	else if (a & 0xff0000)
	{
	  if (a & 0xff000000)
	    continue;
	}

	if ((!extra_ins) || !ArmInstructionIsNOOP(extra_ins))
	{
	  if(!INS_DATA_POOL(ins))
	    FATAL(("no NOOP 7, val_invert = %x\n",val_invert));
	  else
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("Estimated load, but could do it in 2 ins from scratch: @I, val_invert = %x",ins, val_invert));
#endif
	    not_needed = TRUE;
	    break;
	  }
	}
	else
	{
	  ARMINSOPCODE(ins) = ARM_MVN;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = val_invert & rotate_left(0xff,32-i);
	  ARMINSFLAGS(ins) = FL_IMMED;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_TYPE(ins) = IT_DATAPROC;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

	  ARMINSOPCODE(extra_ins) = ARM_BIC;
	  ARMINSREGA(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGB(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGC(extra_ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(extra_ins) = val_invert & ~rotate_left(0xff,32-i);
	  ARMINSFLAGS(extra_ins) = FL_IMMED;
	  INS_ATTRIB(extra_ins) |= IF_CHANGED;
	  INS_TYPE(extra_ins) = IT_DATAPROC;
	  ARMINSCONDITION(extra_ins) = ARMINSCONDITION(ins);
	  INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
	  INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);
	  return;
	}
      }

  }/*}}}*/
  if ((0<=offset && offset<= (mask+(mask<<8))))
  {
    t_ins * extra_ins = INS_NEXT(ins);

    if ((!extra_ins) || !ArmInstructionIsNOOP(extra_ins))
    {
      if(!INS_DATA_POOL(ins))
        FATAL(("no NOOP 1 @I, @iB produced: %x, offset: %x\n",ins,INS_BBL(ins),produced_new_address,offset));
      else
      {
#ifdef DEBUG_LAYOUT
	VERBOSE(0,("Loading, but could do it in 2 ins from PC: @I, offset = %x",ins,offset));
#endif
	not_needed = TRUE;
      }
    }
    else
    {
      ArmInstructionMakeAdd(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,offset&mask,ARMINSCONDITION(ins));
      ArmInstructionMakeAdd(extra_ins,ARMINSREGA(ins),ARMINSREGA(ins),ARM_REG_NONE,offset & (mask<<8),ARMINSCONDITION(ins));
      return;
    }
  }
  if ((-(mask+(mask<<8))<=offset && offset<=0))
  {
    t_ins * extra_ins = INS_NEXT(ins);

    if ((!extra_ins) || !ArmInstructionIsNOOP(extra_ins))
    {
      if(!INS_DATA_POOL(ins))
	FATAL(("no NOOP 2 for @I in @iB %x\n",ins,INS_BBL(ins),offset));
      else
      {
#ifdef DEBUG_LAYOUT
	VERBOSE(0,("Loading, but could do it in 2 ins from PC: @I, offset = %x",ins,offset));
#endif
	not_needed = TRUE;
      }
    }
    else
    {
      ArmInstructionMakeSub(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,(-offset)&mask,ARMINSCONDITION(ins));
      ArmInstructionMakeSub(extra_ins,ARMINSREGA(ins),ARMINSREGA(ins),ARM_REG_NONE,(-offset) & (mask<<8),ARMINSCONDITION(ins));
      return;
    }
  }

  if(not_needed) unefficient++;

  if(INS_DATA_POOL(ins))
  {
    t_bool dir_up = TRUE;
    t_ins * data_ins = NULL;
    t_uint32 load_offset = 0;

#ifdef DEBUG_LAYOUT
    VERBOSE(0,("Datapool for @I is @B\n",ins,INS_DATA_POOL_BBL(ins)));
#endif

    data_ins = InsFindIndexInPool(ins);

    if(!data_ins) FATAL(("No NOOP in datapool for @I, data_pool is @B",ins,INS_DATA_POOL_BBL(ins)));
    
    if(GENERIC_GE(INS_CADDRESS(ins),INS_CADDRESS(data_ins)))
    {
      load_offset = G_T_UINT32(GENERIC_SUB(INS_CADDRESS(ins),INS_CADDRESS(data_ins))) + 8;
      dir_up = FALSE;
    }
    else
      load_offset = G_T_UINT32(GENERIC_SUB(INS_CADDRESS(data_ins),INS_CADDRESS(ins))) - 8;
      
    ARMINSOPCODE(ins) = ARM_LDR;
    ARMINSREGB(ins) = 15;
    ARMINSREGC(ins) = ARM_REG_NONE;
    ARMINSIMMEDIATE(ins) = load_offset;
    ARMINSFLAGS(ins) = FL_IMMED | FL_PREINDEX;
    if(dir_up) ARMINSFLAGS(ins) |= FL_DIRUP;
    INS_ATTRIB(ins) |= IF_CHANGED;
    INS_TYPE(ins) = IT_LOAD;
    INS_REGS_USE(ins) = ArmUsedRegisters(ins);
    INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

    ARMINSOPCODE(data_ins) = ARM_DATA;
    INS_TYPE(data_ins) = IT_DATA;
    ARMINSIMMEDIATE(data_ins) = produced_new_address;

    return;
  }

  FATAL(("Don't know how to generate the address for @I\nval = %x, offset = %x",ins, val, offset));

}
/*}}}*/
/*!
 * \todo Document
 *
 * \param ins
 *
 * \return void 
*/
/*GenerateInstructionsForConstProd {{{ */
static void GenerateInstructionsForConstProd(t_ins *ins)
{
  t_int32 val = ARMINSIMMEDIATE(ins);
  t_int32 val_neg = -val;
  t_int32 val_invert = ~val;
  t_int32 val_diff, val_abs;
  t_uint32 caddress, uval, max_offset;
  t_bool not_needed = FALSE;

  t_ins * j_ins;
  t_bbl * bbl = INS_BBL(ins);

  t_uint32 a;
  t_uint32 i;

  /* Try it in one instruction from scratch */
  {
  for (i = 0; i < 32; i += 2)
    if ((a = rotate_left (val, i)) <= 0xff)
      {
	ARMINSOPCODE(ins) = ARM_MOV;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = val;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	if(INS_DATA_POOL(ins))
	{
	  VERBOSE(0,("Really weird, this should be done in 1 ins : @I",ins));
	  not_needed = TRUE;
	}
	return;
      }

  for (i = 0; i < 32; i += 2)
    if ((a = rotate_left (val_neg-1, i)) <= 0xff)
      {
	ARMINSOPCODE(ins) = ARM_MVN;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = val_neg-1;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	if(INS_DATA_POOL(ins))
	{
	  VERBOSE(0,("Really weird, this should be done in 1 ins : @I",ins));
	  not_needed = TRUE;
	}
	return;
      }
  

  for (i = 0; i < 32; i += 2)
    if ((a = rotate_left (val_invert, i)) <= 0xff)
      {
	ARMINSOPCODE(ins) = ARM_MVN;
	ARMINSREGB(ins) = ARM_REG_NONE;
	ARMINSREGC(ins) = ARM_REG_NONE;
	ARMINSIMMEDIATE(ins) = val_invert;
	ARMINSFLAGS(ins) = FL_IMMED;
	INS_ATTRIB(ins) |= IF_CHANGED;
	INS_TYPE(ins) = IT_DATAPROC;
	INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	if(INS_DATA_POOL(ins))
	{
	  VERBOSE(0,("Really weird, this should be done in 1 ins : @I",ins));
	  not_needed = TRUE;
	}
	return;
      }
  }

  /* Maybe there is another constantproducer preceding this one, that we can use?*/
  BBL_FOREACH_INS(bbl,j_ins)
    if(ARMINSOPCODE(j_ins) == ARM_CONSTANT_PRODUCER && ARMINSCONDITION(j_ins) == ARM_CONDITION_AL && j_ins != ins)
    {
      t_ins * r_ins = INS_NEXT(j_ins);
      while(r_ins != ins)
      {
	if(RegsetIn(INS_REGS_DEF(r_ins),ARMINSREGA(j_ins))) break;
	r_ins = INS_NEXT(r_ins);
      }
      
      if(r_ins == ins)
      {
	val_abs = val_diff = ARMINSIMMEDIATE(j_ins) - val;
	if(val_diff < 0) val_abs = -val_diff; 

	for (i = 0; i < 32; i += 2)
	  if ((a = rotate_left (val_abs, i)) <= 0xff)
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("Using a previous constant producer, @I, to produce @I",j_ins,ins));
#endif
	    if(val_diff < 0)
	    {
	      ArmInstructionMakeAdd(ins,ARMINSREGA(ins),ARMINSREGA(j_ins),ARM_REG_NONE,val_abs,ARMINSCONDITION(ins));
	    }
	    else
	    {
	      ArmInstructionMakeSub(ins,ARMINSREGA(ins),ARMINSREGA(j_ins),ARM_REG_NONE,val_abs,ARMINSCONDITION(ins));
	    }
	    if(INS_DATA_POOL(ins))
	    {
#ifdef DEBUG_LAYOUT
	      VERBOSE(0,("Estimated load, but doing it in one ins using previous constprod: @I",ins));
#endif
	      not_needed = TRUE;
	    }
	    return;
	  }
      }
    }

  

  caddress = GENERIC_EXTRACT_UINT32(INS_CADDRESS(ins)) + 8;
  uval = ARMINSIMMEDIATE(ins);
  if(uval <= caddress) max_offset = caddress - uval;
  else max_offset = uval - caddress;

  /* Try to generate it from the pc */
  
#if 0
  if(max_offset <256)
  {
    if(uval < caddress)
    {
      ArmInstructionMakeSub(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,max_offset,ARMINSCONDITION(ins));
    }
    else
    {
      ArmInstructionMakeAdd(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,max_offset,ARMINSCONDITION(ins));
    }
    /*DiabloPrint(stdout,"Constprod from pc in 1 ins\n");*/
    if(INS_DATA_POOL(ins))
    {
#ifdef DEBUG_LAYOUT
      VERBOSE(0,("Estimated load, but doing it in one ins from PC: @I",ins));
#endif
      not_needed = TRUE;
    }
    return;
  }
#endif


  {
  for (i = 0; i < 32; i += 2)
    if (((a = rotate_left (val, i)) & 0xff) != 0)
      {
	t_ins * extra_ins = INS_NEXT(ins);

	if (a & 0xff00)
	  {
	    if (a & ~ 0xffff)
	      continue;
	  }
	else if (a & 0xff0000)
	  {
	    if (a & 0xff000000)
	      continue;
	  }

	if (!extra_ins || !ArmInstructionIsNOOP(extra_ins))
	{
	  if(!INS_DATA_POOL(ins))
	    FATAL(("no NOOP 5 for @I\n",ins));
	  else
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("1. Loading but could have done it in 2 ins from scratch: @I",ins));
#endif
	    not_needed = TRUE;
	  }
	}
	else
	{
	  ARMINSOPCODE(ins) = ARM_MOV;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = val & rotate_left(0xff,32-i);
	  ARMINSFLAGS(ins) = FL_IMMED;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_TYPE(ins) = IT_DATAPROC;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

	  ARMINSOPCODE(extra_ins) = ARM_ORR;
	  ARMINSREGB(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGC(extra_ins) = ARM_REG_NONE;
	  ARMINSREGA(extra_ins) = ARMINSREGA(ins);
	  ARMINSIMMEDIATE(extra_ins) = val & ~rotate_left(0xff,32-i);
	  ARMINSFLAGS(extra_ins) = FL_IMMED;
	  INS_ATTRIB(extra_ins) |= IF_CHANGED;
	  INS_TYPE(extra_ins) = IT_DATAPROC;
	  ARMINSCONDITION(extra_ins) = ARMINSCONDITION(ins);
	  INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
	  INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);
	  return;
	}
      }
  
  for (i = 0; i < 32; i += 2)
    if (((a = rotate_left (val_neg-1, i)) & 0xff) != 0)
      {
	t_ins * extra_ins = INS_NEXT(ins);
	if (a & 0xff00)
	  {
	    if (a & ~ 0xffff)
	      continue;
	  }
	else if (a & 0xff0000)
	  {
	    if (a & 0xff000000)
	      continue;
	  }
	
	if (!extra_ins || !ArmInstructionIsNOOP(extra_ins))
	{
	  if(!INS_DATA_POOL(ins))
	    FATAL(("no NOOP 6 for @I\n",ins));
	  else
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("4. Loading but could have done it in 2 ins from scratch: @I",ins));
#endif
	    not_needed = TRUE;
	  }
	}
	else
	{
	  ARMINSOPCODE(ins) = ARM_MVN;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = (val_neg-1) & rotate_left(0xff,32-i);
	  ARMINSFLAGS(ins) = FL_IMMED;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_TYPE(ins) = IT_DATAPROC;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

	  ARMINSREGA(extra_ins) = ARMINSREGA(ins);
	  ARMINSOPCODE(extra_ins) = ARM_BIC;
	  ARMINSREGB(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGC(extra_ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(extra_ins) = (val_neg-1) & ~rotate_left(0xff,32-i);
	  ARMINSFLAGS(extra_ins) = FL_IMMED;
	  INS_ATTRIB(extra_ins) |= IF_CHANGED;
	  INS_TYPE(extra_ins) = IT_DATAPROC;
	  ARMINSCONDITION(extra_ins) = ARMINSCONDITION(ins);
	  INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
	  INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);
	  return;
	}
      }
  

  for (i = 0; i < 32; i += 2)
    if (((a = rotate_left (val_invert, i)) & 0xff) != 0)
      {
	t_ins * extra_ins = INS_NEXT(ins);
	if (a & 0xff00)
	  {
	    if (a & ~ 0xffff)
	      continue;
	  }
	else if (a & 0xff0000)
	  {
	    if (a & 0xff000000)
	      continue;
	  }

	if (!extra_ins || !ArmInstructionIsNOOP(extra_ins))
	{
	  if(!INS_DATA_POOL(ins))
	    FATAL(("no NOOP 7 for @I\n"));
	  else
	  {
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("3. Loading but could have done it in 2 ins from scratch: @I",ins));
#endif
	    not_needed = TRUE;
	  }
	}
	else
	{
	  ARMINSOPCODE(ins) = ARM_MVN;
	  ARMINSREGB(ins) = ARM_REG_NONE;
	  ARMINSREGC(ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(ins) = val_invert & rotate_left(0xff,32-i);
	  ARMINSFLAGS(ins) = FL_IMMED;
	  INS_ATTRIB(ins) |= IF_CHANGED;
	  INS_TYPE(ins) = IT_DATAPROC;
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

	  ARMINSOPCODE(extra_ins) = ARM_BIC;
	  ARMINSREGA(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGB(extra_ins) = ARMINSREGA(ins);
	  ARMINSREGC(extra_ins) = ARM_REG_NONE;
	  ARMINSIMMEDIATE(extra_ins) = val_invert & ~rotate_left(0xff,32-i);
	  ARMINSFLAGS(extra_ins) = FL_IMMED;
	  INS_ATTRIB(extra_ins) |= IF_CHANGED;
	  INS_TYPE(extra_ins) = IT_DATAPROC;
	  ARMINSCONDITION(extra_ins) = ARMINSCONDITION(ins);
	  INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
	  INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);
	  return;
	}
      }
  }

  if(max_offset < 65536)
  {
    t_ins * extra_ins = INS_NEXT(ins);
    if (extra_ins) 
    {
      if (!ArmInstructionIsNOOP(extra_ins))
      {
	if(!INS_DATA_POOL(ins))
	  FATAL(("no NOOP 70\n"));
	else
	{
#ifdef DEBUG_LAYOUT
	  VERBOSE(0,("Loading but can do it in two inses from PC: @I",ins));
#endif
	  not_needed = TRUE;
	}
      }
      else
      {

	if(uval < caddress)
	{
	  ArmInstructionMakeSub(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,max_offset & 0xff,ARMINSCONDITION(ins));
	  ArmInstructionMakeSub(extra_ins,ARMINSREGA(ins),ARMINSREGA(ins),ARM_REG_NONE,max_offset & 0xff00,ARMINSCONDITION(ins));
	}
	else
	{
	  ArmInstructionMakeAdd(ins,ARMINSREGA(ins),ARM_REG_R15,ARM_REG_NONE,max_offset & 0xff,ARMINSCONDITION(ins));
	  ArmInstructionMakeAdd(extra_ins,ARMINSREGA(ins),ARMINSREGA(ins),ARM_REG_NONE,max_offset & 0xff00,ARMINSCONDITION(ins));
	}
	/*DiabloPrint(stdout,"Constprod from pc in 2 ins\n");*/
	return;
      }
    }
  }

  /* Maybe there is another const prod preceding this one, that we can use?*/
  BBL_FOREACH_INS(bbl,j_ins)
    if(ARMINSOPCODE(j_ins) == ARM_CONSTANT_PRODUCER && ARMINSCONDITION(j_ins) == ARM_CONDITION_AL && j_ins != ins)
    {
      t_ins * r_ins = INS_NEXT(j_ins);
      while(r_ins != ins)
      {
	if(RegsetIn(INS_REGS_DEF(r_ins),ARMINSREGA(j_ins))) break;
	r_ins = INS_NEXT(r_ins);
      }

      if(r_ins == ins)
      {
	t_ins * extra_ins;
	val_abs = val_diff = ARMINSIMMEDIATE(j_ins) - val;
	if(val_diff < 0) val_abs = -val_diff; 

	for (i = 0; i < 32; i += 2)
	  if (((a = rotate_left (val_abs, i)) & 0xff) != 0)
	  {
	    if (a & 0xff00)
	    {
	      if (a & ~ 0xffff)
		continue;
	    }
	    else if (a & 0xff0000)
	    {
	      if (a & 0xff000000)
		continue;
	    }
	    else
	    {
	    }
	    
	    extra_ins = INS_NEXT(ins);
	    if (!extra_ins || !ArmInstructionIsNOOP(extra_ins))
	    {
	      if(!INS_DATA_POOL(ins))
		FATAL(("no NOOP x\n"));
	      else
	      {
#ifdef DEBUG_LAYOUT
		VERBOSE(0,("2. Loading but could have done it in 2 ins from previous constprod: @I",ins));
#endif
		not_needed = TRUE;
	      }
	    }
	    else
	    {
#ifdef DEBUG_LAYOUT
	      VERBOSE(0,("Generated constant from previous in 2 instructions, %x %x, @I, @I", val_abs, (val_abs & rotate_left(0xff,32-i))+(val_abs & ~rotate_left(0xff,32-i)),j_ins,ins));
#endif
	      if(val_diff > 0)
	      {
		ArmInstructionMakeSub(ins,ARMINSREGA(ins),ARMINSREGA(j_ins),ARM_REG_NONE,val_abs & rotate_left(0xff,32-i),ARMINSCONDITION(ins));
		ArmInstructionMakeSub(extra_ins,ARMINSREGA(ins),ARMINSREGA(ins),ARM_REG_NONE,val_abs & ~rotate_left(0xff,32-i),ARMINSCONDITION(ins));
	      }
	      else
	      {
		ArmInstructionMakeAdd(ins,ARMINSREGA(ins),ARMINSREGA(j_ins),ARM_REG_NONE,val_abs & rotate_left(0xff,32-i),ARMINSCONDITION(ins));
		ArmInstructionMakeAdd(extra_ins,ARMINSREGA(ins),ARMINSREGA(ins),ARM_REG_NONE,val_abs & ~rotate_left(0xff,32-i),ARMINSCONDITION(ins));
	      }
	      return;
	    }
	  }
      }
    }

  if(not_needed && (ARMINSFLAGS(ins) & FL_ORIG_CONST_PROD)) unefficient++;

  if(INS_DATA_POOL(ins))
  {
    t_bool dir_up = TRUE;
    t_ins * data_ins = NULL;
    t_uint32 load_offset = 0;

    /*VERBOSE(0,("Datapool at @B\n",INS_DATA_POOL_BBL(ins)));*/

    data_ins = InsFindIndexInPool(ins);

    /*    data_ins = BBL_INS_FIRST(INS_DATA_POOL_BBL(ins));
    BBL_FOREACH_INS(INS_DATA_POOL_BBL(ins),data_ins)
      if(ARMINSIMMEDIATE(data_ins) == 0) break;
    */
    if(!data_ins) FATAL(("No NOOP 4 for @I, data_pool is @B",ins,INS_DATA_POOL_BBL(ins)));
    
    if(GENERIC_GE(INS_CADDRESS(ins),INS_CADDRESS(data_ins)))
    {
      load_offset = G_T_UINT32(GENERIC_SUB(INS_CADDRESS(ins),INS_CADDRESS(data_ins))) + 8;
      dir_up = FALSE;
    }
    else
      load_offset = G_T_UINT32(GENERIC_SUB(INS_CADDRESS(data_ins),INS_CADDRESS(ins))) - 8;
      
    ARMINSOPCODE(ins) = ARM_LDR;
    ARMINSREGB(ins) = 15;
    ARMINSREGC(ins) = ARM_REG_NONE;
    ARMINSIMMEDIATE(ins) = load_offset;
    ARMINSFLAGS(ins) = FL_IMMED | FL_PREINDEX;
    if(dir_up) ARMINSFLAGS(ins) |= FL_DIRUP;
    INS_ATTRIB(ins) |= IF_CHANGED;
    INS_TYPE(ins) = IT_LOAD;
    INS_REGS_USE(ins) = ArmUsedRegisters(ins);
    INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

    ARMINSOPCODE(data_ins) = ARM_DATA;
    INS_TYPE(data_ins) = IT_DATA;
    ARMINSIMMEDIATE(data_ins) = val;
    return;
  }

  FATAL(("Don't know how to generate constant for @I",ins));
#if 0
  {
    t_ins * extra_ins = INS_NEXT(ins);
    t_ins * data_ins = INS_NEXT(extra_ins);

    if (!extra_ins) 
      FATAL(("No extra ins!"));
    if (!ArmInstructionIsNOOP(extra_ins))
      FATAL(("no NOOP 8 for @I\n",ins));

    if (!data_ins) 
      FATAL(("No data ins!"));
    if (!ArmInstructionIsNOOP(data_ins))
      FATAL(("no NOOP 9 for @I\n,MAx offset = %x\n",ins,max_offset));


    ARMINSOPCODE(ins) = ARM_LDR;
    ARMINSREGB(ins) = 15;
    ARMINSREGC(ins) = ARM_REG_NONE;
    ARMINSIMMEDIATE(ins) = 0;
    ARMINSFLAGS(ins) = FL_IMMED | FL_PREINDEX | FL_DIRUP;
    INS_ATTRIB(ins) |= IF_CHANGED;
    INS_TYPE(ins) = IT_LOAD;
    INS_REGS_USE(ins) = ArmUsedRegisters(ins);
    INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

    ARMINSOPCODE(extra_ins) = ARM_B;
    ARMINSREGA(extra_ins) = ARM_REG_NONE;
    ARMINSREGB(extra_ins) = ARM_REG_NONE;
    ARMINSREGC(extra_ins) = ARM_REG_NONE;
    ARMINSIMMEDIATE(extra_ins) = 0;
    ARMINSFLAGS(extra_ins) = FL_IMMED;
    INS_ATTRIB(extra_ins) |= IF_CHANGED;
    INS_TYPE(extra_ins) = IT_BRANCH;
    ARMINSCONDITION(extra_ins) = ARM_CONDITION_AL;
    INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
    INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);

    ARMINSOPCODE(data_ins) = ARM_DATA;
    INS_TYPE(data_ins) = IT_DATA;
    ARMINSIMMEDIATE(data_ins) = val;
    ARMINSIMMEDIATE(data_ins) = val;
  }
#endif
}
/*}}}*/
/*!
 * \todo Document
 *
 * \param ins
 *
 * \return void 
*/
/*GenerateInstructionsForFloatProd {{{ */
static void GenerateInstructionsForFloatProd(t_ins *ins)
{
  char * data=ARMINSDATA(ins);
  t_ins * extra_ins = INS_NEXT(ins);
  t_ins * data_ins = INS_NEXT(extra_ins);

  if (!extra_ins) 
    FATAL(("No extra ins!"));
  if (!ArmInstructionIsNOOP(extra_ins))
    FATAL(("no NOOP 8\n"));

  if (!data_ins) 
    FATAL(("No data ins!"));
  if (!ArmInstructionIsNOOP(data_ins))
    FATAL(("no NOOP 9\n"));

      
  ARMINSOPCODE(ins) = ARM_LDF;
  ARMINSREGB(ins) = 15;
  ARMINSREGC(ins) = ARM_REG_NONE;
  ARMINSIMMEDIATE(ins) = 0;
  ARMINSFLAGS(ins) |= FL_IMMED | FL_PREINDEX | FL_DIRUP;
  INS_ATTRIB(ins) |= IF_CHANGED;
  INS_TYPE(ins) = IT_FLT_LOAD;
  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);

  ARMINSOPCODE(extra_ins) = ARM_B;
  ARMINSREGA(extra_ins) = ARM_REG_NONE;
  ARMINSREGB(extra_ins) = ARM_REG_NONE;
  ARMINSREGC(extra_ins) = ARM_REG_NONE;
  ARMINSIMMEDIATE(extra_ins) = 0;
  ARMINSFLAGS(extra_ins) = FL_IMMED;
  INS_ATTRIB(extra_ins) |= IF_CHANGED;
  INS_TYPE(extra_ins) = IT_BRANCH;
  ARMINSCONDITION(extra_ins) = ARM_CONDITION_AL;
  INS_REGS_USE(extra_ins) = ArmUsedRegisters(extra_ins);
  INS_REGS_DEF(extra_ins) = ArmDefinedRegisters(extra_ins);

  ARMINSOPCODE(data_ins) = ARM_DATA;
  INS_TYPE(data_ins) = IT_DATA;
  ARMINSIMMEDIATE(data_ins) = *((t_uint32 *) data);

  if ((ARMINSFLAGS(ins)&FL_FLT_DOUBLE) || (ARMINSFLAGS(ins)&FL_FLT_DOUBLE_EXTENDED) || (ARMINSFLAGS(ins)&FL_FLT_PACKED))
  {
    data+=4;
    ARMINSIMMEDIATE(extra_ins) += 4;
    data_ins=INS_NEXT(data_ins);
    ARMINSOPCODE(data_ins) = ARM_DATA;
    INS_TYPE(data_ins) = IT_DATA;
    ARMINSIMMEDIATE(data_ins) = *((t_uint32 *) data);
  }

  if ((ARMINSFLAGS(ins)&FL_FLT_DOUBLE_EXTENDED) || (ARMINSFLAGS(ins)&FL_FLT_PACKED))
  {
    data+=4;
    ARMINSIMMEDIATE(extra_ins) += 4;
    data_ins=INS_NEXT(data_ins);
    ARMINSOPCODE(data_ins) = ARM_DATA;
    INS_TYPE(data_ins) = IT_DATA;
    ARMINSIMMEDIATE(data_ins) = *((t_uint32 *) data);
  }

}
/*}}}*/
/*!
 * \todo Document
 *
 * \param edge
 *
 * \return void 
*/
/*ArmUpdateBranchDisplacement{{{*/
static void ArmUpdateBranchDisplacement(t_cfg_edge * edge)
{
  t_bbl * head = CFG_EDGE_HEAD(edge);
  t_bbl * tail = CFG_EDGE_TAIL(edge);
  t_ins * ins = BBL_INS_LAST(head);

  
  ARMINSIMMEDIATE(ins) = G_T_UINT32(BBL_CADDRESS(tail)) - (G_T_UINT32(BBL_CADDRESS(head)) + BBL_NINS(head) * 4 - 4) - 8;
  INS_ATTRIB(ins) |= IF_CHANGED;
  /* minus 4 because then we have the address where the jump starts, minus 8 because pc points 8 too far */
}
/*}}}*/
/*!
 * \todo Document
 *
 * \param cfg
 *
 * \return void 
*/
/*!
 * \todo Document
 *
 * \param cfg
 *
 * \return void 
*/
/* ArmRelocate {{{ */
static void ArmRelocate(t_cfg * cfg, t_uint32 mode)
{
  t_reloc * reloc;
  t_bbl * i_bbl;
  t_ins * i_ins;
  t_cfg_edge * i_edge;

  /* adapt the displacements encoded in direct control flow transfer instructions {{{ */
  ANALYSIS_START("Relocating branch displacements");
  CFG_FOREACH_EDGE(cfg,i_edge)
    if (EDGE_CAT(i_edge)==ET_JUMP || EDGE_CAT(i_edge)==ET_IPJUMP || EDGE_CAT(i_edge)==ET_CALL)
      if (BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)) && ( ARMINSOPCODE(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)))==ARM_B || ARMINSOPCODE(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)))==ARM_BL))
      {
	ArmUpdateBranchDisplacement(i_edge);
      }
  ANALYSIS_SIMPLE_END("Relocating branch displacements");
  /* }}} */
  /* relocate the addresses stored in non-code sections {{{ */
  ANALYSIS_START("Relocating in non-code sections");
  OBJECT_FOREACH_RELOC(cfg->sec->obj,reloc)
  {
    if (!REL_TO_RELOCATABLE(reloc)) continue;
    if (REL_TO_RELOCATABLE(reloc)->type==RT_INS) FATAL(("relocs to ins"));
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_SUBSECTION)
    {
      if ((REL_TO_RELOCATABLE(reloc)->type==RT_BBL) || (REL_TO_RELOCATABLE(reloc)->type==RT_SUBSECTION)|| (REL_TO_RELOCATABLE(reloc)->type==RT_SECTION))
      {
	t_address ret;
	if (GENERIC_EXTRACT_UINT32(SECTION_GET_OLD_ADDRESS(T_SECTION(REL_FROM(reloc))))==0x0)
	  ret=StackExec(reloc->code,reloc,T_SECTION(REL_FROM(reloc))->data,TRUE,0);
	else
	  ret=StackExec(reloc->code,reloc,T_SECTION(REL_FROM(reloc))->data,TRUE,mode);
      }
      else FATAL(("IMPLEMENT 2 > %d",REL_TO_RELOCATABLE(reloc)->type));
    }
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_BBL) continue;
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_INS) continue;
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_SECTION) FATAL(("reloc from section"));
    else FATAL(("Unknown reloc @R\n",reloc));
  }
  ANALYSIS_SIMPLE_END("Relocating in non-code sections");
  /* }}} */
  /* relocate the addresses stored in data blocks in the text section {{{ */
  ANALYSIS_START("Relocating in code sections");
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_FOREACH_INS(i_bbl,i_ins)
    {      
      if (RELOCATABLE_REFERS_TO(i_ins))
      {
	t_arm_ins rins;
	char tmp[4];
	t_address ret;
	t_reloc * reloc = RELOCATABLE_REFERS_TO(i_ins)->rel;

	ArmAssembleOne(i_ins,tmp);
	if (GENERIC_EXTRACT_UINT32(INS_OLD_ADDRESS(i_ins))==0xffffffff)
	  ret=StackExec(reloc->code,reloc,tmp,TRUE,0);
	else
	  ret=StackExec(reloc->code,reloc,tmp,TRUE,mode);

	if (INS_TYPE(i_ins)==IT_DATA)
	{
	  ArmDisassembleData(T_INS(&rins),*(t_uint32 *) tmp,ARM_DATA);
	  ARMINSIMMEDIATE(i_ins)=ARMINSIMMEDIATE(&rins);
	}
	else
	{
	  arm_opcode_table[ARMINSOPCODE(i_ins)].Disassemble((t_ins *) &rins,*(t_uint32 *) tmp,ARMINSOPCODE(i_ins));
	}
	INS_SECTION(&rins)=INS_SECTION(i_ins);
	INS_BBL(&rins)=INS_BBL(i_ins);
	INS_CADDRESS(((t_ins *) &rins))=INS_CADDRESS(i_ins);
	INS_OLD_ADDRESS(((t_ins *) &rins))=INS_OLD_ADDRESS(i_ins);
      }
    }
  }
  ANALYSIS_SIMPLE_END("Relocating in code sections");
  /* }}} */

  /* set new entry point for the object */
  cfg->sec->obj->entry =  BBL_CADDRESS(cfg->entries->entry_bbl);
}
/*}}}*/
/*!
 * 
 *
 * \param cfg
 *
 * \return t_bbl *
*/
/*  ArmChain {{{ */
static t_bbl * ArmChain(t_cfg * cfg)
{
  t_bbl ** chains=NULL;
  int nchains=0, tel;
  t_bbl * i_bbl;

  t_bbl * heads=NULL;
  
  t_cfg_edge * i_edge;
  /* Create chains of one block long */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_NEXT_IN_CHAIN(i_bbl) = NULL;
    BBL_PREV_IN_CHAIN(i_bbl) = NULL;
  }

  VERBOSE(0,("Basic block fallthrough chaining"));




  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_bbl * j_bbl;
    t_bool switch_bbl=FALSE;
    t_cfg_edge * ft=NULL;
    t_cfg_edge * ft_like=NULL;
    t_int32 switch_max=0;
    if(!NODE_PRED_FIRST(i_bbl)) continue;
    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      if ((EDGE_CAT(i_edge)==ET_SWITCH) || (EDGE_CAT(i_edge)==ET_SWITCH))
      {
	switch_bbl=TRUE;
	/*	DiabloPrint(stdout,"@E = %d\n",i_edge,i_edge->switchvalue);*/
	if (i_edge->switchvalue>switch_max)
	  switch_max=i_edge->switchvalue;
      }
      else if ((EDGE_CAT(i_edge)==ET_FALLTHROUGH)||(EDGE_CAT(i_edge)==ET_IPFALLTHRU))
      {
	if (ft_like && (CFG_EDGE_TAIL(i_edge)!=CFG_EDGE_TAIL(ft_like))) FATAL(("More than one ft-like edge in bbl @eB",i_bbl));
	if (!ft_like) ft_like=i_edge;
	ft=i_edge;
      }
      else if ((EDGE_CAT(i_edge)==ET_FUNLINK) ||
	       (EDGE_CAT(i_edge)==ET_SWILINK) ||
	       (EDGE_CAT(i_edge)==ET_IPFUNLINK) ||
	       (EDGE_CAT(i_edge)==ET_IPSWILINK))
      {
	if (ft_like && (CFG_EDGE_TAIL(i_edge)!=CFG_EDGE_TAIL(ft_like))) FATAL(("More than one ft-like edge in bbl @eB",i_bbl));
	ft_like=ft=i_edge;
      }
    }
    if (switch_bbl && (ARMINSOPCODE(BBL_INS_LAST(i_bbl))==ARM_ADD))
/* {{{ Handle switch statements */
    {
      t_bbl ** array= Calloc(sizeof(t_bbl *),switch_max+2); /* 0 holds fallthrough */
      t_bbl * prev=i_bbl;
      DiabloPrint(stdout,"Found switch block @B, max switch value = %d\n",i_bbl,switch_max);
     
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
      {
	if ((EDGE_CAT(i_edge)==ET_SWITCH) || (EDGE_CAT(i_edge)==ET_IPSWITCH))
	{
	  if (array[i_edge->switchvalue+1]) FATAL(("Switch values corrupt!"));
	  array[i_edge->switchvalue+1]=CFG_EDGE_TAIL(i_edge);
	}
	else if ((EDGE_CAT(i_edge)==ET_FALLTHROUGH)||(EDGE_CAT(i_edge)==ET_IPFALLTHRU))
	{
	  if (array[0]) FATAL(("Multiple fallthroughs in switch"));
	  array[0]=CFG_EDGE_TAIL(i_edge);
	}
	else
	{
	  FATAL(("Weird edge in switch: @E\n",i_edge));
	}
      }

      for (tel=0; tel<switch_max+2; tel++)
      {
	if (!array[tel])
	{
	  t_ins * noop;
	  printf("Added bbl for case %d\n",tel);
	  array[tel]=BblNew(cfg);
	  BblInsertInFunction(array[tel],BBL_FUN(i_bbl));
	  noop=InsNewForBbl(array[tel]);
	  ArmInstructionMakeMov(noop,ARM_REG_R0,ARM_REG_R0,0,ARM_CONDITION_AL);
	  InsAppendToBbl(noop,array[tel]); 
	}
	/*	else
		printf("%d ok\n",tel);*/
      }
	
      for (tel=0; tel<switch_max+2; tel++)
      {
	if(BBL_NEXT_IN_CHAIN(prev)) FATAL(("Already next for @B, i.e. @B",i_bbl,BBL_NEXT_IN_CHAIN(i_bbl)));
	if(BBL_PREV_IN_CHAIN(array[tel])) FATAL(("Already prev for @B, i.e. @B",array[tel],BBL_PREV_IN_CHAIN(array[tel])));
	BBL_NEXT_IN_CHAIN(prev)=array[tel]; 
	BBL_PREV_IN_CHAIN(array[tel])=prev;
	prev=array[tel];
      }
      
      Free(array);
    }
    else if (switch_bbl && (ARMINSOPCODE(BBL_INS_LAST(i_bbl))==ARM_LDR))
    {
      if (ft_like)
      {
	j_bbl=CFG_EDGE_TAIL(ft_like);

	if(BBL_NEXT_IN_CHAIN(i_bbl) && (BBL_NEXT_IN_CHAIN(i_bbl)!=j_bbl)) FATAL(("Already next for @B, i.e. @B",i_bbl,BBL_NEXT_IN_CHAIN(i_bbl)));
	if(BBL_PREV_IN_CHAIN(j_bbl) && (BBL_PREV_IN_CHAIN(j_bbl)!=i_bbl)) FATAL(("Already prev for @B, i.e. @B != @B",j_bbl,BBL_PREV_IN_CHAIN(j_bbl),i_bbl));
	BBL_NEXT_IN_CHAIN(i_bbl)=j_bbl; 
	BBL_PREV_IN_CHAIN(j_bbl)=i_bbl;
	if (!(IS_SWITCH_TABLE(BBL_NEXT(j_bbl))))
	{
	  FATAL(("No switch table found for switch at @B with FT @B\n",i_bbl,j_bbl));
	}
	else 
	{
	  BBL_NEXT_IN_CHAIN(j_bbl)=BBL_NEXT(j_bbl); 
	  BBL_PREV_IN_CHAIN(BBL_NEXT(j_bbl))=j_bbl;
	}
      }
      else
      {
	if (!(IS_SWITCH_TABLE(BBL_NEXT(i_bbl))))
	{
	  FATAL(("No switch table found for switch at @B\n",i_bbl));
	}
	else
	{
	  t_ins * noop;
	  j_bbl=BblNew(cfg);
	  BblInsertInFunction(j_bbl,BBL_FUN(i_bbl));
	  noop=InsNewForBbl(j_bbl);
	  ArmInstructionMakeMov(noop,ARM_REG_R0,ARM_REG_R0,0,ARM_CONDITION_AL);
	  InsAppendToBbl(noop,j_bbl); 
	  BBL_NEXT_IN_CHAIN(i_bbl)=j_bbl; 
	  BBL_PREV_IN_CHAIN(j_bbl)=i_bbl;
	  BBL_NEXT_IN_CHAIN(j_bbl)=BBL_NEXT(i_bbl); 
	  BBL_PREV_IN_CHAIN(BBL_NEXT(i_bbl))=j_bbl;

	}
      }
      
    }
/* }}} */    
    else if (ft_like)
    {
      /* Normal fall through chaining */
      j_bbl=CFG_EDGE_TAIL(ft_like);

      if(BBL_NEXT_IN_CHAIN(i_bbl) && (BBL_NEXT_IN_CHAIN(i_bbl)!=j_bbl)) FATAL(("Already next for @B, i.e. @B",i_bbl,BBL_NEXT_IN_CHAIN(i_bbl)));
      if(BBL_PREV_IN_CHAIN(j_bbl) && (BBL_PREV_IN_CHAIN(j_bbl)!=i_bbl)) 
      {
	DiabloPrint(stdout,"@ieB -> @ieB\n",CFG_EDGE_HEAD(ft_like),j_bbl);
	DiabloPrint(stdout,"Previous = @ieB\n",BBL_PREV_IN_CHAIN(j_bbl));
	FATAL(("Already prev for @B, i.e. @B != @B",j_bbl,BBL_PREV_IN_CHAIN(j_bbl),i_bbl));
      }

      

      
      BBL_NEXT_IN_CHAIN(i_bbl)=j_bbl; 
      BBL_PREV_IN_CHAIN(j_bbl)=i_bbl;
    }

  }




#if 0  /*{{{*/
  /* Append Fallthrough paths */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    j_bbl = NULL;
    if(NODE_PRED_FIRST(i_bbl)) /* Only do this, when the node has predecessors, because otherwise its an unreachable block and
				  causes troubles in lay-outing the chains */
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
      {
	/* chain fallthrough paths together.
	 * for switches it's possible that the fallthrough case
	 * has been removed. in that case we should chain in the
	 * first switch case. a noop representing the fallthrough
	 * case will be inserted below */
	if (EDGE_CAT(i_edge)==ET_FUNLINK ||
	    EDGE_CAT(i_edge)==ET_SWILINK ||
	    EDGE_CAT(i_edge)==ET_IPFUNLINK ||
	    EDGE_CAT(i_edge)==ET_IPSWILINK ||
	    EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
	    EDGE_CAT(i_edge)==ET_IPFALLTHRU
	   )
	{
	  j_bbl = CFG_EDGE_TAIL(i_edge);
	  break;
	}

	/*  if ((EDGE_CAT(i_edge)==ET_SWITCH) && (i_edge->switchvalue==0))
	    {
	    j_bbl = CFG_EDGE_TAIL(i_edge);
	    }
	    */
      }

    if (j_bbl)
    {
      if(BBL_NEXT_IN_CHAIN(i_bbl) && (BBL_NEXT_IN_CHAIN(i_bbl)!=j_bbl)) FATAL(("Already next for @B, i.e. @B",i_bbl,BBL_NEXT_IN_CHAIN(i_bbl)));
      if(BBL_PREV_IN_CHAIN(j_bbl) && (BBL_PREV_IN_CHAIN(j_bbl)!=i_bbl)) FATAL(("Already prev for @B, i.e. @B != @B",j_bbl,BBL_PREV_IN_CHAIN(j_bbl),i_bbl));
      BBL_NEXT_IN_CHAIN(i_bbl)=j_bbl; 
      BBL_PREV_IN_CHAIN(j_bbl)=i_bbl;
    }

    if (BBL_INS_LAST(i_bbl) && (INS_ATTRIB(BBL_INS_LAST(i_bbl)) & IF_SWITCHJUMP) && (ARMINSOPCODE(BBL_INS_LAST(i_bbl)) == ARM_ADD))
    {
#if 0
      /* ads-style switch: chain together the default case with the switch cases */
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	if (EDGE_CAT(i_edge) == ET_SWITCH && i_edge->switchvalue == 0)
	  break;
      VERIFY_FASSERT(i_edge,("Could not find first switch edge"));

      if (!j_bbl)
      {
	j_bbl=CFG_EDGE_TAIL(i_edge);
      if((BBL_NEXT_IN_CHAIN(i_bbl)&&(BBL_NEXT_IN_CHAIN(i_bbl)!=j_bbl)) || (BBL_PREV_IN_CHAIN(j_bbl)&&(BBL_PREV_IN_CHAIN(j_bbl)!=i_bbl))) FATAL(("Already chained! @B and @B",i_bbl,j_bbl));
	BBL_NEXT_IN_CHAIN(i_bbl)=j_bbl; BBL_PREV_IN_CHAIN(j_bbl)=i_bbl; 
      }
      else
      {
	if(BBL_NEXT_IN_CHAIN(j_bbl) || BBL_PREV_IN_CHAIN(CFG_EDGE_TAIL(i_edge)))
	  FATAL(("Already chained! @B and @B",j_bbl,CFG_EDGE_TAIL(i_edge)));
	BBL_NEXT_IN_CHAIN(j_bbl) = CFG_EDGE_TAIL(i_edge);
	BBL_PREV_IN_CHAIN(CFG_EDGE_TAIL(i_edge)) = j_bbl;
      }
#endif
    }
  }
#endif /*}}}*/
/*  VERBOSE(0,("Switch Chaining"));*/
#if 0 /* {{{ */
  /* support for switch tables of the gcc type: the table consists of addresses and 
   * the pc is loaded directly from that table */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (IS_SWITCH_TABLE(i_bbl))
    {
      DiabloPrint(stdout,"Chaining switch table @B, @B\n",i_bbl,BBL_PREV(i_bbl));
      if (G_T_UINT32(BBL_OLD_ADDRESS(i_bbl))==4+G_T_UINT32(BBL_OLD_ADDRESS(BBL_PREV(i_bbl))))
      {
	if(BBL_NEXT_IN_CHAIN(BBL_PREV(i_bbl)) || BBL_PREV_IN_CHAIN(i_bbl))
	  FATAL(("Already chained! @B and @B",BBL_PREV(i_bbl),i_bbl));
	BBL_NEXT_IN_CHAIN(BBL_PREV(i_bbl))=i_bbl;
	BBL_PREV_IN_CHAIN(i_bbl)=BBL_PREV(i_bbl);
      }
      else if (G_T_UINT32(BBL_OLD_ADDRESS(i_bbl))==8+G_T_UINT32(INS_OLD_ADDRESS(BBL_INS_LAST(BBL_PREV(i_bbl)))))
      {
	DiabloPrint(stdout,"Changed jump @I\n",(BBL_INS_LAST(BBL_PREV(i_bbl))));

	if(BBL_NEXT_IN_CHAIN(BBL_PREV(i_bbl)) || BBL_PREV_IN_CHAIN(i_bbl))
	  FATAL(("Already chained! @B and @B",BBL_PREV(i_bbl),i_bbl));
	BBL_NEXT_IN_CHAIN(BBL_PREV(i_bbl))=i_bbl;
	BBL_PREV_IN_CHAIN(i_bbl)=BBL_PREV(i_bbl);
      }else
      {
	BblPrint(BBL_PREV(i_bbl),PRINT_INSTRUCTIONS);
	printf("-------------------------------\n");
	BblPrint(i_bbl,PRINT_INSTRUCTIONS);
	FATAL(("potential problem with layouting switch table !!!: switch table without preceeding load and jump\n"));
      }
    }
  }
  VERBOSE(0,("Switch Noops"));


  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      if (INS_ATTRIB(i_ins)&IF_SWITCHJUMP)
      {
	/*DiabloPrint(stdout,"Found switch: @I\n",i_ins);*/
	if (ARMINSCONDITION(i_ins)==ARM_CONDITION_AL)  
	{
	  t_ins * j_ins = InsNewForCfg(cfg);
	  GenericSet_uint32(INS_CSIZE(j_ins),4);
	  GenericSet_uint32(INS_OLD_SIZE(j_ins),0);
	  ArmInstructionMakeNoop(j_ins);
	  InsInsertAfter(j_ins,i_ins);
	}
#ifdef INSTRUMENTATION
	BBL_FOREACH_SUCC_EDGE(INS_BBL(i_ins),i_edge)
	  DiabloPrint(stdout, "succ @E switchvalue %d\n",i_edge,i_edge->switchvalue);
	BBL_FOREACH_SUCC_EDGE(INS_BBL(i_ins),i_edge)
	  if(EDGE_CAT(i_edge) == ET_FALLTHROUGH) break;
	if(i_edge)
	{
	  if(BBL_NEXT_IN_CHAIN(i_bbl) != CFG_EDGE_TAIL(i_edge))
	      FATAL(("Switch fallthrough not next in chain!"));
	}
	
	else/* if (0) */
	{
	  t_uint32 switchvalue = 0;
	  t_uint32 maxvalue = 0;
	  t_bbl * curr = i_bbl, *next = NULL;
	  BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	    if(EDGE_CAT(i_edge)==ET_SWITCH && i_edge->switchvalue > maxvalue) maxvalue = i_edge->switchvalue;
	  while(switchvalue <= maxvalue)
	  {
	    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	      if(EDGE_CAT(i_edge)==ET_SWITCH && i_edge->switchvalue == switchvalue)
	      {
		if(BBL_PREV_IN_CHAIN(CFG_EDGE_TAIL(i_edge)) && BBL_PREV_IN_CHAIN(CFG_EDGE_TAIL(i_edge)) != curr) FATAL(("@iB already has prev! @iB",CFG_EDGE_TAIL(i_edge),BBL_PREV_IN_CHAIN(CFG_EDGE_TAIL(i_edge))));
		BBL_NEXT_IN_CHAIN(curr) = CFG_EDGE_TAIL(i_edge);
		BBL_PREV_IN_CHAIN(CFG_EDGE_TAIL(i_edge)) = curr;
		curr = CFG_EDGE_TAIL(i_edge);
		if(BBL_NEXT_IN_CHAIN(curr))
		  FATAL(("Already next! @iB,  @iB",curr,BBL_NEXT_IN_CHAIN(curr)));
	      }
	    switchvalue++;
	  }
	}
#endif /*INSTRUMENTATION*/
	    
      }
    }
  }
#endif /* }}} */
  VERBOSE(0,("Chain count"));

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (BBL_PREV_IN_CHAIN(i_bbl)==NULL) 
      nchains++;
  }

  chains=Malloc(sizeof(t_bbl *) * nchains);

  VERBOSE(0,("  %d chains found",nchains));
  VERBOSE(0,("Set head of chains"));
  nchains=0;
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_bbl * tmp;
    if (BBL_PREV_IN_CHAIN(i_bbl)==NULL) 
    {
      tmp=i_bbl;
      while(tmp) 
      { 
/*	BBL_FIRST_IN_CHAIN(tmp)=nchains;*/
	tmp=BBL_NEXT_IN_CHAIN(tmp); 
      }
      chains[nchains]=i_bbl;
      nchains++;
    }
  }


  VERBOSE(0,("Chain the chains"));
  /* We want the original data to be placed near the blocks that reference them {{{*/
  for (tel=0; tel<nchains; tel++)
  {
    if(chains[tel] && IS_DATABBL(chains[tel]))
    {
      t_bbl * tail;
      t_bbl * other_tail = NULL;
/*      DiabloPrint(stdout,"-----------> data chain:\n @iB\n",chains[tel]);*/
      
      if(BBL_NEXT_IN_CHAIN(chains[tel])) FATAL(("Data chain with more than 1 bbl! @iB",chains[tel]));
      if(!RELOCATABLE_REFED_BY(chains[tel]))
      {
	VERBOSE(1,("Chain is not refed by, skipping"));
	continue; 
      }
      
      if(RELOCATABLE_TYPE(REL_FROM(RELOCATABLE_REFED_BY(chains[tel])->rel)) == RT_INS)
      {
	if(RELOCATABLE_REFED_BY(chains[tel])->next)
	{
	  if(RELOCATABLE_TYPE(REL_FROM(RELOCATABLE_REFED_BY(chains[tel])->next->rel)) == RT_INS)
	    other_tail = INS_BBL(REL_FROM(RELOCATABLE_REFED_BY(chains[tel])->next->rel));
	  while(BBL_NEXT_IN_CHAIN(other_tail)) other_tail = BBL_NEXT_IN_CHAIN(other_tail);
	  if(RELOCATABLE_REFED_BY(chains[tel])->next->next) VERBOSE(1,("Lots of references!"));
	}
	tail = INS_BBL(REL_FROM(RELOCATABLE_REFED_BY(chains[tel])->rel));
	while(BBL_NEXT_IN_CHAIN(tail)) tail = BBL_NEXT_IN_CHAIN(tail);
#if 0 /* Not debugged yet */
	if(other_tail && other_tail != tail)
	{
	  int tel2;
	  while(BBL_PREV_IN_CHAIN(other_tail)) other_tail = BBL_PREV_IN_CHAIN(other_tail);
	  BBL_PREV_IN_CHAIN(other_tail) = chains[tel];
	  BBL_NEXT_IN_CHAIN(chains[tel]) = other_tail;
	  for(tel2 = 0; tel2 < nchains; tel2++)
	    if(chains[tel2] == other_tail)
	    {
	      chains[tel2] = NULL;
	      break;
	    }
	}
#endif
	/* Now we have got the end of the chain */
	BBL_NEXT_IN_CHAIN(tail) = chains[tel];
	BBL_PREV_IN_CHAIN(chains[tel]) = tail;
	chains[tel] = NULL;
      }

/*      else VERBOSE(0,("From type == %d",RELOCATABLE_TYPE(REL_FROM(RELOCATABLE_REFED_BY(chains[tel])->rel))));*/
    }
  }/*}}}*/

  /* We reorder the chains such that hot BBL's that reference the data section are placed at the end of the code section {{{*/
  /*ArmReorderChains(cfg,chains,nchains);*/
  /*}}}*/
  for (tel=0; tel<nchains; tel++)
  {
    if(!chains[tel]) continue;
    if(BBL_PREV_IN_CHAIN(chains[tel])) FATAL(("Already prev @B",chains[tel]));
    BBL_PREV_IN_CHAIN(chains[tel])=heads;
    heads=chains[tel];
    chains[tel]=NULL;
  }

  VERBOSE(0,("Correct previous in chain"));
  {
    t_bbl * tmp;
    t_bbl * ret=heads;
    t_bbl * prev=NULL;
    while (heads)
    {
      t_bbl * tmp=heads;
      t_bbl * nexthead=BBL_PREV_IN_CHAIN(heads);




      /* Sanity check */
      while(BBL_NEXT_IN_CHAIN(tmp)) { 
	if(tmp==BBL_NEXT_IN_CHAIN(tmp)) FATAL(("@B is its own nextinchain!",tmp));
	tmp=BBL_NEXT_IN_CHAIN(tmp); }

	/* tmp is now the last block from the chain */

	BBL_NEXT_IN_CHAIN(tmp)=nexthead;
	BBL_PREV_IN_CHAIN(heads)=prev;
	prev=tmp;
	heads=nexthead;
    }
    tmp=ret;
    VERBOSE(0,("Chain created"));

    /* Now remove the empty (return-) blocks as they need not be layouted! */
    /*while(tmp)
    {
      t_bbl * tmp2 = BBL_NEXT_IN_CHAIN(tmp);
      if(!BBL_INS_FIRST(tmp))
      {
	if(!BBL_INS_LAST(tmp))
	{
	  if(!BBL_NINS(tmp))
	  {
	    //DiabloPrint(stdout,"Removing @iB from chain, its empty!\n",tmp);
	    if (BBL_PREV_IN_CHAIN(tmp)) BBL_NEXT_IN_CHAIN(BBL_PREV_IN_CHAIN(tmp)) = BBL_NEXT_IN_CHAIN(tmp);
	    else ret = BBL_NEXT_IN_CHAIN(tmp);
	    if (BBL_NEXT_IN_CHAIN(tmp)) BBL_PREV_IN_CHAIN(BBL_NEXT_IN_CHAIN(tmp)) = BBL_PREV_IN_CHAIN(tmp);
	    BBL_NEXT_IN_CHAIN(tmp) = BBL_PREV_IN_CHAIN(tmp) = NULL;
	  }
	  else
	    FATAL(("No first and last but still %d ins? @iB",BBL_NINS(tmp),tmp));
	}
	else
	  FATAL(("No first, but still last? @iB",tmp));
      }
      
      tmp=tmp2;
    }*/

    Free(chains);
    return ret;
  }
}
/* }}} */
/*!
 * \todo Document
 *
 * \param head
 * \param cfg
 *
 * \return void 
*/
/* ArmAddressProducers {{{ */
extern t_int32 nr_loaders;
extern t_int32 nr_addr_prod;
extern t_int32 unique_addr;
static void ArmAddressProducers(t_bbl * head, t_cfg * cfg, t_uint32 mode)
{
  t_address section_begin_address, section_old_begin_address;
  t_address old_address, new_address;
  t_bbl * i_bbl;
  t_bbl * new_bbl = NULL;
  t_int32 nr_of_datapools = 0, nr_loaders_left = 0,nr_addr_in_two = 0;
  int entries = 0,count = 0;
  t_ins * i_ins;
  t_ins * j_ins;

  t_bool change,allow_shrink, allow_load_to_two;

  if (global_options.blockprofilefile)
    CfgComputeHotBblThreshold(cfg,0.95);

  section_begin_address = old_address = new_address = SECTION_GET_CADDRESS(cfg->sec);
  section_old_begin_address = SECTION_GET_CADDRESS(cfg->sec);

  if(mode == 2)
    /* Change all addressproducers from the original code to constantproducers, since we have to produce the old addresses {{{*/
  {
    i_bbl = head;
    while(i_bbl)
    {
      if(!IS_DATABBL(i_bbl))
      {
	BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  if(ARMINSOPCODE(i_ins) == ARM_ADDRESS_PRODUCER && GENERIC_EXTRACT_UINT32(INS_OLD_ADDRESS(i_ins))!=0xffffffff)
	  {
	    t_reloc * reloc = RELOCATABLE_REFERS_TO(i_ins)->rel;
	    ARMINSIMMEDIATE(i_ins) = G_T_UINT32(StackExec(reloc->code,reloc,NULL,FALSE,2));
	    ARMINSOPCODE(i_ins) = ARM_CONSTANT_PRODUCER;
	  }
	}
      }
    }
  }/*}}}*/

  i_bbl = head; 
  /* Initialize {{{ */
  VERBOSE(0,("Text section starts at @G/@G",section_begin_address,section_old_begin_address));
  while(i_bbl)
  {
    BBL_ATTRIB(i_bbl) &= ~BBL_DATA_POOL;
    BBL_ATTRIB(i_bbl) &= ~BBL_DATA_POOL_MIGRATED;
    i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
  }

  i_bbl = head; 

  while (i_bbl)
  {
    if (!IS_DATABBL(i_bbl))
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	GenericSet_uint32(INS_MIN_SIZE(i_ins),4);
	INS_DATA_POOL(i_ins) = NULL;
	ARMINSFLAGS(i_ins) &= ~FL_NO_LOAD;

	if (ARMINSOPCODE(i_ins)==ARM_ADDRESS_PRODUCER)
	{
	  /* Set the maximal instruction size for an address producer to 8, because we will produce
	   * it in 2 instructions or we load the data from a datapool somewhere */
	  GenericSet_uint32(INS_CSIZE(i_ins),8);
	  count++;
	}
	else if (ARMINSOPCODE(i_ins)==ARM_CONSTANT_PRODUCER)
	{
	  t_uint32 csize = 4 * ConstProdGuessNumberOfInstructionsNeeded(i_ins, NULL_GENERIC,NULL_GENERIC);

	  if(csize)
	  {
	    ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
	    GenericSet_uint32(INS_CSIZE(i_ins),csize);
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("No doubt: should not be loaded! @I",i_ins));
#endif
	  }
	  else
	    GenericSet_uint32(INS_CSIZE(i_ins),8);
	  if(ARMINSFLAGS(i_ins) & FL_ORIG_CONST_PROD) count++;
	}
	else if (ARMINSOPCODE(i_ins)==ARM_FLOAT_PRODUCER)  /* {{{  */
	{
	  j_ins = InsNewForCfg(cfg);
	  GenericSet_uint32(INS_CSIZE(j_ins),4);
	  GenericSet_uint32(INS_MIN_SIZE(j_ins),4);
	  ArmInstructionMakeNoop(j_ins);
	  InsInsertAfter(j_ins,i_ins);

	  j_ins = InsNewForCfg(cfg);
	  GenericSet_uint32(INS_CSIZE(j_ins),4);
	  GenericSet_uint32(INS_MIN_SIZE(j_ins),4);
	  ArmInstructionMakeNoop(j_ins);
	  InsInsertAfter(j_ins,i_ins);

	  if ((ARMINSFLAGS(i_ins)&FL_FLT_DOUBLE)  || (ARMINSFLAGS(i_ins)&FL_FLT_DOUBLE_EXTENDED) || (ARMINSFLAGS(i_ins)&FL_FLT_PACKED))
	  {
	    j_ins = InsNewForCfg(cfg);
	    GenericSet_uint32(INS_CSIZE(j_ins),4);
	    GenericSet_uint32(INS_MIN_SIZE(j_ins),4);
	    ArmInstructionMakeNoop(j_ins);
	    InsInsertAfter(j_ins,i_ins);
	  }

	  if ((ARMINSFLAGS(i_ins)&FL_FLT_DOUBLE_EXTENDED) || (ARMINSFLAGS(i_ins)&FL_FLT_PACKED))
	  {
	    j_ins = InsNewForCfg(cfg);
	    GenericSet_uint32(INS_CSIZE(j_ins),4);
	    GenericSet_uint32(INS_MIN_SIZE(j_ins),4);
	    ArmInstructionMakeNoop(j_ins);
	    InsInsertAfter(j_ins,i_ins);
	  }
	}/*}}}*/
      }
    }

    BblSetAllAddressSuper(i_bbl,old_address,new_address);
    new_address=GENERIC_ADD(new_address,BBL_CSIZE(i_bbl));
    old_address=GENERIC_ADD(old_address,BBL_MIN_SIZE(i_bbl));
    i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
  }
  VERBOSE(0,("And runs till @G/@G",new_address,old_address));
  /*}}}*/

  PlaceSections(cfg->sec->obj,FALSE);
  PlaceSections(cfg->sec->obj,TRUE);

  /* Create datapools for all address producers and for constant producers that cannot be produced
   * in 1 or 2 instructions without the pc *//*{{{*/
  i_bbl = head;
  while(i_bbl)
  {
    if (!IS_DATABBL(i_bbl))
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
      {

	if (ARMINSOPCODE(i_ins)==ARM_ADDRESS_PRODUCER || (ARMINSOPCODE(i_ins)==ARM_CONSTANT_PRODUCER && !(ARMINSFLAGS(i_ins) & FL_NO_LOAD)))
	{
	  t_bool pool_found = FALSE;
	  t_bbl * data_pool = INS_BBL(i_ins);
	  t_bbl * forward_candidate = NULL, *backward_candidate = NULL;
	  t_uint32 offset;
	  t_ins * an_ins;
	  /* Find a data_pool first {{{*/

	  /* Initialize offset for backward datapool. Since we would add an element before this
	   * address producer, the c_address would lie 4 bytes further */
	  offset= G_T_UINT32(GENERIC_SUB(INS_CADDRESS(i_ins),BBL_CADDRESS(data_pool))) + 12;
	  /* Search backward... */
	  while(data_pool && offset < LOAD_OFFSET)
	  {
	    /* If this bbl is marked as having a datapool as predeccessor, use this datapool */
	    if(BBL_PREV_DATA_POOL(data_pool))
	    {
	      pool_found = TRUE;
	      backward_candidate = data_pool;
	      break;
	    }
	    /* If the block preceding this block has an unconditional control transfer, we can put data before
	     * the current block. We should however check if there are switch edges, since we can not place
	     * data in between switchblocks. We mark this block as candidate if we have no candidate yet. */
	    if(BBL_PREV_IN_CHAIN(data_pool) && !BblGetSuccEdgeOfType(BBL_PREV_IN_CHAIN(data_pool),ET_FT_BLOCK) && !backward_candidate)
	      if(BBL_PREV_IN_CHAIN(BBL_PREV_IN_CHAIN(data_pool)) &&
		  !BblGetSuccEdgeOfType(BBL_PREV_IN_CHAIN(BBL_PREV_IN_CHAIN(data_pool)),ET_SWITCH) &&
		  !BblGetPredEdgeOfType(data_pool,ET_SWITCH))
		backward_candidate = data_pool;

	    BBL_FOREACH_INS(data_pool,an_ins)
	    {
	      if(INS_DATA_POOL(an_ins))
	      {
		/* We find an ins with a datapool: we check if this pool can be used an stop if this is so */
		if((G_T_UINT32(GENERIC_SUB(INS_CADDRESS(i_ins),BBL_CADDRESS(INS_DATA_POOL_BBL(an_ins))))+12) < LOAD_OFFSET)
		{
		  data_pool = INS_DATA_POOL_BBL(an_ins);
		  pool_found = TRUE;
		  backward_candidate = data_pool;
		}
		else
		/* If we cannot use the pool, we must stop since otherwise the distance between an_ins and its data_pool can
		 * increase, which is not allowed. */
		{
		  data_pool = NULL;
		}
	      }
	    }
	    if(pool_found || !data_pool ) break;
	    data_pool = BBL_PREV_IN_CHAIN(data_pool);
	    if(data_pool)
	      offset += G_T_UINT32(BBL_CSIZE(data_pool));
	  }

	  /* Search forward if necessary */
	  if(!pool_found && BBL_NEXT_IN_CHAIN(INS_BBL(i_ins)))
	  {
	    data_pool = BBL_NEXT_IN_CHAIN(INS_BBL(i_ins));
	    offset = G_T_UINT32(GENERIC_SUB(BBL_CADDRESS(data_pool),INS_CADDRESS(i_ins))) - 8;
	    while(data_pool && ((t_int32)offset) < LOAD_OFFSET)
	    {
	      if(BBL_PREV_DATA_POOL(data_pool))
	      {
		pool_found = TRUE;
		forward_candidate = data_pool;
		break;
	      }


	      /* If the block preceding this block has an unconditional control transfer, we can put data before
	       * the current block. We should however check if there are switch edges, since we can not place
	       * data in between switchblocks. We mark this block as candidate even if we have a candidate yet.
	       * Doing so will result in data_pool that are spread more evenly accross the program */
	      if(!BblGetSuccEdgeOfType(BBL_PREV_IN_CHAIN(data_pool),ET_FT_BLOCK))
		/* If there are switch edges, we can not place data in between the blocks! */
		if(BBL_PREV_IN_CHAIN(BBL_PREV_IN_CHAIN(data_pool)) &&
		    !BblGetSuccEdgeOfType(BBL_PREV_IN_CHAIN(BBL_PREV_IN_CHAIN(data_pool)),ET_SWITCH) &&
		    !BblGetPredEdgeOfType(data_pool,ET_SWITCH))
		  forward_candidate = data_pool;

	      offset += G_T_UINT32(BBL_CSIZE(data_pool));
	      data_pool = BBL_NEXT_IN_CHAIN(data_pool);
	    }
	  }

	  /* If we still have no pool found, we take the forward candidate, if it exists, else the backward
	   * candidate. If none of these exist, we have to split a basic block. */
	  if(!pool_found)
	  {
	    if(forward_candidate)
	    {
	      data_pool = forward_candidate;
	      pool_found = TRUE;
	    }
	    else if(backward_candidate) 
	    {
	      data_pool = backward_candidate;
	      pool_found = TRUE; 
	    }
	  }

	  if(!pool_found) /* Split a block... */
	  {
	    t_ins * j_ins;
	    t_cfg_edge * edge;

	    j_ins = InsNewForCfg(cfg);
	    InsInsertAfter(j_ins,i_ins);
	    SECTION_NINS(cfg->sec)++;
	    if(BBL_INS_LAST(i_bbl) != j_ins)
	    {
	      data_pool = BblSplitBlock(i_bbl,j_ins,FALSE);
	      EDGE_CAT(BBL_SUCC_FIRST(i_bbl)) = ET_JUMP;
	    }
	    else
	    {
	      BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
		if(EDGE_CAT(edge) == ET_FALLTHROUGH) EDGE_CAT(edge) = ET_JUMP;
	      data_pool = BBL_NEXT_IN_CHAIN(i_bbl);
	    }
	    ArmInstructionMakeUncondBranch(j_ins);
	    GenericSet_uint32(INS_CSIZE(i_ins),8);

	    BblSetAllAddressSuper(i_bbl,BBL_MIN_ADDRESS(i_bbl),BBL_CADDRESS(i_bbl));
	    BblSetAllAddressSuper(data_pool,GENERIC_ADD(BBL_MIN_ADDRESS(i_bbl),BBL_MIN_SIZE(i_bbl)),GENERIC_ADD(BBL_CADDRESS(i_bbl),BBL_CSIZE(i_bbl)));

	    /*DiabloPrint(stdout,"Block splitted @iB\n and @iB\n",i_bbl,data_pool);*/
	  }
	  /*}}}*/

	  BBL_ATTRIB(data_pool) |= BBL_DATA_POOL;

	  {
	    t_int32 slack;

	    if (data_pool==forward_candidate)
	      {
		slack = 100000;
	      }
	    else if (data_pool==backward_candidate)
	      {
		slack = LOAD_OFFSET -8 - GENERIC_EXTRACT_UINT32(GENERIC_SUB(INS_CADDRESS(i_ins),BBL_CADDRESS(data_pool)));
		if (slack<0)
		  FATAL(("OOPS"));
	      }
	    else
	      {
		slack = 0;
	      }

	    if (InsAddToPool(i_ins,data_pool,slack))
	    {
	      /* Set the instruction size to 4, since we can reuse a dataslot*/
	      GenericSet_uint32(INS_CSIZE(i_ins),4);
#ifdef DEBUG_LAYOUT
	      VERBOSE(0,("7:Down to 4 for addr @I with datapool @B and slack %x",i_ins,data_pool,slack));
#endif
	    }
#ifdef DEBUG_LAYOUT
	    else
	      VERBOSE(0,("7:Down to 8 for addr @I with datapool @B and slack %x",i_ins,data_pool,slack));
#endif
	  }
	}
      }
    }
    i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);

  }/*}}}*/

  {
    i_bbl=head;
    new_address = SECTION_GET_CADDRESS(cfg->sec);
    old_address = SECTION_GET_CADDRESS(cfg->sec);
    while (i_bbl)
    {
      BblSetAllAddressSuper(i_bbl,old_address,new_address);
      new_address=GENERIC_ADD(new_address,BBL_CSIZE(i_bbl));
      old_address=GENERIC_ADD(old_address,BBL_MIN_SIZE(i_bbl));
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }

  }
 
  /* Now we try to get rid of the address and constant producers that load the data, and change them into
   * data processing instructions. When using c_address, we have to adjust it because when using datapools,
   * the data is not stored where the addressproducer is stored. */
#ifndef LOAD_EVERYTHING
  allow_shrink = TRUE;
  allow_load_to_two = FALSE;
  do
  {
    change = FALSE;
    i_bbl = head;

    while(i_bbl)
    {
      /*t_bool hotbll = BblIsHot(i_bbl);*/
      if(!IS_DATABBL(i_bbl))
      {
	BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  if (ARMINSOPCODE(i_ins)==ARM_ADDRESS_PRODUCER)/*{{{*/
	  {
	    t_uint32 my_min_value,my_max_value=GENERIC_EXTRACT_UINT32(INS_CADDRESS(i_ins))+8;
	    t_uint32 swap;

	    t_uint32 max_value,min_value;

	    t_uint32 offset;

	    t_bool pass_shrink = FALSE;

	    /* Skip this one if already minimal sized */
	    if ((GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins))==4) && !INS_DATA_POOL(i_ins)) continue;

#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("________________________________"));
#endif
	    /* Set the my_min and my_max. This is the current address as in the CADDRESS-field, and the current address adjusted
	     * to include instruction-shifts due to data-pool stuff */
	    my_min_value = my_max_value - NumberOfPrevInstructionsWithForwardDatapool(i_ins)*4 + NumberOfNextInstructionsWithBackwardDatapool(i_ins)*4;
	    if(my_min_value > my_max_value){ swap = my_max_value; my_max_value = my_min_value; my_min_value = swap;}

	    /* Get the c_address of the destination */
	    max_value=GENERIC_EXTRACT_UINT32(StackExec(RELOCATABLE_REFERS_TO(i_ins)->rel->code,RELOCATABLE_REFERS_TO(i_ins)->rel,NULL,FALSE,0));

	    min_value = max_value;

	    /* If the destination address is in the code section, adjust it to keep track of data_pool stuff */
	    if(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)) == RT_BBL)
	    {
#ifdef DEBUG_LAYOUT
	      VERBOSE(0,("To: max_value %x, @B, offset: %d",max_value,T_BBL(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)),RELOCATABLE_REFERS_TO(i_ins)->rel->to_offset));
#endif
	      if(BBL_INS_FIRST(T_BBL(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel))))
	      {
		min_value += NumberOfNextInstructionsWithBackwardDatapool(BBL_INS_FIRST(T_BBL(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel))))*4;
		min_value -=NumberOfPrevInstructionsWithForwardDatapool(BBL_INS_FIRST(T_BBL(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel))))*4;
	      }
	      else
	      {
		min_value += NumberOfNextInstructionsWithBackwardDatapool(BBL_INS_FIRST(T_BBL(BBL_NEXT_IN_CHAIN(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)))))*4;
		min_value -= NumberOfPrevInstructionsWithForwardDatapool(BBL_INS_FIRST(T_BBL(BBL_PREV_IN_CHAIN(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)))))*4;
	      }
	      if (min_value > max_value){ swap = max_value; max_value =min_value;min_value = swap;}
	    }
	    /*else if(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)) == RT_SECTION)
	    {
	      DiabloPrint(stdout,"To section %s: aligment = @G\n",SECTION_NAME(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel))?SECTION_NAME(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)):"no name",REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)->alignment);
	    }
	    else if(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)) == RT_SUBSECTION)
	    {
	      DiabloPrint(stdout,"To section %s, parent %s: aligment = @G\n",SECTION_NAME(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel))?SECTION_NAME(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)):"no name",SECTION_NAME(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel))),T_RELOCATABLE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(RELOCATABLE_REFERS_TO(i_ins)->rel)))->alignment);
	    }*/

	    if (my_max_value<max_value)
	      offset=max_value - my_min_value;
	    else
	      offset=my_max_value - min_value;



#ifdef DEBUG_LAYOUT
	    DiabloPrint(stdout,"Produced new address = %x, offset = %x, datapool? @G\n",max_value,offset,INS_DATA_POOL(i_ins)?BBL_CADDRESS(INS_DATA_POOL_BBL(i_ins)):NULL_GENERIC);
#endif

	    if(allow_shrink && ((max_value&0x3)==0)&&(offset<1020))
	    {
	      GenericSet_uint32(INS_CSIZE(i_ins),4);
	      change=TRUE;
	      pass_shrink=TRUE;
	      InsRemoveFromPool(__FILE__,__LINE__,i_ins);
	      ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
	      VERBOSE(0,("1:Down to 4 for addr @I",i_ins));
#endif
	    }
	    if(!pass_shrink && allow_shrink && offset < 256)
	    {
	      GenericSet_uint32(INS_CSIZE(i_ins),4);
	      change=TRUE;
	      pass_shrink=TRUE;
	      InsRemoveFromPool(__FILE__,__LINE__,i_ins);
	      ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
	      VERBOSE(0,("2:Down to 4 for addr @I",i_ins));
#endif
	    }
	    /*TODO: Add another case here!*/
	    if(allow_load_to_two && !pass_shrink &&
		(
		 ((max_value&0x3)==0 &&   (offset<0x3fffc))
		 || (offset<0xffff) 
		))
	    {
	      if (INS_DATA_POOL(i_ins))
	      {
		change=TRUE;
		pass_shrink=TRUE;
		INS_DATA_POOL(i_ins) = NULL;
		ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
	        if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins)) != 8) FATAL(("Should be size 8 when we get here! %d, @I",GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins)),i_ins));
		VERBOSE(0,("3:Without datapool for addr @I",i_ins));
#endif
	      }

	    }
	    if(allow_load_to_two && !pass_shrink && ((max_value&0x3)==0) && (max_value < 0x3fffc))
	    {
	      if (INS_DATA_POOL(i_ins))
	      {
		change=TRUE;
		pass_shrink=TRUE;
		INS_DATA_POOL(i_ins) = NULL;
		ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
		if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins)) != 8) FATAL(("Should be size 8 when we get here!"));
		VERBOSE(0,("4:Without datapool for addr @I",i_ins));
#endif
	      }
	    }
	    if(allow_load_to_two && !allow_shrink && !pass_shrink && (max_value == min_value) && ArmConstantIsProducableInTwoInstructions(min_value))
	    {
	      if (INS_DATA_POOL(i_ins))
	      {
		change=TRUE;
		pass_shrink=TRUE;
		INS_DATA_POOL(i_ins) = NULL;
		ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
		if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins)) != 8) FATAL(("Should be size 8 when we get here!"));
		VERBOSE(0,("5:Without datapool for addr @I",i_ins));
#endif
	      }
	    }
	    if(allow_load_to_two && !allow_shrink && !pass_shrink && (max_value == min_value) && (my_max_value == my_min_value) && ArmConstantIsProducableInTwoInstructions(min_value))
	    {
	      if (INS_DATA_POOL(i_ins))
	      {
		change=TRUE;
		pass_shrink=TRUE;
		INS_DATA_POOL(i_ins) = NULL;
		ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
		if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins)) != 8) FATAL(("Should be size 8 when we get here!"));
		VERBOSE(0,("6:Without datapool for addr @I",i_ins));
#endif
	      }
	    }

	  }/*}}}*/
	  if (ARMINSOPCODE(i_ins)==ARM_CONSTANT_PRODUCER)/*{{{*/
	  {
	    t_uint32 my_min_value,my_max_value=GENERIC_EXTRACT_UINT32(INS_CADDRESS(i_ins))+8;
	    t_uint32 swap;

	    t_uint32 old_size;
	    t_uint32 new_size;

	    if ((GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins))==4)) continue;
#ifdef DEBUG_LAYOUT
	    VERBOSE(0,("________________________________"));
#endif
	    /* Set the my_min and my_max. This is the current address as in the CADDRESS-field, and the current address adjusted
	     * to include instruction-shifts due to data-pool stuff */
	    my_min_value = my_max_value - NumberOfPrevInstructionsWithForwardDatapool(i_ins)*4 + NumberOfNextInstructionsWithBackwardDatapool(i_ins)*4;
	    if(my_min_value > my_max_value){ swap = my_max_value; my_max_value = my_min_value; my_min_value = swap;}

	    old_size = GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins));
	    new_size = ConstProdGuessNumberOfInstructionsNeeded(i_ins,my_min_value,my_max_value)*4;

	    if(!new_size)
	    {
	      if(!INS_DATA_POOL(i_ins)) FATAL(("Should have data_pool! @I",i_ins));
	    }
	    else
	    {

	      if (new_size < old_size && allow_shrink)
	      {
		if(old_size != 8) FATAL(("Should be 8!"));
		GenericSet_uint32(INS_CSIZE(i_ins),new_size);
		change = TRUE;
		InsRemoveFromPool(__FILE__,__LINE__,i_ins);
		ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
		VERBOSE(0,("8:Down from %d to %d for @I",old_size,new_size,i_ins));
#endif
	      }
	      if (allow_load_to_two && new_size == 8 && INS_DATA_POOL(i_ins))
	      {
		/* We can do it without datapool now! */
		if(GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins)) != 8) FATAL(("Should be 8!"));
		change = TRUE;
		INS_DATA_POOL(i_ins) = NULL;
		ARMINSFLAGS(i_ins) |= FL_NO_LOAD;
#ifdef DEBUG_LAYOUT
		VERBOSE(0,("9:Without datapool for @I",i_ins));
#endif
	      }
	    }
#ifdef DEBUG_LAYOUT
	    if(INS_DATA_POOL(i_ins)) DiabloPrint(stdout,"@I has pool @G\n",i_ins,BBL_CADDRESS(INS_DATA_POOL_BBL(i_ins)));
#endif
	  }/*}}}*/
	}
      }
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }


    /* Update the c_address and min_address fields of instructions and bbl's */
    {
      i_bbl=head;
      new_address = section_begin_address;
      old_address = section_begin_address;
      while (i_bbl)
      {
	BblSetAllAddressSuper(i_bbl,old_address,new_address);
	new_address=GENERIC_ADD(new_address,BBL_CSIZE(i_bbl));
	old_address=GENERIC_ADD(old_address,BBL_MIN_SIZE(i_bbl));
	i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
      }
      PlaceSections(cfg->sec->obj,FALSE);
      PlaceSections(cfg->sec->obj,TRUE);
      /*OBJECT_FOREACH_SECTION(cfg->sec->obj,sec,tel)*/
      /*  DiabloPrint(stdout,"Section %s at @G/@G\n",sec->name,SECTION_GET_CADDRESS(sec),SECTION_GET_MIN_ADDRESS(sec));*/
      VERBOSE(0,("And runs till @G/@G",new_address,old_address));
    }
    /* If there are no more changes, but it is still allowed to shrink sizes, do another round without shrinking.
     * This way, quite some loads can still be changed into 2 dataproc inses. */
    if(!change && allow_shrink)
    {
      allow_shrink = FALSE;
      change = TRUE;
    }
      
  } while(change || allow_shrink);

#endif /*LOAD_EVERYTHING*/

  /* Insert enough noops now at the right place *//*{{{*/
  i_bbl = head;
  while (i_bbl)
  {
    if (!IS_DATABBL(i_bbl))
    {
      new_bbl = NULL;
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	if(INS_DATA_POOL(i_ins))
	{
	  /* Make new block to insert data (if necessary), and set the data_pool of the ins to the new block */
          if(!BBL_DATA_POOL_IS_MIGRATED(INS_DATA_POOL_BBL(i_ins)))
	    /*	  if(!BBL_PREV_IN_CHAIN(INS_DATA_POOL_BBL(i_ins)) || !BBL_PREV_DATA_POOL(BBL_PREV_IN_CHAIN(INS_DATA_POOL_BBL(i_ins)))) */
	  {
	    t_int32 i;
	    t_data_pool_entry * entry;
	    new_bbl = BblNew(BBL_CFG(i_bbl));
	    nr_of_datapools++;
	    BBL_ATTRIB(new_bbl) |= BBL_DATA_POOL_MIGRATED;
	    BBL_NEXT_IN_CHAIN(new_bbl) = INS_DATA_POOL_BBL(i_ins);
	    BBL_PREV_IN_CHAIN(new_bbl) = BBL_PREV_IN_CHAIN(INS_DATA_POOL_BBL(i_ins));
	    BBL_PREV_IN_CHAIN(INS_DATA_POOL_BBL(i_ins)) = new_bbl;
	    if(BBL_PREV_IN_CHAIN(new_bbl))
	      BBL_NEXT_IN_CHAIN(BBL_PREV_IN_CHAIN(new_bbl)) = new_bbl;
	    else
	    {
	      head = new_bbl;
	      BBL_CADDRESS(new_bbl) = BBL_CADDRESS(INS_DATA_POOL_BBL(i_ins));
	    }

	    BBL_ATTRIB(new_bbl) |= BBL_DATA_POOL;

	    entry = INS_DATA_POOL(i_ins)->list;
	    
	    for (i=0;i<INS_DATA_POOL(i_ins)->nr_entries;i++,entry=entry->next)
	      {
		if (entry->nr_ins>0)
		  {
		    t_inslist * inslist;
		    j_ins = InsNewForCfg(cfg);
		    GenericSet_uint32(INS_CSIZE(j_ins),4);
		    GenericSet_uint32(INS_MIN_SIZE(j_ins),4);
		    ArmInstructionMakeNoop(j_ins);
		    InsPrependToBbl(j_ins,new_bbl);
		    if (INS_REFERS_TO(POOLENTRY_FIRSTINS(entry)))
		      {
			ArmInstructionMakeData(j_ins,0);
		      }
		    else
		      {
			ArmInstructionMakeData(j_ins,ARMINSIMMEDIATE(POOLENTRY_FIRSTINS(entry)));
		      }
		    inslist = entry->inslist;
		    while(inslist)
		    {
		      /* Count this entry when it is used for an addr producer or an original const prod */
		      if(ARMINSOPCODE(inslist->ins) == ARM_ADDRESS_PRODUCER || ARMINSFLAGS(inslist->ins) & FL_ORIG_CONST_PROD)
		      {
			entries++;
			break;
		      }
		      inslist = inslist->next;
		    }
		  }/*
		else
		  DiabloPrint(stdout, "SAVED ONE ENTRY @R\n",entry->rel->rel); */
	      }

	    INS_DATA_POOL_BBL(i_ins) = new_bbl;

	  /*VERBOSE(0,("Datapool for @I is @B",i_ins,new_bbl));*/
	    
	    

	  }
	  else new_bbl = INS_DATA_POOL_BBL(i_ins);

	  GenericSet_uint32(INS_CSIZE(i_ins),4);
	  GenericSet_uint32(INS_MIN_SIZE(i_ins),4);

          /* Insert a noop in this block */
	  if(ARMINSOPCODE(i_ins) == ARM_ADDRESS_PRODUCER || ARMINSFLAGS(i_ins) & FL_ORIG_CONST_PROD)
	    nr_loaders_left++;
	  
	}
	else if (ARMINSOPCODE(i_ins)==ARM_ADDRESS_PRODUCER)
	{
	  if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins))>8) FATAL(("Size > 8 is not possible anymore!"));

	  if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins))== 8)
	  {
	    nr_addr_in_two++;
	    j_ins = InsNewForCfg(cfg);
	    GenericSet_uint32(INS_CSIZE(j_ins),4);
	    GenericSet_uint32(INS_MIN_SIZE(j_ins),4);
	    ArmInstructionMakeNoop(j_ins);
	    InsInsertAfter(j_ins,i_ins);
	    SECTION_NINS(cfg->sec)++;
	    GenericSet_uint32(INS_CSIZE(i_ins),4);
	    GenericSet_uint32(INS_MIN_SIZE(i_ins),4);
	  }
	}
	else if (ARMINSOPCODE(i_ins)==ARM_CONSTANT_PRODUCER)
	{
	  if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins))>8) FATAL(("Size 12 is not good for constants anymore!"));
	  if (GENERIC_EXTRACT_UINT32(INS_CSIZE(i_ins))== 8)
	  {
	    nr_addr_in_two++;
	    j_ins = InsNewForCfg(cfg);
	    GenericSet_uint32(INS_CSIZE(j_ins),4);
	    GenericSet_uint32(INS_MIN_SIZE(j_ins),4);
	    ArmInstructionMakeNoop(j_ins);
	    InsInsertAfter(j_ins,i_ins);
	    SECTION_NINS(cfg->sec)++;
	    GenericSet_uint32(INS_CSIZE(i_ins),4);
	    GenericSet_uint32(INS_MIN_SIZE(i_ins),4);
	  }
	}
      }
    }
    i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
  }/*}}}*/
 
  {
    i_bbl=head;
    new_address = SECTION_GET_CADDRESS(cfg->sec);
    old_address = SECTION_GET_CADDRESS(cfg->sec);
    while (i_bbl)
    {
#ifdef DEBUG_LAYOUT
      if(BBL_ATTRIB(i_bbl) & BBL_DATA_POOL) VERBOSE(0,("Datapool: size = @G\n@B",BBL_CSIZE(i_bbl),i_bbl));
#endif
      BblSetAllAddressSuper(i_bbl,old_address,new_address);
      new_address=GENERIC_ADD(new_address,BBL_CSIZE(i_bbl));
      old_address=GENERIC_ADD(old_address,BBL_MIN_SIZE(i_bbl));
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }

  }
 
  PlaceSections(cfg->sec->obj,FALSE);
 
#if 0
  DistributionProducedAddresses(cfg);
#endif

  i_bbl = head;
  while (i_bbl)
  {
    if (!IS_DATABBL(i_bbl))
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	if (ARMINSOPCODE(i_ins)==ARM_ADDRESS_PRODUCER) GenerateInstructionsForAddrProd(i_ins);
	else if (ARMINSOPCODE(i_ins)==ARM_FLOAT_PRODUCER) GenerateInstructionsForFloatProd(i_ins);
      }
      BBL_FOREACH_INS_R(i_bbl,i_ins)
      {
	if (ARMINSOPCODE(i_ins)==ARM_CONSTANT_PRODUCER) GenerateInstructionsForConstProd(i_ins);
      }
    }
    i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
  }

  i_bbl = head;
  while(i_bbl)
  {
    if(BBL_INS_FIRST(i_bbl) && (INS_TYPE(BBL_INS_FIRST(i_bbl)) != IT_DATA) && (BBL_ATTRIB(i_bbl) & BBL_DATA_POOL))
    {
      /*DiabloPrint(stdout,"@iB cannot be datapool!\n",i_bbl);*/
      BBL_ATTRIB(i_bbl) &= ~BBL_DATA_POOL;
    }
#ifdef DEBUG_LAYOUT
    if(BBL_ATTRIB(i_bbl) & BBL_DATA_POOL)
    {
      DiabloPrint(stdout,"________________!!!!!!!!!!!!!!!!!!!___________________\n");
      BBL_FOREACH_INS(i_bbl,i_ins)
	DiabloPrint(stdout,"%x | @I\n",ARMINSIMMEDIATE(i_ins),i_ins);
    }
#endif
    i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
  }
  VERBOSE(0,("____________Address statistics____________"));
  VERBOSE(0,("Original number of address producers : %d",nr_addr_prod));
  VERBOSE(0,("Remaining number of address producers: %d (%.2f)",count, ((float)count/nr_addr_prod)*100.0));
  VERBOSE(0,("Original number of loaders : %d",nr_loaders));
  VERBOSE(0,("Remaining number of loaders: %d (%.2f)",nr_loaders_left, ((float)nr_loaders_left/nr_loaders)*100.0));
  VERBOSE(0,("Original number of entries : %d",unique_addr));
  VERBOSE(0,("Remaining number of entries: %d (%.2f)",entries, ((float)entries/unique_addr)*100.0));
  VERBOSE(0,("Number of datapools: %d",nr_of_datapools));
  VERBOSE(0,("Number of unneccesary loads: %d",unefficient));
  VERBOSE(0,("In two inses: %d",nr_addr_in_two));
  VERBOSE(0,("__________________________________________"));
}

/*}}}*/
/*!
 * \todo Document
 *
 * \param head
 * \param cfg
 *
 * \return void 
*/
/* ArmCfgLayout {{{ */
t_uint32 ArmCfgLayout(t_cfg * cfg, t_uint32 mode)
{
  t_bbl * i_bbl;
  t_bbl * chain_head;
  t_uint32 size,old_size;

  old_size = G_T_UINT32(SECTION_CSIZE(cfg->sec));
 
  CfgVerifyCorrectness(cfg);
 
  chain_head=ArmChain(cfg);

  PlaceSections(cfg->sec->obj,FALSE);

  ArmAddressProducers(chain_head,cfg,mode);

#ifdef SCHEDULE
  ScheduleCfg(cfg);
  /*  SchedulerCfgListSchedule(cfg);*/
#endif
  
  ArmRelocate(cfg,mode);
  
  /* calculate the size of the ER_RO section */
  size = 0;
  CFG_FOREACH_BBL(cfg,i_bbl)
    size+=BBL_CSIZE(i_bbl);
 
  printf("%d total number of 4-byte words of the code section\n",size/4);
  if((t_int32)(old_size - size) >= 0)
  printf("GAIN: %lf\n", ((double) (old_size - size))/((double)old_size) );
  else
  printf("GAIN: -%lf\n", ((double) (size - old_size))/((double)old_size) );
  DiabloPrint(stdout,"New entry point @G\n",cfg->sec->obj->entry);

#if 1
  {
    FILE *f = fopen("./final.list","w");
    t_ins * i_ins;

    i_bbl = chain_head; 
    while (i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
	DiabloPrint(f,"@I\n",i_ins);
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }
  }
#endif

  FreePoolData();

  return size;
}
/* }}} */

static t_data_pool pools[4096]; 
static int number_of_pools = 0;

void FreePoolData()
{
  t_data_pool_entry * iterator1;
  t_inslist * iterator2;
  int i;

  for (i=0;i<number_of_pools;i++)
    {
      iterator1 = pools[i].list;
      while (iterator1)
	{
	  t_data_pool_entry * old1 = iterator1;
	  iterator2 = iterator1->inslist;
	  while(iterator2)
	    {
	      t_inslist * old2=iterator2;
	      iterator2=iterator2->next;
	      Free(old2);
	    }
	  iterator1=iterator1->next;
	  Free(old1);
	}
    }
}

t_bool RelocEqual(t_reloc * reloc1, t_reloc * reloc2)
{
  /*  DiabloPrint(stdout,"compare\n @R\n @R\n",reloc1,reloc2);*/

  if (reloc1==reloc2)
    return TRUE;
  if (REL_TO_RELOCATABLE(reloc1)!=REL_TO_RELOCATABLE(reloc2))
    return FALSE;
  if (GENERIC_EQ(GENERIC_ADD(reloc1->to_offset,reloc1->addend),GENERIC_ADD(reloc2->to_offset,reloc2->addend)))
    return TRUE;
  return FALSE;
}

t_bool InsAddToPool(t_ins * ins, t_bbl * bbl, t_int32 slack)
{
  int i;
  t_bool ret = FALSE;
  t_data_pool * pool;
  t_data_pool_entry * pool_entry;
  t_inslist * pool_ins;

  for (i=0;i<number_of_pools;i++)
  {
    if(pools[i].bbl == bbl) break;
  }
  
  if(i == number_of_pools)
  {
    pools[i].bbl = bbl;
    pools[i].list = NULL;
    pools[i].end = NULL;
    pools[i].nr_entries = 0;
    number_of_pools++;
    /*    printf("number of pools: %d\n",number_of_pools);*/
  }
  pool = &(pools[i]);
  /* We have a datapool now, lets look for the reloc */
  pool_entry = pool->list;

  /*  DiabloPrint(stdout,"@I %x\n",ins,INS_REFERS_TO(ins));fflush(stdout);*/
  
  for (i=0;i<pool->nr_entries-(slack/4);i++)
    {
      pool_entry=pool_entry->next;
    }

  while(pool_entry)
    {
      
      if (ARMINSOPCODE(ins) == ARM_CONSTANT_PRODUCER && ARMINSOPCODE(POOLENTRY_FIRSTINS(pool_entry)) == ARM_CONSTANT_PRODUCER)
	{
	  if (ARMINSIMMEDIATE(POOLENTRY_FIRSTINS(pool_entry))==ARMINSIMMEDIATE(ins))
	    break;
	}
      else if (ARMINSOPCODE(ins) == ARM_ADDRESS_PRODUCER && ARMINSOPCODE(POOLENTRY_FIRSTINS(pool_entry)) == ARM_ADDRESS_PRODUCER)
	{
	  if (RelocEqual(INS_REFERS_TO(POOLENTRY_FIRSTINS(pool_entry))->rel,INS_REFERS_TO(ins)->rel))
	      break;
	}    
      pool_entry = pool_entry->next;
    }
  
  if(!pool_entry)
  {
    pool_entry = Malloc(sizeof(t_data_pool_entry));
/*    pool_entry->rel = INS_REFERS_TO(ins);*/
/*    pool_entry->data = ARMINSIMMEDIATE(ins);*/
    pool_entry->next = NULL;
    pool_entry->inslist = NULL;
    pool_entry->nr_ins = 0;
    if(pool->end == NULL)
      pool->list = pool_entry;
    else
      pool->end->next = pool_entry;
    pool->end = pool_entry;
    pool->nr_entries++;
  }
  else ret = TRUE;
  
  pool_ins = Malloc(sizeof(t_inslist)); 
  pool_ins->ins = ins;
  pool_ins->next = pool_entry->inslist;
  pool_entry->inslist = pool_ins;
  pool_entry->nr_ins++;

  INS_DATA_POOL(ins) = pool;

  return ret;
}

t_ins* InsFindIndexInPool(t_ins * ins)
{
  t_data_pool * pool = INS_DATA_POOL(ins);

  t_data_pool_entry * pool_entry;

  t_ins * result = BBL_INS_FIRST(pool->bbl);
  pool_entry = pool->list;
  
  while(pool_entry)
    {
      t_inslist * pool_ins = pool_entry->inslist;
      if (pool_entry->nr_ins==0)
	{
	  pool_entry = pool_entry->next;
	  continue;
	}

      while (pool_ins)
	{
	  if (pool_ins->ins==ins)
	    return result;
	  pool_ins = pool_ins->next;
	}
      pool_entry = pool_entry->next;
      result=INS_NEXT(result);
    }
  return NULL;
}  

void InsRemoveFromPool(char * file, int line, t_ins *ins)
{
  t_data_pool * pool = INS_DATA_POOL(ins);
  t_data_pool_entry * pool_entry;
  
  pool_entry = pool->list;
  
  while(pool_entry)
    {
      t_inslist * pool_ins = pool_entry->inslist;
      while (pool_ins)
	{
	  if (pool_ins->ins==ins)
	    goto found;
	  pool_ins = pool_ins->next;
	}
      pool_entry = pool_entry->next;
    }

  FATAL(("OOPS"));

 found:
  pool_entry->nr_ins--;
  if (pool_entry->nr_ins==0)
    {
      /*      printf("SAVED2 \n");*/
    }
  INS_DATA_POOL(ins) = NULL;
#ifdef DEBUG_LAYOUT
  DiabloPrint(stdout,"00. removed from pool in %s (%d): @I\n", file,line,ins);
#endif
}

#if 0
void DistributionProducedAddresses(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  t_uint32 produced_new_address;
  t_uint32 total_size = 0x8000000;
  t_uint32 * distribution_load = malloc(total_size); 
  t_uint32 * distribution_single = malloc(total_size); 
  t_uint32 i;

  CFG_FOREACH_BBL(cfg,bbl)
    {
      BBL_FOREACH_INS(bbl,ins)
	{
	  if (ARMINSOPCODE(ins)==ARM_ADDRESS_PRODUCER)
	    {
	      t_reloc * reloc = RELOCATABLE_REFERS_TO(ins)->rel; 
	      produced_new_address = G_T_UINT32(StackExec(reloc->code,reloc,NULL,FALSE,0));
	      if (produced_new_address>=total_size)
		FATAL(("OOPS"));
	      if (INS_DATA_POOL(ins))
		distribution_load[produced_new_address/4]++;
	      else
		distribution_single[produced_new_address/4]++;
	      DiabloPrint(stdout,"PROD 0x%8x INS @I\n",produced_new_address,ins); 
	    }
	  else
	    DiabloPrint(stdout,  "NOPROD          INS @I\n",ins); 
	}
      printf("BORDER -------------------------\n");
    }

  for (i=0;i<total_size/4;i++)
    {
      if (distribution_single[i]>0 || distribution_load[i]>0)
	printf("DISTR: 0x%x: %d %d\n",i*4,distribution_single[i],distribution_load[i]);
    }
}
#endif

t_int32 ComputeNrDep(t_ins * ins1, t_ins * ins2, t_ins * ins3, t_ins * ins4)
{
  t_int32 nr_deps = 0;
  
  if (ins1 && (INS_TYPE(ins1)==IT_LOAD || INS_TYPE(ins1)==IT_LOAD_MULTIPLE || INS_TYPE(ins1)==IT_FLT_LOAD))
    if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins1),INS_REGS_USE(ins2))))
      nr_deps += 1;
  
  if (ins4 && (INS_TYPE(ins3)==IT_LOAD || INS_TYPE(ins3)==IT_LOAD_MULTIPLE || INS_TYPE(ins3)==IT_FLT_LOAD ))
    if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins3),INS_REGS_USE(ins4))))
      nr_deps += 1;

  return nr_deps;
}

t_bool InsIsLoad(t_ins * ins)
{
  return INS_TYPE(ins)==IT_LOAD || INS_TYPE(ins)==IT_FLT_LOAD || INS_TYPE(ins)==IT_LOAD_MULTIPLE;
}

t_bool InsIsSimpleLoad(t_ins * ins)
{
  return INS_TYPE(ins)==IT_LOAD || INS_TYPE(ins)==IT_FLT_LOAD;
}

t_bool InsIsStore(t_ins * ins)
{
  return INS_TYPE(ins)==IT_STORE || INS_TYPE(ins)==IT_FLT_STORE || INS_TYPE(ins)==IT_STORE_MULTIPLE;
}

t_bool InsIsSimpleStore(t_ins * ins)
{
  return INS_TYPE(ins)==IT_STORE || INS_TYPE(ins)==IT_FLT_STORE;
}

t_bool NoAliasing(t_ins *ins1, t_ins *ins2)
{
  if (ARMINSREGB(ins1)!=ARMINSREGB(ins2)) return FALSE;
  if (ARMINSIMMEDIATE(ins1)==ARMINSIMMEDIATE(ins2)) return FALSE;
  return TRUE;
}

t_ins * PreviousInsFT(t_ins* ins)
{
  t_ins * ret = INS_PREV(ins);

  t_cfg_edge * edge;

  if (ret) 
    return ret;

  BBL_FOREACH_PRED_EDGE(INS_BBL(ins),edge)
    {
      if (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	return BBL_INS_LAST(CFG_EDGE_HEAD(edge));
    }
  
  return NULL;
}

t_ins * NextInsFT(t_ins* ins)
{
  t_ins * ret = INS_NEXT(ins);

  t_cfg_edge * edge;

  if (ret) 
    return ret;

  BBL_FOREACH_SUCC_EDGE(INS_BBL(ins),edge)
    {
      if (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	return BBL_INS_FIRST(CFG_EDGE_TAIL(edge));
    }
  
  return NULL;
}

void ScheduleCfg(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  t_ins * prev_ins = NULL;
  t_ins * prev_prev_ins = NULL;
  t_ins * next_ins = NULL;

  if (global_options.blockprofilefile)
    CfgComputeHotBblThreshold(cfg,0.30);

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (!BBL_FUN(bbl)) continue;
    restart:
      BBL_FOREACH_INS_R(bbl,ins)
	{
	  if (ArmIsControlflow(ins)) continue;
	  if (RegsetIn(INS_REGS_USE(ins),ARM_REG_R15)) continue;
	  if (INS_TYPE(ins)==IT_DATA) continue;
	  if (!(prev_ins=INS_PREV(ins))) continue;
	  if (RegsetIn(INS_REGS_USE(prev_ins),ARM_REG_R15)) continue;
	  if (INS_TYPE(prev_ins)==IT_DATA) continue;
	  if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(prev_ins),INS_REGS_USE(ins)))) continue;
	  if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(prev_ins),INS_REGS_DEF(ins)))) continue;
	  if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(prev_ins),INS_REGS_DEF(ins)))) continue;
	  
	  if ((InsIsLoad(ins) && InsIsStore(prev_ins)) ||
	      (InsIsStore(ins) && InsIsStore(prev_ins)) ||
	      (InsIsStore(ins) && InsIsLoad(prev_ins))
	      )
	    {
	      if ((InsIsSimpleLoad(ins) || InsIsSimpleStore(ins)) &&
		  (InsIsSimpleLoad(prev_ins) || InsIsSimpleStore(prev_ins)) &&
		  NoAliasing(ins,prev_ins)
		  )
		{
		  prev_prev_ins = PreviousInsFT(prev_ins);
		  next_ins = NextInsFT(ins);
		  if (ComputeNrDep(prev_prev_ins,prev_ins,ins,next_ins)>ComputeNrDep(prev_prev_ins,ins,prev_ins,next_ins))
		    {
#ifdef PRINT_LAYOUT
		      if (BblIsHot(bbl))
			DiabloPrint(stdout,"BEFORE @iB\n",bbl);
#endif
		      BblMoveInstructionBefore(ins,prev_ins);
#ifdef PRINT_LAYOUT
		      if (BblIsHot(bbl))
			DiabloPrint(stdout,"AFTER @iB\n",bbl);
#endif
		      goto restart;
		    }
		}
	      else
		continue;
	    }

	  prev_prev_ins = PreviousInsFT(prev_ins);
	  next_ins = NextInsFT(ins);
	  if (ComputeNrDep(prev_prev_ins,prev_ins,ins,next_ins)>ComputeNrDep(prev_prev_ins,ins,prev_ins,next_ins))
	    {
#ifdef PRINT_LAYOUT
	      if (BblIsHot(bbl))
		DiabloPrint(stdout,"BEFORE @iB\n",bbl);
#endif
	      BblMoveInstructionBefore(ins,prev_ins);
#ifdef PRINT_LAYOUT
	      if (BblIsHot(bbl))
		DiabloPrint(stdout,"AFTER @iB\n",bbl);
#endif
	      goto restart;
	    }
	}
    }
}


void ArmReorderChains(t_cfg * cfg,t_bbl ** chains, int nchains)
{
  t_function * ifun;
  t_bbl * ibbl;
  t_ins * iins;

  if (global_options.blockprofilefile)
    CfgComputeHotBblThreshold(cfg,0.0001);

  DiabloPrint(stdout,"IN REORDERING\n");
  
  CFG_FOREACH_FUN(cfg,ifun)
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      if(BblIsHot(ibbl))
      {
	VERBOSE(0,("Reorder?"));
	BBL_FOREACH_INS(ibbl,iins)
	{
	  if(INS_REFERS_TO(iins))
	  {
	    if(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel)) == RT_SECTION)
	    {
	      if(SECTION_TYPE(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel)) == DATA_SECTION
		  || SECTION_TYPE(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel)) == RO_DATA_SECTION
		  || SECTION_TYPE(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel)) == BSS_SECTION )
		VERBOSE(0,("%c :Put me at the end of the chains @iB",SECTION_TYPE(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel)),INS_BBL(iins)));
	    }
	    else if(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel)) == RT_SUBSECTION)
	    {
	      if(SECTION_TYPE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel))) == DATA_SECTION
		  || SECTION_TYPE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel))) == RO_DATA_SECTION
		  || SECTION_TYPE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel))) == BSS_SECTION)
		VERBOSE(0,("%c :Put me at the end of the chains @iB",SECTION_TYPE(SECTION_PARENT_SECTION(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel))),INS_BBL(iins)));
	    }
	    else VERBOSE(0,("Hot ins refers to %d",RELOCATABLE_TYPE(REL_TO_RELOCATABLE(INS_REFERS_TO(iins)->rel))));
	  }
	}
      }
    }

}
/* vim: set shiftwidth=2 foldmethod=marker : */
