/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_opcodes.h"
#include "arm_disassemble_one.h"
/* }}} */
/*! \todo Document */
/* arm_opcode_table {{{ */
const arm_opcode arm_opcode_table[] = {

  /* Branches */
  {0x0f000000, 0x0a000000, "B", ArmDisassembleBranch},
  {0x0f000000, 0x0b000000, "BL", ArmDisassembleBranch},
  {0x0ffffff0, 0x012fff10, "BX", ArmDisassembleBranch},
  {0x0ffffff0, 0x012fff30, "BLX", ArmDisassembleBranch},

  /* Software Interrupt */
  {0x0f000000, 0x0f000000, "SWI", ArmDisassembleSWI},

  /* Multiplication */
  {0x0fe000f0, 0x00000090, "MUL", ArmDisassembleMultiplication},
  {0x0fe000f0, 0x00200090, "MLA", ArmDisassembleMultiplication},
  {0x0fe000f0, 0x00800090, "UMULL", ArmDisassembleMultiplication},
  {0x0fe000f0, 0x00a00090, "UMLAL", ArmDisassembleMultiplication},
  {0x0fe000f0, 0x00c00090, "SMULL", ArmDisassembleMultiplication},
  {0x0fe000f0, 0x00e00090, "SMLAL", ArmDisassembleMultiplication},

   /* Swap memory and register instructions */
  {0x0ff00ff0, 0x01000090, "SWP", ArmDisassembleSwap},
  {0x0ff00ff0, 0x01400090, "SWPB", ArmDisassembleSwap},

  /* Status register to general register Transfer instructions */
  {0x0fbf0fff, 0x010f0000, "MRS", ArmDisassembleMRS},

  /* General register to status register Transfer instructions */
  {0x0db0f000, 0x0120f000, "MSR", ArmDisassembleMSR},

  /* Half-word and signed byte data Transfer instructions,
   * mind that the order is important: otherwise, some of them are recognized as 
   * dataprocessing instructions*/

  {0x0e1000f0, 0x000000b0, "STRH", ArmDisassembleDataTransfer},
  {0x0e1000f0, 0x001000b0, "LDRH", ArmDisassembleDataTransfer},
/*    {0x0e1000f0, 0x000000f0, "STRSH", ArmDisassembleDataTransfer}, */
  {0x0e1000f0, 0x001000f0, "LDRSH", ArmDisassembleDataTransfer},
/*    {0x0e1000f0, 0x000000d0, "STRSB", ArmDisassembleDataTransfer}, */
  {0x0e1000f0, 0x001000d0, "LDRSB", ArmDisassembleDataTransfer},


  /* Data Manipulation */
  {0x0de00000, 0x00000000, "AND", ArmDisassembleDataProcessing},
  {0x0de00000, 0x00200000, "EOR", ArmDisassembleDataProcessing},
  {0x0de00000, 0x00400000, "SUB", ArmDisassembleDataProcessing},
  {0x0de00000, 0x00600000, "RSB", ArmDisassembleDataProcessing},
  {0x0de00000, 0x00800000, "ADD", ArmDisassembleDataProcessing},
  {0x0de00000, 0x00a00000, "ADC", ArmDisassembleDataProcessing},
  {0x0de00000, 0x00c00000, "SBC", ArmDisassembleDataProcessing},
  {0x0de00000, 0x00e00000, "RSC", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01000000, "TST", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01200000, "TEQ", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01400000, "CMP", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01600000, "CMN", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01800000, "ORR", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01a00000, "MOV", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01c00000, "BIC", ArmDisassembleDataProcessing},
  {0x0de00000, 0x01e00000, "MVN", ArmDisassembleDataProcessing},

  /* Single word and unsigned byte data Transfer instructions */
  {0x0c500000, 0x04000000, "STR", ArmDisassembleDataTransfer},
  {0X0c500000, 0x04100000, "LDR", ArmDisassembleDataTransfer},
  {0x0c500000, 0x04400000, "STRB", ArmDisassembleDataTransfer},
  {0X0c500000, 0x04500000, "LDRB", ArmDisassembleDataTransfer},

  /* Multiple register Transfer instructions */
  {0x0e100000, 0x08000000, "STM", ArmDisassembleMultipleTransfer},
  {0x0e100000, 0x08100000, "LDM", ArmDisassembleMultipleTransfer},

  /* Coprocessor instructions - overgenomen uit de GNU binutils */

  /* Floating point coprocessor datatransfer: see ARM DDI 0077B = ARM7500FE Data Sheet */
  {0x0ff08f10, 0x0e000100, "ADF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e100100, "MUF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e200100, "SUF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e300100, "RSF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e400100, "DVF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e500100, "RDF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e600100, "POW", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e700100, "RPW", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e800100, "RMF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e900100, "FML", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0ea00100, "FDV", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0eb00100, "FRD", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0ec00100, "POL", ArmDisassembleCPDO},

  {0x0ff08f10, 0x0e008100, "MVF", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e108100, "MNF", ArmDisassembleCPDO},

  {0x0ff08f10, 0x0e208100, "ABS", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e308100, "RND", ArmDisassembleCPDO},

  {0x0ff08f10, 0x0e408100, "SQT", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0e508100, "LOG", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0e608100, "LGN", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0e708100, "EXP", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0e808100, "SIN", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0e908100, "COS", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0ea08100, "TAN", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0eb08100, "ASN", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0ec08100, "ACS", ArmDisassembleUnsupported},
  {0x0ff08f10, 0x0ed08100, "ATN", ArmDisassembleUnsupported},

  {0x0ff08f10, 0x0ee08100, "URD", ArmDisassembleCPDO},
  {0x0ff08f10, 0x0ef08100, "NRM", ArmDisassembleCPDO},
  {0x0ff00f1f, 0x0e000110, "FLT", ArmDisassembleCPRT},
  {0x0fff0f98, 0x0e100110, "FIX", ArmDisassembleCPRT},
  {0x0fff0fff, 0x0e200110, "WFS", ArmDisassembleCPRT},
  {0x0fff0fff, 0x0e300110, "RFS", ArmDisassembleCPRT},
  {0x0fff0fff, 0x0e400110, "WFC", ArmDisassembleCPRT},
  {0x0fff0fff, 0x0e500110, "RFC", ArmDisassembleCPRT},

  {0x0ff8fff0, 0x0e90f110, "CMF", ArmDisassembleCPRT},
  {0x0ff8fff0, 0x0eb0f110, "CNF", ArmDisassembleCPRT},
  {0x0ff8fff0, 0x0ed0f110, "CMFE", ArmDisassembleCPRT},
  {0x0ff8fff0, 0x0ef0f110, "CNFE", ArmDisassembleCPRT},

  /* Floating point coprocessor datatransfer */
  {0x0e100f00, 0x0c000100, "STF", ArmDisassembleCPDT}, /* Store float (p. 10 - 2) */
  {0x0e100f00, 0x0c100100, "LDF", ArmDisassembleCPDT}, /* Load float (p. 10 - 2) */
  {0x0e100f00, 0x0c000200, "SFM", ArmDisassembleCPDT}, /* Store float multiple (p. 10 - 4) */
  {0x0e100f00, 0x0c100200, "LFM", ArmDisassembleCPDT}, /* Load float multiple (p. 10 -4) */
  
  /* Generic coprocessor instructions */
  {0x0f000010, 0x0e000000, "CDP", ArmDisassembleCPDO}, /* FP CP Data operations (p. 10-7) */
  {0x0f100010, 0x0e100010, "MRC", ArmDisassembleUnsupported}, /* */ 
  {0x0f100010, 0x0e000010, "MCR", ArmDisassembleUnsupported}, /* */
  {0x0e100000, 0x0c000000, "STC", ArmDisassembleUnsupported}, /* Store floating Point (p. 10 - 2) */
  {0x0e100000, 0x0c100000, "LDC", ArmDisassembleUnsupported}, /* Load floating Point (p. 10 -2 ) */

  /* opvang voor niet-herkende instructies */
  {0x00000000, 0x00000000, "UNDEF", ArmDisassembleUnsupported}
  
};
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
