/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/*!\addtogroup ARM_BACKEND */
/*@{ */
#include "../../config.h"
#include "../../simple_types.h"
#include "../../datastructures/data_string.h"
#include "../../kernel/diablo_ins.h"

#ifndef ARM_OPCODES_TYPEDEFS
#define ARM_OPCODES_TYPEDEFS
/*! \todo Document */
typedef enum _t_arm_condition_code 
{
  /*! \todo Document */
  ARM_CONDITION_EQ=0,
  /*! \todo Document */
  ARM_CONDITION_NE=1,
  /*! \todo Document */
  ARM_CONDITION_CS=2,
  /*! \todo Document */
  ARM_CONDITION_CC=3,
  /*! \todo Document */
  ARM_CONDITION_MI=4,
  /*! \todo Document */
  ARM_CONDITION_PL=5,
  /*! \todo Document */
  ARM_CONDITION_VS=6,
  /*! \todo Document */
  ARM_CONDITION_VC=7,
  /*! \todo Document */
  ARM_CONDITION_HI=8,
  /*! \todo Document */
  ARM_CONDITION_LS=9,
  /*! \todo Document */
  ARM_CONDITION_GE=10,
  /*! \todo Document */
  ARM_CONDITION_LT=11,
  /*! \todo Document */
  ARM_CONDITION_GT=12,
  /*! \todo Document */
  ARM_CONDITION_LE=13,
  /*! \todo Document */
  ARM_CONDITION_AL=14,
  /*! \todo Document */
  ARM_CONDITION_NV=15
} t_arm_condition_code;

/*! \todo Document */
typedef enum _t_arm_shift_type
{
  /*! \todo Document */
  ARM_SHIFT_TYPE_NONE=255,
  /*! \todo Document */
  ARM_SHIFT_TYPE_LSL_IMM=0,
  /*! \todo Document */
  ARM_SHIFT_TYPE_LSR_IMM=1,
  /*! \todo Document */
  ARM_SHIFT_TYPE_ASR_IMM=2,
  /*! \todo Document */
  ARM_SHIFT_TYPE_ROR_IMM=3,
  /*! \todo Document */
  ARM_SHIFT_TYPE_LSL_REG=4,
  /*! \todo Document */
  ARM_SHIFT_TYPE_LSR_REG=5,
  /*! \todo Document */
  ARM_SHIFT_TYPE_ASR_REG=6,
  /*! \todo Document */
  ARM_SHIFT_TYPE_ROR_REG=7,
  /*! \todo Document */
  ARM_SHIFT_TYPE_RRX=8	/*speciaal geval, geen shiftlengte nodig*/
} t_arm_shift_type;
#endif

#ifndef ARM_OPCODES_H
#define ARM_OPCODES_H
/*! The structure of the different disassemble functions */
typedef void (*ArmDisassembleFunction) (t_ins *, t_uint32 instr, t_uint16 opc);
/*! The opcode structure, used in the opcode table to specify what disassembly
 * function corresponds to what bitpattern */
typedef struct _arm_opcode
{
  /*! \todo Document */
  t_uint32 mask;
  /*! \todo Document */
  t_uint32 opcode;
  /*! \todo Document */
  t_string desc;
  /*! \todo Document */
  ArmDisassembleFunction Disassemble;
} arm_opcode;
/*! The opcode table */
extern const arm_opcode arm_opcode_table[];
/* The opcodes {{{ */
typedef enum _t_arm_opcode
{
  /*! An Arm Branch instruction */
  ARM_B,
  /*! An Arm Branch And Link instruction */
  ARM_BL,
  /*! An Arm Branch And Exchange instruction */
  ARM_BX,
  /*! An Arm Branch, Link And Exchange instruction */
  ARM_BLX,
  /*! An Arm Software Interrupt */
  ARM_SWI,
  /*! An Arm Multiply instruction */
  ARM_MUL,
  /*! An Arm Multiply and accumulate instruction */
  ARM_MLA,
  /*! An Unsigned Arm Multiply Long */
  ARM_UMULL,
  /*! An Unsigned Arm Multiply and Accumulate Long */
  ARM_UMLAL,
  ARM_SMULL,
  ARM_SMLAL,
  /*! An Arm Swap instruction */
  ARM_SWP,
  ARM_SWPB,
  ARM_MRS,
  ARM_MSR,
  ARM_STRH,
  ARM_LDRH,
  ARM_LDRSH,
  ARM_LDRSB,
  ARM_AND,
  ARM_EOR,
  ARM_SUB,
  ARM_RSB,
  ARM_ADD,
  ARM_ADC,
  ARM_SBC,
  ARM_RSC,
  ARM_TST,
  ARM_TEQ,
  ARM_CMP,
  ARM_CMN,
  ARM_ORR,
  ARM_MOV,
  ARM_BIC,
  ARM_MVN,
  ARM_STR,
  ARM_LDR,
  ARM_STRB,
  ARM_LDRB,
  ARM_STM,
  ARM_LDM,
  ARM_ADF,
  ARM_MUF,
  ARM_SUF,
  ARM_RSF,
  ARM_DVF,
  ARM_RDF,
  ARM_POW,
  ARM_RPW,
  ARM_RMF,
  ARM_FML,
  ARM_FDV,
  ARM_FRD,
  ARM_POL,
  ARM_MVF,
  ARM_MNF,
  ARM_ABS,
  ARM_RND,
  ARM_SQT,
  ARM_LOG,
  ARM_LGN,
  ARM_EXP,
  ARM_SIN,
  ARM_COS,
  ARM_TAN,
  ARM_ASN,
  ARM_ACS,
  ARM_ATN,
  ARM_URD,
  ARM_NRM,
  ARM_FLT,
  ARM_FIX,
  ARM_WFS,
  ARM_RFS,
  ARM_WFC,
  ARM_RFC,
  ARM_CMF,
  ARM_CNF,
  ARM_CMFE,
  ARM_CNFE,
  ARM_STF,
  ARM_LDF,
  ARM_SFM,
  ARM_LFM,
  ARM_CDP,
  ARM_MRC,
  ARM_MCR,
  ARM_STC,
  ARM_LDC,
  /*! Pseudo op, used in the opcode table */
  ARM_UNDEF,
  /*! Pseudo instruction, used to represent data in basic blocks (=data blocks,
   * for data in the code section)*/
  ARM_DATA		,
  /*! Pseudo instruction that produces an address (=integer value+ relocation) */
  ARM_ADDRESS_PRODUCER,
  /*! Pseudo instruction that produces an constant(=integer value) */
  ARM_CONSTANT_PRODUCER,
  /*! Pseudo instruction that produces a floating point value in one of the
   * possible encodings */
  ARM_FLOAT_PRODUCER,
  /*! Pseudo instruction, used to add calls during instrumentation */
  ARM_PSEUDO_CALL
} t_arm_opcode;
/*
#define TH_BEQ                     0
#define TH_BNE                     1
#define TH_BCS                     2
#define TH_BCC                     3
#define TH_BMI                     4
#define TH_BPL                     5
#define TH_BVS                     6
#define TH_BVC                     7
#define TH_BHI                     8
#define TH_BLS                     9
#define TH_BGE                    10
#define TH_BLT                    11
#define TH_BGT                    12
#define TH_BLE                    13
#define TH_BAL                    14
#define TH_B                      15
#define TH_BL                     16
#define TH_BX                     17
#define TH_SWI                    18
#define TH_SUB_3R                 19
#define TH_ADD_3R                 20
#define TH_SUB_3I                 21
#define TH_ADD_3I                 22
#define TH_MOV_I                  23
#define TH_CMP_I                  24
#define TH_ADD_2I                 25
#define TH_SUB_2I                 26
#define TH_LSL_I                  27
#define TH_LSR_I                  28
#define TH_ASR_I                  29
#define TH_AND                    30
#define TH_EOR                    31
#define TH_LSL_R                  32
#define TH_LSR_R                  33
#define TH_ASR_R                  34
#define TH_ADC                    35
#define TH_SBC                    36
#define TH_ROR                    37
#define TH_TST                    38
#define TH_NEG                    39
#define TH_CMP_R                  40
#define TH_CMN                    41
#define TH_ORR                    42
#define TH_MUL                    43
#define TH_BIC                    44
#define TH_MVN                    45
#define TH_ADD_2RH                46
#define TH_CMP_RH                 47
#define TH_MOV_RH                 48
#define TH_ADD_PC                 49
#define TH_ADD_SP                 50
#define TH_ADD_SPI                51
#define TH_SUB_SPI                52
#define TH_STR_I                  53
#define TH_LDR_I                  54
#define TH_STRB_I                 55
#define TH_LDRB_I                 56
#define TH_STRH_I                 57
#define TH_LDRH_I                 58
#define TH_STRH_R                 59
#define TH_LDRH_R                 60
#define TH_LDRSB                  61
#define TH_LDRSH                  62
#define TH_STR_R                  63
#define TH_LDR_R                  64
#define TH_STRB_R                 65
#define TH_LDRB_R                 66
#define TH_LDR_PC                 67
#define TH_STR_SP                 68
#define TH_LDR_SP                 69
#define TH_STMIA                  70
#define TH_LDMIA                  71
#define TH_PUSH                   72
#define TH_POP                    73
#define TH_UNDEF                  74
*/
/* }}} */
/* Condition aliases {{{ */
#define ARM_CONDITION_HS		ARM_CONDITION_CS
#define ARM_CONDITION_LO		ARM_CONDITION_CC
/* }}} */
#endif
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
