#include "arm_peephole.h"
#include "../../kernel/diablo_iterators.h"
#include "arm_instructions.h"
#include "arm_registers.h"
#include "arm_utils.h"
#include "../../utils/util_status.h"
#include "../../utils/util_error.h"


t_uint32 ArmPeepholeCombineIdenticalConditionalIns(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins, *jins, *tmp;
  t_uint32 killers = 0;

  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
    if ((jins=INS_NEXT(ins)) && ArmInstructionIsConditional(ins) && ArmCompareInstructionsOppositeCond(ins,jins) && !ArmInsUpdatesCond(ins))
      {
	/*	DiabloPrint(stdout,"optimizing\n           @I\n           @I\n",ins,jins);*/
	INS_EXEC_COUNT(jins)+=INS_EXEC_COUNT(ins);
	InsKill(ins);
	killers++;
	ArmInsUnconditionalize(jins);
	INS_REGS_USE(jins) = ArmUsedRegisters(jins);
	INS_REGS_DEF(jins) = ArmDefinedRegisters(jins);
	/*	DiabloPrint(stdout,"to         @I\n",jins);*/
      }
  return killers;
}

t_uint32 ArmPeepholeCombineAdds(t_cfg * cfg)
{
  t_function * fun;
  t_bbl * bbl;
  t_ins * i_ins, *j_ins;
  t_uint32 killers = 0;

  CFG_FOREACH_FUN(cfg,fun)
    FUN_FOREACH_BBL(fun,bbl)
      BBL_FOREACH_INS(bbl,i_ins)
      {
	if(ARMINSOPCODE(i_ins) == ARM_ADD && ARMINSREGA(i_ins) == ARMINSREGB(i_ins))
	{
	  /*if(INS_PREV(i_ins) &&
	      (ARMINSOPCODE(INS_PREV(i_ins)) == ARM_ADD || ARMINSOPCODE(INS_PREV(i_ins)) == ARM_MOV)
	      && ARMINSREGA(i_ins) == ARMINSREGA(INS_PREV(i_ins))
	    )
	  {
	  DiabloPrint(stdout,"Yep: @I, @I\nARMREGC = %d\n",INS_PREV(i_ins),i_ins,ARMINSREGC(i_ins));
	  }*/
	  if(/*ARMINSREGC(i_ins) == ARM_REG_NONE */ (ARMINSFLAGS(i_ins)&FL_IMMED) && ARMINSSHIFTTYPE(i_ins) == ARM_SHIFT_TYPE_NONE)
	  {
	    if(INS_PREV(i_ins) && ARMINSOPCODE(INS_PREV(i_ins)) == ARM_ADD && ARMINSREGA(i_ins) == ARMINSREGA(INS_PREV(i_ins)) && ARMINSREGC(INS_PREV(i_ins)) == ARM_REG_NONE && ARMINSSHIFTTYPE(INS_PREV(i_ins)) == ARM_SHIFT_TYPE_NONE && ARMINSCONDITION(i_ins) == ARMINSCONDITION(INS_PREV(i_ins)))
	    {
	      t_uint32 sum = ARMINSIMMEDIATE(i_ins) + ARMINSIMMEDIATE(INS_PREV(i_ins));
	      if(ArmIsEncodableConstantForOpcode(sum,ARM_ADD))
	      {
		j_ins = INS_PREV(i_ins);
		/*DiabloPrint(stdout,"ArmPeepholeCombineAdds: @I and @I\n",i_ins,j_ins);*/
		ARMINSIMMEDIATE(j_ins) = sum;
		InsKill(i_ins);
		i_ins = j_ins;
		killers++;
		INS_REGS_USE(j_ins) = ArmUsedRegisters(j_ins);
		INS_REGS_DEF(j_ins) = ArmDefinedRegisters(j_ins);
		/*DiabloPrint(stdout,"after: @I\n",j_ins);*/
	      }
	    }
	  }
	  /*DiabloPrint(stdout,"Na : @I\n\n",i_ins);*/
	}
      }
  if(killers)
    VERBOSE(0,("ArmPeepholeCombineAdds killed %d",killers));
  return killers;
}


t_uint32 ArmPeepholeUselessAdds(t_cfg * cfg)
{
  t_function * fun;
  t_bbl * bbl;
  t_ins * i_ins;
  t_uint32 killers = 0;

  CFG_FOREACH_FUN(cfg,fun)
    FUN_FOREACH_BBL(fun,bbl)
      BBL_FOREACH_INS(bbl,i_ins)
      {
	if(ARMINSOPCODE(i_ins) == ARM_ADD)
	{
	  if((ARMINSFLAGS(i_ins)&FL_IMMED) /*ARMINSREGC(i_ins) == ARM_REG_NONE*/ && ARMINSSHIFTTYPE(i_ins) == ARM_SHIFT_TYPE_NONE && !(ARMINSFLAGS(i_ins)&FL_S) &&  ARMINSIMMEDIATE(i_ins) == 0)
	  {
	    /*DiabloPrint(stdout,"ArmPeepholeUselessAdds: @I\n",i_ins);*/
	    ArmInstructionMakeMov(i_ins,ARMINSREGA(i_ins),ARMINSREGB(i_ins),0,ARMINSCONDITION(i_ins));
	    INS_REGS_USE(i_ins) = ArmUsedRegisters(i_ins);
	    INS_REGS_DEF(i_ins) = ArmDefinedRegisters(i_ins);
	  }
	}
      }

  return killers;
}

t_uint32 ArmPeepholeUselessAdd_Sub(t_cfg * cfg)
{
  t_function * fun;
  t_bbl * bbl;
  t_ins * i_ins, *j_ins, *tmp;
  t_uint32 killers = 0;

  CFG_FOREACH_FUN(cfg,fun)
    FUN_FOREACH_BBL(fun,bbl)
      BBL_FOREACH_INS_SAFE(bbl,i_ins,tmp)
      {
	if(ARMINSOPCODE(i_ins) == ARM_ADD
	    && INS_NEXT(i_ins)
	    && ARMINSOPCODE(INS_NEXT(i_ins)) == ARM_SUB
	    && ARMINSREGA(i_ins) == ARMINSREGB(i_ins)
	    && (ARMINSFLAGS(i_ins) & FL_IMMED)
	    && !(ARMINSFLAGS(i_ins) & FL_S))
	{
	  j_ins = INS_NEXT(i_ins);
	  if(ARMINSREGA(j_ins) == ARMINSREGA(i_ins)
	      && ARMINSREGB(j_ins) == ARMINSREGA(i_ins)
	      && (ARMINSFLAGS(j_ins) & FL_IMMED)
	      && !(ARMINSFLAGS(j_ins) & FL_S)
	      && ARMINSIMMEDIATE(i_ins) == ARMINSIMMEDIATE(j_ins)
	      && ARMINSCONDITION(i_ins) == ARMINSCONDITION(j_ins))
	  {
	    tmp = INS_NEXT(INS_NEXT(i_ins));
	    InsKill(i_ins);
	    InsKill(j_ins);
	    killers +=2;
	  }
	}
      }
  if(killers)
    VERBOSE(0,("ArmPeepholeUselessAddSub killed %d",killers));
  return killers;
}

t_uint32 ArmPeepholeWriteBackAnomaly(t_cfg * cfg)
{
  t_function * fun;
  t_bbl * bbl;
  t_ins * i_ins, *j_ins;
  t_uint32 killers = 0;

  CFG_FOREACH_FUN(cfg,fun)
    FUN_FOREACH_BBL(fun,bbl)
      BBL_FOREACH_INS(bbl,i_ins)
      {
	if(ArmInsWriteBackHappens(i_ins))
	{
	  j_ins = INS_NEXT(i_ins);
	  while(j_ins)
	  {
	    if(ARMINSOPCODE(j_ins) == ARM_ADD
		&& ARMINSREGA(j_ins) == ARMINSREGB(j_ins)
		&& ARMINSREGA(j_ins) == ARMINSREGB(i_ins)
		&& !(ARMINSFLAGS(i_ins) & FL_DIRUP)
		&& !(ARMINSFLAGS(j_ins) & FL_S) 
		&& ARMINSCONDITION(j_ins) == ARMINSCONDITION(i_ins) )
	    {
	      if(ARMINSOPCODE(i_ins) == ARM_LDM || ARMINSOPCODE(i_ins) == ARM_STM)
	      {
		t_uint32 nr_regs;
		t_regset regs = NullRegs;
		
		RegsetAddMultipleRegs(regs,ARMINSIMMEDIATE(i_ins));
		nr_regs = RegsetCountRegs(regs) * 4;
		if(nr_regs == ARMINSIMMEDIATE(j_ins))
		{
		  InsKill(j_ins);
		  ARMINSFLAGS(i_ins) &= ~FL_WRITEBACK;
		  INS_REGS_USE(i_ins) = ArmUsedRegisters(i_ins);
		  INS_REGS_DEF(i_ins) = ArmDefinedRegisters(i_ins);
		  killers++;
		}
	      }
	      else if(ARMINSOPCODE(i_ins) == ARM_LDR || ARMINSOPCODE(i_ins) == ARM_STR)
	      {
		if(ARMINSIMMEDIATE(j_ins) == 4)
		{
		  InsKill(j_ins);
		  killers++;
		  if(!(ARMINSFLAGS(i_ins) & FL_PREINDEX))
		  {
		    ARMINSIMMEDIATE(i_ins) = 0;
		    ARMINSFLAGS(i_ins) |= FL_PREINDEX;
		  }
		  else ARMINSFLAGS(i_ins) &= ~FL_WRITEBACK;
		  INS_REGS_USE(i_ins) = ArmUsedRegisters(i_ins);
		  INS_REGS_DEF(i_ins) = ArmDefinedRegisters(i_ins);
		}
	      }
	      break;
	    }
	    if(ARMINSOPCODE(j_ins) == ARM_SUB
		&& ARMINSREGA(j_ins) == ARMINSREGB(j_ins)
		&& ARMINSREGA(j_ins) == ARMINSREGB(i_ins)
		&& (ARMINSFLAGS(i_ins) & FL_DIRUP)
		&& !(ARMINSFLAGS(j_ins) & FL_S) 
		&& ARMINSCONDITION(j_ins) == ARMINSCONDITION(i_ins) )
	    {
	      if(ARMINSOPCODE(i_ins) == ARM_LDM || ARMINSOPCODE(i_ins) == ARM_STM)
	      {
		t_uint32 nr_regs;
		t_regset regs = NullRegs;
		
		RegsetAddMultipleRegs(regs,ARMINSIMMEDIATE(i_ins));
		nr_regs = RegsetCountRegs(regs) * 4;
		if(nr_regs == ARMINSIMMEDIATE(j_ins))
		{
		  InsKill(j_ins);
		  ARMINSFLAGS(i_ins) &= ~FL_WRITEBACK;
		  INS_REGS_USE(i_ins) = ArmUsedRegisters(i_ins);
		  INS_REGS_DEF(i_ins) = ArmDefinedRegisters(i_ins);
		  killers++;
		}
	      }
	      else if(ARMINSOPCODE(i_ins) == ARM_LDR || ARMINSOPCODE(i_ins) == ARM_STR)
	      {
		if(ARMINSIMMEDIATE(j_ins) == 4)
		{
		  InsKill(j_ins);
		  killers++;
		  if(!(ARMINSFLAGS(i_ins) & FL_PREINDEX))
		  {
		    ARMINSIMMEDIATE(i_ins) = 0;
		    ARMINSFLAGS(i_ins) |= FL_PREINDEX;
		  }
		  else ARMINSFLAGS(i_ins) &= ~FL_WRITEBACK;
		  INS_REGS_USE(i_ins) = ArmUsedRegisters(i_ins);
		  INS_REGS_DEF(i_ins) = ArmDefinedRegisters(i_ins);
		}
	      }
	      break;
	    }
	    if(RegsetIn(INS_REGS_DEF(j_ins),ARMINSREGB(i_ins))) break;
	    if(RegsetIn(INS_REGS_USE(j_ins),ARMINSREGB(i_ins))) break;
	    j_ins = INS_NEXT(j_ins);
	  }
	}
      }
  
  if(killers)
    VERBOSE(0,("ArmPeepholeWritebackAnomaly killed %d",killers));
  return killers;
}

void ArmPeepholeOptimizations(t_cfg * cfg)
{
  t_uint32 killers = 0;
  killers += ArmPeepholeCombineAdds(cfg);
  killers += ArmPeepholeUselessAdds(cfg);
  killers += ArmPeepholeUselessAdd_Sub(cfg);
  killers += ArmPeepholeWriteBackAnomaly(cfg);
  killers += ArmPeepholeCombineIdenticalConditionalIns(cfg);
/*  return killers;*/
}


void SwitchMoves(t_cfg * cfg)
{ 
  t_bbl * bbl;
  t_ins * ins;
  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      t_reg src_reg;
      if (ARMINSOPCODE(ins)==ARM_MOV && (src_reg=ARMINSREGC(ins))!=ARM_REG_NONE && ARMINSSHIFTTYPE(ins)==ARM_SHIFT_TYPE_NONE && (!(ARMINSFLAGS(ins) & (FL_S))))
	{
	  t_ins * jins = INS_PREV(ins);

	  if (!jins)  continue;
	  if ((ARMINSFLAGS(jins) & (FL_S))) continue;
	  if (ARMINSOPCODE(jins)==ARM_MUL) continue;
	  if (ARMINSOPCODE(jins)==ARM_MLA) continue;
	  if (src_reg==ARM_REG_R15) continue;
	  if (jins && RegsetIn(INS_REGS_DEF(jins),src_reg) && ARMINSREGA(jins)==src_reg && ARMINSCONDITION(jins)==ARMINSCONDITION(ins))
	    {
	      t_reg dest_reg = ARMINSREGA(ins);
	      if (dest_reg==ARM_REG_R15) continue;
	      

	      /*	      DiabloPrint(stdout,"BEFORE @I @I\n",jins,ins);*/

	      
	      ARMINSREGA(jins)=ARMINSREGC(ins)=dest_reg;
	      ARMINSREGA(ins)=src_reg;
	      
	      INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	      INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	      INS_REGS_USE(jins) = ArmUsedRegisters(jins);
	      INS_REGS_DEF(jins) = ArmDefinedRegisters(jins);

	      /*DiabloPrint(stdout,"AFTER  @I @I\n",jins,ins);*/

	    }
	}
    }
}
/* vim: set shiftwidth=2 foldmethod=marker : */
