/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "arm_print.h"
#include "arm_opcodes.h"
#include "arm_registers.h"
#include "../../utils/util_error.h"
#include <string.h>
#include <ctype.h>
#include "thumb_print.h"
/* }}} */
/*! \todo Move me and document */
static char *Conditions [] = {"EQ","NE","CS","CC","MI","PL","VS","VC","HI","LS","GE","LT","GT", "LE","","NV"};
/*! \todo Found this definition in 3 different files... Keep one of them */ 
static char *Regs[] = {"r0","r1","r2","r3","r4","r5","r6","r7","r8","r9","r10","r11","r12","r13", "r14","pc","cpsr","spsr","q","c","v","z","n","f0","f1","f2","f3","f4","f5","f6","f7","fpsr"};
/*! \todo Move me and document */
static char *Shifts[] = {"LSL","LSR","ASR","ROR","LSL","LSR","ASR","ROR","RRX"};
/*!
 * \todo Document
 *
 * \param data
 * \param outputstring
 *
 * \return void 
*/
/* ArmInstructionPrint {{{ */
void ArmInstructionPrint(t_ins * data, t_string outputstring) 
{
  /* assume outputstring always has enough room - say at least 80 characters (this is way too much but it's safe */
  t_arm_ins * instruction=(t_arm_ins *) data;

  char opcode[10]="";
  char oper1[10]="", oper2[10]="\0\0\0\0\0\0\0\0\0\0", oper3[30]="\0", shift[6]="", shiftarg[10]="";
  char operands[60]="";

  if (ARMINSFLAGS(data) & FL_THUMB) {
    ThumbInstructionPrint(data, outputstring);
    return;
  }
  /* VERY SPECIAL CASE: data in the code */
  if (ARMINSOPCODE(instruction) == ARM_DATA) {
    sprintf(opcode,"DATA");
    sprintf(operands, "|%08x|", ARMINSIMMEDIATE(instruction));
    if (isprint(((char *) &ARMINSIMMEDIATE(instruction))[0])) sprintf(operands+10, " %c  ", ((char *) &ARMINSIMMEDIATE(instruction))[0]);
    else sprintf(operands+10,"\\%2x ",((unsigned char *) &ARMINSIMMEDIATE(instruction))[0]);

    if (isprint(((char *) &ARMINSIMMEDIATE(instruction))[1])) sprintf(operands+14, " %c  ", ((char *) &ARMINSIMMEDIATE(instruction))[1]);
    else sprintf(operands+14,"\\%2x ",((unsigned char *) &ARMINSIMMEDIATE(instruction))[1]); 

    if (isprint(((char *) &ARMINSIMMEDIATE(instruction))[2])) sprintf(operands+18, " %c  ", ((char *) &ARMINSIMMEDIATE(instruction))[2]);
    else sprintf(operands+18,"\\%2x ",((unsigned char *) &ARMINSIMMEDIATE(instruction))[2]); 

    if (isprint(((char *) &ARMINSIMMEDIATE(instruction))[3])) sprintf(operands+22, " %c  ", ((char *) &ARMINSIMMEDIATE(instruction))[3]);
    else sprintf(operands+22,"\\%2.2x ",((unsigned char *) &ARMINSIMMEDIATE(instruction))[3]); 

    sprintf(operands+26,"|");
  }
  else if (ARMINSOPCODE(instruction) == ARM_PSEUDO_CALL)
  {
      sprintf(opcode,"CALL%s %s",Conditions[ARMINSCONDITION(instruction)],((t_function *) instruction->data)->name);
  }
  else if (ARMINSOPCODE(instruction) == ARM_ADDRESS_PRODUCER)
    {
      sprintf(opcode,"ADR%s",Conditions[ARMINSCONDITION(instruction)]);
      sprintf(operands,"r%d %x", ARMINSREGA(instruction), ARMINSIMMEDIATE(instruction));

    } 
  else if (ARMINSOPCODE(instruction) == ARM_CONSTANT_PRODUCER)
    {
      sprintf(opcode,"CONST%s",Conditions[ARMINSCONDITION(instruction)]);
      sprintf(operands,"r%d %x (%d)", ARMINSREGA(instruction), ARMINSIMMEDIATE(instruction),ARMINSIMMEDIATE(instruction));
    }
  else if (ARMINSOPCODE(instruction) == ARM_FLOAT_PRODUCER)
    {
      sprintf(opcode,"FLOAT%s",Conditions[ARMINSCONDITION(instruction)]);
      sprintf(operands,"f%d", ARMINSREGA(instruction) -ARM_REG_F0);
    }
  else
    {
      /* format opcode + condition code + set condition codes flag */
      sprintf(opcode,"%s%s%s%c",arm_opcode_table[ARMINSOPCODE(instruction)].desc,Conditions[ARMINSCONDITION(instruction)],(ARMINSFLAGS(instruction)&FL_FLT_DOUBLE)?"D":((ARMINSFLAGS(instruction)&FL_FLT_DOUBLE_EXTENDED)?"DE":""),
	      (ARMINSFLAGS(instruction) & FL_S) ? 'S' : ' ');

      if ((ARMINSOPCODE(instruction) == ARM_CMP) || (ARMINSOPCODE(instruction) == ARM_CMN) ||
	  (ARMINSOPCODE(instruction) == ARM_TST)) 
	/* remove trailing S from the opcode (not usually displayed here) */
	opcode[strlen(opcode) - 1] = '\0';


      /* format operand 1 */
      if (ARMINSREGA(instruction) != ARM_REG_NONE) 
      {
	sprintf(oper1,"%s,",Regs[ARMINSREGA(instruction)]);
      }
      /* format operand 2 */
      if (ARMINSREGB(instruction) != ARM_REG_NONE) 
	sprintf(oper2,"%s,",Regs[ARMINSREGB(instruction)]);
      /* format operand 3 and shift and shiftarg*/
      if ((ARMINSFLAGS(instruction) & FL_IMMED) || ARMINSIMMEDIATE(instruction)) {
	sprintf(oper3,"%s%#x",(INS_TYPE(instruction) == IT_BRANCH) ? "" : "#",ARMINSIMMEDIATE(instruction));
	/*	      sprintf(shift, "");
		      sprintf(shiftarg, "");
	*/
	if (INS_TYPE(instruction) == IT_BRANCH && INS_BBL(instruction)) 
	  sprintf(oper3+strlen(oper3)," abs: %x",ARMINSIMMEDIATE(instruction)+G_T_UINT32(BBL_OLD_ADDRESS(INS_BBL(instruction)))+INS_BBL(instruction)->nins*4+4);
      }
      else {
	sprintf(oper3,"%s", (ARMINSREGC(instruction) == ARM_REG_NONE) ? "" : Regs[ARMINSREGC(instruction)]);
	/*	      sprintf(shift, "");
		      sprintf(shiftarg, "");
	*/	      
	if (ARMINSSHIFTTYPE(instruction) != ARM_SHIFT_TYPE_NONE) {
	  sprintf(shift, ",%s ", Shifts[ARMINSSHIFTTYPE(instruction)]);
	  if (ARMINSSHIFTTYPE(instruction) & 0x4) 
	    sprintf(shiftarg, "%s",Regs[ARMINSREGS(instruction)]);
	  else
	    sprintf(shiftarg, "#%d",ARMINSSHIFTLENGTH(instruction));
	}
	if(ARMINSOPCODE(instruction) == ARM_MLA)
	  sprintf(oper3+strlen(oper3),",%s",Regs[ARMINSREGS(instruction)]);
      };

      /* SPECIAL CASE: MRS and MSR instructions use different registers */ 
      if (ARMINSOPCODE(instruction) == ARM_MRS)
	{
	  if (ARMINSFLAGS(instruction) & FL_SPSR)
	    sprintf(oper2,"SPSR");
	  else
	    sprintf(oper2,"CPSR");
	}

      if (ARMINSOPCODE(instruction) == ARM_MSR)
	sprintf(oper1,"%cPSR_%s%s%s%s,", (ARMINSFLAGS(instruction) & FL_SPSR) ? 'S' : 'C',
		(ARMINSFLAGS(instruction) & (1U << 28)) ? "c" : "",(ARMINSFLAGS(instruction) & (1U << 29)) ? "x" : "",
		(ARMINSFLAGS(instruction) & (1U << 30U)) ? "s" : "",(ARMINSFLAGS(instruction) & (1U << 31U)) ? "f" : "");


      sprintf(operands,"%s%s%s%s%s",oper1,oper2,oper3,shift,shiftarg);

      /* special case: load and store instructions -> modify operands according to flags */
      if ((INS_TYPE(instruction) == IT_LOAD) || (INS_TYPE(instruction) == IT_STORE)) {
	char oper3temp[20];

	if (!(ARMINSFLAGS(instruction) & FL_DIRUP)) {
	  /* put a minus in front of the third argument */
	  if (ARMINSFLAGS(instruction) & FL_IMMED) 
	    sprintf(oper3temp,"#-%s",oper3+1);
	  else 
	    sprintf(oper3temp,"-%s",oper3);
	  sprintf(oper3,oper3temp);
	};
	/* strip comma from the end of oper */
	oper2[strlen(oper2)-1] = '\0';
	sprintf(operands,"%s[%s%s%s%s%s%s%c",oper1,oper2,(ARMINSFLAGS(instruction) & FL_PREINDEX) ? "," : "],",
		oper3,shift,shiftarg,(ARMINSFLAGS(instruction) & FL_PREINDEX) ? "]" : "",
		(ARMINSFLAGS(instruction) & FL_WRITEBACK) ? '!' : ' ');
      };

      /* SPECIAL CASE: load and store multiple */
      if ((INS_TYPE(instruction) == IT_LOAD_MULTIPLE) || (INS_TYPE(instruction) == IT_STORE_MULTIPLE))
      {
	/* add a suffix to the opcode, indicating the addressing mode */
	char suffix[3];
	char reglist[65] = "{";
	int i;
	t_uint32 fl = ARMINSFLAGS(instruction);

	suffix[0] = (fl & FL_DIRUP) ? 'I' : 'D';
	suffix[1] = (fl & FL_PREINDEX) ? 'B' : 'A';
	suffix[2] = '\0';
	sprintf(opcode,"%s%s%s%c",arm_opcode_table[ARMINSOPCODE(instruction)].desc,Conditions[ARMINSCONDITION(instruction)],suffix,
		(fl & FL_S) ? 'S' : ' ');

	/* add a ! to operand 2 if writeback is enabled */
	sprintf(oper2,"%s%s",Regs[ARMINSREGB(instruction)],(fl & FL_WRITEBACK) ? "!" : "");

	/* print out register list in a human-readable form */
	for (i = 0; i < 16; i++)
	  if (ARMINSIMMEDIATE(instruction) & (1 << i)) {
	    strcat(reglist,Regs[i]);
	    strcat(reglist,",");
	  };
	reglist[strlen(reglist) - 1] = '}';

	sprintf(operands,"%s,%s",oper2,reglist);

      }
      if ((ARMINSOPCODE(instruction) == ARM_LFM) || (ARMINSOPCODE(instruction) == ARM_SFM))
      {
	char reglist[65] = "{";
	int i;
	t_uint32 fl = ARMINSFLAGS(instruction);
	sprintf(oper2,"%s%s",Regs[ARMINSREGB(instruction)],(fl & FL_WRITEBACK) ? "!" : "");
	for (i = ARM_REG_F0; i < ARM_REG_F7+1; i++)
	  if (RegsetIn(ARMINSMULTIPLE(instruction),i)) {
	    strcat(reglist,Regs[i]);
	    strcat(reglist,",");
	  };
	reglist[strlen(reglist) - 1] = '}';

	if (ARMINSIMMEDIATE(instruction))
	{
	  sprintf(operands,"[%s %c %x],%s",oper2,(ARMINSFLAGS(instruction) & FL_DIRUP)?'+':'-',ARMINSIMMEDIATE(instruction),reglist);
	}
	else
	  sprintf(operands,"%s,%s",oper2,reglist);

      }
    }

  sprintf(outputstring,"%-10s %-69s",opcode,operands);
  StringTrim(outputstring);
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
