/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../../kernel/diablo_function.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_print.h"
#include "../../utils/util_error.h"
#include "../../diablo_options.h"
#include "arm_opcodes.h"
#include "arm_instructions.h"
#include "arm_utils.h"
/*}}} */
/** This function computes the registers that a procedure saved on the stack */
/* XXX: in fact this is unsafe! (hand-written assembly code can easily change
 * the saved registers on the stack) */

extern void ArmFunComputeStackSavedRegisters(t_function * ifun)
{
  t_bbl*  i_bbl;
  t_ins* iins;
  t_int32 count;
  t_int32 SavedRegOffset[64];
  t_regset use;
  t_regset def;
  t_regset saved;
  t_reg reg;
  t_uint32 offset = 0;
  t_cfg_edge * i_edge;


  FUN_REGS_SAVED(ifun)=0;
  goto end;
  
#if MAX_NR_REG_SUBSETS > 4
  FATAL(("Immediate regs must be real regsets!"));
#else
  REGSET_FOREACH_REG(AllRegs(),reg)
    {
      SavedRegOffset[reg] = -1;
    }
#endif  
  i_bbl = FUN_BBL_FIRST(ifun);
  saved = RegsetNew();
  use  = RegsetNew();
  def  = RegsetNew();
    
  /* do the stores as a first approximation - we look in the very first bbl of the function*/

  BBL_FOREACH_INS(i_bbl,iins)
    {
      if(ARMINSOPCODE(iins) == ARM_STF && ARMINSREGB(iins) == ARM_REG_R13 && ArmInsWriteBackHappens(iins))
	{
	  offset -= ARMINSIMMEDIATE(iins);
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
	}
      else if((ARMINSOPCODE(iins) == ARM_LDR || ARMINSOPCODE(iins) == ARM_LDF) && ARMINSREGB(iins) == ARM_REG_R13 && ArmInsWriteBackHappens(iins))
	{
	  offset += ARMINSIMMEDIATE(iins);
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
	}
      else if(ARMINSOPCODE(iins) == ARM_SUB && ARMINSREGA(iins) == ARM_REG_R13 && ARMINSREGB(iins) == ARM_REG_R13 && ARMINSREGC(iins) == ARM_REG_NONE) 
	{
	  offset -= ARMINSIMMEDIATE(iins);
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
	}
      else if(ARMINSOPCODE(iins) == ARM_ADD && ARMINSREGA(iins) == ARM_REG_R13 && ARMINSREGB(iins) == ARM_REG_R13 && ARMINSREGC(iins) == ARM_REG_NONE) 
	{
	  offset += ARMINSIMMEDIATE(iins);
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
	}
      else if(ARMINSOPCODE(iins) == ARM_RSB && ARMINSREGA(iins) == ARM_REG_R13 && ARMINSREGC(iins) == ARM_REG_R13 && ARMINSREGB(iins) == ARM_REG_NONE) 
	{
	  offset -= ARMINSIMMEDIATE(iins);
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
	}
      else if ((ARMINSOPCODE(iins) == ARM_RSB || ARMINSOPCODE(iins) == ARM_SUB  || ARMINSOPCODE(iins) == ARM_LDR || ARMINSOPCODE(iins) == ARM_MOV) && ARMINSREGA(iins) == ARM_REG_R13) 
	{
	  /* WE LOSE TRACK OF THE STACK POINTER, THEREFORE THERE IS NO USE IN ADDING OTHER REGISTERS FURTHER DOWN IN THIS BLOCK*/
	  while (iins)
	    {
	      RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	      REGSET_FOREACH_REG(AllRegs(),reg)
		{
		  if (RegsetIn(saved,reg)) continue;
		  SavedRegOffset[reg] = -1;
		}
	      RegsetSetUnion(def,INS_REGS_DEF(iins));
	      RegsetSetUnion(use,INS_REGS_USE(iins));
	      iins = INS_NEXT(iins);
	    }
	  break;
	}
      else if (ARMINSOPCODE(iins) == ARM_LDM && ARMINSREGB(iins) == ARM_REG_R13 && ArmInsWriteBackHappens(iins))
	{
	  REGSET_FOREACH_REG(ARMINSIMMEDIATE(iins),reg)
	    {
	      if (ARMINSFLAGS(iins) & FL_DIRUP)
		offset +=4;
	      else 
		offset -=4;
	    }
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
	}
      else if (ARMINSOPCODE(iins) == ARM_LDM && RegsetIn(INS_REGS_DEF(iins),ARM_REG_R13) && RegsetIn(INS_REGS_DEF(iins),ARM_REG_R15))
	{
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
	}
      else if (ARMINSOPCODE(iins) == ARM_SWI)
	{
	}
      else if (ARMINSOPCODE(iins) == ARM_SFM && ArmInsWriteBackHappens(iins))
	{
	  t_uint32 local_offset = offset;

	  if (!ArmInsWriteBackHappens(iins))
	    FATAL(("no writeback @I\n",iins));
	  else
	  {
	    t_regset saved_regs ;
	    if (ARMINSFLAGS(iins) & FL_PREINDEX)
	      {
		if (ARMINSFLAGS(iins) & FL_DIRUP)
		  offset +=4*ARMINSIMMEDIATE(iins);
		else 
		  offset -=4*ARMINSIMMEDIATE(iins);;
	      }
	    
	    local_offset = offset;
	    
	    saved_regs = ARMINSMULTIPLE(iins);
	    REGSET_FOREACH_REG_INVERSE(saved_regs,reg)
	      {
		if (!RegsetIn(use,reg) && !RegsetIn(def,reg))
		  {
		    if (RegsetIn(saved,reg))
		      {
			RegsetSubReg(saved,reg);
			RegsetAddReg(use,reg);
			SavedRegOffset[reg] = -1;
		      }
		    else
		      {
			RegsetAddReg(saved,reg);
			SavedRegOffset[reg] = local_offset;
		      }
		  }
		if (ARMINSFLAGS(iins) & FL_DIRUP)
		  local_offset-=4;
		else
		  local_offset+=4;
	      }
	    
	    if (!(ARMINSFLAGS(iins) & FL_PREINDEX))
	      {
		if (ARMINSFLAGS(iins) & FL_DIRUP)
		  offset +=4*ARMINSIMMEDIATE(iins);
		else 
		  offset -=4*ARMINSIMMEDIATE(iins);
	      }
	    
	  }
	}
      else if (ARMINSOPCODE(iins) == ARM_LFM && ArmInsWriteBackHappens(iins))
	{
	  t_uint32 local_offset = offset;
	  t_regset saved_regs;
	  if (!ArmInsWriteBackHappens(iins))
	    FATAL(("no writeback @I\n",iins));
	  else
	  {
	    if (ARMINSFLAGS(iins) & FL_PREINDEX)
	      {
		if (ARMINSFLAGS(iins) & FL_DIRUP)
		  offset +=4*ARMINSIMMEDIATE(iins);
		else 
		  offset -=4*ARMINSIMMEDIATE(iins);;
	      }
	    
	    local_offset = offset;
	    
	    saved_regs = ARMINSMULTIPLE(iins);
	    REGSET_FOREACH_REG_INVERSE(saved_regs,reg)
	      {
		if (ARMINSFLAGS(iins) & FL_DIRUP)
		  local_offset-=4;
		else
		  local_offset+=4;
	      }
	    
	    if (!(ARMINSFLAGS(iins) & FL_PREINDEX))
	      {
		if (ARMINSFLAGS(iins) & FL_DIRUP)
		  offset +=4*ARMINSIMMEDIATE(iins);
		else 
		  offset -=4*ARMINSIMMEDIATE(iins);
	      }
	    
	  }
	}
      else if (ARMINSOPCODE(iins)==ARM_STR && ARMINSREGB(iins) == ARM_REG_R13)
	{
	  t_uint32 imm = ARMINSIMMEDIATE(iins);

	  if (ArmInsWriteBackHappens(iins))
	  {
	    if (ARMINSFLAGS(iins) & FL_DIRUP)
	      FATAL(("dirup store @I\n",iins));
	    if (!(ARMINSFLAGS(iins) & FL_PREINDEX))
	      FATAL(("no preindex @I\n",iins));
	    offset-=imm;
	  }

	  reg = ARMINSREGA(iins);

	  /* only if reg has not been used and not been defined
	     we have a true save */ 

	  if (!RegsetIn(use,reg) && !RegsetIn(def,reg))
            {
	      if (RegsetIn(saved,reg))
                {
		  RegsetSubReg(saved,reg);
		  RegsetAddReg(use,reg);
		  SavedRegOffset[reg] = -1;
                }
	      else
                {
		  RegsetAddReg(saved,reg);
		  if(ArmInsWriteBackHappens(iins))
		    SavedRegOffset[reg] = offset; 
		  else 
		    SavedRegOffset[reg] = offset+imm;
                }
            }

	  /* do not consider this a use ! */
        }
      else if (ARMINSOPCODE(iins) == ARM_STM && ARMINSREGB(iins) == ARM_REG_R13 &&
	       !(ARMINSFLAGS(iins) & FL_DIRUP) &&
	       (ARMINSFLAGS(iins) & FL_PREINDEX)
	       )
	{
	  t_int32 old_offset = offset;  

	  t_regset saved_regs = ARMINSIMMEDIATE(iins);
	  REGSET_FOREACH_REG_INVERSE(saved_regs,reg)
	    {
	      offset-=4;
	      if (!RegsetIn(use,reg) && !RegsetIn(def,reg))
		{
		  if (RegsetIn(saved,reg))
		    {
		      RegsetSubReg(saved,reg);
		      RegsetAddReg(use,reg);
		      SavedRegOffset[reg] = -1;
		    }
		  else
		    {
		      RegsetAddReg(saved,reg);
		      SavedRegOffset[reg] = offset; 
		    }
		}
            }
	  if (!ArmInsWriteBackHappens(iins))
	    offset = old_offset;
	}
      else
        {
	  if (RegsetIn(INS_REGS_DEF(iins),ARM_REG_R13))
	    {
	      WARNING(("ins: @I\n bbl:@iB\nstack pointer manipulation\n",iins,i_bbl));
	      break;
	    }
	  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(INS_REGS_USE(iins)),def));
	  REGSET_FOREACH_REG(AllRegs(),reg)
            {
	      if (RegsetIn(saved,reg)) continue;
	      SavedRegOffset[reg] = -1;
            }
	  RegsetSetUnion(def,INS_REGS_DEF(iins));
	  RegsetSetUnion(use,INS_REGS_USE(iins));
        }
    }

  /* now make sure that nothing used subsequently */

  RegsetSetIntersect(saved,RegsetNewUnion(RegsetNewInvers(BBL_REGS_LIVE_OUT(i_bbl)),def));

  REGSET_FOREACH_REG(AllRegs(),reg)
    {
      if(RegsetIn(saved,reg)) continue;
      SavedRegOffset[reg] = -1;
    }
#if 0
  printf("========================================================================================================\n");

  DiabloPrintArch(stdout,BBL_DESCRIPTOR(i_bbl),"@iB @A\n",i_bbl,BBL_REGS_LIVE_OUT(i_bbl));

  REGSET_FOREACH_REG(saved,reg)
    {
      printf("%d: %d\n",reg,SavedRegOffset[reg]);
    }
#endif
  /*now check the loads - we look at all the return bbls*/
  
  count = 0;
    
  BBL_FOREACH_PRED_EDGE(FUN_BBL_LAST(ifun),i_edge)
    {
      if (EDGE_CAT(i_edge)!=ET_JUMP) continue;
      i_bbl = CFG_EDGE_HEAD(i_edge);

      count++;

      def = RegsetNew();
      saved = RegsetNew();
        
      /*      printf("--------------------------------------------------------------------\n");*/

      offset = 0;

      BBL_FOREACH_INS_R(i_bbl,iins)
        {
	  /*	  DiabloPrint(stdout,"offset %d @I\n",offset,iins);*/
	  t_regset def_regs = INS_REGS_DEF(iins);

	  if( !RegsetIsEmpty(RegsetNewIntersect(def,def_regs)))
	  {
	    if(ARMINSOPCODE(iins) == ARM_LDR && ARMINSREGB(iins) == ARM_REG_R13)
	    {

	    }
	    continue;
	  }

	  RegsetSetUnion(def,INS_REGS_USE(iins));
	  RegsetSetUnion(def,INS_REGS_DEF(iins));

	  /*	  printf("%d %d %d\n",ARMINSOPCODE(iins) == ARM_STR,ARMINSREGB(iins) == ARM_REG_R13,ArmInsWriteBackHappens(iins));*/
	  
	  if(ARMINSOPCODE(iins) == ARM_STR && ARMINSREGB(iins) == ARM_REG_R13 && ArmInsWriteBackHappens(iins))
	    {
	      /* opposite operation since we are running backwards */
	      if (ARMINSFLAGS(iins) & FL_DIRUP)
		offset -= ARMINSIMMEDIATE(iins);
	      else
		offset += ARMINSIMMEDIATE(iins);
	    }
	  else if(ARMINSOPCODE(iins) == ARM_LDR && ARMINSREGB(iins) == ARM_REG_R13 && ArmInsWriteBackHappens(iins))
	    {
	      /* opposite operation since we are running backwards */

	      if (ARMINSFLAGS(iins) & FL_DIRUP)
	      {
		offset -= ARMINSIMMEDIATE(iins);
	      }
	      else
	      {
		offset += ARMINSIMMEDIATE(iins);
	      }

	    }
	  else if(ARMINSOPCODE(iins) == ARM_SUB && ARMINSREGA(iins) == ARM_REG_R13 && ARMINSREGB(iins) == ARM_REG_R13 && ARMINSREGC(iins) == ARM_REG_NONE) 
	    {
	      /* opposite operation since we are running backwards */
	      offset += ARMINSIMMEDIATE(iins);
	    }
	  else if(ARMINSOPCODE(iins) == ARM_ADD && ARMINSREGA(iins) == ARM_REG_R13 && ARMINSREGB(iins) == ARM_REG_R13 && ARMINSREGC(iins) == ARM_REG_NONE) 
	    {
	      /* opposite operation since we are running backwards */
	      offset -= ARMINSIMMEDIATE(iins);
	    }
	  else if(ARMINSOPCODE(iins) == ARM_RSB && ARMINSREGA(iins) == ARM_REG_R13 && ARMINSREGC(iins) == ARM_REG_R13 && ARMINSREGB(iins) == ARM_REG_NONE) 
	    {
	      /* opposite operation since we are running backwards */
	      offset += ARMINSIMMEDIATE(iins);
	    }
	  else if (ARMINSOPCODE(iins)==ARM_LDR && ARMINSREGA(iins)==ARM_REG_R15 && ARMINSIMMEDIATE(iins)==4 && ARMINSREGB(iins) == ARM_REG_R13)
	    {
	      if (!ArmInsWriteBackHappens(iins))
		FATAL(("no writeback @I\n",iins));
	      if (!(ARMINSFLAGS(iins) & FL_DIRUP))
		FATAL(("dirup store @I\n",iins));
	      if (ARMINSFLAGS(iins) & FL_PREINDEX)
		FATAL(("no preindex @I\n",iins));
	      if( SavedRegOffset[ARM_REG_R15] != offset)
		{
		  SavedRegOffset[ARM_REG_R15] = -1;
		  continue;
		}
	      else
		{
		  RegsetAddReg(saved,ARM_REG_R15);
		}            
	      offset-=4;
	    }
	  else if (
		   ARMINSREGB(iins)==ARM_REG_R13 && 
		   ARMINSOPCODE(iins)==ARM_LDM &&
		   ARMINSFLAGS(iins) & FL_DIRUP &&
		   !(ARMINSFLAGS(iins) & FL_PREINDEX)
		   )
	    {
	      t_regset saved_regs = ARMINSIMMEDIATE(iins);
	      if (!ArmInsWriteBackHappens(iins))
		{
		  REGSET_FOREACH_REG_INVERSE(saved_regs,reg)
		    SavedRegOffset[reg] = -1;
		  continue;
		  /*		  FATAL(("no writeback @I\n",iins));*/
		}
	     
	      REGSET_FOREACH_REG_INVERSE(saved_regs,reg)
		{
		  offset -=4;
		  if( SavedRegOffset[reg] != offset )
		    {
		      /*		      printf("not saved %d - offset %d \n",reg,offset);*/
		      SavedRegOffset[reg] = -1;
		    }
		  else
		    {
		      /*		      printf("saved += %d\n",reg);*/
		      RegsetAddReg(saved,reg);
		    }            
		  /*		  offset -= 4;*/
		}
	    }
	  else if (
		   ARMINSREGB(iins)==ARM_REG_R11 && 
		   ARMINSOPCODE(iins)==ARM_LDM &&
		   !(ARMINSFLAGS(iins) & FL_DIRUP) &&
		   ARMINSFLAGS(iins) & FL_PREINDEX &&
		   RegsetIn(INS_REGS_DEF(iins),ARM_REG_R15) &&
		   RegsetIn(INS_REGS_DEF(iins),ARM_REG_R13)
		   )
	    {
	      t_regset saved_regs = ARMINSIMMEDIATE(iins);
	      offset -=8;
	     
	      REGSET_FOREACH_REG_INVERSE(saved_regs,reg)
		{
		  if( SavedRegOffset[reg] != offset )
		    {
		      /*		      printf("not saved %d - offset %d \n",reg,offset);*/
		      SavedRegOffset[reg] = -1;
		    }
		  else
		    {
		      /*		      printf("saved += %d\n",reg);*/
		      RegsetAddReg(saved,reg);
		    }            
		  offset -= 4;
		}
	    }
	  else if ((ARMINSOPCODE(iins)==ARM_LDF) && RegsetIn(INS_REGS_DEF(iins),ARM_REG_R13))
	    {
	      REGSET_FOREACH_REG(AllRegs(),reg)
		{
		  saved = RegsetNew();
		  SavedRegOffset[reg] = -1;
		}
	    }
	  else if ((ARMINSOPCODE(iins)==ARM_STM) && RegsetIn(INS_REGS_DEF(iins),ARM_REG_R13))
	    {
	      REGSET_FOREACH_REG(AllRegs(),reg)
		{
		  saved = RegsetNew();
		  SavedRegOffset[reg] = -1;
		}
	    }
	  else if ((ARMINSOPCODE(iins)==ARM_BIC) && RegsetIn(INS_REGS_DEF(iins),ARM_REG_R13))
	    {
	      REGSET_FOREACH_REG(AllRegs(),reg)
		{
		  saved = RegsetNew();
		  SavedRegOffset[reg] = -1;
		}
	    }
	  else if (ARMINSOPCODE(iins) == ARM_LFM && ArmInsWriteBackHappens(iins))
	    {
	      t_uint32 local_offset = offset;
	      
	      if (!ArmInsWriteBackHappens(iins))
		FATAL(("no writeback @I\n",iins));
	      else
		{
		  t_regset saved_regs;
		  if (ARMINSFLAGS(iins) & FL_PREINDEX)
		    {
		      if (ARMINSFLAGS(iins) & FL_DIRUP)
			offset -=4*ARMINSIMMEDIATE(iins);
		      else 
			offset +=4*ARMINSIMMEDIATE(iins);;
		    }
		  
		  local_offset = offset;
		  
		  saved_regs = ARMINSMULTIPLE(iins);
		  REGSET_FOREACH_REG(saved_regs,reg)
		    {
		      if( SavedRegOffset[reg] != offset )
			{
			  /*		      printf("not saved %d - offset %d \n",reg,offset);*/
			  SavedRegOffset[reg] = -1;
			}
		      else
			{
			  /*		      printf("saved += %d\n",reg);*/
			  RegsetAddReg(saved,reg);
			}          
		      if (ARMINSFLAGS(iins) & FL_DIRUP)
			local_offset+=4;
		      else
			local_offset-=4;
		    }
		  
		  if (!(ARMINSFLAGS(iins) & FL_PREINDEX))
		    {
		      if (ARMINSFLAGS(iins) & FL_DIRUP)
			offset -=4*ARMINSIMMEDIATE(iins);
		      else 
			offset +=4*ARMINSIMMEDIATE(iins);
		    }
		}
	    }
	  else if ((ARMINSOPCODE(iins)==ARM_LDR)&&(ARMINSREGB(iins)==ARM_REG_R13))
	  {
	    DiabloPrint(stdout,"aaaaaaaaaaaaaaaaaa @iB\n",INS_BBL(iins));
	  }
	  else if (RegsetIn(INS_REGS_DEF(iins),ARM_REG_R13))
	    {
	      DiabloPrint(stdout,"ins: @I\n bbl:@iB\n",iins,i_bbl);	
	      FATAL(("DEFINITION OF THE SP"));
	    }
        }

      /*      PrintRegset(stdout,ifun->cfg->sec->obj->description,saved);printf("\n");*/
      
      REGSET_FOREACH_REG(AllRegs(),reg)
	{
	  if (RegsetIn(saved,reg)) continue;
	  SavedRegOffset[reg] = -1;
	}
    }
  
  
  if( count )
    {
      FUN_REGS_SAVED(ifun) = RegsetNew();
      /*      PrintRegset(stdout,ifun->cfg->sec->obj->description,FUN_REGS_SAVED(ifun));*/
      
      REGSET_FOREACH_REG(AllRegs(),reg)
        {
	  if( SavedRegOffset[reg]!=-1 )
	    RegsetAddReg(FUN_REGS_SAVED(ifun),reg);
        }
    }
  else
    {
      FUN_REGS_SAVED(ifun) = RegsetNew();
    }

  /* WIJZIGING*/

  /*  PrintRegset(stdout,ifun->cfg->sec->obj->description,FUN_REGS_SAVED(ifun));*/

  /*  printf("saved before ");*/
  /*  PrintRegset(stdout,ifun->cfg->sec->obj->description,FUN_REGS_SAVED(ifun));*/

 end:
#ifdef AGGRESSIVE_LIVENESS
  if (global_options.rely_on_calling_conventions)
    if (ifun->function_flags & FF_IS_EXPORTED)
      {      
	/*      RegsetAddReg(FUN_REGS_SAVED(ifun),);*/
	
	FUN_REGS_SAVED(ifun) |=  FUNCTION_DESCRIPTOR(ifun)->callee_saved &~ FUNCTION_DESCRIPTOR(ifun)->callee_may_use;
	/*            FUN_REGS_SAVED(ifun) |=  FUNCTION_DESCRIPTOR(ifun)->callee_saved; */
      }
  /*  printf("saved after ");*/
  /*  PrintRegset(stdout,ifun->cfg->sec->obj->description,FUN_REGS_SAVED(ifun));*/
#endif
  return;
}


t_bool ArmFunIsGlobal(t_function * fun)
{
  if (fun->function_flags & FF_IS_EXPORTED)
  {
    if(FunctionBehaves(fun))
      return TRUE;
    else return FALSE;
  }
  return FALSE;
}
/* vim: set shiftwidth=2 foldmethod=marker: */
