/* Includes {{{ */
#include "thumb_assemble.h"
#include "arm_opcodes.h"
#include "arm_registers.h"
#include "arm_instructions.h"
/*#include "thumb_print.h" */
#include "../../utils/util_error.h"
/*#include "../../utils/util_status.h" */
/*#include "../../utils/util_bitmanipulation.h" */
/* }}} */
/* ThumbAssembleCondBranch {{{ */
void ThumbAssembleCondBranch(t_ins * ins, t_uint16 * instr)
{
	*instr = 0xd000;  /* Set the opcode */
	*instr |= (ARMINSCONDITION(ins)) << 8;
	*instr |= (ARMINSIMMEDIATE(ins)) >> 1;
}
/* }}} */

/* ThumbAssembleBranchLink {{{ */
void ThumbAssembleBranchLink(t_ins * ins, t_uint32 * instr)
{
	*instr = (ARMINSOPCODE(ins) == ARM_BL) ? 0xf800f000 : 0xe800f000; /* Set the opcode */
	*instr |= (ARMINSIMMEDIATE(ins) & 0x00000ffe) << 15;
	*instr |= (ARMINSIMMEDIATE(ins) & 0x007ff000) >> 12;
}
/* }}} */

/* ThumbAssembleBranch {{{ */
void ThumbAssembleBranch(t_ins * ins, t_uint16 * instr)
{
	if (ARMINSFLAGS(ins) & FL_IMMED) {
		*instr = 0xe000;  /* Set the opcode */
		*instr |= (ARMINSIMMEDIATE(ins)) >> 1;
	}
	else {
		*instr = 0x4700;  /* Set the opcode */
		*instr |= (ARMINSREGB(ins)) << 3;
	}
}
/* }}} */

/* ThumbAssembleSWI {{{ */
void ThumbAssembleSWI(t_ins * ins, t_uint16 * instr)
{
	*instr = 0xdf00; /* Set the opcode */
	*instr |= ARMINSIMMEDIATE(ins);
}
/* }}} */

/* ThumbAssemble3Bit {{{ */
void ThumbAssemble3Bit(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x1800; /* Set the opcode */
	if (ARMINSFLAGS(ins) & FL_IMMED) {
		*instr |= 0x1 << 10;
		*instr |= (ARMINSIMMEDIATE(ins)) << 6;
	}
	else
		*instr |= (ARMINSREGC(ins)) << 6;
	if (ARMINSOPCODE(ins) == ARM_SUB)
		*instr |= 0x1 << 9;
	*instr |= (ARMINSREGB(ins)) << 3;
	*instr |= ARMINSREGA(ins);
}
/* }}} */

/* ThumbAssembleImm {{{ */
void ThumbAssembleImm(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x2000;
	*instr |= ARMINSIMMEDIATE(ins);
	switch(ARMINSOPCODE(ins)) {
	case ARM_MOV:
		*instr |= (ARMINSREGA(ins)) << 8;
		break;
	case ARM_CMP:
		*instr |= (ARMINSREGB(ins)) << 8;
		*instr |= 0x1 << 11;
		break;
	case ARM_ADD:
		*instr |= (ARMINSREGA(ins)) << 8;
		*instr |= 0x2 << 11;
		break;
	case ARM_SUB:
		*instr |= (ARMINSREGB(ins)) << 8;
		*instr |= 0x3 << 11;
		break;
	default:
		FATAL(("No thumb instruction"));
	}
}
/* }}} */

/* ThumbAssembleShifted {{{ */
void ThumbAssembleShifted(t_ins * ins, t_uint16 * instr)
{
	*instr = ARMINSSHIFTLENGTH(ins) << 6;
	*instr |= (ARMINSREGC(ins)) << 3;
	*instr |= ARMINSREGA(ins);
	if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_LSR_IMM)
		*instr |= 0x1 << 11;
	if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_ASR_IMM)
		*instr |= 0x2 << 11;
}
/* }}} */

/* ThumbAssembleALU {{{ */
void ThumbAssembleALU(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x4000;
	switch(ARMINSOPCODE(ins)) {
	case ARM_AND:
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_EOR:
		*instr |= 0x1 << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_MOV:
		if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_LSL_REG)
			*instr |= 0x2 << 6;
		if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_LSR_REG)
			*instr |= 0x3 << 6;
		if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_ASR_REG)
			*instr |= 0x4 << 6;
		if (ARMINSSHIFTTYPE(ins) == ARM_SHIFT_TYPE_ROR_REG)
			*instr |= 0x7 << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGS(ins)) << 3;
		break;
	case ARM_ADC:
		*instr |= 0x5 << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_SBC:
		*instr |= 0x6 << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_TST:
		*instr |= 0x8 << 6;
		*instr |= ARMINSREGB(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_RSB:
		*instr |= 0x9 << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGB(ins)) << 3;
		break;
	case ARM_CMP:
		*instr |= 0xa << 6;
		*instr |= ARMINSREGB(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_CMN:
		*instr |= 0xb << 6;
		*instr |= ARMINSREGB(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_ORR:
		*instr |= 0xc << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_MUL:
		*instr |= 0xd << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGB(ins)) << 3;
		break;
	case ARM_BIC:
		*instr |= 0xe << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	case ARM_MVN:
		*instr |= 0xf << 6;
		*instr |= ARMINSREGA(ins);
		*instr |= (ARMINSREGC(ins)) << 3;
		break;
	default:
		FATAL(("No thumb instruction"));
	}
}
/* }}} */

/* ThumbAssembleHiReg {{{ */
void ThumbAssembleHiReg(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x4400;
	if (ARMINSOPCODE(ins) == ARM_ADD || ARMINSOPCODE(ins) == ARM_MOV) {
		*instr |= ((ARMINSREGA(ins) & 0x8) << 4);
		*instr |= ARMINSREGA(ins) & 0x7;
		if (ARMINSOPCODE(ins) == ARM_MOV)
			*instr |= 0x1 << 9;
	}
	if (ARMINSOPCODE(ins) == ARM_CMP) {
		*instr |= ((ARMINSREGB(ins) & 0x8) << 4);
		*instr |= ARMINSREGB(ins) & 0x7;
		*instr |= 0x1 << 8;
	}
	*instr |= (ARMINSREGC(ins)) << 3;
}
/* }}} */

/* ThumbAssembleLoadAddress {{{ */
void ThumbAssembleLoadAddress(t_ins * ins, t_uint16 * instr)
{
	*instr = 0xa000;
	*instr |= (ARMINSREGA(ins)) << 8;
	if (ARMINSREGB(ins) == ARM_REG_R13)
		*instr |= 0x0800;
	*instr |= (ARMINSIMMEDIATE(ins)) >> 2;
}
/* }}} */

/* ThumbAssembleStack {{{ */
void ThumbAssembleStack(t_ins * ins, t_uint16 * instr)
{
	*instr = 0xb000;
	if (ARMINSOPCODE(ins) == ARM_SUB)
		*instr |= 0x1 << 7;
	*instr |= (ARMINSIMMEDIATE(ins)) >> 2;
}
/* }}} */

/* ThumbAssembleTransferImm {{{ */
void ThumbAssembleTransferImm(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x6000;
	*instr |= ARMINSREGA(ins);
	*instr |= (ARMINSREGB(ins)) << 3;
	switch(ARMINSOPCODE(ins)) {
	case ARM_LDR:
		*instr |= 0x1 << 11;
	case ARM_STR:
		*instr |= ((ARMINSIMMEDIATE(ins)) >> 2) << 6;
		break;
	case ARM_LDRB:
		*instr |= 0x1 << 11;
	case ARM_STRB:
		*instr |= 0x1 << 12;
		*instr |= (ARMINSIMMEDIATE(ins)) << 6;
		break;
	default:
		FATAL(("No thumb instruction"));
	}
}
/* }}} */

/* ThumbAssembleTransferHalf {{{ */
void ThumbAssembleTransferHalf(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x8000;
	*instr |= ARMINSREGA(ins);
	*instr |= (ARMINSREGB(ins)) << 3;
	if (ARMINSOPCODE(ins) == ARM_LDRH)
		*instr |= 0x1 << 11;
	*instr |= ((ARMINSIMMEDIATE(ins)) >> 1) << 6;
}
/* }}} */

/* ThumbAssembleTransferSign {{{ */
void ThumbAssembleTransferSign(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x5200;
	*instr |= ARMINSREGA(ins);
	*instr |= (ARMINSREGB(ins)) << 3;
	*instr |= (ARMINSREGC(ins)) << 6;
	if (ARMINSOPCODE(ins) == ARM_LDRSB)
		*instr |= 0x1 << 10;
	if (ARMINSOPCODE(ins) == ARM_LDRH)
		*instr |= 0x2 << 10;
	if (ARMINSOPCODE(ins) == ARM_LDRSH)
		*instr |= 0x3 << 10;
}
/* }}} */

/* ThumbAssembleTransferRegOff {{{ */
void ThumbAssembleTransferRegOff(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x5000;
	*instr |= ARMINSREGA(ins);
	*instr |= (ARMINSREGB(ins)) << 3;
	*instr |= (ARMINSREGC(ins)) << 6;
	if (ARMINSOPCODE(ins) == ARM_STRB)
		*instr |= 0x1 << 10;
	if (ARMINSOPCODE(ins) == ARM_LDR)
		*instr |= 0x2 << 10;
	if (ARMINSOPCODE(ins) == ARM_LDRB)
		*instr |= 0x3 << 10;
}
/* }}} */

/* ThumbAssembleTransferPC {{{ */
void ThumbAssembleTransferPC(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x4800;
	*instr |= (ARMINSREGA(ins)) << 8;
	*instr |= (ARMINSIMMEDIATE(ins)) >> 2;
}
/* }}} */

/* ThumbAssembleTransferSP {{{ */
void ThumbAssembleTransferSP(t_ins * ins, t_uint16 * instr)
{
	*instr = 0x9000;
	*instr |= (ARMINSREGA(ins)) << 8;
	*instr |= (ARMINSIMMEDIATE(ins)) >> 2;
	if (ARMINSOPCODE(ins) == ARM_LDR)
		*instr |= 0x1 << 11;
}
/* }}} */

/* ThumbAssembleMultipleTransfer {{{ */
void ThumbAssembleMultipleTransfer(t_ins * ins, t_uint16 * instr)
{
	*instr = 0xc000;
	*instr |= (ARMINSREGB(ins)) << 8;
	*instr |= ARMINSIMMEDIATE(ins);
	if (ARMINSOPCODE(ins) == ARM_LDM)
		*instr |= 0x1 << 11;
}
/* }}} */

/* ThumbAssemblePP {{{ */
void ThumbAssemblePP(t_ins * ins, t_uint16 * instr)
{
	*instr = 0xb400;
	*instr |= (ARMINSIMMEDIATE(ins) & 0x00ff);
	if (ARMINSOPCODE(ins) == ARM_LDM) {
		*instr |= 0x1 << 11;
		*instr |= (ARMINSIMMEDIATE(ins) & 0x8000) >> 7;
	}
	else
		*instr |= (ARMINSIMMEDIATE(ins) & 0x4000) >> 6;
}
/* }}} */
/* ThumbAssembleData {{{ */
void ThumbAssembleData(t_ins * ins, t_uint16 * instr)
{
	*instr = ARMINSIMMEDIATE(ins);
}
/* }}} */
/* ThumbAssembleUnsupported {{{ */
void ThumbAssembleUnsupported(t_ins * ins, t_uint16 instr, t_uint16 opc)
{
        FATAL(("Unsupported thumb instruction. Instr = 0x%x\n", instr));
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
