/* Includes {{{ */
#include "thumb_opcodes.h"
#include "thumb_disassemble_one.h"
/* }}} */
/*! \todo Document */
/* thumb_opcode_table {{{ */
const thumb_opcode thumb_opcode_table[] = {

	/* Conditional Branch */
	{0xff00, 0xd000, "BEQ", ThumbDisassembleCondBranch},
	{0xff00, 0xd100, "BNE", ThumbDisassembleCondBranch},
	{0xff00, 0xd200, "BCS", ThumbDisassembleCondBranch},
	{0xff00, 0xd300, "BCC", ThumbDisassembleCondBranch},
	{0xff00, 0xd400, "BMI", ThumbDisassembleCondBranch},
	{0xff00, 0xd500, "BPL", ThumbDisassembleCondBranch},
	{0xff00, 0xd600, "BVS", ThumbDisassembleCondBranch},
	{0xff00, 0xd700, "BVC", ThumbDisassembleCondBranch},
	{0xff00, 0xd800, "BHI", ThumbDisassembleCondBranch},
	{0xff00, 0xd900, "BLS", ThumbDisassembleCondBranch},
	{0xff00, 0xda00, "BGE", ThumbDisassembleCondBranch},
	{0xff00, 0xdb00, "BLT", ThumbDisassembleCondBranch},
	{0xff00, 0xdc00, "BGT", ThumbDisassembleCondBranch},
	{0xff00, 0xdd00, "BLE", ThumbDisassembleCondBranch},

	/* Branches */
	{0xff00, 0xde00, "BAL", ThumbDisassembleBranch},
	{0xf800, 0xe000, "B", ThumbDisassembleBranch},
	{0xf800, 0xf000, "BL", ThumbDisassembleBranchLink},
	{0xff80, 0x4700, "BX", ThumbDisassembleBranch},
/*	{0xf800, 0xf800, "BLX", ThumbDisassembleBranchLink}, */

	/* Software Interrupt */
	{0xff00, 0xdf00, "SWI", ThumbDisassembleSWI},

	/* Add/Subtract */
	{0xfe00, 0x1a00, "SUB_3R", ThumbDisassemble3Bit},
	{0xfe00, 0x1800, "ADD_3R", ThumbDisassemble3Bit},
	{0xfe00, 0x1e00, "SUB_3I", ThumbDisassemble3Bit},
	{0xfe00, 0x1c00, "ADD_3I", ThumbDisassemble3Bit},

	/* Move, compare, add, subtract immediate */
	{0xf800, 0x2000, "MOV_I", ThumbDisassembleImm},
	{0xf800, 0x2800, "CMP_I", ThumbDisassembleImm},
	{0xf800, 0x3000, "ADD_2I", ThumbDisassembleImm},
	{0xf800, 0x3800, "SUB_2I", ThumbDisassembleImm},

	/* Move shifted register */
	{0xf800, 0x0000, "LSL_I", ThumbDisassembleShifted},
	{0xf800, 0x0800, "LSR_I", ThumbDisassembleShifted},
	{0xf800, 0x1000, "ASR_I", ThumbDisassembleShifted},

	/* ALU operations */
	{0xffc0, 0x4000, "AND", ThumbDisassembleALU},
	{0xffc0, 0x4040, "EOR", ThumbDisassembleALU},
	{0xffc0, 0x4080, "LSL", ThumbDisassembleALU},
	{0xffc0, 0x40c0, "LSR", ThumbDisassembleALU},
	{0xffc0, 0x4100, "ASR", ThumbDisassembleALU},
	{0xffc0, 0x4140, "ADC", ThumbDisassembleALU},
	{0xffc0, 0x4180, "SBC", ThumbDisassembleALU},
	{0xffc0, 0x41c0, "ROR", ThumbDisassembleALU},
	{0xffc0, 0x4200, "TST", ThumbDisassembleALU},
	{0xffc0, 0x4240, "NEG", ThumbDisassembleALU},
	{0xffc0, 0x4280, "CMP", ThumbDisassembleALU},
	{0xffc0, 0x42c0, "CMN", ThumbDisassembleALU},
	{0xffc0, 0x4300, "ORR", ThumbDisassembleALU},
	{0xffc0, 0x4340, "MUL", ThumbDisassembleALU},
	{0xffc0, 0x4380, "BIC", ThumbDisassembleALU},
	{0xffc0, 0x43c0, "MVN", ThumbDisassembleALU},

	/* Hi register operations */
	{0xff00, 0x4400, "ADD_2RH", ThumbDisassembleHiReg},
	{0xff00, 0x4500, "CMP_2RH", ThumbDisassembleHiReg},
	{0xff00, 0x4600, "MOV_2RH", ThumbDisassembleHiReg},

	/* Load address */
	{0xf800, 0xa000, "ADDPC", ThumbDisassembleLoadAddress},
	{0xf800, 0xa800, "ADDSP", ThumbDisassembleLoadAddress},

	/* Add offset to stack pointer */
	{0xff80, 0xb000, "ADD_SPI", ThumbDisassembleStack},
	{0xff80, 0xb080, "SUB_SPI", ThumbDisassembleStack},

	/* Load, store with immediate offset */
	{0xf800, 0x6000, "STRI", ThumbDisassembleTransferImm},
	{0xf800, 0x6800, "LDRI", ThumbDisassembleTransferImm},
	{0xf800, 0x7000, "STRBI", ThumbDisassembleTransferImm},
	{0xf800, 0x7800, "LDRBI", ThumbDisassembleTransferImm},

	/* Load, store halfword */
	{0xf800, 0x8000, "STRHI", ThumbDisassembleTransferHalf},
	{0xf800, 0x8800, "LDRHI", ThumbDisassembleTransferHalf},

	/* Load, store sign-extended byte/halfword */
	{0xfe00, 0x5200, "STRHR", ThumbDisassembleTransferSign},
	{0xfe00, 0x5a00, "LDRHR", ThumbDisassembleTransferSign},
	{0xfe00, 0x5600, "LDRSB", ThumbDisassembleTransferSign},
	{0xfe00, 0x5e00, "LDRSH", ThumbDisassembleTransferSign},

	/* Load, store with register offset */
	{0xfe00, 0x5000, "STRR", ThumbDisassembleTransferRegOff},
	{0xfe00, 0x5800, "LDRR", ThumbDisassembleTransferRegOff},
	{0xfe00, 0x5400, "STRBR", ThumbDisassembleTransferRegOff},
	{0xfe00, 0x5c00, "LDRBR", ThumbDisassembleTransferRegOff},

	/* PC-relative load */
	{0xf800, 0x4800, "LDR_PC", ThumbDisassembleTransferPC},

	/* SP-relative load/store */
	{0xf800, 0x9000, "STR_SP", ThumbDisassembleTransferSP},
	{0xf800, 0x9800, "LDR_SP", ThumbDisassembleTransferSP},

	/* Multiple load/store */
	{0xf800, 0xc000, "STMIA", ThumbDisassembleMultipleTransfer},
	{0xf800, 0xc800, "LDMIA", ThumbDisassembleMultipleTransfer},

	/* Push, pop registers */
	{0xfe00, 0xb400, "PUSH", ThumbDisassemblePP},
	{0xfe00, 0xbc00, "POP", ThumbDisassemblePP},

	/* opvang voor niet-herkende instructies */
	{0x0000, 0x0000, "UNDEF", ThumbDisassembleUnsupported}

};
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
