#include "../../config.h"
#include "../../simple_types.h"
#include "../../datastructures/data_string.h"
#include "../../kernel/diablo_ins.h"

#ifndef THUMB_OPCODES_H
#define THUMB_OPCODES_H
/*! The structure of the different disassemble functions */
typedef void (*ThumbDisassembleFunction) (t_ins *, t_uint32 instr, t_uint16 opc);
/*! The opcode structure, used in the opcode table to specify what disassembly
 * function corresponds to what bitpattern */
typedef struct _thumb_opcode
{
	/*! \todo Document */
	t_uint16 mask;
	/*! \todo Document */
	t_uint16 opcode;
	/*! \todo Document */
	t_string desc;
	/*! \todo Document */
	ThumbDisassembleFunction Disassemble;
} thumb_opcode;
/*! The opcode table */
extern const thumb_opcode thumb_opcode_table[];
/* The opcodes {{{ */
typedef enum _t_thumb_opcode
{
	TH_BEQ, /* 0 */
	TH_BNE, /* 1 */
	TH_BCS, /* 2 */
	TH_BCC, /* 3 */
	TH_BMI, /* 4 */
	TH_BPL, /* 5 */
	TH_BVS, /* 6 */
	TH_BVC, /* 7 */
	TH_BHI, /* 8 */
	TH_BLS, /* 9 */
	TH_BGE, /* 10 */
	TH_BLT, /* 11 */
	TH_BGT, /* 12 */
	TH_BLE, /* 13 */
	TH_BAL, /* 14 */
	TH_B, /* 15 */
	TH_BL, /* 16 */
	TH_BX, /* 17 */
/*	TH_BLX, */ /* 17b */
	TH_SWI, /* 18 */
	TH_SUB_3R, /* 19 */
	TH_ADD_3R, /* 20 */
	TH_SUB_3I, /* 21 */
	TH_ADD_3I, /* 22 */
	TH_MOV_I, /* 23 */
	TH_CMP_I, /* 24 */
	TH_ADD_2I, /* 25 */
	TH_SUB_2I, /* 26 */
	TH_LSL_I, /* 27 */
	TH_LSR_I, /* 28 */
	TH_ASR_I, /* 29 */
	TH_AND, /* 30 */
	TH_EOR, /* 31 */
	TH_LSL_R, /* 32 */
	TH_LSR_R, /* 33 */
	TH_ASR_R, /* 34 */
	TH_ADC, /* 35 */
	TH_SBC, /* 36 */
	TH_ROR, /* 37 */
	TH_TST, /* 38 */
	TH_NEG, /* 39 */
	TH_CMP_R, /* 40 */
	TH_CMN, /* 41 */
	TH_ORR, /* 42 */
	TH_MUL, /* 43 */
	TH_BIC, /* 44 */
	TH_MVN, /* 45 */
	TH_ADD_2RH, /* 46 */
	TH_CMP_RH, /* 47 */
	TH_MOV_RH, /* 48 */
	TH_ADD_PC, /* 49 */
	TH_ADD_SP, /* 50 */
	TH_ADD_SPI, /* 51 */
	TH_SUB_SPI, /* 52 */
	TH_STR_I, /* 53 */
	TH_LDR_I, /* 54 */
	TH_STRB_I, /* 55 */
	TH_LDRB_I, /* 56 */
	TH_STRH_I, /* 57 */
	TH_LDRH_I, /* 58 */
	TH_STRH_R, /* 59 */
	TH_LDRH_R, /* 60 */
	TH_LDRSB, /* 61 */
	TH_LDRSH, /* 62 */
	TH_STR_R, /* 63 */
	TH_LDR_R, /* 64 */
	TH_STRB_R, /* 65 */
	TH_LDRB_R, /* 66 */
	TH_LDR_PC, /* 67 */
	TH_STR_SP, /* 68 */
	TH_LDR_SP, /* 69 */
	TH_STMIA, /* 70 */
	TH_LDMIA, /* 71 */
	TH_PUSH, /* 72 */
	TH_POP, /* 73 */
	/*! Pseudo op, used in the opcode table */
	TH_UNDEF /* 74 */
	/*! Pseudo instruction, used to add calls during instrumentation */
/*	TH_PSEUDO_CALL */
} t_thumb_opcode;
/* }}} */
#endif
/* vim: set shiftwidth=2 foldmethod=marker : */
