/* Includes {{{ */
#include "thumb_print.h"
#include "thumb_opcodes.h"
#include "arm_registers.h"
#include "../../utils/util_error.h"
#include <string.h>
#include <ctype.h>
/* }}} */
static char *Conditions [] = {"EQ","NE","CS","CC","MI","PL","VS","VC","HI","LS","GE","LT","GT", "LE"};
static char *Regs[] = {"r0","r1","r2","r3","r4","r5","r6","r7","r8","r9","r10","r11","r12","r13", "r14","pc","cpsr"};
static char *Shifts[] = {"LSL","LSR","ASR","ROR","LSL","LSR","ASR","ROR","RRX"};
/*!
 * \todo Document
 *
 * \param data
 * \param outputstring
 *
 * \return void
 */
/* ThumbInstructionPrint {{{ */
void ThumbInstructionPrint(t_ins * data, t_string outputstring)
{
	t_arm_ins * instruction=(t_arm_ins *) data;
	char opcode[8]="";
	char oper1[8]="", oper2[10]="", oper3[30]="";
	char operands[60]="";
	
	if ((INS_TYPE(instruction) == IT_LOAD) || (INS_TYPE(instruction) == IT_STORE))
	{
		sprintf(oper1,"%s,",Regs[ARMINSREGA(instruction)]);
		if (ARMINSFLAGS(instruction) & FL_IMMED) {
			if (ARMINSREGB(instruction) == ARM_REG_R15)
				sprintf(oper3,"%x",ARMINSIMMEDIATE(instruction));
			else {
				sprintf(oper2,"[%s,",Regs[ARMINSREGB(instruction)]);
				if (ARMINSIMMEDIATE(instruction) > 0x9)
					sprintf(oper3,"#%#x]",ARMINSIMMEDIATE(instruction));
				else
					sprintf(oper3,"#%x]",ARMINSIMMEDIATE(instruction));
			}
		}
		else {
			sprintf(oper2,"[%s,",Regs[ARMINSREGB(instruction)]);
			sprintf(oper3,"%s]",Regs[ARMINSREGB(instruction)]);
		}
		if (ARMINSOPCODE(instruction) == ARM_STR)
			sprintf(opcode, "STR");
		if (ARMINSOPCODE(instruction) == ARM_STRB)
			sprintf(opcode, "STRB");
		if (ARMINSOPCODE(instruction) == ARM_STRH)
			sprintf(opcode, "STRH");
		if (ARMINSOPCODE(instruction) == ARM_LDR)
			sprintf(opcode, "LDR");
		if (ARMINSOPCODE(instruction) == ARM_LDRB)
			sprintf(opcode, "LDRB");
		if (ARMINSOPCODE(instruction) == ARM_LDRH)
			sprintf(opcode, "LDRH");
		if (ARMINSOPCODE(instruction) == ARM_LDRSB)
			sprintf(opcode, "LDSB");
		if (ARMINSOPCODE(instruction) == ARM_LDRSH)
			sprintf(opcode, "LDSH");
		sprintf(operands,"%s%s%s",oper1,oper2,oper3);
	}
	if (INS_TYPE(instruction) == IT_SWI) {
		sprintf(opcode, "SWI");
		sprintf(operands,"#%#x",ARMINSIMMEDIATE(instruction));
	}
	if (INS_TYPE(instruction) == IT_BRANCH) {
		sprintf(operands,"%#x",ARMINSIMMEDIATE(instruction));
		if (ARMINSCONDITION(instruction) != ARM_CONDITION_AL)
			sprintf(opcode, "B%s",Conditions[ARMINSCONDITION(instruction)]);
		else if (ARMINSFLAGS(instruction) & FL_IMMED)
			sprintf(opcode, "B");
		if (ARMINSOPCODE(instruction) == ARM_BL)
			sprintf(opcode, "BL");
		if (ARMINSOPCODE(instruction) == ARM_BX) {
			sprintf(opcode, "BX");
			sprintf(operands,"%s",Regs[ARMINSREGB(instruction)]);
		}
	}
	if (INS_TYPE(instruction) == IT_DATAPROC || INS_TYPE(instruction) == IT_MUL)
	{
		sprintf(oper1,"%s,",Regs[ARMINSREGA(instruction)]);
		if (ARMINSFLAGS(instruction) & FL_IMMED) {
			if (ARMINSIMMEDIATE(instruction) > 0x9)
				sprintf(oper3,"#%#x",ARMINSIMMEDIATE(instruction));
			else
				sprintf(oper3,"#%x",ARMINSIMMEDIATE(instruction));
		}
		else
			sprintf(oper3,"%s",Regs[ARMINSREGC(instruction)]);
		if (ARMINSOPCODE(instruction) == ARM_AND)
			sprintf(opcode, "AND");
		if (ARMINSOPCODE(instruction) == ARM_EOR)
			sprintf(opcode, "EOR");
		if (ARMINSOPCODE(instruction) == ARM_ADC)
			sprintf(opcode, "ADC");
		if (ARMINSOPCODE(instruction) == ARM_SBC)
			sprintf(opcode, "SBC");
		if (ARMINSOPCODE(instruction) == ARM_ORR)
			sprintf(opcode, "ORR");
		if (ARMINSOPCODE(instruction) == ARM_BIC)
			sprintf(opcode, "BIC");
		if (ARMINSOPCODE(instruction) == ARM_MVN)
			sprintf(opcode, "MVN");
		if (ARMINSOPCODE(instruction) == ARM_RSB) {
			sprintf(oper3,"%s",Regs[ARMINSREGB(instruction)]);
			sprintf(opcode, "NEG");
		}
		if (ARMINSOPCODE(instruction) == ARM_MUL) {
			sprintf(oper3,"%s",Regs[ARMINSREGB(instruction)]);
			sprintf(opcode, "MUL");
		}
		if (ARMINSOPCODE(instruction) == ARM_TST) {
			sprintf(oper1,"%s,",Regs[ARMINSREGB(instruction)]);
			sprintf(opcode, "TST");
		}
		if (ARMINSOPCODE(instruction) == ARM_CMN) {
			sprintf(oper1,"%s,",Regs[ARMINSREGB(instruction)]);
			sprintf(opcode, "CMN");
		}
		if (ARMINSOPCODE(instruction) == ARM_CMP) {
			sprintf(oper1,"%s,",Regs[ARMINSREGB(instruction)]);
			sprintf(opcode, "CMP");
		}
		if (ARMINSOPCODE(instruction) == ARM_MOV) {
			if (ARMINSSHIFTTYPE(instruction) == ARM_SHIFT_TYPE_NONE)
				sprintf(opcode, "MOV");
			else {
				sprintf(opcode, "%s", Shifts[ARMINSSHIFTTYPE(instruction)]);
				if (ARMINSREGS(instruction) == ARM_REG_NONE) {
					sprintf(oper2,"%s,",Regs[ARMINSREGC(instruction)]);
					sprintf(oper3,"#%x",ARMINSSHIFTLENGTH(instruction));
				}
				else
					sprintf(oper3,"#%x",ARMINSREGS(instruction));
			}
		}
		if (ARMINSOPCODE(instruction) == ARM_ADD) {
			sprintf(opcode, "ADD");
			if (ARMINSREGB(instruction) == ARM_REG_R15) {
				sprintf(opcode, "ADR");
				sprintf(oper3,"%#x", ARMINSIMMEDIATE(instruction));
			}
			else if (ARMINSREGA(instruction) != ARMINSREGB(instruction))
				sprintf(oper2,"%s,",Regs[ARMINSREGB(instruction)]);
		}
		if (ARMINSOPCODE(instruction) == ARM_SUB) {
			sprintf(opcode, "SUB");
			if (ARMINSREGA(instruction) != ARMINSREGB(instruction))
				sprintf(oper2, "%s,",Regs[ARMINSREGB(instruction)]);
		}
		if ((ARMINSOPCODE(instruction) == ARM_SUB || ARMINSOPCODE(instruction) == ARM_ADD) && ARMINSREGA(instruction) != ARMINSREGB(instruction) && ARMINSFLAGS(instruction) & FL_S && ARMINSFLAGS(instruction) & FL_IMMED)
		{ /* ifimmediate-value = 0 --> move instruction */
			sprintf(opcode, "MOV");
			oper2[strlen(oper2)-1] = '\0';
			oper3[0] = '\0';
		}
		sprintf(operands,"%s%s%s",oper1,oper2,oper3);
	}
	if ((INS_TYPE(instruction) == IT_LOAD_MULTIPLE) || (INS_TYPE(instruction) == IT_STORE_MULTIPLE))
	{
		char reglist[65] = "{";
		int i, teller;

		for (i = 0, teller = 0; i < 16; i++) {
			if (ARMINSIMMEDIATE(instruction) & (1 << i)) {
				if (++teller == 1)
					strcat(reglist,Regs[i]);
			}
			else {
				if (teller > 2)
					strcat(reglist,"-");
				if (teller == 2)
					strcat(reglist,",");
				if (teller > 1)
					strcat(reglist,Regs[i-1]);
				if (teller > 0)
					strcat(reglist,",");
				teller = 0;
			}
		}
		if (teller == 1)
			strcat(reglist,","); /* to avoid removal of 'c' in 'pc' with pop instruction */
		reglist[strlen(reglist) - 1] = '}';
		if (ARMINSREGB(instruction) == ARM_REG_R13) {
			(ARMINSOPCODE(instruction) == ARM_STM) ? sprintf(opcode,"PUSH") : sprintf(opcode,"POP");
			sprintf(operands,"%s",reglist);
		}
		else {
			(ARMINSOPCODE(instruction) == ARM_STM) ? sprintf(opcode,"STMIA") : sprintf(opcode,"LDMIA");
			sprintf(oper2,"%s!",Regs[ARMINSREGB(instruction)]);
			sprintf(operands,"%s,%s",oper2,reglist);
		}
	}
	sprintf(outputstring,"%-10s %-69s",opcode,operands);
	StringTrim(outputstring);
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker : */
