/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef I386_ACTIONS_H
#define I386_ACTIONS_H
/* Includes {{{ */ 
#include "../../kernel/diablo_action.h"
#include "i386_actions.h"
#include "i386_assemble.h"
#include "i386_constant_propagation.h"
#include "i386_deflowgraph.h"
#include "i386_disassemble.h"
#include "i386_emulation.h"
#include "i386_flowgraph.h"
#include "i386_map.h"
#include "i386_print.h"
#include "i386_registers.h"
#include "i386_datamap.h"
#include "i386_write.h"
#include "i386_factor.h"
#include "i386_stack.h"
#include "i386_copy_propagation.h"
#ifdef STEGO
#include "../../watermark/i386_superoptimizer.h"
#endif
/* }}} */
const t_action_struct i386_actions={
  NULL,                                 /* Group Info */
  I386ReadMap,                      /* Read Map */
  NULL,                                 /* Disassemble */
  I386DisassembleSection,           /* Disassemble Section */
  NULL,                                 /* Flowgraph */
  I386FlowgraphSection,             /* Flowgraph Section */
  I386MakeAddressProducers,	    /* Make address producers */
  NULL,                                 /* Deflowgraph */
  I386DeflowgraphSection,           /* Deflowgraph Section */
  NULL,                                 /* Assemble */ 
  I386AssembleSection,              /* Assemble Section */
  I386Write,                                 /* Write Object */

  /* Instruction related i386 callbacks*/
  I386InstructionPrint,             /* Print Instruction */
  I386InstructionMakeNoop,          /* Make a noop from the instruction */

  I386RegisterName,                 /* return the register name (from a static array of registers )*/
  I386InstructionEmulator,          /* Instruction Emulation */
  I386EdgePropagator,               /* */
  I386InstructionUnconditionalizer, /* */
  I386InstructionsConstantOptimizer, /* */
  NULL,                             /* */
  I386ConvertToBlocksSec,             /* */
  NULL,                             /* */
  NULL,
  I386InsHasSideEffect,
  /* Factorisation related i386 callbacks*/
  I386BblCmpExtReg,
  I386BblAbstract,
  I386FunComputeStackSavedRegisters,
  I386CopyInstructionPropagator,    /* Propagate copy analysis information */
  I386IsSyscallExit,
  I386InsIsControlTransfer,
  /* 30 */  I386InsIsIndirectCall,
  /* 31 */ I386FunIsGlobal,
  NULL,
  NULL,
  
#ifdef STEGO
  I386InsFindAlternatives,
  I386InsCompare,
  I386MorphInstructionTo,
  I386MarkLeadersThatCanBeGuessed,
#endif
};
#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
