/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "i386_constant_propagation.h"
#include "i386_constantoptimizations.h"
#include "i386_registers.h"
#include "i386_instructions.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_iterators.h"


void i386TestAndSetConditionAllPaths(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
  ProcStateDup(*state_true,state);
}

void i386TestAndSetConditionAlways(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = state;
  *state_false = NULL;
}

void i386TestAndSetConditionNever(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_false  = state;
  *state_true = NULL;
}

void AlphaTestAndSetConditionSetAllToBot(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
  ProcStateSetAllBot(*state_true,AllRegs());
}


void i386TestAndSetConditionTailCall(t_cfg_edge * edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  t_regset regs = RegsetNew();
  t_reg reg;
  t_procstate * corr_state = EDGE_CORR(edge)->procstate;
  t_lattice_level level;
  register_content value;
  t_reloc * reloc;
  *state_true  = ProcStateNewDup(state);
  *state_false = state;
  RegsetDup(regs,FUN_REGS_CHANGED(BBL_FUN(CFG_EDGE_HEAD(edge))));
  RegsetSetInvers(regs);
  
  REGSET_FOREACH_REG(regs,reg)
    { 
      level= ProcStateGetReg(corr_state,reg, &value);
      if(level==BOT)
	{
	  ProcStateSetRegBot(*state_true,reg);
	}
      else if(level==TOP)
	{
	  ProcStateSetRegTop(*state_true,reg);
	}
      else
	{
	  ProcStateSetReg(*state_true,reg,value);
	}

      level= ProcStateGetTag(corr_state,reg, &reloc);

      if(level==BOT)
	{
	  ProcStateSetTagBot(*state_true,reg);
	}
      else if(level==TOP)
	{
	  ProcStateSetTagTop(*state_true,reg);
	}
      else
	{
	  ProcStateSetTag(*state_true,reg,reloc);
	}
    }
}



void I386EdgePropagator(t_cfg_edge * edge, t_ins * ins)
{
  t_uint16 edgetype  = EDGE_CAT(edge);
  
  if (IsHellNode(EDGE_HEAD(edge)))
    {
      edge->testcondition = &i386TestAndSetConditionAllPaths;
      return;
    }
  
  switch (edgetype)
    {
    case ET_JUMP:
      if (CFG_EDGE_FLAGS(edge) & EF_FROM_SWITCH_TABLE)
	{
	  edge->testcondition = &i386TestAndSetConditionAllPaths;
	  break;
	}
    case ET_CALL:
    case ET_IPJUMP:
      edge->testcondition = &i386TestAndSetConditionAllPaths;
      break;
    case ET_IPFALLTHRU:
    case ET_FALLTHROUGH:
    case ET_UNKNOWN:
    case ET_IPUNKNOWN:
      edge->testcondition = &i386TestAndSetConditionAlways;
      break;
      
    case ET_COMPENSATING:
      {
	t_cfg_edge * j_edge = EDGE_CORR(edge);
	if (BBL_IS_FIRST(CFG_EDGE_TAIL(j_edge)) && EDGE_CAT(j_edge) == ET_IPJUMP && !IsHellNode(CFG_EDGE_TAIL(j_edge)))
	  {
	    /*	    DiabloPrintArch(stdout,cfg->sec->obj->description,"I found what I think is a tail call @E @A\n!",j_edge,RegsetNewInvers(FUN_REGS_CHANGED(BBL_FUN(EDGE_HEAD(edge))))); */
	    edge->testcondition = &i386TestAndSetConditionTailCall;
	    break;
	    
	  }
      }
    case ET_SWITCH:
    case ET_RETURN:
    case ET_IPSWITCH:
      edge->testcondition = &i386TestAndSetConditionAllPaths;
      break;
      
      /* Over een FUNLINK wordt NOOIT gepropageerd */    
    case ET_FUNLINK:
    case ET_IPFUNLINK:
    case ET_SWI:
      edge->testcondition = &i386TestAndSetConditionNever;
      break;
    case ET_SWILINK:
    case ET_IPSWILINK:
      edge->testcondition = &AlphaTestAndSetConditionSetAllToBot;
      break;

    default:
      FATAL(("MipsSetConditionToProp invalid type %lx  %d\n",edgetype,edgetype));
      break;
    }
}
/*
{
  if (IsHellNode(EDGE_HEAD(edge)))
  {
    edge->testcondition = &i386TestAndSetConditionAllPaths;
    return;
  }
  edge->testcondition = &i386TestAndSetConditionAllPaths;
  
}*/

/* unconditionalize conditional instructions of which constant propagation has determined that
 * they will always be executed */
t_bool I386InstructionUnconditionalizer(t_ins * ins)
{
 return FALSE;
}

t_bool I386InstructionsConstantOptimizer(t_ins * ins, t_procstate * prev_state, t_procstate * next_state)
{
 return TRUE;
}

/* vim: set shiftwidth=2: */
