/******************************************************************************
 * Copyright 2001,2002,2003:  {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/

#include "i386_instructions.h"
#include "i386_registers.h"
#include "i386_assemble.h"
#include "../../kernel/diablo_copy_analysis.h"
#include "../../kernel/diablo_iterators.h"
#include "../../utils/util_error.h"
#include "../../diablo_options.h"
#include "../../kernel/diablo_print.h"

/*#define I386COPYDEBUG */

void I386ReplaceCopiedByCopyBackward(t_ins * prev_ins, t_reg copy, t_reg copied){

  while(prev_ins){
    if(RegsetIn(INS_REGS_DEF(prev_ins),copied)&&!RegsetIn(INS_REGS_USE(prev_ins),copied))
    {
      if(I386_OP_BASE(I386_INS_DEST(prev_ins)) == copied)
	I386_OP_BASE(I386_INS_DEST(prev_ins)) = copy;
      else {
	DiabloPrint(stdout,"%d %d @I\n",copy, copied, prev_ins);
	FATAL(("this shouldn't happen"));
      } 
      INS_REGS_USE(prev_ins)=I386InsUsedRegisters(prev_ins);;
      INS_REGS_DEF(prev_ins)=I386InsDefinedRegisters(prev_ins);;
      return;
    }
    else
    {
      if(I386_OP_BASE(I386_INS_DEST(prev_ins))== copied)
	I386_OP_BASE(I386_INS_DEST(prev_ins)) = copy;
      if(I386_OP_BASE(I386_INS_SOURCE1(prev_ins))== copied)
	I386_OP_BASE(I386_INS_SOURCE1(prev_ins)) = copy;
      if(I386_OP_BASE(I386_INS_SOURCE2(prev_ins))== copied)
	I386_OP_BASE(I386_INS_SOURCE2(prev_ins)) = copy;
    }
    INS_REGS_USE(prev_ins)=I386InsUsedRegisters(prev_ins);;
    INS_REGS_DEF(prev_ins)=I386InsDefinedRegisters(prev_ins);;
      
    prev_ins = INS_PREV(prev_ins);
  }
}

void I386ReplaceCopiedByCopyForward(t_ins * next_ins, t_reg copy, t_reg copied){

  while(next_ins){
    if(RegsetIn(INS_REGS_DEF(next_ins),copied)&&!RegsetIn(INS_REGS_USE(next_ins),copied))
    {
      return;
    }
    else
    {
      if(I386_OP_BASE(I386_INS_DEST(next_ins))== copied)
	I386_OP_BASE(I386_INS_DEST(next_ins)) = copy;
      if(I386_OP_BASE(I386_INS_SOURCE1(next_ins))== copied)
	I386_OP_BASE(I386_INS_SOURCE1(next_ins)) = copy;
      if(I386_OP_BASE(I386_INS_SOURCE2(next_ins))== copied)
	I386_OP_BASE(I386_INS_SOURCE2(next_ins)) = copy;
    }
    INS_REGS_USE(next_ins)=I386InsUsedRegisters(next_ins);;
    INS_REGS_DEF(next_ins)=I386InsDefinedRegisters(next_ins);;
      
    next_ins = INS_NEXT(next_ins);
  }
}

t_bool I386CopiedIsReplacableByCopyInIns(t_ins * ins, t_reg copy, t_reg copied)
{
/*  DiabloPrint(stdout,"=><=@I, %d, %d, %d, %d\n", ins, copied, I386_OP_BASE(I386_INS_DEST(ins)),I386_OP_BASE(I386_INS_SOURCE1(ins)),I386_OP_BASE(I386_INS_SOURCE2(ins))); */
  if(RegsetIn(INS_REGS_USE(ins),copied)||RegsetIn(INS_REGS_DEF(ins),copied)){
    /* implicit */
    if(!(
	  (I386_OP_TYPE(I386_INS_DEST(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_DEST(ins))==copied)
	  ||
	  (I386_OP_TYPE(I386_INS_SOURCE1(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_SOURCE1(ins))==copied)
	  ||
	  (I386_OP_TYPE(I386_INS_SOURCE2(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_SOURCE2(ins))==copied)
	)){
      	return FALSE;
    }
    /* explicit, check if can be replaced */
    else{
      t_i386_opcode_entry * forms[10];   /* this is certainly large enough */
      t_uint8 nfitting;
      
      if(I386_OP_TYPE(I386_INS_DEST(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_DEST(ins))==copied)
	I386_OP_BASE(I386_INS_DEST(ins))=copy;
      if(I386_OP_TYPE(I386_INS_SOURCE1(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_SOURCE1(ins))==copied)
	I386_OP_BASE(I386_INS_SOURCE1(ins))=copy;
      if(I386_OP_TYPE(I386_INS_SOURCE2(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_SOURCE2(ins))==copied)
	I386_OP_BASE(I386_INS_SOURCE2(ins))=copy;
      
      nfitting = I386GetPossibleEncodings(ins,forms);

      if(I386_OP_TYPE(I386_INS_DEST(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_DEST(ins))==copy)
	I386_OP_BASE(I386_INS_DEST(ins))=copied;
      if(I386_OP_TYPE(I386_INS_SOURCE1(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_SOURCE1(ins))==copy)
	I386_OP_BASE(I386_INS_SOURCE1(ins))=copied;
      if(I386_OP_TYPE(I386_INS_SOURCE2(ins))==i386_optype_reg&&I386_OP_BASE(I386_INS_SOURCE2(ins))==copy)
	I386_OP_BASE(I386_INS_SOURCE2(ins))=copied;

      if(nfitting==0) return FALSE;
    } 
  }
      
  return TRUE;
}

/*
 * This function checks if the copied register can be replaced by the copy in the preceeding instructions
 * */ 
t_bool I386CopiedIsReplacableByCopyBackward(t_ins * prev_ins, t_reg copy, t_reg copied){
  
  while(prev_ins){

    if(RegsetIn(INS_REGS_USE(prev_ins),copy))
      return FALSE;
    if(RegsetIn(INS_REGS_DEF(prev_ins),copy)){
      return FALSE;
    }
    
    if(!I386CopiedIsReplacableByCopyInIns(prev_ins, copy, copied))
      return FALSE;
    
    if(RegsetIn(INS_REGS_DEF(prev_ins),copied)&&!RegsetIn(INS_REGS_USE(prev_ins),copied))
      return TRUE;
    
    prev_ins = INS_PREV(prev_ins);
  }
  
  return FALSE;
  /*
  BBL_FOREACH_PRED_EDGE(INS_BBL(prev_ins),prev_edge){
    t_bbl * prev_bbl = T_CFG_HEAD(prev_edge);
    
    if(IsHellNode(prev_bbl))
      return FALSE;

    if(!I386CopiedIsReplacableByCopyBackward(BBL_INS_LAST(prev_bbl),copy,copied))
      return FALSE; 

    BBL_FOREACH_SUCC_EDGE(prev_bbl, next_edge){
      t_bbl * next_bbl = CFG_EDGE_TAIL(next_edge);
    
      if(IsHellNode(next_bbl))
	return FALSE;
      if(!I386CopiedIsReplacableForward(BBL_INS_FIRST(next_bbl), copy, copied)) 
	return FALSE;

    }
  }
  return TRUE;*/
}

/*
 * This function checks if the copied register can be replaced by the copy in the following instructions 
 * */

t_bool I386CopiedIsReplacableByCopyForward(t_ins * ins, t_reg copy, t_reg copied){
  while(ins){
    if(!RegsetIn(InsRegsLiveBefore(ins), copied))
      return TRUE;

    if(RegsetIn(INS_REGS_DEF(ins),copy)){
      return FALSE;
    }

    if(!I386CopiedIsReplacableByCopyInIns(ins, copy, copied))
      return FALSE;
 
    ins = INS_NEXT(ins);
  }
  return FALSE;
/*  
  BBL_FOREACH_SUCC_EDGE(INS_BBL(ins),edge){
    t_bbl * next_bbl = CFG_EDGE_TAIL(edge);
    
    if(IsHellNode(next_bbl))
      return FALSE;
    if(!I386CopiedIsReplacableForward(BBL_INS_FIRST(next_bbl), copy, copied)) 
      return FALSE;
  }
  return TRUE;
  */
}


static t_uint32 killed=0;
/* This optimization searches for a copy instruction in which the copied register is no longer alive after the 
 * copy instruction. In this case, it is sometimes possible not to use the copied register, but to assign its 
 * value immediately to the copy where it is declared.
 * */

void I386BackwardCopyPropagation(t_section * sec){
  t_cfg *cfg = T_CFG(sec->data);
  t_reg copy, copied;
  t_bbl * bbl;
  t_ins * ins;
  t_i386_operand * dest;
  t_i386_operand * src1;

  InitNodeMarking();
  InitEdgeMarking();

  CFG_FOREACH_BBL(cfg,bbl){
    BBL_FOREACH_INS(bbl,ins){
      dest = I386_INS_DEST(ins);
      src1 = I386_INS_SOURCE1(ins);
      if(I386_INS_OPCODE(ins)==I386_MOV){
	if (I386_OP_TYPE(dest) == i386_optype_reg && I386IsGeneralPurposeReg(I386_OP_BASE(dest))){
	  if (I386_OP_TYPE(src1) == i386_optype_reg && I386IsGeneralPurposeReg(I386_OP_BASE(src1))){
	    copy = I386_OP_BASE(dest);
	    copied = I386_OP_BASE(src1);
	    if(I386CopiedIsReplacableByCopyBackward(INS_PREV(ins), copy, copied)&&I386CopiedIsReplacableByCopyForward(INS_NEXT(ins), copy, copied)){

#ifdef I386COPYDEBUG
	      DiabloPrint(stdout,"voor---\n@I:\n@iB\n---\n",ins,INS_BBL(ins));
#endif
	      I386ReplaceCopiedByCopyBackward(INS_PREV(ins), copy, copied);
	      I386ReplaceCopiedByCopyForward(INS_NEXT(ins), copy, copied);
	      InsKill(ins);
#ifdef I386COPYDEBUG
	      DiabloPrint(stdout,"na---\n@I:\n@iB\n---\n",ins,INS_BBL(ins));
#endif
	      killed++;
	    }
	  }
	}
      }
    }
  }
  printf("%d victims\n",killed);
}

/* {{{ instruction propagator for copy analysis */
void I386CopyInstructionPropagator(t_ins * ins, t_equations eqs)
{
  t_reg reg;
  t_i386_operand * dest = I386_INS_DEST(ins);
  t_i386_operand * src1 = I386_INS_SOURCE1(ins);
  t_regset tmp;
  /*t_i386_operand * src2 = I386_INS_SOURCE2(ins); */

  /*DiabloPrint(stdout,"copyprop @I\n", ins); */

  switch (I386_INS_OPCODE(ins))
  {
    case I386_MOV:
      if (I386_OP_TYPE(dest) == i386_optype_reg 
	  && I386_OP_REGMODE(dest) == i386_regmode_full32 
	  && I386IsGeneralPurposeReg(I386_OP_BASE(dest)))
      {
	if (I386_OP_TYPE(src1) == i386_optype_reg 
	    && I386_OP_REGMODE(src1) == i386_regmode_full32 
	    && I386IsGeneralPurposeReg(I386_OP_BASE(src1)))
	{
	  /* mov %reg1,%reg2 */
	  EquationsAdd(eqs,I386_OP_BASE(dest),I386_OP_BASE(src1),0,NULL,NULL);
	  return;
	}
	else if (I386_OP_TYPE(src1) == i386_optype_imm)
	{
	  /* mov $imm,%reg */
	  if (I386_OP_FLAGS(src1) & I386_OPFLAG_ISRELOCATED)
	    EquationsAdd(eqs,I386_OP_BASE(dest),null_register,0,RELOCATABLE_REFERS_TO(ins)->rel,NULL);
	  else
	    EquationsAdd(eqs,I386_OP_BASE(dest),null_register,0-I386_OP_IMMEDIATE(src1),NULL,NULL);
	  return;
	}
      }
      break;
    case I386_ADD:
      if (I386_OP_TYPE(dest) == i386_optype_reg 
	  && I386_OP_REGMODE(dest) == i386_regmode_full32 
	  && I386IsGeneralPurposeReg(I386_OP_BASE(dest)))
      {
	if (I386_OP_TYPE(src1) == i386_optype_imm
	    && !(I386_OP_FLAGS(src1) & I386_OPFLAG_ISRELOCATED))
	{
	  /* add $imm,%reg */
	  EquationsAdd(eqs,I386_OP_BASE(dest),I386_OP_BASE(dest),-I386_OP_IMMEDIATE(src1),NULL,NULL);
	  return;
	}
      }
      break;
    case I386_SUB:
      if (I386_OP_TYPE(dest) == i386_optype_reg 
	  && I386_OP_REGMODE(dest) == i386_regmode_full32 
	  && I386IsGeneralPurposeReg(I386_OP_BASE(dest)))
      {
	if (I386_OP_TYPE(src1) == i386_optype_imm
	    && !(I386_OP_FLAGS(src1) & I386_OPFLAG_ISRELOCATED))
	{
	  /* sub $imm,%reg */
	  EquationsAdd(eqs,I386_OP_BASE(dest),I386_OP_BASE(dest),I386_OP_IMMEDIATE(src1),NULL,NULL);
	  return;
	}
      }
      break;
    case I386_LEA:
      if (I386_OP_INDEX(src1) == I386_REG_NONE)
      {
	if (I386_OP_BASE(src1) == I386_REG_NONE)
	{
	  /* lea addr,%reg */
	  if (I386_OP_FLAGS(src1) & I386_OPFLAG_ISRELOCATED)
	    EquationsAdd(eqs,I386_OP_BASE(dest),null_register,0,RELOCATABLE_REFERS_TO(ins)->rel,NULL);
	  else
	    EquationsAdd(eqs,I386_OP_BASE(dest),null_register,0-I386_OP_IMMEDIATE(src1),NULL,NULL);
	}
	else
	{
	  /* lea offset(%reg1),%reg2 */
	  if (I386_OP_FLAGS(src1) & I386_OPFLAG_ISRELOCATED)
	    EquationsAdd(eqs,I386_OP_BASE(dest),I386_OP_BASE(src1),0,RELOCATABLE_REFERS_TO(ins)->rel,NULL);
	  else
	    EquationsAdd(eqs,I386_OP_BASE(dest),I386_OP_BASE(src1),0-I386_OP_IMMEDIATE(src1),NULL,NULL);
	}
	return;
      }
      break;
    case I386_INSB: case I386_INSD:
    case I386_OUTSB: case I386_OUTSD:
    case I386_CMPSB: case I386_CMPSD:
    case I386_SCASB: case I386_SCASD:
    case I386_MOVSB: case I386_MOVSD:
    case I386_LODSB: case I386_LODSD:
    case I386_STOSB: case I386_STOSD:
      /* we don't know if the D flag is set, so we don't know whether 
       * %esi and %edi increase or decrease */
      break;
    case I386_POP:
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,-4,NULL,NULL);
      /* invalidate register into which the stack is popped */
      if (I386IsGeneralPurposeReg(I386_OP_BASE(I386_INS_DEST(ins))))
	EquationsInvalidate(eqs,I386_OP_BASE(I386_INS_DEST(ins)));
      return;
    case I386_POPF:
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,-4,NULL,NULL);
      return;
    case I386_POPA:
      /* invalidate all registers apart from %esp, to which 32 is added */
      RegsetDup(tmp,INS_OBJECT(ins)->description->int_registers);
      RegsetSetIntersect(tmp,INS_REGS_DEF(ins));
      RegsetSubReg(tmp,I386_REG_ESP);
      REGSET_FOREACH_REG(tmp,reg)
	EquationsInvalidate(eqs,reg);
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,-32,NULL,NULL);
      break;
    case I386_PUSH:
    case I386_PUSHF:
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,+4,NULL,NULL);
      return;
    case I386_PUSHA:
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,+32,NULL,NULL);
      return;
    case I386_LEAVE:
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_EBP,0,NULL,NULL);
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,-4,NULL,NULL);
      EquationsInvalidate(eqs,I386_REG_EBP);
      return;
    case I386_CALL:
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,+4,NULL,NULL);
      return;
    case I386_RET:
      EquationsAdd(eqs,I386_REG_ESP,I386_REG_ESP,-4,NULL,NULL);
      return;
    default:
      /* keep the compiler happy */
      break;
  }

  RegsetDup(tmp,INS_OBJECT(ins)->description->int_registers);
  RegsetSetIntersect(tmp,INS_REGS_DEF(ins));
  REGSET_FOREACH_REG(tmp,reg)
    EquationsInvalidate(eqs,reg);
} /* }}} */

/* vim: set shiftwidth=2 foldmethod=marker : */
