/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "i386_instructions.h"
#include "i386_registers.h"
#include "../../kernel/diablo_constprop.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_print.h"
#include <string.h>
#include <strings.h>

void I386InstructionEmulator(t_ins * ins, t_procstate * state)
{
  register_content reg_cont_dest0, reg_cont_dest1;
  INS_ATTRIB(ins)|=IF_EXECED;

  bzero(&reg_cont_dest0,sizeof(register_content));
  bzero(&reg_cont_dest1,sizeof(register_content));
  GenericSet_uint32(reg_cont_dest0.i,0);
  GenericSet_uint32(reg_cont_dest1.i,1);

  switch (I386_INS_OPCODE(ins))
  {
    case I386_AAA: 	/* 3 */
    case I386_AAD:
    case I386_AAM:
    case I386_AAS:
    case I386_ADC:	/* 7 */
/*      DiabloPrint(stdout,"@I\n",ins);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    case I386_ADD:	/* 8 */
      if((I386_INS_DEST(ins))->type == i386_optype_reg)
      {
	t_reg reg_s1, reg_s2;
	t_uint32 u_src1,u_src2;
	t_int32 temp32;
	register_content reg_cont_src1, reg_cont_src2, reg_cont_dest;

	if((I386_INS_SOURCE1(ins))->type == i386_optype_reg)
	{
	  reg_s1 = I386_OP_BASE(I386_INS_SOURCE1(ins));
	  reg_s2 = I386_OP_BASE(I386_INS_DEST(ins));

	  if(ProcStateGetReg(state,reg_s1,&reg_cont_src1)!=0||
	      ProcStateGetReg(state,reg_s2,&reg_cont_src2)!=0)
	  {
/*            DiabloPrint(stdout,"@I\n",ins);*/
	    ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*            DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
	    break;
	  }
	  u_src1=G_T_UINT32(reg_cont_src1.i);
	  switch (I386_OP_REGMODE(I386_INS_SOURCE1(ins)))
	  {
	    case i386_regmode_full32:
	      temp32=(t_int32)u_src1;
	      break;
	    case i386_regmode_lo16:
	      temp32=(t_int32)((t_int16)u_src1);
	      break;
	    case i386_regmode_lo8:
	      temp32=(t_int32)((t_int8)u_src1);
	      break;
	    default:
	      FATAL(("Implossible!"));
	      break;
	  }
	}
	else if((I386_INS_SOURCE1(ins))->type == i386_optype_imm)
	{

	  reg_s2 = I386_OP_BASE(I386_INS_DEST(ins));
	  u_src1=I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins));
	  switch (I386_OP_IMMEDSIZE(I386_INS_SOURCE1(ins)))
	  {
	    case 4:
	      temp32=(t_int32)u_src1;
	      break;
	    case 2:
	      temp32=(t_int32)((t_int16)u_src1);
	      break;
	    case 1:
	      temp32=(t_int32)((t_int8)u_src1);
	      break;
	    default:
	      FATAL(("Implossible!"));
	      break;
	  }
	}
	else 
	{
/*          DiabloPrint(stdout,"@I\n",ins);*/
	  ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*          DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
	  break;
	}

	u_src2=G_T_UINT32(reg_cont_src2.i);
	bzero(&reg_cont_dest,sizeof(register_content));
	switch (I386_OP_REGMODE(I386_INS_DEST(ins)))
	{
	  case i386_regmode_full32:
	    temp32+=(t_int32)u_src2;
	    break;
	  case i386_regmode_lo16:
	    temp32+=(t_int32)((t_int16)u_src2);
	    break;
	  case i386_regmode_lo8:
	    temp32+=(t_int32)((t_int8)u_src2);
	    break;
	  default:
	    FATAL(("Implossible!"));
	    break;
	}
	GenericSet_uint32(reg_cont_dest.i,(t_uint32)temp32);
	ProcStateSetReg(state,reg_s2,reg_cont_dest);
	if (RELOCATABLE_REFERS_TO(ins))
	  ProcStateSetTag(state,reg_s2,RELOCATABLE_REFERS_TO(ins)->rel);
/*        DiabloPrint(stdout,"ADD: @I s:%x s2:%x d:%x\n",ins,u_src1,u_src2,temp32);*/
	break;
      }
/*      DiabloPrint(stdout,"@I\n",ins);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    case I386_AND:	/* 9 */
    case I386_ARPL:
    case I386_BOUND:
    case I386_BSF:	/* 12 */
    case I386_BSR:	/* 13 */
    case I386_BSWAP:
    case I386_BT:
    case I386_BTC:
    case I386_BTR:
    case I386_BTS:
    case I386_CALL:	/* 19 */
    case I386_CALLF:
    case I386_CDQ:	/* 21 */
    case I386_CLC:
    case I386_CLD:	/* 23 */
    case I386_CLI:
    case I386_CLTS:
    case I386_CMC:
    case I386_CMOVcc:	/* 27 */
    case I386_CMP:	/* 28 */
    case I386_CMPSB:	/* 29 */
    case I386_CMPSD:
    case I386_CMPXCHG:
    case I386_CPUID:
    case I386_CWDE:
    case I386_DAA:
    case I386_DAS:
    case I386_DEC:	/* 36 */
    case I386_DIV:	/* 37 */
    case I386_ENTER:
    case I386_F2XM1:
    case I386_FABS:
    case I386_FADD:
    case I386_FADDP:
    case I386_FBLD:
    case I386_FBSTP:
    case I386_FCHS:
    case I386_FCLEX:
    case I386_FCMOVcc:
    case I386_FCOM:
    case I386_FCOMI:
    case I386_FCOMIP:
    case I386_FCOMP:
    case I386_FCOMPP:
    case I386_FCOS:
    case I386_FDECSTP:
    case I386_FDIV:
    case I386_FDIVP:
    case I386_FDIVR:
    case I386_FDIVRP:
    case I386_FFREE:
    case I386_FIADD:
    case I386_FICOM:
    case I386_FICOMP:
    case I386_FIDIV:
    case I386_FIDIVR:
    case I386_FILD:
    case I386_FIMUL:
    case I386_FINCSTP:
    case I386_FINIT:
    case I386_FIST:
    case I386_FISTP:
    case I386_FISUB:
    case I386_FISUBR:
    case I386_FLD:	/* 73 */
    case I386_FLD1:
    case I386_FLDCW:
    case I386_FLDENV:
    case I386_FLDL2E:
    case I386_FLDL2T:
    case I386_FLDLG2:
    case I386_FLDLN2:
    case I386_FLDPI:
    case I386_FLDZ:	/* 82 */
    case I386_FMUL:
    case I386_FMULP:
    case I386_FNOP:
    case I386_FPATAN:
    case I386_FPREM:
    case I386_FPREM1:
    case I386_FPTAN:
    case I386_FRNDINT:
    case I386_FRSTOR:
    case I386_FSAVE:
    case I386_FSCALE:
    case I386_FSIN:
    case I386_FSINCOS:
    case I386_FSQRT:
    case I386_FST:
    case I386_FSTCW:
    case I386_FSTENV:
    case I386_FSTP:
    case I386_FSTSW:	/* 101 */
    case I386_FSUB:	
    case I386_FSUBP:
    case I386_FSUBR:
    case I386_FSUBRP:
    case I386_FTST:
    case I386_FUCOM:
    case I386_FUCOMI:
    case I386_FUCOMIP:
    case I386_FUCOMP:
    case I386_FUCOMPP:	/* 111 */
    case I386_FXAM:
    case I386_FXCH:	/* 113 */
    case I386_FXTRACT:
    case I386_FYL2X:
    case I386_FYL2XP1:
    case I386_HLT:	/* 117 */
    case I386_IDIV:	/* 118 */
    case I386_IMUL:	/* 119 */
    case I386_IMULexp1:	/* 120 */
    case I386_IMULexp2:	/* 121 */
    case I386_IN:
    case I386_INC:	/* 123 */
    case I386_INSB:
    case I386_INSD:
    case I386_INT:	/* 126 */
    case I386_INT3:
    case I386_INTO:
    case I386_INVD:
    case I386_INVLPG:
    case I386_IRET:
    case I386_JECXZ:	/* 132 */
    case I386_JMP:	/* 133 */
    case I386_JMPF:
    case I386_Jcc:	/* 135 */
    case I386_LAHF:
    case I386_LAR:
    case I386_LDS:
    case I386_LEA:	/* 139 */
    case I386_LEAVE:
    case I386_LES:
    case I386_LFS:
    case I386_LGDT:
    case I386_LGS:
    case I386_LIDT:
    case I386_LLDT:
    case I386_LMSW:
    case I386_LODSB:
    case I386_LODSD:
    case I386_LOOP:
    case I386_LOOPNZ:
    case I386_LOOPZ:
    case I386_LSL:
    case I386_LSS:
    case I386_LTR:
/*      DiabloPrint(stdout,"@I\n",ins);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    case I386_MOV:	/* 156 */
      if((I386_INS_DEST(ins))->type == i386_optype_reg)
      {
	t_reg /*reg_s1,*/ reg_s2;
	t_uint32 u_src1/*,u_src2*/;
	t_int32 temp32;
	register_content /*reg_cont_src1, reg_cont_src2,*/ reg_cont_dest;

	if((I386_INS_SOURCE1(ins))->type == i386_optype_imm)
	{

	  reg_s2 = I386_OP_BASE(I386_INS_DEST(ins));
	  u_src1=I386_OP_IMMEDIATE(I386_INS_SOURCE1(ins));
	  switch (I386_OP_IMMEDSIZE(I386_INS_SOURCE1(ins)))
	  {
	    case 4:
	      temp32=(t_int32)u_src1;
	      break;
	    case 2:
	      temp32=(t_int32)((t_int16)u_src1);
	      break;
	    case 1:
	      temp32=(t_int32)((t_int8)u_src1);
	      break;
	    default:
	      FATAL(("Implossible!%d\n",I386_OP_IMMEDSIZE(I386_INS_SOURCE1(ins))));
	      break;
	  }

	  bzero(&reg_cont_dest,sizeof(register_content));
	  switch (I386_OP_REGMODE(I386_INS_DEST(ins)))
	  {
	    case i386_regmode_full32:
	      GenericSet_uint32(reg_cont_dest.i,(t_uint32)temp32);
	      break;
	    case i386_regmode_lo16:
	      GenericSet_uint32(reg_cont_dest.i,(((t_uint32)temp32) & 0x0000ffff));
	      break;
	    case i386_regmode_lo8:
	      GenericSet_uint32(reg_cont_dest.i,(((t_uint32)temp32) & 0x000000ff));
	      break;
	    case i386_regmode_hi8:
	      GenericSet_uint32(reg_cont_dest.i,(((t_uint32)temp32) & 0x0000ff00) >> 8);
	      break;
	    default:
	      FATAL(("Impossible! (regmode == %d) for @I", I386_OP_REGMODE(I386_INS_DEST(ins)), ins));
	      break;
	  }

	  ProcStateSetReg(state,reg_s2,reg_cont_dest);
	  if (RELOCATABLE_REFERS_TO(ins))
	    ProcStateSetTag(state,reg_s2,RELOCATABLE_REFERS_TO(ins)->rel);
/*          DiabloPrint(stdout,"MOV: @I s:%x s2:%x d:%x\n",ins,u_src1,u_src2,temp32);*/
	  break;
	}
	else if ((I386_INS_SOURCE1(ins))->type == i386_optype_mem)
	{
	  /*INS_SECTION(ins)
	  TODO Ophalen als het in Read-Only geheugen zit! MAAR: relocs staan nog niet juist!
	 */ 
/*          DiabloPrint(stdout,"@I\n", ins);*/
/*          DiabloPrint(stdout,"Flags:%x\n",(I386_INS_SOURCE1(ins))->flags);*/
/*          if(INS_REFERS_TO(ins)!=NULL)*/
/*            DiabloPrint(stdout,"Reloc: @R\n",INS_REFERS_TO(ins)->rel);*/
	  /*(I386_INS_SOURCE1(ins))->type  */
	}
      }
/*      DiabloPrint(stdout,"@I SourceType:%d\n",ins,(I386_INS_SOURCE1(ins))->type);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    case I386_MOVNTI:
    case I386_MOVSB:	/* 158 */
    case I386_MOVSD:	/* 159 */
    case I386_MOVSX:	/* 160 */
    case I386_MOVZX:	/* 161 */
    case I386_MUL:	/* 162 */
    case I386_NEG:	/* 163 */
    case I386_NOP:	/* 164 */
    case I386_NOT:	/* 165 */
    case I386_OR:	/* 166 */
    case I386_OUT:
    case I386_OUTSB:
    case I386_OUTSD:
    case I386_POP:	/* 170 */
    case I386_POPA:
    case I386_POPF:
    case I386_PUSH:	/* 173 */
    case I386_PUSHA:
    case I386_PUSHF:
    case I386_RCL:
    case I386_RCR:
    case I386_RDMSR:
    case I386_RDPMC:
    case I386_RDTSC:
    case I386_RET:	/* 181 */
    case I386_RETF:
    case I386_ROL:
    case I386_ROR:
    case I386_RSM:	
    case I386_SAHF:
    case I386_SAR:	/* 187 */
    case I386_SBB:	/* 188 */
    case I386_SCASB:
    case I386_SCASD:
    case I386_SETcc:	/* 191 */
    case I386_SGDT:
    case I386_SHL:	/* 193 */
    case I386_SHLD:	/* 194 */
    case I386_SHR:
    case I386_SHRD:	/* 196 */
    case I386_SIDT:
    case I386_SLDT:
    case I386_SMSW:
    case I386_STC:
    case I386_STD:	/* 201 */
    case I386_STI:
    case I386_STOSB:	/* 203 */
    case I386_STOSD:	/* 204 */
    case I386_STR:
    case I386_SUB:	/* 206 */
    case I386_SYSENTER:
    case I386_SYSEXIT:
/*      DiabloPrint(stdout,"@I\n",ins);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    case I386_TEST:	/* 209 */
      {
	t_reg reg_s1, reg_s2;
	t_uint32 u_src1,u_src2,u_temp32;
	register_content reg_cont_src1, reg_cont_src2, reg_cont_dest;

	reg_s1 = I386_OP_BASE(I386_INS_SOURCE1(ins));
	reg_s2 = I386_OP_BASE(I386_INS_DEST(ins));

	if(ProcStateGetReg(state,reg_s1,&reg_cont_src1)!=0||
	    ProcStateGetReg(state,reg_s2,&reg_cont_src2)!=0)
	{
/*          DiabloPrint(stdout,"@I\n",ins);*/
	  ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*                      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
	  break;
	}
	u_src1=G_T_UINT32(reg_cont_src1.i);
	u_temp32=(t_uint32)u_src1;

	u_src2=G_T_UINT32(reg_cont_src2.i);
	u_temp32&=(t_uint32)u_src2;

	bzero(&reg_cont_dest,sizeof(register_content));
	GenericSet_uint32(reg_cont_dest.i,u_temp32);
	ProcStateSetReg(state,reg_s2,reg_cont_dest);

	if(u_temp32==0)
	  ProcStateSetReg(state, I386_CONDREG_ZF,reg_cont_dest1);
	else
	  ProcStateSetReg(state, I386_CONDREG_ZF,reg_cont_dest0);
	  
	  ProcStateSetReg(state, I386_CONDREG_CF,reg_cont_dest0);
	  ProcStateSetReg(state, I386_CONDREG_OF,reg_cont_dest0);

	  if(u_temp32 & 0x80000000)
	  ProcStateSetReg(state, I386_CONDREG_SF,reg_cont_dest1);
	  else
	  ProcStateSetReg(state, I386_CONDREG_SF,reg_cont_dest0);

	  ProcStateSetReg(state, I386_CONDREG_PF,reg_cont_dest1);

	  ProcStateSetRegBot(state, I386_CONDREG_AF);
	
	if (RELOCATABLE_REFERS_TO(ins))
	  ProcStateSetTag(state,reg_s2,RELOCATABLE_REFERS_TO(ins)->rel);
/*        DiabloPrint(stdout,"TEST: @I s:%x s2:%x d:%x\n",ins,u_src1,u_src2,u_temp32);*/
	break;
      }
/*      DiabloPrint(stdout,"@I\n",ins);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*            DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    case I386_UD2:
    case I386_VERR:
    case I386_VERW:
    case I386_WAIT:
    case I386_WBINVD:
    case I386_WRMSR:
    case I386_XADD:	/* 216 */
    case I386_XCHG:	/* 217 */
    case I386_XLAT:
/*      DiabloPrint(stdout,"@I\n",ins);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    case I386_XOR:	/* 219 */
      if((I386_INS_DEST(ins))->type == i386_optype_reg &&
	  (I386_INS_SOURCE1(ins))->type == i386_optype_reg)
      {
	t_reg reg_s1, reg_s2;
	reg_s1 = I386_OP_BASE(I386_INS_SOURCE1(ins));
	reg_s2 = I386_OP_BASE(I386_INS_DEST(ins));
	if( reg_s1==reg_s2)
	{
	  ProcStateSetReg(state, I386_CONDREG_OF,reg_cont_dest0);
	  ProcStateSetReg(state, I386_CONDREG_CF,reg_cont_dest0);
	  ProcStateSetReg(state, I386_CONDREG_SF,reg_cont_dest0);
	  ProcStateSetReg(state, reg_s1,reg_cont_dest0);
	  ProcStateSetReg(state, I386_CONDREG_ZF,reg_cont_dest1);
	  ProcStateSetReg(state, I386_CONDREG_PF,reg_cont_dest1);
	  ProcStateSetRegBot(state, I386_CONDREG_AF);
/*          DiabloPrint(stdout,"XOR: @I -> 0! \n",ins);*/
	  break;
	}
      }
/*      DiabloPrint(stdout,"@I\n",ins);*/
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
/*      DiabloPrint(stdout,"OUT: %d \n",I386_INS_OPCODE(ins));*/
      break;
    default:
      FATAL(("Implement %d",I386_INS_OPCODE(ins)));
  }
}

/* vim: set shiftwidth=2: */
