/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

/* Includes {{{ */
#include <stdlib.h>
#include "../../utils/util_status.h"
#include "../../kernel/diablo_iterators.h"
/*#include "../../obfuscation/obf_opposite.h" */
#include "i386_opcode.h"
#include "i386_registers.h"
#include "i386_opposite.h"
/* }}} */

void GetArchs(t_bbl * bbl,t_section *sec)
{
  t_ins * ins;

  ins=InsNewForBbl(bbl);
  I386InstructionMakeCondJump(ins, I386_CONDITION_NZ);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeRM32IMM32(ins, I386_REG_EBX, I386_REG_NONE, 1,I386_SUB);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeRM32IMM32(ins, I386_REG_ECX, I386_REG_NONE, 4,I386_ADD);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeRM32IMM32(ins, I386_REG_EAX, I386_REG_NONE, 4,I386_ADD);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeMovToMem(ins,0, I386_REG_ECX, I386_REG_NONE ,I386_SCALE_1, I386_REG_EDI,0);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeRM32IMM32(ins,I386_REG_EDI, I386_REG_ESI,0,I386_OR);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeRM32IMM8(ins,I386_REG_ESI, I386_REG_NONE, 0x4, I386_SHL);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeRM32IMM32(ins,I386_REG_EDI, I386_REG_NONE, 0x0f0f0f0f, I386_AND);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeRM32IMM32(ins,I386_REG_ESI, I386_REG_NONE, 0x0f0f0f0f, I386_AND);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeMovFromMem(ins,I386_REG_EDI, 0x4 , I386_REG_EAX, I386_REG_NONE ,I386_SCALE_1);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeMovFromMem(ins,I386_REG_ESI, 0 , I386_REG_EAX, I386_REG_NONE ,I386_SCALE_1);
  InsPrependToBbl(ins,bbl);


}

void PostArchs(t_bbl * bbl,t_section *sec)
{
}

void PreArchs(t_bbl * bbl,t_section *sec)
{

  t_section * psec=SECTION_PARENT_SECTION(sec);

  t_ins * ins;

  ins=InsNewForBbl(bbl);
  I386InstructionMakeMovToReg(ins,I386_REG_ECX,I386_REG_NONE, 0x0);
  InsPrependToBbl(ins,bbl);
  I386_OP_FLAGS(I386_INS_SOURCE1(ins)) =I386_OPFLAG_ISRELOCATED;
  RelocTableAddRelocToRelocatable(psec->obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(sec),NULL_GENERIC,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32, FALSE);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeMovFromMem(ins,I386_REG_EAX, 0 , I386_REG_ESP, I386_REG_EAX ,I386_SCALE_4);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeMovFromMem(ins,I386_REG_EAX, 0 , I386_REG_ESP, I386_REG_NONE ,I386_SCALE_1);
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  I386InstructionMakeMovToReg(ins,I386_REG_EBX,I386_REG_NONE, 0x1);
  InsPrependToBbl(ins,bbl);

}

t_uint32 FindFreeReg(t_regset dead)
{
  t_uint32 i=0;

  REGSET_FOREACH_REG(dead,i)
    break;

  return i;
}

t_bool RegisterCanBeUsed(t_uint32 reg)
{
  if(reg<=I386_REG_ESP)
    return TRUE;
  return FALSE;
}

t_bbl * IndirectJumpBbl(t_cfg * cfg,t_section * added, t_uint32 free_reg,t_cfg_edge * jump_edge)
{
  t_object * obj=cfg->sec->obj;
  t_bbl * new_bbl= BblNew(cfg);
  t_ins * ins;
  
  ins=InsNewForBbl(new_bbl);
  I386InstructionMakeMovFromMem(ins,free_reg, 0x4 , I386_REG_NONE, I386_REG_NONE ,I386_SCALE_1);
  InsAppendToBbl(ins,new_bbl);
  I386_OP_FLAGS(I386_INS_SOURCE1(ins)) =I386_OPFLAG_ISRELOCATED;
  RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(added),NULL_GENERIC,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);

  ins=InsNewForBbl(new_bbl);
  I386InstructionMakeRM32IMM32(ins, free_reg, I386_REG_NONE, 0x4,I386_ADD);
  InsAppendToBbl(ins,new_bbl);
  I386_OP_FLAGS(I386_INS_SOURCE1(ins)) =I386_OPFLAG_ISRELOCATED;
  RelocTableAddRelocToRelocatable(obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,REL_TO_RELOCATABLE(jump_edge),NULL_GENERIC,FALSE,NULL,NULL,NULL,"SA-\\" WRITE_S16, TRUE);
  INS_ATTRIB(ins)|=IF_BRANCH_OBF;

  ins=InsNewForBbl(new_bbl);
  I386InstructionMakeJumpReg(ins, free_reg);
  InsAppendToBbl(ins,new_bbl);

  return new_bbl;
}


/* vim: set shiftwidth=2 foldmethod=marker: */
