/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "i386_opposite.h"
#include "i386_opcode.h"

const t_i386_opposite i386_opposite_table[]={
  {I386_CONDITION_A,I386_CONDITION_BE},
  {I386_CONDITION_AE,I386_CONDITION_B},
  {I386_CONDITION_Z,I386_CONDITION_NZ},
  {I386_CONDITION_NP,I386_CONDITION_P},
  {I386_CONDITION_GE,I386_CONDITION_L},
  {I386_CONDITION_LE,I386_CONDITION_G},
  {I386_CONDITION_NS,I386_CONDITION_S},
  {I386_CONDITION_LOOP,I386_CONDITION_ECXZ},
  {I386_CONDITION_NONE,I386_CONDITION_NONE}
};  

t_bool OppositeExist(t_bbl * i_bbl)
{
  if(I386_INS_OPCODE(i_bbl->ins_last)==I386_Jcc)
  if(OppositeCondExist(I386_INS_CONDITION(i_bbl->ins_last)))
    return TRUE;
  return FALSE;
}

t_bool MakeOpposite(t_bbl * i_bbl)
{
  if(OppositeExist(i_bbl))
  {
    I386_INS_CONDITION(i_bbl->ins_last)=GiveOppositeCondition(I386_INS_CONDITION(i_bbl->ins_last));
    return TRUE;
  }
  return FALSE;
}

t_bool OppositeCondExist(t_uint32 test_cond)
{
  t_uint32 i=0;
  for(i=0;i386_opposite_table[i].cond0!=I386_CONDITION_NONE;i++)
    if(i386_opposite_table[i].cond0 == test_cond)
      return TRUE;
  for(i=0;i386_opposite_table[i].cond1!=I386_CONDITION_NONE;i++)
    if(i386_opposite_table[i].cond1 == test_cond)
      return TRUE;
  return FALSE;
}

t_uint32 GiveOppositeCondition(t_uint32 test_cond)
{
  t_uint32 i=0;
  for(i=0;i386_opposite_table[i].cond0!=I386_CONDITION_NONE;i++)
  {
    if(i386_opposite_table[i].cond0 == test_cond)
      return i386_opposite_table[i].cond1;
  }
  for(i=0;i386_opposite_table[i].cond1!=I386_CONDITION_NONE;i++)
  {
    if(i386_opposite_table[i].cond1 == test_cond)
      return i386_opposite_table[i].cond0;
  }
  return I386_CONDITION_NONE;
}

/* vim: set shiftwidth=2: */
