/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef IA64_ACTIONS_H
#define IA64_ACTIONS_H
#include "../../kernel/diablo_action.h"
#include "ia64_actions.h"
#include "ia64_assemble.h"
#include "ia64_constant_propagation.h"
#include "ia64_deflowgraph.h"
#include "ia64_disassemble.h"
#include "ia64_emulation.h"
#include "ia64_flowgraph.h"
#include "ia64_map.h"
#include "ia64_print.h"
#include "ia64_registers.h"
#include "ia64_datamap.h"
#include "ia64_write.h" 

#include "ia64_factor.h"

const t_action_struct ia64_actions={
  /*  0 */   NULL,                             /* Group Info */
  /*  1 */   Ia64ReadMap,                      /* Read Map */
  /*  2 */   NULL,                             /* Disassemble */
  /*  3 */   Ia64DisassembleSection,           /* Disassemble Section */
  /*  4 */   NULL,                             /* Flowgraph */
  /*  5 */   Ia64FlowgraphSection,             /* Flowgraph Section */
  /*  6 */   Ia64MakeAddressProducers,         /* Make address producers */
  /*  7 */   NULL,                             /* Deflowgraph */
  /*  8 */   Ia64DeflowgraphSection,           /* Deflowgraph Section */
  /*  9 */   NULL,                             /* Assemble */ 
  /* 10 */   Ia64AssembleSection,              /* Assemble Section */
  /* 11 */   Ia64Write,                        /* Write Object */
  /* Instruction related ia64 callbacks*/
  /* 12 */   Ia64InstructionPrint,             /* Print Instruction */
  /* 13 */   Ia64InstructionMakeNoop,          /* Make a noop from the instruction */

  /* 14 */   Ia64RegisterName,                 /* return the register name (from a static array of registers )*/
  /* 15 */   Ia64InstructionEmulator,          /* Instruction Emulation */
  /* 16 */   Ia64EdgePropagator,               /* */
  /* 17 */   Ia64InstructionUnconditionalizer, /* */
  /* 18 */   Ia64InstructionsConstantOptimizer,/* */
  /* 19 */   NULL,                             /* */
  /* 20 */   Ia64ConvertToBlocksSec,           /* */
  /* 21 */   NULL,                             /* */
  /* 22 */   NULL,			     /* */ 	
  /* 23 */   Ia64InsHasSideEffect,             /* */
  /* 24 */   Ia64BblCmpExtReg,
  /* 25 */   NULL,
  /* 26 */   NULL,
  /* 27 */   NULL,
  /* 28 */   Ia64IsSyscallExit,
  /* 29 */   NULL,
  /* 30 */   Ia64InsIsIndirectCall,              /* InsIsIndirectCall */
  /* 31 */   Ia64FunIsGlobal,              /* FunIsGlobal */
  NULL,
  NULL
};
#endif
/* vim: set shiftwidth=2: */
