/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"

#ifndef IA64_ASSEMBLE_H
#define IA64_ASSEMBLE_H
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_section.h"
#include "ia64_bundle.h"
void Ia64AssembleSection(t_object * obj, t_section * sec);

/* The structure of the different assemble functions */
  
typedef void (*Ia64AssembleFunction) (t_ins * ins, t_uint64 * data, t_bundle * bundle);
 

void Ia64AssembleUnsupported(t_ins * ins, t_uint64 * instr, t_bundle * bundle);
 
/*
 * bevat de lijst van alle assemble functions, 1 voor elk instructie formaat
 * de indexering in deze tabel moet dan ook volgens t_ia64_instr_format gebeuren...
 */
extern const Ia64AssembleFunction Ia64AssembleFunction_table [];


#define TO_UINT64(value) ((t_uint64)(value))


/* shorthands used in assemble_one_?.c */

#define SET_PRED(instr, value)  (instr |= (TO_UINT64(value)) - IA64_PR)
#define SET_GR_R1(instr, value) (instr |= ((TO_UINT64(value)) - IA64_GR)<<6)
#define SET_GR_R2(instr, value) (instr |= ((TO_UINT64(value)) - IA64_GR)<<13)
#define SET_GR_R3(instr, value) (instr |= ((TO_UINT64(value)) - IA64_GR)<<20)

#define SET_FR_F1(instr, value) (instr |= ((TO_UINT64(value)) - IA64_FR)<<6)
#define SET_FR_F2(instr, value) (instr |= ((TO_UINT64(value)) - IA64_FR)<<13)
#define SET_FR_F3(instr, value) (instr |= ((TO_UINT64(value)) - IA64_FR)<<20)
#define SET_FR_F4(instr, value) (instr |= ((TO_UINT64(value)) - IA64_FR)<<27)

#define SET_PR_P1(instr, value) (instr |= ((TO_UINT64(value)) - IA64_PR)<<6)
#define SET_PR_P2(instr, value) (instr |= ((TO_UINT64(value)) - IA64_PR)<<27)

#define SET_AR_R3(instr, value) (instr |= ((TO_UINT64(value)) - IA64_AR)<<20)
#define SET_CR_R3(instr, value) (instr |= ((TO_UINT64(value)) - IA64_CR)<<20)

#define SET_BR_B1(instr, value) (instr |= ((TO_UINT64(value)) - IA64_BR)<<6)
#define SET_BR_B2(instr, value) (instr |= ((TO_UINT64(value)) - IA64_BR)<<13)

#define SET_S_I_BIT(instr, value) (instr |= (TO_UINT64(value))<<36)

#define SET_IMM2c(instr, value) (instr |= (TO_UINT64(value))<<33)
#define SET_IMM2d(instr, value) (instr |= (TO_UINT64(value))<<30)
#define SET_IMM4c(instr, value) (instr |= (TO_UINT64(value))<<20)
#define SET_IMM4d(instr, value) (instr |= (TO_UINT64(value))<<27)
#define SET_IMM5b(instr, value) (instr |= (TO_UINT64(value))<<14)
#define SET_IMM5c(instr, value) (instr |= (TO_UINT64(value))<<22)
#define SET_IMM5d(instr, value) (instr |= (TO_UINT64(value))<<20)
#define SET_IMM6b(instr, value) (instr |= (TO_UINT64(value))<<14)
#define SET_IMM6c(instr, value) (instr |= (TO_UINT64(value))<<20)
#define SET_IMM6d(instr, value) (instr |= (TO_UINT64(value))<<27)
#define SET_IMM6e(instr, value) (instr |= (TO_UINT64(value))<<31)
#define SET_IMM7a(instr, value) (instr |= (TO_UINT64(value))<<6)
#define SET_IMM7b(instr, value) (instr |= (TO_UINT64(value))<<13)
#define SET_IMM7c(instr, value) (instr |= (TO_UINT64(value))<<20)
#define SET_IMM8c(instr, value) (instr |= (TO_UINT64(value))<<20)
#define SET_IMM8d(instr, value) (instr |= (TO_UINT64(value))<<24)
#define SET_IMM9c(instr, value) (instr |= (TO_UINT64(value))<<24)
#define SET_IMM9d(instr, value) (instr |= (TO_UINT64(value))<<27)
#define SET_IMM13c(instr, value) (instr |= (TO_UINT64(value))<<20)
#define SET_IMM20a(instr, value) (instr |= (TO_UINT64(value))<<6)
#define SET_IMM20b(instr, value) (instr |= (TO_UINT64(value))<<13)
#define SET_IMM27a(instr, value) (instr |= (TO_UINT64(value))<<6)


#define SET_IMM2_A2_A10(instr, value) \
    (instr |= ((TO_UINT64(value))-1)<<27)
#define SET_IMM8_A3_A8_I27_M30(instr, value) \
    ( instr |= SET_S_I_BIT(instr,GETBITS64(TO_UINT64(value), 7, 1)) | SET_IMM7b(instr,GETBITS64(TO_UINT64(value),0,7)) )
#define SET_IMM14_A4(instr, value) \
    (instr |=   SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 13, 1)) \
                | SET_IMM6d(instr, GETBITS64(TO_UINT64(value), 7, 6)) \
                | SET_IMM7b(instr, GETBITS64(TO_UINT64(value), 0, 7)) )
#define SET_IMM22_A5(instr, value) \
    (instr |=   SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 21, 1)) \
                | SET_IMM5c(instr, GETBITS64(TO_UINT64(value), 16, 5)) \
                | SET_IMM9d(instr, GETBITS64(TO_UINT64(value), 7, 9)) \
                | SET_IMM7b(instr, GETBITS64(TO_UINT64(value), 0, 7)) )
#define SET_IMM2_I1(instr, value)\
    ( instr |= SET_IMM2d(instr, (TO_UINT64(value)==15)?2:(TO_UINT64(value)==7)?1:(TO_UINT64(value)==0)?0:3) )
#define SET_IMM4_I3(instr, value)\
    ( instr |= SET_IMM4c(instr, TO_UINT64(value)) )
#define SET_IMM8_I4(instr, value)\
    ( instr |= SET_IMM8c(instr, TO_UINT64(value)) )
#define SET_IMM5_I6(instr, value)\
    ( instr |= SET_IMM5b(instr, TO_UINT64(value)) )
#define SET_IMM5_I8(instr, value)\
    ( instr |= SET_IMM5d(instr, 31 - TO_UINT64(value)) )
#define SET_IMM6_I10(instr, value)\
    ( instr |= SET_IMM6d(instr, TO_UINT64(value)) )
#define SET_IMM6b_I11(instr, value)\
    ( instr |= SET_IMM6b(instr, TO_UINT64(value)) )
#define SET_IMM6d_I11(instr, value)\
    ( instr |= SET_IMM6d(instr, TO_UINT64(value) - 1) )
#define SET_IMM6c_I12(instr, value)\
    ( instr |= SET_IMM6c(instr, 63 - TO_UINT64(value)) )
#define SET_IMM6d_I12(instr, value)\
    ( instr |= SET_IMM6d(instr, TO_UINT64(value) - 1) )
#define SET_IMM6c_I13(instr, value)\
    ( instr |= SET_IMM6c(instr, 63 - TO_UINT64(value)) )
#define SET_IMM6d_I13(instr, value)\
    ( instr |= SET_IMM6d(instr, TO_UINT64(value) - 1) )
#define SET_IMM8_I13(instr, value) \
    ( instr |= SET_S_I_BIT(instr,GETBITS64(TO_UINT64(value), 7, 1)) | SET_IMM7b(instr,GETBITS64(TO_UINT64(value),0,7)) )
#define SET_IMM1_I14(instr, value)\
    ( instr |= SET_S_I_BIT(instr,GETBITS64(TO_UINT64(value),1,1)) )
#define SET_IMM6b_I14(instr, value)\
    ( instr |= SET_IMM6b(instr, 63 - TO_UINT64(value)) )
#define SET_IMM6d_I14(instr, value)\
    ( instr |= SET_IMM6d(instr, TO_UINT64(value) - 1) )
#define SET_IMM4_I15(instr, value)\
    ( instr |= SET_IMM4d(instr, TO_UINT64(value) - 1) )
#define SET_IMM6_I15(instr, value)\
    ( instr |= SET_IMM6e(instr, 63 - TO_UINT64(value)) )
#define SET_IMM6_I16(instr, value)\
    ( instr |= SET_IMM6b(instr, TO_UINT64(value)) )
#define SET_IMM21_I19_M37(instr, value) \
    ( instr |= SET_S_I_BIT(instr,GETBITS64(TO_UINT64(value), 20, 1)) | SET_IMM20a(instr,GETBITS64(TO_UINT64(value),0,20)) )
#define SET_TAG13_I21(instr, value)\
    ( instr |= SET_IMM9c(instr,GETBITS64((TO_UINT64(value)) >> 4,0,9)) )
#define SET_MASK17_I23(instr, value) \
    (instr |=   SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 16, 1)) \
                | SET_IMM8d(instr, GETBITS64(TO_UINT64(value), 8, 8)) \
                | SET_IMM7a(instr, GETBITS64(TO_UINT64(value), 1, 7)) )
#define SET_IMM44_I24(instr, value) \
    (instr |=   SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 43, 1)) \
                | SET_IMM27a(instr, GETBITS64(TO_UINT64(value), 16, 27)) )
#define SET_IMM9_F5(instr, value) \
    (instr |=   SET_IMM7c(instr, GETBITS64(TO_UINT64(value), 2, 7)) \
                | SET_IMM2c(instr, GETBITS64(TO_UINT64(value), 0, 2)) )
#define SET_IMM7b_F12(instr, value)\
    ( instr |= SET_IMM7b(instr, TO_UINT64(value)) )
#define SET_IMM7c_F12(instr, value)\
    ( instr |= SET_IMM7c(instr, TO_UINT64(value)) )
#define SET_TARGET25_F14(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64((TO_UINT64(value))>>4, 20, 1)) \
           | SET_IMM20a(instr, GETBITS64((TO_UINT64(value))>>4, 0, 20)))
#define SET_IMM21_F15(instr, value) \
    (instr |=   SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 20, 1)) \
                | SET_IMM20a(instr, GETBITS64(TO_UINT64(value), 0, 20)) )

#define SET_IMM2_A2(instr, value) SET_IMM2_A2_A10(instr, TO_UINT64(value))
#define SET_IMM2_A10(instr, value) SET_IMM2_A2_A10(instr, TO_UINT64(value))
#define SET_IMM8_A3(instr, value) SET_IMM8_A3_A8_I27_M30(instr, TO_UINT64(value))
#define SET_IMM8_A8(instr, value) SET_IMM8_A3_A8_I27_M30(instr, TO_UINT64(value))
#define SET_IMM8_I27(instr, value) SET_IMM8_A3_A8_I27_M30(instr, TO_UINT64(value))
#define SET_IMM8_M30(instr, value) SET_IMM8_A3_A8_I27_M30(instr, TO_UINT64(value))
#define SET_IMM21_I19(instr, value) SET_IMM21_I19_M37(instr, TO_UINT64(value))
#define SET_IMM21_M37(instr, value) SET_IMM21_I19_M37(instr, TO_UINT64(value))


#define SET_IMM9_M3_M8_M15(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 8, 1)) \
           | GETBITS64(TO_UINT64(value), 7, 1)<<27 \
           | SET_IMM7b(instr, GETBITS64(TO_UINT64(value), 0, 7)))

#define SET_IMM9_M5_M10(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 8, 1)) \
           | GETBITS64(TO_UINT64(value), 7, 1)<<27 \
           | SET_IMM7a(instr, GETBITS64(TO_UINT64(value), 0, 7)))

#undef SET_IMM21_M37
#define SET_IMM21_M37(instr, value) \
    ( instr |= SET_S_I_BIT(instr,GETBITS64(TO_UINT64(value), 20, 1)) | SET_IMM20a(instr,GETBITS64(TO_UINT64(value),0,20)) )

#define SET_IMM24_M44(instr, value) \
    ( instr |= SET_S_I_BIT(instr,GETBITS64(TO_UINT64(value), 23, 1)) \
                        | GETBITS64(TO_UINT64(value),21,2)<<31 \
                        | GETBITS64(TO_UINT64(value),0,21)<<6 )

#define SET_TARGET25_I20_M20_M21(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64((TO_UINT64(value))>>4, 20, 1)) \
           | SET_IMM13c(instr, GETBITS64((TO_UINT64(value))>>4, 7, 13)) \
           | SET_IMM7a(instr, GETBITS64((TO_UINT64(value))>>4, 0, 7)))

#define SET_TARGET25_M22_M23(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64((TO_UINT64(value))>>4, 20, 1)) \
           | SET_IMM20b(instr, GETBITS64((TO_UINT64(value))>>4, 0, 20)) )


#define SET_TARGET25_B1_B2_B3_B6(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64((TO_UINT64(value))>>4, 20, 1)) \
           | SET_IMM20b(instr, GETBITS64((TO_UINT64(value))>>4, 0, 20)) )

#define SET_TAG13_B6_B7(instr, value) \
    (instr |= GETBITS64((TO_UINT64(value))>>4, 7, 2)<<33 \
           | SET_IMM7a(instr, GETBITS64((TO_UINT64(value))>>4, 0, 7)) )

#define SET_IMM21_B9(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 20, 1)) \
           | SET_IMM20a(instr, GETBITS64(TO_UINT64(value), 0, 20)) )

#define SET_IMM62_X1(instr, value) \
                (instr |= SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 20, 1)) \
           | SET_IMM20a(instr, GETBITS64(TO_UINT64(value), 0, 20)) )

#define SET_IMM64_X2(instr, value) \
                (instr |= SET_S_I_BIT(instr, GETBITS64(TO_UINT64(value), 63, 1)) \
           | GETBITS64(TO_UINT64(value), 21, 1)<<21 \
           | GETBITS64(TO_UINT64(value), 16, 5)<<22 \
           | GETBITS64(TO_UINT64(value), 7, 9)<<27 \
           | SET_IMM7b(instr, GETBITS64(TO_UINT64(value), 0, 7)) )

#define SET_TARGET64_X3_X4(instr, value) \
    (instr |= SET_S_I_BIT(instr, GETBITS64((TO_UINT64(value))>>4, 59, 1)) \
           | SET_IMM20b(instr, GETBITS64((TO_UINT64(value))>>4, 0, 20)) )


#define SET_IMM9_M3(instr, value) SET_IMM9_M3_M8_M15(instr, TO_UINT64(value))
#define SET_IMM9_M8(instr, value) SET_IMM9_M3_M8_M15(instr, TO_UINT64(value))
#define SET_IMM9_M15(instr, value) SET_IMM9_M3_M8_M15(instr, TO_UINT64(value))

#define SET_IMM9_M5(instr, value) SET_IMM9_M5_M10(instr, TO_UINT64(value))
#define SET_IMM9_M10(instr, value) SET_IMM9_M5_M10(instr, TO_UINT64(value))

#define SET_TARGET25_I20(instr, value) SET_TARGET25_I20_M20_M21(instr, TO_UINT64(value))
#define SET_TARGET25_M20(instr, value) SET_TARGET25_I20_M20_M21(instr, TO_UINT64(value))
#define SET_TARGET25_M21(instr, value) SET_TARGET25_I20_M20_M21(instr, TO_UINT64(value))

#define SET_TARGET25_M22(instr, value) SET_TARGET25_M22_M23(instr, TO_UINT64(value))
#define SET_TARGET25_M23(instr, value) SET_TARGET25_M22_M23(instr, TO_UINT64(value))

#define SET_TARGET25_B1(instr, value) SET_TARGET25_B1_B2_B3_B6(instr, TO_UINT64(value))
#define SET_TARGET25_B2(instr, value) SET_TARGET25_B1_B2_B3_B6(instr, TO_UINT64(value))
#define SET_TARGET25_B3(instr, value) SET_TARGET25_B1_B2_B3_B6(instr, TO_UINT64(value))
#define SET_TARGET25_B6(instr, value) SET_TARGET25_B1_B2_B3_B6(instr, TO_UINT64(value))

#define SET_TAG13_B6(instr, value) SET_TAG13_B6_B7(instr, TO_UINT64(value))
#define SET_TAG13_B7(instr, value) SET_TAG13_B6_B7(instr, TO_UINT64(value))

#define SET_TARGET64_X3(instr, value) SET_TARGET64_X3_X4(instr, TO_UINT64(value))
#define SET_TARGET64_X4(instr, value) SET_TARGET64_X3_X4(instr, TO_UINT64(value))


#endif
/* vim: set shiftwidth=2: */
