/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_assemble_one_a.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"



#ifdef _DEBUG_ASSEMBLE_
#include "../ia64_disassemble/ia64_disassemble_one_a.h"
#endif


/* A1: Integer Alu-Register-Register */
void Ia64AssembleA1(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA1(ins, *instr,bundle);
#endif
}

/* A2: Integer Alu-shift left and add */
void Ia64AssembleA2(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM2_A2(*instr,IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA2(ins, *instr,bundle);
#endif
}


/* A3: Integer Alu-immediate8 register */
void Ia64AssembleA3(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM8_A3(*instr,IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA3(ins, *instr,bundle);
#endif
}


/* A4: Integer Alu-add immediate14*/
void Ia64AssembleA4(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM14_A4(*instr,IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA4(ins, *instr,bundle);
#endif

}

/* A5: Integer Alu-add immediate22*/
void Ia64AssembleA5(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  if(IA64INSIN1(ins)>IA64_R3)
  {
    //InsPrint(ins,0);
    FATAL(("addl rx=imm22,ry, met y in [0-3]"));
  }
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM22_A5(*instr,IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA5(ins, *instr,bundle);
#endif
}

/* A6: Integer compare-register-register*/
void Ia64AssembleA6(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_PR_P1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA6(ins, *instr,bundle);
#endif
}


/* A7: Integer comparetozero-register*/
void Ia64AssembleA7(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_PR_P1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA7(ins, *instr,bundle);
#endif
}


/* A8: Integer compare-immediate-register*/
void Ia64AssembleA8(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_PR_P1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

  SET_IMM8_A8(*instr,IA64INSIMM1(ins));
  
#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA8(ins, *instr,bundle);
#endif
}


/* A9: Multimedia ALU*/
void Ia64AssembleA9(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA9(ins, *instr,bundle);
#endif
}


/* A10: Multimedia shift and add*/
void Ia64AssembleA10(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM2_A10(*instr,IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleA10(ins, *instr,bundle);
#endif
}


/* vim: set shiftwidth=2: */

