/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_assemble_one_b.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"

#ifdef _DEBUG_ASSEMBLE_
#include "../ia64_disassemble/ia64_disassemble_one_b.h"
#endif


/* B1: IP-Relative Branch */
void Ia64AssembleB1(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_TARGET25_B1(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB1(ins, *instr,  bundle);
#endif

}


/* B2: IP-Relative Counted Branch */
void Ia64AssembleB2(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_TARGET25_B2(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB2(ins, *instr,  bundle);
#endif


}


/* B3: IP-Relative Call */
void Ia64AssembleB3(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_BR_B1(*instr, IA64INSOUT1(ins));

  SET_TARGET25_B3(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB3(ins, *instr,  bundle);
#endif


}


/* B4: Indirect Branch */
void Ia64AssembleB4(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_BR_B2(*instr, IA64INSIN1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB4(ins, *instr,  bundle);
#endif


}


/* B5: Indirect Call */
void Ia64AssembleB5(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_BR_B2(*instr, IA64INSIN1(ins));

  SET_BR_B1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB5(ins, *instr,  bundle);
#endif


}


/* B6: IP-Relative Predict */
void Ia64AssembleB6(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
	/* B6 en B7 zijn (voorlopig) de enige die de laagste 6 bits voor iets anders dan qp/0 gebruiken... */
	
  SET_TARGET25_B6(*instr, IA64INSIMM1(ins));
  SET_TAG13_B6(*instr, IA64INSIMM2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB6(ins, *instr,  bundle);
#endif


}

/* B7: Indirect Predict */
void Ia64AssembleB7(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
	/* B6 en B7 zijn (voorlopig) de enige die de laagste 6 bits voor iets anders dan qp/0 gebruiken... */

  SET_BR_B2(*instr, IA64INSIN1(ins));

  SET_TAG13_B7(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB7(ins, *instr,  bundle);
#endif


}


/* B8: Miscellaneous (M-Unit) */
void Ia64AssembleB8(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB8(ins, *instr,  bundle);
#endif


}


/* B9: Break/Nop (B-Unit) */
void Ia64AssembleB9(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_IMM21_B9(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleB9(ins, *instr,  bundle);
#endif


}

/* vim: set shiftwidth=2: */
