/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_assemble_one_f.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"



#ifdef _DEBUG_ASSEMBLE_
#include "../ia64_disassemble/ia64_disassemble_one_f.h"
#endif


/* F1: Floating-point multiply add */
void Ia64AssembleF1(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F3(*instr,IA64INSIN1(ins));
  SET_FR_F4(*instr,IA64INSIN2(ins));
  SET_FR_F2(*instr,IA64INSIN3(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF1(ins, *instr,bundle);
#endif  
}

/* F2: Fixed-point multiply add */
void Ia64AssembleF2(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F3(*instr,IA64INSIN1(ins));
  SET_FR_F4(*instr,IA64INSIN2(ins));
  SET_FR_F2(*instr,IA64INSIN3(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF2(ins, *instr,bundle);
#endif  
}

/* F3: Parallel Floating-point select */
void Ia64AssembleF3(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F3(*instr,IA64INSIN1(ins));
  SET_FR_F4(*instr,IA64INSIN2(ins));
  SET_FR_F2(*instr,IA64INSIN3(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF3(ins, *instr,bundle);
#endif  
}

/* F4: Floating-point compare */
void Ia64AssembleF4(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr,IA64INSIN1(ins));
  SET_FR_F3(*instr,IA64INSIN2(ins));

  SET_PR_P1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF4(ins, *instr,bundle);
#endif  
}

/* F5: Floating-point class */
void Ia64AssembleF5(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr,IA64INSIN1(ins));

  SET_PR_P1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

  SET_IMM9_F5(*instr, IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF5(ins, *instr,bundle);
#endif  
}

/* F6: Floating-point reciprocal approxiimation */
void Ia64AssembleF6(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr,IA64INSIN1(ins));
  SET_FR_F3(*instr,IA64INSIN2(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF6(ins, *instr,bundle);
#endif  
}

/* F7: Floating-point reciprocal square root approximation */
void Ia64AssembleF7(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F3(*instr,IA64INSIN1(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF7(ins, *instr,bundle);
#endif  
}

/* F8: minimum/maximum and parallel compare */
void Ia64AssembleF8(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr,IA64INSIN1(ins));
  SET_FR_F3(*instr,IA64INSIN2(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF8(ins, *instr,bundle);
#endif  
}

/* F9: merge and logical */
void Ia64AssembleF9(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr,IA64INSIN1(ins));
  SET_FR_F3(*instr,IA64INSIN2(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF9(ins, *instr,bundle);
#endif  
}

/* F10: convert floating-point to fixed-point */
void Ia64AssembleF10(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr,IA64INSIN1(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF10(ins, *instr,bundle);
#endif  
}

/* F11: convert fixed-point to floating-point */
void Ia64AssembleF11(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr,IA64INSIN1(ins));

  SET_FR_F1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF11(ins, *instr,bundle);
#endif  
}

/* F12: floating-point set controls*/
void Ia64AssembleF12(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_IMM7b_F12(*instr,IA64INSIMM1(ins));
  SET_IMM7c_F12(*instr,IA64INSIMM2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF12(ins, *instr,bundle);
#endif  
}

/* F13: floating-point clear flags*/
void Ia64AssembleF13(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF13(ins, *instr,bundle);
#endif  
}

/* F14: floating-point check flags*/
void Ia64AssembleF14(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_TARGET25_F14(*instr,IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF14(ins, *instr,bundle);
#endif  
}

/* F15: break/nop*/
void Ia64AssembleF15(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_IMM21_F15(*instr,IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleF15(ins, *instr,bundle);
#endif  
}



/* vim: set shiftwidth=2: */
