/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_assemble_one_i.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"



#ifdef _DEBUG_ASSEMBLE_
#include "../ia64_disassemble/ia64_disassemble_one_i.h"
#endif


/* I1: Multimedia multiply and shift */
void Ia64AssembleI1(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM2_I1(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI1(ins, *instr,bundle);
#endif
}

/* I2: Multimedia multiply/mix/pack/unpack */
void Ia64AssembleI2(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI2(ins, *instr,bundle);
#endif
}


/* I3: Multimedia mux1 */
void Ia64AssembleI3(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM4_I3(*instr, IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI3(ins, *instr,bundle);
#endif
}

/* I4: Multimedia mux2 */
void Ia64AssembleI4(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM8_I4(*instr, IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI4(ins, *instr,bundle);
#endif
}

/* I5: Multimedia mux2 */
void Ia64AssembleI5(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));
  SET_GR_R2(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI5(ins, *instr,bundle);
#endif
}

/* I6: Multimedia shift right fixec */
void Ia64AssembleI6(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM5_I6(*instr, IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI6(ins, *instr,bundle);
#endif
}

/* I7: Multimedia shift left variable */
void Ia64AssembleI7(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI7(ins, *instr,bundle);
#endif
}

/* I8: Multimedia shift left fixed */
void Ia64AssembleI8(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));
 
  SET_IMM5_I8(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI8(ins, *instr,bundle);
#endif
}

/* I9: Population count */
void Ia64AssembleI9(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));
  
  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI9(ins, *instr,bundle);
#endif
}

/* I10: shift right pair */
void Ia64AssembleI10(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM6_I10(*instr, IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI10(ins, *instr,bundle);
#endif
}

/* I11: extract */
void Ia64AssembleI11(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM6b_I11(*instr, IA64INSIMM1(ins));
  SET_IMM6d_I11(*instr, IA64INSIMM2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI11(ins, *instr,bundle);
#endif
}

/* I12: zero and deposit */
void Ia64AssembleI12(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM6c_I12(*instr, IA64INSIMM1(ins));
  SET_IMM6d_I12(*instr, IA64INSIMM2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI12(ins, *instr,bundle);
#endif
}

/* I13: zero and deposit immediate*/
void Ia64AssembleI13(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM8_I13(*instr, IA64INSIMM1(ins));
  SET_IMM6c_I13(*instr, IA64INSIMM2(ins));
  SET_IMM6d_I13(*instr, IA64INSIMM3(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI13(ins, *instr,bundle);
#endif
}

/* I14: deposit immediate*/
void Ia64AssembleI14(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM1_I14(*instr, IA64INSIMM1(ins));
  SET_IMM6b_I14(*instr, IA64INSIMM2(ins));
  SET_IMM6d_I14(*instr, IA64INSIMM3(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI14(ins, *instr,bundle);
#endif
}

/* I15: deposit*/
void Ia64AssembleI15(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));
  SET_GR_R3(*instr,IA64INSIN2(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

  SET_IMM6_I15(*instr, IA64INSIMM1(ins));
  SET_IMM4_I15(*instr, IA64INSIMM2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI15(ins, *instr,bundle);
#endif
}


/* I16: test bit*/
void Ia64AssembleI16(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_PR_P1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

  SET_IMM6_I16(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI16(ins, *instr,bundle);
#endif
}

/* I17: test naT*/
void Ia64AssembleI17(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_PR_P1(*instr,IA64INSOUT1(ins));
  SET_PR_P2(*instr,IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI17(ins, *instr,bundle);
#endif
}

/* I19: break\nop*/
void Ia64AssembleI19(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_IMM21_I19(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI19(ins, *instr,bundle);
#endif
}

/* I20: integer speculation unit*/
void Ia64AssembleI20(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_TARGET25_I20_M20_M21(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI20(ins, *instr,bundle);
#endif
}

/* I21: move to BR*/
void Ia64AssembleI21(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_BR_B1(*instr,IA64INSOUT1(ins));

  SET_TAG13_I21(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI21(ins, *instr,bundle);
#endif
}

/* I22: move from BR*/
void Ia64AssembleI22(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_BR_B2(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI22(ins, *instr,bundle);
#endif
}

/* I23: move to predicates -register*/
void Ia64AssembleI23(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_MASK17_I23(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI23(ins, *instr,bundle);
#endif
}

/* I24: move to predicates - immediate44*/
void Ia64AssembleI24(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_IMM44_I24(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI24(ins, *instr,bundle);
#endif
}

/* I25: move from predicates ip*/
void Ia64AssembleI25(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI25(ins, *instr,bundle);
#endif
}

/* I26: move to AR - register*/
void Ia64AssembleI26(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr,IA64INSIN1(ins));

  SET_AR_R3(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI26(ins, *instr,bundle);
#endif
}


/* I27: move to AR - immediate8*/
void Ia64AssembleI27(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_AR_R3(*instr,IA64INSOUT1(ins));

  SET_IMM8_I27(*instr, IA64INSIMM1(ins));
#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI27(ins, *instr,bundle);
#endif
}

/* I28: move from AR - immediate8*/
void Ia64AssembleI28(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_AR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI27(ins, *instr,bundle);
#endif
}

/* I29: sign/zero extend/compute zero index */
void Ia64AssembleI29(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr,IA64INSIN1(ins));

  SET_GR_R1(*instr,IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
        Ia64DisassembleI27(ins, *instr,bundle);
#endif
}

/* vim: set shiftwidth=2: */
