/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_assemble_one_m.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"


#ifdef _DEBUG_ASSEMBLE_
#include "../ia64_disassemble/ia64_disassemble_one_m.h"
#endif


/* M1: Integer Load */
void Ia64AssembleM1(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM1(ins, *instr,  bundle);
#endif

}


/* M2: Integer Load - Increment by Register */
void Ia64AssembleM2(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));
  SET_GR_R2(*instr, IA64INSIN2(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM2(ins, *instr,  bundle);
#endif


}

/* M3: Integer Load - Increment by immediate */
void Ia64AssembleM3(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

  SET_IMM9_M3(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM3(ins, *instr,  bundle);
#endif


}



/* M4: Integer Store */
void Ia64AssembleM4(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

  SET_GR_R3(*instr, IA64INSIN2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM4(ins, *instr,  bundle);
#endif


}


/* M5: Integer Store - Increment by immediate */
void Ia64AssembleM5(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

  SET_GR_R3(*instr, IA64INSIN2(ins));

  SET_IMM9_M5(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM5(ins, *instr,  bundle);
#endif

}

/* M6: Floating-point Load */
void Ia64AssembleM6(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_FR_F1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM6(ins, *instr,  bundle);
#endif


}


/* M7: Floating-point Load - Increment by Register */
void Ia64AssembleM7(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));
  SET_GR_R2(*instr, IA64INSIN2(ins));

  SET_FR_F1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM7(ins, *instr,  bundle);
#endif


}

/* M8: Floating-point Load - Increment by immediate */
void Ia64AssembleM8(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_FR_F1(*instr, IA64INSOUT1(ins));

  SET_IMM9_M8(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM8(ins, *instr,  bundle);
#endif


}



/* M9: Floating-point Store */
void Ia64AssembleM9(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr, IA64INSIN1(ins));

  SET_GR_R3(*instr, IA64INSIN2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM9(ins, *instr,  bundle);
#endif

}


/* M10: Floating-point Store - Increment by immediate */
void Ia64AssembleM10(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr, IA64INSIN1(ins));

  SET_GR_R3(*instr, IA64INSIN2(ins));

  SET_IMM9_M10(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM10(ins, *instr,  bundle);
#endif


}




/* M11: Floating-point Load Pair */
void Ia64AssembleM11(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_FR_F1(*instr, IA64INSOUT1(ins));
  SET_FR_F2(*instr, IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM11(ins, *instr,  bundle);
#endif


}


/* M12: Floating-point Load Pair - Increment by Immediate*/
void Ia64AssembleM12(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_FR_F1(*instr, IA64INSOUT1(ins));
  SET_FR_F2(*instr, IA64INSOUT2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM12(ins, *instr,  bundle);
#endif


}



/* M13: Line Prefetch */
void Ia64AssembleM13(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM13(ins, *instr,  bundle);
#endif


}

/* M14: Line Prefetch - Increment by Register */
void Ia64AssembleM14(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));
  SET_GR_R2(*instr, IA64INSIN2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM14(ins, *instr,  bundle);
#endif


}


/* M15: Line Prefetch - Increment by Immediate */
void Ia64AssembleM15(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_IMM9_M15(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM15(ins, *instr,  bundle);
#endif


}


/* M16: Exchange/Compare and Exchange */
void Ia64AssembleM16(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));
  SET_GR_R2(*instr, IA64INSIN2(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM16(ins, *instr,  bundle);
#endif


}


/* M17: Fetch and Add - Immediate */
void Ia64AssembleM17(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
	t_uint64 value = IA64INSIMM1(ins);
	
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

	if(value<0)
	{
		*instr |= (t_uint64)1<<15;
		value = -value;
	}	
	
	switch(value)
	{
		case 1:
			*instr |= (t_uint64)3<<13;
			break;
		case 4:
			*instr |= (t_uint64)2<<13;
			break;
		case 8:
			*instr |= (t_uint64)1<<13;
			break;
		case 16:
			break;
		default: /* zou niet mogen gebeuren !!! */
				FATAL(("Foute waarde in immediate van instructie M17 !"));
	}

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM17(ins, *instr,  bundle);
#endif


}

/* M18: Set FR */
void Ia64AssembleM18(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

  SET_FR_F1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM18(ins, *instr,  bundle);
#endif


}


/* M19: Get FR */
void Ia64AssembleM19(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM19(ins, *instr,  bundle);
#endif


}

/* M20: Integer Speculation Check (M-Unit) */
void Ia64AssembleM20(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

  SET_TARGET25_M20(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM20(ins, *instr,  bundle);
#endif


}

/* M21: Floating-point Speculation Check */
void Ia64AssembleM21(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F2(*instr, IA64INSIN1(ins));

  SET_TARGET25_M21(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM21(ins, *instr,  bundle);
#endif


}


/* M22: Integer Advanced Load Check */
void Ia64AssembleM22(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R1(*instr, IA64INSIN1(ins));

  SET_TARGET25_M22(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM22(ins, *instr,  bundle);
#endif


}


/* M23: Floating-point Advanced Load Check */
void Ia64AssembleM23(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F1(*instr, IA64INSIN1(ins));

  SET_TARGET25_M23(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM23(ins, *instr,  bundle);
#endif


}

/* M24: Sync/Fence/Serialize/ALAT Control */
void Ia64AssembleM24(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM24(ins, *instr,  bundle);
#endif

}


/* M25: RSE Control */
void Ia64AssembleM25(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM25(ins, *instr,  bundle);
#endif


}

/* M26: Integer ALAT Entry Invalidate */
void Ia64AssembleM26(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R1(*instr, IA64INSIN1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM26(ins, *instr,  bundle);
#endif


}

/* M27: Floating-point ALAT Entry Invalidate */
void Ia64AssembleM27(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_FR_F1(*instr, IA64INSIN1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM27(ins, *instr,  bundle);
#endif


}

/* M28: Flush Cache/Purge Translation Cache Entry */
void Ia64AssembleM28(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM28(ins, *instr,  bundle);
#endif


}

/* M29: Move to AR - Register (M-Unit) */
void Ia64AssembleM29(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

  SET_AR_R3(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM29(ins, *instr,  bundle);
#endif


}

/* M30: Move to AR - Immediate8 (M-Unit) */
void Ia64AssembleM30(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_AR_R3(*instr, IA64INSOUT1(ins));

  SET_IMM8_M30(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM30(ins, *instr,  bundle);
#endif


}


/* M31: Move from AR (M-Unit) */
void Ia64AssembleM31(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_AR_R3(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM31(ins, *instr,  bundle);
#endif


}


/* M32: Move to CR */
void Ia64AssembleM32(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

  SET_CR_R3(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM32(ins, *instr,  bundle);
#endif


}


/* M33: Move from CR */
void Ia64AssembleM33(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_CR_R3(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM33(ins, *instr,  bundle);
#endif


}


/* M34: Allocate Register Stack Frame */
void Ia64AssembleM34(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
	/* zie ook Ia64DisassembleM34 voor meer info */
  
  SET_GR_R1(*instr, IA64INSOUT1(ins));

  *instr |= ((t_uint64)IA64INSIMM1(ins))<<20; 			/*sol*/
  *instr |= ((t_uint64)IA64INSIMM2(ins))<<13; 			/*sof*/
  *instr |= (((t_uint64)IA64INSIMM3(ins))>>3)<<27; 	/*sor*/	

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM34(ins, *instr,  bundle);
#endif

}


/* M35: Move to PSR */
void Ia64AssembleM35(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM35(ins, *instr,  bundle);
#endif


}


/* M36: Move from PSR */
void Ia64AssembleM36(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM36(ins, *instr,  bundle);
#endif


}


/* M37: Break/Nop (M-Unit) */
void Ia64AssembleM37(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_IMM21_M37(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM37(ins, *instr,  bundle);
#endif


}

/* M38: Probe - Register */
void Ia64AssembleM38(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));
  SET_GR_R2(*instr, IA64INSIN2(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM38(ins, *instr,  bundle);
#endif


}


/* M39: Probe - Immediate2 */
void Ia64AssembleM39(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

  *instr |= IA64INSIMM1(ins)<<13;

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM39(ins, *instr,  bundle);
#endif


}


/* M40: Probe Fault - Immediate2 */
void Ia64AssembleM40(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  *instr |= IA64INSIMM1(ins)<<13;

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM40(ins, *instr,  bundle);
#endif


}


/* M41: Translation Cache Insert */
void Ia64AssembleM41(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R2(*instr, IA64INSIN1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM41(ins, *instr,  bundle);
#endif


}


/* M42: Move to Indirect Register/Translation Register Insert */
void Ia64AssembleM42(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));
  SET_GR_R2(*instr, IA64INSIN2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM42(ins, *instr,  bundle);
#endif


}


/* M43: Move from Indirect Register */
void Ia64AssembleM43(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
	/* zie ook Ia64DisassembleM43 */
  SET_GR_R3(*instr, IA64INSIN2(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM43(ins, *instr,  bundle);
#endif


}


/* M44: Set/Rest User/System Mask */
void Ia64AssembleM44(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_IMM24_M44(*instr, IA64INSIMM1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM44(ins, *instr,  bundle);
#endif


}


/* M45: Translation Purge */
void Ia64AssembleM45(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));
  SET_GR_R2(*instr, IA64INSIN2(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM45(ins, *instr,  bundle);
#endif


}


/* M46: Translation Access */
void Ia64AssembleM46(t_ins * ins, t_uint64 * instr, t_bundle * bundle)
{
  SET_GR_R3(*instr, IA64INSIN1(ins));

  SET_GR_R1(*instr, IA64INSOUT1(ins));

#ifdef _DEBUG_ASSEMBLE_
	Ia64DisassembleM46(ins, *instr,  bundle);
#endif


}







/* vim: set shiftwidth=2: */
