/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ia64_bwh.h"
#include <string.h>
#include "../../datastructures/data_string.h"
#include "../../utils/util_error.h"
#include "ia64_opcode_label.h"
#include "../../kernel/diablo_iterators.h"

t_bool Ia64IsCircular(t_bbl * boven, t_bbl * onder);

static int adaptdebug=0;

t_cfg_edge * Ia64FindLargest(t_bbl * bbl, t_uint64 ft_count)
{
  t_cfg_edge * return_edge=NULL, * edge;
  t_uint64 max=ft_count;
  max+=(max==0)?0:10;
//  if(adaptdebug == 37) printf("in Findlargestedge, max: %d\n", max);
  BBL_FOREACH_PRED_EDGE(bbl, edge)
  {
    if(EDGE_CAT(edge)==ET_JUMP || EDGE_CAT(edge) ==ET_IPJUMP)
      if(EDGE_TAKEN(edge) > max)
	if(IA64INSPRED(BBL_INS_LAST(CFG_EDGE_HEAD(edge)))==IA64_P0)
//	  if(CFG_EDGE_HEAD(edge)!=CFG_EDGE_TAIL(edge))
	    if(!Ia64IsCircular(bbl, CFG_EDGE_HEAD(edge)))
	  {
	    return_edge = edge;
	    max = EDGE_TAKEN(edge);	    
	  }
  }
  return return_edge;
}

t_bool Ia64HasFallthrough(t_bbl * bbl)
{
  t_cfg_edge * edge;
  BBL_FOREACH_SUCC_EDGE(bbl, edge)
  {
    if(EDGE_CAT(edge)==ET_FALLTHROUGH) return TRUE;
  }
  return FALSE;
}

t_bool Ia64IsCircular(t_bbl * boven, t_bbl * onder)
{
  t_cfg_edge * i_edge;
  t_cfg_edge * ft_edge=NULL;
  t_bbl * i_bbl;
  t_bbl * ft_bbl=onder;;
  
  
  if(ft_bbl == boven) return TRUE;
  
  BBL_FOREACH_SUCC_EDGE(onder, i_edge)
  {
    if(EDGE_CAT(i_edge)==ET_FUNLINK ||
          EDGE_CAT(i_edge)==ET_SWILINK ||
          EDGE_CAT(i_edge)==ET_IPFUNLINK ||
          EDGE_CAT(i_edge)==ET_IPSWILINK ||
          EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
          EDGE_CAT(i_edge)==ET_IPFALLTHRU )
      ft_edge = i_edge;
  }
  while(ft_edge)
  {
//   CfgEdgePrint(ft_edge);
    ft_bbl = CFG_EDGE_TAIL(ft_edge);
    if(ft_bbl == boven) return TRUE;
    ft_edge = NULL;
    BBL_FOREACH_SUCC_EDGE(ft_bbl, i_edge)
    {
      if(EDGE_CAT(i_edge)==ET_FUNLINK ||
          EDGE_CAT(i_edge)==ET_SWILINK ||
          EDGE_CAT(i_edge)==ET_IPFUNLINK ||
          EDGE_CAT(i_edge)==ET_IPSWILINK ||
          EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
          EDGE_CAT(i_edge)==ET_IPFALLTHRU )
      ft_edge = i_edge;
    }
  }
  return FALSE;
}

t_ins * Ia64SearchForCompare(t_bbl *bbl);

t_bool Ia64ItIsWorthIt(t_uint64 jump, t_uint64 comp_ft)
{
  return FALSE;
}

void Ia64AdaptFallthrough(t_cfg * cfg)
{
  t_cfg_edge * jump_ft_edge, * comp_ft_edge, * jump_edge;
  t_cfg_edge * i_edge, * j_edge;
  t_ia64_ins * cmp_ins, * jump_ins;
  t_ia64_ins * new_ins;
  t_bbl * new_bbl;
  t_bbl * comp_bbl;
  t_bbl * comp_ft_bbl;

  CFG_FOREACH_EDGE(cfg, i_edge)
  {
/*    if(adaptdebug++ >= global_options.debugcounter)
continue;
//      FATAL(("stop"));*/
//    printf("------\n");
    if(EDGE_CAT(i_edge)==ET_JUMP || EDGE_CAT(i_edge)==ET_IPJUMP)
    {
      jump_edge = i_edge;
      if(Ia64IsCircular(CFG_EDGE_HEAD(jump_edge),CFG_EDGE_TAIL(jump_edge)))
	continue;
      if(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge))&&
	  IA64INSPRED(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)))!=IA64_P0&&
	  EDGE_EVALUATED(i_edge)>5&& 
	  EDGE_EVALUATED(i_edge)-EDGE_TAKEN(i_edge) < EDGE_TAKEN(i_edge))
      {
	comp_ft_edge =NULL;
	BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(jump_edge),j_edge)
	{
	  if(EDGE_CAT(j_edge)==ET_FALLTHROUGH || EDGE_CAT(j_edge)==ET_IPFALLTHRU || EDGE_CAT(j_edge)==ET_FUNLINK||EDGE_CAT(j_edge)==ET_IPFUNLINK||EDGE_CAT(j_edge)== ET_SWILINK || EDGE_CAT(j_edge) == ET_IPSWILINK)
	    comp_ft_edge = j_edge;
	}

	jump_ft_edge = NULL;
	BBL_FOREACH_SUCC_EDGE(CFG_EDGE_HEAD(jump_edge), j_edge)
	{
	  if(EDGE_CAT(j_edge)==ET_FALLTHROUGH || EDGE_CAT(j_edge)==ET_IPFALLTHRU) 
	    jump_ft_edge=j_edge;
	}

	if(!jump_ft_edge) FATAL(("help"));

	if(comp_ft_edge)
	{
	  if(BBL_NINS(CFG_EDGE_HEAD(comp_ft_edge))==0)
	    continue;//FATAL(("stop"));
	  if(! Ia64ItIsWorthIt(EDGE_TAKEN(jump_edge),EDGE_TAKEN(comp_ft_edge)))
	    continue;
	}

//	printf("i'm here\n");
	
	cmp_ins = (t_ia64_ins *) Ia64SearchForCompare(CFG_EDGE_HEAD(jump_edge));
	if(!cmp_ins) continue;
	jump_ins = (t_ia64_ins *) BBL_INS_LAST(CFG_EDGE_HEAD(jump_edge));

	BblPrint(INS_BBL(cmp_ins),2);
	//change predicate
	if(IA64INSOUT1(cmp_ins)==IA64INSPRED(jump_ins))
	{
	  if(IA64INSOUT2(cmp_ins)==IA64_P0)
	    continue;
	  IA64INSPRED(jump_ins)=IA64INSOUT2(cmp_ins);
	}
	else if(IA64INSOUT2(cmp_ins)==IA64INSPRED(jump_ins))
	{
	  if(IA64INSOUT1(cmp_ins)==IA64_P0)
	    continue;
	  IA64INSPRED(jump_ins)=IA64INSOUT1(cmp_ins);
	}
	else continue;
	
//	printf("--\n");
	
	BblPrint(INS_BBL(cmp_ins),2);

	//CfgEdgePrint(jump_edge);
	//change jump to fallthrough and vice versa
	if(EDGE_CAT(jump_edge)==ET_JUMP)
	  EDGE_CAT(jump_edge) = ET_FALLTHROUGH;
	else if(EDGE_CAT(jump_edge)==ET_IPJUMP)
	  EDGE_CAT(jump_edge) = ET_IPFALLTHRU;
	else FATAL(("we have a problem"));//CfgEdgePrint(jump_edge);

	if(EDGE_CAT(jump_ft_edge)==ET_FALLTHROUGH)
	  EDGE_CAT(jump_ft_edge) = ET_JUMP;
	else if(EDGE_CAT(jump_ft_edge)==ET_IPFALLTHRU)
	  EDGE_CAT(jump_ft_edge) = ET_IPJUMP;
	else FATAL(("we have a problem"));//CfgEdgePrint(fallthrough_edge);



	if(comp_ft_edge)
	{
	  FATAL(("nouw zeg\n"));
	  new_bbl = BblNew(cfg);
	  new_ins = InsNewForBbl(new_bbl);
	  
	  comp_bbl = CFG_EDGE_HEAD(comp_ft_edge);
	  comp_ft_bbl = CFG_EDGE_TAIL(comp_ft_edge);

	  new_ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)new_ins, (t_uint64)0, IA64_TYPE_B, IA64_BR_OPCODE, IA64_P0);

	  BBL_INS_FIRST(new_bbl)=BBL_INS_LAST(new_bbl)=(t_ins*)new_ins;
	  BBL_NINS(new_bbl)=1;

	  BBL_FUN(new_bbl)=BBL_FUN(comp_bbl);

	  BBL_NEXT_IN_FUN(new_bbl) = BBL_NEXT_IN_FUN(comp_bbl);
	  if(BBL_NEXT_IN_FUN(comp_bbl))
	    BBL_PREV_IN_FUN(BBL_NEXT_IN_FUN(comp_bbl)) = new_bbl;

	  BBL_PREV_IN_FUN(new_bbl) = comp_bbl;
	  BBL_NEXT_IN_FUN(comp_bbl) = new_bbl;

	  BBL_PROFILE_INFO(new_bbl) = EDGE_TAKEN(comp_ft_edge);

	  if(EDGE_IS_INTERPROC(comp_ft_edge))
	    CfgEdgeNew(cfg, new_bbl, comp_ft_bbl, ET_IPJUMP);
	  else
	    CfgEdgeNew(cfg, new_bbl, comp_ft_bbl, ET_JUMP);

	  CfgEdgeKill(comp_ft_edge);

	  CfgEdgeNew(cfg, comp_bbl, new_bbl, ET_FALLTHROUGH);


	}

      }

    }
  }
}

void Ia64AdaptFallthrough2(t_cfg * cfg)
{
  t_bbl * bbl, * ft_bbl, * largest_bbl;
  t_bool mogelijk;
  t_cfg_edge * edge, * ft_edge = NULL, * largest_edge, * i_edge, * j_edge, * new_edge;
  t_uint64 ft_count;
  t_ia64_ins * ins;
  t_cfg_edge * test_edge;
  t_ins * cmp_ins, * jump_ins;
  
  CFG_FOREACH_EDGE(cfg, i_edge)
  {
    if(EDGE_CAT(i_edge)==ET_JUMP || EDGE_CAT(i_edge)==ET_IPJUMP)
      if(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge))&&IA64INSPRED(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)))!=IA64_P0&&EDGE_EVALUATED(i_edge)>5&& EDGE_EVALUATED(i_edge)-EDGE_TAKEN(i_edge) < EDGE_TAKEN(i_edge))
      {
	ft_edge =NULL;
	BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(i_edge),j_edge)
	{
	  if(EDGE_CAT(j_edge)==ET_FALLTHROUGH || EDGE_CAT(j_edge)==ET_IPFALLTHRU || EDGE_CAT(j_edge)==ET_FUNLINK||EDGE_CAT(j_edge)==ET_IPFUNLINK||EDGE_CAT(j_edge)== ET_SWILINK || EDGE_CAT(j_edge) == ET_IPSWILINK)
	    ft_edge = j_edge;
	}
	if(Ia64IsCircular(CFG_EDGE_HEAD(i_edge),CFG_EDGE_TAIL(i_edge)))continue;
	if(ft_edge)
	{
	  if(EDGE_TAKEN(ft_edge)<EDGE_TAKEN(i_edge))
	  {
	    cmp_ins = Ia64SearchForCompare((t_bbl*)EDGE_HEAD(i_edge));
	    //if (cmp_ins) InsPrint(cmp_ins,0);
	    //InsPrint(BBL_INS_LAST(EDGE_HEAD(i_edge)));
//	    printf("edge: not-taken %d, taken %d, fallthrough %d\n", EDGE_EVALUATED(i_edge)-EDGE_TAKEN(i_edge), EDGE_TAKEN(i_edge), EDGE_TAKEN(ft_edge));
	    if(!BBL_INS_LAST(CFG_EDGE_HEAD(ft_edge))) FATAL(("trouble\n"));

	    {
	      t_bbl * ft_bbl = CFG_EDGE_HEAD(ft_edge);
	      t_bbl * new_bbl = BblNew(cfg);
	      t_ia64_ins * new_ins = InsNewForBbl(new_bbl);
	      bbl = CFG_EDGE_TAIL(ft_edge);

	      new_ins->instr_group = GetCurrentInstrGroupAndInc();
	      Ia64InitializeInstruction((t_ins *)new_ins, (t_uint64)0, IA64_TYPE_B, IA64_BR_OPCODE, IA64_P0);

	      BBL_INS_FIRST(new_bbl)=BBL_INS_LAST(new_bbl)=(t_ins*)new_ins;
	      BBL_NINS(new_bbl)=1;

	      BBL_FUN(new_bbl)=BBL_FUN(ft_bbl);

	      BBL_NEXT_IN_FUN(new_bbl) = BBL_NEXT_IN_FUN(ft_bbl);
	      if(BBL_NEXT_IN_FUN(ft_bbl))
		BBL_PREV_IN_FUN(BBL_NEXT_IN_FUN(ft_bbl)) = new_bbl;

	      BBL_PREV_IN_FUN(new_bbl) = ft_bbl;
	      BBL_NEXT_IN_FUN(ft_bbl) = new_bbl;

	      BBL_PROFILE_INFO(new_bbl) = EDGE_TAKEN(ft_edge);

	      if(EDGE_IS_INTERPROC(ft_edge))
		CfgEdgeNew(cfg, new_bbl, bbl, ET_IPJUMP);
	      else
		CfgEdgeNew(cfg, new_bbl, bbl, ET_JUMP);

	      CfgEdgeKill(ft_edge);

	      CfgEdgeNew(cfg, ft_bbl, new_bbl, ET_FALLTHROUGH);
	    }

	  }
	  else continue;
	}
	else if(!ft_edge)
	{
	  t_cfg_edge * jump_edge = i_edge;
	  t_cfg_edge * fallthrough_edge;
	  
	  /*
	  cmp_ins = Ia64SearchForCompare(EDGE_HEAD(i_edge));
	  if (cmp_ins) InsPrint(cmp_ins,0);
	  InsPrint(BBL_INS_LAST(EDGE_HEAD(i_edge)));
	  printf("edge: not-taken %d, taken %d\n", EDGE_EVALUATED(i_edge)-EDGE_TAKEN(i_edge), EDGE_TAKEN(i_edge));*/
	  cmp_ins = Ia64SearchForCompare((t_bbl *)EDGE_HEAD(i_edge));
	  jump_ins = BBL_INS_LAST(CFG_EDGE_HEAD(i_edge));
	  if (cmp_ins)
	  {

	    //change predicate
	    if(Ia64IsCircular(CFG_EDGE_HEAD(i_edge),CFG_EDGE_TAIL(i_edge)))FATAL(("circular"));
	    if(IA64INSOUT1(cmp_ins)==IA64INSPRED(jump_ins))
	    {
	      if(IA64INSOUT2(cmp_ins)==IA64_P0)
		continue;//FATAL(("only one predicate"));
	      IA64INSPRED(jump_ins)=IA64INSOUT2(cmp_ins);
	    }
	    else if(IA64INSOUT2(cmp_ins)==IA64INSPRED(jump_ins))
	    {
	      if(IA64INSOUT1(cmp_ins)==IA64_P0)
		continue;//FATAL(("only one predicate"));
	      IA64INSPRED(jump_ins)=IA64INSOUT1(cmp_ins);
	    }
	    else continue;//FATAL(("wrong cmp"));

	    fallthrough_edge = NULL;
	    BBL_FOREACH_SUCC_EDGE(CFG_EDGE_HEAD(i_edge), j_edge)
	      if(EDGE_CAT(j_edge)==ET_FALLTHROUGH || EDGE_CAT(j_edge)==ET_IPFALLTHRU) fallthrough_edge=j_edge;

	    if(!fallthrough_edge) FATAL(("help"));
	    if(EDGE_CAT(fallthrough_edge)!=ET_FALLTHROUGH && EDGE_CAT(fallthrough_edge)!=ET_IPFALLTHRU) FATAL(("bad boy"));
	    //change jump to fallthrough and vice versa
	    if(EDGE_CAT(jump_edge)==ET_JUMP)
	      EDGE_CAT(jump_edge) = ET_FALLTHROUGH;
	    else if(EDGE_CAT(jump_edge)==ET_IPJUMP)
	      EDGE_CAT(jump_edge) = ET_IPFALLTHRU;
	    else FATAL(("we have a problem"));//CfgEdgePrint(jump_edge);

	    if(EDGE_CAT(fallthrough_edge)==ET_FALLTHROUGH)
	      EDGE_CAT(fallthrough_edge) = ET_JUMP;
	    else if(EDGE_CAT(fallthrough_edge)==ET_IPFALLTHRU)
	      EDGE_CAT(fallthrough_edge) = ET_IPJUMP;
	    else FATAL(("we have a problem"));//CfgEdgePrint(fallthrough_edge);

	   // InsPrint(cmp_ins,0);
	   // InsPrint(jump_ins,0);
	   // printf("edge: not-taken %d, taken %d\n", EDGE_EVALUATED(i_edge)-EDGE_TAKEN(i_edge), EDGE_TAKEN(i_edge));
	  }
	  
	}
      }
  }
  
 // FATAL(("stop"));
  CFG_FOREACH_BBL(cfg, bbl)
  {    
    mogelijk = TRUE;
    ft_edge = NULL;
    ft_bbl = NULL;
    largest_edge = NULL;
    largest_bbl = NULL;
    BBL_FOREACH_PRED_EDGE(bbl, edge)
    {
      if(EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU || EDGE_CAT(edge)==ET_FUNLINK||EDGE_CAT(edge)==ET_IPFUNLINK||EDGE_CAT(edge)== ET_SWILINK || EDGE_CAT(edge) == ET_IPSWILINK)
	ft_edge = edge;
    }
      if(!ft_edge) ft_count = 0;
      else 
      {
	ft_count = EDGE_TAKEN(ft_edge);
	ft_bbl = CFG_EDGE_HEAD(ft_edge);
      }

      largest_edge = Ia64FindLargest(bbl, ft_count);
      if(largest_edge)
	largest_bbl = CFG_EDGE_HEAD(largest_edge);
      
      if(largest_edge == NULL) continue;
      else if(Ia64HasFallthrough(largest_bbl)) continue;
      //case: largest_edge != NULL, ft_edge == NULL, NINS(bbl) == 0
      else if(ft_edge == NULL && BBL_NINS(bbl)>0)
      {
	
//	printf("1: van 0 naar %d\n", EDGE_TAKEN(largest_edge));
	InsKill(BBL_INS_LAST(largest_bbl));
	
	if(EDGE_CAT(largest_edge)==ET_JUMP)
	  EDGE_CAT(largest_edge) = ET_FALLTHROUGH;
	else if(EDGE_CAT(largest_edge)==ET_IPJUMP)
	  EDGE_CAT(largest_edge) = ET_IPFALLTHRU;

//	new_edge = CfgEdgeNew(cfg, largest_bbl, bbl, ET_FALLTHROUGH);
//	EDGE_TAKEN(new_edge) = EDGE_TAKEN(largest_edge);
//	EDGE_EVALUATED(new_edge) = EDGE_EVALUATED(largest_edge);
//	CfgEdgeKill(largest_edge);
	
      }
      //case: largest_edge != NULL, ft_edge == NULL, NINS(bbl)==0
      else if(ft_edge == NULL)
      {
//	printf("1: missed chance %d\n", EDGE_TAKEN(largest_edge));

	//implement me
      }
      //case: largest_edge != NULL, ft_edge != NULL
      else if(
	  BBL_NINS(ft_bbl)!=0 && 
	  !Ia64IsBranch(BBL_INS_LAST(ft_bbl)) 
	)
      {
//	printf("2: van %d naar %d\n", EDGE_TAKEN(ft_edge), EDGE_TAKEN(largest_edge));

	InsKill(BBL_INS_LAST(largest_bbl));
	
	if(EDGE_IS_INTERPROC(largest_edge))
	  EDGE_CAT(largest_edge) = ET_IPFALLTHRU;
	else
	  EDGE_CAT(largest_edge) = ET_FALLTHROUGH;

	ft_bbl = CFG_EDGE_HEAD(ft_edge);
	ins = InsNewForBbl(ft_bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_B, IA64_BR_OPCODE, IA64_P0);
	InsInsertAfter((t_ins*)ins, (t_ins *)BBL_INS_LAST(ft_bbl));

	if(EDGE_IS_INTERPROC(ft_edge))
	  EDGE_CAT(ft_edge)=ET_IPJUMP;
	else
	  EDGE_CAT(ft_edge)=ET_JUMP;
      }
      else if(BBL_NINS(CFG_EDGE_HEAD(ft_edge))!=0 && 
	  Ia64IsBranch(BBL_INS_LAST(CFG_EDGE_HEAD(ft_edge))) )
      {
//	printf("3: van %d naar %d\n", EDGE_TAKEN(ft_edge), EDGE_TAKEN(largest_edge));
	
	t_bbl * new_bbl = BblNew(cfg);
	t_ia64_ins * new_ins = InsNewForBbl(new_bbl);

	new_ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)new_ins, (t_uint64)0, IA64_TYPE_B, IA64_BR_OPCODE, IA64_P0);

	BBL_INS_FIRST(new_bbl)=BBL_INS_LAST(new_bbl)=(t_ins*)new_ins;
	BBL_NINS(new_bbl)=1;

	BBL_FUN(new_bbl)=BBL_FUN(ft_bbl);

	BBL_NEXT_IN_FUN(new_bbl) = BBL_NEXT_IN_FUN(ft_bbl);
	if(BBL_NEXT_IN_FUN(ft_bbl))
	  BBL_PREV_IN_FUN(BBL_NEXT_IN_FUN(ft_bbl)) = new_bbl;

	BBL_PREV_IN_FUN(new_bbl) = ft_bbl;
	BBL_NEXT_IN_FUN(ft_bbl) = new_bbl;

	BBL_PROFILE_INFO(new_bbl) = EDGE_TAKEN(ft_edge);

	InsKill(BBL_INS_LAST(largest_bbl));
	
	if(EDGE_IS_INTERPROC(largest_edge))
	  EDGE_CAT(largest_edge) = ET_IPFALLTHRU;
	else
	  EDGE_CAT(largest_edge) = ET_FALLTHROUGH;

	if(EDGE_IS_INTERPROC(ft_edge))
	  CfgEdgeNew(cfg, new_bbl, bbl, ET_IPJUMP);
	else
	  CfgEdgeNew(cfg, new_bbl, bbl, ET_JUMP);
	
	CfgEdgeKill(ft_edge);


	CfgEdgeNew(cfg, ft_bbl, new_bbl, ET_FALLTHROUGH);
      }
      
      //case: largest_edge != NULL, ft_edge != NULL, BBL_NINS(ft_bbl) == 0
      else if(BBL_NINS(ft_bbl)==0)
      {
//	printf("2: missed chance %d to %d\n", EDGE_TAKEN(ft_edge), EDGE_TAKEN(largest_edge));
	//implement me
      }
      else FATAL(("unhandled case"));
    }
//  printf("uit adaptfallthrough\n");
}
void Ia64AdaptBranch(t_cfg * cfg){
  t_bbl * bbl;
  t_edge * edge;
  char * token;
  t_string_array * strings;

  //Ia64AdaptBranchMakeTable();
  CFG_FOREACH_BBL(cfg,bbl)
  {
    if(!IsHellNode(bbl) && 
	(!bbl->ins_first || 
	 IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA))

      BBL_FOREACH_SUCC_EDGE(bbl,(t_cfg_edge*)edge)
      {
	if ((EDGE_CAT(edge)==ET_JUMP || 
	      EDGE_CAT(edge)==ET_IPJUMP || 
	      EDGE_CAT(edge)==ET_CALL)&& !Ia64IsReturn(bbl->ins_last)&&!Ia64IsCheck(bbl->ins_last))
	{
	  if(EDGE_EVALUATED(edge)==0)
	    continue;
	  //printf("geevalueerd: %d, genomen: %d\n",EDGE_EVALUATED(edge), EDGE_TAKEN(edge));
	  strings=StringDivider(IA64OPCODEINFO(bbl->ins_last).desc,".",TRUE,FALSE);
	  //printf("%s\n",strings->strings[0]);
	 if(EDGE_EVALUATED(edge)==1 && (strings->nstrings<3||strings->strings[2] != "clr") && (strings->nstrings<4||strings->strings[3] != "clr") )
	    IA64INSOPCODE(bbl->ins_last)+=1;

	 if(EDGE_TAKEN(edge) > EDGE_EVALUATED(edge)-EDGE_TAKEN(edge))
	 {
	//   printf("---\n");
	//   printf("geevalueerd: %d, genomen: %d\n",EDGE_EVALUATED(edge), EDGE_TAKEN(edge));
//	   InsPrint(bbl->ins_last,0);
	   if(strcmp(strings->strings[2],"spnt")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=2;
	   } 
	   if(strcmp(strings->strings[2],"dpnt")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=2;
	   }
//	   InsPrint(bbl->ins_last,0);
	 }
	 
	 if(EDGE_TAKEN(edge) < EDGE_EVALUATED(edge)-EDGE_TAKEN(edge))
	 {
//	   printf("---\n");
//	   printf("geevalueerd: %d, genomen: %d\n",EDGE_EVALUATED(edge), EDGE_TAKEN(edge));
//	   InsPrint(bbl->ins_last,0);
	   if(strcmp(strings->strings[2],"sptk")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)+=2;
	   } 
	   if(strcmp(strings->strings[2],"dptk")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)+=2;
	   }
//	   InsPrint(bbl->ins_last,0);
	 }

	 if(EDGE_TAKEN(edge)==EDGE_EVALUATED(edge))
	 {
//	   printf("---\n");
//	   printf("geevalueerd: %d, genomen: %d\n",EDGE_EVALUATED(edge), EDGE_TAKEN(edge));
//	   InsPrint(bbl->ins_last,0);
	   if(strcmp(strings->strings[2],"spnt")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=2;
	   } 
	   if(strcmp(strings->strings[2],"dpnt")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=6;
	   }
	   if(strcmp(strings->strings[2],"dptk")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=4;
	   }
	   if(strcmp(strings->strings[3],"few")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)+=8;
	   }
//	   InsPrint(bbl->ins_last,0);
	 }
	 
	 if(EDGE_TAKEN(edge)==0)
	 {
//	   printf("---\n");
//	   printf("geevalueerd: %d, genomen: %d\n",EDGE_EVALUATED(edge), EDGE_TAKEN(edge));
//	   InsPrint(bbl->ins_last,0);
	   if(strcmp(strings->strings[2],"sptk")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)+=2;
	   } 
	   if(strcmp(strings->strings[2],"dptk")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=2;
	   }
	   if(strcmp(strings->strings[2],"dpnt")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=4;
	   }
	   if(strcmp(strings->strings[3],"many")==0)
	   {
	     IA64INSOPCODE(bbl->ins_last)-=8;
	   }
//	   InsPrint(bbl->ins_last,0);
	  }
	}
      }
  }
  //FATAL(("stop"));
}

/* vim: set shiftwidth=2: */
