/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be)
 ***********************************************************/

#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_liveness.h"
#include "../../diablo_options.h"
#include "ia64_instructions.h"
#include "ia64_registers.h"
#include "ia64_utils.h"
#include "ia64_opcodes.h"
#include "ia64_opcode_label.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_iterators.h"

#if 0
t_regset DefinitionCanBeRenamed(t_ins * ins)
{
  t_regset defined_reg;
  t_regset free_regs;
  t_ins * i_ins;

  RegsetSetSingleton(defined_reg, ARMINSREGA(ins));
  
  RegsetDup(free_regs,InsRegsLiveAfter(ins));
  RegsetSetInvers(free_regs);
  RegsetSetIntersect(free_regs,Registers(1));
  i_ins = INS_NEXT(ins);

  if (ARMINSREGA(ins)==ARM_REG_NONE)
    return 0;

  while (i_ins && !RegsetIsEmpty(defined_reg))
  {
     RegsetSetDiff(free_regs, InsRegsLiveAfter(i_ins));

     if (!INS_IS_CONDITIONAL(i_ins))
	RegsetSetDiff(defined_reg,INS_REGS_DEF(i_ins));

     i_ins = INS_NEXT(i_ins);
  }

  return free_regs;

}
#endif

static int Ia64OptCopyInner(t_ins * j_ins, t_reg copy, t_reg copied, t_uint32 *pcnt);
static void Ia64OptCopyRecurse(t_bbl * i_bbl, t_ia64_reg copy, t_ia64_reg copied, t_uint32 *pcnt);
void Ia64CleanupForward(t_ia64_reg copy, t_ia64_reg copied, t_bbl * i_bbl);
void Ia64OptReversePropagation(t_ins * i_ins);
static int reken=0;
static int cyclus=0;
static int resultaat=0;
void Ia64OptCopyPropagation(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_ins * i_ins, *j_ins;
  t_ia64_reg copied;
  t_ia64_reg copy;
  t_uint32 cnt = 0;
  t_uint32 teller=0;
  t_uint32 teller2=0;

  reken = 0;
//  if(cyclus==1)
//    return;
//  cyclus++;
  
  /*TODO: Turn backward copy propagation back on*/
/*  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (!BBL_FUN(i_bbl)) continue;

    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      if (INS_IS_CONDITIONAL(i_ins)) continue;

      if (!(IA64INS_IS_GR_MOV(i_ins) && IA64INSIMM1(i_ins)==0)) continue;

      if(!RegsetIn(InsRegsLiveAfter(i_ins),IA64INSIN1(i_ins)))
      {
	  reken++;
	if(reken <= global_options.debugcounter)
	{
	  Ia64OptReversePropagation(i_ins);	    
	}
      }
    }
  }*/
//  FATAL(("stop"));
  
//  if(cyclus==3)
//    FATAL(("stop"));
//  printf("%d\n",reken);
//  FATAL(("leg het moa stille;"));
  printf("Backward copy propagation killed %d instructions\n", resultaat);

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (!BBL_FUN(i_bbl)) continue;

    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      if (INS_IS_CONDITIONAL(i_ins)) continue;

      if (!(IA64INS_IS_GR_MOV(i_ins) && IA64INSIMM1(i_ins)==0)) continue;

      teller2++;
      //	  printf("Found a copy-instruction: ");
      //	  InsPrint(i_ins, 0);
      //	  if(!RegsetIn(InsRegsLiveAfter(i_ins),IA64INSIN1(i_ins)))
      //	  {
      //	    teller++;
      //	    printf("found one\n");
      //	    InsPrint(i_ins,0);
      //	    BblPrint(i_bbl,2);
      //	    if(reken < global_options.debugcounter)
      //	      Ia64OptReversePropagation(i_ins);	    
      //	  }

      //if (ARMINSREGC(i_ins)==ARM_REG_NONE) continue;
      // Detect a constant move C->A 

      copied = IA64INSIN1(i_ins); // Original value 
      copy = IA64INSOUT1(i_ins);   // Copy 

      j_ins=INS_NEXT(i_ins);

      if (Ia64OptCopyInner(j_ins,copy,copied, &cnt)) continue;
      InitNodeMarking();
      InitEdgeMarking();
      Ia64OptCopyRecurse(i_bbl, copy, copied, &cnt);
    }
  }

  //FATAL(("stop"));

  printf("Copy propagation updated %d registers.\n", cnt);
}

static int Ia64OptCopyInner(t_ins * j_ins, t_reg copy, t_reg copied, t_uint32 *pcnt)
{
   static int tel=0;
   while (j_ins!=NULL)
   {
      /*reg krijgt nieuwe waarde...*/
      /*ik zet dit ervoor, vb voor de postinc-meminstructies...*/
      /*nog alles eens checken wanneer het eig kwaad kan...*/
     //InsPrint(j_ins,0);
      if (RegsetIn(INS_REGS_DEF(j_ins),copy))
      {
	return -1;
      }
      
      if (IA64INSIN1(j_ins)==copy)
      {
	/* ADDL mag niet gerenamed worden! BRUNO */
	if ((IA64INSOPCODE(j_ins)!=IA64_ADDL_OPCODE))
	{
	  tel++;
	  //printf("2. propagating original r%d in stead of new r%d in: ", copied-IA64_GR, copy-IA64_GR);
	  //InsPrint(j_ins, 0);
	  IA64INSIN1(j_ins) = copied; /* use original */
	  INS_REGS_USE(j_ins) = Ia64UsedRegisters(j_ins);
	  //InsPrint(j_ins, 0);
	  (*pcnt)++;
	}
      }
      if (IA64INSIN2(j_ins)==copy)
      {
	 tel++;
	 //printf("1. propagating original r%d in stead of new r%d in: ", copied-IA64_GR, copy-IA64_GR);
	 //InsPrint(j_ins, 0);
	 IA64INSIN2(j_ins) = copied; /* use original */
	 INS_REGS_USE(j_ins) = Ia64UsedRegisters(j_ins);
	 //InsPrint(j_ins, 0);
	 (*pcnt)++;
      }
      if (IA64INSIN3(j_ins)==copy)
      {
	 tel++;
	 IA64INSIN3(j_ins) = copied; /* use original */
	 INS_REGS_USE(j_ins) = Ia64UsedRegisters(j_ins);
	 (*pcnt)++;
      }

     //InsPrint(j_ins,0);

      /*reg krijgt nieuwe waarde...*/
      if (RegsetIn(INS_REGS_DEF(j_ins),copied))
      {
	 return -1;
      }
      j_ins=INS_NEXT(j_ins);
   }
   return 0;
}
#if 0
static void OptCopyRecurse(t_bbl * i_bbl, t_reg copy, t_reg copied, t_uint32 *pcnt)
{
  t_cfg_edge * edge;
  if(!i_bbl)
	  return;
	  
   BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
   {
      if (T_NODE(CFG_EDGE_TAIL(edge))->npreds==1)
      {
	 if (!OptCopyInner(BBL_INS_FIRST(CFG_EDGE_TAIL(edge)),copy,copied, pcnt))
	 {
	    OptCopyRecurse(CFG_EDGE_TAIL(edge),copy,copied, pcnt);
	 }
      }
   }
}

t_bool Ia64CopyPropagationSkipEdge(t_cfg_edge * edge)
{
  return (EdgeIsMarked(edge) || EDGE_CAT(edge)==ET_FUNLINK || EDGE_CAT(edge)==ET_IPFUNLINK || EDGE_CAT(edge)==ET_SWITCH);
}

t_bool Ia64IsAlsoCopy(t_ia64_reg copy, t_ia64_reg copied, t_bbl * bbl)
{
  //printf("---\n %d\t%d\n", copy, copied);
  //BblPrint(bbl, 2);
  return FALSE;
  
}
t_bool Ia64CopyPropagationIsOkay(t_ia64_reg copy, t_ia64_reg copied, t_cfg_edge * edge)
{
  t_ia64_reg i_copy = copy;
  t_ia64_reg i_copied = copied;
  //if(Ia64CopyPropagationSkipEdge(edge)) return TRUE;
  if(copied == IA64_R1) return TRUE;
/*  if(EDGE_CAT(edge) == ET_CALL && 
      FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_HEAD(edge))) && 
      copied >= IA64_R32+FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(edge))))
    return TRUE;
  if(EDGE_CAT(edge) == ET_RETURN &&
      FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_TAIL(edge))) &&
      copied >= IA64_R32 &&
      copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge))))
    return TRUE;
*/  
  /*if(EDGE_CAT(edge)==ET_CALL)
  {
    if()
  }
  if(EDGE_CAT(edge)==ET_RETURN)
  {
    
  }*/
/*  if(EDGE_CAT(edge)!=ET_CALL && EDGE_CAT(edge)!=ET_RETURN)
  {
    if(Ia64IsAlsoCopy(i_copy, i_copied, CFG_EDGE_HEAD(edge)));
    return TRUE;
  }
*/  
  return FALSE;
}

static void Ia64OptCopyRecurseBerre(t_bbl * i_bbl, t_ia64_reg copy, t_ia64_reg copied, t_uint32 *pcnt)
{
  t_cfg_edge * edge, * edge2;
  t_ia64_reg i_copy, i_copied;
  t_bool mogelijk;
  t_bbl * j_bbl;
  
  if (NodeIsMarked(i_bbl)) return;
  
  BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
  {
//    InitNodeMarking();
    //if(EDGE_CAT(edge)==ET_CALL || EDGE_CAT(edge)==ET_FUNLINK || EDGE_CAT(edge)==ET_IPFUNLINK )
    //InitEdgeMarking();

    EdgeMark(edge);

    mogelijk=TRUE;

    BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge), edge2)
    {
      if(!Ia64CopyPropagationIsOkay(copy, copied, edge2))
	mogelijk = FALSE;
    }

    if(mogelijk)
    {
      i_copy = copy;
      i_copied = copied;
      if((copied >= IA64_R32 || copy >=IA64_R32))
      {

	if(EDGE_CAT(edge)==ET_CALL)
	{
	  mogelijk=FALSE;
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_HEAD(edge))))
	    mogelijk = FALSE;//gevonden=FALSE;

	  if(copied>=IA64_R32)
	    i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge)));
	  if(copy >=IA64_R32)
	    i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge)));

	  if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge))) && copied >= IA64_R32)  
	    mogelijk = FALSE;//i_copied = IA64_R0;

	  if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge))) && copy >= IA64_R32)  
	    mogelijk = FALSE;//i_copy = IA64_R0;
	}
	else //return FALSE;
	if (EDGE_CAT(edge)==ET_RETURN)
	{
	  mogelijk = FALSE;
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_TAIL(edge))))
	    mogelijk = FALSE;//gevonden=FALSE;

	  if(copied>=IA64_R32)
	    i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(edge)));
	  if(copy >=IA64_R32)
	    i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(edge)));

	  if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(edge))) )  
	    mogelijk = FALSE; //i_copied = IA64_R0;
	  if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(edge))) )  
	    mogelijk = FALSE;//i_copy = IA64_R0;
	}
      }
      if(mogelijk) 
      {
	NodeMark(i_bbl);
	if(! Ia64OptCopyInner(BBL_INS_FIRST(CFG_EDGE_TAIL(edge)),i_copy,i_copied,pcnt))
	  Ia64OptCopyRecurse(CFG_EDGE_TAIL(edge), i_copy, i_copied, pcnt);
      }
    }
  }
}
#endif

static void Ia64OptCopyRecurse(t_bbl * i_bbl, t_ia64_reg copy, t_ia64_reg copied, t_uint32 *pcnt)
{
  t_cfg_edge * edge,* edge2;
  t_uint32 real_preds;
  t_ia64_reg i_copy;
  t_ia64_reg i_copied;
  if (NodeIsMarked(i_bbl)) return;
  BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
  {
//    InitEdgeMarking();
    i_copy = copy;
    i_copied = copied;

    if(EDGE_CAT(edge)==ET_RETURN)
    {
      if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_TAIL(edge))))
	continue;

      if(copied>=IA64_R32)
	i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(edge)));
      if(copy >=IA64_R32)
	i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(edge)));

      if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(edge)))||
       	copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(edge))))
      {	
	continue;
      }
    }

    if(EDGE_CAT(edge)==ET_CALL)
    {
      if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_HEAD(edge))))
	continue;

      if(copied>=IA64_R32)
	i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge)));
      if(copy >=IA64_R32)
	i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge)));

      if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge))) && copied >= IA64_R32 ||
	  copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(edge))) && copy >= IA64_R32) 
      {	
	continue;
      }
    }
  
    EdgeMark(edge);

    real_preds=0;

    BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),edge2)
    {
      /* Fun link edges don't count as predecessor edges for copy propagation */
      //says who???????????????????
      //CHANGED !
      //if ((EDGE_CAT(edge2)==ET_FUNLINK) || (EDGE_CAT(edge2)==ET_IPFUNLINK)) continue;
      //if (EDGE_CAT(edge2)==ET_SWITCH)  continue;
      if (EdgeIsMarked(edge2)) continue;
      real_preds++;
    }
    if (real_preds==0)
    {
      NodeMark(i_bbl);
      if (!Ia64OptCopyInner(BBL_INS_FIRST(CFG_EDGE_TAIL(edge)),i_copy,i_copied,pcnt))
      {
	Ia64OptCopyRecurse(CFG_EDGE_TAIL(edge),i_copy,i_copied,pcnt);
      }
    }
  }
}

t_bool Ia64ReverseCopyEdge(t_cfg_edge * i_edge)
{
  return //(EDGE_CAT(i_edge) == ET_FALLTHROUGH || EDGE_CAT(i_edge)==ET_JUMP);

    (EDGE_CAT(i_edge) != ET_CALL && EDGE_CAT(i_edge)!=ET_RETURN);
    /*(EDGE_CAT(i_edge) == ET_FALLTHROUGH
      || EDGE_CAT(i_edge)==ET_JUMP
      || EDGE_CAT(i_edge)==ET_IPJUMP
      || EDGE_CAT(i_edge)==ET_IPFALLTHRU
      || EDGE_CAT(i_edge)==ET_FUNLINK
      || EDGE_CAT(i_edge)==ET_SWI
      || EDGE_CAT(i_edge)==ET_SWILINK
      || EDGE_CAT(i_edge)==ET_IPSWILINK
      || EDGE_CAT(i_edge)==ET_IPFUNLINK
      );*/
}
t_bool Ia64IsLink(t_cfg_edge * i_edge)
{
  return (
      EDGE_CAT(i_edge)==ET_FUNLINK
//      || EDGE_CAT(i_edge)==ET_SWILINK
//      || EDGE_CAT(i_edge)==ET_IPSWILINK
      || EDGE_CAT(i_edge)==ET_IPFUNLINK
      );
}
t_bool Ia64IsDead(t_bbl * i_bbl, t_ia64_reg i_reg)
{
  if(i_reg == IA64_R0)
    return TRUE;
  else return !RegsetIn(InsRegsLiveBefore(BBL_INS_FIRST(i_bbl)),i_reg);
}

t_bool Ia64IsUsed(t_ins * i_ins, t_ia64_reg i_reg)
{
  if(i_reg == IA64_R0)
    return FALSE;
  else return RegsetIn(INS_REGS_USE(i_ins),i_reg);
}

t_bool Ia64CopiedIsReplacableForward(t_ia64_reg copy, t_ia64_reg copied, t_bbl * i_bbl)
{
  t_ins * i_ins = BBL_INS_FIRST(i_bbl);
  t_cfg_edge * i_edge;
  t_bool mogelijk=TRUE;
  t_bool inthisbbl = FALSE;
  t_ia64_reg i_copy;
  t_ia64_reg i_copied;
//  printf("called replacableforward %d %d\n", copy, copied);
//  BblPrint(i_bbl,2);
  /*als hellnode:onmogelijk*/
  if(IsHellNode(i_bbl)) return FALSE;
  
  if(BBL_INS_FIRST(i_bbl)&&Ia64IsDead(i_bbl, copy)&&Ia64IsDead(i_bbl, copied))
    return TRUE;

  if(BBL_INS_FIRST(i_bbl)&&!Ia64IsDead(i_bbl, copy))
    return FALSE;
  /*ga na of copied geherdefinieerd wordt in dit bbl voor copy gebruikt of overschreven wordt*/
  while(i_ins)
  {
    if(Ia64IsUsed(i_ins, copy))
    {
      return FALSE;
    }
    if(RegsetIn(INS_REGS_DEF(i_ins),copy))
    {
      return FALSE;
    }

    if(RegsetIn(INS_REGS_USE(i_ins), copied) && IA64INSOPCODE(i_ins)==IA64_ADDL_OPCODE && copied <=IA64_R3)
      if(copy >IA64_R3)
	return FALSE;
    
    if(RegsetIn(INS_REGS_DEF(i_ins),copied))
    {
      //new
      //if(IA64INSOPCODE(i_ins)==IA64_ADDL_OPCODE)
//	return FALSE;
      //end new
      if(IA64INSOUT1(i_ins)==copied&&!INS_IS_CONDITIONAL(i_ins))
      {
	inthisbbl=TRUE;
	break;
      }
    }
    i_ins = INS_NEXT(i_ins);
  }
  if(inthisbbl) return TRUE;

  /*ga na of verder kunnen gaan.*/
  BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
  {
    if(Ia64IsLink(i_edge))
      continue;
    if(!EdgeIsMarked(i_edge))
    {
      i_copy = copy;
      i_copied = copied;
      if((copied >= IA64_R32 || copy >=IA64_R32) && !Ia64ReverseCopyEdge(i_edge))
      {
	//	  return FALSE;

	if(EDGE_CAT(i_edge)==ET_CALL)
	{
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_HEAD(i_edge))))
	    return FALSE;//gevonden=FALSE;

	  if(copied>=IA64_R32)
	    i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	  if(copy >=IA64_R32)
	    i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));

	  if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copied >= IA64_R32)  
	    i_copied = IA64_R0;
	  //	      FATAL(("this is strange"));
	  if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copy >= IA64_R32)  
	    i_copy = IA64_R0;

	}
	else //return FALSE;
	if (EDGE_CAT(i_edge)==ET_RETURN)
	{
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_TAIL(i_edge))))
	    return FALSE;//gevonden=FALSE;

	  if(copied>=IA64_R32)
	    i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	  if(copy >=IA64_R32)
	    i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));

	  if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	    i_copied = IA64_R0;
	  //FATAL(("this is strange"));
	  if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	    i_copy = IA64_R0;
	  //FATAL(("this is strange"));

	  //  FATAL(("ik ben hier"));
	  //berresaysjump
	}
	else FATAL(("shouldn't get here. see ia64reversecopyedge"));


	/*try to adapt if not possible: false*/
      }
//      CfgEdgePrint(i_edge,0);
      EdgeMark(i_edge);
      if(!Ia64CopiedIsReplacableForward(i_copy, i_copied, CFG_EDGE_TAIL(i_edge)))
	return FALSE;
    }
  }
  return TRUE;
}

t_bool Ia64CopiedIsReplacableBackward(t_ia64_reg copy, t_ia64_reg copied, t_bbl * i_bbl)
{
  t_bool mogelijk=TRUE;
  t_bool inthisbbl=FALSE;
  t_cfg_edge * i_edge;
  t_ins * i_ins;
  t_ia64_reg i_copy;
  t_ia64_reg i_copied;
  
//  printf("entered Replacable Backward %d %d\n", copy, copied);
//  BblPrint(i_bbl,2);
  /*als hellnode: onmogelijk*/
  if(IsHellNode(i_bbl)) return FALSE;

  if(!FUN_ISALLOCCLEAN(BBL_FUN(i_bbl))) return FALSE;
  
  /*ga na of kan vervangen worden bij opvolgers*/
  BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
  {
    if(Ia64IsLink(i_edge))
      continue;
    if(!EdgeIsMarked(i_edge))
    {
      i_copy=copy;
      i_copied=copied;
      if((copied >= IA64_R32 || copy >=IA64_R32) && !Ia64ReverseCopyEdge(i_edge))
      {
	//return FALSE;
	
	if(EDGE_CAT(i_edge)==ET_CALL)
	{
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_HEAD(i_edge))))
	    return FALSE;//gevonden=FALSE;
	  
	  if(copied>=IA64_R32)
	    i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	  if(copy >=IA64_R32)
	    i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	  
	  if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copied >= IA64_R32)  
	    i_copied = IA64_R0;
	  //	      FATAL(("this is strange"));
	  if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copy >= IA64_R32)  
	    i_copy = IA64_R0;

	}
	else //return FALSE;
	if (EDGE_CAT(i_edge)==ET_RETURN)
	{
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_TAIL(i_edge))))
	    return FALSE;//gevonden=FALSE;

	  if(copied>=IA64_R32)
	    i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	  if(copy >=IA64_R32)
	    i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	  
	  if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	    i_copied = IA64_R0;
	  //FATAL(("this is strange"));
	  if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	    i_copy = IA64_R0;
	  //FATAL(("this is strange"));

	  //  FATAL(("ik ben hier"));
//berresaysjump
	}
	else FATAL(("shouldn't get here. see ia64reversecopyedge"));
      }
//      CfgEdgePrint(i_edge,0);
//      EdgeMark(i_edge);
//      printf("checking replacableforward with %d and %d", i_copy, i_copied);
      if(!Ia64CopiedIsReplacableForward(i_copy, i_copied, CFG_EDGE_TAIL(i_edge)))
	return FALSE;
    }
  }

  
  /*zoek definitie in dit bbl*/
  i_ins = BBL_INS_LAST(i_bbl);
  while(i_ins)
  {
    if(Ia64IsUsed(i_ins, copy))
    {
      return FALSE;
    }
    if(RegsetIn(INS_REGS_DEF(i_ins),copy))
    {
      return FALSE;
    }
    if(RegsetIn(INS_REGS_USE(i_ins), copied) && IA64INSOPCODE(i_ins)==IA64_ADDL_OPCODE && copied <=IA64_R3)
      if(copy >IA64_R3)
	return FALSE;
    if(RegsetIn(INS_REGS_DEF(i_ins),copied))
    {
//new
//if(IA64INSOPCODE(i_ins)==IA64_ADDL_OPCODE)
//	return FALSE;
//end new
      
      if( IA64INSOUT1(i_ins)==copied )
      {
	if(copy == IA64_R0)
	  return FALSE;
	if(!INS_IS_CONDITIONAL(i_ins))
	{
	  inthisbbl=TRUE;
	  break;
	}
      }
    }
    i_ins = INS_PREV(i_ins);
  }
  
 
  if(inthisbbl) return TRUE;
  
  BBL_FOREACH_PRED_EDGE(i_bbl, i_edge)
  {
    if(Ia64IsLink(i_edge))
      continue;
    if(!EdgeIsMarked(i_edge))
    {
      i_copy=copy;
      i_copied=copied;
      if((copied >= IA64_R32 || copy >=IA64_R32) && !Ia64ReverseCopyEdge(i_edge))
      {
	if(EDGE_CAT(i_edge)==ET_CALL)
	{
//	  return FALSE;
	  // printf("een\n");
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_HEAD(i_edge))))
	    return FALSE;//gevonden=FALSE;

	  if(copied>=IA64_R32)
	    i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	  if(copy >=IA64_R32)
	    i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	  
	  if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	    i_copied = IA64_R0;
	  //	      FATAL(("this is strange"));
	  if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	    i_copy = IA64_R0;
	  //FATAL(("this is strange"));

	  // FATAL(("ik ben hier"));
	  //  printf("ten\n");
	}
	else if(EDGE_CAT(i_edge)==ET_RETURN)
	{
	  //return FALSE;
	  
	  if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_TAIL(i_edge))))
	    return FALSE;//gevonden=FALSE;

	  if(copied>=IA64_R32)
	    i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	  if(copy >=IA64_R32)
	    i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	  
	  if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copied >= IA64_R32)  
	    i_copied = IA64_R0;
	  //FATAL(("this is strange"));
	  if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copy >= IA64_R32)  
	    i_copy = IA64_R0;
	  //FATAL(("this is strange"));

	  //  FATAL(("ik ben hier"));
	}
	else FATAL(("shouldn't get here. see ia64reversecopyedge"));
      }
      /*if((copied >= IA64_R32 || copy >=IA64_R32) && !Ia64ReverseCopyEdge(i_edge))
	return FALSE;*/
//      CfgEdgePrint(i_edge,0);
      EdgeMark(i_edge);
//      printf("calling replacable backward with %d and %d\n", i_copy, i_copied);
      if(!Ia64CopiedIsReplacableBackward(i_copy, i_copied, CFG_EDGE_HEAD(i_edge)))
	return  FALSE;
    }
  }
  return TRUE;
}

void Ia64CleanupBackward(t_ia64_reg copy, t_ia64_reg copied, t_bbl * i_bbl)
{
  t_cfg_edge * i_edge;
  t_bool done=FALSE;
  t_ins * i_ins;
  t_ia64_reg i_copy;
  t_ia64_reg i_copied;

//  printf("called cleanupbackward met %d en %d\n", copy, copied);
//  BblPrint(i_bbl,2);
  
  BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
  {
    if(Ia64IsLink(i_edge))
      continue;
    if(!EdgeIsMarked(i_edge))
    {
//      CfgEdgePrint(i_edge,0);
      //      !!!!!!!!!!!!!!
      i_copy=copy;
      i_copied=copied;

      if(EDGE_CAT(i_edge)==ET_CALL)
      {
	if(copied>=IA64_R32)
	  i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	if(copy >=IA64_R32)
	  i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));

	if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copied >= IA64_R32)  
	  return;//i_copied = IA64_R0;
	if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copy >= IA64_R32)  
	  return;//i_copy = IA64_R0;
      }
      if(EDGE_CAT(i_edge)==ET_RETURN)
      {
	if(copied>=IA64_R32)
	  i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	if(copy >=IA64_R32)
	  i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	
	if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	  return;//i_copied = IA64_R0;
	if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	  return;//i_copy = IA64_R0;
      }
      
      EdgeMark(i_edge);
      //!!!!!!!!!!!!!!!!!!!!
//      printf("b calling cleanupforward with %d and %d\n", i_copy, i_copied);
      Ia64CleanupForward(i_copy, i_copied, CFG_EDGE_TAIL(i_edge));
    }
  }
  
  i_ins = BBL_INS_LAST(i_bbl);
  while(i_ins)
  {
    if(RegsetIn(INS_REGS_DEF(i_ins),copied))
    {
      if(IA64INSOUT1(i_ins)==copied)
      {
	IA64INSOUT1(i_ins)=copy;
  	INS_REGS_USE(i_ins) = Ia64UsedRegisters(i_ins);
	INS_REGS_DEF(i_ins) = Ia64DefinedRegisters(i_ins);
	if(!INS_IS_CONDITIONAL(i_ins))
	{
//	  printf("na cleanupbackward\n");
//	  BblPrint(i_bbl,2);
	  return;
	}
      }
      else 
      {
	if(IA64INSIN1(i_ins)==copied)
	{
	  IA64INSIN1(i_ins)=copy;
	}
	if(IA64INSIN2(i_ins)==copied)
	{
	  IA64INSIN2(i_ins)=copy;
	}
	if(IA64INSIN3(i_ins)==copied)
	{
	  IA64INSIN3(i_ins)=copy;
	}
	INS_REGS_USE(i_ins) = Ia64UsedRegisters(i_ins);
	INS_REGS_DEF(i_ins) = Ia64DefinedRegisters(i_ins);
      }
    }
    if(Ia64IsUsed(i_ins, copied))
    {
      if(IA64INSIN1(i_ins)==copied)
	IA64INSIN1(i_ins)=copy;
      if(IA64INSIN2(i_ins)==copied)
	IA64INSIN2(i_ins)=copy;
      if(IA64INSIN3(i_ins)==copied)
	IA64INSIN3(i_ins)=copy;
      INS_REGS_USE(i_ins) = Ia64UsedRegisters(i_ins);
      INS_REGS_DEF(i_ins) = Ia64DefinedRegisters(i_ins);
    }
    i_ins = INS_PREV(i_ins);
  }
  BBL_FOREACH_PRED_EDGE(i_bbl, i_edge)
  {
    if(Ia64IsLink(i_edge))
      continue;
    
    if(!EdgeIsMarked(i_edge))
    {
      
//      CfgEdgePrint(i_edge,0);
      i_copy=copy;
      i_copied=copied;

      if(EDGE_CAT(i_edge)==ET_CALL)
      {
	if(copied>=IA64_R32)
	  i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	if(copy >=IA64_R32)
	  i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	
	if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	  return;//i_copied = IA64_R0;
	if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	  return;//i_copy = IA64_R0;
      }
      if(EDGE_CAT(i_edge)==ET_RETURN)
      {
	if(copied>=IA64_R32)
	  i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	if(copy >=IA64_R32)
	  i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));

	if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copied >= IA64_R32)  
	  return;//i_copied = IA64_R0;
	if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copy >= IA64_R32)  
	  return;//i_copy = IA64_R0;
      }
      
      EdgeMark(i_edge);
  //    printf("calling cleanupbackward with %d and %d\n", i_copy, i_copied);
      Ia64CleanupBackward(i_copy, i_copied, CFG_EDGE_HEAD(i_edge));
    }
  }
//  printf("na cleanupbackward\n");
//  BblPrint(i_bbl,2);
}
void Ia64CleanupForward(t_ia64_reg copy, t_ia64_reg copied, t_bbl * i_bbl)
{
  t_cfg_edge * i_edge;
  t_bool done=FALSE;
  t_ins * i_ins;
  t_ia64_reg i_copy;
  t_ia64_reg i_copied;

//  printf("called cleanupforward met %d en %d", copy, copied);
//  BblPrint(i_bbl, 2);

  if(BBL_INS_FIRST(i_bbl)&&!RegsetIn(InsRegsLiveBefore(BBL_INS_FIRST(i_bbl)),copy)&&!RegsetIn(InsRegsLiveBefore(BBL_INS_FIRST(i_bbl)),copied))
    return;
  
  
//  printf("nog in de running\n");
  
  i_ins = BBL_INS_FIRST(i_bbl);
  
  while(i_ins)
  {
    if(RegsetIn(INS_REGS_DEF(i_ins),copied))
    {
      if(IA64INSOUT1(i_ins)==copied)
      {
	if(!INS_IS_CONDITIONAL(i_ins))
	{
//	  printf("na cleanupforward\n");
//	  BblPrint(i_bbl,2);
	  return;
	}
      }
      else
      {
	if(IA64INSIN1(i_ins)==copied)
	{
	  IA64INSIN1(i_ins)=copy;
	}
	if(IA64INSIN2(i_ins)==copied)
	{
	  IA64INSIN2(i_ins)=copy;
	}
	if(IA64INSIN3(i_ins)==copied)
	{
	  IA64INSIN3(i_ins)=copy;
	}
	INS_REGS_USE(i_ins) = Ia64UsedRegisters(i_ins);
	INS_REGS_DEF(i_ins) = Ia64DefinedRegisters(i_ins);
      }
    }
    if(Ia64IsUsed(i_ins, copied))
    {
      if(IA64INSIN1(i_ins)==copied)
	IA64INSIN1(i_ins)=copy;
      if(IA64INSIN2(i_ins)==copied)
	IA64INSIN2(i_ins)=copy;
      if(IA64INSIN3(i_ins)==copied)
	IA64INSIN3(i_ins)=copy;
      INS_REGS_USE(i_ins) = Ia64UsedRegisters(i_ins);
      INS_REGS_DEF(i_ins) = Ia64DefinedRegisters(i_ins);
    }
    i_ins = INS_NEXT(i_ins);
  }
//  printf("nog in de running\n");
  BBL_FOREACH_SUCC_EDGE(i_bbl, i_edge)
  {
    
//    CfgEdgePrint(i_edge, 0);
    if(Ia64IsLink(i_edge))
      continue;
//    CfgEdgePrint(i_edge, 0);
    if(!EdgeIsMarked(i_edge))
    {
//      printf("ik ben niet gemarkeerd\n");
      //!!!!!!!!!
      i_copy=copy;
      i_copied=copied;

      if(EDGE_CAT(i_edge)==ET_CALL)
      {
	if(copied>=IA64_R32)
	  i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	if(copy >=IA64_R32)
	  i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	
	if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copied >= IA64_R32)  
	  return;//i_copied = IA64_R0;
	if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))) && copy >= IA64_R32)  
	  return;//i_copy = IA64_R0;
      }
      if(EDGE_CAT(i_edge)==ET_RETURN)
      {
	if(copied>=IA64_R32)
	  i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	if(copy >=IA64_R32)
	  i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));

	if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	  return;//i_copied = IA64_R0;
	if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_TAIL(i_edge))) )  
	  return;//i_copy = IA64_R0;
      }
      
      EdgeMark(i_edge);
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//      printf("f calling cleanupforward with %d and %d\n", copy, copied);
      Ia64CleanupForward(i_copy, i_copied, CFG_EDGE_TAIL(i_edge));
    }
  }
//  printf("na cleanupforward\n");
//  BblPrint(i_bbl,2);
}

void Ia64Cleanup(t_ia64_reg copy, t_ia64_reg copied, t_ins * i_ins)
{
  t_bbl * i_bbl = INS_BBL(i_ins);
  t_ins * j_ins = INS_PREV(i_ins);
  t_cfg_edge * i_edge;
  t_bool done=FALSE;
  t_ia64_reg i_copy;
  t_ia64_reg i_copied;  
  
  printf("decided to clean up: %d %d\n", copy, copied);
  DiabloPrint(stdout,"@I\n",i_ins);
  printf("in bbl:\n");
  BblPrint(INS_BBL(i_ins),2);
  
  while(j_ins)
  {
      if(RegsetIn(INS_REGS_DEF(j_ins),copied))
      {
	if(IA64INSOUT1(j_ins)==copied)
	{
	  IA64INSOUT1(j_ins)=copy;
	  INS_REGS_USE(j_ins) = Ia64UsedRegisters(j_ins);
	  INS_REGS_DEF(j_ins) = Ia64DefinedRegisters(j_ins);
	  if(!INS_IS_CONDITIONAL(j_ins))
	  {
	    done=TRUE;
	    break;
	  }
	}
	else
	{
	  if(IA64INSIN1(j_ins)==copied)
	  {
	    IA64INSIN1(j_ins)=copy;
	  }
	  if(IA64INSIN2(j_ins)==copied)
	  {
	    IA64INSIN2(j_ins)=copy;
	  }
	  if(IA64INSIN3(j_ins)==copied)
	  {
	    IA64INSIN3(j_ins)=copy;
	  }
	  INS_REGS_USE(j_ins) = Ia64UsedRegisters(j_ins);
	  INS_REGS_DEF(j_ins) = Ia64DefinedRegisters(j_ins);
	}
      }
    if(Ia64IsUsed(j_ins, copied))
    {
      if(IA64INSIN1(j_ins)==copied)
	IA64INSIN1(j_ins)=copy;
      if(IA64INSIN2(j_ins)==copied)
	IA64INSIN2(j_ins)=copy;
      if(IA64INSIN3(j_ins)==copied)
	IA64INSIN3(j_ins)=copy;
      INS_REGS_USE(j_ins) = Ia64UsedRegisters(j_ins);
      INS_REGS_DEF(j_ins) = Ia64DefinedRegisters(j_ins);
    }
    j_ins = INS_PREV(j_ins);
  }

  
  if(done)
  { 
//    printf("killed (executed %d times): ",BBL_PROFILE_INFO(INS_BBL(i_ins)));
    resultaat++;
//    InsPrint(i_ins,0);
    InsKill(i_ins);
    
//    printf("na cleanup1:\n");
//    BblPrint(INS_BBL(i_ins),2);
    return;
  }
  
  InitEdgeMarking();

  BBL_FOREACH_PRED_EDGE(i_bbl, i_edge)
  {
    i_copy=copy;
    i_copied=copied;

    if(Ia64IsLink(i_edge))
      continue;
    if(EDGE_CAT(i_edge)==ET_CALL)
    {
      //      printf();
      if(copied>=IA64_R32)
	i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
      if(copy >=IA64_R32)
	i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
      
      if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	i_copied = IA64_R0;
      //	      FATAL(("this is strange"));
      if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	i_copy = IA64_R0;
    }
    if(EDGE_CAT(i_edge)==ET_RETURN)
    {
      //      printf();
      if(copied>=IA64_R32)
	i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
      if(copy >=IA64_R32)
	i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
      
      if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copied >= IA64_R32)  
	i_copied = IA64_R0;
      //FATAL(("this is strange"));
      if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copy >= IA64_R32)  
	i_copy = IA64_R0;
    }
//    FATAL(("strange"));
    
//    CfgEdgePrint(i_edge,0);
//    printf("%d, %d",i_copy,i_copied);
    EdgeMark(i_edge);
//    printf("calling clenaupbackward withc %d and %d\n", i_copy, i_copied);

    Ia64CleanupBackward(i_copy, i_copied, CFG_EDGE_HEAD(i_edge));
  }
//  printf("killed (executed %d times): ",BBL_PROFILE_INFO(INS_BBL(i_ins)));
    resultaat++;
//  InsPrint(i_ins,0);
  InsKill(i_ins);

//  printf("na cleanup2:\n");
//  BblPrint(INS_BBL(i_ins),2);
}
void Ia64OptReversePropagation(t_ins * i_ins)
{
  t_ia64_reg copy = IA64INSOUT1(i_ins);
  t_ia64_reg copied = IA64INSIN1(i_ins);
  t_ia64_reg i_copy = IA64INSOUT1(i_ins);
  t_ia64_reg i_copied = IA64INSIN1(i_ins);
  t_ins * j_ins = INS_PREV(i_ins);
  t_bool gevonden=FALSE;
  t_bool mogelijk=TRUE;
  t_bbl * i_bbl = INS_BBL(i_ins);
  t_cfg_edge * i_edge, * j_edge;

//  printf("--------\n");
  
  while(j_ins)
  {
    if(RegsetIn(INS_REGS_USE(j_ins),copy))
    {
      return;
    }
    if(RegsetIn(INS_REGS_DEF(j_ins),copy))
    {
      return;
    }
    if(RegsetIn(INS_REGS_USE(j_ins), copied) && IA64INSOPCODE(j_ins)==IA64_ADDL_OPCODE && copied <=IA64_R3)
      if(copy >IA64_R3)
	return;
    if(RegsetIn(INS_REGS_DEF(j_ins),copied))
    {
      if(IA64INSOUT1(j_ins)==copied&&!INS_IS_CONDITIONAL(j_ins))
      {
	gevonden=TRUE;
	break;
      }
    }
    j_ins = INS_PREV(j_ins);
  }
  
  if(gevonden)
  {
    Ia64Cleanup(copy,copied,i_ins);
    return;
  }
//  printf("trying to eliminate\n");
//  InsPrint(i_ins,0);
//  printf("in BBL:\n");
//  BblPrint(i_bbl,2);
  if(!gevonden )
  {
      InitEdgeMarking();

      gevonden = TRUE;
      BBL_FOREACH_PRED_EDGE(i_bbl, i_edge)
      {
	i_copy=copy;
	i_copied=copied;

	if(Ia64IsLink(i_edge))
	  continue;
	if((copied >= IA64_R32 || copy >=IA64_R32) && !Ia64ReverseCopyEdge(i_edge))
	{
	  if(EDGE_CAT(i_edge)==ET_CALL)
	  {
	   // return;
	   // printf("een\n");
	    if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_HEAD(i_edge))))
	      return;//gevonden=FALSE;
	    
	    if(copied>=IA64_R32)
	      i_copied += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	    
	    if(copy >=IA64_R32)
	      i_copy += FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge)));
	    
	    if(copied >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	      i_copied = IA64_R0;
//	      FATAL(("this is strange"));
	    if(copy >= IA64_R32 + FUN_SOO(BBL_FUN(CFG_EDGE_HEAD(i_edge))) )  
	      i_copy = IA64_R0;
	      //FATAL(("this is strange"));
	    
	   // FATAL(("ik ben hier"));
	  //  printf("ten\n");*/
	  }
	  else if(EDGE_CAT(i_edge)==ET_RETURN)
	  {
	    //return;
	    if(!FUN_ISALLOCVALID(BBL_FUN(CFG_EDGE_TAIL(i_edge))))
	      return;//gevonden=FALSE;
	   
	    if(copied>=IA64_R32)
	      i_copied -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	    if(copy >=IA64_R32)
	      i_copy -= FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge)));
	    
	    if(copied < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copied >= IA64_R32)  
	      i_copied = IA64_R0;
	      //FATAL(("this is strange"));
	    if(copy < IA64_R32 + FUN_SOL(BBL_FUN(CFG_EDGE_TAIL(i_edge))) && copy >= IA64_R32)  
	      i_copy = IA64_R0;
	      //FATAL(("this is strange"));
	   
	  //  FATAL(("ik ben hier"));*/
	  }
	  else FATAL(("shouldn't get here. see ia64reversecopyedge"));
	}
	EdgeMark(i_edge);
//	CfgEdgePrint(i_edge, 0);
//	printf("%d, %d, %d, %d, %d\n",copy,i_copy,copied,i_copied,FUN_SOL(BBL_FUN(CFG_EDGE_HEAD(i_edge))));
	if (!Ia64CopiedIsReplacableBackward(i_copy, i_copied, CFG_EDGE_HEAD(i_edge)))
	  gevonden=FALSE;
      }
      if(gevonden)
      {
	if(1/*reken++==global_options.debugcounter*/)
	{
 	  Ia64Cleanup(copy, copied, i_ins);
	  //FATAL(("stop"));
	}
	
      }
  }

  return;
}

/* vim: set shiftwidth=2: */
