/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_deflowgraph.h"
#include "ia64_instructions.h"
#include "ia64_opcodes.h"

#include "../../utils/util_error.h"
#include "../../utils/util_malloc.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_stack_machine.h"
#include <string.h>

/*staat in ia64layout.c gemplementeerd*/
t_uint32 Ia64CfgLayout(t_object * obj, t_section *code, t_cfg * cfg);

#define IA64INS_IS_MOV_FROM_IP(ins) (IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_MOV_IP  )

#if 0
void Ia64UpdateRelocations(t_object * obj, t_section * code)
{
#define WARNING_RELOCTYPE(reloc, a) \
  printf("StartOffset: 0x%llx; Addend: 0x%llx\n", reloc->from->address, reloc->addend);  \
    RelocPrint(reloc); \
    printf("WARNING: implement reloc-type %s; offset = 0x%llx !\n\n", a, G_T_UINT64(reloc->offset));
#define WARNING_RELOCTYPE2(reloc, a) \
  printf("WARNING: skipped reloc-type %s...\n", a);


  t_bbl *bbl;
  t_reloc * reloc;
  t_reloc_ref * ref;
  t_function * function;

  {
    /* first apply a patch for the opd entry that may exist for this bbl */
    t_section * opd=SectionGetFromObjectByName(obj,".opd");


    if (opd)
    {
      t_uint64 *data = (t_uint64*)opd->data;
      t_uint32 opd_size = GENERIC_EXTRACT_UINT32(SECTION_CSIZE(opd));
      t_uint32 currsize =0;
   //   t_uint64 max_address = G_T_UINT64(GENERIC_ADD(SECTION_GET_CADDRESS(code), SECTION_SIZE(code)));

      while(currsize < opd_size)
      {
	if(*data != 0 /*&& *data < max_address*/)
	{
	  t_address addr;
	  GenericSet_uint64(addr,*data);
	  FOREACH_FUNCTION(code, function)
	  {
	    bbl = function->bbl_first;
	    if(GENERIC_EQ(BBL_OLD_ADDRESS(bbl), addr))
	    {
	      *data = G_T_UINT64(BBL_CADDRESS(bbl));
#ifdef IA64_DEBUG_RELOCS
	      printf("Found an opd-entry for this bbl: changed 0x%llx => 0x%llx\n", BBL_OLD_ADDRESS(bbl), BBL_CADDRESS(bbl));
#endif	      
	      break;
	    }

	  }
	}

	currsize +=16;
	data +=2;
      }

    }
    else
      FATAL(("Section .opd NOT found !!!"));  
  }


  CFG_FOREACH_BBL(SECTION_CFG(code), bbl)
  {
    for(ref=RELOCATABLE_REFED_BY(bbl); ref != NULL; ref=ref->next)
    {
      t_reloc *reloc = ref->rel;

      FATAL(("Reimplement"));
#if 0
      switch(reloc->write_type)
      {
	case ABS64: 
	  /* associated implemented relocs: 
	  *  - DIR64LSB -> REPRESENT_VALUE
	  *  - GPREL64LSB -> REPRESENT_GPREL_VALUE
	  *  - FPTR64LSB -> REPRESENT_OPD_POINTER
	  *  - PCREL64LSB -> REPRESENT_SELFREL_VALUE
	  *  - SEGREL64LSB -> REPRESENT_SEGREL_VALUE
	   */
	  switch(reloc->type)
	  {
	    case REPRESENT_VALUE:
	      {
		t_section *sec = T_SECTION(reloc->from)->parent_section;
		t_address addr = GENERIC_ADD(GENERIC_SUB(reloc->from->address, SECTION_GET_CADDRESS(sec)), reloc->offset);
		t_uint64 *value_addr = (t_uint64*)((char*)sec->data + G_T_UINT64(addr));
		t_bbl * target_bbl=T_BBL(reloc->to->base);
#if 0
		t_address target_addr = BBL_OLD_ADDRESS(T_BBL(reloc->to->base));
		//t_address target_addr = GENERIC_ADD(BBL_OLD_ADDRESS(T_BBL(reloc->to->base)), reloc->addend);
		/* now search for the Bbl - that's the only way I know how to do it pretty fast */
		for(target_bbl = T_BBL(reloc->to->base); !GENERIC_EQ(BBL_OLD_ADDRESS(target_bbl), target_addr); target_bbl=BBL_NEXT(target_bbl));
#endif		  
//		printf("old data: 0x%llx\n", *value_addr);
		*value_addr = BBL_CADDRESS(target_bbl);
//		printf("new data: 0x%llx\n", *value_addr);
//		printf("Succesfully adjusted DIR64LSB relocation\n");
		//		  WARNING_RELOCTYPE(reloc, "DIR64LSB");
		break;
	      }
	    case REPRESENT_GPREL_VALUE:
#ifdef IA64_DEBUG_RELOCS
	      WARNING_RELOCTYPE(reloc, "REPRESENT_GPREL_VALUE");
#endif	      
	      break;
	    case REPRESENT_OPD_POINTER:
	      {
		t_section *opd = SectionGetFromObjectByName(obj,".opd"); 
		t_section *sec = T_SECTION(reloc->from)->parent_section;
		t_address addr = GENERIC_ADD(GENERIC_SUB(reloc->from->address, SECTION_GET_CADDRESS(sec)), reloc->offset);
		t_address fptr_addr;
		if(reloc->addend != 0)
		  FATAL(("Nonzero addend with FPTR64LSB: don't know what to do with it !"));
		GenericSet_uint64(fptr_addr, *(t_uint64*)((char*)sec->data + G_T_UINT64(addr)));
		if(fptr_addr)
		{
		  t_uint64 *opd_addr = 
		    (t_uint64*)((char*)opd->data + G_T_UINT64(GENERIC_SUB(fptr_addr, SECTION_GET_CADDRESS(opd))));
		  if(*opd_addr != BBL_OLD_ADDRESS(T_BBL(reloc->to->base)) && *opd_addr != BBL_CADDRESS(T_BBL(reloc->to->base)))
		  {
		    printf("old BBL-address <in opd>: 0x%llx\n", *opd_addr);
		    printf("old BBL-address <in bbl>: 0x%llx\n", BBL_OLD_ADDRESS(T_BBL(reloc->to->base)));
		    FATAL(("Error in reloc FPTR64LSB"));
		  }
//		  printf("old BBL-address <in opd>: 0x%llx\n", *opd_addr);
		  *opd_addr = BBL_CADDRESS(T_BBL(reloc->to->base));
//		  printf("old BBL-address <in bbl>: 0x%llx\n", BBL_OLD_ADDRESS(T_BBL(reloc->to->base)));
//		  printf("Succesfully adjusted FPTR64LSB relocation\n");
		  //		    WARNING_RELOCTYPE(reloc, "FPTR64LSB");
		}
		else
		  printf("Warning: address to fptr_entry was null !\n");
		break;
	      }
	    case REPRESENT_SELFREL_VALUE:
	      {
		t_section *sec = T_SECTION(reloc->from)->parent_section;
		t_address addr = GENERIC_ADD(GENERIC_SUB(reloc->from->address, SECTION_GET_CADDRESS(sec)), reloc->offset);
		t_bbl * switch_bbl = bbl;
		t_uint64 * disp_addr = (t_uint64*)((char*)sec->data + G_T_UINT64(addr));
#if 0
		t_address switch_addr = BBL_OLD_ADDRESS(T_BBL(reloc->to->base));
		//t_address switch_addr = GENERIC_ADD(BBL_OLD_ADDRESS(T_BBL(reloc->to->base)), reloc->addend);
		/* now search for the Bbl - that's the only way I know how to do it pretty fast */
		for(switch_bbl = bbl; !GENERIC_EQ(BBL_OLD_ADDRESS(switch_bbl), switch_addr); switch_bbl=BBL_NEXT(switch_bbl));
#endif		  
//		printf("old displacement: 0x%llx\n", *disp_addr);
		*disp_addr += G_T_UINT64(GENERIC_SUB(BBL_CADDRESS(switch_bbl), BBL_OLD_ADDRESS(switch_bbl)));
//		printf("new displacement: 0x%llx\n", *disp_addr);
//		printf("Succesfully adjusted PCREL64LSB relocation\n");
		//		  WARNING_RELOCTYPE(reloc, "PCREL64LSB");
		break;
	      }
	    case REPRESENT_SEGREL_VALUE:
#ifdef IA64_DEBUG_RELOCS
	      WARNING_RELOCTYPE(reloc, "REPRESENT_SEGREL_VALUE");
#endif	      
	      break;
	    default:
	      RelocPrint(reloc);
	      FATAL(("Invalid or not yet implemented reloc-type: %d !\n", reloc->type));
	      break;
	  }
	  break;
	case INSTRUCTION_IMMEDIATE64: 
	  /* associated implemented relocs:
	  *  - GPREL64I -> REPRESENT_GPREL_VALUE
	  *  - PCREL64I -> REPRESENT_SELFREL_VALUE
	   */
	  switch(reloc->type)
	  {
	    case REPRESENT_GPREL_VALUE:
	      {
		t_bool isNeg = FALSE;
		t_int64 cnt, nBundlesFromStart1=0, nBundlesFromStart2=0, diff=0;
		t_ins *tmp_ins, *ins = (t_ins*)reloc->from;
		t_address target_addr=
		  GENERIC_ADD_UINT32(BBL_OLD_ADDRESS(T_BBL(reloc->to->base)), 16); /*+16 aangezien de mov_from_ip normaal in volgende bundel pas zit => wordt hierna gecontroleerd bij offset*/

		t_bbl * ip_bbl = INS_BBL(ins);
		t_ins *ip_ins = ip_bbl->ins_first;

		if(((t_uint64)(-IA64INSIMM1(ins)))>>60 ==0 )
		  break;
		
		while(ip_ins && !IA64INS_IS_MOV_FROM_IP(ip_ins))
		{
		  ip_ins = INS_NEXT(ip_ins);
#ifdef IA64_DEBUG_RELOCS
		  if(ip_ins)
		    InsPrint(ip_ins, 0);
#endif		  
		}

		if(!ip_ins && (ip_bbl = BBL_NEXT(ip_bbl)))
		{
		  ip_ins = ip_bbl->ins_first;
		  while(ip_ins && !IA64INS_IS_MOV_FROM_IP(ip_ins))
		  {
		    ip_ins = INS_NEXT(ip_ins);
#ifdef IA64_DEBUG_RELOCS
		    if(ip_ins)
		      InsPrint(ip_ins, 0);
#endif		    
		  }

		}

		/*no ip !*/
		if(!ip_ins)
		{
		  ip_bbl = INS_BBL(ins);
		  ip_ins = ins;
		}

		
		if(reloc->from_type != RELOC_FROM_INSTRUCTION)
		  FATAL(("relocation type GPREL64I expected an instruction as relocation !"));
#ifdef IA64_DEBUG_RELOCS
		printf("GPREL64I: old displacement: 0x%llx\n", IA64INSIMM1(ins));
#endif
		isNeg = IA64INSIMM1(ins) < 0;
		
		IA64INSIMM1(ins)+=G_T_UINT64(GENERIC_SUB(BBL_CADDRESS(INS_BBL(ip_ins)), BBL_OLD_ADDRESS(INS_BBL(ip_ins))));

		nBundlesFromStart1 = G_T_UINT64(GENERIC_SUB(INS_CADDRESS(ip_ins), BBL_OLD_ADDRESS(INS_BBL(ip_ins))))>>4;


		
		/*kijken hoeveel bundels de instructie tov begin van BBl verschoven is...*/
//		cnt = (G_T_UINT64(GENERIC_SUB(INS_CADDRESS(ip_ins), BBL_OLD_ADDRESS(INS_BBL(ip_ins))))>>4)*3 + 1;
#ifdef IA64_DEBUG_RELOCS
		printf("cnt = %lld\n", cnt);
#endif		
		tmp_ins = ip_bbl->ins_first;
		while(tmp_ins != ip_ins && tmp_ins)
		{
		  nBundlesFromStart2++;
#ifdef IA64_DEBUG_RELOCS
	          InsPrint(tmp_ins, 0);
#endif
		  tmp_ins = INS_NEXT(tmp_ins);
//		  cnt--; /*zal negatief zijn als instructie verder in bbl staat...*/
		}
		if(!tmp_ins)
		  FATAL(("Error in relocs: didn't find instruction that should be there !!!"));
		
		nBundlesFromStart2/=3;
#if 0		
		printf("cnt = %lld\n", cnt);
		cnt/=3;
		printf("cnt = %lld\n", cnt);
		
		if(isNeg) /*IA64INSIMM1(ins) < 0*/
  		  cnt = -cnt;
		
		if(cnt != 0)
		  printf("position of instruction in BBl has changed with %lld bundles !\n", cnt);
#endif
		diff = nBundlesFromStart2 - nBundlesFromStart1;

		IA64INSIMM1(ins)+=diff*16;
		
		//IA64INSIMM1(ins)+=cnt*16; /*cnt neg als verder in bbl, maw dichter bij doel...=>offset verminderen*/

#ifdef IA64_DEBUG_RELOCS
		printf("\n");
#endif
		/*symbool kan ook verplaatst zijn !!!*/

		break;
	      }

	      
	      //WARNING_RELOCTYPE(reloc, "REPRESENT_GPREL_VALUE");
	      
	      break;
	    case REPRESENT_SELFREL_VALUE:
	      {
		t_bool isNeg = FALSE;
		t_int64 cnt, nBundlesFromStart1=0, nBundlesFromStart2=0, diff=0;
		t_ins *tmp_ins, *ins = (t_ins*)reloc->from;
		t_bbl * target_bbl = T_BBL(reloc->to->base);
		t_address target_addr=
		  GENERIC_ADD_UINT32(BBL_OLD_ADDRESS(T_BBL(reloc->to->base)), 16); /*+16 aangezien de mov_from_ip normaal in volgende bundel pas zit => wordt hierna gecontroleerd bij offset*/

		t_bbl * ip_bbl = INS_BBL(ins);
		t_ins *ip_ins = ip_bbl->ins_first;

		while(!GENERIC_EQ(target_addr, BBL_OLD_ADDRESS(target_bbl)))
		  target_bbl = BBL_NEXT(target_bbl);

		while(ip_ins && !IA64INS_IS_MOV_FROM_IP(ip_ins))
		{
		  ip_ins = INS_NEXT(ip_ins);
#ifdef IA64_DEBUG_RELOCS
		  if(ip_ins)
		    InsPrint(ip_ins, 0);
#endif		  
		}

		if(!ip_ins && (ip_bbl = BBL_NEXT(ip_bbl)))
		{
		  ip_ins = ip_bbl->ins_first;
		  while(ip_ins && !IA64INS_IS_MOV_FROM_IP(ip_ins))
		  {
		    ip_ins = INS_NEXT(ip_ins);
#ifdef IA64_DEBUG_RELOCS
		    if(ip_ins)
		      InsPrint(ip_ins, 0);
#endif		    
		  }

		}

		/*no ip !*/
		if(!ip_ins)
		{
		  ip_bbl = INS_BBL(ins);
		  ip_ins = ins;
		}


		if(reloc->offset != 2) /* als deze fout zou optreden, betekent dit volgens mij dat de mov_from_ip niet in de bundel direct na de movl zit => dan nog opvangen !!!*/
		  FATAL(("relocation type PCREL64I expected value 2 as offset !"));
		if(reloc->from_type != RELOC_FROM_INSTRUCTION)
		  FATAL(("relocation type PCREL64I expected an instruction as relocation !"));

#ifdef IA64_DEBUG_RELOCS
		printf("PCREL64I: old displacement: 0x%llx\n", IA64INSIMM1(ins));
#endif		
	        //isNeg = IA64INSIMM1(ins)<0;
		
		IA64INSIMM1(ins)-=G_T_UINT64(GENERIC_SUB(BBL_CADDRESS(INS_BBL(ip_ins)), BBL_OLD_ADDRESS(INS_BBL(ip_ins))));

		nBundlesFromStart1 = G_T_UINT64(GENERIC_SUB(INS_CADDRESS(ip_ins), BBL_OLD_ADDRESS(INS_BBL(ip_ins))))>>4;
#if 0		
		/*kijken hoeveel bundels de instructie tov begin van BBl verschoven is...*/
		cnt = (G_T_UINT64(GENERIC_SUB(INS_CADDRESS(ip_ins), BBL_OLD_ADDRESS(INS_BBL(ip_ins))))>>4)*3 + 1;
#endif
//		printf("cnt = %lld\n", cnt);
		tmp_ins = ip_bbl->ins_first;
		while(tmp_ins != ip_ins && tmp_ins)
		{
		  nBundlesFromStart2++;
#ifdef IA64_DEBUG_RELOCS
		  InsPrint(tmp_ins, 0);
#endif
		  tmp_ins = INS_NEXT(tmp_ins);
//		  cnt--; /*zal negatief zijn als instructie verder in bbl staat...*/
		}
		if(!tmp_ins)
		  FATAL(("Error in relocs: didn't find instruction that should be there !!!"));
//		printf("cnt = %lld\n", cnt);
		
		nBundlesFromStart2/=3;
//		cnt/=3;

		diff = nBundlesFromStart2 - nBundlesFromStart1;
		
//		printf("cnt = %lld\n", cnt);
#if 0		
          	if(isNeg) /* IA64INSIMM1(ins)<0 */
   		  cnt = -cnt;

		
		if(cnt != 0)
		  printf("position of instruction in BBl has changed with %lld bundles !\n", cnt);
#endif		
		IA64INSIMM1(ins)-=diff*16; 
		
		//IA64INSIMM1(ins)+=cnt*16; /*cnt neg als verder in bbl, maw dichter bij doel...=>offset verminderen*/

		/*symbool kan ook verplaatst zijn !!!*/
		/*BBL van de juiste offset nodig !!!*/
		IA64INSIMM1(ins)+=G_T_UINT64(GENERIC_SUB(BBL_CADDRESS(target_bbl), BBL_OLD_ADDRESS(target_bbl)));

#ifdef IA64_DEBUG_RELOCS
		printf("\n");
#endif		
		break;
	      }
	    default:
	      RelocPrint(reloc);
	      FATAL(("Invalid or not yet implemented reloc-type: %d !\n", reloc->type));
	      break;
	  }
	  break;
	case INSTRUCTION_IMMEDIATE22: 
	  /* associated implemented relocs:
	  *  - special gp-optimizations relocs: GOTREL and RODATAREL 
	  * 
	  *  - GPREL22 -> REPRESENT_GPREL_VALUE
	  *  - LTOFF22 -> REPRESENT_LT_POINTER
	  *  - LTOFF22_FPTR22 -> REPRESENT_LT_OPD_POINTER
	   */ 
	  switch(reloc->type)
	  {
	    case REPRESENT_GOTREL_VALUE:
#ifdef IA64_DEBUG_RELOCS
	      WARNING_RELOCTYPE(reloc, "REPRESENT_GOTREL_VALUE");
#endif	      
	      break;
	      
	    case REPRESENT_RODATAREL_VALUE:
#ifdef IA64_DEBUG_RELOCS
	      WARNING_RELOCTYPE(reloc, "REPRESENT_RODATAREL_VALUE");
#endif	      
	      break;
      
	    
	    case REPRESENT_GPREL_VALUE:
#ifdef IA64_DEBUG_RELOCS
	      WARNING_RELOCTYPE(reloc, "REPRESENT_GPREL_VALUE");
#endif	      
	      break;
	    case REPRESENT_LT_POINTER:
	      {
		/*
		   t_section *sec = T_SECTION(reloc->from)->parent_section;
		   t_address addr = GENERIC_ADD(GENERIC_SUB(reloc->from->address, SECTION_GET_CADDRESS(sec)), reloc->offset);
		   t_uint64 *value_addr = (t_uint64*)((char*)sec->data + G_T_UINT64(addr));
		   t_bbl * target_bbl;
		   t_address target_addr = GENERIC_ADD(BBL_OLD_ADDRESS(T_BBL(reloc->to->base)), reloc->addend);
		 */	

		t_ins * ins = T_INS(reloc->from);
		t_bbl * target_bbl = T_BBL(reloc->to->base);
		t_uint64 base_addr = G_T_UINT64(BBL_OLD_ADDRESS(ins->bbl->fun->bbl_first));
		t_uint64 new_addr = G_T_UINT64(BBL_CADDRESS(ins->bbl->fun->bbl_first));

		/*now search for base addr in opd -> we need the associated gp-value !!! */
		t_section * opd=SectionGetFromObjectByName(obj,".opd");
		t_section * got=SectionGetFromObjectByName(obj,".got");

#ifdef IA64_DEBUG_RELOCS
		WARNING_RELOCTYPE(reloc, "REPRESENT_LT_POINTER");
#endif
		if (opd)
		{
		  t_uint64 *data = (t_uint64*)opd->data;
		  t_uint32 opd_size = GENERIC_EXTRACT_UINT32(SECTION_SIZE(opd));
		  t_uint32 got_size = GENERIC_EXTRACT_UINT32(SECTION_SIZE(got));
		  t_uint32 currsize =0;
		  t_uint64 max_address = G_T_UINT64(GENERIC_ADD(SECTION_GET_CADDRESS(code), SECTION_SIZE(code)));
		  t_uint64 gp;
		  t_uint64 * gp_value;

#ifdef IA64_DEBUG_RELOCS
		  printf("Func: old: 0x%llx; new: 0x%llx\n", base_addr, new_addr);
#endif
		  while(currsize < opd_size)
		  {
		    if(/*data == base_addr ||*/ *data == new_addr) /*addr should already be adapted*/
		      break;

		    currsize +=16;
		    data +=2;
		  }

		  /*do not look after opd...*/
		  //currsize=opd_size;

		  if(currsize>=opd_size)
		  {
		    /*see Conventions; p.13-1 13.1.2 */
		    gp = G_T_UINT64(SECTION_GET_CADDRESS(got));
		    //FATAL(("Function address @G not found in .opd (needed for gp).", base_addr));
		  }
		  else
		    gp = *(data+1);

		  gp += IA64INSIMM1(ins);
#ifdef IA64_DEBUG_RELOCS
	          printf("gp-value: 0x%llx\n", gp);
#endif		  
		  gp -= G_T_UINT64(SECTION_GET_CADDRESS(got));
#ifdef IA64_DEBUG_RELOCS
		  printf("gp-disp: 0x%llx\n", gp);
#endif
		  if(gp >= got_size)
		    FATAL(("Problem: gp > got_size"));

		  gp_value = (t_uint64*)((char*)got->data + gp);
#ifdef IA64_DEBUG_RELOCS
		  printf("old gp-content: 0x%llx; old bbl_addr: 0x%llx\n", *gp_value,  G_T_UINT64(BBL_OLD_ADDRESS(target_bbl)));
#endif		  
		  if(*gp_value != G_T_UINT64(BBL_OLD_ADDRESS(target_bbl)))
		  {
#ifdef IA64_DEBUG_RELOCS
		    printf("WARNING: Something NOT right -> gp NOT changed!!!");
#endif		    
		  }
		  else
		  *gp_value = G_T_UINT64(BBL_CADDRESS(target_bbl));

		}
		else
		  FATAL(("Section .opd NOT found !!!"));   


		break;
	      }
	    case REPRESENT_LT_OPD_POINTER:
#ifdef IA64_DEBUG_RELOCS
	      WARNING_RELOCTYPE2(reloc, "REPRESENT_LT_OPD_POINTER");
#endif	      
	      break;

	    default:
	      RelocPrint(reloc);
	      FATAL(("Invalid or not yet implemented reloc-type: %d !\n", reloc->type));
	      break;
	  }
	  break;
	case INSTRUCTION_IMMEDIATE21B:
	  /* associated implemented relocs:
	  *  - PCREL21B -> REPRESENT_SELFREL_VALUE
	   */

	  /*those should already be dealt with with UpdateDisplacement... - 
	  * unless there is an instruction with a imm21b field that isn't a jump !!!*/
	  break;
	default:
	  RelocPrint(reloc);
	  FATAL(("Invalid or not yet implemented reloc-writetype: %d !\n", reloc->write_type));
	  break;
      }
#endif
    }



  }


  {
    t_object *tmp, *tmp2;
    t_symbol *symptr;

    /*seeking for special $diablo-symbols*/
    for (tmp2=obj->mapped_first;tmp2!=NULL; tmp2=tmp2->next)
    {
     FATAL(("Reimplement"));
#if 0
      for (tmp=tmp2; tmp!=NULL; tmp=tmp->equal)
      {     
	t_address generic;
	t_uint32 tel;

	for (symptr=tmp->symbol_table->first; symptr!=NULL; symptr=symptr->next)
	{
	  t_bbl * last_diablo_bbl = NULL;

	  if(!symptr->name)
	    continue;


	  /*PATCH: skip these*/
	  if (strncmp(symptr->name,"$diablo:pow",11)==0)
	    continue;

	  if (strncmp(symptr->name,"$diablo:",8)==0)
	  {
	    int begin = strstr(symptr->name,"@-@")-symptr->name;
	    char save = symptr->name[begin];
	    t_symbol *s1, *s2;
	    t_address ad, a1, a2, texta1, texta2;
	    t_bbl *bbl1, *bbl2, *bbldata;
	    t_ins *data_ins;


	    ad = GENERIC_ADD(symptr->base->address, symptr->current_address);

#ifdef IA64_DEBUG_RELOCS
	    printf("symbol: %s @ 0x%llx\n", symptr->name, ad);
#endif
	    symptr->name[begin] = '\0';


	    /*TODO ook checken als het symbool van zelfde .o-file komt*/
	    s1=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+8);
	    s2=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+begin+3);

	    a1 = GENERIC_ADD(symptr->base->address, s1->current_address);
	    a2 = GENERIC_ADD(symptr->base->address, s2->current_address);

	    //texta1 = GENERIC_SUB();

	    /*char herstellen*/
	    symptr->name[begin] = save;

	    /*controleren of wel bestaat !*/
	    if(!s1|| !s2)
	      FATAL(("Invalid special diablo symbol !!!"));

	    if(!last_diablo_bbl || !GENERIC_EQ(BBL_OLD_ADDRESS(last_diablo_bbl), symptr->base->address))
	      last_diablo_bbl = T_BBL(GRAPH_NODE_FIRST(SECTION_CFG(code)));

	    while(!GENERIC_EQ(BBL_OLD_ADDRESS(last_diablo_bbl), symptr->base->address))
	      last_diablo_bbl = BBL_NEXT(last_diablo_bbl);

	    for(bbl1 = last_diablo_bbl; !GENERIC_EQ(BBL_OLD_ADDRESS(bbl1), a1); bbl1 = BBL_NEXT(bbl1));
	    for(bbl2 = last_diablo_bbl; !GENERIC_EQ(BBL_OLD_ADDRESS(bbl2), a2); bbl2 = BBL_NEXT(bbl2));

#ifdef IA64_DEBUG_RELOCS
	    printf(">>>> old bbl: 0x%llx; new bbl: 0x%llx\n", BBL_OLD_ADDRESS(bbl1), BBL_CADDRESS(bbl1));
	    printf(">>>> old bbl: 0x%llx; new bbl: 0x%llx\n", BBL_OLD_ADDRESS(bbl2), BBL_CADDRESS(bbl2));
#endif
            for(bbldata = last_diablo_bbl; GENERIC_LT(BBL_OLD_ADDRESS(BBL_NEXT(bbldata)), ad); bbldata = BBL_NEXT(bbldata));

	    
#if 0
	    for(bbldata = last_diablo_bbl; 
		!(GENERIC_LE(BBL_OLD_ADDRESS(bbldata), ad)) || GENERIC_LT(BBL_OLD_ADDRESS(BBL_NEXT(bbldata)), ad)/*GENERIC_GT(GENERIC_ADD_UINT32(BBL_OLD_ADDRESS(bbldata), (bbldata->nins/3)<<4), ad)*/;
		bbldata = BBL_NEXT(bbldata));
#endif
#ifdef IA64_DEBUG_RELOCS
	    printf(">>>> old bbl: 0x%llx; new bbl: 0x%llx; size = %d\n", BBL_OLD_ADDRESS(bbldata), BBL_CADDRESS(bbldata), bbldata->nins);
#endif
	    if(G_T_UINT64(ad) % 16 == 8)
	      GenericSet_uint64(ad, (G_T_UINT64(ad) & ~0x0FULL) + 1);
	    
	    /*nu nog juiste instr vinden...*/
	    for(data_ins = bbldata->ins_first; !GENERIC_EQ(INS_CADDRESS(data_ins), ad); data_ins = INS_NEXT(data_ins))
	    {
#ifdef IA64_DEBUG_RELOCS
	      printf("searching data @ %llx\n", INS_CADDRESS(data_ins)); fflush(stdout);
#endif	      
	    }

	    IA64INSIMM1(data_ins) = G_T_UINT64(GENERIC_SUB(BBL_CADDRESS(bbl1), BBL_CADDRESS(bbl2)));

	  }
	}

      }
#endif
    }

  }

#undef WARNING_RELOCTYPE
}

#endif



void Ia64UpdateRelocations(t_object * obj, t_section * code, t_cfg *cfg)
{
  t_reloc * reloc, *rel;
  t_bbl * i_bbl;
  t_ins * i_ins;
  t_cfg_edge * i_edge;


  OBJECT_FOREACH_RELOC(obj,rel)
  {
    if (RELOCATABLE_TYPE(REL_FROM(rel)) == RT_INS)
    {
    
      if(RELOCATABLE_REFERS_TO(REL_FROM(rel)))
      {
	t_address ret;
	t_reloc * rel2 = RELOCATABLE_REFERS_TO(REL_FROM(rel))->rel;

	/*special patch for the gp-optim-relocations*/
	if(strcmp(rel2->code, "s0000$") != 0)
	{
	  
	  ret = StackExec(rel2->code,rel2,NULL,FALSE,0);

//	DiabloPrint(stdout, "ins:@I old:0x%llx new:@G from_offset: @G\n", REL_FROM(rel), IA64INSIMM1(REL_FROM(rel)), ret, rel->from_offset);

	  IA64INSIMM1(REL_FROM(rel2)) = GENERIC_EXTRACT_UINT64(ret);
	}
      }
      else
	FATAL(("no refers to !!!"));
      
    }
    
    else
    {
      //DiabloPrint(stdout,"relocing in section %s at @G old 0x%x new ",T_SECTION(REL_FROM(rel))->name,GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset), *(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset));
      t_address ret = StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,0);
      //DiabloPrint(stdout,"%x\n",*(t_uint32 *)(T_SECTION(REL_FROM(rel))->data + rel->from_offset));
      VERIFY_FASSERT(GENERICISNULL(ret),("Could not properly execute @R",rel));
    }
    
  }

//#if 0
  {
    t_object *tmp, *tmp2;
    t_symbol *symptr;

    /*seeking for special $diablo-symbols*/
    for (tmp2=obj->mapped_first;tmp2!=NULL; tmp2=tmp2->next)
    {
      for (tmp=tmp2; tmp!=NULL; tmp=tmp->equal)
      {     
	t_address generic;
	t_uint32 tel;

	for (symptr=tmp->symbol_table->first; symptr!=NULL; symptr=symptr->next)
	{
	  t_bbl * last_diablo_bbl = NULL;

	  if(!symptr->name)
	    continue;


	  /*PATCH: skip these*/
	  if (strncmp(symptr->name,"$diablo:pow",11)==0)
	    continue;

	  if (strncmp(symptr->name,"$diablo:",8)==0)
	  {
	    int begin = strstr(symptr->name,"@-@")-symptr->name;
	    char save = symptr->name[begin];
	    t_symbol *s1, *s2;
	    t_address ad, a1, a2, texta1, texta2;
	    t_bbl *bbl1, *bbl2, *bbldata;
	    t_ins *data_ins;


	    ad = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), symptr->current_address);
//#define IA64_DEBUG_RELOCS
#ifdef IA64_DEBUG_RELOCS
	    printf("symbol: %s @ 0x%llx\n", symptr->name, ad);
#endif
	    symptr->name[begin] = '\0';


	    /*TODO ook checken als het symbool van zelfde .o-file komt*/
	    s1=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+8);
	    s2=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+begin+3);

	    a1 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s1->current_address);
	    a2 = GENERIC_ADD(SECTION_GET_CADDRESS(symptr->base), s2->current_address);

	    //texta1 = GENERIC_SUB();

	    /*char herstellen*/
	    symptr->name[begin] = save;

	    /*controleren of wel bestaat !*/
	    if(!s1|| !s2)
	      FATAL(("Invalid special diablo symbol !!!"));

//	    if(!last_diablo_bbl || !GENERIC_EQ(BBL_OLD_ADDRESS(last_diablo_bbl), SECTION_GET_CADDRESS(symptr->base)))
	      last_diablo_bbl = T_BBL(GRAPH_NODE_FIRST(SECTION_CFG(code)));

	    while(!GENERIC_EQ(BBL_OLD_ADDRESS(last_diablo_bbl), SECTION_GET_CADDRESS(symptr->base)))
	      last_diablo_bbl = BBL_NEXT(last_diablo_bbl);

	    for(bbl1 = last_diablo_bbl; bbl1!= NULL && !GENERIC_EQ(BBL_OLD_ADDRESS(bbl1), a1); bbl1 = BBL_NEXT(bbl1));
	    for(bbl2 = last_diablo_bbl; bbl2!= NULL && !GENERIC_EQ(BBL_OLD_ADDRESS(bbl2), a2); bbl2 = BBL_NEXT(bbl2));

	    if(bbl1 == NULL || bbl2 == NULL)
	      continue;

#ifdef IA64_DEBUG_RELOCS
	    printf(">>>> old bbl: 0x%llx; new bbl: 0x%llx\n", BBL_OLD_ADDRESS(bbl1), BBL_CADDRESS(bbl1));
	    printf(">>>> old bbl: 0x%llx; new bbl: 0x%llx\n", BBL_OLD_ADDRESS(bbl2), BBL_CADDRESS(bbl2));
#endif
            for(bbldata = last_diablo_bbl; GENERIC_LE(BBL_OLD_ADDRESS(BBL_NEXT(bbldata)), ad); bbldata = BBL_NEXT(bbldata));

	    
#if 0
	    for(bbldata = last_diablo_bbl; 
		!(GENERIC_LE(BBL_OLD_ADDRESS(bbldata), ad)) || GENERIC_LT(BBL_OLD_ADDRESS(BBL_NEXT(bbldata)), ad)/*GENERIC_GT(GENERIC_ADD_UINT32(BBL_OLD_ADDRESS(bbldata), (bbldata->nins/3)<<4), ad)*/;
		bbldata = BBL_NEXT(bbldata));
#endif
#ifdef IA64_DEBUG_RELOCS
	    printf(">>>> old bbl: 0x%llx; new bbl: 0x%llx; size = %d\n", BBL_OLD_ADDRESS(bbldata), BBL_CADDRESS(bbldata), bbldata->nins);
#endif
	    if(G_T_UINT64(ad) % 16 == 8)
	      GenericSet_uint64(ad, (G_T_UINT64(ad) & ~0x0FULL) + 1);
	    
	    /*nu nog juiste instr vinden...*/
	    for(data_ins = bbldata->ins_first; !GENERIC_EQ(INS_OLD_ADDRESS(data_ins), ad); data_ins = INS_NEXT(data_ins))
	    {
#ifdef IA64_DEBUG_RELOCS
	      printf("searching data @ %llx\n", INS_OLD_ADDRESS(data_ins)); fflush(stdout);
#endif	      
	    }

	    IA64INSIMM1(data_ins) = G_T_UINT64(GENERIC_SUB(BBL_CADDRESS(bbl1), BBL_CADDRESS(bbl2)));

	  }
	}

      }
    }

  }
//#endif
  
#if 0  
  /* relocate the addresses stored in non-code sections */
//  ANALYSIS_START("Relocating in non-code sections");
  OBJECT_FOREACH_RELOC(obj,reloc)
  {
    if (!REL_TO_RELOCATABLE(reloc)) continue;
    if (REL_TO_RELOCATABLE(reloc)->type==RT_INS) FATAL(("relocs to ins"));
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_SUBSECTION)
    {
      if ((REL_TO_RELOCATABLE(reloc)->type==RT_BBL) || (REL_TO_RELOCATABLE(reloc)->type==RT_SUBSECTION)|| (REL_TO_RELOCATABLE(reloc)->type==RT_SECTION))
      {
	t_address ret;
	ret=StackExec(reloc->code,reloc,T_SECTION(REL_FROM(reloc))->data,TRUE,0);
      }
      else FATAL(("IMPLEMENT 2 > %d",REL_TO_RELOCATABLE(reloc)->type));
    }
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_BBL) continue;
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_INS) continue;
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_SECTION) FATAL(("reloc from section"));
    else FATAL(("Unknown reloc @R\n",reloc));
  }
//  ANALYSIS_SIMPLE_END("Relocating in non-code sections");
  
  /* relocate the addresses stored in data blocks in the text section */
//  ANALYSIS_START("Relocating in code sections");
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_FOREACH_INS(i_bbl,i_ins)
    {      
      if (RELOCATABLE_REFERS_TO(i_ins))
      {
	t_address ret;
	t_reloc * rel = RELOCATABLE_REFERS_TO(i_ins)->rel;

	ret = StackExec(rel->code,rel,NULL,FALSE,0);

//	DiabloPrint(stdout, "ins:@I old:0x%llx new:@G from_offset: @G\n", REL_FROM(rel), IA64INSIMM1(REL_FROM(rel)), ret, rel->from_offset);

	IA64INSIMM1(REL_FROM(rel)) = GENERIC_EXTRACT_UINT64(ret);
	
      }
    }
  }
//  ANALYSIS_SIMPLE_END("Relocating in code sections");
#endif  
}


void Ia64DeflowgraphSection(t_object * obj, t_section * code)
{
  t_bbl * bbl;
  t_uint32 size=0;
  t_uint32 tel=0;
  t_reloc * reloc;
  t_reloc_ref * ref;
  t_ins * ins;
  /*t_bool data;*/

  size=Ia64CfgLayout(obj, code, T_CFG(code->data));
  

#ifdef IA64_NEW_BBL_ADDRESS
  /*relocation-adaption not needed if everything stays put*/
  Ia64UpdateRelocations(obj, code, T_CFG(code->data));
#endif

  
  SECTION_TMP_BUF(code)=Malloc(sizeof(t_ia64_ins)*((size>>4)*3 /*+ (3*GENERIC_SUB(obj->entry, SECTION_GET_CADDRESS(code))>>4)*/ ));
  memset(SECTION_TMP_BUF(code), 0xFF, sizeof(t_ia64_ins)*((size>>4)*3 )); 
  /*\TODO: nog mooier oplossen*/
 

  SymbolTableFree(obj->symbol_table);
  obj->symbol_table=NULL;

  obj->symbol_table=SymbolTableNew(obj);

  printf("%llx\n", SECTION_GET_CADDRESS(code));
//  printf("\n\n\n!!! WARNING !!!: in Ia64DeflowgraphSection: BBL_CADDRESS = tijdelijk BBL_OLD_ADDRESS\n\n\n\n");
   
  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
  {
      t_uint64 addr;
      t_ia64_ins * current=(t_ia64_ins *) bbl->ins_first;
      
      if(bbl->nins == 0)
	continue;
      
      if (!current) continue;

//      InsPrint(current,0);

//#define BBL_CADDRESS BBL_OLD_ADDRESS

//      printf("%llx\n",G_T_UINT64(BBL_CADDRESS(bbl)));fflush(stdout);

#ifdef IA64_NEW_BBL_ADDRESS
      addr=(G_T_UINT64(GENERIC_SUB(BBL_CADDRESS(bbl),SECTION_GET_CADDRESS(code)))>>4)*3;/*********/
#endif
    
      while(current)
      {
        tel++;
//        InsPrint(current,0);
        //((t_ins *) current)->copy=(t_ins *) &((t_ia64_ins *) SECTION_TMP_BUF(code))[addr];
	
#ifndef IA64_NEW_BBL_ADDRESS	
	if(INS_CADDRESS(current)%16 > 2)
	{
//	  if(IA64INSTYPE(current) == IA64_TYPE_DATA && INS_CADDRESS(current)%16 == 8)
//	    addr=1 + (G_T_UINT64(GENERIC_SUB(INS_CADDRESS(current),SECTION_GET_CADDRESS(code)))>>4)*3;
//	  else
  	    FATAL(("Wrong instruction address"));
	}
	else
	  addr=INS_CADDRESS(current)%16 + (G_T_UINT64(GENERIC_SUB(INS_CADDRESS(current),SECTION_GET_CADDRESS(code)))>>4)*3;
#endif	
	INS_COPY(current)=&((t_ia64_ins *) SECTION_TMP_BUF(code))[addr];
        memcpy(&((t_ia64_ins *) SECTION_TMP_BUF(code))[addr],current,sizeof(t_ia64_ins));
#ifdef IA64_NEW_BBL_ADDRESS
	addr++;
#endif
	current=(t_ia64_ins *) INS_NEXT(current);
      }
    }



  
  printf("Done, wrote %d\n",tel);



  
//  printf("\n\n\n!!! WARNING !!!: in Ia64DeflowgraphSection: BBL_CADDRESS = tijdelijk BBL_OLD_ADDRESS\n\n\n\n");
  
  for(tel = 0; tel<(size>>4)*3; tel++)
  {
    INS_NEXT(&((t_ia64_ins *) SECTION_TMP_BUF(code))[tel])=(t_ins *) (&((t_ia64_ins *) SECTION_TMP_BUF(code))[tel+1]);
    INS_BBL(&((t_ia64_ins *) SECTION_TMP_BUF(code))[tel]) = NULL;
  }



  
  
#if 0

  data=INS_TYPE((&((t_ia64_ins *) SECTION_TMP_BUF(code))[0]))==IT_DATA?FALSE:TRUE;
  for (tel=0; tel<size/4-1; tel++)
    {
      if ((data) && (INS_TYPE((&((t_ia64_ins *) SECTION_TMP_BUF(code))[tel]))!=IT_DATA))
	{
	  t_address tmp;
	  GenericSet_uint32(tmp,tel*4);
	  tmp=GENERIC_ADD(tmp,SECTION_GET_CADDRESS(code));
	  SymbolTableAddSymbolToSection(obj->symbol_table, 0, T_RELOCATABLE(code),tmp,StringDup("$switch"), TRUE);
	  //	DiabloPrint(stdout,"Code at %x, @I\n",tel*4+0x8000, (&((t_ia64_ins *) SECTION_TMP_BUF(code))[tel]));
	  data=!data;
	}
      else if ((!data) && (INS_TYPE((&((t_arm_ins *) SECTION_TMP_BUF(code))[tel]))==IT_DATA))
	{
	  t_address tmp;
	  GenericSet_uint32(tmp,tel*4);
	  tmp=GENERIC_ADD(tmp,SECTION_GET_CADDRESS(code));
	  SymbolTableAddSymbolToSection(obj->symbol_table, 0, T_RELOCATABLE(code),tmp,StringDup("$switch"), TRUE);
	  //	DiabloPrint(stdout,"Data at %x, @I\n",tel*4+0x8000, (&((t_arm_ins *) SECTION_TMP_BUF(code))[tel]));
	  data=!data;
	}
      INS_NEXT(&((t_arm_ins *) SECTION_TMP_BUF(code))[tel])=(t_ins *) (&((t_arm_ins *) SECTION_TMP_BUF(code))[tel+1]);
      INS_BBL(&((t_arm_ins *) SECTION_TMP_BUF(code))[tel]) = NULL;
    }
  /* Sort the $switch symbols */
  {
    t_symbol_he * top=( t_symbol_he *) HashLookup(obj->symbol_table->hash_table, "$switch");
    t_symbol_he * he;
    t_symbol *tmp;
    if (top)
      {
	while(top->node.equal)
	  {
	    for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
	      {
		if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
		  {
		    tmp=top->sym;
		    top->sym=he->sym;
		    he->sym=tmp;
		  }
	      }
	    top=(t_symbol_he *) top->node.equal;
	  }
      }
  }
#endif


  INS_NEXT(&((t_ia64_ins *) SECTION_TMP_BUF(code))[(size>>4)*3 - 1])=NULL;

  /*Adapt relocations */
  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
  {
    t_ins * ins;

    BBL_FOREACH_INS(bbl,ins)
    {
      while (RELOCATABLE_REFED_BY(ins))
	FATAL(("all relocations to instructions should be converted to bbls at this point")); 

      if (RELOCATABLE_REFERS_TO(ins))
      {
	RELOCATABLE_REFERS_TO(INS_COPY(ins))=NULL;
      }
      while (RELOCATABLE_REFERS_TO(ins)){
	t_reloc * relptr = RELOCATABLE_REFERS_TO(ins)->rel;

	RelocSetFrom(relptr,T_RELOCATABLE(INS_COPY(ins)));
	relptr->edge = NULL;
      }
    }

  while(RELOCATABLE_REFED_BY(bbl))
    {
      t_reloc * relptr = RELOCATABLE_REFED_BY(bbl)->rel;

      if(BBL_INS_FIRST(bbl)==NULL){
	FATAL(("reloc to empty BBL"));
      }
      RelocSetToRelocatable(relptr, T_RELOCATABLE(INS_COPY(BBL_INS_FIRST(bbl))));

      relptr->addend = GENERIC_SUB(relptr->addend, GENERIC_SUB(INS_CADDRESS(BBL_INS_FIRST(bbl)),BBL_CADDRESS(bbl)));
      relptr->edge = NULL;
    } 

    while(RELOCATABLE_REFERS_TO(bbl))
      FATAL(("relocation from BBL, this shouldn't happen"));
  }

  /* {{{ compute code size gain */
  {
    double gain, os, ns;
    int i;
    t_address osize, nsize;

    osize = nsize = NULL_GENERIC;
    for (i=0; i < obj->ncodes; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->code[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->code[i]));
    }
    DiabloPrint(stdout,"Code section old size @G new size @G\n",osize,nsize);
    os = (double) G_T_UINT64(osize);
    ns = (double) G_T_UINT64(nsize);
    gain = (os - ns)/os;
    printf("GAIN %lf%% in code section\n",100.0 * gain);
    for (i=0; i < obj->nrodatas; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->rodata[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->rodata[i]));
    }
    for (i=0; i < obj->ndatas; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->data[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->data[i]));
    }
    for (i=0; i < obj->nbsss; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->bss[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->bss[i]));
    }
    os = (double) G_T_UINT64(osize);
    ns = (double) G_T_UINT64(nsize);
    gain = (os - ns)/os;
    printf("GAIN %lf%% total\n",100.0 * gain);
  } /* }}} */
//#if 0  
// TODO !!! tijdelijk afschakelen
 
  CFG_SECTION(SECTION_CFG(code)) = NULL; 	/* prevent CfgFree from freeing code->data, which we still need */
  CfgFree(SECTION_CFG(code));
//#endif
  code->data=SECTION_TMP_BUF(code);
  code->type=DISASSEMBLED_CODE_SECTION;
  
  printf("Counted till %d\n",tel);
}
/* vim: set shiftwidth=2: */
