/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ia64_description.h"
#include "ia64_instructions.h"

static char * names[] =    {

  "R0", "R1", "R2", "R3", "R4", "R5", "R6", "R7",  "R8", "R9", "R10", "R11", "R12", "R13", "R14", "R15",
  "R16", "R17", "R18", "R19", "R20", "R21", "R22", "R23",  "R24", "R25", "R26", "R27", "R28", "R29", "R30", "R31",
  "R32", "R33", "R34", "R35", "R36", "R37", "R38", "R39",  "R40", "R41", "R42", "R43", "R44", "R45", "R46", "R47",
  "R48", "R49", "R50", "R51", "R52", "R53", "R54", "R55",  "R56", "R57", "R58", "R59", "R60", "R61", "R62", "R63",
  "R64", "R65", "R66", "R67", "R68", "R69", "R70", "R71",  "R72", "R73", "R74", "R75", "R76", "R77", "R78", "R79",
  "R80", "R81", "R82", "R83", "R84", "R85", "R86", "R87",  "R88", "R89", "R90", "R91", "R92", "R93", "R94", "R95",
  "R96", "R97", "R98", "R99", "R100", "R101", "R102", "R103",  "R104", "R105", "R106", "R107", "R108", "R109", "R110", "R111",
  "R112", "R113", "R114", "R115", "R116", "R117", "R118", "R119",  "R120", "R121", "R122", "R123", "R124", "R125", "R126", "R127",
  
  /* All global registers - nat bit */
  "R0_NAT", "R1_NAT", "R2_NAT", "R3_NAT", "R4_NAT", "R5_NAT", "R6_NAT", "R7_NAT",   "R8_NAT", "R9_NAT", "R10_NAT", "R11_NAT", "R12_NAT", "R13_NAT", "R14_NAT", "R15_NAT",
  "R16_NAT", "R17_NAT", "R18_NAT", "R19_NAT", "R20_NAT", "R21_NAT", "R22_NAT", "R23_NAT",  "R24_NAT", "R25_NAT", "R26_NAT", "R27_NAT", "R28_NAT", "R29_NAT", "R30_NAT", "R31_NAT",
  "R32_NAT", "R33_NAT", "R34_NAT", "R35_NAT", "R36_NAT", "R37_NAT", "R38_NAT", "R39_NAT",  "R40_NAT", "R41_NAT", "R42_NAT", "R43_NAT", "R44_NAT", "R45_NAT", "R46_NAT", "R47_NAT",
  "R48_NAT", "R49_NAT", "R50_NAT", "R51_NAT", "R52_NAT", "R53_NAT", "R54_NAT", "R55_NAT",  "R56_NAT", "R57_NAT", "R58_NAT", "R59_NAT", "R60_NAT", "R61_NAT", "R62_NAT", "R63_NAT",
  "R64_NAT", "R65_NAT", "R66_NAT", "R67_NAT", "R68_NAT", "R69_NAT", "R70_NAT", "R71_NAT",  "R72_NAT", "R73_NAT", "R74_NAT", "R75_NAT", "R76_NAT", "R77_NAT", "R78_NAT", "R79_NAT",
  "R80_NAT", "R81_NAT", "R82_NAT", "R83_NAT", "R84_NAT", "R85_NAT", "R86_NAT", "R87_NAT",  "R88_NAT", "R89_NAT", "R90_NAT", "R91_NAT", "R92_NAT", "R93_NAT", "R94_NAT", "R95_NAT",
  "R96_NAT", "R97_NAT", "R98_NAT", "R99_NAT", "R100_NAT", "R101_NAT", "R102_NAT", "R103_NAT",  "R104_NAT", "R105_NAT", "R106_NAT", "R107_NAT", "R108_NAT", "R109_NAT", "R110_NAT", "R111_NAT",
  "R112_NAT", "R113_NAT", "R114_NAT", "R115_NAT", "R116_NAT", "R117_NAT", "R118_NAT", "R119_NAT",  "R120_NAT", "R121_NAT", "R122_NAT", "R123_NAT", "R124_NAT", "R125_NAT", "R126_NAT", "R127_NAT",

  "F0", "F1", "F2", "F3", "F4", "F5", "F6", "F7",   "F8", "F9", "F10", "F11", "F12", "F13", "F14", "F15",
  "F16", "F17", "F18", "F19", "F20", "F21", "F22", "F23",  "F24", "F25", "F26", "F27", "F28", "F29", "F30", "F31",
  "F32", "F33", "F34", "F35", "F36", "F37", "F38", "F39",  "F40", "F41", "F42", "F43", "F44", "F45", "F46", "F47",
  "F48", "F49", "F50", "F51", "F52", "F53", "F54", "F55",  "F56", "F57", "F58", "F59", "F60", "F61", "F62", "F63",
  "F64", "F65", "F66", "F67", "F68", "F69", "F70", "F71",  "F72", "F73", "F74", "F75", "F76", "F77", "F78", "F79",
  "F80", "F81", "F82", "F83", "F84", "F85", "F86", "F87",  "F88", "F89", "F90", "F91", "F92", "F93", "F94", "F95",
  "F96", "F97", "F98", "F99", "F100", "F101", "F102", "F103",  "F104", "F105", "F106", "F107", "F108", "F109", "F110", "F111",
  "F112", "F113", "F114", "F115", "F116", "F117", "F118", "F119",  "F120", "F121", "F122", "F123", "F124", "F125", "F126", "F127",

  /* All predicate registers */
  "P0", "P1", "P2", "P3", "P4", "P5", "P6", "P7",  "P8", "P9", "P10", "P11", "P12", "P13", "P14", "P15",
  "P16", "P17", "P18", "P19", "P20", "P21", "P22", "P23",  "P24", "P25", "P26", "P27", "P28", "P29", "P30", "P31",
  "P32", "P33", "P34", "P35", "P36", "P37", "P38", "P39",  "P40", "P41", "P42", "P43", "P44", "P45", "P46", "P47",
  "P48", "P49", "P50", "P51", "P52", "P53", "P54", "P55",  "P56", "P57", "P58", "P59", "P60", "P61", "P62", "P63",

  /* All branch registers */
  "B0", "B1", "B2", "B3", "B4", "B5", "B6", "B7",

  "AR0", "AR1", "AR2", "AR3", "AR4", "AR5", "AR6", "AR7",  "AR8", "AR9", "AR10", "AR11", "AR12", "AR13", "AR14", "AR15",
  "AR16", "AR17", "AR18", "AR19", "AR20", "AR21", "AR22", "AR23",  "AR24", "AR25", "AR26", "AR27", "AR28", "AR29", "AR30", "AR31",
  "AR32", "AR33", "AR34", "AR35", "AR36", "AR37", "AR38", "AR39",  "AR40", "AR41", "AR42", "AR43", "AR44", "AR45", "AR46", "AR47",
  "AR48", "AR49", "AR50", "AR51", "AR52", "AR53", "AR54", "AR55",  "AR56", "AR57", "AR58", "AR59", "AR60", "AR61", "AR62", "AR63",
  "AR64", "AR65", "AR66", "AR67", "AR68", "AR69", "AR70", "AR71",  "AR72", "AR73", "AR74", "AR75", "AR76", "AR77", "AR78", "AR79",
  "AR80", "AR81", "AR82", "AR83", "AR84", "AR85", "AR86", "AR87",  "AR88", "AR89", "AR90", "AR91", "AR92", "AR93", "AR94", "AR95",
  "AR96", "AR97", "AR98", "AR99", "AR100", "AR101", "AR102", "AR103",  "AR104", "AR105", "AR106", "AR107", "AR108", "AR109", "AR110", "AR111",
  "AR112", "AR113", "AR114", "AR115", "AR116", "AR117", "AR118", "AR119",  "AR120", "AR121", "AR122", "AR123", "AR124", "AR125", "AR126", "AR127",

  "IP", "CFM", "UM", "CPUID", "PMD",

  "CR0", "CR1", "CR2", "CR3", "CR4", "CR5", "CR6", "CR7",  "CR8", "CR9", "CR10", "CR11", "CR12", "CR13", "CR14", "CR15",
  "CR16", "CR17", "CR18", "CR19", "CR20", "CR21", "CR22", "CR23",  "CR24", "CR25", "CR26", "CR27", "CR28", "CR29", "CR30", "CR31",
  "CR32", "CR33", "CR34", "CR35", "CR36", "CR37", "CR38", "CR39",  "CR40", "CR41", "CR42", "CR43", "CR44", "CR45", "CR46", "CR47",
  "CR48", "CR49", "CR50", "CR51", "CR52", "CR53", "CR54", "CR55",  "CR56", "CR57", "CR58", "CR59", "CR60", "CR61", "CR62", "CR63",
  "CR64", "CR65", "CR66", "CR67", "CR68", "CR69", "CR70", "CR71",  "CR72", "CR73", "CR74", "CR75", "CR76", "CR77", "CR78", "CR79",
  "CR80", "CR81",

  "PSR",  "PSR_L",  "PSR_UM",

  "RR0", "RR1", "RR2", "RR3", "RR4", "RR5", "RR6", "RR7",

  "PKR",  "ITR",  "ITC",  "DTR",  "DTC",  "IBR",  "DBR",  "PMC",  "RESOURCE_MEM",


  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  
  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",
  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",
  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",
  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?",  "?"

};
const t_architecture_description ia64_description =
{
   41, /* encoded minimal size */
   41, /* encoded maximum size */
   41, /* mod size */
   3, /* bundel size */
   5, /* bundel template size */
   sizeof(t_ia64_ins),/* size of an instruction */
   ADDRSIZE64,
   128, /* int regs */
   128, /* float regs */
   64, /* pred regs */
   8, /* branch regs */
   
   /* special regs */
   128 +  /*gr-nat bits*/
   128 +  /*application registers*/ 
   1 +    /* IP */
   1 +    /* CFM */
   1 +    /* UM */
   1 +    /* CPUID */
   1 +    /* PMD */
   82 +   /* Control Registers*/
   1 +    /* PSR */
   1 +    /* PSR.L */ /*\TODO dit moet nog anders opgelost worden*/
   1 +    /* PSR.UM */ /*\TODO dit moet nog anders opgelost worden*/
   8 +    /* Region Registers*/
   8,     /* indirect system registers*/
   names,
   {MAX_REG_ITERATOR, {~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL}},/* all */
   {MAX_REG_ITERATOR, {~0x0ULL,~0x0ULL,~0x0ULL,~0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL}}, /* int */
   {MAX_REG_ITERATOR, {0x0ULL,0x0ULL,0x0ULL,0x0ULL,~0x0ULL,~0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL}}, /* flt */
   {MAX_REG_ITERATOR, {0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,~0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL}}, /* cond */
   {MAX_REG_ITERATOR, {0x0ULL}}, /* callee_saved */
   {MAX_REG_ITERATOR, {0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,~0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL}}, /* callee_may_use */
   {MAX_REG_ITERATOR, {0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,~0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL}}, /* callee_may_change*/
   {MAX_REG_ITERATOR, {0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,~0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL}}, /*callee_may_return*/
   {MAX_REG_ITERATOR, {0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,~0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL,0x0ULL}}, /* always_live*/
   {MAX_REG_ITERATOR, {0x0ULL}},/* registers_prop_over_hell */
   {MAX_REG_ITERATOR, {0x0ULL}},/* const_registers */
   {MAX_REG_ITERATOR, {0x0ULL}},/* Dead over call */
   REG_NONE, /* program counter */
   NULL, /* opts */
#ifdef INSTRUMENTATION
   NULL, /* atom_actions */
#endif
   NULL,	/* An array with names of unbehaved functions */
   NULL,	/* add jump instruction*/
   Ia64IsCall,	/* is procedure call */
   {MAX_REG_ITERATOR, {0x0ULL}},	/* link_registers */
   NULL		/*computeLiveRegsBefore*/
};


/* vim: set shiftwidth=2: */


