/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble.h"
#include "ia64_disassemble/ia64_disassemble_one.h"
#include "ia64_types.h"
#include "ia64_utils.h"
#include "ia64_instructions.h"
#include "ia64_registers.h"
#include "ia64_opcodes.h"
#include "ia64_bundle.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_addresses.h"
#include "../../kernel/diablo_iterators.h"
#include "../../kernel/diablo_stack_machine.h"

#include "ia64_assemble.h"


t_bundle SectionGetBundle(t_section * sec, t_address offset);
t_uint64 SectionGetData(t_section * sec, t_address offset);



void * Ia64DisassembleOneInstruction(t_object * obj, t_address start, int * size_ret)
{
	FATAL(("Implement"));
   return NULL; /* Keep the compiler happy */
}

void * Ia64DisassembleOneInstructionForSection(t_section * sec, int offset, int * size_ret)
{
	FATAL(("Implement"));
   return NULL; /* Keep the compiler happy */
}

void Ia64InitializeInstruction(t_ins * ins_s, t_uint64 orig_instr,
    t_ia64_instr_type instr_type, t_uint16 opcode, t_ia64_reg pred)
{
  IA64INSTYPE(ins_s) = instr_type;        
  IA64INSOPCODE(ins_s) = opcode;
  
  IA64INSPRED(ins_s) = pred;
  
  IA64INSIN1(ins_s) = IA64_NONE;
  IA64INSIN2(ins_s) = IA64_NONE;
  IA64INSIN3(ins_s) = IA64_NONE;
  
  IA64INSOUT1(ins_s) = IA64_NONE;
  IA64INSOUT2(ins_s) = IA64_NONE;
  
  IA64INSIMM1(ins_s) = 0;
  IA64INSIMM2(ins_s) = 0;
  IA64INSIMM3(ins_s) = 0;

#ifndef IA64_NEW_BBL_ADDRESS
  IA64INSORIGINSTR(ins_s)=orig_instr;
#endif 	  
}


/*
typedef struct __ia64_address_item
{
  t_uint64 address;
  struct __ia64_address_item *next;
  
} ia64_address_item;

static ia64_address_item *powAddresses = NULL;
static powAddressCount = 0;
*/

#if 0
static __powPatchActive = FALSE;
static t_uint64 __lowAddress = (t_uint64)-1ULL;
static t_uint64 __highAddress = 0;

void Ia64PatchPowFunctionTable(t_object * obj)
{
  t_symbol *symptr;
  t_object * tmp2, * tmp;

  /*seeking for special $diablo-symbols - pow-addresses*/
  for (tmp2=obj->mapped_first;tmp2!=NULL; tmp2=tmp2->next)
  { 
    for (tmp=tmp2; tmp!=NULL; tmp=tmp->equal)
    {     
      t_address generic;
      t_uint32 tel;
      
      for (symptr=tmp->symbol_table->first; symptr!=NULL; symptr=symptr->next)
      {
	if(!symptr->name)
	  continue;

	if(strcmp(symptr->name,"__libm_error_region") == 0)
	{
	  t_uint64 addr = G_T_UINT64(GENERIC_ADD(symptr->base->address, symptr->current_address));
	  
	  __lowAddress = MIN(__lowAddress, addr);
	  
	  __highAddress = MAX(__highAddress, addr);
	  
	}
	
	if (strncmp(symptr->name,"$diablo:pow",11)==0) /* !!!!!! pow added !!!!! */
	{
	  int begin = strstr(symptr->name,"@-@")-symptr->name;
	  char save = symptr->name[begin];
	  t_symbol *s1, *s2;
	  t_address a1, a2;
#ifdef IA64_DEBUG_RELOCS
	  printf("symbol: %s @ 0x%llx\n", symptr->name, symptr->base->address);
#endif	  
	  symptr->name[begin] = '\0';
	  

	  /*TODO ook checken als het symbool van zelfde .o-file komt*/
	  s1=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+8);
	  s2=SymbolTableGetSymbolByName(tmp->symbol_table, symptr->name+begin+3);

	  a1 = GENERIC_ADD(symptr->base->address, s1->current_address);
	  a2 = GENERIC_ADD(symptr->base->address, s2->current_address);
/*
	  printf("%s @ 0x%llx;  %s @ 0x%llx\n", s1->name, a1, s2->name, a2);
          fflush(stdout);
*/	  
	  /*char herstellen*/
	  symptr->name[begin] = save;
	  
	  /*controleren of wel bestaat !*/
	  if(!s1|| !s2)
	    FATAL(("Invalid special diablo symbol !!!"));

	  __powPatchActive = TRUE;

	  __lowAddress = MIN(__lowAddress, G_T_UINT64(a1));
	  __lowAddress = MIN(__lowAddress, G_T_UINT64(a2));
	  
	  __highAddress = MAX(__highAddress, a1);
	  __highAddress = MAX(__highAddress, a2);
	  
	}
      }

    }
  }
  
}

t_bool Ia64PatchIsPowAddress(t_uint64 addr)
{
  return __powPatchActive && __lowAddress<=addr && addr <__highAddress;
}

#endif

void Ia64DisassembleSection(t_object * obj, t_section * code)
{
  t_address offset;
  t_uint16 i, slot;
  t_ins * ins_s;
  t_ins * prev;
  t_bundle bundle;

  t_ia64_unit * units;
  t_ia64_instr_type instr_type;

#if 0
  /*before we do anything: remove annoying relocations, for which the to-relocations is nonexistent(null)*/
  t_reloc *rel, *tmp;

  OBJECT_FOREACH_RELOC_SAFE(obj, rel, tmp)
  {
    if(GENERIC_EQ(REL_TO_RELOCATABLE(rel)->caddress, NULL_GENERIC))   
//    if(G_T_UINT64(REL_TO_RELOCATABLE(rel)->caddress) == 0x4000000000000000ull)   
    {
      t_address ret;
      /*eerst relocatie opnieuw herstellen...*/
      REL_TO_RELOCATABLE(rel)->caddress = GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset);
      ret=StackExec(rel->code,rel,T_SECTION(REL_FROM(rel))->data,TRUE,0);

      if (GENERIC_EXTRACT_UINT32(ret)!=0) FATAL(("Relocation failed!"));

      /*den verwijderen*/
      RelocTableRemoveReloc(obj->reloc_table, rel);
    }
  }
#endif

  GenericSet_uint64(offset,0);

  /*special patch*/
  //Ia64PatchPowFunctionTable(obj);
  
//  printf("halloooo @ 0x%llx", G_T_UINT64(SECTION_GET_CADDRESS(code)));
//  fflush(stdout);
  
//  DiabloPrint("Code section addr: @G\n", SECTION_GET_CADDRESS(code));
//  fflush(stdout);

  for ( prev = NULL; GENERIC_LT(offset, SECTION_CSIZE(code)); offset=GENERIC_ADD_UINT32(offset,16))
  {
//    printf("offset = 0x%llx\n",  G_T_UINT64(offset));
//    fflush(stdout);

    bundle = SectionGetBundle(code, offset);

    /*special marker: skip holes in code (viz: __evoke_link_warning_sys_errlist etc)*/
/*    
    if(bundle.template == 16 && bundle.slot[0] == 0x8484848484ull && bundle.slot[1] == 0x4242424242ull && bundle.slot[2] == 0x12121212121ull &&
      SymbolTableGetDataType(obj, GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) != ADDRESS_IS_DATA)
    {
      printf("hallo\n");
    }
*/    
    units = DecodeTemplate(&bundle);

    for(slot = 0; slot < 3 ; slot++)
    {
      ins_s = InsNewForSec(code);
      if (!ins_s) FATAL(("No instructions!"));

      RELOCATABLE_REFERS_TO(ins_s)=NULL;  /* TODO: Check if this is necessary */
      RELOCATABLE_REFED_BY(ins_s)=NULL; /* TODO: Check if this is necessary */


      if (/*Ia64PatchIsPowAddress(G_T_UINT64(GENERIC_ADD(SECTION_GET_CADDRESS(code),offset))) || */
	  SymbolTableGetDataType(obj, GENERIC_ADD(SECTION_GET_CADDRESS(code),offset)) == ADDRESS_IS_DATA || (bundle.template == 16 && bundle.slot[0] == 0x8484848484ull && bundle.slot[1] == 0x4242424242ull && bundle.slot[2] == 0x12121212121ull))
      {
//	DiabloPrint(stdout, "hallo, ik ben gearriveerd @G.\n", GENERIC_ADD(SECTION_GET_CADDRESS(code),offset));
	Ia64InitializeInstruction(ins_s, bundle.slot[slot], IA64_TYPE_DATA, DATA_OPCODE, IA64_P0);
	//INS_TYPE(ins_s) = IT_DATA;

	if(slot==0)
	{
	  INS_CADDRESS(ins_s) = GENERIC_ADD(SECTION_GET_CADDRESS(code),offset);
	  INS_OLD_ADDRESS(ins_s) = INS_CADDRESS(ins_s);
	  IA64INSIMM1(ins_s)= SectionGetData(code, offset);
	}
	else if(slot==1)
	{
	  INS_CADDRESS(ins_s) = GENERIC_ADD_UINT32(GENERIC_ADD(SECTION_GET_CADDRESS(code),offset), 1);
	  INS_OLD_ADDRESS(ins_s) = INS_CADDRESS(ins_s);
	  //T_RELOCATABLE(ins_s)->address = GENERIC_ADD_UINT32(GENERIC_ADD(T_RELOCATABLE(code)->address,offset), 8);
	  IA64INSIMM1(ins_s)= SectionGetData(code, GENERIC_ADD_UINT32(offset,8));
	}
	else /*slot = 2*/
	  INS_CADDRESS(ins_s) = GENERIC_ADD_UINT32(GENERIC_ADD(SECTION_GET_CADDRESS(code),offset), 2);
	  INS_OLD_ADDRESS(ins_s) = INS_CADDRESS(ins_s);
	//IA64DisassembleData(ins_s, data1, data2, slot, IA64_DATA);
      }
      else
      {

	instr_type = unit2instr_type(bundle.slot[slot], units[slot]);

	/* data-gedeelte overslaan... */
	if(instr_type == IA64_TYPE_X) 
	  slot++;

	/* 
	* kijken of we geen dummy instructie moeten toevoegen 
	* o.a. nodig bij opbouwen flowgraph, opdat de bundlegrenzen 
	* altijd zouden overeenstemmen met 3 (opeenvolgende) instructies
	* Deze dummy plaatsen we vr de X-type-instructie, 
	* zodat deze niet in een andere BBL zou terecht komen...
	 */
	if(instr_type == IA64_TYPE_X)
	{
	  Ia64InitializeInstruction(ins_s, bundle.slot[slot-1], IA64_TYPE_DUMMY, DUMMY_X_OPCODE, IA64_P0);
//	  InsPrint(ins_s,0);

	  /* zal zelfde instructiegroepnr moeten hebben als volgende X  */
	  IA64INSINSTRGROUP(ins_s) = GetCurrentInstrGroup();

	  INS_PREV(ins_s) = T_INS(prev);

	  INS_REGS_USE(ins_s)=Ia64UsedRegisters(ins_s);
	  INS_REGS_DEF(ins_s)=Ia64DefinedRegisters(ins_s);

	  if (prev) INS_NEXT(prev) = ins_s;

	  prev=ins_s;      	

	  INS_CADDRESS(ins_s) = GENERIC_ADD_UINT32(GENERIC_ADD(SECTION_GET_CADDRESS(code),offset),slot-1);
	  INS_OLD_ADDRESS(ins_s) = INS_CADDRESS(ins_s);

	  ins_s = InsNewForSec(code);
	  if (!ins_s) FATAL(("No instructions!"));

	  RELOCATABLE_REFERS_TO(ins_s)=NULL;  /* TODO: Check if this is necessary */
	  RELOCATABLE_REFED_BY(ins_s)=NULL; /* TODO: Check if this is necessary */

	}


	/* 
	   Een interessante versnelling is voorsorteren op major opcode: 
	   In de meeste tabellen staan ze gesorteerd volgens major opcode: wat je kan doen
	   is om ipv for(i=0... for(i=GetStartIndex(instr_type, bundle.slot[slot])...
	 */

	for (i = GetStartOffset(instr_type, GETBITS64(bundle.slot[slot], 37, 4)); 
	    ((bundle.slot[slot]) & ia64_opcode_table[instr_type][i].mask) != ia64_opcode_table[instr_type][i].opcode; 
	    i++)
	{
/*	  
	  if(G_T_UINT64(offset) == 0x13a0 && slot == 2)
	  {
	    printf("mask = 0x%llx, bundle = 0x%llx, bundle2 = 0x%llx, opcode = 0x%llx\n", ia64_opcode_table[instr_type][i].mask, 
		bundle.slot[slot], (bundle.slot[slot]) & ia64_opcode_table[instr_type][i].mask, ia64_opcode_table[instr_type][i].opcode);
	  }
*/	  
	}

	/* Alles op voorhand initialiseren... */
	Ia64InitializeInstruction(ins_s, bundle.slot[slot], instr_type, i, GET_PRED(bundle.slot[slot]));

//	printf("0x%llx\n", bundle.slot[slot]);
//	printf("instr_type = %d, i = %d, instr_format = %d\n", instr_type, i, ia64_opcode_table[instr_type][i].instr_format);
	Ia64DisassembleFunction_table[ia64_opcode_table[instr_type][i].instr_format](ins_s,(bundle.slot[slot]), &bundle);
	/*
	   Ia64AssembleFunction_table[IA64OPCODEINFO(ins_s).instr_format](ins_s, &bundle.slot[slot], &bundle);
	 */      
	if(IsFollowedByStop(&bundle, slot))
	  IA64INSINSTRGROUP(ins_s) = GetCurrentInstrGroupAndInc();
	else
	  IA64INSINSTRGROUP(ins_s) = GetCurrentInstrGroup();

	if (IA64INSPRED(ins_s) != IA64_P0) 
	  INS_ATTRIB(ins_s)|=IF_CONDITIONAL;

	INS_CADDRESS(ins_s) = GENERIC_ADD_UINT32(GENERIC_ADD(SECTION_GET_CADDRESS(code),offset),slot);
	INS_OLD_ADDRESS(ins_s) = INS_CADDRESS(ins_s);

      }

      INS_PREV(ins_s) = T_INS(prev);
	
      INS_REGS_USE(ins_s)=Ia64UsedRegisters(ins_s);
      INS_REGS_DEF(ins_s)=Ia64DefinedRegisters(ins_s);
	
      if (prev) INS_NEXT(prev) = ins_s;

      prev=ins_s;
     

      if(Ia64IsBranchOrCall(ins_s) ||Ia64IsReturn(ins_s)||Ia64IsCheck(ins_s) || Ia64IsBreak(ins_s))
	INS_TYPE(ins_s) = IT_BRANCH;
//      DiabloPrint(stdout, "New ins: @I\n", ins_s);

    }

 
  }

  printf("Need %d instructions\n",SECTION_NINS(code));
}

t_uint64 SectionGetData(t_section * sec, t_address offset)
{
  t_uint32 data0, data1;
  t_uint64 data;
  data0 = SectionGetData32(sec,offset);
  offset = GENERIC_ADD_UINT32(offset,4);
  data1 = SectionGetData32(sec,offset);
  offset = GENERIC_ADD_UINT32(offset,4);

  data = data0 | (((t_uint64) data1)<<32);
  return data;
  
}
t_bundle SectionGetBundle(t_section * sec, t_address offset)
{
	t_uint32 data0, data1, data2, data3;
	t_bundle bundle;

	data0 = SectionGetData32(sec,offset);
	offset = GENERIC_ADD_UINT32(offset,4);
	data1 = SectionGetData32(sec,offset);
	offset = GENERIC_ADD_UINT32(offset,4);

	data2 = SectionGetData32(sec,offset);
	offset = GENERIC_ADD_UINT32(offset,4);
	data3 = SectionGetData32(sec,offset);

	bundle.template = data0 & 0x0000001F;
	bundle.slot[0] =
		((data0 >> 5) & 0x07FFFFFF) | ((t_uint64)(data1 & 0x00003FFF) << 27);
	bundle.slot[1] =
		((data1 >> 14) & 0x0003FFFF) | ((t_uint64)(data2 & 0x007FFFFF) << 18);
	bundle.slot[2] =
		((data2 >> 23) & 0x000001FF) | ((t_uint64)data3 << 9);

	return bundle;
}

/* vim: set shiftwidth=2: */

