/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"
#ifndef IA64_DISASSEMBLE_H
#define IA64_DISASSEMBLE_H
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_addresses.h"
#include "../../kernel/diablo_section.h"
void * Ia64DisassembleOneInstruction(t_object * obj, t_address start, int * size_ret);
void * Ia64DisassembleOneInstructionForSection(t_section * sec, int offset, int * size_ret);
void Ia64DisassembleSection(t_object * obj, t_section * code);

#include "ia64_instructions.h"

void Ia64InitializeInstruction(t_ins * ins_s, t_uint64 orig_instr,
    t_ia64_instr_type instr_type, t_uint16 opcode, t_ia64_reg pred);

#include "ia64_utils.h"

/* shorthands used in disassemble.c */

#define GET_PRED(instr) (IA64_PR + ((instr) & 0x3F))
#define GET_GR_R1(instr) (IA64_GR + GETBITS64(instr, 6, 7))
#define GET_GR_R2(instr) (IA64_GR + GETBITS64(instr, 13, 7))
#define GET_GR_R3(instr) (IA64_GR + GETBITS64(instr, 20, 7))

#define GET_FR_F1(instr) (IA64_FR + GETBITS64(instr, 6, 7))
#define GET_FR_F2(instr) (IA64_FR + GETBITS64(instr, 13, 7))
#define GET_FR_F3(instr) (IA64_FR + GETBITS64(instr, 20, 7))
#define GET_FR_F4(instr) (IA64_FR + GETBITS64(instr, 27, 7))

#define GET_PR_P1(instr) (IA64_PR + GETBITS64(instr, 6, 6))
#define GET_PR_P2(instr) (IA64_PR + GETBITS64(instr, 27, 6))

#define GET_AR_R3(instr) (IA64_AR + GETBITS64(instr, 20, 7))
#define GET_CR_R3(instr) (IA64_CR + GETBITS64(instr, 20, 7))

#define GET_BR_B1(instr) (IA64_BR + GETBITS64(instr, 6, 3))
#define GET_BR_B2(instr) (IA64_BR + GETBITS64(instr, 13, 3))

#define IMM_S_I_BIT(instr) (GETBITS64(instr, 36, 1))

#define IMM_IMM2c(instr) (GETBITS64(instr, 33, 2))
#define IMM_IMM2d(instr) (GETBITS64(instr, 30, 2))
#define IMM_IMMct2d(instr) (GETBITS64(instr,27,2))
#define IMM_IMM4c(instr) (GETBITS64(instr, 20, 4))
#define IMM_IMM4d(instr) (GETBITS64(instr, 27, 4))
#define IMM_IMM5b(instr) (GETBITS64(instr, 14, 5))
#define IMM_IMM5c(instr) (GETBITS64(instr, 20, 5))
#define IMM_IMM5cA(instr) (GETBITS64(instr, 22, 5))/********/
#define IMM_IMM6b(instr) (GETBITS64(instr, 14, 6))
#define IMM_IMM6c(instr) (GETBITS64(instr, 20, 6))
#define IMM_IMM6d(instr) (GETBITS64(instr, 27, 6))/*******/
#define IMM_IMM6e(instr) (GETBITS64(instr, 31, 6))/***is in feite ook d****/
#define IMM_IMM7a(instr) (GETBITS64(instr, 6, 7))
#define IMM_IMM7b(instr) (GETBITS64(instr, 13, 7))
#define IMM_IMM7c(instr) (GETBITS64(instr, 20, 7))
#define IMM_IMM8c(instr) (GETBITS64(instr, 20, 8))/******/
#define IMM_IMM8d(instr) (GETBITS64(instr, 24, 8))/***idem***/
#define IMM_IMM9c(instr) (GETBITS64(instr, 24, 9))
#define IMM_IMM9d(instr) (GETBITS64(instr, 27, 9))
#define IMM_IMM20a(instr) (GETBITS64(instr, 6, 20))
#define IMM_IMM20b(instr) (GETBITS64(instr, 13, 20))
#define IMM_IMM27a(instr) ((t_uint64)GETBITS64(instr, 6, 27))

#define IMM_IMM21_F15(instr) \
    ( IMM_S_I_BIT(instr) << 20 | IMM_IMM20a(instr) )

#define IMM3_A2_A10(instr)\
    (IMM_IMMct2d(instr) + 1)

#define IMM8_A3_A8_I27_M30(instr) \
    (SIGN_EXT(IMM_S_I_BIT(instr)<<7 | IMM_IMM7b(instr), 8))

#define IMM9_F5(instr) \
    ( IMM_IMM7c(instr)<<2 | IMM_IMM2c(instr) )

#define IMM9_M3_M8_M15(instr) \
    (SIGN_EXT(IMM_S_I_BIT(instr)<<8 | GETBITS64(instr, 27, 1)<<7 | IMM_IMM7b(instr), 9))

#define IMM9_M5_M10(instr) \
    (SIGN_EXT(IMM_S_I_BIT(instr)<<8 | GETBITS64(instr, 27, 1)<<7 | IMM_IMM7a(instr), 9))

#define IMM14_A4(instr)\
    (SIGN_EXT(IMM_S_I_BIT(instr)<<13|IMM_IMM6d(instr)<<7| IMM_IMM7b(instr),14))

#define IMM21_B9(instr) \
    (IMM_S_I_BIT(instr)<<20 | IMM_IMM20a(instr))

#define IMM22_A5(instr)\
    (SIGN_EXT(IMM_S_I_BIT(instr)<<21|IMM_IMM5cA(instr)<<16|IMM_IMM9d(instr)<<7|IMM_IMM7b(instr),22))

#define IMM21_I19_M37(instr) \
    (IMM_S_I_BIT(instr)<<20 | IMM_IMM20a(instr))

#define IMM24_M44(instr) \
    (IMM_S_I_BIT(instr)<<23 | GETBITS64(instr, 31, 2)<<21 | GETBITS64(instr, 6, 21))

#define IMM44_I24(instr)\
    ( SIGN_EXT(IMM_S_I_BIT(instr)<<43|IMM_IMM27a(instr)<<16,44) )

#define IMM62_X1(instr, data) \
    ( (data)<<21 | IMM_S_I_BIT(instr)<<20 | IMM_IMM20a(instr))

#define IMM64_X2(instr, data) \
    (IMM_S_I_BIT(instr)<<63 | (data)<<22 | GETBITS64(instr, 21, 1)<<21 | \
     GETBITS64(instr, 22, 5)<<16 | GETBITS64(instr, 27, 9)<<7 | GETBITS64(instr, 13, 7))

/*Opgelet: bij de IP-relatieve instructies, worden enkel de displacements bijgehouden !!! */

#define TARGET25_B1_B2_B3_B6(instr) \
    (SIGN_EXT(IMM_S_I_BIT(instr)<<20 | GETBITS64(instr, 13, 20), 21)<<4)

#define TAG13_B6_B7(instr) \
    (SIGN_EXT(GETBITS64(instr, 33, 2)<<7 | GETBITS64(instr, 6, 7), 9)<<4)

#define TARGET25_B1(instr) TARGET25_B1_B2_B3_B6(instr)
#define TARGET25_B2(instr) TARGET25_B1_B2_B3_B6(instr)
#define TARGET25_B3(instr) TARGET25_B1_B2_B3_B6(instr)
#define TARGET25_B6(instr) TARGET25_B1_B2_B3_B6(instr)

#define TAG13_B6(instr) TAG13_B6_B7(instr)
#define TAG13_B7(instr) TAG13_B6_B7(instr)

#define TARGET25_F14(instr) \
    (SIGN_EXT(IMM_S_I_BIT(instr)<<20 | IMM_IMM20a(instr)<<7 , 21)<<4)


#define TARGET25_I20_M20_M21(instr) \
    (SIGN_EXT(IMM_S_I_BIT(instr)<<20 | GETBITS64(instr, 20, 13)<<7 | IMM_IMM7a(instr), 21)<<4)

#define TAG13_I21(instr)\
    (SIGN_EXT(IMM_IMM9c(instr),9)<<4)

#define MASK17_I23(instr)\
    ( SIGN_EXT(IMM_S_I_BIT(instr)<<16|IMM_IMM8d(instr)<<8|IMM_IMM7a(instr)<<1 ,17) )

#define TARGET64_X3_X4(instr, data) \
    (( IMM_S_I_BIT(instr)<<59 | GETBITS64((data), 2, 39)<<20 | IMM_IMM20b(instr) )<<4)

#define TARGET64_X3(instr, data) TARGET64_X3_X4(instr, data)
#define TARGET64_X4(instr, data) TARGET64_X3_X4(instr, data)

#define INC3_M17(instr) \
    (SIGN_EXT( ( (GETBITS64(instr, 15, 1))? -1:1) * ((GETBITS64(instr, 13, 2) == 3)? 1:1<< (4-GETBITS64(instr, 13,2))) ,6))

#define TARGET25_M22_M23(instr) \
    (SIGN_EXT(IMM_S_I_BIT(instr)<<20 | GETBITS64(instr, 13, 20), 21)<<4)


#define IMM9_M3(instr) IMM9_M3_M8_M15(instr)
#define IMM9_M5(instr) IMM9_M5_M10(instr)
#define IMM9_M8(instr) IMM9_M3_M8_M15(instr)
#define IMM9_M10(instr) IMM9_M5_M10(instr)
#define IMM9_M15(instr) IMM9_M3_M8_M15(instr)
#define IMM8_M30(instr) IMM8_A3_A8_I27_M30(instr)
#define IMM21_M37(instr) IMM21_I19_M37(instr)

#define TARGET25_M20(instr) TARGET25_I20_M20_M21(instr)
#define TARGET25_M21(instr) TARGET25_I20_M20_M21(instr)
#define TARGET25_M22(instr) TARGET25_M22_M23(instr)
#define TARGET25_M23(instr) TARGET25_M22_M23(instr)
#endif
/* vim: set shiftwidth=2: */
