/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble_one.h"
#include "../ia64_instructions.h"

#include "../../../utils/util_error.h"

/*#include <stdio.h>*/



void Ia64DisassembleUnsupported(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
/*  printf("Unsupported ia64 instruction. Opcode = %d\n", opc);*/
  DiabloPrint(stdout,"Instruction: @I\n", ins);
  fflush(stdout);
  FATAL(("Unsupported ia64 instruction. Opcode = %d\n", IA64INSOPCODE(ins)));
}





/* 
 * bevat de lijst van alle disassemble functions, 1 voor elk instructie formaat
 * de indexering in deze tabel moet dan ook volgens t_ia64_instr_format gebeuren...
 */
const Ia64DisassembleFunction Ia64DisassembleFunction_table [] =
{
	Ia64DisassembleA1, Ia64DisassembleA2, Ia64DisassembleA3, Ia64DisassembleA4, Ia64DisassembleA5,
	Ia64DisassembleA6, Ia64DisassembleA7, Ia64DisassembleA8, Ia64DisassembleA9, Ia64DisassembleA10,
	
	Ia64DisassembleI1, Ia64DisassembleI2, Ia64DisassembleI3, Ia64DisassembleI4, Ia64DisassembleI5,
	Ia64DisassembleI6, Ia64DisassembleI7, Ia64DisassembleI8, Ia64DisassembleI9, Ia64DisassembleI10,
	Ia64DisassembleI11, Ia64DisassembleI12, Ia64DisassembleI13, Ia64DisassembleI14, Ia64DisassembleI15,
	Ia64DisassembleI16, Ia64DisassembleI17, Ia64DisassembleUnsupported, Ia64DisassembleI19, Ia64DisassembleI20,
	Ia64DisassembleI21, Ia64DisassembleI22, Ia64DisassembleI23, Ia64DisassembleI24, Ia64DisassembleI25,
	Ia64DisassembleI26, Ia64DisassembleI27, Ia64DisassembleI28, Ia64DisassembleI29, 

	Ia64DisassembleM1, Ia64DisassembleM2, Ia64DisassembleM3, Ia64DisassembleM4, Ia64DisassembleM5,
	Ia64DisassembleM6, Ia64DisassembleM7, Ia64DisassembleM8, Ia64DisassembleM9, Ia64DisassembleM10,
	Ia64DisassembleM11, Ia64DisassembleM12, Ia64DisassembleM13, Ia64DisassembleM14, Ia64DisassembleM15,
	Ia64DisassembleM16, Ia64DisassembleM17, Ia64DisassembleM18, Ia64DisassembleM19, Ia64DisassembleM20,
	Ia64DisassembleM21, Ia64DisassembleM22, Ia64DisassembleM23, Ia64DisassembleM24, Ia64DisassembleM25,
	Ia64DisassembleM26, Ia64DisassembleM27, Ia64DisassembleM28, Ia64DisassembleM29, Ia64DisassembleM30,
	Ia64DisassembleM31, Ia64DisassembleM32, Ia64DisassembleM33, Ia64DisassembleM34, Ia64DisassembleM35,
	Ia64DisassembleM36, Ia64DisassembleM37, Ia64DisassembleM38, Ia64DisassembleM39, Ia64DisassembleM40,
	Ia64DisassembleM41, Ia64DisassembleM42, Ia64DisassembleM43, Ia64DisassembleM44, Ia64DisassembleM45,
	Ia64DisassembleM46, 
	
	Ia64DisassembleB1, Ia64DisassembleB2, Ia64DisassembleB3, Ia64DisassembleB4, Ia64DisassembleB5,
	Ia64DisassembleB6, Ia64DisassembleB7, Ia64DisassembleB8, Ia64DisassembleB9,	

	Ia64DisassembleF1, Ia64DisassembleF2, Ia64DisassembleF3, Ia64DisassembleF4, Ia64DisassembleF5,
	Ia64DisassembleF6, Ia64DisassembleF7, Ia64DisassembleF8, Ia64DisassembleF9, Ia64DisassembleF10,
	Ia64DisassembleF11, Ia64DisassembleF12, Ia64DisassembleF13, Ia64DisassembleF14, Ia64DisassembleF15,

	Ia64DisassembleX1, Ia64DisassembleX2, Ia64DisassembleX3, Ia64DisassembleX4, 
	Ia64DisassembleUnsupported, Ia64DisassembleUnsupported
	
};







/* vim: set shiftwidth=2: */
