/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble_one_b.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"

#ifdef _DEBUG_DISASSEMBLE_
#include "../ia64_print/ia64_print_b.h"
static char buffer[256];
#endif


/* B1: IP-Relative Branch */
void Ia64DisassembleB1(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = TARGET25_B1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB1(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* B2: IP-Relative Counted Branch */
void Ia64DisassembleB2(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = TARGET25_B2(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB2(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* B3: IP-Relative Call */
void Ia64DisassembleB3(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSOUT1(ins) = GET_BR_B1(instr);

  IA64INSIMM1(ins) = TARGET25_B3(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB3(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* B4: Indirect Branch */
void Ia64DisassembleB4(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_BR_B2(instr);

  if (GETBITS64(instr,6,3)==1) FATAL(("ATTEMPT TO SWITCH TO IA32"));

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB4(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* B5: Indirect Call */
void Ia64DisassembleB5(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  IA64INSIN1(ins) = GET_BR_B2(instr);

  IA64INSOUT1(ins) = GET_BR_B1(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB5(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* B6: IP-Relative Predict */
void Ia64DisassembleB6(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
	/* B6 en B7 zijn (voorlopig) de enige die de laagste 6 bits voor iets anders dan qp/0 gebruiken... */
	IA64INSPRED(ins) = IA64_P0;	
	

  IA64INSIMM2(ins) = TAG13_B6(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB6(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* B7: Indirect Predict */
void Ia64DisassembleB7(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
	/* B6 en B7 zijn (voorlopig) de enige die de laagste 6 bits voor iets anders dan qp/0 gebruiken... */
	IA64INSPRED(ins) = IA64_P0;	
	
  IA64INSIN1(ins) = GET_BR_B2(instr);

  IA64INSIMM1(ins) = TAG13_B7(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB7(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* B8: Miscellaneous (M-Unit) */
void Ia64DisassembleB8(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB8(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}


/* B9: Break/Nop (B-Unit) */
void Ia64DisassembleB9(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = IMM21_B9(instr);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintB9(ins, buffer);
	printf(buffer);
	printf("\n");
#endif


}

/* vim: set shiftwidth=2: */
