/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_disassemble_one_x.h"
#include "../../../utils/util_error.h"


#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"

#ifdef _DEBUG_DISASSEMBLE_
#include "../ia64_print/ia64_print_x.h"
static char buffer[256];
#endif


/* X1: Break/Nop (X-Unit) */
void Ia64DisassembleX1(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSIMM1(ins) = IMM62_X1(instr, bundle->slot[1]);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintX1(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}

/* X2: Move Long Immediate64 */
void Ia64DisassembleX2(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSOUT1(ins) = GET_GR_R1(instr);

  IA64INSIMM1(ins) = IMM64_X2(instr, bundle->slot[1]);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintX2(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* X3: Long Branch */
void Ia64DisassembleX3(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{
  
  IA64INSIMM1(ins) = TARGET64_X3(instr, bundle->slot[1]);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintX3(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}


/* X4: Long Call */
void Ia64DisassembleX4(t_ins * ins, t_uint64 instr, t_bundle *bundle)
{

  IA64INSOUT1(ins) = GET_BR_B1(instr);

  IA64INSIMM1(ins) = TARGET64_X4(instr, bundle->slot[1]);

#ifdef _DEBUG_DISASSEMBLE_
	Ia64PrintX4(ins, buffer);
	printf(buffer);
	printf("\n");
#endif

}



/* vim: set shiftwidth=2: */
