/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ia64_dummy.h" 
#include "../../utils/util_error.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_iterators.h"
#include "ia64_disassemble.h"
#include "ia64_instructions.h"
#include "ia64_opcodes.h"
#include "ia64_registers.h"
#include "../../utils/util_malloc.h"




/*
* alloc:
* - imm1 = sol
* - imm2 = sof
* - imm3 = sor (sor al * 8 gedaan !)
*
* dummy:
* - imm1 : bevat tijdens constantenpropagatie een pointer naar een tijdelijke procstate
* - imm2 : sol (IA64_ALLOC_MISSING als alloc niet gekend)
* - imm3 : sof (IA64_ALLOC_MISSING als alloc niet gekend)
*
* 
* */



void Ia64InitializeDummies(t_cfg * cfg)
{
  t_edge * edge, * funlink_edge;
  t_bbl * head;
  t_bbl * tail;
  t_bbl * i_bbl;
  t_bbl * ft_bbl;

  t_ins * ins_dummy1;
  t_ins * ins_dummy2;
  t_ins * tmp_ins;

  t_ins * probably_alloc;
  t_ins * ins_ft;
  t_uint32 countAllocs;

  CFG_FOREACH_EDGE(cfg, T_CFG_EDGE(edge))             
  {
    if(EDGE_CAT(edge)&ET_CALL)
    {
      head=CFG_EDGE_HEAD(edge);
      tail=CFG_EDGE_TAIL(edge);
      if(head->ins_last)
      {
	int i;

	probably_alloc = Ia64SearchForAlloc(head->fun);

	countAllocs = CountAllocsInFun(head->fun);

	if(probably_alloc == NULL || !FUN_ISALLOCCLEAN(head->fun))
	  probably_alloc = head->fun->bbl_first->ins_first;
	
	/* aanmaken dummy1 aan callside */
	ins_dummy1 = InsNewForBbl(head);
	
	if (!ins_dummy1) FATAL(("No instructions!"));

	Ia64InitializeInstruction(ins_dummy1, 0, IA64_TYPE_DUMMY, DUMMY_LIVE_OPCODE, IA64_P0);

	INS_BBL(ins_dummy1)=head;
	head->nins++; 

	IA64INSINSTRGROUP(ins_dummy1) = IA64INSINSTRGROUP(head->ins_last);

	/*hier zal er altijd een instructie zijn (minstens een call...)*/
	INS_PREV(ins_dummy1) = INS_PREV(head->ins_last);
	INS_NEXT(ins_dummy1) = T_INS(head->ins_last);
	if(INS_PREV(head->ins_last))
	{
	  INS_NEXT(INS_PREV(head->ins_last)) = T_INS(ins_dummy1);
	}
	INS_PREV(head->ins_last) = T_INS(ins_dummy1);

	RegsetSetEmpty(INS_REGS_USE(ins_dummy1));
	for(i=0;i<96;i++)
	{
	  RegsetAddReg(INS_REGS_USE(ins_dummy1),IA64_R32+i);	  
	  RegsetAddReg(INS_REGS_USE(ins_dummy1),IA64_R32_NAT+i); /* to be safe: also do the nat bits !!!*/
	}

	if(countAllocs < 2 && IA64ISALLOC(probably_alloc))
	{
	  IA64INSIMM1(ins_dummy1) = 0ULL;
	  IA64INSIMM2(ins_dummy1) = IA64INSIMM1(probably_alloc);
	  IA64INSIMM3(ins_dummy1) = IA64INSIMM2(probably_alloc);
	}
	else
	{
          IA64INSIMM1(ins_dummy1) = 0ULL;
	  IA64INSIMM2(ins_dummy1) = IA64_ALLOC_MISSING;
	  IA64INSIMM3(ins_dummy1) = IA64_ALLOC_MISSING;
	  
	}
	  
	
	BBL_FOREACH_SUCC_EDGE(head, T_CFG_EDGE(funlink_edge))
	{
	  if(EDGE_CAT(funlink_edge)&ET_FUNLINK)
	    ft_bbl=CFG_EDGE_TAIL(funlink_edge);
	}
	
	if(ft_bbl)
	{
	  //InsPrint(ins_ft, 0);
	  /* aanmaken dummy2 */
	  ins_dummy2 = InsNewForBbl(ft_bbl);

	  if (!ins_dummy2) FATAL(("No instructions!"));

	  Ia64InitializeInstruction(ins_dummy2, 0, IA64_TYPE_DUMMY, DUMMY_LIVE_OPCODE, IA64_P0);
	  INS_BBL(ins_dummy2)=ft_bbl;
	  INS_BBL(ins_dummy2)->nins++; 

	  IA64INSINSTRGROUP(ins_dummy2) = GetCurrentInstrGroupAndInc();

	  ins_ft = ft_bbl->ins_first;
	  ft_bbl->ins_first = T_INS(ins_dummy2);

	  INS_PREV(ins_dummy2) = NULL;
	  INS_NEXT(ins_dummy2) = T_INS(ins_ft);
	  if(ins_ft != NULL)
	    INS_PREV(ins_ft) = T_INS(ins_dummy2);
	  else
	    ft_bbl->ins_last = T_INS(ins_dummy2);

	  if(countAllocs < 2 && IA64ISALLOC(probably_alloc))
	  {
	    int i;
	    RegsetSetEmpty(INS_REGS_USE(ins_dummy2));
	    for(i=0;(i<IA64INSIMM2(probably_alloc)-IA64INSIMM1(probably_alloc));i++)
	    {
	      RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32+i);
	      RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32_NAT+i);
	    }

	    IA64INSIMM1(ins_dummy2) = -1LL;
	    IA64INSIMM2(ins_dummy2) = IA64INSIMM1(probably_alloc);
	    IA64INSIMM3(ins_dummy2) = IA64INSIMM2(probably_alloc);
	    
	  }
	  else
	  {
	    int i;
	    RegsetSetEmpty(INS_REGS_USE(ins_dummy2));
	    for(i=0;i<96;i++)
	    {
	      RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32+i);
	      RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32_NAT+i);
	    }

	    IA64INSIMM1(ins_dummy2) = -1LL;
	    IA64INSIMM2(ins_dummy2) = IA64_ALLOC_MISSING;
	    IA64INSIMM3(ins_dummy2) = IA64_ALLOC_MISSING;
	    
	  }

	}
	else FATAL(("NO FUNLINK"));
      }
    }
  }                                 
}  


void Ia64FinalizeDummies(t_cfg * cfg)
{
  t_bbl * i_bbl;
  CFG_FOREACH_BBL(cfg,i_bbl)            
  {      
    t_ins * tmp_ins;
    t_ins *ins;
    BBL_FOREACH_INS_SAFE(i_bbl, ins, tmp_ins)
    {
      if(ins && IA64ISLIVEDUMMY(ins))
      {
	InsKill(ins);
      }
    }
  }
}
/* vim: set shiftwidth=2: */
