/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"
#ifndef IA64_INSTRUCTION_TYPEDEFS
#define IA64_INSTRUCTION_TYPEDEFS
typedef struct _t_ia64_ins t_ia64_ins;
#endif

#ifndef IA64_INSTRUCTION_H
#define IA64_INSTRUCTION_H

#include "ia64_registers.h"
#include "ia64_bundle.h"
#include "../../kernel/diablo_ins.h"




typedef enum _t_ia64_instr_type
  {
    IA64_TYPE_A = 0,	/* integer alu (I- or M-unit) */
    IA64_TYPE_I,	    /* non-alu integer (I-unit) */
    IA64_TYPE_M,	    /* memory (M-unit) */
    IA64_TYPE_B,	    /* branch (B-unit) */
    IA64_TYPE_F,	    /* floating-point (F-unit) */
    IA64_TYPE_X,	    /* long encoding (X-unit) */
    /* Dient om de vulling voor een X-type instructie mee aan te duiden. */
    /* Of om de vulling voor liveness analyse */
    IA64_TYPE_DUMMY,
    IA64_TYPE_DATA,     /* data */
    IA64_TYPE_NIL	    /* illegal type */
  } t_ia64_instr_type;


/* bij wijziging van dit, moeten vb ook de arrays in ia64_opcodes.c aangepast worden*/
typedef enum _t_ia64_unit
  {
    IA64_UNIT_I = 0,	/* integer unit */
    IA64_UNIT_M,	/* memory unit */
    IA64_UNIT_B,	/* branching unit */
    IA64_UNIT_F,	/* floating-point unit */
    IA64_UNIT_X,	/* may be integer or branch unit */
    IA64_UNIT_L,	/* long "unit" */
    IA64_UNIT_NIL	/* illegal unit */
  } t_ia64_unit;

#define IA64_NUM_TYPES 6




typedef enum _t_ia64_instr_format
  {
    IA64_INSTR_FORMAT_A = 0,	
    IA64_INSTR_FORMAT_A1 = IA64_INSTR_FORMAT_A,	IA64_INSTR_FORMAT_A2,	IA64_INSTR_FORMAT_A3,	IA64_INSTR_FORMAT_A4,	
    IA64_INSTR_FORMAT_A5,	IA64_INSTR_FORMAT_A6,	IA64_INSTR_FORMAT_A7,	IA64_INSTR_FORMAT_A8,	
    IA64_INSTR_FORMAT_A9,	IA64_INSTR_FORMAT_A10,	

    IA64_INSTR_FORMAT_I,	
    IA64_INSTR_FORMAT_I1 = IA64_INSTR_FORMAT_I,	IA64_INSTR_FORMAT_I2,	IA64_INSTR_FORMAT_I3,	IA64_INSTR_FORMAT_I4,	
    IA64_INSTR_FORMAT_I5,	IA64_INSTR_FORMAT_I6,	IA64_INSTR_FORMAT_I7,	IA64_INSTR_FORMAT_I8,	
    IA64_INSTR_FORMAT_I9,	IA64_INSTR_FORMAT_I10, IA64_INSTR_FORMAT_I11, IA64_INSTR_FORMAT_I12,	
    IA64_INSTR_FORMAT_I13, IA64_INSTR_FORMAT_I14,	IA64_INSTR_FORMAT_I15, IA64_INSTR_FORMAT_I16,	
    IA64_INSTR_FORMAT_I17, IA64_INSTR_FORMAT_I18,	IA64_INSTR_FORMAT_I19, IA64_INSTR_FORMAT_I20,	
    IA64_INSTR_FORMAT_I21, IA64_INSTR_FORMAT_I22,	IA64_INSTR_FORMAT_I23, IA64_INSTR_FORMAT_I24,	
    IA64_INSTR_FORMAT_I25, IA64_INSTR_FORMAT_I26,	IA64_INSTR_FORMAT_I27, IA64_INSTR_FORMAT_I28,	
    IA64_INSTR_FORMAT_I29,   

    IA64_INSTR_FORMAT_M,	
    IA64_INSTR_FORMAT_M1 = IA64_INSTR_FORMAT_M,	IA64_INSTR_FORMAT_M2,	IA64_INSTR_FORMAT_M3,	IA64_INSTR_FORMAT_M4,	
    IA64_INSTR_FORMAT_M5,	IA64_INSTR_FORMAT_M6,	IA64_INSTR_FORMAT_M7,	IA64_INSTR_FORMAT_M8,	
    IA64_INSTR_FORMAT_M9,	IA64_INSTR_FORMAT_M10, IA64_INSTR_FORMAT_M11, IA64_INSTR_FORMAT_M12,	
    IA64_INSTR_FORMAT_M13, IA64_INSTR_FORMAT_M14,	IA64_INSTR_FORMAT_M15, IA64_INSTR_FORMAT_M16,	
    IA64_INSTR_FORMAT_M17, IA64_INSTR_FORMAT_M18,	IA64_INSTR_FORMAT_M19, IA64_INSTR_FORMAT_M20,	
    IA64_INSTR_FORMAT_M21, IA64_INSTR_FORMAT_M22,	IA64_INSTR_FORMAT_M23, IA64_INSTR_FORMAT_M24,	
    IA64_INSTR_FORMAT_M25, IA64_INSTR_FORMAT_M26,	IA64_INSTR_FORMAT_M27, IA64_INSTR_FORMAT_M28,	
    IA64_INSTR_FORMAT_M29, IA64_INSTR_FORMAT_M30, IA64_INSTR_FORMAT_M31, IA64_INSTR_FORMAT_M32, 
    IA64_INSTR_FORMAT_M33, IA64_INSTR_FORMAT_M34,	IA64_INSTR_FORMAT_M35, IA64_INSTR_FORMAT_M36,
    IA64_INSTR_FORMAT_M37, IA64_INSTR_FORMAT_M38, IA64_INSTR_FORMAT_M39, IA64_INSTR_FORMAT_M40, 
    IA64_INSTR_FORMAT_M41, IA64_INSTR_FORMAT_M42, IA64_INSTR_FORMAT_M43, IA64_INSTR_FORMAT_M44,
    IA64_INSTR_FORMAT_M45, IA64_INSTR_FORMAT_M46,

    IA64_INSTR_FORMAT_B,	
    IA64_INSTR_FORMAT_B1 = IA64_INSTR_FORMAT_B,	IA64_INSTR_FORMAT_B2,	IA64_INSTR_FORMAT_B3,	IA64_INSTR_FORMAT_B4,	
    IA64_INSTR_FORMAT_B5,	IA64_INSTR_FORMAT_B6,	IA64_INSTR_FORMAT_B7,	IA64_INSTR_FORMAT_B8,	
    IA64_INSTR_FORMAT_B9,	
    
    IA64_INSTR_FORMAT_F,	
    IA64_INSTR_FORMAT_F1 = IA64_INSTR_FORMAT_F,	IA64_INSTR_FORMAT_F2,	IA64_INSTR_FORMAT_F3,	IA64_INSTR_FORMAT_F4,	
    IA64_INSTR_FORMAT_F5,	IA64_INSTR_FORMAT_F6,	IA64_INSTR_FORMAT_F7,	IA64_INSTR_FORMAT_F8,	
    IA64_INSTR_FORMAT_F9,	IA64_INSTR_FORMAT_F10, IA64_INSTR_FORMAT_F11, IA64_INSTR_FORMAT_F12,	
    IA64_INSTR_FORMAT_F13, IA64_INSTR_FORMAT_F14,	IA64_INSTR_FORMAT_F15,    

    IA64_INSTR_FORMAT_X,	
    IA64_INSTR_FORMAT_X1 = IA64_INSTR_FORMAT_X,	IA64_INSTR_FORMAT_X2,	IA64_INSTR_FORMAT_X3,	IA64_INSTR_FORMAT_X4,
    	
    IA64_INSTR_FORMAT_UNSUPPORTED,

    IA64_INSTR_FORMAT_DUMMY,

    IA64_INSTR_FORMAT_DATA
    
    
  } t_ia64_instr_format;



#define IA64_INSTR_FLAG_USE_M_FOR_A  0x8000
#define IA64_INSTR_FLAG_USE_I_FOR_A  0x4000


struct _t_ia64_ins
{
  t_ins generic;
  t_ia64_instr_type instr_type;
  t_uint16 opcode; /* Change type if necessary */

  /* waarsch zal t_uint32 al meer dan genoeg zijn...*/
  t_uint64 instr_group; 

  t_instruction_flags attributes;    /* instructiegebonden vlaggen (set CPSR, ...) */

  t_ia64_reg predicate;

  t_ia64_reg in1;
  t_ia64_reg in2;
  t_ia64_reg in3;

  t_ia64_reg out1;
  t_ia64_reg out2;

  t_int64 imm1;
  t_int16 imm2;
  t_int16 imm3;

#ifndef IA64_NEW_BBL_ADDRESS
  t_uint64 orig_instr;
#endif  
  /* Add extra (ia64 specific) fields here */
};



typedef enum
{
  IA64_DV_RAW,
  IA64_DV_WAW,
  IA64_DV_WAR,
}t_ia64_dependency_mode;
    
typedef enum
{
  IA64_DVS_NONE,
  IA64_DVS_IMPLIED,
  IA64_DVS_IMPLIEDF,
  IA64_DVS_DATA,
  IA64_DVS_INSTR,
  IA64_DVS_SPECIFIC,
  IA64_DVS_STOP,
  IA64_DVS_OTHER,
} t_ia64_dependency_semantics;
    

typedef enum 
{
  IA64_RS_ANY,
  IA64_RS_AR_K,
  IA64_RS_AR_UNAT,
  IA64_RS_AR, /* 8-15, 20, 22-23, 31, 33-35, 37-39, 41-43, 45-47, 67-111 */
  IA64_RS_ARb, /* 48-63, 112-127 */
  IA64_RS_BR,
  IA64_RS_CFM,
  IA64_RS_CPUID,
  IA64_RS_CR_IRR,
  IA64_RS_CR_LRR,
  IA64_RS_CR, /* 3-7,10-15,18,26-63,75-79,82-127 */
  IA64_RS_DBR,
  IA64_RS_FR,
  IA64_RS_FRb,
  IA64_RS_GR0,
  IA64_RS_GR,
  IA64_RS_IBR,
  IA64_RS_INSERVICE, /* CR[EOI] or CR[IVR] */
  IA64_RS_MSR,
  IA64_RS_PKR,
  IA64_RS_PMC,
  IA64_RS_PMD,
  IA64_RS_PR,  /* non-rotating, 1-15 */
  IA64_RS_PRr, /* rotating, 16-62 */
  IA64_RS_PR63,
  IA64_RS_RR,

  IA64_RS_ARX, /* ARs not in RS_AR or RS_ARb */
  IA64_RS_CRX, /* CRs not in RS_CR */
  IA64_RS_PSR, /* PSR bits */
  IA64_RS_RSE, /* implementation-specific RSE resources */
  IA64_RS_AR_FPSR,
} t_ia64_resource_specifier;

typedef enum
{
  IA64_RSE_N_STACKED_PHYS,
  IA64_RSE_BOF,
  IA64_RSE_STORE_REG,
  IA64_RSE_LOAD_REG,
  IA64_RSE_BSPLOAD,
  IA64_RSE_RNATBITINDEX,
  IA64_RSE_CFLE,
  IA64_RSE_NDIRTY,
} t_ia64_rse_resource;

/* Information about a given resource dependency */
typedef struct 
{
  /* Name of the resource */
  const char *name;
  /* Does this dependency need further specification? */
  t_ia64_resource_specifier specifier;
  /* Mode of dependency */
  t_ia64_dependency_mode mode;
  /* Dependency semantics */
  t_ia64_dependency_semantics semantics;
  /* Register index, if applicable (distinguishes AR, CR, and PSR deps) */
#define REG_NONE (-1)
  int regindex;
  /* Special info on semantics */
  const char *info;
} t_ia64_dependency;

/* Two arrays of indexes into the ia64_dependency table.
   chks are dependencies to check for conflicts when an opcode is
   encountered; regs are dependencies to register (mark as used) when an
   opcode is used.  chks correspond to readers (RAW) or writers (WAW or
   WAR) of a resource, while regs correspond to writers (RAW or WAW) and
   readers (WAR) of a resource.  */
typedef struct
{
  int nchks;
  const unsigned short *chks;
  int nregs;
  const unsigned short *regs;
} t_ia64_opcode_dependency;





extern const t_ia64_unit templateField [32] [3];

t_ia64_unit * DecodeTemplate(const t_bundle *bundle);

t_ia64_instr_type unit2instr_type(t_uint64 instruction, t_ia64_unit unit);

t_bool IsFollowedByStop(const t_bundle *bundle, t_uint8 slot);

void MakeTemplate(t_bundle * bundle, const t_ia64_instr_type *instr_types, const t_bool *followedByStop);



/* Often used shorthands */


#define IA64INSOPCODE(ia64_ins) (((t_ia64_ins *) ia64_ins)->opcode)
#define IA64INSTYPE(ia64_ins)    (((t_ia64_ins *) ia64_ins)->instr_type)
#define IA64INSINSTRGROUP(ia64_ins)    (((t_ia64_ins *) ia64_ins)->instr_group)
#define IA64INSATTR(ia64_ins)    (((t_ia64_ins *) ia64_ins)->attributes)
#define IA64INSPRED(ia64_ins)    (((t_ia64_ins *) ia64_ins)->predicate)
#define IA64INSIN1(ia64_ins)    (((t_ia64_ins *) ia64_ins)->in1)
#define IA64INSIN2(ia64_ins)    (((t_ia64_ins *) ia64_ins)->in2)
#define IA64INSIN3(ia64_ins)    (((t_ia64_ins *) ia64_ins)->in3)
#define IA64INSOUT1(ia64_ins)    (((t_ia64_ins *) ia64_ins)->out1)
#define IA64INSOUT2(ia64_ins)    (((t_ia64_ins *) ia64_ins)->out2)
#define IA64INSIMM1(ia64_ins)    (((t_ia64_ins *) ia64_ins)->imm1)
#define IA64INSIMM2(ia64_ins)    (((t_ia64_ins *) ia64_ins)->imm2)
#define IA64INSIMM3(ia64_ins)    (((t_ia64_ins *) ia64_ins)->imm3)

#ifndef IA64_NEW_BBL_ADDRESS
#define IA64INSORIGINSTR(ia64_ins)    (((t_ia64_ins *) ia64_ins)->orig_instr)
#endif 

#define T_IA64_INS(ia64_ins)            ((t_ia64_ins *) ia64_ins)





void Ia64InstructionMakeNoop(t_ins * ins);

/*enkel voor addl of adds (met constante)*/
/*predicaat blijft behouden*/
void Ia64InstructionMakeAdd(t_ins * ins, t_bool useAddl, t_ia64_reg out, t_int64 imm, t_ia64_reg in);

t_ins * Ia64InstructionCreateNoop(t_section *code, t_ia64_unit unit);

void Ia64InstructionMakeCondBranch(t_ins * ins, t_ia64_reg pred);


#define IA64ISALLOC(ins) (IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_ALLOC)

#define IA64ISLIVEDUMMY(ins) (IA64INSTYPE(ins) == IA64_TYPE_DUMMY && IA64INSOPCODE(ins) == DUMMY_LIVE_OPCODE)




t_bool Ia64IsBranch(t_ins *ins);
t_bool Ia64IsCall(t_ins *ins);
t_bool Ia64IsBranchOrCall(t_ins *ins);

t_bool Ia64IsIPRelativeBranch(t_ins *ins);
t_bool Ia64IsIPRelativeCall(t_ins *ins);
t_bool Ia64IsIPRelativeBranchOrCall(t_ins *ins);

t_bool Ia64IsCheck(t_ins *ins);

/* kan nog als marco gedaan worden */
t_bool Ia64IsReturn(t_ins *ins);

t_bool Ia64IsModOrLoopBranch(t_ins *ins);

#define Ia64IsBreak(ins) (IA64OPCODEINFO(ins).opcode == 0x0ULL)

t_bool Ia64InsHasSideEffect(t_ins * ins);
t_tristate Ia64IsSyscallExit(t_ins * ins);
t_bool Ia64InsIsIndirectCall(t_ins * ins);


#endif
/* vim: set shiftwidth=2: */
