/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../kernel/diablo_edge.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_print.h"
#include "ia64_opcodes.h"
#include "ia64_registers.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_iterators.h"
#include "../../diablo_options.h"
#include <string.h>

#include "ia64_scheduler.h"
#include "ia64_statusbar.h"
#include "ia64_prograph.h"
#include "ia64_profiler.h"

#define rotate_left(v, n) (((v) << (n)) | ((t_uint64)(v))>>(((t_uint64)(64-(n)))))

typedef struct 
{
  t_bbl * bbl;
  int links_to_next_chain;
  int links_to_prev_chain;
  int links_to_all_prev_chains;
  int links_to_all_next_chains;
  int profile_info;
} layouted_chain;

static layouted_chain** layout;

static int number_of_layouted_chains;

static int ComputeLinks(int a, int b);
#define IS_LARGE_DATABBL(x)   (IS_DATABBL(x) && BBL_NINS(x)>200)
/*tagmerge*/
static void MergeTwoChains(t_bbl * head, t_bbl * tail)
{
  t_bbl * i_bbl;
  t_bbl * head_chain_first = BBL_FIRST_IN_CHAIN(head);
  t_bbl * tail_chain_first = BBL_FIRST_IN_CHAIN(tail);
  t_bbl * head_chain_last = BBL_LAST_IN_CHAIN(head);
  t_bbl * tail_chain_last = BBL_LAST_IN_CHAIN(tail);
  
  i_bbl = head_chain_first;


  if(head_chain_first == tail_chain_first)
    return;
  i_bbl = head_chain_first;


  while (i_bbl)
    {
      BBL_LAST_IN_CHAIN(i_bbl)=tail_chain_last;
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }

  
  i_bbl = tail_chain_first;

  while (i_bbl)
    {
      BBL_FIRST_IN_CHAIN(i_bbl)=head_chain_first;
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }

  BBL_NEXT_IN_CHAIN(head_chain_last)=tail_chain_first;
  BBL_PREV_IN_CHAIN(tail_chain_first)=head_chain_last;
}

static void UpdateBranchDisplacement(t_cfg_edge * edge)
{
  t_bbl * head = CFG_EDGE_HEAD(edge);
  t_bbl * tail = CFG_EDGE_TAIL(edge);
  t_ins * ins = BBL_INS_LAST(head);
  t_uint64 cnt=0;


  while(ins && IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_NOP)
  {
    cnt++;
    ins = INS_PREV(ins);
  }

  if (ins && 
      (Ia64IsIPRelativeBranchOrCall(ins) || Ia64IsCheck(ins) ) )
  {
    IA64INSIMM1(ins) = G_T_UINT64(BBL_CADDRESS(tail)) - (G_T_UINT64(BBL_CADDRESS(head)) + (((BBL_NINS(head))/3)<<4) - 16);
/*    InsPrint(ins, 0); */
    INS_ATTRIB(ins) |= IF_CHANGED;

  }
}

static void BepaalProfileInfo()
{
  int i;
  t_bbl * bbl;
  for (i=0;i<number_of_layouted_chains;i++)
  {
    layout[i]->profile_info=0;
    bbl = layout[i]->bbl;
    
    while(bbl)
    {
      layout[i]->profile_info+=BBL_PROFILE_INFO(bbl);
      bbl = BBL_NEXT_IN_CHAIN(bbl);
    }
  }
}
void PfgDetermineLayoutOrder(t_cfg * cfg, t_pfg * graaf)
{
  t_node * node;
  t_pfg_node * iter;
  t_bbl * leider;
  t_bbl * volger;
  t_bbl * i_bbl;
  int i=0;

  number_of_layouted_chains = 0;
  
  GRAPH_FOREACH_NODE(T_GRAPH(graaf),node)
  {
    if(!PFG_NODE_PREV(node))
    {
      leider = PFG_BBL(node);
      iter = PFG_NODE_NEXT(node);
      while(iter)
      {
	volger = PFG_BBL(iter);
	MergeTwoChains(leider,volger);
	iter = PFG_NODE_NEXT(iter);	
      }
    }
  }
  
  /*GRAPH_FOREACH_NODE(T_GRAPH(graaf),node){
    if(!PFG_NODE_PREV(node) &&(!BBL_FUN(PFG_BBL(node))||!BBL_IS_LAST(PFG_BBL(node))))
    {
      number_of_layouted_chains++;
    }
  }*/
    CFG_FOREACH_BBL(cfg,i_bbl)
    if (!BBL_PREV_IN_CHAIN(i_bbl) && (!BBL_FUN(i_bbl) || !BBL_IS_LAST(i_bbl)))
    {
      number_of_layouted_chains++;
    }

  layout = (layouted_chain **) Malloc (number_of_layouted_chains*sizeof(layouted_chain*));

  number_of_layouted_chains = 0;
/*  GRAPH_FOREACH_NODE(T_GRAPH(graaf),node)
  {
    if (!PFG_NODE_PREV(node) && (!BBL_FUN(PFG_BBL(node)) || !BBL_IS_LAST(PFG_BBL(node))))
    {
      layout[number_of_layouted_chains] = (layouted_chain*) Malloc (sizeof(layouted_chain));
      layout[number_of_layouted_chains]->links_to_all_prev_chains = 0;
      layout[number_of_layouted_chains]->links_to_all_next_chains = 0;
      layout[number_of_layouted_chains]->links_to_next_chain = 0;
      layout[number_of_layouted_chains]->links_to_prev_chain = 0;
      layout[number_of_layouted_chains]->bbl = PFG_BBL(node);
      PFG_BBL(node)->layout_number = number_of_layouted_chains;
      number_of_layouted_chains++;
    }
  }
  */
   CFG_FOREACH_BBL(cfg,i_bbl)
    if (!BBL_PREV_IN_CHAIN(i_bbl) && (!BBL_FUN(i_bbl) || !BBL_IS_LAST(i_bbl)))
    {
      layout[number_of_layouted_chains] = (layouted_chain*) Malloc (sizeof(layouted_chain));
      layout[number_of_layouted_chains]->links_to_all_prev_chains = 0;
      layout[number_of_layouted_chains]->links_to_all_next_chains = 0;
      layout[number_of_layouted_chains]->links_to_next_chain = 0;
      layout[number_of_layouted_chains]->links_to_prev_chain = 0;
      layout[number_of_layouted_chains]->bbl = i_bbl;
      i_bbl->layout_number = number_of_layouted_chains;
      number_of_layouted_chains++;
    }
 
  for (i=1;i<number_of_layouted_chains;i++)
    MergeTwoChains(layout[0]->bbl,layout[i]->bbl);
}
void CfgInitLayoutOrder(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_function * fun;
  t_bbl * j_bbl;
  /*t_ins * i_ins;*/
  /*int own_rank;*/
  /*int other_rank;*/
  /*t_bbl * own_head;*/
  /*t_bbl * other_head;*/
  
  number_of_layouted_chains = 0;

  /*
  *samennemen per functie
  * */
  CFG_FOREACH_FUN(cfg, fun)
  {
    i_bbl = FUN_BBL_FIRST(fun);
    j_bbl = BBL_NEXT_IN_FUN(i_bbl);
    while (j_bbl)
    {
      MergeTwoChains(i_bbl,j_bbl);
      j_bbl =BBL_NEXT_IN_FUN(j_bbl);
    }
  }

  CFG_FOREACH_BBL(cfg,i_bbl)
    if (!BBL_PREV_IN_CHAIN(i_bbl) && (!BBL_FUN(i_bbl) || FunctionGetExitBlock(BBL_FUN(i_bbl))!=i_bbl))
    {
      number_of_layouted_chains++;
    }

  layout = (layouted_chain**) Malloc (number_of_layouted_chains*sizeof(layouted_chain*));
  
  number_of_layouted_chains = 0;

  CFG_FOREACH_BBL(cfg,i_bbl)
    if (!BBL_PREV_IN_CHAIN(i_bbl) && (!BBL_FUN(i_bbl) || FunctionGetExitBlock(BBL_FUN(i_bbl))!=i_bbl))
    {
      layout[number_of_layouted_chains] = (layouted_chain*) Malloc (sizeof(layouted_chain));
      layout[number_of_layouted_chains]->links_to_all_prev_chains = 0;
      layout[number_of_layouted_chains]->links_to_all_next_chains = 0;
      layout[number_of_layouted_chains]->links_to_next_chain = 0;
      layout[number_of_layouted_chains]->links_to_prev_chain = 0;
      layout[number_of_layouted_chains]->bbl = i_bbl;
      i_bbl->layout_number = number_of_layouted_chains;
      number_of_layouted_chains++;
    }
}

static void CfgDetermineLayoutOrder(t_cfg * cfg)
{ 
  //t_bool change = FALSE;
  int total_distance = 0;
  int i;
 //  int j,k;
  //t_bbl * i_bbl;

  CfgInitLayoutOrder(cfg);
/*  BepaalProfileInfo();

  if (global_options.debugcounter == 0) global_options.debugcounter=5;

  j=0;

  for (i=0; i<number_of_layouted_chains; i++) 
  {
    if(layout[j]->profile_info > 306672 && layout[j]->profile_info <= 766728)
    {
      for (k=j;k<number_of_layouted_chains-1;k++)
        Ia64Swap(k,k+1);
    }
    else j++;
  }
*/
  /*
     <<<<<<< ia64_layout.c

     if (global_options.debugcounter == 0) global_options.debugcounter=5;

     j=0;
     for (i=0; i<number_of_layouted_chains; i++) 
     {
     if(layout[j]->profile_info > 0 && layout[j]->profile_info <= 1000)
     =======

     if (global_options.debugcounter == 0) global_options.debugcounter=5;

     j=0;
     for (i=0; i<number_of_layouted_chains; i++) 
     {
     if(layout[j]->profile_info > 0 && layout[j]->profile_info <= global_options.debugcounter)
     >>>>>>> 1.44
     {
     <<<<<<< ia64_layout.c
     for (k=j;k<number_of_layouted_chains-1;k++)
     Ia64Swap(k,k+1);
     }
     else j++;
     }

     j=0;
     for (i=0; i<number_of_layouted_chains; i++) 
     {
     if(layout[j]->profile_info ==0)
     {
     for (k=j;k<number_of_layouted_chains-1;k++)
     Ia64Swap(k,k+1);
     =======
     for (k=j;k<number_of_layouted_chains-1;k++)
     Ia64Swap(k,k+1);
     }
     else j++;
     }
     j=0;
     for (i=0; i<number_of_layouted_chains; i++) 
     {
     if(layout[j]->profile_info ==0)
     {
     for (k=j;k<number_of_layouted_chains-1;k++)
     Ia64Swap(k,k+1);
     >>>>>>> 1.44
     }
     else j++;
     }

  //    for (j=0; j<number_of_layouted_chains-1-i; j++)
   */

  /*
  //sorteer binnen functie
  if(layout[j+1]->bbl->fun == layout[j]->bbl->fun && layout[j+1]->profile_info > layout[j]->profile_info)
  Ia64Swap(j,j+1);         
   */

  //zet niet gebruikte kettingen achteraan
  //     if (layout[i]->profile_info ==0)  
  //	for(j=i;j<number_of_layouted_chains-1;j++)
  //	  Ia64Swap(j,j+1);        
  //      if(layout[i]->profile_info == 0) i--;

  /*    
  //sorteer alles
  if (layout[j+1]->profile_info > layout[j]->profile_info)  
  Ia64Swap(j,j+1);       
   */
  //  }
  //zet alles achteraan dat niet meer dan x keer wordt opgeroepen
  //ev. i.s.m nul blokken helemaan achteraan
  //  if (global_options.debugcounter == 0) global_options.debugcounter=5;
  //printf("%d",global_options.debugcounter);FATAL(("ik gebruik hier debugcounter"));
  /*for (i=0; i<number_of_layouted_chains-1; i++) 
    {
    for (j=i; j<number_of_layouted_chains-1//-i//; j++)
    {
    if (layout[j]->profile_info <= global_options.debugcounter && //als lege helemaal achteraan//layout[j+1]->profile_info!=0)  
    Ia64Swap(j,j+1); 
    }
    }*/
  /*  for (i=0; i<number_of_layouted_chains; i++)
      {
      printf("%d\t%llx\n",layout[i]->profile_info,layout[i]->bbl->fun);
      }
      FATAL(("stop"));
   */

  
  for (i=1;i<number_of_layouted_chains;i++)
    MergeTwoChains(layout[0]->bbl,layout[i]->bbl);

  printf("final total distance %d\n",total_distance);

}

/* dit is de eigenlijke lay-out procedure */

t_uint32 Ia64CfgLayout(t_object * obj, t_section *code, t_cfg * cfg)
{
  t_bbl * i_bbl, * j_bbl;
  t_cfg_edge * i_edge;
  /*t_ins * i_ins, * j_ins, * tmp_ins;*/
  t_uint64 address;
  /*int count_old, count_new;*/
  /*t_bool changed;*/
  t_uint32 size;
  //t_uint32 count=0;
  //int i=0,j=0;
  /*berre*/
  FILE * fp;
  t_pfg * graaf;
  int k=0;
  //t_node * node;
  int debugteller=0;
  /*einde berre*/

  if(global_options.placement)
    graaf = Ia64InitializeGraph(cfg);
//  MeetAantalBblsPerCluster(cfg);
  /* put each block in separate chain */

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    //      printf("HOLALALA\n");
    //      BblPrint(i_bbl,PRINT_INSTRUCTIONS);
    BBL_NEXT_IN_CHAIN(i_bbl) = BBL_PREV_IN_CHAIN(i_bbl) = NULL;
    BBL_FIRST_IN_CHAIN(i_bbl) = BBL_LAST_IN_CHAIN(i_bbl) = i_bbl;
  }

  /* check for more then one incoming fall_through path */

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
//    printf("in: %d\n", debugteller++);
    j_bbl = NULL;
    BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
    {
      if (EDGE_CAT(i_edge)==ET_FUNLINK ||
          EDGE_CAT(i_edge)==ET_SWILINK ||
          EDGE_CAT(i_edge)==ET_IPFUNLINK ||
          EDGE_CAT(i_edge)==ET_IPSWILINK ||
          EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
          EDGE_CAT(i_edge)==ET_IPFALLTHRU
         )
      {
        if (j_bbl)
          if (j_bbl!=CFG_EDGE_HEAD(i_edge))
            FATAL(("two succ edges\n"));

        j_bbl = CFG_EDGE_HEAD(i_edge);
      }
    }
  }

  /* check for more then one outgoing fall_through path */
  /* also combine blocks involved in fall_through paths in one chain */

  debugteller=0;
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
//    printf("out: %d\n",debugteller++);
    j_bbl = NULL;
    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      if (EDGE_CAT(i_edge)==ET_FUNLINK ||
          EDGE_CAT(i_edge)==ET_SWILINK ||
          EDGE_CAT(i_edge)==ET_IPFUNLINK ||
          EDGE_CAT(i_edge)==ET_IPSWILINK ||
          EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
          EDGE_CAT(i_edge)==ET_IPFALLTHRU
         )
      {
        if (j_bbl)
          if (j_bbl!=CFG_EDGE_TAIL(i_edge))
	  {
	    BblPrint(i_bbl,2);
            FATAL(("two succ edges\n"));
	  }

        j_bbl = CFG_EDGE_TAIL(i_edge);
      }
    }
    if (j_bbl)
    {
      if(global_options.placement)
	PfgMergeByBbl(graaf, i_bbl, j_bbl);
      MergeTwoChains(i_bbl,j_bbl);
    }
  }
  //Ia64PrintPrograph("test.dot",graaf);
  //FATAL(("stopstop"));
  /*stop gisteren...*/
/*  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if(IA64INSTYPE(BBL_NEXT(i_bbl)) == IA64_TYPE_DATA)
    {
      	MergeTwoChains(i_bbl,BBL_NEXT(i_bbl));
    }
  }
*/

//  printf("ten\n");
  if(global_options.placement)
    PfgClosestIsBest(graaf);
//  printf("ten\n");
  /* now the order in which chains will be linked into 1 big chain is
     determined, and the chains are merged */

  if(global_options.placement)
    PfgDetermineLayoutOrder(cfg, graaf);
  else 
    CfgDetermineLayoutOrder(cfg);

  CFG_FOREACH_BBL(cfg,i_bbl){
    if((BBL_NEXT_IN_CHAIN(i_bbl)==BBL_PREV_IN_CHAIN(i_bbl))){
      DiabloPrint(stdout,"%p\n",BBL_NEXT_IN_CHAIN(i_bbl));
      DiabloPrint(stdout,"@iB\n",i_bbl);
      FATAL(("hey"));
    }
  }
  

  
#ifdef IA64_NEW_BBL_ADDRESS
  Ia64ScheduleCfg(code, cfg);
#endif


#ifdef IA64_NEW_BBL_ADDRESS

  i_bbl = T_BBL(GRAPH_NODE_FIRST(cfg));//cfg->entries->entry_bbl;

  /*we zullen de code na alle constante data gaan plaatsen */
  {
    t_object * obj = SECTION_OBJECT(code);
    t_uint64 maxAddress = 0, addend = 0;
    t_int32 i;

    for(i = 0; i<obj->ncodes; i++)
    {
      t_uint64 tmpAddress = SECTION_GET_CADDRESS(obj->code[i]) + SECTION_CSIZE(obj->code[i]);

      maxAddress = MAX(maxAddress, tmpAddress);
    } 

    for(i = 0; i<obj->nrodatas; i++)
    {
      t_uint64 tmpAddress = SECTION_GET_CADDRESS(obj->rodata[i]) + SECTION_CSIZE(obj->rodata[i]);

      maxAddress = MAX(maxAddress, tmpAddress);
    } 

    maxAddress = (maxAddress<<4)>>4; /*cut off region bits*/
    i = 0;
    while((maxAddress & 0x8000000000000000ULL) == 0)
    {
      i++;
      maxAddress<<=1;
    }

    i = 63-i;
    i++; /*volgende bit nemen*/

    addend = 1ULL<<i;  
    address = G_T_UINT64(BBL_CADDRESS(i_bbl)) + addend;

    SectionSetCAddressSuper(code,SECTION_GET_CADDRESS(code)+addend);

    /*        
              address = 0xd00000+G_T_UINT64(BBL_OLD_ADDRESS(i_bbl));
              SECTION_GET_CADDRESS(code)+=0xd00000;
     */    

  }

  if(global_options.makeprofile)
    fp = fopen("nieuwoud","w+");
  
  i_bbl=layout[0]->bbl;
  while (i_bbl)
  //CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_ins *tmpins;
    int cnt = 0;
    t_uint8 slot = 0;
    t_address insaddress;

    k++;
    GenericSet_uint64(BBL_CADDRESS(i_bbl),address);
//    printf("%llx\t%llx\n",BBL_CADDRESS(i_bbl) ,BBL_OLD_ADDRESS(i_bbl));
    if(global_options.makeprofile)
      fprintf(fp,"%llx\t%llx\n",BBL_CADDRESS(i_bbl) ,BBL_OLD_ADDRESS(i_bbl));
    if(GENERIC_EQ(BBL_OLD_ADDRESS(cfg->entries->entry_bbl),BBL_OLD_ADDRESS(i_bbl) ))
      obj->entry = BBL_CADDRESS(i_bbl);

    insaddress = BBL_CADDRESS(i_bbl);
    
    BBL_FOREACH_INS(i_bbl, tmpins)
    {
      INS_CADDRESS(tmpins) = insaddress;
//      DiabloPrint(stdout, "@I\n", tmpins);
      
      cnt++;
      if(slot == 2)
      {
	slot = 0;
	insaddress = GENERIC_ADD_UINT64(insaddress, 14);
      }
      else
      {
	slot++;
	insaddress = GENERIC_ADD_UINT64(insaddress, 1);
      }
    }
    //TODO: er klopt blijkbaar nog iets niet..

    if(cnt != BBL_NINS(i_bbl))
    {
      printf("Warning: BBL_NINS(%d) does not match with actual number of instructions(%d) !!\n", BBL_NINS(i_bbl), cnt);
      BblPrint(i_bbl, 0);
    }

    /* BBL_NINS zou altijd veelvoud van drie moeten zijn */
    address+=(cnt/3)<<4;
    //address+=(BBL_NINS(i_bbl)/3)<<4;
   i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
  }
//  printf("->%d\n",k);
//  FATAL(("stop"));
  if(global_options.makeprofile)
    fclose(fp);


  /*check if all blls have been given a new address*/
/*  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if(GENERIC_EQ(BBL_OLD_ADDRESS(i_bbl),BBL_CADDRESS(i_bbl)))
      FATAL(("no new address for bbl @B\n", i_bbl));
  }*/
  
  printf("Given new addresses to the Basic Blocks...\n");
#else

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_CADDRESS(i_bbl)= BBL_OLD_ADDRESS(i_bbl);
  }

#endif

#ifdef IA64_NEW_BBL_ADDRESS
  /* adapt the displacements encoded in direct control flow transfer instructions */
//#if 0 
  /*the relocations should take care of that from now on !!!*/ 
  CFG_FOREACH_EDGE(cfg,i_edge)
    if (EDGE_CAT(i_edge)==ET_JUMP || EDGE_CAT(i_edge)==ET_IPJUMP || EDGE_CAT(i_edge)==ET_CALL)
      UpdateBranchDisplacement(i_edge); /*possibly adaptation needed ...*/
//#endif
#endif

  /*  exit(0);*/
  size = 0;
  //  printf("bijna op einde"); 

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_ins *tmpins;
    int cnt = 0;
    BBL_FOREACH_INS(i_bbl, tmpins)
      cnt++;
    //TODO: er klopt blijkbaar nog iets niet..
    size += (cnt/3)<<4;
    if(cnt != BBL_NINS(i_bbl))
    {
      printf("Warning: BBL_NINS(%d) does not match with actual number of instructions(%d) !!\n", BBL_NINS(i_bbl), cnt);
      BblPrint(i_bbl, 0);
#if 0
      {
        t_ins *tmp_ins = i_bbl->ins_first;

        while(tmp_ins)
        {
          InsPrint(tmp_ins, 0);
          tmp_ins = INS_NEXT(tmp_ins);
        }
        printf("\n");
      }
#endif
    }
    //size+=(BBL_NINS(i_bbl)/3)<<4;
  }
//  printf("%d total number of 4-byte words of the code section\n",size/4);

#ifdef IA64_NEW_BBL_ADDRESS
  SECTION_NINS(code) = (size>>4)*3;
  SECTION_CSIZE(code) = size;
  return size;
#else
  SECTION_NINS(code) = (GENERIC_EXTRACT_UINT32(SECTION_CSIZE(code))>>4)*3;
  return GENERIC_EXTRACT_UINT32(SECTION_CSIZE(code));
#endif
}

static t_bool Ia64Swap(int a, int b)
{
  layouted_chain * x = layout[a], * y = layout[b];
  layouted_chain * new_y;
  layouted_chain * new_x; 

  new_y = (layouted_chain*) Malloc(sizeof(layouted_chain));
  new_x = (layouted_chain*) Malloc(sizeof(layouted_chain));

  new_y->bbl = x->bbl;
  new_x->bbl = y->bbl;
  
  new_y->links_to_prev_chain  = x->links_to_next_chain;
  new_y->links_to_all_prev_chains            = x->links_to_all_prev_chains + x->links_to_next_chain;
  new_y->links_to_all_next_chains           = x->links_to_all_next_chains - x->links_to_next_chain;
  new_y->links_to_next_chain = ComputeLinks(a,b+1);
  new_y->profile_info = x->profile_info;
  
  new_x->links_to_next_chain = y->links_to_prev_chain;
  new_x->links_to_all_next_chains           = y->links_to_all_next_chains + y->links_to_prev_chain;
  new_x->links_to_all_prev_chains            = y->links_to_all_prev_chains - y->links_to_prev_chain;
  new_x->links_to_prev_chain  = ComputeLinks(a-1,b);
  new_x->profile_info = y->profile_info;
  
  if (a>0)
    layout[a-1]->links_to_next_chain = new_x->links_to_prev_chain;

  if (b<number_of_layouted_chains-1)
    layout[b+1]->links_to_prev_chain = new_y->links_to_next_chain;
  
  new_y->bbl->layout_number = b;
  new_x->bbl->layout_number = a;
  
  Free(x);
  Free(y);

  layout[a] = new_x; 
  layout[b] = new_y;

  return TRUE;
}

static int ComputeLinks(int a, int b)
{ 
 return 0;
/* 
  t_ins * i_ins;
  t_bbl * bbl_a, * bbl_b;

  int count = 0;

  if (a<0 || b == number_of_layouted_chains)
    return 0;

  bbl_a = BBL_FIRST_IN_CHAIN(layout[a]->bbl);
  bbl_b = BBL_FIRST_IN_CHAIN(layout[b]->bbl);

  while (bbl_a)
    {
      BBL_FOREACH_INS(bbl_a,i_ins)
	{
	  if (ARMINSOPCODE(i_ins) == ARM_ADDRESS_PRODUCER)
	    {
	      t_reloc * reloc = INS_REFERS_TO(i_ins)->rel;
	      if (reloc->to->ftype == TO_BBL)
		if (BBL_FIRST_IN_CHAIN(reloc->to->base) == bbl_b)
		  count++;
	    }
	}
      bbl_a = BBL_NEXT_IN_CHAIN(bbl_a);
    }
  bbl_a = BBL_FIRST_IN_CHAIN(layout[a]->bbl);

  while (bbl_b)
    {
      BBL_FOREACH_INS(bbl_b,i_ins)
	{
	  if (ARMINSOPCODE(i_ins) == ARM_ADDRESS_PRODUCER)
	    {
	      t_reloc * reloc = INS_REFERS_TO(i_ins)->rel;
	      if (reloc->to->ftype == TO_BBL)
		if (BBL_FIRST_IN_CHAIN(reloc->to->base) == bbl_a)
		  count++;
	    }
	}
      bbl_b = BBL_NEXT_IN_CHAIN(bbl_b);
    }
  return count;*/
}

static void PrintChain(t_bbl * bbl,t_uint32 flags)
{
  t_bbl * i_bbl;
  t_ins * ins;
  
  printf("---------------------- CHAIN ---------------------------\n ");
  //BblPrint(bbl,0);
  //BBL_FOREACH_INS(bbl,ins)
  //{
  // InsPrint(ins,0);
  //}

  i_bbl = BBL_FIRST_IN_CHAIN(bbl);
  while (i_bbl)
    {
      printf("F %4llx L %4llx P %4llx ",
       G_T_UINT64(BBL_OLD_ADDRESS(BBL_FIRST_IN_CHAIN(i_bbl))),
       G_T_UINT64(BBL_OLD_ADDRESS(BBL_LAST_IN_CHAIN(i_bbl))),
       BBL_PREV_IN_CHAIN(i_bbl)?G_T_UINT64(BBL_OLD_ADDRESS(BBL_PREV_IN_CHAIN(i_bbl))):0x0ULL
       );
      fflush(stdout);
      BblPrint(i_bbl,flags);
      
      BBL_FOREACH_INS(i_bbl,ins)
      {
	FATAL(("Implement"));
	//InsPrint(ins,0);
      }
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }
}

/*
static t_bool Swap(int a, int b)
{

  int gain;

  layouted_chain * x = layout[a], * y = layout[b];
  layouted_chain * new_y;
  layouted_chain * new_x; 
 
  gain = x->links_to_all_next_chains + y->links_to_all_prev_chains - (x->links_to_all_prev_chains + y->links_to_all_next_chains) - (x->links_to_next_chain+y->links_to_prev_chain) * (1 + IS_LARGE_DATABBL(y->bbl)?1:0 - IS_LARGE_DATABBL(x->bbl)?1:0);
  
  if (gain<=0)
    return FALSE;

  new_y = (layouted_chain*) Malloc(sizeof(layouted_chain));
  new_x = (layouted_chain*) Malloc(sizeof(layouted_chain));

  new_y->bbl = x->bbl;
  new_x->bbl = y->bbl;
  
  new_y->links_to_prev_chain  = x->links_to_next_chain;
  new_y->links_to_all_prev_chains            = x->links_to_all_prev_chains + x->links_to_next_chain;
  new_y->links_to_all_next_chains           = x->links_to_all_next_chains - x->links_to_next_chain;
  new_y->links_to_next_chain = ComputeLinks(a,b+1);
  
  new_x->links_to_next_chain = y->links_to_prev_chain;
  new_x->links_to_all_next_chains           = y->links_to_all_next_chains + y->links_to_prev_chain;
  new_x->links_to_all_prev_chains            = y->links_to_all_prev_chains - y->links_to_prev_chain;
  new_x->links_to_prev_chain  = ComputeLinks(a-1,b);
  
  if (a>0)
    layout[a-1]->links_to_next_chain = new_x->links_to_prev_chain;

  if (b<number_of_layouted_chains-1)
    layout[b+1]->links_to_prev_chain = new_y->links_to_next_chain;
  
  new_y->bbl->layout_number = b;
  new_x->bbl->layout_number = a;
  
  Free(x);
  Free(y);

  layout[a] = new_x; 
  layout[b] = new_y;

  return TRUE;
}
*/
#if 0

static void InitializePossibleTemplatesList(t_bool *possibleTemplates)
{
  t_uint8 i;
  
  for(i=0; i<16; i++)
    if(i == 0x03 || i == 0x0A || i == 0x0D || i == 0x0F)
      possibleTemplates[i] = 0;
    else
      possibleTemplates[i] = 1;
}



static void Ia64InsertInstruction(t_ins * ins, t_ins * new_ins)
{
  t_ins * next = INS_NEXT(ins);
  
  if(next == NULL)
  {
    INS_NEXT(ins) = new_ins;
    INS_PREV(new_ins) = ins;
  }
  else
  {
    INS_NEXT(ins) = new_ins;
    INS_NEXT(new_ins) = next;

    INS_PREV(next) = new_ins;
    INS_PREV(new_ins) = ins;
        
  }
  
}

static t_ins * Ia64InsertNops(t_section *code, t_bbl * bbl, t_ins * lastFromPrevBundle, 
                              t_bool *mustBeNop, t_uint8 templateNr)
{
  t_uint8 slot;
  t_ins * curr;
  t_ins * nop;
  
  static int c=0;
  curr = lastFromPrevBundle;

//  printf("slot0:%d, slot1: %d,slot2: %d\n",templateField[templateNr][0],templateField[templateNr][1],templateField[templateNr][2]);

  /* debugging
  if(curr && INS_CADDRESS(INS_NEXT(curr)) == 0x4000000000001240ULL)
    *(int *)0 = 0;
  */
  
  for(slot = 0; slot < 3; slot++)
  {
    if(!mustBeNop[slot])
    {
      if(slot == 0 && curr == NULL)
        curr = bbl->ins_first;
      else
        //if(slot != 2)
          curr = INS_NEXT(curr);
      continue;
    }
    
    if(slot == 0)
    { /*\TODO ook nog de dingen (zoals BBL-address, is het een BBL-leader, ...) in generic juist zetten !!!*/

      printf("CREATE %d\n",c++);
      nop = Ia64InstructionCreateNoop(code, templateField[templateNr][slot]);
    }
    else
    {
      printf("CREATE %d\n",c++);
      nop = Ia64InstructionCreateNoop(code, templateField[templateNr][slot]);
    }

    INS_BBL(nop) = bbl;   
    BBL_OLD_ADDRESS(nop)=BBL_OLD_ADDRESS(bbl);
    //BBL_OLD_ADDRESS(nop)=(curr==NULL?BBL_OLD_ADDRESS(bbl->ins_first):BBL_OLD_ADDRESS(curr));
	    
    
    bbl->nins++;
    
    if(slot == 0 && curr == NULL)
    {
      if(bbl->ins_first)
      {
	INS_PREV(bbl->ins_first) = nop;
	INS_NEXT(nop) = bbl->ins_first;

	/* is nop; mag zeker in dezelfde instr groep zitten */
	IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(bbl->ins_first);
      }
      else
      {
	IA64INSINSTRGROUP(nop) = GetCurrentInstrGroupAndInc();
	bbl->ins_last = nop;
      }
      curr = bbl->ins_first = nop;
    }
    else
      if(slot == 2 && INS_NEXT(curr) == NULL)
      {
        INS_NEXT(bbl->ins_last) = nop;
        INS_PREV(nop) = bbl->ins_last;

        /* is nop; mag zeker in dezelfde instr groep zitten */
        IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(bbl->ins_last);

        curr = bbl->ins_last = nop;
      }
      else
      {
        /* is nop; mag zeker in dezelfde instr groep zitten :wel de juiste kiezen... */
	if(slot == 0 && INS_NEXT(curr))
	  IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(INS_NEXT(curr)); //deze van dezelfde bundel nemen
	else
        IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(curr);
        
        Ia64InsertInstruction(curr, nop);
	
	if(bbl->ins_last == curr)
	  bbl->ins_last=nop;
	
	curr = nop;
      }
        
//    if(slot != 2)
//      curr = INS_NEXT(curr);
        
  }
  
  return curr;
}

static t_bool StopMatch(t_uint8 slot, t_ins * ins, t_uint8 templateNr)
{
//  t_ins * next_ins=INS_NEXT(ins);
  t_bool mustBeStopped;
  
  /* je hebt hiervoor toch altijd de twee opties */
  if(slot == 2)
    return 1;
    
  mustBeStopped = 
    INS_NEXT(ins) == NULL || (IA64INSINSTRGROUP(ins) != IA64INSINSTRGROUP(INS_NEXT(ins)));
    
  if(mustBeStopped)
  {
    if((slot == 1 && templateNr == 0x02) || (slot == 0 && templateNr == 0x0A))
      return 1;
    else
      return 0;
  }
      
  return 1;
}

static t_bool InstrTypeMatch(t_ia64_instr_type instr_type, t_ia64_unit unit)
{
  switch(instr_type)
  {
    case IA64_TYPE_A:
      return unit == IA64_UNIT_I || unit == IA64_UNIT_M;
    case IA64_TYPE_I:
      return unit == IA64_UNIT_I;
    case IA64_TYPE_M:
      return unit == IA64_UNIT_M;
    case IA64_TYPE_B:
      return unit == IA64_UNIT_B;
    case IA64_TYPE_F:
      return unit == IA64_UNIT_F;
    case IA64_TYPE_DUMMY:
      return unit == IA64_UNIT_L;
    case IA64_TYPE_X:
      return unit == IA64_UNIT_X;
    case IA64_TYPE_DATA:
      return 1;
    default:
//      printf("%d\n",instr_type);
      FATAL(("Wrong instruction match !!!"));
      /* Keep the compiler happy*/
      return 0;
  }
  
}

static void Ia64BBLBundeling(t_section *code, t_bbl * bbl)
{
  t_ins * ins;
  t_ins * lastFromPrevBundle = NULL;
  t_bool possibleTemplates[16];
  t_bool mustBeNop[3];
  t_uint8 slot = 0, i;
  t_bool isMatch;
  t_uint32 fullNOPBundle = 0;
 
  /* standaard moet alles nopkes zijn... */
  mustBeNop[0] = mustBeNop[1] = mustBeNop[2] = 1;
  
  InitializePossibleTemplatesList(possibleTemplates);
/*
  {
    t_ins *tmp_ins = bbl->ins_first;

    while(tmp_ins)
    {
      InsPrint(tmp_ins, 0);
      tmp_ins = INS_NEXT(tmp_ins);
    }

*/
  
  for(ins=bbl->ins_first; ins!=NULL;)  
  {
    isMatch = 0;

    /* first check if there is a dummy without a X-type instruction: this is possible if the X-type instruction was removed*/
    if(IA64INSTYPE(ins) == IA64_TYPE_DUMMY && (INS_NEXT(ins) == NULL || IA64INSTYPE(INS_NEXT(ins)) != IA64_TYPE_DUMMY  ))
    {
      t_ins * next = INS_NEXT(ins);
      InsKill(ins);
      ins = next; 
    }

    if(ins != NULL)
    {

      for(i = 0; i<16; i++)
      {
	if(possibleTemplates[i] && InstrTypeMatch(IA64INSTYPE(ins), templateField[i*2][slot]) 
	    && StopMatch(slot, ins, i*2))
	{
	  isMatch = 1;
	  break;
	}
      }   

      /*overrule...*/

      if(slot != 2 && (IA64OPCODEINFO(ins).flags & IA64_OPCODE_SLOT2))
	isMatch = 0;

      if(isMatch)
      { 
	for(i = 0; i<16; i++)
	  if(!(possibleTemplates[i] && InstrTypeMatch(IA64INSTYPE(ins), templateField[i*2][slot])
		&& StopMatch(slot, ins, i*2)))
	    possibleTemplates[i] = 0;
	ins=INS_NEXT(ins);

	/* we hebben een instructie die erin past => moet geen nop meer in */
	mustBeNop[slot] = 0;
      }

    }

    slot++;
    /* ook kijken of we aan het einde van de instructies gekomen zijn !!!*/
    if(slot == 3 || ins == NULL)
    {
      for(i = 0; i<16; i++)
	if(possibleTemplates[i])
	  break;

      if(i == 16)
	FATAL(("Something wrong with bundeling !!!"));

      if(mustBeNop[0] && mustBeNop[1] && mustBeNop[2])
      {
	fullNOPBundle++;
	if(fullNOPBundle >= 128)
	{
	  printf("Something wrong with bundeling: encountered a bundle full of nops ?!!\n");
	  if(ins == NULL)
	    printf("Current instruction was NULL [usually meaning end of BBL]\n");
	  else
	  {
	    printf("Current instruction was:\n");
	    InsPrint(ins, 0);
	  }
	  FATAL(("Please Correct !"));
	}

      }
      else
	fullNOPBundle = 0;

      //      printf("inserting nops\n");  
      lastFromPrevBundle = Ia64InsertNops(code, bbl, lastFromPrevBundle, mustBeNop, i*2);


      mustBeNop[0] = mustBeNop[1] = mustBeNop[2] = 1;
      slot = 0; 

      InitializePossibleTemplatesList(possibleTemplates);
    }
  }
}

#endif


#if 0
static void UpdateBranchDisplacement(t_cfg_edge * edge)
{
  t_bbl * head = CFG_EDGE_HEAD(edge);
  t_bbl * tail = CFG_EDGE_TAIL(edge);
  t_ins * ins = BBL_INS_LAST(head);

//  DiabloPrint(stdout,"@G @G\n",BBL_CADDRESS(tail),BBL_CADDRESS(head));

//  IA64INSIMM1(ins) = G_T_UINT64(BBL_CADDRESS(tail)) - (G_T_UINT64(INS_CADDRESS(ins)) & 0xFFFFFFFFFFFFFFF0ULL);

  IA64INSIMM1(ins) = G_T_UINT64(BBL_CADDRESS(tail)) - (G_T_UINT64(BBL_CADDRESS(head)) + ((BBL_NINS(head)/3)<<4) - 16);
//  printf("%llx\n",  IA64INSIMM1(ins));
  INS_ATTRIB(ins) |= IF_CHANGED;
}

#endif




#if 0
static int CfgTotalDistance(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_ins * i_ins;
  int i = 0;
  int own_rank;
  int other_rank;
  t_bbl * own_head;
  t_bbl * other_head;

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    own_head = BBL_FIRST_IN_CHAIN(i_bbl);
    own_rank = own_head->layout_number;

    i+=number_of_layouted_chains - own_rank;
  }
  return i;
}
#endif


/* vim: set shiftwidth=2: */
