/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include "ia64_liveness.h" 
#include "../../utils/util_error.h"
#include "../../kernel/diablo_flowgraph.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_iterators.h"
#include "ia64_disassemble.h"
#include "ia64_instructions.h"
#include "ia64_opcodes.h"
#include "ia64_registers.h"
#include "../../utils/util_malloc.h"

#include "ia64_dummy.h"


/*nodig, omdat de dummies anders in de war geraken... als een alloc door inlining zopas uitgebreid werd */
void Ia64ResetLivenessDummies(t_cfg * cfg)
{
  t_edge * edge, * funlink_edge;
  t_bbl * funlink_bbl;
  t_bbl * head;
  t_bbl * tail;
  t_bbl * i_bbl;

  t_ins * ins_dummy1;
  t_ins * ins_dummy2;
  t_ins * tmp_ins;

  t_ins * probably_alloc;
  t_ins * ins_ft;

  CFG_FOREACH_EDGE(cfg, T_CFG_EDGE(edge))             
  {
    if(EDGE_CAT(edge)&ET_CALL)
    {
      head=CFG_EDGE_HEAD(edge);
      tail=CFG_EDGE_TAIL(edge);

      BBL_FOREACH_SUCC_EDGE(head, T_CFG_EDGE(funlink_edge))
      {
	if(EDGE_CAT(funlink_edge)&ET_FUNLINK)
	  funlink_bbl=CFG_EDGE_TAIL(funlink_edge);
      }
      
      if(head->ins_last)
      {
	int i;
	probably_alloc = Ia64SearchForAlloc(head->fun);

	if(probably_alloc == NULL)
	  probably_alloc = head->fun->bbl_first->ins_first;

	
	ins_dummy1 = INS_PREV(head->ins_last);
	if(! ins_dummy1 || !IA64ISLIVEDUMMY(ins_dummy1))
	  FATAL(("NO DUMMY 1"));
	ins_dummy2 = funlink_bbl->ins_first;

	/*no dummy 2 is possible if the called function does not return and so the blocks after the
	* call are removed. */
	if(!ins_dummy2)
	  printf("WARNING: NO DUMMY 2 => see source file !\n");
	else
	  if(!IA64ISLIVEDUMMY(ins_dummy2))
	    FATAL(("NO DUMMY 2"));

	
	RegsetSetEmpty(INS_REGS_USE(ins_dummy1));
	for(i=0;i<96;i++)
	  RegsetAddReg(INS_REGS_USE(ins_dummy1),IA64_R32+i);	  

	
	if(ins_dummy2)
	{

	  if(IA64OPCODEINFO(probably_alloc).instr_opcode == IA64_INSTR_ALLOC)
	  {
	    int i;
	    RegsetSetEmpty(INS_REGS_USE(ins_dummy2));
	    for(i=0;(i<IA64INSIMM2(probably_alloc)-IA64INSIMM1(probably_alloc));i++)
	      RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32+i);	  
	  }
	  else
	  {
	    int i;
	    RegsetSetEmpty(INS_REGS_USE(ins_dummy2));
	    for(i=0;i<96;i++)
	      RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32+i);	  
	  }
	}
      }
    }
  }                                 
       
}

void Ia64UpdateLiveness(t_cfg * cfg)
{  
  t_edge * edge, * funlink_edge;
  t_bbl * head;
  t_bbl * tail;
  t_bbl * funlink_bbl;
  t_bbl * i_bbl;

  t_ins * ins_dummy1;
  t_ins * ins_dummy2;
  t_ins * tmp_ins;

  t_ins * probably_alloc;
  t_ins * ins_ft;

  CFG_FOREACH_EDGE(cfg, T_CFG_EDGE(edge))             
  {
    if(EDGE_CAT(edge)&ET_CALL)
    {
      head=CFG_EDGE_HEAD(edge);
      tail=CFG_EDGE_TAIL(edge);

      BBL_FOREACH_SUCC_EDGE(head, T_CFG_EDGE(funlink_edge))
      {
	if(EDGE_CAT(funlink_edge)&ET_FUNLINK)
	  funlink_bbl=CFG_EDGE_TAIL(funlink_edge);
      }

      if(head->ins_last)
      {
	probably_alloc = Ia64SearchForAlloc(head->fun);

	if(CountAllocsInFun(head->fun)>1)
	{
	  //printf("WARNING: cannot update dummies because of multiple allocs !\n");
	  continue;
	}

	if(probably_alloc == NULL || !FUN_ISALLOCCLEAN(head->fun))
	  probably_alloc = head->fun->bbl_first->ins_first;
	
	ins_dummy1 = INS_PREV(head->ins_last);
	if(! ins_dummy1 || !IA64ISLIVEDUMMY(ins_dummy1))
	  FATAL(("NO DUMMY 1"));
	ins_dummy2 = funlink_bbl->ins_first;

	/*no dummy 2 is possible if the called function does not return and so the blocks after the
	* call are removed. */
	if(!ins_dummy2)
	{
	  //printf("WARNING: NO DUMMY 2 => see source file !\n");
	}
	else
	  if(!IA64ISLIVEDUMMY(ins_dummy2))
	    FATAL(("NO DUMMY 2"));

	if(IA64OPCODEINFO(probably_alloc).instr_opcode == IA64_INSTR_ALLOC)
	{
	  int i;
	  int reg;
	  t_regset liveafter,livebefore;

	  if(IA64INSIMM3(probably_alloc))
	  {  
	    //InsPrint(probably_alloc,0);
	    FATAL(("Rotating registers"));
	  }
	  /* definen */
	  for(i=0;i<IA64INSIMM2(probably_alloc)-IA64INSIMM1(probably_alloc);i++)
	  {
	    RegsetAddReg(INS_REGS_DEF(ins_dummy1),IA64_R32+i);
	    RegsetAddReg(INS_REGS_DEF(ins_dummy1),IA64_R32_NAT+i);
	  }

	  if(ins_dummy2)
	    for(i=0;i<IA64INSIMM2(probably_alloc);i++)
	    {
	      RegsetAddReg(INS_REGS_DEF(ins_dummy2),IA64_R32+i);
	      RegsetAddReg(INS_REGS_DEF(ins_dummy2),IA64_R32_NAT+i);
	    }

	  /* dummy 1 updaten */
	  
	  RegsetSetEmpty(INS_REGS_USE(ins_dummy1));

	  if(ins_dummy2)
	    for(i=0;i<IA64INSIMM1(probably_alloc);i++)
	    {
	      liveafter = InsRegsLiveAfter(ins_dummy2);
	      if(RegsetIn(liveafter,IA64_R32+i))
		RegsetAddReg(INS_REGS_USE(ins_dummy1),IA64_R32+i);
	      if(RegsetIn(liveafter,IA64_R32_NAT+i))
		RegsetAddReg(INS_REGS_USE(ins_dummy1),IA64_R32_NAT+i);
	    }

	  for(i=0;i<(IA64INSIMM2(probably_alloc)-IA64INSIMM1(probably_alloc));i++)
	  {
	    livebefore = tail->ins_first ? InsRegsLiveBefore(tail->ins_first) : RegsetNew();
	    if(!tail->ins_first || RegsetIn(livebefore,IA64_R32+i))
	      RegsetAddReg(INS_REGS_USE(ins_dummy1),IA64_R32+IA64INSIMM1(probably_alloc)+i);	  
	    if(!tail->ins_first || RegsetIn(livebefore,IA64_R32_NAT+i))
	      RegsetAddReg(INS_REGS_USE(ins_dummy1),IA64_R32_NAT+IA64INSIMM1(probably_alloc)+i);	  
	  }

	  if(ins_dummy2)
	  {

	    /* dummy 2 updaten */
	    RegsetSetEmpty(INS_REGS_USE(ins_dummy2));

	    for(i=0;i<(IA64INSIMM2(probably_alloc)-IA64INSIMM1(probably_alloc));i++)
	    {
	      liveafter = InsRegsLiveAfter(ins_dummy2);

	      if(RegsetIn(liveafter,IA64_R32+IA64INSIMM1(probably_alloc)+i))
	      {
		RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32+i);
		RegsetAddReg(INS_REGS_USE(ins_dummy2),IA64_R32_NAT+i);
	      }
	    }
	  }
	}
      }
    }
  }                                 
}

/* vim: set shiftwidth=2: */
