/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdlib.h>
#include "ia64_opcodes.h"

#include "ia64_opcodes/ia64_opcodes_a.h"
#include "ia64_opcodes/ia64_opcodes_i.h"
#include "ia64_opcodes/ia64_opcodes_m.h"
#include "ia64_opcodes/ia64_opcodes_b.h"
#include "ia64_opcodes/ia64_opcodes_f.h"
#include "ia64_opcodes/ia64_opcodes_x.h"
#include "ia64_utils.h"

//#include "../../utils/util_error.h"

#include <string.h>

const t_ia64_opcode ia64_opcode_table_dummy[] =
{
  /* Dummy-type instruction encodings (sorted according to major opcode) */
  {"dummy", 0xFFFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFFULL, IA64_INSTR_FORMAT_DUMMY,IA64_INSTR_DUMMY},
  {"dummy", 0xFFFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFFULL, IA64_INSTR_FORMAT_DUMMY,IA64_INSTR_DUMMY},
    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
};  

const t_ia64_opcode ia64_opcode_table_data[] =
{
  /* Dummy-type instruction encodings (sorted according to major opcode) */
  {"data", 0xFFFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFFULL, IA64_INSTR_FORMAT_DATA, IA64_INSTR_DATA},
    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
};  



/* Deze tabel nog aanvullen, in de volgorde van enum t_ia64_instr_type !!! */
const t_ia64_opcode *ia64_opcode_table[] =
{
  ia64_opcode_table_a,
  ia64_opcode_table_i,
  ia64_opcode_table_m,
  ia64_opcode_table_b,
  ia64_opcode_table_f,
  ia64_opcode_table_x, 
  ia64_opcode_table_dummy, 
  ia64_opcode_table_data 

};



#define IA64_IS_A_FORMAT(format) ((format)>=IA64_INSTR_FORMAT_A && (format)<= IA64_INSTR_FORMAT_A10)
#define IA64_IS_B_FORMAT(format) ((format)>=IA64_INSTR_FORMAT_B && (format)<= IA64_INSTR_FORMAT_B9)
#define IA64_IS_F_FORMAT(format) ((format)>=IA64_INSTR_FORMAT_F && (format)<= IA64_INSTR_FORMAT_F15)
#define IA64_IS_I_FORMAT(format) ((format)>=IA64_INSTR_FORMAT_I && (format)<= IA64_INSTR_FORMAT_I29)
#define IA64_IS_M_FORMAT(format) ((format)>=IA64_INSTR_FORMAT_M && (format)<= IA64_INSTR_FORMAT_M46)
#define IA64_IS_X_FORMAT(format) ((format)>=IA64_INSTR_FORMAT_X && (format)<= IA64_INSTR_FORMAT_X4)



t_ia64_instr_type Ia64GetInstrType(t_ia64_opcode* opc)
{
  if(IA64_IS_A_FORMAT(opc->instr_format))
    return IA64_TYPE_A;
  
  if(IA64_IS_B_FORMAT(opc->instr_format))
    return IA64_TYPE_B;
  
  if(IA64_IS_F_FORMAT(opc->instr_format))
    return IA64_TYPE_F;

  if(IA64_IS_I_FORMAT(opc->instr_format))
    return IA64_TYPE_I;

  if(IA64_IS_M_FORMAT(opc->instr_format))
    return IA64_TYPE_M;

  if(IA64_IS_X_FORMAT(opc->instr_format))
    return IA64_TYPE_X;

  fprintf(stderr, "\n\nimplement Ia64GetInstrType !!!!\n\n");
  exit(0);

  return 0; /* keep compiler happy */
}


t_uint16 GetOpcode(t_ia64_instr_type instr_type, t_string string)
{
  t_uint16 opcode = 0;
//  t_ia64_instr_type instr_type = unit2instr_type(0,unit);
  while(strncmp(string, ia64_opcode_table[instr_type][opcode].desc,15) != 0)
    opcode++;
  return opcode;
}

t_uint16 GetNopOpcode(t_ia64_unit unit)
{
  static t_uint16 nopOpcodes [5] = {0, 0, 0, 0, 0};
  
  t_uint16 opcode = 0;
  t_ia64_instr_type instr_type;

  if(unit == IA64_UNIT_L)
    return 0;
  
  if(nopOpcodes[(int)unit] != 0)
    return nopOpcodes[(int)unit];
    
  instr_type = unit2instr_type(0, unit);
  while(strncmp("nop.", ia64_opcode_table[instr_type][opcode].desc, 4) != 0)
    opcode++;
    
//  printf("%d\n",unit);
//  printf("%s\n",ia64_opcode_table[instr_type][opcode].desc);
//  printf("%d\n",instr_type);
  
  return nopOpcodes[(int)unit] = opcode;
}

t_uint16 GetBrCondOpcode(t_string str)
{
  t_uint16 opcode = 0;
  
  while((ia64_opcode_table[IA64_TYPE_B][opcode].instr_format != IA64_INSTR_FORMAT_B1 ||
	strcmp(ia64_opcode_table[IA64_TYPE_B][opcode].desc, str) != 0))
    opcode++;

  return opcode;
}




t_uint16 GetAddsOpcode()
{
  static t_uint16 addsOpcode = 0;
  
  t_uint16 opcode = 0;

  if(addsOpcode != 0)
    return addsOpcode;

  while(ia64_opcode_table[IA64_TYPE_A][opcode].opcode != IA64INS_ADDS_OPCODE)
    opcode++;

  return addsOpcode = opcode;
}
t_uint16 GetAndOpcode()
{
  static t_uint16 andOpcode = 0;
  
  t_uint16 opcode = 0;

  if(andOpcode != 0)
    return andOpcode;

  while(ia64_opcode_table[IA64_TYPE_A][opcode].opcode != IA64INS_AND_OPCODE)
  {
    //printf("%d",opcode);
    opcode++;
  }

  return andOpcode = opcode;
}
t_uint16 GetMoviOpcode()
{
  static t_uint16 movIOpcode = 0;
  
  t_uint16 opcode = 0;

  if(movIOpcode != 0)
    return movIOpcode;

  while(ia64_opcode_table[IA64_TYPE_I][opcode].opcode != IA64INS_MOVI2_OPCODE)
  {
    //printf("%d",opcode);
    opcode++;
  }

  return movIOpcode = opcode;
}
t_uint16 GetMovbOpcode()
{
  static t_uint16 movBOpcode = 0;
  
  t_uint16 opcode = 0;

  if(movBOpcode != 0)
    return movBOpcode;

  while(ia64_opcode_table[IA64_TYPE_I][opcode].instr_format != IA64_INSTR_FORMAT_I22)
  {
//    printf("%d",opcode);
    opcode++;
  }

  return movBOpcode = opcode;
}
t_uint16 GetStfOpcode()
{
  static t_uint16 stfOpcode = 0;
  
  t_uint16 opcode = 0;

  if(stfOpcode != 0)
    return stfOpcode;

  while(ia64_opcode_table[IA64_TYPE_M][opcode].opcode != IA64INS_STIF_OPCODE)
  {
    //printf("%d",opcode);
    opcode++;
  }

  return stfOpcode = opcode;
}
t_uint16 GetStOpcode()
{
  static t_uint16 stOpcode = 0;
  
  t_uint16 opcode = 0;

  if(stOpcode != 0)
    return stOpcode;

  while(ia64_opcode_table[IA64_TYPE_M][opcode].opcode != IA64INS_STII_OPCODE)
  {
    //printf("%d",opcode);
    opcode++;
  }

  return stOpcode = opcode;
}
t_uint16 GetMvOpcode()
{
  static t_uint16 mvOpcode = 0;
  
  t_uint16 opcode = 0;

  if(mvOpcode != 0)
    return mvOpcode;

  while(ia64_opcode_table[IA64_TYPE_X][opcode].opcode != IA64INS_MOVL_OPCODE)
  {
    //printf("%d",opcode);
    opcode++;
  }

  return mvOpcode = opcode;
}

t_uint16 GetLdfOpcode()
{
  static t_uint16 ldfOpcode = 0;
  
  t_uint16 opcode = 0;

  if(ldfOpcode != 0)
    return ldfOpcode;

  while(ia64_opcode_table[IA64_TYPE_M][opcode].opcode != IA64INS_LDIF_OPCODE)
  {
    //printf("%d",opcode);
    opcode++;
  }

  return ldfOpcode = opcode;
}

t_uint16 GetLdOpcode()
{
  static t_uint16 ldOpcode = 0;
  
  t_uint16 opcode = 0;

  if(ldOpcode != 0)
    return ldOpcode;

  while(ia64_opcode_table[IA64_TYPE_M][opcode].opcode != IA64INS_LDII_OPCODE)
  {
    //printf("%d",opcode);
    opcode++;
  }

  return ldOpcode = opcode;
}

t_uint16 GetAddlOpcode()
{
  static t_uint16 addlOpcode = 0;
  
  t_uint16 opcode = 0;

  if(addlOpcode != 0)
    return addlOpcode;

  while(ia64_opcode_table[IA64_TYPE_A][opcode].opcode != IA64INS_ADDL_OPCODE)
    opcode++;

  return addlOpcode = opcode;
}











static t_uint16 opcodeOffsets [6][16] = 
{{0}, {0}, {0}, {0}, {0}, {0}};

static t_bool opcodeOffsetsOK [6][16] = 
{{0}, {0}, {0}, {0}, {0}, {0}};



t_uint16 GetStartOffset(t_ia64_instr_type instr_type, t_uint8 majorOpc)
{
/*
  static t_uint32 total_offset = 0;
*/
  t_uint16 offset = 0;
  
  if(opcodeOffsetsOK[instr_type][majorOpc])
      return opcodeOffsets[instr_type][majorOpc];
  
  if (majorOpc == 0)
   return 0;  
  
  /*majorOpc >0 => majorOpc-1>=0*/
  /* offset zal zeker maar kunnen beginnen vanaf de vorige (als deze al ingevuld staat)...*/
  offset = opcodeOffsets[instr_type][majorOpc-1];
  
  
  
  while(GETBITS64(ia64_opcode_table[instr_type][offset].opcode, 37, 4) != majorOpc || 
      (GETBITS64(ia64_opcode_table[instr_type][offset].opcode, 37, 4) == 0x0 && offset > 0))
      offset++;
/*  
  if(offset == 0)
    printf("%d\n", majorOpc);
  total_offset += offset;
  printf("Current total offset: %d\n", total_offset);  
*/  
  opcodeOffsetsOK[instr_type][majorOpc] = 1;
  return opcodeOffsets[instr_type][majorOpc] = offset;  
}


/* vim: set shiftwidth=2: */
