/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"
#ifndef IA64_OPCODES_H
#define IA64_OPCODES_H

#include "../../datastructures/data_string.h"
#include "../../kernel/diablo_ins.h"

#include "ia64_bundle.h"
#include "ia64_instructions.h"
#include <string.h>


typedef enum _t_ia64_instr_opcode
{
  IA64_INSTR_ADD,
  IA64_INSTR_ADDP4,
  IA64_INSTR_ALLOC,
  IA64_INSTR_AND,
  IA64_INSTR_ANDCM,
  IA64_INSTR_BR,
  IA64_INSTR_BREAK,
  IA64_INSTR_BRL,
  IA64_INSTR_BRP,
  IA64_INSTR_BSW,
  IA64_INSTR_CHK,
  IA64_INSTR_CLRRRB,
  IA64_INSTR_CMP,
  IA64_INSTR_CMP4,
  IA64_INSTR_CMPXCHG,
  IA64_INSTR_COVER,
  IA64_INSTR_CZX,
  IA64_INSTR_DEP,
  IA64_INSTR_EPC,
  IA64_INSTR_EXTR,
  IA64_INSTR_FAMAX,
  IA64_INSTR_FAMIN,
  IA64_INSTR_FAND,
  IA64_INSTR_FANDCM,
  IA64_INSTR_FC,
  IA64_INSTR_FCHKF,
  IA64_INSTR_FCLASS,
  IA64_INSTR_FCLRF,
  IA64_INSTR_FCMP,
  IA64_INSTR_FCVT_FX,
  IA64_INSTR_FCVT_XF,
  IA64_INSTR_FETCHADD,
  IA64_INSTR_FLUSHRS,
  IA64_INSTR_FMA,
  IA64_INSTR_FMAX,
  IA64_INSTR_FMERGE,
  IA64_INSTR_FMIN,
  IA64_INSTR_FMIX,
  IA64_INSTR_FMS,
  IA64_INSTR_FNMA,
  IA64_INSTR_FOR,
  IA64_INSTR_FPACK,
  IA64_INSTR_FPAMAX,
  IA64_INSTR_FPAMIN,
  IA64_INSTR_FPCMP,
  IA64_INSTR_FPCVT_FX,
  IA64_INSTR_FPMA,
  IA64_INSTR_FPMAX,
  IA64_INSTR_FPMERGE,
  IA64_INSTR_FPMIN,
  IA64_INSTR_FPMS,
  IA64_INSTR_FPNMA,
  IA64_INSTR_FPRCPA,
  IA64_INSTR_FPRSQRTA,
  IA64_INSTR_FRCPA,
  IA64_INSTR_FRSQRTA,
  IA64_INSTR_FSELECT,
  IA64_INSTR_FSETC,
  IA64_INSTR_FSWAP,
  IA64_INSTR_FSXT,
  IA64_INSTR_FWB,
  IA64_INSTR_FXOR,
  IA64_INSTR_GETF,
  IA64_INSTR_INVALA,
  IA64_INSTR_ITC,
  IA64_INSTR_ITR,
  IA64_INSTR_LD,
  IA64_INSTR_LDF,
  IA64_INSTR_LDFP,
  IA64_INSTR_LFETCH,
  IA64_INSTR_LOADRS,
  IA64_INSTR_MF,
  IA64_INSTR_MIX,
  IA64_INSTR_MOV_AR,
  IA64_INSTR_MOV_BR,
  IA64_INSTR_MOV_CR,
  IA64_INSTR_MOV_IREG,
  IA64_INSTR_MOV_IP,
  IA64_INSTR_MOV_PR,
  IA64_INSTR_MOV_PSR,
  IA64_INSTR_MOV_UM,
  IA64_INSTR_MOVL,
  IA64_INSTR_MUX,
  IA64_INSTR_NOP,
  IA64_INSTR_OR,
  IA64_INSTR_PACK,
  IA64_INSTR_PADD,
  IA64_INSTR_PAVG,
  IA64_INSTR_PAVGSUB,
  IA64_INSTR_PCMP,
  IA64_INSTR_PMAX,
  IA64_INSTR_PMIN,
  IA64_INSTR_PMPY2,
  IA64_INSTR_PMPYSHR2,
  IA64_INSTR_POPCNT,
  IA64_INSTR_PROBE,
  IA64_INSTR_PSAD1,
  IA64_INSTR_PSHL,
  IA64_INSTR_PSHLADD2,
  IA64_INSTR_PSHR,
  IA64_INSTR_PSHRADD2,
  IA64_INSTR_PSUB,
  IA64_INSTR_PTC_E,
  IA64_INSTR_PTC_G,
  IA64_INSTR_PTC_L,
  IA64_INSTR_PTR,
  IA64_INSTR_RFI,
  IA64_INSTR_RSM,
  IA64_INSTR_RUM,
  IA64_INSTR_SETF,
  IA64_INSTR_SHL,
  IA64_INSTR_SHLADD,
  IA64_INSTR_SHLADDP4,
  IA64_INSTR_SHR,
  IA64_INSTR_SHRP,
  IA64_INSTR_SRLZ,
  IA64_INSTR_SSM,
  IA64_INSTR_ST,
  IA64_INSTR_STF,
  IA64_INSTR_SUB,
  IA64_INSTR_SUM,
  IA64_INSTR_SXT,
  IA64_INSTR_SYNC,
  IA64_INSTR_TAK,
  IA64_INSTR_TBIT,
  IA64_INSTR_THASH,
  IA64_INSTR_TNAT,
  IA64_INSTR_TPA,
  IA64_INSTR_TTAG,
  IA64_INSTR_UNPACK,
  IA64_INSTR_XCHG,
  IA64_INSTR_XMA,
  IA64_INSTR_XOR,
  IA64_INSTR_ZXT,
  
  IA64_INSTR_DUMMY,
  IA64_INSTR_DATA,
  IA64_INSTR_UNSUPPORTED
} t_ia64_instr_opcode;








/* The opcode structure */

typedef struct _t_ia64_opcode
{
  t_string desc;
  t_uint64 opcode;
  t_uint64 mask;
  t_ia64_instr_format instr_format; 
  t_ia64_instr_opcode instr_opcode;
  t_uint32 flags; 
} t_ia64_opcode;


/* IA64-Opcode-flags: provide some extra info about the instruction-opcode */
#define IA64_OPCODE_FIRST	(1<<0)	/* must be first in an insn group */
#define IA64_OPCODE_X_IN_MLX	(1<<1)	/* insn is allowed in X slot of MLX */
#define IA64_OPCODE_LAST	(1<<2)	/* must be last in an insn group */
#define IA64_OPCODE_PRIV	(1<<3)	/* privileged instruct */
#define IA64_OPCODE_SLOT2	(1<<4)	/* insn allowed in slot 2 only */
#define IA64_OPCODE_NO_PRED	(1<<5)	/* insn cannot be predicated */
#define IA64_OPCODE_PSEUDO	(1<<6)	/* insn is a pseudo-op */
#define IA64_OPCODE_F2_EQ_F3	(1<<7)	/* constraint: F2 == F3 */
#define IA64_OPCODE_LEN_EQ_64MCNT	(1<<8)	/* constraint: LEN == 64-CNT */
#define IA64_OPCODE_MOD_RRBS    (1<<9)	/* modifies all rrbs in CFM */
#define IA64_OPCODE_POSTINC	(1<<10)	/* postincrement MR3 operand */



/* 
 * 16bit vertaalde Dummy-opcode 
 */
#define DUMMY_X_OPCODE 0x0
#define DUMMY_LIVE_OPCODE 0x1
#define DATA_OPCODE 0x0


extern const t_ia64_opcode *ia64_opcode_table[];

t_ia64_instr_type Ia64GetInstrType(t_ia64_opcode* opc);

/* nuttige shorthands...->nog vervangen waar gebruikt...*/
#define IA64OPCODEINFO(ins) (ia64_opcode_table[IA64INSTYPE(ins)][IA64INSOPCODE(ins)])


/*gemeenschappelijke shorthands*/

#define bOp(x)  (((t_uint64) ((x) & 0xf)) << 37)
#define mOp     bOp (-1)
#define Op(x)   bOp (x), mOp

#define Is_Ia64_Nop(instr_type, opcode) (strncmp("nop.", ia64_opcode_table[instr_type][opcode].desc), 4)

t_uint16 GetNopOpcode(t_ia64_unit unit);




#define _bX2a(x)		(((t_uint64) ((x) & 0x3)) << 34)
#define _bVe(x)		(((t_uint64) ((x) & 0x1)) << 33)
#define bVc(x)		(((t_uint64) ((x) & 0x1)) << 20)
#define bX3(x)		(((t_uint64) ((x) & 0x7)) << 33)
#define bX4(x)		(((t_uint64) ((x) & 0xf)) << 29)
#define bX6(x)		(((t_uint64) ((x) & 0x3f)) << 26)
#define bX2b(x)		(((t_uint64) ((x) & 0x3)) << 27)


#define mX6a	bX6a (-1)
#define mHint	bHint (-1)
#define bX6a(x)		(((t_uint64) ((x) & 0x3f)) << 30)
#define bHint(x)	(((t_uint64) ((x) & 0x3)) << 28)


#define _OpX2aVe(a,b,c) (bOp (a) | _bX2a (b) | _bVe (c))

#define _OpX6aHint(a,b,c) 	(bOp (a) | bX6a (b) | bHint (c))

#define _OpVc(a,b)		(bOp (a) | bVc (b))

#define _OpX3X6(a,b,c)		(bOp (a) | bX3 (b) | bX6(c))

#define _STINCIMMED(c,h) _OpX6aHint (5, c, h)

#define _LDINCIMMED(c,h) _OpX6aHint (5, c, h)

#define _OpX2aVeX4X2b(a,b,c,d,e) (bOp (a) | _bX2a (b) | _bVe (c) | bX4 (d) | bX2b (e))
  

#define IA64INS_ADDL_OPCODE  (bOp(9))
#define IA64INS_ADDS_OPCODE  (_OpX2aVe (8, 2, 0))
#define IA64INS_STIF_OPCODE  (_STINCIMMED (0x3b, 0))
#define IA64INS_STII_OPCODE  (_STINCIMMED (0x33, 0))
#define IA64INS_LDII_OPCODE  (_LDINCIMMED (0x03, 0))
#define IA64INS_LDIF_OPCODE  (_LDINCIMMED (0x1b, 0))
#define IA64INS_MOVL_OPCODE  (_OpVc (6, 0))
#define IA64INS_AND_OPCODE  (_OpX2aVeX4X2b (8, 0, 0, 0xb, 0))
#define IA64INS_BREAK_OPCODE (_OpX3X6(0,0,0x0))
#define IA64INS_MOVI2_OPCODE (_OpX3X6(0,0,0x0a))
#define IA64INS_MOVB_OPCODE (_OpX3X6(0,0,0x31))

#define IA64INS_GR_MOV_OPCODE  IA64INS_ADDS_OPCODE
#define IA64INS_IS_GR_MOV(ins) (IA64OPCODEINFO(ins).opcode == IA64INS_GR_MOV_OPCODE)

t_uint16 GetBrCondOpcode(t_string str);

t_uint16 GetAddsOpcode();
t_uint16 GetAddlOpcode();
t_uint16 GetStOpcode();
t_uint16 GetStfOpcode();
t_uint16 GetLdOpcode();
t_uint16 GetLdfOpcode();
t_uint16 GetMvOpcode();
t_uint16 GetOrOpcode();
t_uint16 GetMoviOpcode();
t_uint16 GetMovbOpcode();







t_uint16 GetStartOffset(t_ia64_instr_type instr_type, t_uint8 majorOpc);


#endif
/* vim: set shiftwidth=2: */

