/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_opcodes_b.h"
#include "../../../utils/util_error.h"

#include "../ia64_instructions.h"
#include "../ia64_disassemble/ia64_disassemble_one.h"
#include "../../../simple_types.h"


/* instruction bit fields: */
#define bBtype(x)	(((t_uint64) ((x) & 0x7)) << 6)
#define bD(x)		(((t_uint64) ((x) & 0x1)) << 35)
#define bIh(x)		(((t_uint64) ((x) & 0x1)) << 35)
#define bPa(x)		(((t_uint64) ((x) & 0x1)) << 12)
#define bPr(x)		(((t_uint64) ((x) & 0x3f)) << 0)
#define bWha(x)		(((t_uint64) ((x) & 0x3)) << 33)
#define bWhb(x)		(((t_uint64) ((x) & 0x3)) << 3)
#define bWhc(x)		(((t_uint64) ((x) & 0x7)) << 32)
#define bX6_(x)		(((t_uint64) ((x) & 0x3f)) << 27)

#define mBtype		bBtype (-1)
#define mD		bD (-1)
#define mIh		bIh (-1)
#define mPa		bPa (-1)
#define mPr		bPr (-1)
#define mWha		bWha (-1)
#define mWhb		bWhb (-1)
#define mWhc		bWhc (-1)
#define mX6		bX6_ (-1)

#define OpX6(a,b) 	(bOp (a) | bX6_ (b)), (mOp | mX6)
#define OpPaWhaD(a,b,c,d) \
	(bOp (a) | bPa (b) | bWha (c) | bD (d)), (mOp | mPa | mWha | mD)
#define OpPaWhcD(a,b,c,d) \
	(bOp (a) | bPa (b) | bWhc (c) | bD (d)), (mOp | mPa | mWhc | mD)
#define OpBtypePaWhaD(a,b,c,d,e) \
	(bOp (a) | bBtype (b) | bPa (c) | bWha (d) | bD (e)), \
	(mOp | mBtype | mPa | mWha | mD)
#define OpBtypePaWhaDPr(a,b,c,d,e,f) \
	(bOp (a) | bBtype (b) | bPa (c) | bWha (d) | bD (e) | bPr (f)), \
	(mOp | mBtype | mPa | mWha | mD | mPr)
#define OpX6BtypePaWhaD(a,b,c,d,e,f) \
	(bOp (a) | bX6_ (b) | bBtype (c) | bPa (d) | bWha (e) | bD (f)), \
	(mOp | mX6 | mBtype | mPa | mWha | mD)
#define OpX6BtypePaWhaDPr(a,b,c,d,e,f,g) \
   (bOp (a) | bX6_ (b) | bBtype (c) | bPa (d) | bWha (e) | bD (f) | bPr (g)), \
	(mOp | mX6 | mBtype | mPa | mWha | mD | mPr)
#define OpIhWhb(a,b,c) \
	(bOp (a) | bIh (b) | bWhb (c)), \
	(mOp | mIh | mWhb)
#define OpX6IhWhb(a,b,c,d) \
	(bOp (a) | bX6_ (b) | bIh (c) | bWhb (d)), \
	(mOp | mX6 | mIh | mWhb)


const t_ia64_opcode ia64_opcode_table_b[] =
  {
    /* B-type instruction encodings (sorted according to major opcode) */

/* zijn pseudo's
#define BR(a,b)		OpX6BtypePaWhaDPr (0, 0x20, 0, a, 0, b, 0)
    {"br.few",		BR (0, 0), IA64_INSTR_FORMAT_Unsupported},
    {"br",		BR (0, 0), IA64_INSTR_FORMAT_Unsupported},
    {"br.few.clr",	BR (0, 1), IA64_INSTR_FORMAT_Unsupported},
    {"br.clr",		BR (0, 1), IA64_INSTR_FORMAT_Unsupported},
    {"br.many",		BR (1, 0), IA64_INSTR_FORMAT_Unsupported},
    {"br.many.clr",	BR (1, 1), IA64_INSTR_FORMAT_Unsupported},
#undef BR
*/

#define BR(a,b,c,d,e)	OpX6BtypePaWhaD (0, a, b, c, d, e)
    {"br.cond.sptk.few",	BR (0x20, 0, 0, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.sptk.few.clr",	BR (0x20, 0, 0, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.spnt.few",	BR (0x20, 0, 0, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.spnt.few.clr",	BR (0x20, 0, 0, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dptk.few",	BR (0x20, 0, 0, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dptk.few.clr",	BR (0x20, 0, 0, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dpnt.few",	BR (0x20, 0, 0, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dpnt.few.clr",	BR (0x20, 0, 0, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.sptk.many",	BR (0x20, 0, 1, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.sptk.many.clr",	BR (0x20, 0, 1, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.spnt.many",	BR (0x20, 0, 1, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.spnt.many.clr",	BR (0x20, 0, 1, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dptk.many",	BR (0x20, 0, 1, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dptk.many.clr",	BR (0x20, 0, 1, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dpnt.many",	BR (0x20, 0, 1, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.cond.dpnt.many.clr",	BR (0x20, 0, 1, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.sptk.few",		BR (0x20, 0, 0, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.sptk.few.clr",		BR (0x20, 0, 0, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.spnt.few",		BR (0x20, 0, 0, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.spnt.few.clr",		BR (0x20, 0, 0, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dptk.few",		BR (0x20, 0, 0, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dptk.few.clr",		BR (0x20, 0, 0, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dpnt.few",		BR (0x20, 0, 0, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dpnt.few.clr",		BR (0x20, 0, 0, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.sptk.many",		BR (0x20, 0, 1, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.sptk.many.clr",	BR (0x20, 0, 1, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.spnt.many",		BR (0x20, 0, 1, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.spnt.many.clr",	BR (0x20, 0, 1, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dptk.many",		BR (0x20, 0, 1, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dptk.many.clr",	BR (0x20, 0, 1, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dpnt.many",		BR (0x20, 0, 1, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.dpnt.many.clr",	BR (0x20, 0, 1, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.sptk.few",		BR (0x20, 1, 0, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.sptk.few.clr",	BR (0x20, 1, 0, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.spnt.few",		BR (0x20, 1, 0, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.spnt.few.clr",	BR (0x20, 1, 0, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dptk.few",		BR (0x20, 1, 0, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dptk.few.clr",	BR (0x20, 1, 0, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dpnt.few",		BR (0x20, 1, 0, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dpnt.few.clr",	BR (0x20, 1, 0, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.sptk.many",		BR (0x20, 1, 1, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.sptk.many.clr",	BR (0x20, 1, 1, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.spnt.many",		BR (0x20, 1, 1, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.spnt.many.clr",	BR (0x20, 1, 1, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dptk.many",		BR (0x20, 1, 1, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dptk.many.clr",	BR (0x20, 1, 1, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dpnt.many",		BR (0x20, 1, 1, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ia.dpnt.many.clr",	BR (0x20, 1, 1, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR},
    {"br.ret.sptk.few",		BR (0x21, 4, 0, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.sptk.few.clr",	BR (0x21, 4, 0, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.spnt.few",		BR (0x21, 4, 0, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.spnt.few.clr",	BR (0x21, 4, 0, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dptk.few",		BR (0x21, 4, 0, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR,  
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dptk.few.clr",	BR (0x21, 4, 0, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dpnt.few",		BR (0x21, 4, 0, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dpnt.few.clr",	BR (0x21, 4, 0, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.sptk.many",	BR (0x21, 4, 1, 0, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.sptk.many.clr",	BR (0x21, 4, 1, 0, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.spnt.many",	BR (0x21, 4, 1, 1, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.spnt.many.clr",	BR (0x21, 4, 1, 1, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dptk.many",	BR (0x21, 4, 1, 2, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dptk.many.clr",	BR (0x21, 4, 1, 2, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dpnt.many",	BR (0x21, 4, 1, 3, 0), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
    {"br.ret.dpnt.many.clr",	BR (0x21, 4, 1, 3, 1), IA64_INSTR_FORMAT_B4, IA64_INSTR_BR, 
      IA64_OPCODE_MOD_RRBS},
#undef BR

    {"cover",		OpX6 (0, 0x02), IA64_INSTR_FORMAT_B8, IA64_INSTR_COVER, 
      IA64_OPCODE_NO_PRED | IA64_OPCODE_LAST | IA64_OPCODE_MOD_RRBS},
    {"clrrrb",		OpX6 (0, 0x04), IA64_INSTR_FORMAT_B8, IA64_INSTR_CLRRRB, 
      IA64_OPCODE_NO_PRED | IA64_OPCODE_LAST | IA64_OPCODE_MOD_RRBS},
    {"clrrrb.pr",	OpX6 (0, 0x05), IA64_INSTR_FORMAT_B8, IA64_INSTR_CLRRRB, 
      IA64_OPCODE_NO_PRED | IA64_OPCODE_LAST | IA64_OPCODE_MOD_RRBS},
    {"rfi",		OpX6 (0, 0x08), IA64_INSTR_FORMAT_B8, IA64_INSTR_RFI, 
      IA64_OPCODE_NO_PRED | IA64_OPCODE_LAST | IA64_OPCODE_PRIV | IA64_OPCODE_MOD_RRBS},
    {"bsw.0",		OpX6 (0, 0x0c), IA64_INSTR_FORMAT_B8, IA64_INSTR_BSW, 
      IA64_OPCODE_NO_PRED | IA64_OPCODE_LAST | IA64_OPCODE_PRIV},
    {"bsw.1",		OpX6 (0, 0x0d), IA64_INSTR_FORMAT_B8, IA64_INSTR_BSW, 
      IA64_OPCODE_NO_PRED | IA64_OPCODE_LAST | IA64_OPCODE_PRIV},
    {"epc",		OpX6 (0, 0x10), IA64_INSTR_FORMAT_B8, IA64_INSTR_EPC, 
      IA64_OPCODE_NO_PRED},

    {"break.b",		OpX6 (0, 0x00), IA64_INSTR_FORMAT_B9, IA64_INSTR_BREAK},

    {"br.call.sptk.few",	OpPaWhcD (1, 0, 1, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.sptk.few.clr",	OpPaWhcD (1, 0, 1, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.spnt.few",	OpPaWhcD (1, 0, 3, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.spnt.few.clr",	OpPaWhcD (1, 0, 3, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dptk.few",	OpPaWhcD (1, 0, 5, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dptk.few.clr",	OpPaWhcD (1, 0, 5, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dpnt.few",	OpPaWhcD (1, 0, 7, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dpnt.few.clr",	OpPaWhcD (1, 0, 7, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    /* GCC compiler fout */
    {"br.call.sptk.many",	OpPaWhcD (1, 1, 0, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR}, 
    {"br.call.sptk.many",	OpPaWhcD (1, 1, 1, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.sptk.many.clr",	OpPaWhcD (1, 1, 1, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.spnt.many",	OpPaWhcD (1, 1, 2, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.spnt.many",	OpPaWhcD (1, 1, 3, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.spnt.many.clr",	OpPaWhcD (1, 1, 3, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.spnt.many",	OpPaWhcD (1, 1, 4, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dptk.many",	OpPaWhcD (1, 1, 5, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dptk.many.clr",	OpPaWhcD (1, 1, 5, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dpnt.many",	OpPaWhcD (1, 1, 6, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dpnt.many",	OpPaWhcD (1, 1, 7, 0), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},
    {"br.call.dpnt.many.clr",	OpPaWhcD (1, 1, 7, 1), IA64_INSTR_FORMAT_B5, IA64_INSTR_BR},

#define BRP(a,b,c)		OpX6IhWhb (2, a, b, c)
    {"brp.sptk",		BRP (0x10, 0, 0), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.dptk",		BRP (0x10, 0, 2), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.sptk.imp",		BRP (0x10, 1, 0), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.dptk.imp",		BRP (0x10, 1, 2), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.ret.sptk",		BRP (0x11, 0, 0), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.ret.dptk",		BRP (0x11, 0, 2), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.ret.sptk.imp",	BRP (0x11, 1, 0), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.ret.dptk.imp",	BRP (0x11, 1, 2), IA64_INSTR_FORMAT_B7, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
#undef BRP

    {"nop.b",		OpX6 (2, 0x00), IA64_INSTR_FORMAT_B9, IA64_INSTR_NOP},

/* zijn pseudo's
#define BR(a,b)			OpBtypePaWhaDPr (4, 0, a, 0, b, 0)
    {"br.few",		BR (0, 0), IA64_INSTR_FORMAT_B},
    {"br",		BR (0, 0), IA64_INSTR_FORMAT_B},
    {"br.few.clr",	BR (0, 1), IA64_INSTR_FORMAT_B},
    {"br.clr",		BR (0, 1), IA64_INSTR_FORMAT_B},
    {"br.many",		BR (1, 0), IA64_INSTR_FORMAT_B},
    {"br.many.clr",	BR (1, 1), IA64_INSTR_FORMAT_B},
#undef BR
*/


#define BR(a,b,c)		OpBtypePaWhaD (4, 0, a, b, c)
    {"br.cond.sptk.few",	BR (0, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.sptk.few.clr",	BR (0, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.spnt.few",	BR (0, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.spnt.few.clr",	BR (0, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dptk.few",	BR (0, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dptk.few.clr",	BR (0, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dpnt.few",	BR (0, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dpnt.few.clr",	BR (0, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.sptk.many",	BR (1, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.sptk.many.clr",	BR (1, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.spnt.many",	BR (1, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.spnt.many.clr",	BR (1, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dptk.many",	BR (1, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dptk.many.clr",	BR (1, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dpnt.many",	BR (1, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.cond.dpnt.many.clr",	BR (1, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.sptk.few",		BR (0, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.sptk.few.clr",		BR (0, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.spnt.few",		BR (0, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.spnt.few.clr",		BR (0, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dptk.few",		BR (0, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dptk.few.clr",		BR (0, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dpnt.few",		BR (0, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dpnt.few.clr",		BR (0, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.sptk.many",		BR (1, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.sptk.many.clr",	BR (1, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.spnt.many",		BR (1, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.spnt.many.clr",	BR (1, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dptk.many",		BR (1, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dptk.many.clr",	BR (1, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dpnt.many",		BR (1, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
    {"br.dpnt.many.clr",	BR (1, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR},
#undef BR

#define BR(a,b,c,d) 		OpBtypePaWhaD (4, a, b, c, d)
    {"br.wexit.sptk.few",	BR (2, 0, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.sptk.few.clr",	BR (2, 0, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.spnt.few",	BR (2, 0, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.spnt.few.clr",	BR (2, 0, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dptk.few",	BR (2, 0, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dptk.few.clr",	BR (2, 0, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dpnt.few",	BR (2, 0, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dpnt.few.clr",	BR (2, 0, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.sptk.many",	BR (2, 1, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.sptk.many.clr",	BR (2, 1, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.spnt.many",	BR (2, 1, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.spnt.many.clr",	BR (2, 1, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dptk.many",	BR (2, 1, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dptk.many.clr",	BR (2, 1, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dpnt.many",	BR (2, 1, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wexit.dpnt.many.clr",	BR (2, 1, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.sptk.few",	BR (3, 0, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.sptk.few.clr",	BR (3, 0, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.spnt.few",	BR (3, 0, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.spnt.few.clr",	BR (3, 0, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dptk.few",	BR (3, 0, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dptk.few.clr",	BR (3, 0, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dpnt.few",	BR (3, 0, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dpnt.few.clr",	BR (3, 0, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.sptk.many",	BR (3, 1, 0, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.sptk.many.clr",	BR (3, 1, 0, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.spnt.many",	BR (3, 1, 1, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.spnt.many.clr",	BR (3, 1, 1, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dptk.many",	BR (3, 1, 2, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dptk.many.clr",	BR (3, 1, 2, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dpnt.many",	BR (3, 1, 3, 0), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},
    {"br.wtop.dpnt.many.clr",	BR (3, 1, 3, 1), IA64_INSTR_FORMAT_B1, IA64_INSTR_BR, 
      IA64_OPCODE_SLOT2 | IA64_OPCODE_MOD_RRBS},

#undef BR
#define BR(a,b,c,d) 		OpBtypePaWhaD (4, a, b, c, d)
    {"br.cloop.sptk.few",	BR (5, 0, 0, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.sptk.few.clr",	BR (5, 0, 0, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.spnt.few",	BR (5, 0, 1, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.spnt.few.clr",	BR (5, 0, 1, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dptk.few",	BR (5, 0, 2, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dptk.few.clr",	BR (5, 0, 2, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dpnt.few",	BR (5, 0, 3, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dpnt.few.clr",	BR (5, 0, 3, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.sptk.many",	BR (5, 1, 0, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.sptk.many.clr",	BR (5, 1, 0, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.spnt.many",	BR (5, 1, 1, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.spnt.many.clr",	BR (5, 1, 1, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dptk.many",	BR (5, 1, 2, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dptk.many.clr",	BR (5, 1, 2, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dpnt.many",	BR (5, 1, 3, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cloop.dpnt.many.clr",	BR (5, 1, 3, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED},
    {"br.cexit.sptk.few",	BR (6, 0, 0, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.sptk.few.clr",	BR (6, 0, 0, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.spnt.few",	BR (6, 0, 1, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.spnt.few.clr",	BR (6, 0, 1, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dptk.few",	BR (6, 0, 2, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dptk.few.clr",	BR (6, 0, 2, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dpnt.few",	BR (6, 0, 3, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dpnt.few.clr",	BR (6, 0, 3, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.sptk.many",	BR (6, 1, 0, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.sptk.many.clr",	BR (6, 1, 0, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.spnt.many",	BR (6, 1, 1, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.spnt.many.clr",	BR (6, 1, 1, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dptk.many",	BR (6, 1, 2, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dptk.many.clr",	BR (6, 1, 2, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dpnt.many",	BR (6, 1, 3, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.cexit.dpnt.many.clr",	BR (6, 1, 3, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.sptk.few",	BR (7, 0, 0, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.sptk.few.clr",	BR (7, 0, 0, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.spnt.few",	BR (7, 0, 1, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.spnt.few.clr",	BR (7, 0, 1, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dptk.few",	BR (7, 0, 2, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dptk.few.clr",	BR (7, 0, 2, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dpnt.few",	BR (7, 0, 3, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dpnt.few.clr",	BR (7, 0, 3, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.sptk.many",	BR (7, 1, 0, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.sptk.many.clr",	BR (7, 1, 0, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.spnt.many",	BR (7, 1, 1, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.spnt.many.clr",	BR (7, 1, 1, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dptk.many",	BR (7, 1, 2, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dptk.many.clr",	BR (7, 1, 2, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dpnt.many",	BR (7, 1, 3, 0), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},
    {"br.ctop.dpnt.many.clr",	BR (7, 1, 3, 1), IA64_INSTR_FORMAT_B2, IA64_INSTR_BR,
      IA64_OPCODE_SLOT2 | IA64_OPCODE_NO_PRED | IA64_OPCODE_MOD_RRBS},

#undef BR
//#define BR(a,b,c,d)		OpBtypePaWhaD (4, a, b, c, d)
    {"br.call.sptk.few",	OpPaWhaD (5, 0, 0, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.sptk.few.clr",	OpPaWhaD (5, 0, 0, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.spnt.few",	OpPaWhaD (5, 0, 1, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.spnt.few.clr",	OpPaWhaD (5, 0, 1, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dptk.few",	OpPaWhaD (5, 0, 2, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dptk.few.clr",	OpPaWhaD (5, 0, 2, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dpnt.few",	OpPaWhaD (5, 0, 3, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dpnt.few.clr",	OpPaWhaD (5, 0, 3, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.sptk.many",	OpPaWhaD (5, 1, 0, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.sptk.many.clr",	OpPaWhaD (5, 1, 0, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.spnt.many",	OpPaWhaD (5, 1, 1, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.spnt.many.clr",	OpPaWhaD (5, 1, 1, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dptk.many",	OpPaWhaD (5, 1, 2, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dptk.many.clr",	OpPaWhaD (5, 1, 2, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dpnt.many",	OpPaWhaD (5, 1, 3, 0), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
    {"br.call.dpnt.many.clr",	OpPaWhaD (5, 1, 3, 1), IA64_INSTR_FORMAT_B3, IA64_INSTR_BR},
//#undef BR

    /* branch predict */
#define BRP(a,b)		OpIhWhb (7, a, b)
    {"brp.sptk",		BRP (0, 0), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.loop",		BRP (0, 1), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.dptk",		BRP (0, 2), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.exit",		BRP (0, 3), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.sptk.imp",		BRP (1, 0), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.loop.imp",		BRP (1, 1), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.dptk.imp",		BRP (1, 2), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
    {"brp.exit.imp",		BRP (1, 3), IA64_INSTR_FORMAT_B6, IA64_INSTR_BRP, 
      IA64_OPCODE_NO_PRED},
#undef BRP

    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
};

#undef bBtype
#undef bD
#undef bIh
#undef bPa
#undef bPr
#undef bWha
#undef bWhb
#undef bWhc
#undef bX6_
#undef mBtype
#undef mD
#undef mIh
#undef mPa
#undef mPr
#undef mWha
#undef mWhb
#undef mWhc
#undef mX6
#undef OpX6
#undef OpPaWhaD
#undef OpPaWhcD
#undef OpBtypePaWhaD
#undef OpBtypePaWhaDPr
#undef OpX6BtypePaWhaD
#undef OpX6BtypePaWhaDPr
#undef OpIhWhb
#undef OpX6IhWhb

/* vim: set shiftwidth=2: */



