/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimiz$
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published$
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1$
 *
 */
#include "ia64_opcodes_i.h"
#include "../../../utils/util_error.h"

#include "../ia64_instructions.h"
#include "../ia64_disassemble/ia64_disassemble_one.h"
#include "../../../simple_types.h"

/* instruction bit fields: */
#define bC(x)           (((t_uint64) ((x) & 0x1)) << 12)
#define bIh(x)          (((t_uint64) ((x) & 0x1)) << 23)
#define bTa(x)          (((t_uint64) ((x) & 0x1)) << 33)
#define bTag13(x)       (((t_uint64) ((x) & 0x1)) << 33)
#define bTb(x)          (((t_uint64) ((x) & 0x1)) << 36)
#define bVc(x)          (((t_uint64) ((x) & 0x1)) << 20)
#define bVe(x)          (((t_uint64) ((x) & 0x1)) << 32)
#define bWh(x)          (((t_uint64) ((x) & 0x3)) << 20)
#define bX(x)           (((t_uint64) ((x) & 0x1)) << 33)
#define bXb(x)          (((t_uint64) ((x) & 0x1)) << 22)
#define bX2(x)          (((t_uint64) ((x) & 0x3)) << 34)
#define bX2a(x)         (((t_uint64) ((x) & 0x3)) << 34)
#define bX2b_(x)         (((t_uint64) ((x) & 0x3)) << 28)
#define bX2c(x)         (((t_uint64) ((x) & 0x3)) << 30)
#define bX3(x)          (((t_uint64) ((x) & 0x7)) << 33)
#define bX6_(x)          (((t_uint64) ((x) & 0x3f)) << 27)
#define bYa(x)          (((t_uint64) ((x) & 0x1)) << 13)
#define bYb(x)          (((t_uint64) ((x) & 0x1)) << 26)
#define bZa(x)          (((t_uint64) ((x) & 0x1)) << 36)
#define bZb(x)          (((t_uint64) ((x) & 0x1)) << 33)

/* instruction bit masks: */
#define mC	bC (-1)
#define mIh	bIh (-1)
#define mTa	bTa (-1)
#define mTag13	bTag13 (-1)
#define mTb	bTb (-1)
#define mVc	bVc (-1)
#define mVe	bVe (-1)
#define mWh	bWh (-1)
#define mX	bX (-1)
#define mXb	bXb (-1)
#define mX2	bX2 (-1)
#define mX2a	bX2a (-1)
#define mX2b	bX2b_ (-1)
#define mX2c	bX2c (-1)
#define mX3	bX3 (-1)
#define mX6	bX6_ (-1)
#define mYa	bYa (-1)
#define mYb	bYb (-1)
#define mZa	bZa (-1)
#define mZb	bZb (-1)

#define OpZaZbVeX2aX2b(a,b,c,d,e,f) \
	(bOp (a) | bZa (b) | bZb (c) | bVe (d) | bX2a (e) | bX2b_ (f)), \
	(mOp | mZa | mZb | mVe | mX2a | mX2b)
#define OpZaZbVeX2aX2bX2c(a,b,c,d,e,f,g) \
  (bOp (a) | bZa (b) | bZb (c) | bVe (d) | bX2a (e) | bX2b_ (f) | bX2c (g)), \
	(mOp | mZa | mZb | mVe | mX2a | mX2b | mX2c)
#define OpX2X(a,b,c)		(bOp (a) | bX2 (b) | bX (c)), (mOp | mX2 | mX)
#define OpX2XYa(a,b,c,d)	(bOp (a) | bX2 (b) | bX (c) | bYa (d)), \
				(mOp | mX2 | mX | mYa)
#define OpX2XYb(a,b,c,d)	(bOp (a) | bX2 (b) | bX (c) | bYb (d)), \
				(mOp | mX2 | mX | mYb)
#define OpX2TaTbYaC(a,b,c,d,e,f) \
	(bOp (a) | bX2 (b) | bTa (c) | bTb (d) | bYa (e) | bC (f)), \
	(mOp | mX2 | mTa | mTb | mYa | mC)
#define OpX3(a,b)		(bOp (a) | bX3 (b)), (mOp | mX3)
#define OpX3X6(a,b,c)		(bOp (a) | bX3 (b) | bX6_(c)), \
				(mOp | mX3 | mX6)
#define OpX3XbIhWh(a,b,c,d,e) \
  (bOp (a) | bX3 (b) | bXb (c) | bIh (d) | bWh (e)), \
  (mOp | mX3 | mXb | mIh | mWh)
#define OpX3XbIhWhTag13(a,b,c,d,e,f) \
     (bOp (a) | bX3 (b) | bXb (c) | bIh (d) | bWh (e) | bTag13 (f)), \
     (mOp | mX3 | mXb | mIh | mWh | mTag13)

const t_ia64_opcode ia64_opcode_table_i[] =
  {
    /* I-type instruction encodings (sorted according to major opcode) */

    {"break.i", OpX3X6 (0, 0, 0x00), IA64_INSTR_FORMAT_I19, IA64_INSTR_BREAK, 
      IA64_OPCODE_X_IN_MLX},
    {"nop.i",   OpX3X6 (0, 0, 0x01), IA64_INSTR_FORMAT_I19, IA64_INSTR_NOP, 
      IA64_OPCODE_X_IN_MLX},
    {"chk.s.i", OpX3 (0, 1), IA64_INSTR_FORMAT_I20, IA64_INSTR_CHK, 
      IA64_OPCODE_X_IN_MLX},

#define MOV(a,b,c,d)    OpX3XbIhWh (0, a, b, c, d)
    {"mov.sptk",                MOV (7, 0, 0, 0), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.sptk.imp",            MOV (7, 0, 1, 0), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov",                     MOV (7, 0, 0, 1), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.imp",                 MOV (7, 0, 1, 1), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.dptk",                MOV (7, 0, 0, 2), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.dptk.imp",            MOV (7, 0, 1, 2), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.ret.sptk",            MOV (7, 1, 0, 0), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.ret.sptk.imp",        MOV (7, 1, 1, 0), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.ret",                 MOV (7, 1, 0, 1), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.ret.imp",             MOV (7, 1, 1, 1), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.ret.dptk",            MOV (7, 1, 0, 2), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
    {"mov.ret.dptk.imp",        MOV (7, 1, 1, 2), IA64_INSTR_FORMAT_I21, IA64_INSTR_MOV_BR},
#undef MOV
    {"mov",     OpX3X6 (0, 0, 0x31), IA64_INSTR_FORMAT_I22, IA64_INSTR_MOV_BR},
    {"mov",     OpX3 (0, 3), IA64_INSTR_FORMAT_I23, IA64_INSTR_MOV_PR},
    {"mov",     OpX3 (0, 2), IA64_INSTR_FORMAT_I24, IA64_INSTR_MOV_PR},
    {"mov",     OpX3X6 (0, 0, 0x30), IA64_INSTR_FORMAT_I25, IA64_INSTR_MOV_IP},
    {"mov",     OpX3X6 (0, 0, 0x33), IA64_INSTR_FORMAT_I25, IA64_INSTR_MOV_PR},
    {"mov.i",   OpX3X6 (0, 0, 0x2a), IA64_INSTR_FORMAT_I26, IA64_INSTR_MOV_AR},
    {"mov.i",   OpX3X6 (0, 0, 0x0a), IA64_INSTR_FORMAT_I27, IA64_INSTR_MOV_AR},
    {"mov.i",   OpX3X6 (0, 0, 0x32), IA64_INSTR_FORMAT_I28, IA64_INSTR_MOV_AR},
    {"zxt1",    OpX3X6 (0, 0, 0x10), IA64_INSTR_FORMAT_I29, IA64_INSTR_ZXT},
    {"zxt2",    OpX3X6 (0, 0, 0x11), IA64_INSTR_FORMAT_I29, IA64_INSTR_ZXT},
    {"zxt4",    OpX3X6 (0, 0, 0x12), IA64_INSTR_FORMAT_I29, IA64_INSTR_ZXT},
    {"sxt1",    OpX3X6 (0, 0, 0x14), IA64_INSTR_FORMAT_I29, IA64_INSTR_SXT},
    {"sxt2",    OpX3X6 (0, 0, 0x15), IA64_INSTR_FORMAT_I29, IA64_INSTR_SXT},
    {"sxt4",    OpX3X6 (0, 0, 0x16), IA64_INSTR_FORMAT_I29, IA64_INSTR_SXT},
    {"czx1.l",  OpX3X6 (0, 0, 0x18), IA64_INSTR_FORMAT_I29, IA64_INSTR_CZX},
    {"czx2.l",  OpX3X6 (0, 0, 0x19), IA64_INSTR_FORMAT_I29, IA64_INSTR_CZX},
    {"czx1.r",  OpX3X6 (0, 0, 0x1c), IA64_INSTR_FORMAT_I29, IA64_INSTR_CZX},
    {"czx2.r",  OpX3X6 (0, 0, 0x1d), IA64_INSTR_FORMAT_I29, IA64_INSTR_CZX},

    {"dep",     Op (4), IA64_INSTR_FORMAT_I15, IA64_INSTR_DEP},

    {"shrp",    OpX2X (5, 3, 0), IA64_INSTR_FORMAT_I10, IA64_INSTR_SHRP},

    {"extr.u",  OpX2XYa (5, 1, 0, 0), IA64_INSTR_FORMAT_I11, IA64_INSTR_EXTR},

    {"extr",    OpX2XYa (5, 1, 0, 1), IA64_INSTR_FORMAT_I11, IA64_INSTR_EXTR},

    {"dep.z",   OpX2XYb (5, 1, 1, 0), IA64_INSTR_FORMAT_I12, IA64_INSTR_DEP},
    {"dep.z",   OpX2XYb (5, 1, 1, 1), IA64_INSTR_FORMAT_I13, IA64_INSTR_DEP},
    {"dep",     OpX2X (5, 3, 1), IA64_INSTR_FORMAT_I14, IA64_INSTR_DEP},
#define TBIT(a,b,c,d)   OpX2TaTbYaC (5, 0, a, b, c, d)

    {"tbit.z",           TBIT   (0, 0, 0, 0), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
    {"tbit.z.unc",       TBIT   (0, 0, 0, 1), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
    {"tbit.z.and",       TBIT   (0, 1, 0, 0), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
    {"tbit.nz.and",      TBIT   (0, 1, 0, 1), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
    {"tbit.z.or",        TBIT   (1, 0, 0, 0), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
    {"tbit.nz.or",       TBIT   (1, 0, 0, 1), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
    {"tbit.z.or.andcm",  TBIT   (1, 1, 0, 0), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
    {"tbit.nz.or.andcm", TBIT   (1, 1, 0, 1), IA64_INSTR_FORMAT_I16, IA64_INSTR_TBIT},
#undef TBIT

#define TNAT(a,b,c,d)   OpX2TaTbYaC (5, 0, a, b, c, d)
    {"tnat.z",           TNAT   (0, 0, 1, 0), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
    {"tnat.z.unc",       TNAT   (0, 0, 1, 1), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
    {"tnat.z.and",       TNAT   (0, 1, 1, 0), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
    {"tnat.nz.and",      TNAT   (0, 1, 1, 1), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
    {"tnat.z.or",        TNAT   (1, 0, 1, 0), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
    {"tnat.nz.or",       TNAT   (1, 0, 1, 1), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
    {"tnat.z.or.andcm",  TNAT   (1, 1, 1, 0), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
    {"tnat.nz.or.andcm", TNAT   (1, 1, 1, 1), IA64_INSTR_FORMAT_I17, IA64_INSTR_TNAT},
#undef TNAT
    {"pmpyshr2",   OpZaZbVeX2aX2b (7, 0, 1, 0, 0, 3), IA64_INSTR_FORMAT_I1, IA64_INSTR_PMPYSHR2},
    {"pmpyshr2.u", OpZaZbVeX2aX2b (7, 0, 1, 0, 0, 1), IA64_INSTR_FORMAT_I1, IA64_INSTR_PMPYSHR2},
    {"pmpy2.r",    OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 1, 3), IA64_INSTR_FORMAT_I2, IA64_INSTR_PMPY2},
    {"pmpy2.l",    OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 3, 3), IA64_INSTR_FORMAT_I2, IA64_INSTR_PMPY2},
    {"mix1.r",     OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 2, 0, 2), IA64_INSTR_FORMAT_I2, IA64_INSTR_MIX},
    {"mix2.r",     OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 0, 2), IA64_INSTR_FORMAT_I2, IA64_INSTR_MIX},
    {"mix4.r",     OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 2, 0, 2), IA64_INSTR_FORMAT_I2, IA64_INSTR_MIX},
    {"mix1.l",     OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 2, 2, 2), IA64_INSTR_FORMAT_I2, IA64_INSTR_MIX},
    {"mix2.l",     OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 2, 2), IA64_INSTR_FORMAT_I2, IA64_INSTR_MIX},
    {"mix4.l",     OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 2, 2, 2), IA64_INSTR_FORMAT_I2, IA64_INSTR_MIX},
    {"pack2.uss",  OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 0, 0), IA64_INSTR_FORMAT_I2, IA64_INSTR_PACK},
    {"pack2.sss",  OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 2, 0), IA64_INSTR_FORMAT_I2, IA64_INSTR_PACK},
    {"pack4.sss",  OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 2, 2, 0), IA64_INSTR_FORMAT_I2, IA64_INSTR_PACK},
    {"unpack1.h",  OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 2, 0, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_UNPACK},
    {"unpack2.h",  OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 0, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_UNPACK},
    {"unpack4.h",  OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 2, 0, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_UNPACK},
    {"unpack1.l",  OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 2, 2, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_UNPACK},
    {"unpack2.l",  OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 2, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_UNPACK},
    {"unpack4.l",  OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 2, 2, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_UNPACK},
    {"pmin1.u",    OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 2, 1, 0), IA64_INSTR_FORMAT_I2, IA64_INSTR_PMIN},
    {"pmax1.u",    OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 2, 1, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_PMAX},
    {"pmin2",      OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 3, 0), IA64_INSTR_FORMAT_I2, IA64_INSTR_PMIN},
    {"pmax2",      OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 2, 3, 1), IA64_INSTR_FORMAT_I2, IA64_INSTR_PMAX},
    {"psad1",      OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 2, 3, 2), IA64_INSTR_FORMAT_I2, IA64_INSTR_PSAD1},
    {"mux1", OpZaZbVeX2aX2bX2c (7, 0, 0, 0, 3, 2, 2), IA64_INSTR_FORMAT_I3, IA64_INSTR_MUX},
    {"mux2", OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 3, 2, 2), IA64_INSTR_FORMAT_I4, IA64_INSTR_MUX},
    {"pshr2",   OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 0, 2, 0), IA64_INSTR_FORMAT_I5, IA64_INSTR_PSHR},
    {"pshr4",   OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 0, 2, 0), IA64_INSTR_FORMAT_I5, IA64_INSTR_PSHR},
    {"shr",     OpZaZbVeX2aX2bX2c (7, 1, 1, 0, 0, 2, 0), IA64_INSTR_FORMAT_I5, IA64_INSTR_SHR},
    {"pshr2.u", OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 0, 0, 0), IA64_INSTR_FORMAT_I5, IA64_INSTR_PSHR},
    {"pshr4.u", OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 0, 0, 0), IA64_INSTR_FORMAT_I5, IA64_INSTR_PSHR},
    {"shr.u",   OpZaZbVeX2aX2bX2c (7, 1, 1, 0, 0, 0, 0), IA64_INSTR_FORMAT_I5, IA64_INSTR_SHR},
    {"pshr2",   OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 1, 3, 0), IA64_INSTR_FORMAT_I6, IA64_INSTR_PSHR},
    {"pshr4",   OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 1, 3, 0), IA64_INSTR_FORMAT_I6, IA64_INSTR_PSHR},
    {"pshr2.u", OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 1, 1, 0), IA64_INSTR_FORMAT_I6, IA64_INSTR_PSHR},
    {"pshr4.u", OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 1, 1, 0), IA64_INSTR_FORMAT_I6, IA64_INSTR_PSHR},
    {"pshl2",   OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 0, 0, 1), IA64_INSTR_FORMAT_I7, IA64_INSTR_PSHL},
    {"pshl4",   OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 0, 0, 1), IA64_INSTR_FORMAT_I7, IA64_INSTR_PSHL},
    {"shl",     OpZaZbVeX2aX2bX2c (7, 1, 1, 0, 0, 0, 1), IA64_INSTR_FORMAT_I7, IA64_INSTR_SHL},
    {"pshl2",   OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 3, 1, 1), IA64_INSTR_FORMAT_I8, IA64_INSTR_PSHL},
    {"pshl4",   OpZaZbVeX2aX2bX2c (7, 1, 0, 0, 3, 1, 1), IA64_INSTR_FORMAT_I8, IA64_INSTR_PSHL},
    {"popcnt",  OpZaZbVeX2aX2bX2c (7, 0, 1, 0, 1, 1, 2), IA64_INSTR_FORMAT_I9, IA64_INSTR_POPCNT},

    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
  };


#undef L
#undef bC
#undef bIh
#undef bTa
#undef bTag13
#undef bTb
#undef bVc
#undef bVe
#undef bWh
#undef bX
#undef bXb
#undef bX2
#undef bX2a
#undef bX2b_
#undef bX2c
#undef bX3
#undef bX6_
#undef bY
#undef bZa
#undef bZb
#undef mC
#undef mIh
#undef mTa
#undef mTag13
#undef mTb
#undef mVc
#undef mVe
#undef mWh
#undef mX
#undef mXb
#undef mX2
#undef mX2a
#undef mX2b
#undef mX2c
#undef mX3
#undef mX6
#undef mY
#undef mZa
#undef mZb
#undef OpZaZbVeX2aX2b
#undef OpZaZbVeX2aX2bX2c
#undef OpX2X
#undef OpX2XYa
#undef OpX2XYb
#undef OpX2TaTbYaC
#undef OpX3
#undef OpX3X6
#undef OpX3XbIhWh
#undef OpX3XbIhWhTag13


/* vim: set shiftwidth=2: */
