/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_opcodes_x.h"
#include "../../../utils/util_error.h"

#include "../ia64_instructions.h"
#include "../ia64_disassemble/ia64_disassemble_one.h"
#include "../../../simple_types.h"

/* instruction bit fields: */
#define bBtype(x)	(((t_uint64) ((x) & 0x7)) << 6)
#define bD(x)		(((t_uint64) ((x) & 0x1)) << 35)
#define bPa(x)		(((t_uint64) ((x) & 0x1)) << 12)
#define bPr(x)		(((t_uint64) ((x) & 0x3f)) << 0)
#define bVc(x)		(((t_uint64) ((x) & 0x1)) << 20)
#define bWha(x)		(((t_uint64) ((x) & 0x3)) << 33)
#define bX3(x)		(((t_uint64) ((x) & 0x7)) << 33)
#define bX6_(x)		(((t_uint64) ((x) & 0x3f)) << 27)

#define mBtype		bBtype (-1)
#define mD		bD (-1)
#define mPa		bPa (-1)
#define mPr		bPr (-1)
#define mVc             bVc (-1)
#define mWha		bWha (-1)
#define mX3             bX3 (-1)
#define mX6		bX6_ (-1)

#define OpX3X6(a,b,c)		(bOp (a) | bX3 (b) | bX6_(c)), \
				(mOp | mX3 | mX6)
#define OpVc(a,b)		(bOp (a) | bVc (b)), (mOp | mVc)
#define OpPaWhaD(a,b,c,d) \
	(bOp (a) | bPa (b) | bWha (c) | bD (d)), (mOp | mPa | mWha | mD)
#define OpBtypePaWhaD(a,b,c,d,e) \
	(bOp (a) | bBtype (b) | bPa (c) | bWha (d) | bD (e)), \
	(mOp | mBtype | mPa | mWha | mD)
#define OpBtypePaWhaDPr(a,b,c,d,e,f) \
	(bOp (a) | bBtype (b) | bPa (c) | bWha (d) | bD (e) | bPr (f)), \
	(mOp | mBtype | mPa | mWha | mD | mPr)

const t_ia64_opcode ia64_opcode_table_x[] =
  {
    {"break.x", OpX3X6 (0, 0, 0x00), IA64_INSTR_FORMAT_X1, IA64_INSTR_BREAK},
    {"nop.x",   OpX3X6 (0, 0, 0x01), IA64_INSTR_FORMAT_X1, IA64_INSTR_NOP},
    
    {"movl",	OpVc (6, 0), IA64_INSTR_FORMAT_X2, IA64_INSTR_MOVL},

#define BRL(a,b,c)	OpBtypePaWhaD (0xC, 0, a, b, c)
    {"brl.cond.sptk.few",	BRL (0, 0, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.sptk.few.clr",	BRL (0, 0, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.spnt.few",	BRL (0, 1, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.spnt.few.clr",	BRL (0, 1, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dptk.few",	BRL (0, 2, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dptk.few.clr",	BRL (0, 2, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dpnt.few",	BRL (0, 3, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dpnt.few.clr",	BRL (0, 3, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.sptk.many",	BRL (1, 0, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.sptk.many.clr",	BRL (1, 0, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.spnt.many",	BRL (1, 1, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.spnt.many.clr",	BRL (1, 1, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dptk.many",	BRL (1, 2, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dptk.many.clr",	BRL (1, 2, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dpnt.many",	BRL (1, 3, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.cond.dpnt.many.clr",	BRL (1, 3, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.sptk.few",		BRL (0, 0, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.sptk.few.clr",	BRL (0, 0, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.spnt.few",		BRL (0, 1, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.spnt.few.clr",	BRL (0, 1, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dptk.few",		BRL (0, 2, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dptk.few.clr",	BRL (0, 2, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dpnt.few",		BRL (0, 3, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dpnt.few.clr",	BRL (0, 3, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.sptk.many",		BRL (1, 0, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.sptk.many.clr",	BRL (1, 0, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.spnt.many",		BRL (1, 1, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.spnt.many.clr",	BRL (1, 1, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dptk.many",		BRL (1, 2, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dptk.many.clr",	BRL (1, 2, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dpnt.many",		BRL (1, 3, 0), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
    {"brl.dpnt.many.clr",	BRL (1, 3, 1), IA64_INSTR_FORMAT_X3, IA64_INSTR_BRL},
#undef BRL
#define BRL(a,b,c) OpPaWhaD (0xD, a, b, c)
    {"brl.call.sptk.few",	BRL (0, 0, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.sptk.few.clr",	BRL (0, 0, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.spnt.few",	BRL (0, 1, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.spnt.few.clr",	BRL (0, 1, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dptk.few",	BRL (0, 2, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dptk.few.clr",	BRL (0, 2, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dpnt.few",	BRL (0, 3, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dpnt.few.clr",	BRL (0, 3, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.sptk.many",	BRL (1, 0, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.sptk.many.clr",	BRL (1, 0, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.spnt.many",	BRL (1, 1, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.spnt.many.clr",	BRL (1, 1, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dptk.many",	BRL (1, 2, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dptk.many.clr",	BRL (1, 2, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dpnt.many",	BRL (1, 3, 0), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
    {"brl.call.dpnt.many.clr",	BRL (1, 3, 1), IA64_INSTR_FORMAT_X4, IA64_INSTR_BRL},
#undef BRL
    {"Unsupported", 0x0ULL, 0x0ULL, IA64_INSTR_FORMAT_UNSUPPORTED, IA64_INSTR_UNSUPPORTED}
  };

#undef bBtype
#undef bD
#undef bPa
#undef bPr
#undef bVc
#undef bWha
#undef bX3
#undef bX6_

#undef mBtype
#undef mD
#undef mPa
#undef mPr
#undef mVc
#undef mWha
#undef mX3
#undef mX6

#undef OpX3X6
#undef OpVc
#undef OpPaWhaD
#undef OpBtypePaWhaD
#undef OpBtypePaWhaDPr

/* vim: set shiftwidth=2: */

