/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_print.h"
#include "../../utils/util_error.h"

#include "ia64_opcodes.h"

#include "ia64_print/ia64_print_a.h"
#include "ia64_print/ia64_print_b.h"
#include "ia64_print/ia64_print_f.h"
#include "ia64_print/ia64_print_i.h"
#include "ia64_print/ia64_print_m.h"
#include "ia64_print/ia64_print_x.h"
#include "ia64_dummy.h"

void Ia64InstructionPrint(t_ins * ins, t_string outputstring) 
{
	Ia64PrintFunction_table[IA64OPCODEINFO(ins).instr_format](ins, outputstring);

	if(INS_NEXT(ins) == NULL || IA64INSINSTRGROUP(ins) != IA64INSINSTRGROUP(INS_NEXT(ins)))
	  strcat(outputstring, ";;");
}

void Ia64PrintUnsupported(t_ins * ins, t_string outputstring)
{
  FATAL(("Unsupported ia64 instruction; cannot print."));
}


void Ia64PrintDummy(t_ins * ins, t_string outputstring)
{
  if(IA64INSOPCODE(ins)==DUMMY_X_OPCODE)
    sprintf(outputstring, "dummy (X-type)");
  else if(IA64INSOPCODE(ins)==DUMMY_LIVE_OPCODE)
  {
    if(IA64INSIMM2(ins) == IA64_ALLOC_MISSING)
      sprintf(outputstring, "dummy (liveness/constprop) %s; alloc missing", IA64INSIMM1(ins)==0?"1":"2");
    else
      sprintf(outputstring, "dummy (liveness/constprop) %s; sol = %d; sof = %d", IA64INSIMM1(ins)==0?"1":"2", IA64INSIMM2(ins), IA64INSIMM3(ins));
  }
  else
  {
    FATAL(("unknown dummy opcode"));
  }
}
void Ia64PrintData(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "data");
}





/* 
 * bevat de lijst van alle print functions, 1 voor elk instructie formaat
 * de indexering in deze tabel moet dan ook volgens t_ia64_instr_format gebeuren...
 */
const Ia64PrintFunction Ia64PrintFunction_table [] =
{
	Ia64PrintA1, Ia64PrintA2, Ia64PrintA3, Ia64PrintA4, Ia64PrintA5,
	Ia64PrintA6, Ia64PrintA7, Ia64PrintA8, Ia64PrintA9, Ia64PrintA10,
	
	Ia64PrintI1, Ia64PrintI2, Ia64PrintI3, Ia64PrintI4, Ia64PrintI5,
	Ia64PrintI6, Ia64PrintI7, Ia64PrintI8, Ia64PrintI9, Ia64PrintI10,
	Ia64PrintI11, Ia64PrintI12, Ia64PrintI13, Ia64PrintI14, Ia64PrintI15,
	Ia64PrintI16, Ia64PrintI17, Ia64PrintUnsupported, Ia64PrintI19, Ia64PrintI20,
	Ia64PrintI21, Ia64PrintI22, Ia64PrintI23, Ia64PrintI24, Ia64PrintI25,
	Ia64PrintI26, Ia64PrintI27, Ia64PrintI28, Ia64PrintI29, 

	Ia64PrintM1, Ia64PrintM2, Ia64PrintM3, Ia64PrintM4, Ia64PrintM5,
	Ia64PrintM6, Ia64PrintM7, Ia64PrintM8, Ia64PrintM9, Ia64PrintM10,
	Ia64PrintM11, Ia64PrintM12, Ia64PrintM13, Ia64PrintM14, Ia64PrintM15,
	Ia64PrintM16, Ia64PrintM17, Ia64PrintM18, Ia64PrintM19, Ia64PrintM20,
	Ia64PrintM21, Ia64PrintM22, Ia64PrintM23, Ia64PrintM24, Ia64PrintM25,
	Ia64PrintM26, Ia64PrintM27, Ia64PrintM28, Ia64PrintM29, Ia64PrintM30,
	Ia64PrintM31, Ia64PrintM32, Ia64PrintM33, Ia64PrintM34, Ia64PrintM35,
	Ia64PrintM36, Ia64PrintM37, Ia64PrintM38, Ia64PrintM39, Ia64PrintM40,
	Ia64PrintM41, Ia64PrintM42, Ia64PrintM43, Ia64PrintM44, Ia64PrintM45,
	Ia64PrintM46, 
	
	Ia64PrintB1, Ia64PrintB2, Ia64PrintB3, Ia64PrintB4, Ia64PrintB5,
	Ia64PrintB6, Ia64PrintB7, Ia64PrintB8, Ia64PrintB9,	

	Ia64PrintF1, Ia64PrintF2, Ia64PrintF3, Ia64PrintF4, Ia64PrintF5,
	Ia64PrintF6, Ia64PrintF7, Ia64PrintF8, Ia64PrintF9, Ia64PrintF10,
	Ia64PrintF11, Ia64PrintF12, Ia64PrintF13, Ia64PrintF14, Ia64PrintF15,

	Ia64PrintX1, Ia64PrintX2, Ia64PrintX3, Ia64PrintX4, 

	Ia64PrintUnsupported,
	Ia64PrintDummy,
	Ia64PrintData
	
};

/* vim: set shiftwidth=2: */
