/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_print_a.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"
#include "../ia64_opcodes/ia64_opcodes_a.h"

/* A1: Integer Alu-Register-Register */
void Ia64PrintA1(t_ins * ins, t_string outputstring)
{
  /* bevat het opcodegedeelte van de instructie */
  t_uint64 instr = ia64_opcode_table_a[IA64INSOPCODE(ins)].opcode;
  
  if(  ((GETBITS64(instr,29,4)==0)&&(GETBITS64(instr,27,2)==1)) 
    || ((GETBITS64(instr,29,4)==1) && (GETBITS64(instr,27,2)==0)) ) 
    sprintf(outputstring, "(p%d) %s r%d = r%d, r%d, 1",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSOUT1(ins) - IA64_GR,
      IA64INSIN1(ins) - IA64_GR,
      IA64INSIN2(ins) - IA64_GR
    );
  else
    sprintf(outputstring, "(p%d) %s r%d = r%d, r%d",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSOUT1(ins) - IA64_GR,
      IA64INSIN1(ins) - IA64_GR,
      IA64INSIN2(ins) - IA64_GR
    );
}

/* A2: Integer Alu-shift left and add */
void Ia64PrintA2(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIN2(ins) - IA64_GR
  );
}


/* A3: Integer Alu-immediate8 register */
void Ia64PrintA3(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = %lld, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIN1(ins) - IA64_GR
  );
}


/* A4: Integer Alu-add immediate14*/
void Ia64PrintA4(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = %lld, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIN1(ins) - IA64_GR
  );
}

/* A5: Integer Alu-add immediate22*/
void Ia64PrintA5(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = %lld, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIN1(ins) - IA64_GR
  );
}

/* A6: Integer compare-register-register*/
void Ia64PrintA6(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s p%d, p%d = r%d, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_PR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* A7: Integer comparetozero-register*/
void Ia64PrintA7(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s p%d, p%d = r0, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_PR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* A8: Integer compare-immediate-register*/
void Ia64PrintA8(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s p%d, p%d = %lld, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_PR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIMM1(ins),
    IA64INSIN1(ins) - IA64_GR
  );
}


/* A9: Multimedia ALU*/
void Ia64PrintA9(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins)- IA64_GR,
    IA64INSIN1(ins)- IA64_GR,
    IA64INSIN2(ins)- IA64_GR
  );
}


/* A10: Multimedia shift and add*/
void Ia64PrintA10(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIN2(ins) - IA64_GR
  );
}


/* vim: set shiftwidth=2: */

