/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_print_b.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"


/* B1: IP-Relative Branch */
void Ia64PrintB1(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}


/* B2: IP-Relative Counted Branch */
void Ia64PrintB2(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}


/* B3: IP-Relative Call */
void Ia64PrintB3(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s b%d = %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_BR,
    IA64INSIMM1(ins)
  );
}


/* B4: Indirect Branch */
void Ia64PrintB4(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s b%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_BR
  );
}


/* B5: Indirect Call */
void Ia64PrintB5(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s b%d = b%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_BR,
    IA64INSIN1(ins) - IA64_BR
  );
}


/* B6: IP-Relative Predict */
void Ia64PrintB6(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld, %d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins),
    IA64INSIMM2(ins)
  );
}

/* B7: Indirect Predict */
void Ia64PrintB7(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s b%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_BR,
    IA64INSIMM1(ins)
  );
}


/* B8: Miscellaneous (M-Unit) */
void Ia64PrintB8(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc
  );
}


/* B9: Break/Nop (B-Unit) */
void Ia64PrintB9(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}

/* vim: set shiftwidth=2: */
