/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_print_f.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"


/* F1: Floating-point multiply add */
void Ia64PrintF1(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = f%d, f%d, f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIN2(ins) - IA64_FR,
    IA64INSIN3(ins) - IA64_FR
  );
}

/* F2: Fixed-point multiply add */
void Ia64PrintF2(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = f%d, f%d, f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIN2(ins) - IA64_FR,
    IA64INSIN3(ins) - IA64_FR
  );
}

/* F3: Parallel Floating-point select */
void Ia64PrintF3(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = f%d, f%d, f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIN2(ins) - IA64_FR,
    IA64INSIN3(ins) - IA64_FR
  );
}

/* F4: Floating-point compare */
void Ia64PrintF4(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s p%d, p%d = f%d, f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_PR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIN2(ins) - IA64_FR
  );
}


/* F5: Floating-point class */
void Ia64PrintF5(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s p%d, p%d = f%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_PR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIMM1(ins)
  );
}


/* F6: Floating-point reciprocal approxiimation */
void Ia64PrintF6(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d, p%d = f%d, f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIN2(ins) - IA64_FR
  );
}

/* F7: Floating-point reciprocal square root approximation */
void Ia64PrintF7(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d, p%d = f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_FR
  );
}

/* F8: minimum/maximum and parallel compare */
void Ia64PrintF8(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = f%d, f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIN2(ins) - IA64_FR
  );
}

/* F9: merge and logical */
void Ia64PrintF9(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = f%d, f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIN2(ins) - IA64_FR
  );
}

/* F10: convert floating-point to fixed-point */
void Ia64PrintF10(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_FR
  );
}


/* F11: convert fixed-point to floating-point */
void Ia64PrintF11(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_FR
  );
}

/* F12: floating-point set controls*/
void Ia64PrintF12(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld %d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins),
    IA64INSIMM2(ins)
  );
}


/* F13: floating-point clear flags*/
void Ia64PrintF13(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc
  );
}


/* F14: floating-point check flags*/
void Ia64PrintF14(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}

/* F15: break/nop*/
void Ia64PrintF15(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}



/* vim: set shiftwidth=2: */
