/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_print_i.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"
#include "../ia64_opcodes/ia64_opcodes_i.h"

/* I1: Multimedia multiply and shift */
void Ia64PrintI1(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}

/* I2: Multimedia multiply/mix/pack/unpack */
void Ia64PrintI2(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* I3: Multimedia mux1 */
void Ia64PrintI3(t_ins * ins, t_string outputstring)
{
  t_string mbtype;
  switch(IA64INSIMM1(ins))
  {
    case 0:
      mbtype="@brcst";
      break;
    case 8:
      mbtype="@mix";
      break;
    case 9:
      mbtype="@shuf";
      break;
    case 0xA:
      mbtype="@alt";
      break;
    case 0xB:
      mbtype="@rev";
      break;
    default:
      FATAL(("unknown mtype I3"));
  }
  sprintf(outputstring, "(p%d) %s r%d = r%d, %s",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    mbtype
  );
}


/* I4: Multimedia mux2 */
void Ia64PrintI4(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}

/* I5: Multimedia mux2 */
void Ia64PrintI5(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* I6: Multimedia shift right fixec */
void Ia64PrintI6(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* I7: Multimedia shift left variable */
void Ia64PrintI7(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* I8: Multimedia shift left fixed */
void Ia64PrintI8(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* I9: Population count */
void Ia64PrintI9(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR
  );
}

/* I10: shift right pair */
void Ia64PrintI10(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* I11: extract */
void Ia64PrintI11(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld, %d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIMM2(ins)
  );
}

/* I12: zero and deposit */
void Ia64PrintI12(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld, %d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIMM2(ins)
  );
}


/* I13: zero and deposit immediate*/
void Ia64PrintI13(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = %lld, %d, %d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIMM2(ins),
    IA64INSIMM3(ins)
  );
}


/* I14: deposit immediate*/
void Ia64PrintI14(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = %lld, r%d, %d, %d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM2(ins),
    IA64INSIMM3(ins)
  );
}


/* I15: deposit*/
void Ia64PrintI15(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d, %lld, %d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIMM2(ins)
  );
}


/* I16: test bit*/
void Ia64PrintI16(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s p%d, p%d = r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_PR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}

/* I17: test naT*/
void Ia64PrintI17(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s p%d, p%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_PR,
    IA64INSOUT2(ins) - IA64_PR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* I19: break\nop*/
void Ia64PrintI19(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}

/* I20: integer speculation unit*/
void Ia64PrintI20(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* I21: move to BR*/
void Ia64PrintI21(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s b%d = r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_BR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* I22: move from BR*/
void Ia64PrintI22(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = b%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_BR
  );
}


/* I23: move to predicates -register*/
void Ia64PrintI23(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s pr = r%d  %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}

/* I24: move to predicates - immediate44*/
void Ia64PrintI24(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s pr.rot = %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}


/* I25: move from predicates ip*/
void Ia64PrintI25(t_ins * ins, t_string outputstring)
{
  /* bevat het opcodegedeelte van de instructie */
  t_uint64 instr = ia64_opcode_table_i[IA64INSOPCODE(ins)].opcode;

  if (GETBITS64(instr,27,6) == 0x30)  
    sprintf(outputstring, "(p%d) %s r%d = ip",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSOUT1(ins) - IA64_GR
    );

  if (GETBITS64(instr,27,6) == 0x33)  
    sprintf(outputstring, "(p%d) %s r%d = pr",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSOUT1(ins) - IA64_GR
    );
}

/* I26: move to AR - register*/
void Ia64PrintI26(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s ar%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_AR,
    IA64INSIN1(ins) - IA64_GR
  );
}

/* I27: move to AR - immediate8*/
void Ia64PrintI27(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s ar%d = %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_AR,
    IA64INSIMM1(ins)
  );
}


/* I28: move from AR - immediate8*/
void Ia64PrintI28(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = ar%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_AR
  );
}

/* I29: sign/zero extend/compute zero index */
void Ia64PrintI29(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR
  );
}









/* vim: set shiftwidth=2: */
