/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_print_m.h"
#include "../ia64_disassemble/ia64_disassemble_one_m.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"
#include "../ia64_opcodes/ia64_opcodes_m.h"

/* M1: Integer Load */
void Ia64PrintM1(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = [r%d]",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M2: Integer Load - Increment by Register */
void Ia64PrintM2(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = [r%d], r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}

/* M3: Integer Load - Increment by immediate */
void Ia64PrintM3(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = [r%d], %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}



/* M4: Integer Store */
void Ia64PrintM4(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s [r%d] = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN2(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M5: Integer Store - Increment by immediate */
void Ia64PrintM5(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s [r%d] = r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN2(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}

/* M6: Floating-point Load */
void Ia64PrintM6(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = [r%d]",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M7: Floating-point Load - Increment by Register */
void Ia64PrintM7(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = [r%d], r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}

/* M8: Floating-point Load - Increment by immediate */
void Ia64PrintM8(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = [r%d], %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}



/* M9: Floating-point Store */
void Ia64PrintM9(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s [r%d] = f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN2(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_FR
  );
}


/* M10: Floating-point Store - Increment by immediate */
void Ia64PrintM10(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s [r%d] = f%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN2(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIMM1(ins)
  );
}




/* M11: Floating-point Load Pair */
void Ia64PrintM11(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d, f%d = [r%d]",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSOUT2(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M12: Floating-point Load Pair - Increment by Immediate*/
void Ia64PrintM12(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d, f%d = [r%d], %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSOUT2(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}



/* M13: Line Prefetch */
void Ia64PrintM13(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s [r%d]",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR
  );
}

/* M14: Line Prefetch - Increment by Register */
void Ia64PrintM14(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s [r%d], r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* M15: Line Prefetch - Increment by Immediate */
void Ia64PrintM15(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s [r%d], %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* M16: Exchange/Compare and Exchange */
void Ia64PrintM16(t_ins * ins, t_string outputstring)
{
  /* bevat het opcodegedeelte van de instructie */
  t_uint64 instr = ia64_opcode_table_m[IA64INSOPCODE(ins)].opcode;
    
  if(GETBITS64(instr, 30, 6)<0x08)
    sprintf(outputstring, "(p%d) %s r%d = [r%d], r%d, ar.ccv",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSOUT1(ins) - IA64_GR,
      IA64INSIN1(ins) - IA64_GR,
      IA64INSIN2(ins) - IA64_GR
    );
  else
    sprintf(outputstring, "(p%d) %s r%d = [r%d], r%d",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSOUT1(ins) - IA64_GR,
      IA64INSIN1(ins) - IA64_GR,
      IA64INSIN2(ins) - IA64_GR
    );
  
}


/* M17: Fetch and Add - Immediate */
void Ia64PrintM17(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = [r%d], %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}

/* M18: Set FR */
void Ia64PrintM18(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_FR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M19: Get FR */
void Ia64PrintM19(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_FR
  );
}

/* M20: Integer Speculation Check (M-Unit) */
void Ia64PrintM20(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}

/* M21: Floating-point Speculation Check */
void Ia64PrintM21(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIMM1(ins)
  );
}


/* M22: Integer Advanced Load Check */
void Ia64PrintM22(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* M23: Floating-point Advanced Load Check */
void Ia64PrintM23(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_FR,
    IA64INSIMM1(ins)
  );
}

/* M24: Sync/Fence/Serialize/ALAT Control */
void Ia64PrintM24(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc
  );
}


/* M25: RSE Control */
void Ia64PrintM25(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc
  );
}

/* M26: Integer ALAT Entry Invalidate */
void Ia64PrintM26(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR
  );
}

/* M27: Floating-point ALAT Entry Invalidate */
void Ia64PrintM27(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s f%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_FR
  );
}

/* M28: Flush Cache/Purge Translation Cache Entry */
void Ia64PrintM28(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR
  );
}

/* M29: Move to AR - Register (M-Unit) */
void Ia64PrintM29(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s ar%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_AR,
    IA64INSIN1(ins) - IA64_GR
  );
}

/* M30: Move to AR - Immediate8 (M-Unit) */
void Ia64PrintM30(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s ar%d = %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_AR,
    IA64INSIMM1(ins)
  );
}


/* M31: Move from AR (M-Unit) */
void Ia64PrintM31(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = ar%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_AR
  );
}


/* M32: Move to CR */
void Ia64PrintM32(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s cr%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_CR,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M33: Move from CR */
void Ia64PrintM33(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = cr%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_CR
  );
}


/* M34: Allocate Register Stack Frame */
void Ia64PrintM34(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = ar.pfs, %lld, %d, %d", 
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIMM1(ins),
    IA64INSIMM2(ins),
    IA64INSIMM3(ins)
    );
}


/* M35: Move to PSR */
void Ia64PrintM35(t_ins * ins, t_string outputstring)
{
  /* bevat het opcodegedeelte van de instructie */
  t_uint64 instr = ia64_opcode_table_m[IA64INSOPCODE(ins)].opcode;
    
  if(GETBITS64(instr, 27, 6) == 0x2D)
    sprintf(outputstring, "(p%d) %s psr.l = r%d",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSIN1(ins) - IA64_GR
    );
  else
    sprintf(outputstring, "(p%d) %s psr.um = r%d",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSIN1(ins) - IA64_GR
    );
}


/* M36: Move from PSR */
void Ia64PrintM36(t_ins * ins, t_string outputstring)
{
  /* bevat het opcodegedeelte van de instructie */
  t_uint64 instr = ia64_opcode_table_m[IA64INSOPCODE(ins)].opcode;
    
  if(GETBITS64(instr, 27, 6) == 0x25)
    sprintf(outputstring, "(p%d) %s r%d = psr",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSIN1(ins) - IA64_GR
    );
  else
    sprintf(outputstring, "(p%d) %s r%d = psr.um",
      IA64INSPRED(ins) - IA64_PR,
      IA64OPCODEINFO(ins).desc,
      IA64INSIN1(ins) - IA64_GR
    );
}


/* M37: Break/Nop (M-Unit) */
void Ia64PrintM37(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}

/* M38: Probe - Register */
void Ia64PrintM38(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* M39: Probe - Immediate2 */
void Ia64PrintM39(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* M40: Probe Fault - Immediate2 */
void Ia64PrintM40(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d, %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* M41: Translation Cache Insert */
void Ia64PrintM41(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M42: Move to Indirect Register/Translation Register Insert */
void Ia64PrintM42(t_ins * ins, t_string outputstring)
{
  char *indirectReg = "";

  /* bevat het opcodegedeelte van de instructie */
  t_uint64 instr = ia64_opcode_table_m[IA64INSOPCODE(ins)].opcode;
  
  switch(GETBITS64(instr, 27, 6))
  {
    case 0x00:
      indirectReg = "rr";
      break;
    case 0x01:
      indirectReg = "dbr";
      break;
    case 0x02:
      indirectReg = "ibr";
      break;
    case 0x03:
      indirectReg = "pkr";
      break;
    case 0x04:
      indirectReg = "pmc";
      break;
    case 0x05:
      indirectReg = "pmd";
      break;
    case 0x0E:
      indirectReg = "dtr";
      break;
    case 0x0F:
      indirectReg = "itr";
      break;
  } 

  sprintf(outputstring, "(p%d) %s %s[r%d] = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    indirectReg,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* M43: Move from Indirect Register */
void Ia64PrintM43(t_ins * ins, t_string outputstring)
{
  char *indirectReg = "";

  /* bevat het opcodegedeelte van de instructie */
  t_uint64 instr = ia64_opcode_table_m[IA64INSOPCODE(ins)].opcode;
    
  switch(GETBITS64(instr, 27, 6))
  {
    case 0x10:
      indirectReg = "rr";
      break;
    case 0x11:
      indirectReg = "dbr";
      break;
    case 0x12:
      indirectReg = "ibr";
      break;
    case 0x13:
      indirectReg = "pkr";
      break;
    case 0x14:
      indirectReg = "pmc";
      break;
    case 0x15:
      indirectReg = "pmd";
      break;
    case 0x17:
      indirectReg = "cpuid";
      break;
  }

  sprintf(outputstring, "(p%d) %s r%d = %s[r%d]",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    indirectReg,
    IA64INSIN1(ins) - IA64_GR
  );
}


/* M44: Set/Rest User/System Mask */
void Ia64PrintM44(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s %lld",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}


/* M45: Translation Purge */
void Ia64PrintM45(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d, r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIN1(ins) - IA64_GR,
    IA64INSIN2(ins) - IA64_GR
  );
}


/* M46: Translation Access */
void Ia64PrintM46(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = r%d",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIN1(ins) - IA64_GR
  );
}







/* vim: set shiftwidth=2: */
