/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_print_x.h"
#include "../../../utils/util_error.h"


#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"


/* X1: Break/Nop (X-Unit) */
void Ia64PrintX1(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s 0x%llx",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}

/* X2: Move Long Immediate64 */
void Ia64PrintX2(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s r%d = 0x%llx",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_GR,
    IA64INSIMM1(ins)
  );
}


/* X3: Long Branch */
void Ia64PrintX3(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s 0x%llx",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSIMM1(ins)
  );
}


/* X4: Long Call */
void Ia64PrintX4(t_ins * ins, t_string outputstring)
{
  sprintf(outputstring, "(p%d) %s b%d = 0x%llx",
    IA64INSPRED(ins) - IA64_PR,
    IA64OPCODEINFO(ins).desc,
    IA64INSOUT1(ins) - IA64_BR,
    IA64INSIMM1(ins)
  );
}



/* vim: set shiftwidth=2: */
