/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_profiler.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_iterators.h"
#include "ia64_prograph.h"
#include "ia64_opcode_label.h"

//#define GCC
#define AANTAL_GCC 3


void Ia64MaakFiguurtje(t_cfg * cfg)
{
  FILE * out;
  static char buffer1[512];
  static char buffer2[512];
  t_bbl * bbl;
  t_edge * edge;
  //printf("in figuur\n");
  out = fopen("figuur.dot","w");
  VERIFY_FASSERT(out,("Could not open figuur.dot for writing!"));
  fprintf(out, "digraph \"%llx\" {\n\tnode [shape=box]\n",0x0000000000000000);
 
  CFG_FOREACH_BBL(cfg, bbl)
  {
    if(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)
      if(BBL_PROFILE_INFO(bbl)>0)
      {
	Ia64InstructionPrint(bbl->ins_first, buffer1);
	Ia64InstructionPrint(bbl->ins_last, buffer2);
	fprintf(out, "\t\"%llx\" [label=\"%d - (0x%llx)\\n%s\\n...\\n%s\"]\n",
	    BBL_OLD_ADDRESS(bbl), BBL_PROFILE_INFO(bbl),BBL_OLD_ADDRESS(bbl), buffer1, buffer2);
      }
  }
  CFG_FOREACH_BBL(cfg,bbl)
  {
    if(!IsHellNode(bbl) && 
	bbl->ins_first && 
	IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)

      BBL_FOREACH_SUCC_EDGE(bbl,(t_cfg_edge*)edge)
      {
	if ((EDGE_CAT(edge)==ET_JUMP || 
	      EDGE_CAT(edge)==ET_IPJUMP || 
	      EDGE_CAT(edge)==ET_CALL)&& !Ia64IsReturn(bbl->ins_last)&&!Ia64IsCheck(bbl->ins_last))
	{
	    if(BBL_PROFILE_INFO(bbl)>0)
	  fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d-%d\",style=solid,color=\"black\"];\n",BBL_OLD_ADDRESS(bbl), EDGE_TARGET(edge),EDGE_TAKEN(edge), EDGE_EVALUATED(edge));
	}
	else if((EDGE_CAT(edge)==ET_UNKNOWN||EDGE_CAT(edge)==ET_SWITCH||EDGE_CAT(edge)==ET_IPUNKNOWN))
	{
	  FATAL(("found one"));
	}
	else if(EDGE_CAT(edge)==ET_FUNLINK&&BBL_PROFILE_INFO(bbl)!=0)
	{
	    fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d-%d\",style=dashed,color=\"black\"];\n",BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge)), EDGE_TARGET(edge), EDGE_TAKEN(edge), EDGE_EVALUATED(edge));
	}
	else if(EDGE_CAT(edge)==ET_IPFUNLINK&&BBL_PROFILE_INFO(bbl)!=0)
	{
	    fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d-%d\",style=dashed,color=\"black\"];\n",BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge)), EDGE_TARGET(edge), EDGE_TAKEN(edge), EDGE_EVALUATED(edge));
	}
	else if(EDGE_CAT(edge)==ET_SWILINK&&BBL_PROFILE_INFO(bbl)!=0)
	{
	    fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d-%d\",style=dashed,color=\"black\"];\n",BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge)), EDGE_TARGET(edge), EDGE_TAKEN(edge), EDGE_EVALUATED(edge));
	}
	else if(EDGE_CAT(edge)==ET_IPFALLTHRU&&BBL_PROFILE_INFO(bbl)!=0)
	{
	    fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d-%d\",style=dotted,color=\"black\"];\n",BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge)), EDGE_TARGET(edge), EDGE_TAKEN(edge), EDGE_EVALUATED(edge));
	}
	else if(EDGE_CAT(edge)==ET_FALLTHROUGH&&BBL_PROFILE_INFO(bbl)!=0)
	{
	    fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d-%d\",style=dotted,color=\"black\"];\n",BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge)), EDGE_TARGET(edge), EDGE_TAKEN(edge), EDGE_EVALUATED(edge));
	}
	else if(EDGE_CAT(edge)==ET_IPSWILINK&&BBL_PROFILE_INFO(bbl)!=0)
	{
	    fprintf(out, "\t\"%llx\" -> \"%llx\" [label=\"%d-%d\",style=dashed,color=\"black\"];\n",BBL_OLD_ADDRESS(CFG_EDGE_HEAD(edge)), EDGE_TARGET(edge), EDGE_TAKEN(edge), EDGE_EVALUATED(edge));
	}
      }
    
  }
  fprintf(out, "}\n");
  fclose(out);
}


void Ia64ProfilePrintHotBbls(t_cfg * cfg)
{
  t_bbl *bbl;
  t_uint64 threshold=0;
  
  CFG_FOREACH_BBL(cfg,bbl){
    if(bbl->nins != 0 && 
	(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)) 
    {
      threshold=BBL_PROFILE_INFO(bbl)>threshold?BBL_PROFILE_INFO(bbl):threshold;
    }
  }

  threshold=threshold*0.8;
  
  CFG_FOREACH_BBL(cfg,bbl){
    if(BBL_PROFILE_INFO(bbl)>threshold)
    {
      printf("\nwerd %d keer uitgevoerd\n", BBL_PROFILE_INFO(bbl));
      BblPrint(bbl,2);
    }
  }
  
}


/* Neem profileringsinformatie op in basic blocks */
void Ia64VerwerkProfile(t_object * obj)
{
  FILE * fp_tellers_bbl = fopen("bblvoer","r");
  FILE * fp_tellers_edg = fopen("edgvoer","r");
  FILE * fp_adressen = fopen("uitvoer.bbl","r");
  //meting verhouding statisch vs dynamisch aantal uitgevoerde instructies
//  FILE * percbestand = fopen("percentage","w+");
  //einde
  FILE * fp_nieuwoud;
  t_uint32 i;
  t_uint64 adres;
  t_uint64 teller, teller2, voorpro, napro, doel;
  t_bbl * bbl;
  t_cfg * cfg = SECTION_CFG(SectionGetFromObjectByName(obj, ".init"));
  t_edge * edge, *ft_edge;
  t_bool gemeten;
  char s [16];
  
  //meting niet uitgevoerde (+reeds verwijderde)
//  int totaal_aantal_instructies=0;
// int aantal_niet_uitgevoerde_instructies=0;
  //einde meting

  VERIFY_FASSERT(fp_tellers_bbl,("Could not open bblvoer!"));
  VERIFY_FASSERT(fp_tellers_edg,("Could not open edgvoer!"));
  VERIFY_FASSERT(fp_adressen,("Could not open uitvoer.bbl!"));

  



  //FATAL(("start\n"));
  CFG_FOREACH_BBL(cfg,bbl){
    if(bbl->nins != 0 && 
	(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)) 
    {
#ifndef GCC
      fscanf(fp_adressen,"%d\t%llx\n",&i,&adres);
      if(adres != BBL_OLD_ADDRESS(bbl))
	FATAL(("profileringsinformatie lijkt niet afkomstig van dit programma."));
#endif

      fread(&teller, 8, 1,fp_tellers_bbl);
      //meting niet uitgevoerde
//      if(teller==0)
//	aantal_niet_uitgevoerde_instructies += BBL_NINS(bbl);
//      totaal_aantal_instructies += BBL_NINS(bbl);
      //einde meting
  
      //meting verhouding statisch vs dynamisch aantal uitgevoerde instructies
//      fprintf(percbestand, "%d\t%d\t%d\n", teller, teller * BBL_NINS(bbl), BBL_NINS(bbl));
      //einde
      BBL_PROFILE_INFO(bbl)=teller;
    }
  }
  
  //meting verhouding statisch vs dynamisch aantal uitgevoerde instructies
//  fclose(percbestand);
//  FATAL(("stop"));
  //einde
//  meting niet uitgevoerde
//  printf("%d %d\n", aantal_niet_uitgevoerde_instructies, totaal_aantal_instructies);
//  FATAL(("stop"));
//  einde meting
  

  CFG_FOREACH_BBL(cfg,bbl)
  {
    gemeten=FALSE;
    if(!IsHellNode(bbl) && 
	bbl->ins_first && 
	IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)
    {

      BBL_FOREACH_SUCC_EDGE(bbl,T_CFG_EDGE(edge))
      {
	if ((EDGE_CAT(edge)==ET_JUMP || 
	      EDGE_CAT(edge)==ET_IPJUMP || 
	      EDGE_CAT(edge)==ET_CALL)&& !Ia64IsReturn(bbl->ins_last)&&!Ia64IsCheck(bbl->ins_last))
	{
	  gemeten=TRUE;
	  //printf("testing\n");
	  fread(&teller,8,1,fp_tellers_edg);
	  fread(&teller2,8,1,fp_tellers_edg);
	  fread(&doel,8,1,fp_tellers_edg);
	  EDGE_TAKEN(edge)=teller;
	  EDGE_EVALUATED(edge)=teller2;
	  if(doel>0x4000000000000000ULL&&doel<0x5000000000000000ULL)
	  {
#ifndef GCC
	    fp_nieuwoud=fopen("nieuwoud","r");
	    VERIFY_FASSERT(fp_nieuwoud,("Could not open nieuwoud!"));
	    do 
	    {
	      fscanf(fp_nieuwoud, "%llx\t%llx",&napro, &voorpro); 
	    }while (napro != doel||voorpro==0);
	    fclose(fp_nieuwoud);

	    /*if(BBL_OLD_ADDRESS(bbl)==0x40000000000068c0)
	      {
	      printf("%lld\n",voorpro);
	      printf("%lld\n",doel);
	      FATAL(("stop"));
	      }*/
	    EDGE_TARGET(edge)=voorpro;
#endif
#ifdef GCC
	    EDGE_TARGET(edge)=doel;
#endif
	    //printf("testing2\n");

	  }
	  else 
	  {
	    doel = (INS_CADDRESS(bbl->ins_last)&0xfffffffffffffff0ull)+IA64INSIMM1(bbl->ins_last);
	    EDGE_TARGET(edge)=doel;
	  }
	}
	BBL_FOREACH_SUCC_EDGE(bbl,T_CFG_EDGE(ft_edge))
	{
	  if (EDGE_CAT(ft_edge)==ET_FALLTHROUGH ||
	      EDGE_CAT(ft_edge)==ET_IPFALLTHRU)
	  {
	    if(gemeten)
	    {
	      EDGE_TAKEN(ft_edge)=teller2-teller;
	      EDGE_EVALUATED(ft_edge)=teller2;
	      EDGE_TARGET(ft_edge)=BBL_OLD_ADDRESS(CFG_EDGE_TAIL(ft_edge));
	    }
	    else if(!Ia64IsReturn(bbl->ins_last))
	    {
	      EDGE_TAKEN(ft_edge)=BBL_PROFILE_INFO(bbl);
	      EDGE_EVALUATED(ft_edge)=BBL_PROFILE_INFO(bbl);
	      EDGE_TARGET(ft_edge)=BBL_OLD_ADDRESS(CFG_EDGE_TAIL(ft_edge));
	    }
	    else
	    {
	      EDGE_TAKEN(ft_edge)=BBL_PROFILE_INFO(CFG_EDGE_TAIL(ft_edge))<=BBL_PROFILE_INFO(bbl)?BBL_PROFILE_INFO(CFG_EDGE_TAIL(ft_edge)):BBL_PROFILE_INFO(bbl);
	      EDGE_EVALUATED(ft_edge)=BBL_PROFILE_INFO(bbl);
	      EDGE_TARGET(ft_edge)=BBL_OLD_ADDRESS(CFG_EDGE_TAIL(ft_edge));
	    }

	  }
	  if (EDGE_CAT(ft_edge)==ET_SWILINK ||
	      EDGE_CAT(ft_edge)==ET_IPSWILINK)
	  {
	    EDGE_TAKEN(ft_edge)=BBL_PROFILE_INFO(CFG_EDGE_TAIL(ft_edge));
	    EDGE_EVALUATED(ft_edge)=BBL_PROFILE_INFO(CFG_EDGE_TAIL(ft_edge));
	    EDGE_TARGET(ft_edge)=BBL_OLD_ADDRESS(CFG_EDGE_TAIL(ft_edge));
	  }
	  if (EDGE_CAT(ft_edge)==ET_FUNLINK ||
	      EDGE_CAT(ft_edge)==ET_IPFUNLINK)
	  {
	    int tellertje=BBL_PROFILE_INFO(BBL_NEXT_IN_FUN(CFG_EDGE_TAIL(ft_edge)));
	    t_cfg_edge * edge2;
	    BBL_FOREACH_PRED_EDGE(BBL_NEXT_IN_FUN(CFG_EDGE_TAIL(ft_edge)),edge2)
	      if(EDGE_CAT(edge2)!=ET_FUNLINK&&EDGE_CAT(edge2)!=ET_IPFUNLINK)
		tellertje -= EDGE_TAKEN(edge2);
	    BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(ft_edge),edge2)
	      if(EDGE_CAT(edge2)!=ET_FUNLINK&&EDGE_CAT(edge2)!=ET_IPFUNLINK)
		tellertje -= EDGE_TAKEN(edge2);
	    EDGE_TAKEN(ft_edge)=tellertje;
	    EDGE_EVALUATED(ft_edge)=teller2;//BBL_PROFILE_INFO(BBL_NEXT_IN_FUN(CFG_EDGE_TAIL(ft_edge)));
	    EDGE_TARGET(ft_edge)=BBL_OLD_ADDRESS(CFG_EDGE_TAIL(ft_edge));
	    /*int y = BBL_PROFILE_INFO(bbl)<BBL_PROFILE_INFO(BBL_NEXT_IN_FUN(CFG_EDGE_TAIL(ft_edge)))? 
	      BBL_PROFILE_INFO(bbl): BBL_PROFILE_INFO(BBL_NEXT_IN_FUN(CFG_EDGE_TAIL(ft_edge)));
	      int z = teller < y ? teller : y;
	      EDGE_TAKEN(ft_edge)=z;//BBL_PROFILE_INFO(BBL_NEXT_IN_FUN(CFG_EDGE_TAIL(ft_edge)));
	      EDGE_EVALUATED(ft_edge)=z;//BBL_PROFILE_INFO(BBL_NEXT_IN_FUN(CFG_EDGE_TAIL(ft_edge)));
	      EDGE_TARGET(ft_edge)=BBL_OLD_ADDRESS(CFG_EDGE_TAIL(ft_edge));*/
	  }
	}

      }
    }
  }
  CFG_FOREACH_EDGE(cfg, T_CFG_EDGE(edge))
  {
    int teller;
    t_cfg_edge * edge2;
    if(CFG_EDGE_HEAD(edge)->ins_last && IA64INSOPCODE(CFG_EDGE_HEAD(edge)->ins_last) >= br_cloop_sptk_few_B_8000000140 && IA64INSOPCODE(CFG_EDGE_HEAD(edge)->ins_last)<= br_cloop_dpnt_many_clr_B_8e00001140 && BBL_PROFILE_INFO(CFG_EDGE_HEAD(edge))>0)
    {
      if(EDGE_CAT(edge)==ET_FALLTHROUGH)
      {
	//InsPrint(CFG_EDGE_HEAD(edge)->ins_last,0);
	teller=BBL_PROFILE_INFO(CFG_EDGE_TAIL(edge));
	BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),edge2)
	  teller -= EDGE_TAKEN(edge2);
	EDGE_TAKEN(edge)+=teller;//printf("%d\n",teller);
	//printf("fallthrough: %d\n",EDGE_TAKEN(edge));
      }

      if(EDGE_CAT(edge)==ET_JUMP)
      {
	//InsPrint(CFG_EDGE_HEAD(edge)->ins_last,0);
	//printf("jump: %d\n",EDGE_TAKEN(edge));
	EDGE_TAKEN(edge)-=teller;
	//printf("jump: %d\n",EDGE_TAKEN(edge));
      }
    }
  }
  //FATAL(("stop"));
}

static int gcc_counter;
static int aantal_bbl_totaal;

/*Voeg profilering toe per edge (en per bbl op einde)*/
t_uint32 Ia64MakeProfile(t_object * obj)
{
  t_bbl * bbl;
  t_cfg_edge * edge;
  t_cfg * cfg = SECTION_CFG(SectionGetFromObjectByName(obj, ".init"));
  int aantal_edge=0;
  t_ia64_ins * ins, * break_ins, * ins_last, * ins_first;
  t_symbol * bss_end;
  t_section * bss = SectionGetFromObjectByName(obj, ".bss");
  t_uint64 einde_bss = SECTION_GET_CADDRESS(bss)+SECTION_CSIZE(bss); 
  int i = 0;
  FILE * fp_adressen = fopen("uitvoer.edge","w+");
  t_bbl * end_bbl;
  
  /*tel aantal edges */
  gcc_counter=0;
  aantal_bbl_totaal=0;
  
  CFG_FOREACH_BBL(cfg,bbl)
  {
    aantal_bbl_totaal++;
  }
  
  CFG_FOREACH_BBL(cfg,bbl)
  {
#ifdef GCC
    if(!( gcc_counter >= global_options.debugcounter*aantal_bbl_totaal/AANTAL_GCC && gcc_counter < (global_options.debugcounter+1)*aantal_bbl_totaal/AANTAL_GCC))
    { 
      //      printf("%d %d\n",aantal_bbl,gcc_counter);
      //      FATAL(("ik ben hier"));
      gcc_counter++;      
      continue;
    }
#endif
    if(!IsHellNode(bbl) && 
	bbl->ins_first && 
	IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)

      BBL_FOREACH_SUCC_EDGE(bbl,edge)
      {
	if ((EDGE_CAT(edge)==ET_JUMP || 
	      EDGE_CAT(edge)==ET_IPJUMP || 
	      EDGE_CAT(edge)==ET_CALL)&& !Ia64IsReturn(bbl->ins_last)&&!Ia64IsCheck(bbl->ins_last))
	{
	  aantal_edge++;
	}
      }
    gcc_counter++;
  }

//  printf("aantal edges: %d\n",aantal_edge);

  gcc_counter=0;
  CFG_FOREACH_BBL(cfg,bbl)
  {

    /* r15 krijgt de waarde van de exit systeemoproep */
    if(bbl->ins_first && IA64OPCODEINFO(bbl->ins_first).opcode==IA64INS_ADDS_OPCODE && IA64INSIMM1(bbl->ins_first)==0x401)
    {
    /* zet instructies om bestand te schrijven net voor exit */
      ins_first =(t_ia64_ins *) bbl->ins_first;
      break_ins=NULL;
      /* we zoeken de systeemoproep */
      BBL_FOREACH_INS(bbl, T_INS(ins))
      {
	/* break heeft opcode 0 */
	if(IA64OPCODEINFO(ins).opcode==IA64INS_BREAK_OPCODE)
	  break_ins = ins;
      }
      
      if (break_ins==NULL) FATAL(("break niet gevonden"));

      end_bbl = bbl;
      
      /* uitschrijven van tellertjes naar het bestand uitvoer */
      
      // alloc, we reserveren 5 registers 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetOpcode(IA64_TYPE_M,"alloc"), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIMM2(ins) = 5;
      IA64INSIMM3(ins) = 0;
      IA64INSOUT1(ins) = IA64_R2; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // we bewaren de exit code in R36 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R32;
      IA64INSOUT1(ins) = IA64_R36; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
      
      // dummy voor X 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // waar staat filenaam 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
      IA64INSIMM1(ins) = einde_bss + aantal_edge*24;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R32; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // dummy voor X 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // bestandsnaam hexa 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
      IA64INSIMM1(ins) = 0x0072656f76676465ull;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R33; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // instructie om naam te schrijven 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
      IA64INSIN1(ins) = IA64_R33;
      IA64INSIN2(ins) = IA64_R32;
      IA64INSIMM1(ins) = 8;
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // dummy voor X 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // openingspermissie 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
      IA64INSIMM1(ins) = 0x0000000000002b77ull;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R33; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

     // instructie om permissie te schrijven 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
      IA64INSIN1(ins) = IA64_R33;
      IA64INSIN2(ins) = IA64_R32;
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      //  zet arg voor open 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0x1b6;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R34; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // zet arg voor open 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0x242;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R33; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // dummy voor X 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // waar staat filenaam 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
      IA64INSIMM1(ins) = einde_bss+aantal_edge*24;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R32; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // welke systeemoproep voor open 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 1028;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R15; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // systeemoproep open
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
      IA64INSIMM1(ins) = 0x100000; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // we bewaren de file handler in R35 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R8;
      IA64INSOUT1(ins) = IA64_R35; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // zet argumenten fcntl 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R34; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // zet file handler voor fcntl
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R35;
      IA64INSOUT1(ins) = IA64_R33; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // zet argument voor fcntl 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R32; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // welke systeemoproep voor fcntl 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 1066;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R15; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
 
      // systeemoproep fcntl 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
      IA64INSIMM1(ins) = 0x100000; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
    
      // zet inhoud LC
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddlOpcode(), IA64_P0);
      IA64INSIMM1(ins) = aantal_edge*3-1;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R32; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // zet LC 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I, "mov.i"), IA64_P0);
      IA64INSIN1(ins) = IA64_R32;
      IA64INSOUT1(ins) = IA64_AR_LC; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
      
      // dummy voor X 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // waar staat data 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
      IA64INSIMM1(ins) = einde_bss;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R33; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
      
      // zet file handler 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R35;
      IA64INSOUT1(ins) = IA64_R32; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // zet lengte string 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 8;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R34; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
  
      // welke systeemoproep voor write 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 1027;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R15; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
 
      // systeemoproep write 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
      IA64INSIMM1(ins) = 0x100000; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // incrementeer adres
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 8;
      IA64INSIN1(ins)=IA64_R33;
      IA64INSOUT1(ins) = IA64_R33; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // speel lus
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_B, GetOpcode(IA64_TYPE_B,"br.cloop.sptk.many"), IA64_P0);
      IA64INSIMM1(ins) = -0x20;
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
     
      // zet file handler voor close
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R35;
      IA64INSOUT1(ins) = IA64_R32; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
  
      // welke systeemoproep voor close
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroupAndInc();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 1029;
      IA64INSIN1(ins)=IA64_R0;
      IA64INSOUT1(ins) = IA64_R15; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
  
      // systeemoproep close 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
      IA64INSIMM1(ins) = 0x100000; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      // terugzetten exit code 
      ins = InsNewForBbl(bbl);
      ins->instr_group = GetCurrentInstrGroup();
      Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
      IA64INSIMM1(ins) = 0;
      IA64INSIN1(ins)=IA64_R36;
      IA64INSOUT1(ins) = IA64_R32; 
      InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
      
    }
#ifdef GCC
    if(!( gcc_counter >= global_options.debugcounter*aantal_bbl_totaal/AANTAL_GCC && gcc_counter < (global_options.debugcounter+1)*aantal_bbl_totaal/AANTAL_GCC))
    { 
      gcc_counter++;      
      continue;
    }
#endif
    if(!IsHellNode(bbl) && 
	bbl->ins_first && 
	IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)

      BBL_FOREACH_SUCC_EDGE(bbl,edge)
      {
	if ((EDGE_CAT(edge)==ET_JUMP || 
	      EDGE_CAT(edge)==ET_IPJUMP || 
	      EDGE_CAT(edge)==ET_CALL )&& !Ia64IsReturn(bbl->ins_last)&&!Ia64IsCheck(bbl->ins_last))
	{

          fprintf(fp_adressen,"%d\t%llx\n",i,BBL_OLD_ADDRESS(bbl));
    /* zet instructies om te profileren per edge */

	  ins_last = (t_ia64_ins *)bbl->ins_last;
	  
	  // instructie om stack_pointer te verhogen 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R12;
	  IA64INSOUT1(ins) = IA64_R12;
	  IA64INSIMM1(ins) = -16;
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om r15 op de stapel te zetten 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStfOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R15;
	  IA64INSIN2(ins) = IA64_R12;
	  IA64INSIMM1(ins) = -16; 
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om r14 op de stapel te zetten 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStfOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R14;
	  IA64INSIN2(ins) = IA64_R12;
	  IA64INSIMM1(ins) = 0; 
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);
	  

	  // dummy voor X 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroup();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om adres in r14 te steken 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetMvOpcode(), IA64_P0);
	  IA64INSOUT1(ins) = IA64_R14;
	  IA64INSIMM1(ins) = einde_bss+i*24; 
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  //printf("edge op basisadres %llx schrijft zijn teller op %llx\n",BBL_OLD_ADDRESS(bbl),einde_bss+i*24);

	  // instructie om teller in r15 te steken 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetLdOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R14;
	  IA64INSOUT1(ins) = IA64_R15;
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om teller te verhogen 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64INSPRED(ins_last));
	  IA64INSIN1(ins) = IA64_R15;
	  IA64INSOUT1(ins) = IA64_R15;
	  IA64INSIMM1(ins) = 1;
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om teller terug te schrijven 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R15;
	  IA64INSIN2(ins) = IA64_R14;
	  IA64INSIMM1(ins) = 8;
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om teller in r15 te steken 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetLdOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R14;
	  IA64INSOUT1(ins) = IA64_R15;
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om teller te verhogen 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R15;
	  IA64INSOUT1(ins) = IA64_R15;
	  IA64INSIMM1(ins) = 1;
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om teller terug te schrijven 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R15;
	  IA64INSIN2(ins) = IA64_R14;
	  IA64INSIMM1(ins) = 8;
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);


	  if(IA64INSIN1(ins_last)==IA64_NONE)
	  {
	    // dummy voor X 
	    ins = InsNewForBbl(bbl);
	    ins->instr_group = GetCurrentInstrGroup();
	    Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	    InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	    // instructie om doeladres in r15 te steken 
	    ins = InsNewForBbl(bbl);
	    ins->instr_group = GetCurrentInstrGroupAndInc();
	    Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetMvOpcode(), IA64_P0);
	    IA64INSOUT1(ins) = IA64_R15;
	    IA64INSIMM1(ins) = IA64INSIMM1(ins_last);//einde_bss+i*8; 
	    InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	    // instructie om doeladres te schrijven 
	    ins = InsNewForBbl(bbl);
	    ins->instr_group = GetCurrentInstrGroupAndInc();
	    Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	    IA64INSIN1(ins) = IA64_R15;
	    IA64INSIN2(ins) = IA64_R14;
	    InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);
	  }
	  
	  else
	  {
	    // instructie om doeladres in R15 te schrijven
	    ins = InsNewForBbl(bbl);
	    ins->instr_group = GetCurrentInstrGroupAndInc();
	    Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetMovbOpcode(), IA64_P0);
	    IA64INSIN1(ins) = IA64INSIN1(ins_last);
	    IA64INSOUT1(ins) = IA64_R15;
	    InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	    // instructie om doeladres te schrijven 
	    ins = InsNewForBbl(bbl);
	    ins->instr_group = GetCurrentInstrGroupAndInc();
	    Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	    IA64INSIN1(ins) = IA64_R15;
	    IA64INSIN2(ins) = IA64_R14;
	    InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);
	  }

	  //instructie om r14 van de stapel te halen 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetLdfOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R12;
	  IA64INSOUT1(ins) = IA64_R14;
	  IA64INSIMM1(ins) = 16; 
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	  // instructie om r15 van de stapel te halen 
	  ins = InsNewForBbl(bbl);
	  ins->instr_group = GetCurrentInstrGroupAndInc();
	  Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetLdfOpcode(), IA64_P0);
	  IA64INSIN1(ins) = IA64_R12;
	  IA64INSOUT1(ins) = IA64_R15;
	  IA64INSIMM1(ins) = 16; 
	  InsInsertBefore((t_ins*)ins, (t_ins *)ins_last);

	i++;
	}
      }
    gcc_counter++;
  }
  fclose(fp_adressen);
  
  SECTION_CSIZE(bss)+=(aantal_edge*24)+16;

  bss_end = SymbolTableGetSymbolByName(obj->symbol_table, "_end");

  if(bss_end == 0)
    FATAL(("We have a problem"));
  else 
  {
    bss_end->current_address += (aantal_edge*24)+16;
    //printf("%llx",bss_end->current_address);
  }
  Ia64MakeProfileBbl(obj, end_bbl);
}
/*Voeg profilering toe per basic block  */
void Ia64MakeProfileBbl(t_object * obj, t_bbl * end_bbl)
{
  t_bbl * bbl;
  t_ia64_ins * ins, * break_ins, * ins_first;
  int i = 0;
  int aantal_bbl=0;
  t_section * bss = SectionGetFromObjectByName(obj, ".bss");
  t_cfg * cfg = SECTION_CFG(SectionGetFromObjectByName(obj, ".init"));
  t_symbol * bss_end;
  t_uint64 einde_bss = SECTION_GET_CADDRESS(bss)+SECTION_CSIZE(bss); 
  FILE * fp;
  fp = fopen("uitvoer.bbl","w+");
  
 
  gcc_counter = 0;
  CFG_FOREACH_BBL(cfg,bbl){
#ifdef GCC
    if(!( gcc_counter >= global_options.debugcounter*aantal_bbl_totaal/AANTAL_GCC && gcc_counter < (global_options.debugcounter+1)*aantal_bbl_totaal/AANTAL_GCC))
    { 
      gcc_counter++;      
      continue;
    }
#endif
    if(bbl->nins != 0 && 
	(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)) 
      /*enkel de niet lege bbls die geen DATA zijn*/
      aantal_bbl++;
    gcc_counter++;
  }
  //printf("aantal bbl's: %d\n",aantal_bbl);
  //FATAL(("stop"));
 
  //printf("Aantal niet lege bbls: %d\n", aantal_bbl);

  gcc_counter=0;
  CFG_FOREACH_BBL(cfg,bbl){
    if(bbl->nins != 0 && 
	(bbl->ins_first && IA64INSTYPE(bbl->ins_first) != IA64_TYPE_DATA)) 
    {

      fprintf(fp,"%d\t%llx\n",i,BBL_OLD_ADDRESS(bbl));

      ins_first = (t_ia64_ins *)bbl->ins_first;

      /* r15 krijgt de waarde van de exit systeemoproep */
      if(end_bbl == bbl /*bbl->ins_first && IA64OPCODEINFO(bbl->ins_first).opcode==IA64INS_ADDS_OPCODE && IA64INSIMM1(bbl->ins_first)==0x401*/)
      {
	break_ins=NULL;
	/* we zoeken de systeemoproep */
	BBL_FOREACH_INS(bbl, T_INS(ins))
	{
	  /* break heeft opcode 0 */
	  if(IA64OPCODEINFO(ins).opcode==IA64INS_BREAK_OPCODE)
	    break_ins = ins;
	}

	if (break_ins==NULL) FATAL(("break niet gevonden"));

	/* uitschrijven van tellertjes naar het bestand uitvoer */

	// alloc, we reserveren 5 registers 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetOpcode(IA64_TYPE_M,"alloc"), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIMM2(ins) = 5;
	IA64INSIMM3(ins) = 0;
	IA64INSOUT1(ins) = IA64_R2; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// dit bbl zal 1 maal opgeroepen worden 

	// dummy voor X 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om adres in r14 te steken 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetMvOpcode(), IA64_P0);
	IA64INSOUT1(ins) = IA64_R36;
	IA64INSIMM1(ins) = einde_bss+i*8; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om teller te verhogen 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R0;
	IA64INSOUT1(ins) = IA64_R35;
	IA64INSIMM1(ins) = 1;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
    
#ifdef GCC
    if(( gcc_counter >= global_options.debugcounter*aantal_bbl_totaal/AANTAL_GCC && gcc_counter < (global_options.debugcounter+1)*aantal_bbl_totaal/AANTAL_GCC))
    { 
#endif
	// instructie om teller terug te schrijven 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R35;
	IA64INSIN2(ins) = IA64_R36;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
#ifdef GCC
    }
    else i--;
#endif

	// we bewaren de exit code in R36 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R32;
	IA64INSOUT1(ins) = IA64_R36; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// dummy voor X 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// waar staat filenaam 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
	IA64INSIMM1(ins) = einde_bss + aantal_bbl*8;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R32; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// dummy voor X 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// bestandsnaam hexa 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
	IA64INSIMM1(ins) = 0x0072656f766c6262ull;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R33; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om naam te schrijven 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R33;
	IA64INSIN2(ins) = IA64_R32;
	IA64INSIMM1(ins) = 8;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// dummy voor X 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// openingspermissie 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
	IA64INSIMM1(ins) = 0x0000000000002b77ull;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R33; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om permissie te schrijven 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R33;
	IA64INSIN2(ins) = IA64_R32;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	//  zet arg voor open 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0x1b6;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R34; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet arg voor open 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0x242;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R33; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// dummy voor X 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// waar staat filenaam 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
	IA64INSIMM1(ins) = einde_bss+aantal_bbl*8;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R32; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// welke systeemoproep voor open 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 1028;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R15; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// systeemoproep open
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
	IA64INSIMM1(ins) = 0x100000; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// we bewaren de file handler in R35 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R8;
	IA64INSOUT1(ins) = IA64_R35; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet argumenten fcntl 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R34; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet file handler voor fcntl
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R35;
	IA64INSOUT1(ins) = IA64_R33; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet argument voor fcntl 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R32; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// welke systeemoproep voor fcntl 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 1066;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R15; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// systeemoproep fcntl 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
	IA64INSIMM1(ins) = 0x100000; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet inhoud LC
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddlOpcode(), IA64_P0);
	IA64INSIMM1(ins) = aantal_bbl-1;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R32; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet LC 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I, "mov.i"), IA64_P0);
	IA64INSIN1(ins) = IA64_R32;
	IA64INSOUT1(ins) = IA64_AR_LC; 
	//InsPrint(ins,0);
	//FATAL(("stop"));
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// dummy voor X 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// waar staat data 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetOpcode(IA64_TYPE_X,"movl"), IA64_P0);
	IA64INSIMM1(ins) = einde_bss;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R33; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet file handler 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R35;
	IA64INSOUT1(ins) = IA64_R32; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet lengte string 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 8;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R34; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// welke systeemoproep voor write 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 1027;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R15; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// systeemoproep write 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
	IA64INSIMM1(ins) = 0x100000; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// incrementeer adres
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 8;
	IA64INSIN1(ins)=IA64_R33;
	IA64INSOUT1(ins) = IA64_R33; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// speel lus
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_B, GetOpcode(IA64_TYPE_B,"br.cloop.sptk.many"), IA64_P0);
	IA64INSIMM1(ins) = -0x20;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// zet file handler voor close
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R35;
	IA64INSOUT1(ins) = IA64_R32; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// welke systeemoproep voor close
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 1029;
	IA64INSIN1(ins)=IA64_R0;
	IA64INSOUT1(ins) = IA64_R15; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// systeemoproep close 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetOpcode(IA64_TYPE_I,"break.i"), IA64_P0);
	IA64INSIMM1(ins) = 0x100000; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// terugzetten exit code 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIMM1(ins) = 0;
	IA64INSIN1(ins)=IA64_R36;
	IA64INSOUT1(ins) = IA64_R32; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

      }
      else 
      {

#ifdef GCC
	if(!( gcc_counter >= global_options.debugcounter*aantal_bbl_totaal/AANTAL_GCC && gcc_counter < (global_options.debugcounter+1)*aantal_bbl_totaal/AANTAL_GCC))
	{ 
	  gcc_counter++;      
	  continue;
	}
#endif
	// instructie om stack_pointer te verhogen 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R12;
	IA64INSOUT1(ins) = IA64_R12;
	IA64INSIMM1(ins) = -16;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om r15 op de stapel te zetten 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStfOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R15;
	IA64INSIN2(ins) = IA64_R12;
	IA64INSIMM1(ins) = -16; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om r14 op de stapel te zetten 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStfOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R14;
	IA64INSIN2(ins) = IA64_R12;
	IA64INSIMM1(ins) = 0; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// dummy voor X 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_DUMMY, DUMMY_X_OPCODE,IA64_P0);
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om adres in r14 te steken 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_X, GetMvOpcode(), IA64_P0);
	IA64INSOUT1(ins) = IA64_R14;
	IA64INSIMM1(ins) = einde_bss+i*8; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	//printf("bbl op basisadres %llx, schrijft op %llx\n",BBL_OLD_ADDRESS(bbl), einde_bss+i*8);

	// instructie om teller in r15 te steken 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetLdOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R14;
	IA64INSOUT1(ins) = IA64_R15;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om teller te verhogen 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_A, GetAddsOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R15;
	IA64INSOUT1(ins) = IA64_R15;
	IA64INSIMM1(ins) = 1;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om teller terug te schrijven 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetStOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R15;
	IA64INSIN2(ins) = IA64_R14;
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	//instructie om r14 van de stapel te halen 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetLdfOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R12;
	IA64INSOUT1(ins) = IA64_R14;
	IA64INSIMM1(ins) = 16; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	// instructie om r15 van de stapel te halen 
	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroup();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetLdfOpcode(), IA64_P0);
	IA64INSIN1(ins) = IA64_R12;
	IA64INSOUT1(ins) = IA64_R15;
	IA64INSIMM1(ins) = 16/*ddd*/; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);

	/*      ins = InsNewForBbl(bbl);
		ins->instr_group = GetCurrentInstrGroup();
		Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_M, GetNopOpcode(IA64_UNIT_M), IA64_P0);
		IA64INSIN1(ins) = 0;
		IA64INSOUT1(ins) = 0;
		IA64INSIMM1(ins) = 0; 
		InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);
	 */

	ins = InsNewForBbl(bbl);
	ins->instr_group = GetCurrentInstrGroupAndInc();
	Ia64InitializeInstruction((t_ins *)ins, (t_uint64)0, IA64_TYPE_I, GetNopOpcode(IA64_UNIT_I), IA64_P0);
	IA64INSIN1(ins) = 0;
	IA64INSOUT1(ins) = 0;
	IA64INSIMM1(ins) = 0; 
	InsInsertBefore((t_ins*)ins, (t_ins *)ins_first);


      }
      i++;  
    }
    gcc_counter++;
  }
  //FATAL(("stop"));
  fclose(fp);

  SECTION_CSIZE(bss)+=(aantal_bbl*8)+16;

  bss_end = SymbolTableGetSymbolByName(obj->symbol_table, "_end");
  
  if(bss_end == 0)
    FATAL(("We have a problem"));
  else 
  {
    bss_end->current_address += (aantal_bbl*8)+16;
  }


}


/* vim: set shiftwidth=2 foldmethod=marker: */
