/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Headers for flowgraph manipulation
 *  
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 ***********************************************************/

#ifndef MY_PROGRAPH_TYPEDEFS
#define MY_PROGRAPH_TYPEDEFS
typedef struct _t_pfg t_pfg;
typedef struct _t_pfg_node t_pfg_node;
typedef struct _t_pfg_edge t_pfg_edge;
#endif /* MY_FLOWGRAH_TYPEDEFS */

//#ifndef DIABLO_INS_TYPEDEFS
//#define DIABLO_INS_TYPEDEFS
//#include "../../simple_types.h"
//#endif

#ifndef _GRAPH_TYPEDEFS
#define _GRAPH_TYPEDEFS
typedef struct _t_edge t_edge;
typedef struct _t_node t_node;
typedef struct _t_graph t_graph;
#endif

#ifndef MY_PROGRAPH_H
#define MY_PROGRAPH_H
#include "../../datastructures/data_graph.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_object.h"
#include "../../utils/util_error.h"
#include "ia64_instructions.h"
#include "ia64_opcodes.h"
#include "ia64_disassemble.h"
#include "ia64_prograph.h"

struct _t_pfg_node
{
	t_node node;
	t_bbl * bbl;
	t_bool breakable_next;
	t_pfg_node * layout_next;
	t_bool breakable_prev;
	t_pfg_node * layout_prev;
	t_pfg_node * layout_last;
	t_pfg_node * layout_first;
	t_bool header;
	
};
struct _t_pfg_edge
{
	t_edge edge;
	t_uint64 counter;
};
struct _t_pfg
{
	t_graph graph;
};

#define PFG_HEAD(edge) ((t_pfg_node *)(((t_edge *)edge)->head))
#define PFG_TAIL(edge) ((t_pfg_node *)(((t_edge *)edge)->tail))
#define PFG_BBL(node) (((t_pfg_node *)node)->bbl)
#define PFG_COUNTER(edge) (((t_pfg_edge *) edge)->counter)
#define PFG_NODE_NEXT(node) ((t_pfg_node *) (((t_pfg_node *) node)->layout_next))
#define PFG_NODE_PREV(node) ((t_pfg_node *) (((t_pfg_node *) node)->layout_prev))
		
#define T_PFG(cfg) ((t_pfg*)(cfg))

t_pfg * PfgNew();
void PfgFree(t_pfg * pfg);

void PfgInitNode(t_pfg * graph, t_pfg_node * temp);
t_pfg_node * PfgNewNode(t_pfg * graph, t_bbl * bbl);
t_pfg_edge * PfgNewEdge(t_pfg * graph, t_pfg_node * from, t_pfg_node * to, t_uint64 counter);

void PfgMergeByNodes(t_pfg * graaf, t_pfg_node * head, t_pfg_node * tail);
void PfgMergeByEdge(t_pfg * graaf, t_pfg_edge * the_edge);
void PfgMergeByBbl(t_pfg * graaf, t_bbl * head, t_bbl * tail);

void PfgClosestIsBest(t_pfg * graaf);
t_pfg_edge * PfgFindLargestEdge(t_pfg * graaf);
void PfgAddEdge(t_pfg * graaf, t_bbl * bbl, t_uint64 address, t_uint64 counter);
t_pfg * Ia64InitializeGraph(t_cfg * cfg);
void Ia64PrintPrograph(t_string bestand, t_pfg * pfg);
#endif

