/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"
#ifndef IA64_REGISTERS_H
#define IA64_REGISTERS_H
#include "../../datastructures/data_string.h"
#include "../../kernel/diablo_regset.h"


#define IA64GETNAT(ia64_gr) ((ia64_gr)-IA64_GR+IA64_GR_NAT)



typedef enum _t_ia64_reg
{
  /* **** APPLICATION REGISTERS **** */


  /* All global registers */
  IA64_GR=0,
  IA64_R0 = IA64_GR, IA64_R1, IA64_R2, IA64_R3, IA64_R4, IA64_R5, IA64_R6, IA64_R7,
  IA64_R8, IA64_R9, IA64_R10, IA64_R11, IA64_R12, IA64_R13, IA64_R14, IA64_R15,
  IA64_R16, IA64_R17, IA64_R18, IA64_R19, IA64_R20, IA64_R21, IA64_R22, IA64_R23,
  IA64_R24, IA64_R25, IA64_R26, IA64_R27, IA64_R28, IA64_R29, IA64_R30, IA64_R31,
  IA64_R32, IA64_R33, IA64_R34, IA64_R35, IA64_R36, IA64_R37, IA64_R38, IA64_R39,
  IA64_R40, IA64_R41, IA64_R42, IA64_R43, IA64_R44, IA64_R45, IA64_R46, IA64_R47,
  IA64_R48, IA64_R49, IA64_R50, IA64_R51, IA64_R52, IA64_R53, IA64_R54, IA64_R55,
  IA64_R56, IA64_R57, IA64_R58, IA64_R59, IA64_R60, IA64_R61, IA64_R62, IA64_R63,
  IA64_R64, IA64_R65, IA64_R66, IA64_R67, IA64_R68, IA64_R69, IA64_R70, IA64_R71,
  IA64_R72, IA64_R73, IA64_R74, IA64_R75, IA64_R76, IA64_R77, IA64_R78, IA64_R79,
  IA64_R80, IA64_R81, IA64_R82, IA64_R83, IA64_R84, IA64_R85, IA64_R86, IA64_R87,
  IA64_R88, IA64_R89, IA64_R90, IA64_R91, IA64_R92, IA64_R93, IA64_R94, IA64_R95,
  IA64_R96, IA64_R97, IA64_R98, IA64_R99, IA64_R100, IA64_R101, IA64_R102, IA64_R103,
  IA64_R104, IA64_R105, IA64_R106, IA64_R107, IA64_R108, IA64_R109, IA64_R110, IA64_R111,
  IA64_R112, IA64_R113, IA64_R114, IA64_R115, IA64_R116, IA64_R117, IA64_R118, IA64_R119,
  IA64_R120, IA64_R121, IA64_R122, IA64_R123, IA64_R124, IA64_R125, IA64_R126, IA64_R127,
  
  /* All global registers - nat bit */
  IA64_GR_NAT,
  IA64_R0_NAT = IA64_GR_NAT, IA64_R1_NAT, IA64_R2_NAT, IA64_R3_NAT, IA64_R4_NAT, IA64_R5_NAT, IA64_R6_NAT, IA64_R7_NAT, 
  IA64_R8_NAT, IA64_R9_NAT, IA64_R10_NAT, IA64_R11_NAT, IA64_R12_NAT, IA64_R13_NAT, IA64_R14_NAT, IA64_R15_NAT,
  IA64_R16_NAT, IA64_R17_NAT, IA64_R18_NAT, IA64_R19_NAT, IA64_R20_NAT, IA64_R21_NAT, IA64_R22_NAT, IA64_R23_NAT,
  IA64_R24_NAT, IA64_R25_NAT, IA64_R26_NAT, IA64_R27_NAT, IA64_R28_NAT, IA64_R29_NAT, IA64_R30_NAT, IA64_R31_NAT,
  IA64_R32_NAT, IA64_R33_NAT, IA64_R34_NAT, IA64_R35_NAT, IA64_R36_NAT, IA64_R37_NAT, IA64_R38_NAT, IA64_R39_NAT,
  IA64_R40_NAT, IA64_R41_NAT, IA64_R42_NAT, IA64_R43_NAT, IA64_R44_NAT, IA64_R45_NAT, IA64_R46_NAT, IA64_R47_NAT,
  IA64_R48_NAT, IA64_R49_NAT, IA64_R50_NAT, IA64_R51_NAT, IA64_R52_NAT, IA64_R53_NAT, IA64_R54_NAT, IA64_R55_NAT,
  IA64_R56_NAT, IA64_R57_NAT, IA64_R58_NAT, IA64_R59_NAT, IA64_R60_NAT, IA64_R61_NAT, IA64_R62_NAT, IA64_R63_NAT,
  IA64_R64_NAT, IA64_R65_NAT, IA64_R66_NAT, IA64_R67_NAT, IA64_R68_NAT, IA64_R69_NAT, IA64_R70_NAT, IA64_R71_NAT,
  IA64_R72_NAT, IA64_R73_NAT, IA64_R74_NAT, IA64_R75_NAT, IA64_R76_NAT, IA64_R77_NAT, IA64_R78_NAT, IA64_R79_NAT,
  IA64_R80_NAT, IA64_R81_NAT, IA64_R82_NAT, IA64_R83_NAT, IA64_R84_NAT, IA64_R85_NAT, IA64_R86_NAT, IA64_R87_NAT,
  IA64_R88_NAT, IA64_R89_NAT, IA64_R90_NAT, IA64_R91_NAT, IA64_R92_NAT, IA64_R93_NAT, IA64_R94_NAT, IA64_R95_NAT,
  IA64_R96_NAT, IA64_R97_NAT, IA64_R98_NAT, IA64_R99_NAT, IA64_R100_NAT, IA64_R101_NAT, IA64_R102_NAT, IA64_R103_NAT,
  IA64_R104_NAT, IA64_R105_NAT, IA64_R106_NAT, IA64_R107_NAT, IA64_R108_NAT, IA64_R109_NAT, IA64_R110_NAT, IA64_R111_NAT,
  IA64_R112_NAT, IA64_R113_NAT, IA64_R114_NAT, IA64_R115_NAT, IA64_R116_NAT, IA64_R117_NAT, IA64_R118_NAT, IA64_R119_NAT,
  IA64_R120_NAT, IA64_R121_NAT, IA64_R122_NAT, IA64_R123_NAT, IA64_R124_NAT, IA64_R125_NAT, IA64_R126_NAT, IA64_R127_NAT,


  /* All floating point registers */
  IA64_FR,
  IA64_F0 = IA64_FR, IA64_F1, IA64_F2, IA64_F3, IA64_F4, IA64_F5, IA64_F6, IA64_F7,
  IA64_F8, IA64_F9, IA64_F10, IA64_F11, IA64_F12, IA64_F13, IA64_F14, IA64_F15,
  IA64_F16, IA64_F17, IA64_F18, IA64_F19, IA64_F20, IA64_F21, IA64_F22, IA64_F23,
  IA64_F24, IA64_F25, IA64_F26, IA64_F27, IA64_F28, IA64_F29, IA64_F30, IA64_F31,
  IA64_F32, IA64_F33, IA64_F34, IA64_F35, IA64_F36, IA64_F37, IA64_F38, IA64_F39,
  IA64_F40, IA64_F41, IA64_F42, IA64_F43, IA64_F44, IA64_F45, IA64_F46, IA64_F47,
  IA64_F48, IA64_F49, IA64_F50, IA64_F51, IA64_F52, IA64_F53, IA64_F54, IA64_F55,
  IA64_F56, IA64_F57, IA64_F58, IA64_F59, IA64_F60, IA64_F61, IA64_F62, IA64_F63,
  IA64_F64, IA64_F65, IA64_F66, IA64_F67, IA64_F68, IA64_F69, IA64_F70, IA64_F71,
  IA64_F72, IA64_F73, IA64_F74, IA64_F75, IA64_F76, IA64_F77, IA64_F78, IA64_F79,
  IA64_F80, IA64_F81, IA64_F82, IA64_F83, IA64_F84, IA64_F85, IA64_F86, IA64_F87,
  IA64_F88, IA64_F89, IA64_F90, IA64_F91, IA64_F92, IA64_F93, IA64_F94, IA64_F95,
  IA64_F96, IA64_F97, IA64_F98, IA64_F99, IA64_F100, IA64_F101, IA64_F102, IA64_F103,
  IA64_F104, IA64_F105, IA64_F106, IA64_F107, IA64_F108, IA64_F109, IA64_F110, IA64_F111,
  IA64_F112, IA64_F113, IA64_F114, IA64_F115, IA64_F116, IA64_F117, IA64_F118, IA64_F119,
  IA64_F120, IA64_F121, IA64_F122, IA64_F123, IA64_F124, IA64_F125, IA64_F126, IA64_F127,

  /* All predicate registers */
  IA64_PR,
  IA64_P0 = IA64_PR, IA64_P1, IA64_P2, IA64_P3, IA64_P4, IA64_P5, IA64_P6, IA64_P7,
  IA64_P8, IA64_P9, IA64_P10, IA64_P11, IA64_P12, IA64_P13, IA64_P14, IA64_P15,
  IA64_P16, IA64_P17, IA64_P18, IA64_P19, IA64_P20, IA64_P21, IA64_P22, IA64_P23,
  IA64_P24, IA64_P25, IA64_P26, IA64_P27, IA64_P28, IA64_P29, IA64_P30, IA64_P31,
  IA64_P32, IA64_P33, IA64_P34, IA64_P35, IA64_P36, IA64_P37, IA64_P38, IA64_P39,
  IA64_P40, IA64_P41, IA64_P42, IA64_P43, IA64_P44, IA64_P45, IA64_P46, IA64_P47,
  IA64_P48, IA64_P49, IA64_P50, IA64_P51, IA64_P52, IA64_P53, IA64_P54, IA64_P55,
  IA64_P56, IA64_P57, IA64_P58, IA64_P59, IA64_P60, IA64_P61, IA64_P62, IA64_P63,

  /* All branch registers */
  IA64_BR,
  IA64_B0 = IA64_BR, IA64_B1, IA64_B2, IA64_B3, IA64_B4, IA64_B5, IA64_B6, IA64_B7,


  /* All Application registers - By number */
  IA64_AR,
  IA64_AR0 = IA64_AR, IA64_AR1, IA64_AR2, IA64_AR3, IA64_AR4, IA64_AR5, IA64_AR6, IA64_AR7,
  IA64_AR8, IA64_AR9, IA64_AR10, IA64_AR11, IA64_AR12, IA64_AR13, IA64_AR14, IA64_AR15,
  IA64_AR16, IA64_AR17, IA64_AR18, IA64_AR19, IA64_AR20, IA64_AR21, IA64_AR22, IA64_AR23,
  IA64_AR24, IA64_AR25, IA64_AR26, IA64_AR27, IA64_AR28, IA64_AR29, IA64_AR30, IA64_AR31,
  IA64_AR32, IA64_AR33, IA64_AR34, IA64_AR35, IA64_AR36, IA64_AR37, IA64_AR38, IA64_AR39,
  IA64_AR40, IA64_AR41, IA64_AR42, IA64_AR43, IA64_AR44, IA64_AR45, IA64_AR46, IA64_AR47,
  IA64_AR48, IA64_AR49, IA64_AR50, IA64_AR51, IA64_AR52, IA64_AR53, IA64_AR54, IA64_AR55,
  IA64_AR56, IA64_AR57, IA64_AR58, IA64_AR59, IA64_AR60, IA64_AR61, IA64_AR62, IA64_AR63,
  IA64_AR64, IA64_AR65, IA64_AR66, IA64_AR67, IA64_AR68, IA64_AR69, IA64_AR70, IA64_AR71,
  IA64_AR72, IA64_AR73, IA64_AR74, IA64_AR75, IA64_AR76, IA64_AR77, IA64_AR78, IA64_AR79,
  IA64_AR80, IA64_AR81, IA64_AR82, IA64_AR83, IA64_AR84, IA64_AR85, IA64_AR86, IA64_AR87,
  IA64_AR88, IA64_AR89, IA64_AR90, IA64_AR91, IA64_AR92, IA64_AR93, IA64_AR94, IA64_AR95,
  IA64_AR96, IA64_AR97, IA64_AR98, IA64_AR99, IA64_AR100, IA64_AR101, IA64_AR102, IA64_AR103,
  IA64_AR104, IA64_AR105, IA64_AR106, IA64_AR107, IA64_AR108, IA64_AR109, IA64_AR110, IA64_AR111,
  IA64_AR112, IA64_AR113, IA64_AR114, IA64_AR115, IA64_AR116, IA64_AR117, IA64_AR118, IA64_AR119,
  IA64_AR120, IA64_AR121, IA64_AR122, IA64_AR123, IA64_AR124, IA64_AR125, IA64_AR126, IA64_AR127,

  
  /* All Application registers - By name */

  /* Kernel registers */
  IA64_AR_K0 = IA64_AR0,
  IA64_AR_K1 = IA64_AR1,
  IA64_AR_K2 = IA64_AR2,
  IA64_AR_K3 = IA64_AR3,
  IA64_AR_K4 = IA64_AR4,
  IA64_AR_K5 = IA64_AR5,
  IA64_AR_K6 = IA64_AR6,
  IA64_AR_K7 = IA64_AR7,

  IA64_AR_RSC = IA64_AR16, 	    /* Register Stack Configuration Register */
  IA64_AR_BSP = IA64_AR17, 	    /* Backing Store Pointer */
  IA64_AR_BSPSTORE = IA64_AR18,	/* Backing Store Pointer for Memory Stores */
  IA64_AR_RNAT = IA64_AR19,	    /* RSE NaT Collection Register */

  IA64_AR_FCR = IA64_AR21,	    /* IA-32 Floating-point Control Register */

  IA64_AR_EFLAG = IA64_AR24,	  /* IA-32 EFLAG Register */
  IA64_AR_CSD = IA64_AR25,	    /* IA-32 Code Segment Descriptor */
  IA64_AR_SSD = IA64_AR26,	    /* IA-32 Stack Segment Descriptor */
  IA64_AR_CFLG = IA64_AR27,	    /* IA-32 Combined CR0 and CR4 Register */
  IA64_AR_FSR = IA64_AR28,	    /* IA-32 Floating-point Status Register */
  IA64_AR_FIR = IA64_AR29,	    /* IA-32 Floating-point Instruction Register */
  IA64_AR_FDR = IA64_AR30,	    /* IA-32 Floating-point Data Register */

  IA64_AR_CCV = IA64_AR32,	    /* Compare and Exchange Compare Value Register */

  IA64_AR_UNAT = IA64_AR36,	    /* User NaT Collection Register */
  
  IA64_AR_FPSR = IA64_AR40,	    /* Floating-point Status Register */

  IA64_AR_ITC = IA64_AR44,	    /* Interval Time Counter */

  IA64_AR_PFS = IA64_AR64,	    /* Previous Function State */
  IA64_AR_LC = IA64_AR65,	      /* Loop Count Register */
  IA64_AR_EC = IA64_AR66,	      /* Epilog Count Register */


  IA64_AR_LAST = IA64_AR127,         /* to set enum number correct */

  /* Instruction pointer */
  IA64_IP,

  /* Current Frame Marker */
  IA64_CFM,

  /* User Mask */
  IA64_UM,



  /* 
   * Indirecte registers: duiden enkel aan dat het om die soort gaat; indexering  
   * moet via een ander register...
   */

  /* CPUID Processor Identifiers indirecte registers */ 
  IA64_CPUID,
  
  /* PMD Performance Monitor Data Registers indirecte registers */
  IA64_PMD,




 
  /* **** SYSTEM REGISTERS **** */

  /* All Control registers - By number */
  IA64_CR,
  IA64_CR0 = IA64_CR, IA64_CR1, IA64_CR2, IA64_CR3, IA64_CR4, IA64_CR5, IA64_CR6, IA64_CR7,
  IA64_CR8, IA64_CR9, IA64_CR10, IA64_CR11, IA64_CR12, IA64_CR13, IA64_CR14, IA64_CR15,
  IA64_CR16, IA64_CR17, IA64_CR18, IA64_CR19, IA64_CR20, IA64_CR21, IA64_CR22, IA64_CR23,
  IA64_CR24, IA64_CR25, IA64_CR26, IA64_CR27, IA64_CR28, IA64_CR29, IA64_CR30, IA64_CR31,
  IA64_CR32, IA64_CR33, IA64_CR34, IA64_CR35, IA64_CR36, IA64_CR37, IA64_CR38, IA64_CR39,
  IA64_CR40, IA64_CR41, IA64_CR42, IA64_CR43, IA64_CR44, IA64_CR45, IA64_CR46, IA64_CR47,
  IA64_CR48, IA64_CR49, IA64_CR50, IA64_CR51, IA64_CR52, IA64_CR53, IA64_CR54, IA64_CR55,
  IA64_CR56, IA64_CR57, IA64_CR58, IA64_CR59, IA64_CR60, IA64_CR61, IA64_CR62, IA64_CR63,
  /* External Interrupt Control Registers */
  IA64_CR64, IA64_CR65, IA64_CR66, IA64_CR67, IA64_CR68, IA64_CR69, IA64_CR70, IA64_CR71,
  IA64_CR72, IA64_CR73, IA64_CR74, IA64_CR75, IA64_CR76, IA64_CR77, IA64_CR78, IA64_CR79,
  IA64_CR80, IA64_CR81,


  
  /* All Control registers - By name */
  
  /* Global Control Registers */
  IA64_CR_DCR = IA64_CR0,   /* Default Control Register */
  IA64_CR_ITM = IA64_CR1,   /* Interval Timer Match Register */
  IA64_CR_IVA = IA64_CR2,   /* Interruption Vector Address */

  IA64_CR_PTA = IA64_CR8,   /* Page Table Address */


  /* Interruption Control Registers */
  IA64_CR_IPSR = IA64_CR16, /* Interruption Processor Status Register */
  IA64_CR_ISR = IA64_CR17,  /* Interruption Status Register */
  
  IA64_CR_IIP = IA64_CR19,  /* Interruption Instruction Pointer */
  IA64_CR_IFA = IA64_CR20,  /* Interruption Faulting Address */
  IA64_CR_ITIR = IA64_CR21, /* Interruption TLB Insertion Register */
  IA64_CR_IIPA = IA64_CR22, /* Interruption Instruction Previous Address */
  IA64_CR_IFS = IA64_CR23,  /* Interruption Function State */
  IA64_CR_IIM = IA64_CR24,  /* Interruption Immediate Register */
  IA64_CR_IHA = IA64_CR25,  /* Interruption Hash Address */

  IA64_CR_LAST = IA64_CR81, /* to set enum number correct*/
  /* PSR Processor Status Register */
  IA64_PSR,
  
  /* PSR.l Lower PSR part*/  /*\TODO dit moet nog anders opgelost worden !!!*/
  IA64_PSR_L,

  /* PSR.um PSR User Mask part */
  IA64_PSR_UM,

  
  
  /* All Region Registers */
  IA64_RR,
  IA64_RR0 = IA64_RR, IA64_RR1, IA64_RR2, IA64_RR3, IA64_RR4, IA64_RR5, IA64_RR6, IA64_RR7,
    


  /* 
   * Indirecte registers: duiden enkel aan dat het om die soort gaat; indexering  
   * moet via een ander register...
   */

  /* PKR Protection Key Register indirecte registers */ 
  IA64_PKR,

  /* ITR Instruction Translation Lookaside Buffer - Instruction Translation Registers indirecte registers */ 
  IA64_ITR,

  /* ITC Instruction Translation Lookaside Buffer - Instruction Translation Cache */ 
  IA64_ITC,

  /* ITR Data Translation Lookaside Buffer - Data Translation Registers indirecte registers */ 
  IA64_DTR,

  /* ITC Data Translation Lookaside Buffer - Data Translation Cache */ 
  IA64_DTC,

  /* IBR Instruction Debug Breakpoint Registers indirecte registers */
  IA64_IBR,

  /* DBR Data Debug Breakpoint Registers indirecte registers */
  IA64_DBR,

  /* PMC Performance Monitor Configuration Registers indirecte registers */
  IA64_PMC,
  
  

  /* **** MISCELLANEOUS **** */

  /*MEM: for all instructions, dealing with memory*/
  IA64_RESOURCE_MEM,
  
  /* Immediate */
/*  IA64_IMM, */
  /* Empty */
  IA64_NONE

} t_ia64_reg;


#define IA64_COUNT_GR 128
#define IA64_COUNT_FR 128
#define IA64_COUNT_PR 64
#define IA64_COUNT_BR 8
#define IA64_COUNT_AR 128
#define IA64_COUNT_CR 82

/*moet waarsch ook nog geupdated worden in ia64_descriptions*/

#define IA64_TOTALREGS  \
( \
   IA64_COUNT_GR + /* int regs */ \
   IA64_COUNT_FR + /* float regs */ \
   IA64_COUNT_PR + /* pred regs */ \
   IA64_COUNT_BR + /* branch regs */ \
      \
      \
   /* special regs */ \
   IA64_COUNT_GR +  /*gr-nat bits*/ \
   IA64_COUNT_AR +  /*application registers*/ \
   1 +    /* IP */ \
   1 +    /* CFM */ \
   1 +    /* UM */ \
   1 +    /* CPUID */ \
   1 +    /* PMD */ \
   IA64_COUNT_CR +   /* Control Registers*/ \
   1 +    /* PSR */ \
   1 +    /* PSR.L */ /*\TODO dit moet nog anders opgelost worden*/ \
   1 +    /* PSR.UM */ /*\TODO dit moet nog anders opgelost worden*/ \
   8 +    /* Region Registers*/ \
   8     /* indirect system registers*/    \
)


#define IA64_MAX_GR IA64_R127
#define IA64_MAX_FR IA64_F127
#define IA64_MAX_PR IA64_P63
#define IA64_MAX_BR IA64_B7
#define IA64_MAX_AR IA64_AR127
#define IA64_MAX_CR IA64_CR81
#define IA64_MAX_RR IA64_RR7

#define IA64_IS_GR(reg) ((reg)>=IA64_GR && (reg)<= IA64_MAX_GR)
#define IA64_IS_FR(reg) ((reg)>=IA64_FR && (reg)<= IA64_MAX_FR)
#define IA64_IS_PR(reg) ((reg)>=IA64_PR && (reg)<= IA64_MAX_PR)
#define IA64_IS_BR(reg) ((reg)>=IA64_BR && (reg)<= IA64_MAX_BR)
#define IA64_IS_AR(reg) ((reg)>=IA64_AR && (reg)<= IA64_MAX_AR)
#define IA64_IS_CR(reg) ((reg)>=IA64_CR && (reg)<= IA64_MAX_CR)
#define IA64_IS_RR(reg) ((reg)>=IA64_RR && (reg)<= IA64_MAX_RR)

#define IA64_IS_INDIRECT_REG(reg) \
(IA64_IS_RR(reg) || (reg) == IA64_DBR || (reg) == IA64_IBR || \
 (reg) == IA64_PKR || (reg) == IA64_PMC || (reg) == IA64_PMD || \
 /*(reg) == IA64_MSR ||*/ (reg) == IA64_CPUID)

/* TODO: aanvullen naar gebruik*/


t_string Ia64RegisterName(t_reg reg);



#endif


/* vim: set shiftwidth=2: */
