/* tc-ia64.c -- Assembler for the HP/Intel IA-64 architecture.
   Copyright 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
   Contributed by David Mosberger-Tang <davidm@hpl.hp.com>

   This file is part of GAS, the GNU Assembler.

   GAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GAS; see the file COPYING.  If not, write to
   the Free Software Foundation, 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */


#include "ia64_instructions.h"
#include "ia64_registers.h"
#include "ia64_opcodes.h"
#include "ia64-asmtab.c"

#include "ia64_bundle.h"
#include "ia64_disassemble/ia64_disassemble_one.h"
#include "ia64_disassemble.h"

#include "../../utils/util_error.h"

#define DEP(X) ((X)&0x7FF)


t_ia64_ins* slot[3];

int currslot;

#define CURR_SLOT (slot[currslot])




/* The best template for a particular sequence of up to three
   instructions:  */
#define N	IA64_NUM_TYPES
static unsigned char best_template[N][N][N];
#undef N

/* Resource dependencies currently in effect */
static struct rsrc {
  int depind;                       /* dependency index */
  const t_ia64_dependency *dependency; /* actual dependency */
  unsigned specific:1,              /* is this a specific bit/regno? */
    link_to_qp_branch:1;           /* will a branch on the same QP clear it?*/
  int index;                        /* specific regno/bit within dependency */
  int note;                         /* optional qualifying note (0 if none) */
#define STATE_NONE 0
#define STATE_STOP 1
#define STATE_SRLZ 2
  int insn_srlz;                    /* current insn serialization state */
  int data_srlz;                    /* current data serialization state */
  int qp_regno;                     /* qualifying predicate for this usage */
  char *file;                       /* what file marked this dependency */
  unsigned int line;                /* what line marked this dependency */
//  struct mem_offset mem_offset;     /* optional memory offset hint */
  enum { CMP_NONE, CMP_OR, CMP_AND } cmp_type; /* OR or AND compare? */
  int path;                         /* corresponding code entry index */
} *regdeps = NULL;
static int regdepslen = 0;
static int regdepstotlen = 0;
static const char *dv_mode[] = { "RAW", "WAW", "WAR" };
static const char *dv_sem[] = { "none", "implied", "impliedf",
				"data", "instr", "specific", "stop", "other" };
static const char *dv_cmp_type[] = { "none", "OR", "AND" };

/* Current state of PR mutexation */
static struct qpmutex {
  t_uint64 prmask;
  int path;
} *qp_mutexes = NULL;          /* QP mutex bitmasks */
static int qp_mutexeslen = 0;
static int qp_mutexestotlen = 0;
static t_uint64 qp_safe_across_calls = 0;

/* Current state of PR implications */
static struct qp_imply {
  unsigned p1:6;
  unsigned p2:6;
  unsigned p2_branched:1;
  int path;
} *qp_implies = NULL;
static int qp_implieslen = 0;
static int qp_impliestotlen = 0;

/* Keep track of static GR values so that indirect register usage can
   sometimes be tracked.  */
static struct gr {
  unsigned known:1;
  int path;
  t_uint64 value;
} gr_values[128] = {{ 1, 0, 0 }};




static t_ia64_ins tmp_ins;
static t_bundle tmp_bundle;

t_ia64_ins *Ia64MakeTmpIns(t_ia64_opcode * opc)
{
  t_ia64_instr_type itype = Ia64GetInstrType(opc);
  Ia64InitializeInstruction((t_ins*)&tmp_ins, itype, ia64_opcode_table[itype] - opc,IA64_P0);
  Ia64DisassembleFunction_table[opc->instr_format]((t_ins*)&tmp_ins, opc->opcode, &tmp_bundle);
  
  return &tmp_ins;
}


const t_ia64_dependency *
ia64_find_dependency (index)
    int index;
{
  index = DEP(index);

  if (index < 0
      || index >= (int)(sizeof(dependencies) / sizeof(dependencies[0])))
    return NULL;

  return &dependencies[index];
}






const t_ia64_opcode_dependency* ia64_get_dependencies(t_ia64_opcode * desc)
{
  t_ia64_instr_type itype = Ia64GetInstrType(desc);
  t_uint32 index = (ia64_opcode_table[itype] - desc);// /sizeof(t_ia64_opcode);
  
  if(&ia64_opcode_table[itype][index] != desc)
    FATAL(("Something wrong in ia64_get_dependencies"));

  return &op_dependencies[ia64_ins_dependencies_index[itype][index]];
}






#if 0



static void
emit_one_bundle ()
{
  unsigned int manual_bundling_on = 0, manual_bundling_off = 0;
  unsigned int manual_bundling = 0;
  enum ia64_unit required_unit, insn_unit = 0;
  enum ia64_insn_type type[3], insn_type;
  unsigned int template, orig_template;
  bfd_vma insn[3] = { -1, -1, -1 };
  t_ia64_opcode *idesc;
  int end_of_insn_group = 0, user_template = -1;
  int n, i, j, first, curr;
  unw_rec_list *ptr;
  bfd_vma t0 = 0, t1 = 0;
  struct label_fix *lfix;
  struct insn_fix *ifix;
  char mnemonic[16];
  fixS *fix;
  char *f;

  first = (md.curr_slot + NUM_SLOTS - md.num_slots_in_use) % NUM_SLOTS;
  know (first >= 0 & first < NUM_SLOTS);
  n = MIN (3, md.num_slots_in_use);

  /* Determine template: user user_template if specified, best match
otherwise:  */

  if (md.slot[first].user_template >= 0)
    user_template = template = md.slot[first].user_template;
  else
  {
    /* Auto select appropriate template.  */
    memset (type, 0, sizeof (type));
    curr = first;
    for (i = 0; i < n; ++i)
    {
      if (md.slot[curr].label_fixups && i != 0)
	break;
      type[i] = md.slot[curr].idesc->type;
      curr = (curr + 1) % NUM_SLOTS;
    }
    template = best_template[type[0]][type[1]][type[2]];
  }

  /* initialize instructions with appropriate nops:  */
  for (i = 0; i < 3; ++i)
    insn[i] = nop[ia64_templ_desc[template].exec_unit[i]];

  f = frag_more (16);

  /* now fill in slots with as many insns as possible:  */
  curr = first;
  idesc = md.slot[curr].idesc;
  end_of_insn_group = 0;
  for (i = 0; i < 3 && md.num_slots_in_use > 0; ++i)
  {
    /* Set the slot number for prologue/body records now as those
       refer to the current point, not the point after the
       instruction has been issued:  */
    /* Don't try to delete prologue/body records here, as that will cause
       them to also be deleted from the master list of unwind records.  */
    for (ptr = md.slot[curr].unwind_record; ptr; ptr = ptr->next)
      if (ptr->r.type == prologue || ptr->r.type == prologue_gr
	  || ptr->r.type == body)
      {
	ptr->slot_number = (unsigned long) f + i;
	ptr->slot_frag = frag_now;
      }

    if (idesc->flags & IA64_OPCODE_SLOT2)
    {
      if (manual_bundling && i != 2)
	as_bad_where (md.slot[curr].src_file, md.slot[curr].src_line,
	    "`%s' must be last in bundle", idesc->desc);
      else
	i = 2;
    }
    if (idesc->flags & IA64_OPCODE_LAST)
    {
      int required_slot;
      unsigned int required_template;

      /* If we need a stop bit after an M slot, our only choice is
	 template 5 (M;;MI).  If we need a stop bit after a B
	 slot, our only choice is to place it at the end of the
	 bundle, because the only available templates are MIB,
	 MBB, BBB, MMB, and MFB.  We don't handle anything other
	 than M and B slots because these are the only kind of
	 instructions that can have the IA64_OPCODE_LAST bit set.  */
      required_template = template;
      switch (idesc->type)
      {
	case IA64_TYPE_M:
	  required_slot = 0;
	  required_template = 5;
	  break;

	case IA64_TYPE_B:
	  required_slot = 2;
	  break;

	default:
	  as_bad_where (md.slot[curr].src_file, md.slot[curr].src_line,
	      "Internal error: don't know how to force %s to end"
	      "of instruction group", idesc->desc);
	  required_slot = i;
	  break;
      }
      if (manual_bundling && i != required_slot)
	as_bad_where (md.slot[curr].src_file, md.slot[curr].src_line,
	    "`%s' must be last in instruction group",
	    idesc->desc);
      if (required_slot < i)
	/* Can't fit this instruction.  */
	break;

      i = required_slot;
      if (required_template != template)
      {
	/* If we switch the template, we need to reset the NOPs
	   after slot i.  The slot-types of the instructions ahead
	   of i never change, so we don't need to worry about
	   changing NOPs in front of this slot.  */
	for (j = i; j < 3; ++j)
	  insn[j] = nop[ia64_templ_desc[required_template].exec_unit[j]];
      }
      template = required_template;
    }
    if (curr != first && md.slot[curr].label_fixups)
    {
      if (manual_bundling_on)
	as_bad_where (md.slot[curr].src_file, md.slot[curr].src_line,
	    "Label must be first in a bundle");
      /* This insn must go into the first slot of a bundle.  */
      break;
    }

    manual_bundling_on = md.slot[curr].manual_bundling_on;
    manual_bundling_off = md.slot[curr].manual_bundling_off;

    if (manual_bundling_on)
    {
      if (curr == first)
	manual_bundling = 1;
      else
	break;			/* need to start a new bundle */
    }

    if (end_of_insn_group && md.num_slots_in_use >= 1)
    {
      /* We need an instruction group boundary in the middle of a
	 bundle.  See if we can switch to an other template with
	 an appropriate boundary.  */

      orig_template = template;
      if (i == 1 && (user_template == 4
	    || (user_template < 0
	      && (ia64_templ_desc[template].exec_unit[0]
		== IA64_UNIT_M))))
      {
	template = 5;
	end_of_insn_group = 0;
      }
      else if (i == 2 && (user_template == 0
	    || (user_template < 0
	      && (ia64_templ_desc[template].exec_unit[1]
		== IA64_UNIT_I)))
	  /* This test makes sure we don't switch the template if
	     the next instruction is one that needs to be first in
	     an instruction group.  Since all those instructions are
	     in the M group, there is no way such an instruction can
	     fit in this bundle even if we switch the template.  The
	     reason we have to check for this is that otherwise we
	     may end up generating "MI;;I M.." which has the deadly
	     effect that the second M instruction is no longer the
	     first in the bundle! --davidm 99/12/16  */
	  && (idesc->flags & IA64_OPCODE_FIRST) == 0)
      {
	template = 1;
	end_of_insn_group = 0;
      }
      else if (curr != first)
	/* can't fit this insn */
	break;

      if (template != orig_template)
	/* if we switch the template, we need to reset the NOPs
	   after slot i.  The slot-types of the instructions ahead
	   of i never change, so we don't need to worry about
	   changing NOPs in front of this slot.  */
	for (j = i; j < 3; ++j)
	  insn[j] = nop[ia64_templ_desc[template].exec_unit[j]];
    }
    required_unit = ia64_templ_desc[template].exec_unit[i];

    /* resolve dynamic opcodes such as "break" and "nop":  */
    if (idesc->type == IA64_TYPE_DYN)
    {
      if ((strcmp (idesc->desc, "nop") == 0)
	  || (strcmp (idesc->desc, "break") == 0))
	insn_unit = required_unit;
      else if (strcmp (idesc->desc, "chk.s") == 0)
      {
	insn_unit = IA64_UNIT_M;
	if (required_unit == IA64_UNIT_I)
	  insn_unit = IA64_UNIT_I;
      }
      else
	as_fatal ("emit_one_bundle: unexpected dynamic op");

      sprintf (mnemonic, "%s.%c", idesc->desc, "?imbf??"[insn_unit]);
      ia64_free_opcode (idesc);
      md.slot[curr].idesc = idesc = ia64_find_opcode (mnemonic);
#if 0
      know (!idesc->next);	/* no resolved dynamic ops have collisions */
#endif
    }
    else
    {
      insn_type = idesc->type;
      insn_unit = IA64_UNIT_NIL;
      switch (insn_type)
      {
	case IA64_TYPE_A:
	  if (required_unit == IA64_UNIT_I || required_unit == IA64_UNIT_M)
	    insn_unit = required_unit;
	  break;
	case IA64_TYPE_X: insn_unit = IA64_UNIT_L; break;
	case IA64_TYPE_I: insn_unit = IA64_UNIT_I; break;
	case IA64_TYPE_M: insn_unit = IA64_UNIT_M; break;
	case IA64_TYPE_B: insn_unit = IA64_UNIT_B; break;
	case IA64_TYPE_F: insn_unit = IA64_UNIT_F; break;
	default:				       break;
      }
    }

    if (insn_unit != required_unit)
    {
      if (required_unit == IA64_UNIT_L
	  && insn_unit == IA64_UNIT_I
	  && !(idesc->flags & IA64_OPCODE_X_IN_MLX))
      {
	/* we got ourselves an MLX template but the current
	   instruction isn't an X-unit, or an I-unit instruction
	   that can go into the X slot of an MLX template.  Duh.  */
	if (md.num_slots_in_use >= NUM_SLOTS)
	{
	  as_bad_where (md.slot[curr].src_file,
	      md.slot[curr].src_line,
	      "`%s' can't go in X slot of "
	      "MLX template", idesc->desc);
	  /* drop this insn so we don't livelock:  */
	  --md.num_slots_in_use;
	}
	break;
      }
      continue;		/* try next slot */
    }

    {
      bfd_vma addr;

      addr = frag_now->fr_address + frag_now_fix () - 16 + i;
      dwarf2_gen_line_info (addr, &md.slot[curr].debug_line);
    }

    if (errata_nop_necessary_p (md.slot + curr, insn_unit))
      as_warn (_("Additional NOP may be necessary to workaround Itanium processor A/B step errata"));

    build_insn (md.slot + curr, insn + i);

    /* Set slot counts for non prologue/body unwind records.  */
    for (ptr = md.slot[curr].unwind_record; ptr; ptr = ptr->next)
      if (ptr->r.type != prologue && ptr->r.type != prologue_gr
	  && ptr->r.type != body)
      {
	ptr->slot_number = (unsigned long) f + i;
	ptr->slot_frag = frag_now;
      }
    md.slot[curr].unwind_record = NULL;

    if (required_unit == IA64_UNIT_L)
    {
      know (i == 1);
      /* skip one slot for long/X-unit instructions */
      ++i;
    }
    --md.num_slots_in_use;

    /* now is a good time to fix up the labels for this insn:  */
    for (lfix = md.slot[curr].label_fixups; lfix; lfix = lfix->next)
    {
      S_SET_VALUE (lfix->sym, frag_now_fix () - 16);
      symbol_set_frag (lfix->sym, frag_now);
    }
    /* and fix up the tags also.  */
    for (lfix = md.slot[curr].tag_fixups; lfix; lfix = lfix->next)
    {
      S_SET_VALUE (lfix->sym, frag_now_fix () - 16 + i);
      symbol_set_frag (lfix->sym, frag_now);
    }

    for (j = 0; j < md.slot[curr].num_fixups; ++j)
    {
      ifix = md.slot[curr].fixup + j;
      fix = fix_new_exp (frag_now, frag_now_fix () - 16 + i, 8,
	  &ifix->expr, ifix->is_pcrel, ifix->code);
      fix->tc_fix_data.opnd = ifix->opnd;
      fix->fx_plt = (fix->fx_r_type == BFD_RELOC_IA64_PLTOFF22);
      fix->fx_file = md.slot[curr].src_file;
      fix->fx_line = md.slot[curr].src_line;
    }

    end_of_insn_group = md.slot[curr].end_of_insn_group;

    if (end_of_insn_group)
    {
      md.group_idx = (md.group_idx + 1) % 3;
      memset (md.last_groups + md.group_idx, 0, sizeof md.last_groups[0]);
    }

    /* clear slot:  */
    ia64_free_opcode (md.slot[curr].idesc);
    memset (md.slot + curr, 0, sizeof (md.slot[curr]));
    md.slot[curr].user_template = -1;

    if (manual_bundling_off)
    {
      manual_bundling = 0;
      break;
    }
    curr = (curr + 1) % NUM_SLOTS;
    idesc = md.slot[curr].idesc;
  }
  if (manual_bundling)
  {
    if (md.num_slots_in_use > 0)
      as_bad_where (md.slot[curr].src_file, md.slot[curr].src_line,
	  "`%s' does not fit into %s template",
	  idesc->desc, ia64_templ_desc[template].name);
    else
      as_bad_where (md.slot[curr].src_file, md.slot[curr].src_line,
	  "Missing '}' at end of file");
  }
  know (md.num_slots_in_use < NUM_SLOTS);

  t0 = end_of_insn_group | (template << 1) | (insn[0] << 5) | (insn[1] << 46);
  t1 = ((insn[1] >> 18) & 0x7fffff) | (insn[2] << 23);

  number_to_chars_littleendian (f + 0, t0, 8);
  number_to_chars_littleendian (f + 8, t1, 8);

  unwind.next_slot_number = (unsigned long) f + 16;
  unwind.next_slot_frag = frag_now;
}




#endif














#if 0
/* Return true if TYPE fits in TEMPL at SLOT.  */

static int
match (int templ, int type, int slot)
{
  t_ia64_unit unit;
  int result;

  unit = ia64_templ_desc[templ].exec_unit[slot];
  switch (type)
    {
//    case IA64_TYPE_DYN: result = 1; break; /* for nop and break */
    case IA64_TYPE_A:
      result = (unit == IA64_UNIT_I || unit == IA64_UNIT_M);
      break;
    case IA64_TYPE_X:	result = (unit == IA64_UNIT_L); break;
    case IA64_TYPE_I:	result = (unit == IA64_UNIT_I); break;
    case IA64_TYPE_M:	result = (unit == IA64_UNIT_M); break;
    case IA64_TYPE_B:	result = (unit == IA64_UNIT_B); break;
    case IA64_TYPE_F:	result = (unit == IA64_UNIT_F); break;
    default:		result = 0; break;
    }
  return result;
}

#endif






#if 0 


/* Add a bit of extra goodness if a nop of type F or B would fit
   in TEMPL at SLOT.  */

static inline int
extra_goodness (int templ, int slot)
{
  if (slot == 1 && match (templ, IA64_TYPE_F, slot))
    return 2;
  if (slot == 2 && match (templ, IA64_TYPE_B, slot))
    return 1;
  return 0;
}

/* This function is called once, at assembler startup time.  It sets
   up all the tables, etc. that the MD part of the assembler will need
   that can be determined before arguments are parsed.  */
void
md_begin ()
{
  int i, j, k, t, total, ar_base, cr_base, goodness, best, regnum, ok;
  const char *err;
  char name[8];

  md.auto_align = 1;
  md.explicit_mode = md.default_explicit_mode;

  bfd_set_section_alignment (stdoutput, text_section, 4);

  target_big_endian = TARGET_BYTES_BIG_ENDIAN;
  pseudo_func[FUNC_DTP_MODULE].u.sym =
    symbol_new (".<dtpmod>", undefined_section, FUNC_DTP_MODULE,
	&zero_address_frag);

  pseudo_func[FUNC_DTP_RELATIVE].u.sym =
    symbol_new (".<dtprel>", undefined_section, FUNC_DTP_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_FPTR_RELATIVE].u.sym =
    symbol_new (".<fptr>", undefined_section, FUNC_FPTR_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_GP_RELATIVE].u.sym =
    symbol_new (".<gprel>", undefined_section, FUNC_GP_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_LT_RELATIVE].u.sym =
    symbol_new (".<ltoff>", undefined_section, FUNC_LT_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_PC_RELATIVE].u.sym =
    symbol_new (".<pcrel>", undefined_section, FUNC_PC_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_PLT_RELATIVE].u.sym =
    symbol_new (".<pltoff>", undefined_section, FUNC_PLT_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_SEC_RELATIVE].u.sym =
    symbol_new (".<secrel>", undefined_section, FUNC_SEC_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_SEG_RELATIVE].u.sym =
    symbol_new (".<segrel>", undefined_section, FUNC_SEG_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_TP_RELATIVE].u.sym =
    symbol_new (".<tprel>", undefined_section, FUNC_TP_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_LTV_RELATIVE].u.sym =
    symbol_new (".<ltv>", undefined_section, FUNC_LTV_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_LT_FPTR_RELATIVE].u.sym =
    symbol_new (".<ltoff.fptr>", undefined_section, FUNC_LT_FPTR_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_LT_DTP_MODULE].u.sym =
    symbol_new (".<ltoff.dtpmod>", undefined_section, FUNC_LT_DTP_MODULE,
	&zero_address_frag);

  pseudo_func[FUNC_LT_DTP_RELATIVE].u.sym =
    symbol_new (".<ltoff.dptrel>", undefined_section, FUNC_LT_DTP_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_LT_TP_RELATIVE].u.sym =
    symbol_new (".<ltoff.tprel>", undefined_section, FUNC_LT_TP_RELATIVE,
	&zero_address_frag);

  pseudo_func[FUNC_IPLT_RELOC].u.sym =
    symbol_new (".<iplt>", undefined_section, FUNC_IPLT_RELOC,
	&zero_address_frag);

  /* Compute the table of best templates.  We compute goodness as a
     base 4 value, in which each match counts for 3, each F counts
     for 2, each B counts for 1.  This should maximize the number of
     F and B nops in the chosen bundles, which is good because these
     pipelines are least likely to be overcommitted.  */
  for (i = 0; i < IA64_NUM_TYPES; ++i)
    for (j = 0; j < IA64_NUM_TYPES; ++j)
      for (k = 0; k < IA64_NUM_TYPES; ++k)
      {
	best = 0;
	for (t = 0; t < NELEMS (ia64_templ_desc); ++t)
	{
	  goodness = 0;
	  if (match (t, i, 0))
	  {
	    if (match (t, j, 1))
	    {
	      if (match (t, k, 2))
		goodness = 3 + 3 + 3;
	      else
		goodness = 3 + 3 + extra_goodness (t, 2);
	    }
	    else if (match (t, j, 2))
	      goodness = 3 + 3 + extra_goodness (t, 1);
	    else
	    {
	      goodness = 3;
	      goodness += extra_goodness (t, 1);
	      goodness += extra_goodness (t, 2);
	    }
	  }
	  else if (match (t, i, 1))
	  {
	    if (match (t, j, 2))
	      goodness = 3 + 3;
	    else
	      goodness = 3 + extra_goodness (t, 2);
	  }
	  else if (match (t, i, 2))
	    goodness = 3 + extra_goodness (t, 1);

	  if (goodness > best)
	  {
	    best = goodness;
	    best_template[i][j][k] = t;
	  }
	}
      }

  for (i = 0; i < NUM_SLOTS; ++i)
    md.slot[i].user_template = -1;

  md.pseudo_hash = hash_new ();
  for (i = 0; i < NELEMS (pseudo_opcode); ++i)
  {
    err = hash_insert (md.pseudo_hash, pseudo_opcode[i].name,
	(void *) (pseudo_opcode + i));
    if (err)
      as_fatal ("ia64.md_begin: can't hash `%s': %s",
	  pseudo_opcode[i].name, err);
  }

  md.reg_hash = hash_new ();
  md.dynreg_hash = hash_new ();
  md.const_hash = hash_new ();
  md.entry_hash = hash_new ();

  /* general registers:  */

  total = 128;
  for (i = 0; i < total; ++i)
  {
    sprintf (name, "r%d", i - REG_GR);
    md.regsym[i] = declare_register (name, i);
  }

  /* floating point registers:  */
  total += 128;
  for (; i < total; ++i)
  {
    sprintf (name, "f%d", i - REG_FR);
    md.regsym[i] = declare_register (name, i);
  }

  /* application registers:  */
  total += 128;
  ar_base = i;
  for (; i < total; ++i)
  {
    sprintf (name, "ar%d", i - REG_AR);
    md.regsym[i] = declare_register (name, i);
  }

  /* control registers:  */
  total += 128;
  cr_base = i;
  for (; i < total; ++i)
  {
    sprintf (name, "cr%d", i - REG_CR);
    md.regsym[i] = declare_register (name, i);
  }

  /* predicate registers:  */
  total += 64;
  for (; i < total; ++i)
  {
    sprintf (name, "p%d", i - REG_P);
    md.regsym[i] = declare_register (name, i);
  }

  /* branch registers:  */
  total += 8;
  for (; i < total; ++i)
  {
    sprintf (name, "b%d", i - REG_BR);
    md.regsym[i] = declare_register (name, i);
  }

  md.regsym[REG_IP] = declare_register ("ip", REG_IP);
  md.regsym[REG_CFM] = declare_register ("cfm", REG_CFM);
  md.regsym[REG_PR] = declare_register ("pr", REG_PR);
  md.regsym[REG_PR_ROT] = declare_register ("pr.rot", REG_PR_ROT);
  md.regsym[REG_PSR] = declare_register ("psr", REG_PSR);
  md.regsym[REG_PSR_L] = declare_register ("psr.l", REG_PSR_L);
  md.regsym[REG_PSR_UM] = declare_register ("psr.um", REG_PSR_UM);

  for (i = 0; i < NELEMS (indirect_reg); ++i)
  {
    regnum = indirect_reg[i].regnum;
    md.regsym[regnum] = declare_register (indirect_reg[i].name, regnum);
  }

  /* define synonyms for application registers:  */
  for (i = REG_AR; i < REG_AR + NELEMS (ar); ++i)
    md.regsym[i] = declare_register (ar[i - REG_AR].name,
	REG_AR + ar[i - REG_AR].regnum);

  /* define synonyms for control registers:  */
  for (i = REG_CR; i < REG_CR + NELEMS (cr); ++i)
    md.regsym[i] = declare_register (cr[i - REG_CR].name,
	REG_CR + cr[i - REG_CR].regnum);

  declare_register ("gp", REG_GR +  1);
  declare_register ("sp", REG_GR + 12);
  declare_register ("rp", REG_BR +  0);

  /* pseudo-registers used to specify unwind info:  */
  declare_register ("psp", REG_PSP);

  declare_register_set ("ret", 4, REG_GR + 8);
  declare_register_set ("farg", 8, REG_FR + 8);
  declare_register_set ("fret", 8, REG_FR + 8);

  for (i = 0; i < NELEMS (const_bits); ++i)
  {
    err = hash_insert (md.const_hash, const_bits[i].name,
	(PTR) (const_bits + i));
    if (err)
      as_fatal ("Inserting \"%s\" into constant hash table failed: %s",
	  name, err);
  }

  /* Set the architecture and machine depending on defaults and command line
     options.  */
  if (md.flags & EF_IA_64_ABI64)
    ok = bfd_set_arch_mach (stdoutput, bfd_arch_ia64, bfd_mach_ia64_elf64);
  else
    ok = bfd_set_arch_mach (stdoutput, bfd_arch_ia64, bfd_mach_ia64_elf32);

  if (! ok)
    as_warn (_("Could not set architecture and machine"));

  /* Set the pointer size and pointer shift size depending on md.flags */

  if (md.flags & EF_IA_64_ABI64)
  {
    md.pointer_size = 8;         /* pointers are 8 bytes */
    md.pointer_size_shift = 3;   /* alignment is 8 bytes = 2^2 */
  }
  else
  {
    md.pointer_size = 4;         /* pointers are 4 bytes */
    md.pointer_size_shift = 2;   /* alignment is 4 bytes = 2^2 */
  }

  md.mem_offset.hint = 0;
  md.path = 0;
  md.maxpaths = 0;
  md.entry_labels = NULL;
}

#endif























/* Return true if idesc is a conditional branch instruction.  This excludes
   the modulo scheduled branches, and br.ia.  Mod-sched branches are excluded
   because they always read/write resources regardless of the value of the
   qualifying predicate.  br.ia must always use p0, and hence is always
   taken.  Thus this function returns true for branches which can fall
   through, and which use no resources if they do fall through.  */

static int
is_conditional_branch (idesc)
     t_ia64_opcode *idesc;
{
  /* br is a conditional branch.  Everything that starts with br. except
     br.ia, br.c{loop,top,exit}, and br.w{top,exit} is a conditional branch.
     Everything that starts with brl is a conditional branch.  */
  return (idesc->desc[0] == 'b' && idesc->desc[1] == 'r'
	  && (idesc->desc[2] == '\0'
	      || (idesc->desc[2] == '.' && idesc->desc[3] != 'i'
		  && idesc->desc[3] != 'c' && idesc->desc[3] != 'w')
	      || idesc->desc[2] == 'l'
	      /* br.cond, br.call, br.clr  */
	      || (idesc->desc[2] == '.' && idesc->desc[3] == 'c'
		  && (idesc->desc[4] == 'a' || idesc->desc[4] == 'o'
		      || (idesc->desc[4] == 'l' && idesc->desc[5] == 'r')))));
}

/* Return whether the given opcode is a taken branch.  If there's any doubt,
   returns zero.  */

static int
is_taken_branch (idesc)
     t_ia64_opcode *idesc;
{
  return ((is_conditional_branch (idesc) && IA64INSPRED(CURR_SLOT) == IA64_P0)
	  || strncmp (idesc->desc, "br.ia", 5) == 0);
}

/* Return whether the given opcode is an interruption or rfi.  If there's any
   doubt, returns zero.  */

static int
is_interruption_or_rfi (idesc)
     t_ia64_opcode *idesc;
{
  if (strcmp (idesc->desc, "rfi") == 0)
    return 1;
  return 0;
}

















/* Returns the index of the given dependency in the opcode's list of chks, or
   -1 if there is no dependency.  */

static int
depends_on (depind, idesc)
     int depind;
     t_ia64_opcode *idesc;
{
  int i;
  const t_ia64_opcode_dependency *dep = ia64_get_dependencies(idesc);
  for (i = 0; i < dep->nchks; i++)
    {
      if (depind == DEP (dep->chks[i]))
	return i;
    }
  return -1;
}

/* Determine a set of specific resources used for a particular resource
   class.  Returns the number of specific resources identified  For those
   cases which are not determinable statically, the resource returned is
   marked nonspecific.

   Meanings of value in 'NOTE':
   1) only read/write when the register number is explicitly encoded in the
   insn.
   2) only read CFM when accessing a rotating GR, FR, or PR.  mov pr only
   accesses CFM when qualifying predicate is in the rotating region.
   3) general register value is used to specify an indirect register; not
   determinable statically.
   4) only read the given resource when bits 7:0 of the indirect index
   register value does not match the register number of the resource; not
   determinable statically.
   5) all rules are implementation specific.
   6) only when both the index specified by the reader and the index specified
   by the writer have the same value in bits 63:61; not determinable
   statically.
   7) only access the specified resource when the corresponding mask bit is
   set
   8) PSR.dfh is only read when these insns reference FR32-127.  PSR.dfl is
   only read when these insns reference FR2-31
   9) PSR.mfl is only written when these insns write FR2-31.  PSR.mfh is only
   written when these insns write FR32-127
   10) The PSR.bn bit is only accessed when one of GR16-31 is specified in the
   instruction
   11) The target predicates are written independently of PR[qp], but source
   registers are only read if PR[qp] is true.  Since the state of PR[qp]
   cannot statically be determined, all source registers are marked used.
   12) This insn only reads the specified predicate register when that
   register is the PR[qp].
   13) This reference to ld-c only applies to teh GR whose value is loaded
   with data returned from memory, not the post-incremented address register.
   14) The RSE resource includes the implementation-specific RSE internal
   state resources.  At least one (and possibly more) of these resources are
   read by each instruction listed in IC:rse-readers.  At least one (and
   possibly more) of these resources are written by each insn listed in
   IC:rse-writers.
   15+16) Represents reserved instructions, which the assembler does not
   generate.

   Memory resources (i.e. locations in memory) are *not* marked or tracked by
   this code; there are no dependency violations based on memory access.
*/

#define MAX_SPECS 256
#define DV_CHK 1
#define DV_REG 0

static int
specify_resource (dep, idesc, type, specs, note, path)
     const t_ia64_dependency *dep;
     t_ia64_opcode *idesc;
     int type;                         /* is this a DV chk or a DV reg? */
     struct rsrc specs[MAX_SPECS];     /* returned specific resources */
     int note;                         /* resource note for this insn's usage */
     int path;                         /* which execution path to examine */
{
  int count = 0;
  int i;
  int rsrc_write = 0;
  struct rsrc tmpl;

  if (dep->mode == IA64_DV_WAW
      || (dep->mode == IA64_DV_RAW && type == DV_REG)
      || (dep->mode == IA64_DV_WAR && type == DV_CHK))
    rsrc_write = 1;

  /* template for any resources we identify */
  tmpl.dependency = dep;
  tmpl.note = note;
  tmpl.insn_srlz = tmpl.data_srlz = 0;
  tmpl.qp_regno = IA64INSPRED(CURR_SLOT)-IA64_P0;
  tmpl.link_to_qp_branch = 1;
//  tmpl.mem_offset.hint = 0;
  tmpl.specific = 1;
  tmpl.index = 0;
  tmpl.cmp_type = CMP_NONE;

#define UNHANDLED \
  FATAL (("Unhandled dependency %s for %s (%s), note %d", \
      dep->name, idesc->desc, (rsrc_write?"write":"read"), note))
//known gaan we (nog) niet simuleren
//#define KNOWN(REG) (gr_values[REG].known && gr_values[REG].path >= path)

    /* we don't need to track these */
    if (dep->semantics == IA64_DVS_NONE)
      return 0;

  switch (dep->specifier)
  {
    case IA64_RS_AR_K:
      if (note == 1)
      {
	if (IA64_IS_AR(rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc)))
	    /*idesc->operands[!rsrc_write] == IA64_OPND_AR3*/)
	{
	  int regno = (rsrc_write?IA64INSOUT1(CURR_SLOT):IA64INSIN1(CURR_SLOT))-IA64_AR; //CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_AR;
	  if (regno >= 0 && regno <= 7)
	  {
	    specs[count] = tmpl;
	    specs[count++].index = regno;
	  }
	}
      }
      else if (note == 0)
      {
	for (i = 0; i < 8; i++)
	{
	  specs[count] = tmpl;
	  specs[count++].index = i;
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_AR_UNAT:
      /* This is a mov =AR or mov AR= instruction.  */
      if (IA64_IS_AR(rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1((idesc))) 
	  /*idesc->operands[!rsrc_write] == IA64_OPND_AR3*/)
      {
	int regno = (rsrc_write?IA64INSOUT1(CURR_SLOT):IA64INSIN1(CURR_SLOT))-IA64_AR; //CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_AR;
	if (regno == IA64_AR_UNAT)
	{
	  specs[count++] = tmpl;
	}
      }
      else
      {
	/* This is a spill/fill, or other instruction that modifies the
	   unat register.  */

	/* Unless we can determine the specific bits used, mark the whole
	   thing; bits 8:3 of the memory address indicate the bit used in
	   UNAT.  The .mem.offset hint may be used to eliminate a small
	   subset of conflicts.  */
	specs[count] = tmpl;
#if 0	
	if (md.mem_offset.hint)
	{
	  if (md.debug_dv)
	    fprintf (stderr, "  Using hint for spill/fill\n");
	  /* The index isn't actually used, just set it to something
	     approximating the bit index.  */
	  specs[count].index = (md.mem_offset.offset >> 3) & 0x3F;
	  specs[count].mem_offset.hint = 1;
	  specs[count].mem_offset.offset = md.mem_offset.offset;
	  specs[count++].mem_offset.base = md.mem_offset.base;
	}
	else
	{
	  specs[count++].specific = 0;
	}
#endif	
      }
      break;

    case IA64_RS_AR:
      if (note == 1)
      {
	if (IA64_IS_AR(rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) 
	    /*idesc->operands[!rsrc_write] == IA64_OPND_AR3*/)
	{
	  int regno = (rsrc_write?IA64INSOUT1(CURR_SLOT):IA64INSIN1(CURR_SLOT))-IA64_AR; //CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_AR;
	  if ((regno >= 8 && regno <= 15)
	      || (regno >= 20 && regno <= 23)
	      || (regno >= 31 && regno <= 39)
	      || (regno >= 41 && regno <= 47)
	      || (regno >= 67 && regno <= 111))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = regno;
	  }
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_ARb:
      if (note == 1)
      {
	if (IA64_IS_AR(rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) 
	    /*idesc->operands[!rsrc_write] == IA64_OPND_AR3*/)
	{
	  int regno = (rsrc_write?IA64INSOUT1(CURR_SLOT):IA64INSIN1(CURR_SLOT))-IA64_AR; //CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_AR;
	  if ((regno >= 48 && regno <= 63)
	      || (regno >= 112 && regno <= 127))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = regno;
	  }
	}
      }
      else if (note == 0)
      {
	for (i = 48; i < 64; i++)
	{
	  specs[count] = tmpl;
	  specs[count++].index = i;
	}
	for (i = 112; i < 128; i++)
	{
	  specs[count] = tmpl;
	  specs[count++].index = i;
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_BR:
      if (note != 1)
      {
	UNHANDLED;
      }
      else
      {
	if (rsrc_write)
	{
	  if(IA64_IS_BR(IA64INSOUT1(Ia64MakeTmpIns(idesc))))
	  {
	    specs[count] = tmpl;
	    specs[count++].index =
	      IA64INSOUT1(CURR_SLOT) - IA64_BR;
	  }
	  if(IA64_IS_BR(IA64INSOUT2(Ia64MakeTmpIns(idesc))))
	  {
	    specs[count] = tmpl;
	    specs[count++].index =
	      IA64INSOUT2(CURR_SLOT) - IA64_BR;
	  }
 
/*	  
	  for (i = 0; i < idesc->num_outputs; i++)
	    if (idesc->operands[i] == IA64_OPND_B1
		|| idesc->operands[i] == IA64_OPND_B2)
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		CURR_SLOT.opnd[i].X_add_number - REG_BR;
	    }
*/	    
	}
	else
	{
          if(IA64_IS_BR(IA64INSIN1(Ia64MakeTmpIns(idesc))))
	  {
	    specs[count] = tmpl;
	    specs[count++].index =
	      IA64INSIN1(CURR_SLOT) - IA64_BR;
	  }
          if(IA64_IS_BR(IA64INSIN2(Ia64MakeTmpIns(idesc))))
	  {
	    specs[count] = tmpl;
	    specs[count++].index =
	      IA64INSIN2(CURR_SLOT) - IA64_BR;
	  }
          if(IA64_IS_BR(IA64INSIN3(Ia64MakeTmpIns(idesc))))
	  {
	    specs[count] = tmpl;
	    specs[count++].index =
	      IA64INSIN3(CURR_SLOT) - IA64_BR;
	  }


/*	  
	  for (i = idesc->num_outputs; i < NELEMS (idesc->operands); i++)
	    if (idesc->operands[i] == IA64_OPND_B1
		|| idesc->operands[i] == IA64_OPND_B2)
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		CURR_SLOT.opnd[i].X_add_number - REG_BR;
	    }
*/	    
	}
      }
      break;

    case IA64_RS_CPUID: /* four or more registers */
      if (note == 3)
      {
	if ((rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) == IA64_CPUID 
	    /*idesc->operands[!rsrc_write] == IA64_OPND_CPUID_R3*/)
	{
/*	  
	  int regno = (rsrc_write?IA64INSOUT1(CURR_SLOT):IA64INSIN1(CURR_SLOT))-IA64_GR; 
	              //CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_GR;

	  if (regno >= 0 && regno < IA64_COUNT_GR
	      && KNOWN (regno))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = gr_values[regno].value & 0xFF;
	  }
	  else
	  {
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  }
*/
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_DBR: /* four or more registers */
      if (note == 3)
      {
	if ((rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) == IA64_DBR
	      /*idesc->operands[!rsrc_write] == IA64_OPND_DBR_R3*/)
	{
/*	
	  int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_GR;
	  if (regno >= 0 && regno < NELEMS (gr_values)
	      && KNOWN (regno))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = gr_values[regno].value & 0xFF;
	  }
	  else
	  {
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  }
*/	  
	  specs[count] = tmpl;
          specs[count++].specific = 0;

	  
	}
      }
      else if (note == 0 && !rsrc_write)
      {
	specs[count] = tmpl;
	specs[count++].specific = 0;
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_IBR: /* four or more registers */
      if (note == 3)
      {
	if ((rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) == IA64_IBR
	    /*idesc->operands[!rsrc_write] == IA64_OPND_IBR_R3*/)
	{
/*	  
	  int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_GR;
	  if (regno >= 0 && regno < NELEMS (gr_values)
	      && KNOWN (regno))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = gr_values[regno].value & 0xFF;
	  }
	  else
	  {
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  }
*/
          specs[count] = tmpl;
          specs[count++].specific = 0;
	  
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_MSR:
      if (note == 5)
      {
	/* These are implementation specific.  Force all references to
	   conflict with all other references.  */
	specs[count] = tmpl;
	specs[count++].specific = 0;
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_PKR: /* 16 or more registers */
      if (note == 3 || note == 4)
      {
	if ((rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) == IA64_PKR
	    /*idesc->operands[!rsrc_write] == IA64_OPND_PKR_R3*/)
	{
#if 0  
	  int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_GR;
	  if (regno >= 0 && regno < NELEMS (gr_values)
	      && KNOWN (regno))
	  {
	    if (note == 3)
	    {
	      specs[count] = tmpl;
	      specs[count++].index = gr_values[regno].value & 0xFF;
	    }
	    else
	      for (i = 0; i < NELEMS (gr_values); i++)
	      {
		/* Uses all registers *except* the one in R3.  */
		if ((unsigned)i != (gr_values[regno].value & 0xFF))
		{
		  specs[count] = tmpl;
		  specs[count++].index = i;
		}
	      }
	  }
	  else
	  {
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  }
#endif
	  specs[count] = tmpl;
	  specs[count++].specific = 0;
	}
      }
      else if (note == 0)
      {
	/* probe et al.  */
	specs[count] = tmpl;
	specs[count++].specific = 0;
      }
      break;

    case IA64_RS_PMC: /* four or more registers */
      if (note == 3)
      {
	if ((rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) == IA64_PMC ||
	    (!rsrc_write && IA64INSIN1(Ia64MakeTmpIns(idesc)) == IA64_PMD)
	    /*idesc->operands[!rsrc_write] == IA64_OPND_PMC_R3
	    || (!rsrc_write && idesc->operands[1] == IA64_OPND_PMD_R3)*/)

	{
/*	  
	  int index = ((idesc->operands[1] == IA64_OPND_R3 && !rsrc_write)
	      ? 1 : !rsrc_write);
	  int regno = CURR_SLOT.opnd[index].X_add_number - REG_GR;
	  if (regno >= 0 && regno < NELEMS (gr_values)
	      && KNOWN (regno))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = gr_values[regno].value & 0xFF;
	  }
	  else
	  {
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  }
*/
	  specs[count] = tmpl;
          specs[count++].specific = 0;

	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_PMD: /* four or more registers */
      if (note == 3)
      {
	if ((rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc))) == IA64_PMD
	    /*idesc->operands[!rsrc_write] == IA64_OPND_PMD_R3*/)
	{
/*	  
	  int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_GR;
	  if (regno >= 0 && regno < NELEMS (gr_values)
	      && KNOWN (regno))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = gr_values[regno].value & 0xFF;
	  }
	  else
	  {
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  }
*/
	  specs[count] = tmpl;
          specs[count++].specific = 0;

	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_RR: /* eight registers */
      if (note == 6)
      {
	if (IA64_IS_RR(rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc)))
            /*idesc->operands[!rsrc_write] == IA64_OPND_RR_R3*/)
	{
/*	  
	  int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_GR;
	  if (regno >= 0 && regno < NELEMS (gr_values)
	      && KNOWN (regno))
	  {
	    specs[count] = tmpl;
	    specs[count++].index = (gr_values[regno].value >> 61) & 0x7;
	  }
	  else
	  {
	    specs[count] = tmpl;
	    specs[count++].specific = 0;
	  }
*/
	  specs[count] = tmpl;
          specs[count++].specific = 0;

	}
      }
      else if (note == 0 && !rsrc_write)
      {
	specs[count] = tmpl;
	specs[count++].specific = 0;
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_CR_IRR:
      UNHANDLED;
#if 0      
      if (note == 0)
      {
	/* handle mov-from-CR-IVR; it's a read that writes CR[IRR] */
	int regno = IA64INSCURR_SLOT.opnd[1].X_add_number - REG_CR;
	if (rsrc_write
	    && idesc->operands[1] == IA64_OPND_CR3
	    && regno == CR_IVR)
	{
	  for (i = 0; i < 4; i++)
	  {
	    specs[count] = tmpl;
	    specs[count++].index = CR_IRR0 + i;
	  }
	}
      }
      else if (note == 1)
      {
	int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_CR;
	if (idesc->operands[!rsrc_write] == IA64_OPND_CR3
	    && regno >= CR_IRR0
	    && regno <= CR_IRR3)
	{
	  specs[count] = tmpl;
	  specs[count++].index = regno;
	}
      }
      else
      {
	UNHANDLED;
      }
#endif      
      break;

    case IA64_RS_CR_LRR:
      UNHANDLED;
#if 0      
      if (note != 1)
      {
	UNHANDLED;
      }
      else
      {
	int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_CR;
	if (idesc->operands[!rsrc_write] == IA64_OPND_CR3
	    && (regno == CR_LRR0 || regno == CR_LRR1))
	{
	  specs[count] = tmpl;
	  specs[count++].index = regno;
	}
      }
#endif      
      break;

    case IA64_RS_CR:
      if (note == 1)
      {
	if (IA64_IS_CR(rsrc_write?IA64INSOUT1(Ia64MakeTmpIns(idesc)):IA64INSIN1(Ia64MakeTmpIns(idesc)))
	    /*idesc->operands[!rsrc_write] == IA64_OPND_CR3*/)
	{
	  specs[count] = tmpl;
	  specs[count++].index =(rsrc_write?IA64INSOUT1(CURR_SLOT):IA64INSIN1(CURR_SLOT))-IA64_CR;
          //CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_CR;
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_FR:
    case IA64_RS_FRb:
      if (note != 1)
      {
	UNHANDLED;
      }
      else if (rsrc_write)
      {
	if (dep->specifier == IA64_RS_FRb
	    && IA64_IS_FR(IA64INSOUT1(Ia64MakeTmpIns(idesc))) /*idesc->operands[0] == IA64_OPND_F1*/)
	{
	  specs[count] = tmpl;
	  specs[count++].index = IA64INSOUT1(CURR_SLOT) - IA64_FR;//CURR_SLOT.opnd[0].X_add_number - REG_FR;
	}
      }
      else
      {
	if(IA64_IS_FR(IA64INSIN1(Ia64MakeTmpIns(idesc))))
	{
	  specs[count] = tmpl;
	  specs[count++].index =
	    IA64INSIN1(CURR_SLOT) - IA64_FR;
	}
	if(IA64_IS_FR(IA64INSIN2(Ia64MakeTmpIns(idesc))))
	{
	  specs[count] = tmpl;
	  specs[count++].index =
	    IA64INSIN2(CURR_SLOT) - IA64_FR;
	}
	if(IA64_IS_FR(IA64INSIN3(Ia64MakeTmpIns(idesc))))
	{
	  specs[count] = tmpl;
	  specs[count++].index =
	    IA64INSIN3(CURR_SLOT) - IA64_FR;
	}

	/*	
		for (i = idesc->num_outputs; i < NELEMS (idesc->operands); i++)
		{
		if (idesc->operands[i] == IA64_OPND_F2
		|| idesc->operands[i] == IA64_OPND_F3
		|| idesc->operands[i] == IA64_OPND_F4)
		{
		specs[count] = tmpl;
		specs[count++].index =
		CURR_SLOT.opnd[i].X_add_number - REG_FR;
		}
		}
	 */	
      }
      break;

    case IA64_RS_GR:
      if (note == 13)
      {
	/* This reference applies only to the GR whose value is loaded with
	   data returned from memory.  */
	specs[count] = tmpl;
	specs[count++].index = IA64INSOUT1(CURR_SLOT) - IA64_GR; //CURR_SLOT.opnd[0].X_add_number - REG_GR;
      }
      else if (note == 1)
      {
	if (rsrc_write)
	{
	  if(IA64_IS_GR(IA64INSOUT1(Ia64MakeTmpIns(idesc))))
	  {
	    specs[count] = tmpl;
	    specs[count++].index =
	      IA64INSOUT1(CURR_SLOT) - IA64_GR;
	  }
          if(IA64_IS_GR(IA64INSOUT2(Ia64MakeTmpIns(idesc))))
	  {
	    specs[count] = tmpl;
	    specs[count++].index =
	      IA64INSOUT2(CURR_SLOT) - IA64_GR;
	  }

/*	  
	  for (i = 0; i < idesc->num_outputs; i++)
	    if (idesc->operands[i] == IA64_OPND_R1
		|| idesc->operands[i] == IA64_OPND_R2
		|| idesc->operands[i] == IA64_OPND_R3)
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		CURR_SLOT.opnd[i].X_add_number - REG_GR;
	    }
*/	    

	  if (idesc->flags & IA64_OPCODE_POSTINC)
	  {
	    if(IA64_IS_GR(IA64INSOUT1(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSOUT1(CURR_SLOT) - IA64_GR;
	    }
            if(IA64_IS_GR(IA64INSOUT2(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSOUT2(CURR_SLOT) - IA64_GR;
	    }
            if(IA64_IS_GR(IA64INSIN1(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSIN1(CURR_SLOT) - IA64_GR;
	    }
            if(IA64_IS_GR(IA64INSIN2(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSIN2(CURR_SLOT) - IA64_GR;
	    }
	    if(IA64_IS_GR(IA64INSIN3(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSIN3(CURR_SLOT) - IA64_GR;
	    }


	  }
/*	  
	  if (idesc->flags & IA64_OPCODE_POSTINC)
	    for (i = 0; i < NELEMS (idesc->operands); i++)
	      if (idesc->operands[i] == IA64_OPND_MR3)
	      {
		specs[count] = tmpl;
		specs[count++].index =
		  CURR_SLOT.opnd[i].X_add_number - REG_GR;
	      }
*/	      
	}
	else
	{
	  /* Look for anything that reads a GR.  */

            if(IA64_IS_GR(IA64INSIN1(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSIN1(CURR_SLOT) - IA64_GR;
	    }
	    if(IA64_IS_GR(IA64INSIN2(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSIN2(CURR_SLOT) - IA64_GR;
	    }
	    if(IA64_IS_GR(IA64INSIN3(Ia64MakeTmpIns(idesc))))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		IA64INSIN3(CURR_SLOT) - IA64_GR;
	    }


#if 0
	  for (i = 0; i < NELEMS (idesc->operands); i++)
	  {
	    if (idesc->operands[i] == IA64_OPND_MR3
		|| idesc->operands[i] == IA64_OPND_CPUID_R3
		|| idesc->operands[i] == IA64_OPND_DBR_R3
		|| idesc->operands[i] == IA64_OPND_IBR_R3
		|| idesc->operands[i] == IA64_OPND_MSR_R3
		|| idesc->operands[i] == IA64_OPND_PKR_R3
		|| idesc->operands[i] == IA64_OPND_PMC_R3
		|| idesc->operands[i] == IA64_OPND_PMD_R3
		|| idesc->operands[i] == IA64_OPND_RR_R3
		|| ((i >= idesc->num_outputs)
		  && (idesc->operands[i] == IA64_OPND_R1
		    || idesc->operands[i] == IA64_OPND_R2
		    || idesc->operands[i] == IA64_OPND_R3
		    /* addl source register.  */
		    || idesc->operands[i] == IA64_OPND_R3_2)))
	    {
	      specs[count] = tmpl;
	      specs[count++].index =
		CURR_SLOT.opnd[i].X_add_number - REG_GR;
	    }
	  }
#endif	
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

      /* This is the same as IA64_RS_PRr, except that the register range is
	 from 1 - 15, and there are no rotating register reads/writes here.  */
    case IA64_RS_PR:
      if (note == 0)
      {
	for (i = 1; i < 16; i++)
	{
	  specs[count] = tmpl;
	  specs[count++].index = i;
	}
      }
      else if (note == 7)
      {
	t_uint64 mask = 0;
	/* Mark only those registers indicated by the mask.  */
	if (rsrc_write)
	{
	  mask = CURR_SLOT.opnd[2].X_add_number;
	  for (i = 1; i < 16; i++)
	    if (mask & ((t_uint64) 1 << i))
	    {
	      specs[count] = tmpl;
	      specs[count++].index = i;
	    }
	}
	else
	{
	  UNHANDLED;
	}
      }
      else if (note == 11) /* note 11 implies note 1 as well */
      {
	if (rsrc_write)
	{
	  for (i = 0; i < idesc->num_outputs; i++)
	  {
	    if (idesc->operands[i] == IA64_OPND_P1
		|| idesc->operands[i] == IA64_OPND_P2)
	    {
	      int regno = CURR_SLOT.opnd[i].X_add_number - REG_P;
	      if (regno >= 1 && regno < 16)
	      {
		specs[count] = tmpl;
		specs[count++].index = regno;
	      }
	    }
	  }
	}
	else
	{
	  UNHANDLED;
	}
      }
      else if (note == 12)
      {
	if (CURR_SLOT.qp_regno >= 1 && CURR_SLOT.qp_regno < 16)
	{
	  specs[count] = tmpl;
	  specs[count++].index = CURR_SLOT.qp_regno;
	}
      }
      else if (note == 1)
      {
	if (rsrc_write)
	{
	  int p1 = CURR_SLOT.opnd[0].X_add_number - REG_P;
	  int p2 = CURR_SLOT.opnd[1].X_add_number - REG_P;
	  int or_andcm = strstr (idesc->desc, "or.andcm") != NULL;
	  int and_orcm = strstr (idesc->desc, "and.orcm") != NULL;

	  if ((idesc->operands[0] == IA64_OPND_P1
		|| idesc->operands[0] == IA64_OPND_P2)
	      && p1 >= 1 && p1 < 16)
	  {
	    specs[count] = tmpl;
	    specs[count].cmp_type =
	      (or_andcm ? CMP_OR : (and_orcm ? CMP_AND : CMP_NONE));
	    specs[count++].index = p1;
	  }
	  if ((idesc->operands[1] == IA64_OPND_P1
		|| idesc->operands[1] == IA64_OPND_P2)
	      && p2 >= 1 && p2 < 16)
	  {
	    specs[count] = tmpl;
	    specs[count].cmp_type =
	      (or_andcm ? CMP_AND : (and_orcm ? CMP_OR : CMP_NONE));
	    specs[count++].index = p2;
	  }
	}
	else
	{
	  if (CURR_SLOT.qp_regno >= 1 && CURR_SLOT.qp_regno < 16)
	  {
	    specs[count] = tmpl;
	    specs[count++].index = CURR_SLOT.qp_regno;
	  }
	  if (idesc->operands[1] == IA64_OPND_PR)
	  {
	    for (i = 1; i < 16; i++)
	    {
	      specs[count] = tmpl;
	      specs[count++].index = i;
	    }
	  }
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

      /* This is the general case for PRs.  IA64_RS_PR and IA64_RS_PR63 are
	 simplified cases of this.  */
    case IA64_RS_PRr:
      if (note == 0)
      {
	for (i = 16; i < 63; i++)
	{
	  specs[count] = tmpl;
	  specs[count++].index = i;
	}
      }
      else if (note == 7)
      {
	t_uint64 mask = 0;
	/* Mark only those registers indicated by the mask.  */
	if (rsrc_write
	    && idesc->operands[0] == IA64_OPND_PR)
	{
	  mask = CURR_SLOT.opnd[2].X_add_number;
	  if (mask & ((t_uint64) 1 << 16))
	    for (i = 16; i < 63; i++)
	    {
	      specs[count] = tmpl;
	      specs[count++].index = i;
	    }
	}
	else if (rsrc_write
	    && idesc->operands[0] == IA64_OPND_PR_ROT)
	{
	  for (i = 16; i < 63; i++)
	  {
	    specs[count] = tmpl;
	    specs[count++].index = i;
	  }
	}
	else
	{
	  UNHANDLED;
	}
      }
      else if (note == 11) /* note 11 implies note 1 as well */
      {
	if (rsrc_write)
	{
	  for (i = 0; i < idesc->num_outputs; i++)
	  {
	    if (idesc->operands[i] == IA64_OPND_P1
		|| idesc->operands[i] == IA64_OPND_P2)
	    {
	      int regno = CURR_SLOT.opnd[i].X_add_number - REG_P;
	      if (regno >= 16 && regno < 63)
	      {
		specs[count] = tmpl;
		specs[count++].index = regno;
	      }
	    }
	  }
	}
	else
	{
	  UNHANDLED;
	}
      }
      else if (note == 12)
      {
	if (CURR_SLOT.qp_regno >= 16 && CURR_SLOT.qp_regno < 63)
	{
	  specs[count] = tmpl;
	  specs[count++].index = CURR_SLOT.qp_regno;
	}
      }
      else if (note == 1)
      {
	if (rsrc_write)
	{
	  int p1 = CURR_SLOT.opnd[0].X_add_number - REG_P;
	  int p2 = CURR_SLOT.opnd[1].X_add_number - REG_P;
	  int or_andcm = strstr (idesc->desc, "or.andcm") != NULL;
	  int and_orcm = strstr (idesc->desc, "and.orcm") != NULL;

	  if ((idesc->operands[0] == IA64_OPND_P1
		|| idesc->operands[0] == IA64_OPND_P2)
	      && p1 >= 16 && p1 < 63)
	  {
	    specs[count] = tmpl;
	    specs[count].cmp_type =
	      (or_andcm ? CMP_OR : (and_orcm ? CMP_AND : CMP_NONE));
	    specs[count++].index = p1;
	  }
	  if ((idesc->operands[1] == IA64_OPND_P1
		|| idesc->operands[1] == IA64_OPND_P2)
	      && p2 >= 16 && p2 < 63)
	  {
	    specs[count] = tmpl;
	    specs[count].cmp_type =
	      (or_andcm ? CMP_AND : (and_orcm ? CMP_OR : CMP_NONE));
	    specs[count++].index = p2;
	  }
	}
	else
	{
	  if (CURR_SLOT.qp_regno >= 16 && CURR_SLOT.qp_regno < 63)
	  {
	    specs[count] = tmpl;
	    specs[count++].index = CURR_SLOT.qp_regno;
	  }
	  if (idesc->operands[1] == IA64_OPND_PR)
	  {
	    for (i = 16; i < 63; i++)
	    {
	      specs[count] = tmpl;
	      specs[count++].index = i;
	    }
	  }
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_PSR:
      /* Verify that the instruction is using the PSR bit indicated in
	 dep->regindex.  */
      if (note == 0)
      {
	if (idesc->operands[!rsrc_write] == IA64_OPND_PSR_UM)
	{
	  if (dep->regindex < 6)
	  {
	    specs[count++] = tmpl;
	  }
	}
	else if (idesc->operands[!rsrc_write] == IA64_OPND_PSR)
	{
	  if (dep->regindex < 32
	      || dep->regindex == 35
	      || dep->regindex == 36
	      || (!rsrc_write && dep->regindex == PSR_CPL))
	  {
	    specs[count++] = tmpl;
	  }
	}
	else if (idesc->operands[!rsrc_write] == IA64_OPND_PSR_L)
	{
	  if (dep->regindex < 32
	      || dep->regindex == 35
	      || dep->regindex == 36
	      || (rsrc_write && dep->regindex == PSR_CPL))
	  {
	    specs[count++] = tmpl;
	  }
	}
	else
	{
	  /* Several PSR bits have very specific dependencies.  */
	  switch (dep->regindex)
	  {
	    default:
	      specs[count++] = tmpl;
	      break;
	    case PSR_IC:
	      if (rsrc_write)
	      {
		specs[count++] = tmpl;
	      }
	      else
	      {
		/* Only certain CR accesses use PSR.ic */
		if (idesc->operands[0] == IA64_OPND_CR3
		    || idesc->operands[1] == IA64_OPND_CR3)
		{
		  int index =
		    ((idesc->operands[0] == IA64_OPND_CR3)
		     ? 0 : 1);
		  int regno =
		    CURR_SLOT.opnd[index].X_add_number - REG_CR;

		  switch (regno)
		  {
		    default:
		      break;
		    case CR_ITIR:
		    case CR_IFS:
		    case CR_IIM:
		    case CR_IIP:
		    case CR_IPSR:
		    case CR_ISR:
		    case CR_IFA:
		    case CR_IHA:
		    case CR_IIPA:
		      specs[count++] = tmpl;
		      break;
		  }
		}
	      }
	      break;
	    case PSR_CPL:
	      if (rsrc_write)
	      {
		specs[count++] = tmpl;
	      }
	      else
	      {
		/* Only some AR accesses use cpl */
		if (idesc->operands[0] == IA64_OPND_AR3
		    || idesc->operands[1] == IA64_OPND_AR3)
		{
		  int index =
		    ((idesc->operands[0] == IA64_OPND_AR3)
		     ? 0 : 1);
		  int regno =
		    CURR_SLOT.opnd[index].X_add_number - REG_AR;

		  if (regno == AR_ITC
		      || (index == 0
			&& (regno == AR_ITC
			  || regno == AR_RSC
			  || (regno >= AR_K0
			    && regno <= AR_K7))))
		  {
		    specs[count++] = tmpl;
		  }
		}
		else
		{
		  specs[count++] = tmpl;
		}
		break;
	      }
	  }
	}
      }
      else if (note == 7)
      {
	t_uint64 mask = 0;
	if (idesc->operands[0] == IA64_OPND_IMMU24)
	{
	  mask = CURR_SLOT.opnd[0].X_add_number;
	}
	else
	{
	  UNHANDLED;
	}
	if (mask & ((t_uint64) 1 << dep->regindex))
	{
	  specs[count++] = tmpl;
	}
      }
      else if (note == 8)
      {
	int min = dep->regindex == PSR_DFL ? 2 : 32;
	int max = dep->regindex == PSR_DFL ? 31 : 127;
	/* dfh is read on FR32-127; dfl is read on FR2-31 */
	for (i = 0; i < NELEMS (idesc->operands); i++)
	{
	  if (idesc->operands[i] == IA64_OPND_F1
	      || idesc->operands[i] == IA64_OPND_F2
	      || idesc->operands[i] == IA64_OPND_F3
	      || idesc->operands[i] == IA64_OPND_F4)
	  {
	    int reg = CURR_SLOT.opnd[i].X_add_number - REG_FR;
	    if (reg >= min && reg <= max)
	    {
	      specs[count++] = tmpl;
	    }
	  }
	}
      }
      else if (note == 9)
      {
	int min = dep->regindex == PSR_MFL ? 2 : 32;
	int max = dep->regindex == PSR_MFL ? 31 : 127;
	/* mfh is read on writes to FR32-127; mfl is read on writes to
	   FR2-31 */
	for (i = 0; i < idesc->num_outputs; i++)
	{
	  if (idesc->operands[i] == IA64_OPND_F1)
	  {
	    int reg = CURR_SLOT.opnd[i].X_add_number - REG_FR;
	    if (reg >= min && reg <= max)
	    {
	      specs[count++] = tmpl;
	    }
	  }
	}
      }
      else if (note == 10)
      {
	for (i = 0; i < NELEMS (idesc->operands); i++)
	{
	  if (idesc->operands[i] == IA64_OPND_R1
	      || idesc->operands[i] == IA64_OPND_R2
	      || idesc->operands[i] == IA64_OPND_R3)
	  {
	    int regno = CURR_SLOT.opnd[i].X_add_number - REG_GR;
	    if (regno >= 16 && regno <= 31)
	    {
	      specs[count++] = tmpl;
	    }
	  }
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_AR_FPSR:
      if (idesc->operands[!rsrc_write] == IA64_OPND_AR3)
      {
	int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_AR;
	if (regno == AR_FPSR)
	{
	  specs[count++] = tmpl;
	}
      }
      else
      {
	specs[count++] = tmpl;
      }
      break;

    case IA64_RS_ARX:
      /* Handle all AR[REG] resources */
      if (note == 0 || note == 1)
      {
	int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_AR;
	if (idesc->operands[!rsrc_write] == IA64_OPND_AR3
	    && regno == dep->regindex)
	{
	  specs[count++] = tmpl;
	}
	/* other AR[REG] resources may be affected by AR accesses */
	else if (idesc->operands[0] == IA64_OPND_AR3)
	{
	  /* AR[] writes */
	  regno = CURR_SLOT.opnd[0].X_add_number - REG_AR;
	  switch (dep->regindex)
	  {
	    default:
	      break;
	    case AR_BSP:
	    case AR_RNAT:
	      if (regno == AR_BSPSTORE)
	      {
		specs[count++] = tmpl;
	      }
	    case AR_RSC:
	      if (!rsrc_write &&
		  (regno == AR_BSPSTORE
		   || regno == AR_RNAT))
	      {
		specs[count++] = tmpl;
	      }
	      break;
	  }
	}
	else if (idesc->operands[1] == IA64_OPND_AR3)
	{
	  /* AR[] reads */
	  regno = CURR_SLOT.opnd[1].X_add_number - REG_AR;
	  switch (dep->regindex)
	  {
	    default:
	      break;
	    case AR_RSC:
	      if (regno == AR_BSPSTORE || regno == AR_RNAT)
	      {
		specs[count++] = tmpl;
	      }
	      break;
	  }
	}
	else
	{
	  specs[count++] = tmpl;
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_CRX:
      /* Handle all CR[REG] resources */
      if (note == 0 || note == 1)
      {
	if (idesc->operands[!rsrc_write] == IA64_OPND_CR3)
	{
	  int regno = CURR_SLOT.opnd[!rsrc_write].X_add_number - REG_CR;
	  if (regno == dep->regindex)
	  {
	    specs[count++] = tmpl;
	  }
	  else if (!rsrc_write)
	  {
	    /* Reads from CR[IVR] affect other resources.  */
	    if (regno == CR_IVR)
	    {
	      if ((dep->regindex >= CR_IRR0
		    && dep->regindex <= CR_IRR3)
		  || dep->regindex == CR_TPR)
	      {
		specs[count++] = tmpl;
	      }
	    }
	  }
	}
	else
	{
	  specs[count++] = tmpl;
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_INSERVICE:
      /* look for write of EOI (67) or read of IVR (65) */
      if ((idesc->operands[0] == IA64_OPND_CR3
	    && CURR_SLOT.opnd[0].X_add_number - REG_CR == CR_EOI)
	  || (idesc->operands[1] == IA64_OPND_CR3
	    && CURR_SLOT.opnd[1].X_add_number - REG_CR == CR_IVR))
      {
	specs[count++] = tmpl;
      }
      break;

    case IA64_RS_GR0:
      if (note == 1)
      {
	specs[count++] = tmpl;
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_CFM:
      if (note != 2)
      {
	specs[count++] = tmpl;
      }
      else
      {
	/* Check if any of the registers accessed are in the rotating region.
	   mov to/from pr accesses CFM only when qp_regno is in the rotating
	   region */
	for (i = 0; i < NELEMS (idesc->operands); i++)
	{
	  if (idesc->operands[i] == IA64_OPND_R1
	      || idesc->operands[i] == IA64_OPND_R2
	      || idesc->operands[i] == IA64_OPND_R3)
	  {
	    int num = CURR_SLOT.opnd[i].X_add_number - REG_GR;
	    /* Assumes that md.rot.num_regs is always valid */
	    if (md.rot.num_regs > 0
		&& num > 31
		&& num < 31 + md.rot.num_regs)
	    {
	      specs[count] = tmpl;
	      specs[count++].specific = 0;
	    }
	  }
	  else if (idesc->operands[i] == IA64_OPND_F1
	      || idesc->operands[i] == IA64_OPND_F2
	      || idesc->operands[i] == IA64_OPND_F3
	      || idesc->operands[i] == IA64_OPND_F4)
	  {
	    int num = CURR_SLOT.opnd[i].X_add_number - REG_FR;
	    if (num > 31)
	    {
	      specs[count] = tmpl;
	      specs[count++].specific = 0;
	    }
	  }
	  else if (idesc->operands[i] == IA64_OPND_P1
	      || idesc->operands[i] == IA64_OPND_P2)
	  {
	    int num = CURR_SLOT.opnd[i].X_add_number - REG_P;
	    if (num > 15)
	    {
	      specs[count] = tmpl;
	      specs[count++].specific = 0;
	    }
	  }
	}
	if (CURR_SLOT.qp_regno > 15)
	{
	  specs[count] = tmpl;
	  specs[count++].specific = 0;
	}
      }
      break;

      /* This is the same as IA64_RS_PRr, except simplified to account for
	 the fact that there is only one register.  */
    case IA64_RS_PR63:
      if (note == 0)
      {
	specs[count++] = tmpl;
      }
      else if (note == 7)
      {
	t_uint64 mask = 0;
	if (idesc->operands[2] == IA64_OPND_IMM17)
	  mask = CURR_SLOT.opnd[2].X_add_number;
	if (mask & ((t_uint64) 1 << 63))
	  specs[count++] = tmpl;
      }
      else if (note == 11)
      {
	if ((idesc->operands[0] == IA64_OPND_P1
	      && CURR_SLOT.opnd[0].X_add_number - REG_P == 63)
	    || (idesc->operands[1] == IA64_OPND_P2
	      && CURR_SLOT.opnd[1].X_add_number - REG_P == 63))
	{
	  specs[count++] = tmpl;
	}
      }
      else if (note == 12)
      {
	if (CURR_SLOT.qp_regno == 63)
	{
	  specs[count++] = tmpl;
	}
      }
      else if (note == 1)
      {
	if (rsrc_write)
	{
	  int p1 = CURR_SLOT.opnd[0].X_add_number - REG_P;
	  int p2 = CURR_SLOT.opnd[1].X_add_number - REG_P;
	  int or_andcm = strstr (idesc->desc, "or.andcm") != NULL;
	  int and_orcm = strstr (idesc->desc, "and.orcm") != NULL;

	  if (p1 == 63
	      && (idesc->operands[0] == IA64_OPND_P1
		|| idesc->operands[0] == IA64_OPND_P2))
	  {
	    specs[count] = tmpl;
	    specs[count++].cmp_type =
	      (or_andcm ? CMP_OR : (and_orcm ? CMP_AND : CMP_NONE));
	  }
	  if (p2 == 63
	      && (idesc->operands[1] == IA64_OPND_P1
		|| idesc->operands[1] == IA64_OPND_P2))
	  {
	    specs[count] = tmpl;
	    specs[count++].cmp_type =
	      (or_andcm ? CMP_AND : (and_orcm ? CMP_OR : CMP_NONE));
	  }
	}
	else
	{
	  if (CURR_SLOT.qp_regno == 63)
	  {
	    specs[count++] = tmpl;
	  }
	}
      }
      else
      {
	UNHANDLED;
      }
      break;

    case IA64_RS_RSE:
      /* FIXME we can identify some individual RSE written resources, but RSE
	 read resources have not yet been completely identified, so for now
	 treat RSE as a single resource */
      if (strncmp (idesc->desc, "mov", 3) == 0)
      {
	if (rsrc_write)
	{
	  if (idesc->operands[0] == IA64_OPND_AR3
	      && CURR_SLOT.opnd[0].X_add_number - REG_AR == AR_BSPSTORE)
	  {
	    specs[count] = tmpl;
	    specs[count++].index = 0; /* IA64_RSE_BSPLOAD/RNATBITINDEX */
	  }
	}
	else
	{
	  if (idesc->operands[0] == IA64_OPND_AR3)
	  {
	    if (CURR_SLOT.opnd[0].X_add_number - REG_AR == AR_BSPSTORE
		|| CURR_SLOT.opnd[0].X_add_number - REG_AR == AR_RNAT)
	    {
	      specs[count++] = tmpl;
	    }
	  }
	  else if (idesc->operands[1] == IA64_OPND_AR3)
	  {
	    if (CURR_SLOT.opnd[1].X_add_number - REG_AR == AR_BSP
		|| CURR_SLOT.opnd[1].X_add_number - REG_AR == AR_BSPSTORE
		|| CURR_SLOT.opnd[1].X_add_number - REG_AR == AR_RNAT)
	    {
	      specs[count++] = tmpl;
	    }
	  }
	}
      }
      else
      {
	specs[count++] = tmpl;
      }
      break;

    case IA64_RS_ANY:
      /* FIXME -- do any of these need to be non-specific? */
      specs[count++] = tmpl;
      break;

    default:
      as_bad (_("Unrecognized dependency specifier %d\n"), dep->specifier);
      break;
  }

  return count;
}


















/* Clear branch flags on marked resources.  This breaks the link between the
   QP of the marking instruction and a subsequent branch on the same QP.  */

static void
clear_qp_branch_flag (mask)
     t_uint64 mask;
{
  int i;
  for (i = 0; i < regdepslen; i++)
    {
      t_uint64 bit = ((t_uint64) 1 << regdeps[i].qp_regno);
      if ((bit & mask) != 0)
	{
	  regdeps[i].link_to_qp_branch = 0;
	}
    }
}

/* Remove any mutexes which contain any of the PRs indicated in the mask.

   Any changes to a PR clears the mutex relations which include that PR.  */

static void
clear_qp_mutex (mask)
     t_uint64 mask;
{
  int i;

  i = 0;
  while (i < qp_mutexeslen)
    {
      if ((qp_mutexes[i].prmask & mask) != 0)
	{
	  if (md.debug_dv)
	    {
	      fprintf (stderr, "  Clearing mutex relation");
	      print_prmask (qp_mutexes[i].prmask);
	      fprintf (stderr, "\n");
	    }
	  qp_mutexes[i] = qp_mutexes[--qp_mutexeslen];
	}
      else
	++i;
    }
}

/* Clear implies relations which contain PRs in the given masks.
   P1_MASK indicates the source of the implies relation, while P2_MASK
   indicates the implied PR.  */

static void
clear_qp_implies (p1_mask, p2_mask)
     t_uint64 p1_mask;
     t_uint64 p2_mask;
{
  int i;

  i = 0;
  while (i < qp_implieslen)
    {
      if ((((t_uint64) 1 << qp_implies[i].p1) & p1_mask) != 0
	  || (((t_uint64) 1 << qp_implies[i].p2) & p2_mask) != 0)
	{
	  if (md.debug_dv)
	    fprintf (stderr, "Clearing implied relation PR%d->PR%d\n",
		     qp_implies[i].p1, qp_implies[i].p2);
	  qp_implies[i] = qp_implies[--qp_implieslen];
	}
      else
	++i;
    }
}

/* Add the PRs specified to the list of implied relations.  */

static void
add_qp_imply (p1, p2)
     int p1, p2;
{
  t_uint64 mask;
  t_uint64 bit;
  int i;

  /* p0 is not meaningful here.  */
  if (p1 == 0 || p2 == 0)
    abort ();

  if (p1 == p2)
    return;

  /* If it exists already, ignore it.  */
  for (i = 0; i < qp_implieslen; i++)
    {
      if (qp_implies[i].p1 == p1
	  && qp_implies[i].p2 == p2
	  && qp_implies[i].path == md.path
	  && !qp_implies[i].p2_branched)
	return;
    }

  if (qp_implieslen == qp_impliestotlen)
    {
      qp_impliestotlen += 20;
      qp_implies = (struct qp_imply *)
	xrealloc ((void *) qp_implies,
		  qp_impliestotlen * sizeof (struct qp_imply));
    }
  if (md.debug_dv)
    fprintf (stderr, "  Registering PR%d implies PR%d\n", p1, p2);
  qp_implies[qp_implieslen].p1 = p1;
  qp_implies[qp_implieslen].p2 = p2;
  qp_implies[qp_implieslen].path = md.path;
  qp_implies[qp_implieslen++].p2_branched = 0;

  /* Add in the implied transitive relations; for everything that p2 implies,
     make p1 imply that, too; for everything that implies p1, make it imply p2
     as well.  */
  for (i = 0; i < qp_implieslen; i++)
    {
      if (qp_implies[i].p1 == p2)
	add_qp_imply (p1, qp_implies[i].p2);
      if (qp_implies[i].p2 == p1)
	add_qp_imply (qp_implies[i].p1, p2);
    }
  /* Add in mutex relations implied by this implies relation; for each mutex
     relation containing p2, duplicate it and replace p2 with p1.  */
  bit = (t_uint64) 1 << p1;
  mask = (t_uint64) 1 << p2;
  for (i = 0; i < qp_mutexeslen; i++)
    {
      if (qp_mutexes[i].prmask & mask)
	add_qp_mutex ((qp_mutexes[i].prmask & ~mask) | bit);
    }
}

/* Add the PRs specified in the mask to the mutex list; this means that only
   one of the PRs can be true at any time.  PR0 should never be included in
   the mask.  */

static void
add_qp_mutex (mask)
     t_uint64 mask;
{
  if (mask & 0x1)
    abort ();

  if (qp_mutexeslen == qp_mutexestotlen)
    {
      qp_mutexestotlen += 20;
      qp_mutexes = (struct qpmutex *)
	xrealloc ((void *) qp_mutexes,
		  qp_mutexestotlen * sizeof (struct qpmutex));
    }
  if (md.debug_dv)
    {
      fprintf (stderr, "  Registering mutex on");
      print_prmask (mask);
      fprintf (stderr, "\n");
    }
  qp_mutexes[qp_mutexeslen].path = md.path;
  qp_mutexes[qp_mutexeslen++].prmask = mask;
}

static int
has_suffix_p (name, suffix)
     const char *name;
     const char *suffix;
{
  size_t namelen = strlen (name);
  size_t sufflen = strlen (suffix);

  if (namelen <= sufflen)
    return 0;
  return strcmp (name + namelen - sufflen, suffix) == 0;
}

static void
clear_register_values ()
{
  int i;
  if (md.debug_dv)
    fprintf (stderr, "  Clearing register values\n");
  for (i = 1; i < NELEMS (gr_values); i++)
    gr_values[i].known = 0;
}

/* Keep track of register values/changes which affect DV tracking.

   optimization note: should add a flag to classes of insns where otherwise we
   have to examine a group of strings to identify them.  */

static void
note_register_values (idesc)
     t_ia64_opcode *idesc;
{
  t_uint64 qp_changemask = 0;
  int i;

  /* Invalidate values for registers being written to.  */
  for (i = 0; i < idesc->num_outputs; i++)
    {
      if (idesc->operands[i] == IA64_OPND_R1
	  || idesc->operands[i] == IA64_OPND_R2
	  || idesc->operands[i] == IA64_OPND_R3)
	{
	  int regno = CURR_SLOT.opnd[i].X_add_number - REG_GR;
	  if (regno > 0 && regno < NELEMS (gr_values))
	    gr_values[regno].known = 0;
	}
      else if (idesc->operands[i] == IA64_OPND_R3_2)
	{
	  int regno = CURR_SLOT.opnd[i].X_add_number - REG_GR;
	  if (regno > 0 && regno < 4)
	    gr_values[regno].known = 0;
	}
      else if (idesc->operands[i] == IA64_OPND_P1
	       || idesc->operands[i] == IA64_OPND_P2)
	{
	  int regno = CURR_SLOT.opnd[i].X_add_number - REG_P;
	  qp_changemask |= (t_uint64) 1 << regno;
	}
      else if (idesc->operands[i] == IA64_OPND_PR)
	{
	  if (idesc->operands[2] & (t_uint64) 0x10000)
	    qp_changemask = ~(t_uint64) 0x1FFFF | idesc->operands[2];
	  else
	    qp_changemask = idesc->operands[2];
	  break;
	}
      else if (idesc->operands[i] == IA64_OPND_PR_ROT)
	{
	  if (idesc->operands[1] & ((t_uint64) 1 << 43))
	    qp_changemask = ~(t_uint64) 0xFFFFFFFFFFF | idesc->operands[1];
	  else
	    qp_changemask = idesc->operands[1];
	  qp_changemask &= ~(t_uint64) 0xFFFF;
	  break;
	}
    }

  /* Always clear qp branch flags on any PR change.  */
  /* FIXME there may be exceptions for certain compares.  */
  clear_qp_branch_flag (qp_changemask);

  /* Invalidate rotating registers on insns which affect RRBs in CFM.  */
  if (idesc->flags & IA64_OPCODE_MOD_RRBS)
    {
      qp_changemask |= ~(t_uint64) 0xFFFF;
      if (strcmp (idesc->desc, "clrrrb.pr") != 0)
	{
	  for (i = 32; i < 32 + md.rot.num_regs; i++)
	    gr_values[i].known = 0;
	}
      clear_qp_mutex (qp_changemask);
      clear_qp_implies (qp_changemask, qp_changemask);
    }
  /* After a call, all register values are undefined, except those marked
     as "safe".  */
  else if (strncmp (idesc->desc, "br.call", 6) == 0
	   || strncmp (idesc->desc, "brl.call", 7) == 0)
    {
      /* FIXME keep GR values which are marked as "safe_across_calls"  */
      clear_register_values ();
      clear_qp_mutex (~qp_safe_across_calls);
      clear_qp_implies (~qp_safe_across_calls, ~qp_safe_across_calls);
      clear_qp_branch_flag (~qp_safe_across_calls);
    }
  else if (is_interruption_or_rfi (idesc)
	   || is_taken_branch (idesc))
    {
      clear_register_values ();
      clear_qp_mutex (~(t_uint64) 0);
      clear_qp_implies (~(t_uint64) 0, ~(t_uint64) 0);
    }
  /* Look for mutex and implies relations.  */
  else if ((idesc->operands[0] == IA64_OPND_P1
	    || idesc->operands[0] == IA64_OPND_P2)
	   && (idesc->operands[1] == IA64_OPND_P1
	       || idesc->operands[1] == IA64_OPND_P2))
    {
      int p1 = CURR_SLOT.opnd[0].X_add_number - REG_P;
      int p2 = CURR_SLOT.opnd[1].X_add_number - REG_P;
      t_uint64 p1mask = (t_uint64) 1 << p1;
      t_uint64 p2mask = (t_uint64) 1 << p2;

      /* If one of the PRs is PR0, we can't really do anything.  */
      if (p1 == 0 || p2 == 0)
	{
	  if (md.debug_dv)
	    fprintf (stderr, "  Ignoring PRs due to inclusion of p0\n");
	}
      /* In general, clear mutexes and implies which include P1 or P2,
	 with the following exceptions.  */
      else if (has_suffix_p (idesc->desc, ".or.andcm")
	       || has_suffix_p (idesc->desc, ".and.orcm"))
	{
	  add_qp_mutex (p1mask | p2mask);
	  clear_qp_implies (p2mask, p1mask);
	}
      else if (has_suffix_p (idesc->desc, ".andcm")
	       || has_suffix_p (idesc->desc, ".and"))
	{
	  clear_qp_implies (0, p1mask | p2mask);
	}
      else if (has_suffix_p (idesc->desc, ".orcm")
	       || has_suffix_p (idesc->desc, ".or"))
	{
	  clear_qp_mutex (p1mask | p2mask);
	  clear_qp_implies (p1mask | p2mask, 0);
	}
      else
	{
	  clear_qp_implies (p1mask | p2mask, p1mask | p2mask);
	  if (has_suffix_p (idesc->desc, ".unc"))
	    {
	      add_qp_mutex (p1mask | p2mask);
	      if (CURR_SLOT.qp_regno != 0)
		{
		  add_qp_imply (CURR_SLOT.opnd[0].X_add_number - REG_P,
				CURR_SLOT.qp_regno);
		  add_qp_imply (CURR_SLOT.opnd[1].X_add_number - REG_P,
				CURR_SLOT.qp_regno);
		}
	    }
	  else if (CURR_SLOT.qp_regno == 0)
	    {
	      add_qp_mutex (p1mask | p2mask);
	    }
	  else
	    {
	      clear_qp_mutex (p1mask | p2mask);
	    }
	}
    }
  /* Look for mov imm insns into GRs.  */
  else if (idesc->operands[0] == IA64_OPND_R1
	   && (idesc->operands[1] == IA64_OPND_IMM22
	       || idesc->operands[1] == IA64_OPND_IMMU64)
	   && (strcmp (idesc->desc, "mov") == 0
	       || strcmp (idesc->desc, "movl") == 0))
    {
      int regno = CURR_SLOT.opnd[0].X_add_number - REG_GR;
      if (regno > 0 && regno < NELEMS (gr_values))
	{
	  gr_values[regno].known = 1;
	  gr_values[regno].value = CURR_SLOT.opnd[1].X_add_number;
	  gr_values[regno].path = md.path;
	  if (md.debug_dv)
	    {
	      fprintf (stderr, "  Know gr%d = ", regno);
	      fprintf_vma (stderr, gr_values[regno].value);
	      fputs ("\n", stderr);
	    }
	}
    }
  else
    {
      clear_qp_mutex (qp_changemask);
      clear_qp_implies (qp_changemask, qp_changemask);
    }
}

/* Return whether the given predicate registers are currently mutex.  */

static int
qp_mutex (p1, p2, path)
     int p1;
     int p2;
     int path;
{
  int i;
  t_uint64 mask;

  if (p1 != p2)
    {
      mask = ((t_uint64) 1 << p1) | (t_uint64) 1 << p2;
      for (i = 0; i < qp_mutexeslen; i++)
	{
	  if (qp_mutexes[i].path >= path
	      && (qp_mutexes[i].prmask & mask) == mask)
	    return 1;
	}
    }
  return 0;
}


















/* Return whether the given resource is in the given insn's list of chks
   Return 1 if the conflict is absolutely determined, 2 if it's a potential
   conflict.  */

static int
resources_match (rs, idesc, note, qp_regno, path)
     struct rsrc *rs;
     t_ia64_opcode *idesc;
     int note;
     int qp_regno;
     int path;
{
  struct rsrc specs[MAX_SPECS];
  int count;

  /* If the marked resource's qp_regno and the given qp_regno are mutex,
     we don't need to check.  One exception is note 11, which indicates that
     target predicates are written regardless of PR[qp].  */
  if (qp_mutex (rs->qp_regno, qp_regno, path)
      && note != 11)
    return 0;

  count = specify_resource (rs->dependency, idesc, DV_CHK, specs, note, path);
  while (count-- > 0)
    {
      /* UNAT checking is a bit more specific than other resources */
      if (rs->dependency->specifier == IA64_RS_AR_UNAT
	  && specs[count].mem_offset.hint
	  && rs->mem_offset.hint)
	{
	  if (rs->mem_offset.base == specs[count].mem_offset.base)
	    {
	      if (((rs->mem_offset.offset >> 3) & 0x3F) ==
		  ((specs[count].mem_offset.offset >> 3) & 0x3F))
		return 1;
	      else
		continue;
	    }
	}

      /* Skip apparent PR write conflicts where both writes are an AND or both
	 writes are an OR.  */
      if (rs->dependency->specifier == IA64_RS_PR
	  || rs->dependency->specifier == IA64_RS_PRr
	  || rs->dependency->specifier == IA64_RS_PR63)
	{
	  if (specs[count].cmp_type != CMP_NONE
	      && specs[count].cmp_type == rs->cmp_type)
	    {
	      if (md.debug_dv)
		fprintf (stderr, "  %s on parallel compare allowed (PR%d)\n",
			 dv_mode[rs->dependency->mode],
			 rs->dependency->specifier != IA64_RS_PR63 ?
			 specs[count].index : 63);
	      continue;
	    }
	  if (md.debug_dv)
	    fprintf (stderr,
		     "  %s on parallel compare conflict %s vs %s on PR%d\n",
		     dv_mode[rs->dependency->mode],
		     dv_cmp_type[rs->cmp_type],
		     dv_cmp_type[specs[count].cmp_type],
		     rs->dependency->specifier != IA64_RS_PR63 ?
		     specs[count].index : 63);

	}

      /* If either resource is not specific, conservatively assume a conflict
       */
      if (!specs[count].specific || !rs->specific)
	return 2;
      else if (specs[count].index == rs->index)
	return 1;
    }
#if 0
  if (md.debug_dv)
    fprintf (stderr, "  No %s conflicts\n", rs->dependency->name);
#endif

  return 0;
}













/* Indicate an instruction group break; if INSERT_STOP is non-zero, then
   insert a stop to create the break.  Update all resource dependencies
   appropriately.  If QP_REGNO is non-zero, only apply the break to resources
   which use the same QP_REGNO and have the link_to_qp_branch flag set.
   If SAVE_CURRENT is non-zero, don't affect resources marked by the current
   instruction.  */

static void
insn_group_break (insert_stop, qp_regno, save_current)
     int insert_stop;
     int qp_regno;
     int save_current;
{
  int i;

  if (insert_stop && md.num_slots_in_use > 0)
    PREV_SLOT.end_of_insn_group = 1;

  if (md.debug_dv)
    {
      fprintf (stderr, "  Insn group break%s",
	       (insert_stop ? " (w/stop)" : ""));
      if (qp_regno != 0)
	fprintf (stderr, " effective for QP=%d", qp_regno);
      fprintf (stderr, "\n");
    }

  i = 0;
  while (i < regdepslen)
    {
      const t_ia64_dependency *dep = regdeps[i].dependency;

      if (qp_regno != 0
	  && regdeps[i].qp_regno != qp_regno)
	{
	  ++i;
	  continue;
	}

      if (save_current
	  && CURR_SLOT.src_file == regdeps[i].file
	  && CURR_SLOT.src_line == regdeps[i].line)
	{
	  ++i;
	  continue;
	}

      /* clear dependencies which are automatically cleared by a stop, or
	 those that have reached the appropriate state of insn serialization */
      if (dep->semantics == IA64_DVS_IMPLIED
	  || dep->semantics == IA64_DVS_IMPLIEDF
	  || regdeps[i].insn_srlz == STATE_SRLZ)
	{
	  print_dependency ("Removing", i);
	  regdeps[i] = regdeps[--regdepslen];
	}
      else
	{
	  if (dep->semantics == IA64_DVS_DATA
	      || dep->semantics == IA64_DVS_INSTR
	      || dep->semantics == IA64_DVS_SPECIFIC)
	    {
	      if (regdeps[i].insn_srlz == STATE_NONE)
		regdeps[i].insn_srlz = STATE_STOP;
	      if (regdeps[i].data_srlz == STATE_NONE)
		regdeps[i].data_srlz = STATE_STOP;
	    }
	  ++i;
	}
    }
}











/* Add the given resource usage spec to the list of active dependencies.  */

static void
mark_resource (idesc, dep, spec, depind, path)
     t_ia64_opcode *idesc ATTRIBUTE_UNUSED;
     const t_ia64_dependency *dep ATTRIBUTE_UNUSED;
     struct rsrc *spec;
     int depind;
     int path;
{
  if (regdepslen == regdepstotlen)
    {
      regdepstotlen += 20;
      regdeps = (struct rsrc *)
	xrealloc ((void *) regdeps,
		  regdepstotlen * sizeof (struct rsrc));
    }

  regdeps[regdepslen] = *spec;
  regdeps[regdepslen].depind = depind;
  regdeps[regdepslen].path = path;
  regdeps[regdepslen].file = CURR_SLOT.src_file;
  regdeps[regdepslen].line = CURR_SLOT.src_line;

  print_dependency ("Adding", regdepslen);

  ++regdepslen;
}










static void
print_dependency (action, depind)
     const char *action;
     int depind;
{
  if (md.debug_dv)
    {
      fprintf (stderr, "  %s %s '%s'",
	       action, dv_mode[(regdeps[depind].dependency)->mode],
	       (regdeps[depind].dependency)->name);
      if (regdeps[depind].specific && regdeps[depind].index != 0)
	fprintf (stderr, " (%d)", regdeps[depind].index);
      if (regdeps[depind].mem_offset.hint)
	{
	  fputs (" ", stderr);
	  fprintf_vma (stderr, regdeps[depind].mem_offset.base);
	  fputs ("+", stderr);
	  fprintf_vma (stderr, regdeps[depind].mem_offset.offset);
	}
      fprintf (stderr, "\n");
    }
}











static void
instruction_serialization ()
{
  int i;
  if (md.debug_dv)
    fprintf (stderr, "  Instruction serialization\n");
  for (i = 0; i < regdepslen; i++)
    if (regdeps[i].insn_srlz == STATE_STOP)
      regdeps[i].insn_srlz = STATE_SRLZ;
}









static void
data_serialization ()
{
  int i = 0;
  if (md.debug_dv)
    fprintf (stderr, "  Data serialization\n");
  while (i < regdepslen)
    {
      if (regdeps[i].data_srlz == STATE_STOP
	  /* Note: as of 991210, all "other" dependencies are cleared by a
	     data serialization.  This might change with new tables */
	  || (regdeps[i].dependency)->semantics == IA64_DVS_OTHER)
	{
	  print_dependency ("Removing", i);
	  regdeps[i] = regdeps[--regdepslen];
	}
      else
	++i;
    }
}











/* Insert stops and serializations as needed to avoid DVs.  */

static void
remove_marked_resource (rs)
     struct rsrc *rs;
{
  switch (rs->dependency->semantics)
    {
    case IA64_DVS_SPECIFIC:
      if (md.debug_dv)
	fprintf (stderr, "Implementation-specific, assume worst case...\n");
      /* ...fall through...  */
    case IA64_DVS_INSTR:
      if (md.debug_dv)
	fprintf (stderr, "Inserting instr serialization\n");
      if (rs->insn_srlz < STATE_STOP)
	insn_group_break (1, 0, 0);
      if (rs->insn_srlz < STATE_SRLZ)
	{
	  int oldqp = CURR_SLOT.qp_regno;
	  t_ia64_opcode *oldidesc = CURR_SLOT.idesc;
	  /* Manually jam a srlz.i insn into the stream */
	  CURR_SLOT.qp_regno = 0;
	  CURR_SLOT.idesc = ia64_find_opcode ("srlz.i");
	  instruction_serialization ();
	  md.curr_slot = (md.curr_slot + 1) % NUM_SLOTS;
	  if (++md.num_slots_in_use >= NUM_SLOTS)
	    emit_one_bundle ();
	  CURR_SLOT.qp_regno = oldqp;
	  CURR_SLOT.idesc = oldidesc;
	}
      insn_group_break (1, 0, 0);
      break;
    case IA64_DVS_OTHER: /* as of rev2 (991220) of the DV tables, all
			    "other" types of DV are eliminated
			    by a data serialization */
    case IA64_DVS_DATA:
      if (md.debug_dv)
	fprintf (stderr, "Inserting data serialization\n");
      if (rs->data_srlz < STATE_STOP)
	insn_group_break (1, 0, 0);
      {
	int oldqp = CURR_SLOT.qp_regno;
	t_ia64_opcode *oldidesc = CURR_SLOT.idesc;
	/* Manually jam a srlz.d insn into the stream */
	CURR_SLOT.qp_regno = 0;
	CURR_SLOT.idesc = ia64_find_opcode ("srlz.d");
	data_serialization ();
	md.curr_slot = (md.curr_slot + 1) % NUM_SLOTS;
	if (++md.num_slots_in_use >= NUM_SLOTS)
	  emit_one_bundle ();
	CURR_SLOT.qp_regno = oldqp;
	CURR_SLOT.idesc = oldidesc;
      }
      break;
    case IA64_DVS_IMPLIED:
    case IA64_DVS_IMPLIEDF:
      if (md.debug_dv)
	fprintf (stderr, "Inserting stop\n");
      insn_group_break (1, 0, 0);
      break;
    default:
      break;
    }
}

















/* Check the resources used by the given opcode against the current dependency
   list.

   The check is run once for each execution path encountered.  In this case,
   a unique execution path is the sequence of instructions following a code
   entry point, e.g. the following has three execution paths, one starting
   at L0, one at L1, and one at L2.

   L0:     nop
   L1:     add
   L2:     add
   br.ret
*/

static void
check_dependencies (idesc)
     t_ia64_opcode *idesc;
{
  const t_ia64_opcode_dependency *opdeps = idesc->dependencies;
  int path;
  int i;

  /* Note that the number of marked resources may change within the
     loop if in auto mode.  */
  i = 0;
  while (i < regdepslen)
    {
      struct rsrc *rs = &regdeps[i];
      const t_ia64_dependency *dep = rs->dependency;
      int chkind;
      int note;
      int start_over = 0;

      if (dep->semantics == IA64_DVS_NONE
	  || (chkind = depends_on (rs->depind, idesc)) == -1)
	{
	  ++i;
	  continue;
	}

      note = NOTE (opdeps->chks[chkind]);

      /* Check this resource against each execution path seen thus far.  */
      for (path = 0; path <= md.path; path++)
	{
	  int matchtype;

	  /* If the dependency wasn't on the path being checked, ignore it.  */
	  if (rs->path < path)
	    continue;

	  /* If the QP for this insn implies a QP which has branched, don't
	     bother checking.  Ed. NOTE: I don't think this check is terribly
	     useful; what's the point of generating code which will only be
	     reached if its QP is zero?
	     This code was specifically inserted to handle the following code,
	     based on notes from Intel's DV checking code, where p1 implies p2.

		  mov r4 = 2
	     (p2) br.cond L
	     (p1) mov r4 = 7
	  */
	  if (CURR_SLOT.qp_regno != 0)
	    {
	      int skip = 0;
	      int implies;
	      for (implies = 0; implies < qp_implieslen; implies++)
		{
		  if (qp_implies[implies].path >= path
		      && qp_implies[implies].p1 == CURR_SLOT.qp_regno
		      && qp_implies[implies].p2_branched)
		    {
		      skip = 1;
		      break;
		    }
		}
	      if (skip)
		continue;
	    }

	  if ((matchtype = resources_match (rs, idesc, note,
					    CURR_SLOT.qp_regno, path)) != 0)
	    {
	      char msg[1024];
	      char pathmsg[256] = "";
	      char indexmsg[256] = "";
	      int certain = (matchtype == 1 && CURR_SLOT.qp_regno == 0);

	      if (path != 0)
		sprintf (pathmsg, " when entry is at label '%s'",
			 md.entry_labels[path - 1]);
	      if (rs->specific && rs->index != 0)
		sprintf (indexmsg, ", specific resource number is %d",
			 rs->index);
	      sprintf (msg, "Use of '%s' %s %s dependency '%s' (%s)%s%s",
		       idesc->desc,
		       (certain ? "violates" : "may violate"),
		       dv_mode[dep->mode], dep->name,
		       dv_sem[dep->semantics],
		       pathmsg, indexmsg);

	      if (md.explicit_mode)
		{
		  as_warn ("%s", msg);
		  if (path < md.path)
		    as_warn (_("Only the first path encountering the conflict "
			       "is reported"));
		  as_warn_where (rs->file, rs->line,
				 _("This is the location of the "
				   "conflicting usage"));
		  /* Don't bother checking other paths, to avoid duplicating
		     the same warning */
		  break;
		}
	      else
		{
		  if (md.debug_dv)
		    fprintf (stderr, "%s @ %s:%d\n", msg, rs->file, rs->line);

		  remove_marked_resource (rs);

		  /* since the set of dependencies has changed, start over */
		  /* FIXME -- since we're removing dvs as we go, we
		     probably don't really need to start over...  */
		  start_over = 1;
		  break;
		}
	    }
	}
      if (start_over)
	i = 0;
      else
	++i;
    }
}













/* Register new dependencies based on the given opcode.  */

static void
mark_resources (idesc)
     t_ia64_opcode *idesc;
{
  int i;
  const t_ia64_opcode_dependency *opdeps = idesc->dependencies;
  int add_only_qp_reads = 0;

  /* A conditional branch only uses its resources if it is taken; if it is
     taken, we stop following that path.  The other branch types effectively
     *always* write their resources.  If it's not taken, register only QP
     reads.  */
  if (is_conditional_branch (idesc) || is_interruption_or_rfi (idesc))
    {
      add_only_qp_reads = 1;
    }

  if (md.debug_dv)
    fprintf (stderr, "Registering '%s' resource usage\n", idesc->desc);

  for (i = 0; i < opdeps->nregs; i++)
    {
      const t_ia64_dependency *dep;
      struct rsrc specs[MAX_SPECS];
      int note;
      int path;
      int count;

      dep = ia64_find_dependency (opdeps->regs[i]);
      note = NOTE (opdeps->regs[i]);

      if (add_only_qp_reads
	  && !(dep->mode == IA64_DV_WAR
	       && (dep->specifier == IA64_RS_PR
		   || dep->specifier == IA64_RS_PRr
		   || dep->specifier == IA64_RS_PR63)))
	continue;

      count = specify_resource (dep, idesc, DV_REG, specs, note, md.path);

#if 0
      if (md.debug_dv && !count)
	fprintf (stderr, "  No %s %s usage found (path %d)\n",
		 dv_mode[dep->mode], dep->name, md.path);
#endif

      while (count-- > 0)
	{
	  mark_resource (idesc, dep, &specs[count],
			 DEP (opdeps->regs[i]), md.path);
	}

      /* The execution path may affect register values, which may in turn
	 affect which indirect-access resources are accessed.  */
      switch (dep->specifier)
	{
	default:
	  break;
	case IA64_RS_CPUID:
	case IA64_RS_DBR:
	case IA64_RS_IBR:
	case IA64_RS_MSR:
	case IA64_RS_PKR:
	case IA64_RS_PMC:
	case IA64_RS_PMD:
	case IA64_RS_RR:
	  for (path = 0; path < md.path; path++)
	    {
	      count = specify_resource (dep, idesc, DV_REG, specs, note, path);
	      while (count-- > 0)
		mark_resource (idesc, dep, &specs[count],
			       DEP (opdeps->regs[i]), path);
	    }
	  break;
	}
    }
}



















/* Remove dependencies when they no longer apply.  */

static void
update_dependencies (idesc)
     t_ia64_opcode *idesc;
{
  int i;

  if (strcmp (idesc->desc, "srlz.i") == 0)
    {
      instruction_serialization ();
    }
  else if (strcmp (idesc->desc, "srlz.d") == 0)
    {
      data_serialization ();
    }
  else if (is_interruption_or_rfi (idesc)
	   || is_taken_branch (idesc))
    {
      /* Although technically the taken branch doesn't clear dependencies
	 which require a srlz.[id], we don't follow the branch; the next
	 instruction is assumed to start with a clean slate.  */
      regdepslen = 0;
      md.path = 0;
    }
  else if (is_conditional_branch (idesc)
	   && CURR_SLOT.qp_regno != 0)
    {
      int is_call = strstr (idesc->desc, ".call") != NULL;

      for (i = 0; i < qp_implieslen; i++)
	{
	  /* If the conditional branch's predicate is implied by the predicate
	     in an existing dependency, remove that dependency.  */
	  if (qp_implies[i].p2 == CURR_SLOT.qp_regno)
	    {
	      int depind = 0;
	      /* Note that this implied predicate takes a branch so that if
		 a later insn generates a DV but its predicate implies this
		 one, we can avoid the false DV warning.  */
	      qp_implies[i].p2_branched = 1;
	      while (depind < regdepslen)
		{
		  if (regdeps[depind].qp_regno == qp_implies[i].p1)
		    {
		      print_dependency ("Removing", depind);
		      regdeps[depind] = regdeps[--regdepslen];
		    }
		  else
		    ++depind;
		}
	    }
	}
      /* Any marked resources which have this same predicate should be
	 cleared, provided that the QP hasn't been modified between the
	 marking instruction and the branch.  */
      if (is_call)
	{
	  insn_group_break (0, CURR_SLOT.qp_regno, 1);
	}
      else
	{
	  i = 0;
	  while (i < regdepslen)
	    {
	      if (regdeps[i].qp_regno == CURR_SLOT.qp_regno
		  && regdeps[i].link_to_qp_branch
		  && (regdeps[i].file != CURR_SLOT.src_file
		      || regdeps[i].line != CURR_SLOT.src_line))
		{
		  /* Treat like a taken branch */
		  print_dependency ("Removing", i);
		  regdeps[i] = regdeps[--regdepslen];
		}
	      else
		++i;
	    }
	}
    }
}














/* Examine the current instruction for dependency violations.  */

static int
check_dv (idesc)
     t_ia64_opcode *idesc;
{
  if (md.debug_dv)
    {
      fprintf (stderr, "Checking %s for violations (line %d, %d/%d)\n",
	       idesc->desc, CURR_SLOT.src_line,
	       idesc->dependencies->nchks,
	       idesc->dependencies->nregs);
    }

  /* Look through the list of currently marked resources; if the current
     instruction has the dependency in its chks list which uses that resource,
     check against the specific resources used.  */
  check_dependencies (idesc);

  /* Look up the instruction's regdeps (RAW writes, WAW writes, and WAR reads),
     then add them to the list of marked resources.  */
  mark_resources (idesc);

  /* There are several types of dependency semantics, and each has its own
     requirements for being cleared

     Instruction serialization (insns separated by interruption, rfi, or
     writer + srlz.i + reader, all in separate groups) clears DVS_INSTR.

     Data serialization (instruction serialization, or writer + srlz.d +
     reader, where writer and srlz.d are in separate groups) clears
     DVS_DATA. (This also clears DVS_OTHER, but that is not guaranteed to
     always be the case).

     Instruction group break (groups separated by stop, taken branch,
     interruption or rfi) clears DVS_IMPLIED and DVS_IMPLIEDF.
   */
  update_dependencies (idesc);

  /* Sometimes, knowing a register value allows us to avoid giving a false DV
     warning.  Keep track of as many as possible that are useful.  */
  note_register_values (idesc);

  /* We don't need or want this anymore.  */
  md.mem_offset.hint = 0;

  return 0;
}











#if 0




/* Translate one line of assembly.  Pseudo ops and labels do not show
   here.  */
void
md_assemble (str)
     char *str;
{
  char *saved_input_line_pointer, *mnemonic;
  const struct pseudo_opcode *pdesc;
  t_ia64_opcode *idesc;
  unsigned char qp_regno;
  unsigned int flags;
  int ch;

  saved_input_line_pointer = input_line_pointer;
  input_line_pointer = str;

  /* extract the opcode (mnemonic):  */

  mnemonic = input_line_pointer;
  ch = get_symbol_end ();
  pdesc = (struct pseudo_opcode *) hash_find (md.pseudo_hash, mnemonic);
  if (pdesc)
    {
      *input_line_pointer = ch;
      (*pdesc->handler) (pdesc->arg);
      goto done;
    }

  /* Find the instruction descriptor matching the arguments.  */

  idesc = ia64_find_opcode (mnemonic);
  *input_line_pointer = ch;
  if (!idesc)
    {
      as_bad ("Unknown opcode `%s'", mnemonic);
      goto done;
    }

  idesc = parse_operands (idesc);
  if (!idesc)
    goto done;

  /* Handle the dynamic ops we can handle now:  */
  if (idesc->type == IA64_TYPE_DYN)
    {
      if (strcmp (idesc->desc, "add") == 0)
	{
	  if (CURR_SLOT.opnd[2].X_op == O_register
	      && CURR_SLOT.opnd[2].X_add_number < 4)
	    mnemonic = "addl";
	  else
	    mnemonic = "adds";
	  ia64_free_opcode (idesc);
	  idesc = ia64_find_opcode (mnemonic);
#if 0
	  know (!idesc->next);
#endif
	}
      else if (strcmp (idesc->desc, "mov") == 0)
	{
	  enum ia64_opnd opnd1, opnd2;
	  int rop;

	  opnd1 = idesc->operands[0];
	  opnd2 = idesc->operands[1];
	  if (opnd1 == IA64_OPND_AR3)
	    rop = 0;
	  else if (opnd2 == IA64_OPND_AR3)
	    rop = 1;
	  else
	    abort ();
	  if (CURR_SLOT.opnd[rop].X_op == O_register
	      && ar_is_in_integer_unit (CURR_SLOT.opnd[rop].X_add_number))
	    mnemonic = "mov.i";
	  else
	    mnemonic = "mov.m";
	  ia64_free_opcode (idesc);
	  idesc = ia64_find_opcode (mnemonic);
	  while (idesc != NULL
		 && (idesc->operands[0] != opnd1
		     || idesc->operands[1] != opnd2))
	    idesc = get_next_opcode (idesc);
	}
    }

  qp_regno = 0;
  if (md.qp.X_op == O_register)
    {
      qp_regno = md.qp.X_add_number - REG_P;
      md.qp.X_op = O_absent;
    }

  flags = idesc->flags;



  //WATCH

  
  if ((flags & IA64_OPCODE_FIRST) != 0)
    insn_group_break (1, 0, 0);





  
  if ((flags & IA64_OPCODE_NO_PRED) != 0 && qp_regno != 0)
    {
      as_bad ("`%s' cannot be predicated", idesc->desc);
      goto done;
    }

  
  /* Build the instruction.  */
  CURR_SLOT.qp_regno = qp_regno;
  CURR_SLOT.idesc = idesc;
  as_where (&CURR_SLOT.src_file, &CURR_SLOT.src_line);
  dwarf2_where (&CURR_SLOT.debug_line);

  /* Add unwind entry, if there is one.  */
  if (unwind.current_entry)
    {
      CURR_SLOT.unwind_record = unwind.current_entry;
      unwind.current_entry = NULL;
    }

//WATCH

  
  /* Check for dependency violations.  */
  if (md.detect_dv)
    check_dv (idesc);

  md.curr_slot = (md.curr_slot + 1) % NUM_SLOTS;
  if (++md.num_slots_in_use >= NUM_SLOTS)
    emit_one_bundle ();

//WATCH
  
  if ((flags & IA64_OPCODE_LAST) != 0)
    insn_group_break (1, 0, 0);








  
  md.last_text_seg = now_seg;

 done:
  input_line_pointer = saved_input_line_pointer;
}

#endif

/* vim: set shiftwidth=2: */
