/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_utils.h"
#include "ia64_registers.h"
#include "ia64_instructions.h"
#include "ia64_opcodes.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_iterators.h"

#include "../../utils/util_error.h"


#include "../../simple_types.h"

#include "ia64_utils/ia64_utils_a.h"
#include "ia64_utils/ia64_utils_b.h"
#include "ia64_utils/ia64_utils_f.h"
#include "ia64_utils/ia64_utils_i.h"
#include "ia64_utils/ia64_utils_m.h"
#include "ia64_utils/ia64_utils_x.h"

#include "ia64_opcodes/ia64_opcodes_b.h"
#include "ia64_opcodes/ia64_opcodes_m.h"
#include "ia64_opcodes/ia64_opcodes_i.h"
#include "ia64_opcodes/ia64_opcodes_f.h"
#include "ia64_opcodes/ia64_opcodes_x.h"
#include "ia64_opcodes/ia64_opcodes_a.h"

//#define MIN(a, b) ((a) < (b) ? (a) : (b))
//#define MAX(a, b) ((a) < (b) ? (b) : (a))


#define IA64_REGSET_ADDREG(mask, reg) if((reg)!= IA64_NONE) RegsetAddReg(mask, reg)

#define IA64_REGSET_ADDREG_NAT(mask, reg) if(IA64_IS_GR(reg)) RegsetAddReg(mask, (reg)-IA64_GR+IA64_GR_NAT)

#define IA64ISROTATABLE_REG_GR(reg) ( (reg)>=IA64_R32 && (reg)<= IA64_MAX_GR)


t_uint32 Ia64CalcNinsWithOpcodeInCfg(t_cfg *cfg, t_ia64_instr_opcode opcode)
{
  t_bbl *bbl;

  t_uint32 cnt = 0;

  CFG_FOREACH_BBL(cfg, bbl)
  {
    t_ins *ins;

    BBL_FOREACH_INS(bbl, ins)
      if(IA64OPCODEINFO(ins).instr_opcode == opcode)
        cnt++;
  }

  return cnt;
}



t_bool Ia64DoesThisFunctionAskForTrouble(t_function *fun)
{
  t_bbl *i_bbl;
  t_bool hasReturn = FALSE;
  t_cfg_edge *i_edge;
  
  FUN_FOREACH_BBL(fun, i_bbl)
  {
    /*skip dangerous functions...*/
    if(i_bbl->ins_last && Ia64IsBranchOrCall(i_bbl->ins_last))
    {
      if(IA64_IS_BR(IA64INSIN1(i_bbl->ins_last)))
	return TRUE;
    }

    if(i_bbl->ins_last && Ia64IsReturn(i_bbl->ins_last))
      hasReturn = TRUE;

    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	return TRUE;
    }    
    BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
    {
      if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	return TRUE;
    }    
    

  }

  return !hasReturn;
}


t_ins *Ia64SearchForAlloc(t_function *fun)
{
  t_ins *alloc = NULL;
  t_bbl *i_bbl;

  FUN_FOREACH_BBL(fun, i_bbl)
  {
    BBL_FOREACH_INS(i_bbl, alloc)
      if(IA64ISALLOC(alloc))
	return alloc;
  }
  return NULL;
}



t_uint32 CountAllocsInFun(t_function *fun)
{
  t_uint32 cnt = 0;
  t_ins *alloc = NULL;
  t_bbl *i_bbl;

  FUN_FOREACH_BBL(fun, i_bbl)
  {
    BBL_FOREACH_INS(i_bbl, alloc)
      if(IA64ISALLOC(alloc))
	cnt++;
  }
  
  return cnt;
  
}




/*overloop alle functies, zoek de allocs en vul deze in*/
void Ia64CacheAllocsInFunctions(t_cfg *cfg)
{
  t_function *fun;
  /*berre*/
  //t_edge * edge;

  CFG_FOREACH_FUN(cfg, fun)
  {
    t_uint32 cnt = CountAllocsInFun(fun);

    if(cnt != 1)
    {
      //printf("Alloc count was %d for function %s.\n", cnt, fun->name);
      FUN_ISALLOCVALID(fun) = FALSE;
      FUN_SOL(fun) = IA64_INVALID_ALLOC;
      FUN_SOO(fun) = IA64_INVALID_ALLOC;
      FUN_SOF(fun) = IA64_INVALID_ALLOC;
      FUN_SOR(fun) = IA64_INVALID_ALLOC;

    }
    else
    {
      t_ins *alloc = Ia64SearchForAlloc(fun);

      if(alloc == NULL)
	FATAL(("This should not be happening !!!"));

      Ia64SyncAllocCacheInFunction(fun, alloc);

#if 0      
      printf("function %s has following alloc: sol = %d; soo = %d; sof = %d; sor = %d.\n", fun->name, 
	  FUN_SOL(fun), FUN_SOO(fun), FUN_SOF(fun), FUN_SOR(fun));
#endif      
    }
  }
#if 0
  /*begin berre*/
  CFG_FOREACH_EDGE(cfg, edge)
  {
    t_edge * i_edge;
    if(EDGE_CAT(edge)==ET_IPJUMP)
    {
	CfgEdgePrint(edge,0);
	BblPrint(CFG_EDGE_HEAD(edge),2);
      /*if(BBL_INS_LAST(CFG_EDGE_HEAD(edge))&&Ia64IsReturn(BBL_INS_LAST(CFG_EDGE_HEAD(edge))))
      {
	printf("--\n");
	CfgEdgePrint(edge,0);
	BblPrint(CFG_EDGE_HEAD(edge),2);
	printf("-\n");
	BblPrint(CFG_EDGE_TAIL(edge),2);
	BBL_FOREACH_SUCC_EDGE(CFG_EDGE_TAIL(edge),i_edge)
	{
	  CfgEdgePrint(i_edge,0);
	}
	
      }*/
    }
  }
  FATAL(("just checking"));
  /*einde berre*/
#endif
}



void Ia64SyncAllocCacheInFunction(t_function *fun, t_ins *alloc)
{
  if(IA64OPCODEINFO(alloc).instr_opcode != IA64_INSTR_ALLOC)
    FATAL(("instruction should be an alloc"));


  if(BBL_FUN(INS_BBL(alloc)) != fun)
    FATAL(("alloc is not of the given function"));


  FUN_ISALLOCVALID(fun) = TRUE;
  FUN_SOL(fun) = IA64INSIMM1(alloc);
  FUN_SOO(fun) = IA64INSIMM2(alloc) - IA64INSIMM1(alloc);
  FUN_SOF(fun) = IA64INSIMM2(alloc);
  FUN_SOR(fun) = IA64INSIMM3(alloc);

  FUN_ISALLOCCLEAN(fun) = FUN_BBL_FIRST(fun)->ins_first == alloc;
}




void Ia64ShrinkFrameInFunction(t_function *fun)
{
  FATAL(("implement"));
}

void Ia64ShrinkFrameInAllFunctions(t_cfg *cfg)
{
  t_function *fun;

  CFG_FOREACH_FUN(cfg, fun)
  {
    Ia64ShrinkFrameInFunction(fun);
  }
  
}


t_bool Ia64CanExpandFrameInFunction(t_function *fun, t_uint32 expansion)
{
  return !Ia64DoesThisFunctionAskForTrouble(fun)
    && FUN_ISALLOCVALID(fun) && FUN_SOF(fun) + expansion <= 96;
}

/*TODO functie die sol en soo dynamisch uitbreidt (maw ook alle instructies in de functie aanpassen)*/
/*false als hij niet kon (alloc <=96) of alloc niet gekend, of bad function (indirect calls/jumps e.d.)*/
t_bool Ia64ExpandLocalsInFunction(t_function *fun, t_uint32 expansion)
{
  t_ins *alloc;
  t_bbl *bbl;
  t_uint8 locals;
  
  if(!Ia64CanExpandFrameInFunction(fun, expansion))
    return FALSE;

  alloc = Ia64SearchForAlloc(fun);

  if(alloc == NULL)
    return FALSE;

  locals = IA64INSIMM1(alloc);


  /*eerst alle outputregisters van de functie manueel verschuiven*/
  
#define IA64_IS_OUT_REG(reg, shift) ((reg) >= IA64_R32 + shift && (reg) <= IA64_R127)
  
  FUN_FOREACH_BBL(fun, bbl)
  {
    t_ins *ins;

    BBL_FOREACH_INS(bbl, ins)
    {
      if(IA64_IS_OUT_REG(IA64INSIN1(ins), locals))
	IA64INSIN1(ins)+= expansion;

      if(IA64_IS_OUT_REG(IA64INSIN2(ins), locals))
	IA64INSIN2(ins)+= expansion;

      if(IA64_IS_OUT_REG(IA64INSIN3(ins), locals))
	IA64INSIN3(ins)+= expansion;

      if(IA64_IS_OUT_REG(IA64INSOUT1(ins), locals))
	IA64INSOUT1(ins)+= expansion;

      if(IA64_IS_OUT_REG(IA64INSOUT2(ins), locals))
	IA64INSOUT2(ins)+= expansion;

      INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
      INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);
      
      
    }
  }

  
  
  IA64INSIMM1(alloc) += expansion;
  IA64INSIMM2(alloc) += expansion;
  FUN_SOF(fun) += expansion;
  FUN_SOL(fun) += expansion;
  
  /* ook alle dummies van deze functie at callsite aanpassen !*/
  FUN_FOREACH_BBL(fun,bbl)
  {
    t_ins *ins;
    
    BBL_FOREACH_INS(bbl, ins)
    {
      if(IA64ISLIVEDUMMY(ins))
      {
	IA64INSIMM2(ins) = IA64INSIMM1(alloc);
	IA64INSIMM3(ins) = IA64INSIMM2(alloc);
      }
    }
  }
  
  
  return TRUE;
}

t_bool Ia64ExpandOutputsInFunction(t_function *fun, t_uint32 expansion)
{
  t_ins *alloc;
  t_bbl *bbl;
  
  if(!Ia64CanExpandFrameInFunction(fun, expansion))
    return FALSE;

  alloc = Ia64SearchForAlloc(fun);

  if(alloc == NULL)
    return FALSE;

  IA64INSIMM2(alloc) += expansion;
  FUN_SOF(fun) += expansion;
  FUN_SOO(fun) += expansion;

  /* ook alle dummies van deze functie at callsite aanpassen !*/
  FUN_FOREACH_BBL(fun,bbl)
  {
    t_ins *ins;
    
    BBL_FOREACH_INS(bbl, ins)
    {
      if(IA64ISLIVEDUMMY(ins))
      {
	IA64INSIMM2(ins) = IA64INSIMM1(alloc);
	IA64INSIMM3(ins) = IA64INSIMM2(alloc);
      }
    }
  }
  
  
  return TRUE;
}




void Ia64ExpandLocalsInAllFunctions(t_cfg *cfg, t_uint8 expansion)
{
  t_function *fun;

  CFG_FOREACH_FUN(cfg, fun)
  {
    Ia64ExpandLocalsInFunction(fun, expansion);
  }
}









/* The structure of the different UsedRegisters functions */

typedef void (*Ia64UsedRegistersFunction) (t_regset * mask, t_ins * ins);

/* 
 * bevat de lijst van alle used_registers functions, 1 voor elk instructie formaat
 * de indexering in deze tabel moet dan ook volgens t_ia64_instr_format gebeuren...
 */
extern const Ia64UsedRegistersFunction Ia64UsedRegistersFunction_table [];




/* The structure of the different DefinedRegisters functions */

typedef void (*Ia64DefinedRegistersFunction) (t_regset * mask, t_ins * ins);

/* 
 * bevat de lijst van alle defined_registers functions, 1 voor elk instructie formaat
 * de indexering in deze tabel moet dan ook volgens t_ia64_instr_format gebeuren...
 */
extern const Ia64DefinedRegistersFunction Ia64DefinedRegistersFunction_table [];



t_regset Ia64UsedRegisters(t_ins * ins)
{
  t_regset mask = NullRegs;

  IA64_REGSET_ADDREG(mask, IA64INSPRED(ins));

  IA64_REGSET_ADDREG(mask, IA64INSIN1(ins));
  
  if(IA64ISROTATABLE_REG_GR(IA64INSIN1(ins)))
    RegsetAddReg(mask, IA64_CFM);
  
  IA64_REGSET_ADDREG(mask, IA64INSIN2(ins));
  
  if(IA64ISROTATABLE_REG_GR(IA64INSIN2(ins)))
    RegsetAddReg(mask, IA64_CFM);
  
  IA64_REGSET_ADDREG(mask, IA64INSIN3(ins));
  
  if(IA64ISROTATABLE_REG_GR(IA64INSIN3(ins)))
    RegsetAddReg(mask, IA64_CFM);

  
  /* voor CFM ook kijken naar defined reg ! */
  
  if(IA64ISROTATABLE_REG_GR(IA64INSOUT1(ins)))
    RegsetAddReg(mask, IA64_CFM);  
  if(IA64ISROTATABLE_REG_GR(IA64INSOUT2(ins)))
    RegsetAddReg(mask, IA64_CFM); 
  
  /*CFM: moet niet meer voor NAT: zal al voor de gewone */
  IA64_REGSET_ADDREG_NAT(mask, IA64INSIN1(ins));   
  IA64_REGSET_ADDREG_NAT(mask, IA64INSIN2(ins)); 
  IA64_REGSET_ADDREG_NAT(mask, IA64INSIN3(ins)); 
  
  Ia64UsedRegistersFunction_table[IA64OPCODEINFO(ins).instr_format](&mask, ins);

  return mask;
}

t_regset Ia64DefinedRegisters(t_ins * ins)
{
  t_regset mask = NullRegs;

  IA64_REGSET_ADDREG(mask, IA64INSOUT1(ins));
  IA64_REGSET_ADDREG(mask, IA64INSOUT2(ins));  
    
  IA64_REGSET_ADDREG_NAT(mask, IA64INSOUT1(ins)); 
  IA64_REGSET_ADDREG_NAT(mask, IA64INSOUT2(ins)); 
  
  Ia64DefinedRegistersFunction_table[IA64OPCODEINFO(ins).instr_format](&mask, ins);

  return mask;
}



t_uint64 GetBits128(t_uint128 input, t_uint8 offset, t_uint8 length)
{
  t_uint64 ret = 0;  

  if(offset < 64)
    ret = (input.lower>>offset) & CREATE_MASK(MIN(length, 64 - offset));

  if(offset + length >= 64)
    ret += ((input.upper>>MAX(0, offset - 64)) & CREATE_MASK(length + MIN(offset - 64, 0)))<<MAX(0, 64 - offset);

  return ret;
}





t_bool Ia64InsHasSideEffect(t_ins *ins)
{
//  printf("WARNING !!! IMPLEMENT function Ia64InsHasSideEffect !!!\n");


  /*ev nog verfijnen: voorlopig worst case scenario...(+nog checken of dat wel klopt)*/
  switch(IA64OPCODEINFO(ins).instr_opcode)
  {
    case IA64_INSTR_ADD:
    case IA64_INSTR_ADDP4:
//    case IA64_INSTR_ALLOC:
    case IA64_INSTR_AND:
    case IA64_INSTR_ANDCM:
//    case IA64_INSTR_BR:
//    case IA64_INSTR_BREAK:
//    case IA64_INSTR_BRL:
//    case IA64_INSTR_BRP:
//    case IA64_INSTR_BSW:
//    case IA64_INSTR_CHK:
//    case IA64_INSTR_CLRRRB:
    case IA64_INSTR_CMP:
    case IA64_INSTR_CMP4:
//    case IA64_INSTR_CMPXCHG:
//    case IA64_INSTR_COVER:
    case IA64_INSTR_CZX:
    case IA64_INSTR_DEP:
//    case IA64_INSTR_EPC:
    case IA64_INSTR_EXTR:
    case IA64_INSTR_FAMAX:
    case IA64_INSTR_FAMIN:
    case IA64_INSTR_FAND:
    case IA64_INSTR_FANDCM:
//    case IA64_INSTR_FC:
//    case IA64_INSTR_FCHKF:
//    case IA64_INSTR_FCLASS:
//    case IA64_INSTR_FCLRF:
    case IA64_INSTR_FCMP:
    case IA64_INSTR_FCVT_FX:
    case IA64_INSTR_FCVT_XF:
    case IA64_INSTR_FETCHADD:
//    case IA64_INSTR_FLUSHRS:
    case IA64_INSTR_FMA:
    case IA64_INSTR_FMAX:
    case IA64_INSTR_FMERGE:
    case IA64_INSTR_FMIN:
    case IA64_INSTR_FMIX:
    case IA64_INSTR_FMS:
    case IA64_INSTR_FNMA:
    case IA64_INSTR_FOR:
    case IA64_INSTR_FPACK:
    case IA64_INSTR_FPAMAX:
    case IA64_INSTR_FPAMIN:
    case IA64_INSTR_FPCMP:
    case IA64_INSTR_FPCVT_FX:
    case IA64_INSTR_FPMA:
    case IA64_INSTR_FPMAX:
    case IA64_INSTR_FPMERGE:
    case IA64_INSTR_FPMIN:
    case IA64_INSTR_FPMS:
    case IA64_INSTR_FPNMA:
    case IA64_INSTR_FPRCPA:
    case IA64_INSTR_FPRSQRTA:
    case IA64_INSTR_FRCPA:
    case IA64_INSTR_FRSQRTA:
//    case IA64_INSTR_FSELECT:
//    case IA64_INSTR_FSETC:
    case IA64_INSTR_FSWAP:
    case IA64_INSTR_FSXT:
//    case IA64_INSTR_FWB:
    case IA64_INSTR_FXOR:
//    case IA64_INSTR_GETF:
//    case IA64_INSTR_INVALA:
//    case IA64_INSTR_ITC:
//    case IA64_INSTR_ITR:
    case IA64_INSTR_LD:
    case IA64_INSTR_LDF:
    case IA64_INSTR_LDFP:
//    case IA64_INSTR_LFETCH:
//    case IA64_INSTR_LOADRS:
//    case IA64_INSTR_MF:
    case IA64_INSTR_MIX:
    case IA64_INSTR_MOV_AR:
    case IA64_INSTR_MOV_BR:
//    case IA64_INSTR_MOV_CR:
//    case IA64_INSTR_MOV_IREG:
//    case IA64_INSTR_MOV_IP:
    case IA64_INSTR_MOV_PR:
    case IA64_INSTR_MOV_PSR:
//    case IA64_INSTR_MOV_UM:
    case IA64_INSTR_MOVL:
    case IA64_INSTR_MUX:
//    case IA64_INSTR_NOP:
    case IA64_INSTR_OR:
    case IA64_INSTR_PACK:
    case IA64_INSTR_PADD:
    case IA64_INSTR_PAVG:
    case IA64_INSTR_PAVGSUB:
    case IA64_INSTR_PCMP:
    case IA64_INSTR_PMAX:
    case IA64_INSTR_PMIN:
    case IA64_INSTR_PMPY2:
    case IA64_INSTR_PMPYSHR2:
    case IA64_INSTR_POPCNT:
//    case IA64_INSTR_PROBE:
    case IA64_INSTR_PSAD1:
    case IA64_INSTR_PSHL:
    case IA64_INSTR_PSHLADD2:
    case IA64_INSTR_PSHR:
    case IA64_INSTR_PSHRADD2:
    case IA64_INSTR_PSUB:
//    case IA64_INSTR_PTC_E:
//    case IA64_INSTR_PTC_G:
//    case IA64_INSTR_PTC_L:
//    case IA64_INSTR_PTR:
//    case IA64_INSTR_RFI:
//    case IA64_INSTR_RSM:
//    case IA64_INSTR_RUM:
//    case IA64_INSTR_SETF:
    case IA64_INSTR_SHL:
    case IA64_INSTR_SHLADD:
    case IA64_INSTR_SHLADDP4:
    case IA64_INSTR_SHR:
    case IA64_INSTR_SHRP:
    case IA64_INSTR_SRLZ:
//    case IA64_INSTR_SSM:
//    case IA64_INSTR_ST:
//    case IA64_INSTR_STF:
    case IA64_INSTR_SUB:
//    case IA64_INSTR_SUM:
    case IA64_INSTR_SXT:
//    case IA64_INSTR_SYNC:
//    case IA64_INSTR_TAK:
    case IA64_INSTR_TBIT:
//    case IA64_INSTR_THASH:
    case IA64_INSTR_TNAT:
//    case IA64_INSTR_TPA:
//    case IA64_INSTR_TTAG:
    case IA64_INSTR_UNPACK:
    case IA64_INSTR_XCHG:
    case IA64_INSTR_XMA:
    case IA64_INSTR_XOR:
    case IA64_INSTR_ZXT:
//    case IA64_INSTR_DUMMY:
    case IA64_INSTR_UNSUPPORTED:  
      return FALSE;
    default:
      return TRUE;
  }
  
  return TRUE;
}














void Ia64UsedRegistersUnsupported(t_regset * mask, t_ins * ins)
{
  FATAL(("Unsupported ia64 instruction; cannot calculate used registers."));
}



void Ia64UsedRegistersDummy(t_regset * mask, t_ins * ins)
{
}
void Ia64UsedRegistersData(t_regset * mask, t_ins * ins)
{
}




/* 
 * bevat de lijst van alle used_registers functions, 1 voor elk instructie formaat
 * de indexering in deze tabel moet dan ook volgens t_ia64_instr_format gebeuren...
 */
const Ia64UsedRegistersFunction Ia64UsedRegistersFunction_table [] =
{
	Ia64UsedRegistersA1, Ia64UsedRegistersA2, Ia64UsedRegistersA3, Ia64UsedRegistersA4, Ia64UsedRegistersA5,
	Ia64UsedRegistersA6, Ia64UsedRegistersA7, Ia64UsedRegistersA8, Ia64UsedRegistersA9, Ia64UsedRegistersA10,
	
	Ia64UsedRegistersI1, Ia64UsedRegistersI2, Ia64UsedRegistersI3, Ia64UsedRegistersI4, Ia64UsedRegistersI5,
	Ia64UsedRegistersI6, Ia64UsedRegistersI7, Ia64UsedRegistersI8, Ia64UsedRegistersI9, Ia64UsedRegistersI10,
	Ia64UsedRegistersI11, Ia64UsedRegistersI12, Ia64UsedRegistersI13, Ia64UsedRegistersI14, Ia64UsedRegistersI15,
	Ia64UsedRegistersI16, Ia64UsedRegistersI17, Ia64UsedRegistersUnsupported, Ia64UsedRegistersI19, Ia64UsedRegistersI20,
	Ia64UsedRegistersI21, Ia64UsedRegistersI22, Ia64UsedRegistersI23, Ia64UsedRegistersI24, Ia64UsedRegistersI25,
	Ia64UsedRegistersI26, Ia64UsedRegistersI27, Ia64UsedRegistersI28, Ia64UsedRegistersI29, 

	Ia64UsedRegistersM1, Ia64UsedRegistersM2, Ia64UsedRegistersM3, Ia64UsedRegistersM4, Ia64UsedRegistersM5,
	Ia64UsedRegistersM6, Ia64UsedRegistersM7, Ia64UsedRegistersM8, Ia64UsedRegistersM9, Ia64UsedRegistersM10,
	Ia64UsedRegistersM11, Ia64UsedRegistersM12, Ia64UsedRegistersM13, Ia64UsedRegistersM14, Ia64UsedRegistersM15,
	Ia64UsedRegistersM16, Ia64UsedRegistersM17, Ia64UsedRegistersM18, Ia64UsedRegistersM19, Ia64UsedRegistersM20,
	Ia64UsedRegistersM21, Ia64UsedRegistersM22, Ia64UsedRegistersM23, Ia64UsedRegistersM24, Ia64UsedRegistersM25,
	Ia64UsedRegistersM26, Ia64UsedRegistersM27, Ia64UsedRegistersM28, Ia64UsedRegistersM29, Ia64UsedRegistersM30,
	Ia64UsedRegistersM31, Ia64UsedRegistersM32, Ia64UsedRegistersM33, Ia64UsedRegistersM34, Ia64UsedRegistersM35,
	Ia64UsedRegistersM36, Ia64UsedRegistersM37, Ia64UsedRegistersM38, Ia64UsedRegistersM39, Ia64UsedRegistersM40,
	Ia64UsedRegistersM41, Ia64UsedRegistersM42, Ia64UsedRegistersM43, Ia64UsedRegistersM44, Ia64UsedRegistersM45,
	Ia64UsedRegistersM46, 
	
	Ia64UsedRegistersB1, Ia64UsedRegistersB2, Ia64UsedRegistersB3, Ia64UsedRegistersB4, Ia64UsedRegistersB5,
	Ia64UsedRegistersB6, Ia64UsedRegistersB7, Ia64UsedRegistersB8, Ia64UsedRegistersB9,	

	Ia64UsedRegistersF1, Ia64UsedRegistersF2, Ia64UsedRegistersF3, Ia64UsedRegistersF4, Ia64UsedRegistersF5,
	Ia64UsedRegistersF6, Ia64UsedRegistersF7, Ia64UsedRegistersF8, Ia64UsedRegistersF9, Ia64UsedRegistersF10,
	Ia64UsedRegistersF11, Ia64UsedRegistersF12, Ia64UsedRegistersF13, Ia64UsedRegistersF14, Ia64UsedRegistersF15,

	Ia64UsedRegistersX1, Ia64UsedRegistersX2, Ia64UsedRegistersX3, Ia64UsedRegistersX4, 

	Ia64UsedRegistersUnsupported,
	Ia64UsedRegistersDummy,
	Ia64UsedRegistersData
	
};






void Ia64DefinedRegistersUnsupported(t_regset * mask, t_ins * ins)
{
  FATAL(("Unsupported ia64 instruction; cannot calculate defined registers."));
}


void Ia64DefinedRegistersDummy(t_regset * mask, t_ins * ins)
{
}
void Ia64DefinedRegistersData(t_regset * mask, t_ins * ins)
{
}





/* 
 * bevat de lijst van alle defined_registers functions, 1 voor elk instructie formaat
 * de indexering in deze tabel moet dan ook volgens t_ia64_instr_format gebeuren...
 */
const Ia64DefinedRegistersFunction Ia64DefinedRegistersFunction_table [] =
{
	Ia64DefinedRegistersA1, Ia64DefinedRegistersA2, Ia64DefinedRegistersA3, Ia64DefinedRegistersA4, Ia64DefinedRegistersA5,
	Ia64DefinedRegistersA6, Ia64DefinedRegistersA7, Ia64DefinedRegistersA8, Ia64DefinedRegistersA9, Ia64DefinedRegistersA10,
	
	Ia64DefinedRegistersI1, Ia64DefinedRegistersI2, Ia64DefinedRegistersI3, Ia64DefinedRegistersI4, Ia64DefinedRegistersI5,
	Ia64DefinedRegistersI6, Ia64DefinedRegistersI7, Ia64DefinedRegistersI8, Ia64DefinedRegistersI9, Ia64DefinedRegistersI10,
	Ia64DefinedRegistersI11, Ia64DefinedRegistersI12, Ia64DefinedRegistersI13, Ia64DefinedRegistersI14, Ia64DefinedRegistersI15,
	Ia64DefinedRegistersI16, Ia64DefinedRegistersI17, Ia64DefinedRegistersUnsupported, Ia64DefinedRegistersI19, Ia64DefinedRegistersI20,
	Ia64DefinedRegistersI21, Ia64DefinedRegistersI22, Ia64DefinedRegistersI23, Ia64DefinedRegistersI24, Ia64DefinedRegistersI25,
	Ia64DefinedRegistersI26, Ia64DefinedRegistersI27, Ia64DefinedRegistersI28, Ia64DefinedRegistersI29, 

	Ia64DefinedRegistersM1, Ia64DefinedRegistersM2, Ia64DefinedRegistersM3, Ia64DefinedRegistersM4, Ia64DefinedRegistersM5,
	Ia64DefinedRegistersM6, Ia64DefinedRegistersM7, Ia64DefinedRegistersM8, Ia64DefinedRegistersM9, Ia64DefinedRegistersM10,
	Ia64DefinedRegistersM11, Ia64DefinedRegistersM12, Ia64DefinedRegistersM13, Ia64DefinedRegistersM14, Ia64DefinedRegistersM15,
	Ia64DefinedRegistersM16, Ia64DefinedRegistersM17, Ia64DefinedRegistersM18, Ia64DefinedRegistersM19, Ia64DefinedRegistersM20,
	Ia64DefinedRegistersM21, Ia64DefinedRegistersM22, Ia64DefinedRegistersM23, Ia64DefinedRegistersM24, Ia64DefinedRegistersM25,
	Ia64DefinedRegistersM26, Ia64DefinedRegistersM27, Ia64DefinedRegistersM28, Ia64DefinedRegistersM29, Ia64DefinedRegistersM30,
	Ia64DefinedRegistersM31, Ia64DefinedRegistersM32, Ia64DefinedRegistersM33, Ia64DefinedRegistersM34, Ia64DefinedRegistersM35,
	Ia64DefinedRegistersM36, Ia64DefinedRegistersM37, Ia64DefinedRegistersM38, Ia64DefinedRegistersM39, Ia64DefinedRegistersM40,
	Ia64DefinedRegistersM41, Ia64DefinedRegistersM42, Ia64DefinedRegistersM43, Ia64DefinedRegistersM44, Ia64DefinedRegistersM45,
	Ia64DefinedRegistersM46, 
	
	Ia64DefinedRegistersB1, Ia64DefinedRegistersB2, Ia64DefinedRegistersB3, Ia64DefinedRegistersB4, Ia64DefinedRegistersB5,
	Ia64DefinedRegistersB6, Ia64DefinedRegistersB7, Ia64DefinedRegistersB8, Ia64DefinedRegistersB9,	

	Ia64DefinedRegistersF1, Ia64DefinedRegistersF2, Ia64DefinedRegistersF3, Ia64DefinedRegistersF4, Ia64DefinedRegistersF5,
	Ia64DefinedRegistersF6, Ia64DefinedRegistersF7, Ia64DefinedRegistersF8, Ia64DefinedRegistersF9, Ia64DefinedRegistersF10,
	Ia64DefinedRegistersF11, Ia64DefinedRegistersF12, Ia64DefinedRegistersF13, Ia64DefinedRegistersF14, Ia64DefinedRegistersF15,

	Ia64DefinedRegistersX1, Ia64DefinedRegistersX2, Ia64DefinedRegistersX3, Ia64DefinedRegistersX4, 

	Ia64DefinedRegistersUnsupported,
	Ia64DefinedRegistersDummy,
	Ia64DefinedRegistersData
	
};

void Ia64GenerateOpcodeLabels()
{
  int teller;
  FILE * bestand = fopen("ia64_opcode_label.h","w+");
  
  teller=0;
  while(strcmp(ia64_opcode_table_a[teller].desc,"Unsupported")!=0)
  {
    fprintf(bestand, "#define %s_A_%llx %d\n",ia64_opcode_table_a[teller].desc,ia64_opcode_table_a[teller].opcode,teller++);
  }
  
  teller=0;
  while(strcmp(ia64_opcode_table_b[teller].desc,"Unsupported")!=0)
  {
    fprintf(bestand, "#define %s_B_%llx %d\n",ia64_opcode_table_b[teller].desc,ia64_opcode_table_b[teller].opcode,teller++);
  }
  
  teller=0;
  while(strcmp(ia64_opcode_table_x[teller].desc,"Unsupported")!=0)
  {
    fprintf(bestand, "#define %s_X_%llx %d\n",ia64_opcode_table_x[teller].desc,ia64_opcode_table_x[teller].opcode,teller++);
  }
  
  teller=0;
  while(strcmp(ia64_opcode_table_m[teller].desc,"Unsupported")!=0)
  {
    fprintf(bestand, "#define %s_M_%llx %d\n",ia64_opcode_table_m[teller].desc,ia64_opcode_table_m[teller].opcode,teller++);
  }


  teller=0;
  while(strcmp(ia64_opcode_table_f[teller].desc,"Unsupported")!=0)
  {
    fprintf(bestand, "#define %s_F_%llx %d\n",ia64_opcode_table_f[teller].desc,ia64_opcode_table_f[teller].opcode,teller++);
  }
  

  teller=0;
  while(strcmp(ia64_opcode_table_i[teller].desc,"Unsupported")!=0)
  {
    fprintf(bestand, "#define %s_I_%llx %d\n",ia64_opcode_table_i[teller].desc,ia64_opcode_table_i[teller].opcode,teller++);
  }

  fclose(bestand);
  FATAL(("hey"));
}



/* vim: set shiftwidth=2: */
