/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"
#ifndef IA64_UTILS_H
#define IA64_UTILS_H

#include "ia64_types.h"
#include "ia64_instructions.h"

#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_ins.h"

#include "ia64_opcodes.h"

t_uint32 Ia64CalcNinsWithOpcodeInCfg(t_cfg *cfg, t_ia64_instr_opcode opcode);

t_bool Ia64DoesThisFunctionAskForTrouble(t_function *fun);

t_ins *Ia64SearchForAlloc(t_function *fun);
t_uint32 CountAllocsInFun(t_function *fun);


/*evt ook nog check schrijven die voor disassemberen nagaat of de waarden in t_fun en de allocinstructie wel nog compatibel zijn*/
void Ia64CacheAllocsInFunctions(t_cfg *cfg);
  
void Ia64SyncAllocCacheInFunction(t_function *fun, t_ins *alloc);


/*TODO shrink operatie die zowel evt gaten opvult als alloc zo compact mogelijk maakt*/
void Ia64ShrinkFrameInFunction(t_function *fun);

void Ia64ShrinkFrameInAllFunctions(t_cfg *cfg);


t_bool Ia64CanExpandFrameInFunction(t_function *fun, t_uint32 expansion);

/*TODO functie die sol en soo dynamisch uitbreidt (maw ook alle instructies in de functie aanpassen)*/
/*false als hij niet kon (alloc <=96) of alloc niet gekend*/

t_bool Ia64ExpandLocalsInFunction(t_function *fun, t_uint32 expansion);

t_bool Ia64ExpandOutputsInFunction(t_function *fun, t_uint32 expansion);

/*debugging-function for Ia64ExpandLocalsInFunction and Ia64ShrinkFrameInAllFunctions*/
void Ia64ExpandLocalsInAllFunctions(t_cfg *cfg, t_uint8 expansion);


t_regset Ia64UsedRegisters(t_ins * ins);
t_regset Ia64DefinedRegisters(t_ins * ins);


t_bool Ia64InsHasSideEffect(t_ins *ins);


/*#define _DEBUG_DISASSEMBLE_*/
#define _DEBUG_ASSEMBLE_
/*#define _DEBUG_FLOWGRAPH_ */


#define SIGN_EXT(x, p) ((((x)&(CREATE_MASK(p))) ^ ((t_uint64)1 << ((p) - 1))) - ((t_uint64)1 << ((p) - 1)))

#define ZERO_EXT(x, p) (((t_uint64)x)&CREATE_MASK(p))

#define CREATE_MASK(aantal) (((t_uint64)1 << (aantal)) - 1)

#define GETBITS64(input, offset, length) (((input) >> (offset)) & CREATE_MASK(length))

t_uint64 GetBits128(t_uint128 input, t_uint8 offset, t_uint8 length);



#endif
/* vim: set shiftwidth=2: */
