/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_utils_b.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"




/* TODO: nog alle verborgen gebruikte en gewijzigde registers toevoegen !!!*/



/* B1: IP-Relative Branch */
void Ia64UsedRegistersB1(t_regset * mask, t_ins * ins)
{
  
  /* blijkbaar leest hij ook nog van PSR, enkele vlagjes*/
  RegsetAddReg(*mask, IA64_PSR);
  RegsetAddReg(*mask, IA64_IP);
   
  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 6, 3))
  {
    case 0:/* br.cond */
      break;

    case 2:/* br.wexit */
    case 3:/* br.wtop */
      RegsetAddReg(*mask, IA64_AR_EC);
      /* wijzigt de RRB's in CFM */
      RegsetAddReg(*mask, IA64_CFM);
      break;
    default:
      FATAL(("This shouldn't happen: invalid B1-type instruction"));
  }
  
}


/* B2: IP-Relative Counted Branch */
void Ia64UsedRegistersB2(t_regset * mask, t_ins * ins)
{

  /* blijkbaar leest hij ook nog van PSR, enkele vlagjes*/
  RegsetAddReg(*mask, IA64_PSR);

  RegsetAddReg(*mask, IA64_IP);
  
  RegsetAddReg(*mask, IA64_AR_LC);	
  
  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 6, 3))
  {
    case 5:/* br.cloop */
      break;

    case 6:/* br.cexit */
    case 7:/* br.ctop */
      RegsetAddReg(*mask, IA64_AR_EC);
      /* wijzigt de RRB's in CFM */
      RegsetAddReg(*mask, IA64_CFM);
      break;
    default:
      FATAL(("This shouldn't happen: invalid B2-type instruction"));
  }


}


/* B3: IP-Relative Call */
void Ia64UsedRegistersB3(t_regset * mask, t_ins * ins)
{

  /* blijkbaar leest hij ook nog van PSR, enkele vlagjes*/
  RegsetAddReg(*mask, IA64_PSR);

  /* leest den IP om het te bewaren in het opgegeven branch register*/
  RegsetAddReg(*mask, IA64_IP);

  /* leest CFM om het te bewaren in PFS*/
  RegsetAddReg(*mask, IA64_CFM);
  /* leest EC om het te bewaren in PFS*/
  RegsetAddReg(*mask, IA64_AR_EC);
  /* leest PSR.cpl om het te bewaren in PFS*/
  RegsetAddReg(*mask, IA64_PSR);

  
}


/* B4: Indirect Branch */
void Ia64UsedRegistersB4(t_regset * mask, t_ins * ins)
{
  

  /* blijkbaar leest hij ook nog van PSR, enkele vlagjes*/
  RegsetAddReg(*mask, IA64_PSR);
  
  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 6, 3))
  {
    case 0:/* br.cond */
      break;
    case 1:/* br.ia */
      FATAL(("This shouldn't happen: br.ia NOT supported !!!"));
      break;
    case 4:/* br.ret */
      RegsetAddReg(*mask, IA64_AR_PFS);
      RegsetAddReg(*mask, IA64_CFM);
      
      /* zie rse_penable_current_frame_load */  
      RegsetAddReg(*mask, IA64_AR_BSP);
      
      break;
    default:
      FATAL(("This shouldn't happen: invalid B4-type instruction"));
  }


  

}


/* B5: Indirect Call */
void Ia64UsedRegistersB5(t_regset * mask, t_ins * ins)
{
  

  /* blijkbaar leest hij ook nog van PSR, enkele vlagjes*/
  RegsetAddReg(*mask, IA64_PSR);

  /* leest den IP om het te bewaren in het opgegeven branch register*/
  RegsetAddReg(*mask, IA64_IP);

  /* leest CFM om het te bewaren in PFS*/
  RegsetAddReg(*mask, IA64_CFM);
  /* leest EC om het te bewaren in PFS*/
  RegsetAddReg(*mask, IA64_AR_EC);
  /* leest PSR.cpl om het te bewaren in PFS*/
  RegsetAddReg(*mask, IA64_PSR);

  

}


/* B6: IP-Relative Predict */
void Ia64UsedRegistersB6(t_regset * mask, t_ins * ins)
{

  RegsetAddReg(*mask, IA64_IP);
  
}

/* B7: Indirect Predict */
void Ia64UsedRegistersB7(t_regset * mask, t_ins * ins)
{


}


/* B8: Miscellaneous (M-Unit) */
void Ia64UsedRegistersB8(t_regset * mask, t_ins * ins)
{
  
  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 6))
  {
    case 0x02: /* cover */
      RegsetAddReg(*mask, IA64_CFM);
      RegsetAddReg(*mask, IA64_PSR);
      break;
    case 0x04: /* clrrrb */
      break;
    case 0x05: /* clrrrb.pr */
      break;
    case 0x08: /* rfi */
      RegsetAddReg(*mask, IA64_PSR);
      RegsetAddReg(*mask, IA64_CR_IPSR);
      RegsetAddReg(*mask, IA64_CR_IIP);
      RegsetAddReg(*mask, IA64_AR_CSD);
      RegsetAddReg(*mask, IA64_CFM);
      RegsetAddReg(*mask, IA64_CR_IFS);
      /* zie rse_enable_current_frame_load */  
      RegsetAddReg(*mask, IA64_AR_BSP);      
      break;
    case 0x0C: /* bsw.0 */
      RegsetAddReg(*mask, IA64_PSR);
      break;
    case 0x0D: /* bsw.1 */
      RegsetAddReg(*mask, IA64_PSR);
      break;
    case 0x10: /* epc */
      RegsetAddReg(*mask, IA64_AR_PFS);
      RegsetAddReg(*mask, IA64_PSR);
      break;
    default:
      FATAL(("This shouldn't happen: invalid B8-type instruction"));
  }
  
}


/* B9: Break/Nop (B-Unit) */
void Ia64UsedRegistersB9(t_regset * mask, t_ins * ins)
{

}






/*----------------------------------------------------------------------------*/





/* B1: IP-Relative Branch */
void Ia64DefinedRegistersB1(t_regset * mask, t_ins * ins)
{

  /* overschijft misschien den IP */
  RegsetAddReg(*mask, IA64_IP);

  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 6, 3))
  {
    case 0:/* br.cond */
      break;

    case 2:/* br.wexit */
    case 3:/* br.wtop */
      RegsetAddReg(*mask, IA64_AR_EC);
      RegsetAddReg(*mask, IA64_P63);
      /* wijzigt de RRB's in CFM */
      RegsetAddReg(*mask, IA64_CFM);
      break;
    default:
      FATAL(("This shouldn't happen: invalid B1-type instruction"));
  }


}


/* B2: IP-Relative Counted Branch */
void Ia64DefinedRegistersB2(t_regset * mask, t_ins * ins)
{


  /* overschijft misschien den IP */
  RegsetAddReg(*mask, IA64_IP);

  RegsetAddReg(*mask, IA64_AR_LC);	

  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 6, 3))
  {
    case 5:/* br.cloop */
      break;

    case 6:/* br.cexit */
    case 7:/* br.ctop */
      RegsetAddReg(*mask, IA64_AR_EC);
      RegsetAddReg(*mask, IA64_P63);
      /* wijzigt de RRB's in CFM */
      RegsetAddReg(*mask, IA64_CFM);
      break;
    default:
      FATAL(("This shouldn't happen: invalid B2-type instruction"));
  }


}


/* B3: IP-Relative Call */
void Ia64DefinedRegistersB3(t_regset * mask, t_ins * ins)
{
  
  /* overschijft misschien den IP */
  RegsetAddReg(*mask, IA64_IP);

  RegsetAddReg(*mask, IA64_AR_PFS);
  RegsetAddReg(*mask, IA64_CFM);

  /* zie rse_preserve_frame */  
  RegsetAddReg(*mask, IA64_AR_BSP);

}


/* B4: Indirect Branch */
void Ia64DefinedRegistersB4(t_regset * mask, t_ins * ins)
{

  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 6, 3))
  {
    case 0:/* br.cond */
      break;
    case 1:/* br.ia */
      FATAL(("This shouldn't happen: br.ia NOT supported !!!"));
      break;
    case 4:/* br.ret */
      RegsetAddReg(*mask, IA64_CFM);
      RegsetAddReg(*mask, IA64_AR_EC);
      RegsetAddReg(*mask, IA64_PSR);
      
      /* zie rse_restore_frame */  
      RegsetAddReg(*mask, IA64_AR_BSP);
      
      break;
    default:
      FATAL(("This shouldn't happen: invalid B4-type instruction"));
  }

  
}


/* B5: Indirect Call */
void Ia64DefinedRegistersB5(t_regset * mask, t_ins * ins)
{
 
  /* overschijft misschien den IP */
  RegsetAddReg(*mask, IA64_IP);

  RegsetAddReg(*mask, IA64_AR_PFS);
  RegsetAddReg(*mask, IA64_CFM);
  
  /* zie rse_preserve_frame */  
  RegsetAddReg(*mask, IA64_AR_BSP);

}


/* B6: IP-Relative Predict */
void Ia64DefinedRegistersB6(t_regset * mask, t_ins * ins)
{

}

/* B7: Indirect Predict */
void Ia64DefinedRegistersB7(t_regset * mask, t_ins * ins)
{

}


/* B8: Miscellaneous (M-Unit) */
void Ia64DefinedRegistersB8(t_regset * mask, t_ins * ins)
{
  
  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 6))
  {
    case 0x02: /* cover */
      RegsetAddReg(*mask, IA64_CR_IFS);
      RegsetAddReg(*mask, IA64_CFM);

      /* zie rse_preserve_frame */  
      RegsetAddReg(*mask, IA64_AR_BSP);
      
      break;
    case 0x04: /* clrrrb */
      RegsetAddReg(*mask, IA64_CFM);
      break;
    case 0x05: /* clrrrb.pr */
      RegsetAddReg(*mask, IA64_CFM);
      break;
    case 0x08: /* rfi */
      RegsetAddReg(*mask, IA64_PSR);
      RegsetAddReg(*mask, IA64_IP);
      RegsetAddReg(*mask, IA64_CFM);
      /* zie rse_restore_frame */  
      RegsetAddReg(*mask, IA64_AR_BSP);
      break;
    case 0x0C: /* bsw.0 */
      RegsetAddReg(*mask, IA64_PSR);
      break;
    case 0x0D: /* bsw.1 */
      RegsetAddReg(*mask, IA64_PSR);
      break;
    case 0x10: /* epc */
      RegsetAddReg(*mask, IA64_PSR);
      break;
    default:
      FATAL(("This shouldn't happen: invalid B8-type instruction"));
  }


}


/* B9: Break/Nop (B-Unit) */
void Ia64DefinedRegistersB9(t_regset * mask, t_ins * ins)
{
 
}




/* vim: set shiftwidth=2: */
