/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_utils_m.h"
#include "../../../utils/util_error.h"

#include "../ia64_utils.h"
#include "../ia64_instructions.h"
#include "../ia64_registers.h"

#include "../ia64_opcodes.h"



/* TODO: nog alle verborgen gebruikte en gewijzigde registers toevoegen !!!*/



/* M1: Integer Load */
void Ia64UsedRegistersM1(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);

  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 30, 6) == 0x1B)
    RegsetAddReg(*mask, IA64_AR_UNAT);
    
  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}


/* M2: Integer Load - Increment by Register */
void Ia64UsedRegistersM2(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);

  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 30, 6) == 0x1B)
    RegsetAddReg(*mask, IA64_AR_UNAT);
    
  RegsetAddReg(*mask, IA64_RESOURCE_MEM);


}

/* M3: Integer Load - Increment by immediate */
void Ia64UsedRegistersM3(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);

  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 30, 6) == 0x1B)
    RegsetAddReg(*mask, IA64_AR_UNAT);
    
  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
  

}



/* M4: Integer Store */
void Ia64UsedRegistersM4(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

}


/* M5: Integer Store - Increment by immediate */
void Ia64UsedRegistersM5(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  
  

}

/* M6: Floating-point Load */
void Ia64UsedRegistersM6(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);    

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}


/* M7: Floating-point Load - Increment by Register */
void Ia64UsedRegistersM7(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);    

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
}

/* M8: Floating-point Load - Increment by immediate */
void Ia64UsedRegistersM8(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);    
  
  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}



/* M9: Floating-point Store */
void Ia64UsedRegistersM9(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);

}


/* M10: Floating-point Store - Increment by immediate */
void Ia64UsedRegistersM10(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

}




/* M11: Floating-point Load Pair */
void Ia64UsedRegistersM11(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
}


/* M12: Floating-point Load Pair - Increment by Immediate*/
void Ia64UsedRegistersM12(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
}



/* M13: Line Prefetch */
void Ia64UsedRegistersM13(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}

/* M14: Line Prefetch - Increment by Register */
void Ia64UsedRegistersM14(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}


/* M15: Line Prefetch - Increment by Immediate */
void Ia64UsedRegistersM15(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}


/* M16: Exchange/Compare and Exchange */
void Ia64UsedRegistersM16(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 30, 6) < 0x08)
  RegsetAddReg(*mask, IA64_AR_CCV);

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
}


/* M17: Fetch and Add - Immediate */
void Ia64UsedRegistersM17(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
}

/* M18: Set FR */
void Ia64UsedRegistersM18(t_regset * mask, t_ins * ins)
{
  

}


/* M19: Get FR */
void Ia64UsedRegistersM19(t_regset * mask, t_ins * ins)
{
  

}

/* M20: Integer Speculation Check (M-Unit) */
void Ia64UsedRegistersM20(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);
  RegsetAddReg(*mask, IA64_PSR);    

}

/* M21: Floating-point Speculation Check */
void Ia64UsedRegistersM21(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);
  RegsetAddReg(*mask, IA64_PSR);      

}


/* M22: Integer Advanced Load Check */
void Ia64UsedRegistersM22(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);
  RegsetAddReg(*mask, IA64_PSR);      

}


/* M23: Floating-point Advanced Load Check */
void Ia64UsedRegistersM23(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);
  RegsetAddReg(*mask, IA64_PSR);      

}

/* M24: Sync/Fence/Serialize/ALAT Control */
void Ia64UsedRegistersM24(t_regset * mask, t_ins * ins)
{
  

}


/* M25: RSE Control */
void Ia64UsedRegistersM25(t_regset * mask, t_ins * ins)
{
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 4) == 0xC)
  {
    RegsetAddReg(*mask, IA64_AR_BSPSTORE);
    RegsetAddReg(*mask, IA64_AR_BSP);
  }
  else
  {
    /* zie rse_ensure_regs_loaded*/
    RegsetAddReg(*mask, IA64_AR_BSP);
    RegsetAddReg(*mask, IA64_AR_RSC);
    RegsetAddReg(*mask, IA64_CFM);
  }

}

/* M26: Integer ALAT Entry Invalidate */
void Ia64UsedRegistersM26(t_regset * mask, t_ins * ins)
{
  

}

/* M27: Floating-point ALAT Entry Invalidate */
void Ia64UsedRegistersM27(t_regset * mask, t_ins * ins)
{
  

}

/* M28: Flush Cache/Purge Translation Cache Entry */
void Ia64UsedRegistersM28(t_regset * mask, t_ins * ins)
{
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 6) == 0x34)
    RegsetAddReg(*mask, IA64_PSR);

}

/* M29: Move to AR - Register (M-Unit) */
void Ia64UsedRegistersM29(t_regset * mask, t_ins * ins)
{
  if(IA64INSOUT1(ins) == IA64_AR_BSPSTORE || IA64INSOUT1(ins) == IA64_AR_RNAT)
    RegsetAddReg(*mask, IA64_AR_RSC);
  
  RegsetAddReg(*mask, IA64_PSR);
  

}

/* M30: Move to AR - Immediate8 (M-Unit) */
void Ia64UsedRegistersM30(t_regset * mask, t_ins * ins)
{
  if(IA64INSOUT1(ins) == IA64_AR_BSPSTORE || IA64INSOUT1(ins) == IA64_AR_RNAT)
    RegsetAddReg(*mask, IA64_AR_RSC);

  RegsetAddReg(*mask, IA64_PSR);

}


/* M31: Move from AR (M-Unit) */
void Ia64UsedRegistersM31(t_regset * mask, t_ins * ins)
{
  if(IA64INSOUT1(ins) == IA64_AR_BSPSTORE || IA64INSOUT1(ins) == IA64_AR_RNAT)
    RegsetAddReg(*mask, IA64_AR_RSC);

  RegsetAddReg(*mask, IA64_PSR);
  

}


/* M32: Move to CR */
void Ia64UsedRegistersM32(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);
  

}


/* M33: Move from CR */
void Ia64UsedRegistersM33(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);
  

}


/* M34: Allocate Register Stack Frame */
void Ia64UsedRegistersM34(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_CFM);
  RegsetAddReg(*mask, IA64_AR_PFS);
  
}


/* M35: Move to PSR */
void Ia64UsedRegistersM35(t_regset * mask, t_ins * ins)
{
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 6) == 0x2D)
    RegsetAddReg(*mask, IA64_PSR);
  

}


/* M36: Move from PSR */
void Ia64UsedRegistersM36(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);
  

}


/* M37: Break/Nop (M-Unit) */
void Ia64UsedRegistersM37(t_regset * mask, t_ins * ins)
{
  
}

/* M38: Probe - Register */
void Ia64UsedRegistersM38(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

}


/* M39: Probe - Immediate2 */
void Ia64UsedRegistersM39(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);

}


/* M40: Probe Fault - Immediate2 */
void Ia64UsedRegistersM40(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

}


/* M41: Translation Cache Insert */
void Ia64UsedRegistersM41(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);
  RegsetAddReg(*mask, IA64_CR_ITIR);
  RegsetAddReg(*mask, IA64_CR_IFA);

  RegsetAddReg(*mask, IA64_RR);
  

}


/* M42: Move to Indirect Register/Translation Register Insert */
void Ia64UsedRegistersM42(t_regset * mask, t_ins * ins)
{
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 6) >= 0x0E)
  {
    RegsetAddReg(*mask, IA64_PSR);
    RegsetAddReg(*mask, IA64_CR_ITIR);
    RegsetAddReg(*mask, IA64_CR_IFA);

    RegsetAddReg(*mask, IA64_RR);
  }
  else
  {
    RegsetAddReg(*mask, IA64_PSR);
  }
}


/* M43: Move from Indirect Register */
void Ia64UsedRegistersM43(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);

}


/* M44: Set/Rest User/System Mask */
void Ia64UsedRegistersM44(t_regset * mask, t_ins * ins)
{
  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 4))
  {  
    case 0x4:
    case 0x5:
      break;
    case 0x6:
    case 0x7:
      RegsetAddReg(*mask, IA64_PSR);
      break;
    default:
      FATAL(("This shouldn't happen !!! Invalid opcode extension in M44 !!!"));
  }
}


/* M45: Translation Purge */
void Ia64UsedRegistersM45(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);
  RegsetAddReg(*mask, IA64_RR);
  

}


/* M46: Translation Access */
void Ia64UsedRegistersM46(t_regset * mask, t_ins * ins)
{
  switch(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 6))
  {  
    case 0x1A:
    case 0x1B:
      break;
    case 0x1E:
    case 0x1F:
      RegsetAddReg(*mask, IA64_PSR);
      break;
    default:
      FATAL(("This shouldn't happen !!! Invalid opcode extension in M46 !!!"));
  }
  

}







/*----------------------------------------------------------------------------*/





/* M1: Integer Load */
void Ia64DefinedRegistersM1(t_regset * mask, t_ins * ins)
{
  

}


/* M2: Integer Load - Increment by Register */
void Ia64DefinedRegistersM2(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64INSIN1(ins));
}

/* M3: Integer Load - Increment by immediate */
void Ia64DefinedRegistersM3(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64INSIN1(ins));
}



/* M4: Integer Store */
void Ia64DefinedRegistersM4(t_regset * mask, t_ins * ins)
{
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 30, 6) == 0x3B)
    RegsetAddReg(*mask, IA64_AR_UNAT);  

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
}


/* M5: Integer Store - Increment by immediate */
void Ia64DefinedRegistersM5(t_regset * mask, t_ins * ins)
{
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 30, 6) == 0x3B)
    RegsetAddReg(*mask, IA64_AR_UNAT);  
  RegsetAddReg(*mask, IA64INSIN2(ins));
  
  RegsetAddReg(*mask, IA64_RESOURCE_MEM);
}

/* M6: Floating-point Load */
void Ia64DefinedRegistersM6(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);    

}


/* M7: Floating-point Load - Increment by Register */
void Ia64DefinedRegistersM7(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);    
  RegsetAddReg(*mask, IA64INSIN1(ins));
}

/* M8: Floating-point Load - Increment by immediate */
void Ia64DefinedRegistersM8(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);    
  RegsetAddReg(*mask, IA64INSIN1(ins));
}



/* M9: Floating-point Store */
void Ia64DefinedRegistersM9(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}


/* M10: Floating-point Store - Increment by immediate */
void Ia64DefinedRegistersM10(t_regset * mask, t_ins * ins)
{
  

  RegsetAddReg(*mask, IA64INSIN2(ins));

  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}




/* M11: Floating-point Load Pair */
void Ia64DefinedRegistersM11(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  

}


/* M12: Floating-point Load Pair - Increment by Immediate*/
void Ia64DefinedRegistersM12(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);  
  RegsetAddReg(*mask, IA64INSIN1(ins));

}



/* M13: Line Prefetch */
void Ia64DefinedRegistersM13(t_regset * mask, t_ins * ins)
{
  
}

/* M14: Line Prefetch - Increment by Register */
void Ia64DefinedRegistersM14(t_regset * mask, t_ins * ins)
{
  
  RegsetAddReg(*mask, IA64INSIN1(ins));
}


/* M15: Line Prefetch - Increment by Immediate */
void Ia64DefinedRegistersM15(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64INSIN1(ins));
  
}


/* M16: Exchange/Compare and Exchange */
void Ia64DefinedRegistersM16(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_RESOURCE_MEM);

}


/* M17: Fetch and Add - Immediate */
void Ia64DefinedRegistersM17(t_regset * mask, t_ins * ins)
{
  

}

/* M18: Set FR */
void Ia64DefinedRegistersM18(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);    

}


/* M19: Get FR */
void Ia64DefinedRegistersM19(t_regset * mask, t_ins * ins)
{
  

}

/* M20: Integer Speculation Check (M-Unit) */
void Ia64DefinedRegistersM20(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);    
}

/* M21: Floating-point Speculation Check */
void Ia64DefinedRegistersM21(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);    
  
}


/* M22: Integer Advanced Load Check */
void Ia64DefinedRegistersM22(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);    
  
}


/* M23: Floating-point Advanced Load Check */
void Ia64DefinedRegistersM23(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_IP);    
  
}

/* M24: Sync/Fence/Serialize/ALAT Control */
void Ia64DefinedRegistersM24(t_regset * mask, t_ins * ins)
{
  

}


/* M25: RSE Control */
void Ia64DefinedRegistersM25(t_regset * mask, t_ins * ins)
{
  if(GETBITS64(IA64OPCODEINFO(ins).opcode, 27, 4) == 0xC)
  {
    RegsetAddReg(*mask, IA64_AR_BSPSTORE);
  }
  else
  {
    RegsetAddReg(*mask, IA64_AR_RNAT);
  }
  

}

/* M26: Integer ALAT Entry Invalidate */
void Ia64DefinedRegistersM26(t_regset * mask, t_ins * ins)
{
  
}

/* M27: Floating-point ALAT Entry Invalidate */
void Ia64DefinedRegistersM27(t_regset * mask, t_ins * ins)
{
  
}

/* M28: Flush Cache/Purge Translation Cache Entry */
void Ia64DefinedRegistersM28(t_regset * mask, t_ins * ins)
{
  
}

/* M29: Move to AR - Register (M-Unit) */
void Ia64DefinedRegistersM29(t_regset * mask, t_ins * ins)
{
  if(IA64INSOUT1(ins) == IA64_AR_BSPSTORE)
  {
    RegsetAddReg(*mask, IA64_AR_RNAT);
    RegsetAddReg(*mask, IA64_AR_BSP);

    /* zie rse_update_internal_stack_pointers */  
    RegsetAddReg(*mask, IA64_AR_BSP);
  }
  

}

/* M30: Move to AR - Immediate8 (M-Unit) */
void Ia64DefinedRegistersM30(t_regset * mask, t_ins * ins)
{
  if(IA64INSOUT1(ins) == IA64_AR_BSPSTORE)
  {
    RegsetAddReg(*mask, IA64_AR_RNAT);
    RegsetAddReg(*mask, IA64_AR_BSP);
    /* zie rse_update_internal_stack_pointers */  
    RegsetAddReg(*mask, IA64_AR_BSP);
  }
  

}


/* M31: Move from AR (M-Unit) */
void Ia64DefinedRegistersM31(t_regset * mask, t_ins * ins)
{
  

}


/* M32: Move to CR */
void Ia64DefinedRegistersM32(t_regset * mask, t_ins * ins)
{
  

}


/* M33: Move from CR */
void Ia64DefinedRegistersM33(t_regset * mask, t_ins * ins)
{
  

}


/* M34: Allocate Register Stack Frame */
void Ia64DefinedRegistersM34(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_CFM);
  

}


/* M35: Move to PSR */
void Ia64DefinedRegistersM35(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);
  
  
}


/* M36: Move from PSR */
void Ia64DefinedRegistersM36(t_regset * mask, t_ins * ins)
{
  

}


/* M37: Break/Nop (M-Unit) */
void Ia64DefinedRegistersM37(t_regset * mask, t_ins * ins)
{
  
}

/* M38: Probe - Register */
void Ia64DefinedRegistersM38(t_regset * mask, t_ins * ins)
{
  

}


/* M39: Probe - Immediate2 */
void Ia64DefinedRegistersM39(t_regset * mask, t_ins * ins)
{
  

}


/* M40: Probe Fault - Immediate2 */
void Ia64DefinedRegistersM40(t_regset * mask, t_ins * ins)
{
  
}


/* M41: Translation Cache Insert */
void Ia64DefinedRegistersM41(t_regset * mask, t_ins * ins)
{
  
}


/* M42: Move to Indirect Register/Translation Register Insert */
void Ia64DefinedRegistersM42(t_regset * mask, t_ins * ins)
{
  

}


/* M43: Move from Indirect Register */
void Ia64DefinedRegistersM43(t_regset * mask, t_ins * ins)
{
  

}


/* M44: Set/Rest User/System Mask */
void Ia64DefinedRegistersM44(t_regset * mask, t_ins * ins)
{
  RegsetAddReg(*mask, IA64_PSR);
  
}


/* M45: Translation Purge */
void Ia64DefinedRegistersM45(t_regset * mask, t_ins * ins)
{
  
}


/* M46: Translation Access */
void Ia64DefinedRegistersM46(t_regset * mask, t_ins * ins)
{
  

}




/* vim: set shiftwidth=2: */
