/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "mips_assemble.h"
#include "mips_assemble_one.h"
#include "mips_opcodes.h"
#include "../../utils/util_error.h"

void MipsAssembleOne(t_ins * i_ins, char * data)
{
    if (INS_TYPE(i_ins) != IT_DATA)
    {
      if (MIPSINSPROCESSOR(i_ins)==CPU)
	MipsAssembleCPU(i_ins, (t_uint32*)data); 
      else if(MIPSINSPROCESSOR(i_ins)==COP0)
	MipsAssembleCOP0(i_ins, (t_uint32*)data); 
      else if(MIPSINSPROCESSOR(i_ins)==COP1)
	switch(INS_TYPE(i_ins))
	{
	  case IT_BRANCH:
	    MipsAssembleCOP1Branch(i_ins, (t_uint32*)data); 
	    break;
	  case IT_STORE:
	  case IT_LOAD:
	    MipsAssembleCOP1LoadStore(i_ins, (t_uint32*)data); 
	    break;
	  case IT_DATAPROC:
	    MipsAssembleCOP1Dataproc(i_ins, (t_uint32*)data);
	    break;
	  default:
	    FATAL(("Assemble Fault; No implementation for %d\n",MIPSINSOPCODE(i_ins)));
	    break;
	}
    }
    else 
      *((t_uint32 *) data)=MIPSINSIMMEDIATE(i_ins);
}

void MipsAssembleSection(t_object * obj, t_section * sec)
{
  t_uint32 address=0;
  int nins = 0;
  t_ins * i_ins;
  char * data = SECTION_TMP_BUF(sec);

  i_ins = T_INS(sec->data); 
  address=G_T_UINT32(SECTION_GET_OLD_ADDRESS(sec));


  while(i_ins != NULL)
  {
    MipsAssembleOne(i_ins, data);

    nins++;
    data += 4;
    address+=4;
    i_ins = INS_NEXT(i_ins); 
  }

  printf("Assembled %d instructions\n",nins);
}
/* vim: set shiftwidth=2: */
