/*
 * Copyright 2001,2002{{{ : Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 }}}*/
#include "mips_utils.h"
#include "mips_constant_propagation.h"
#include "mips_registers.h"
#include "mips_instructions.h"
#include "mips_opcodes.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_print.h"
#include "../../datastructures/data_graph.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_iterators.h"

t_bool MipsInsOptConstantResultInTwoRegisters(t_ins * ins, t_procstate * before, t_procstate * after);
t_bool MipsInsOptMakeIndirectJumpsDirect(t_ins * ins, t_procstate * procstate1, t_procstate * procstate2);
t_bool MipsInsOptMakeRegisterJumpsDirect(t_ins * ins, t_procstate * procstate1, t_procstate * procstate2);
t_bool MipsInsOptConstantResultInOneRegisterAndImmediate(t_ins * ins, t_procstate * before, t_procstate * after);
void RemoveJumpEdges(t_ins * ins);
void RemoveCallEdges(t_ins * ins);
void MakeConstant(t_ins * ins,t_reg register1,t_uint32 waarde);
void MakeCall(t_ins * ins);
void MipsTestAndSetOneRegister(t_cfg_edge* edge,t_procstate * state,t_procstate ** state_true,t_procstate ** state_false)
{
  register_content value1;
  t_lattice_level reg1;
  t_reg register1;
  t_ins * jump_instr;
  
  if(INS_TYPE(BBL_INS_LAST(CFG_EDGE_HEAD(edge)))==IT_BRANCH)
    jump_instr=BBL_INS_LAST(CFG_EDGE_HEAD(edge));
  else if(INS_TYPE(INS_PREV(BBL_INS_LAST(CFG_EDGE_HEAD(edge))))==IT_BRANCH)
    jump_instr=INS_PREV(BBL_INS_LAST(CFG_EDGE_HEAD(edge)));
  else FATAL(("Geen Branch??"));

  reg1=ProcStateGetReg(state,MIPS_BRANCH_REG1,&value1);

  register1=MIPSINSRS(jump_instr);

#ifdef EXTENDED_CONSTPROP

  {
    t_bool plus,minus,zero,ok;
    plus=ProcStateGetPlus(state,MIPS_BRANCH_REG1);
    zero=ProcStateGetZero(state,MIPS_BRANCH_REG1);
    minus=ProcStateGetMinus(state,MIPS_BRANCH_REG1);
    ok=FALSE;

    switch(MIPSINSOPCODE(jump_instr))
    {
      case MIPS_BLTZ:
      case MIPS_BLTZAL:
	if (!plus && !zero)
	{
	  *state_true=state;
	  *state_false=NULL;
	  ok=TRUE;
	}
	else if (!minus)
	{
          *state_true=NULL;
          *state_false=state;
	  ok=TRUE;
	}
	if(ok==FALSE)
	{
	  *state_true  = ProcStateNew();
	  *state_false = state;
	  ProcStateDup(*state_true,state);
	  ProcStateSetRegBotExtend(*state_true,register1,FALSE,FALSE,TRUE);
	  ProcStateSetRegBotExtend(*state_false,register1,TRUE,TRUE,FALSE);
	  ok=TRUE;
	}
	break;
      case MIPS_BGEZ:
      case MIPS_BGEZAL:
	if (!minus)
	{
	  *state_true=state;
	  *state_false=NULL;
	  ok=TRUE;
	}
	else if (!plus && !zero)
	{
          *state_true=NULL;
          *state_false=state;
	  ok=TRUE;
	}
	if(ok==FALSE)
	{
	  *state_true  = ProcStateNew();
	  *state_false = state;
	  ProcStateDup(*state_true,state);
	  ProcStateSetRegBotExtend(*state_true,register1,TRUE,TRUE,FALSE);
	  ProcStateSetRegBotExtend(*state_false,register1,FALSE,FALSE,TRUE);
	  ok=TRUE;
	}
	break;
      case MIPS_BLEZ:
	if (!plus)
	{
	  *state_true=state;
	  *state_false=NULL;
	  ok=TRUE;
	}
	else if (!minus && !zero)
	{
          *state_true=NULL;
          *state_false=state;
	  ok=TRUE;
	}
	if(ok==FALSE)
	{
	  *state_true  = ProcStateNew();
	  *state_false = state;
	  ProcStateDup(*state_true,state);
	  ProcStateSetRegBotExtend(*state_true,register1,FALSE,TRUE,TRUE);
	  ProcStateSetRegBotExtend(*state_false,register1,TRUE,FALSE,FALSE);
	  ok=TRUE;
	}
	break;
      case MIPS_BGTZ:
	if (!minus && !zero)
	{
	  *state_true=state;
	  *state_false=NULL;
	  ok=TRUE;
	}
	else if (!plus)
	{
          *state_true=NULL;
          *state_false=state;
	  ok=TRUE;
	}
	if(ok==FALSE)
	{
	  *state_true  = ProcStateNew();
	  *state_false = state;
	  ProcStateDup(*state_true,state);
	  ProcStateSetRegBotExtend(*state_true,register1,TRUE,FALSE,FALSE);
	  ProcStateSetRegBotExtend(*state_false,register1,FALSE,TRUE,TRUE);
	  ok=TRUE;
	}
	break;
      default:
        FATAL(("Implement"));
	//ok=FALSE;
	break;
    }
    if(ok==FALSE)
    {
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
    }
  }
#else
    
  /* BOT wijst op verschillende producers => geen constante waarde */
  if (BOT==reg1)
  {
    *state_true  = ProcStateNew();
    *state_false = state;
    ProcStateDup(*state_true,state);
  }
  else
  {
    switch(MIPSINSOPCODE(jump_instr))
    {
      case MIPS_BLTZ:
      case MIPS_BLTZAL:
	if ((t_int32)G_T_UINT32(value1.i)<0)
	{
	  *state_true=state;
	  *state_false=NULL;
	}
	else
	{
          *state_true=NULL;
          *state_false=state;
	}
	break;
      case MIPS_BGEZ:
      case MIPS_BGEZAL:
	if ((t_int32)G_T_UINT32(value1.i)>=0)
	{
	  *state_true=state;
	  *state_false=NULL;
	}
	else
	{
          *state_true=NULL;
          *state_false=state;
	}
	break;
      case MIPS_BLEZ:
	if ((t_int32)G_T_UINT32(value1.i)<=0)
	{
	  *state_true=state;
	  *state_false=NULL;
	}
	else
	{
          *state_true=NULL;
          *state_false=state;
	}
	break;
      case MIPS_BGTZ:
	if ((t_int32)G_T_UINT32(value1.i)>0)
	{
	  *state_true=state;
	  *state_false=NULL;
	}
	else
	{
          *state_true=NULL;
          *state_false=state;
	}
	break;
      default:
        FATAL(("Implement"));
	break;
    }
  }
  
#endif

}


//-------------------------

void MipsTestAndSetTwoRegisters(t_cfg_edge* edge,t_procstate * state,t_procstate ** state_true,t_procstate ** state_false)
{
  register_content value1,value2;
  t_lattice_level reg1,reg2;
  t_reg register1,register2;
  t_ins * jump_instr;
  
  if(INS_TYPE(BBL_INS_LAST(CFG_EDGE_HEAD(edge)))==IT_BRANCH)
    jump_instr=BBL_INS_LAST(CFG_EDGE_HEAD(edge));
  else if(INS_TYPE(INS_PREV(BBL_INS_LAST(CFG_EDGE_HEAD(edge))))==IT_BRANCH)
    jump_instr=INS_PREV(BBL_INS_LAST(CFG_EDGE_HEAD(edge)));
  else FATAL(("Geen Branch??"));
  
  reg1=ProcStateGetReg(state,MIPS_BRANCH_REG1,&value1);
  reg2=ProcStateGetReg(state,MIPS_BRANCH_REG2,&value2);
  
  register1=MIPSINSRS(jump_instr);
  register2=MIPSINSRT(jump_instr);

  /* BOT wijst op verschillende producers => geen constante waarde */
  if (BOT==reg1 && BOT==reg2)
    {
#ifdef EXTENDED_CONSTPROP
    t_bool plus,minus,zero,ok;
    plus=ProcStateGetPlus(state,MIPS_BRANCH_REG1);
    zero=ProcStateGetPlus(state,MIPS_BRANCH_REG1);
    minus=ProcStateGetPlus(state,MIPS_BRANCH_REG1);
    //if(INS_ADDRESS(jump_instr)==0xa002499c)
      //DiabloPrint(stdout,"BNE:eerste %d %d %d\n",plus,zero,minus); 
    ok=FALSE;
    if(MIPSINSRS(jump_instr)==MIPS_REG_R0)
    switch(MIPSINSOPCODE(jump_instr))
    {
      case MIPS_BEQ:
	//DiabloPrint(stdout,"BEQ: %d %d %d\n",plus,zero,minus); 
	if (!zero)
	{
          *state_true=NULL;
          *state_false=state;
	  //DiabloPrint(stdout,"Different: %d %d %d\n",plus,zero,minus); 
	  ok=TRUE;
	}
	break;
      default:
	break;
    }
    if(ok==FALSE)
    {
	*state_true  = ProcStateNew();
	*state_false = state;
	ProcStateDup(*state_true,state);
    }
#else
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
#endif
    }
  else if (BOT==reg1)
    {
      t_ins * last_instr;
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
      last_instr=BBL_INS_LAST(CFG_EDGE_HEAD(edge));
      //DiabloPrint(stdout,"UITVOERING reg1\n");
     
      switch(MIPSINSOPCODE(jump_instr))
      {
	case MIPS_BEQ:
	  if(jump_instr==last_instr)
	    ProcStateSetReg(*state_true,register1,value2);
	  else if((MIPSINSRS(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRS(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRS(jump_instr)==MIPS_REG_R0)
	    if((MIPSINSRT(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRT(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRT(jump_instr)==MIPS_REG_R0)
	      ProcStateSetReg(*state_true,register1,value2);
	  break;
	case MIPS_BNE:
	  if(jump_instr==last_instr)
	    ProcStateSetReg(*state_false,register1,value2);
	  else if((MIPSINSRS(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRS(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRS(jump_instr)==MIPS_REG_R0)
	    if((MIPSINSRT(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRT(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRT(jump_instr)==MIPS_REG_R0)
	      ProcStateSetReg(*state_false,register1,value2);
	  break;
	default:
          FATAL(("Implement"));
	  break;
      }
    }
  else if (BOT==reg2)
    {
      t_ins * last_instr;
      *state_true  = ProcStateNew();
      *state_false = state;
      ProcStateDup(*state_true,state);
      last_instr=BBL_INS_LAST(CFG_EDGE_HEAD(edge));
      //DiabloPrint(stdout,"UITVOERING reg2\n");

      switch(MIPSINSOPCODE(jump_instr))
      {
	case MIPS_BEQ:
	  if(jump_instr==last_instr)
	    ProcStateSetReg(*state_true,register2,value1);
	  else if((MIPSINSRS(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRS(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRS(jump_instr)==MIPS_REG_R0)
	    if((MIPSINSRT(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRT(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRT(jump_instr)==MIPS_REG_R0)
	      ProcStateSetReg(*state_true,register2,value1);
	  break;
	case MIPS_BNE:
	  if(jump_instr==last_instr)
	    ProcStateSetReg(*state_false,register2,value1);
	  else if((MIPSINSRS(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRS(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRS(jump_instr)==MIPS_REG_R0)
	    if((MIPSINSRT(jump_instr)!=MIPSINSRD(last_instr) && MIPSINSRT(jump_instr)!=MIPSINSRT(last_instr))||MIPSINSRT(jump_instr)==MIPS_REG_R0)
	      ProcStateSetReg(*state_false,register2,value1);
	  break;
	default:
          FATAL(("Implement"));
	  break;
      }
    }
  else 
  {
    switch(MIPSINSOPCODE(jump_instr))
    {
      case MIPS_BEQ:
	if (G_T_UINT32(value1.i)==G_T_UINT32(value2.i))
	{
	  *state_true=state;
	  *state_false=NULL;
	}
	else
	{
          *state_true=NULL;
          *state_false=state;
	}
	break;
      case MIPS_BNE:
	if (G_T_UINT32(value1.i)!=G_T_UINT32(value2.i))
	{
	  *state_true=state;
	  *state_false=NULL;
	}
	else
	{
          *state_true=NULL;
          *state_false=state;
	}
	break;
      default:
        FATAL(("Implement"));
	break;
    }
  }
}

//---------------------------------------------


void MipsTestAndSetOneRegisterCOP1(t_cfg_edge* edge,t_procstate * state,t_procstate ** state_true,t_procstate ** state_false)
{
  t_uint8 bit;
  t_ins * jump_instr;
  t_lattice_level reg1;
  register_content value1;
  
  if(INS_TYPE(BBL_INS_LAST(CFG_EDGE_HEAD(edge)))==IT_BRANCH)
    jump_instr=BBL_INS_LAST(CFG_EDGE_HEAD(edge));
  else
    jump_instr=INS_PREV(BBL_INS_LAST(CFG_EDGE_HEAD(edge)));

  bit=MIPSINSCC(jump_instr);

  if(bit!=0)
  {
    *state_true  = ProcStateNew();
    *state_false = state;
    ProcStateDup(*state_true,state);
    return;
  }

  reg1=ProcStateGetReg(state,MIPS_REG_FCCR,&value1);

  if (BOT==reg1)
  {
    *state_true  = ProcStateNew();
    *state_false = state;
    ProcStateDup(*state_true,state);
  }
  else

  switch(MIPSINSOPCODE(jump_instr))
  {
    case MIPS_BC1F:
      if (((t_int32) G_T_UINT32(value1.i))==0)
      {
	*state_true=state;
	*state_false=NULL;
      }
      else
      {
        *state_true=NULL;
        *state_false=state;
      }
      break;
    case MIPS_BC1T:
      if (((t_int32) G_T_UINT32(value1.i))==1)
      {
	*state_true=state;
	*state_false=NULL;
      }
      else
      {
        *state_true=NULL;
        *state_false=state;
      }
      break;
    default:
      FATAL(("Implement"));
      break;
  }
}


void MipsTestAndSetCondition_BGEZ(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetOneRegister(edge,state, state_true, state_false);
}

void MipsTestAndSetCondition_BLTZ(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetOneRegister(edge,state, state_true, state_false);
}

void MipsTestAndSetCondition_BEQ(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetTwoRegisters(edge,state, state_true, state_false);
}  

void MipsTestAndSetCondition_BNE(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetTwoRegisters(edge,state, state_true, state_false);
}  

void MipsTestAndSetCondition_BLEZ(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetOneRegister(edge,state, state_true, state_false);
}

void MipsTestAndSetCondition_BGTZ(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetOneRegister(edge,state, state_true, state_false);
}

void MipsTestAndSetCondition_BC1F(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetOneRegisterCOP1(edge,state, state_true, state_false);
}

void MipsTestAndSetCondition_BC1T(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  MipsTestAndSetOneRegisterCOP1(edge,state, state_true, state_false);
}

void MipsSet_TsF0(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true=state;
  *state_false=NULL;
}

void MipsSet_T0Fs(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true=NULL;
  *state_false=state;
}

void MipsSWI(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
}

void MipsTestAndSetConditionAllPaths(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
  ProcStateDup(*state_true,state);
}

void MipsTestAndSetConditionSetAllToBot(t_cfg_edge* edge,t_procstate* state, t_procstate **state_true, t_procstate** state_false)
{
  *state_true  = ProcStateNew();
  *state_false = state;
  ProcStateSetAllBot(*state_true,AllRegs());
}

void MipsEdgePropagator(t_cfg_edge * edge, t_ins * ins)
{
  t_uint8  condition;
  t_uint16 edgetype  = EDGE_CAT(edge);
  
  /* Als er een edge van de Hellnode komt, dan moeten we over alle paden propageren */
      //DiabloPrint(stdout,"Edge: @E \n",edge);
      //DiabloPrint(stdout,"head BBL: @B \n",EDGE_HEAD(edge));
  if (IsHellNode(EDGE_HEAD(edge)))
  {
    edge->testcondition = &MipsTestAndSetConditionAllPaths;
    return;
  }
  
  switch (edgetype)
  {
    case ET_CALL:
    case ET_IPJUMP:
    case ET_JUMP:
      {
    if (ins && INS_TYPE(ins)==IT_BRANCH)
      condition = MIPSINSOPCODE(T_MIPS_INS(ins));
    else if (ins && INS_PREV(ins) && INS_TYPE(INS_PREV(ins))==IT_BRANCH)
      condition = MIPSINSOPCODE(T_MIPS_INS(INS_PREV(ins)));
    else
    {
      DiabloPrint(stdout,"Edge: @E \n",edge);
      DiabloPrint(stdout,"head BBL: @B \n",EDGE_HEAD(edge));
      BblPrint(CFG_EDGE_HEAD(edge),PRINT_INSTRUCTIONS);
            FATAL(("ONMOGELIJK"));
      break;
    }
	switch (condition)
	{
	  case MIPS_JR:
	  case MIPS_J:
	  case MIPS_JAL:
	  case MIPS_JALR:
            edge->testcondition = &MipsSet_TsF0;
	    break;
	  case MIPS_BLTZ:
	  case MIPS_BLTZAL:
            edge->testcondition = &MipsTestAndSetCondition_BLTZ;
	    break;
	  case MIPS_BGEZ:
	  case MIPS_BGEZAL:
            edge->testcondition = &MipsTestAndSetCondition_BGEZ;
	    break;
	  case MIPS_BEQ:
            edge->testcondition = &MipsTestAndSetCondition_BEQ;
	    break;
	  case MIPS_BNE:
            edge->testcondition = &MipsTestAndSetCondition_BNE;
	    break;
	  case MIPS_BLEZ:
            edge->testcondition = &MipsTestAndSetCondition_BLEZ;
	    break;
	  case MIPS_BGTZ:
            edge->testcondition = &MipsTestAndSetCondition_BGTZ;
	    break;
	  case MIPS_BC1F:
            edge->testcondition = &MipsTestAndSetCondition_BC1F;
	    break;
	  case MIPS_BC1T:
            edge->testcondition = &MipsTestAndSetCondition_BC1T;
	    break;
	  default:
	    DiabloPrint(stdout,"@I\n",ins);
            FATAL(("Jump-conditie niet gevonden:%d \n",condition));
	    break;
	}
      }
      break;
    case ET_IPFALLTHRU:
    case ET_FALLTHROUGH:
      edge->testcondition = &MipsSet_TsF0;
      break;
      
    case ET_COMPENSATING:
    case ET_RETURN:
      edge->testcondition = &MipsTestAndSetConditionAllPaths;
      break;
      
      /* Over een FUNLINK wordt NOOIT gepropageerd */    
    case ET_FUNLINK:
    case ET_IPFUNLINK:
    case ET_SWI:
      //edge->testcondition = &MipsSet_T0Fs;
      edge->testcondition = &MipsSWI;
      //edge->testcondition = &MipsTestAndSetConditionAllPaths;
      break;
    case ET_SWILINK:
    case ET_IPSWILINK:
      edge->testcondition = &MipsTestAndSetConditionSetAllToBot;
      break;

    default:
      FATAL(("MipsSetConditionToProp invalid type %lx  %d\n",edgetype,edgetype));
      break;
  }
}

t_bool MipsInstructionUnconditionalizer(t_ins * ins)
{
  return FALSE;
}

t_bool MipsInstructionsConstantOptimizer(t_ins * ins, t_procstate * procstate1, t_procstate * procstate2)
{
  if (!MipsInsOptMakeIndirectJumpsDirect(ins,procstate1,procstate2)) return FALSE;
  if (!MipsInsOptMakeRegisterJumpsDirect(ins,procstate1,procstate2)) return FALSE;
//  if (!MipsInsOptConstantResultInTwoRegisters(ins,procstate1,procstate2)) return FALSE;
//  if (!MipsInsOptConstantResultInOneRegisterAndImmediate(ins,procstate1,procstate2)) return FALSE;
  return TRUE;
}

t_bool MipsInsOptConstantResultInOneRegisterAndImmediate(t_ins * ins, t_procstate * before, t_procstate * after)
{
#ifdef EXTENDED_CONSTPROP
  register_content value1,value2;
  t_lattice_level reg1,reg2;
  t_reg register1,register2;
  t_int32 s_src1;
  
  if (RELOCATABLE_REFERS_TO(ins))
  {
    return TRUE;
  }

  register1=MIPSINSRS(ins);
  register2=MIPSINSRT(ins);

  if (register1 == MIPS_REG_NONE || register2 == MIPS_REG_NONE) return TRUE;
  reg1=ProcStateGetReg(before,register1,&value1);
  reg2=ProcStateGetReg(after,register2,&value2);

  if(reg2!= VALUE)
    return TRUE;

  s_src1=(t_int32)G_T_UINT32(value2.i);

  if(register1 != MIPS_REG_R0 && register2 != MIPS_REG_R0)
    if(register1 <= MIPS_REG_R31 && register2 <= MIPS_REG_R31)
      if(MIPSINSIMMEDIATE(ins)!=0)
	if(INS_TYPE(ins)==IT_DATAPROC)
	{
//	  DiabloPrint(stdout,"1: @I value:%x\n",ins,s_src1);
	  if(s_src1 >=0 && s_src1<32767)
	  {
//	    DiabloPrint(stdout,"Change\n");
	    MakeConstant(ins,register2,(t_uint32)s_src1);
	    return TRUE;
	  }
	}
#endif
  return TRUE;
}

t_bool MipsInsOptConstantResultInTwoRegisters(t_ins * ins, t_procstate * before, t_procstate * after)
{
#ifdef EXTENDED_CONSTPROP
  register_content value1,value2,value3;
  t_lattice_level reg1,reg2,reg3;
  t_reg register1,register2,register3;
  t_bool plus1,minus1,zero1;
  t_bool plus2,minus2,zero2;
  //t_uint32 u_src1;
  t_int32 s_src1,s_src2,s_src3;
  
  register1=MIPSINSRS(ins);
  register2=MIPSINSRT(ins);
  register3=MIPSINSRD(ins);

  if (RELOCATABLE_REFERS_TO(ins))
    return TRUE;

  if (register1 == MIPS_REG_NONE || register2 == MIPS_REG_NONE || register3 == MIPS_REG_NONE) return TRUE;
  reg1=ProcStateGetReg(before,register1,&value1);
  reg2=ProcStateGetReg(before,register2,&value2);
  reg3=ProcStateGetReg(after,register3,&value3);

  if(reg1!=VALUE && reg2!= VALUE && reg3!=VALUE)
    return TRUE;

  if (reg3 == VALUE)
  {
  plus1=ProcStateGetPlus(before,register1);
  zero1=ProcStateGetZero(before,register1);
  minus1=ProcStateGetMinus(before,register1);
  plus2=ProcStateGetPlus(before,register2);
  zero2=ProcStateGetZero(before,register2);
  minus2=ProcStateGetMinus(before,register2);

  s_src1=(t_int32)G_T_UINT32(value3.i);
  s_src2=(t_int32)G_T_UINT32(value2.i);
  s_src3=(t_int32)G_T_UINT32(value1.i);

  if(register1 != MIPS_REG_R0 && register2 != MIPS_REG_R0 && register3 != MIPS_REG_R0)
    if(register1 <= MIPS_REG_R31 && register2 <= MIPS_REG_R31 && register3 <= MIPS_REG_R31)
    {
//      DiabloPrint(stdout,"2: @I value:%x = %x OP %x\n",ins,s_src1,s_src2,s_src3);
      if(s_src1 >=0 && s_src1<32767)
      {
//	DiabloPrint(stdout,"Change\n");
	MakeConstant(ins,register3,(t_uint32)s_src1);
	  return TRUE;
      }
    }
  }
  else if (reg1 == VALUE)
  {
    s_src3=(t_int32)G_T_UINT32(value1.i);
    //DiabloPrint(stdout, "2:Value1 : %d, @I\n",s_src3,ins);
  }
  else if (reg2 == VALUE && register2 != MIPS_REG_R0)
  {
    s_src2=(t_int32)G_T_UINT32(value2.i);
    //DiabloPrint(stdout, "2:Value2 : %d, @I\n",s_src2,ins);
  }
#endif
  return TRUE;
}

t_bool MipsInsOptMakeRegisterJumpsDirect(t_ins * ins, t_procstate * procstate1, t_procstate * procstate2)
{
#ifdef EXTENDED_CONSTPROP

  if(MIPSINSOPCODE(ins)==MIPS_JR)
  {
    register_content value1;
    t_lattice_level reg1;
    t_uint32 u_src1;

    reg1=ProcStateGetReg(procstate2,MIPS_BRANCH_REG1,&value1);

    u_src1=G_T_UINT32(value1.i);

    if(reg1==VALUE && ((u_src1 & 0xf0000000)==(INS_OLD_ADDRESS(ins)&0xf0000000)) && MIPSINSRS(ins)!=MIPS_REG_R31 )
    {
      MakeUnConditional(ins);
      teller++;
    }
  }
#else 
    return TRUE;
#endif
  return TRUE;
}

t_bool MipsInsOptMakeIndirectJumpsDirect(t_ins * ins, t_procstate * procstate1, t_procstate * procstate2)
{
  register_content value1,value2,value3;
  t_lattice_level reg1,reg2,reg3;
  t_uint8 bit;

  if(INS_TYPE(ins)!=IT_BRANCH)
    return TRUE;

  bit=MIPSINSCC(ins);

  reg1=ProcStateGetReg(procstate2,MIPS_BRANCH_REG1,&value1);
  reg2=ProcStateGetReg(procstate2,MIPS_BRANCH_REG2,&value2);
  reg3=ProcStateGetReg(procstate2,MIPS_REG_FCCR,&value3);


#ifdef EXTENDED_CONSTPROP

  {
    t_bool plus,minus,zero,ok;
    plus=ProcStateGetPlus(procstate2,MIPS_BRANCH_REG1);
    zero=ProcStateGetZero(procstate2,MIPS_BRANCH_REG1);
    minus=ProcStateGetMinus(procstate2,MIPS_BRANCH_REG1);
    ok=FALSE;

    switch(MIPSINSOPCODE(ins))
    {
      case MIPS_BLTZ:
	if (!plus && !zero)
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	else if (!minus)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	break;
      case MIPS_BLTZAL:
	if (!plus && !zero)
	{
	  MakeCall(ins);
	  return TRUE;
	}
	else if (!minus)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  if(ins!=INS_BBL(ins)->cfg->entries->entry_bbl->ins_first)
	  {
	    RemoveCallEdges(ins);
	    InsKill(ins);
	  }
	  else
//	    DiabloPrint(stdout,"No entry-point kill\n");
	  return FALSE;
	}
	break;
      case MIPS_BGEZ:
	if (!minus)
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	else if (!plus && !zero)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	break;
      case MIPS_BGEZAL:
	if (!minus)
	{
	  MakeCall(ins);
	  return TRUE;
	}
	else if (!plus && !zero)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveCallEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	break;
      case MIPS_BLEZ:
	if (!plus)
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	else if (!minus && !zero)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
//	  DiabloPrint(stdout,"minus:%d zero:%d plus:%d\n",minus,zero,plus);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	break;
      case MIPS_BGTZ:
	if (!minus && !zero)
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	else if (!plus)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	break;
#else
  /* BOT wijst op verschillende producers => geen constante waarde */
  {
    switch(MIPSINSOPCODE(ins))
    {
      case MIPS_BLTZ:
	if (BOT==reg1)
	{
	  return TRUE;
	}
	if ((t_int32) G_T_UINT32(value1.i)>=0)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	else
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	break;
      case MIPS_BLTZAL:
	if (BOT==reg1)
	{
	  return TRUE;
	}
	if ((t_int32) G_T_UINT32(value1.i)>=0)
	{
	  if(ins!=INS_BBL(ins)->cfg->entries->entry_bbl->ins_first)
	  {
//	    DiabloPrint(stdout,"Kill: @I\n",ins);
	    RemoveCallEdges(ins);
	    InsKill(ins);
	  }
	  else
//	    DiabloPrint(stdout,"NO Kill: @I\n",ins);
	  return FALSE;
	}
	else
	{
	  if(ins!=INS_BBL(ins)->cfg->entries->entry_bbl->ins_first)
	  {
//	    DiabloPrint(stdout,"Make Call: @I\n",ins);
	    MakeCall(ins);
	  }
	  else
	    DiabloPrint(stdout,"NO CHANGES: @I\n",ins);
	  return TRUE;
	}
	break;
      case MIPS_BGEZ:
	if (BOT==reg1)
	{
	  return TRUE;
	}
	if ((t_int32) G_T_UINT32(value1.i)<0)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	else
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	break;
      case MIPS_BGEZAL:
	if (BOT==reg1)
	{
	  return TRUE;
	}
	if ((t_int32) G_T_UINT32(value1.i)<0)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveCallEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	else
	{
	  if(ins!=INS_BBL(ins)->cfg->entries->entry_bbl->ins_first)
	  {
//	    DiabloPrint(stdout,"Make Call: @I\n",ins);
	    MakeCall(ins);
	  }
	  else
//	    DiabloPrint(stdout,"NO CHANGES: @I\n",ins);
	  return TRUE;
	}
	break;
      case MIPS_BLEZ:
	if (BOT==reg1)
	{
	  return TRUE;
	}
	if ((t_int32) G_T_UINT32(value1.i)>0)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	else
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	break;
      case MIPS_BGTZ:
	if (BOT==reg1)
	{
	  return TRUE;
	}
	if ((t_int32) G_T_UINT32(value1.i)<=0)
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	else
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	break;
#endif
      case MIPS_BEQ:
	if (BOT==reg1 || BOT==reg2)
	{
	  return TRUE;
	}
	if (G_T_UINT32(value1.i)!=G_T_UINT32(value2.i))
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	else
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	break;
      case MIPS_BNE:
	if (BOT==reg1 || BOT==reg2)
	{
	  return TRUE;
	}
	if (G_T_UINT32(value1.i)==G_T_UINT32(value2.i))
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
//	  DiabloPrint(stdout,"%x %x \n",G_T_UINT32(value1.i),G_T_UINT32(value2.i));
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	else
	{
//	  DiabloPrint(stdout,"UnCond: @I\n",ins);
//	  DiabloPrint(stdout,"%x %x \n",G_T_UINT32(value1.i),G_T_UINT32(value2.i));
	  MakeUnConditional(ins);
	  return TRUE;
	}
	break;
      case MIPS_BC1F:
	if (BOT==reg3)
	{
	  return TRUE;
	}
	if(bit!=0)
	  return TRUE;
	if (((t_int32) G_T_UINT32(value3.i))==0)
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	else
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	break;
    case MIPS_BC1T:
	if (BOT==reg3)
	{
	  return TRUE;
	}
	if(bit!=0)
	  return TRUE;
	if (((t_int32) G_T_UINT32(value1.i))==1)
	{
	  MakeUnConditional(ins);
	  return TRUE;
	}
	else
	{
//	  DiabloPrint(stdout,"Kill @I\n",ins);
	  RemoveJumpEdges(ins);
	  InsKill(ins);
	  return FALSE;
	}
	break;
      default:
	return TRUE;
	break;
    }
  }
	return TRUE;
}


void MakeCall(t_ins * ins)
{
   INS_TYPE(ins)=IT_BRANCH;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=MIPS_JAL;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSRS(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSINSTR_INDEX(ins)=0;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins)=0;
   INS_ATTRIB(ins)|=IF_PCDEF;
   INS_ATTRIB(ins)|=IF_CHANGED; 

  INS_REGS_USE(ins) = MipsUsedRegisters(ins);
  INS_REGS_DEF(ins) = MipsDefinedRegisters(ins);
}


void RemoveCallEdges(t_ins * ins)
{
  t_cfg_edge * edge;
  t_bool x=TRUE;

  BBL_FOREACH_SUCC_EDGE(INS_BBL(ins),edge)
  {
    if( EDGE_CAT(edge) == ET_CALL)
//	EDGE_CAT(edge) == ET_IPJUMP)
    {
  //    DiabloPrint(stdout,"Corr Edge Killed:@E\n",edge->corr_edge);
      CfgEdgeKill(edge->corr_edge);
    }
    if( x && (EDGE_CAT(edge) == ET_FUNLINK))
    {
      x=FALSE;
      EDGE_CAT(edge) =ET_FALLTHROUGH;
    }
    else
      if(EDGE_CAT(edge) == ET_FALLTHROUGH ||
	  EDGE_CAT(edge) == ET_IPFALLTHRU)
      {
	x=FALSE;
      }
      else
      {
//	DiabloPrint(stdout,"Edge Killed:@E\n",edge);
	CfgEdgeKill(edge);
      }
  }
  if(x)
    FATAL(("Must be at least 1 FALLTHROUGH"));

}

void RemoveJumpEdges(t_ins * ins)
{
  t_cfg_edge * edge;

  BBL_FOREACH_SUCC_EDGE(INS_BBL(ins),edge)
  {
    if( EDGE_CAT(edge) == ET_JUMP || 
	EDGE_CAT(edge) == ET_IPJUMP)
    {
  //    DiabloPrint(stdout,"Edge Killed:@E\n",edge);
      CfgEdgeKill(edge);
    }
  }
}



void MakeUnConditional(t_ins * ins)
{
  t_cfg_edge * edge;

   INS_TYPE(ins)=IT_BRANCH;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=MIPS_J;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSRS(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSINSTR_INDEX(ins)=0;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins)=0;
   INS_ATTRIB(ins)|=IF_PCDEF;
   INS_ATTRIB(ins)|=IF_CHANGED; 

   BBL_FOREACH_SUCC_EDGE(INS_BBL(ins),edge)
   {
     if(EDGE_CAT(edge) == ET_FALLTHROUGH ||
	EDGE_CAT(edge) == ET_IPFALLTHRU)
     {
//       DiabloPrint(stdout,"Edge Killed:@E\n",edge);
       CfgEdgeKill(edge);
     }
   }

  INS_REGS_USE(ins) = MipsUsedRegisters(ins);
  INS_REGS_DEF(ins) = MipsDefinedRegisters(ins);
}


void MakeConstant(t_ins * ins,t_reg register1,t_uint32 waarde)
{
   INS_TYPE(ins)=IT_DATAPROC;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=MIPS_ADDIU;
   MIPSINSRT(ins)=register1;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSRS(ins)=MIPS_REG_R0;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSINSTR_INDEX(ins)=0;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins)=(waarde&0x0000ffff);
   INS_ATTRIB(ins)|=IF_CHANGED; 

  INS_REGS_USE(ins) = MipsUsedRegisters(ins);
  INS_REGS_DEF(ins) = MipsDefinedRegisters(ins);

}
	  
/* vim: set shiftwidth=2 foldmethod=marker: */
