/*
 * Copyright 2001,2002,2003{{{:
 * 			Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *}}}
 */
#include "mips_deflowgraph.h"
#include "mips_opcodes.h"
#include "mips_instructions.h"
#include "mips_registers.h"
#include "mips_layout.h"
#include "../../utils/util_error.h"
#include "../../utils/util_malloc.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_iterators.h"
#include <string.h>

t_int32 InsertNops(t_cfg * graph);
void MipsDeflowgraphSection(t_object * obj, t_section * code)
{
  t_bbl * bbl;
  t_uint32 size=0;
  t_uint32 tel=0;
  t_uint32 addr2=0;
  t_uint32 i=0;
  t_cfg * cfg = SECTION_CFG(code);

  tel=InsertNops(T_CFG(code->data));
  
  printf("Nops inserted: %d\n",tel);

  size=MipsCfgLayout(obj,code,T_CFG(code->data));
  SECTION_TMP_BUF(code)=Malloc(sizeof(t_mips_ins)*(size/4+2));
 
  /* this is temporary: free the refers_to and refed_by of bbls before killing the cfg */
  CFG_FOREACH_BBL(cfg,bbl)
  {
    t_ins * ins;
    
    BBL_FOREACH_INS(bbl,ins)
    {
      while (RELOCATABLE_REFED_BY(ins))
	RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFED_BY(ins)->rel);
      while (RELOCATABLE_REFERS_TO(ins))
	RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFERS_TO(ins)->rel);
    }
    
    while(RELOCATABLE_REFED_BY(bbl))
      RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFED_BY(bbl)->rel);

    while(RELOCATABLE_REFERS_TO(bbl))
      RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFERS_TO(bbl)->rel);
  }


  SymbolTableFree(obj->symbol_table);
  obj->symbol_table=NULL;

  obj->symbol_table=SymbolTableNew(obj);

  MipsDifferentInstructions2(code);

  CFG_FOREACH_BBL(T_CFG(code->data),bbl)
    {
      t_uint32 addr;
      t_mips_ins * current=(t_mips_ins *) bbl->ins_first;
      if (!current) continue;

      if(i==0)
      {
	addr2=G_T_UINT32(GENERIC_SUB(BBL_CADDRESS(bbl),SECTION_GET_OLD_ADDRESS(code)))/4;
	i++;
      }
      addr=G_T_UINT32(GENERIC_SUB(BBL_CADDRESS(bbl),SECTION_GET_OLD_ADDRESS(code)))/4-addr2;

    
      while(current)
	{
	  void * result;
	  tel++;
	  result=memcpy(&((t_mips_ins *) SECTION_TMP_BUF(code))[addr],current,sizeof(t_mips_ins));
	  addr++;
	  current=(t_mips_ins *) INS_NEXT(current);
	}
    }

  for (tel=0; tel<size/4-1; tel++)
  {
    INS_NEXT(&((t_mips_ins *) SECTION_TMP_BUF(code))[tel])=(t_ins *) (&((t_mips_ins *) SECTION_TMP_BUF(code))[tel+1]);
    INS_BBL(&((t_mips_ins *) SECTION_TMP_BUF(code))[tel]) = NULL;
  }

  printf("Done, wrote %d\n",tel);
  

  INS_NEXT(&((t_mips_ins *) SECTION_TMP_BUF(code))[size/4-1])=NULL;
  
  /* {{{ compute code size gain */
  {
    double gain, os, ns;
    int i;
    t_address osize, nsize;

    osize = nsize = NULL_GENERIC;
    for (i=0; i < obj->ncodes; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->code[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->code[i]));
    }
    DiabloPrint(stdout,"Code section old size @G new size @G\n",osize,nsize);
    os = (double) G_T_UINT32(osize);
    ns = (double) G_T_UINT32(nsize);
    gain = (os - ns)/os;
    printf("GAIN %lf%% in code section\n",100.0 * gain);
    for (i=0; i < obj->nrodatas; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->rodata[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->rodata[i]));
    }
    for (i=0; i < obj->ndatas; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->data[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->data[i]));
    }
    for (i=0; i < obj->nbsss; i++)
    {
      osize = GENERIC_ADD(osize,SECTION_OLD_SIZE(obj->bss[i]));
      nsize = GENERIC_ADD(nsize,SECTION_CSIZE(obj->bss[i]));
    }
    os = (double) G_T_UINT32(osize);
    ns = (double) G_T_UINT32(nsize);
    gain = (os - ns)/os;
    printf("Size O:%d\n",G_T_UINT32(osize));
    printf("Size N:%d\n",G_T_UINT32(nsize));
    printf("GAIN %lf%% total\n",100.0 * gain);
  } /* }}} */

  /* Sometimes crahes here: */
  CfgFree(SECTION_CFG(code));

  
  code->data=SECTION_TMP_BUF(code);
  code->type=DISASSEMBLED_CODE_SECTION;
  code->nins=(tel+1);
  
  printf("Counted till %d\n",tel);
}

t_int32 InsertNops(t_cfg * graph)
{
  t_ins * ins;
  t_bbl * bbl;
  t_int32 count=0;

  CFG_FOREACH_BBL(graph,bbl)
  {
    // Try to fill the delay slot with a usefull instruction
    if(bbl!=NULL)
      if(BBL_INS_LAST(bbl)!=NULL)
	if(INS_TYPE(BBL_INS_LAST(bbl))==IT_BRANCH)
	{
	  t_ins * prev_ins;
	  t_ins * last_ins;
	  t_bool ok;
	  last_ins=BBL_INS_LAST(bbl);
	  ok=FALSE;


	  if(INS_PREV(last_ins)!=NULL && INS_TYPE(INS_PREV(last_ins))!=IT_LOAD && INS_TYPE(INS_PREV(last_ins))!=IT_STORE)
	  {
	    prev_ins=INS_PREV(BBL_INS_LAST(bbl));
	    switch(MIPSINSOPCODE(last_ins))
	    {
	      case MIPS_J:
		InsSwap(prev_ins,last_ins);
		ok=TRUE;
		break;
	      case MIPS_JAL:
		if(
		    MIPSINSRT(prev_ins)!=MIPS_REG_RA&&
		    MIPSINSRS(prev_ins)!=MIPS_REG_RA&&
		    MIPSINSRD(prev_ins)!=MIPS_REG_RA
		  )
		  InsSwap(prev_ins,last_ins);
		ok=TRUE;
		break;
	      case MIPS_JR:
	      case MIPS_BLEZ:
	      case MIPS_BLTZ:
	      case MIPS_BGEZ:
	      case MIPS_BGTZ:
		if(
		    MIPSINSOPCODE(prev_ins)!=MIPS_MFHI && 
		    MIPSINSOPCODE(prev_ins)!=MIPS_MFLO &&
		    // Otherwise it will crash on some benchmarks (jpeg for example); but I don't see the dependency!
		    ((MIPSINSRS(last_ins)!=MIPSINSRD(prev_ins) && 
		      MIPSINSRS(last_ins)!=MIPSINSRT(prev_ins)
		     )
		     ||MIPSINSRS(last_ins)==MIPS_REG_R0))
		{
		  InsSwap(prev_ins,last_ins);
		  ok=TRUE;
		}
		break;
	      case MIPS_BLTZAL:
	      case MIPS_BGEZAL:
		if(
		    MIPSINSOPCODE(prev_ins)!=MIPS_MFHI && 
		    MIPSINSOPCODE(prev_ins)!=MIPS_MFLO &&
		    // Otherwise it will crash on some benchmarks (jpeg for example); but I don't see the dependency!
		    MIPSINSRT(prev_ins)!=MIPS_REG_RA&&
		    MIPSINSRS(prev_ins)!=MIPS_REG_RA&&
		    MIPSINSRD(prev_ins)!=MIPS_REG_RA&&
		    (
		     (MIPSINSRS(last_ins)!=MIPSINSRD(prev_ins) && 
		      MIPSINSRS(last_ins)!=MIPSINSRT(prev_ins)
		     )
		     ||MIPSINSRS(last_ins)==MIPS_REG_R0))
		{
		  InsSwap(prev_ins,last_ins);
		  ok=TRUE;
		}
		break;
	      case MIPS_JALR:
		if(
		    MIPSINSRS(last_ins)!=MIPSINSRD(prev_ins)&& 
		    MIPSINSRS(last_ins)!=MIPSINSRT(prev_ins)&&
		    MIPSINSRD(last_ins)!=MIPSINSRD(prev_ins)&& 
		    MIPSINSRD(last_ins)!=MIPSINSRT(prev_ins)&&
		    MIPSINSRT(prev_ins)!=MIPS_REG_RA&&
		    MIPSINSRS(prev_ins)!=MIPS_REG_RA&&
		    MIPSINSRD(prev_ins)!=MIPS_REG_RA)
		{
		  InsSwap(prev_ins,last_ins);
		  ok=TRUE;
		}
		else 
		  /* Construction:
		   * X r1,r2,r3
		   * MOVE r4,r5 r5<>r1 and r5<>r2 and r5<>r3
		   * JALR r4
		   * is equal to
		   * MOVE r5,r4
		   * JALR r4
		   * X r1,r2,r3 r1<>ra and r2<>ra and r3<>ra
		   * 
		   * In the next stage there will be no Nop-instruction added
		   * */
		  if(
		      INS_PREV(prev_ins)!=NULL&&
		      MIPSINSOPCODE(prev_ins)==MIPS_ADDU&&
		      MIPSINSRT(prev_ins)==MIPS_REG_R0 &&
		      MIPSINSRT(INS_PREV(prev_ins))!=MIPSINSRS(prev_ins)&&
		      MIPSINSRS(INS_PREV(prev_ins))!=MIPSINSRS(prev_ins)&&
		      MIPSINSRD(INS_PREV(prev_ins))!=MIPSINSRS(prev_ins)&&
		      MIPSINSRT(INS_PREV(prev_ins))!=MIPS_REG_RA&&
		      MIPSINSRS(INS_PREV(prev_ins))!=MIPS_REG_RA&&
		      MIPSINSRD(INS_PREV(prev_ins))!=MIPS_REG_RA
		    )
		  {
		    InsSwap(INS_PREV(prev_ins),prev_ins);
		    prev_ins=INS_PREV(BBL_INS_LAST(bbl));
		    InsSwap(prev_ins,last_ins);
		    ok=TRUE;
		  }
		break;
	      case MIPS_BEQ:
	      case MIPS_BNE:
		if((MIPSINSRS(last_ins)!=MIPSINSRD(prev_ins) && MIPSINSRS(last_ins)!=MIPSINSRT(prev_ins))||MIPSINSRS(last_ins)==MIPS_REG_R0)
		if((MIPSINSRT(last_ins)!=MIPSINSRD(prev_ins) && MIPSINSRT(last_ins)!=MIPSINSRT(prev_ins))||MIPSINSRT(last_ins)==MIPS_REG_R0)
		{
		  InsSwap(prev_ins,last_ins);
		  ok=TRUE;
		}
		break;
	      default:
		ok=FALSE;
		break;

	    }
	  }

	  if(ok==FALSE)
	  {
	    t_ins * nop_ins;
	    nop_ins= InsNewForBbl(bbl);
	    RELOCATABLE_REFERS_TO(nop_ins)=NULL;
	    INS_OLD_ADDRESS(nop_ins)=GENERIC_ADD_UINT32(INS_OLD_ADDRESS(BBL_INS_LAST(bbl)),4);
	    MipsInstructionMakeNoop(nop_ins);
	    InsInsertAfter(nop_ins,BBL_INS_LAST(bbl));
	    count++;
	  }
	}

    /*  Try to fill in the load delay-slot */
    BBL_FOREACH_INS(bbl,ins)
    {
      /* General method:
       * X r1,r2,r3
       * LW r4,0(r5)
       * Y r4
       * Try to become:
       * LW r4,0(r5)
       * X r1,r2,r3
       * Y r4
       *
       * 1st:
       * r1<>r4 and r2<>r4 and r3<>r4
       * r1<>r5 and r2<>r5 and r3<>r5
       * X <> IT_STORE
       */
      if(
	   INS_PREV(ins)!=NULL&&
	   INS_TYPE(INS_PREV(ins))!=IT_STORE&&
	  (
	  (INS_NEXT(ins)!=NULL&&
	   INS_TYPE(ins)==IT_LOAD&&
	   MIPSINSRT(ins)==MIPSINSRS(INS_NEXT(ins)))
	  ||
	  (INS_NEXT(ins)!=NULL&&
	   INS_TYPE(ins)==IT_LOAD&&
	   MIPSINSRT(ins)==MIPSINSRT(INS_NEXT(ins)))
	  ||
	  (INS_NEXT(ins)!=NULL&&
	   INS_TYPE(ins)==IT_LOAD&&
	   INS_TYPE(INS_NEXT(ins))==IT_STORE)
	  )
	  &&
	   MIPSINSRS(INS_PREV(ins))!=MIPSINSBASE(ins)&&
	   MIPSINSRT(INS_PREV(ins))!=MIPSINSBASE(ins)&&
	   MIPSINSRD(INS_PREV(ins))!=MIPSINSBASE(ins)&&
	   MIPSINSRS(INS_PREV(ins))!=MIPSINSRT(ins)&&
	   MIPSINSRT(INS_PREV(ins))!=MIPSINSRT(ins)&&
	   MIPSINSRD(INS_PREV(ins))!=MIPSINSRT(ins)
	)
      {
	InsChangePlace(INS_PREV(ins),ins);
	ins=INS_NEXT(ins);
      }
      else
	/* Construction:
	 * LW r1,0(r2)
	 * LW r3,0(r4) with: r2<>r3 and r1<>r4
	 * SW r3,0(r5) or SW r5,0(r3)
	 * is equal to
	 * LW r3,0(r4)
	 * LW r1,0(r2)
	 * SW r3,0(r5) or SW r5,0(r3)
	 * 
	 * In the next stage there will be no Nop-instruction added
	 * */
	if(
	  INS_PREV(ins)!=NULL&&
	  INS_NEXT(ins)!=NULL&&
	  INS_TYPE(INS_PREV(ins))==IT_LOAD&&
	  INS_TYPE(ins)==IT_LOAD&&
	  INS_TYPE(INS_NEXT(ins))==IT_STORE&&
	  MIPSINSRT(ins)!=MIPSINSBASE(INS_PREV(ins))&&
	  MIPSINSBASE(ins)!=MIPSINSRT(INS_PREV(ins))&&
	  (MIPSINSRT(ins)==MIPSINSBASE(INS_NEXT(ins)) || MIPSINSRT(ins)==MIPSINSRT(INS_NEXT(ins)))
	  )
      {
	InsChangePlace(INS_PREV(ins),ins);
	ins=INS_NEXT(ins);
      }
	
      else
	/* Construction:
	 * SW r1,X(r2)
	 * LW r3,Y(r2) with: r2<>r3 and r1<>r3
	 * SW r3,0(r5) or SW r5,0(r3)
	 * is equal to
	 * LW r3,Y(r2)
	 * SW r1,X(r2)
	 * SW r3,0(r5) or SW r5,0(r3)
	 * 
	 * In the next stage there will be no Nop-instruction added
	 * */
	if(
	  INS_PREV(ins)!=NULL&&
	  INS_NEXT(ins)!=NULL&&
	  MIPSINSBASE(ins)==MIPSINSBASE(INS_PREV(ins))&&
	  MIPSINSIMMEDIATE(ins)!=MIPSINSIMMEDIATE(INS_PREV(ins))&&
	  INS_TYPE(INS_PREV(ins))==IT_STORE&&
	  INS_TYPE(ins)==IT_LOAD&&
	  INS_TYPE(INS_NEXT(ins))==IT_STORE&&
	  MIPSINSRT(ins)!=MIPSINSBASE(INS_PREV(ins))&&
	  MIPSINSRT(ins)!=MIPSINSRT(INS_PREV(ins))&&
	  (MIPSINSRT(ins)==MIPSINSBASE(INS_NEXT(ins)) || MIPSINSRT(ins)==MIPSINSRT(INS_NEXT(ins)))
	  )
      {
	InsChangePlace(INS_PREV(ins),ins);
	ins=INS_NEXT(ins);
      }
      else 
	/* Construction:
	 * LW r1,0(r2)
	 * LW r3,0(r4) with: r2<>r3 and r1<>r4
	 * JR r3
	 * is equal to
	 * LW r3,0(r4)
	 * LW r1,0(r2)
	 * JR r3
	 * 
	 * In the next stage there will be no Nop-instruction added
	 * */
	if(
	  INS_PREV(ins)!=NULL&&
	  INS_NEXT(ins)!=NULL&&
	  INS_TYPE(INS_PREV(ins))==IT_LOAD&&
	  INS_TYPE(ins)==IT_LOAD&&
	  MIPSINSOPCODE(INS_NEXT(ins))==MIPS_JR&&
	  MIPSINSRT(ins)!=MIPSINSBASE(INS_PREV(ins))&&
	  MIPSINSBASE(ins)!=MIPSINSRT(INS_PREV(ins))&&
	  MIPSINSRT(ins)==MIPSINSRS(INS_NEXT(ins))
	  )
      {
	InsChangePlace(INS_PREV(ins),ins);
	ins=INS_NEXT(ins);
      }
	  
    }
   
    BBL_FOREACH_INS(bbl,ins)
    {
      t_ins * nop_ins;
      if(
	  (INS_NEXT(ins)!=NULL&&
	   INS_TYPE(ins)==IT_LOAD&&
	   MIPSINSRT(ins)==MIPSINSRS(INS_NEXT(ins)))
	  ||
	  (INS_NEXT(ins)!=NULL&&
	   INS_TYPE(ins)==IT_LOAD&&
	   MIPSINSRT(ins)==MIPSINSRT(INS_NEXT(ins)))
	  ||
	  (INS_NEXT(ins)!=NULL&&
	   MIPSINSOPCODE(ins)==MIPS_CFC1&&
	   MIPSINSRT(ins)==MIPSINSRS(INS_NEXT(ins)))
	  ||
	  (INS_NEXT(ins)!=NULL&&
	   MIPSINSOPCODE(ins)==MIPS_CFC1&&
	   MIPSINSRT(ins)==MIPSINSRT(INS_NEXT(ins)))
//	  ||
//	  (INS_NEXT(ins)!=NULL&&
//	   INS_TYPE(ins)==IT_LOAD&&
//	   INS_TYPE(INS_NEXT(ins))==IT_STORE)//&&
//	   MIPSINSRT(ins)==MIPSINSRT(INS_NEXT(ins)))
	  ||
	  /* No Nop-instruction needed: CPU stops untill data arrives
	  (INS_NEXT(ins)!=NULL&&
	   MIPSINSOPCODE(ins)==MIPS_C&&
	   (MIPSINSOPCODE(INS_NEXT(ins))==MIPS_BC1F||
	   MIPSINSOPCODE(INS_NEXT(ins))==MIPS_BC1F))
	  ||
	  (INS_NEXT(ins)!=NULL&&
	   MIPSINSOPCODE(ins)==MIPS_MTC1&&
	   MIPSINSOPCODE(INS_NEXT(ins))==MIPS_C)
	  ||*/
	  (INS_NEXT(ins)!=NULL&&
	   MIPSINSOPCODE(ins)==MIPS_MTC1&&
	   MIPSINSOPCODE(INS_NEXT(ins))==MIPS_SWC1&&
	   MIPSINSFS(ins)!=MIPSINSFT(INS_NEXT(ins)))
	  ||
	  (INS_NEXT(ins)!=NULL&&
	   (MIPSINSOPCODE(ins)==MIPS_MFLO||MIPSINSOPCODE(ins)==MIPS_MFHI)&&
	   (MIPSINSOPCODE(INS_NEXT(ins))==MIPS_MULTU||MIPSINSOPCODE(INS_NEXT(ins))==MIPS_MULT))
	  )
	  {
	    nop_ins= InsNewForBbl(bbl);
	    RELOCATABLE_REFERS_TO(nop_ins)=NULL;
	    INS_OLD_ADDRESS(nop_ins)=GENERIC_ADD_UINT32(INS_OLD_ADDRESS(ins),4);
	    MipsInstructionMakeNoop(nop_ins);
	    InsInsertAfter(nop_ins,ins);
	    count++;
	  }
      else
	if 
	  /* Attention:
	   *   00001812        mflo    v1
	   *   1000000a        b       412ed0 <$L4>
	   *   00430018        mult    v0,v1
	   * */
	(INS_NEXT(ins)!=NULL&&
	 INS_PREV(ins)!=NULL&&
	 (MIPSINSOPCODE(INS_PREV(ins))==MIPS_MFLO||MIPSINSOPCODE(INS_PREV(ins))==MIPS_MFHI)&&
	 (MIPSINSOPCODE(INS_NEXT(ins))==MIPS_MULTU||MIPSINSOPCODE(INS_NEXT(ins))==MIPS_MULT))
	  {
	    nop_ins= InsNewForBbl(bbl);
	    RELOCATABLE_REFERS_TO(nop_ins)=NULL;
	    INS_OLD_ADDRESS(nop_ins)=GENERIC_ADD_UINT32(INS_OLD_ADDRESS(ins),4);
	    MipsInstructionMakeNoop(nop_ins);
	    InsInsertBefore(nop_ins,ins);
	    count++;
	  }
    }
  }

  return count;
}

/* vim: set shiftwidth=2 foldmethod=marker: */
