/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mips_description.h"
#include "mips_instructions.h"
#include "mips_opt.h"
#include "mips_inline.h"
#include "../../kernel/diablo_opt_misc.h"

static char * names[] =    {  "r0",  "r1",  "r2",  "r3",  "r4",  "r5",  "r6",  "r7",
			      "r8",  "r9",  "r10",  "r11",  "r12",  "r13",  "r14",  "r15",
			      "r16",  "r17",  "r18",  "r19",  "r20",  "r21",  "r22",  "r23",
			      "r24",  "r25",  "r26",  "r27",  "r28",  "r29",  "r30",  "r31",
			      "sr",   "lo",   "hi",  "bad",  "cause", "pc",
			      "f0",  "f1",  "f2",  "f3",  "f4",  "f5",  "f6",  "f7",
			      "f8",  "f9",  "f10",  "f11",  "f12",  "f13",  "f14",  "f15",
			      "f16",  "f17",  "f18",  "f19",  "f20",  "f21",  "f22",  "f23",
			      "f24",  "f25",  "f26",  "f27",  "f28",  "f29",  "f30",  "f31",
			      "fsr",  "fir",  "fp",  "inx",  "rand",  "tlblo", "ctxt", "tlbhi", "epc",  "prid"
};

t_arch_specific_optimization mips_opts[]={
//  { "BranchForwarding", BranchForwarding}, \todo Insert BranchForwarding again
//  { "OptimizeStackLoadAndStores", OptimizeStackLoadAndStores},
  { "GeneralInlining", MipsGeneralInlining}, /* \todo Make a MipsGeneralInlining */
  { "InlineTrivial", MipsInlineTrivial},
//  { "OptCopyPropagation", OptCopyPropagation},
//  { "ArmLoadStoreFwd", ArmLoadStoreFwd},
  { "MergeBbls", MergeBbls}, 
  { "InstructieOptimalisatie", OptInstructions}, 
  { NULL,NULL}
} ;

const t_architecture_description mips_description =
{
   32, /* encoded minimal size */
   32, /* encoded maximum size */
   32, /* mod size */
   1, /* bundel size */
   0, /* bundel template size */
   sizeof(t_mips_ins),/* size of an instruction */
   ADDRSIZE32,
   32,  /* int regs */
   0,  /* float regs */
   0,  /* predicate regs */
   0,  /* branch regs */
   0,   /* special regs */
   names,
#if MAX_REG_ITERATOR > 64

#ifndef BIT64ADDRSUPPORT
   {MAX_REG_ITERATOR, {0xffff,0xffff,0xffff,0xffff,0xffff,0}},
   {MAX_REG_ITERATOR, {0xffff,0xffff,0x003f,0x0000,0xffc0,0}},
   {MAX_REG_ITERATOR, {0x0000,0x0000,0xffc0,0xffff,0x003f,0}},
   {MAX_REG_ITERATOR, {0}},
   {MAX_REG_ITERATOR, {0x0000,0x00ff,0x0000,0x0000,0x0000,0x0}},/* callee-saved */
   {MAX_REG_ITERATOR, {0x00f0,0xf200,0xffff,0xffff,0xffff,0x0}},/* callee may use */
   {MAX_REG_ITERATOR, {0xff0c,0x0100,0x0000,0x0000,0x0000,0x0}},/* callee may change */
   {MAX_REG_ITERATOR, {0x000c,0x0000,0x0000,0x0000,0x0000,0x0}},/* callee may contain return value */
   {MAX_REG_ITERATOR, {0}}, /* prop over hell */
   {MAX_REG_ITERATOR, {0}}, /* prop over hell */
   {MAX_REG_ITERATOR, {0}}, /* prop over hell */
   {MAX_REG_ITERATOR, {0}},/* Dead over call */
#else
   {MAX_REG_ITERATOR, {0xffffffffULL,0xffffffffULL,0x0000ffffULL,0x0ULL}},
   {MAX_REG_ITERATOR, {0xffffffffULL,0x0000003fULL,0x0000ffc0ULL,0x0ULL}},
   {MAX_REG_ITERATOR, {0x00000000ULL,0xffffffc0ULL,0x0000003fULL,0x0ULL}},
   {MAX_REG_ITERATOR, {0ULL}},
   {MAX_REG_ITERATOR, {0x00ff0000ULL,0x00000000ULL,0x00000000ULL,0x0ULL}},/* callee-saved */
//   {MAX_REG_ITERATOR, {0xf00000f0ULL,0xffffffffULL,0x0000ffffULL,0x0ULL}},/* callee may use */
   {MAX_REG_ITERATOR, {0xf20000f0ULL,0xffffffffULL,0x0000ffffULL,0x0ULL}},/* callee may use */
//   {MAX_REG_ITERATOR, {0x0300ff0cULL,0x00000000ULL,0x00000000ULL,0x0ULL}},/* callee may change */
   {MAX_REG_ITERATOR, {0x0100ff0cULL,0x00000000ULL,0x00000000ULL,0x0ULL}},/* callee may change */
   {MAX_REG_ITERATOR, {0x0000000cULL,0x00000000ULL,0x00000000ULL,0x0ULL}},/* callee may contain return value */
   {MAX_REG_ITERATOR, {0ULL}}, /* prop over hell */
   {MAX_REG_ITERATOR, {0ULL}}, /* prop over hell */
   {MAX_REG_ITERATOR, {0ULL}}, /* prop over hell */
   {MAX_REG_ITERATOR, {0ULL}}, /* Dead over call */
#endif

#elif MAX_REG_ITERATOR > 32
#error The regsets on the calling conventions are incorrect, see above
   0x7fffffLL,
   0xffffLL,
   0x0LL,
   0x7c0000LL,
   0x000000LL,/* callee-saved */
   0x0LL,
   0x0LL, /* Dead over call */
#else
#error The regsets on the calling conventions are incorrect, see above
   0x7fffff,
   0x00ffff,
   0x000000,
   0x7c0000, 
   0x000000,/* callee-saved */
   0x7fffff,/* callee may use */
   0x7fffff,/* callee may change */
   0x0,
   0x0, /* Dead over call */
#endif
   REG_NONE,
   mips_opts,
#ifdef INSTRUMENTATION
   NULL,
#endif
   NULL,   /* An array with names of unbehaved functions */
   NULL,
   MIPSInsIsProcedureCall,
#if MAX_REG_ITERATOR > 64

#ifndef BIT64ADDRSUPPORT
   {MAX_REG_ITERATOR, {0x0000,0x8000,0x0}},/* return address */
#else
   {MAX_REG_ITERATOR, {0x80000000ULL,0x0ULL}},/* return address */
#endif
#elif MAX_REG_ITERATOR > 32
#error The regsets on the calling conventions are incorrect, see above
#else
#error The regsets on the calling conventions are incorrect, see above
#endif
   NULL
};
/* vim: set shiftwidth=2: */
