/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "mips_disassemble.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_ins.h"
#include "mips_disassemble_one.h"
#include "mips_instructions.h"
#include "mips_opcodes.h"
#include "mips_utils.h"
#include "../../utils/util_error.h"
#include "../../utils/util_malloc.h"

void * MipsDisassembleOneInstruction(t_object * obj, t_address start, int * size_ret)
{
  FATAL(("Implement"));
  return NULL; /* Keep the compiler happy */
}

void * MipsDisassembleOneInstructionForSection(t_section * sec, int offset, int * size_ret)
{
  t_ins * ret=NULL;
  t_uint32 encoded_instruction;
  t_uint16 i;
  t_address goffset;

  GenericSet_uint32(goffset,offset);
  if (G_T_UINT32(SECTION_OLD_SIZE(sec))>offset)
  {
    ret=Malloc(sizeof(t_mips_ins));
    encoded_instruction=SectionGetData32(sec,goffset);
    /* Find the opcode in the opcode table */
    for (i = 0; (encoded_instruction & mips_opcode_table[i].mask) != mips_opcode_table[i].opcode; i++);
    mips_opcode_table[i].Disassemble(ret,encoded_instruction,i);

    /* Initialize the other fields */
    INS_PREV(ret) = INS_NEXT(ret) = NULL; 
    INS_REGS_USE(ret)=RegsetNewDup(MipsUsedRegisters(ret));
    INS_REGS_DEF(ret)=RegsetNewDup(MipsDefinedRegisters(ret));
  }
  return ret;
}

void MipsDisassembleSection(t_object * obj, t_section * code)
{
  t_address offset;
  t_uint16 i;
  t_ins * ins_s;
  t_ins * prev;
  t_uint32 data;
  t_int32 cpu,cop0,cop1;

  cpu=0;cop0=0;cop1=0;

  GenericSet_uint32(offset,0);

  for ( prev = NULL; GENERIC_LT(offset, SECTION_CSIZE(code)); offset=GENERIC_ADD_UINT32(offset,4) , prev=ins_s)
  {
    ins_s = InsNewForSec(code);
    data=SectionGetData32(code,offset);
    if (!ins_s) FATAL(("No instructions!"));
    RELOCATABLE_REFERS_TO(ins_s)=NULL;  /* TODO: Check if this is necessary */
    RELOCATABLE_REFED_BY(ins_s)=NULL; /* TODO: Check if this is necessary */


    /* Use the symbol table to find out if this is data or code */
    if (SymbolTableGetDataType(obj, GENERIC_ADD(SECTION_GET_OLD_ADDRESS(code),offset)) == ADDRESS_IS_DATA)  
      MipsDisassembleData(ins_s,(data),MIPS_DATA);
    else
    {
      for (i = 0; ((data) & mips_opcode_table[i].mask) != mips_opcode_table[i].opcode; i++);
      mips_opcode_table[i].Disassemble(ins_s,(data),i);
    //DiabloPrint(stdout,"@I, %d \n",ins_s,MIPSINSPROCESSOR(ins_s));
    }

    INS_PREV(ins_s) = T_INS(prev);
    INS_REGS_USE(ins_s)=RegsetNewDup(MipsUsedRegisters(ins_s));
    INS_REGS_DEF(ins_s)=RegsetNewDup(MipsDefinedRegisters(ins_s));
    INS_OLD_ADDRESS(ins_s)=GENERIC_ADD(SECTION_GET_OLD_ADDRESS(code),offset);
    INS_CADDRESS(ins_s)=GENERIC_ADD(SECTION_GET_OLD_ADDRESS(code),offset);
    GenericSet_uint32(INS_CSIZE(ins_s),4);
    GenericSet_uint32(INS_OLD_SIZE(ins_s),4);

   // DiabloPrint(stdout,"@I, %d \n",ins_s,MIPSINSPROCESSOR(ins_s));
    if (prev) INS_NEXT(prev) = ins_s;
  }
  INS_NEXT(ins_s) = NULL;

    
  DiabloPrint(stdout,"Disassembled %d instructions\n",G_T_UINT32(SECTION_OLD_SIZE(code))/4);
}
/* vim: set shiftwidth=2: */
