/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../utils/util_error.h"
#include "../../utils/util_bitmanipulation.h"
#include "mips_disassemble_one.h"
#include "mips_instructions.h"
#include "mips_registers.h"
#include "mips_opcodes.h"

/* Condition Codes */
#define MIPS_EXTRACT_CONDITION_CODE(i) (((i) & 0xf0000000) >> 28)
/* Utility Function */

void MipsDisassemble(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_UNKNOWN;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=MIPS_REG_NONE;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins) =0;
   MIPSINSINSTR_INDEX(ins)=0;
}
void MipsDisassembleBreak(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_SWI;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSIMMEDIATE(ins) =(instr & 0x03ffffc0) >> 6;
}
void MipsDisassembleData(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATA;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=MIPS_DATA;
   MIPSINSRS(ins)=MIPS_REG_NONE;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins) =instr;
   MIPSINSINSTR_INDEX(ins)=0;
}
void MipsDisassembleUnsupported(t_ins * ins, t_uint32 instr, t_uint16 opc ) 
{
   FATAL(("Unsupported MIPS instruction. Opcode = %x\n",instr));
}

void MipsDisassembleNOP(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_NOP;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=MIPS_REG_NONE;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins) =0;
   MIPSINSINSTR_INDEX(ins)=0;
//   DiabloPrint(stdout,"%d\n",MIPSINSPROCESSOR(ins));
}
void MipsDisassembleDataProcessingI(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins)=(instr & 0x0000ffff);
   MIPSINSINSTR_INDEX(ins)=0;
   if(opc==MIPS_LUI)
     MIPSINSRS(ins)=MIPS_REG_NONE;
}
void MipsDisassembleDataProcessing(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=(instr & 0x0000003f);
   MIPSINSIMMEDIATE(ins)=0;
   MIPSINSINSTR_INDEX(ins)=0;
}
void MipsDisassembleShift(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
   if(( (instr & 0x03e00000) >> 21) == 0 ) 
   {
     MIPSINSRS(ins)=MIPS_REG_NONE;
     MIPSINSSA(ins)=(instr & 0x000007c0) >> 6;
   }
   else 
   {
     MIPSINSSA(ins)=MIPS_REG_NONE;
     MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   }
   MIPSINSFUNCTION(ins)=(instr & 0x0000003f);
   MIPSINSIMMEDIATE(ins)=0;
   MIPSINSINSTR_INDEX(ins)=0;
}
void MipsDisassembleMultiply(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_MUL;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=(instr & 0x0000003f);
   MIPSINSIMMEDIATE(ins)=0;
   MIPSINSINSTR_INDEX(ins)=0;
   switch(opc)
   {
     case MIPS_MFHI:
     case MIPS_MFLO:
       MIPSINSRS(ins)=MIPS_REG_NONE;
       MIPSINSRT(ins)=MIPS_REG_NONE;
       MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
       break;
     case MIPS_MTHI:
     case MIPS_MTLO:
       MIPSINSRT(ins)=MIPS_REG_NONE;
       break;
     case MIPS_MUL:
       MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
       break;
   }
}
void MipsDisassembleUnconditionalJump(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_BRANCH;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSINSTR_INDEX(ins)=0;
   /* default is JALR, now undo if it's not default */
   switch(opc)
   {
     case MIPS_JR:
       MIPSINSRD(ins)=MIPS_REG_NONE;
       break;
     case MIPS_J:
     case MIPS_JAL:
       MIPSINSRS(ins)=MIPS_REG_NONE;
       MIPSINSRD(ins)=MIPS_REG_NONE;
       MIPSINSINSTR_INDEX(ins)=(instr & 0x03ffffff);
       break;
   }
   MIPSINSFUNCTION(ins)=(instr & 0x0000003f);
   MIPSINSIMMEDIATE(ins)=0;
   INS_ATTRIB(ins)|=IF_PCDEF;
}
void MipsDisassembleBranch(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_BRANCH;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins)=(instr & 0x0000ffff);
   MIPSINSINSTR_INDEX(ins)=0;
   if(opc==MIPS_BEQ||opc==MIPS_BNE)
     MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   INS_ATTRIB(ins)|=IF_PCDEF;
}
void MipsDisassembleTrapR(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_BRANCH;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=(instr & 0x0000003f);
   MIPSINSIMMEDIATE(ins)=0;
   MIPSINSINSTR_INDEX(ins)=0;
}
void MipsDisassembleTrapI(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_BRANCH;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=0;
   MIPSINSIMMEDIATE(ins)=(instr & 0x0000ffff);
   MIPSINSINSTR_INDEX(ins)=0;
}
void MipsDisassembleConditionalMove(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSFUNCTION(ins)=(instr & 0x0000003f);
   MIPSINSIMMEDIATE(ins)=0;
   MIPSINSINSTR_INDEX(ins)=0;
}
void MipsDisassembleDataTransferS(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_STORE;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSBASE(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSOFFSET(ins)=(instr & 0x0000ffff);
}
void MipsDisassembleDataTransferL(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_LOAD;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=CPU;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSBASE(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSA(ins)=MIPS_REG_NONE;
   MIPSINSOFFSET(ins)=(instr & 0x0000ffff);
}

/* COP0 */

void MipsDisassembleCOP0DataProcessing(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP0;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
   MIPSINSSEL(ins)=(instr & 0x00000007);
}
void MipsDisassembleCOP0(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   FATAL(("Unsupported MIPS instruction for COP0. Opcode = %d\n"));
   INS_TYPE(ins)=IT_UNKNOWN;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP0;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRT(ins)=MIPS_REG_NONE;
   MIPSINSRD(ins)=MIPS_REG_NONE;
   MIPSINSSEL(ins)=0;
}

/* COP1 */

void MipsDisassembleCOP1DataTransferS(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_STORE;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSBASE(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSFT(ins)=((instr & 0x001f0000) >> 16)+MIPS_REG_F0;
   MIPSINSOFFSET(ins)=(instr & 0x0000ffff);
}
void MipsDisassembleCOP1DataTransferL(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_LOAD;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSBASE(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSFT(ins)=((instr & 0x001f0000) >> 16)+MIPS_REG_F0;
   MIPSINSOFFSET(ins)=(instr & 0x0000ffff);
}
void MipsDisassembleCOP1DataProcessingFT(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSFS(ins)=((instr & 0x0000f800) >> 11)+MIPS_REG_F0;
}
void MipsDisassembleCOP1DataProcessing(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSFMT(ins)=(instr & 0x03e00000) >> 21;
   if(opc!=MIPS_ABS&&opc!=MIPS_NEG&&opc!=MIPS_SQRT)
     MIPSINSFT(ins)=((instr & 0x001f0000) >> 16)+MIPS_REG_F0;
   MIPSINSFS(ins)=((instr & 0x0000f800) >> 11)+MIPS_REG_F0;
   if(opc!=MIPS_C)
   {
     MIPSINSFD(ins)=((instr & 0x000007c0) >> 6)+MIPS_REG_F0;
     MIPSINSCC(ins)=MIPS_REG_NONE;
   }
   else
   {
     MIPSINSCC(ins)=(instr & 0x00000700) >> 8;
     MIPSINSCOND(ins)=(instr & 0x0000000f);
   }
}
void MipsDisassembleCOP1DataProcessingRound(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSFMT(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSFS(ins)=((instr & 0x0000f800) >> 11)+MIPS_REG_F0;
   MIPSINSFD(ins)=((instr & 0x000007c0) >> 6)+MIPS_REG_F0;
}
void MipsDisassembleCOP1DataProcessingMOV(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSFMT(ins)=(instr & 0x03e00000) >> 21;
   if(opc==MIPS_MOVF_||opc==MIPS_MOVT_)
     MIPSINSCC(ins)=(instr & 0x001c0000) >> 18;
   else
     MIPSINSRT(ins)=(instr & 0x001f0000) >> 16;
   MIPSINSFS(ins)=((instr & 0x0000f800) >> 11)+MIPS_REG_F0;
   MIPSINSFD(ins)=((instr & 0x000007c0) >> 6)+MIPS_REG_F0;
}
void MipsDisassembleCOP1Branch(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_BRANCH;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSCC(ins)=(instr & 0x001c0000) >> 18;
   MIPSINSOFFSET(ins)=(instr & 0x0000ffff);
   INS_ATTRIB(ins)|=IF_PCDEF;
}
void MipsDisassembleCOP1DataProcessingMoveTF(t_ins * ins,t_uint32 instr, t_uint16 opc ) 
{
   INS_TYPE(ins)=IT_DATAPROC;
   INS_ATTRIB(ins)=0x0;
   MIPSINSPROCESSOR(ins)=COP1;
   MIPSINSOPCODE(ins)=opc;
   MIPSINSRS(ins)=(instr & 0x03e00000) >> 21;
   MIPSINSCC(ins)=(instr & 0x001c0000) >> 18;
   MIPSINSRD(ins)=(instr & 0x0000f800) >> 11;
}
/* vim: set shiftwidth=2: */
