/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <string.h>
#include "mips_emulation.h"
#include "mips_registers.h"
#include "mips_opcodes.h"
#include "mips_instructions.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_addresses.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_procstate.h"

//static int tellertje=0;

void MipsInsEmulDataproc(t_ins * ins, t_procstate * procstate);
void MipsInsEmulLoad(t_ins * ins, t_procstate * procstate);
void MipsInsEmulStore(t_ins * ins, t_procstate * procstate);
void MipsInsEmulBranch(t_ins * ins, t_procstate * procstate);
void MipsInsEmulMul(t_ins * ins, t_procstate * procstate);
void MipsInsEmulSWI(t_ins * ins, t_procstate * procstate);
void MipsInstructionEmulator(t_ins * ins, t_procstate * procstate)
{

  register_content dest;
  INS_ATTRIB(ins) |= IF_EXECED;

  GenericSet_uint32(dest.i,(t_uint32)0);
#ifdef EXTENDED_CONSTPROP
  ProcStateSetRegZero(procstate,0,dest);

//  DiabloPrint(stdout,"@I\n",ins);
//  ProcStateGetRegBotExtend(procstate,MIPS_REG_A2,&plus1,&zero1,&minus1);
//  DiabloPrint(stdout,"minus:%d zero:%d plus:%d\n",minus1,zero1,plus1);
//  ProcStateGetRegBotExtend(procstate,MIPS_BRANCH_REG1,&plus1,&zero1,&minus1);
//  DiabloPrint(stdout,"minus:%d zero:%d plus:%d\n",minus1,zero1,plus1);
#else
  ProcStateSetReg(procstate,0,dest);
#endif
  
  switch (INS_TYPE(ins))
  {
    case IT_LOAD:
      MipsInsEmulLoad(ins, procstate);
      break;
    case IT_STORE:
      MipsInsEmulStore(ins, procstate);
      break;
    case IT_DATAPROC:
      MipsInsEmulDataproc(ins, procstate);
      break;
    case IT_BRANCH:
      MipsInsEmulBranch(ins, procstate);
      break;
    case IT_MUL:
      MipsInsEmulMul(ins, procstate);
      break;
    case IT_SWI:
      MipsInsEmulSWI(ins, procstate);
      break;
    case IT_NOP:
      break;
    default:
        ProcStateSetRegsetBot(procstate, INS_REGS_DEF(ins));
      FATAL(("Can not be emulated: instrcutie:@I INS-TYPE:%d\n",ins,INS_TYPE(ins)));
      break;
  }
}
/* De link invullen en anders niets doen, denk ik */
void MipsInsEmulBranch(t_ins * ins, t_procstate * procstate)
{
  register_content dest,value1,value2;
  t_reg register1,register2;
  t_lattice_level reg1,reg2;
  //int i,j;
  
  switch(MIPSINSOPCODE(ins))
  {
    case MIPS_BLTZAL:
    case MIPS_BGEZAL:
    case MIPS_BLTZ:
    case MIPS_BLEZ:
    case MIPS_BGEZ:
    case MIPS_BGTZ:
#ifdef EXTENDED_CONSTPROP
      ProcStateRegDup(procstate,MIPSINSRS(ins),MIPS_BRANCH_REG1);
#else
      register1=MIPSINSRS(ins);
      reg1=ProcStateGetReg(procstate,register1,&value1);
      if(reg1==BOT)
	ProcStateSetRegBot(procstate,MIPS_BRANCH_REG1);
      else if(reg1==TOP)
	ProcStateSetRegTop(procstate,MIPS_BRANCH_REG1);
      else if(reg1==VALUE)
	ProcStateSetReg(procstate,MIPS_BRANCH_REG1,value1);
#endif
      break;
    case MIPS_BEQ:
    case MIPS_BNE:
#ifdef EXTENDED_CONSTPROP
      ProcStateRegDup(procstate,MIPSINSRS(ins),MIPS_BRANCH_REG1);
      ProcStateRegDup(procstate,MIPSINSRT(ins),MIPS_BRANCH_REG2);
#else
      register1=MIPSINSRS(ins);
      register2=MIPSINSRT(ins);
      reg1=ProcStateGetReg(procstate,register1,&value1);
      reg2=ProcStateGetReg(procstate,register2,&value2);
      if(reg1==BOT)
	ProcStateSetRegBot(procstate,MIPS_BRANCH_REG1);
      else if(reg1==TOP)
	ProcStateSetRegTop(procstate,MIPS_BRANCH_REG1);
      else if(reg1==VALUE)
	ProcStateSetReg(procstate,MIPS_BRANCH_REG1,value1);
      if(reg2==BOT)
	ProcStateSetRegBot(procstate,MIPS_BRANCH_REG2);
      if(reg2==TOP)
	ProcStateSetRegTop(procstate,MIPS_BRANCH_REG2);
      else if(reg2==VALUE)
	ProcStateSetReg(procstate,MIPS_BRANCH_REG2,value2);
#endif
      break;
    case MIPS_JR:
#ifdef EXTENDED_CONSTPROP
      ProcStateRegDup(procstate,MIPSINSRS(ins),MIPS_BRANCH_REG1);
      break;
#endif
    case MIPS_JAL:
    case MIPS_JALR:
    case MIPS_J:
    case MIPS_BC1F:
    case MIPS_BC1T:
      break;
    default:
      FATAL(("FATAAL JA; %d ",MIPSINSOPCODE(ins)));
      break;
  }
    
  switch(MIPSINSOPCODE(ins))
  {
    case MIPS_JALR:
      bzero(&dest,sizeof(register_content));
      dest.i=GENERIC_ADD_UINT32(INS_OLD_ADDRESS(ins),8);
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
      break;
    case MIPS_JAL:
    case MIPS_BLTZAL:
    case MIPS_BGEZAL:
      bzero(&dest,sizeof(register_content));
      dest.i=GENERIC_ADD_UINT32(INS_OLD_ADDRESS(ins),8);
      ProcStateSetReg(procstate,MIPS_REG_R31,dest);
      break;
  }
}

void MipsInsEmulLoad(t_ins * ins, t_procstate * procstate)
{
  t_reloc * base_reloc;
  t_lattice_level base_address_unknown;
  t_lattice_level offset_unknown;
  t_lattice_level base_reloc_unknown;
  register_content dest,tmp_base_address,tmp_offset;
  t_uint32 base_address,offset;
  
  base_reloc_unknown = ProcStateGetTag(procstate,MIPSINSBASE(ins),&base_reloc); //BOT,VALUE,TOP
  base_address_unknown = ProcStateGetReg(procstate,MIPSINSBASE(ins),&tmp_base_address);
  base_address = G_T_UINT32(tmp_base_address.i);

  offset_unknown=VALUE;
  GenericSet_uint32(tmp_offset.i,(t_int32)((t_int16)MIPSINSIMMEDIATE(ins)));

  offset=G_T_UINT32(tmp_offset.i);

  if(base_address_unknown==BOT)
  {
    ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
    ProcStateSetTagsetBot(procstate,INS_REGS_DEF(ins));
    return;
  }

  bzero(&dest,sizeof(register_content));

  if (!base_reloc_unknown && !base_address_unknown && !offset_unknown)
  {
    if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(base_reloc))==RT_SUBSECTION)
    {
      t_section *sec = T_SECTION(REL_TO_RELOCATABLE(base_reloc));
      if (sec->type != RO_DATA_SECTION)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	ProcStateSetTagsetBot(procstate,INS_REGS_DEF(ins));
      }
      else
      {
	t_address tmp_addr;
	t_uint32 tmp = base_address + (t_int32) offset;
	GenericSet_uint32(tmp_addr,tmp);
	offset = G_T_UINT32(GENERIC_SUB(tmp_addr,SECTION_GET_OLD_ADDRESS(sec)));
	if (offset > G_T_UINT32(SECTION_OLD_SIZE(sec))) 
	{
	  ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	  ProcStateSetTagsetBot(procstate,INS_REGS_DEF(ins));
	} 
	else 
	{
	  t_uint32 offset2 = G_T_UINT32(GENERIC_SUB(tmp_addr,SECTION_GET_OLD_ADDRESS(SECTION_PARENT_SECTION(sec))));
	  t_uint32 loaded;
	  loaded = *(((t_uint32 *) SECTION_PARENT_SECTION(sec)->data) + offset2/4);
	  if(MIPSINSOPCODE(ins)==MIPS_LB)
	    loaded&=0x000000ff;
	  if(MIPSINSOPCODE(ins)==MIPS_LH)
	    loaded&=0x0000ffff;
	  GenericSet_uint32( dest.i, loaded);
#ifdef EXTENDED_CONSTPROP
	  if(loaded==0)
	    ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
	  else if(((t_int32)loaded)>0)
	  {
	    ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
	  }
	  else if(((t_int32)loaded)<0)
	    ProcStateSetRegMinus(procstate,MIPSINSRT(ins),dest);
	  else
	    FATAL(("FOUT"));
#else
	  ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
	}
      }
    }
    else if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(base_reloc))==RT_BBL)
    {
      t_bbl * bbl;
      t_address tmp_addr;
      t_uint32 tmp = base_address + (t_int16) offset;
      bbl=T_BBL(REL_TO_RELOCATABLE(base_reloc));
      GenericSet_uint32(tmp_addr,tmp);
      offset = G_T_UINT32(GENERIC_SUB(tmp_addr,BBL_OLD_ADDRESS(bbl)));
      if (offset & 3) 
      {
        ProcStateSetRegsetBot(procstate, INS_REGS_DEF(ins));
	ProcStateSetTagsetBot(procstate, INS_REGS_DEF(ins));
      }
      if ((offset < 0) || (offset > 4*bbl->nins))
      {
	ProcStateSetRegsetBot(procstate, INS_REGS_DEF(ins));
	ProcStateSetTagsetBot(procstate, INS_REGS_DEF(ins));
      }
      else
      {
	int i;
	t_ins * i_ins;
	t_uint32 loaded;
	i_ins = bbl->ins_first;
	for (i=0; i < offset; i+=4)
	  i_ins = INS_NEXT(i_ins);
	loaded = MIPSINSIMMEDIATE(i_ins);
	GenericSet_uint32(dest.i, loaded);
	if (RELOCATABLE_REFERS_TO(ins))
	  ProcStateSetTag(procstate, MIPSINSRT(ins), RELOCATABLE_REFERS_TO(ins)->rel);
#ifdef EXTENDED_CONSTPROP
	  if(loaded==0)
	    ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
	  else if(((t_int32)loaded)>0)
	  {
	    ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
	  }
	  else if(((t_int32)loaded)<0)
	    ProcStateSetRegMinus(procstate,MIPSINSRT(ins),dest);
	  else
	    FATAL(("FOUT"));
#else
	ProcStateSetReg(procstate, MIPSINSRT(ins), dest);
#endif
      }
    }
    else
    {
      ProcStateSetRegsetBot(procstate, INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(procstate, INS_REGS_DEF(ins));
    }
    
    if (base_address_unknown || offset_unknown)
    {
      ProcStateSetRegsetBot(procstate, INS_REGS_DEF(ins));
      ProcStateSetTagsetBot(procstate, INS_REGS_DEF(ins));
    }
  }
}

void MipsInsEmulStore(t_ins * ins, t_procstate * state)
{
  t_reloc * base_reloc;
  register_content tmp_base_address, tmp_offset;// dest;
  t_uint32 base_address,offset;
  t_lattice_level base_address_unknown;
  t_lattice_level offset_unknown;
  t_lattice_level base_reloc_unknown;
  register_content tmp_src;

  /* get the base address and reloc from the state */
  
  base_reloc_unknown = ProcStateGetTag(state,MIPSINSBASE(ins),&base_reloc);
  base_address_unknown = ProcStateGetReg(state,MIPSINSBASE(ins),&tmp_base_address);
  base_address = G_T_UINT32(tmp_base_address.i);
  
  
  /* get the offset from the state */

  offset_unknown = VALUE;
  GenericSet_uint32(tmp_offset.i,(t_int32)((t_int16)MIPSINSIMMEDIATE(ins)));
  
  offset = G_T_UINT32(tmp_offset.i);
  
  if(ProcStateGetReg(state,MIPSINSRT(ins),&tmp_src)!=0)
  {
    ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
    return;
  }
  
  if (TOP==base_reloc_unknown)
    {
 //     FATAL(("problem: no reloc in base register (ins: @I)\n",ins));
    }
  
  if (base_address_unknown == BOT)
    {
      ProcStateSetRegsetBot(state,INS_REGS_DEF(ins));
      return;
    }
}



void MipsInsEmulDataproc(t_ins * ins, t_procstate * procstate)
{
#ifdef EXTENDED_CONSTPROP
  t_lattice_level level1,level2;
  t_bool plus1,zero1,minus1;
  t_bool plus2,zero2,minus2;
#endif
  register_content tmp_src1,tmp_src2, dest;
  t_uint32 src1,src2;
  t_int32 temp32;
  t_uint32 u_temp32,u_temp32b;

  switch (MIPSINSOPCODE(ins))
  {
    //difference: ADD overflow-control; ADDU no overflow control
    //Voorlopig nog geen rekening mee gehouden
    //CORRECT
    case MIPS_ADD:
    case MIPS_ADDU:
//      DiabloPrint(stdout,"@I\n",ins);
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE||level2!=VALUE)
      {
	if(minus1 && !zero1 && !plus1)
	  if(minus2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	  else if(!minus2 && zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(minus1 && zero1 && !plus1)
	  if(minus2 && zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,TRUE,TRUE);
	  else if(minus2 && !zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && !zero1 && plus1)
	  if(!minus2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,FALSE,FALSE);
	  else if(!minus2 && zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,FALSE,FALSE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && zero1 && plus1)
	  if(!minus2 && !zero2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,FALSE,FALSE);
	  else if(!minus2 && zero2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && zero1 && !plus1)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),plus2,zero2,minus2);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_int32)src1;
      temp32+=(t_int32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      //DiabloPrint(stdout,"ADDU: arg1:0x%08x, arg2:0x%08x, result:0x%08x\n",src1,src2,dest.i);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      if (RELOCATABLE_REFERS_TO(ins))
	ProcStateSetTag(procstate,MIPSINSRT(ins),RELOCATABLE_REFERS_TO(ins)->rel);
      break;
      
/*    case MIPS_SUB_:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSFT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      f_src1=FP_VALUE(tmp_src1.f);
      f_src2=tmp_src2.f.value;
      bzero(&dest,sizeof(register_content));
      f_temp=f_src1-f_src2;
      FP_VALUE(dest.f)=f_temp;
      FP_FLAGS(dest.f)=FP_FLAGS(tmp_src1.f);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
      //src1=G_T_UINT32(tmp_src1.i);
      //src2=G_T_UINT32(tmp_src2.i);
      //temp32=(t_int32)src1;
      //temp32-=(t_int32)src2;
      //GenericSet_uint32(dest.i,(t_uint32)temp32);
     // DiabloPrint(stdout,"ADD_: %d, %d, %d",src1,src2,temp32);
      break;
      
    case MIPS_ADD_:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSFT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_int32)src1;
      temp32+=(t_int32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
     // DiabloPrint(stdout,"ADD_: %d, %d, %d",src1,src2,temp32);
      break;
      
    case MIPS_MOV:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_int32)src1;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
     // DiabloPrint(stdout,"MOV: %d, %d",src1,temp32);
      break;
 */     
      //MIPS_ADDIU CORRECT
    case MIPS_ADDIU:
#ifdef EXTENDED_CONSTPROP
      //DiabloPrint(stdout,"@I\n",ins);
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      
      GenericSet_uint32(tmp_src2.i,(t_int32)(t_int16)MIPSINSIMMEDIATE(ins));
      s_src2=(t_int32)G_T_UINT32(tmp_src2.i);
      if(s_src2>0)
	plus2=TRUE;
      else
	plus2=FALSE;
      if(s_src2<0)
	minus2=TRUE;
      else
	minus2=FALSE;
      if(s_src2==0)
	zero2=TRUE;
      else
	zero2=FALSE;
      
      if(level1!=VALUE)
      {
	if(minus1 && !zero1 && !plus1)
	  if(minus2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),FALSE,FALSE,TRUE);
	  else if(!minus2 && zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),FALSE,FALSE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	else if(minus1 && zero1 && !plus1)
	  if(minus2 && zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),FALSE,TRUE,TRUE);
	  else if(minus2 && !zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),FALSE,FALSE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && !zero1 && plus1)
	  if(!minus2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,FALSE,FALSE);
	  else if(!minus2 && zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,FALSE,FALSE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && zero1 && plus1)
	  if(!minus2 && !zero2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,FALSE,FALSE);
	  else if(!minus2 && zero2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,FALSE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && zero1 && !plus1)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),plus2,zero2,minus2);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      GenericSet_uint32(tmp_src2.i,(t_int32)(t_int16)MIPSINSIMMEDIATE(ins));
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_int32)src1;
      temp32+=(t_int32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      //DiabloPrint(stdout,"ADDIU: arg1:0x%08x, arg2:0x%08x, result:0x%08x\n",src1,src2,dest.i);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRT(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
      if (RELOCATABLE_REFERS_TO(ins))
	ProcStateSetTag(procstate,MIPSINSRT(ins),RELOCATABLE_REFERS_TO(ins)->rel);
      break;
      
      //CORRECT
   case MIPS_AND:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE||level2!=VALUE)
      {
	if(!minus1 || !minus2)
	{
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	}
	else if((!zero1 &&!plus1)&&(!zero2 && !plus2))
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_uint32)src1;
      temp32&=(t_uint32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;
      
      //CORRECT
    case MIPS_ANDI:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      
      GenericSet_uint32(tmp_src2.i,(t_int32)(t_int16)MIPSINSIMMEDIATE(ins));
      s_src2=(t_int32)G_T_UINT32(tmp_src2.i);
      if(s_src2>0)
	plus2=TRUE;
      else
	plus2=FALSE;
      if(s_src2<0)
	minus2=TRUE;
      else
	minus2=FALSE;
      if(s_src2==0)
	zero2=TRUE;
      else
	zero2=FALSE;
      if(level1!=VALUE)
      {
	if(!minus1 || !minus2)
	{
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,FALSE);
	}
	else if((!zero1 &&!plus1)&&(!zero2 && !plus2))
	{
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),FALSE,FALSE,TRUE);
	}
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      GenericSet_uint32(tmp_src2.i,(t_uint16)MIPSINSIMMEDIATE(ins));
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_uint32)src1;
      temp32&=(t_uint32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRT(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
      if (RELOCATABLE_REFERS_TO(ins))
	ProcStateSetTag(procstate,MIPSINSRT(ins),RELOCATABLE_REFERS_TO(ins)->rel);
      break;
      
      //CORRECT
    case MIPS_LUI:
      GenericSet_uint32(tmp_src1.i,(MIPSINSIMMEDIATE(ins)<<16));
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_uint32)src1;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
     // DiabloPrint(stdout,"@I\n",ins);
     // DiabloPrint(stdout,"LUI: arg1:0x%08x, result:0x%08x\n",src1,dest.i);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
      	ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
      else if(temp32<0)
      	ProcStateSetRegMinus(procstate,MIPSINSRT(ins),dest);
      else if(temp32==0)
      	ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
      if (RELOCATABLE_REFERS_TO(ins))
	ProcStateSetTag(procstate,MIPSINSRT(ins),RELOCATABLE_REFERS_TO(ins)->rel);
      break;

    case MIPS_OR:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE||level2!=VALUE)
      {
	if((!plus1 && !zero1)||(!plus2 && !zero2))
	{
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	}
	else if(!minus1 && !minus2)
	{
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	}
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_uint32)src1;
      temp32|=(t_uint32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

      //CORRECT
    case MIPS_ORI:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      
      GenericSet_uint32(tmp_src2.i,(t_int32)(t_int16)MIPSINSIMMEDIATE(ins));
      s_src2=(t_int32)G_T_UINT32(tmp_src2.i);
      if(s_src2>0)
	plus2=TRUE;
      else
	plus2=FALSE;
      if(s_src2<0)
	minus2=TRUE;
      else
	minus2=FALSE;
      if(s_src2==0)
	zero2=TRUE;
      else
	zero2=FALSE;
      if(level1!=VALUE)
      {
	if((!plus1 && !zero1)||(!plus2 && !zero2))
	{
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),FALSE,FALSE,TRUE);
	}
	else if(!minus1 && !minus2)
	{
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,FALSE);
	}
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      GenericSet_uint32(tmp_src2.i,(t_uint16)MIPSINSIMMEDIATE(ins));
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_uint32)src1;
      temp32|=(t_uint32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRT(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
      break;

    case MIPS_SRL:
#ifdef EXTENDED_CONSTPROP
      GenericSet_uint32(tmp_src1.i,(t_uint16)MIPSINSSA(ins));
      src1=G_T_UINT32(tmp_src1.i);
      if(ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2))
      {
	if(src1!=0)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      GenericSet_uint32(tmp_src1.i,(t_uint16)MIPSINSSA(ins));
      src1=G_T_UINT32(tmp_src1.i);
      if(ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      u_temp32=(t_uint32)src2;
      u_temp32>>=src1;
      GenericSet_uint32(dest.i,u_temp32);
#ifdef EXTENDED_CONSTPROP
      if(((t_int32)u_temp32)>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

    case MIPS_SRA: //OK
      GenericSet_uint32(tmp_src1.i,(t_uint16)MIPSINSSA(ins));
      src1=G_T_UINT32(tmp_src1.i);
//	DiabloPrint(stdout,"SRA @I\n",ins);
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);

      if(level2!=VALUE)
      {
//	DiabloPrint(stdout,"SRA NO VALUE: m:%d z:%d p:%d\n",minus2,zero2,plus2);
	if(!plus2 && !zero2)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	else if(!minus2)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      u_temp32=src2;
      u_temp32>>=src1;
      if((0x80000000&src2)==0x80000000)
      {
	u_temp32|=(0xffffffff<<(32-src1));
      }
      GenericSet_uint32(dest.i,u_temp32);
  //    DiabloPrint(stdout,"SRA: arg1:0x%08x, arg2:0x%08x, result:0x%08x\n",src1,src2,dest.i);
#ifdef EXTENDED_CONSTPROP
      if(((t_int32)u_temp32)>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

    case MIPS_SRAV:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE||level2!=VALUE)
      {
	if(minus2 && !zero2 && !plus2)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	else if(!minus2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      src1&=0x0000000f;
      bzero(&dest,sizeof(register_content));
      temp32=(t_int32)src2;
      temp32>>=src1;
      if((0x80000000&src2)==0x80000000)
      {
	temp32|=(0xffffffff<<(32-src1));
      }
      DiabloPrint(stdout,"SRAV: arg1:0x%08x, arg2:0x%08x, result:0x%08x\n",src1,src2,dest.i);
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;
      
    case MIPS_SRLV:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      else
      {
	s_src2=(t_int32)G_T_UINT32(tmp_src2.i);
	s_src2&=0x000000ff;
	if(s_src2>0)
	  plus2=TRUE;
	else
	  plus2=FALSE;
	if(s_src2<0)
	  minus2=TRUE;
	else
	  minus2=FALSE;
	if(s_src2==0)
	  zero2=TRUE;
	else
	  zero2=FALSE;
	if(!zero2)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      src1&=0x0000000f;
      bzero(&dest,sizeof(register_content));
      u_temp32=src2;
      u_temp32>>=src1;
      GenericSet_uint32(dest.i,u_temp32);
#ifdef EXTENDED_CONSTPROP
      if(((t_int32)u_temp32)>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      DiabloPrint(stdout,"SRLV: %x, %x, %x\n",src1,src2,u_temp32);
      break;

    case MIPS_SLTI: //OK
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);

      if(level1!=VALUE)
      {
	GenericSet_uint32(tmp_src2.i,(t_int32)(t_int16)MIPSINSIMMEDIATE(ins));
	s_src2=(t_int32)G_T_UINT32(tmp_src2.i);
	bzero(&dest,sizeof(register_content));
	if(s_src2>0)
	  plus2=TRUE;
	else
	  plus2=FALSE;
	if(s_src2<0)
	  minus2=TRUE;
	else
	  minus2=FALSE;
	if(s_src2==0)
	  zero2=TRUE;
	else
	  zero2=FALSE;
	if((!plus1 && !minus2 && !zero2 &&plus2)|| 
	    (!plus1 && !zero1 && !minus2 && zero2))
	{
	  temp32=1;
	  GenericSet_uint32(dest.i,(t_uint32)temp32);
	  ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
	}
	else if((plus1 && !zero1 && !minus1 && !plus2)||
	    (plus1 && zero1 && !minus1 && !plus2 && !zero2))
	{
	  temp32=0;
	  GenericSet_uint32(dest.i,(t_uint32)temp32);
	  ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
	}
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,FALSE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      GenericSet_uint32(tmp_src2.i,(t_int32)(t_int16)MIPSINSIMMEDIATE(ins));
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
#ifdef EXTENDED_CONSTPROP
      if((t_int32)src1<(t_int32)src2)
      {
	temp32=1;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
      }
      else
      {
	temp32=0;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
      }
#else
      if((t_int32)src1<(t_int32)src2)
	temp32=1;
      else
	temp32=0;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
//      DiabloPrint(stdout,"SLTI: @I\n",ins);
//      DiabloPrint(stdout,"SLTI: arg1:0x%08x, arg2:0x%08x, result:0x%08x\n",src1,src2,dest.i);
      break;

    case MIPS_SUB:
    case MIPS_SUBU:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE||level2!=VALUE)
      {
	if(minus1 && !zero1 && !plus1)
	  if(!minus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(minus1 && zero1 && !plus1)
	  if(!minus2 && !zero2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	  else if(!minus2 && zero2 && plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,TRUE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && !zero1 && plus1)
	  if(!plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,FALSE,FALSE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && zero1 && plus1)
	  if(minus2 && !zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,FALSE,FALSE);
	  else if(minus2 && zero2 && !plus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && zero1 && !plus1)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),minus2,zero2,plus2);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_int32)src1;
      temp32-=(t_int32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;
    case MIPS_MTC0:
      break;
      /*De waarde die naar een fp register wordt geschreven, wordt getransformeerd
       * in de juiste vorm*/
/*    case MIPS_MTC1:
       GEEN FLOATING POINT MEER !!! (toch geen contstanten vinden)
      if(ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      bzero(&dest,sizeof(register_content));
      src1=G_T_UINT32(tmp_src1.i);
      FP_VALUE(dest.f)=(long double)(t_int32)src1;
      ProcStateSetReg(procstate,MIPSINSFS(ins),dest);

      break;
     
    case MIPS_MFC1:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      bzero(&dest,sizeof(register_content));
      s_src1=(t_int32)(long double)FP_VALUE(dest.f);
      GenericSet_uint32(dest.i,(t_uint32)s_src1);
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
      break;
*/      
    case MIPS_SLT:
#ifdef EXTENDED_CONSTPROP
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=VALUE||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=VALUE)
      {
	ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
#ifdef EXTENDED_CONSTPROP
      if((t_int32)src1<(t_int32)src2)
      {
	temp32=1;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      }
      else
      {
	temp32=0;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      }
#else
      if((t_int32)src1<(t_int32)src2)
	temp32=1;
      else
	temp32=0;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

      //CORRECT
    case MIPS_SLTU:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
#ifdef EXTENDED_CONSTPROP
      if(src1<src2)
      {
	temp32=1;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      }
      else
      {
	temp32=0;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      }
#else
      if(src1<src2)
	temp32=1;
      else
	temp32=0;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

    case MIPS_SLTIU:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      GenericSet_uint32(tmp_src2.i,MIPSINSIMMEDIATE(ins));
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
#ifdef EXTENDED_CONSTPROP
      if(src1<src2)
      {
	temp32=1;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
      }
      else
      {
	temp32=0;
	GenericSet_uint32(dest.i,(t_uint32)temp32);
	ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
      }
#else
      if(src1<src2)
	temp32=1;
      else
	temp32=0;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
      break;

      //CORRECT
    case MIPS_SLL:
      GenericSet_uint32(tmp_src1.i,(t_uint16)MIPSINSSA(ins));
      src1=G_T_UINT32(tmp_src1.i);
      if(ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2))
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32=(t_uint32)src2;
      temp32<<=src1;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

      //CORRECT
    case MIPS_SLLV:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      src1&=0x000f;
      bzero(&dest,sizeof(register_content));
      temp32=src2;
      temp32<<=src1;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

      //CORRECT
    case MIPS_NOR:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE||level2!=VALUE)
      {
	if(minus1 && !minus2)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	else if((!zero1 && !plus1)||(!zero2 && !plus2))
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32 = src1;
      temp32 |= src2;
      temp32 = ~temp32;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      //DiabloPrint(stdout,"@I\n",ins);
      //DiabloPrint(stdout,"NOR: arg1:0x%08x, arg2:0x%08x, result:0x%08x\n",src1,src2,temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;
/*
    case MIPS_DIV_:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSFT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      //DiabloPrint(stdout,"DIV_: %x, %x, %x\n",src1,src2,temp32);
      temp32 = src1;
      if(src2==0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      temp32 /= src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
      //DiabloPrint(stdout,"DIV_: %x, %x, %x\n",src1,src2,temp32);
      break;
*/
    case MIPS_MUL:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32 = (t_int32)src1;
      temp32 *= (t_int32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(temp32<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(temp32==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;
/*
    case MIPS_MUL_:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSFT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      temp32 = (t_int32)src1;
      temp32 *= (t_int32)src2;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
      //DiabloPrint(stdout,"MUL_: %x, %x, %x\n",src1,src2,temp32);
      break;
  */    
    case MIPS_XOR:
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      ProcStateGetRegBotExtend(procstate,MIPSINSRT(ins),&plus2,&zero2,&minus2);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);
      level2=ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2);
      if(level1!=VALUE||level2!=VALUE)
      {
	if(!plus1 && !zero1)
	  if (!plus2 && !zero2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	  else if (!minus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),FALSE,FALSE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && !minus2)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,FALSE);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRD(ins),TRUE,TRUE,TRUE);
	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      u_temp32 = src1;
      u_temp32 |= src2;
      u_temp32b = src1;
      u_temp32b &= src2;
      u_temp32b = !u_temp32b;
      u_temp32 &= u_temp32b;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
#ifdef EXTENDED_CONSTPROP
      if(((t_int32)u_temp32)>0)
	ProcStateSetRegPlus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)<0)
	ProcStateSetRegMinus(procstate,MIPSINSRD(ins),dest);
      else if(((t_int32)u_temp32)==0)
	ProcStateSetRegZero(procstate,MIPSINSRD(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;
      
    case MIPS_XORI: //OK!
#ifdef EXTENDED_CONSTPROP
      ProcStateGetRegBotExtend(procstate,MIPSINSRS(ins),&plus1,&zero1,&minus1);
      level1=ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1);

      if(level1!=VALUE)
      {
	GenericSet_uint32(tmp_src2.i,MIPSINSIMMEDIATE(ins));
	src2=G_T_UINT32(tmp_src2.i);
	if(src2>0)
	  plus2=TRUE;
	else
	  plus2=FALSE;
	minus2=FALSE;
	if(src2==0)
	  zero2=TRUE;
	else
	  zero2=FALSE;

	if(!plus1 && !zero1)
	  if (!plus2 && !zero2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,FALSE);
	  else if (!minus2)
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),FALSE,FALSE,TRUE);
	  else
	    ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);
	else if(!minus1 && !minus2)
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,FALSE);
	else
	  ProcStateSetRegBotExtend(procstate,MIPSINSRT(ins),TRUE,TRUE,TRUE);

	break;
      }
#else
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
#endif
      src1=G_T_UINT32(tmp_src1.i);
      GenericSet_uint32(tmp_src2.i,MIPSINSIMMEDIATE(ins));
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      u_temp32 = src1;
      u_temp32 |= src2;
      u_temp32b = src1;
      u_temp32b &= src2;
      u_temp32b = ~u_temp32b;
      u_temp32 &= u_temp32b;
      GenericSet_uint32(dest.i,(t_uint32)u_temp32);
//      DiabloPrint(stdout,"@I\n",ins);
//      DiabloPrint(stdout,"XORI: arg1:0x%08x, arg2:0x%08x, result:0x%08x\n",src1,src2,u_temp32);
#ifdef EXTENDED_CONSTPROP
      if(((t_int32)u_temp32)>0)
	ProcStateSetRegPlus(procstate,MIPSINSRT(ins),dest);
      else if(((t_int32)u_temp32)<0)
	ProcStateSetRegMinus(procstate,MIPSINSRT(ins),dest);
      else if(((t_int32)u_temp32)==0)
	ProcStateSetRegZero(procstate,MIPSINSRT(ins),dest);
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPSINSRT(ins),dest);
#endif
      break;
/*      
    case MIPS_NEG:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      temp32 = -((t_int32)src1);
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
    //  DiabloPrint(stdout,"NEG: %d, %d\n",(t_int32)src1,temp32);
      break;

    case MIPS_ABS:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      u_temp32=src1;
      if((31>>src1&0x80000000)==1)
	u_temp32 = (!src1)+1;ERROR ! != ~ !!!!!!!!!!
      GenericSet_uint32(dest.i,u_temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
      DiabloPrint(stdout,"ABS: %d, %d\n",(t_int32)src1,(t_int32)u_temp32);
      break;
*//*
    case MIPS_C:
	  //DiabloPrint(stdout,"REGS Uitgevoerd: %d %d\n",MIPSINSFS(ins),MIPSINSFT(ins));
      if(MIPSINSCC(ins)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      bzero(&dest,sizeof(register_content));
      switch(MIPSINSCOND(ins))
      {
	case C_F:
	case C_SF:
	case C_UN:
	case C_NGLE:
          GenericSet_uint32(dest.i,0);
          ProcStateSetReg(procstate,MIPS_REG_FCCR,dest);
	  break;
      }
      if(ProcStateGetReg(procstate,MIPSINSFS(ins)+1,&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSFT(ins)+1,&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      f_src1=FP_VALUE(tmp_src1.f);
      f_src2=FP_VALUE(tmp_src2.f);
      //src1=G_T_UINT32(tmp_src1.i);
      //src2=G_T_UINT32(tmp_src2.i);
      //s_src1 = (t_int32)src1;
      //s_src2 = (t_int32)src2;

      //GenericSet_uint32(dest.i,(t_uint32)temp32);
      //ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
      switch(MIPSINSCOND(ins))
      {
	case C_EQ:
	case C_UEQ:
	case C_SEQ:
	case C_NGL:
	  //if(s_src1==s_src2)
	  if(f_src1==f_src2)
	  {
	    GenericSet_uint32(dest.i,1);
	    ProcStateSetReg(procstate,MIPS_REG_FCCR,dest);
	  }
	  else
	  {
	    GenericSet_uint32(dest.i,0);
	    ProcStateSetReg(procstate,MIPS_REG_FCCR,dest);
	  }
	  break;
	case C_OLT:
	case C_ULT:
	case C_LT:
	case C_NGE:
	  //if(s_src1<s_src2)
	  if(f_src1<f_src2)
	  {
	    GenericSet_uint32(dest.i,1);
	    ProcStateSetReg(procstate,MIPS_REG_FCCR,dest);
	  }
	  else
	  {
	    GenericSet_uint32(dest.i,0);
	    ProcStateSetReg(procstate,MIPS_REG_FCCR,dest);
	  }
	  //DiabloPrint(stdout,"Uitgevoerd: %d %d\n",ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1),ProcStateGetReg(procstate,MIPSINSFT(ins),&tmp_src2));
	  DiabloPrint(stdout,"Uitgevoerd: %ld %ld",f_src1,f_src2);
	  DiabloPrint(stdout," %d %d\n",MIPSINSFS(ins),MIPSINSFT(ins));
	  break;
	case C_OLE:
	case C_ULE:
	case C_LE:
	case C_NGT:
	  //if(s_src1==s_src2||s_src1<s_src2)
	  if(f_src1==f_src2||f_src1<f_src2)
	  {
	    GenericSet_uint32(dest.i,1);
	    ProcStateSetReg(procstate,MIPS_REG_FCCR,dest);
	  }
	  else
	  {
	    GenericSet_uint32(dest.i,0);
	    ProcStateSetReg(procstate,MIPS_REG_FCCR,dest);
	  }
	  break;
	default:
	//  ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
          FATAL(("Can not be emulated: Not implemented condition code for C_:%d\n",MIPSINSCOND(ins)));
	  break;
      }
//	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
      break;

      
    */  /*
    case MIPS_CVT_D:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      u_temp32=src1;
      GenericSet_uint32(dest.i,u_temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
      DiabloPrint(stdout,"CVT_D_: %d, %d\n",(t_int32)src1,(t_int32)u_temp32);
      break;

    case MIPS_CVT_W:
    case MIPS_CVT_S:
      if(ProcStateGetReg(procstate,MIPSINSFS(ins),&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      u_temp32=src1;
      GenericSet_uint32(dest.i,u_temp32);
      ProcStateSetReg(procstate,MIPSINSFD(ins),dest);
      DiabloPrint(stdout,"CVT_D_: %d, %d\n",(t_int32)src1,(t_int32)u_temp32);
      break;
*/
    case MIPS_CFC1:
    case MIPS_CTC1:
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
    default:
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	// FATAL(("Can not be emulated: Not implemented:%d\n",MIPSINSOPCODE(ins)));
	// DiabloPrint(stdout,"STDOUT: %d\n",MIPSINSOPCODE(ins));
      break;
  }
  
}




void MipsInsEmulMul(t_ins * ins, t_procstate * procstate)
{
  register_content tmp_src1,tmp_src2, dest,dest2;
  t_uint32 src1,src2;
  t_uint64 u_temp64;
  t_uint32 u_temp32,u_temp32b;
  t_int32 s_src1,s_src2;
  t_int64 temp64;
  t_int32 temp32,temp32b;

  switch (MIPSINSOPCODE(ins))
  {
    case MIPS_MULT:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      s_src1=(t_int64)(t_int32)src1;
      s_src2=(t_int64)(t_int32)src2;
      bzero(&dest,sizeof(register_content));
      bzero(&dest2,sizeof(register_content));
      temp64=s_src1;
      temp64*=s_src2;
      temp32=(t_uint32)(temp64&0x0000ffff);
      temp32b=(t_uint32)((temp64&0xffff0000)>>32);
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      GenericSet_uint32(dest2.i,(t_uint32)temp32b);
#ifdef EXTENDED_CONSTPROP
      if(temp64>0)
      {
	ProcStateSetRegPlus(procstate,MIPS_REG_LO,dest);
	ProcStateSetRegPlus(procstate,MIPS_REG_HI,dest2);
      }
      else if(temp64<0)
      {
	ProcStateSetRegMinus(procstate,MIPS_REG_LO,dest);
	ProcStateSetRegMinus(procstate,MIPS_REG_HI,dest2);
      }
      else if(temp64==0)
      {
	ProcStateSetRegZero(procstate,MIPS_REG_LO,dest);
	ProcStateSetRegZero(procstate,MIPS_REG_HI,dest2);
      }
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPS_REG_LO,dest);
      ProcStateSetReg(procstate,MIPS_REG_HI,dest2);
#endif
      break;

    case MIPS_MFLO:
#ifdef EXTENDED_CONSTPROP
      ProcStateRegDup(procstate,MIPS_REG_LO,MIPSINSRD(ins));
#else
      if(ProcStateGetReg(procstate,MIPS_REG_LO,&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      temp32=src1;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

    case MIPS_MFHI:
#ifdef EXTENDED_CONSTPROP
      ProcStateRegDup(procstate,MIPS_REG_HI,MIPSINSRD(ins));
#else
      if(ProcStateGetReg(procstate,MIPS_REG_HI,&tmp_src1)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      bzero(&dest,sizeof(register_content));
      temp32=src1;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      ProcStateSetReg(procstate,MIPSINSRD(ins),dest);
#endif
      break;

    case MIPS_DIV:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      bzero(&dest2,sizeof(register_content));
      s_src1=(t_int32)src1;
      s_src2=(t_int32)src2;
      temp32=s_src1;
      temp32/=s_src2;
      temp32b=s_src1;
      temp32b%=s_src1;
      GenericSet_uint32(dest.i,(t_uint32)temp32);
      GenericSet_uint32(dest2.i,(t_uint32)temp32b);
#ifdef EXTENDED_CONSTPROP
      if(temp32>0)
      {
	ProcStateSetRegPlus(procstate,MIPS_REG_LO,dest);
      }
      else if(temp32<0)
      {
	ProcStateSetRegMinus(procstate,MIPS_REG_LO,dest);
      }
      else if(temp32==0)
      {
	ProcStateSetRegZero(procstate,MIPS_REG_LO,dest);
      }
      else FATAL(("ONMOGELIJK!!"));
      if(temp32b>0)
      {
	ProcStateSetRegPlus(procstate,MIPS_REG_HI,dest2);
      }
      else if(temp32b<0)
      {
	ProcStateSetRegMinus(procstate,MIPS_REG_HI,dest2);
      }
      else if(temp32b==0)
      {
	ProcStateSetRegZero(procstate,MIPS_REG_HI,dest2);
      }
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPS_REG_LO,dest);
      ProcStateSetReg(procstate,MIPS_REG_HI,dest2);
#endif
      break;
    case MIPS_DIVU:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      bzero(&dest2,sizeof(register_content));
      u_temp32=src1;
      if(src2==0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      u_temp32/=src2;
      u_temp32b=src1;
      u_temp32b%=src1;
      GenericSet_uint32(dest.i,u_temp32);
      GenericSet_uint32(dest2.i,u_temp32b);
#ifdef EXTENDED_CONSTPROP
      ProcStateSetRegPlus(procstate,MIPS_REG_LO,dest);
      ProcStateSetRegPlus(procstate,MIPS_REG_HI,dest2);
#else
      ProcStateSetReg(procstate,MIPS_REG_LO,dest);
      ProcStateSetReg(procstate,MIPS_REG_HI,dest2);
#endif
      break;
    case MIPS_MULTU:
      if(ProcStateGetReg(procstate,MIPSINSRS(ins),&tmp_src1)!=0||
	  ProcStateGetReg(procstate,MIPSINSRT(ins),&tmp_src2)!=0)
      {
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	break;
      }
      src1=G_T_UINT32(tmp_src1.i);
      src2=G_T_UINT32(tmp_src2.i);
      bzero(&dest,sizeof(register_content));
      bzero(&dest2,sizeof(register_content));
      u_temp64=src1;
      u_temp64*=src2;
      u_temp32=temp64&0x0000ffff;
      u_temp32b=(temp64&0xffff0000)>>32;
      GenericSet_uint32(dest.i,u_temp32);
      GenericSet_uint32(dest2.i,u_temp32b);
#ifdef EXTENDED_CONSTPROP
      if(u_temp64>0)
      {
	ProcStateSetRegPlus(procstate,MIPS_REG_LO,dest);
	ProcStateSetRegPlus(procstate,MIPS_REG_HI,dest2);
      }
      else if(u_temp64==0)
      {
	ProcStateSetRegZero(procstate,MIPS_REG_LO,dest);
	ProcStateSetRegZero(procstate,MIPS_REG_HI,dest2);
      }
      else FATAL(("ONMOGELIJK!!"));
#else
      ProcStateSetReg(procstate,MIPS_REG_LO,dest);
      ProcStateSetReg(procstate,MIPS_REG_HI,dest2);
#endif
      break;
    default:
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
	// FATAL(("Can not be emulated: Not implemented:%d\n",MIPSINSOPCODE(ins)));
	// DiabloPrint(stdout,"STDOUT: %d\n",MIPSINSOPCODE(ins));
      break;
  }

}

void MipsInsEmulSWI(t_ins * ins, t_procstate * procstate)
{
	ProcStateSetRegsetBot(procstate,INS_REGS_DEF(ins));
}



/* vim: set shiftwidth=2: */
