/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

#include <stdlib.h>
#include "../../utils/util_error.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_function.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_liveness.h"
#include "../../kernel/diablo_constprop.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_edge.h"
#include "../../kernel/diablo_opt_misc.h"
#include "../../kernel/diablo_iterators.h"
#ifdef MIPS_SUPPORT
#include "mips_opcodes.h"
#include "mips_registers.h"
#include "mips_constant_propagation.h"
#endif

void InlineFunAtCallsite(t_function * fun, t_bbl *call_site);
t_uint32 MipsInlineSmallProcedure(t_object * obj, t_cfg * cfg, t_function * fun);

/* MipsInlineTrivial {{{ */
void MipsInlineTrivial(t_cfg * cfg)
{
  t_object * obj=CFG_OBJECT(cfg);
  t_ins * i_ins;
  t_cfg_edge * i_edge;
  t_function * i_fun;
  t_bbl * i_bbl;
  t_function* tmp;

  int nr_ins;
  int nr_edges;
  int nr_callees = 0;
  int nr_callers = 0;

  DiabloPrint(stdout,"INLINE TRIVIAL\n");
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    nr_ins = nr_edges = 0;

    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	nr_ins++;
      }

      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
      {
	if (!CfgEdgeIsForwardInterproc(i_edge)) continue;
	if (EDGE_CAT(i_edge)==ET_CALL) nr_edges+=100;
	if (IsHellNode(EDGE_HEAD(i_edge))) nr_edges+=100;
	nr_edges++;
      }
    }

    if ((nr_ins<=10 && nr_ins>=1)&&(BBL_INS_LAST(i_fun->bbl_first)))
    {
      if(( MIPSINSOPCODE(BBL_INS_LAST(i_fun->bbl_first))==MIPS_JR && MIPSINSRS(BBL_INS_LAST(i_fun->bbl_first))==MIPS_REG_RA) ||
	  (INS_PREV(BBL_INS_LAST(i_fun->bbl_first)) && MIPSINSOPCODE(INS_PREV(BBL_INS_LAST(i_fun->bbl_first)))==MIPS_JR && MIPSINSRS(INS_PREV(BBL_INS_LAST(i_fun->bbl_first)))==MIPS_REG_RA))
      {
	int nr = MipsInlineSmallProcedure(obj, cfg, i_fun);
	if (nr)
	  nr_callees++;
	nr_callers+=nr;
      }
    }
    else
      if (nr_edges<2)
      {
	if (nr_edges==1)
	{
	  nr_callees++;
	  nr_callers++;
	}
      }
  }

  printf("[inlining            ] %d callees inlined at %d callers\n",nr_callees,nr_callers);
}
/* }}} */

/* MipsInlineSmallProcedure {{{ */
/* #define PRINT_INLINE */
t_uint32 MipsInlineSmallProcedure(t_object * obj, t_cfg * cfg, t_function * fun)
{
  t_bbl* entry;
  t_bbl* i_bbl;
  t_cfg_edge * call_edge, * tmp;
  int nr_bbls=0;
  t_ins * new_ins;

  DiabloPrint(stdout,"INLINE SMALL PROCEDURE\n");
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    nr_bbls++;
  }

#ifdef PRINT_INLINE
  if (nr_bbls!=2)
  {
    printf("not inlined: number of bbls is %d\n",nr_bbls++);
  }
#endif

  nr_bbls=0;

  entry = fun->bbl_first;


  BBL_FOREACH_PRED_EDGE_SAFE(entry,call_edge,tmp)
  {
    t_bbl* call_site = CFG_EDGE_HEAD(call_edge);
    t_ins* call_ins = BBL_INS_LAST(call_site);
    if (EDGE_CAT(call_edge)!=ET_CALL) continue;
    if (IsHellNode(call_site)) continue;
    if ((MIPSINSOPCODE(call_ins)!=MIPS_JAL))
    {
      if(INS_PREV(call_ins) && MIPSINSOPCODE(INS_PREV(call_ins))!=MIPS_JAL) continue;
      call_ins=INS_PREV(call_ins);
    }
    if ((BBL_NINS(entry)>=1 && BBL_NINS(entry)<=10) )
    {
      t_cfg_edge * link_edge = NULL;
      t_cfg_edge * fall_through_edge = NULL;
      t_cfg_edge * edge;
      nr_bbls++;

#ifdef PRINT_INLINE
      DiabloPrint(stdout,"Call:@eiB\n",call_site);
      DiabloPrint(stdout,"Bbl IN:@eiB\n",entry);
#endif

      Export_Function(BBL_FUN(call_site),"inline.dot");

      BBL_EXEC_COUNT(entry)-=EDGE_EXEC_COUNT(call_edge);

      {

	t_ins * h_ins;
	h_ins=BBL_INS_FIRST(entry);

	for(h_ins=BBL_INS_FIRST(entry);;h_ins=INS_NEXT(h_ins))
	{
	  if(h_ins==NULL)
	  {
#ifdef PRINT_INLINE
	    DiabloPrint(stdout,"Break!");
#endif
	    break;
	  }
	  if(MIPSINSOPCODE(h_ins)==MIPS_JR)
	  {
#ifdef PRINT_INLINE
	    DiabloPrint(stdout,"JR:continue");
#endif
	    continue;
	  }
#ifdef PRINT_INLINE
	  DiabloPrint(stdout,"Toegevoegd: @I\n",h_ins);
#endif

	  new_ins = InsDup(h_ins);
	  RELOCATABLE_REFED_BY(new_ins)=NULL;
	  InsAppendToBbl(new_ins,call_site);
	  if (INS_IS_CONDITIONAL(call_ins))
	  {
	    FATAL(("No cond in MIPS?"));
	  }
	}
      }


      InsKill(call_ins);

      if (tmp==call_edge->corr_edge)
	tmp = T_CFG_EDGE(EDGE_PRED_NEXT(tmp));

      CfgEdgeKill(call_edge->corr_edge);
      CfgEdgeKill(call_edge);

      BBL_FOREACH_SUCC_EDGE(call_site,edge)
      {
	if  (EDGE_CAT(edge)==ET_FUNLINK || EDGE_CAT(edge)==ET_IPFUNLINK)
	  link_edge = edge;
	else if  (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	  fall_through_edge = edge;
	else
	  FATAL(("OOPS: EDGE OTHER THAN LINK OR FALLTHROUGH EDGE DURING INLINING\n"));
      }

      if (fall_through_edge)
      {
	CfgEdgeKill(link_edge);
      }
      else
      {
	if (!link_edge) FATAL(("OOPS: NO LINK EDGE IN INLINING\n"));
	if (EDGE_CAT(link_edge)==ET_FUNLINK)
	  EDGE_CAT(link_edge)=ET_FALLTHROUGH;
	else
	  EDGE_CAT(link_edge)=ET_IPFALLTHRU;
      }
#ifdef PRINT_INLINE
      DiabloPrint(stdout,"Become:@eiB\n",call_site);
#endif
    }
  }
  return nr_bbls;

}
/* }}} */


void MipsGeneralInlining(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_cfg_edge * i_edge;
  t_bool flag = TRUE;
  t_function * i_fun;
  int teller;
  t_bbl * call_site=NULL;
  t_ins * last_ins;
  t_function * tmp;

  DiabloPrint(stdout,"INLINE GENERAL\n");
#ifdef PRINT_INLINE
  DiabloPrint(stdout,"inlining ");
#endif
  CFG_FOREACH_FUN_SAFE(cfg,i_fun,tmp)
  {
    flag = TRUE;
    teller = 0;
    FUN_FOREACH_BBL(i_fun,i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	if (CfgEdgeIsForwardInterproc(i_edge) && EDGE_CAT(i_edge)!=ET_CALL)
	  flag = FALSE;

      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
      {
	if (CfgEdgeIsForwardInterproc(i_edge))
	{
	  if (EDGE_CAT(i_edge)!=ET_CALL)
	    flag = FALSE;
	  else if (IsHellNode(EDGE_HEAD(i_edge)))
	    teller+=2;
	  else
	  {
	    teller++;
	    call_site = CFG_EDGE_HEAD(i_edge);
	  }
	}
      }
    }

    if (call_site == cfg->unique_entry_node) continue;
    if (!flag || teller!=1)
      continue;

    BBL_FOREACH_PRED_EDGE(FUN_BBL_LAST(i_fun),i_edge)
    {
      last_ins = BBL_INS_LAST(CFG_EDGE_HEAD(i_edge));
      if (!last_ins)
      {
	flag=FALSE;
      }
      else if (MIPSINSOPCODE(last_ins)!=MIPS_JR)
      {
	last_ins = INS_PREV(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)));
	if (MIPSINSOPCODE(last_ins)!=MIPS_JR)
	  flag=FALSE;
      }
    }


#if 0
    {
      t_bbl* i_bbl;
      t_int32 aantal=0;

      FUN_FOREACH_BBL(i_fun,i_bbl)
      {
	t_ins * i_ins;
	BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  if (MIPSINSOPCODE(i_ins)==MIPS_JR)
	  {
	    aantal++;
	  }
	}
      }
      if(aantal!=1)
	flag=FALSE;
    }
#endif


    if (!flag)
    {
#ifdef PRINT_INLINE
      FunctionPrint(i_fun,PRINT_INSTRUCTIONS);
      printf("NO INLINE: PROBLEM WITH RETURN\n");
#endif
    }
    else if (teller==1 && (MIPSINSOPCODE(BBL_INS_LAST(call_site))==MIPS_JAL|| 
	  ((INS_PREV(BBL_INS_LAST(call_site)))!=NULL 
	   && MIPSINSOPCODE(INS_PREV(BBL_INS_LAST(call_site)))==MIPS_JAL)) )
    {
#ifdef PRINT_INLINE
      DiabloPrint(stdout,", %s in %s", i_fun->name, BBL_FUN((call_site))->name); 
#endif
      InlineFunAtCallsite(i_fun,call_site);
    }
  }	  
#ifdef PRINT_INLINE
  DiabloPrint(stdout,"\n");
#endif
}

void InlineFunAtCallsite(t_function * fun, t_bbl *call_site)
{
  t_bbl* entry;
  t_bbl* i_bbl;
  t_cfg_edge * call_edge, * tmp;
  t_cfg_edge * link_edge;  
  t_cfg_edge * return_edge;  
  t_cfg_edge * i_edge, *j_edge;

  t_bbl * tmp_bbl;
  t_bbl * return_point;

  t_ins * jump;

  t_bool last_is_ip=FALSE;
//  int gain = 0;

  if(MIPSINSOPCODE(BBL_INS_LAST(call_site))==MIPS_JAL)
  {
    jump=BBL_INS_LAST(call_site);
  }
  else if (INS_PREV(BBL_INS_LAST(call_site))!=NULL&&MIPSINSOPCODE(INS_PREV(BBL_INS_LAST(call_site)))==MIPS_JAL)
  {
    jump=INS_PREV(BBL_INS_LAST(call_site));
  }
  else 
  {
    BblPrint(call_site,PRINT_INSTRUCTIONS);
    FATAL(("ERROR"));
  }


  /* Change the call of the last ins into a branch instruction */

  MakeUnConditional(jump);

  /* Load and store of the return-address can be deleted */
  {
    t_bbl* i_bbl;
    t_int32 aantal=0;

    FUN_FOREACH_BBL(fun,i_bbl)
    {
      t_ins * i_ins;
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	if (MIPSINSOPCODE(i_ins)==MIPS_JAL)
	{
	  aantal++;
	}
      }
    }
    if(aantal==0)
    {
      FUN_FOREACH_BBL(fun,i_bbl)
      {
	t_ins * i_ins;
	BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  if (MIPSINSOPCODE(i_ins)==MIPS_LW && MIPSINSRT(i_ins)==MIPS_REG_R31)
	    InsKill(i_ins);
	  else if (MIPSINSOPCODE(i_ins)==MIPS_SW && MIPSINSRT(i_ins)==MIPS_REG_R31)
	    InsKill(i_ins);
	}
      }
    }
  }
  /* Look at all the predecessors of the entry block of the function, to find
   * the call edge */

  entry = fun->bbl_first;
  BBL_FOREACH_PRED_EDGE_SAFE(entry,call_edge,tmp)
  {
    t_bbl* site = CFG_EDGE_HEAD(call_edge);
    if (site!=call_site) continue;
    break;
  }      

  /* Look at all the successor edges of the callsite to find the link edge (and
   * thus the returnsite) */

  BBL_FOREACH_SUCC_EDGE_SAFE(call_site,link_edge,tmp)
  {
    if (EDGE_CAT(link_edge)==ET_FUNLINK || EDGE_CAT(link_edge)==ET_IPFUNLINK)
      break;
  }

  return_edge = EDGE_CORR(call_edge);
  return_point = CFG_EDGE_TAIL(link_edge);



  /* If the link edge is interprocedural we need to remember this (the added
   * branches that replace the returns must be also interprocedural */

  if (CfgEdgeIsInterproc(link_edge))
    last_is_ip=TRUE;

  /* Kill the edges */
  if (last_is_ip)
    CfgEdgeKill(link_edge->corr_edge);
  CfgEdgeKill(link_edge);
  CfgEdgeKill(call_edge);
  CfgEdgeKill(return_edge);

  /* Add the jump that replaces the call */

  CfgEdgeNew(fun->cfg,call_site,FUN_BBL_FIRST(fun),ET_JUMP);

  /* Add all blocks but the exit block to the calling function */
  FUN_FOREACH_BBL_SAFE_R(fun,i_bbl,tmp_bbl)
  {
    if (i_bbl!=FUN_BBL_LAST(fun))
    {	
      i_bbl->next_in_fun=call_site->next_in_fun;
      call_site->next_in_fun->prev_in_fun=i_bbl;
      call_site->next_in_fun = i_bbl;
      i_bbl->prev_in_fun = call_site;
      i_bbl->fun=call_site->fun;
    }
  }
  if(fun->bbl_last==NULL)
    FATAL(("ERROR"));
  fun->bbl_first = fun->bbl_last;
  fun->bbl_last->prev_in_fun = NULL;
  fun->bbl_last->next_in_fun = NULL;


  /* Look at the predecessors of the exit block, to identify all returnsites */

  BBL_FOREACH_PRED_EDGE_SAFE(FUN_BBL_LAST(fun),i_edge,tmp)
  {

    t_bbl * head = CFG_EDGE_HEAD(i_edge);
    t_ins * last_ins;
    last_ins = BBL_INS_LAST(head);
    if (!last_ins)
    {
      FunctionPrint(fun,PRINT_INSTRUCTIONS);
      FATAL(("This should be impossible: we encountered a returnsite without instructions. As a return needs an explicit return instruction this is impossible\n"));
    }

    /* Change last instruction of return site */


    if (MIPSINSOPCODE(last_ins)==MIPS_JR)
    {
      MakeUnConditional(last_ins);
    }
    else if(MIPSINSOPCODE(INS_PREV(BBL_INS_LAST(head)))==MIPS_JR)
    {
      MakeUnConditional(INS_PREV(BBL_INS_LAST(head)));
    }
    else FATAL(("ERROR"));

    /* Kill the return edge, add a (IP)JUMP edge */
    CfgEdgeKill(i_edge);
    if (last_is_ip)
    {
      j_edge = CfgEdgeNew(fun->cfg,head,return_point,ET_IPJUMP);
      CfgEdgeNewCompensating(fun->cfg,j_edge);
    }
    else
    {
      j_edge=  CfgEdgeNew(fun->cfg,head,return_point,ET_JUMP);
    }
    j_edge->cfg = fun->cfg;
  }

//  printf("[gain= %d]",gain);
}


