/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"
#ifndef MIPS_INSTRUCTION_TYPEDEFS
#define MIPS_INSTRUCTION_TYPEDEFS
typedef struct _t_mips_ins t_mips_ins;
#endif

#ifndef MIPS_INSTRUCTION_H
#define MIPS_INSTRUCTION_H
#include "../../kernel/diablo_ins.h"
struct _t_mips_ins
{
   t_ins generic;
   t_uint16 opcode; /* Change type if necessary */
   t_uint8 processor; /* Processor die de functie uitvoerd */

   /*Behorende bij de CPU-instructies */
   t_uint8 rs; /* Source register */
   t_uint8 rt; /* target register (Source/destination) */
   t_uint8 rd; /* Destination register */
   t_uint8 sa; /* Shift amount */
   t_uint8 function; /* Funtion in the primary opcode FUNTION */
   t_uint32 immediate; /* immediate register */
   t_uint32 instr_index; /* 26 bit index shifted left two bits to supply the low-order 28 bits of the jump target adress*/

   /*Behorende bij CO0 instructies */
   t_uint8 sel;

   /*Behorende bij CO1 instructies */
   t_uint8 cc;
   
};

/* Often used shorthands */
#define MIPSINSPROCESSOR(mips_ins) (((t_mips_ins *) mips_ins) -> processor)
#define MIPSINSOPCODE(mips_ins) (((t_mips_ins *) mips_ins)->opcode)
/* CPU */
#define MIPSINSRS(mips_ins) (((t_mips_ins *) mips_ins) -> rs )
#define MIPSINSBASE(mips_ins) (((t_mips_ins *) mips_ins) -> rs )
#define MIPSINSRT(mips_ins) (((t_mips_ins *) mips_ins) -> rt )
#define MIPSINSRD(mips_ins) (((t_mips_ins *) mips_ins) -> rd )
#define MIPSINSSA(mips_ins) (((t_mips_ins *) mips_ins) -> sa )
#define MIPSINSIMMEDIATE(mips_ins) (((t_mips_ins *) mips_ins) -> immediate )
#define MIPSINSOFFSET(mips_ins) (((t_mips_ins *) mips_ins) -> immediate )
#define MIPSINSINSTR_INDEX(mips_ins) (((t_mips_ins *) mips_ins) -> instr_index )
#define MIPSINSFUNCTION(mips_ins) (((t_mips_ins *) mips_ins) -> function )
/* CO0 */
#define MIPSINSSEL(mips_ins) (((t_mips_ins *) mips_ins) -> sel )
/* CO1 */
#define MIPSINSFT(mips_ins) (((t_mips_ins *) mips_ins) -> rt )
#define MIPSINSFS(mips_ins) (((t_mips_ins *) mips_ins) -> rd )
#define MIPSINSFD(mips_ins) (((t_mips_ins *) mips_ins) -> sa )
#define MIPSINSFMT(mips_ins) (((t_mips_ins *) mips_ins) -> rs )
#define MIPSINSSUB(mips_ins) (((t_mips_ins *) mips_ins) -> rs )
#define MIPSINSCC(mips_ins) (((t_mips_ins *) mips_ins) -> cc )
#define MIPSINSCOND(mips_ins) (((t_mips_ins *) mips_ins) -> sa )

#define T_MIPS_INS(mips_ins)            ((t_mips_ins *) mips_ins)
/* processoren of co- */
#define CPU	0
#define COP0	1
#define COP1	2
#define COP2	3
#define COP3	4

t_bool MipsInsHasSideEffect(t_ins * ins);

void MipsInstructionMakeNoop(t_ins * ins);
void MipsMakeAddressProducer(t_ins * ins);
t_bool MipsIsControlflow(t_ins * ins);
void MipsDifferentInstructions(t_object * obj, t_section * code);
void MipsDifferentInstructions2(t_section * code); 
t_tristate MipsIsSyscallExit(t_ins * ins);
t_bool MipsInsIsIndirectCall(t_ins * ins);
t_bool MipsInsIsIndirectJump(t_ins * ins);
t_bool MIPSInsIsProcedureCall(t_ins * ins);

#endif
/* vim: set shiftwidth=2: */
