/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../kernel/diablo_edge.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_layout.h"
#include "../../kernel/diablo_stack_machine.h"
#include "mips_opcodes.h"
#include "mips_registers.h"
#include "mips_utils.h"
#include "mips_layout.h"
#include "mips_assemble.h"
#include "mips_disassemble_one.h"

#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_section.h"
#include "../../kernel/diablo_iterators.h"
#include "../../utils/util_status.h"
#include <string.h>

#ifdef OBFUSCATION_SUPPORT
#include "../../obfuscation/obf_branch.h"
#endif

typedef struct 
{
  t_bbl * bbl;
  int links_to_next_chain;
  int links_to_prev_chain; 
  int links_to_all_prev_chains;
  int links_to_all_next_chains;
} layouted_chain;

static layouted_chain** layout;

static int number_of_layouted_chains;

void MipsAddressProducers(t_bbl * head, t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_ins * i_ins;
  t_address new_address;
  t_address old_address;
#ifdef AVOID_BRANCHES
  t_ins *prev_triple = NULL;
  t_ins * load_ins;
    int teller = 0;
#endif

  t_address gaddress;
  i_bbl = head; 
  gaddress = SECTION_GET_CADDRESS(cfg->sec);
  new_address = SECTION_GET_CADDRESS(cfg->sec);
  old_address = SECTION_GET_CADDRESS(cfg->sec);


  VERBOSE(0,("Text section starts at @G/@G",new_address,old_address));
  while (i_bbl)
  {
    if (!IS_DATABBL(i_bbl))
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
      {
	GenericSet_uint32(INS_MIN_SIZE(i_ins),4);

      }
    }

    BblSetAllAddressSuper(i_bbl,old_address,new_address);
    new_address=GENERIC_ADD(new_address,BBL_CSIZE(i_bbl));
    old_address=GENERIC_ADD(old_address,BBL_MIN_SIZE(i_bbl));
    i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
  }
  VERBOSE(0,("And runs till @G/@G",new_address,old_address));
}


//===============================================================================================


static void MergeTwoChains(t_bbl * head, t_bbl * tail)
{
  t_bbl * i_bbl;
  t_bbl * head_chain_first = BBL_FIRST_IN_CHAIN(head);
  t_bbl * tail_chain_first = BBL_FIRST_IN_CHAIN(tail);
  t_bbl * head_chain_last = BBL_LAST_IN_CHAIN(head);
  t_bbl * tail_chain_last = BBL_LAST_IN_CHAIN(tail);
  
  i_bbl = head_chain_first;


  while (i_bbl)
    {
      BBL_LAST_IN_CHAIN(i_bbl)=tail_chain_last;
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }

  i_bbl = tail_chain_first;

  while (i_bbl)
    {
      BBL_FIRST_IN_CHAIN(i_bbl)=head_chain_first;
      i_bbl = BBL_NEXT_IN_CHAIN(i_bbl);
    }

  BBL_NEXT_IN_CHAIN(head_chain_last)=tail_chain_first;
  BBL_PREV_IN_CHAIN(tail_chain_first)=head_chain_last;
}

static void UpdateJDisplacement(t_cfg_edge * edge, t_ins * h_ins)
{
  t_bbl * head = CFG_EDGE_HEAD(edge);
  t_bbl * tail = CFG_EDGE_TAIL(edge);
  t_address hulp_addr;
  GenericSet_uint32(hulp_addr,0xf0000000);
  
  MIPSINSINSTR_INDEX(h_ins) = G_T_UINT32(GENERIC_SUB(BBL_CADDRESS(tail),GENERIC_AND(hulp_addr,INS_OLD_ADDRESS(BBL_INS_LAST(head)))))>>2; 

}
static void UpdateBranchDisplacement(t_cfg_edge * edge, t_ins * h_ins)
{
  t_bbl * head = CFG_EDGE_HEAD(edge);
  t_bbl * tail = CFG_EDGE_TAIL(edge);
  t_address hulp_addr;
  GenericSet_uint32(hulp_addr,0xf0000000);

 MIPSINSOFFSET(h_ins)=(GENERIC_DIV_UINT32(GENERIC_SUB(BBL_CADDRESS(tail),GENERIC_ADD_UINT32(BBL_CADDRESS(head),(BBL_NINS(head)-1)<<2)),4))&0x0000ffff;
 /* CHECK ! */
// DiabloPrint(stdout,"@I\n %x %x\n",h_ins,BBL_CADDRESS(tail),MIPSINSOFFSET(h_ins));
// DiabloPrint(stdout,"@I\n %x %x\n",h_ins,BBL_CADDRESS(tail),  (((t_uint32)((t_int16)MIPSINSIMMEDIATE(h_ins)))<<2)+GENERIC_ADD_UINT32(BBL_CADDRESS(head),(BBL_NINS(head)-1)<<2) );
 if(BBL_CADDRESS(tail)!=(((t_uint32)((t_int16)MIPSINSIMMEDIATE(h_ins)))<<2)+GENERIC_ADD_UINT32(BBL_CADDRESS(head),(BBL_NINS(head)-1)<<2) )
 {
   DiabloPrint(stdout,"@I\n %x %x\n",h_ins,BBL_CADDRESS(tail),  (((t_uint32)((t_int16)MIPSINSIMMEDIATE(h_ins)))<<2)+GENERIC_ADD_UINT32(BBL_CADDRESS(head),(BBL_NINS(head)-1)<<2) );
   FATAL(("ERROR, not possible to insert"));
 }
     //GENERIC_ADD_UINT32(BBL_CADDRESS(head),(BBL_NINS(head)-1)<<2)
 
}

void MipsCfgInitLayoutOrder(t_cfg * cfg)
{
  t_bbl * i_bbl;
  
  number_of_layouted_chains = 0;

  CFG_FOREACH_BBL(cfg,i_bbl)
    if (!BBL_PREV_IN_CHAIN(i_bbl) && (!BBL_FUN(i_bbl) || !BBL_IS_LAST(i_bbl)))
      {
	number_of_layouted_chains++;
      }

  layout = (layouted_chain**) malloc (number_of_layouted_chains*sizeof(layouted_chain*));
  
  number_of_layouted_chains = 0;

  CFG_FOREACH_BBL(cfg,i_bbl)
    if (!BBL_PREV_IN_CHAIN(i_bbl) && (!BBL_FUN(i_bbl) || !BBL_IS_LAST(i_bbl)))
      {
	layout[number_of_layouted_chains] = (layouted_chain*) malloc (sizeof(layouted_chain));
	layout[number_of_layouted_chains]->links_to_all_prev_chains = 0;
	layout[number_of_layouted_chains]->links_to_all_next_chains = 0;
	layout[number_of_layouted_chains]->links_to_next_chain = 0;
	layout[number_of_layouted_chains]->links_to_prev_chain = 0;
	layout[number_of_layouted_chains]->bbl = i_bbl;
	i_bbl->layout_number = number_of_layouted_chains;
	number_of_layouted_chains++;
      }
}


static void CfgDetermineLayoutOrder(t_cfg * cfg)
{ 
  int i;

  MipsCfgInitLayoutOrder(cfg);

  
  for (i=1;i<number_of_layouted_chains;i++)
    MergeTwoChains(layout[0]->bbl,layout[i]->bbl);


}


/* dit is de eigenlijke lay-out procedure */

t_uint32 MipsCfgLayout(t_object * obj, t_section *code, t_cfg * cfg)
{
  t_bbl * i_bbl, * j_bbl;
  t_cfg_edge * i_edge;
  t_uint32 size;
  t_uint32 nops_count=0;

  /* put each block in separate chain */

  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      BBL_NEXT_IN_CHAIN(i_bbl) = BBL_PREV_IN_CHAIN(i_bbl) = NULL;
      BBL_FIRST_IN_CHAIN(i_bbl) = BBL_LAST_IN_CHAIN(i_bbl) = i_bbl;
    }

  /* check for more than one incoming fall_through path */

  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      j_bbl = NULL;
      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  if (EDGE_CAT(i_edge)==ET_FUNLINK ||
	      EDGE_CAT(i_edge)==ET_SWILINK ||
	      EDGE_CAT(i_edge)==ET_IPFUNLINK ||
	      EDGE_CAT(i_edge)==ET_IPSWILINK ||
	      EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
	      EDGE_CAT(i_edge)==ET_IPFALLTHRU
	      )
	    {
	      if (j_bbl)
		if (j_bbl!=CFG_EDGE_HEAD(i_edge))
		  {
		    BblPrint(i_bbl,PRINT_INSTRUCTIONS|PRINT_SUCCESSOR);
		    BblPrint(j_bbl,PRINT_INSTRUCTIONS|PRINT_SUCCESSOR);
		    BblPrint(CFG_EDGE_HEAD(i_edge),PRINT_INSTRUCTIONS|PRINT_SUCCESSOR);
		    FATAL(("two pred edges\n"));
		  }

	      j_bbl = CFG_EDGE_HEAD(i_edge);
	    }
	}
    }

  /* check for more then one outgoing fall_through path */
  /* also combine blocks involved in fall_through paths in one chain */

  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      j_bbl = NULL;
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	{
	  if (EDGE_CAT(i_edge)==ET_FUNLINK ||
	      EDGE_CAT(i_edge)==ET_SWILINK ||
	      EDGE_CAT(i_edge)==ET_IPFUNLINK ||
	      EDGE_CAT(i_edge)==ET_IPSWILINK ||
	      EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
	      EDGE_CAT(i_edge)==ET_IPFALLTHRU
	      )
	    {
	      if (j_bbl)
		if (j_bbl!=CFG_EDGE_TAIL(i_edge))
		  FATAL(("two succ edges\n"));

	      j_bbl = CFG_EDGE_TAIL(i_edge);
	    }
	}
      if (j_bbl)
	{
	  MergeTwoChains(i_bbl,j_bbl);
	  if(
	      i_bbl->ins_last && j_bbl->ins_first &&
	      (
 	       (
	       INS_TYPE(i_bbl->ins_last)==IT_LOAD &&
	       MIPSINSRT(i_bbl->ins_last)==MIPSINSRS(j_bbl->ins_first)
	       )
	       ||
	       (
	       INS_TYPE(i_bbl->ins_last)==IT_LOAD &&
	       MIPSINSRT(i_bbl->ins_last)==MIPSINSRT(j_bbl->ins_first)
	       )
	      )
	    )
	  {
	    t_ins * nop_ins;
	    nop_ins= InsNewForBbl(i_bbl);
	    RELOCATABLE_REFERS_TO(nop_ins)=NULL;
	    INS_OLD_ADDRESS(nop_ins)=GENERIC_ADD_UINT32(INS_OLD_ADDRESS(i_bbl->ins_last),4);
	    MipsInstructionMakeNoop(nop_ins);
	    InsInsertAfter(nop_ins,i_bbl->ins_last);
	    nops_count++;
	  }
	}
    }
  printf("Nops inserted, introduced through fallthrough edges: %d\n",nops_count);

#ifdef OBFUSCATION_SUPPORT
  if(!strcmp(global_options.obfuscation,"branch_obf"))
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      j_bbl = NULL;
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	{
	  if (EDGE_CAT(i_edge)==ET_JUMP ||
	      EDGE_CAT(i_edge)==ET_IPJUMP
	      )
	    {
	      if(i_bbl->ins_last && !MipsInsIsIndirectJump(i_bbl->ins_last))
	      {
		j_bbl = CFG_EDGE_TAIL(i_edge);
	      if(i_bbl && j_bbl && BBL_FIRST_IN_CHAIN(i_bbl)!=BBL_FIRST_IN_CHAIN(j_bbl))
		MergeTwoChains(i_bbl,j_bbl);
	      }
	    }
	}
    }

#endif

  /* now the order in which chains will be linked into 1 big chain is
     determined, and the chains are merged */

  CfgDetermineLayoutOrder(cfg);

  MipsAddressProducers(layout[0]->bbl/*T_BBL(GRAPH_NODE_FIRST(cfg))*/,cfg);
  PlaceSections(obj,FALSE);
  /* chain should be complete now */
  MipsAddressProducers(layout[0]->bbl/*T_BBL(GRAPH_NODE_FIRST(cfg))*/,cfg);

  DiabloPrint(stdout,"Given new addresses to the Basic Blocks...\n");

  
  t_section * sec;
  int i;
  OBJECT_FOREACH_SECTION(obj,sec,i)
  {
    DiabloPrint(stdout,"Section %s type %c old @G new @G (osize @G nsize @G)\n",sec->name,
	sec->type,SECTION_GET_OLD_ADDRESS(sec), SECTION_GET_CADDRESS(sec),
	SECTION_OLD_SIZE(sec), SECTION_CSIZE(sec));
  }

  MipsRelocate(cfg);
  
  cfg->sec->obj->entry =  BBL_CADDRESS(cfg->entries->entry_bbl);
  DiabloPrint(stdout,"Setting entry point @G\n", cfg->sec->obj->entry);

  size = 0;
  
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    size+=BBL_NINS(i_bbl)*4;
  }
  
  printf("%d total number of 4-byte words of the code section\n",size/4);
  DiabloPrint(stdout,"New entry point @G\n",cfg->sec->obj->entry);

#if 1
  {
    FILE *f = fopen("./final.list","w");

      CFG_FOREACH_BBL(cfg,i_bbl)
    {
      DiabloPrint(f,"bbl @B\n",i_bbl);
    }
  }
#endif


  return size;
}



void MipsRelocate(t_cfg * cfg)
{
  t_reloc * reloc;
  t_bbl * i_bbl;
  t_ins * i_ins;
  t_cfg_edge * i_edge;

  /* adapt the displacements encoded in direct control flow transfer instructions {{{ */
  ANALYSIS_START("Relocating branch displacements");
  
/* replacement for jumps */
  CFG_FOREACH_EDGE(cfg,i_edge)
  {
  t_bbl * head = CFG_EDGE_HEAD(i_edge);
  t_ins * ins = head->ins_first;
  t_ins * h_ins;
  int i=0;
  if(ins!=0){
  }
    if (EDGE_CAT(i_edge)==ET_JUMP || EDGE_CAT(i_edge)==ET_IPJUMP || EDGE_CAT(i_edge)==ET_CALL) 
      if((CFG_EDGE_HEAD(i_edge))->nins>1)
      {
    BBL_FOREACH_INS(head,i_ins)
    {
	    if(i==head->nins-2)
	    h_ins=i_ins;
	    i++;
	    
    }
      if (MIPSINSOPCODE(h_ins)==MIPS_J||
        MIPSINSOPCODE(h_ins)==MIPS_JAL)
	{
		  UpdateJDisplacement(i_edge,h_ins);
	}
      if (MIPSINSOPCODE(h_ins)==MIPS_BLTZ||
        MIPSINSOPCODE(h_ins)==MIPS_BGEZ||
        MIPSINSOPCODE(h_ins)==MIPS_BLTZAL||
        MIPSINSOPCODE(h_ins)==MIPS_BGEZAL||
        MIPSINSOPCODE(h_ins)==MIPS_BEQ||
        MIPSINSOPCODE(h_ins)==MIPS_BNE||
        MIPSINSOPCODE(h_ins)==MIPS_BLEZ||
        MIPSINSOPCODE(h_ins)==MIPS_BC1T||
        MIPSINSOPCODE(h_ins)==MIPS_BC1F||
        MIPSINSOPCODE(h_ins)==MIPS_BGTZ
	)
	{
	  UpdateBranchDisplacement(i_edge,h_ins);
	}
      }
  }
  
  ANALYSIS_SIMPLE_END("Relocating branch displacements");
  /* }}} */
  /* relocate the addresses stored in non-code sections {{{ */
  ANALYSIS_START("Relocating in non-code sections");
  OBJECT_FOREACH_RELOC(cfg->sec->obj,reloc)
  {
    if (!REL_TO_RELOCATABLE(reloc)) continue;
    if (REL_TO_RELOCATABLE(reloc)->type==RT_INS) FATAL(("relocs to ins"));
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_SUBSECTION)
    {
      if ((REL_TO_RELOCATABLE(reloc)->type==RT_BBL) || (REL_TO_RELOCATABLE(reloc)->type==RT_SUBSECTION)|| (REL_TO_RELOCATABLE(reloc)->type==RT_SECTION))
      {
	t_address ret;
	ret=StackExec(reloc->code,reloc,T_SECTION(REL_FROM(reloc))->data,TRUE,0);
      }
      else FATAL(("IMPLEMENT 2 > %d",REL_TO_RELOCATABLE(reloc)->type));
    }
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_BBL) continue;
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_INS) continue;
    else if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_SECTION) FATAL(("reloc from section"));
    else FATAL(("Unknown reloc @R\n",reloc));
  }
  ANALYSIS_SIMPLE_END("Relocating in non-code sections");
  /* }}} */
  /* relocate the addresses stored in data blocks in the text section {{{ */

  ANALYSIS_START("Relocating in code sections");
#ifdef OBFUSCATION_SUPPORT

  if(!strcmp(global_options.obfuscation,"branch_obf"))
  {
    t_uint32 X=100;
    t_uint32 relocs [X];
    t_uint32 tel=0;
    char out[8];
    for(tel;tel<X;tel++)
      relocs[tel]=0;
    CFG_FOREACH_BBL(cfg,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
      {      
	if (RELOCATABLE_REFERS_TO(i_ins))
	{
	  t_mips_ins rins;
	  char tmp[4];
	  t_address ret;
	  t_reloc * reloc = RELOCATABLE_REFERS_TO(i_ins)->rel;

	  MipsAssembleOne(i_ins,tmp);
	  
	  tel=0;

	  if((INS_ATTRIB(i_ins)&IF_BRANCH_OBF)!=0)
	  {
	    if(relocs[0]==0)
	      relocs[0]=GENERIC_ADD_UINT32(GENERIC_ADD(REL_TO_RELOCATABLE(reloc)->caddress,reloc->to_offset),8);
	    reloc->addend=relocs[0];
	  }

	  ret=StackExec(reloc->code,reloc,tmp,TRUE,0);
	  if (INS_TYPE(i_ins)==IT_DATA)
	  {
	    MipsDisassembleData(T_INS(&rins),*(t_uint32 *) tmp,MIPS_DATA);
	  }
	  else
	    mips_opcode_table[MIPSINSOPCODE(i_ins)].Disassemble((t_ins *) &rins,*(t_uint32 *) tmp,MIPSINSOPCODE(i_ins));
	  MIPSINSIMMEDIATE(i_ins)=MIPSINSIMMEDIATE(&rins);

	  if((INS_ATTRIB(i_ins)&IF_BRANCH_OBF)!=0)
	  {
	    t_bool satisfy=FALSE;
	    do{
	      if(reloc->addend+(t_int32)((t_int16)MIPSINSIMMEDIATE(i_ins))!=GENERIC_ADD(REL_TO_RELOCATABLE(reloc)->caddress,reloc->to_offset))
	      {
		tel++;
		if(relocs[tel]==0)
		  relocs[tel]=GENERIC_ADD_UINT32(GENERIC_ADD(REL_TO_RELOCATABLE(reloc)->caddress,reloc->to_offset),8);
		reloc->addend=relocs[tel];
		if(INS_PREV(i_ins) && INS_PREV(INS_PREV(i_ins)) && MIPSINSOPCODE(INS_PREV(INS_PREV(i_ins)))==MIPS_LW &&
		    MIPSINSRT(INS_PREV(INS_PREV(i_ins)))==MIPSINSRT(i_ins))
		{
		  MIPSINSIMMEDIATE(INS_PREV(INS_PREV(i_ins)))+=4;
		}
		else
		  FATAL(("ERROR"));
		ret=StackExec(reloc->code,reloc,tmp,TRUE,0);
		mips_opcode_table[MIPSINSOPCODE(i_ins)].Disassemble((t_ins *) &rins,*(t_uint32 *) tmp,MIPSINSOPCODE(i_ins));
		MIPSINSIMMEDIATE(i_ins)=MIPSINSIMMEDIATE(&rins);
		if(reloc->addend+(t_int32)((t_int16)MIPSINSIMMEDIATE(i_ins))!=GENERIC_ADD(REL_TO_RELOCATABLE(reloc)->caddress,reloc->to_offset))
		  satisfy=TRUE;
		else
		  satisfy=FALSE;
		  
	      }
	    } while(satisfy);
	  }
	}
      }
    }
//    DiabloPrint(stdout,"TELLER:%d\n",tel);
    //AddressSpecialPrint()
    {
      t_uint32 j,l;
      t_uint32 k[]={4,0,5,1,6,2,7,3};

      DiabloPrint(stdout,"CODE:");
      l=-1;
      do
      {
	l++;
	PrintCode(relocs[l]);
      }while(relocs[l+1]!=0);
      DiabloPrint(stdout,"\n");
    DiabloPrint(stdout,"TELLER:%d\n",l);
    l=-1;
      do
      {
	l++;
	DiabloPrint(stdout,"address %d: %x\n",l,relocs[l]);
      }while(relocs[l+1]!=0);

      //Aantal inputs correct zetten.
      {
	t_ins * start=cfg->entries->entry_bbl->ins_first;
	for(;;)
	{
	  if(MIPSINSOPCODE(start)==MIPS_ADDIU&&MIPSINSRT(start)==MIPS_REG_T4)
	  {
	    MIPSINSIMMEDIATE(start)=l+1;
	    break;
	  }
	  start=INS_NEXT(start);
	}
      }
    }

  }
  else
#endif
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    BBL_FOREACH_INS(i_bbl,i_ins)
    {      
      if (RELOCATABLE_REFERS_TO(i_ins))
      {
	t_mips_ins rins;
	char tmp[4];
	t_address ret;
	t_reloc * reloc = RELOCATABLE_REFERS_TO(i_ins)->rel;

	MipsAssembleOne(i_ins,tmp);
	ret=StackExec(reloc->code,reloc,tmp,TRUE,0);
	if (INS_TYPE(i_ins)==IT_DATA)
	{
	  MipsDisassembleData(T_INS(&rins),*(t_uint32 *) tmp,MIPS_DATA);
	}
	else
	  mips_opcode_table[MIPSINSOPCODE(i_ins)].Disassemble((t_ins *) &rins,*(t_uint32 *) tmp,MIPSINSOPCODE(i_ins));
	MIPSINSIMMEDIATE(i_ins)=MIPSINSIMMEDIATE(&rins);
      }
    }
  }
  ANALYSIS_SIMPLE_END("Relocating in code sections");
  
  /* }}} */
  /* set new entry point for the object */
  cfg->sec->obj->entry =  BBL_CADDRESS(cfg->entries->entry_bbl);
}
/*}}}*/


