/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

/* Includes {{{ */
#include <stdlib.h>
#include "../../utils/util_status.h"
#include "../../kernel/diablo_iterators.h"
//#include "../../obfuscation/obf_opposite.h"
#ifdef MIPS_SUPPORT
#include "mips_opcodes.h"
#include "mips_registers.h"
#endif
/* }}} */

void GetArchs(t_bbl * bbl,t_section *sec)
{

  t_ins * ins;

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADDIU;
  MIPSINSRS(ins)=MIPS_REG_T3;
  MIPSINSRT(ins)=MIPS_REG_T3;
  MIPSINSIMMEDIATE(ins)=0x4;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_BRANCH;
  MIPSINSOPCODE(ins)=MIPS_BNE;
  MIPSINSRS(ins)=MIPS_REG_ZERO;
  MIPSINSRT(ins)=MIPS_REG_T4;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADDIU;
  MIPSINSRS(ins)=MIPS_REG_T4;
  MIPSINSRT(ins)=MIPS_REG_T4;
  MIPSINSIMMEDIATE(ins)=0xffff;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADDIU;
  MIPSINSRS(ins)=MIPS_REG_T8;
  MIPSINSRT(ins)=MIPS_REG_T8;
  MIPSINSIMMEDIATE(ins)=0x8;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_STORE;
  MIPSINSOPCODE(ins)=MIPS_SW;
  MIPSINSRS(ins)=MIPS_REG_T3;
  MIPSINSRT(ins)=MIPS_REG_T7;
  InsPrependToBbl(ins,bbl);


  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_OR;
  MIPSINSRS(ins)=MIPS_REG_T6;
  MIPSINSRD(ins)=MIPS_REG_T7;
  MIPSINSRT(ins)=MIPS_REG_T7;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_SLL;
  MIPSINSRD(ins)=MIPS_REG_T7;
  MIPSINSSA(ins)=0x4;
  MIPSINSRT(ins)=MIPS_REG_T7;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_AND;
  MIPSINSRS(ins)=MIPS_REG_T7;
  MIPSINSRD(ins)=MIPS_REG_T7;
  MIPSINSRT(ins)=MIPS_REG_T5;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSPROCESSOR(ins)=CPU;
  MIPSINSOPCODE(ins)=MIPS_AND;
  MIPSINSRS(ins)=MIPS_REG_T6;
  MIPSINSRD(ins)=MIPS_REG_T6;
  MIPSINSRT(ins)=MIPS_REG_T5;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADDIU;
  MIPSINSRS(ins)=MIPS_REG_T5;
  MIPSINSRT(ins)=MIPS_REG_T5;
  MIPSINSIMMEDIATE(ins)=0x0f0f;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_LUI;
  MIPSINSRT(ins)=MIPS_REG_T5;
  MIPSINSIMMEDIATE(ins)=0x0f0f;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_LOAD;
  MIPSINSOPCODE(ins)=MIPS_LW;
  MIPSINSBASE(ins)=MIPS_REG_T8;
  MIPSINSRT(ins)=MIPS_REG_T6;
  MIPSINSIMMEDIATE(ins)=0x4;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_LOAD;
  MIPSINSOPCODE(ins)=MIPS_LW;
  MIPSINSBASE(ins)=MIPS_REG_T8;
  MIPSINSRT(ins)=MIPS_REG_T7;
  InsPrependToBbl(ins,bbl);

/*  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_LOAD;
  MIPSINSOPCODE(ins)=MIPS_LW;
  MIPSINSBASE(ins)=MIPS_REG_SP;
  MIPSINSRT(ins)=MIPS_REG_T8;
  MIPSINSIMMEDIATE(ins)=0x8;
  InsPrependToBbl(ins,bbl);*/

}

void PostArchs(t_bbl * bbl,t_section *sec)
{
}

void PreArchs(t_bbl * bbl,t_section *sec)
{

  t_section * psec=SECTION_PARENT_SECTION(sec);
  
  t_ins * ins;

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADDIU;
  MIPSINSRS(ins)=MIPS_REG_T3;
  MIPSINSRT(ins)=MIPS_REG_T3;
  InsPrependToBbl(ins,bbl);
  //RelocTableAddRelocToRelocatable(sec->obj->reloc_table,0x0,T_RELOCATABLE(ins),0x0,T_RELOCATABLE(sec),0x0,FALSE,NULL,NULL,T_RELOCATABLE(sec),"AS+ \\ sffff& = l iffff0000 &|w \\l sffff &-$", TRUE);
  RelocTableAddRelocToRelocatable(psec->obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(sec),NULL_GENERIC,FALSE,NULL,NULL,NULL,"AS+ \\ sffff& = l iffff0000 &|w \\l sffff &-$", TRUE);
  //NULL_GENERIC

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_LUI;
  MIPSINSRT(ins)=MIPS_REG_T3;
  InsPrependToBbl(ins,bbl);
  //RelocTableAddRelocToRelocatable(sec->obj->reloc_table,0x0,T_RELOCATABLE(ins),0x0,T_RELOCATABLE(sec),0x0,FALSE,NULL,NULL,T_RELOCATABLE(sec),"A #0f S+ = sffff& #0f  - s0010> \\ sffff& = l iffff0000&|w\\l sffff &-$", TRUE);
  RelocTableAddRelocToRelocatable(psec->obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(sec),NULL_GENERIC,FALSE,NULL,NULL,NULL,"A #0f S+ = sffff& #0f  - s0010> \\ sffff& = l iffff0000&|w\\l sffff &-$", TRUE);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_LOAD;
  MIPSINSOPCODE(ins)=MIPS_LW;
  MIPSINSBASE(ins)=MIPS_REG_T8;
  MIPSINSRT(ins)=MIPS_REG_T8;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADD;
  MIPSINSRT(ins)=MIPS_REG_T8;
  MIPSINSRD(ins)=MIPS_REG_T8;
  MIPSINSRS(ins)=MIPS_REG_SP;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_SLL;
  MIPSINSRT(ins)=MIPS_REG_T8;
  MIPSINSRD(ins)=MIPS_REG_T8;
  MIPSINSSA(ins)=0x2;
  InsPrependToBbl(ins,bbl);

  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_LOAD;
  MIPSINSOPCODE(ins)=MIPS_LW;
  MIPSINSBASE(ins)=MIPS_REG_SP;
  MIPSINSRT(ins)=MIPS_REG_T8;
  InsPrependToBbl(ins,bbl);
  
  ins=InsNewForBbl(bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADDIU;
  MIPSINSRS(ins)=MIPS_REG_ZERO;
  MIPSINSRT(ins)=MIPS_REG_T4;
  MIPSINSIMMEDIATE(ins)=0x2;
  InsPrependToBbl(ins,bbl);

}

t_uint32 FindFreeReg(t_regset dead)
{
  t_uint32 i=0;
  
  REGSET_FOREACH_REG(dead,i)
  {
    if(i!=MIPS_REG_R0 && i !=MIPS_REG_K0 && i != MIPS_REG_K1)
      break;
  }

  return i;
}

t_bool RegisterCanBeUsed(t_uint32 reg)
{
  if(reg>MIPS_REG_R0 && reg<=MIPS_REG_RA)
    return TRUE;
  return FALSE;
}

t_bbl * IndirectJumpBbl(t_cfg * cfg,t_section * added, t_uint32 free_reg,t_cfg_edge * jump_edge)
{
  t_object * obj=cfg->sec->obj;
  t_bbl * new_bbl= BblNew(cfg);
  t_ins * ins;
  t_reloc * corr, *corr2;

  ins=InsNewForBbl(new_bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSPROCESSOR(ins)=CPU;
  MIPSINSOPCODE(ins)=MIPS_LUI;
  MIPSINSRT(ins)=free_reg;
  InsAppendToBbl(ins,new_bbl);
  corr=RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(added),NULL_GENERIC,FALSE,NULL,NULL,NULL,"A #0f S+ = sffff& #0f  - s0010> \\ sffff& = l iffff0000&|w\\l sffff &-$", TRUE);

  ins=InsNewForBbl(new_bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_LOAD;
  MIPSINSOPCODE(ins)=MIPS_LW;
  MIPSINSRS(ins)=free_reg;
  MIPSINSRT(ins)=free_reg;
  InsAppendToBbl(ins,new_bbl);
  corr2=RelocTableAddRelocToRelocatable(cfg->sec->obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,T_RELOCATABLE(added),NULL_GENERIC,FALSE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);

  ins=InsNewForBbl(new_bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_DATAPROC;
  MIPSINSOPCODE(ins)=MIPS_ADDIU;
  MIPSINSRS(ins)=free_reg;
  MIPSINSRT(ins)=free_reg;
  InsAppendToBbl(ins,new_bbl);
  corr=RelocTableAddRelocToRelocatable(obj->reloc_table,NULL_GENERIC,T_RELOCATABLE(ins),NULL_GENERIC,REL_TO_RELOCATABLE(jump_edge),NULL_GENERIC,FALSE,NULL,NULL,NULL,"SA-\\" WRITE_S16, TRUE);
  INS_ATTRIB(ins)|=IF_BRANCH_OBF;

  ins=InsNewForBbl(new_bbl);
  MipsInstructionMakeNoop(ins);
  INS_TYPE(ins)=IT_BRANCH;
  MIPSINSPROCESSOR(ins)=CPU;
  MIPSINSOPCODE(ins)=MIPS_JR;
  MIPSINSRS(ins)=free_reg;
  INS_ATTRIB(ins)|=IF_PCDEF;
  InsAppendToBbl(ins,new_bbl);

  return new_bbl;
}


/* vim: set shiftwidth=2 foldmethod=marker: */
