/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "mips_opcodes.h"
#include "mips_disassemble_one.h"

const mips_opcode mips_opcode_table[]={
  /* NOP */
  {0xffffffff, 0x00000000,"NOP", MipsDisassembleNOP },
  /*Aligned Load/Store instructions  */
  {0xfc000000, 0x80000000,"LB",MipsDisassembleDataTransferL },
  {0xfc000000, 0x84000000,"LH",MipsDisassembleDataTransferL },
  {0xfc000000, 0x8c000000,"LW",MipsDisassembleDataTransferL },
  {0xfc000000, 0x90000000,"LBU",MipsDisassembleDataTransferL },
  {0xfc000000, 0x94000000,"LHU",MipsDisassembleDataTransferL },
  {0xfc000000, 0xa0000000,"SB",MipsDisassembleDataTransferS },
  {0xfc000000, 0xa4000000,"SH",MipsDisassembleDataTransferS },
  {0xfc000000, 0xac000000,"SW",MipsDisassembleDataTransferS },
  /* Unaligned Load/Store Instructions */
  {0xfc000000, 0xb8000000,"SWR",MipsDisassembleDataTransferS },
  {0xfc000000, 0x88000000,"LWL",MipsDisassembleDataTransferL },
  {0xfc000000, 0x98000000,"LWR",MipsDisassembleDataTransferL },
  {0xfc000000, 0xa8000000,"SWL",MipsDisassembleDataTransferS },
  /* DataProcessingImmediate */
  {0xfc000000, 0x20000000,"ADDI",MipsDisassembleDataProcessingI },
  {0xfc000000, 0x24000000,"ADDIU",MipsDisassembleDataProcessingI },
  {0xfc000000, 0x28000000,"SLTI",MipsDisassembleDataProcessingI },
  {0xfc000000, 0x2c000000,"SLTIU",MipsDisassembleDataProcessingI },
  {0xfc000000, 0x30000000,"ANDI",MipsDisassembleDataProcessingI },
  {0xfc000000, 0x34000000,"ORI",MipsDisassembleDataProcessingI },
  {0xfc000000, 0x38000000,"XORI",MipsDisassembleDataProcessingI },
  {0xfc000000, 0x3c000000,"LUI",MipsDisassembleDataProcessingI },
  /* DataProcessing */
  {0xfc00003f, 0x00000020,"ADD",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x00000021,"ADDU",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x00000022,"SUB",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x00000023,"SUBU",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x00000024,"AND",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x00000025,"OR",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x00000026,"XOR",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x00000027,"NOR",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x0000002a,"SLT",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x0000002b,"SLTU",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x70000020,"CLZ",MipsDisassembleDataProcessing },
  {0xfc00003f, 0x70000021,"CLO",MipsDisassembleDataProcessing} ,
  /* Shift Instructions */ 
  {0xfc00003f, 0x00000000,"SLL", MipsDisassembleShift },
  {0xfc00003f, 0x00000002,"SRL",MipsDisassembleShift },
  {0xfc00003f, 0x00000003,"SRA",MipsDisassembleShift },
  {0xfc00003f, 0x00000004,"SLLV",MipsDisassembleShift },
  {0xfc00003f, 0x00000006,"SRLV",MipsDisassembleShift },
  {0xfc00003f, 0x00000007,"SRAV",MipsDisassembleShift },
  /* Multiply/Divide */
  {0xfc00003f, 0x00000010,"MFHI",MipsDisassembleMultiply },
  {0xfc00003f, 0x00000011,"MTHI",MipsDisassembleMultiply },
  {0xfc00003f, 0x00000012,"MFLO",MipsDisassembleMultiply },
  {0xfc00003f, 0x00000013,"MTLO",MipsDisassembleMultiply },
  {0xfc00003f, 0x00000018,"MULT",MipsDisassembleMultiply },
  {0xfc00003f, 0x00000019,"MULTU",MipsDisassembleMultiply },
  {0xfc00003f, 0x0000001a,"DIV",MipsDisassembleMultiply },
  {0xfc00003f, 0x0000001b,"DIVU",MipsDisassembleMultiply },
  {0xfc00003f, 0x70000000,"MADD",MipsDisassembleMultiply },
  {0xfc00003f, 0x70000001,"MADDU",MipsDisassembleMultiply },
  {0xfc00003f, 0x70000002,"MUL",MipsDisassembleMultiply },
  {0xfc00003f, 0x70000004,"MSUB",MipsDisassembleMultiply },
  {0xfc00003f, 0x70000005,"MSUBU",MipsDisassembleMultiply },
  /* Unconditional Jump */
  {0xfc00003f, 0x00000008,"JR",MipsDisassembleUnconditionalJump } ,
  {0xfc00003f, 0x00000009,"JALR",MipsDisassembleUnconditionalJump },
  {0xfc000000, 0x08000000,"J",MipsDisassembleUnconditionalJump },
  {0xfc000000, 0x0c000000,"JAL",MipsDisassembleUnconditionalJump },
  /* Branch */
  {0xfc1f0000, 0x04000000,"BLTZ",MipsDisassembleBranch },
  {0xfc1f0000, 0x04010000,"BGEZ",MipsDisassembleBranch },
  {0xfc1f0000, 0x04100000,"BLTZAL",MipsDisassembleBranch },
  {0xfc1f0000, 0x04110000,"BGEZAL",MipsDisassembleBranch },
  {0xfc000000, 0x10000000,"BEQ",MipsDisassembleBranch },
  {0xfc000000, 0x14000000,"BNE",MipsDisassembleBranch },
  {0xfc000000, 0x18000000,"BLEZ",MipsDisassembleBranch },
  {0xfc000000, 0x1c000000,"BGTZ",MipsDisassembleBranch },
  
  /* Controle instructies die we niet ver uitwerken */
  
  /* Trap on condition comparing two registers*/
  /* Wensen we dit wel uit te werken: MipsDisassembleTrapR */
  {0xfc00003f, 0x00000030,"TGE",MipsDisassembleUnsupported  },
  {0xfc00003f, 0x00000031,"TGEU",MipsDisassembleUnsupported  },
  {0xfc00003f, 0x00000032,"TLT",MipsDisassembleUnsupported  },
  {0xfc00003f, 0x00000033,"TLTU",MipsDisassembleUnsupported  },
  {0xfc00003f, 0x00000034,"TEQ",MipsDisassembleUnsupported  },
  {0xfc00003f, 0x00000036,"TNE",MipsDisassembleUnsupported  },
  /* Trap on condition comparing an immediate value */
  /* Wensen we dit wel uit te werken: MipsDisassembleTrapI */
  {0xfc1f0000, 0x04080000,"TGEI",MipsDisassembleUnsupported },
  {0xfc1f0000, 0x04090000,"TGEIU",MipsDisassembleUnsupported },
  {0xfc1f0000, 0x040a0000,"TLTI",MipsDisassembleUnsupported },
  {0xfc1f0000, 0x040b0000,"TLTIU",MipsDisassembleUnsupported },
  {0xfc1f0000, 0x040c0000,"TEQI",MipsDisassembleUnsupported },
  {0xfc1f0000, 0x040e0000,"TNEI",MipsDisassembleUnsupported },
  
  /* Conditional move instructions */
  {0xfc00003f, 0x0000000a,"MOVZ",MipsDisassembleConditionalMove},
  {0xfc00003f, 0x0000000b,"MOVN",MipsDisassembleConditionalMove},
  /* Atomic Update CPU Load/Store Instructions */
  {0xfc000000, 0xc0000000,"LL",MipsDisassembleDataTransferL},
  {0xfc000000, 0xe0000000,"SC",MipsDisassembleDataTransferS},
 
  /* COP0 volgens rs */
  {0xffe00000, 0x40000000,"MFC0",MipsDisassembleCOP0DataProcessing},
  {0xffe00000, 0x40800000,"MTC0",MipsDisassembleCOP0DataProcessing},
  /* COP0 volgens rs=CO */
  {0xfe00003f, 0x42000001,"TLBR",MipsDisassembleCOP0},
  {0xfe00003f, 0x42000002,"TLBWI",MipsDisassembleCOP0},
  {0xfe00003f, 0x42000006,"TLBWR",MipsDisassembleCOP0},
  {0xfe00003f, 0x42000008,"TLBP",MipsDisassembleCOP0},
  {0xfe00003f, 0x42000018,"ERET",MipsDisassembleCOP0},
  {0xfe00003f, 0x4200001f,"DERET",MipsDisassembleCOP0},
  {0xfe00003f, 0x42000020,"WAIT",MipsDisassembleCOP0},
  
  /* COP1 Load Store */
  {0xfc000000, 0xc4000000,"LWC1",MipsDisassembleCOP1DataTransferL},
  {0xfc000000, 0xd4000000,"LDC1",MipsDisassembleCOP1DataTransferL},
  {0xfc000000, 0xe4000000,"SWC1",MipsDisassembleCOP1DataTransferS},
  {0xfc000000, 0xf4000000,"SDC1",MipsDisassembleCOP1DataTransferS},
  /* COP1 volgens rs move from/to CPU/COP1 
   * opgepast: deze moeten voor de CVT's en Arithmetic Operations komen */
  {0xffe00000, 0x44000000,"MFC1",MipsDisassembleCOP1DataProcessingFT},
  {0xffe00000, 0x44400000,"CFC1",MipsDisassembleCOP1DataProcessingFT},
  {0xffe00000, 0x44800000,"MTC1",MipsDisassembleCOP1DataProcessingFT},
  {0xffe00000, 0x44c00000,"CTC1",MipsDisassembleCOP1DataProcessingFT},
  /* COP1 volgens rs=BC1 Opgespast: deze moeten voor de CVT's komen!! */
  {0xffe30000, 0x45000000,"BC1F",MipsDisassembleCOP1Branch},
  {0xffe30000, 0x45010000,"BC1T",MipsDisassembleCOP1Branch},
  /* COP1 IEEE Arithmetic Operations */
  {0xfc00003f, 0x44000000,"ADD.",MipsDisassembleCOP1DataProcessing},
  {0xfc00003f, 0x44000001,"SUB.",MipsDisassembleCOP1DataProcessing},
  {0xfc00003f, 0x44000002,"MUL.",MipsDisassembleCOP1DataProcessing},
  {0xfc00003f, 0x44000003,"DIV.",MipsDisassembleCOP1DataProcessing},
  {0xfc00003f, 0x44000004,"SQRT.",MipsDisassembleCOP1DataProcessing},
  {0xfc00003f, 0x44000005,"ABS.",MipsDisassembleCOP1DataProcessing},
  {0xfc00003f, 0x44000007,"NEG.",MipsDisassembleCOP1DataProcessing},
  {0xfc000030, 0x44000030,"C.",MipsDisassembleCOP1DataProcessing},
  /* COP1 Conversion Operations Using the FCSR Rounding Mode */
  {0xfc00003f, 0x44000021,"CVT.D.",MipsDisassembleCOP1DataProcessingRound},
  {0xfc00003f, 0x44000020,"CVT.S.",MipsDisassembleCOP1DataProcessingRound},
  {0xfc00003f, 0x44000024,"CVT.W.",MipsDisassembleCOP1DataProcessingRound},
  /* COP1 Conversion Operations Using a Directed Rounding Mode */ 
  {0xfc00003f, 0x4400000c,"ROUND.W.",MipsDisassembleCOP1DataProcessingRound},
  {0xfc00003f, 0x4400000d,"TRUNC.W.",MipsDisassembleCOP1DataProcessingRound},
  {0xfc00003f, 0x4400000e,"CEIL.W.",MipsDisassembleCOP1DataProcessingRound},
  {0xfc00003f, 0x4400000f,"FLOOR.W.",MipsDisassembleCOP1DataProcessingRound},
  /* COP1 Formatted Operand Move Instruction */
  {0xfc00003f, 0x44000006,"MOV.",MipsDisassembleCOP1DataProcessingRound},
  /* COP1 volgens rs=S en MOVCF */
  {0xfc01003f, 0x44000011,"MOVF.",MipsDisassembleCOP1DataProcessingMOV},
  {0xfc01003f, 0x44010011,"MOVT.",MipsDisassembleCOP1DataProcessingMOV},
  {0xfc00003f, 0x44000012,"MOVZ.",MipsDisassembleCOP1DataProcessingMOV},
  {0xfc00003f, 0x44000013,"MOVN.",MipsDisassembleCOP1DataProcessingMOV},
  /* COP1 Conditional move on COP1 True/False Instruction */
  {0xfc01003f, 0x00000001,"MOVF",MipsDisassembleCOP1DataProcessingMoveTF },
  {0xfc01003f, 0x00010001,"MOVT",MipsDisassembleCOP1DataProcessingMoveTF },
  
  /* Systeem functies */
  {0xfc00003f, 0x0000000c,"SYSCALL",MipsDisassembleBreak },
  {0xfc00003f, 0x0000000d,"BREAK",MipsDisassembleBreak },
  {0xfc00003f, 0x0000000f,"SYNC",MipsDisassembleUnsupported },
  {0xfc000000, 0xbc000000,"CACHE",MipsDisassembleUnsupported },
  {0xfc000000, 0xcc000000,"PREF",MipsDisassembleUnsupported },
 
  /* if it is an other instruction ... */
  {0x00000000, 0x00000000,"DATA?",MipsDisassembleUnsupported},
  
};  
/* vim: set shiftwidth=2: */
