/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../../config.h"
#ifndef MIPS_OPCODES_H
#define MIPS_OPCODES_H
#include "mips_instructions.h"

/*The structure of the different disassemble functions */

typedef void (* MipsDisassembleFunction) (t_ins *,t_uint32 instr,t_uint16 opc);

/* The opcode structure */

typedef struct _mips_opcode
{
  t_uint32 mask;
  t_uint32 opcode;
  t_string desc;
  MipsDisassembleFunction Disassemble;
} mips_opcode;

/* The opcode table */

extern const mips_opcode mips_opcode_table[];

/* The opcodes */
#define MIPS_NOP	0
#define MIPS_LB		1
#define MIPS_LH		2
#define MIPS_LW		3
#define MIPS_LBU	4
#define MIPS_LHU	5
#define MIPS_SB		6
#define MIPS_SH		7
#define MIPS_SW		8
#define MIPS_SWR	9
#define MIPS_LWL	10
#define MIPS_LWR	11
#define MIPS_SWL	12
#define MIPS_ADDI	13
#define MIPS_ADDIU	14
#define MIPS_SLTI	15
#define MIPS_SLTIU	16
#define MIPS_ANDI	17
#define MIPS_ORI	18
#define MIPS_XORI	19
#define MIPS_LUI	20
#define MIPS_ADD	21
#define MIPS_ADDU	22
#define MIPS_SUB	23
#define MIPS_SUBU	24
#define MIPS_AND	25
#define MIPS_OR		26
#define MIPS_XOR	27
#define MIPS_NOR	28
#define MIPS_SLT	29
#define MIPS_SLTU	30
#define MIPS_CLZ	31
#define MIPS_CLO	32
#define MIPS_SLL	33
#define MIPS_SRL	34
#define MIPS_SRA	35
#define MIPS_SLLV	36
#define MIPS_SRLV	37
#define MIPS_SRAV 	38
#define MIPS_MFHI	39
#define MIPS_MTHI	40
#define MIPS_MFLO	41
#define MIPS_MTLO	42
#define MIPS_MULT	43
#define MIPS_MULTU	44
#define MIPS_DIV	45
#define MIPS_DIVU	46
#define MIPS_MADD	47
#define MIPS_MADDU	48
#define MIPS_MUL	49
#define MIPS_MSUB	50
#define MIPS_MSUBU	51
#define MIPS_JR		52
#define MIPS_JALR	53
#define MIPS_J		54
#define MIPS_JAL	55
#define MIPS_BLTZ	56
#define MIPS_BGEZ	57
#define MIPS_BLTZAL	58
#define MIPS_BGEZAL	59
#define MIPS_BEQ	60
#define MIPS_BNE	61	
#define MIPS_BLEZ	62
#define MIPS_BGTZ	63
#define MIPS_TGE	64
#define MIPS_TGEU	65
#define MIPS_TLT	66
#define MIPS_TLTU	67
#define MIPS_TEQ	68
#define MIPS_TNE	69
#define MIPS_TGEI	70
#define MIPS_TGEIU	71
#define MIPS_TLTI	72
#define MIPS_TLTIU	73
#define MIPS_TEQI	74
#define MIPS_TNEI	75
#define MIPS_MOVZ	76
#define MIPS_MOVN	77
#define MIPS_LL		78
#define MIPS_SC		79
#define MIPS_MFC0	80
#define MIPS_MTC0	81
#define MIPS_TLBR	82
#define MIPS_TLBWI	83
#define MIPS_TLBWR	84
#define MIPS_TLBP	85
#define MIPS_ERET	86
#define MIPS_DERET	87
#define MIPS_WAIT	88
#define MIPS_LWC1	89
#define MIPS_LDC1	90
#define MIPS_SWC1	91
#define MIPS_SDC1	92
#define MIPS_MFC1	93
#define MIPS_CFC1	94
#define MIPS_MTC1	95
#define MIPS_CTC1	96
#define MIPS_BC1F	97
#define MIPS_BC1T	98
#define MIPS_ADD_	99
#define MIPS_SUB_	100
#define MIPS_MUL_	101
#define MIPS_DIV_	102
#define MIPS_SQRT	103
#define MIPS_ABS	104
#define MIPS_NEG	105
#define MIPS_C		106
#define MIPS_CVT_D	107
#define MIPS_CVT_S	108
#define MIPS_CVT_W	109
#define MIPS_ROUND_W	110
#define MIPS_TRUNC_W	111
#define MIPS_CEIL_W	112
#define MIPS_FLOOR_W	113
#define MIPS_MOV 	114
#define MIPS_MOVF_	115
#define MIPS_MOVT_	116
#define MIPS_MOVZ_	117
#define MIPS_MOVN_	118
#define MIPS_MOVF	119
#define MIPS_MOVT	120
#define MIPS_SYSCALL	121
#define MIPS_BREAK	122
#define MIPS_SYNC	123
#define MIPS_CACHE	124
#define MIPS_PREF	125
#define MIPS_DATA_	126
#define MIPS_DATA	127
#define MIPS_ADDRESS_PRODUCER 	128
#define MIPS_CONSTANT_PRODUCER	129
#define MIPS_NONE	130

#define C_F	0
#define C_UN	1
#define C_EQ	2
#define C_UEQ	3
#define C_OLT	4
#define C_ULT	5
#define C_OLE	6
#define C_ULE	7
#define C_SF	8
#define C_NGLE	9
#define C_SEQ	10
#define C_NGL	11
#define C_LT	12
#define C_NGE	13
#define C_LE	14
#define C_NGT	15

#endif
/* vim: set shiftwidth=2: */
