/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "mips_opposite.h"

const t_mips_opposite mips_opposite_table[]={
  {MIPS_BNE, MIPS_BEQ},
  {MIPS_BLEZ, MIPS_BGTZ},
  {MIPS_BGEZ, MIPS_BLTZ},
//  {MIPS_BLTZAL,MIPS_BGEZAL},
  {MIPS_NONE,MIPS_NONE}
};  

t_bool OppositeExist(t_bbl * i_bbl)
{
  if(OppositeOpcodeExist(MIPSINSOPCODE(i_bbl->ins_last))||
      (INS_PREV(i_bbl->ins_last)&&OppositeOpcodeExist(MIPSINSOPCODE(INS_PREV(i_bbl->ins_last)))))
    return TRUE;
  return FALSE;
}

t_bool MakeOpposite(t_bbl * i_bbl)
{
  if(OppositeOpcodeExist(MIPSINSOPCODE(i_bbl->ins_last)))
  {
    MIPSINSOPCODE(i_bbl->ins_last)=GiveOppositeOpcode(MIPSINSOPCODE(i_bbl->ins_last));
    return TRUE;
  }
  if(INS_PREV(i_bbl->ins_last)&&OppositeOpcodeExist(MIPSINSOPCODE(INS_PREV(i_bbl->ins_last))))
  {
    MIPSINSOPCODE(INS_PREV(i_bbl->ins_last))=GiveOppositeOpcode(MIPSINSOPCODE(INS_PREV(i_bbl->ins_last)));
    return TRUE;
  }
  return FALSE;
}

t_bool OppositeOpcodeExist(t_uint32 test_opcode)
{
  t_uint32 i=0;
  for(i=0;mips_opposite_table[i].opcode0!=MIPS_NONE;i++)
    if(mips_opposite_table[i].opcode0 == test_opcode)
      return TRUE;
  for(i=0;mips_opposite_table[i].opcode1!=MIPS_NONE;i++)
    if(mips_opposite_table[i].opcode1 == test_opcode)
      return TRUE;
  return FALSE;
}

t_uint32 GiveOppositeOpcode(t_uint32 test_opcode)
{
  t_uint32 i=0;
  for(i=0;mips_opposite_table[i].opcode0!=MIPS_NONE;i++)
  {
    if(mips_opposite_table[i].opcode0 == test_opcode)
      return mips_opposite_table[i].opcode1;
  }
  for(i=0;mips_opposite_table[i].opcode1!=MIPS_NONE;i++)
  {
    if(mips_opposite_table[i].opcode1 == test_opcode)
      return mips_opposite_table[i].opcode0;
  }
  return MIPS_NONE;
}

/* vim: set shiftwidth=2: */
