/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "mips_print.h"
#include "mips_opcodes.h"
#include "mips_registers.h"
#include "mips_instructions.h"
#include "../../utils/util_error.h"
#include "string.h"

static const char * const Regs[] = 
{
  "zero", "at",	  "v0",	 "v1",	 "a0",	  "a1",	   "a2",   "a3",
  "t0",	  "t1",	  "t2",	 "t3",	 "t4",	  "t5",	   "t6",   "t7",
  "s0",	  "s1",	  "s2",	 "s3",	 "s4",	  "s5",	   "s6",   "s7",
  "t8",	  "t9",	  "k0",	 "k1",	 "gp",	  "sp",	   "s8",   "ra",
  "sr",	  "lo",	  "hi",	 "bad",	 "cause", "pc",
  "fv0",  "$f1",  "fv1", "$f3",  "ft0",   "$f5",   "ft1",  "$f7",
  "ft2",  "$f9",  "ft3", "$f11", "fa0",   "$f13",  "fa1",  "$f15",
  "ft4",  "f17",  "ft5", "f19",  "fs0",   "f21",   "fs1",  "f23",
  "fs2",  "$f25", "fs3", "$f27", "fs4",   "$f29",  "fs5",  "$f31",
  "fsr",  "fir",  "fp",  "inx",  "rand",  "tlblo", "ctxt", "tlbhi",
  "epc",  "prid"
};
static const char * const Fmt[]=
{
  "O","1","2","3","4","5","6","7",
  "8","9","10","11","12","13","14","15",
  "S","D","26","27","W","29","30","31",
  "","","","","","","",""
    
};
static const char * const Cond[]=
{
  "F","UN","EQ","UEQ","OLT","ULT","OLE","ULE",
  "SF","NGLE","SEQ","NGL","LT","NGE","LE","NGT"
};

void MipsInstructionPrint(t_ins * data, t_string outputstring) 
{
  t_mips_ins * instruction=(t_mips_ins *) data;
  char opcode[10]="";
  char base[10]="";
  char rs[10]="",rt[10]="",rd[10]="",sa[10]="",sel[10]="",cond[10]="",
  fs[10]="",ft[10]="",fd[10]="",cc[10]="",fmt[10]="";
  char immediate[20]="",offset[20]="";
  char instr_index[30]="";
  char operands[60]="";
  t_address hulp_addr,hulp_addr2;
  GenericSet_uint32(hulp_addr,0xf0000000);
  
  sprintf(opcode,"%s",mips_opcode_table[MIPSINSOPCODE(instruction)].desc);

  if(MIPSINSPROCESSOR(instruction)==CPU)
  {
    switch(INS_TYPE(instruction))
    {
      case IT_LOAD:
      case IT_STORE:
	sprintf(base,"%s",Regs[MIPSINSRS(instruction)]);
	sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	sprintf(offset,"0x%x",MIPSINSIMMEDIATE(instruction));
	sprintf(operands,"%s,%s(%s)",rt,offset,base);
	break;
      case IT_DATAPROC:
	if(MIPSINSRD(instruction) ==MIPS_REG_NONE)
	{
	  /* We hebben te maken met een immediate data*/
	  if(MIPSINSOPCODE(instruction)!=MIPS_LUI)
	  {
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(immediate,"Ox%x",MIPSINSIMMEDIATE(instruction));
	    sprintf(operands,"%s, %s, %s",rt,rs,immediate);
	  }
	  else
	  {
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(immediate,"Ox%x",MIPSINSIMMEDIATE(instruction));
	    sprintf(operands,"%s, %s",rt,immediate);
	  }
	}
	else if( ( MIPSINSFUNCTION(instruction)& 0x38 ) == 0 )
	  if ( MIPSINSRS(instruction) == MIPS_REG_NONE )
	  {
	    /* SLL,SRL,SRA */
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(rd,"%s",Regs[MIPSINSRD(instruction)]);
	    sprintf(sa,"%d",MIPSINSSA(instruction));
	    sprintf(operands,"%s, %s, %s",rd,rt,sa);
	  }
	  else
	  {
	    /* SLLV,SRLV,SRAV */
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(rd,"%s",Regs[MIPSINSRD(instruction)]);
	    sprintf(operands,"%s, %s, %s",rd,rt,rs);
	  }
	else
	{
	  /* register Data */
	  sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	  sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	  sprintf(rd,"%s",Regs[MIPSINSRD(instruction)]);
	  sprintf(operands,"%s, %s, %s",rd,rs,rt);
	}
	break;
      case IT_MUL:
	switch(MIPSINSOPCODE(instruction))
	{
	  case MIPS_MFHI:
	  case MIPS_MFLO:
	    sprintf(rd,"%s",Regs[MIPSINSRD(instruction)]);
	    sprintf(operands,"%s",rd);
	    break;
	  case MIPS_MTHI:
	  case MIPS_MTLO:
	    sprintf(rd,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(operands,"%s",rd);
	    break;
	  case MIPS_MUL:
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(rd,"%s",Regs[MIPSINSRD(instruction)]);
	    sprintf(operands,"%s, %s, %s",rd,rs,rt);
	    break;
	  case MIPS_MULT:
	  case MIPS_MULTU:
	  case MIPS_DIV:
	  case MIPS_DIVU:
	  case MIPS_MADD:
	  case MIPS_MADDU:
	  case MIPS_MSUB:
	  case MIPS_MSUBU:
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(operands,"%s, %s",rs,rt);
	    break;
	}
	break;
      case IT_BRANCH:
	switch(MIPSINSOPCODE(instruction))
	{
	  case MIPS_J:
	  case MIPS_JAL:
	    GenericSet_uint32(hulp_addr2,MIPSINSINSTR_INDEX(instruction)<<2);
	    sprintf(instr_index,"0x%x",GENERIC_OR(GENERIC_AND((GENERIC_ADD_INT32(INS_OLD_ADDRESS(instruction),4)),hulp_addr),hulp_addr2));
	    sprintf(operands,"%s",instr_index);
	    break;
	  case MIPS_JALR:
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(rd,"%s",Regs[MIPSINSRD(instruction)]);
	    sprintf(operands,"%s, %s",rd,rs);
	    break;
	  case MIPS_JR:
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(operands,"%s",rs);
	    break;
	  case MIPS_BEQ:
	  case MIPS_BNE:
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(offset,"Ox%x",GENERIC_ADD_INT32(INS_OLD_ADDRESS(instruction),4+(t_int32)((t_int16)MIPSINSIMMEDIATE(instruction))*4));
	    sprintf(operands,"%s, %s, %s",rs,rt,offset);
	    break;
	  case MIPS_BLTZ:
	  case MIPS_BGEZ:
	  case MIPS_BLEZ:
	  case MIPS_BGTZ:
	  case MIPS_BLTZAL:
	  case MIPS_BGEZAL:
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(offset,"0x%x",GENERIC_ADD_INT32(INS_OLD_ADDRESS(instruction),4+(t_int32)((t_int16)MIPSINSIMMEDIATE(instruction))*4));
	    sprintf(operands,"%s, %s",rs,offset);
	    break;
	}
	break;
      case IT_SWI:
	sprintf(operands,"0x%x",MIPSINSIMMEDIATE(instruction)>>6);
	break;
      default:
	break;
    }
  }
  else if(MIPSINSPROCESSOR(instruction)==COP0)
  {
    switch(MIPSINSOPCODE(instruction))
    {
      case MIPS_MFC0:
      case MIPS_MTC0:
	sprintf(rd,"$%d",MIPSINSRD(instruction));
	sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	sprintf(sel,"%d",MIPSINSSEL(instruction));
	sprintf(operands,"%s, %s, %s",rt,rd,sel);
	break;
    }
  }
  else if(MIPSINSPROCESSOR(instruction)==COP1)
  {
    switch(INS_TYPE(instruction))
    {
      case IT_LOAD:
      case IT_STORE:
	sprintf(base,"%s",Regs[MIPSINSBASE(instruction)]);
	sprintf(ft,"%s",Regs[MIPSINSFT(instruction)]);
	sprintf(offset,"%d",MIPSINSOFFSET(instruction));
	sprintf(operands,"%s, %s(%s)",ft,offset,base);
	break;
      case IT_BRANCH:
	sprintf(cc,"%d",MIPSINSCC(instruction));
	sprintf(offset,"%d",MIPSINSOFFSET(instruction));
	sprintf(operands,"%s, %s",cc,offset);
	break;
      case IT_DATAPROC:
	switch(MIPSINSOPCODE(instruction))
	{
	  case MIPS_MFC1:
	  case MIPS_CFC1:
	  case MIPS_MTC1:
	  case MIPS_CTC1:
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(fs,"%s",Regs[MIPSINSFS(instruction)]);
	    sprintf(operands,"%s, %s",rt,fs);
	    break;
	  case MIPS_ABS:
	  case MIPS_SQRT:
	  case MIPS_NEG:
	  case MIPS_CVT_D :
	  case MIPS_CVT_S :
	  case MIPS_CVT_W :
	  case MIPS_TRUNC_W :
	  case MIPS_ROUND_W :
	  case MIPS_CEIL_W :
	  case MIPS_FLOOR_W :
	  case MIPS_MOV :
	    sprintf(fs,"%s",Regs[MIPSINSFS(instruction)]);
	    sprintf(fmt,"%s",Fmt[MIPSINSFMT(instruction)]);
	    sprintf(fd,"%s",Regs[MIPSINSFD(instruction)]);
	    sprintf(operands,"%s %s, %s",fmt,fd,fs);
	    break;
	  case MIPS_ADD_:
	  case MIPS_SUB_:
	  case MIPS_MUL_:
	  case MIPS_DIV_:
	    sprintf(fs,"%s",Regs[MIPSINSFS(instruction)]);
	    sprintf(fmt,"%s",Fmt[MIPSINSFMT(instruction)]);
	    sprintf(fd,"%s",Regs[MIPSINSFD(instruction)]);
	    sprintf(ft,"%s",Regs[MIPSINSFT(instruction)]);
	    sprintf(operands,"%s %s, %s, %s",fmt,fd,fs,ft);
	    break;
	  case MIPS_C:
	    sprintf(fs,"%s",Regs[MIPSINSFS(instruction)]);
	    sprintf(fmt,"%s",Fmt[MIPSINSFMT(instruction)]);
	    sprintf(cc,"%d",MIPSINSCC(instruction));
	    sprintf(ft,"%s",Regs[MIPSINSFT(instruction)]);
	    sprintf(cond,"%s",Cond[MIPSINSCOND(instruction)]);
	    sprintf(operands,"%s.%s %s, %s, %s",cond,fmt,cc,fs,ft);
	    break;

	  case MIPS_MOVF_:
	  case MIPS_MOVT_:
	    sprintf(fs,"%s",Regs[MIPSINSFS(instruction)]);
	    sprintf(fmt,"%s",Fmt[MIPSINSFMT(instruction)]);
	    sprintf(cc,"%d",MIPSINSCC(instruction));
	    sprintf(fd,"%s",Regs[MIPSINSFD(instruction)]);
	    sprintf(operands,"%s %s, %s, %s",fmt,fd,fs,cc);
	    break;
	  case MIPS_MOVN_:
	  case MIPS_MOVZ_:
	    sprintf(fs,"%s",Regs[MIPSINSFS(instruction)]);
	    sprintf(fmt,"%s",Fmt[MIPSINSFMT(instruction)]);
	    sprintf(fd,"%s",Regs[MIPSINSFD(instruction)]);
	    sprintf(rt,"%s",Regs[MIPSINSRT(instruction)]);
	    sprintf(operands,"%s %s, %s, %s",fmt,fd,fs,rt);
	    break;
	  case MIPS_MOVF:
	  case MIPS_MOVT:
	    sprintf(rs,"%s",Regs[MIPSINSRS(instruction)]);
	    sprintf(cc,"%d",MIPSINSCC(instruction));
	    sprintf(rd,"%s",Regs[MIPSINSRD(instruction)]);
	    sprintf(operands,"%s, %s, %s",rd,rs,cc);
	    break;
	}
	break;
    }
  }

  sprintf(outputstring,"%s %s",opcode,operands);
  StringTrim(outputstring);
}
/* vim: set shiftwidth=2: */
