
#ifndef DIABLO_BBL_TYPEDEFS
#define DIABLO_BBL_TYPEDEFS
typedef struct _t_bbl t_bbl;
#endif

#ifndef ATOM_TYPEDEFS
#define ATOM_TYPEDEFS
typedef enum 
{
  ObjBefore,
  ObjAfter,
  ProgramBefore,
  ProgramAfter,
  ProcBefore,
  ProcAfter,
  BlockBefore,
  BlockAfter,
  InstBefore,
  InstAfter
} t_atom_instrument_when;

typedef enum
{
   InstTypeCondBr,
   InstTypeLoad,
   InstTypeStore,
   InstTypeMem,
   InstTypeConditional
} t_atom_inst_type;

typedef enum
{
   ClassIbranch,
   ClassFbranch,
   ClassSubroutine
} t_inst_class;

/* define aliases for i386 registers */
#define REG_I386_EAX	REG_R0
#define REG_I386_EBX	REG_R1
#define REG_I386_ECX	REG_R2
#define REG_I386_EDX	REG_R3
#define REG_I386_ESI	REG_R4
#define REG_I386_EDI	REG_R5
#define REG_I386_EBP	REG_R6
#define REG_I386_ESP	REG_R7

typedef enum
{
   REG_PC,
   REG_SP,
   REG_R0,
   REG_R1,
   REG_R2,
   REG_R3,
   REG_R4,
   REG_R5,
   REG_R6,
   REG_R7,
   REG_R8,
   REG_R9,
   REG_R10,
   REG_R11,
   REG_R12,
   REG_R13,
   REG_R14,
   REG_R15,
   REG_R16,
   REG_R17,
   REG_R18,
   REG_R19,
   REG_R20,
   REG_R21,
   REG_R22,
   REG_R23,
   REG_R24,
   REG_R25,
   REG_R26,
   REG_R27,
   REG_R28,
   REG_R29,
   REG_R30,
   REG_R31
} t_atom_regs;

#define BrCondValue CondValue
typedef enum
{
   CondValue,
   FunctionName,
   ObjectName,
   EffAddrValue,	/* effective load or store address, only valid if unambiguous */
   EffStoreAddrValue,	/* effective store address for an instruction */
   EffLoadAddrValue,	/* effective load address for an instruction */
   EffLoadAddrValue2	/* FOR INTERNAL USE ONLY -- i386 has instructions with 2 source memory operands (e.g. cmpsb) */
} t_atom_ins_queries;

typedef enum
{
   InstOpcode,
   InstBinary
} t_atom_inst_info;

typedef enum
{
   BlockNumberInsts
} t_atom_block_info;

typedef enum
{
  ATOM_REG_VALUE,
  ATOM_ARCHSPECIFIC_REG_VALUE,
  ATOM_CHAR_VALUE,
  ATOM_INT_VALUE,
  ATOM_LONG_VALUE,
  ATOM_LONG_LONG_VALUE,
  ATOM_SPECIAL_VALUE,
  ATOM_POINTER_VALUE
} t_atom_argument_type;

typedef struct _t_atom_proto t_atom_proto;
#endif

#ifndef ATOM_H
#define ATOM_H
#include "kernel/diablo_object.h"

#ifndef ATOM_TYPEDEFS_2
#define ATOM_TYPEDEFS_2
typedef t_object Obj;
typedef t_function Proc;
typedef t_bbl Block;
typedef t_ins Inst;
#endif

#define GetFirstBlock(x) ((x)->bbl_first)
#define GetLastBlock(x) ((x)->bbl_last)
#define GetLastInst(x)  ((x)->ins_last)
#define GetFirstInst(x)  ((x)->ins_first)
#define GetNextInst(x)   (INS_NEXT(x))
#define GetNextProc(x)	  ((x)->instrument_next)
#define GetFirstObjProc(x)	((SECTION_CFG(T_SECTION(T_OBJECT(x)->code[0])))->fun_instrument_first)
#define GetFirstProc()	  GetFirstObjProc(obj)	
#define AddCallProto(x)   AtomProtoNew(x)
#define BlockPC(x) 	BBL_OLD_ADDRESS(x)
#define BlockNinst(x) 	BBL_NINS(x)
#define InstPC(x) 	INS_OLD_ADDRESS(x)
Block * GetNextBlock(Block * block);

/* Prototypes for instrumentation functions {{{ */
struct _t_atom_proto
{
   char * name;
   int nargs;
   t_atom_argument_type * args;
   struct _t_atom_proto * next;
   t_bbl * bbl;
};
/* }}} */

t_bool IsInstType(Inst * ins, t_atom_inst_type type);
t_atom_proto * AtomProtoNew(char * def);
void AtomProtosFree();
void AtomProtosResolve(t_object * obj);
t_bbl * AtomSplitBlock(t_bbl * orig_bbl, t_ins * where, t_bool before);
t_bbl * AtomMoveBbl(t_bbl * bbl);
void AtomPatchPseudoCalls(t_object * obj);

void AddArrayProto(char * name);

void InitTransTables(t_object * obj);
t_uint32 TransTableCalcSize(t_object * obj);
void TransTableAddElem32(t_object * ob, t_uint32 s, t_uint32 e, t_uint32 t);
void TransTableAddElem64(t_object * ob, t_uint64 s, t_uint64 e, t_uint64 t);
void TransTableFinish();

void AddCallProgram(t_atom_instrument_when when, char * callbackname,...);
void AddCallProc(Proc * proc, t_atom_instrument_when when,char * callbackname,...);
void AddCallBlock(Block * block, t_atom_instrument_when when,char * callbackname,...);
void AddCallInst(Inst *ins,t_atom_instrument_when when,char * callbackname,...);
void SetArraySize(char * name,int nelems, int elemsize);
void SetArrayElem32(char * name,int elem, int count, ...);

void InstrumentInit(int argc, char **argv);
void Instrument(int argc, char **argv, Obj *obj);
void RealInstrument(t_object * obj);
#endif
/* vim: set shiftwidth=2 foldmethod=marker:*/
