/***********************************************************
 * Copyright 2001,2002,2003: {{{
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * }}}
 **********************************************************/

#ifndef ATOM_INTERNAL_TYPEDEFS
#define ATOM_INTERNAL_TYPEDEFS

#include "atom.h"
#include "simple_types.h"
/*! An argument value */
typedef union 
{
  t_atom_regs reg;
  t_reg archspecific_reg;
  t_atom_ins_queries query;
  t_uint32 uint32;
  t_int32 int32;
  t_uint64 uint64;
  t_int64 int64;
} t_atom_value;

typedef struct _t_atom_instrument_marker t_atom_instrument_marker;
typedef struct _t_atom_actions t_atom_actions;
#endif

#ifndef ATOM_TYPEDEFS_2
#define ATOM_TYPEDEFS_2
typedef t_object Obj;
typedef t_function Proc;
typedef t_bbl Block;
typedef t_ins Inst;
#endif

#ifndef ATOM_INTERNAL_H
#define ATOM_INTERNAL_H

#include "atom.h"
#include "simple_types.h"

/*! Marker for instrumentation, a post pass removes these and inserts the real call to the instrumentation functions */
struct _t_atom_instrument_marker
{
  t_atom_instrument_when when;
  t_atom_proto * proto;
  t_atom_value * values;
  struct _t_atom_instrument_marker * next;
  struct _t_atom_instrument_marker * prev;
};

struct _t_atom_actions
{
  void (*Init)(t_object *);
  t_regset (*EstimateChangedRegisters)(t_atom_instrument_marker * m);
  t_uint32 (*AddSave)(t_bbl * in, t_regset liveregs, t_regset changedregs);
  void (*AddRestore)(t_bbl * out,t_regset liveregs,t_regset changedregs,t_uint32 stack_change) ;
  void (*AddParams)(t_bbl * in, t_ins * ins, t_bbl * orig, t_atom_instrument_marker * marker,t_uint32 stack_change, t_address pc);
  void (*AddKillParams)(t_bbl * out, t_atom_instrument_marker * marker);
  void (*AddCheckForExitCall)(t_bbl * check, t_bbl * call, t_bbl * restore);
  void (*AddCall)(t_bbl * in, t_bbl * out, t_atom_proto * proto);
  t_bool (*IsInstType)(Inst * ins, t_atom_inst_type type);
};

void AtomReorderParameters(t_atom_proto * proto, t_ins ** first_inses_for_params, t_ins * parstart,t_bbl * in);
t_atom_instrument_marker *AtomInstrumentMarkerNew(t_atom_instrument_when when, t_atom_proto * proto);
t_bbl * AtomMoveBbl(t_bbl * bbl);
t_section *  InstrumentAddExtraSection(t_object * parent, t_section * parent_section, char type, char * objectname, char * sectionname, t_address total_size, t_address startaddress);

/* the transtable sections */
extern t_section * added_section;
extern t_section * added_section2;
/* the translate functions */
extern t_function * translate;
extern t_function * reversetranslate;
/* stuff for relocating with old addresses */
extern t_address * bbl_translate_list;
extern t_bool use_old_addresses;

#endif

/* vim: set shiftwidth=2 foldmethod=marker:*/
