/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* This code is based on libavl: The copyright is included:
 *
 * libavl: manipulates AVL trees.  Copyright (C) 1998, 1999 Free Software
 * Foundation, Inc.  
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.  
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.  
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * The author may be contacted at <pfaffben@pilot.msu.edu> on the Internet, or
 * as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA through more mundane
 * means. */ 


/* This is file avlt.h in libavl. */

#ifndef AVL_TREE_TYPEDEFS
#define AVL_TREE_TYPEDEFS
typedef int (*t_avl_comparison_func) (const void *, const void *);
typedef void (*t_avl_node_func) (void *);
typedef void (*t_avl_node_print_func) (void *, void *);
typedef void *(*t_avl_copy_func) (void *, void *);
#endif


#ifndef AVLT_H 
#define AVLT_H 

#ifndef AVL_MAX_HEIGHT
/*! The default maximum height of 32 allows for AVL trees having between
 * 5,704,880 and 4,294,967,295 nodes, depending on order of insertion.  You may
 * change this compile-time constant as you wish. */
#define AVL_MAX_HEIGHT	32
#endif

/*! Structure for a node in a threaded AVL tree. */
typedef struct _t_threaded_avl_tree_node
{
  struct _t_threaded_avl_tree_node *parent;
  /*! Subtrees or threads. */
  struct _t_threaded_avl_tree_node *link[2];	
  /*! Balance factor. */
  signed char bal;
  /*! Used during insertion. */
  char cache;			
  /*! Left and right thread tags. */
  signed char tag[2];		
}
t_threaded_avl_tree_node;

/*! Used for traversing a threaded AVL tree. */
typedef struct _t_threaded_avl_tree_traverser
{
  int init;			/* Initialized? */
  const t_threaded_avl_tree_node *p;		/* Last node returned. */
}
t_threaded_avl_tree_traverser;

/* Initializer for t_threaded_avl_tree_traverser */ 
#define ThreadedAvlTreeInit {0}

/* Structure which holds information about a threaded AVL tree. */
typedef struct _t_threaded_avl_tree
{
  t_threaded_avl_tree_node root;		/* Tree root node. */
  t_avl_comparison_func cmp;	/* Used to compare keys. */
  int count;			/* Number of nodes in the tree. */
  t_uint32 node_offset;
}
t_threaded_avl_tree;

/* General functions. {{{ */
t_threaded_avl_tree *ThreadedAvlTreeNew (t_uint32, t_avl_comparison_func);
void ThreadedAvlTreeDestroy (t_threaded_avl_tree *, t_avl_node_func);
void ThreadedAvlTreeFree (t_threaded_avl_tree *);
int ThreadedAvlTreeCount (const t_threaded_avl_tree *);
t_threaded_avl_tree *ThreadedAvlTreeCopy (const t_threaded_avl_tree *, t_avl_copy_func);
/*}}}*/
/* Walk the tree. {{{ */
void ThreadedAvlTreeWalk (const t_threaded_avl_tree *, t_avl_node_func);
void *ThreadedAvlTreeTraverse (const t_threaded_avl_tree *, t_threaded_avl_tree_traverser *);
#define ThreadedAvlTreeInitTraverser(TRAVERSER) ((TRAVERSER)->init = 0)
void *ThreadedAvlTreeNext (const t_threaded_avl_tree *tree, void *item);
void *ThreadedAvlTreePrev (const t_threaded_avl_tree *tree, void *item);
/*}}}*/
/* Search for a given item.{{{ */
void *ThreadedAvlTreeProbe (t_threaded_avl_tree *, void *);
void *ThreadedAvlTreeDelete (t_threaded_avl_tree *, const void *);
void **ThreadedAvlTreeFind (const t_threaded_avl_tree *, const void *);
void **ThreadedAvlTreeFindClose (const t_threaded_avl_tree *, const void *);
void *ThreadedAvlTreeInsert (t_threaded_avl_tree *, void *);
void *ThreadedAvlTreeReplace (t_threaded_avl_tree *, void *);
/*}}}*/
#endif 
/* vim: set shiftwidth=2 foldmethod=marker: */
