/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef MY_HASH_H
#define MY_HASH_H
#include "../simple_types.h"
 
typedef struct _data_hash_node
{
  struct _data_hash_node * next;
  struct _data_hash_node * equal; 
  void * key;
} data_hash_node;


typedef struct _data_hash_table
{
  data_hash_node ** nodes; 
  t_uint32 tsize;
  t_uint32 offset;
  t_uint32 (*hash_func) (struct _data_hash_table *, void *);
  my_cmpresult (*hash_cmp)(void *, void *);
  void (*node_free)(void *,void *);
} data_hash_table;

data_hash_table * HashTableNew(t_uint32 size, t_uint32 offset, t_uint32 (*func)(data_hash_table *, void *), my_cmpresult (*cmp)(void *, void *), void (*node_free)(void *,void *)); 
void HashInsert (data_hash_table *,  void *);
void HashPrepend (data_hash_table *,  void *);
void HashDelete (data_hash_table *, void *);	/* Caution!: second argument is the address of the node that has to be deleted, not the key!*/
void HashFree(data_hash_table * );
void * HashLookup (data_hash_table *,  void *);
t_uint32 string_hash (data_hash_table * , void * );
void HashWalk(data_hash_table * table, void (*func)(void *,void *),void *);
void HashStats(data_hash_table * table);
my_cmpresult string_cmp(void * a , void * b);

#endif



