/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "../config.h" 
#include <stdlib.h>
#include "data_ll.h"
#include "../utils/util_error.h"

void RealAppendLL(ll * prev, ll * insert, ll ** head)
{
	if (prev)
	{
		LLNext(insert)=LLNext(prev);
		LLNext(prev)=insert;
	}
	else if (head)
	{
		LLNext(insert)=*head;
		*head=insert;
	}
	else
	{
	    FATAL(("Could not append node!"));	
	}
}


void RealAppendLLSortedCached(ll *insert, ll ** head)
{
  static data_ll_elem* cached_elem = NULL;
  static ll* cached_ll = NULL;
  data_ll_elem *iter = NULL, *iter2 = NULL;

  if (cached_ll==*head && cached_elem && cached_elem->data < ((data_ll_elem*)insert)->data)
    {
      iter2=cached_elem;
      iter = LLNext(cached_elem);
    }
  else  
    iter = (data_ll_elem*)*head;
  
  for (; iter && iter->data < ((data_ll_elem*)insert)->data; iter2=iter,iter = LLNext(iter)) ;
  
  LLAppend(iter2,insert,head);
  
  cached_elem = (data_ll_elem*)insert;
  cached_ll = *head;
}

void RealAppendLLSorted(ll *insert, ll ** head)
{
  data_ll_elem *iter, *iter2 = NULL;
  for (iter = (data_ll_elem*)*head; iter && iter->data < ((data_ll_elem*)insert)->data; iter2=iter,iter = LLNext(iter)) ;

  LLAppend(iter2,insert,head);
}

data_ll_elem * list=NULL;

data_ll_elem * LLCalloc(void * param)
{
  data_ll_elem * ret;
  if(list)
  {
    ret = list;
    list = LLNext(list);
    LLNext(ret) = NULL;
    ret->data = param;
    return ret;
  }
  ret = (data_ll_elem *) Calloc(1,sizeof(data_ll_elem));
  ret->data = param;
  return ret;
}


void LLFree(data_ll_elem* to_free)
{
  LLNext(to_free) = list;
  list = to_free;
}

void LLFinalise(void)
{
  data_ll_elem * to_free;
  while(list)
  {
    to_free = list;
    list = LLNext(list);
    Free(to_free);
  }
}

void LLListCleanup(data_ll_elem ** start)
{
  data_ll_elem * iter;
  while(*start)
  {
    iter = *start;
    *start = LLNext(*start);
    Free(iter);
  }
}

void RealUnlinkLL(ll* prev, ll* unlink, ll** head)
{
  if(!prev && *head != unlink) FATAL(("No prev and unlink is not head"));
  if(prev)
  {
    if(LLNext(prev) != unlink) FATAL(("Wrong prev for unlink"));
    LLNext(prev) = LLNext(unlink);
  }
  else
  {
    *head = LLNext(unlink);
  }
  LLNext(unlink) = NULL;
}
