/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* contains some usefull functions to parse a file.
   Maybe it could be included in data_string?? */

#include "../config.h"
#include "../utils/util_error.h"
#include "../utils/util_malloc.h"
#include "../datastructures/data_string.h"
#include "data_parse_func.h"

#include <string.h>
#include <ctype.h>

/* return-value points to the next non-white character in in_string */
char * skip_white(const char * in_string)
{
  char * ret = (char *) in_string;
  t_int16 i = 0,j = 0;

  j = strlen(in_string);
  while(isspace(ret[i]) && (i <= j)) {i++;}
  return (ret+i);
}

/* returns string containing the next field in in_string.
   Memory needs to be freed! */
char * extract_field(char * in_string)
{
  char * new_string, * tmp_string = NULL;
  t_int16 i = 0;

  tmp_string = skip_white(in_string);
  while(!isspace(tmp_string[i])) i++;
  if (i == 0) return NULL;
  new_string = (char *) Malloc(i+1); /* strdup() cannot be used here! */
  strncpy(new_string, (const char *) tmp_string, i);
  new_string[i] = '\0';
  return new_string;
}

/* return-value points to non-space character following whitespace, one or 
more non space characters, and additional whitespace */
char * skip_field_and_white(char * in_string)
{
  char * tmp_string;
  t_int16 i = 0;

  tmp_string = skip_white(in_string);
  while(!isspace(tmp_string[i])) i++;
  tmp_string = skip_white(tmp_string+i);
  return tmp_string;
}

