/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* $Header: /home/cvs/cvsroot/diablo/datastructures/data_string.c,v 1.2 2004/02/09 00:49:45 bdebus Exp $ */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "../config.h"
#include "../simple_types.h"
#include "data_string.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"

#ifdef DEBUG_MALLOC
char * RealStringDup(char * file, t_uint32 lnno, const t_string in)
#else
char * RealStringDup(const char * in)
#endif
{
   char * ret;
   if (in==NULL) return NULL;
#ifdef DEBUG_MALLOC
   ret=RealCalloc(file,lnno,sizeof(char) , (strlen(in)+1));
#else
   ret=Calloc(sizeof(char) , (strlen(in)+1));
#endif
   strcpy(ret,in);
   return ret;
}

char * Collate(const char * in1, const char * in2)
{
	char * out=(char *) Malloc(strlen(in1)+strlen(in2)+1);
	sprintf(out,"%s%s",in1,in2);
	return out;
}

#ifdef DEBUG_MALLOC
char * RealCollate3(char * file, t_uint32 lnno, const char * in1, const char * sep, const char * in2)
#else
char * RealCollate3(const char * in1, const char * sep, const char * in2)
#endif
{
#ifdef DEBUG_MALLOC
	char * out=(char *) RealMalloc(file,lnno,strlen(in1)+strlen(in2)+strlen(sep)+1);
#else
	char * out=(char *) Malloc(strlen(in1)+strlen(in2)+strlen(sep)+1);
#endif
	sprintf(out,"%s%s%s",in1,sep,in2);
	out[strlen(in1)+strlen(in2)+strlen(sep)]=0;
	return out;
}

/* remove trailing newlines from a string */

void Chomp(t_string in)
{
   int len=strlen(in);
   if (in[len-1] == '\n') in[len-1] = '\0';
}

t_bool Chop(t_string in,char c)
{
   int len=strlen(in);
   if (in[len-1] == c) in[len-1] = '\0';
   else return FALSE;
   return TRUE;
}
/* Divide a string seperated by the chars (ifndef CHAR_PATTERNS) or the
 * char pattern (ifdef CHAR_PATTERNS) into an array of strings */ 

#ifdef DEBUG_MALLOC
void RealPushString(char * file, t_uint32 lnno, t_string_array * array, t_string in)
#else
void RealPushString(t_string_array * array, t_string in)
#endif
{
#ifdef DEBUG_MALLOC
   array->strings=(t_string *) RealRealloc(file,lnno,array->strings,sizeof(t_string) * (array->nstrings+1));
#else
   array->strings=(t_string *) Realloc(array->strings,sizeof(t_string) * (array->nstrings+1));
#endif
   array->strings[array->nstrings]=in;
   array->nstrings++;
}


#define CHAR_PATTERNS




#ifdef DEBUG_MALLOC
t_string_array * RealStringDivider(char * file, t_uint32 lnno, t_string in, t_string separators, t_bool strict,t_bool keep_sep)
#else
t_string_array * RealStringDivider(t_string in, t_string separators, t_bool strict,t_bool keep_sep)
#endif
{
   int len=strlen(in);
#ifndef CHAR_PATTERNS
   int len2=strlen(separators), tel2;
#endif
   int tel, fields=0, sublength=0;
   t_bool prev_is_separator=TRUE;
   t_bool this_is_separator=FALSE;
   t_string start_prev=in;
#ifdef DEBUG_MALLOC
   t_string_array * ret=RealMalloc(file, lnno,sizeof(t_string_array));
#else
   t_string_array * ret=Malloc(sizeof(t_string_array));
#endif
   ret->nstrings=0;
   ret->strings=NULL;

   for (tel=0; tel<len; tel++)
   {
#ifdef CHAR_PATTERNS
      this_is_separator=CharPatternMatch(separators,in[tel]);
      if ((this_is_separator)&& ((strict)||(!prev_is_separator)))
      {
#ifdef DEBUG_MALLOC
	 t_string tmp=RealMalloc(file,lnno,sublength+1);
#else
	 t_string tmp=Malloc(sublength+1);
#endif
	 strncpy(tmp,start_prev,sublength);
	 tmp[sublength]='\0';
#ifdef DEBUG_MALLOC
	 RealPushString(file,lnno,ret,tmp);
#else
	 PushString(ret,tmp);
#endif
	 fields++;
	 start_prev+=sublength-(keep_sep?1:0);
	 sublength=keep_sep?1:0;
      }
#else
      this_is_separator=FALSE;
      for (tel2=0; tel2<len2;tel2++)
      {
	 if (in[tel]==separators[tel2])
	 {
	    if ((strict)||(!prev_is_separator))
	    {
	       t_string tmp=Malloc(sublength+1);
	       strncpy(tmp,start_prev,sublength);
	       tmp[sublength]='\0';
	       PushString(ret,tmp);
	       fields++;
	       start_prev+=sublength-(keep_sep?1:0);
	       sublength=keep_sep?1:0;
	    }
	    this_is_separator=TRUE;
	 }
      }
#endif
      if (!this_is_separator)
      {
	 sublength++;
      }
      else
      {
	 start_prev++;
	 sublength=keep_sep?1:0;
      }

      prev_is_separator=this_is_separator;
   }
   if (sublength)
   {
      t_string tmp=Malloc(sublength+1);
      strncpy(tmp,start_prev,sublength);
      tmp[sublength]='\0';
      PushString(ret,tmp);
      fields++;
      start_prev+=sublength;
      sublength=0;
   }
   return ret;
}


/* Join an array of strings in on strings, separated by con */
#ifdef DEBUG_MALLOC
t_string RealStringArrayJoiner(char * file, t_uint32 lnno, t_string_array * in, t_string con)
#else
t_string StringArrayJoiner(t_string_array * in, t_string con)
#endif
{
   t_string * tel;
   t_string ret;
   int nstr=0;
   int len=0;
   FORALL_STRINGS(in, tel)
   {
      len+=strlen(*tel);
      nstr++;
   }
   len+=((nstr-1)*strlen(con))+1;

#ifdef DEBUG_MALLOC
   ret=RealMalloc(file, lnno,len);
#else
   ret=Malloc(len);
#endif

   len=0;
   FORALL_STRINGS(in, tel)
   {
      sprintf(ret+len,"%s",*tel);
      len+=strlen(*tel);
      nstr--;
      if (nstr!=0) 
      {
	 sprintf(ret+len,"%s",con);
	 len+=strlen(con);
      }
   }
   ret[len]='\0';
   return ret;
}

void StringArrayFree(t_string_array * free_me)
{
   t_string * tel;
   FORALL_STRINGS(free_me, tel)
   {
      Free(*tel);
   }
   if (free_me->strings) Free(free_me->strings);
   Free(free_me);
}

t_uint16 StringArrayNElems(t_string_array * in)
{
   return in->nstrings;
}

t_bool CharPatternMatch(t_string pattern, char match)
{
   t_uint32 tel=0;
   t_uint32 len=strlen(pattern);
   t_uint16 prev=256;
   t_bool negate=(pattern[0]=='^');

   if (negate) tel++;

   for (;tel<len; tel++)
   {
      if (pattern[tel]=='-') 
      {
	 if ((prev!=256)&&(tel+1<len))
	 {
	    if ((match>=prev)&&(match<=pattern[tel+1]))
	    {
	       return (!negate);
	    }
	 }
	 else
	 {
	    if (match=='-') return (!negate);
	 }
      }
      else
      {
	 if (match==pattern[tel]) return (!negate);
      }
      prev=pattern[tel];
   }
   return negate;
}

t_uint32 PatternMatch(const char* mask, const char* s) {
   const char * cp =0, * mp=0;
   for (; *s&&*mask!='*'; mask++,s++) if (*mask!=*s&&*mask!='?') return 0;
   for (;;) {
      if (!*s) { while (*mask=='*') mask++; return !*mask; }
      if (*mask=='*') { if (!*++mask) return 1; mp=mask; cp=s+1; continue; }
      if (*mask==*s||*mask=='?') { mask++, s++; continue; }
      mask=mp; s=cp++;
   }
}

void StringTrim(t_string string)
{
   int i;
   for (i = strlen(string) - 1; i >= 0; i--)
      if (!isspace(string[i])) break;
   string[i+1] = '\0';
}


t_uint32 StringToUint32(char * str, t_uint32 maxlen)
{
	t_uint32 val=0, tel=0, base=10;
	if ((str[0]=='0')&&((str[1]=='x')||(str[1]=='X'))) { 
	   base=16; 
	   tel=2;
	}
	
	while ((!maxlen)||(tel<maxlen))
	{
	   if (base==10)
	   {
	      if ((str[tel]>='0')&&(str[tel]<='9'))
		val=val*10+(str[tel]-'0');
	      else break;
	   }
	   else
	   {
	      if ((str[tel]>='0')&&(str[tel]<='9'))
		val=val*16+(str[tel]-'0');
	      else if ((str[tel]>='a')&&(str[tel]<='f'))
		val=val*16+10+(str[tel]-'a');
	      else if ((str[tel]>='A')&&(str[tel]<='F'))
		val=val*16+10+(str[tel]-'A');
	      else 
		 break;
	   }
	   tel++;
	}

	if ((!maxlen) && (str[tel]!=0)) FATAL(("Trailing characters in string!"));

	return val;
}

t_uint64 StringToUint64(char * str, t_uint32 maxlen)
{
	t_uint64 val=0, tel=0, base=10;
	if ((str[0]=='0')&&((str[1]=='x')||(str[1]=='X'))) { 
	   base=16; 
	   tel=2;
	}
	
	while ((!maxlen)||(tel<maxlen))
	{
	   if (base==10)
	   {
	      if ((str[tel]>='0')&&(str[tel]<='9'))
		val=val*10+(str[tel]-'0');
	      else break;
	   }
	   else
	   {
	      if ((str[tel]>='0')&&(str[tel]<='9'))
		val=val*16+(str[tel]-'0');
	      else if ((str[tel]>='a')&&(str[tel]<='f'))
		val=val*16+10+(str[tel]-'a');
	      else if ((str[tel]>='A')&&(str[tel]<='F'))
		val=val*16+10+(str[tel]-'A');
	      else 
		 break;
	   }
	   tel++;
	}

	if ((!maxlen) && (str[tel]!=0)) FATAL(("Trailing characters in string!"));

	return val;
}

