/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*!\addtogroup GENERIC_DATATYPES  Generic Datatypes*/
/*@{ */
/*! \defgroup STRING Strings */
/*@{ */

#ifndef DATA_STRING_TYPEDEFS
#define DATA_STRING_TYPEDEFS
#include "../simple_types.h"
typedef struct _t_string_array t_string_array;

typedef char * t_string; 
#endif

#ifndef DATA_STRING_H 
#define DATA_STRING_H
struct _t_string_array 
{
   t_uint16 nstrings;
   t_string * strings;
};

#define FORALL_STRINGS(x,y) for (y=x->strings; ((y))!=(x->strings+x->nstrings); y++)

#ifdef DEBUG_MALLOC
#define StringDup(x) RealStringDup(__FILE__,__LINE__,x)
char * RealStringDup(char * file, t_uint32 lnno, const t_string in);
#define Collate3(x,y,z) RealCollate3(__FILE__,__LINE__,x,y,z)
char * RealCollate3(char *, t_uint32, const char * , const char *, const char * );
#define PushString(x,y) RealPushString(__FILE__,__LINE__,x,y)
void RealPushString(char *, t_uint32, t_string_array * array, t_string in);
#else
#define StringDup(x) RealStringDup(x)
char * RealStringDup(const char * in);
#define Collate3(x,y,z) RealCollate3(x,y,z)
char * RealCollate3(const char * , const char *, const char * );
#define PushString(x,y) RealPushString(x,y)
void RealPushString(t_string_array * array, t_string in);
#endif

char * Collate(const char * , const char * );
t_uint32 PatternMatch(const char* mask, const char* s); 
t_bool CharPatternMatch(t_string pattern, char match);
t_uint16 StringArrayNElems(t_string_array * in);
#ifdef DEBUG_MALLOC
#define StringArrayJoiner(in,con) RealStringArrayJoiner(__FILE__,__LINE__,in,con)
t_string RealStringArrayJoiner(char * file, t_uint32 lnno, t_string_array * in, t_string con);
#else
t_string StringArrayJoiner(t_string_array * in, t_string con);
#endif
#ifdef DEBUG_MALLOC
#define StringDivider(in,separators,strict,keep_sep) RealStringDivider(__FILE__,__LINE__,in,separators,strict,keep_sep)
t_string_array * RealStringDivider(char * file, t_uint32 lnno, t_string in, t_string separators, t_bool strict, t_bool keep_sep);
#else
#define StringDivider(in,separators,strict,keep_sep) RealStringDivider(in,separators,strict,keep_sep)
t_string_array * RealStringDivider(t_string in, t_string separators, t_bool strict, t_bool keep_sep);
#endif
void StringArrayFree(t_string_array * free_me);
void Chomp(t_string in);
t_bool Chop(t_string in,char c);
void StringTrim(t_string string);
t_uint32 StringToUint32(char * str, t_uint32 maxlen);
t_uint64 StringToUint64(char * str, t_uint32 maxlen);
#endif
/*@}*/
/*@}*/
