/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../../config.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_object.h"
#include "ar_arch.h"
#include "../../datastructures/data_hash.h"
#include "../../utils/util_file.h"
#include "../../utils/util_error.h"
#include "../../utils/util_options.h"
#include "../../utils/util_malloc.h"
#include "../../datastructures/data_string.h"

t_bool IsArchAr(FILE * fp)
{
   char buffer[SARMAG+16]; /* Read magic string + the name of the first archive header */
   long fpos=ftell(fp);
   int ret=fread(buffer,SARMAG+16,1,fp); 
   fseek(fp,fpos,SEEK_SET);
   if (ret!=1) return FALSE;

   if (strncmp("________64E",buffer+SARMAG,11)==0) return FALSE;
   if (strncmp(ARMAG,buffer,SARMAG)==0) return TRUE;
   return FALSE;
}

t_bool IsArchAlphaAr(FILE * fp)
{
   char buffer[SARMAG+16]; /* Read magic string + the name of the first archive header */
   long fpos=ftell(fp);
   int ret=fread(buffer,SARMAG+16,1,fp); 
   fseek(fp,fpos,SEEK_SET);
   if (ret!=1) return FALSE;
   if ((strncmp("________64E",buffer+SARMAG,11)==0)&& (strncmp(ARMAG,buffer,SARMAG)==0)) return TRUE;
   return FALSE;
}


void ArchArOpen(FILE * fp, t_archive * ar)
{
   char * strtbl=NULL;
   t_ar_archive_header tmp;
   t_uint32 max;
   t_uint32 stblsize=0;  
   t_arch_obj_hash_node * aohn=NULL;
   /* Eventually, we'll need subtypes since different arch's exist but normally
    * they'll only differ in symbob table lookup functions and stuff ... */

   ar->open_fp=fp;
   ar->objects=HashTableNew(47,0,string_hash,string_cmp,ArchObjHashNodeFree);
   fseek(fp,SARMAG,SEEK_SET);
   fread(&tmp,1,sizeof(t_ar_archive_header),fp);
   max=StringToUint32(tmp.ar_size,10);
   fseek(fp,max,SEEK_CUR);
   while(1)
   {
      VERIFY_FASSERT (strncmp(tmp.ar_fmag,ARFMAG,2)==0, ("Corrupted archive or unsupported archivetype %s",ar->name));
      if (fread(&tmp,sizeof(t_ar_archive_header),1,fp)!=1) break;
      if (strncmp(tmp.ar_name,"//",2)==0)
      {
	 strtbl=(char *) Malloc(StringToUint32(tmp.ar_size,10)+1);
	 fread(strtbl,StringToUint32(tmp.ar_size,10),1,fp);
	 stblsize=StringToUint32(tmp.ar_size,10);
      }
      else
      {
	 char * rname=NULL;
	 if (tmp.ar_name[0]!='/')
	 {
	    long size=0;
	    while(tmp.ar_name[size]!='/')
	    {
	       size++;
	    }

	    rname=(char *) Malloc(size+1);
	    strncpy(rname,tmp.ar_name,size);
	    rname[size]=0;
	 }
	 else
	 {
	    t_uint32 size=0;
	    t_uint32 start=StringToUint32(tmp.ar_name+1,15);

	    VERIFY_FASSERT(strtbl!=NULL,("Long name, but no string table found in %s",ar->name));
	    while((strtbl[start+size]!='/')&&(stblsize>start+size))
	    {
	       size++;
	    }

	    rname=(char *) Malloc(size+1);
	    strncpy(rname,strtbl+start,size);
	    rname[size]=0;
	 }

	 aohn=Malloc(sizeof(t_arch_obj_hash_node));
	 aohn->node.key=rname;
	 aohn->fpos=ftell(fp);
	 aohn->size=StringToUint32(tmp.ar_size,10);
	 aohn->flags=0;
	 HashInsert(ar->objects,(data_hash_node *) aohn);
	 max=StringToUint32(tmp.ar_size,10);
	 fseek(fp,2*(max/2+max%2),SEEK_CUR); /* Hey, it does the trick.... Needed this on linux. */
      }
   }
   if (strtbl) Free(strtbl);
}

void ArchAlphaArOpen(FILE * fp, t_archive * ar)
{
   /* OK, Tru64Unix archives have support for compressed objects and stuff, 
    * Need to handle that different */ 
   char * strtbl=NULL;
   t_ar_archive_header tmp;
   t_uint32 max;
   t_uint32 stblsize=0; 
   t_arch_obj_hash_node * aohn=NULL;

   ar->open_fp=fp;
   ar->objects=HashTableNew(47,0,string_hash,string_cmp,ArchObjHashNodeFree);
   fseek(fp,SARMAG,SEEK_SET);
   fread(&tmp,1,sizeof(t_ar_archive_header),fp);
   max=StringToUint32(tmp.ar_size,10);
   fseek(fp,max,SEEK_CUR);
   while(1)
   {
      VERIFY_FASSERT ((strncmp(tmp.ar_fmag,ARFZMAG,2)==0)||(strncmp(tmp.ar_fmag,ARFMAG,2)==0), ("Corrupted archive or unsupported archivetype %s (magix=%.35s)",ar->name,tmp.ar_fmag));
      if (fread(&tmp,sizeof(t_ar_archive_header),1,fp)!=1) break;
      if (strncmp(tmp.ar_name,"//",2)==0)
      {
	 strtbl=(char *) Malloc(StringToUint32(tmp.ar_size,10)+1);
	 fread(strtbl,StringToUint32(tmp.ar_size,10),1,fp);
	 stblsize=StringToUint32(tmp.ar_size,10);
	 
      }
      else
      {
	 char * rname=NULL;
	 if (tmp.ar_name[0]!='/')
	 {
	    long size=0;
	    while(tmp.ar_name[size]!=' ')
	    {
	       size++;
	    }

	    rname=(char *) Malloc(size+1);
/*#ifdef VISUALSTUDIO
		_snprintf(rname,size+1,"%s",tmp.ar_name);
#else
	    snprintf(rname,size+1,"%s",tmp.ar_name); */
/*#endif */

	    strncpy(rname,tmp.ar_name,size);
	    rname[size]=0;
	 }
	 else
	 {
	    t_uint32 size=0;
	    t_uint32 start=StringToUint32(tmp.ar_name+1,15);

	    VERIFY_FASSERT(strtbl!=NULL,("Long name, but no string table found in %s",ar->name));
	    while((strtbl[start+size]!='/')&&(stblsize>start+size))
	    {
	       size++;
	    }

	    rname=(char *) Malloc(size+1);
	    strncpy(rname,strtbl+start,size);
	    rname[size]=0;
	 }


	 aohn=Malloc(sizeof(t_arch_obj_hash_node));
	 aohn->node.key=rname;
	 aohn->fpos=ftell(fp);
	 aohn->size=StringToUint32(tmp.ar_size,10);
	 if (strncmp(tmp.ar_fmag,ARFZMAG,2)==0)
	    aohn->flags=COMPRESSED_ALPHA;
	 else if (strncmp(tmp.ar_fmag,ARFMAG,2)==0)
	    aohn->flags=0;
	 else FATAL(("MAGIC IS WRONG!"));
	 HashInsert(ar->objects,(data_hash_node *) aohn);
	 max=StringToUint32(tmp.ar_size,10);
	 fseek(fp,aohn->fpos+2*(max/2+max%2),SEEK_SET); /* Hey, it does the trick.... Needed this on linux and Alpha*/
      }
   }
   if (strtbl) Free(strtbl);
}


