/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "elf_main.h"
#include "../../utils/util_endian.h"

/* Endianness swapping routines */

void Elf32HdrSwitchEndian(Elf32_Ehdr * hdr)
{
   hdr->e_type      = UINT16_SWAP_ENDIAN(hdr->e_type);
   hdr->e_machine   = UINT16_SWAP_ENDIAN(hdr->e_machine);
   hdr->e_version   = UINT32_SWAP_ENDIAN(hdr->e_version);
   hdr->e_entry     = UINT32_SWAP_ENDIAN(hdr->e_entry);
   hdr->e_phoff     = UINT32_SWAP_ENDIAN(hdr->e_phoff);
   hdr->e_shoff     = UINT32_SWAP_ENDIAN(hdr->e_shoff);
   hdr->e_flags     = UINT32_SWAP_ENDIAN(hdr->e_flags);
   hdr->e_ehsize    = UINT16_SWAP_ENDIAN(hdr->e_ehsize);
   hdr->e_phentsize = UINT16_SWAP_ENDIAN(hdr->e_phentsize);
   hdr->e_phnum     = UINT16_SWAP_ENDIAN(hdr->e_phnum);
   hdr->e_shentsize = UINT16_SWAP_ENDIAN(hdr->e_shentsize);
   hdr->e_shnum     = UINT16_SWAP_ENDIAN(hdr->e_shnum);
   hdr->e_shstrndx  = UINT16_SWAP_ENDIAN(hdr->e_shstrndx);
}

void  Elf64ShdrSwitchEndian(Elf64_Shdr * shdr)
{
   shdr->sh_name      = UINT32_SWAP_ENDIAN(shdr->sh_name);
   shdr->sh_type      = UINT32_SWAP_ENDIAN(shdr->sh_type);
   shdr->sh_flags     = UINT64_SWAP_ENDIAN(shdr->sh_flags);
   shdr->sh_addr      = UINT64_SWAP_ENDIAN(shdr->sh_addr);
   shdr->sh_offset    = UINT64_SWAP_ENDIAN(shdr->sh_offset);
   shdr->sh_size      = UINT64_SWAP_ENDIAN(shdr->sh_size);
   shdr->sh_link      = UINT32_SWAP_ENDIAN(shdr->sh_link);
   shdr->sh_info      = UINT32_SWAP_ENDIAN(shdr->sh_info);
   shdr->sh_addralign = UINT64_SWAP_ENDIAN(shdr->sh_addralign);
   shdr->sh_entsize   = UINT64_SWAP_ENDIAN(shdr->sh_entsize);
}
void  Elf32ShdrSwitchEndian(Elf32_Shdr * shdr)
{
   shdr->sh_name      = UINT32_SWAP_ENDIAN(shdr->sh_name);
   shdr->sh_type      = UINT32_SWAP_ENDIAN(shdr->sh_type);
   shdr->sh_flags     = UINT32_SWAP_ENDIAN(shdr->sh_flags);
   shdr->sh_addr      = UINT32_SWAP_ENDIAN(shdr->sh_addr);
   shdr->sh_offset    = UINT32_SWAP_ENDIAN(shdr->sh_offset);
   shdr->sh_size      = UINT32_SWAP_ENDIAN(shdr->sh_size);
   shdr->sh_link      = UINT32_SWAP_ENDIAN(shdr->sh_link);
   shdr->sh_info      = UINT32_SWAP_ENDIAN(shdr->sh_info);
   shdr->sh_addralign = UINT32_SWAP_ENDIAN(shdr->sh_addralign);
   shdr->sh_entsize   = UINT32_SWAP_ENDIAN(shdr->sh_entsize);
}
void  Elf32SymSwitchEndian(Elf32_Sym * sym)
{
   sym->st_name  = UINT32_SWAP_ENDIAN(sym->st_name);
   sym->st_value = UINT32_SWAP_ENDIAN(sym->st_value);
   sym->st_size  = UINT32_SWAP_ENDIAN(sym->st_size);
   sym->st_shndx = UINT16_SWAP_ENDIAN(sym->st_shndx);
}

void  Elf32RelSwitchEndian(Elf32_Rel * rel)
{
   rel->r_offset = UINT32_SWAP_ENDIAN(rel->r_offset);
   rel->r_info   = UINT32_SWAP_ENDIAN(rel->r_info);
}
