/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef I386_SUPPORT 
#include "../../config.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "elf_main.h"
#include "elf_common32.h"
#include "elf_armsupport.h"
#include "../../utils/util_machine.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_section.h"

#include <string.h>
#include "elf_types.h"
#include "elf_endian.h"

typedef struct 
{
  t_section * sec;
  Elf32_Shdr hdr;
  t_int32 described_by_pdr;
  t_int32 prepadding; 
} sec_store_info;


t_bool IsElfI386SameEndianOnLsb(FILE * fp)
{
  Elf32_Byte buffer[EI_NIDENT];
  /*my_byteorder byte_order; */
  Elf32_Ehdr hdr;
  /*int tel=0;*/
  Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
  Filepos save=ftell(fp);
  
  if (global_machine_info.endian!=LSB)
    return FALSE;
  
  if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
  
  fseek(fp,save,SEEK_SET);
  
  if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
    FATAL(("Not an elf objectfile"));
  
  switch (buffer[EI_DATA])
  {
    default:
    case ELFDATANONE:
    case ELFDATA2MSB:
      return FALSE;
    case ELFDATA2LSB:
      break;
  }
  
  switch (buffer[EI_CLASS])
  {
    case ELFCLASS32: 
      if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
      {
        FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
      }
      fseek(fp,save,SEEK_SET);
      machine=hdr.e_machine;
      break;
    case ELFCLASS64: 
    case ELFCLASSNONE:
    default:
      return FALSE;
      break;
  }

  if (machine!=EM_386)
    return FALSE; 
  
  return TRUE;
}

t_bool IsElfI386SwitchedEndianOnMsb(FILE * fp)
{
  Elf32_Byte buffer[EI_NIDENT];
  /*my_byteorder byte_order; */
  Elf32_Ehdr hdr;
  /*int tel=0;*/
  Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
  Filepos save=ftell(fp);
  
  if (global_machine_info.endian!=MSB)
    return FALSE;
  
  if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
  
  fseek(fp,save,SEEK_SET);

  if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
    FATAL(("Not an elf objectfile"));
  
  switch (buffer[EI_DATA])
  {
    default:
    case ELFDATANONE:
    case ELFDATA2MSB:
      return FALSE;
    case ELFDATA2LSB:
      break;
  }
  
  switch (buffer[EI_CLASS])
  {
    case ELFCLASS32: 
      if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
      {
        FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
      }
      fseek(fp,save,SEEK_SET);
      machine=hdr.e_machine;
      break;
    case ELFCLASS64: 
    case ELFCLASSNONE:
    default:
      return FALSE;
      break;
  }
  
  if (machine!=EM_SWITCHED_386)
    return FALSE; 
  
  return TRUE;
}

void ElfWriteI386SameEndian(t_object * obj, t_string filename)
{
  FILE * fp=fopen(filename,"w");
  int tel=0;
  Elf32_Ehdr hdr;
  ANALYSIS_START("Writing I386 binary");
  hdr.e_ident[EI_MAG0]=ELFMAG0;
  hdr.e_ident[EI_MAG1]=ELFMAG1;
  hdr.e_ident[EI_MAG2]=ELFMAG2;
  hdr.e_ident[EI_MAG3]=ELFMAG3;
  hdr.e_ident[EI_VERSION]=EV_CURRENT;
  hdr.e_ident[EI_DATA]=ELFDATA2LSB;
  hdr.e_ident[EI_CLASS]=ELFCLASS32;
   
  /* Padding : EI_PAD=7 from e_ident[7] to e_ident[15] */
  for (tel=EI_PAD; tel<16; tel++)
  {
    hdr.e_ident[tel]=0;
  }

  /* Size of e_ident[] is already declared */
  
  hdr.e_type=ET_EXEC;
  hdr.e_machine=EM_386;
  hdr.e_version=EV_CURRENT;
  hdr.e_entry=G_T_UINT32(obj->entry); 
  hdr.e_phoff=0;
  /* The 32-bit Intel Architecture defines no flags; so this member contains zero. */
  hdr.e_flags=0;
  hdr.e_ehsize=sizeof(Elf32_Ehdr);
  /* Set in common */
  hdr.e_phentsize=0;
  hdr.e_phnum=0;
  hdr.e_shentsize=0; 
  hdr.e_shnum=0; 
  hdr.e_shstrndx=0; 
  
  ElfWriteCommonSameEndian(fp,&hdr,obj,1<<12); 
  ANALYSIS_SIMPLE_END("Writing I386 binary");
}

void ElfReadI386SameEndian(FILE * fp, t_object * obj)
{
  Elf32_Shdr * shdrs=NULL;
  Elf32_Sym * symbol_table=NULL;
  t_section ** sec_ret=NULL;
  char * strtab=NULL;
  int numsyms=0;
  int tel;
  char * sechdrstrtab=NULL;
  Elf32_Ehdr hdr32;
  t_symbol ** table; 
  if (fread((&hdr32),sizeof(Elf32_Ehdr),1,fp)!=1)
  {
    FATAL(("Could not read elf header!"));
  }

   /* Reading elf headers does not depend on the type of objectfile */
  ElfReadCommon32(fp, &hdr32, obj, &shdrs, &symbol_table, &numsyms, &strtab, &sechdrstrtab,&sec_ret,&table,FALSE); 
  obj->reloc_table=RelocTableNew();
  if (hdr32.e_shnum)
  {
    tel=0;
    /* STEP 5: Find the relocations */ 
    while (tel<hdr32.e_shnum)
    {
      if (PatternMatch(".rel.debug*",sechdrstrtab + shdrs[tel].sh_name))
      {
        /* if we ignore debug info, we can also ignore debug relocations */
      }
      else if (PatternMatch(".rela*",sechdrstrtab + shdrs[tel].sh_name))
      {
        FATAL(("RELA relocations not implemented!"));
      }
      else if (PatternMatch(".rel.stab*",sechdrstrtab + shdrs[tel].sh_name))
      {
        /* Lets ignore these */
      }
      else if (PatternMatch(".rel*",sechdrstrtab + shdrs[tel].sh_name))
      {
        Elf32_Rel * rel=Malloc(shdrs[tel].sh_size);
        int tel2;
        t_section_hash_node * sh;


        fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
        fread(rel,shdrs[tel].sh_size,1,fp);
        /*printf("-- Relocations (%s for %s)\n",sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name,sechdrstrtab + shdrs[tel].sh_name); */

        sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);
        
        VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
        
        for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rel)); tel2++)
        {
          VERIFY_FASSERT(symbol_table && strtab,("Symtab or stringtab not found\n Implement me"));
          VERIFY_FASSERT(ELF32_R_TYPE(rel[tel2].r_info)<=16, ("Implement relocation %d",ELF32_R_TYPE(rel[tel2].r_info)));
          switch(ELF32_R_TYPE(rel[tel2].r_info))
          {
            case R_386_32:
            {
	       t_uint32 data;
	       t_symbol * sym;
	       t_address generic,addend;
	       sym=table[ELF32_R_SYM(rel[tel2].r_info)];
	       if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
	       GenericSet_uint32(generic,rel[tel2].r_offset);
	       data=SectionGetData32(sh->section,generic);
	       GenericSet_uint32(addend,data);
	       RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,FALSE);
	       /*DiabloPrint(stdout,"!!! ABS32 reloc in sec %s offset @G of file %s\n",sh->section->name,generic,obj->name); */
	    }
            break;
            case R_386_PC32:
            {
	      t_uint32 data;
	      t_address generic;
	      t_address addend;
	      t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

	      GenericSet_uint32(generic,rel[tel2].r_offset);
	      data=SectionGetData32(sh->section,generic);
	      GenericSet_uint32(addend,data);

	      RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,CALC_REL "\\" WRITE_32,FALSE);
	      /*DiabloPrint(stdout,"!!! PC32 reloc in sec %s offset @G of file %s\n",sh->section->name,generic,obj->name); */
            }
            break;
            case R_386_GOT32:
              /* toevoegen van datasectie -> adres invullen, relocatie naar daar, daarna relocatie van daar naar uiteindelijk doel*/
            {
	       t_uint32 data;
	      t_address generic;
	      t_address addend;
	      t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

	      GenericSet_uint32(generic,rel[tel2].r_offset);
	      data=SectionGetData32(sh->section,generic);
	      GenericSet_uint32(addend,data);

	      RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,"GA+" "\\" WRITE_32,FALSE);

	      /*DiabloPrint(stdout,"!!! GOT32 reloc in sec %s offset @G of file %s\n",sh->section->name,generic,obj->name); */
            }
              break;
            case R_386_GOTPC:
            {
	       t_uint32 data;
	       t_address generic;
	       t_address addend;
	       t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

	       GenericSet_uint32(generic,rel[tel2].r_offset);
	       data=SectionGetData32(sh->section,generic);
	       GenericSet_uint32(addend,data);

	       RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,"SA+P-" "\\" WRITE_32,FALSE);
	       /*DiabloPrint(stdout,"!!! GOTPC reloc to sym @S in sec %s offset @G of file %s\n",sym,sh->section->name,generic,obj->name); */
#if 0
              if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
              GenericSet_uint32(generic,rel[tel2].r_offset);
              RelocTableAddReloc(obj->reloc_table,sh->section,generic, sym,REPRESENT_SELFREL_VALUE, GOTPC,RELOC_FROM_SECTION,NULL);
#endif
            }
              /*printf("GOTPC\n"); */
              break;
            case R_386_GOTOFF:
	      {
		t_uint32 data;
		t_symbol * sym;
		t_address generic;
		t_address addend;

		sym=table[ELF32_R_SYM(rel[tel2].r_info)];
		GenericSet_uint32(generic,rel[tel2].r_offset);

		data=SectionGetData32(sh->section,generic);
		GenericSet_uint32(addend,data);

		RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,"SA+g-\\" WRITE_32,FALSE);
	      }
              /*printf("GOTOFF\n"); */
              break;
            case R_386_PLT32:
	      {
		/* PLT32 in a statically linked program == REL32 */
		t_uint32 data;
		t_address generic;
		t_address addend;
		t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

		GenericSet_uint32(generic,rel[tel2].r_offset);
		data=SectionGetData32(sh->section,generic);
		GenericSet_uint32(addend,data);

		RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,CALC_REL "\\" WRITE_32,FALSE);
		/*DiabloPrint(stdout,"!!! PLT32 reloc in sec %s offset @G of file %s\n",sh->section->name,generic,obj->name); */
	      }
              break;
            case R_386_COPY:
            case R_386_GLOB_DAT:
            case R_386_JMP_SLOT:
            default:
              FATAL(("Implement relocation %d!",ELF32_R_TYPE(rel[tel2].r_info)));
          }
        }
        Free(rel);
	 }
      tel++;
    }
  }
  
   if (shdrs) Free(shdrs);
   if (sechdrstrtab) Free(sechdrstrtab);
   if (symbol_table) Free(symbol_table);
   if (strtab) Free(strtab);
   if (table) Free(table);
   
}

void ElfReadI386SwitchedEndian(FILE * fp,  t_object * obj)
{
  FATAL(("386 ELF on an machine with different endianness currently not supported. Support for 386 ELF on machines with the same endianness is very rudementary"));
}
#endif
