/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef IA64_SUPPORT
#include "../../config.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "elf_main.h"
#include "elf_ia64support.h"
#include "../../utils/util_machine.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_object.h"

t_bool IsElfIA64SameEndianOnLsb(FILE *fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf64_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=LSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS64: 
	 if (fread(&(hdr),sizeof(Elf64_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS32: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_IA_64)
      return FALSE; 

   return TRUE;
}

t_bool IsElfIA64SwitchedEndianOnMsb (FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf64_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=MSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS64: 
	 if (fread(&(hdr),sizeof(Elf64_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS32: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_SWITCHED_IA_64)
      return FALSE; 

   return TRUE;
}
//Elf64ShdrSwitchEndian


void ElfWriteIA64SameEndian(t_object * obj, t_string filename)
{
   FILE * fp=fopen(filename,"w");
   int tel=0;
   Elf64_Ehdr hdr;
   ANALYSIS_START("Writing IA64 binary");
   hdr.e_ident[EI_MAG0]=ELFMAG0;
   hdr.e_ident[EI_MAG1]=ELFMAG1;
   hdr.e_ident[EI_MAG2]=ELFMAG2;
   hdr.e_ident[EI_MAG3]=ELFMAG3;
   hdr.e_ident[EI_VERSION]=EV_CURRENT;
   if (global_machine_info.endian==LSB)
      hdr.e_ident[EI_DATA]=ELFDATA2LSB;
   else
      hdr.e_ident[EI_DATA]=ELFDATA2MSB;

   hdr.e_ident[EI_CLASS]=ELFCLASS64;
   for (tel=EI_PAD; tel<16; tel++)
   {
      hdr.e_ident[tel]=0;
   }
/*TODO: de velden nog juist zetten... */
   hdr.e_type=ET_EXEC;
   hdr.e_machine=EM_IA_64;
   hdr.e_version=EV_CURRENT;
   hdr.e_entry=G_T_UINT64(obj->entry); 
   hdr.e_phoff=0;
   hdr.e_flags=EF_IA_64_ABI64;
   hdr.e_ehsize=0;
   hdr.e_phentsize=0;
   hdr.e_phnum=0;
   hdr.e_shentsize=0; /* Set in common */
   hdr.e_shnum=0; /* Set in common */
   hdr.e_shstrndx=0; /* Set in common */

   ElfWriteCommonSameEndian64(fp,&hdr,obj, 1<<14/*getpagesize()*/); 
   ANALYSIS_SIMPLE_END("Writing IA64 binary");
}

/* example from test
$1 = {
e_ident = "\177ELF\002\001\001\0\0\0\0\0\0\0\0", 
e_type = 2, 
e_machine = 50, 
e_version = 1, 
e_entry = 4611686018427388448, 
e_phoff = 64, 
e_shoff = 2992024, 
e_flags = 16, 
e_ehsize = 64, 
e_phentsize = 56, 
e_phnum = 4, 
e_shentsize = 64, 
e_shnum = 31, 
e_shstrndx = 28
}
*/


void ElfReadIA64SameEndian(FILE * fp, t_object * obj)
{
   Elf64_Shdr * shdrs=NULL;
   Elf64_Sym * symbol_table=NULL;
   char * strtab=NULL;
   int numsyms=0;
   t_uint32 tel;
   char * sechdrstrtab=NULL;
   Elf64_Ehdr hdr64;
   t_symbol ** table;

   if (fread((&hdr64),sizeof(Elf64_Ehdr),1,fp)!=1)
   {
      FATAL(("Could not read elf header!"));
   }


   /* Reading elf headers does not depend on the type of objectfile */
   ElfReadCommon64(fp, &hdr64, obj, &shdrs, &symbol_table, &numsyms, &strtab, &sechdrstrtab,&table,FALSE); 
   obj->reloc_table=RelocTableNew();
   if (hdr64.e_shnum)
   {
      tel=0;
      /* STEP 5: Find the relocations */ 
      while (tel<hdr64.e_shnum)
      {
	 if (PatternMatch(".rel.debug*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    /* if we ignore debug info, we can also ignore debug relocations */
	 }
	 else if (PatternMatch(".rela*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    Elf64_Rela * rel=Malloc(shdrs[tel].sh_size);
	    int tel2;
	    t_section_hash_node * sh;
	    tel2=fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
//	    printf("%s ======%lld======== %lld ==========%d\n",obj->name, obj->streampos+shdrs[tel].sh_offset,shdrs[tel].sh_size,tel2);
	    fread(rel,shdrs[tel].sh_size,1,fp);
	    if (!(shdrs[shdrs[tel].sh_info].sh_flags & SHF_ALLOC)) { tel++; Free(rel); continue;}
	    sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);
	    VERIFY_FASSERT( sh!=NULL ,("Base section for relocations (%s - %s) not found!",sechdrstrtab + shdrs[tel].sh_name,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name));
	    for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf64_Rela)); tel2++)
	    {
	       switch(ELF64_R_TYPE(rel[tel2].r_info))
	       {
		 case R_IA64_NONE:
		   break;
		 case R_IA64_IPLTLSB:	        /* dynamic reloc, imported PLT, LSB */
		   FATAL(("Implement R_IA64_IPLTLSB"));
		   break;
		 case R_IA64_LTOFF_FPTR22:	
		   {
		     /* @ltoff(@fptr(s+a)), imm22: Ask for the creation of a GOT
		     * entry, containing the address of the official procedure
		     * descriptor of the function at (s+a): We present this as
		     * a pointer to the symbol of type  */
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);
		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,"q" ,FALSE);
		   }
		   break;
		 case R_IA64_PCREL21B:		
		   /* @pcrel(sym + add), ptb, call: Symbol + Addend - Address */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     //DiabloPrint(stdout, "@S\n", sym);
		     //if(sym->to_type != TO_NONE)
		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,CALC_REL_ALIGN "\\" WRITE_IMM21,FALSE);
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_SELFREL_VALUE,INSTRUCTION_IMMEDIATE21B,RELOC_FROM_SECTION,NULL);
		   }

		   break;
		 case R_IA64_GPREL22:		/* @gprel(sym + add), add imm22 */
		   /* The value, sym, write gp relative / imm22 */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,"SA+g-" "\\"  WRITE_IMM22,FALSE);
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_GPREL_VALUE,INSTRUCTION_IMMEDIATE22,RELOC_FROM_SECTION,NULL);
		   }

		   break;
		 case R_IA64_LTOFF22:		/* @ltoff(sym + add), add imm22 */
		   /*  Ask for the creation of a GOT entry, containing the address of the sym + add */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,"u",FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_LT_POINTER,INSTRUCTION_IMMEDIATE22,RELOC_FROM_SECTION,NULL);
		   }

		   break;
		 case R_IA64_DIR64LSB:	        /* symbol + addend, data8 LSB */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,"SA+" "\\" WRITE_64 ,FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_VALUE,ABS64,RELOC_FROM_SECTION,NULL);
		   }
		   break;
		 case R_IA64_SEGREL64LSB:      /* @segrel(sym + add), data8 LSB */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,"SA+T-" "\\" WRITE_64 ,FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_SEGREL_VALUE,ABS64,RELOC_FROM_SECTION,NULL);
		   }
		   break;
		 case R_IA64_GPREL64I:		/* @gprel(sym + add), mov imm64 */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     //DiabloPrint(stdout,"---------------- @S\n",sym);
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,
			 "SA+g-" "\\" WRITE_IMM64 ,FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_GPREL_VALUE,INSTRUCTION_IMMEDIATE64,RELOC_FROM_SECTION,NULL);
		   }
		   break;
		 case R_IA64_FPTR64LSB:	/* @fptr(sym + add), data8 LSB */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol %d",ELF64_R_SYM(rel[tel2].r_info)));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,"p" "\\" WRITE_64 ,FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_OPD_POINTER,ABS64,RELOC_FROM_SECTION,NULL);
		   }
		   break;
		 case R_IA64_GPREL64LSB:	/* @gprel(sym + add), data8 LSB */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,"SA+g-" "\\" WRITE_64 ,FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_GPREL_VALUE,ABS64,RELOC_FROM_SECTION,NULL);
		   }
		   break;
		 case R_IA64_PCREL64I:		/* @pcrel(sym + add), 64bit inst */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,CALC_REL_ALIGN "\\" WRITE_IMM64 ,FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_SELFREL_VALUE,INSTRUCTION_IMMEDIATE64,RELOC_FROM_SECTION,NULL);
		   }
		   break;
		 case R_IA64_PCREL64LSB:       /* @pcrel(sym + add), data8 LSB */
		   {
		     t_address generic;
		     t_address addend;
		     t_symbol * sym=table[ELF64_R_SYM(rel[tel2].r_info)];
		     VERIFY_FASSERT(sym,("No symbol"));
		     GenericSet_uint64(generic,rel[tel2].r_offset);
		     GenericSet_uint64(addend,rel[tel2].r_addend);
		     /*		      
		     //printf("pcrel64lsb %llx %llx:\n", rel[tel2].r_offset,rel[tel2].r_addend  );
		     printf("Type: %20s; offset: %16llx; addend: %16llx; section: %s\n", "R_IA64_PCREL64LSB", 
		     rel[tel2].r_offset, addend, sh->section->name);
		     printf("%20s %20s\n", "REPRESENT_SELFREL_VALUE", "ABS64");
		     SymbolPrint(sym);
		      */		      
		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,(t_relocatable *)sh->section,generic,sym,TRUE,NULL,NULL,NULL,CALC_REL_ALIGN "\\" WRITE_64 ,FALSE);
		     //FATAL(("Implement"));
		     //RelocTableAddRelocWithAddend(obj->reloc_table,sh->section, generic,addend,sym,REPRESENT_SELFREL_VALUE,ABS64,RELOC_FROM_SECTION,NULL);
		   }
		   break;
		 default:
		   FATAL(("Implement relocation type 0x%x\n",ELF64_R_TYPE(rel[tel2].r_info)));
	       }
	    }
	    Free(rel);
	 }
	 else if (PatternMatch(".rel.stab*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	   /* Lets ignore these */
	 }
	 else if (PatternMatch(".rel*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    Elf64_Rel * rel=Malloc(shdrs[tel].sh_size);
	    t_section_hash_node * sh;
	    fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
	    fread(rel,shdrs[tel].sh_size,1,fp);
	    sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);
            VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
	    FATAL(("Implement relocation without addends for Itanium"));
	    Free(rel);
	 }
	 tel++;
      }
   }

   if (shdrs) Free(shdrs);
   if (sechdrstrtab) Free(sechdrstrtab);
   if (symbol_table) Free(symbol_table);
   if (strtab) Free(strtab);

}

void ElfReadIA64SwitchedEndian(FILE * fp, t_object * obj)
{



   FATAL(("You are trying to optimize an Itanium binary. That is currently not supported"));
}
#endif
/* vim: set shiftwidth=2: */
