/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef MIPS_SUPPORT
#include "../../kernel/diablo_reloc.h"
#include "../../config.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_print.h"
#include "../../arch/mips/mips_opcodes.h"
#include "../../arch/mips/mips_disassemble.h"
#include "../../arch/mips/mips_assemble.h"
#include "../../arch/mips/mips_instructions.h"
#include "elf_mips.h"
#include "elf_main.h"
#include "elf_mipssupport.h"
#include "elf_common32.h"
#include "../../utils/util_machine.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_object.h"
#include <string.h>
#define sign_extend(x,p)      (((x) ^ (1 << (p))) - (1 << (p)))
#define SELECT_BITS(i,b,e) ((i&((1<<(b+1))-1))>>e)
/* Common operations on objectfiles */

/* Write an elf object file. The elfhdr is supplied by the arch-specific code */

t_bool IsElfMipsSwitchedEndianOnMsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf32_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=MSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_SWITCHED_MIPS)
      return FALSE; 

   return TRUE;
}


t_bool IsElfMipsSameEndianOnLsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   Elf32_Ehdr hdr;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=LSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_MIPS)
      return FALSE; 

   return TRUE;
}

void ElfWriteMipsSameEndian(t_object * obj, t_string filename)
{
   FILE * fp=fopen(filename,"w");
   int tel=0;
   Elf32_Ehdr hdr;
   ANALYSIS_START("Writing Mips binary");
   hdr.e_ident[EI_MAG0]=ELFMAG0;
   hdr.e_ident[EI_MAG1]=ELFMAG1;
   hdr.e_ident[EI_MAG2]=ELFMAG2;
   hdr.e_ident[EI_MAG3]=ELFMAG3;
   hdr.e_ident[EI_VERSION]=EV_CURRENT;
   if (global_machine_info.endian==LSB)
      hdr.e_ident[EI_DATA]=ELFDATA2LSB;
   else
      hdr.e_ident[EI_DATA]=ELFDATA2MSB;

   hdr.e_ident[EI_CLASS]=ELFCLASS32;
   for (tel=EI_PAD; tel<16; tel++)
   {
      hdr.e_ident[tel]=0;
   }
   hdr.e_type=ET_EXEC;
   hdr.e_machine=EM_MIPS;
   hdr.e_version=EV_CURRENT;
   hdr.e_entry=G_T_UINT32(obj->entry); 
   hdr.e_phoff=0;
   hdr.e_flags=0;
   hdr.e_ehsize=0;
   hdr.e_phentsize=0;
   hdr.e_phnum=0;
   hdr.e_shentsize=0; /* Set in common */
   hdr.e_shnum=0; /* Set in common */
   hdr.e_shstrndx=0; /* Set in common */

   ElfWriteCommonSameEndian(fp,&hdr,obj,4096); 
   ANALYSIS_SIMPLE_END("Writing Mips binary");
}

void ElfReadMipsSameEndian(FILE * fp, t_object * obj)
{
   Elf32_Shdr * shdrs=NULL;
   Elf32_Sym * symbol_table=NULL;
   char * strtab=NULL;
   int numsyms=0;
   int tel;
   char * sechdrstrtab=NULL;
   t_section ** sec_table=NULL;
   Elf32_Ehdr hdr32;
   t_symbol ** table;
   t_symbol * gp;
   t_symbol * gp0;
   if (fread((&hdr32),sizeof(Elf32_Ehdr),1,fp)!=1)
   {
      FATAL(("Could not read elf header!"));
   }

   /* Reading elf headers does not depend on the type of objectfile */
   ElfReadCommon32(fp, &hdr32, obj, &shdrs, &symbol_table, &numsyms, &strtab, &sechdrstrtab, &sec_table, &table,FALSE); 

   gp0 = SymbolTableGetSymbolByName(obj->symbol_table,"_gp");
    gp = SymbolTableAddUndefinedSymbol(obj->symbol_table,"__global");

   obj->reloc_table=RelocTableNew();
   if (hdr32.e_shnum)
   {
     tel=0;
     while (tel<hdr32.e_shnum)
     {
         /* {{{ Ignoring */
         if (PatternMatch(".rela.debug*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    /* if we ignore debug info, we can also ignore debug relocations */
	 }
	 else if (PatternMatch(".rela.pdr*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    /* if we ignore debug info, we can also ignore debug relocations */
	 }
	 else if (PatternMatch(".rel.debug*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    /* if we ignore debug info, we can also ignore debug relocations */
	 }
	 else if (PatternMatch(".rel.pdr*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    /* if we ignore debug info, we can also ignore debug relocations */
	 }
	 else if (PatternMatch(".rela*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	    Elf32_Rela * rel=Malloc(shdrs[tel].sh_size);
	    int tel2;
	    t_section_hash_node * sh;

	    if (strcmp(sechdrstrtab + shdrs[tel].sh_name,".rela.dyn")==0)
	    {
	      FATAL(("Dynamic relocations not implemented!"));
	    }
	    
	    fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
	    fread(rel,shdrs[tel].sh_size,1,fp);

	    sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);
	   
            VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
	    
	    for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rela)); tel2++)
	    {
	       VERIFY_FASSERT(symbol_table && strtab,("Symtab or stringtab not found\n Implement me"));
	       VERIFY_FASSERT(ELF32_R_TYPE(rel[tel2].r_info)<=16, ("Implement relocation %d",ELF32_R_TYPE(rel[tel2].r_info)));
	       switch(ELF32_R_TYPE(rel[tel2].r_info))
	       {
		 default:
		   FATAL(("Implement relocation %d\n for mips",ELF32_R_TYPE(rel[tel2].r_info)));
		   exit(0);
		   break;
	       }
	    }
	    Free(rel);
	 }
	 else if (PatternMatch(".rel.stab*",sechdrstrtab + shdrs[tel].sh_name))
	 {
	   /* Lets ignore these */
	 } /* }}} */
	 else if (PatternMatch(".rel*",sechdrstrtab + shdrs[tel].sh_name) && (shdrs[tel].sh_size != 0))
	 {
	    Elf32_Rel * rel=Malloc(shdrs[tel].sh_size);
	    t_bool prev_hi16 = FALSE;
	    t_bool prev_hi16_added = FALSE;
	    t_bool prev_got = FALSE;
	    t_relocatable *relocatable_hi16;
	    t_address addend_hi16,generic_hi16;
	    t_symbol *sym_hi16;
	    t_reloc * got16_corr;
	    t_reloc *temp;
	    int tel2;
	    t_section_hash_node * sh;

	    if (strcmp(sechdrstrtab + shdrs[tel].sh_name,".rel.dyn")==0)
	    {
	      FATAL(("Dynamic relocations not implemented!"));
	    }
	    
	    fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
	    fread(rel,shdrs[tel].sh_size,1,fp);
	    sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);
            VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
	    
	    for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rel)); tel2++)
	    {
	       VERIFY_FASSERT(symbol_table && strtab,("Symtab or stringtab not found\n Implement me"));
	       switch(ELF32_R_TYPE(rel[tel2].r_info))
	       {
		 case R_MIPS_32:
		   {
                     t_address generic,addend;
		     t_uint32 data;
                     t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
                     GenericSet_uint32(generic,rel[tel2].r_offset);
                     if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));

                     data = SectionGetData32(sh->section,generic);
		     GenericSet_uint32(addend,data);

		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,TRUE);
		     prev_got = FALSE;
		   }
		   break;
		 case R_MIPS_26:
		   {
                     t_address generic,addend;
		     t_uint32 data;
		     t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
                     if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
                     GenericSet_uint32(generic,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic);
		     data = SELECT_BITS(data,25,0);
		     GenericSet_uint32(addend,data);
		     if (sym->local)
		     {
                       RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,FALSE,NULL,NULL,NULL,"As0002< Pif0000000& | S+ s0002 > \\ i03ffffff & = l ifc000000 &|w \\l i03ffffff &-$", TRUE);
		     }
		     else /* external symbol */
		     {
		       RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,FALSE,NULL,NULL,NULL,"As0002< #0f S+ s0002 > \\ i03ffffff & = l ifc000000 &|w \\l i03ffffff &-$", TRUE);
		     }
                     //prev_hi16 = FALSE;
		     prev_got = FALSE;
		   }
		   break;
		 case R_MIPS_HI16:
		   {
		     t_uint32 data;
		     sym_hi16=table[ELF32_R_SYM(rel[tel2].r_info)];
                     if (!sym_hi16) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
                     GenericSet_uint32(generic_hi16,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic_hi16);
		     data <<= 16;
		     GenericSet_uint32(addend_hi16,data);
		     relocatable_hi16 =T_RELOCATABLE(sh->section); 
                     prev_hi16 = TRUE;
		     prev_hi16_added = FALSE;
		     prev_got = FALSE;
		   }
		   break;
		 case R_MIPS_LO16:
		   {
		     t_address generic,addend;
		     t_uint32 data;
		     t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
		     if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
		     GenericSet_uint32(generic,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic);
		     data &= 0x0000ffff;
		     GenericSet_uint32(addend,data);
		     if(!prev_hi16) FATAL(("No previous HI16 for LO16!"));
		     if(prev_got)
		     {
		       temp=RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,got16_corr,NULL,"AS+ \\ sffff& = l iffff0000 &|w \\l sffff &-$", TRUE);
		       got16_corr->corresponding=temp;
		     }

		     else
		     {
		       addend = GENERIC_ADD(addend_hi16,addend);

		       if (!prev_hi16_added)
		       {
			 if (!strcmp(sym_hi16->name,"_gp_disp"))
			 {
			   RelocTableAddRelocToSymbol(obj->reloc_table,addend,relocatable_hi16,generic_hi16,sym_hi16,TRUE,NULL,NULL,gp,"A #0f R+ P- = sffff& #0f - s0010>  \\ sffff& = l iffff0000&|w\\l sffff &-$", TRUE);
			 }
			 else
			 {
			   RelocTableAddRelocToSymbol(obj->reloc_table,addend,relocatable_hi16,generic_hi16,sym_hi16,TRUE,NULL,NULL,NULL,"A #0f S+ = sffff& #0f - s0010> \\ sffff& = l iffff0000&|w\\l sffff &-$", TRUE);
			 }
			 prev_hi16_added = TRUE;
		       }

		       if (!strcmp(sym->name,"_gp_disp"))
		       {
			 RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,gp,"A R+ P- i00000004 + \\ sffff& = l iffff0000 &|w \\l sffff &-$", TRUE);
		       }
		       else /* not _gp_disp */
		       {
			 RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,"AS+ \\ sffff& = l iffff0000 &|w \\l sffff &-$", TRUE);
		       }
		     }
		     prev_got = FALSE;
		   }
		   break;
		 case R_MIPS_GPREL16:
		   {
		     t_address generic,addend;
		     t_uint32 data;
		     t_symbol * sym = table[ELF32_R_SYM(rel[tel2].r_info)];
		     if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
		     GenericSet_uint32(generic,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic);
		     GenericSet_uint32(addend,data);

		     if (sym->local)
		     {
		       /* Assuming here that GP0, the gp-value used to create the relocatable object, was zero (see
			* also System V ABI for mips) */


		       RelocTableAddRelocToSymbol(obj->reloc_table,GENERIC_ADD_INT32(addend,obj->gp),T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,gp,"A #0f S+ R- \\ sffff& = l iffff0000 &|w l sffff &-$",TRUE);
		     }
		     else  /* global */
		     {
		       RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,gp,"A #0f S+ R-  \\ sffff& = l iffff0000 &|w l sffff &-$",TRUE);
		     }

		     prev_got = FALSE;

		   }
		   break;
		 case R_MIPS_LITERAL:
		   {
                     t_address generic,addend;
		     t_uint32 data;
		     t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
                     if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
                     GenericSet_uint32(generic,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic);
		     GenericSet_uint32(addend,data);
		     RelocTableAddRelocToSymbol(obj->reloc_table,GENERIC_ADD_INT32(addend,obj->gp),T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,gp,"A #0f S+ R- r+\\ sffff& = l iffff0000 &|w \\l sffff &-$",TRUE);
		   }
		     prev_got = FALSE;
		   break;
		 case R_MIPS_REL32:
		   FATAL(("MIPS_REL32 relocation only for use with dynamic linker!"));
		   break;
		 case R_MIPS_GOT16:
		 case R_MIPS_CALL16:
		   {
                     t_address generic,addend;
		     t_uint32 data;
		     t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
                     if (!sym) 
		       FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
                     GenericSet_uint32(generic,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic);
		     data &= 0x0000ffff;
		     GenericSet_uint32(addend,data);
		     if (sym->local)
		     {
		       got16_corr=RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,"H \\ sffff& = l iffff0000 &|w \\l sffff &-$", TRUE);
		     }
		     else
		     {
		       got16_corr=RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL,"H \\ sffff& = l iffff0000 &|w \\l sffff &-$", TRUE);
		     }
		     prev_got = TRUE;
		     break;
		   }
		 case R_MIPS_GPREL32:
		   {
		     t_address generic,addend;
		     t_uint32 data;
		     t_symbol * sym = table[ELF32_R_SYM(rel[tel2].r_info)];
		     if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
		     GenericSet_uint32(generic,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic);
		     GenericSet_uint32(addend,data);
		     sym=SymbolTableAddSymbol(obj->symbol_table,sym->to_type,sym->type,sym->base,GENERIC_ADD(addend,SECTION_GET_CADDRESS(T_SECTION(sym->base))),sym->name,sym->local);
		       RelocTableAddRelocToSymbol(obj->reloc_table,obj->gp,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,gp,"A S+ R- \\" WRITE_32,TRUE);
		     
		     prev_got = FALSE;
		     break;
		   }
		 case R_MIPS_PC32:
		   {
                     t_address generic,addend;
		     t_uint32 data;
		     t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];
                     if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
                     GenericSet_uint32(generic,rel[tel2].r_offset);
		     data = SectionGetData32(sh->section,generic);
		     GenericSet_uint32(addend,data);
		     RelocTableAddRelocToSymbol(obj->reloc_table,addend,T_RELOCATABLE(sh->section),generic,sym,TRUE,NULL,NULL,NULL, CALC_REL "\\" WRITE_32,TRUE);
		     prev_got = FALSE;
		     break;
		   }
		 case R_MIPS_PC16:
		 case R_MIPS_16:
		 default:
		   FATAL(("Implement relocation %d\n for mips",ELF32_R_TYPE(rel[tel2].r_info)));
		   exit(0);
	       }
	    }
	    Free(rel);
	 }
       tel++;
     }
   }
   if (shdrs) Free(shdrs);
   if (sechdrstrtab) Free(sechdrstrtab);
   if (symbol_table) Free(symbol_table);
   if (strtab) Free(strtab);
   if (table) Free(table);
}

void ElfReadMipsSwitchedEndian(FILE * fp, t_object * obj)
{
	FATAL(("MIPS Switched endian not supported"));
}
#endif

/* vim : set shiftwidth=2 foldmethod=marker: */
