/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef POWERPC_SUPPORT
#include "../../config.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "../../arch/powerpc/powerpc_opcodes.h"
#include "../../arch/powerpc/powerpc_disassemble.h"
#include "../../arch/powerpc/powerpc_assemble.h"
#include "../../arch/powerpc/powerpc_instructions.h"
#include "elf_powerpc.h"
#include "elf_powerpcsupport.h"
#include "elf_main.h"
#include "../../utils/util_machine.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_object.h"
#include "elf_i386.h"


/* Common operations on objectfiles */

/* Write an elf object file. The elfhdr is supplied by the arch-specific code */


typedef struct 
{
  t_section * sec;
  Elf32_Shdr hdr;
  t_int32 described_by_pdr;
  t_int32 prepadding; 
} sec_store_info;



void   ElfWritePowerpcSameEndian (t_object * obj, t_string filename) {
   FILE * fp = fopen(filename,"w");

   int tel=0;
   Elf32_Ehdr hdr;
   ANALYSIS_START("Writing Powerpc binary");
   hdr.e_ident[EI_MAG0]=ELFMAG0;
   hdr.e_ident[EI_MAG1]=ELFMAG1;
   hdr.e_ident[EI_MAG2]=ELFMAG2;
   hdr.e_ident[EI_MAG3]=ELFMAG3;
      hdr.e_ident[EI_VERSION]=EV_CURRENT;
   if (global_machine_info.endian==LSB)
      hdr.e_ident[EI_DATA]=ELFDATA2LSB;
   else
      hdr.e_ident[EI_DATA]=ELFDATA2MSB;

   hdr.e_ident[EI_CLASS]=ELFCLASS32;
   for (tel=EI_PAD; tel<16; tel++)
   {
      hdr.e_ident[tel]=0;
   }
   hdr.e_type=ET_EXEC;
   hdr.e_machine=EM_PPC;


  hdr.e_version=EV_CURRENT;
  hdr.e_entry=G_T_UINT32(obj->entry); 
  hdr.e_phoff=0;
  /* The 32-bit Intel Architecture defines no flags; so this member contains zero. */
  hdr.e_flags=0;
  hdr.e_ehsize=sizeof(Elf32_Ehdr);
  /* Set in common */
  hdr.e_phentsize=0;
  hdr.e_phnum=0;
  hdr.e_shentsize=0; 
  hdr.e_shnum=0; 
  hdr.e_shstrndx=0; 
  
  PowerpcElfWriteSameEndian(fp,&hdr,obj,1<<12); 
  ANALYSIS_SIMPLE_END("Writing POWERPC  binary");

}



void   ElfReadPowerpcSwitchedEndian(FILE * fp, t_object * obj) {
  FATAL(("ReadSwitchedEndian : Don't know how"));
}

t_bool IsElfPowerpcSameEndianOnMsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   my_byteorder byte_order;
   Elf32_Ehdr hdr;
   int tel=0;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=MSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2LSB:
	 return FALSE;
      case ELFDATA2MSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_PPC)
      return FALSE; 

   return TRUE;
}

t_bool IsElfPowerpcSwitchedEndianOnMsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   my_byteorder byte_order;
   Elf32_Ehdr hdr;
   int tel=0;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=MSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_SWITCHED_PPC)
      return FALSE; 

   return TRUE;
}

t_bool IsElfPowerpcSwitchedEndianOnLsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   my_byteorder byte_order;
   Elf32_Ehdr hdr;
   int tel=0;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=LSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2LSB:
	 return FALSE;
      case ELFDATA2MSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_SWITCHED_SH)
      return FALSE; 

   return TRUE;
}



t_bool IsElfPowerpcSameEndianOnLsb(FILE * fp)
{
   Elf32_Byte buffer[EI_NIDENT];
   my_byteorder byte_order;
   Elf32_Ehdr hdr;
   int tel=0;
   Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
   Filepos save=ftell(fp);

   if (global_machine_info.endian!=LSB)
      return FALSE;

   if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
      FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

   fseek(fp,save,SEEK_SET);

   if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
      FATAL(("Not an elf objectfile"));

   switch (buffer[EI_DATA])
   {
      default:
      case ELFDATANONE:
      case ELFDATA2MSB:
	 return FALSE;
      case ELFDATA2LSB:
	 break;
   }

   switch (buffer[EI_CLASS])
   {
      case ELFCLASS32: 
	 if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
	 {
	    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
	 }
	 fseek(fp,save,SEEK_SET);
	 machine=hdr.e_machine;
	 break;
      case ELFCLASS64: 
      case ELFCLASSNONE:
      default:
	 return FALSE;
	 break;
   }

   if (machine!=EM_SH)
      return FALSE; 

   return TRUE;
}

void ElfReadPowerpcSameEndian(FILE * fp, t_object * obj)
{
   Elf32_Shdr * shdrs=NULL;
   Elf32_Sym * symbol_table=NULL;
   char * strtab=NULL;
   int numsyms=0;
   int tel;
   char * sechdrstrtab=NULL;
   Elf32_Ehdr hdr32;
   t_symbol ** table;
   Filepos save=ftell(fp);
   if (fread((&hdr32),sizeof(Elf32_Ehdr),1,fp)!=1)
   {
      FATAL(("Could not read elf header!"));
   }

   /* Reading elf headers does not depend on the type of objectfile */
   ElfReadCommon32(fp, &hdr32, obj, &shdrs, &symbol_table, &numsyms, &strtab, &sechdrstrtab,&table,FALSE); 
   obj->reloc_table=RelocTableNew();

	 /* Read relocations */
	 if (hdr32.e_shnum)
	 {
		 tel = 0;
		 while (tel<hdr32.e_shnum)
		 {
			 if (PatternMatch(".rela.debug*",sechdrstrtab + shdrs[tel].sh_name))
			 {
				 /* if we ignore debug info, we can also ignore debug relocations */
			 }
			 else if (PatternMatch(".rela.stab*",sechdrstrtab + shdrs[tel].sh_name))
			 {
				 /* Lets ignore these */
			 }
			 else if (PatternMatch(".rela*",sechdrstrtab + shdrs[tel].sh_name))
			 {
				 Elf32_Rela * rel=Malloc(shdrs[tel].sh_size);
				 int tel2;
				 t_section_hash_node * sh;

				 fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
				 fread(rel,shdrs[tel].sh_size,1,fp);

				 DiabloPrint(stdout, "Entered REL section %s (base %s): ",sechdrstrtab + shdrs[tel].sh_name,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);

				 sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);

				 DiabloPrint(stdout,"%d relocs\n",(shdrs[tel].sh_size/sizeof(Elf32_Rela)));

				 VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
				 for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rela)); tel2++)
				 {
					 VERIFY_FASSERT(symbol_table && strtab,("Symtab or stringtab not found. Implement me"));
					 switch(ELF32_R_TYPE(rel[tel2].r_info))
					 {
					   // komt orgineel van de sh3 code. Dank aan Luc Claes. U rule
					   
				  case R_PPC_LOCAL24PC	:

					    {
			
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_LOCAL24PC,RELOC_FROM_SECTION,NULL);
					    }
			break;
				 case R_PPC_GOT16	: 
				   {
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_GOT16,RELOC_FROM_SECTION,NULL);
				   }
			break;	   
				 case R_PPC_ADDR16_HA		:
				   {
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_ADDR16_HA,RELOC_FROM_SECTION,NULL);  
				   }
				   break ;
			case R_PPC_ADDR16_LO		:
			   {
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_ADDR16_LO,RELOC_FROM_SECTION,NULL);  
				   }
				   break ;

			 case R_PPC_REL24		:	/* PC relative 26 bit */
			   {
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_REL24,RELOC_FROM_SECTION,NULL);  
				   }
				   break ;
			case R_PPC_ADDR32: /* 32bit absolute address */
			  {
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_ADDR32,RELOC_FROM_SECTION,NULL);  
				   }
				   break ;
			case R_PPC_REL32	:
			  {
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_REL32,RELOC_FROM_SECTION,NULL);  
				   }
				   break ;

				   	 case R_PPC_PLTREL24	:
					   {
			t_symbol * sym;
			t_address generic;
			sym=table[ELF32_R_SYM(rel[tel2].r_info)];
			if (!sym) FATAL(("Ajaj: symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));
			GenericSet_uint32(generic,rel[tel2].r_offset);
			RelocTableAddReloc(obj->reloc_table,sh->section, generic,sym,REPRESENT_VALUE,R_PPC_REL32,RELOC_FROM_SECTION,NULL);  
				   }
				   break ;					  
					     
					   

					 case R_PPC_NONE: FATAL(("Relocs: R_PPC_NONE found\n")); break;
					 
					 case R_PPC_ADDR24		:	/* 26bit address, 2 bits ignored.  */
					 case R_PPC_ADDR16		:	/* 16bit absolute address */
					 	/* lower 16bit of absolute address */
					 case R_PPC_ADDR16_HI		:	/* high 16bit of absolute address */
					/* adjusted high 16bit */
					 case R_PPC_ADDR14		:	/* 16bit address, 2 bits ignored */
					 case R_PPC_ADDR14_BRTAKEN	:
					 case R_PPC_ADDR14_BRNTAKEN	:
					
					 case R_PPC_REL14		:	/* PC relative 16 bit */
					 case R_PPC_REL14_BRTAKEN	:
					 case R_PPC_REL14_BRNTAKEN	:
				
					 case R_PPC_GOT16_LO	:
					 case R_PPC_GOT16_HI	:
					 case R_PPC_GOT16_HA	:
					   //case R_PPC_PLTREL24	:
					 case R_PPC_COPY	:
					 case R_PPC_GLOB_DAT	:
					 case R_PPC_JMP_SLOT	:
					 case R_PPC_RELATIVE	:
					
					 case R_PPC_UADDR32	:
					 case R_PPC_UADDR16	:
				
					 case R_PPC_PLT32	:
					 case R_PPC_PLTREL32	:
					 case R_PPC_PLT16_LO	:
					 case R_PPC_PLT16_HI	:
					 case R_PPC_PLT16_HA	:
					 case R_PPC_SDAREL16	:
					 case R_PPC_SECTOFF	:
					 case R_PPC_SECTOFF_LO	:
					 case R_PPC_SECTOFF_HI	:
					 case R_PPC_SECTOFF_HA	:	 
					

			
						 default:
														 FATAL(("Implement RELOC_TYPE %d",ELF32_R_TYPE(rel[tel2].r_info)));
					 }
				 }
				 Free(rel);
			 }
			 else if (PatternMatch(".rel*",sechdrstrtab + shdrs[tel].sh_name))
			 {
				 FATAL(("REL relocations not implemented!"));
			 }
			 tel++;
		 }
	 }

	 if (shdrs) Free(shdrs);
	 if (sechdrstrtab) Free(sechdrstrtab);
	 if (symbol_table) Free(symbol_table);
	 if (strtab) Free(strtab);
}

void I386ElfWriteSameEndian(FILE * fp, Elf32_Ehdr * hdr, t_object * obj, int pagesize)
{
  Elf32_Phdr * phdrs=NULL;
  Elf32_Shdr nullshdr;
  Elf32_Shdr sectabshdr;
  int tel;
  int nphdrs=0;
  int run=-1;
  long offset;
  t_uint32 can_be_placed_at; 
  sec_store_info * secs=NULL;
  
  char * secstrtab;
  long secstrtabsz=1;
  long nsecs=0;
  
  /* Proberen secties perfect naast elkaar te krijgen {{{*/
  for (tel=0; tel<obj->ncodes; tel++)
  {
    printf("Sectie adres van %d-e codesectie 0x%x\n",tel,SECTION_ADDRESS(obj->code[tel]));
    printf("Sectie grootte van %d-e codesectie 0x%x\n",tel,SECTION_SIZE(obj->code[tel]));
  }
  
  for (tel=0; tel<obj->nrodatas; tel++)
  {
    printf("Sectie adres van %d-e read-only sectie 0x%x\n",tel,SECTION_ADDRESS(obj->rodata[tel]));
    printf("Sectie grootte van %d-e read only sectie 0x%x\n",tel,SECTION_SIZE(obj->rodata[tel])); 
  }
  
  for (tel=0; tel<obj->ndatas; tel++)
  {
    printf("Sectie adres van %d-e data sectie 0x%x\n",tel,SECTION_ADDRESS(obj->data[tel]));
    printf("Sectie grootte van %d-e data sectie 0x%x\n",tel,SECTION_SIZE(obj->data[tel]));
  }
  
  for (tel=0; tel<obj->nbsss; tel++)
  {
    printf("Sectie adres van %d-e bss sectie 0x%x\n",tel,SECTION_ADDRESS(obj->bss[tel]));
    printf("Sectie grootte van %d-e bss sectie 0x%x\n",tel,SECTION_SIZE(obj->bss[tel]));
  }
  
  /* End of testing }}} */
  
  /* Calculate the section - string table size "\0.section\0... and the number of sections" {{{ */
 
  for (tel=0; tel<obj->ncodes; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->code[tel];
    secstrtabsz+=strlen(obj->code[tel]->name)+1;
    nsecs++;
  }
  printf("%d codesecties gevonden\n",tel);
  
  for (tel=0; tel<obj->nrodatas; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->rodata[tel];
    secstrtabsz+=strlen(obj->rodata[tel]->name)+1;
    nsecs++;
  }
  printf("%d rodatasecties gevonden\n",tel);
  
  for (tel=0; tel<obj->ndatas; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->data[tel];
    secstrtabsz+=strlen(obj->data[tel]->name)+1;
    nsecs++;
  }
  printf("%d datasecties gevonden\n",tel);

  for (tel=0; tel<obj->nbsss; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->bss[tel];
    secstrtabsz+=strlen(obj->bss[tel]->name)+1;
    nsecs++;
  }
  printf("%d bsssecties gevonden\n",tel);
  

  secstrtabsz+=strlen(".shstrtab")+1;

  /* End of Calculate the section - string table size }}} */
  
  /* Allocate space for string table and initialize it {{{ */
  secstrtab=Malloc(secstrtabsz);
  secstrtab[0]='\0';
  /* }}} */

  /* Calculate section headers {{{ */
  secstrtabsz=1; 
  nsecs=0;

  /* the zero sectionheader */
  nullshdr.sh_name=0;
  nullshdr.sh_type=SHT_NULL;
  nullshdr.sh_flags=0;
  nullshdr.sh_addr=0;
  nullshdr.sh_offset=0;
  nullshdr.sh_size=0;
  nullshdr.sh_link=SHN_UNDEF;
  nullshdr.sh_info=0;
  nullshdr.sh_addralign=0;
  nullshdr.sh_entsize=0;
  
  /* The normal sectionheaders */
  for (tel=0; tel<obj->ncodes; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_EXECINSTR;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->code[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->code[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF; /*ter duidelijkheid MACRO gebruiken en niet 0*/
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->code[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+secstrtabsz,"%s",obj->code[tel]->name);
    
    printf("Section %s at 0x%x size 0x%x\n",obj->code[tel]->name, secs[nsecs].hdr.sh_addr, secs[nsecs].hdr.sh_size);
    secstrtabsz+=strlen(obj->code[tel]->name)+1;

    nsecs++; 
  }

  for (tel=0; tel<obj->nrodatas; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->rodata[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->rodata[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->rodata[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    
    sprintf(secstrtab+secstrtabsz,"%s",obj->rodata[tel]->name);
    secstrtabsz+=strlen(obj->rodata[tel]->name)+1;
    nsecs++; 
  }

  for (tel=0; tel<obj->ndatas; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_WRITE;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->data[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->data[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->data[tel]));
    secs[nsecs].hdr.sh_entsize=0;

    sprintf(secstrtab+secstrtabsz,"%s",obj->data[tel]->name);
    secstrtabsz+=strlen(obj->data[tel]->name)+1;
    nsecs++; 
  }

  for (tel=0; tel<obj->nbsss; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_NOBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_WRITE;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->bss[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->bss[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->bss[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+secstrtabsz,"%s",obj->bss[tel]->name);
    secstrtabsz+=strlen(obj->bss[tel]->name)+1;
    nsecs++; 
  }

  /* data_string table */
  sectabshdr.sh_name=secstrtabsz;
  sectabshdr.sh_type=SHT_STRTAB;
  sectabshdr.sh_addr=0;
  sectabshdr.sh_size=secstrtabsz+strlen(".shstrtab")+1;
  sectabshdr.sh_flags=0;
  sectabshdr.sh_offset=0;
  sectabshdr.sh_link=0;
  sectabshdr.sh_info=0;
  sectabshdr.sh_addralign=0;
  sectabshdr.sh_entsize=0;
  sprintf(secstrtab+secstrtabsz,"%s",".shstrtab");
  secstrtabsz+=strlen(".shstrtab")+1;
  /* }}} */

  
  /* Calculate the number of phdrs needed. The phdrs describes which segments
   * the OS/debugger/whatever needs to make to run the program. So it describes
   * a collection of sections */
  
  for (tel=0; tel<nsecs; tel++)
  {
     if (secs[tel].hdr.sh_flags&SHF_ALLOC)
     {
	/* Alloc new segment when necessary */ 
	if (secs[tel].hdr.sh_addr>run+20)
	{
	   printf("Allocing for %s, diff = %d\n",secstrtab+secs[tel].hdr.sh_name,secs[tel].hdr.sh_addr-run);
	   phdrs=Realloc(phdrs,sizeof(Elf32_Phdr)*(nphdrs+1));
	   phdrs[nphdrs].p_type=PT_LOAD;
	   phdrs[nphdrs].p_offset=0; /* Depends on nphdrs */

	   phdrs[nphdrs].p_paddr=phdrs[nphdrs].p_vaddr=secs[tel].hdr.sh_addr;
	   phdrs[nphdrs].p_memsz=0;
	   phdrs[nphdrs].p_align=pagesize;
	   phdrs[nphdrs].p_flags=PF_R;
	   phdrs[nphdrs].p_filesz=0;
	   nphdrs++;
	   secs[tel].prepadding=-1;
	}
	else if (secs[tel].hdr.sh_addr<run)
	{
	   FATAL(("Sections overlap (or are not sorted as expected) for section %s at @G!",secs[tel].sec->name,SECTION_ADDRESS(secs[tel].sec)));
	}
	else
	{
	   secs[tel].prepadding=secs[tel].hdr.sh_addr-run;
	   printf("Hoisting %s in pdr, diff = %d \n",secstrtab+secs[tel].hdr.sh_name,secs[tel].hdr.sh_addr-run);
	}
	phdrs[nphdrs-1].p_filesz+=(secs[tel].hdr.sh_type&SHT_PROGBITS)?secs[tel].hdr.sh_size:0;
	phdrs[nphdrs-1].p_memsz+=secs[tel].hdr.sh_size;

	secs[tel].described_by_pdr=nphdrs-1;
	if (secs[tel].hdr.sh_flags & SHF_WRITE) phdrs[nphdrs-1].p_flags|=PF_W;
	if (secs[tel].hdr.sh_flags & SHF_EXECINSTR) phdrs[nphdrs-1].p_flags|=PF_X;
	run=secs[tel].hdr.sh_addr+secs[tel].hdr.sh_size;
     }
     else
     {
	secs[tel].described_by_pdr=-1;
	printf("SKIP: %s\n",secstrtab+secs[tel].hdr.sh_name);
     }
  }


  /* Now place all segments in the file: segments file offsets need to be
   * congruent with the virtual address modulo the pagesize */  

  can_be_placed_at=sizeof(Elf32_Ehdr)+nphdrs*sizeof(Elf32_Phdr)+(nsecs+2)*sizeof(Elf32_Shdr);


  for (tel=0; tel<nsecs; tel++)
  {
     if (secs[tel].hdr.sh_flags&SHF_ALLOC)
     {
	if (secs[tel].prepadding==-1)
	{
	   if ((can_be_placed_at&4095)<(secs[tel].hdr.sh_addr&4095))
	   {
	      secs[tel].prepadding=(secs[tel].hdr.sh_addr&4095)-(can_be_placed_at&4095);
	      
	      printf("Segment prepadding =%d\n",secs[tel].prepadding);
	   }
	   else  if ((can_be_placed_at&4095)>(secs[tel].hdr.sh_addr&4095))
	   {
	      secs[tel].prepadding=4096+(secs[tel].hdr.sh_addr&4095)-(can_be_placed_at&4095);
	      printf("Segment prepadding =%d\n",secs[tel].prepadding);
	   }
	   else
	   {
	      secs[tel].prepadding=0;
	   }
	   can_be_placed_at+=secs[tel].prepadding;
	   phdrs[secs[tel].described_by_pdr].p_offset=can_be_placed_at;
	   secs[tel].hdr.sh_offset=can_be_placed_at;
	   if (secs[tel].hdr.sh_type&SHT_PROGBITS) can_be_placed_at+=secs[tel].hdr.sh_size;
	}
	else
	{
	   can_be_placed_at+=secs[tel].prepadding;
	   secs[tel].hdr.sh_offset=can_be_placed_at;
	   if (secs[tel].hdr.sh_type&SHT_PROGBITS) can_be_placed_at+=secs[tel].hdr.sh_size;
	}
     }
  }



  sectabshdr.sh_offset=can_be_placed_at;



  /* Set ELFheaders */
  /* Ident is already set */
  /* Type is already set */
  /* Machine is already set */
  /* Version is already set */
  /* Entry is already set */
  hdr->e_phoff=sizeof(Elf32_Ehdr);
  hdr->e_shoff=sizeof(Elf32_Ehdr)+nphdrs*sizeof(Elf32_Phdr);
  /* Flags : TODO */
  /* Ehsize is already set */ 
  hdr->e_phentsize=sizeof(Elf32_Phdr);
  hdr->e_phnum=nphdrs;
  hdr->e_shentsize=sizeof(Elf32_Shdr);
  /* #entries in section header table =
   * #normal sections + 1 zero-section + 1 section name string table */
  hdr->e_shnum=nsecs+2;
  hdr->e_shstrndx=nsecs+1;

  /* Write ELF header + Program headers {{{ */
  fwrite(hdr,1,sizeof(Elf32_Ehdr),fp);
  fwrite(phdrs,sizeof(Elf32_Phdr),nphdrs,fp);
  /* }}} */
  
  /* Write section headers {{{ */

  /* Verify correctness: section table should be at hdr->e_shoff */
  VERIFY_FASSERT(ftell(fp)==hdr->e_shoff,("Section header table is not at %x as said in the program header, but at %x",hdr->e_shoff,ftell(fp)));
  /* 1 - zero header */
  fwrite(&nullshdr,sizeof(Elf32_Shdr),1,fp);
  /* 2 - normal section headers */
  for (tel=0; tel<nsecs; tel++)
    fwrite(&secs[tel].hdr,sizeof(Elf32_Shdr),1,fp);
  /* 3 - section name string table header */ 
  fwrite(&sectabshdr,sizeof(Elf32_Shdr),1,fp);
  /* }}} */
  
  /* Write section data {{{ */
  for (tel=0; tel<nsecs; tel++)
  {
    if (secs[tel].hdr.sh_flags&SHF_ALLOC)
    {
      static char padd[1<<12]; /*todo..aanpassen aan pagesize=4kB [OK]*/
      fwrite(padd, sizeof(char), secs[tel].prepadding, fp);
      VERIFY_FASSERT(secs[tel].hdr.sh_offset==ftell(fp),("Section %d not at the right place (should be %x, is %x)",tel,secs[tel].hdr.sh_offset,ftell(fp)));
       if (secs[tel].hdr.sh_type&SHT_PROGBITS)fwrite(secs[tel].sec->data,sizeof(char),G_T_UINT32(SECTION_SIZE(secs[tel].sec)),fp);
    }
  }
  /* }}} */

  /* Write section string table {{{ */

  VERIFY_FASSERT(sectabshdr.sh_offset==ftell(fp),("String table not at the right place: should be %x but is %x",sectabshdr.sh_offset,ftell(fp)));
  printf("Writing secstrtab at Ox%x\n",(t_int32) ftell(fp));
  fwrite(secstrtab,sizeof(char),secstrtabsz,fp);
 
  /* }}} */

  if (nphdrs) Free(phdrs);
  Free(secs);
  Free(secstrtab);
}




void PowerpcElfWriteSameEndian(FILE * fp, Elf32_Ehdr * hdr, t_object * obj, int pagesize)
{
  Elf32_Phdr * phdrs=NULL;
  Elf32_Shdr nullshdr;
  Elf32_Shdr sectabshdr;
  int tel;
  int nphdrs=0;
  int run=-1;
  long offset;
  t_uint32 can_be_placed_at; 
  sec_store_info * secs=NULL;
  
  char * secstrtab;
  long secstrtabsz=1;
  long nsecs=0;
  
  /* Proberen secties perfect naast elkaar te krijgen {{{*/
  for (tel=0; tel<obj->ncodes; tel++)
  {
    printf("Sectie adres van %d-e codesectie 0x%x\n",tel,SECTION_ADDRESS(obj->code[tel]));
    printf("Sectie grootte van %d-e codesectie 0x%x\n",tel,SECTION_SIZE(obj->code[tel]));
  }
  
  for (tel=0; tel<obj->nrodatas; tel++)
  {
    printf("Sectie adres van %d-e read-only sectie 0x%x\n",tel,SECTION_ADDRESS(obj->rodata[tel]));
    printf("Sectie grootte van %d-e read only sectie 0x%x\n",tel,SECTION_SIZE(obj->rodata[tel])); 
  }
  
  for (tel=0; tel<obj->ndatas; tel++)
  {
    printf("Sectie adres van %d-e data sectie 0x%x\n",tel,SECTION_ADDRESS(obj->data[tel]));
    printf("Sectie grootte van %d-e data sectie 0x%x\n",tel,SECTION_SIZE(obj->data[tel]));
  }
  
  for (tel=0; tel<obj->nbsss; tel++)
  {
    printf("Sectie adres van %d-e bss sectie 0x%x\n",tel,SECTION_ADDRESS(obj->bss[tel]));
    printf("Sectie grootte van %d-e bss sectie 0x%x\n",tel,SECTION_SIZE(obj->bss[tel]));
  }
  
  /* End of testing }}} */
  
  /* Calculate the section - string table size "\0.section\0... and the number of sections" {{{ */
 
  for (tel=0; tel<obj->ncodes; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->code[tel];
    secstrtabsz+=strlen(obj->code[tel]->name)+1;
    nsecs++;
  }
  printf("%d codesecties gevonden\n",tel);
  
  for (tel=0; tel<obj->nrodatas; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->rodata[tel];
    secstrtabsz+=strlen(obj->rodata[tel]->name)+1;
    nsecs++;
  }
  printf("%d rodatasecties gevonden\n",tel);
  
  for (tel=0; tel<obj->ndatas; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->data[tel];
    secstrtabsz+=strlen(obj->data[tel]->name)+1;
    nsecs++;
  }
  printf("%d datasecties gevonden\n",tel);

  for (tel=0; tel<obj->nbsss; tel++)
  {
    secs=Realloc(secs,sizeof(sec_store_info)*(nsecs+1));
    secs[nsecs].sec=obj->bss[tel];
    secstrtabsz+=strlen(obj->bss[tel]->name)+1;
    nsecs++;
  }
  printf("%d bsssecties gevonden\n",tel);
  

  secstrtabsz+=strlen(".shstrtab")+1;

  /* End of Calculate the section - string table size }}} */
  
  /* Allocate space for string table and initialize it {{{ */
  secstrtab=Malloc(secstrtabsz);
  secstrtab[0]='\0';
  /* }}} */

  /* Calculate section headers {{{ */
  secstrtabsz=1; 
  nsecs=0;

  /* the zero sectionheader */
  nullshdr.sh_name=0;
  nullshdr.sh_type=SHT_NULL;
  nullshdr.sh_flags=0;
  nullshdr.sh_addr=0;
  nullshdr.sh_offset=0;
  nullshdr.sh_size=0;
  nullshdr.sh_link=SHN_UNDEF;
  nullshdr.sh_info=0;
  nullshdr.sh_addralign=0;
  nullshdr.sh_entsize=0;
  
  /* The normal sectionheaders */
  for (tel=0; tel<obj->ncodes; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_EXECINSTR;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->code[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->code[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF; /*ter duidelijkheid MACRO gebruiken en niet 0*/
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->code[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+secstrtabsz,"%s",obj->code[tel]->name);
    
    printf("Section %s at 0x%x size 0x%x\n",obj->code[tel]->name, secs[nsecs].hdr.sh_addr, secs[nsecs].hdr.sh_size);
    secstrtabsz+=strlen(obj->code[tel]->name)+1;

    nsecs++; 
  }

  for (tel=0; tel<obj->nrodatas; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->rodata[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->rodata[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->rodata[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    
    sprintf(secstrtab+secstrtabsz,"%s",obj->rodata[tel]->name);
    secstrtabsz+=strlen(obj->rodata[tel]->name)+1;
    nsecs++; 
  }

  for (tel=0; tel<obj->ndatas; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_PROGBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_WRITE;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->data[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->data[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->data[tel]));
    secs[nsecs].hdr.sh_entsize=0;

    sprintf(secstrtab+secstrtabsz,"%s",obj->data[tel]->name);
    secstrtabsz+=strlen(obj->data[tel]->name)+1;
    nsecs++; 
  }

  for (tel=0; tel<obj->nbsss; tel++)
  {
    secs[nsecs].hdr.sh_name=secstrtabsz;
    secs[nsecs].hdr.sh_type=SHT_NOBITS;
    secs[nsecs].hdr.sh_flags=SHF_ALLOC|SHF_WRITE;
    secs[nsecs].hdr.sh_addr=G_T_UINT32(SECTION_ADDRESS(obj->bss[tel]));
    secs[nsecs].hdr.sh_size=G_T_UINT32(SECTION_SIZE(obj->bss[tel]));
    secs[nsecs].hdr.sh_offset=0;
    secs[nsecs].hdr.sh_link=SHN_UNDEF;
    secs[nsecs].hdr.sh_info=0;
    secs[nsecs].hdr.sh_addralign=G_T_UINT32(SECTION_ALIGN(obj->bss[tel]));
    secs[nsecs].hdr.sh_entsize=0;
    sprintf(secstrtab+secstrtabsz,"%s",obj->bss[tel]->name);
    secstrtabsz+=strlen(obj->bss[tel]->name)+1;
    nsecs++; 
  }

  /* data_string table */
  sectabshdr.sh_name=secstrtabsz;
  sectabshdr.sh_type=SHT_STRTAB;
  sectabshdr.sh_addr=0;
  sectabshdr.sh_size=secstrtabsz+strlen(".shstrtab")+1;
  sectabshdr.sh_flags=0;
  sectabshdr.sh_offset=0;
  sectabshdr.sh_link=0;
  sectabshdr.sh_info=0;
  sectabshdr.sh_addralign=0;
  sectabshdr.sh_entsize=0;
  sprintf(secstrtab+secstrtabsz,"%s",".shstrtab");
  secstrtabsz+=strlen(".shstrtab")+1;
  /* }}} */

  
  /* Calculate the number of phdrs needed. The phdrs describes which segments
   * the OS/debugger/whatever needs to make to run the program. So it describes
   * a collection of sections */
  
  for (tel=0; tel<nsecs; tel++)
  {
     if (secs[tel].hdr.sh_flags&SHF_ALLOC)
     {
	/* Alloc new segment when necessary */ 
	if (secs[tel].hdr.sh_addr>run+20)
	{
	   printf("Allocing for %s, diff = %d\n",secstrtab+secs[tel].hdr.sh_name,secs[tel].hdr.sh_addr-run);
	   phdrs=Realloc(phdrs,sizeof(Elf32_Phdr)*(nphdrs+1));
	   phdrs[nphdrs].p_type=PT_LOAD;
	   phdrs[nphdrs].p_offset=0; /* Depends on nphdrs */

	   phdrs[nphdrs].p_paddr=phdrs[nphdrs].p_vaddr=secs[tel].hdr.sh_addr;
	   phdrs[nphdrs].p_memsz=0;
	   phdrs[nphdrs].p_align=pagesize;
	   phdrs[nphdrs].p_flags=PF_R;
	   phdrs[nphdrs].p_filesz=0;
	   nphdrs++;
	   secs[tel].prepadding=-1;
	}
	else if (secs[tel].hdr.sh_addr<run)
	{
	  // FATAL(("Sections overlap (or are not sorted as expected) for section %s at @G!",secs[tel].sec->name,SECTION_ADDRESS(secs[tel].sec)));
	}
	else
	{
	   secs[tel].prepadding=secs[tel].hdr.sh_addr-run;
	   printf("Hoisting %s in pdr, diff = %d \n",secstrtab+secs[tel].hdr.sh_name,secs[tel].hdr.sh_addr-run);
	}
	phdrs[nphdrs-1].p_filesz+=(secs[tel].hdr.sh_type&SHT_PROGBITS)?secs[tel].hdr.sh_size:0;
	phdrs[nphdrs-1].p_memsz+=secs[tel].hdr.sh_size;

	secs[tel].described_by_pdr=nphdrs-1;
	if (secs[tel].hdr.sh_flags & SHF_WRITE) phdrs[nphdrs-1].p_flags|=PF_W;
	if (secs[tel].hdr.sh_flags & SHF_EXECINSTR) phdrs[nphdrs-1].p_flags|=PF_X;
	run=secs[tel].hdr.sh_addr+secs[tel].hdr.sh_size;
     }
     else
     {
	secs[tel].described_by_pdr=-1;
	printf("SKIP: %s\n",secstrtab+secs[tel].hdr.sh_name);
     }
  }


  /* Now place all segments in the file: segments file offsets need to be
   * congruent with the virtual address modulo the pagesize */  

  can_be_placed_at=sizeof(Elf32_Ehdr)+nphdrs*sizeof(Elf32_Phdr)+(nsecs+2)*sizeof(Elf32_Shdr);


  for (tel=0; tel<nsecs; tel++)
  {
     if (secs[tel].hdr.sh_flags&SHF_ALLOC)
     {
	if (secs[tel].prepadding==-1)
	{
	   if ((can_be_placed_at&4095)<(secs[tel].hdr.sh_addr&4095))
	   {
	      secs[tel].prepadding=(secs[tel].hdr.sh_addr&4095)-(can_be_placed_at&4095);
	      
	      printf("Segment prepadding =%d\n",secs[tel].prepadding);
	   }
	   else  if ((can_be_placed_at&4095)>(secs[tel].hdr.sh_addr&4095))
	   {
	      secs[tel].prepadding=4096+(secs[tel].hdr.sh_addr&4095)-(can_be_placed_at&4095);
	      printf("Segment prepadding =%d\n",secs[tel].prepadding);
	   }
	   else
	   {
	      secs[tel].prepadding=0;
	   }
	   can_be_placed_at+=secs[tel].prepadding;
	   phdrs[secs[tel].described_by_pdr].p_offset=can_be_placed_at;
	   secs[tel].hdr.sh_offset=can_be_placed_at;
	   if (secs[tel].hdr.sh_type&SHT_PROGBITS) can_be_placed_at+=secs[tel].hdr.sh_size;
	}
	else
	{
	   can_be_placed_at+=secs[tel].prepadding;
	   secs[tel].hdr.sh_offset=can_be_placed_at;
	   if (secs[tel].hdr.sh_type&SHT_PROGBITS) can_be_placed_at+=secs[tel].hdr.sh_size;
	}
     }
  }



  sectabshdr.sh_offset=can_be_placed_at;



  /* Set ELFheaders */
  /* Ident is already set */
  /* Type is already set */
  /* Machine is already set */
  /* Version is already set */
  /* Entry is already set */
  hdr->e_phoff=sizeof(Elf32_Ehdr);
  hdr->e_shoff=sizeof(Elf32_Ehdr)+nphdrs*sizeof(Elf32_Phdr);
  /* Flags : TODO */
  /* Ehsize is already set */ 
  hdr->e_phentsize=sizeof(Elf32_Phdr);
  hdr->e_phnum=nphdrs;
  hdr->e_shentsize=sizeof(Elf32_Shdr);
  /* #entries in section header table =
   * #normal sections + 1 zero-section + 1 section name string table */
  hdr->e_shnum=nsecs+2;
  hdr->e_shstrndx=nsecs+1;

  /* Write ELF header + Program headers {{{ */
  fwrite(hdr,1,sizeof(Elf32_Ehdr),fp);
  fwrite(phdrs,sizeof(Elf32_Phdr),nphdrs,fp);
  /* }}} */
  
  /* Write section headers {{{ */

  /* Verify correctness: section table should be at hdr->e_shoff */
  VERIFY_FASSERT(ftell(fp)==hdr->e_shoff,("Section header table is not at %x as said in the program header, but at %x",hdr->e_shoff,ftell(fp)));
  /* 1 - zero header */
  fwrite(&nullshdr,sizeof(Elf32_Shdr),1,fp);
  /* 2 - normal section headers */
  for (tel=0; tel<nsecs; tel++)
    fwrite(&secs[tel].hdr,sizeof(Elf32_Shdr),1,fp);
  /* 3 - section name string table header */ 
  fwrite(&sectabshdr,sizeof(Elf32_Shdr),1,fp);
  /* }}} */
  
  /* Write section data {{{ */
  for (tel=0; tel<nsecs; tel++)
  {
    if (secs[tel].hdr.sh_flags&SHF_ALLOC)
    {
      static char padd[1<<12]; /*todo..aanpassen aan pagesize=4kB [OK]*/
      fwrite(padd, sizeof(char), secs[tel].prepadding, fp);
      VERIFY_FASSERT(secs[tel].hdr.sh_offset==ftell(fp),("Section %d not at the right place (should be %x, is %x)",tel,secs[tel].hdr.sh_offset,ftell(fp)));
       if (secs[tel].hdr.sh_type&SHT_PROGBITS)fwrite(secs[tel].sec->data,sizeof(char),G_T_UINT32(SECTION_SIZE(secs[tel].sec)),fp);
    }
  }
  /* }}} */

  /* Write section string table {{{ */

  VERIFY_FASSERT(sectabshdr.sh_offset==ftell(fp),("String table not at the right place: should be %x but is %x",sectabshdr.sh_offset,ftell(fp)));
  printf("Writing secstrtab at Ox%x\n",(t_int32) ftell(fp));
  fwrite(secstrtab,sizeof(char),secstrtabsz,fp);
 
  /* }}} */

  if (nphdrs) Free(phdrs);
  Free(secs);
  Free(secstrtab);
}

#endif

