/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifdef SH_SUPPORT
#include <string.h>
#include "../../config.h"
#include "../../simple_types.h"
#include "../../kernel/diablo_ins.h"
#include "../../arch/sh3/sh3_opcodes.h"
#include "../../arch/sh3/sh3_disassemble.h"
#include "../../arch/sh3/sh3_assemble.h"
#include "../../arch/sh3/sh3_instructions.h"
#include "elf_sh3.h"
#include "elf_sh3support.h"
#include "elf_main.h"
#include "elf_common32.h"
#include "../../utils/util_machine.h"
#include "../../utils/util_status.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "../../kernel/diablo_object.h"
#include "../../kernel/diablo_print.h"
typedef struct 
{
  t_section * sec;
  Elf32_Shdr hdr;
} sec_store_info;


/* Common operations on objectfiles */

t_bool IsElfSh3SameEndianOnMsb(FILE * fp)
{
  Elf32_Byte buffer[EI_NIDENT];
  Elf32_Ehdr hdr;
  Elf32_Half machine;
  Filepos save=ftell(fp);

  if (global_machine_info.endian!=MSB)
    return FALSE;

  if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

  fseek(fp,save,SEEK_SET);

  if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
    FATAL(("Not an elf objectfile"));

  switch (buffer[EI_DATA])
  {
    default:
    case ELFDATANONE:
    case ELFDATA2LSB:
      return FALSE;
    case ELFDATA2MSB:
      break;
  }

  switch (buffer[EI_CLASS])
  {
    case ELFCLASS32: 
      if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
      {
        FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
      }
      fseek(fp,save,SEEK_SET);
      machine=hdr.e_machine;
      break;
    case ELFCLASS64: 
    case ELFCLASSNONE:
    default:
      return FALSE;
      break;
  }

  if (machine!=EM_SH)
    return FALSE; 

  return TRUE;
}

t_bool IsElfSh3SwitchedEndianOnMsb(FILE * fp)
{
  Elf32_Byte buffer[EI_NIDENT];
  Elf32_Ehdr hdr;
  Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
  Filepos save=ftell(fp);

  if (global_machine_info.endian!=MSB)
    return FALSE;

  if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

  fseek(fp,save,SEEK_SET);

  if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
    FATAL(("Not an elf objectfile"));

  switch (buffer[EI_DATA])
  {
    default:
    case ELFDATANONE:
    case ELFDATA2MSB:
      return FALSE;
    case ELFDATA2LSB:
      break;
  }

  switch (buffer[EI_CLASS])
  {
    case ELFCLASS32: 
      if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
      {
        FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
      }
      fseek(fp,save,SEEK_SET);
      machine=hdr.e_machine;
      break;
    case ELFCLASS64: 
    case ELFCLASSNONE:
    default:
      return FALSE;
      break;
  }

  if (machine!=EM_SWITCHED_SH)
    return FALSE; 

  return TRUE;
}

t_bool IsElfSh3SwitchedEndianOnLsb(FILE * fp)
{
  Elf32_Byte buffer[EI_NIDENT];
  Elf32_Ehdr hdr;
  Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
  Filepos save=ftell(fp);

  if (global_machine_info.endian!=LSB)
    return FALSE;

  if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

  fseek(fp,save,SEEK_SET);

  if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
    FATAL(("Not an elf objectfile"));

  switch (buffer[EI_DATA])
  {
    default:
    case ELFDATANONE:
    case ELFDATA2LSB:
      return FALSE;
    case ELFDATA2MSB:
      break;
  }

  switch (buffer[EI_CLASS])
  {
    case ELFCLASS32: 
      if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
      {
        FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
      }
      fseek(fp,save,SEEK_SET);
      machine=hdr.e_machine;
      break;
    case ELFCLASS64: 
    case ELFCLASSNONE:
    default:
      return FALSE;
      break;
  }

  if (machine!=EM_SWITCHED_SH)
    return FALSE; 

  return TRUE;
}



t_bool IsElfSh3SameEndianOnLsb(FILE * fp)
{
  Elf32_Byte buffer[EI_NIDENT];
  Elf32_Ehdr hdr;
  Elf32_Half machine; /* For 64 bits this is Elf64Machine, but currently they are the same size ! */
  Filepos save=ftell(fp);

  if (global_machine_info.endian!=LSB)
    return FALSE;

  if (fread(buffer,sizeof(Elf32_Byte),EI_NIDENT,fp)!=EI_NIDENT) 
    FATAL(("Could not read elf header!!!!!!\nFile corrupt"));

  fseek(fp,save,SEEK_SET);

  if ((buffer[EI_MAG0]!=ELFMAG0) || (buffer[EI_MAG1]!=ELFMAG1) || (buffer[EI_MAG2]!=ELFMAG2) || (buffer[EI_MAG3]!=ELFMAG3)) 
    FATAL(("Not an elf objectfile"));

  switch (buffer[EI_DATA])
  {
    default:
    case ELFDATANONE:
    case ELFDATA2MSB:
      return FALSE;
    case ELFDATA2LSB:
      break;
  }

  switch (buffer[EI_CLASS])
  {
    case ELFCLASS32: 
      if (fread(&(hdr),sizeof(Elf32_Ehdr),1,fp)!=1)
      {
        FATAL(("Could not read elf header!!!!!!\nFile corrupt"));
      }
      fseek(fp,save,SEEK_SET);
      machine=hdr.e_machine;
      break;
    case ELFCLASS64: 
    case ELFCLASSNONE:
    default:
      return FALSE;
      break;
  }

  if (machine!=EM_SH)
    return FALSE; 

  return TRUE;
}

void ElfReadSh3SameEndian(FILE * fp, t_object * obj)
{
  Elf32_Shdr * shdrs=NULL;
  Elf32_Sym * symbol_table=NULL;
  char * strtab=NULL;
  int numsyms=0;
  int tel;
  char * sechdrstrtab=NULL;
  Elf32_Ehdr hdr32;
  t_symbol ** table;

  if (fread((&hdr32),sizeof(Elf32_Ehdr),1,fp)!=1)
  {
    FATAL(("Could not read elf header!"));
  }

  /* Reading elf headers does not depend on the type of objectfile */
  ElfReadCommon32(fp, &hdr32, obj, &shdrs, &symbol_table, &numsyms, &strtab, &sechdrstrtab,&table,FALSE); 
  obj->reloc_table=RelocTableNew();

  /* mark the sections with fixed addresses (currently only .stack) */
  for (tel=0; tel<obj->ndatas; tel++)
  {
    if (strcmp(obj->data[tel]->name,".stack") == 0)
    {
      obj->data[tel]->is_fixed = TRUE;
      printf("fixing .stack...\n");
    }
  }

  /* Read relocations */
  if (hdr32.e_shnum)
  {
    tel = 0;
    while (tel<hdr32.e_shnum)
    {
      if (PatternMatch(".rela.debug*",sechdrstrtab + shdrs[tel].sh_name))
      {
        /* if we ignore debug info, we can also ignore debug relocations */
      }
      else if (PatternMatch(".rela.stab*",sechdrstrtab + shdrs[tel].sh_name))
      {
        /* Lets ignore these */
      }
      else if (PatternMatch(".rela*",sechdrstrtab + shdrs[tel].sh_name))
      {
        Elf32_Rela * rel=Malloc(shdrs[tel].sh_size);
        int tel2;
        t_section_hash_node * sh;

        fseek(fp,obj->streampos+shdrs[tel].sh_offset,SEEK_SET);
        fread(rel,shdrs[tel].sh_size,1,fp);

        //DiabloPrint(stdout, "Entered REL section %s (base %s): ",sechdrstrtab + shdrs[tel].sh_name,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);

        sh=(t_section_hash_node *) HashLookup(obj->section_table,sechdrstrtab + shdrs[shdrs[tel].sh_info].sh_name);

        //DiabloPrint(stdout,"%d relocs\n",(shdrs[tel].sh_size/sizeof(Elf32_Rela)));

        VERIFY_FASSERT( sh!=NULL ,("Base section for relocations not found!"));
        for (tel2=0 ; tel2<(shdrs[tel].sh_size/sizeof(Elf32_Rela)); tel2++)
        {
          VERIFY_FASSERT(symbol_table && strtab,("Symtab or stringtab not found. Implement me"));

          switch(ELF32_R_TYPE(rel[tel2].r_info))
          {
            case R_SH_NONE: FATAL(("Relocs: R_SH_NONE found\n")); break;
            case R_SH_DIR32: 
                            {
                              t_address generic;
                              t_address addend;
                              t_uint32 implicit_addend;
                              t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

                              //DiabloPrint(stdout,"Relocs: R_SH_DIR32 found\n");
                              //DiabloPrint(stdout,"Symbol %s\n",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name);
                              //DiabloPrint(stdout,"Offset: %x, Addend: %d\n",rel[tel2].r_offset,rel[tel2].r_addend);

                              if (!sym) 
                                FATAL(("Symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));

                              GenericSet_uint32(generic,rel[tel2].r_offset);
                              GenericSet_uint32(addend,rel[tel2].r_addend);
                              /* sh3 oddity: account for both explicit and implicit addends */
                              implicit_addend = SectionGetData32(sh->section,generic);
                              addend = GENERIC_ADD_UINT32(addend,implicit_addend);
                              
			      RelocTableAddRelocToSymbol(obj->reloc_table,addend,sh->section,generic,sym,TRUE,NULL,NULL,NULL,CALC_ABS "\\" WRITE_32,TRUE);
                            }
                            break;
            case R_SH_REL32: 
                            {
                              t_address generic;
                              t_address addend;
                              t_uint32 implicit_addend;
                              t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

                              //DiabloPrint(stdout,"Relocs: R_SH_REL32 found\n");
                              //DiabloPrint(stdout,"Symbol %s\n",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name);
                              //DiabloPrint(stdout,"Offset: %x, Addend: %d\n",rel[tel2].r_offset,rel[tel2].r_addend);

                              if (!sym) 
                                FATAL(("Symbol %s not found in symbol table",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name));

                              GenericSet_uint32(generic,rel[tel2].r_offset);
                              GenericSet_uint32(addend,rel[tel2].r_addend);
                              /* sh3 oddity: account for both explicit and implicit addends */
                              implicit_addend = SectionGetData32(sh->section,generic);
                              addend = GENERIC_ADD_UINT32(addend,implicit_addend);

                              if (sh->section->type!=CODE_SECTION)
                              {
                                FATAL(("REL 32 not in code section!"));
                              }

			      RelocTableAddRelocToSymbol(obj->reloc_table,addend,sh->section,generic,sym,TRUE,NULL,NULL,NULL,CALC_REL "\\" WRITE_32,TRUE);
                            }
                            break;

            case R_SH_GOTPC: 
                            {

                              t_address generic;
                              t_address addend;
                              t_uint32 implicit_addend;
                              t_symbol * sym=table[ELF32_R_SYM(rel[tel2].r_info)];

                              //DiabloPrint(stdout,"Relocs: R_SH_GOTPC found\n");
                              //DiabloPrint(stdout,"Symbol %s\n",strtab+symbol_table[ELF32_R_SYM(rel[tel2].r_info)].st_name);
                              //DiabloPrint(stdout,"Offset: %x, Addend: %d\n",rel[tel2].r_offset,rel[tel2].r_addend);

                              GenericSet_uint32(generic,rel[tel2].r_offset);
                              GenericSet_uint32(addend,rel[tel2].r_addend);
                              /* sh3 oddity: account for both explicit and implicit addends */
                              implicit_addend = SectionGetData32(sh->section,generic);
                              addend = GENERIC_ADD_UINT32(addend,implicit_addend);

			      /* nothing special needs to be done for this kind of relocation,
			       * as the symbol it references (_GLOBAL_OFFSET_TABLE) will point 
			       * to the .got section, so 'S' in this case is the same as 'GOT' */
			      RelocTableAddRelocToSymbol(obj->reloc_table,addend,sh->section,generic,sym,FALSE,NULL,NULL,NULL,CALC_REL "\\" WRITE_32,TRUE);
                              break;
                            }

                            /* All other reltypes haven't been encountered yet.
                             * They will be implemented on an as needed basis. */
            case R_SH_DIR8WPN:
            case R_SH_IND12W:
            case R_SH_DIR8WPL:
            case R_SH_DIR8WPZ:
            case R_SH_DIR8BP:
            case R_SH_DIR8W:
            case R_SH_DIR8L:
            case R_SH_SWITCH16:
            case R_SH_SWITCH32:
            case R_SH_USES:
            case R_SH_COUNT:
            case R_SH_ALIGN:
            case R_SH_CODE:
            case R_SH_DATA:
            case R_SH_LABEL:
            case R_SH_SWITCH8:
            case R_SH_GNU_VTINHERIT:
            case R_SH_GNU_VTENTRY:
            case R_SH_PLT32:
            case R_SH_COPY:
            case R_SH_GLOB_DAT:
            case R_SH_JMP_SLOT:
            case R_SH_RELATIVE:
            case R_SH_GOTOFF:
            case R_SH_GOT32:
            default:
                            FATAL(("Implement RELOC_TYPE %d",ELF32_R_TYPE(rel[tel2].r_info)));
          }
        }
        Free(rel);
      }
      else if (PatternMatch(".rel*",sechdrstrtab + shdrs[tel].sh_name))
      {
        FATAL(("REL relocations not implemented!"));
      }
      tel++;
    }
  }

  if (shdrs) Free(shdrs);
  if (sechdrstrtab) Free(sechdrstrtab);
  if (symbol_table) Free(symbol_table);
  if (strtab) Free(strtab);
}

void ElfReadSh3SwitchedEndian(FILE * fp, t_object * obj)
{
  FATAL(("SH3 Switched endian not supported"));
}

#define SYMTABNAME ".symtab"
#define SYMSTRTABNAME ".strtab"
/* #define SYMTAB */
void ElfWriteSh3SameEndian(t_object * obj, t_string filename)
{
#ifdef SYMTAB
  t_symbol_he * top = (t_symbol_he *) HashLookup(obj->symbol_table->hash_table, "$switch");
  Elf32_Shdr symtabshdr;
  Elf32_Shdr symstrtabshdr;
#endif
  FILE * fp = fopen(filename,"w");
  int tel = 0;
  Elf32_Ehdr hdr;
  Elf32_Phdr * phdrs = NULL;
  Elf32_Shdr nullshdr;
  Elf32_Shdr sectabshdr;
  int nphdrs = 0;
  int run = -1;
  long offset;
  sec_store_info * secs = NULL;
  char * secstrtab;
  long secstrtabsz = 1;
  long nsecs = 0;
  int pagesize = 4;
  ANALYSIS_START("Writing sh3 binary");
  if(fp == 0) FATAL(("could not open file: %s\n", filename));
  /* Invullen Ehdr (ELF-header) */
  hdr.e_ident[EI_MAG0] = ELFMAG0;
  hdr.e_ident[EI_MAG1] = ELFMAG1;
  hdr.e_ident[EI_MAG2] = ELFMAG2;
  hdr.e_ident[EI_MAG3] = ELFMAG3;
  hdr.e_ident[EI_VERSION] = EV_CURRENT;
  if (global_machine_info.endian == LSB)
    hdr.e_ident[EI_DATA] = ELFDATA2LSB;
  else
    hdr.e_ident[EI_DATA] = ELFDATA2MSB;

  hdr.e_ident[EI_CLASS] = ELFCLASS32;
  /*   hdr.e_ident[EI_OSABI] = ELFOSABI_NONE;
   *   hdr.e_ident[EI_ABIVERSION] = 0;
   */
  for (tel = EI_PAD; tel<16; tel++)
  {
    hdr.e_ident[tel] = 0;
  }
  hdr.e_type = ET_EXEC;
  hdr.e_machine = EM_SH;
  hdr.e_version = EV_CURRENT;
  hdr.e_flags = EF_SH3; /* sh3 zie binutils.../include/elf/sh.h */
  hdr.e_entry = G_T_UINT32(obj->entry);
  printf("Entry at %x\n", hdr.e_entry);
  hdr.e_phoff = 0;
  hdr.e_ehsize = 52;
  hdr.e_phentsize = 0;
  hdr.e_phnum = 0;
  hdr.e_shentsize = 0;
  hdr.e_shnum = 0;
  hdr.e_shstrndx = 0;

  ElfWriteCommonSameEndian(fp,&hdr,obj,2<<1); 
  ANALYSIS_SIMPLE_END("Writing sh3 binary");
}

#endif
