/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)
 ***********************************************************/


/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup Archives Working with archives */
/*@{ */
#ifndef DIABLO_ARCHIVES_TYPEDEFS
#define DIABLO_ARCHIVES_TYPEDEFS
typedef struct _t_archive t_archive;
#endif


#ifndef DIABLO_ARCHIVES_H
#define DIABLO_ARCHIVES_H
#include "../datastructures/data_hash.h"
#include "diablo_object.h"
#include "diablo_archive_tables.h"
/*! Archives */
struct _t_archive
{
   t_archivetype type;
   char * name;
   FILE * open_fp;
   data_hash_table * objects;
};

/*! TODO: document, move to fileformats */
typedef struct 
{
   char ar_name[16];
   char ar_date[12];
   char ar_uid[6]; 
   char ar_gid[6];
   char ar_mode[8];
   char ar_size[10];
   char ar_fmag[2];
} t_ar_archive_header;

/*! TODO: document */
typedef struct
{
   data_hash_node node;
   t_archive * arch;
} t_archive_hash_node;

/*! TODO: document */
typedef struct
{
   data_hash_node node;
   int flags;
   int fpos;
   t_uint32 size; /* 10 ^ 10 < 2 ^ 32 */
} t_arch_obj_hash_node;


/*! \defgroup GLOBAL_ARCH_ALLOC Global Constructors and destructors */
/*@{*/
void ArchivesInit(void);
void ArchivesFini(void);
/*@}*/



/*! \defgroup ARCH_ALLOC Constructors and destructors */
/*@{*/
void ArchiveFree(t_archive *);
t_archive * ArchiveOpenCached(char * name);
/*@}*/
t_object * ArchiveGetObject(t_archive * ar, char * name);
void ArchObjHashNodeFree(void * tf,void * data);
#endif
/*@}*/
/*@}*/
