/************************************************************
 *{{{ Copyright 2001,2002,2003:
 * 			Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Headers for flowgraph manipulation 
 *
 * Written by Ludo Van Put (lvanput@elis.ugent.be)
 *
 * Copyright (c) 2002 Ludo Van Put
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Ludo Van Put (lvanput@elis.ugent.be) }}}
 ************************************************************/

/*#define DEBUG_COND_BUNDLES*/

#include "diablo_dominator.h"
#include "diablo_loop.h"
#include "diablo_iterators.h"
#include "diablo_print.h"
#include "diablo_profile.h"
#include "../diablo_options.h"
#include "diablo_pre.h"
#include "../utils/util_error.h"
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_instructions.h"
#include "../arch/arm/arm_registers.h"
#include "../arch/arm/arm_utils.h"
#include "../arch/arm/arm_constant_propagation.h"
#endif
void CfgHoistConstantProducingCode(t_cfg * cfg)
{
  t_bbl * head, * tail;
  t_cfg_edge * edge;
  t_ins * ins, *jins;
  t_regset used_regs;
  t_bbl * i_bbl;
  t_loop * loop = NULL;
#ifdef DEBUG_HOISTING1
  static int teller = 0;
#endif

  CfgComputeHotBblThreshold(cfg, 0.90);

  CFG_FOREACH_EDGE(cfg,edge)
  {
    t_loopref* loopref;
    if (
#ifdef ARM_SUPPORT
	EDGE_CAT(edge)!=ET_JUMP &&
#endif
	EDGE_CAT(edge)!=ET_FALLTHROUGH
       ) continue;
    head = CFG_EDGE_HEAD(edge);
    tail = CFG_EDGE_TAIL(edge);
    if (BBL_NINS(tail)==0) continue;
    if (BBL_EXEC_COUNT(tail)<=EDGE_EXEC_COUNT(edge)) continue;
    /*      loop = BBL_LOOP(tail); TODO:FIXME, fix all loop interface things because it has changed a lot */
    BBL_FOREACH_LOOP(tail,loopref,loop)
      if(BBL_FUN(tail) == BBL_FUN(LOOP_HEADER(loop)))
      {
	t_loopiterator * loopiter;
	if (!BblDominates(head,tail)) continue;
	if (LoopContainsBbl(loop,head)) continue;

	ins = BBL_INS_FIRST(tail);
	/* need to test for reloc */
	used_regs = RegsetNew();

	BBL_FOREACH_INS(tail,ins)
	{
	  if (INS_IS_CONSTS(ins))
	  {
	    if (RegsetIsEmpty(RegsetNewIntersect(used_regs,INS_REGS_DEF(ins))))
	      if (RegsetIsEmpty(INS_REGS_USE(ins)))
		break;
	  }
	  RegsetSetUnion(used_regs,INS_REGS_USE(ins));
	  RegsetSetUnion(used_regs,INS_REGS_DEF(ins));
	}

	if (!ins) continue;


	/*      DiabloPrint(stdout,"header @eiB\n",LOOP_HEADER(loop));*/
	/*      DiabloPrint(stdout,"back @E\n",LOOP_BACKEDGE(loop));*/

	LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
	  {
	    
	    if (IsHellNode(i_bbl)) goto end;
	    /*          if (BBL_FUN(i_bbl) != BBL_FUN(LOOP_HEADER)) break;*/
	    /*	  DiabloPrint(stdout,"elem @B\n",LOOPELEM_BBL(elem)); */
	    BBL_FOREACH_INS(i_bbl,jins)
	      if (ins!=jins && !RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(jins),INS_REGS_DEF(ins))))
		goto end;
	  }

end:
	Free(loopiter);
	if (i_bbl)
	  continue;

	/*#define DEBUG_HOISTING1*/
#ifdef DEBUG_HOISTING1
	if(teller < global_options.debugcounter)
	{
	  teller++;
#endif
	DiabloPrint(stdout,"important candidate @I\n for code hoisting1  out of @iB\n",ins,tail);

	{
	  t_bbl * new_bbl = BblNew(cfg);
	  t_ins * new_ins;
	  t_cfg_edge * new_edge;
	  t_cfg_edge * new_edge2;

	  BblInsertInFunction(new_bbl,BBL_FUN(tail));
	  BBL_EXEC_COUNT(new_bbl)=EDGE_EXEC_COUNT(edge);

	  new_edge = CfgEdgeNew(cfg,new_bbl,tail,EDGE_CAT(edge));
	  EDGE_EXEC_COUNT(new_edge)=EDGE_EXEC_COUNT(edge);
	  if (EDGE_CORR(edge))
	    FATAL(("OOPS @E\n",edge));
	  CfgEdgeKill(edge);
	  new_edge2 = CfgEdgeNew(cfg,head,new_bbl,EDGE_CAT(new_edge));
	  EDGE_EXEC_COUNT(new_edge2)=EDGE_EXEC_COUNT(new_edge);

#ifdef ARM_SUPPORT
	  if (EDGE_CAT(new_edge)==ET_JUMP)
	  {
	    new_ins = InsNewForBbl(new_bbl);
	    ArmInstructionMakeUncondBranch(new_ins);		  
	    InsAppendToBbl(new_ins,new_bbl);
	  }
#endif
	  new_ins = InsDup(ins);
	  InsPrependToBbl(new_ins,new_bbl);
	  InsKill(ins);

	  /*	  DiabloPrint(stdout,"@ieB\n@ieB\n@ieB\n",head,new_bbl,tail);*/
	  

	}
#ifdef DEBUG_HOISTING1
	}
#endif
      }
  }
}

t_regset LoopFreeRegisters(t_cfg * cfg, t_loop * loop)
{
  t_bbl * bbl;
  t_regset regs = RegsetNew();
  t_loopiterator * loopiter;

  RegsetDup(regs,cfg->sec->obj->description->int_registers);
  RegsetSetDiff(regs,cfg->sec->obj->description->always_live);

  LOOP_FOREACH_BBL(loop,loopiter,bbl)
  {
    RegsetSetDiff(regs,BblRegsMaybeDef(bbl));
    RegsetSetDiff(regs,BblRegsUse(bbl));
  }
  Free(loopiter);
  return regs;
}


t_regset LoopUsedRegisters(t_cfg * cfg, t_loop * loop)
{
  t_bbl * bbl;
  t_regset regs = RegsetNew();
  t_loopiterator * loopiter;

  RegsetDup(regs,cfg->sec->obj->description->int_registers);
  RegsetSetDiff(regs,cfg->sec->obj->description->always_live);

  LOOP_FOREACH_BBL(loop,loopiter,bbl)
  {
    RegsetSetDiff(regs,BblRegsUse(bbl));
  }
  Free(loopiter);
  return regs;
}

t_ins * LoopContainsConstantProducers(t_cfg * cfg, t_loop * loop)
{
  t_bbl * i_bbl;
  t_ins * ins;
  t_loopiterator * loopiter;

  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
    BBL_FOREACH_INS(i_bbl,ins)
      if (INS_IS_CONSTS(ins))
      {
	Free(loopiter);
        return ins;
      }

  Free(loopiter);
  return NULL;
  
}

void CfgHoistConstantProducingCode2(t_cfg * cfg)
{
  t_ins * ins;
  t_regset free_regs;
  t_loop * loop;
  
  CfgComputeHotBblThreshold(cfg, 0.90);

  printf("ok\n");fflush(stdout);

  CFG_FOREACH_LOOP(cfg,loop)
    {
      t_bbl * header = LOOP_HEADER(loop); 
      DiabloPrint(stdout,"looking at loop beginning with bbl @B\n",header);fflush(stdout);
      continue;
      if (!BblIsHot(header)) continue;
      DiabloPrint(stdout,"1 found loop opportunity with hot header @iB\n",header);
      if (RegsetIsEmpty(free_regs = LoopFreeRegisters(cfg,loop))) continue;
      DiabloPrint(stdout,"2 found loop opportunity with hot header @iB\n",header);
      if (!(ins = LoopContainsConstantProducers(cfg,loop))) continue;
      DiabloPrint(stdout,"3 found loop opportunity with hot header @iB\n",header);
    }

}

void BblMovePredEdgesTo(t_cfg * cfg, t_bbl * from, t_bbl * to)
{
  t_cfg_edge * edge;
  t_cfg_edge * safe;
  
  BBL_FOREACH_PRED_EDGE_SAFE(from,edge,safe)
    {
      t_cfg_edge * copy;
      copy = CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge),to,EDGE_CAT(edge));
      EDGE_EXEC_COUNT(copy) = EDGE_EXEC_COUNT(edge);
      if (EDGE_CORR(edge))
	FATAL(("OOPS EDGE KILL @E\n",edge));
      CfgEdgeKill(edge);
    }
}


t_bbl * LoopAddPreheader(t_cfg * cfg,t_loop * loop)
{
  t_cfg_edge * edge;
  t_bbl * header = LOOP_HEADER(loop);
  t_bbl * fallthrough_preheader;
  t_bbl * new_preheader;
  t_bool is_fallthrough = FALSE;
  t_int32 exec_count = 0;
  t_int32 nr_entry_edges = 0;
  t_cfg_edge * new_edge;

  LOOP_FOREACH_ENTRY_EDGE(loop,edge)
    {
      if (EDGE_CAT(edge)==ET_FALLTHROUGH || EDGE_CAT(edge)==ET_IPFALLTHRU)
	{
	  is_fallthrough = TRUE;
	  fallthrough_preheader = CFG_EDGE_HEAD(edge);
	}
      nr_entry_edges++;
      exec_count += EDGE_EXEC_COUNT(edge);
    }
  
  is_fallthrough = FALSE;

  new_preheader = BblNew(cfg);
  BblInsertInFunction(new_preheader,BBL_FUN(header));
  BBL_EXEC_COUNT(new_preheader)=exec_count;
  
  if (!is_fallthrough)
    {
      t_ins * new_ins = InsNewForBbl(new_preheader);
#ifdef ARM_SUPPORT
      ArmInstructionMakeUncondBranch(new_ins);		  
#endif
      InsAppendToBbl(new_ins,new_preheader);
    }

  {  
    t_cfg_edge * edge;
    t_cfg_edge * safe;
    
    BBL_FOREACH_PRED_EDGE_SAFE(header,edge,safe)
      {
	t_cfg_edge * copy;
	if (LoopContainsBbl(loop,CFG_EDGE_HEAD(edge))) continue;
	copy = CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge),new_preheader,EDGE_CAT(edge));
	EDGE_EXEC_COUNT(copy) = EDGE_EXEC_COUNT(edge);
	if (EDGE_CORR(edge))
	  FATAL(("OOPS EDGE KILL @E\n",edge));
	CfgEdgeKill(edge);
      }
  }
  
  new_edge = CfgEdgeNew(cfg,new_preheader,header,is_fallthrough?ET_FALLTHROUGH:ET_JUMP);
  EDGE_EXEC_COUNT(new_edge)=exec_count;

  return new_preheader;
}

t_bbl * CreateNewTargetBlock(t_cfg * cfg, t_cfg_edge * edge)
{
  t_bbl * new_bbl = BblNew(cfg);
  t_cfg_edge * new_edge;
  t_cfg_edge * new_edge2;

  BblInsertInFunction(new_bbl,BBL_FUN(CFG_EDGE_TAIL(edge)));

  BBL_EXEC_COUNT(new_bbl)=EDGE_EXEC_COUNT(edge);

  if (EDGE_CAT(edge)==ET_JUMP || EDGE_CAT(edge)==ET_IPJUMP)
    {
      t_ins * new_ins = InsNewForBbl(new_bbl);
#ifdef ARM_SUPPORT
      ArmInstructionMakeUncondBranch(new_ins);		  
#endif
      InsAppendToBbl(new_ins,new_bbl);
    }

  new_edge = CfgEdgeNew(cfg,new_bbl,CFG_EDGE_TAIL(edge),EDGE_CAT(edge));
  EDGE_EXEC_COUNT(new_edge)=EDGE_EXEC_COUNT(edge);
  new_edge2 = CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge),new_bbl,EDGE_CAT(new_edge));
  EDGE_EXEC_COUNT(new_edge2)=EDGE_EXEC_COUNT(new_edge);
  if (EDGE_CORR(edge))
    FATAL(("OOPS EDGE KILL @E\n",edge));
  CfgEdgeKill(edge);
  
  return new_bbl;
}

t_bbl * CreateNewTargetBlockWithBranch(t_cfg * cfg, t_cfg_edge * edge)
{
  t_bbl * new_bbl = BblNew(cfg);
  t_ins * new_ins;
  t_cfg_edge * new_edge;
  t_cfg_edge * new_edge2;

  BblInsertInFunction(new_bbl,BBL_FUN(CFG_EDGE_TAIL(edge)));

  BBL_EXEC_COUNT(new_bbl)=EDGE_EXEC_COUNT(edge);

  new_ins = InsNewForBbl(new_bbl);
#ifdef ARM_SUPPORT
  ArmInstructionMakeUncondBranch(new_ins);		  
#endif
  InsAppendToBbl(new_ins,new_bbl);

  new_edge = CfgEdgeNew(cfg,new_bbl,CFG_EDGE_TAIL(edge),ET_JUMP);
  EDGE_EXEC_COUNT(new_edge)=EDGE_EXEC_COUNT(edge);
  new_edge2 = CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge),new_bbl,EDGE_CAT(edge));
  EDGE_EXEC_COUNT(new_edge2)=EDGE_EXEC_COUNT(new_edge);
  if (EDGE_CORR(edge))
    FATAL(("OOPS EDGE KILL @E\n",edge));
  CfgEdgeKill(edge);
  
  return new_bbl;
}

/*#define DEBUG_HOIST3*/

void CfgHoistConstantProducingCode3(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_cfg_edge * edge, * safe;
  t_ins * ins;
  t_loop * loop;
  t_regset free_regs, useable_regs;
  static t_bool done = FALSE;
  t_bbl * i_bbl;

#ifdef DEBUG_HOIST3
  static int teller = 0;
  t_string x = "triangle_preXXX.dot                  ";
  t_string y = "triangle_postXXX.dot                 ";
#endif
  
  /*  if (done) return;*/

  CfgComputeHotBblThreshold(cfg, 0.90);

  CFG_FOREACH_LOOP(cfg,loop)
    {
      t_bbl * header;
      t_bbl * preheader;
      t_reg free_reg;
      t_ins * new_ins;
      t_bool contains_exit_call = FALSE;
      t_loopiterator * loopiter;

      header = LOOP_HEADER(loop);

      if (global_options.blockprofilefile && !BblIsHot(header)) continue;

      LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
	{
	  BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
	    if (CfgEdgeIsInterproc(edge) || EDGE_CAT(edge)==ET_SWITCH || BBL_IS_LAST(CFG_EDGE_TAIL(edge))) contains_exit_call = TRUE;
	  BBL_FOREACH_PRED_EDGE(i_bbl,edge)
	    if (CfgEdgeIsInterproc(edge) || EDGE_CAT(edge)==ET_SWITCH || BBL_IS_LAST(CFG_EDGE_TAIL(edge))) contains_exit_call = TRUE;
	}
      Free(loopiter);

      LOOP_FOREACH_EXIT_EDGE_SAFE(loop,loopiter,i_bbl,edge,safe)      
	{
	  if (EDGE_CAT(edge)==ET_CALL) contains_exit_call = TRUE;
	}
      Free(loopiter);

      if (contains_exit_call) continue;

      do 
	{
	  t_loopiterator * loopiter;
#ifdef DEBUG_HOIST3
	  if (teller>=global_options.debugcounter) return;
#endif
	  ins = LoopContainsConstantProducers(cfg,loop);

	  if (!ins) break;

	  free_regs = LoopFreeRegisters(cfg,loop);
	  
	  RegsetDup(useable_regs,free_regs);
	  RegsetSubReg(useable_regs,ARM_REG_R13);
	  
	  if (RegsetIsEmpty(useable_regs)) break;
	  
	  if (!RegsetIn(free_regs,ARM_REG_R13)) break;
	  	  
	  /*	  DiabloPrintArch(stdout,cfg->sec->obj->description,"trying @iB\n with free regs @A\n ins @I\n",INS_BBL(ins),useable_regs,ins);*/

	  /*	  Export_Function(BBL_FUN(header),"file.dot");*/

#ifdef DEBUG_HOIST3
	  x = sDiabloPrint("hoist3_pre%d.dot",++teller);
	  y = sDiabloPrint("hoist3_post%d.dot",teller);
	  
	  Export_Function(BBL_FUN(INS_BBL(ins)),x);
#endif
	  preheader = LoopAddPreheader(cfg,loop);

	  
	  REGSET_FOREACH_REG(useable_regs,free_reg) 
	    break;
	  
	  InsPrependToBbl(new_ins = InsDup(ins),preheader);
	  
	  ARMINSREGA(new_ins)=free_reg;
	  INS_REGS_DEF(new_ins) = ArmDefinedRegisters(new_ins);
	  
	  ArmInsUnconditionalize(new_ins);

	  new_ins = InsNewForBbl(preheader);
	  ArmInstructionMakeStr(new_ins,free_reg,ARM_REG_R13,ARM_REG_NONE,8,ARM_CONDITION_AL,TRUE,FALSE,FALSE);		  
	  InsPrependToBbl(new_ins,preheader);

	  /*	  LOOP_FOREACH_LOOPELEM(loop,elem)
		  DiabloPrint(stdout,"bbl @B\n",LOOPELEM_BBL(elem));*/

	  LOOP_FOREACH_EXIT_EDGE_SAFE(loop,loopiter,i_bbl,edge,safe)
	    {
	      /*	      DiabloPrint(stdout,"from @eiB\n to @eiB\n",CFG_EDGE_HEAD(edge),CFG_EDGE_TAIL(edge));*/

	      t_bbl * new_target = CreateNewTargetBlock(cfg,edge);
	      DiabloPrint(stdout,"Created new bbl : @B",new_target);
	      new_ins = InsNewForBbl(new_target);
	      ArmInstructionMakeLdr(new_ins,free_reg,ARM_REG_R13,ARM_REG_NONE,8,ARM_CONDITION_AL,TRUE,FALSE,FALSE);		  
	      InsPrependToBbl(new_ins,new_target);	      
	    }
	  Free(loopiter);
	  
	  ArmInstructionMakeMov(ins,ARMINSREGA(ins),free_reg,0,ARMINSCONDITION(ins));
	  INS_REGS_USE(ins) = ArmUsedRegisters(ins);
	  INS_REGS_DEF(ins) = ArmDefinedRegisters(ins);
	  
	  /*	  Export_Function(BBL_FUN(header),"file.dot");*/

#ifdef DEBUG_HOIST3
	  Export_Function(BBL_FUN(INS_BBL(ins)),y);
#endif
	  
	  done = TRUE;
	} while (TRUE);
    }
#endif
}      

t_bool LoopHeaderIsHotter(t_loop * loop)
{
  t_int32 count = 0;
  t_cfg_edge * edge;

  LOOP_FOREACH_ENTRY_EDGE(loop,edge)
    count += EDGE_EXEC_COUNT(edge);

  if (count<BBL_EXEC_COUNT(LOOP_HEADER(loop))/4)
    return TRUE;
  else
    return FALSE;
}

t_bbl * BblDupInFun(t_cfg * cfg,t_bbl * bbl)
{
  t_ins * ins;
  t_bbl * copy;

  copy = BblNew(cfg);

  BblInsertInFunction(copy,BBL_FUN(bbl));

    /* duplicate the instructions in the basic block */
  BBL_FOREACH_INS(bbl,ins)
    {
      t_ins * inscopy = InsDup(ins);
      INS_OLD_ADDRESS(inscopy) = INS_OLD_ADDRESS(ins);
      InsAppendToBbl(inscopy, copy);
    }

  BBL_EXEC_COUNT(copy)=BBL_EXEC_COUNT(bbl);

  bbl->tmp = copy;

  return copy;
}

t_cfg_edge * EdgeDup(t_cfg* cfg,t_cfg_edge * edge, t_bbl * head, t_bbl * tail)
{
  t_cfg_edge * copy = CfgEdgeNew(cfg,head,tail,EDGE_CAT(edge));
  EDGE_EXEC_COUNT(copy)=EDGE_EXEC_COUNT(edge);
  return copy;
}


t_bbl * LoopDuplicate(t_cfg * cfg, t_loop * loop, t_bool share_counts)
{
  t_bbl * copy_bbl;
  t_bbl * header = LOOP_HEADER(loop);
  t_bbl * copy_header;
  t_cfg_edge * edge, *copy_edge;
  t_bbl * i_bbl;
  t_loopiterator * loopiter;

  copy_header = BblDupInFun(cfg,header);
  if (share_counts)
    {
      BBL_EXEC_COUNT(copy_header)/=2;
      BBL_EXEC_COUNT(header)/=2;
    }

  LOOP_FOREACH_BBL_INTRAFUN(loop,loopiter,i_bbl)
    {
      if (i_bbl==header) continue;
      copy_bbl = BblDupInFun(cfg,i_bbl);
      if (share_counts)
	{
	  BBL_EXEC_COUNT(copy_bbl)/=2;
	  BBL_EXEC_COUNT(i_bbl)/=2;
	}
    }
  Free(loopiter);

  LOOP_FOREACH_BBL_INTRAFUN(loop,loopiter,i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
	{
	  t_bbl * tail = CFG_EDGE_TAIL(edge);
	  if (LoopContainsBbl(loop,tail) && BBL_FUN(tail) == BBL_FUN(LOOP_HEADER(loop)))
	  {
	    copy_edge = EdgeDup(cfg,edge,i_bbl->tmp,tail->tmp);
	  }
	  else
	    {
	      copy_edge = EdgeDup(cfg,edge,i_bbl->tmp,tail);
	      if (EDGE_CAT(edge)==ET_FALLTHROUGH)
		{
		  t_bbl * target;
		  if (share_counts)
		    {
		      EDGE_EXEC_COUNT(copy_edge)/=2;
		    }
		  target = CreateNewTargetBlockWithBranch(cfg,copy_edge);
		  copy_edge = BBL_PRED_FIRST(target);
		}
	    }
	  if (share_counts)
	    {
	      EDGE_EXEC_COUNT(copy_edge)/=2;
	      EDGE_EXEC_COUNT(edge)/=2;
	    }
	} 
    }
  Free(loopiter);
  return copy_header;
}

/*#define DEBUG_INVARIANT_ORD*/


t_bool LoopExtractInvariantOrdinary(t_cfg * cfg, t_loop * loop, t_ins *ins)
{
  t_bbl * preheader;

#ifdef DEBUG_INVARIANT_ORD
  static   int counter = 0;
  t_string x = "INVARIANT_ORD_PREXXX.dot                    ";
  t_string y = "INVARIANT_ORD_POSTXXX.dot                   ";

  if (counter>=global_options.debugcounter) return FALSE;

#endif

#ifdef DEBUG_INVARIANT_ORD
  x = sDiabloPrint("INVARIANT_ORD_PRE%d.dot",++counter);
  y = sDiabloPrint("INVARIANT_ORD_POST%d.dot",counter);
  
  Export_Function(BBL_FUN(INS_BBL(ins)),x);

  DiabloPrint(stdout,"LOOP INVARIANT INS @I in loop bbl @iB\n",ins,INS_BBL(ins));
#endif
  preheader = LoopAddPreheader(cfg,loop);
  InsPrependToBbl(InsDup(ins),preheader);
  InsKill(ins);

#ifdef DEBUG_INVARIANT_ORD
  Export_Function(BBL_FUN(INS_BBL(ins)),y);
#endif	  

  return TRUE;
}

t_bool InsInBblDependOnEqNe(t_ins * ins)
{
#ifdef ARM_SUPPORT  
  t_regset regs = INS_REGS_DEF(ins);

  ins = INS_NEXT(ins);

  while (ins)
    {
      if (INS_ATTRIB(ins) & IF_CONDITIONAL)
	if (ARMINSCONDITION(ins)==ARM_CONDITION_NE || ARMINSCONDITION(ins)==ARM_CONDITION_EQ)
	  return TRUE;
      /*      if (!RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(ins),regs)))*/
      RegsetSetDiff(regs,INS_REGS_DEF(ins));
      ins = INS_NEXT(ins);
    }
  return FALSE;
#endif
  return FALSE;
}

/*#define DEBUG_INVARIANT_CMP*/

t_bool LoopExtractInvariantCMP(t_cfg * cfg, t_loop * loop, t_ins * ins)
{
  t_bbl * copy_header;
  t_bbl * split_block;
  t_bbl * preheader;
  t_bbl * copy_preheader;
  t_cfg_edge * copy_entry_edge;
  t_cfg_edge * edge;
#ifdef ARM_SUPPORT  
  t_int32 count;
  t_ins * i_ins, * copy_ins, * ins_new;
#endif

#ifdef DEBUG_INVARIANT_CMP
  static   int counter = 0;
  t_string x = "INVARIANT_CMP_PREXXX.dot                    ";
  t_string y = "INVARIANT_CMP_POSTXXX.dot                   ";

  if (counter>=global_options.debugcounter) return FALSE;
#endif

  if (!global_options.blockprofilefile && !InsInBblDependOnEqNe(ins))
    {
      return FALSE;
    }

  if (!global_options.blockprofilefile)
    {
      t_bbl * i_bbl;
      t_uint32 count = 0;
      t_loopiterator * loopiter;
      LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
	count+=BBL_NINS(i_bbl);
      Free(loopiter);
      if (count>32)
	return FALSE;
    }

  
#ifdef DEBUG_INVARIANT_CMP
  x = sDiabloPrint("INVARIANT_CMP_PRE%d.dot",++counter);
  y = sDiabloPrint("INVARIANT_CMP_POST%d.dot",counter);
  
  Export_Function(BBL_FUN(INS_BBL(ins)),x);

  DiabloPrint(stdout,"LOOP INVARIANT INS @I in loop bbl @iB\n",ins,INS_BBL(ins));
#endif

  copy_header = LoopDuplicate(cfg,loop,TRUE);

  split_block = LoopAddPreheader(cfg,loop);
  preheader = LoopAddPreheader(cfg,loop);
  copy_preheader = BblDupInFun(cfg,preheader);
  copy_entry_edge = EdgeDup(cfg,BBL_SUCC_FIRST(preheader),copy_preheader,copy_header);

  BBL_EXEC_COUNT(preheader)/=2;
  BBL_EXEC_COUNT(copy_preheader) = BBL_EXEC_COUNT(preheader);
  
  EDGE_EXEC_COUNT(BBL_SUCC_FIRST(preheader))/=2;

  EDGE_EXEC_COUNT(copy_entry_edge) = EDGE_EXEC_COUNT(BBL_SUCC_FIRST(preheader));
  
  if (BBL_INS_LAST(preheader))
    InsKill(BBL_INS_LAST(split_block));

  InsAppendToBbl(InsDup(ins),split_block);

  if (EDGE_CORR(BBL_SUCC_FIRST(split_block)))
    FATAL(("OOPS EDGE KILL @E\n",BBL_SUCC_FIRST(split_block)));
  CfgEdgeKill(BBL_SUCC_FIRST(split_block));

  edge = CfgEdgeNew(cfg,split_block,preheader,ET_JUMP);
  EDGE_EXEC_COUNT(edge)=BBL_EXEC_COUNT(split_block)/2;
  edge = CfgEdgeNew(cfg,split_block,copy_preheader,ET_FALLTHROUGH);
  EDGE_EXEC_COUNT(edge)=BBL_EXEC_COUNT(split_block)/2;

#ifdef ARM_SUPPORT  

  ins_new = InsNewForBbl(split_block);
  ArmInstructionMakeCondBranch(ins_new, ARM_CONDITION_EQ);
  InsAppendToBbl(ins_new,split_block);

  count = 0;
  
  BBL_FOREACH_INS(INS_BBL(ins),copy_ins)
    {
      if (copy_ins==ins)
	break;
      count++;
    }
  
  BBL_FOREACH_INS(T_BBL(INS_BBL(ins)->tmp),copy_ins)
    {
      count--;
      if (count<0)
	break;
    }

  i_ins = INS_NEXT(ins);

  while (i_ins)
    {
      if (ARMINSCONDITION(i_ins)==ARM_CONDITION_NE)
	{
	  t_ins * j_ins = INS_NEXT(i_ins);
	  if (ARMINSOPCODE(i_ins)==ARM_B)
	    {
	      EDGE_EXEC_COUNT(FallThroughPath(INS_BBL(i_ins)))+=EDGE_EXEC_COUNT(TakenPath(INS_BBL(i_ins)));
	      CfgEdgeKill(TakenPath(INS_BBL(i_ins)));
	    }
	  InsKill(i_ins);
	  i_ins = j_ins;
	  continue;
	}
      else if (ARMINSCONDITION(i_ins)==ARM_CONDITION_EQ)
	{
	  if (ARMINSOPCODE(i_ins)==ARM_B)
	    {
	      EDGE_EXEC_COUNT(TakenPath(INS_BBL(i_ins)))+=EDGE_EXEC_COUNT(FallThroughPath(INS_BBL(i_ins)));
	      CfgEdgeKill(FallThroughPath(INS_BBL(i_ins)));
	    }
	  ArmInsUnconditionalize(i_ins);
	  INS_REGS_USE(i_ins)=ArmUsedRegisters(i_ins);
	  INS_REGS_DEF(i_ins)=ArmDefinedRegisters(i_ins);
	}
      
      if (ArmInsUpdatesCond(i_ins))
	break;
    
      i_ins = INS_NEXT(i_ins);

    }

  i_ins = INS_NEXT(copy_ins);

  while (i_ins)
    {
      if (ARMINSCONDITION(i_ins)==ARM_CONDITION_EQ)
	{
	  t_ins * j_ins = INS_NEXT(i_ins);
	  if (ARMINSOPCODE(i_ins)==ARM_B)
	    {
	      EDGE_EXEC_COUNT(FallThroughPath(INS_BBL(i_ins)))+=EDGE_EXEC_COUNT(TakenPath(INS_BBL(i_ins)));
	      CfgEdgeKill(TakenPath(INS_BBL(i_ins)));
	    }
	  InsKill(i_ins);
	  i_ins = j_ins;
	  continue;
	}
      else if (ARMINSCONDITION(i_ins)==ARM_CONDITION_NE)
	{
	  if (ARMINSOPCODE(i_ins)==ARM_B)
	    {
	      EDGE_EXEC_COUNT(TakenPath(INS_BBL(i_ins)))+=EDGE_EXEC_COUNT(FallThroughPath(INS_BBL(i_ins)));
	      CfgEdgeKill(FallThroughPath(INS_BBL(i_ins)));
	    }
	  ArmInsUnconditionalize(i_ins);
	  INS_REGS_USE(i_ins)=ArmUsedRegisters(i_ins);
	  INS_REGS_DEF(i_ins)=ArmDefinedRegisters(i_ins);
	}
      
      if (ArmInsUpdatesCond(i_ins))
	break;
    
      i_ins = INS_NEXT(i_ins);
    }
#endif

#ifdef DEBUG_INVARIANT_CMP
  Export_Function(BBL_FUN(INS_BBL(ins)),y);
#endif	      

  return TRUE;
}

t_bool OptimizeLoopInvariantIns(t_cfg * cfg,t_loop * loop, t_ins * ins, t_bool is_single_defined)
{

#ifdef ARM_SUPPORT

  t_bbl * header = LOOP_HEADER(loop);
  t_bbl * bbl = INS_BBL(ins);

  t_bbl * i_bbl;
  t_cfg_edge * exit_edge, *edge;
  t_bbl * head, * tail;
  t_regset live_on_exit;
  t_loopiterator * loopiter;

  /* extra check: must dominate all exit edges, or produced value may not be live on exit edge!!! */

  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
  {

    BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
      if (CfgEdgeIsInterproc(edge) || EDGE_CAT(edge)==ET_SWITCH || BBL_IS_LAST(CFG_EDGE_TAIL(edge)))
      {
	Free(loopiter);
	return FALSE;
      }

    BBL_FOREACH_PRED_EDGE(i_bbl,edge)
      if (CfgEdgeIsInterproc(edge) || EDGE_CAT(edge)==ET_SWITCH || BBL_IS_LAST(CFG_EDGE_TAIL(edge)))
      {
	Free(loopiter);
	return FALSE;
      }
  }
  Free(loopiter);

  if (!global_options.blockprofilefile || (BblIsAlmostHot(bbl) && BblIsAlmostHot(header) && LoopHeaderIsHotter(loop)))
    if (ARMINSOPCODE(ins)!=ARM_CMP && !ArmIsControlflow(ins) && is_single_defined)
      {
	t_loopiterator * loopiter;
	LOOP_FOREACH_EXIT_EDGE(loop,loopiter,i_bbl,exit_edge)
	  {
	    head = CFG_EDGE_HEAD(exit_edge);
	    tail = CFG_EDGE_HEAD(exit_edge);

	    if (BblDominates(bbl,head)) continue;
	    
	    live_on_exit = BBL_INS_FIRST(tail)?InsRegsLiveBefore(BBL_INS_FIRST(tail)):BBL_REGS_LIVE_OUT(tail);
	    
	    if (!RegsetIsEmpty(RegsetNewIntersect(live_on_exit,INS_REGS_DEF(ins)))) 
	      {
		/* 
		   DiabloPrint(stdout,"FUCK, EXTRA CONDITION NOT MET@I\n",ins);
		   Export_Function(BBL_FUN(bbl),"aha.dot");
		   exit(0);
		*/
		return FALSE;
	      }
	  }
	Free(loopiter);
      }
  
  if (!global_options.blockprofilefile || (BblIsAlmostHot(bbl) && BblIsAlmostHot(header) && LoopHeaderIsHotter(loop)))
    {
      if (ARMINSOPCODE(ins)==ARM_CMP)
	{
	  return LoopExtractInvariantCMP(cfg,loop,ins);
	}
      else if (is_single_defined && !ArmIsControlflow(ins))
	{
	  return LoopExtractInvariantOrdinary(cfg,loop,ins);
	}
    }
#endif
  return FALSE;
}

void LoopDetectInvariants(t_cfg * cfg, t_loop* loop)      
{
  t_regset changed_regs = RegsetNew();
  t_regset multiple_defined_regs = RegsetNew();
  t_bbl * i_bbl;
  t_ins * ins;
  t_cfg_edge * edge;
  t_loopiterator * loopiter;
  
  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
	if (CfgEdgeIsInterproc(edge) || EDGE_CAT(edge)==ET_SWITCH || BBL_IS_LAST(CFG_EDGE_TAIL(edge)))
	{
	  Free(loopiter);
	  return;
	}

      BBL_FOREACH_PRED_EDGE(i_bbl,edge)
	if (EDGE_CAT(edge)==ET_COMPENSATING || EDGE_CAT(edge)==ET_RETURN || EDGE_CAT(edge)==ET_SWITCH)
	{
	  Free(loopiter);
	  return;
	}
      
      BBL_FOREACH_INS(i_bbl,ins)
	{
	  if (!RegsetIsEmpty(RegsetNewIntersect(changed_regs,INS_REGS_DEF(ins))))
	    RegsetSetUnion(multiple_defined_regs,INS_REGS_DEF(ins));
	  RegsetSetUnion(changed_regs,INS_REGS_DEF(ins));
	}
    }
  Free(loopiter);
  
  /* THIS COMPUTATION IS OVERLY SIMPLE SO FAR */
  
  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,ins)
	{
	  if ((INS_TYPE(ins)!=IT_LOAD) && (INS_TYPE(ins)!=IT_LOAD_MULTIPLE))
	    if ((INS_TYPE(ins)!=IT_STORE) && (INS_TYPE(ins)!=IT_STORE_MULTIPLE))
	      if ((INS_TYPE(ins)!=IT_FLT_STORE) && (INS_TYPE(ins)!=IT_FLT_LOAD))
		if (RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(ins),changed_regs)))
		  {
		    if (OptimizeLoopInvariantIns(cfg,loop,ins,RegsetIsEmpty(RegsetNewIntersect(INS_REGS_DEF(ins),multiple_defined_regs)))) 
		      {
			Free(loopiter);
			return;
		      }
		  }
	}
    }
  Free(loopiter);
}

void DetectLoopInvariants(t_cfg * cfg)
{
  t_loop * loop;

  CFG_FOREACH_LOOP(cfg,loop)
    LoopDetectInvariants(cfg,loop);
}

void LoopDetectStackSubAdds(t_cfg * cfg, t_loop* loop) 
{
#ifdef ARM_SUPPORT
  t_bbl * i_bbl;
  t_bbl *subbbl, * addbbl;
  t_ins * ins, * subins = NULL, *addins = NULL;
  t_cfg_edge * edge;
  t_loopiterator * loopiter;

  
  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
    {
      if (BBL_FUN(i_bbl)!=BBL_FUN(LOOP_HEADER(loop))) continue;
      BBL_FOREACH_SUCC_EDGE(i_bbl,edge)
	if (CfgEdgeIsInterproc(edge))
	  {
	    if (EDGE_CAT(edge)==ET_CALL &&  BBL_FUN(i_bbl)->cfg->sec->obj->actions->FunIsGlobal(BBL_FUN(i_bbl)))
	      continue;
	    else
	    {
	      Free(loopiter);
	      return;
	    }
	  }
    }  
  Free(loopiter);
  
  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
    {
      if (BBL_FUN(i_bbl)!=BBL_FUN(LOOP_HEADER(loop))) continue;
      BBL_FOREACH_INS(i_bbl,ins)
	{
	  if (ARMINSOPCODE(ins)==ARM_SUB && ARMINSREGA(ins)==ARM_REG_R13 && ARMINSREGB(ins)==ARM_REG_R13 && ARMINSREGC(ins)==ARM_REG_NONE && !ArmInsUpdatesCond(ins) && !ArmInstructionIsConditional(ins))
	    subins = ins;
	  else if (ARMINSOPCODE(ins)==ARM_ADD && ARMINSREGA(ins)==ARM_REG_R13 && ARMINSREGB(ins)==ARM_REG_R13 && ARMINSREGC(ins)==ARM_REG_NONE && !ArmInsUpdatesCond(ins) && !ArmInstructionIsConditional(ins))
	    addins = ins;
	}
    }  
  Free(loopiter);

  if (!addins || !subins) return;

  if (ARMINSIMMEDIATE(addins)!=ARMINSIMMEDIATE(subins)) return;

  subbbl = INS_BBL(subins);
  addbbl = INS_BBL(addins);

  if (!BblDominates(subbbl,addbbl)) return;
  if (!BblPostdominates(addbbl,subbbl)) return;

  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
    {
      if (BBL_FUN(i_bbl)!=BBL_FUN(LOOP_HEADER(loop))) continue;
    }
  Free(loopiter);

  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
    {
      if (BBL_FUN(i_bbl)!=BBL_FUN(LOOP_HEADER(loop))) continue;
      if (!BblDominates(subbbl,i_bbl)) continue;
      if (!BblPostdominates(addbbl,i_bbl)) continue;

      if (i_bbl==subbbl) continue;
      if (i_bbl==addbbl) continue;
      
      BBL_FOREACH_INS(i_bbl,ins)
	{
	  if (RegsetIn(INS_REGS_USE(ins),ARM_REG_R13) || RegsetIn(INS_REGS_DEF(ins),ARM_REG_R13))
	  {
	    Free(loopiter);
	    return;
	  }
	}
    }
  Free(loopiter);

  ins = INS_NEXT(subins);
  while(ins)
    {
      if (RegsetIn(INS_REGS_USE(ins),ARM_REG_R13)) return;
      if (RegsetIn(INS_REGS_DEF(ins),ARM_REG_R13)) return;
      ins=INS_NEXT(ins);
    }

  ins = INS_PREV(addins);
  while(ins)
    {
      if (RegsetIn(INS_REGS_USE(ins),ARM_REG_R13)) return;
      if (RegsetIn(INS_REGS_DEF(ins),ARM_REG_R13)) return;
      ins=INS_PREV(ins);
    }
  
  DiabloPrint(stdout,"found loop with bbls @iB\nand @iB\n",subbbl,addbbl);

  InsKill(subins);
  InsKill(addins);
#endif
}

void DetectLoopStackSubAdds(t_cfg * cfg)
{
  t_loop * loop;
  CFG_FOREACH_LOOP(cfg,loop)
    if (!global_options.blockprofilefile || BblIsHot(LOOP_HEADER(loop)))
      LoopDetectStackSubAdds(cfg,loop);
}

t_bool ExtractConditionalBundle(t_cfg * cfg, t_ins * ins)
{
#ifdef ARM_SUPPORT
  
  t_bbl * first_bbl;
  t_bbl * last_bbl;
  t_bbl * last_bbl2;
  t_bbl * new_bbl;
  t_ins * last_ins = NULL;
  t_ins * first_not_ins;
  t_ins * new_ins;
  t_ins * tmp;
  t_ins * jins;
  t_cfg_edge * new_edge;
  t_uint32 cond = ARMINSCONDITION(ins);
#ifdef DEBUG_COND_BUNDLES
  static int teller = 0;
  t_string x = "bundle_preXXX.dot                  ";
  t_string y = "bundle_postXXX.dot                 ";
#endif  
  jins = ins;
  
  while (jins)
    {
      if (ArmIsControlflow(jins))
	return FALSE;
      if (ARMINSCONDITION(jins)==cond)
	last_ins = jins;
      if (ArmInsUpdatesCond(jins))
	break;
      jins = INS_NEXT(jins);
    }

  first_bbl = INS_BBL(ins);
#ifdef DEBUG_COND_BUNDLES
  x = sDiabloPrint("bundle_pre%d.dot",++teller);
  y = sDiabloPrint("bundle_post%d.dot",teller);
  if (teller>=global_options.debugcounter) return FALSE;
  DiabloPrint(stdout,"DOING @ieB\n@I\n",first_bbl,ins);
  Export_Function(BBL_FUN(first_bbl),x);
#endif
  last_bbl = BblSplitBlock(first_bbl,ins,TRUE);
  new_bbl = BblNew(cfg);

  BblInsertInFunction(new_bbl,BBL_FUN(first_bbl));
  BBL_EXEC_COUNT(new_bbl)=INS_EXEC_COUNT(ins);

  new_ins = InsNewForBbl(first_bbl);
  ArmInstructionMakeCondBranch(new_ins,cond);		  
  InsAppendToBbl(new_ins,first_bbl);
  
  EDGE_EXEC_COUNT(BBL_SUCC_FIRST(first_bbl))=BBL_EXEC_COUNT(first_bbl)-INS_EXEC_COUNT(ins);
    
  new_edge = CfgEdgeNew(cfg,first_bbl,new_bbl,ET_JUMP);
  EDGE_EXEC_COUNT(new_edge)=INS_EXEC_COUNT(ins);
  
  first_not_ins = INS_NEXT(last_ins);

  jins = ins;
  
  BBL_FOREACH_INS_SAFE(last_bbl,jins,tmp)
    {
      if (jins==first_not_ins) break;
      if (ARMINSCONDITION(jins)==ArmInvertCondition(cond)) continue;
      new_ins = InsDup(jins);
      InsAppendToBbl(new_ins,new_bbl);
      ArmInstructionUnconditionalizer(new_ins);
      INS_EXEC_COUNT(new_ins)=BBL_EXEC_COUNT(new_bbl);
      INS_EXEC_COUNT(jins)-=INS_EXEC_COUNT(new_ins);
      if (ARMINSCONDITION(jins)==cond)
	InsKill(jins);
    }

  new_ins = InsNewForBbl(new_bbl);
  ArmInstructionMakeUncondBranch(new_ins);		  
  InsAppendToBbl(new_ins,new_bbl);
  
  INS_EXEC_COUNT(new_ins)=BBL_EXEC_COUNT(new_bbl);

  BBL_EXEC_COUNT(last_bbl)-=BBL_EXEC_COUNT(new_bbl);
  EDGE_EXEC_COUNT(BBL_SUCC_FIRST(last_bbl))-=BBL_EXEC_COUNT(new_bbl);

  if (first_not_ins==BBL_INS_FIRST(last_bbl) || first_not_ins==NULL)
    {
      last_bbl2 = last_bbl;
    }
  else
    {
      last_bbl2 = BblSplitBlock(last_bbl,first_not_ins,TRUE);
    }

  new_edge = CfgEdgeNew(cfg,new_bbl,last_bbl2,ET_JUMP);

  EDGE_EXEC_COUNT(new_edge)=BBL_EXEC_COUNT(new_bbl);
#ifdef DEBUG_COND_BUNDLES
  Export_Function(BBL_FUN(first_bbl),y);
  DiabloPrint(stdout,"DID IT @ieB\n@ieB\n@ieB\n@ieB\n",first_bbl,new_bbl,last_bbl,last_bbl2);
#endif  
  return TRUE;
#endif

  return FALSE;
}

void DetectColdCodeBundles(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins;
  t_uint32 size;

  CFG_FOREACH_BBL(cfg,bbl)
    {
    restart:
      BBL_FOREACH_INS(bbl,ins)
	{
	  if (BblIsHot(bbl))
	    if ((size=ConditionalBundleSize(ins))>=2)
	      {
		
		t_int32 gain = size*(BBL_EXEC_COUNT(bbl)-INS_EXEC_COUNT(ins));
		t_int32 loss = BBL_EXEC_COUNT(bbl) /* for inserting the branch */ 
		               + INS_EXEC_COUNT(ins) /* for missing the branch */ 
           		       + 2 * INS_EXEC_COUNT(ins) /* for adding an unconditional branch and missing it */;
		t_int32 net_gain = gain-loss;
		
		if (net_gain >= BBL_EXEC_COUNT(bbl)/2)
		  {
		    if (ExtractConditionalBundle(cfg,ins))
		      goto restart;
		  }
	      }
	}
    }
}


/*#define DEBUG_LOOP_UNROLLING*/

void LoopUnrollingSimple(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_loop * loop;
  t_uint32 nins,nbbls;
  t_bbl * i_bbl;
  t_bbl * bbl;
  t_uint32 total_nins = 0;
#ifdef DEBUG_LOOP_UNROLLING
  static int teller = 0;
#endif

  CfgEstimateEdgeCounts(cfg);

  CFG_FOREACH_BBL(cfg,bbl)
    total_nins+=BBL_NINS(bbl);

  CFG_FOREACH_LOOP(cfg,loop)
    {
      t_bbl * header = LOOP_HEADER(loop); 
      t_loopiterator * loopiter;
      
#ifdef DEBUG_LOOP_UNROLLING
      if (teller>=global_options.debugcounter) continue;
#endif

      if (!BblIsHot(header)) continue;
      if (!LoopHeaderIsHotter(loop)) continue;
      if (LOOP_COUNT(loop)>4) continue;
      
      nins = nbbls = 0;
      
      LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
	{
	  nins+=BBL_NINS(i_bbl);
	  nbbls++;
	}
      Free(loopiter);
      
      if (nins>total_nins/500) continue;
      if (nbbls>2) continue;

      if (nbbls==1)
	{
	  /*	  Export_Function(BBL_FUN(header),"pre.dot");*/

	  t_bbl * copy = BblDupInFun(cfg,header);
	  t_cfg_edge * fall = FallThroughPath(header);
	  t_cfg_edge * back_edge = TakenPath(header);
	  t_ins * branch_ins = BBL_INS_LAST(header);

#ifdef DEBUG_LOOP_UNROLLING
	  teller++;
#endif	 
	  DiabloPrint(stdout,"HOT LOOP UNROLLING CANDIDATE WITH %d ins and %d bbls @B\n",nins,nbbls,header);
	  
	  BBL_EXEC_COUNT(copy)/=2;
	  BBL_EXEC_COUNT(header)/=2;
	  
	  EDGE_EXEC_COUNT(CfgEdgeNew(cfg,copy,header,ET_JUMP))=EDGE_EXEC_COUNT(back_edge)/2;
	  EDGE_EXEC_COUNT(CfgEdgeNew(cfg,copy,CFG_EDGE_TAIL(fall),ET_FALLTHROUGH))=EDGE_EXEC_COUNT(fall)/2;
	  
	  if (EDGE_CAT(fall)==ET_FALLTHROUGH)
	    EDGE_CAT(fall)=ET_JUMP;
	  else
	    EDGE_CAT(fall)=ET_IPJUMP;
	  
	  EDGE_EXEC_COUNT(fall)/=2;

	  CfgEdgeKill(back_edge);

	  EDGE_EXEC_COUNT(CfgEdgeNew(cfg,header,copy,ET_FALLTHROUGH))=BBL_EXEC_COUNT(header)-EDGE_EXEC_COUNT(fall);
	  
	  
	  ARMINSCONDITION(branch_ins) = ArmInvertCondition(ARMINSCONDITION(branch_ins));
	  /*	  Export_Function(BBL_FUN(header),"post.dot");*/

	  /*	  DiabloPrint(stdout,"DONE @ieB\n @ieB\n",header,copy);*/
	}
      else 
	{	
	  t_bool only_loop_header_exits = TRUE;
	  t_cfg_edge * exit_edge = NULL;
	  t_bbl * other_bbl = NULL;
	  t_uint32 nr_exit_edges = 0;
	  t_loopiterator * loopiter;
	  /*	  DiabloPrint(stdout,"HOT LOOP UNROLLING CANDIDATE WITH %d ins and %d bbls @B\n",nins,nbbls,header);*/
  
	  LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
	    {
	      if (i_bbl!=header && BBL_SUCC_FIRST(i_bbl)!=BBL_SUCC_LAST(i_bbl))
		{
		  only_loop_header_exits = FALSE;
		  /*		  printf("UNROLLING HOLA\n");*/
		}
	      if (i_bbl!=header)
		other_bbl = i_bbl;
	      /*	      DiabloPrint(stdout,"@ieB\n",i_bbl);*/
	    }
	  Free(loopiter);
	  
	  LOOP_FOREACH_EXIT_EDGE(loop,loopiter,i_bbl,exit_edge)
	    nr_exit_edges++;
	  Free(loopiter);
	    
	  LOOP_FOREACH_EXIT_EDGE(loop,loopiter,i_bbl,exit_edge)
	    break;
	  Free(loopiter);
	  
	  if (only_loop_header_exits && EDGE_CAT(exit_edge)==ET_FALLTHROUGH)
	    {
	      /*	      Export_Function(BBL_FUN(header),"pre.dot");*/

	      t_bbl * copy_header = LoopDuplicate(cfg, loop, TRUE);

	      t_cfg_edge * taken_edge = TakenPath(header);

	      t_cfg_edge * taken_edge_copy = TakenPath(copy_header);
	      t_cfg_edge * fall_edge_copy = FallThroughPath(copy_header);

	      t_uint32 tmp_type;

	      t_ins * branch_ins = BBL_INS_LAST(copy_header);
	      
	      DiabloPrint(stdout,"HOT LOOP UNROLLING CANDIDATE WITH %d ins and %d bbls @B\n",nins,nbbls,header);
#ifdef DEBUG_LOOP_UNROLLING
	  teller++;
#endif	 

	      CfgEdgeNew(cfg,header,CFG_EDGE_TAIL(taken_edge)->tmp,ET_JUMP);
	      CfgEdgeKill(taken_edge);

	      tmp_type = EDGE_CAT(taken_edge_copy);
	      EDGE_CAT(taken_edge_copy)=EDGE_CAT(fall_edge_copy);
	      EDGE_CAT(fall_edge_copy)=tmp_type;

	      CfgEdgeKill(taken_edge_copy);
	      CfgEdgeNew(cfg,copy_header,other_bbl,ET_FALLTHROUGH);

	      ARMINSCONDITION(branch_ins)=ArmInvertCondition(ARMINSCONDITION(branch_ins));
	      /*	      Export_Function(BBL_FUN(header),"post.dot");	      */
	    }
	  else if (only_loop_header_exits && EDGE_CAT(exit_edge)==ET_JUMP)
	    {
	      t_loopiterator * loopiter;
	      DiabloPrint(stdout,"OEPS 1 HOT LOOP UNROLLING CANDIDATE WITH %d ins and %d bbls @B\n",nins,nbbls,header);	      
	      LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
		{
		  DiabloPrint(stdout,"@ieB\n",i_bbl);
		}
	      Free(loopiter);
	    }
	  else if (nr_exit_edges==1)
	    {
	      t_loopiterator * loopiter;
	      DiabloPrint(stdout,"OEPS 2 HOT LOOP UNROLLING CANDIDATE WITH %d ins and %d bbls @B\n",nins,nbbls,header);	      
	      LOOP_FOREACH_BBL(loop,loopiter,i_bbl)
		{
		  DiabloPrint(stdout,"@ieB\n",i_bbl);
		}    
	      Free(loopiter);
	    }
	  else 	    
	    {
	      t_bbl * copy_header;
	      t_bbl * copy_other_bbl;
	      t_cfg_edge * edge, * tmp_edge;

	      /*	      Export_Function(BBL_FUN(header),"pre.dot");*/

	      DiabloPrint(stdout,"HOT LOOP UNROLLING CANDIDATE WITH %d ins and %d bbls and %d exit edges @B\n",nins,nbbls,nr_exit_edges,header);	      
#ifdef DEBUG_LOOP_UNROLLING
	  teller++;
#endif	 
	      /*
	      LOOP_FOREACH_BBL(loop,i_bbl)
		{
		  DiabloPrint(stdout,"@ieB\n",i_bbl);
		}
	      */
	      
	      copy_header = LoopDuplicate(cfg, loop, TRUE);
	      copy_other_bbl = T_BBL(other_bbl->tmp);

	      BBL_FOREACH_SUCC_EDGE_SAFE(copy_other_bbl,edge,tmp_edge)
		{
		  if (CFG_EDGE_TAIL(edge)==copy_header)
		    {
		      /* Ludo: this code will not work anymore since loop inteface has changed thoroughly */
		      CfgEdgeNew(cfg,other_bbl->tmp,header,EDGE_CAT(LOOP_BACKEDGES(loop)->data));
		      CfgEdgeKill(edge);
		    }
		}
	      
	      /* Ludo: this code will not work anymore since loop inteface has changed thoroughly */
	      CfgEdgeNew(cfg,other_bbl,copy_header,EDGE_CAT(LOOP_BACKEDGES(loop)->data));
	      CfgEdgeKill(LOOP_BACKEDGES(loop)->data);

	      /*	      Export_Function(BBL_FUN(header),"post.dot");*/
	    }
	}
    }
#endif
}
