/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/
#define DEBUG_DOMINATOR_LEAKS
/*#define DEBUG_DOMINATOR*/
/*{{{ Includes*/
#include "../config.h"
#include "diablo_dominator.h"
#include "diablo_print.h"
#include "diablo_flowgraph.h"
#include "diablo_iterators.h"
#include "../diablo_options.h"
#include "../datastructures/data_ll.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
/*}}}*/

t_bool dominator_info_correct;

#define OPT_DOMINATES
/*#define OPT_DOMINATES2*/
t_bbl all_bbls;
const t_bbl * ALL_BBLS = &all_bbls;

t_uint32 dominator_node_marking_number=1;

typedef struct _t_queue
{
  dll * start;
  dll * end;
} t_queue;

t_uint32 mask_dominator = ET_FALLTHROUGH | ET_CALL | ET_RETURN | ET_JUMP | /*ET_SWI |*/ ET_UNKNOWN | ET_IPUNKNOWN | ET_IPFALLTHRU | ET_IPJUMP /*| ET_FUNLINK | ET_IPFUNLINK*/ | ET_IPSWILINK | ET_SWILINK | ET_COMPENSATING | ET_SWITCH;

t_uint32 mask_postdominator = ET_FALLTHROUGH | ET_CALL | ET_RETURN | ET_JUMP /*| ET_SWI*/ | ET_UNKNOWN | ET_IPUNKNOWN | ET_IPFALLTHRU | ET_IPJUMP /*| ET_FUNLINK | ET_IPFUNLINK*/ | ET_IPSWILINK | ET_SWILINK | ET_COMPENSATING | ET_SWITCH | ET_POSTDOM;

/* {{{ Forward declarations */
t_bool DominatorCallback(t_node * first, t_uint32 mask);
t_bool DominatorIntersect(data_ll_elem ** first, data_ll_elem * second);
t_bool DominatorDiffers(data_ll_elem * first, data_ll_elem * second);
t_bool DominatorContains(data_ll_elem * list, t_bbl * find_me);
t_bool PostDominatorCallback(t_node * first, t_uint32 mask);
void DominatorFree(data_ll_elem ** list);
data_ll_elem * DominatorUnion(data_ll_elem * first, data_ll_elem * second, t_bool * changes);
t_bool QueueContains(t_queue queue,void * data);
void InitNodeMarkingDominator(void);
#define NodeIsMarkedDominator(node) (T_BBL(node)->tmp==(void*)dominator_node_marking_number)
#define NodeMarkDominator(node)     (T_BBL(node)->tmp=(void*)dominator_node_marking_number)
#define NodeUnmarkDominator(node)   (T_BBL(node)->tmp=(void*)0)
t_bool BblDominates2(t_bbl * dominator, t_bbl * dominated);
/* }}} */

void DominatorPrint(data_ll_elem * dom)/*{{{*/
{
  while(dom)
  {
    if(dom->data != ALL_BBLS)
      DiabloPrint(stdout,"@G(@G)  ",BBL_OLD_ADDRESS(dom->data),BBL_FUN(dom->data)?BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(dom->data))):0x0);
    else
      DiabloPrint(stdout,"ALL_BBLS  ");
    dom = LLNext(dom);
  }
  printf("\n");
}/*}}}*/

/*!
 * \param t_cfg
 *
 * Compute dominator information on the graph, ignoring (ip)funlink-edges.
 * BBL_TMP_DOMINATED is set to a linked list of all nodes that dominate
 * the current. If it's empty, then either the node is unreachable 
 * or the last one in the graph. Currently
 * the implementation is straightforward, and is not based on a near-linear
 * algorithm. The performance is still acceptable, but this could change,
 * forcing a more efficient implementation.
 *
 * Also postdominators are being calculated, but first loop detection is done,
 * to connect infinite loops with the unique exit node. While doing loop
 * detection, for each basic block, it's loop_edge field gets filled in with
 * the backedge defining a loop. */


/*#define OLD_COMPUTATIONS *//*|+for the moment this has to be defined, or dominator will not work +|*/

void ComDominators(t_cfg * cfg)
{
  t_bbl * ibbl;
  t_function * ifun;
  ll * looplist = NULL;
  /*t_uint32 iterations = 0;*/
/*  t_uint32 changed = 1;*/
  t_cfg_edge *i_edge;
  t_cfg_edge *j_edge;
#ifdef OLD_COMPUTATIONS
  data_dll_elem *tmp_dll, *elem = NULL;
  t_queue dom_queue = {NULL,NULL};
#else
  t_bool change;
  t_function * fun;
  t_bool big_change;
#endif

  if (dominator_info_correct)
    {
      printf("SKIPPED unnecessary DOM\n");
      return;
    }

  ANALYSIS_START("Dominator calculation");

  /* first some initializations, setting all required fields to some value {{{*/
  /* All blocks are initially dominated by all blocks, except for the entry */
  
  CFG_FOREACH_FUN(cfg, ifun)
  {
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      DominatorFree(&BBL_TMP_DOMINATED(ibbl));
      DominatorFree(&BBL_TMP_PDOMINATED(ibbl));

#ifndef DEBUG_DOMINATOR_LEAKS
      BBL_TMP_DOMINATED(ibbl) = LLCalloc((void*)ALL_BBLS);
      BBL_TMP_PDOMINATED(ibbl) = LLCalloc((void*)ALL_BBLS);
#else
      BBL_TMP_DOMINATED(ibbl) = Calloc(1,sizeof(data_ll_elem));
      BBL_TMP_DOMINATED(ibbl)->data = (void*)ALL_BBLS;
      BBL_TMP_PDOMINATED(ibbl) = Calloc(1,sizeof(data_ll_elem));
      BBL_TMP_PDOMINATED(ibbl)->data = (void*)ALL_BBLS;
#endif
      BBL_DOMINATED_BY(ibbl) = NULL;
      BBL_PDOMINATED_BY(ibbl)= NULL;
    }
  }
#ifndef DEBUG_DOMINATOR_LEAKS  
  if(!BBL_TMP_DOMINATED(cfg->hell_node)) BBL_TMP_DOMINATED(cfg->hell_node) =  LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_DOMINATED(cfg->call_hell_node)) BBL_TMP_DOMINATED(cfg->call_hell_node) = LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_DOMINATED(cfg->exit_hell_node)) BBL_TMP_DOMINATED(cfg->exit_hell_node) = LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_DOMINATED(cfg->exit_call_hell_node)) BBL_TMP_DOMINATED(cfg->exit_call_hell_node) = LLCalloc((void*)ALL_BBLS);
  /* the unique entry point initially is dominated by itself */
  if(!BBL_TMP_DOMINATED(cfg->unique_exit_node)) BBL_TMP_DOMINATED(cfg->unique_exit_node) =  LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_DOMINATED(cfg->unique_entry_node)) BBL_TMP_DOMINATED(cfg->unique_entry_node) =  LLCalloc((void*)cfg->unique_entry_node);
  else BBL_TMP_DOMINATED(cfg->unique_entry_node)->data = (void*)cfg->unique_entry_node; 
#else
  if(!BBL_TMP_DOMINATED(cfg->hell_node)) 
  {
    BBL_TMP_DOMINATED(cfg->hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_DOMINATED(cfg->hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_DOMINATED(cfg->call_hell_node))
  {
    BBL_TMP_DOMINATED(cfg->call_hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_DOMINATED(cfg->call_hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_DOMINATED(cfg->exit_hell_node))
  {
    BBL_TMP_DOMINATED(cfg->exit_hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_DOMINATED(cfg->exit_hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_DOMINATED(cfg->exit_call_hell_node))
  {
    BBL_TMP_DOMINATED(cfg->exit_call_hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_DOMINATED(cfg->exit_call_hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_DOMINATED(cfg->unique_exit_node))
  {
    BBL_TMP_DOMINATED(cfg->unique_exit_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_DOMINATED(cfg->unique_exit_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_DOMINATED(cfg->unique_entry_node))
    BBL_TMP_DOMINATED(cfg->unique_entry_node)=Calloc(1,sizeof(data_ll_elem));
  BBL_TMP_DOMINATED(cfg->unique_entry_node)->data =(void*)cfg->unique_entry_node;
#endif

#ifndef DEBUG_DOMINATOR_LEAKS
  if(!BBL_TMP_PDOMINATED(cfg->hell_node)) BBL_TMP_PDOMINATED(cfg->hell_node) =  LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_PDOMINATED(cfg->call_hell_node)) BBL_TMP_PDOMINATED(cfg->call_hell_node) = LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_PDOMINATED(cfg->exit_hell_node)) BBL_TMP_PDOMINATED(cfg->exit_hell_node) = LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_PDOMINATED(cfg->exit_call_hell_node)) BBL_TMP_PDOMINATED(cfg->exit_call_hell_node) = LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_PDOMINATED(cfg->unique_entry_node)) BBL_TMP_PDOMINATED(cfg->unique_entry_node) =  LLCalloc((void*)ALL_BBLS);
  if(!BBL_TMP_PDOMINATED(cfg->unique_exit_node)) BBL_TMP_PDOMINATED(cfg->unique_exit_node) =  LLCalloc((void*)cfg->unique_exit_node);
  else BBL_TMP_PDOMINATED(cfg->unique_exit_node)->data = (void*)cfg->unique_exit_node; 
#else
  if(!BBL_TMP_PDOMINATED(cfg->hell_node))
  {
    BBL_TMP_PDOMINATED(cfg->hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_PDOMINATED(cfg->hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_PDOMINATED(cfg->call_hell_node))
  {
    BBL_TMP_PDOMINATED(cfg->call_hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_PDOMINATED(cfg->call_hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_PDOMINATED(cfg->exit_hell_node))
  {
    BBL_TMP_PDOMINATED(cfg->exit_hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_PDOMINATED(cfg->exit_hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_PDOMINATED(cfg->exit_call_hell_node))
  {
    BBL_TMP_PDOMINATED(cfg->exit_call_hell_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_PDOMINATED(cfg->exit_call_hell_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_PDOMINATED(cfg->unique_entry_node))
  {
    BBL_TMP_PDOMINATED(cfg->unique_entry_node)=Calloc(1,sizeof(data_ll_elem));
    BBL_TMP_PDOMINATED(cfg->unique_entry_node)->data =(void*)ALL_BBLS;
  }
  if(!BBL_TMP_PDOMINATED(cfg->unique_exit_node))
    BBL_TMP_PDOMINATED(cfg->unique_exit_node)=Calloc(1,sizeof(data_ll_elem));
  BBL_TMP_PDOMINATED(cfg->unique_exit_node)->data =(void*)cfg->unique_exit_node;
#endif
/*}}}*/

  /* iterate until fixpoint */

  ANALYSIS_START("Regular dominator information");
  /*{{{*/

#ifdef OLD_COMPUTATIONS
  BBL_FOREACH_SUCC_EDGE(cfg->unique_entry_node,i_edge)
  {
    if(EDGE_CAT(i_edge) & mask_dominator)
    {
      tmp_dll = (data_dll_elem *)Calloc(1,sizeof(data_dll_elem));
      tmp_dll->data = (void *)CFG_EDGE_TAIL(i_edge);
      DLLAppend(NULL,tmp_dll,&dom_queue.start,&dom_queue.end);
    }
  }

  VERBOSE(0,("Start"));

  /* Calculate dominators interprocedurally */

  while(dom_queue.end)
  {
    elem = (data_dll_elem*)dom_queue.end;
    DLLUnlink(dom_queue.end,&dom_queue.start,&dom_queue.end);

    /*    VERBOSE(0,("Visit @B",elem->data));*/
    if (DominatorCallback(elem->data,mask_dominator))
    {
      BBL_FOREACH_SUCC_EDGE(elem->data,i_edge)
      {
	if(!QueueContains(dom_queue,CFG_EDGE_TAIL(i_edge)))
	{
	  tmp_dll = (data_dll_elem *)Calloc(1,sizeof(data_dll_elem));
	  tmp_dll->data = (void *)CFG_EDGE_TAIL(i_edge);
	  DLLAppend(NULL,tmp_dll,&dom_queue.start,&dom_queue.end);
/*          DiabloPrint(stdout,"@B is in queue\n",tmp_dll->data);*/
	}
	if(EDGE_CORR(i_edge) && !QueueContains(dom_queue,CFG_EDGE_TAIL(EDGE_CORR(i_edge))))
	{
	  tmp_dll = (data_dll_elem *)Calloc(1,sizeof(data_dll_elem));
	  tmp_dll->data = (void *)CFG_EDGE_TAIL(EDGE_CORR(i_edge));
	  DLLAppend(NULL,tmp_dll,&dom_queue.start,&dom_queue.end);
/*          DiabloPrint(stdout,"@B is in queue\n",tmp_dll->data);*/
	}
      }
    }
    Free(elem);
  }
#else
  InitNodeMarking();

  InitEdgeMarking();

  CFG_FOREACH_FUN(cfg,fun)
    FUN_FLAGS(fun)&=~FF_IS_MARKED;

  BBL_FOREACH_SUCC_EDGE(cfg->unique_entry_node,i_edge)
    if(EDGE_CAT(i_edge) & mask_dominator)
      {
	EdgeMark(i_edge);
	NodeMark(CFG_EDGE_TAIL(i_edge));
	FUN_FLAGS(BBL_FUN(CFG_EDGE_TAIL(i_edge)))|=FF_IS_MARKED;
      }

  VERBOSE(0,("Start"));

  do
    {
    restart:
      big_change = FALSE;
      CFG_FOREACH_FUN(cfg,fun)
	{
	  if(!(FUN_FLAGS(fun)&FF_IS_MARKED))
	    continue;
	  
	  FUN_FLAGS(fun)&=~FF_IS_MARKED;
	  
	  /*	  DiabloPrint(stdout,"doing it for @B\n",FUN_BBL_FIRST(fun));*/

	  do
	    {
	      t_bbl * bbl;
	      change = FALSE;
	      
	      FUN_FOREACH_BBL(fun,bbl)
		{
		  if (!NodeIsMarked(bbl)) continue;

		  NodeUnmark(bbl);

		  if (DominatorCallback(T_NODE(bbl),mask_dominator))
		    {
		      BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
			{
			  t_bbl * tail = CFG_EDGE_TAIL(i_edge);
			  t_cfg_edge * edge2 = T_CFG_EDGE(EDGE_CORR(i_edge));
			  EdgeMark(i_edge);
			  if (!NodeIsMarked(tail))
			    {
			      NodeMark(tail);
			      if (CfgEdgeIsInterproc(i_edge))
				{
				  FUN_FLAGS(BBL_FUN(tail))|=FF_IS_MARKED;
				  big_change = TRUE;
				}
			      else
				change = TRUE;
			    }

			  if (edge2)
			    {
			      t_bbl * tail2 = CFG_EDGE_TAIL(edge2);
			      EdgeMark(edge2);
			      if (!NodeIsMarked(tail2))
				{
				  NodeMark(tail2);
				  /*				      if (CfgEdgeIsInterproc(i_edge)) SHOULD ALWAYS BE TRUE*/
				  {
				    FUN_FLAGS(BBL_FUN(tail2))|=FF_IS_MARKED;
				    big_change = TRUE;
				  }
				  /*				      else
								      {
								      change = TRUE;
								      FATAL(("an edge with a corresponding edge should be interprocedural"));
								      }*/
				}
			    }
			}
		    }
		}
	    } while (change);
	  if (big_change) goto restart;
	}
    } while (big_change);
  
#endif
  
#if 0
  FILE* fp = fopen("log_new","w");

  CFG_FOREACH_BBL(cfg,ibbl)
    {
      data_ll_elem * iter = BBL_TMP_DOMINATED(ibbl);
      while (iter)
	{
	  DiabloPrint(fp,"@B dominated by @B\n",ibbl, (t_bbl*) iter->data);
	  iter = LLNext(iter);
	}
    }

  fclose(fp);
#endif

  VERBOSE(0,("Start part 2"));
  
  /* now use the info to fill in the IDOMINATED_BY fields */
  
  CFG_FOREACH_FUN(cfg,ifun)
  {
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      {
	data_ll_elem * iter = BBL_TMP_DOMINATED(ibbl);
#ifdef DEBUG_DOMINATOR
	DiabloPrint(stdout,"Idom, now doing @B",ibbl);
#endif
	if(iter->data == ALL_BBLS)
	{
#ifdef DEBUG_DOMINATOR
	  BblPrint(ibbl,PRINT_INSTRUCTIONS);DiabloPrint(stdout," still dominated by ALL_BBLS\n");
#endif
	  BBL_DOMINATED_BY(ibbl) = NULL;
	  continue;
	}
	if(iter->data == ibbl)
	  /* If only dominated by itself: should only happen for unique entry */
	{
	  iter = LLNext(iter);
	  if(!iter) 
	  {
	    if(ibbl != cfg->unique_entry_node)
	      FATAL(("Immediate dominator troubles\n"));
	    BBL_DOMINATED_BY(ibbl) = NULL;
	    continue;
	  }
	}
	/*iter = LLNext(iter); */
#ifdef OLD_COMPUTATIONS
	while(iter)
	{
	  if (iter->data != ibbl)
	  {
	    t_bool add_it = TRUE;
	    data_ll_elem * iter2 = BBL_TMP_DOMINATED(ibbl);
	    while(iter2)
	    {
	      if(iter2->data == iter->data || iter2->data == ibbl)
	      {
		iter2 = LLNext(iter2);
		continue;
	      }
	      /* If iter2->data dominates iter->data, we don't have to keep iter2->data in the list but iter->data instead. Iter->data is nearer
	       * to ibbl than iter2->data. */
	      if (DominatorContains(BBL_TMP_DOMINATED(iter2->data),iter->data))
	      {
		add_it = FALSE;
		break;
	      }
	      iter2 = LLNext(iter2);

	    }
	    /* Add an entry to the idomlist */
	    if(add_it)
	    {
/*              DiabloPrint(stdout,"Adding for @B an idom @B\n",ibbl,iter->data);*/
	      LLAppend(NULL,LLCalloc(iter->data),&BBL_DOMINATED_BY(ibbl));
	    }
	  }
	  iter = LLNext(iter);
	}
#else
	InitNodeMarking();
	while(iter)
	  {
	    if (iter->data != ibbl && !NodeIsMarked(iter->data))
	      {
		data_ll_elem * iter2 = BBL_TMP_DOMINATED(iter->data);
		while(iter2)
		  {
		    if (iter2->data!=iter->data)
		      NodeMark(iter2->data);
		    iter2 = LLNext(iter2);
		  }
	      }
	    iter = LLNext(iter);
	  }

	iter = BBL_TMP_DOMINATED(ibbl);
	while (iter)
	  {
	    /* Add an entry to the idomlist */
	    if(!NodeIsMarked(iter->data) && iter->data!=ibbl)
	      {
		/*              DiabloPrint(stdout,"Adding for @B an idom @B\n",ibbl,iter->data);*/
		LLAppend(NULL,LLCalloc(iter->data),&BBL_DOMINATED_BY(ibbl));
	      }
	    iter = LLNext(iter);
	  }
#endif
	
	/*        DiabloPrint(stdout,"Idom for @B are:\n",ibbl);*/
	/*        DominatorPrint(BBL_DOMINATED_BY(ibbl));*/
	/*        printf("\n");*/
      }
    }
  }
  VERBOSE(0,("End part 2"));

#ifdef OPT_DOMINATES2
  CFG_FOREACH_BBL(cfg,ibbl)
    BBL_ATTRIB(ibbl) &= ~BBL_CAN_BE_DOMINATOR;
  
  CFG_FOREACH_BBL(cfg,ibbl)
    {
      data_ll_elem * iter = BBL_DOMINATED_BY(ibbl);
      while (iter)
	{
	  BBL_ATTRIB(iter->data) |= BBL_CAN_BE_DOMINATOR;
	  iter = LLNext(iter);
	}
    }
#endif

#if 0
  fp = fopen("log2_new","w");
  
  CFG_FOREACH_BBL(cfg,ibbl)
    {
      data_ll_elem * iter = BBL_DOMINATED_BY(ibbl);
      while (iter)
	{
	  DiabloPrint(fp,"@B dominated by @B\n",ibbl, (t_bbl*) iter->data);
	  iter = LLNext(iter);
	}
    }

  fclose(fp);
  exit(0);
#endif

  /* Exhaustive test for dominators */
#if 0
  {
    t_uint32 modderke = 0;
    CFG_FOREACH_FUN(cfg,ifun)
    {
      FUN_FOREACH_BBL(ifun,ibbl)
      {
	t_function * jfun;
	t_bbl * jbbl;
	CFG_FOREACH_FUN(cfg,jfun)
	  FUN_FOREACH_BBL(jfun,jbbl)
	  {
	    t_bool dom1;
	    t_bool dom2;
	    if(!(modderke++ %16))
	    {
	      dom1 = BblDominates(ibbl,jbbl);
	      dom2 = BblDominates_old(ibbl,jbbl);
	      /*          DominatorPrint(BBL_TMP_DOMINATED(jbbl));*/
	      if(dom1 != dom2)
	      {
		DiabloPrint(stdout,"@B, @B : %s , %s\n----------------------------------------\n",ibbl,jbbl,dom1?" TRUE":"FALSE",dom2?" TRUE":"FALSE");
		FATAL(("Differs for @B and @B",ibbl,jbbl));
	      }
	    }
	  }
      }
    }
  }
#endif


  CFG_FOREACH_FUN(cfg,ifun)
  {
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      DominatorFree(&BBL_TMP_DOMINATED(ibbl));
    }
  }

  /*}}}*/
  ANALYSIS_SIMPLE_END("Regular dominator information");
  


  ANALYSIS_START("Simple Loop detection");
  /*{{{*/
  if (CfgDetectSimpleInfiniteLoops(cfg,&looplist))
  {
    data_ll_elem * inf_iter = NULL;
    t_cfg_edge * new_edge = NULL;
    t_uint32 inf_loops = 0;
    while(looplist)
    {
      inf_iter = (data_ll_elem*)looplist;
      new_edge = CfgEdgeNew(cfg,CFG_EDGE_TAIL(inf_iter->data),cfg->unique_exit_node,ET_POSTDOM);
      looplist = LLNext(looplist);
      Free(inf_iter);
      inf_loops++;
    }
    VERBOSE(0,("%d infinite loops",inf_loops));
  }
  CFG_FOREACH_FUN(cfg,ifun)
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      if(!BBL_SUCC_FIRST(ibbl))
      {
	/* This bbl has no edges out, insert a special edge to the unique exit */
	CfgEdgeNew(cfg,ibbl,cfg->unique_exit_node,ET_POSTDOM);
      }
    }
  /*}}}*/
  ANALYSIS_SIMPLE_END("Simple Loop detection");

#if 0 /* FIXME: Temporarily disabled because it currently is not adjusted to the new dominator calculation yet */
  ANALYSIS_START("Post-dominator information");
  /*{{{ iterate until fixpoint, postdominators */

  BBL_FOREACH_PRED_EDGE(cfg->unique_exit_node,i_edge)
  {
    tmp_dll = (data_dll_elem *)Calloc(1,sizeof(data_dll_elem));
    tmp_dll->data = (void *)CFG_EDGE_HEAD(i_edge);
    DLLAppend(NULL,tmp_dll,&dom_queue.start,&dom_queue.end);
  }
  
#ifdef DEBUG_DOMINATOR
  DiabloPrint(stdout,"%x is de unique exit\n",cfg->unique_exit_node);
  DiabloPrint(stdout,"%x is ALL_BBLS\n",ALL_BBLS);
#endif

  while(dom_queue.end)
  {
    elem = (data_dll_elem*)dom_queue.end;
    DLLUnlink(dom_queue.end,&dom_queue.start,&dom_queue.end);

/*    VERBOSE(0,("Visit @B",elem->data));*/
    if(PostDominatorCallback(elem->data,mask_postdominator))
    {
      BBL_FOREACH_PRED_EDGE(elem->data,i_edge)
      {
	if(!QueueContains(dom_queue,elem->data))
	{
	  tmp_dll = (data_dll_elem *)Calloc(1,sizeof(data_dll_elem));
	  tmp_dll->data = (void *)CFG_EDGE_HEAD(i_edge);
	  DLLAppend(NULL,tmp_dll,&dom_queue.start,&dom_queue.end);
	}
      }
    }
    Free(elem);
  }
 

#if 1
  CFG_FOREACH_FUN(cfg,ifun)
  {
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      t_cfg_edge * iedge;
      t_cfg_edge * call_edge = NULL, *link_edge = NULL;
      BBL_FOREACH_SUCC_EDGE(ibbl,iedge)
      {
	if(EDGE_CAT(iedge) == ET_CALL || EDGE_CAT(iedge) == ET_IPJUMP || EDGE_CAT(iedge) == ET_IPFALLTHRU || EDGE_CAT(iedge) == ET_IPSWITCH || EDGE_CAT(iedge) == ET_IPUNKNOWN) call_edge = iedge;
	else if (EDGE_CAT(iedge) == ET_FUNLINK || EDGE_CAT(iedge) == ET_IPFUNLINK)
	{
	  link_edge = iedge;
	}
	else
	{
	  call_edge = NULL;
	  link_edge = NULL;
	  break;
	}
      }
      if(call_edge)
      {
	BBL_PDOMINATED_BY(ibbl) = LLCalloc(CFG_EDGE_TAIL(call_edge));
	if(EDGE_CORR(call_edge))
	  LLAppend(NULL,LLCalloc(CFG_EDGE_TAIL(EDGE_CORR(call_edge))),&BBL_PDOMINATED_BY(ibbl));
	else
	{
	  DiabloPrint(stdout,"@E has no corresponding!\n",call_edge);
	  WARNING(("No corresponding edge!"));
	}
      }
    }
  }
  CFG_FOREACH_FUN(cfg,ifun)
  {
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      {
	data_ll_elem * iter = BBL_TMP_PDOMINATED(ibbl);
	t_bbl * idom = (t_bbl*)iter->data;
#ifdef DEBUG_DOMINATOR
	DiabloPrint(stdout,"Now doing ");BblPrint(ibbl,0);
#endif
	if(idom == ALL_BBLS)
	{
#ifdef DEBUG_DOMINATOR
	  BblPrint(ibbl,PRINT_INSTRUCTIONS);DiabloPrint(stdout," still pdominated by ALL_BBLS\n");
#endif
	  BBL_PDOMINATED_BY(ibbl) = NULL;
	  continue;
	}
	if(idom == ibbl)
	{
	  iter = LLNext(iter);
	  if(iter) idom = (t_bbl*)iter->data;
	  else
	  {
	    if(ibbl != cfg->unique_exit_node)
	      FATAL(("Post-dominator troubles\n"));
	    BBL_PDOMINATED_BY(ibbl) = NULL;
	    continue;
	  }
	}
	iter = LLNext(iter);
	while(iter)
	{
	  if (iter->data != ibbl)
	  {
	    if (DominatorContains(BBL_TMP_PDOMINATED(iter->data),idom)) idom = iter->data;
/*            else if (!DominatorContains(BBL_TMP_PDOMINATED(idom),BBL_TMP_PDOMINATED(iter->data))) FATAL(("Post-postdominator problem\n"));*/
	  }
	  iter = LLNext(iter);
	}
	BBL_PDOMINATED_BY(ibbl) = LLCalloc(idom);
      }
      if(BBL_PDOMINATED_BY(ibbl))
      {
/*        DominatorPrint(BBL_PDOMINATED_BY(ibbl));*/
      }
      else
	DiabloPrint(stdout,"@B not idom'd by anything!\n",ibbl);
    }
  }
#endif




#if 0

  CFG_FOREACH_FUN(cfg,ifun)
  {
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      DominatorFree(&BBL_TMP_PDOMINATED(ibbl));
    }
  }

  CFG_FOREACH_FUN(cfg,ifun)
  {
    if(ifun != cfg->wrap_function)
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      if(BBL_PDOMINATED_BY(ibbl))
      {
	if(BBL_PDOMINATED_BY(ibbl)!=ALL_BBLS)
	{
#ifndef DEBUG_DOMINATOR_LEAKS
	append = LLCalloc((void*)ibbl);
#else
	append = Calloc(1,sizeof(data_ll_elem));
	append->data = (void*)ibbl;
#endif
	/* insert it, so they are sorted by incrementing pointer-value */
	LLAppendSorted(append,&BBL_TMP_PDOMINATED(BBL_PDOMINATED_BY(ibbl)));
#ifdef DEBUG_DOMINATOR
        BblPrint(ibbl,PRINT_PDOMINATED_BY);
#endif
	}
	else
	{
	  BBL_PDOMINATED_BY(ibbl) = NULL;
	  VERBOSE(1,("Strange, still pdominated by all!! @iB",ibbl));
	}
      }
      else if(ibbl != cfg->unique_exit_node && !BBL_PDOMINATED_BY(ibbl))
      {
#ifdef DEBUG_DOMINATOR
	BblPrint(ibbl,0);
#endif
	FATAL(("@B not post-dominated by anything?",ibbl));
      }
    }
  }
#endif
  
  /* }}}*/
  ANALYSIS_SIMPLE_END("Post-dominator information");
#endif

  /* remove the edges to the unique exit again, because they could interfere
   * with other analysis and optimizations */
  BBL_FOREACH_PRED_EDGE_SAFE(cfg->unique_exit_node,i_edge,j_edge)
    if(EDGE_CAT(i_edge) == ET_POSTDOM)
      CfgEdgeKill(i_edge);

  LLFinalise();
  ANALYSIS_SIMPLE_END("Dominator calculation");

  dominator_info_correct = TRUE;

  return;
}


/*! \param Two nodes of which the second is unused, a extra parameter
 *  \return Bool
 *
 * This function calculates :
 * Dom(first) = UNION(first, INTERSECTION(FORALL(Dom(predecessor))))
 *
 * and always returns false, in order not to stop the DFTraversal
 * It also increments 'changed', if differences are found.
 */
t_bool DominatorCallback(t_node * first, t_uint32 mask)
{
  data_ll_elem * orig = BBL_TMP_DOMINATED(first);
  data_ll_elem * join, *selftmp;
  t_cfg_edge * edge;
  t_bool ret = FALSE;

  if (((t_bbl*)first)->ins_first && INS_TYPE(((t_bbl*)first)->ins_first) == IT_DATA)
    FATAL(("Data in dominators?! @iB",first));
/*  join = LLCalloc((void*)ALL_BBLS);*/

  join = orig;

  if(join->data != ALL_BBLS)
    DominatorRemove(&join,(t_bbl*)first);

  BBL_FOREACH_PRED_EDGE(first,edge)
  {

    if (
#ifndef OLD_COMPUTATIONS
	EdgeIsMarked(edge) &&
#endif
	(EDGE_CAT(edge) & mask) &&

	CFG_EDGE_HEAD(edge)!=T_BBL(first)
	)
      {
#ifndef OLD_COMPUTATIONS
	EdgeUnmark(edge);
#endif
	if(BBL_TMP_DOMINATED(EDGE_HEAD(edge)) == NULL)
	  {
	    BblPrint(CFG_EDGE_HEAD(edge),PRINT_PREDECESSOR|PRINT_DOM_INFO);
	    FATAL(("Empty dominatorset in predecessor: @iB, edge @E",EDGE_HEAD(edge),edge));
	  }
	if(!(EDGE_CAT(edge) & (ET_RETURN | ET_COMPENSATING)))
	  {
	    if(DominatorIntersect(&join,BBL_TMP_DOMINATED(EDGE_HEAD(edge)))) ret = TRUE;
	  }
	else
	  {
	    data_ll_elem * tmp_union;
	    t_bool neglect;
	    if(!EDGE_CORR(edge))
	      {
		WARNING(("NO corresponding edge: @E\n  @eB",edge,first));
		if(DominatorIntersect(&join,BBL_TMP_DOMINATED(EDGE_HEAD(edge)))) ret = TRUE;
	      }
	    else
	      {
		tmp_union = DominatorUnion(BBL_TMP_DOMINATED(CFG_EDGE_HEAD(edge)),BBL_TMP_DOMINATED(CFG_EDGE_HEAD(EDGE_CORR(edge))),&neglect);
		if(DominatorIntersect(&join,tmp_union)) ret = TRUE;
		DominatorFree(&tmp_union);
	      }
	  }
      }
  }

  /* Add myself to the dominator set */
  if(join->data != ALL_BBLS)
  {
    selftmp = LLCalloc((void*)first);
    LLAppendSorted(selftmp,(void*)&join);
  }

/*  if (DominatorDiffers(orig,join))*/
/*    ret = TRUE;*/

/*  DiabloPrint(stdout,"@B\nORIG\n",first);*/
/*  DominatorPrint(orig);*/
/*  DiabloPrint(stdout,"JOIN\n");*/
/*  DominatorPrint(join);*/
/*  printf("-------------------\n");*/

  BBL_TMP_DOMINATED(first) = join;


/*  DominatorFree(&orig);*/
  return ret;
}


/*! \param A pointer to data_ll_elem pointer and a data_ll_elem pointer, both containing dominator info
 *  \return t_bool
 *
 * Returns the intersection of the two list in 'first', frees what was in first before the call. Returns
 * TRUE when first changes */ 
t_bool DominatorIntersect(data_ll_elem ** first, data_ll_elem * second)/*{{{*/
{
  data_ll_elem *iter,*iter2,*tmp;
  t_bool ret = FALSE;

  if (!(*first)) FATAL(("Empty first dominatorset!\n"));
  if (!(second)) FATAL(("Empty second dominatorset!\n"));

  /* first the quick solution : first remains unchanged */
  if (second->data == ALL_BBLS)
  {
    return FALSE;
  }
  /* now the other simple solution : first becomes second. No need to free anything, the only
   * element that should be freed is reused. */
  if ((*first)->data == ALL_BBLS)
  {
    iter2 = (*first);
    iter2->data = second->data;
    iter = LLNext(second);
    while (iter)
    {
#ifndef DEBUG_DOMINATOR_LEAKS
      LLNext(iter2) = LLCalloc((void*)iter->data);
#else
      LLNext(iter2) = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
      ((data_ll_elem*)LLNext(iter2))->data = (void*)iter->data;
#endif
      iter = LLNext(iter);
      iter2 = LLNext(iter2);
    }

    return TRUE;
  }

#ifdef DEBUG_DOMINATOR
  DiabloPrint(stdout,"Verzameling 1:\n");
  for( iter = *first; iter; iter = LLNext(iter))
    if (iter->data!= ALL_BBLS)
    DiabloPrint(stdout,"%x ;",iter->data);
    else DiabloPrint(stdout,"ALL ;");
  DiabloPrint(stdout,"\nVerzameling 2:\n");
  for( iter = second; iter; iter = LLNext(iter))
    if (iter->data!= ALL_BBLS)
    DiabloPrint(stdout,"%x ;",iter->data);
    else DiabloPrint(stdout,"ALL ;");
  DiabloPrint(stdout,"\n\n");
  DiabloPrint(stdout,"\n");
#endif

  /* general case : real intersect */
  iter = (*first);
  iter2 = second;
  
  tmp = NULL;

  while(iter && iter2)
  {
#ifdef DEBUG_DOMINATOR
    if(iter->data == ALL_BBLS || iter2->data == ALL_BBLS) FATAL(("All bbls!"));
#endif
    if (iter->data < iter2->data)
    {
      iter = LLNext(iter);
      if(tmp)
      {
#ifndef DEBUG_DOMINATOR_LEAKS
	LLFree(LLNext(tmp));
#else
	Free(LLNext(tmp));
#endif
	LLNext(tmp) = iter;
      }
      else
      {
#ifndef DEBUG_DOMINATOR_LEAKS
	LLFree(*first);
#else
	Free(*first);
#endif
	(*first) = iter;
      }
      ret = TRUE;
    }
    else if (iter2->data < iter->data)
      {
	iter2 = LLNext(iter2);
      }
    else { 
      tmp = iter; 
      iter = LLNext(iter); 
      iter2 = LLNext(iter2); 
    }
  }

  while(iter)
  {
    iter2 = iter;
    iter = LLNext(iter);
    LLFree(iter2);
    ret = TRUE;
  }
  if(tmp) LLNext(tmp) = NULL;
    
#ifdef DEBUG_DOMINATOR
  DiabloPrint(stdout,"Verzameling 1:\n");
  for( iter = *first; iter; iter = LLNext(iter))
    if (iter->data!= ALL_BBLS)
    DiabloPrint(stdout,"%x ;",iter->data);
    else DiabloPrint(stdout,"ALL ;");
  DiabloPrint(stdout,"\n==========================================\n");
  DiabloPrint(stdout,"\n");
#endif
  if (!*first) FATAL(("Empty first dominatorset!\n"));
  return ret;
}/*}}}*/

data_ll_elem * DominatorUnion(data_ll_elem * first, data_ll_elem * second, t_bool * changes)/*{{{*/
{
  data_ll_elem *iter,*iter2,*tmp = NULL, *ret = NULL;

  if (!first) FATAL(("Empty first dominatorset!\n"));
  if (!second) FATAL(("Empty second dominatorset!\n"));
  
  if (first->data == ALL_BBLS)
  {
    VERBOSE(1,("1: Union of ALL_BBLS!"));
#ifndef DEBUG_DOMINATOR_LEAKS
    return LLCalloc((void*)ALL_BBLS);
#else
    ret = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
    ret->data = (void*)ALL_BBLS;
    return ret;
#endif
  }
  if(second->data == ALL_BBLS)
  {
    VERBOSE(1,("2: Union of ALL_BBLS!"));
    *changes = TRUE;
#ifndef DEBUG_DOMINATOR_LEAKS
    return LLCalloc((void*)ALL_BBLS);
#else
    ret = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
    ret->data = (void*)ALL_BBLS;
    return ret;
#endif
  }

  iter = first;
  iter2 = second;
  while(iter && iter2)
  {
    if(iter->data < iter2->data)
    {
#ifndef DEBUG_DOMINATOR_LEAKS
      tmp = LLCalloc(iter->data);
#else
    tmp = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
    tmp->data = (void*)iter->data;
#endif
    LLAppendSortedCached(tmp,&ret);
      iter = LLNext(iter);
    }
    else if(iter->data > iter2->data)
    {
#ifndef DEBUG_DOMINATOR_LEAKS
      tmp = LLCalloc(iter2->data);
#else
    tmp = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
    tmp->data = (void*)iter2->data;
#endif
      LLAppendSortedCached(tmp,&ret);
      iter2 = LLNext(iter2);
      *changes = TRUE;
    }
    else
    {
#ifndef DEBUG_DOMINATOR_LEAKS
      tmp = LLCalloc(iter->data);
#else
    tmp = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
    tmp->data = (void*)iter->data;
#endif
      LLAppendSortedCached(tmp,&ret);
      iter = LLNext(iter);
      iter2  = LLNext(iter2);
    }
  }
  while(iter)
  {
#ifndef DEBUG_DOMINATOR_LEAKS
    tmp = LLCalloc(iter->data);
#else
    tmp = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
    tmp->data = (void*)iter->data;
#endif
    LLAppendSortedCached(tmp,&ret);
    iter = LLNext(iter);
  }
  while(iter2)
  {
#ifndef DEBUG_DOMINATOR_LEAKS
    tmp = LLCalloc(iter2->data);
#else
    tmp = (data_ll_elem*)Calloc(1,sizeof(data_ll_elem));
    tmp->data = (void*)iter2->data;
#endif
    LLAppendSortedCached(tmp,&ret);
    iter2 = LLNext(iter2);
    *changes = TRUE;
  }
  return ret;
}/*}}}*/

t_bool DominatorDiffers(data_ll_elem * first, data_ll_elem * second)
{
  data_ll_elem *iter,*iter2;

  if (!first) FATAL(("Empty first dominatorset!\n"));
  if (!second) FATAL(("Empty second dominatorset!\n"));
  
  if (first->data == ALL_BBLS && second->data == ALL_BBLS) return FALSE;

  iter = first;
  iter2 = second;
  while (iter && iter2)
  {
    if (iter->data != iter2->data) return TRUE;
    else
    {
      iter = LLNext(iter);
      iter2 = LLNext(iter2);
    }
  }

  if (iter || iter2) return TRUE;
  return FALSE;
}

/*! \param Two bbl's
 *  \ret t_bool
 *
 *  Return TRUE if dominator dominates dominated, otherwise FALSE */
t_bool BblDominates_old(t_bbl * dominator, t_bbl * dominated)
{
  data_ll_elem *iter = BBL_TMP_DOMINATED(dominated);
  
  if(!dominated || !dominator) return FALSE;
#ifdef MIPS_SUPPORT
  if(dominated==NULL || BBL_CFG(dominated)==NULL || BBL_CFG(dominated)->unique_entry_node==NULL)return FALSE;
#endif

  if (dominator == dominated) return TRUE;

  iter = BBL_TMP_DOMINATED(dominated);
  if (!iter) FATAL(("Empty dominatorset! @B\n",dominated));

  if (iter->data == ALL_BBLS) return FALSE;

  while (iter )/*&& iter->data <= second)*/
  {
    if (iter->data == ALL_BBLS) FATAL(("Corrupt dominators"));
    if (iter->data == dominator) return TRUE;
    iter = LLNext(iter);
  }
  return FALSE;
}

t_bool BblDominates(t_bbl * dominator, t_bbl * dominated)
{
  t_queue dom_queue = {NULL,NULL};
  t_bbl * entry_node = BBL_CFG(dominated)->unique_entry_node;
  data_dll_elem * elem, * tmp_dll;
  static t_bbl * prev_or = NULL, *prev_ed = NULL;
  static t_bool prev_result = FALSE;
  data_ll_elem * iter;
#ifdef OPT_DOMINATES
  data_ll_elem * iter2;
  int teller;
#endif
  if (dominator == prev_or && dominated == prev_ed) return prev_result;
  
#ifdef OPT_DOMINATES2
  if (!(BBL_ATTRIB(dominator) & BBL_CAN_BE_DOMINATOR))
    if (dominator!=dominated) return FALSE;
#endif
  
  if(dominator == dominated)
  {
    prev_or = dominator;
    prev_ed = dominated;
    prev_result = TRUE;
    return prev_result;
  }

  iter = BBL_DOMINATED_BY(dominated);

#ifdef OPT_DOMINATES
 retry_first:
  iter2=iter;
  teller = 0;
#endif

  while(iter)
  {
    if(iter->data == dominator)
    {
      prev_or = dominator;
      prev_ed = dominated;
      prev_result = TRUE;
      return prev_result;
    }
#ifdef OPT_DOMINATES
    teller++;
#endif
    iter = LLNext(iter);
  }

#ifdef OPT_DOMINATES
  if (teller==1 && iter2->data != entry_node)
    {
      NodeMarkDominator(iter2->data);
      dominated = iter2->data;
      iter = BBL_DOMINATED_BY(dominated);
      goto retry_first;
    }
#endif

  InitNodeMarkingDominator();

  iter = BBL_DOMINATED_BY(dominated);

  while(iter)
  {
    if(iter->data != entry_node)
    {
      NodeMarkDominator(iter->data);
      tmp_dll = (data_dll_elem *)Calloc(1,sizeof(data_dll_elem));
      tmp_dll->data = (void *)iter->data;
      DLLAppend(NULL,tmp_dll,&dom_queue.start,&dom_queue.end);
      /*DiabloPrint(stdout,"@G(@G), ",BBL_OLD_ADDRESS(iter->data),BBL_FUN(iter->data)?BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(iter->data))):0x0); */
    }
    iter = LLNext(iter);
  }

  while(dom_queue.start)
  {
    elem = (data_dll_elem*)dom_queue.start;
    DLLUnlink(dom_queue.start,&dom_queue.start,&dom_queue.end);
    iter = BBL_DOMINATED_BY(elem->data);

#ifdef OPT_DOMINATES
  retry:
    
    iter2 = iter;
    teller = 0;
#endif
    while(iter)
    {
#ifdef OPT_DOMINATES
      teller++;
#endif
      if(iter->data == dominator)
      {
#ifdef OPT_DOMINATES
	if (elem!=NULL)
	  {
	    Free(elem);
	    elem=NULL;
	  }
#else
	Free(elem);
#endif
	goto cleanup_true;
      }
      iter = LLNext(iter);
    }

#ifdef OPT_DOMINATES
    if (teller==1 && iter2->data != entry_node)
      {
	NodeMarkDominator(iter2->data);
	iter = BBL_DOMINATED_BY(iter2->data);
	if (elem!=NULL)
	  {
	    Free(elem);
	    elem = NULL;
	  }
	goto retry;
      }
#endif

#ifdef OPT_DOMINATES
    iter = BBL_DOMINATED_BY(iter2->data);
#else
    iter = BBL_DOMINATED_BY(elem->data);
#endif
    while(iter)
    {
      if(iter->data != entry_node)
      {
	NodeMarkDominator(iter->data);
	tmp_dll = (data_dll_elem *)Calloc(1,sizeof(data_dll_elem));
	tmp_dll->data = (void *)iter->data;
	DLLAppend(NULL,tmp_dll,&dom_queue.start,&dom_queue.end);
	/*DiabloPrint(stdout,"@G(@G), ",BBL_OLD_ADDRESS(iter->data),BBL_FUN(iter->data)?BBL_OLD_ADDRESS(FUN_BBL_FIRST(BBL_FUN(iter->data))):0x0);*/
      }
      iter = LLNext(iter);
    }
#ifdef OPT_DOMINATES
    if (elem!=NULL)
      {
	Free(elem);
	elem = NULL;
      }
#else
    Free(elem);
#endif
  }

  prev_or = dominator;
  prev_ed = dominated;
  prev_result = FALSE;
/*  printf("\n");*/
  return prev_result;

cleanup_true:
  while(dom_queue.start)
  {
    elem = (data_dll_elem*)dom_queue.start;
    DLLUnlink(dom_queue.start,&dom_queue.start,&dom_queue.end);
    Free(elem);
  }
  prev_or = dominator;
  prev_ed = dominated;
  prev_result = TRUE;
  return prev_result;

}

/*! \param Two bbl's
 *  \ret t_bool
 *
 *  Return TRUE if postdominator postdominates postdominated, otherwise FALSE */
t_bool BblPostdominates(t_bbl * postdominator, t_bbl * postdominated)
{
  data_ll_elem *iter = BBL_TMP_PDOMINATED(postdominated);
  if(!postdominated || !postdominator) return FALSE;
#ifdef MIPS_SUPPORT
  if(postdominated==NULL || BBL_CFG(postdominated)==NULL || BBL_CFG(postdominated)->unique_entry_node==NULL)return FALSE;
#endif

  if (postdominator == postdominated) return TRUE;

  iter = BBL_TMP_PDOMINATED(postdominated);
  if (!iter) FATAL(("Empty postdominatorset!\n"));

  if (iter->data == ALL_BBLS) return FALSE;

  while (iter )/*&& iter->data <= second)*/
  {
    if (iter->data == ALL_BBLS) FATAL(("Corrupt postdominators"));
    if (iter->data == postdominator) return TRUE;
    iter = LLNext(iter);
  }
  return FALSE;
}


/*! Remove a bbl from a linked list of dominated bbl's
 *
 * \todo Should this function go elsewhere? */
void DominatorRemove(data_ll_elem ** list, t_bbl * bbl)
{
  data_ll_elem * iter, *prev;
  iter = *list;
  prev = NULL;
  
  if (*list && (*list)->data == ALL_BBLS) FATAL(("Immediately dominates ALL!"));
  
  while(iter)
  {
    if (iter->data == (void*)bbl)
      {
      if (prev)
	LLNext(prev) = LLNext(iter);
      else
	{
	  *list = LLNext(iter);
	}

#ifndef DEBUG_DOMINATOR_LEAKS
      LLFree(iter);
#else
      Free(iter);
#endif
      return;
    }
    prev = iter;
    iter = LLNext(iter);
  }
}


/*{{{ Graph export functions*/
static t_bool interproc = FALSE;
 
static t_bool node_writeout_dominator(t_node * current, t_node * parent,void * data) {
  /* print out dominator of current node */
  /* used for exporting the flowgraph */
  FILE * out=data;
/*  data_ll_elem * iter;*/
/*  t_bbl * i_bbl;*/

  /* Write out node name of current{{{*/
  DiabloPrint(out, "\t\"%p\" [label=\"",current);
  if (T_BBL(current) ==  T_BBL(current)->cfg->hell_node)
  {
    DiabloPrint(out,"HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->call_hell_node) 
  {
    DiabloPrint(out, "CALL HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->exit_call_hell_node)
  {
    DiabloPrint(out, "EXIT CALL HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->exit_hell_node)
  {
    DiabloPrint(out, "EXIT HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->unique_entry_node)
  {
    DiabloPrint(out, "UNIQUE ENTRY");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->unique_exit_node)
  {
    DiabloPrint(out, "UNIQUE EXIT");
  }
  else if (BBL_IS_LAST(T_BBL(current)) && !BBL_NINS(T_BBL(current)))
  { /* return block */
    DiabloPrint(out, "RETURN FROM F%p", T_BBL(current)->fun);
  } 
  else 
  {
    DiabloPrint(out, "@G\\n", BBL_OLD_ADDRESS(T_BBL(current)));
  }

  DiabloPrint(out,"\"]\n");/*}}}*/
 
  if (T_BBL(current) != T_BBL(current)->cfg->unique_entry_node)
  {
    data_ll_elem * iter = BBL_DOMINATED_BY(current);
    while(iter)
    {
      DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",iter->data,current,"solid",(BBL_FUN(iter->data) == BBL_FUN(current))?"black":"red");
      if(BBL_FUN(iter->data) != BBL_FUN(current) && iter->data != ALL_BBLS)
	DiabloPrint(out,"\t\"%p\" [label=\"@G: %s\",style=filled, color=yellow]\n",iter->data,BBL_OLD_ADDRESS(iter->data),FUN_NAME(BBL_FUN(iter->data)));
      iter = LLNext(iter);
    }
#if 0
    if(BBL_DOMINATED_BY(current))
    {
      DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",BBL_DOMINATED_BY(current)[0],current,
	"solid",
	(BBL_FUN(BBL_DOMINATED_BY(current)[0]) == BBL_FUN(current))?"black":"red");

    if (BBL_FUN(BBL_DOMINATED_BY(current)[0]) != BBL_FUN(current) && /*!interproc &&*/ BBL_DOMINATED_BY(current)[0] != ALL_BBLS)
      DiabloPrint(out,"\t\"%p\" [label=\"@G: %s\",style=filled, color=yellow]\n",
	  BBL_DOMINATED_BY(current)[0], BBL_OLD_ADDRESS(BBL_DOMINATED_BY(current)[0]),
	  FUN_NAME(BBL_FUN(BBL_DOMINATED_BY(current)[0])));
    }
    if(BBL_DOMINATED_BY(current)[1])
    {
      DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",BBL_DOMINATED_BY(current)[1],current,
	"solid",
	(BBL_FUN(BBL_DOMINATED_BY(current)[1]) == BBL_FUN(current))?"black":"red");

    if (BBL_FUN(BBL_DOMINATED_BY(current)[1]) != BBL_FUN(current) && /*!interproc &&*/ BBL_DOMINATED_BY(current)[1] != ALL_BBLS)
      DiabloPrint(out,"\t\"%p\" [label=\"@G: %s\",style=filled, color=yellow]\n",
	  BBL_DOMINATED_BY(current)[1], BBL_OLD_ADDRESS(BBL_DOMINATED_BY(current)[1]),
	  FUN_NAME(BBL_FUN(BBL_DOMINATED_BY(current)[1])));
    }
#endif
    /*BBL_FOREACH_DOMINATES(T_BBL(current),iter,i_bbl)
    {
      if(BBL_FUN(i_bbl) && BBL_FUN(i_bbl) == BBL_FUN(T_BBL(current)) && T_BBL(current) != i_bbl)
      DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",i_bbl,current,
	  "solid",
	  "black");

    }*/
  }

  return FALSE;
}

static t_bool node_writeout_postdominator(t_node * current, t_node * parent,void * data) {
  /* print out postdominator of current node */
  /* used for exportating the flowgraph */
  FILE * out=data;

  /* Write out node name of current{{{*/
  DiabloPrint(out, "\t\"%p\" [label=\"",current);
  if (T_BBL(current) ==  T_BBL(current)->cfg->hell_node)
  {
    DiabloPrint(out,"HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->call_hell_node) 
  {
    DiabloPrint(out, "CALL HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->exit_call_hell_node)
  {
    DiabloPrint(out, "EXIT CALL HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->exit_hell_node)
  {
    DiabloPrint(out, "EXIT HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->unique_entry_node)
  {
    DiabloPrint(out, "UNIQUE ENTRY");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->unique_exit_node)
  {
    DiabloPrint(out, "UNIQUE EXIT");
  }
  else if (BBL_IS_LAST(T_BBL(current)) && !BBL_NINS(T_BBL(current)))
  { /* return block */
    DiabloPrint(out, "RETURN FROM F%p", T_BBL(current)->fun);
  } 
  else 
  {
    DiabloPrint(out, "@G\\n", BBL_OLD_ADDRESS(T_BBL(current)));
  }

  DiabloPrint(out,"\"]\n");/*}}}*/
 
  if (T_BBL(current) != T_BBL(current)->cfg->unique_exit_node)
  {
    data_ll_elem * iter = BBL_PDOMINATED_BY(current);
    while(iter)
    {
      DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",iter->data,current,"solid",(BBL_FUN(iter->data) == BBL_FUN(current))?"black":"red");
      if(BBL_FUN(iter->data) != BBL_FUN(current) && iter->data != ALL_BBLS)
	DiabloPrint(out,"\t\"%p\" [label=\"@G: %s\",style=filled, color=yellow]\n",iter->data,BBL_OLD_ADDRESS(iter->data),FUN_NAME(BBL_FUN(iter->data)));
      iter = LLNext(iter);
    }
#if 0
    if(BBL_PDOMINATED_BY(current)[0])
    {
      DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",BBL_PDOMINATED_BY(current)[0],current,
	  "solid",
	  (BBL_FUN(BBL_PDOMINATED_BY(current)[0]) == BBL_FUN(current))?"black":"red");

      if (BBL_FUN(BBL_PDOMINATED_BY(current)[0]) != BBL_FUN(current) && /*!interproc &&*/ BBL_PDOMINATED_BY(current)[0] != ALL_BBLS)
	DiabloPrint(out,"\t\"%p\" [label=\"@G: %s\",style=filled, color=yellow]\n",
	    BBL_PDOMINATED_BY(current)[0], BBL_OLD_ADDRESS(BBL_PDOMINATED_BY(current)[0]),
	    FUN_NAME(BBL_FUN(BBL_PDOMINATED_BY(current)[0])));
    }
    if(BBL_PDOMINATED_BY(current)[1])
    {
      DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",BBL_PDOMINATED_BY(current)[1],current,
	  "solid",
	  (BBL_FUN(BBL_PDOMINATED_BY(current)[1]) == BBL_FUN(current))?"black":"red");

      if (BBL_FUN(BBL_PDOMINATED_BY(current)[1]) != BBL_FUN(current) && /*!interproc &&*/ BBL_PDOMINATED_BY(current)[1] != ALL_BBLS)
	DiabloPrint(out,"\t\"%p\" [label=\"@G: %s\",style=filled, color=yellow]\n",
	    BBL_PDOMINATED_BY(current)[1], BBL_OLD_ADDRESS(BBL_PDOMINATED_BY(current)[1]),
	    FUN_NAME(BBL_FUN(BBL_PDOMINATED_BY(current)[1])));
    }
#endif
    /*BBL_FOREACH_PDOMINATES(T_BBL(current),iter,i_bbl)
    {
      
      if(BBL_FUN(i_bbl) && BBL_FUN(i_bbl) == BBL_FUN(T_BBL(current)) && (T_BBL(current) != i_bbl))
	DiabloPrint(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",i_bbl,current,
	  "solid",
	  "black");

    }*/
  }

  return FALSE;
}

void Export_FunctionDominator(t_bbl * start, t_uint32 mask, const char * filename)
{
  FILE * out = fopen(filename,"w");
  if (mask & ET_INTERPROC) interproc = TRUE;
  else interproc = FALSE;
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));
  DiabloPrint(out, "digraph \"@G\" {\n\tnode [shape=ellipse]\n",BBL_OLD_ADDRESS(start));
  GraphDFTraversalWithCheckAndData((t_node *) start,node_writeout_dominator, mask,out);
  if(interproc) DiabloPrint(out, "\t\"%p\" [label=\"ALL BBLS\"]\n",ALL_BBLS);
  DiabloPrint(out, "}\n");
  fclose(out);
}

void Export_FunctionPostDominator(t_bbl * start, t_uint32 mask, const char * filename)
{
  FILE * out = fopen(filename,"w");
  if (mask & ET_INTERPROC) interproc = TRUE;
  else interproc = FALSE;
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));
  DiabloPrint(out, "digraph \"@G\" {\n\tnode [shape=ellipse]\n",BBL_OLD_ADDRESS(start));
  GraphDFTraversalWithCheckAndData((t_node *) start,node_writeout_postdominator, mask,out);
  if(interproc) DiabloPrint(out, "\t\"%p\" [label=\"ALL BBLS\"]\n",ALL_BBLS);
  DiabloPrint(out, "}\n");
  fclose(out);
}
/*}}}*/

t_bool PostDominatorCallback(t_node * first, t_uint32 mask)
{
  data_ll_elem * orig = BBL_TMP_PDOMINATED(first);
  data_ll_elem * join, *selftmp;
  t_cfg_edge * edge;
  t_bool ret = FALSE;

  if (BBL_CFG(first)->unique_exit_node == (t_bbl*)first) return FALSE;
  
  join = LLCalloc((void*)ALL_BBLS);

  BBL_FOREACH_SUCC_EDGE(first,edge)
  {
    if (EDGE_CAT(edge) & mask)
    {
      if(!(EDGE_CAT(edge) & (ET_CALL | ET_IPFALLTHRU | ET_IPJUMP | ET_IPUNKNOWN)))
	DominatorIntersect(&join,BBL_TMP_PDOMINATED(EDGE_TAIL(edge)));
      else
      {
	data_ll_elem * tmp_union;
	t_bool neglect;
	if(!EDGE_CORR(edge)) FATAL(("NO corresponding edge"));
	tmp_union = DominatorUnion(BBL_TMP_PDOMINATED(CFG_EDGE_TAIL(edge)),BBL_TMP_PDOMINATED(CFG_EDGE_TAIL(EDGE_CORR(edge))),&neglect);
	DominatorIntersect(&join,tmp_union);
	DominatorFree(&tmp_union);
      }
    }
  }
  if (join->data != ALL_BBLS)
  {
    selftmp = LLCalloc((void*)first);
    LLAppendSorted(selftmp,(void *)&join);
  }
  BBL_TMP_PDOMINATED(first) = join;
  /*else
    FATAL(("Joining problem with @B", first));*/
  
  if (DominatorDiffers(orig,join))
    ret = TRUE;

/*  DiabloPrint(stdout,"@B\nORIG\n",first);
  DominatorPrint(orig);
  DiabloPrint(stdout,"JOIN\n");
  DominatorPrint(join);
  printf("-------------------\n");*/

  DominatorFree(&orig);
  return ret;
}


void DominatorCleanup(t_cfg * cfg)
{
  t_bbl * ibbl;
  t_function * ifun;

  if (dominator_info_correct)
    return;

  CFG_FOREACH_FUN(cfg, ifun)
  {
    FUN_FOREACH_BBL(ifun,ibbl)
    {
      DominatorFree(&BBL_TMP_DOMINATED(ibbl));
      DominatorFree(&BBL_TMP_PDOMINATED(ibbl));
      DominatorFree(&BBL_DOMINATED_BY(ibbl));
      DominatorFree(&BBL_PDOMINATED_BY(ibbl));
    }
  }
  
  LLFinalise();
}

void DominatorFree(data_ll_elem ** list)
{
  data_ll_elem * free;
  
  while(*list)
  {
    free = *list;
    *list = LLNext(*list);
#ifndef DEBUG_DOMINATOR_LEAKS
    LLFree(free);
#else
    Free(free);
#endif
  }
  if(*list) FATAL(("Set me to NULL"));
}

/*! \param A linked list, containing t_bbl pointers, and a bbl 
 *  \return Bool
 *
 *  This function returns whether the bbl is contained in the list */

t_bool DominatorContains(data_ll_elem * list, t_bbl * find_me)
{
  if (!list) FATAL(("Empty dominatorset!\n"));
  if (!find_me) FATAL(("Empty query!\n"));

  if (list->data == ALL_BBLS) return TRUE;

  while (list && list->data < (void*)find_me)
  {
    list = LLNext(list);
  }
  if (list && list->data == find_me) return TRUE;
  
  return FALSE;
}

t_bool QueueContains(t_queue queue,void * data)
{
  data_dll_elem * iter = (data_dll_elem *)queue.start;

  while(iter)
  {
    if(iter->data == data) return TRUE;
    iter = LLNext(iter);
  }
  return FALSE;
}

void InitNodeMarkingDominator(void)
{
  if (++dominator_node_marking_number==0x7fffffff)
    {
      FATAL(("Marking number overflow\n"));
    }
}

/* vim: set shiftwidth=2 foldmethod=marker: */
