/******************************************************************************
 * Copyright 2001,2002, 2003: {{{ 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/
/* Includes {{{ */
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_print.h"
#include "diablo_iterators.h"
#include "../utils/util_error.h"
/* }}} */
/* Query functions {{{ */

extern t_bool dominator_info_correct;

void CfgEdgeMoveToLastSucc(t_cfg_edge * fall)
{
  while (CFG_EDGE_SUCC_NEXT(fall))
    {
      t_cfg_edge * nextnext = NULL;
      t_cfg_edge * prev = CFG_EDGE_SUCC_PREV(fall);
      t_cfg_edge * next = CFG_EDGE_SUCC_NEXT(fall);
      
      if (next)
	nextnext = CFG_EDGE_SUCC_NEXT(next);
      
      CFG_EDGE_SUCC_PREV(next)=prev;
      if (prev)
	CFG_EDGE_SUCC_NEXT(prev)=next;
      else
	BBL_SUCC_FIRST(CFG_EDGE_HEAD(fall))=next;


      CFG_EDGE_SUCC_NEXT(next)=fall;
      CFG_EDGE_SUCC_PREV(fall)=next;
      
      if (nextnext)
	CFG_EDGE_SUCC_PREV(nextnext)=fall;
      else
	BBL_SUCC_LAST(CFG_EDGE_HEAD(fall))=fall;

     
      CFG_EDGE_SUCC_NEXT(fall)=nextnext;
    }
}

t_cfg_edge * CfgEdgeGetFunlinkForCall(t_cfg_edge * call)
{
  t_bbl * call_bbl=CFG_EDGE_HEAD(call);
  t_cfg_edge * it;

  VERIFY_FASSERT(EDGE_CAT(call) == ET_CALL,("Expected call edge, not @E",call));

  BBL_FOREACH_SUCC_EDGE(call_bbl,it)
  {
    if ((EDGE_CAT(it)==ET_FUNLINK) || (EDGE_CAT(it)==ET_IPFUNLINK))
      return it;
  }
  return NULL; /* Keep the compiler happy */
}

t_cfg_edge * CfgEdgeGetReturnForFunlink(t_cfg_edge * funlink)
{
  t_bbl * ret_bbl=CFG_EDGE_TAIL(funlink);
  t_cfg_edge * it;

  VERIFY_FASSERT(EDGE_CAT(funlink) == ET_FUNLINK || EDGE_CAT(funlink) == ET_IPFUNLINK,("Expected funlink edge, not @E",funlink));

  BBL_FOREACH_PRED_EDGE(ret_bbl,it)
  {
    if (EDGE_CAT(it)==ET_RETURN)
      return it;
  }
  return NULL; /* Keep the compiler happy */
}

t_cfg_edge * CfgEdgeGetReturnForCall(t_cfg_edge * call)
{
  return CfgEdgeGetReturnForFunlink(CfgEdgeGetFunlinkForCall(call));
}

/*}}}*/
/* Constructors {{{ */
t_cfg_edge * CfgEdgeNew(t_cfg * flowgraph, t_bbl * from, t_bbl * to,t_uint32 type)
{
  t_cfg_edge * ret;
  if (type==ET_RETURN && !IsHellNode(to)) 
    FATAL(("Cannot use CfgEdgeNew to add return edges! (Use CfgEdgeNewCall)"));
  if (type==ET_COMPENSATING && !IsHellNode(to)) 
    FATAL(("Cannot use CfgEdgeNew to add compensating edges! (Use CfgEdgeNewCompensating)"));
  if (type==ET_CALL && !IsHellNode(to)) 
    {
      FATAL(("Cannot use CfgEdgeNew to add call edges!"));
    }

  ret=CfgAppendNode(flowgraph, from, to,type);
  if (!flowgraph) FATAL(("No flowgraph"));
  ret->cfg=flowgraph;
  ret->refcount=1;

  dominator_info_correct = FALSE;

  return ret; 
}

t_cfg_edge * CfgEdgeNewCompensating(t_cfg * flowgraph, t_cfg_edge * ip)
{
  t_cfg_edge * ret=CfgAppendNode(flowgraph, BBL_FUN(CFG_EDGE_TAIL(ip))->bbl_last, BBL_FUN(CFG_EDGE_HEAD(ip))->bbl_last  ,ET_COMPENSATING);
  ret->corr_edge=ip;
  ip->corr_edge=ret;
  ret->cfg=flowgraph;
  ret->refcount=ip->refcount;
   dominator_info_correct = FALSE;
   if (!flowgraph) FATAL(("No flowgraph"));
  return ret;
}

t_cfg_edge * CfgEdgeNewCall(t_cfg * flowgraph, t_bbl * from, t_bbl * to, t_bbl * ret, t_bbl * exit)
{
  t_cfg_edge * call=CfgAppendNode(flowgraph, from, to ,ET_CALL);
  t_cfg_edge * retu=CfgAppendNode(flowgraph, (!exit)?flowgraph->exit_hell_node:exit , (!ret)?flowgraph->exit_hell_node:ret  ,ET_RETURN);
  call->corr_edge=retu;
  retu->corr_edge=call;
  call->cfg=flowgraph;
  retu->cfg=flowgraph;
  dominator_info_correct = FALSE;
  if (!flowgraph) FATAL(("No flowgraph"));
  return call;
}
/* }}} */
/* Destructor {{{ */
void CfgEdgeKill(t_cfg_edge *edge) 
{
  dominator_info_correct = FALSE;
   edge->refcount--;
   if (edge->refcount==0) 
   {
     if (edge->rel)
     {
       VERIFY_FASSERT(EDGE_CAT(edge) == ET_SWITCH, ("we only expect corresponding relocs with switch edges"));
       RelocTableRemoveReloc(CFG_EDGE_CFG(edge)->sec->obj->reloc_table, edge->rel);
     }
     GraphRemoveEdge(T_GRAPH(CFG_EDGE_CFG(edge)), T_EDGE(edge));
   }
}
/* }}} */

t_cfg_edge * CfgEdgeChangeReturnForCall(t_cfg_edge * call, t_bbl * ret)
{
  if (!BBL_FUN(CFG_EDGE_TAIL(call)))
  {
    FATAL(("Not implemented for non function calls"));
  }
  if (call->corr_edge)
  {
    CfgEdgeKill(call->corr_edge);
    call->corr_edge=NULL;
  }
  call->corr_edge=CfgAppendNode(call->cfg, FUN_BBL_LAST(BBL_FUN(CFG_EDGE_TAIL(call))) , ret  ,ET_RETURN);
  T_CFG_EDGE(call->corr_edge)->corr_edge=call;
  T_CFG_EDGE(call->corr_edge)->cfg=call->cfg;
  return NULL;
}

/* vim: set shiftwidth=2 foldmethod=marker: */
