/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * All functions to manipulate flowgraphs
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/

/*#define NEW_CFG_REMOVE_DEAD_CODE_OR_DATA*/

/*Includes {{{ */
#include <string.h>
#include "../config.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../datastructures/data_string.h"
#include "../datastructures/data_stack.h"
#include "../datastructures/data_ll.h"
#include "../diablo_options.h"
#include "diablo_flowgraph.h"
#include "diablo_function.h"
#include "diablo_bbl.h"
#include "diablo_edge.h"
#include "diablo_ins.h"
#include "diablo_print.h"
#include "diablo_dominator.h"
#include "diablo_reloc.h"
#include "diablo_iterators.h"
#include "diablo_programexit.h"
#include "diablo_profile.h"
#include <string.h>
/*}}}*/
/* CfgNew {{{*/ 
/*! Allocate a new control flow graph for section sec. */

t_cfg * CfgNew(t_section * sec)
{
  t_cfg * ret=Malloc(sizeof(t_cfg));
  ret->entries=NULL;
  ret->fun_first=NULL;
  ret->fun_last=NULL;
  ret->loop_first=NULL;
  ret->loop_last=NULL;
  ret->list_of_marked_funs=NULL;
  CFG_SECTION(ret)=sec;
  ret->udgraph=NULL;
  GraphInit(T_GRAPH(ret),sizeof(t_bbl), sizeof(t_cfg_edge));
  ret->exitblocks=NULL;
  ret->cg = NULL;
  return ret;
}
/*}}}*/
/* Free the cfg {{{ */
static void cfg_helper_free_data(t_bbl * current)
{
  t_ins * ins, * insnext;
  /*  data_ll_elem * llnext;*/
  for (ins = current->ins_first; ins!=NULL; ins=insnext)
  {
    insnext=INS_NEXT(ins);
    if (RELOCATABLE_REFERS_TO(ins)) FATAL(("Still refers to!"));
    if (RELOCATABLE_REFED_BY(ins)) FATAL(("Still refed by!"));
    InsFree(ins);
  }
  LoopBblCleanup(current);

/*  while(BBL_DOMINATES(current))
  {
    llnext = BBL_DOMINATES(current);
    BBL_DOMINATES(current) = LLNext(llnext);
    Free(llnext);
  }*/
}
/*! Free a control flow graph */
/* CfgFree {{{*/
void CfgFree(t_cfg * cfg)
{
  t_section * section=CFG_SECTION(cfg);
  t_node * node, * node_next; 
  t_graph * fg= T_GRAPH(cfg);
  t_edge * edge, * edge_next;
  t_function * func,*funcnext;

  printf("Graph %p\n", fg); printf("has %d nodes\n",fg->nnodes);

  /* first release the associated ud-graph */
  if (cfg->udgraph)
    UdGraphFree(cfg->udgraph);

  if (cfg->cg)
    CgFree(cfg->cg);

  while(CFG_LOOP_FIRST(cfg)) LoopKill(&(CFG_LOOP_FIRST(cfg)));
  /* */
#if 0
  DominatorFree(&BBL_PDOMINATES(cfg->hell_node));
  DominatorFree(&BBL_PDOMINATES(cfg->exit_hell_node));
  DominatorFree(&BBL_PDOMINATES(cfg->call_hell_node));
  DominatorFree(&BBL_PDOMINATES(cfg->exit_call_hell_node));
  DominatorFree(&BBL_PDOMINATES(cfg->swi_hell_node));
  DominatorFree(&BBL_PDOMINATES(cfg->exit_swi_hell_node));
  while(BBL_PDOMINATES(cfg->hell_node))
  {
    data_ll_elem* to_free;
    to_free = BBL_PDOMINATES(cfg->hell_node); 
    BBL_PDOMINATES(cfg->hell_node) = LLNext(BBL_PDOMINATES(cfg->hell_node));
    Free(to_free);
  }
  /* */
  while(BBL_PDOMINATES(cfg->exit_hell_node))
  {
    data_ll_elem* to_free;
    to_free = BBL_PDOMINATES(cfg->exit_hell_node); 
    BBL_PDOMINATES(cfg->exit_hell_node) = LLNext(BBL_PDOMINATES(cfg->exit_hell_node));
    Free(to_free);
  }
  /* */
  while(BBL_PDOMINATES(cfg->swi_hell_node))
  {
    data_ll_elem* to_free;
    to_free = BBL_PDOMINATES(cfg->swi_hell_node); 
    BBL_PDOMINATES(cfg->swi_hell_node) = LLNext(BBL_PDOMINATES(cfg->swi_hell_node));
    Free(to_free);
  }
  /* */
  while(BBL_PDOMINATES(cfg->swi_exit_hell_node))
  {
    data_ll_elem* to_free;
    to_free = BBL_PDOMINATES(cfg->swi_exit_hell_node); 
    BBL_PDOMINATES(cfg->swi_exit_hell_node) = LLNext(BBL_PDOMINATES(cfg->swi_exit_hell_node));
    Free(to_free);
  }
  /* */
  while(BBL_PDOMINATES(cfg->call_hell_node))
  {
    data_ll_elem* to_free;
    to_free = BBL_PDOMINATES(cfg->call_hell_node); 
    BBL_PDOMINATES(cfg->call_hell_node) = LLNext(BBL_PDOMINATES(cfg->call_hell_node));
    Free(to_free);
  }
  /* */
  while(BBL_PDOMINATES(cfg->exit_call_hell_node))
  {
    data_ll_elem* to_free;
    to_free = BBL_PDOMINATES(cfg->exit_call_hell_node); 
    BBL_PDOMINATES(cfg->exit_call_hell_node) = LLNext(BBL_PDOMINATES(cfg->exit_call_hell_node));
    Free(to_free);
  }
#endif
  CfgFreeExitBlockList(cfg);

  for (node=fg->node_first; node!=NULL; node=node_next)
  {
    if ((RELOCATABLE_REFED_BY(node)!=NULL)||(RELOCATABLE_REFERS_TO(node)!=NULL))
      FATAL(("Nodes still used in relocs (eg @R)\n",RELOCATABLE_REFED_BY(node)!=NULL?RELOCATABLE_REFED_BY(node)->rel:RELOCATABLE_REFERS_TO(node)->rel));
    cfg_helper_free_data((t_bbl *) node);
    node_next=node->next;
    Free(node);
  }
  for (edge=fg->edge_first; edge!=NULL; edge=edge_next) {
    edge_next = edge->next;
    Free(edge);
  };
  for (func=cfg->fun_first; func!=NULL; func=funcnext)
  {
    funcnext=func->next;
    if (func->name)
      Free(func->name);
  while(FUN_LOOPS(func))
  {
    t_loopref * delete_me = FUN_LOOPS(func);
    FUN_LOOPS(func) = LLNext(FUN_LOOPS(func));
    Free(delete_me);
  }

    Free(func);
  }

  Free(cfg->entries);
  Free(cfg);
  if (section) section->data=NULL;
}
/*}}}*/

/*}}}*/
/* Functions to patch a flowgraph to function representation {{{ */

/*! this is an easy one: just check all edges to see if they connect blocks 
 * in different functions with a non-interprocedural edge. if so, patch the
 * edge and add a compensating edge */

/* CfgPatchNormalEdges {{{ */
t_uint32 CfgPatchNormalEdges(t_cfg * cfg)
{
  t_cfg_edge * edge;
  t_uint32 ret=0,ret1=0,ret2=0,ret3=0,ret4=0;
  t_bbl * bfrom, * bto;
  t_function * ffrom, * fto;

  /* {{{ avoid interprocedural funlinks */
  /* we want to avoid interprocedural funlink edges, because they require special
   * and hard to understand handling in most analyses. Therefore, for each interprocedural
   * funlink edge we add an empty block to the calling function, and make this the tail 
   * of the funlink edge. From this block, a fallthrough edge (which may be interprocedural)
   * goes to the original tail of the funlink edge */
  CFG_FOREACH_EDGE(cfg,edge)
  {
    t_bbl * head, * tail, * new;
    t_cfg_edge * edge2, * tmp;

    if (EDGE_CAT(edge) != ET_FUNLINK) continue;
    if (BBL_FUN(CFG_EDGE_HEAD(edge)) == BBL_FUN(CFG_EDGE_TAIL(edge))) continue;

    ret2++;

    head = CFG_EDGE_HEAD(edge);
    tail = CFG_EDGE_TAIL(edge);
    new = BblNew(cfg);
    if (BBL_FUN(head))
      BblInsertInFunction(new,BBL_FUN(head));
    else
      BBL_FUN(new) = NULL;
    
    /* move funlink and return and (if present) fallthrough edge */
    BBL_FOREACH_PRED_EDGE_SAFE(tail,edge2,tmp)
    {
      if (EDGE_CAT(edge2) != ET_FUNLINK && EDGE_CAT(edge2) != ET_FALLTHROUGH && EDGE_CAT(edge2) != ET_RETURN) continue;

      /* move the edge */
      CFG_EDGE_TAIL(edge2) = new;
      /* remove from predecessor list of tail */
      if (EDGE_PRED_NEXT(edge2))
	EDGE_PRED_PREV(EDGE_PRED_NEXT(edge2)) = EDGE_PRED_PREV(edge2);
      else
	BBL_PRED_LAST(tail) = T_CFG_EDGE(EDGE_PRED_PREV(edge2));
      if (EDGE_PRED_PREV(edge2))
	EDGE_PRED_NEXT(EDGE_PRED_PREV(edge2)) = EDGE_PRED_NEXT(edge2);
      else
	BBL_PRED_FIRST(tail) = T_CFG_EDGE(EDGE_PRED_NEXT(edge2));
      /* add to predecessor list of new */
      EDGE_PRED_NEXT(edge2) = (t_edge *) BBL_PRED_FIRST(new);
      EDGE_PRED_PREV(edge2) = NULL;
      if (EDGE_PRED_NEXT(edge2)) EDGE_PRED_PREV(EDGE_PRED_NEXT(edge2)) = (t_edge *) edge2;
      BBL_PRED_FIRST(new) = edge2;
      if (!BBL_PRED_LAST(new)) BBL_PRED_LAST(new) = edge2;
    }
    CfgEdgeNew(cfg,new,tail,ET_FALLTHROUGH);
  } /* }}} */

  ANALYSIS_START("Patching normal edges");
  CFG_FOREACH_EDGE(cfg,edge)
  {
    if (EDGE_CAT(edge) == ET_CALL) continue;
    if (EDGE_CAT(edge) == ET_RETURN) continue;
    if (EDGE_CAT(edge) == ET_SWI) continue;
    if (EDGE_CAT(edge) == ET_COMPENSATING) continue;
    if (EDGE_CAT(edge) & ET_INTERPROC) continue;

    bfrom = CFG_EDGE_HEAD(edge);
    bto = CFG_EDGE_TAIL(edge); 
    /*	 if (IsHellNode(bfrom) || IsHellNode(bto)) continue;  */
    if (bfrom->fun == bto->fun) continue; /* It's intraprocedural */
    /* add compensating edge */
    ffrom = BBL_FUN(bfrom); 
    fto = BBL_FUN(bto);
    if ((!ffrom) || (!fto)) continue; 
    ret++;
    CfgEdgeNewCompensating(cfg, edge);
    /* patch type of the interprocedural edge */
    switch (EDGE_CAT(edge)) 
    {
      case ET_UNKNOWN: 
        EDGE_CAT(edge)= ET_IPUNKNOWN; 
        break;
      case ET_FALLTHROUGH: 
        EDGE_CAT(edge)=ET_IPFALLTHRU; 
        ret1++;
        break;
      case ET_SWILINK: 
        EDGE_CAT(edge)=ET_IPSWILINK; 
        ret3++;
        break;
      case ET_JUMP: 
        EDGE_CAT(edge)=ET_IPJUMP; 
        ret4++;
        break;
      case ET_SWITCH:
	EDGE_CAT(edge)=ET_IPSWITCH;
	ret4++;
	break;
      default:
	FATAL(("Uncaught!"));
    }
  }
  ANALYSIS_END("Patching normal edges",("%d ip fallthrough edges, %d ip funlinks, %d ip swilinks, %d ip jumps",ret1,ret2,ret3,ret4));
  return ret;
}
/* }}} */

/*! This function patches edges to the function representation (uniq exit block
 * and IP edges) */

/* CfgPatchInterProcedural {{{ */
t_uint32 CfgPatchInterProcedural(t_cfg * fg)
{
  /* patching interprocedural edges */
  t_uint32 ret=0,ret5=0;
  t_function * function;

  ret=CfgPatchNormalEdges(fg);

  ANALYSIS_START("Moving return edges to return block");
  FOREACH_FUNCTION(CFG_SECTION(fg),function) 
    {
      ret5+=FunctionMoveReturnEdgesToReturnBlock(function);  
    }
  ANALYSIS_END("Moving return edges to return block",("Moved %d edges",ret5));

  ANALYSIS_START("Adding return for normal calls");
  FOREACH_FUNCTION(CFG_SECTION(fg),function) 
    {
      ret5+=FunctionAdjustReturnEdges(function);
    }
  ANALYSIS_END("Adding return for normal calls",("Moved %d edges",ret5));

  ret+=ret5;

  return ret;
}
/* }}} */
/*}}}*/


t_cfg_entrypoints * NewEntry(t_bbl * node, t_cfg_edge * edge)
{
  t_cfg_entrypoints * ret=Malloc(sizeof(t_cfg_entrypoints));
  ret->next=NULL;
  ret->tmp=NULL;
  ret->entry_bbl=node;
  ret->entry_edge=edge;
  return ret;
}

/* Flowgraph export functions: use gviz to convert the output to PostScript
 * files */
/* Export Flowgraphs {{{ */
/* for Export_Flowgraph (icfg = interprocedural control flow graph) */
static t_bool node_writeout_icfg(t_node * current, t_node * parent,void * data) {
  /* print out all successors of a node in the flowgraph */
  /* used for exportation of the flowgraph */
  FILE * out=data;
  t_ins * ins;
  t_cfg_edge  * edge;

  ins = BBL_INS_FIRST(current);
  DiabloPrint(out, "\t\"%p\" [label=\"",current);
  if (T_BBL(current) ==  T_BBL(current)->cfg->hell_node)
  {
    DiabloPrint(out,"HELL");
  } 
  else if (T_BBL(current) ==  T_BBL(current)->cfg->swi_hell_node)
  {
    DiabloPrint(out,"SWI HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->call_hell_node) 
  {
    DiabloPrint(out, "CALL HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->exit_call_hell_node)
  {
    DiabloPrint(out, "EXIT CALL HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->exit_hell_node)
  {
    DiabloPrint(out, "EXIT HELL");
  } 
  else if (T_BBL(current) == T_BBL(current)->cfg->exit_swi_hell_node)
  {
    DiabloPrint(out, "EXIT SWI HELL");
  } 
  else if (BBL_FUN(current)&&(BBL_IS_LAST(T_BBL(current)))) 
  { /* return block */
    DiabloPrint(out, "RETURN FROM F%p", T_BBL(current)->fun);
  } 
  else 
  {
    DiabloPrint(out, "ADDR: @G (F: %p)\\n", BBL_CADDRESS(T_BBL(current)), T_BBL(current)->fun);
    while (ins != NULL)
    {
      DiabloPrint(out,"@I\\l",ins);
      ins = INS_NEXT(ins);
    }
  }

  DiabloPrint(out,"\"]\n");
  BBL_FOREACH_SUCC_EDGE(current,edge) 
  {
    if (global_options.annotate_loops && EDGE_CAT(edge) != ET_FUNLINK)
    {
      DiabloPrint(out, "\t\"%p\" -> \"%p\" [style=%s,color=%s,label=\"%d,%d\"];\n",current,CFG_EDGE_TAIL(edge),
          (EDGE_CAT(edge) == ET_FALLTHROUGH) ? "solid" : "dashed",
          (EDGE_CAT(edge) & ET_INTERPROC) ? "red" : "black", EDGE_EXEC_COUNT(edge),EDGE_EXEC_COUNT_MIN(edge));
    }
    if (/*TOBBL(current->succs[i]->to) != hell_node*/ 1)
    {
       DiabloPrint(out, "\t\"%p\" -> \"%p\" [style=%s,color=%s];\n",current,CFG_EDGE_TAIL(edge),
            (EDGE_CAT(edge) == ET_FALLTHROUGH) ? "solid" : "dashed",
           (EDGE_CAT(edge) & ET_INTERPROC) ? "red" : "black");
    }
  }
  return FALSE;
}


void Export_Flowgraph(t_bbl * start, t_uint16 mask, const char * filename) {
  FILE * out = fopen(filename,"w");
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));
  DiabloPrint(out, "digraph \"@G\" {\n\tnode [shape=box]\n",BBL_OLD_ADDRESS(start));
  GraphDFTraversalWithCheckAndData((t_node *) start,node_writeout_icfg, mask,out);
  DiabloPrint(out, "}\n");
  fclose(out);
}

void Export_Function(t_function * fun, const char * filename)
{
  t_ins * ins;
  t_bbl * bbl;
  t_cfg_edge * edge;
  t_bool show_callers = TRUE;
  int callers = 0;

  FILE * out = fopen(filename,"w");
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));

  /* only show callers if there are less than 10 */
  BBL_FOREACH_PRED_EDGE(FUN_BBL_FIRST(fun),edge)
    if (EDGE_CAT(edge) == ET_CALL)
      callers++;
  show_callers = (callers < 15);

  /* graph header */
  if (FUN_NAME(fun) && strlen(FUN_NAME(fun)))
    DiabloPrint(out, "digraph \"%s\" {\n\tnode [shape=box];\n",FUN_NAME(fun));
  else
    DiabloPrint(out, "digraph \"noname_@G\" {\n\tnode [shape=box]\n",BBL_CADDRESS(FUN_BBL_FIRST(fun)));

  /* graph body */

  /* {{{  regular nodes */
  FUN_FOREACH_BBL(fun,bbl)
  {
    if (bbl == bbl->cfg->hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->call_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"CALL HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->unique_entry_node)
    {
      fprintf(out,"\t\"%p\" [label=\"Entry\"];\n",bbl);
    }
    else if (bbl == FunctionGetExitBlock(fun))
    {
      fprintf(out,"\t\"%p\" [label=\"RETURN\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->exit_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"EXIT HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->exit_call_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"EXIT CALL HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->swi_hell_node)
      {
	fprintf(out,"\t\"%p\" [label=\"SWI HELL\"];\n",bbl);
      }
    else if (bbl == bbl->cfg->exit_swi_hell_node)
      {
	fprintf(out,"\t\"%p\" [label=\"EXIT SWI HELL\"];\n",bbl);
      }
    else
    {
      fprintf(out,"\t\"%p\" ",bbl);
      DiabloPrint(out, "[label=\"@B\\n", bbl);
      BBL_FOREACH_INS(bbl,ins)
	DiabloPrint(out,"@I\\l",ins);
#ifndef INSTRUMENTATION
      fprintf(out,"\" style=filled fillcolor=%s ];\n",((BblIsHot(bbl)&&global_options.blockprofilefile)?"PeachPuff2":((BBL_EXEC_COUNT(bbl)>0&&global_options.blockprofilefile)?"lightyellow":"white")));
#else
      fprintf(out,"\" style=filled fillcolor=white ];\n");
#endif
    }
  } /* }}} */

  /* {{{ successor edges */
  FUN_FOREACH_BBL(fun,bbl)
  {
    BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      t_bbl * tail = CFG_EDGE_TAIL(edge);

      if (!EdgeTestCategoryOr(edge,ET_INTERPROC))
      {
	t_string style, color;

	switch (EDGE_CAT(edge))
	{
	  case ET_FUNLINK:
	  case ET_SWILINK:
	    style = "solid";
	    color = "yellow";
	    break;
	  case ET_FALLTHROUGH:
	    style = "solid";
	    color = "green";
	    break;
	  case ET_UNKNOWN:
	    style = "dashed";
	    color = "purple";
	    break;
	  case ET_JUMP:
	  case ET_SWITCH:
	    style = "solid";
	    color = "black";
	    break;
	  default:
	    style = "dashed";
	    color = "blue";
	}

#ifndef INSTRUMENTATION
	/* intraprocedural edges */
	if (EdgeIsHot(edge) && global_options.blockprofilefile)
	{
	  if (!strcmp(style,"solid"))
	    style = "bold";

	  if (!strcmp(style,"dashed"))
	    style = "dotted";

	  if (EDGE_CAT(edge)==ET_SWITCH) 
	    fprintf(out,"\t\"%p\" -> \"%p\" [weight=%d,style=%s,color=%s,label=\"%d\"];\n",bbl,tail,EDGE_EXEC_COUNT(edge),style,color,edge->switchvalue);
	  else 
	    fprintf(out,"\t\"%p\" -> \"%p\" [weight=%d,style=%s,color=%s,label=\"O(%1.2f)\"];\n",bbl,tail,EDGE_EXEC_COUNT(edge),style,color,ORDER(EDGE_EXEC_COUNT(edge)));
	}
	else if (global_options.annotate_loops)
	{
	  if (EDGE_CAT(edge)==ET_SWITCH) 
	    fprintf(out,"\t\"%p\" -> \"%p\" [weight=%d,style=%s,color=%s,label=\"%d\"];\n",bbl,tail,1+EDGE_EXEC_COUNT(edge),style,color,edge->switchvalue);
	  else 
	    fprintf(out,"\t\"%p\" -> \"%p\" [weight=%d,style=%s,color=%s,label=\"%d,%d\"];\n",bbl,tail,1+EDGE_EXEC_COUNT(edge),style,color,EDGE_EXEC_COUNT(edge),EDGE_EXEC_COUNT_MIN(edge));
	}
	else
#endif
	{
	  if (EDGE_CAT(edge)==ET_SWITCH) 
 	    fprintf(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s,label=\"%d\"];\n",bbl,tail,style,color,edge->switchvalue);
	  else 
 	    fprintf(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s,label=\"O(%1.2f)\"];\n",bbl,tail,style,color,ORDER(EDGE_EXEC_COUNT(edge)));
	}
      }
      else
      {
	/* interprocedural edges */
	t_string estyle, ecolor;
	t_string ncolor;
	t_string nodetext;
	char nodename[20];
	t_bool filled = FALSE;
	t_string destfun = BBL_FUN(tail)?(FUN_NAME(BBL_FUN(tail)) ? FUN_NAME(BBL_FUN(tail)) : "noname"):"nofun";

	/* don't print successors of the hell or call hell nodes (there's no way you can get them all to fit nicely in a drawing) */
	if(!IsHellNode(tail))
	  if (bbl == bbl->cfg->hell_node || bbl == bbl->cfg->call_hell_node || bbl == bbl->cfg->swi_hell_node)
	    continue;

	switch (EDGE_CAT(edge))
	{
	  case ET_CALL:
	    estyle = "solid";
	    ecolor = "red";
	    ncolor = "yellow";
	    sprintf(nodename, "f%p", tail);
	    nodetext = sDiabloPrint("%s (@B)",destfun,tail);
	    filled = TRUE;
	    break;
	  case ET_RETURN:
	    estyle = "solid";
	    ecolor = "black";
	    ncolor = "green";
	    sprintf(nodename, "r%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  case ET_COMPENSATING:
	    estyle = "dashed";
	    ecolor = "black";
	    ncolor = "green";
	    sprintf(nodename, "c%p", tail);
	    nodetext = sDiabloPrint("return of %s",destfun);
	    break;
	  case ET_SWI:
	    estyle = "dashed";
	    ecolor = "red";
	    ncolor = "red";
	    sprintf(nodename,"i%p",tail); /* tail will be the hell node */
	    nodetext = sDiabloPrint("HELL");
	    break;
	  case ET_IPUNKNOWN:
	    estyle = "bold";
	    ecolor = "purple";
	    ncolor = "magenta";
	    filled = TRUE;
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  case ET_IPFALLTHRU:
	  case ET_IPFUNLINK:
	  case ET_IPJUMP:
	  case ET_IPSWILINK:
	  case ET_IPSWITCH:
	    estyle = "solid";
	    ecolor = "red";
	    ncolor = "black";
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  default:
	    /* make it stand out as something we haven't seen yet */
	    estyle = "bold";
	    ecolor = "magenta";
	    ncolor = "blue";
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    filled = TRUE;
	    break;
	}

	/* catch hell node and call hell node */
	if (tail == bbl->cfg->hell_node || tail == bbl->cfg->call_hell_node || tail == bbl->cfg->swi_hell_node)
	{
	  sprintf(nodename,"h%p",tail);
	  Free(nodetext);
	  nodetext = sDiabloPrint("%sHELL", tail == bbl->cfg->hell_node ? "" : tail == bbl->cfg->swi_hell_node ? "SWI " : "CALL ");
	  ncolor = "red";
	  filled = TRUE;
	}

	if (show_callers || EDGE_CAT(edge) != ET_RETURN)
	{
	  if(global_options.annotate_loops)
	  {
	    fprintf(out,"\t\"%p\" -> \"%s\" [style=%s,color=%s,label=\"%d,%d\"];\n",bbl,nodename,estyle,ecolor,EDGE_EXEC_COUNT(edge),EDGE_EXEC_COUNT_MIN(edge));
	    fprintf(out,"\t\"%s\" [label=\"%s\",color=\"%s\"%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
	  }
	  else
	  {
	    fprintf(out,"\t\"%p\" -> \"%s\" [style=%s,color=%s];\n",bbl,nodename,estyle,ecolor);
	    fprintf(out,"\t\"%s\" [label=\"%s\",color=\"%s\"%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
	  }
	}
	Free(nodetext);
      }
    }
  } /* }}} */

  /* {{{ predecessor edges */
  FUN_FOREACH_BBL(fun,bbl)
  {
    BBL_FOREACH_PRED_EDGE(bbl,edge)
    {
      t_bbl * head = CFG_EDGE_HEAD(edge);
      t_string destfun = BBL_FUN(head)?(FUN_NAME(BBL_FUN(head)) ? FUN_NAME(BBL_FUN(head)) : "noname"):"nofun";
      t_string estyle, ecolor;
      t_string ncolor;
      t_string nodetext;
      char nodename[20];
      t_bool filled = FALSE;

      if (!EdgeTestCategoryOr(edge,ET_INTERPROC)) continue; /* all intraprocedural edges have already been handled as predecessors */
      if (bbl == bbl->cfg->hell_node || bbl == bbl->cfg->call_hell_node || bbl == bbl->cfg->swi_hell_node) continue; /* don't do this for hell nodes */

      switch (EDGE_CAT(edge))
      {
	case ET_CALL:
	  estyle = "solid";
	  ecolor = "red";
	  ncolor = "yellow";
	  sprintf(nodename,"F%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_IPUNKNOWN:
	  estyle = "bold";
	  ecolor = "purple";
	  ncolor = "blue";
	  sprintf(nodename,"I%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_IPFALLTHRU:
	case ET_IPFUNLINK:
	case ET_IPSWILINK:
	case ET_IPJUMP:
	case ET_IPSWITCH:
	  estyle = "solid";
	  ecolor = "red";
	  ncolor = "cyan";
	  sprintf(nodename,"I%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_COMPENSATING:
	  estyle = "solid";
	  ecolor = "blue";
	  ncolor = "black";
	  sprintf(nodename,"C%p",head);
	  nodetext = sDiabloPrint("return from %s",destfun);
	  break;
	case ET_RETURN:
	  estyle = "solid";
	  ecolor = "blue";
	  ncolor = NULL;
	  nodetext = NULL;
	  if (FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->hell_node || FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->call_hell_node || FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->swi_hell_node)
	    sprintf(nodename,"h%p", FUN_BBL_FIRST(BBL_FUN(head)));
	  else
	    sprintf(nodename,"f%p", FUN_BBL_FIRST(BBL_FUN(head)));
	  break;
	default:
	  FATAL(("unexpected edge type"));
	  exit(0); /* Keeps the compiler happy */
      }

      /* catch hell node and call hell node */
      if (head == bbl->cfg->hell_node || head == bbl->cfg->call_hell_node || head == bbl->cfg->swi_hell_node)
      {
	sprintf(nodename,"h%p",head);
	if (nodetext)
	  Free(nodetext);
	nodetext = sDiabloPrint("%sHELL", head == bbl->cfg->hell_node ? "" : head == bbl->cfg->swi_hell_node ? "SWI " :"CALL ");
	ncolor = "red";
	filled = TRUE;
      }
      if (head == bbl->cfg->exit_hell_node || head == bbl->cfg->exit_call_hell_node)
      {
	sprintf(nodename,"h%p",head);
	nodetext = sDiabloPrint("EXIT %sHELL", head == bbl->cfg->exit_hell_node? "" : head == bbl->cfg->exit_swi_hell_node ? "SWI " :"CALL ");
	ncolor = "red";
	filled = FALSE;
      }
      /* catch the entry node */
      if (head == bbl->cfg->unique_entry_node)
      {
	sprintf(nodename,"e%p",head);
	if (nodetext)
	  Free(nodetext);
	nodetext = sDiabloPrint("Entry");
	ncolor = "blue";
	filled = TRUE;
      }

      if (show_callers || EDGE_CAT(edge) != ET_CALL)
      {
	if(global_options.annotate_loops)
	{
	  fprintf(out,"\t\"%s\" -> \"%p\" [style=%s,color=%s,label=\"%d,%d\"];\n",nodename,bbl,estyle,ecolor,EDGE_EXEC_COUNT(edge),EDGE_EXEC_COUNT_MIN(edge));
	  if (nodetext)
	    fprintf(out,"\t\"%s\" [label=\"%s\",color=%s%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
	}
	else
	{
	  fprintf(out,"\t\"%s\" -> \"%p\" [style=%s,color=%s];\n",nodename,bbl,estyle,ecolor);
	  if (nodetext)
	    fprintf(out,"\t\"%s\" [label=\"%s\",color=%s%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
	}
      }
      if (nodetext) 
	Free(nodetext);
    }
  } /* }}} */
  
  /* graph footer */
  DiabloPrint(out, "}\n");
  fclose(out);
}

void Export_Function_Annotate(t_function * fun, const char * filename, void (*annotate)(t_bbl * bbl, FILE * fp))
{
  t_ins * ins;
  t_bbl * bbl;
  t_cfg_edge * edge;
  t_bool show_callers = TRUE;
  int callers = 0;

  FILE * out = fopen(filename,"w");
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));

  /* only show callers if there are less than 10 */
  BBL_FOREACH_PRED_EDGE(FUN_BBL_FIRST(fun),edge)
    if (EDGE_CAT(edge) == ET_CALL)
      callers++;
  show_callers = (callers < 10);

  /* graph header */
  if (FUN_NAME(fun) && strlen(FUN_NAME(fun)))
    DiabloPrint(out, "digraph \"%s\" {\n\tnode [shape=box];\n",FUN_NAME(fun));
  else
    DiabloPrint(out, "digraph \"noname_@G\" {\n\tnode [shape=box]\n",BBL_CADDRESS(FUN_BBL_FIRST(fun)));

  /* graph body */

  /* {{{  regular nodes */
  FUN_FOREACH_BBL(fun,bbl)
  {
    if (bbl == bbl->cfg->hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->call_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"CALL HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->swi_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"SWI HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->unique_entry_node)
    {
      fprintf(out,"\t\"%p\" [label=\"Entry\"];\n",bbl);
    }
    else if (bbl == FunctionGetExitBlock(fun))
    {
      fprintf(out,"\t\"%p\" [label=\"",bbl);
      annotate(bbl,out);
      fprintf(out,"RETURN\"];\n");
    }
    else if (bbl == bbl->cfg->exit_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"EXIT HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->exit_call_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"EXIT CALL HELL\"];\n",bbl);
    }
    else if (bbl == bbl->cfg->exit_swi_hell_node)
    {
      fprintf(out,"\t\"%p\" [label=\"EXIT SWI HELL\"];\n",bbl);
    }
    else
    {
      fprintf(out,"\t\"%p\" ",bbl);
      DiabloPrint(out, "[label=\"@B\\n", bbl);
      annotate(bbl, out); fprintf(out, "\\n");
      BBL_FOREACH_INS(bbl,ins)
	DiabloPrint(out,"@I\\l",ins);
#ifndef INSTRUMENTATION
      fprintf(out,"\" style=filled fillcolor=%s ];\n",((BblIsHot(bbl)&&global_options.blockprofilefile)?"PeachPuff2":((BBL_EXEC_COUNT(bbl)>0&&global_options.blockprofilefile)?"lightyellow":"white")));
#else
      fprintf(out,"\" style=filled fillcolor=white ];\n");
#endif
    }
  } /* }}} */

  /* {{{ successor edges */
  FUN_FOREACH_BBL(fun,bbl)
  {
    BBL_FOREACH_SUCC_EDGE(bbl,edge)
    {
      t_bbl * tail = CFG_EDGE_TAIL(edge);

      if (!EdgeTestCategoryOr(edge,ET_INTERPROC))
      {
	t_string style, color;

	switch (EDGE_CAT(edge))
	{
	  case ET_FUNLINK:
	  case ET_SWILINK:
	    style = "solid";
	    color = "yellow";
	    break;
	  case ET_FALLTHROUGH:
	    style = "solid";
	    color = "green";
	    break;
	  case ET_UNKNOWN:
	    style = "dashed";
	    color = "purple";
	    break;
	  case ET_JUMP:
	  case ET_SWITCH:
	    style = "solid";
	    color = "black";
	    break;
	  default:
	    style = "dashed";
	    color = "blue";
	}

#ifndef INSTRUMENTATION
	/* intraprocedural edges */
	if (EdgeIsHot(edge) && global_options.blockprofilefile)
	{
	  if (!strcmp(style,"solid"))
	    style = "bold";

	  if (!strcmp(style,"dashed"))
	    style = "dotted";

	  if (EDGE_CAT(edge)==ET_SWITCH) 
	    fprintf(out,"\t\"%p\" -> \"%p\" [weight=%d,style=%s,color=%s,label=\"%d\"];\n",bbl,tail,EDGE_EXEC_COUNT(edge),style,color,edge->switchvalue);
	  else 
	    fprintf(out,"\t\"%p\" -> \"%p\" [weight=%d,style=%s,color=%s,label=\"O(%1.2f)\"];\n",bbl,tail,EDGE_EXEC_COUNT(edge),style,color,ORDER(EDGE_EXEC_COUNT(edge)));
	}
	else
#endif
	{
	  if (EDGE_CAT(edge)==ET_SWITCH) 
	    fprintf(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s,label=\"%d\"];\n",bbl,tail,style,color,edge->switchvalue);
	  else 
	    fprintf(out,"\t\"%p\" -> \"%p\" [style=%s,color=%s,label=\"O(%1.2f)\"];\n",bbl,tail,style,color,ORDER(EDGE_EXEC_COUNT(edge)));
	}
      }
      else
      {
	/* interprocedural edges */
	t_string estyle, ecolor;
	t_string ncolor;
	t_string nodetext;
	char nodename[20];
	t_bool filled = FALSE;
	t_string destfun = BBL_FUN(tail)?(FUN_NAME(BBL_FUN(tail)) ? FUN_NAME(BBL_FUN(tail)) : "noname"):"nofun";

	/* don't print successors of the hell or call hell nodes (there's no way you can get them all to fit nicely in a drawing) */
	if (bbl == bbl->cfg->hell_node || bbl == bbl->cfg->call_hell_node || bbl == bbl->cfg->swi_hell_node)
	  continue;

	switch (EDGE_CAT(edge))
	{
	  case ET_CALL:
	    estyle = "solid";
	    ecolor = "red";
	    ncolor = "yellow";
	    sprintf(nodename, "f%p", tail);
	    nodetext = sDiabloPrint("%s (@B)",destfun,tail);
	    filled = TRUE;
	    break;
	  case ET_RETURN:
	    estyle = "solid";
	    ecolor = "black";
	    ncolor = "green";
	    sprintf(nodename, "r%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  case ET_COMPENSATING:
	    estyle = "dashed";
	    ecolor = "black";
	    ncolor = "green";
	    sprintf(nodename, "c%p", tail);
	    nodetext = sDiabloPrint("return of %s",destfun);
	    break;
	  case ET_SWI:
	    estyle = "dashed";
	    ecolor = "red";
	    ncolor = "red";
	    sprintf(nodename,"i%p",tail); /* tail will be the hell node */
	    nodetext = sDiabloPrint("HELL");
	    break;
	  case ET_IPUNKNOWN:
	    estyle = "bold";
	    ecolor = "purple";
	    ncolor = "magenta";
	    filled = TRUE;
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  case ET_IPFALLTHRU:
	  case ET_IPFUNLINK:
	  case ET_IPJUMP:
	  case ET_IPSWILINK:
	  case ET_IPSWITCH:
	    estyle = "solid";
	    ecolor = "red";
	    ncolor = "black";
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    break;
	  default:
	    /* make it stand out as something we haven't seen yet */
	    estyle = "bold";
	    ecolor = "magenta";
	    ncolor = "blue";
	    sprintf(nodename, "i%p", tail);
	    nodetext = sDiabloPrint("@B in %s",tail,destfun);
	    filled = TRUE;
	    break;
	}

	/* catch hell node and call hell node */
	if (tail == bbl->cfg->hell_node || tail == bbl->cfg->call_hell_node || tail == bbl->cfg->swi_hell_node)
	{
	  sprintf(nodename,"h%p",tail);
	  Free(nodetext);
	  nodetext = sDiabloPrint("%sHELL", tail == bbl->cfg->hell_node ? "" : tail == bbl->cfg->swi_hell_node ? "SWI " :"CALL ");
	  ncolor = "red";
	  filled = TRUE;
	}

	if (show_callers || EDGE_CAT(edge) != ET_RETURN)
	{
	  fprintf(out,"\t\"%p\" -> \"%s\" [style=%s,color=%s];\n",bbl,nodename,estyle,ecolor);
	  fprintf(out,"\t\"%s\" [label=\"%s\",color=\"%s\"%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
	}
	Free(nodetext);
      }
    }
  } /* }}} */

  /* {{{ predecessor edges */
  FUN_FOREACH_BBL(fun,bbl)
  {
    BBL_FOREACH_PRED_EDGE(bbl,edge)
    {
      t_bbl * head = CFG_EDGE_HEAD(edge);
      t_string destfun = BBL_FUN(head)?(FUN_NAME(BBL_FUN(head)) ? FUN_NAME(BBL_FUN(head)) : "noname"):"nofun";
      t_string estyle, ecolor;
      t_string ncolor;
      t_string nodetext;
      char nodename[20];
      t_bool filled = FALSE;

      if (!EdgeTestCategoryOr(edge,ET_INTERPROC)) continue; /* all intraprocedural edges have already been handled as predecessors */
      if (bbl == bbl->cfg->hell_node || bbl == bbl->cfg->call_hell_node || bbl == bbl->cfg->swi_hell_node) continue; /* don't do this for hell nodes */

      switch (EDGE_CAT(edge))
      {
	case ET_CALL:
	  estyle = "solid";
	  ecolor = "red";
	  ncolor = "yellow";
	  sprintf(nodename,"F%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_IPUNKNOWN:
	  estyle = "bold";
	  ecolor = "purple";
	  ncolor = "blue";
	  sprintf(nodename,"I%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_IPFALLTHRU:
	case ET_IPFUNLINK:
	case ET_IPSWILINK:
	case ET_IPJUMP:
	case ET_IPSWITCH:
	  estyle = "solid";
	  ecolor = "red";
	  ncolor = "cyan";
	  sprintf(nodename,"I%p",head);
	  nodetext = sDiabloPrint("@B in %s",head,destfun);
	  break;
	case ET_COMPENSATING:
	  estyle = "solid";
	  ecolor = "blue";
	  ncolor = "black";
	  sprintf(nodename,"C%p",head);
	  nodetext = sDiabloPrint("return from %s",destfun);
	  break;
	case ET_RETURN:
	  estyle = "solid";
	  ecolor = "blue";
	  ncolor = NULL;
	  nodetext = NULL;
	  if (FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->hell_node || FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->call_hell_node || FUN_BBL_FIRST(BBL_FUN(head)) == bbl->cfg->swi_hell_node)
	    sprintf(nodename,"h%p", FUN_BBL_FIRST(BBL_FUN(head)));
	  else
	    sprintf(nodename,"f%p", FUN_BBL_FIRST(BBL_FUN(head)));
	  break;
	default:
	  FATAL(("unexpected edge type"));
	  exit(0); /* Keeps the compiler happy */
      }

      /* catch hell node and call hell node */
      if (head == bbl->cfg->hell_node || head == bbl->cfg->call_hell_node || head == bbl->cfg->swi_hell_node)
      {
	sprintf(nodename,"h%p",head);
	if (nodetext) Free(nodetext);
	nodetext = sDiabloPrint("%sHELL", head == bbl->cfg->hell_node ? "" : head == bbl->cfg->swi_hell_node ? "SWI " :"CALL ");
	ncolor = "red";
	filled = TRUE;
      }
      /* catch the entry node */
      if (head == bbl->cfg->unique_entry_node)
      {
	sprintf(nodename,"e%p",head);
	Free(nodetext);
	nodetext = sDiabloPrint("Entry");
	ncolor = "blue";
	filled = TRUE;
      }

      if (show_callers || EDGE_CAT(edge) != ET_CALL)
      {
	fprintf(out,"\t\"%s\" -> \"%p\" [style=%s,color=%s];\n",nodename,bbl,estyle,ecolor);
	if (nodetext)
	  fprintf(out,"\t\"%s\" [label=\"%s\",color=%s%s];\n",nodename,nodetext,ncolor,filled?",style=filled":"");
      }
      if (nodetext) 
	Free(nodetext);
    }
  } /* }}} */
  
  /* graph footer */
  DiabloPrint(out, "}\n");
  fclose(out);
}

void ExportFunctionGraph(t_section * code, const t_string dirprefix)
{
  t_function * function;
  t_string dirpref = StringDup(dirprefix);
  char * fname;
  int noname_count = 0;
  char noname[20];

  /* dirprefix is by default ./dots */
  if (!dirpref) dirpref = StringDup("./dots");

  /* remove trailing slash */
  if (dirpref[strlen(dirpref)-1] == '/') dirpref[strlen(dirpref)-1] = '\0';

#ifdef HAVE_DIRS
  DirectoryAssure(dirpref,FALSE);
#endif

  FOREACH_FUNCTION(code,function) 
  {

    if (!FUN_NAME(function))
      sprintf(noname,"-noname-%d-",noname_count++);
      
#ifdef HAVE_DIRS
    if (function->bbl_first)
      fname=sDiabloPrint("%s/@G.func-%s%s.dot",dirpref,BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function)?FUN_NAME(function):noname,(FunIsHot(function)&&global_options.blockprofilefile)?"HOT":"");
    else
      fname=sDiabloPrint("%s/0x%x.func-%s%s.dot",dirpref,0,FUN_NAME(function)?FUN_NAME(function):noname,(FunIsHot(function)&&global_options.blockprofilefile)?"HOT":"");
#else
    if (function->bbl_first)
      fname=sDiabloPrint("./@G.func-%s%s.dot",BBL_OLD_ADDRESS(function->bbl_first),FUN_NAME(function)?FUN_NAME(function):noname,(FunIsHot(function)&&global_options.blockprofilefile)?"HOT":"");
    else
      fname=sDiabloPrint("./0x%x.func-%s%s.dot",0,FUN_NAME(function)?FUN_NAME(function):noname,(FunIsHot(function)&&global_options.blockprofilefile)?"HOT":"");
#endif
    Export_Function(function,fname);
    Free(fname);
  }
  Free(dirpref);

}


/* }}} */
/* CfgCreateBasicBlocks {{{ */

/* helper function: sort symbol array */
static int __helper_sort_syms(const void * a, const void * b)
{
  t_symbol * syma = *((t_symbol **)a);
  t_symbol * symb = *((t_symbol **)b);
  if (GENERIC_GE(symb->base->caddress,syma->base->caddress))
    return -1;
  else
    return 1;
}

t_uint32 CfgCreateBasicBlocks(t_section * section)
{
  t_uint32 nbbls=0;
  t_cfg_edge * edge;
  t_ins *i_ins;
  t_bbl * bbl=NULL;
  t_bbl * entry=NULL;
  t_object * obj=section->obj;
  t_reloc * reloc, * tmp;
  t_address current_address=SECTION_GET_CADDRESS(section);
  t_cfg * flowgraph=CfgNew(section);

  t_symbol * sym, **symarr;
  int nsyms = 0,i;
  enum {COMPILER,HANDWRITTEN} state = HANDWRITTEN;

  section->tmp_buf=flowgraph;

  /* {{{ order all $handwritten and $compiler symbols */
  for (sym = section->obj->sub_symbol_table->first; sym; sym = sym->next)
  {
    if (sym->name && !strncmp(sym->name,"$compiler",9))
    {
      sym->name[9] = '\0';	/* truncate the symbols inserted for inline assembly */
      nsyms++;
    }
    if (sym->name && !strncmp(sym->name,"$handwritten",12))
    {
      sym->name[12] = '\0';	/* truncate the symbols inserted for inline assembly */
      nsyms++;
    }
  }
  symarr = Malloc(nsyms*sizeof(t_symbol *));
  for (i=0, sym = section->obj->sub_symbol_table->first; sym; sym = sym->next)
    if (sym->name && (!strcmp(sym->name,"$compiler") || !strcmp(sym->name,"$handwritten")))
      symarr[i++] = sym;
  qsort(symarr,(size_t)nsyms,sizeof(t_symbol *),__helper_sort_syms);
  /* }}} */

  /* first allocate all basic blocks (so that they are contiguous in memory,
   * not interleaved with instructions) */
  i = 0;
  SECTION_FOREACH_INS(section,i_ins)
  {
    if ((i < nsyms) && (T_RELOCATABLE(i_ins) == symarr[i]->base))
    {
      if (symarr[i]->name[1] == 'c')	/* faster than a strcmp() */
	state = COMPILER;
      else
	state = HANDWRITTEN;
      i++;
    }
    
    if (GENERIC_EQ(INS_CADDRESS(i_ins),obj->entry))
    {
      INS_ATTRIB(i_ins)|= IF_BBL_LEADER;
      if (INS_ATTRIB(i_ins) & IF_DUMMY) FATAL(("Dummy instruction cannot be a bbl leader!"));
      bbl = BblNew(flowgraph);
      if (!entry) entry=bbl;
      BBL_OLD_ADDRESS(bbl) = INS_OLD_ADDRESS(i_ins);
      BBL_CADDRESS(bbl) = INS_CADDRESS(i_ins);
      INS_BBL(i_ins)=bbl;
      nbbls++;
    }
    else if (INS_ATTRIB(i_ins) & IF_BBL_LEADER)
    {
      if (INS_ATTRIB(i_ins) & IF_DUMMY) FATAL(("Dummy instruction cannot be a bbl leader!"));
      bbl = BblNew(flowgraph);
      BBL_OLD_ADDRESS(bbl) = INS_OLD_ADDRESS(i_ins);
      BBL_CADDRESS(bbl) = INS_CADDRESS(i_ins);
      INS_BBL(i_ins)=bbl;
      nbbls++;
    }
    else
      INS_BBL(i_ins) = NULL;

    if (state == HANDWRITTEN)
      BblSetAttributes(bbl,BBL_IS_HANDWRITTEN_ASSEMBLY);
  }
  
  /* now copy the instructions into the basic blocks */
  SECTION_FOREACH_INS(section,i_ins)
  {
    if (INS_ATTRIB(i_ins) & IF_DUMMY) continue;
    if (INS_BBL(i_ins)) bbl = INS_BBL(i_ins);
    InsAppendToBbl(InsDup(i_ins),bbl);
    INS_BBL(i_ins)=bbl;
    /* following is necessary because InsDup doesn't copy address field */
    INS_OLD_ADDRESS(BBL_INS_LAST(bbl))= INS_OLD_ADDRESS(i_ins);
    INS_CADDRESS(BBL_INS_LAST(bbl))= INS_CADDRESS(i_ins);
    INS_OLD_SIZE(BBL_INS_LAST(bbl))= INS_OLD_SIZE(i_ins);
    current_address = GENERIC_ADD(INS_CADDRESS(i_ins),INS_CSIZE(i_ins));
  }
  
  /* create hell nodes */
  flowgraph->hell_node = BblNew(flowgraph);
  BBL_OLD_ADDRESS(flowgraph->hell_node)=current_address;
  BBL_CADDRESS(flowgraph->hell_node)=current_address;

  flowgraph->call_hell_node =BblNew(flowgraph);
  BBL_OLD_ADDRESS(flowgraph->call_hell_node)=GENERIC_ADD_UINT32(current_address,4);
  BBL_CADDRESS(flowgraph->call_hell_node)=GENERIC_ADD_UINT32(current_address,4);

  flowgraph->unique_entry_node = BblNew(flowgraph);
  BBL_OLD_ADDRESS(flowgraph->unique_entry_node)=GENERIC_ADD_UINT32(current_address,8);
  BBL_CADDRESS(flowgraph->unique_entry_node)=GENERIC_ADD_UINT32(current_address,8);

  flowgraph->swi_hell_node =BblNew(flowgraph);
  BBL_OLD_ADDRESS(flowgraph->swi_hell_node)=GENERIC_ADD_UINT32(current_address,12);
  BBL_CADDRESS(flowgraph->swi_hell_node)=GENERIC_ADD_UINT32(current_address,12);

  /* return_hell_node is created by the function creation routines */
  flowgraph->call_hell_function=FunctionNew(flowgraph->call_hell_node,"--CALL-HELL--",FT_HELL);
  flowgraph->swi_hell_function=FunctionNew(flowgraph->swi_hell_node,"--SWI-HELL--",FT_HELL);
  flowgraph->hell_function=FunctionNew(flowgraph->hell_node,"--HELL--",FT_HELL);
  flowgraph->wrap_function=FunctionNew(flowgraph->unique_entry_node,"--WRAPPER--",FT_HELL);

  flowgraph->exit_hell_node = flowgraph->hell_function->bbl_last; 
  flowgraph->exit_call_hell_node = flowgraph->call_hell_function->bbl_last; 
  flowgraph->exit_swi_hell_node = flowgraph->swi_hell_function->bbl_last; 
  flowgraph->unique_exit_node = flowgraph->wrap_function->bbl_last; 

  /* insert a Normal edge between call_hell_node and return_hell_node */

  CfgEdgeNewCall(flowgraph, flowgraph->call_hell_node, flowgraph->hell_node, flowgraph->exit_call_hell_node,flowgraph->exit_hell_node);


  VERBOSE(1,("Adding arrow from hell! (Entry)"));

  if (!entry) FATAL(("Entry point @G not found!",obj->entry));
  edge=CfgEdgeNewCall(flowgraph,  flowgraph->unique_entry_node,  entry,NULL,NULL);
  
  flowgraph->entries=NewEntry(entry,edge);

  /* remove all relocations coming FROM instructions: address producers will be used instead */
  OBJECT_FOREACH_RELOC_SAFE(section->obj,reloc,tmp)
  {
    /* only remove the reloc from the old instructions, ie. the ones in the linear list */
    if ((RELOCATABLE_TYPE(REL_FROM(reloc))==RT_INS)&&(INS_COPY(REL_FROM(reloc))))
    {
      RelocTableRemoveReloc(section->obj->reloc_table,reloc);
    }
  }

  /* Move the relocations to the newly created basic blocks */
  /* Relocations TO instructions should instead point at the instructions' basic blocks.
   * This is because an instruction that has a relocation pointing to it, is per definition
   * a BBL leader, and we still want the relocation to point to that code sequence (BBL) even
   * if the first instruction is removed later on...
   */

  OBJECT_FOREACH_RELOC(section->obj,reloc)
    if (REL_IS_TO_RELOCATABLE(reloc)&&REL_TO_RELOCATABLE(reloc)&&(RELOCATABLE_TYPE(REL_TO_RELOCATABLE(reloc))==RT_INS)&&(INS_COPY(REL_TO_RELOCATABLE(reloc))))
    {
      t_address offset;
      t_ins * ins = T_INS(REL_TO_RELOCATABLE(reloc));

      offset = GENERIC_SUB(INS_CADDRESS(ins),BBL_CADDRESS(INS_BBL(INS_COPY(ins))));

      if (!IS_DATABBL(INS_BBL(INS_COPY(ins))))
        VERIFY_FASSERT(GENERIC_EQ(offset, NULL_GENERIC),("Symbol @R points to middle of code block\n",reloc));

      RelocSetToRelocatable(reloc,T_RELOCATABLE(INS_BBL(INS_COPY(ins))));
      reloc->addend = GENERIC_ADD(offset,reloc->addend);
    }

  OBJECT_FOREACH_RELOC(section->obj,reloc)
    if (REL_E_IS_TO_RELOCATABLE(reloc)&&REL_E_TO_RELOCATABLE(reloc)&&(RELOCATABLE_TYPE(REL_E_TO_RELOCATABLE(reloc))==RT_INS)&&(INS_COPY(REL_E_TO_RELOCATABLE(reloc))))
    {
      t_address offset;
      t_ins * ins = T_INS(REL_E_TO_RELOCATABLE(reloc));

      offset = GENERIC_SUB(INS_CADDRESS(ins),BBL_CADDRESS(INS_BBL(INS_COPY(ins))));

      if (!IS_DATABBL(INS_BBL(INS_COPY(ins))))
        VERIFY_FASSERT(GENERIC_EQ(offset, NULL_GENERIC),("EXTRA Symbol @R points to middle of code block (@I in @iB)\n",reloc,INS_COPY(ins),INS_BBL(INS_COPY(ins))));

      RelocESetToRelocatable(reloc,T_RELOCATABLE(INS_BBL(INS_COPY(ins))));
      reloc->e_offset = GENERIC_ADD(offset,reloc->e_offset);
    }


  /* also migrate the symbols, as these might come in handy later
   * (e.g. during instrumentation, they can be used to look up functions
   * in the program !LUDO turned this off, because off segfault on armlinux binaries!*/
  {
    t_symbol * sym;
    SYMBOLTABLE_FOREACH_SYMBOL(section->obj->sub_symbol_table,sym)
    {
      if (sym->base && RELOCATABLE_TYPE(sym->base) == RT_INS)
      {
	t_ins * ins = T_INS(sym->base);
	sym->base = T_RELOCATABLE(INS_BBL(ins));
	sym->offset_from_start = GENERIC_ADD(sym->offset_from_start,GENERIC_SUB(INS_CADDRESS(ins),BBL_CADDRESS(INS_BBL(ins))));
      }
    }
  }

  Free(symarr);

  return nbbls;
}
/*}}}*/
/* Cfg related printing {{{ */

void CfgCountBlocks(t_cfg * cfg)
{
  t_uint32 ret=0, ret2=0,ret3=0,ret4=0,ret5=0;
  t_bbl * bbl;
  ANALYSIS_START("Counting basic blocks");
  CFG_FOREACH_BBL(cfg,bbl) 
  {
    ret2++;
    if (IS_DATABBL(bbl)) ret++;
    else if (!BBL_FUN(bbl)) ret5++;
    else if (BBL_IS_LAST(bbl)) ret3++;
    else if (BBL_IS_FIRST(bbl)) ret4++;
  }
  ANALYSIS_END("Counting basic blocks", ("%d data blocks, %d return blocks, %d entry blocks, for %d blocks (%d not in a function)",ret,ret3,ret4,ret2,ret5));
}

/*}}}*/

/*! This function adds an edge to the cfg. It should only be called by the edge
 * constructors and manipulators */

t_cfg_edge * CfgAppendNode(t_cfg * cfg, t_bbl * after, t_bbl * insert, t_uint32 edge_category) 
{ 
  t_cfg_edge * t=T_CFG_EDGE(GraphAppendNode(T_GRAPH(cfg), T_NODE(after), T_NODE(insert), edge_category)); 
  t->cfg=cfg; 
  t->refcount=1; 
  return t;
}
/* Structural optimizations on flowgraphs: {{{ */

/* Remove dead code and data blocks {{{ */

static t_bool MarkSelective(t_node * curr, t_node * parent,void * data) 
{
  /* don't propagate over hell nodes */
  if (IsHellNode(curr)) return TRUE;

  if (parent && IsHellNode(parent)) 
  {
    FATAL(("this should not happen"));
    return TRUE;
  }

  if (!NodeIsMarked2(curr))
    {
      NodeMark2(curr);
      return FALSE;
    }
  else 
    return TRUE;
}

void MarkFrom(t_cfg * cfg, t_bbl * from)
{
  t_bool change = TRUE;
  t_bbl * bbl;
  t_cfg_edge * edge;

  NodeMark2(from);

  while(change)
  {
    change = FALSE;
    CFG_FOREACH_BBL(cfg,bbl)
    {
      if (IsHellNode(bbl)) continue;
      
      if (NodeIsMarked2(bbl))
      {
	BBL_FOREACH_SUCC_EDGE(bbl,edge)
	{
	  if ((EDGE_CAT(edge)==ET_FUNLINK) || (EDGE_CAT(edge)==ET_IPFUNLINK)) continue;
	  if (!CfgEdgeIsBackwardInterproc(edge) || (edge->corr_edge && EdgeIsMarked(edge->corr_edge)))
	  {
	    if (!EdgeIsMarked(edge))
	    {
	      EdgeMark(edge);
	      change = TRUE;
	    }

	    if (CfgEdgeIsForwardInterproc(edge) && (IsHellNode(CFG_EDGE_TAIL(edge))) &&  edge->corr_edge && !NodeIsMarked2(CFG_EDGE_TAIL(edge->corr_edge)))
	    {
	      NodeMark2(CFG_EDGE_TAIL(edge->corr_edge));
	      change = TRUE;
	    }

	    if (!NodeIsMarked2(CFG_EDGE_TAIL(edge)))
	    {
	      NodeMark2(CFG_EDGE_TAIL(edge));
	      change = TRUE;
	    }
	  }
	}
      }
    }
  }
}








static t_bool Reviver(t_object * obj, t_cfg * fg, t_reloc * reloc)
{
  t_bool ret=FALSE;
  t_relocatable * points_to = REL_TO_RELOCATABLE(reloc);

  if (REL_IS_TO_SYMBOL(reloc)) { RelocTablePrint(obj->reloc_table); FATAL(("Still to symbol @R",reloc)); }
  if (points_to)
    {
      if (RELOCATABLE_TYPE(points_to)==RT_BBL)
	{
	  if (!NodeIsMarked2(points_to))
	    {
	      t_ins * ins = BBL_INS_FIRST(points_to);
	      if (!ins || (INS_TYPE(ins) != IT_DATA))
		{
		  /* if the reloc does not cause a hell edge (e.g. relocs from switch tables), there is no need to 
		   * revive the block it points to (it will be reachable because of the switch edge that enters it) */
		  if (reloc->hell)
		  {
#ifdef NEW_CFG_REMOVE_DEAD_CODE_OR_DATA
		    MarkFrom(fg,points_to);
#else
		    GraphDFTraversalWithCheckAndData((t_node *)points_to, MarkSelective, ~(ET_RETURN|ET_COMPENSATING),NULL);
#endif
		    NodeMark2(points_to);
		    ret=TRUE;
		  }
		}
	      else if (!NodeIsMarked2(points_to)) 
	      {
		/* data block in code: mark as reachable, loop in CfgRemoveDeadCodeAndDataBlocks will do the rest */
		NodeMark2(points_to);
		ret=TRUE;
	      }
	    }
	}
      else if ((RELOCATABLE_TYPE(points_to)==RT_SUBSECTION) && 
	       ((SECTION_TYPE(points_to)==DATA_SECTION) || (SECTION_TYPE(points_to)==RO_DATA_SECTION)))
	{
	  if (!NodeIsMarked2(points_to))
	    {
	      t_reloc_ref * ref=RELOCATABLE_REFERS_TO(points_to);
	      NodeMark2(points_to);
	      ret=TRUE;
	      /* Search relocations from points_to */

	      while(ref)
		{
		  if (points_to!=REL_FROM(ref->rel))
		    FATAL(("RELOCS @R, searched with @R section %s",ref->rel,reloc,T_SECTION(points_to)->name));
		  Reviver(obj,fg,ref->rel);
		  ref=ref->next;
		}
	    }
	}
      else if ((RELOCATABLE_TYPE(points_to)==RT_SUBSECTION) && (SECTION_TYPE(points_to)==BSS_SECTION))
	{
	}
      else if (RELOCATABLE_TYPE(points_to)==RT_SECTION)
	{
	}
      else
	FATAL(("IMPLEMENT @R",reloc));
    }
#if 0
  if (REL_E_TO_RELOCATABLE(reloc))
    {
      if (RELOCATABLE_TYPE(REL_E_TO_RELOCATABLE(reloc))==RT_BBL)
	{
	  if (!NodeIsMarked2(T_BBL(REL_E_TO_RELOCATABLE(reloc))))
	    {
	      t_ins * ins = BBL_INS_FIRST(REL_E_TO_RELOCATABLE(reloc));
	      ret=TRUE;
	      /* do not explicitly revive the relocations in data bbls here: they
	       * are handled implicitly in the main loop of CfgRemoveDeadCodeAndDataBlocks() */
	      if (!ins || (INS_TYPE(ins) != IT_DATA))
		{
		  /* if the reloc does not cause a hell edge (e.g. relocs from switch tables), there is no need to 
		   * revive the block it points to (it will be reachable because of the switch edge that enters it) */
		  if (reloc->hell)
		    GraphDFTraversalWithCheckAndData((t_node *)REL_E_TO_RELOCATABLE(reloc), MarkSelective, 0xffff & ~ET_RETURN & ~ET_COMPENSATING,NULL);
		}
	      if (!NodeIsMarked2(T_BBL(REL_E_TO_RELOCATABLE(reloc)))) NodeMark2(T_BBL(REL_E_TO_RELOCATABLE(reloc)));
	    }
	}
      else if ((RELOCATABLE_TYPE(REL_E_TO_RELOCATABLE(reloc))==RT_SUBSECTION) && ((((t_section *)REL_E_TO_RELOCATABLE(reloc))->type==DATA_SECTION) || (((t_section *)REL_E_TO_RELOCATABLE(reloc))->type==RO_DATA_SECTION)))
	{
	  if (!NodeIsMarked2(((t_section *)REL_E_TO_RELOCATABLE(reloc))))
	    {
	      t_reloc_ref * ref=RELOCATABLE_REFERS_TO(REL_E_TO_RELOCATABLE(reloc));
	      NodeMark2(((t_section *)REL_E_TO_RELOCATABLE(reloc)));
	      ret=TRUE;
	      /* Search relocations from REL_E_TO_RELOCATABLE(reloc) */

	      while(ref)
		{
		  if (REL_E_TO_RELOCATABLE(reloc)!=REL_FROM(ref->rel))
		    FATAL(("RELOCS @R, searched with @R section %s",ref->rel,reloc,T_SECTION(REL_E_TO_RELOCATABLE(reloc))->name));

		  if (RELOCATABLE_TYPE(REL_FROM(ref->rel))==RT_SUBSECTION)
		    Reviver(obj,fg,ref->rel);
		  ref=ref->next;
		}


	    }
	}
      else if ((RELOCATABLE_TYPE(REL_E_TO_RELOCATABLE(reloc))==RT_SUBSECTION) && (((t_section *)REL_E_TO_RELOCATABLE(reloc))->type==BSS_SECTION))
	{
	}
      else if (RELOCATABLE_TYPE(REL_E_TO_RELOCATABLE(reloc))==RT_SECTION)
	{
	}
      else
	FATAL(("IMPLEMENT @R",reloc));
    }
#endif
  return ret;
}

void CfgRemoveDeadCodeAndDataBlocks(t_cfg * fg)
{
  t_bbl * safe, * bbl;

  t_function * fun, * tmpfun;
  t_object * obj=CFG_OBJECT(fg);
  t_ins * ins;
  t_bool change=FALSE;
  int unmarked=0, unmarked_c=0, killed_ins=0, killed_bbl=0, killed_fun=0, killed_edge=0;
  int total=0;

  InitNodeMarking2();

  InitEdgeMarking();
  
  /* First of all, mark all reachable blocks from the entry point, not
   * including those that can only be reached from the hell node*/
#ifdef NEW_CFG_REMOVE_DEAD_CODE_OR_DATA
  MarkFrom(fg,fg->entries->entry_bbl);
#else
  GraphDFTraversalWithCheckAndData((t_node *) fg->entries->entry_bbl, MarkSelective, ~(ET_RETURN|ET_COMPENSATING),NULL);
#endif
  NodeMark2(fg->unique_exit_node);

  /* Next, look at all the marked blocks, and mark all things that can be
   * reached by using addresses that are produced somewhere in the marked
   * blocks */
  do
  {
    unmarked=0;
    unmarked_c=0;
    total=0;
    change=FALSE;

    CFG_FOREACH_BBL(fg,bbl)
    {
      total++;
      /* if it is not marked, there's no need to follow this block's
       * relocations */
      if (!NodeIsMarked2(bbl))
      {
        if ((bbl->ins_first)&&(INS_TYPE(bbl->ins_first) != IT_DATA))
        {
          unmarked_c++;
        }
        unmarked++;
      }
      /* Else we need to look at the relocation from this block */
      else
      {

        /* Look at each instruction to find out what relocations there are in
         * each block */

        BBL_FOREACH_INS(bbl,ins)
        {
          t_reloc_ref * ref;

          if ((ref=RELOCATABLE_REFERS_TO(ins)))
          {
            while(ref)
            {
	      /*t_bool oldchange=change;*/
              /* Reviver handles the marking of the refered object and
               * will also recursively mark all things that are referenced
               * in the object */
              change|=Reviver(obj,fg,ref->rel);
	   /*   if (change && (!oldchange)) DiabloPrint(stdout,"Revived by @R\n",ref->rel); */
              ref=ref->next;
            }
          }
        } 
      }
    }
    /*printf("Still %d blocks unmarked (%d code) out of %d total\n",unmarked,unmarked_c,total);*/
  } while(change);


  /* remove all instructions and data that are unmarked.
   * to avoid FATALs, we do this in the following order:
   * 1. remove all unreachable data & instructions but don't kill the blocks
   * 2. remove the (now empty) blocks and their incoming and outgoing edges
   * we have to keep this order to avoid things being killed while there are still
   * relocations pointing to them
   */
  /* {{{ kill instructions and data */
  CFG_FOREACH_BBL(fg,bbl)
  {
    /* Every bbl that is not marked can be killed */
    if (!NodeIsMarked2(bbl))
    {
      t_ins * ins, * tmp;
      BBL_FOREACH_INS_SAFE(bbl,ins, tmp)
      {
	InsKill(ins); /* this automatically kills relocs */
	killed_ins++;
      }
    }
  }

  {
    t_object * sub, * tmp;
    t_uint32 tel;

    /* This checks for dead sections in the subobjects. When found, we remove
     * the relocations in those sections */

    OBJECT_FOREACH_SUBOBJECT(obj,sub,tmp)
    {
      for (tel=0; tel<sub->nrodatas; tel++)
      {
        if ((SECTION_IS_MAPPED(sub->rodata[tel]))
	    && !NodeIsMarked2(sub->rodata[tel]) 
	    && !(SECTION_FLAGS(sub->rodata[tel]) & SECTION_FLAG_KEEP))
        {
	  while (RELOCATABLE_REFERS_TO(sub->rodata[tel]))
	    RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFERS_TO(sub->rodata[tel])->rel);

	  /* Set the section size to 0 */
	  SECTION_CSIZE(sub->rodata[tel])=GENERIC_SUB(SECTION_CSIZE(sub->rodata[tel]),SECTION_CSIZE(sub->rodata[tel]));
	  SECTION_OLD_SIZE(sub->rodata[tel])=GENERIC_SUB(SECTION_OLD_SIZE(sub->rodata[tel]),SECTION_OLD_SIZE(sub->rodata[tel]));
	  SECTION_ALIGN(sub->rodata[tel])=GENERIC_SUB(SECTION_OLD_SIZE(sub->rodata[tel]),SECTION_OLD_SIZE(sub->rodata[tel]));
	}
      }

      for (tel=0; tel<sub->ndatas; tel++)
      {
        if ((SECTION_IS_MAPPED(sub->data[tel]))
	    && !NodeIsMarked2(sub->data[tel]) 
	    && !(SECTION_FLAGS(sub->data[tel]) & SECTION_FLAG_KEEP))
        {
	  while (RELOCATABLE_REFERS_TO(sub->data[tel]))
	    RelocTableRemoveReloc(obj->reloc_table,RELOCATABLE_REFERS_TO(sub->data[tel])->rel);

	  /* Set the section size to 0 */
	  SECTION_CSIZE(sub->data[tel])=GENERIC_SUB(SECTION_CSIZE(sub->data[tel]),SECTION_CSIZE(sub->data[tel]));
	  SECTION_OLD_SIZE(sub->data[tel])=GENERIC_SUB(SECTION_OLD_SIZE(sub->data[tel]),SECTION_OLD_SIZE(sub->data[tel]));
	  SECTION_ALIGN(sub->data[tel])=GENERIC_SUB(SECTION_OLD_SIZE(sub->data[tel]),SECTION_OLD_SIZE(sub->data[tel]));
	}
      }
    }
  } /* }}} */
  /* {{{ remove empty blocks and edges */
  CFG_FOREACH_BBL_SAFE(fg,bbl,safe)
  {
    t_cfg_edge * edge, * tmp;

    if (NodeIsMarked2(bbl)) continue;
    if (IsHellNode(bbl)) continue;
    if (bbl == bbl->cfg->unique_entry_node || bbl == bbl->cfg->unique_exit_node) continue;

    if (BBL_NINS(bbl)) FATAL(("@iB unmarked but apparently not empty",bbl));

    BBL_FOREACH_SUCC_EDGE_SAFE(bbl,edge,tmp)
    {
#ifdef NEW_CFG_REMOVE_DEAD_CODE_OR_DATA
      if (edge->corr_edge) 
      {
	if (CfgEdgeIsForwardInterproc(edge))
	{
	  if (edge->corr_edge == tmp) tmp = T_CFG_EDGE(EDGE_SUCC_NEXT(tmp));
	  CfgEdgeKill(edge->corr_edge);
	  killed_edge++;
	}
	else
	  edge->corr_edge->corr_edge=NULL;
      }
      if (EDGE_CAT(edge)==ET_RETURN)
      {
	t_cfg_edge * j_edge=NULL;
	DiabloPrint(stdout,"Killing 2 @E\n",edge);
	BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),j_edge)
	{
	  /*DiabloPrint(stdout,"Looking at @E\n",j_edge);*/
	  if ((EDGE_CAT(j_edge)==ET_FUNLINK) || (EDGE_CAT(j_edge)==ET_IPFUNLINK))
	    break;
	}

	if (j_edge) 
	{
	  if (j_edge == tmp) tmp = T_CFG_EDGE(EDGE_SUCC_NEXT(tmp));
	  CfgEdgeKill(j_edge);
	  killed_edge++;
	}
      }

      CfgEdgeKill(edge);
      killed_edge++;
#else
      if (!CfgEdgeIsInterproc(edge) || CfgEdgeIsForwardInterproc(edge) || !edge->corr_edge)
      {
	if (edge->corr_edge)
	{
	  if (edge->corr_edge == tmp) tmp = T_CFG_EDGE(EDGE_SUCC_NEXT(tmp));
	  CfgEdgeKill(edge->corr_edge);
	  killed_edge++;
	}
	CfgEdgeKill(edge);
	killed_edge++;
      }
#endif
    }
    BBL_FOREACH_PRED_EDGE_SAFE(bbl,edge,tmp)
    {
#ifdef NEW_CFG_REMOVE_DEAD_CODE_OR_DATA
      if (edge->corr_edge) 
      {
	if (CfgEdgeIsForwardInterproc(edge))
	{
	  if (edge->corr_edge == tmp) tmp = T_CFG_EDGE(EDGE_PRED_NEXT(tmp));
	  CfgEdgeKill(edge->corr_edge);
	  killed_edge++;
	}
	else
	  edge->corr_edge->corr_edge=NULL;
      }
      if (EDGE_CAT(edge)==ET_RETURN)
      {
	t_cfg_edge * j_edge=NULL;
	DiabloPrint(stdout,"Killing 1 @E\n",edge);
	BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),j_edge)
	{
	/*DiabloPrint(stdout,"Looking at @E\n",j_edge);*/
	  if ((EDGE_CAT(j_edge)==ET_FUNLINK) || (EDGE_CAT(j_edge)==ET_IPFUNLINK))
	    break;
	}

	if (j_edge) 
	{
	  if (j_edge == tmp) tmp = T_CFG_EDGE(EDGE_PRED_NEXT(tmp));
	  CfgEdgeKill(j_edge);
	  killed_edge++;
	}
      }
      CfgEdgeKill(edge);
      killed_edge++;
#else
      if (!CfgEdgeIsInterproc(edge) || CfgEdgeIsForwardInterproc(edge) || !edge->corr_edge)
      {
	if (edge->corr_edge)
	{
	  if (edge->corr_edge == tmp) tmp = T_CFG_EDGE(EDGE_PRED_NEXT(tmp));
	  CfgEdgeKill(edge->corr_edge);
	  killed_edge++;
	}
	CfgEdgeKill(edge);
	killed_edge++;
      }
#endif
    }

    if (!BBL_PRED_FIRST(bbl) && !BBL_SUCC_FIRST(bbl))
    {
      BblKill(bbl);
      killed_bbl++;
    }
  }
  /* clean up empty functions */
  CFG_FOREACH_FUN_SAFE(fg,fun,tmpfun)
    if (FUN_BBL_FIRST(fun) == NULL)
    {
      FunctionKill(fun);
      killed_fun++;
    }
  /* }}} */

  /* a number of return blocks (e.g. the return block of exit()) are in fact unreachable,
   * but may not be killed because they are the head of return or compensating edges 
   * whose corresponding edge is still live. We'll mark these blocks as unreachable with
   * a special flag, as some analyses can use this information to be more accurate (e.g. 
   * stack height analysis) 
   * The blocks in question are those that are not reachable from already marked exit blocks by 
   * following only compensating edges */

#if 0
 {
   t_cfg_edge * edge;
   CFG_FOREACH_FUN(fg,fun)
     {
       t_bbl * exitblock = FunctionGetExitBlock(fun);
       if (!exitblock || !NodeIsMarked2(exitblock)) continue;
       if (FUN_IS_HELL(fun)) continue;
       GraphDFTraversalWithCheckAndData((t_node *)exitblock,MarkSelective,ET_COMPENSATING,NULL);
     }
   /* mark as reachable all those exit blocks reachable from the exit blocks of the hell functions */
   BBL_FOREACH_SUCC_EDGE(fg->exit_hell_node,edge)
     if (EDGE_CAT(edge) == ET_COMPENSATING)
       GraphDFTraversalWithCheckAndData(EDGE_TAIL(edge),MarkSelective,ET_COMPENSATING,NULL);
   BBL_FOREACH_SUCC_EDGE(fg->exit_call_hell_node,edge)
     if (EDGE_CAT(edge) == ET_COMPENSATING)
       GraphDFTraversalWithCheckAndData(EDGE_TAIL(edge),MarkSelective,ET_COMPENSATING,NULL);
   
   CFG_FOREACH_FUN(fg,fun)
     {
       t_bbl * exitblock = FunctionGetExitBlock(fun);
       if (exitblock && !NodeIsMarked2(exitblock))
	 BblSetAttributes(exitblock,BBL_UNREACHABLE);
     }
 }
#endif

#if 0 /* PROBLEMATIC ???*/
  {
    t_bool change = TRUE;
    t_bbl * bbl;
    t_cfg_edge * edge;
    
    InitNodeMarking();
    InitEdgeMarking();
    NodeMark(fg->entries->entry_bbl);
    
    while(change)
      {
	change = FALSE;
	CFG_FOREACH_BBL(fg,bbl)
	  {
	    if (NodeIsMarked(bbl))
	      BBL_FOREACH_SUCC_EDGE(bbl,edge)
	      {
		if (!CfgEdgeIsBackwardInterproc(edge) || (edge->corr_edge && EdgeIsMarked(edge->corr_edge)))
		  {
		    if (!EdgeIsMarked(edge))
		      {
			EdgeMark(edge);
			change = TRUE;
		      }
		    if (!NodeIsMarked(CFG_EDGE_TAIL(edge)))
		      {
			NodeMark(CFG_EDGE_TAIL(edge));
			change = TRUE;
		      }
		  }
	      }
	  }
      }

    CFG_FOREACH_BBL(fg,bbl)
      {
	if (!NodeIsMarked(bbl) && !IS_DATABBL(bbl) && BBL_FUN(bbl) && strcmp(FUN_NAME(BBL_FUN(bbl)),"--WRAPPER--") && !IsHellNode(bbl))
	  {	
	    t_ins * tmp_ins;
	    t_cfg_edge * tmp_edge;

	   

	    DiabloPrint(stdout,"FORGOT TO KILL @B\n",bbl);

	    BBL_FOREACH_INS_SAFE(bbl,ins,tmp_ins)
	      InsKill(ins);
	    BBL_FOREACH_PRED_EDGE_SAFE(bbl,edge,tmp_edge)
	      {
		if (edge->corr_edge)
		  edge->corr_edge->corr_edge=NULL;
		CfgEdgeKill(edge);
	      }
	    BBL_FOREACH_SUCC_EDGE_SAFE(bbl,edge,tmp_edge)
	      {
		if (edge->corr_edge)
		  edge->corr_edge->corr_edge=NULL;
		CfgEdgeKill(edge);
	      }
	    BblKill(bbl);
	  }
      }
  }
  
  
#endif

  VERBOSE(0,("removed %d instructions, %d bbls, %d edges, %d functions from graph", killed_ins, killed_bbl, killed_edge, killed_fun));
}


/* CfgRemoveDeadCode {{{ */
void CfgRemoveDeadCode(t_cfg * fg)
{
  t_bbl * safe, * bbl;
  t_function * fun, * tmpfun;
  t_bool change=FALSE;
  t_cfg_edge * edge;
  int killed_ins=0, killed_bbl=0, killed_fun=0, killed_edge=0;

  InitNodeMarking2();

  /* First of all, mark all reachable blocks from the entry point or from the hell nodes */

  NodeMark2(fg->entries->entry_bbl);
  NodeMark2(fg->hell_node);
  NodeMark2(fg->call_hell_node);
  NodeMark2(fg->exit_hell_node);
  NodeMark2(fg->exit_call_hell_node);

  do
    {
      change = FALSE;
      CFG_FOREACH_BBL(fg,bbl)
	{
	  if (NodeIsMarked2(bbl))
	    {
	      BBL_FOREACH_SUCC_EDGE(bbl,edge)
		{
		  t_bbl * tail = CFG_EDGE_TAIL(edge);
		    if (EDGE_CAT(edge)==ET_RETURN) continue;
		    if (!tail) continue;
		    if (!NodeIsMarked2(tail))
		      {
			NodeMark2(tail);
			change = TRUE;
		      }
		}
	    }
	}
    }
  while (change);

  NodeMark2(fg->unique_exit_node);

  /* remove all instructions and data that are unmarked.
   * to avoid FATALs, we do this in the following order:
   * 1. remove all unreachable data & instructions but don't kill the blocks
   * 2. remove the (now empty) blocks and their incoming and outgoing edges
   * we have to keep this order to avoid things being killed while there are still
   * relocations pointing to them
   */
  /* {{{ kill instructions and data */
  CFG_FOREACH_BBL(fg,bbl)
  {
    /* Every bbl that is not marked can be killed */
    if (!NodeIsMarked2(bbl))
    {
      t_ins * ins, * tmp;
      BBL_FOREACH_INS_SAFE(bbl,ins, tmp)
      {
	InsKill(ins); /* this automatically kills relocs */
	killed_ins++;
      }
    }
  }
  /* }}} */

  /* {{{ remove empty blocks and edges */
  CFG_FOREACH_BBL_SAFE(fg,bbl,safe)
  {
    t_cfg_edge * edge, * tmp;

    if (NodeIsMarked2(bbl)) continue;
    if (IsHellNode(bbl)) continue;
    if (bbl == bbl->cfg->unique_entry_node || bbl == bbl->cfg->unique_exit_node) continue;

    if (BBL_NINS(bbl)) FATAL(("@iB unmarked but apparently not empty",bbl));

    BBL_FOREACH_SUCC_EDGE_SAFE(bbl,edge,tmp)
    {
      if (!CfgEdgeIsInterproc(edge) || CfgEdgeIsForwardInterproc(edge) || !edge->corr_edge)
      {
	if (edge->corr_edge)
	{
	  if (edge->corr_edge == tmp) tmp = T_CFG_EDGE(EDGE_SUCC_NEXT(tmp));
	  CfgEdgeKill(edge->corr_edge);
	  killed_edge++;
	}
	CfgEdgeKill(edge);
	killed_edge++;
      }
    }
    BBL_FOREACH_PRED_EDGE_SAFE(bbl,edge,tmp)
    {
      if (!CfgEdgeIsInterproc(edge) || CfgEdgeIsForwardInterproc(edge) || !edge->corr_edge)
      {
	if (edge->corr_edge)
	{
	  if (edge->corr_edge == tmp) tmp = T_CFG_EDGE(EDGE_SUCC_NEXT(tmp));
	  CfgEdgeKill(edge->corr_edge);
	  killed_edge++;
	}
	CfgEdgeKill(edge);
	killed_edge++;
      }
    }
    if (!BBL_PRED_FIRST(bbl) && !BBL_SUCC_FIRST(bbl))
    {
      BblKill(bbl);
      killed_bbl++;
    }
  }

  /* clean up empty functions */
  CFG_FOREACH_FUN_SAFE(fg,fun,tmpfun)
    if (FUN_BBL_FIRST(fun) == NULL)
    {
      FunctionKill(fun);
      killed_fun++;
    }
  /* }}} */

  VERBOSE(0,("removed %d instructions, %d bbls, %d edges, %d functions from graph", killed_ins, killed_bbl, killed_edge, killed_fun));
} /* }}} */

/* find calls to functions that don't return and adapt the control flow graph to model this */
/* return TRUE if any changes were made to the cfg (if so, re-running CfgRemoveDeadCodeAndDataBlocks
 * may result in some extra compaction) */
t_bool CfgPatchForNonReturningFunctions(t_cfg * cfg)
{
  t_function * fun;
  t_bool change = FALSE;
  int count = 0;
  t_cfg_edge * edge;

  CFG_FOREACH_FUN(cfg,fun)
  {
    t_bbl * exitblock = FunctionGetExitBlock(fun);
    t_cfg_edge * edge;

    if (!exitblock || !(BBL_ATTRIB(exitblock) & BBL_UNREACHABLE)) continue;
    if (FUN_IS_HELL(fun)) continue;

    BBL_FOREACH_SUCC_EDGE(exitblock,edge)
    {
      t_cfg_edge * funlink, * edge2;
      t_bbl * callsite, * new, * returnsite;

      if (EDGE_CAT(edge) != ET_RETURN) continue;
      funlink = FindFunLinkForReturn(edge);
      if (!funlink) continue;
      callsite = CFG_EDGE_HEAD(funlink);
      returnsite = CFG_EDGE_TAIL(funlink);

      /* only go through with this if the call site has no other edges out */
      BBL_FOREACH_SUCC_EDGE(callsite,edge2)
	if (!EdgeTestCategoryOr(edge2,ET_CALL|ET_FUNLINK))
	  break;
      if (edge2) continue;

      /* check if it hasn't already been done */
      if (!BBL_INS_FIRST(returnsite) && !BBL_SUCC_FIRST(returnsite)) continue;

      /*if (count >= global_options.debugcounter) continue;*/

      new = BblNew(cfg);
      BblInsertInFunction(new,BBL_FUN(returnsite));
      
      /* move the return edge and the funlink edge to the new block */
      CFG_EDGE_TAIL(edge) = new;
      if (EDGE_PRED_PREV(edge))
	EDGE_PRED_NEXT(EDGE_PRED_PREV(edge)) = EDGE_PRED_NEXT(edge);
      else
	BBL_PRED_FIRST(returnsite) = T_CFG_EDGE(EDGE_PRED_NEXT(edge));
      if (EDGE_PRED_NEXT(edge))
	EDGE_PRED_PREV(EDGE_PRED_NEXT(edge)) = EDGE_PRED_PREV(edge);
      else
	BBL_PRED_LAST(returnsite) = T_CFG_EDGE(EDGE_PRED_PREV(edge));
      BBL_PRED_FIRST(new) = BBL_PRED_LAST(new) = edge;
      EDGE_PRED_PREV(edge) = EDGE_PRED_NEXT(edge) = NULL;

      CFG_EDGE_TAIL(funlink) = new;
      if (EDGE_PRED_PREV(funlink))
	EDGE_PRED_NEXT(EDGE_PRED_PREV(funlink)) = EDGE_PRED_NEXT(funlink);
      else
	BBL_PRED_FIRST(returnsite) = T_CFG_EDGE(EDGE_PRED_NEXT(funlink));
      if (EDGE_PRED_NEXT(funlink))
	EDGE_PRED_PREV(EDGE_PRED_NEXT(funlink)) = EDGE_PRED_PREV(funlink);
      else
	BBL_PRED_LAST(returnsite) = T_CFG_EDGE(EDGE_PRED_PREV(funlink));
      BBL_PRED_FIRST(new) = funlink;
      EDGE_PRED_PREV(funlink) = NULL;
      EDGE_PRED_NEXT(funlink) = (t_edge *) edge;
      EDGE_PRED_PREV(edge) = (t_edge *) funlink;

      /*DiabloPrint(stdout,"RETURN SITE %s\n",FUN_NAME(BBL_FUN(returnsite)));*/
      /*DiabloPrint(stdout,"call @ieB\n",callsite);*/
      /*DiabloPrint(stdout,"return @ieB\n",returnsite);*/
      /*DiabloPrint(stdout,"new @ieB\n",new);*/
      /*DiabloPrint(stdout,"exit block @ieB\n",exitblock);*/

      change = TRUE;
      count++;
    }
  }
  DiabloPrint(stdout,"patched %d return sites of non-returning functions\n",count);

  BBL_FOREACH_SUCC_EDGE(cfg->exit_hell_node,edge)
    {
      if (CFG_EDGE_TAIL(edge)==CFG_EDGE_HEAD(edge))
	{
	  /*	  DiabloPrint(stdout,"EDGE @E\n",edge);*/
	  if (edge->corr_edge)
	    {
	      /*	      DiabloPrint(stdout,"EDGE @E\n FROM @B\n TO @B\n",edge->corr_edge,CFG_EDGE_HEAD(edge->corr_edge),CFG_EDGE_TAIL(edge->corr_edge));*/
	      edge->corr_edge->corr_edge=NULL;
	      CfgEdgeKill(edge);
	    }
	}
    }
  return change;
}

/*}}}*/

/* CfgRemoveUselessConditionalJumps {{{ */
void CfgRemoveUselessConditionalJumps(t_cfg * cfg)
{
  t_bbl * bbl;
  CFG_FOREACH_BBL(cfg,bbl)
  {
    if ((NODE_SUCC_FIRST(bbl))
        && (EDGE_SUCC_NEXT(NODE_SUCC_FIRST(bbl)))
        && (!EDGE_SUCC_NEXT(EDGE_SUCC_NEXT(NODE_SUCC_FIRST(bbl)))))
    {
      if (EDGE_TAIL(NODE_SUCC_FIRST(bbl))==EDGE_TAIL(EDGE_SUCC_NEXT(NODE_SUCC_FIRST(bbl))))
      {
        if ((EDGE_CAT(NODE_SUCC_FIRST(bbl))==ET_JUMP)|| (EDGE_CAT(EDGE_SUCC_NEXT(NODE_SUCC_FIRST(bbl))) == ET_JUMP))
          if ((EDGE_CAT(NODE_SUCC_FIRST(bbl))==ET_FALLTHROUGH)|| (EDGE_CAT(EDGE_SUCC_NEXT(NODE_SUCC_FIRST(bbl))) == ET_FALLTHROUGH))
          {
            /* this code should work for both architectures with and without delay slots */
            if(BBL_INS_LAST(bbl)&&INS_TYPE(BBL_INS_LAST(bbl))==IT_BRANCH)
              InsKill(BBL_INS_LAST(bbl));
            else if(INS_PREV(BBL_INS_LAST(bbl)) && INS_TYPE(INS_PREV(BBL_INS_LAST(bbl)))==IT_BRANCH)
              InsKill(INS_PREV(BBL_INS_LAST(bbl)));
            else FATAL(("Could not find jump instruction in block"));

            if (EDGE_CAT(NODE_SUCC_FIRST(bbl))==ET_JUMP)
              CfgEdgeKill(BBL_SUCC_FIRST(bbl));
            else
              CfgEdgeKill(CFG_EDGE_SUCC_NEXT(BBL_SUCC_FIRST(bbl)));
          }
      }
    }
  }
}
/*}}}*/

/* CfgRemoveEmptyBlocks {{{ */

void CfgRemoveEmptyBlocks(t_cfg * cfg)
{
  t_bbl * bbl, * tmp;
  CFG_FOREACH_BBL_SAFE(cfg,bbl,tmp)
  {
    /* skip return blocks */
    if (BBL_FUN(bbl) && (FUN_BBL_LAST(BBL_FUN(bbl))==bbl)) continue;
    /* skip hell nodes */
    if (bbl==cfg->hell_node) continue;
    if (bbl==cfg->call_hell_node) continue;
    /* Fix me, we can change this */
    if (T_RELOCATABLE(bbl)->refedby)  continue;
    if ((NODE_PRED_FIRST(bbl))&& (NODE_SUCC_FIRST(bbl)))
      if (BBL_NINS(bbl)==0) 
      {
        t_cfg_edge * i_edge, * safe;
        t_bbl * tail=NULL;
        t_bbl * head;
        t_uint32 type;
        t_bool ip=FALSE;

        /* the block should have exactly one successor edge, of type fallthrough */
        BBL_FOREACH_SUCC_EDGE_SAFE(bbl,i_edge,safe)
        {
          if (tail) FATAL(("Impossible! : \nEdge:@E\n en tail:@B",i_edge,tail));
          if (EDGE_CAT(i_edge)==ET_IPFALLTHRU)
          {
            ip=TRUE;
          }
          else if (EDGE_CAT(i_edge)!=ET_FALLTHROUGH) 
          {
	    BblPrint(bbl,0x1f);

            DiabloPrint(stdout,"@iB --- @E\n",bbl,i_edge);
            FATAL(("Impossible 2"));
          }
          tail=CFG_EDGE_TAIL(i_edge);
          if (!ip) CfgEdgeKill(i_edge);
        }
        if (ip) continue;
        if (!tail) FATAL(("Impossible 3"));
        BBL_FOREACH_PRED_EDGE_SAFE(bbl,i_edge,safe)
        {
          type=EDGE_CAT(i_edge);
          head=CFG_EDGE_HEAD(i_edge);
          if (type==ET_RETURN)
          {
            CfgEdgeChangeReturnForCall(i_edge->corr_edge,tail);
          }
          else if (type == ET_CALL)
          {
            t_cfg_edge * return_edge = i_edge->corr_edge;
            t_bbl * return_block = CFG_EDGE_HEAD(return_edge);
            t_bbl * return_site = CFG_EDGE_TAIL(return_edge);

            CfgEdgeKill(i_edge->corr_edge);
            CfgEdgeKill(i_edge);

            CfgEdgeNewCall(bbl->fun->cfg,head,tail,return_site,return_block);
          }
          else if (CfgEdgeIsForwardInterproc(i_edge))
          {
            t_cfg_edge * new_edge;

            CfgEdgeKill(i_edge->corr_edge);
            CfgEdgeKill(i_edge);
            new_edge = CfgEdgeNew(bbl->fun->cfg,head,tail,type);
            CfgEdgeNewCompensating(bbl->fun->cfg,new_edge);
          }
          else if (CfgEdgeIsInterproc(i_edge))
          {
            /* this shouldn't happen: all forward interprocedural cases are handled,
             * and return edges are also handled. the only remaining type is compensating
             * edges and those should only arrive in return blocks. return blocks are 
             * skipped so if we fall into such an edge here, something is definitely wrong
             */
            FATAL(("Shouldn't have this kind of edge here! (type = %d)",type));
          }
          else if (CFG_EDGE_FLAGS(i_edge)& EF_FROM_SWITCH_TABLE)
	  {
            CfgEdgeKill(i_edge);
            i_edge=CfgEdgeNew(bbl->fun->cfg,head,tail,type);
	    CFG_EDGE_FLAGS(i_edge)|=EF_FROM_SWITCH_TABLE;
	  }
	  else
          {
            /* regular edges. just kill and replace */
	    t_uint32 count = EDGE_EXEC_COUNT(i_edge);
            CfgEdgeKill(i_edge);
            EDGE_EXEC_COUNT(CfgEdgeNew(bbl->fun->cfg,head,tail,type))=count;

          }
        }
	
	BblKill(bbl);
      }
  }
}
/*}}}*/

/* CfgMoveInsDown {{{ */

void CfgMoveInsDown(t_cfg * cfg)
{
  t_bbl * bbl;
  CFG_FOREACH_BBL(cfg,bbl)
  {
    if (BBL_FUN(bbl) && (FUN_BBL_LAST(BBL_FUN(bbl))==bbl)) continue;
    if (bbl==cfg->hell_node) continue;
    if (bbl==cfg->call_hell_node) continue;
    if (NODE_PRED_FIRST(bbl) && EDGE_PRED_NEXT(NODE_PRED_FIRST(bbl)))
    {
      t_cfg_edge * edge;
      t_ins * i_ins, * j_ins;
      t_bbl * pred, * first=CFG_EDGE_HEAD(NODE_PRED_FIRST(bbl));
      t_bool ok=TRUE,ins_can_move;
      BBL_FOREACH_PRED_EDGE(bbl,edge)
      {
        pred=CFG_EDGE_HEAD(edge);
        if (EDGE_SUCC_NEXT(NODE_SUCC_FIRST(pred)))
        {
	  if ((!EDGE_SUCC_NEXT(EDGE_SUCC_NEXT(NODE_SUCC_FIRST(pred))))&&((EDGE_CAT(EDGE_SUCC_NEXT(NODE_SUCC_FIRST(pred)))==ET_FUNLINK)||(EDGE_CAT(NODE_SUCC_FIRST(pred))==ET_FUNLINK))&&(EDGE_CAT(edge)==ET_CALL))
	  {
	    /* Keep on trying for calls */
	    if (global_options.rely_on_calling_conventions)
	    {
	      ok=FALSE;
	      break;
	    }
	  }
	  else
	  {
	    ok=FALSE;
	    break;
	  }
        }
      }

      if (ok)
      {
        t_bool load_seen=FALSE, store_seen=FALSE;
        t_regset use1=RegsetNew(),  def1=RegsetNew(), tmp=RegsetNew();
        ok=FALSE;
        BBL_FOREACH_INS_R(first,i_ins)
        {
          ins_can_move=TRUE;

          if ((INS_TYPE(i_ins)==IT_LOAD) || (INS_TYPE(i_ins)==IT_LOAD_MULTIPLE))
            if (store_seen) ins_can_move=FALSE;
          if ((INS_TYPE(i_ins)==IT_STORE) || (INS_TYPE(i_ins)==IT_STORE_MULTIPLE))
            if (store_seen||load_seen) ins_can_move=FALSE;
          if ((INS_TYPE(i_ins)==IT_LOAD) || (INS_TYPE(i_ins)==IT_LOAD_MULTIPLE))
            load_seen=TRUE;
          if ((INS_TYPE(i_ins)==IT_STORE) || (INS_TYPE(i_ins)==IT_STORE_MULTIPLE))
            store_seen=TRUE; 

          RegsetDup(tmp,INS_REGS_DEF(i_ins));
          RegsetSetIntersect(tmp,use1);
          if (!RegsetIsEmpty(tmp)) {ins_can_move=FALSE;}
          RegsetDup(tmp,INS_REGS_USE(i_ins));
          RegsetSetIntersect(tmp,def1);
          if (!RegsetIsEmpty(tmp)) {ins_can_move=FALSE;}
          RegsetSetUnion(use1,INS_REGS_USE(i_ins));
          RegsetSetUnion(def1,INS_REGS_DEF(i_ins));
          RegsetSetUnion(use1,INS_REGS_DEF(i_ins));
          if (!ins_can_move) continue;

          ok=TRUE;
          BBL_FOREACH_PRED_EDGE(bbl,edge)
          {
            if (CFG_EDGE_HEAD(edge)==first) continue;
            j_ins=InsFindMovableDownEqualInsInBbl(i_ins, CFG_EDGE_HEAD(edge));
            if (!j_ins) { ok=FALSE; break; }
          }
          if (ok)
          {
            BBL_FOREACH_PRED_EDGE(bbl,edge)
            {
              if (CFG_EDGE_HEAD(edge)==first) continue;
              j_ins=InsFindMovableDownEqualInsInBbl(i_ins, CFG_EDGE_HEAD(edge));
              VERIFY_FASSERT(j_ins,("this should never happen!"));
              InsKill(j_ins);
            }

            {
              t_ins * new_ins = InsDup(i_ins);
              InsPrependToBbl(new_ins,bbl);
              InsKill(i_ins);

            }
            break;
          }
        }

        if (ok)
        {
          BBL_FOREACH_PRED_EDGE(bbl,edge)
          {
            pred=CFG_EDGE_HEAD(edge);
          }
        }
      }
    }
  }
}
/*}}}*/

/*}}}*/

/*! Detection of infinite loops, in a very naive way. We simply look for loops
 * and check if there isn't any path that could escape from the loop.
 * Return value is true if any infinite loop is found and the backedges
 * corresponding to the infinite loops are returned in a linked list */
t_bool CfgDetectSimpleInfiniteLoops(t_cfg * cfg, ll** inf_list)
{
  t_loop * loop;
  t_cfg_edge * backedge = NULL;
  t_bool ret = FALSE;
  t_bbl * header, * footer, *ibbl;
  t_uint32 loops = 0;
  t_function * ifun;

  CfgMarkFunsWithoutIncomingJumps(cfg);
  /* Initialise everything for a new calculation */
/*  printf("Sizeof = %d\n",sizeof(*ibbl));*/
  CFG_FOREACH_BBL(cfg,ibbl)
  {
    LoopBblCleanup(ibbl);
    BBL_LOOPS(ibbl) = NULL;
/*    VERBOSE(0,("%d ",((*((int*)(&ibbl)) & 0xf0) >> 4)));*/
  }
  CFG_FOREACH_FUN(cfg,ifun)
  {
    t_loopref * free_me;
    while(FUN_LOOPS(ifun))
    {
      free_me = FUN_LOOPS(ifun);
      FUN_LOOPS(ifun) = LLNext(FUN_LOOPS(ifun));
      Free(free_me);
    }
  }
/*  printf("\n");*/

  while(CFG_LOOP_FIRST(cfg))
    LoopKill(&(CFG_LOOP_FIRST(cfg)));
  CFG_LOOP_LAST(cfg) = NULL;

  CFG_FOREACH_EDGE(cfg,backedge)
  {
    if(EDGE_CAT(backedge) & (ET_FUNLINK|ET_IPFUNLINK|ET_CALL|ET_IPJUMP|ET_RETURN|ET_COMPENSATING)) continue;
    if(CFG_EDGE_TAIL(backedge) == FunctionGetExitBlock(BBL_FUN(CFG_EDGE_TAIL(backedge)))) continue;
    header = CFG_EDGE_TAIL(backedge);
    footer = CFG_EDGE_HEAD(backedge);
    if(IsHellNode(header)) continue;

    if (BblDominates(header,footer))
      /* We have a loop */
    {
      if(EDGE_CAT(backedge) == ET_SWITCH)
      {
	t_loop *i_loop;
	CFG_FOREACH_LOOP(cfg,i_loop)
	{
	  if(CFG_EDGE_HEAD(LOOP_BACKEDGES(i_loop)->data) == footer && CFG_EDGE_TAIL(LOOP_BACKEDGES(i_loop)->data) == header) 
	    break;
	}
	if (i_loop) continue;
      }

/*      VERBOSE(0,("--------------------------------------------\nLOOP: backedge! @E\n",backedge));*/
      loop = LoopNew(cfg,backedge);
      loops++;

      if (!loop) FATAL(("No loop"));

#ifdef DEBUG_LOOP
      LoopPrint(loop,0);
#endif
    }
  }
  VERBOSE(0,("%d backedges",loops));

  LoopFindNestedLoops(cfg);
  LoopMergeLoopsWithSameHeader(cfg);

  loops=0;
  CFG_FOREACH_LOOP(cfg,loop) {
    loops++;
    /* If we've found an infinite loop, put it on the list */
    if(LOOP_INF(loop))
    {
      data_ll_elem * backedges = LOOP_BACKEDGES(loop);
      data_ll_elem *insert = (data_ll_elem*) Calloc(1,sizeof(data_ll_elem));
      ret = TRUE;
      if(LLNext(backedges)) FATAL(("Didn't expect to have multiple backedges in an infinite loop"));
      insert->data = backedges->data;
      LLAppend(NULL,insert,inf_list);
#ifdef DEBUG_LOOP
      VERBOSE(0,("Infinite loop with edge @E\n",insert->data));
#endif
    }

  }
  VERBOSE(0,("Number of loops: %d",loops));
  return ret;
}

/* Verify correctness of the cfg {{{ */
static void CfgCheckDuplicateFallthroughsIn(t_cfg * cfg)
{
  t_bbl * j_bbl,*i_bbl;
  t_cfg_edge * i_edge;
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    j_bbl = NULL;
    BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
    {
      if (EDGE_CAT(i_edge)==ET_FUNLINK ||
          EDGE_CAT(i_edge)==ET_SWILINK ||
          EDGE_CAT(i_edge)==ET_IPFUNLINK ||
          EDGE_CAT(i_edge)==ET_IPSWILINK ||
          EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
          EDGE_CAT(i_edge)==ET_IPFALLTHRU
         )
      {
        if (j_bbl)
          if (j_bbl!=CFG_EDGE_HEAD(i_edge))
          {
            FATAL(("two pred edges @B, @E\n",j_bbl,i_edge));
          }

        j_bbl = CFG_EDGE_HEAD(i_edge);
      }
    }
  }
}

static void CfgCheckDuplicateFallthroughsOut(t_cfg * cfg)
{
  t_bbl * j_bbl,*i_bbl;
  t_cfg_edge * i_edge;
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    j_bbl = NULL;
    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      if (EDGE_CAT(i_edge)==ET_FUNLINK ||
          EDGE_CAT(i_edge)==ET_SWILINK ||
          EDGE_CAT(i_edge)==ET_IPFUNLINK ||
          EDGE_CAT(i_edge)==ET_IPSWILINK ||
          EDGE_CAT(i_edge)==ET_FALLTHROUGH ||
          EDGE_CAT(i_edge)==ET_IPFALLTHRU
         )
      {
        if (j_bbl)
          if (j_bbl!=CFG_EDGE_TAIL(i_edge))
          {
            FATAL(("two succ edges @iB\n",i_bbl));
          }

        j_bbl = CFG_EDGE_TAIL(i_edge);
      }
    }
  }
}

void CfgVerifyCorrectness(t_cfg * cfg)
{
  CfgCheckDuplicateFallthroughsIn(cfg);
  CfgCheckDuplicateFallthroughsOut(cfg);
}
/*}}}*/

/* {{{ various helper functions */
t_cfg_edge * FindFunLinkForCall(t_cfg_edge * edge)
{
   t_cfg_edge * link_edge;

   if (CFG_EDGE_HEAD(edge)==BBL_CFG(EDGE_HEAD(edge))->hell_node) return NULL;
   BBL_FOREACH_SUCC_EDGE(EDGE_HEAD(edge),link_edge)
   {
      if ((EDGE_CAT(link_edge)==ET_FUNLINK) ||(EDGE_CAT(link_edge)==ET_IPFUNLINK)) 
	 break;
   }

   return link_edge;
}

t_cfg_edge * FindFunLinkForReturn(t_cfg_edge * edge)
{
  return edge->corr_edge ? FindFunLinkForCall(edge->corr_edge) : NULL;
}

t_cfg_edge * FindCallForFunLink(t_cfg_edge * edge)
{
  t_cfg_edge * call;
  BBL_FOREACH_SUCC_EDGE(CFG_EDGE_HEAD(edge),call)
    if (EDGE_CAT(call) == ET_CALL)
      break;
  return call;
}

t_cfg_edge * FindReturnForFunLink(t_cfg_edge * edge)
{
  t_cfg_edge * ret;
  BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge),ret)
    if (EDGE_CAT(ret) == ET_RETURN)
      break;
  return ret;
}
/* }}} */

t_bbl* CfgEdgeInsertBlock(t_cfg * cfg, t_cfg_edge * edge)
{
  t_bbl * bbl = NULL;
  t_function * fun;

  switch (EDGE_CAT(edge))
    {
    case ET_FALLTHROUGH:
      bbl = BblNew(cfg);
      BblInsertInFunction(bbl,BBL_FUN(CFG_EDGE_HEAD(edge)));
      CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge),bbl,ET_FALLTHROUGH);
      CfgEdgeNew(cfg,bbl,CFG_EDGE_TAIL(edge),ET_FALLTHROUGH);
      CfgEdgeKill(edge);
      break;
    case ET_JUMP:
      if (CFG_EDGE_TAIL(edge)==FunctionGetExitBlock(BBL_FUN(CFG_EDGE_TAIL(edge))))
	{
	    FATAL(("JUMPS TO RETURN BLOCK NOT YET IMPLEMENTED\n"));
	  /*	  if (INS_ATTRIB(BBL_INS_LAST(CFG_EDGE_HEAD(edge))) & IF_CONDITIONAL)
	    FATAL(("CONDITIONAL JUMPS TO RETURN BLOCK NOT YET IMPLEMENTED\n"));

	  bbl = BblNew(cfg);
	  cfg->sec->obj->description->moveReturnInstruction(CFG_EDGE_HEAD(edge));
	  CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge),bbl,ET_JUMP);
	  CfgEdgeNew(cfg,bbl,CFG_EDGE_TAIL(edge),ET_JUMP);
	  CfgEdgeKill(edge);
	  cfg->sec->obj->description->addReturnInstruction(bbl);
	  break;
	  */
	}
      else
	{
	  bbl = BblNew(cfg);
	  BblInsertInFunction(bbl,BBL_FUN(CFG_EDGE_HEAD(edge)));
	  CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge),bbl,ET_JUMP);
	  CfgEdgeNew(cfg,bbl,CFG_EDGE_TAIL(edge),ET_JUMP);
	  CfgEdgeKill(edge);      
	  cfg->sec->obj->description->addJumpInstruction(bbl);
	  break;
	}
    case ET_RETURN:
      if (cfg->sec->obj->actions->InsIsIndirectCall(BBL_INS_LAST(CFG_EDGE_HEAD(edge->corr_edge))) || IsHellNode(CFG_EDGE_HEAD(edge)))
	FATAL(("CAN'T INSTRUMENT RETURNS FROM INDIRECT CALLS\n"));
      
      bbl = BblNew(cfg);
      BblInsertInFunction(bbl,BBL_FUN(CFG_EDGE_TAIL(edge)));

      CfgEdgeNewCall(cfg,CFG_EDGE_HEAD(edge->corr_edge),CFG_EDGE_TAIL(edge->corr_edge),bbl,FunctionGetExitBlock(BBL_FUN(CFG_EDGE_HEAD(edge))));
      CfgEdgeNew(cfg,bbl,CFG_EDGE_TAIL(edge),ET_FALLTHROUGH);

      {
	t_cfg_edge * link_edge;
	
	BBL_FOREACH_SUCC_EDGE(CFG_EDGE_HEAD(edge->corr_edge),link_edge)
	  if (EDGE_CAT(link_edge)==ET_FUNLINK) break;

	CfgEdgeNew(cfg,CFG_EDGE_HEAD(edge->corr_edge),bbl,ET_FUNLINK);
	CfgEdgeKill(link_edge);
      }


      CfgEdgeKill(edge->corr_edge);
      CfgEdgeKill(edge);      
      break;

    case ET_CALL:
      if (cfg->sec->obj->actions->InsIsIndirectCall(BBL_INS_LAST(CFG_EDGE_HEAD(edge))) || IsHellNode(CFG_EDGE_TAIL(edge)))
	FATAL(("CAN'T INSTRUMENT INDIRECT CALLS\n"));
      
      /* we will create a new fun, that returns to a new block, and
	 that jumps to the callee */
      
      bbl = BblNew(cfg);
      fun = FunctionNew(bbl,"WRAPPER",0);
      
      if (!edge->corr_edge)
	FATAL(("OOPS NO CORR EDGE\n"));
      
      CfgEdgeNewCall(cfg,CFG_EDGE_HEAD(edge),bbl,CFG_EDGE_TAIL(edge->corr_edge),FunctionGetExitBlock(fun));
      CfgEdgeNewCompensating(cfg, CfgEdgeNew(cfg,bbl,CFG_EDGE_TAIL(edge),ET_IPJUMP));
      CfgEdgeKill(edge->corr_edge);      
      CfgEdgeKill(edge);      
      cfg->sec->obj->description->addJumpInstruction(bbl);
      break;
    default:
      FATAL(("NOT IMPLEMENTED CASE @E\n",edge));
      break;
    }
  
  return bbl;
}

void CfgSortEdges(t_cfg * cfg)
{
  t_cfg_edge * edge, * tmp;
  t_cfg_edge * prev_edge = NULL;

  CFG_FOREACH_EDGE_SAFE(cfg,edge,tmp)
    {
      if (prev_edge && CfgEdgeIsForwardInterproc(edge))
	{
	  EDGE_NEXT(prev_edge) = T_EDGE(tmp);
	  if (tmp)
	    EDGE_PREV(tmp) = T_EDGE(prev_edge);
	  else
	    GRAPH_EDGE_LAST(cfg)=T_EDGE(prev_edge);

	  EDGE_NEXT(edge) = GRAPH_EDGE_FIRST(cfg);
	  EDGE_PREV(GRAPH_EDGE_FIRST(cfg))=T_EDGE(edge);
	  EDGE_PREV(edge) = NULL;

	  GRAPH_EDGE_FIRST(cfg)=T_EDGE(edge);
	}
      else
	prev_edge = edge;
    }
}
/* vim: set shiftwidth=2 foldmethod=marker: */
