/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/

#ifndef MY_OBJECT_TYPEDEFS
#define MY_OBJECT_TYPEDEFS
typedef struct _t_object t_object;
typedef struct _object_hash_node object_hash_node;

#endif /* MY_OBJECT_TYPEDEF */

#ifndef LAYOUT_SCRIPT_H
#define LAYOUT_SCRIPT_H

#include "../datastructures/data_string.h"
#include <stdio.h>

typedef struct _t_layout_exp
{
  int operator;
  struct _t_layout_exp * arg1;
  struct _t_layout_exp * arg2;
  t_string identifier;
  long long constant;
} t_layout_exp;

typedef struct _t_layout_assign
{
  t_string lhs;
  t_layout_exp * rhs;
  t_bool provide;
} t_layout_assign;

typedef struct _t_layout_secspec
{
  t_string name;
  t_layout_exp * address;
  t_layout_assign * internal_rule;
  t_layout_exp * filler_exp;
} t_layout_secspec;

typedef struct _t_layout_rule
{
  enum {ASSIGN, SECTION_SPEC, SEG_START, SEG_END, PUT_SECTIONS} kind;
  union {
    t_layout_secspec * secspec;
    t_layout_assign * assign;
    t_string segment_ident;
    char sectype;
  } u;
  struct _t_layout_rule * next;
} t_layout_rule;

t_layout_rule * LayoutScriptParse(t_string script);
long long LayoutScriptCalcExp(t_layout_exp * exp, long long currpos, t_object * obj);
void LayoutScriptFree(t_layout_rule * script);

#endif

/* vim: set shiftwidth=2 foldmethod=marker: */
