%{
#include "diablo_linkers.h"
#include "diablo_linker_scripts_parser.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../datastructures/data_string.h"
#include <string.h>
char * passon=NULL;
int lexer_line=1;
static int nests=0;

void lexer_error(char * mesg)
{
  FATAL(("Lexer error at line %d in file: %s",lexer_line,mesg));
}

void setpasson(char * in)
{
  if (passon) Free(passon);
  passon=StringDup(in);
}

void clearpasson()
{
   if (!passon) lexer_error("No variable/rule name");
   Free(passon);
   passon=NULL;
}

%}
%x RULE
%x BODY1
%x BODY2
%x STRING_BODY
%x VARIABLE
%x STRING_VARIABLE
%x COMMENT
%x COMMENT_BODY2
%x LAYOUT_SCRIPT
%%

 /* This lexer performs some grammar checks. They could be done in the parser
  * as well, but than this would complicate the parser. This approach also
  * allows more accurate error reporting */

 /* Initial: see if we have variable definition or rules: if we have a rule we
  * must have VAR = , for RULES we have VAR { */
 
 /* Initial: Newlines can be skipped, we just increase the line counter.*/
"\n"                              { lexer_line++; }
 /* Initial: The } should be matched inside <RULE> ! */
"}"                               { lexer_error("Unmatched }"); }
 /* Initial: Whitespace can be skipped */
[ \t]                             { }
"/*"				  { BEGIN(COMMENT); }
<COMMENT>"*/"			  { BEGIN(INITIAL); }
<COMMENT>"\n"			  { lexer_line++; }
<COMMENT>.			  {}
<COMMENT_BODY2>"*/"			  { BEGIN(BODY2); }
<COMMENT_BODY2>"\n"			  { lexer_line++; }
<COMMENT_BODY2>.			  {}


 /* Initial: In case we have a rule. We should already have seen a name (stored
  * in passon), otherwise we report an error  */
"{"                               { 
                                    if (!passon) 
                                      lexer_error("Rule started, but no name for rule!"); 
				    LinkerScriptParserlval.string=StringDup(passon);
                                    clearpasson();
                                    BEGIN(RULE); 
				    return RULE_NAME;
                                  } 
 /* Initial: In case we have a variable definition: We should already have seen
  * a name (stored in passon), otherwise we report an error */
=                                 { 
                                    if (!passon) 
                                      lexer_error("Variable definition, but no name for variable"); 
				    LinkerScriptParserlval.string=StringDup(passon);
                                    clearpasson(); 
                                    BEGIN(VARIABLE); 
				    return VARIABLE_DEFINITION;
                                  }
 /* Initial: skip embedded layout scripts (they are parsed separately) */
 SECTIONS			  { BEGIN(LAYOUT_SCRIPT); }

 /* Initial: Match the variable names and rule names: we should not have
  * encountered another name, otherwise throw an error */ 
[a-zA-Z0-9_]*                     { 
                                    if (passon) 
                                      lexer_error("two names, without intervening rule or definition!"); 
                                    setpasson(yytext); 
                                  }
 /* Initial: All other characters are unmatched! */ 
.                                 { printf("%s\n",yytext); lexer_error("Unmatched character INITIAL"); }

 /* Initial: forks to either <RULE> or <VARIABLE> */

 /* Variable: Matches the value of a variable. For the moment this must be a
  * string, so the only legal values are whitespace, newlines and ". In the
  * future, we could add variables that have a numeric value. A newlines ends
  * the assignment. */

 /* Variable: match the beginning of a string, fork to <STRING_VARIABLE> */
<VARIABLE>"\""                    { BEGIN(STRING_VARIABLE); }
 /* Variable: skip whitespace */
<VARIABLE>[ \t]                   { }
 /* Variable: skip newlines, enter initial state */
<VARIABLE>"\n"                    { lexer_line++; BEGIN(INITIAL); }
 /* Variable: The rest is illegal */
<VARIABLE>.                       { printf("%s\n",yytext); lexer_error("Unmatched character VARIABLE"); }

 /* String variable: Matches a string value for a variable. Strings cannot
  * contain newlines or double quotes (should not be necessary) */

 /* String variable: Newlines in strings are illegal */
<STRING_VARIABLE>"\n"             { lexer_error("Newline in string"); }
 /* String variable: Double quotes end the string */ 
<STRING_VARIABLE>"\""             { BEGIN(VARIABLE); LinkerScriptParserlval.string=StringDup(passon); clearpasson(); return VARIABLE_STRING;  }
 /* String variable: Matches the body of the string */
<STRING_VARIABLE>[a-zA-Z \t0-9.$_]* { if (passon) { printf("%s\n",passon); lexer_error("Strange STRING_VARIABLE"); } setpasson(yytext); }
 /* String variable: The rest are illegal strings */
<STRING_VARIABLE>.                { printf("%s\n",yytext);lexer_error("Unmatched character STRING_VARIABLE"); }

<RULE>"action"                    { BEGIN(BODY1); return ACTION;  }
<RULE>"trigger"                   { BEGIN(BODY1); return TRIGGER; }
<RULE>"address"                   { BEGIN(BODY1); return ADDRESS; }
<RULE>"symbol"                    { BEGIN(BODY1); return SYMBOL;  }
<RULE>"section"                   { BEGIN(BODY1); return SECTION; }
<RULE>"\n"                        { lexer_line++; }
<RULE>"}"                         { BEGIN(INITIAL); }
<RULE>[ \t]                       { }
<RULE>"."                         { 
                                    printf("%s\n",yytext); 
                                    lexer_error("Unmatched character RULE"); 
                                  }
<BODY1>"{"                        { BEGIN(BODY2); }
<BODY1>"\n"                       { lexer_line++; }
<BODY1>[ \t]                      { }
<BODY1>.                          { lexer_error("Unmatched character BODY1"); }
<BODY2>"/*"			  { BEGIN(COMMENT_BODY2); }
<BODY2>"\""                       { BEGIN(STRING_BODY); }
<BODY2>"\n"                       { lexer_line++; }
<BODY2>"("                        { return OPEN_B; }
<BODY2>")"                        { return CLOSE_B;}
<BODY2>"&&"                       { return AND;}
<BODY2>"||"                       { return OR;}
<BODY2>"+"                        { return ADD;}
<BODY2>"-"                        { return SUB;}
<BODY2>","                        { return COMMA; }
<BODY2>"}"                        { BEGIN(RULE);}
<BODY2>[A-Za-z0-9_]*[ \t]*"("     { LinkerScriptParserlval.string=StringDup(yytext); LinkerScriptParserlval.string[strlen(LinkerScriptParserlval.string)-1]=0; unput('('); return FUNCTION_NAME; }
<BODY2>[A-Za-z0-9_]*              { LinkerScriptParserlval.string=StringDup(yytext); return NAME; }
<BODY2>[ \t]                      { }
<BODY2>.                          { printf("--%s = %d--\n",yytext,yytext[0]); lexer_error("Unmatched character BODY2"); }
<STRING_BODY>"\n"                 { lexer_error("Newline in string"); }
<STRING_BODY>"\""                 { BEGIN(BODY2); LinkerScriptParserlval.string=StringDup(passon); clearpasson(); return BODY_STRING;}
<STRING_BODY>[a-zA-Z \t0-9.$_*+\\&|>-]*    { if (passon) { printf("%s\n",passon); lexer_error("Strange STRING_BODY"); } setpasson(yytext);}
<STRING_BODY>.                    { printf("%s\n",yytext);        lexer_error("Unmatched character STRING_BODY"); }

 /* rules for skipping embedded layout scripts */
<LAYOUT_SCRIPT>"{"		  { nests++; }
<LAYOUT_SCRIPT>"}"		  { nests--; if (nests == 0) BEGIN(INITIAL); }
<LAYOUT_SCRIPT>.		  { /* discard */ }
<LAYOUT_SCRIPT>\n		  { lexer_line++; }
%%
int yywrap()
{
  return 1;
}
