%{
#include "diablo_linkers.h"
%}
/* The parser for linker descriptions. See kernel/diablo_linkers.c for more
 * info on linker descriptions */

/* Token definitions */
%token VARIABLE_DEFINITION
%token VARIABLE_STRING
%token RULE_NAME
%token ACTION
%token TRIGGER 
%token SECTION 
%token ADDRESS 
%token SYMBOL
%token NAME
%token FUNCTION_NAME
%token OPEN_B
%token CLOSE_B
%token BODY_STRING

/* Associativity information */
%left COMMA
%left AND
%left OR
%left ADD 
%left SUB 

/* Datatypes used in the parser */
%union 
{
   int number;
   char * string;
   struct {
     char * string;
     int type;
     void * args;
   } arg_and_type;
   void * args;
   t_ast_node * function;
   void * rule;
}

/* Type rules */
%type <string> VARIABLE_DEFINITION
%type <string> VARIABLE_STRING
%type <string> BODY_STRING
%type <string> RULE_NAME
%type <string> NAME
%type <string> FUNCTION_NAME
%type <arg_and_type> argument
%type <args> non_empty_arguments
%type <args> arguments
%type <function> statement_value
%type <function> comma_separated_statement_value
%type <rule> rule_body 

%{
#include "diablo_linkers.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
extern int lexer_line;
void yyerror(char * in)
{
  printf("At line %d: Parse error: %s\n",lexer_line,in);
}
int LinkerScriptParserlex();
%}
%%
linker_file: /* Empty */
		 | linker_file linker_statement { }
		 ;

linker_statement: linker_rule {}
		| linker_definition { /* Definitions are currently ignored */ }
		;

linker_rule: RULE_NAME rule_body { ExecuteRule($1,$2); }
	   ;

rule_body: /* Empty */ 			     		     { $$=LinkerRuleNew();                 }
	 | rule_body ACTION comma_separated_statement_value  { ((t_linker_rule *) $$)->action=$3;  }
	 | rule_body TRIGGER comma_separated_statement_value { ((t_linker_rule *) $$)->trigger=$3; }
	 | rule_body SECTION comma_separated_statement_value { ((t_linker_rule *) $$)->section=$3; }
	 | rule_body ADDRESS comma_separated_statement_value { ((t_linker_rule *) $$)->address=$3; }
	 | rule_body SYMBOL comma_separated_statement_value  { ((t_linker_rule *) $$)->symbol=$3;  }
	 ;

comma_separated_statement_value: comma_separated_statement_value  COMMA statement_value { $$=AstNodeOperatorNew(OPER_COMMA,$1,$3); }
			       | statement_value 					{ $$=$1;                                  }
			       ;

statement_value: statement_value AND statement_value   	{ $$=AstNodeOperatorNew(OPER_AND,$1,$3); }
	       | statement_value OR statement_value    	{ $$=AstNodeOperatorNew(OPER_OR,$1,$3);  }
	       | statement_value ADD statement_value    { $$=AstNodeOperatorNew(OPER_ADD,$1,$3); }
	       | statement_value SUB statement_value    { $$=AstNodeOperatorNew(OPER_SUB,$1,$3); }
	       | OPEN_B statement_value CLOSE_B        	{ $$=$2;                                }
	       | FUNCTION_NAME OPEN_B arguments CLOSE_B { $$=AstNodeFunctionNew($1,$3);         }
	       ;
	
arguments: non_empty_arguments { $$=$1;} 
	 | /* Empty */ 	       { $$=NULL; }
	 ;

non_empty_arguments: non_empty_arguments COMMA argument { $$=$1; AstSuccessorsAddSuccessor($$,$3.string, $3.type,$3.args); }
		   | argument 				{ $$=AstSuccessorsNew(); AstSuccessorsAddSuccessor($$,$1.string, $1.type, $1.args); }
		   ;

argument: BODY_STRING {$$.string=$1; $$.type=TYPE_STRING; $$.args=NULL; }
	| NAME 	      {$$.string=$1; $$.type=TYPE_OTHER; $$.args=NULL; }
	| statement_value { if ($1->type==TYPE_FUNCTION) { $$.string=$1->data.name; $$.type=TYPE_FUNCTION; $$.args=$1->args; Free($1); } else FATAL(("implement non-function argument astnodes"));  }
	;

linker_definition: VARIABLE_DEFINITION VARIABLE_STRING {  Free($1); Free($2); }
		 ;
%%

