/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}}
 *****************************************************************************/
#ifndef DIABLO_LINKERS_H
#define DIABLO_LINKERS_H
#include "../datastructures/data_string.h"
#include "diablo_addresses.h"

typedef enum { NO_LINKER=0, ADS_LINKER, GNU_LINKER } t_linker;

typedef enum
{
  TYPE_BOOL,
  TYPE_VOID,
  TYPE_OPER,
  TYPE_STRING,
  TYPE_NUMERIC,
  TYPE_OTHER,
  TYPE_ADDRESS,
  TYPE_SECTION,
  TYPE_SYMBOL,
  TYPE_FUNCTION,
  TYPE_TRIGGER
} t_ast_types;

typedef enum
{
   OPER_AND,
   OPER_OR,
   OPER_ADD,
   OPER_SUB,
   OPER_COMMA
} t_ast_operators;

typedef struct _t_ast_node t_ast_node;

/*! Struct to hold the successors of an ast node */
typedef struct _t_ast_successors
{
   t_uint32 nargs;
   t_ast_node ** args;
} t_ast_successors;

struct _t_ast_node
{
  t_ast_types type;
  union
  {
    t_string name;
    t_ast_operators oper;
    void * data;
    t_address addr;
  } data; 
  t_ast_successors * args;
};

/*! A linker rule as passed from the linker script parser to the linker script
 * handler*/ 

typedef struct _t_linker_rule
{
   t_ast_node * action;
   t_ast_node * trigger;
   t_ast_node * section;
   t_ast_node * address;
   t_ast_node * symbol;
} t_linker_rule;

/*! An entry in the linker function table */
typedef struct linker_function_table_entry
{
  t_string name;
  t_uint32 nargs;
  t_string arguments;
  t_string ret;
  void * (*fun)(t_ast_successors *,void *);
} t_ast_node_table_entry;

/* Ast nodes */
t_ast_successors * AstSuccessorsNew();
t_ast_node * AstNodeFunctionNew(t_string name, t_ast_successors * args);
void AstSuccessorsAddSuccessor(t_ast_successors * args, t_string argument, t_ast_types type, t_ast_successors * args2);
t_ast_node * AstNodeOperatorNew(t_ast_operators oper, t_ast_node * left, t_ast_node *right);
void FreeAstNode(t_ast_node *);

/* Linker rules */
t_linker_rule * LinkerRuleNew();
void ExecuteRule(t_string name, t_linker_rule * rule);
#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
