/******************************************************************************
 * Copyright 2001,2002,2003{{{: 
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}} 
 *****************************************************************************/

#define LIVENESS_OPTION_CONDITIONAL_CALL_OPT

/* Includes {{{ */
#include "../utils/util_error.h"
#include "../utils/util_malloc.h"
#include "../utils/util_status.h"
#include "../simple_types.h"
#include "diablo_ins.h"
#include "diablo_function.h"
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_liveness.h"
#include "diablo_object.h"
#include "diablo_regset.h"
#include "diablo_print.h"
#include "diablo_edge.h"
#include "diablo_opt_misc.h"
#include "diablo_iterators.h"

#ifdef IA64_SUPPORT
#include "../arch/ia64/ia64_liveness.h"
#endif
/*}}} */

/* Utils {{{ */
/** Check if register reg is defined in [ins1,ins2] */
#if 0
static t_bool IsDefinedInRange(t_reg reg,t_ins * ins1,t_ins * ins2)
{
  if( (!ins1) || (!ins2) )
    return FALSE;

  for(;ins1!=NULL;ins1 = INS_NEXT(ins1))
    {
      if( RegsetIn(INS_REGS_DEF(ins1),reg) ) return TRUE;
      if( ins1 == ins2 ) return FALSE;
    }
   
  FATAL(("We should never get here!"));
  return FALSE;
}
#endif
/** Check if register reg is used in [ins1,ins2] */
#if 0
static t_bool IsUsedInRange(t_reg reg,t_ins * ins1,t_ins * ins2)
{
  if( (!ins1) || (!ins2) )
    return FALSE;
    
  for(;ins1!=NULL;ins1 = INS_NEXT(ins1))
    {
      if( RegsetIn(INS_REGS_USE(ins1),reg) ) return TRUE;
      if( ins1 == ins2 ) return FALSE;
    }
  
  FATAL(("We should never get here!"));
  return FALSE;
}
#endif
/*}}} */
/* Compute changed {{{ */
/** This function computes the registers that a procedure changes,
    ignoring loads that restore a register.  */
static void FunctionComputeChangedRegisters(t_function * fun)
{
    
  t_bbl *  i_bbl;
  t_ins * i_ins;
  t_regset changed;
  
  if( FUN_IS_HELL(fun) )
    {
      /* for now assume very worst-case behavior */
      RegsetSetAll(FUN_REGS_CHANGED(fun));

      if (global_options.rely_on_calling_conventions)
	{
	  if (FUN_BBL_FIRST(fun)==fun->cfg->call_hell_node)
	    {
	      RegsetSetDiff(FUN_REGS_CHANGED(fun),fun->cfg->sec->obj->description->callee_saved);
	    }
	}
      return;

    }
  
  changed = RegsetNew();
  
  FUN_FOREACH_BBL(fun,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
        {
	  RegsetSetUnion(changed,INS_REGS_DEF(i_ins));
        }
    }


  if (global_options.rely_on_calling_conventions)
    if (fun->cfg->sec->obj->actions->FunIsGlobal(fun))
      RegsetSetDiff(changed,fun->cfg->sec->obj->description->callee_saved);
  
  RegsetDup(FUN_REGS_CHANGED(fun),changed);
  RegsetFree(changed);
}

/*}}}*/
/* Compute Saved {{{ */
/** This function computes the registers that a procedure saves, THESE
    ARE ONLY THOSE REGISTERS WHO's VALUE UPON ENTRY IS NOT USED
    OTHERWISE BY THE PROCEDURE */

void FunctionComputeSavedRegisters(t_function * fun)
{
  t_regset saved= RegsetNew();
  RegsetDup(FUN_REGS_SAVED(fun),saved);
  RegsetFree(saved);
  
#ifndef KERNEL_OPT 
  if (fun->cfg->sec->obj->actions->FunComputeStackSavedRegisters)
    fun->cfg->sec->obj->actions->FunComputeStackSavedRegisters(fun);
#endif
}
/*}}} */
/* Compute Saved Changed {{{ */
/** Fix-point computation to find the registers that are not changed,
    but saved by procedures (in themselves or in their call chain). */
void ComputeSavedChangedRegisters(t_section * sec)
{

  t_cfg_edge * i_edge;
  t_function * caller;
  t_function * callee;
  t_regset changed;
  t_regset current;
  t_bool  flag;

  t_function * i_fun;

  CfgSortEdges(T_CFG(sec->data));

  FOREACH_FUNCTION(sec,i_fun)
    {
      FunctionComputeChangedRegisters(i_fun);
      FunctionComputeSavedRegisters(i_fun);
      RegsetSetDiff(FUN_REGS_CHANGED(i_fun),FUN_REGS_SAVED(i_fun));
    }
  
  do
    {
      flag = FALSE;
      SECTION_FOREACH_EDGE(sec,i_edge)
	{
	  if( !CfgEdgeIsForwardInterproc(i_edge) ) break;
	  caller = BBL_FUN(CFG_EDGE_HEAD(i_edge));
	  callee = BBL_FUN(CFG_EDGE_TAIL(i_edge));

#ifdef INSTRUMENTATION
	  if ((!caller)||(!callee)) continue;
#endif

#ifndef INSTRUMENTATION
	  /* this is turned off for instrumentation because
	   * ComputeSavedChangedRegisters is called from 
	   * within OptimizePseudoLoadStores and for that
	   * function, this optimization is not allowed. */

	  /* call hell function _always_ adheres to calling conventions */

	  /*

	  TODO: check this out, it makes lame on linux fail !!!!

	    if (caller==caller->cfg->call_hell_function)
	    continue;
	  */
#endif

	  RegsetDup(current,FUN_REGS_CHANGED(caller));
	  RegsetDup(changed,FUN_REGS_CHANGED(callee));
	  
	  RegsetSetDiff(changed,FUN_REGS_SAVED(caller));
	  
	  if (!RegsetEquals(RegsetNewUnion(changed,current),current))
	    {
	      RegsetDup(FUN_REGS_CHANGED(caller),RegsetNewUnion(current,changed));
	      flag = TRUE;
	    }
	  
        }
    }   
  while( flag );
}
/*}}}*/
/* Liveness Trivial {{{ */
static void ComLivenessTrivial(t_cfg * cfg)
{
  t_bbl * i_bbl;
        
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      RegsetSetAll(BBL_REGS_LIVE_OUT(i_bbl));
    }
}

/*}}}*/
/* Liveness Insensitive {{{ */
/** Perform a straight-forward context-insensitive liveness analysis*/
static void ComLivenessInsensitive(t_cfg * cfg)
{
  /* iterators */

  t_cfg_edge * i_edge;
  t_bbl * i_bbl;
  t_function * i_fun;

  /* local variables */

  t_bool  flag,flag2;
  t_regset in = RegsetNew();
  t_regset tmp;
  t_regset new_out = RegsetNew();
		  
  /* Initialize the def,use,in,out fields for each bbl */
       
  InitNodeMarking();

  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      i_fun = BBL_FUN(i_bbl);
 
      /* initialize it to a ... */

      RegsetDup(i_bbl->regs_used,tmp = BblRegsUse(i_bbl));
      RegsetFree(tmp);
      RegsetDup(i_bbl->regs_defined, tmp = BblRegsDef(i_bbl));
      RegsetFree(tmp);
      
      if( !i_fun || FUN_IS_HELL(i_fun) )
        {
	  /*
	    this information is not really used since
	    AbblRegLiveIn called below has special handling
	    for the PSEUDO functions.
	    It will keep these bbls from beeing unecessarily updated, though
	  */
	  
	  RegsetSetAll(BBL_REGS_LIVE_OUT(i_bbl));
        }
      else
        {
	  RegsetSetEmpty(BBL_REGS_LIVE_OUT(i_bbl));
        }
     
      /* mark all basic blocks */

      BblMark(i_bbl);
    }

  /* initialize: Mark all functions */
  



  CFG_FOREACH_FUN(cfg,i_fun) 
    FUN_FLAGS(i_fun)|=FF_IS_MARKED;

  /* Do cfa0 fixpoint computation */
  do
    {
      flag = 0;
      CFG_FOREACH_FUN(cfg,i_fun)
	{
	 
	  /* Do not propagate non changed information */
	  if( !(FUN_FLAGS(i_fun)&FF_IS_MARKED) ) continue;
	  
	  /* Unmark the function */
	  FUN_FLAGS(i_fun)&=~FF_IS_MARKED;
	  
	  do
	    {
	      flag2 = 0;  

	      FUN_FOREACH_BBL(i_fun,i_bbl)
		{
		  /* Do not propagate non changed information */
		  if( !BblIsMarked(i_bbl) ) continue;
		  
		  /* Unmark the basic block */
		  BblUnmark(i_bbl);

		  flag = 1; /* TODO: Is this necessary ? */
		  
		  RegsetDup(in,BBL_REGS_LIVE_OUT(i_bbl));
		  RegsetSetDiff(in,BBL_REGS_DEFINED(i_bbl));
		  RegsetSetUnion(in,BBL_REGS_USED(i_bbl));
		  
		  BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
		    {
		      t_bbl * hd = (t_bbl *) CFG_EDGE_HEAD(i_edge);
		      /* the following test was added by Bjorn om 05/12/2003 */
		      if (EDGE_CAT(i_edge)==ET_FUNLINK /* || EDGE_CAT(i_edge)==ET_IPFUNLINK LINK_REMOVAL */) continue;
		      RegsetDup(new_out,in);
		      RegsetSetUnion(new_out,BBL_REGS_LIVE_OUT(hd));
		      if(!RegsetEquals(new_out,BBL_REGS_LIVE_OUT(hd)))
			{
			  RegsetDup(BBL_REGS_LIVE_OUT(hd),new_out);
			  BblMark(hd);
			  if (CfgEdgeIsInterproc(i_edge))
			    {
			      if (BBL_FUN(hd))
				{
				  FUN_FLAGS(BBL_FUN(hd))|=FF_IS_MARKED;
				  flag = 1;
				}
			    }
			  else
			    flag2=1;	  /* BBL_FLAGS(hd) |= FF_IS_MARKED; */
			}
		    }
		}
	    }
	  while (flag2);
        }
    }
  while( flag );

  RegsetFree(new_out);
  RegsetFree(in);
}
/* }}} */

static void SensitiveLivenessFixpoint(t_cfg * cfg);
static void ComputeFunctionThroughRegs(t_cfg * cfg);
static void ComputeFunctionUsedRegs(t_cfg * cfg);
static void ComputeMeaningfulRegs(t_cfg * cfg);

/* Liveness Sensitive {{{ */
/** Context-sensitive liveness analyses, in three phases, as described
    in Robert Muth's thesis. */

static void ComLivenessSensitive(t_cfg * cfg)
{
  /* iterators */

  t_function * i_fun;
  t_bbl * i_bbl;
  t_cfg_edge * i_edge;
  /* local variables */

  const t_architecture_description * description = cfg->sec->obj->description;

  t_regset tmp;

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    if (FUN_BBL_FIRST(i_fun) == cfg->call_hell_node)
    {
      RegsetDup(FUN_REGS_THROUGH(i_fun),description->callee_saved);
      RegsetDup(FUN_REGS_USED(i_fun),description->callee_may_use);
    }
    else if( FUN_IS_HELL(i_fun) )
    {
      RegsetSetAll(FUN_REGS_THROUGH(i_fun));
      RegsetSetAll(FUN_REGS_USED(i_fun));
    }
    else
    {
      RegsetSetEmpty(FUN_REGS_THROUGH(i_fun));
      RegsetSetEmpty(FUN_REGS_USED(i_fun));
    }
  }
  /*
  BBL_FOREACH_SUCC_EDGE(cfg->hell_node,i_edge)
    if (EDGE_CAT(i_edge)==ET_CALL && !IsHellNode(CFG_EDGE_TAIL(i_edge)))
      {
#ifdef AGGRESSIVE_LIVENESS
	if (global_options.rely_on_calling_conventions)
	  RegsetSetIntersect(FUN_REGS_USED(BBL_FUN(CFG_EDGE_TAIL(i_edge))),description->callee_may_use);
#endif
      }
  */
  
  /* this information never changes during the liveness analysis! */

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (!BBL_FUN(i_bbl))
      continue; 

    RegsetDup(BBL_REGS_USED(i_bbl),tmp = BblRegsUse(i_bbl));
    RegsetFree(tmp);
    RegsetDup(BBL_REGS_DEFINED(i_bbl), tmp = BblRegsDef(i_bbl));
    RegsetFree(tmp);


    /* We set the exit blocks to 'all registers live' */
    if( BBL_IS_LAST(i_bbl) && !FUN_IS_HELL(BBL_FUN(i_bbl)) && i_bbl == FunctionGetExitBlock(BBL_FUN(i_bbl)))
      {
	RegsetSetAll(BBL_REGS_LIVE_OUT(i_bbl));
      }
    else
      {
	RegsetSetEmpty(BBL_REGS_LIVE_OUT(i_bbl));
      }
  }

  /* FUN_REGS_THROUGH will be set now */

  ComputeFunctionThroughRegs(cfg);


  CFG_FOREACH_FUN(cfg,i_fun)
  {
    if (FUN_BBL_FIRST(i_fun) == cfg->call_hell_node)
    {
      RegsetDup(FUN_REGS_THROUGH(i_fun),description->callee_saved);
      RegsetDup(FUN_REGS_USED(i_fun),description->callee_may_use);
    }
    else if( FUN_IS_HELL(i_fun) )
    {
      RegsetSetAll(FUN_REGS_THROUGH(i_fun));
      RegsetSetAll(FUN_REGS_USED(i_fun));
    }
  }

  BBL_FOREACH_SUCC_EDGE(cfg->hell_node,i_edge)
    if (EDGE_CAT(i_edge)==ET_CALL && !IsHellNode(CFG_EDGE_TAIL(i_edge)))
      {
#ifdef AGGRESSIVE_LIVENESS
	if (global_options.rely_on_calling_conventions)
	  RegsetSetIntersect(FUN_REGS_USED(BBL_FUN(CFG_EDGE_TAIL(i_edge))),description->callee_may_use);
#endif
      }


  /*
   * Now we want to compute FUNCTION_REGS_USED and hence we set the
   * exit blocks to "no register live"
   */


  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    RegsetSetEmpty(BBL_REGS_LIVE_OUT(i_bbl));
  }

  ComputeFunctionUsedRegs(cfg);

  CFG_FOREACH_FUN(cfg,i_fun)
  {
    if (FUN_BBL_FIRST(i_fun) == cfg->call_hell_node)
    {
      RegsetDup(FUN_REGS_THROUGH(i_fun),description->callee_saved);
      RegsetDup(FUN_REGS_USED(i_fun),description->callee_may_use);
    }
    else if( FUN_IS_HELL(i_fun) )
    {
      RegsetSetAll(FUN_REGS_THROUGH(i_fun));
      RegsetSetAll(FUN_REGS_USED(i_fun));
    }
  }

  BBL_FOREACH_SUCC_EDGE(cfg->hell_node,i_edge)
    if (EDGE_CAT(i_edge)==ET_CALL && !IsHellNode(CFG_EDGE_TAIL(i_edge)))
      {
#ifdef AGGRESSIVE_LIVENESS
	if (global_options.rely_on_calling_conventions)
	  RegsetSetIntersect(FUN_REGS_USED(BBL_FUN(CFG_EDGE_TAIL(i_edge))),description->callee_may_use);
#endif
      }

  
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    RegsetSetEmpty(BBL_REGS_LIVE_OUT(i_bbl));
  }


  /*
   * Finally we provide the true information about what is live at the exit block
   * we can use a clever way which requires us to save some information earlier
   * or use another fix point iteration.
   */

  SensitiveLivenessFixpoint(cfg);

  /* and than we add a forward analysis */

  ComputeMeaningfulRegs(cfg);


}
/* }}} */
/* Init liveness fixpoint {{{ */
/** Init a context-sensitive liveness analysis */
void InitSensitiveLivenessFixpoint(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_function * i_fun;
  t_uint32 total=0;
  static t_uint32 oldtotal = 0;
  
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      total+=BBL_NINS(i_bbl);
      BblMark(i_bbl);
    }
  
  if(total != oldtotal)
  {
/*    printf("NINS: %d\n",total);*/
    oldtotal = total;
  }

  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FUN_FLAGS(i_fun)|=FF_IS_MARKED;
    }
}
/*}}}*/
/* Liveness fixpoint {{{ */

/*! Fix-point computation for context-sensitive liveness analysis. Implements
 * all three phases in the algorithm. The equations for the fixpoint can be
 * found in Robert Muth's thesis. A more detailed description can be found in
 * Bjorn De Sutter's thesis.  */

static t_regset BblComputeLiveInFromLiveOut(t_bbl * bbl, t_bool update)
{
  t_regset in = RegsetNew();
  RegsetDup(in,BBL_REGS_LIVE_OUT(bbl));
#if 1
  if (update==0)
    {
      t_ins * i_ins;
      if (bbl->fun && (bbl==bbl->fun->cfg->hell_node || bbl==bbl->fun->cfg->exit_hell_node || bbl==bbl->fun->cfg->swi_hell_node ||  bbl==bbl->fun->cfg->exit_swi_hell_node))
	{
	  RegsetSetDiff(in,BBL_REGS_DEFINED(bbl));
	  RegsetSetUnion(in,BBL_REGS_USED(bbl));
	}
      else
	{
	  BBL_FOREACH_INS_R(bbl,i_ins)
	    {
	      if (InsHasSideEffect(i_ins))
		{
		  if (!INS_IS_CONDITIONAL(i_ins)) 
		    RegsetSetDiff(in,INS_REGS_DEF(i_ins));
		  RegsetSetUnion(in,INS_REGS_USE(i_ins));
		}
	      else
		{
		  t_regset tmp;
		  RegsetDup(tmp,in);
		  RegsetSetIntersect(tmp,INS_REGS_DEF(i_ins));
		  if (!(RegsetIsEmpty(tmp)))
		    {
		      if (!INS_IS_CONDITIONAL(i_ins)) 
			RegsetSetDiff(in,INS_REGS_DEF(i_ins));
		      RegsetSetUnion(in,INS_REGS_USE(i_ins));
		    }
		} 
	    }
	}
    }
  else
    {
      RegsetSetDiff(in,BBL_REGS_DEFINED(bbl));
      RegsetSetUnion(in,BBL_REGS_USED(bbl));
    }
#else
  RegsetSetDiff(in,BBL_REGS_DEFINED(bbl));
  RegsetSetUnion(in,BBL_REGS_USED(bbl));
#endif
  return in;
}

t_bool BblPropLivenessToPredButNoReturnEdge(t_bbl * bbl, t_regset in, const t_architecture_description * description)
{
  t_cfg_edge * i_edge;
  t_regset old_out, out;
  t_function * i_fun = BBL_FUN(bbl);
  t_bool change = FALSE;

  

  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
    {
      register t_bbl * head = CFG_EDGE_HEAD(i_edge);
      t_function * callee;
      t_bbl * returnsite;

      RegsetDup(old_out,BBL_REGS_LIVE_OUT(head));

      if (CfgEdgeIsInterproc(i_edge))
	{
	  /* IP edges {{{ */
	  switch( EDGE_CAT(i_edge) )
	    {
	    case ET_COMPENSATING:
	    case ET_RETURN:
	      /* Just copy the regset, so that we will stop */
	      RegsetDup(out,old_out);

	      if (EDGE_CAT(i_edge)==ET_RETURN && i_edge->corr_edge)
		{
		  t_regset tmp = RegsetNew();
		  t_bbl * callsite = CFG_EDGE_HEAD(i_edge->corr_edge);

		  if (!IsHellNode(callsite) && BBL_FUN(callsite) && !FUN_IS_HELL(BBL_FUN(callsite)))
		    {
		      
		      t_regset callsite_out = RegsetNew();
		      t_regset callsite_oldout = RegsetNew();
		      
		      RegsetDup(callsite_oldout,BBL_REGS_LIVE_OUT(callsite));
		      RegsetDup(callsite_out,callsite_oldout);
		      
		      callee = BBL_FUN(CFG_EDGE_HEAD(i_edge));
		      
		      RegsetDup(tmp,in);
		      RegsetSetIntersect(tmp,FUN_REGS_THROUGH(callee));
		      RegsetSetUnion(callsite_out,tmp);
		      
		      if(!RegsetEquals(callsite_out,callsite_oldout))
			{
			  BblMark(callsite);
			  change = TRUE;
			  RegsetDup(BBL_REGS_LIVE_OUT(callsite),callsite_out);
			}
		    }
		}

	      break;

	    case ET_CALL: 
	      callee = i_fun;

	      returnsite = NULL;

	      if (i_edge->corr_edge)
		returnsite = CFG_EDGE_TAIL(i_edge->corr_edge); 

	      if(returnsite)
		{
		  RegsetDup(out,BBL_REGS_LIVE_OUT(returnsite));
		  RegsetSetDiff(out,BBL_REGS_DEFINED(returnsite));
		  RegsetSetUnion(out,BBL_REGS_USED(returnsite));
		}
	      else
		{
		  out = NullRegs; /* BRUNO */
		  /*		  FATAL(("OOPS, MAYBE THIS SHOULD NOT HAPPEN"));*/
		}

	      RegsetSetIntersect(out,FUN_REGS_THROUGH(callee));
	      RegsetSetUnion(out,FUN_REGS_USED(callee));
	      
#ifdef LIVENESS_OPTION_CONDITIONAL_CALL_OPT
	      /* On the arm, when we have a conditional bl instruction, the link register is not live */
	      {
		t_bbl * callsite = CFG_EDGE_HEAD(i_edge); 
		if(callsite && BBL_INS_LAST(callsite))
		  {
#ifdef MIPS_SUPPORT 
		    // Delay slot! Conditional instructions are always conditional branches.
		    if(description->isProcedureCall(BBL_INS_LAST(callsite)) && (INS_ATTRIB(BBL_INS_LAST(callsite)) & IF_CONDITIONAL || (INS_PREV(BBL_INS_LAST(callsite)) && (INS_ATTRIB(BBL_INS_LAST(callsite)) & IF_CONDITIONAL))))
		      RegsetSetDiff(out,description->link_register);
#else
		    if(description->isProcedureCall(BBL_INS_LAST(callsite)) && INS_ATTRIB(BBL_INS_LAST(callsite)) & IF_CONDITIONAL)
		      RegsetSetDiff(out,description->link_register);
#endif
		  }
	      }
#endif

	      if(FunctionBehaves(BBL_FUN(head)))
		RegsetSetDiff(out,description->dead_over_call);

	      RegsetSetUnion(out,old_out);
	      break;
	    case ET_FUNLINK:
	      FATAL(("SHOULD NEVER COME HERE"));
	      out = NullRegs;
	      break;
	    case ET_SWI:
	      {	
		t_regset tmp = RegsetNew();
		RegsetDup(tmp,in);
		RegsetDup(out,old_out);

		if (description->computeLiveRegsBeforeSwi)
		  description->computeLiveRegsBeforeSwi(&tmp,BBL_INS_LAST(head));
		
		RegsetSetDiff(tmp,description->dead_over_call);
		RegsetSetUnion(out,tmp);
		RegsetFree(tmp);
	      }
	      break;
	    default:
	      RegsetDup(out,old_out);
	      RegsetSetUnion(out,in);
	    }
	  /*}}}*/
	  if(!RegsetEquals(out,old_out))
	    {
	      if (IsHellNode(head)) continue;
	      /*	  if (!BBL_FUN(head)) FATAL(("OOPS 1"));//continue;
	      	  if( FUN_IS_HELL(BBL_FUN(head)))
		  FATAL(("OOPS 2"));//continue;*/
	      
	      BblMark(head);
	      if (BBL_FUN(head)) FUN_FLAGS(BBL_FUN(head))|=FF_IS_MARKED;
	      change = TRUE;
	      RegsetDup(BBL_REGS_LIVE_OUT(head),out);
	    }
      
	}
      else if (EDGE_CAT(i_edge)==ET_FUNLINK)
	{
	  continue;
	 
	}
      else 
	{
	  RegsetDup(out,old_out);

	  if (EDGE_CAT(i_edge)==ET_SWILINK)
	    {
	      if (description->computeLiveRegsBeforeSwi)
		description->computeLiveRegsBeforeSwi(&in,BBL_INS_LAST(head));
	    }
	  RegsetSetUnion(out,in);

	  if(!RegsetEquals(out,old_out))
	    {
	      if (IsHellNode(head)) continue;
	      /*	      if (!BBL_FUN(head)) FATAL(("OOPS 2"));//continue;
	      	      if( FUN_IS_HELL(BBL_FUN(head)))
	      		FATAL(("OOPS 3"));//continue;
	      */
	      BblMark(head);
	      /*	      if (CfgEdgeIsInterproc(i_edge))
	      		{
	      		  FATAL(("OOPS 4"));//continue;
	      		  FUN_FLAGS(BBL_FUN(head))|=FF_IS_MARKED;
			  }*/
	      change = TRUE;
	      RegsetDup(BBL_REGS_LIVE_OUT(head),out);
	      continue;
	    }
	}

      if(!RegsetEquals(out,old_out))
	{
	  if (IsHellNode(head)) continue;
	  /*	  if (!BBL_FUN(head)) FATAL(("OOPS 1"));//continue;
	  	  if( FUN_IS_HELL(BBL_FUN(head)))
		  FATAL(("OOPS 2"));//continue;*/

	  BblMark(head);
	  if (CfgEdgeIsInterproc(i_edge))
	    {
	      if (BBL_FUN(head)) FUN_FLAGS(BBL_FUN(head))|=FF_IS_MARKED;
	    }
	  change = TRUE;
	  RegsetDup(BBL_REGS_LIVE_OUT(head),out);
	}
    }

  return change;

}

t_bool BblPropLivenessToPredIncludingReturnEdge(t_bbl * bbl, t_regset in, const t_architecture_description * description)
{
  t_cfg_edge * i_edge;

  t_regset old_out, out;
  t_function * i_fun = BBL_FUN(bbl);
  t_bool change = FALSE;

  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
    {
      register t_bbl * head = CFG_EDGE_HEAD(i_edge);
      t_function * callee;
      t_bbl * returnsite;

      RegsetDup(old_out,BBL_REGS_LIVE_OUT(head));

      if (CfgEdgeIsInterproc(i_edge))
	{
	  /* IP edges {{{ */
	  switch( EDGE_CAT(i_edge) )
	    {
	    case ET_COMPENSATING:
	    case ET_RETURN:
		{
		  t_regset tmp = RegsetNew();
		  RegsetDup(tmp,in);
		  
		  if (FunctionBehaves(BBL_FUN(head)))
		    RegsetSetDiff(tmp,description->dead_over_call);
		  
		  if (IsHellNode(CFG_EDGE_TAIL(i_edge)))
		    {
		      RegsetSetDiff(tmp,description->callee_may_change);
		      RegsetSetUnion(tmp,description->callee_may_return);
		    }
		  
		  RegsetDup(out,old_out);
		  RegsetSetUnion(out,tmp);
		}

	      if (EDGE_CAT(i_edge)==ET_RETURN && i_edge->corr_edge)
		{
		  t_regset tmp = RegsetNew();
		  t_bbl * callsite = CFG_EDGE_HEAD(i_edge->corr_edge);

		  if (!IsHellNode(callsite) && BBL_FUN(callsite) && !FUN_IS_HELL(BBL_FUN(callsite)))
		    {
		      
		      t_regset callsite_out = RegsetNew();
		      t_regset callsite_oldout = RegsetNew();
		      
		      RegsetDup(callsite_oldout,BBL_REGS_LIVE_OUT(callsite));
		      RegsetDup(callsite_out,callsite_oldout);
		      
		      callee = BBL_FUN(CFG_EDGE_HEAD(i_edge));
		      
		      RegsetDup(tmp,in);
		      RegsetSetIntersect(tmp,FUN_REGS_THROUGH(callee));
		      RegsetSetUnion(callsite_out,tmp);
		      
		      if(!RegsetEquals(callsite_out,callsite_oldout))
			{
			  BblMark(callsite);
			  change = TRUE;
			  RegsetDup(BBL_REGS_LIVE_OUT(callsite),callsite_out);
			}
		    }
		}

	      break;

	    case ET_CALL: 
	      callee = i_fun;

	      returnsite = NULL;

	      if (i_edge->corr_edge)
		returnsite = CFG_EDGE_TAIL(i_edge->corr_edge); 

	      if(returnsite)
		{
		  RegsetDup(out,BBL_REGS_LIVE_OUT(returnsite));
		  RegsetSetDiff(out,BBL_REGS_DEFINED(returnsite));
		  RegsetSetUnion(out,BBL_REGS_USED(returnsite));
		}
	      else
		{
		  out = NullRegs; /* BRUNO */
		  /*		  FATAL(("OOPS, MAYBE THIS SHOULD NOT HAPPEN"));*/
		}

	      RegsetSetIntersect(out,FUN_REGS_THROUGH(callee));
	      RegsetSetUnion(out,FUN_REGS_USED(callee));

#ifdef LIVENESS_OPTION_CONDITIONAL_CALL_OPT
	      /* On the arm, when we have a conditional bl instruction, the link register is not live */
	      {
		t_bbl * callsite = CFG_EDGE_HEAD(i_edge); 
		if(callsite && BBL_INS_LAST(callsite))
		  {
		    if(description->isProcedureCall(BBL_INS_LAST(callsite)) && INS_ATTRIB(BBL_INS_LAST(callsite)) & IF_CONDITIONAL)
		      RegsetSetDiff(out,description->link_register);
		  }
	      }
#endif

	      if(FunctionBehaves(BBL_FUN(head)))
		RegsetSetDiff(out,description->dead_over_call);

	      RegsetSetUnion(out,old_out);
	      break;
	    case ET_FUNLINK:
	      FATAL(("SHOULD NEVER COME HERE"));
	      out = NullRegs;
	      break;
	    case ET_SWI:
	      {	
		t_regset tmp = RegsetNew();
		RegsetDup(tmp,in);
		RegsetDup(out,old_out);

		if (description->computeLiveRegsBeforeSwi)
		  description->computeLiveRegsBeforeSwi(&tmp,BBL_INS_LAST(head));
		
		RegsetSetDiff(tmp,description->dead_over_call);
		RegsetSetUnion(out,tmp);
		RegsetFree(tmp);
	      }
	      break;
	    default:
	      RegsetDup(out,old_out);
	      RegsetSetUnion(out,in);
	    }
	  /*}}}*/
	  if(!RegsetEquals(out,old_out))
	    {
	      if (IsHellNode(head)) continue;
	      /*	  if (!BBL_FUN(head)) FATAL(("OOPS 1"));//continue;
	      	  if( FUN_IS_HELL(BBL_FUN(head)))
		  FATAL(("OOPS 2"));//continue;*/
	      
	      BblMark(head);
	      if (BBL_FUN(head)) FUN_FLAGS(BBL_FUN(head))|=FF_IS_MARKED;
	      change = TRUE;
	      RegsetDup(BBL_REGS_LIVE_OUT(head),out);
	    }
      
	}
      else if (EDGE_CAT(i_edge)==ET_FUNLINK)
	{
	  continue;
	}
      else 
	{
	  RegsetDup(out,old_out);

	  if (EDGE_CAT(i_edge)==ET_SWILINK)
	    {
		if (description->computeLiveRegsBeforeSwi)
		  description->computeLiveRegsBeforeSwi(&in,BBL_INS_LAST(head));
		
	    }

	  RegsetSetUnion(out,in);

	  if(!RegsetEquals(out,old_out))
	    {
	      if (IsHellNode(head)) continue;
	      /*	      if (!BBL_FUN(head)) FATAL(("OOPS 2"));//continue;
	      	      if( FUN_IS_HELL(BBL_FUN(head)))
		      FATAL(("OOPS 3"));//continue;*/
	      
	      BblMark(head);
	      /*	      if (CfgEdgeIsInterproc(i_edge))
	      		{
	      		  FATAL(("OOPS 4"));//continue;
	      		  FUN_FLAGS(BBL_FUN(head))|=FF_IS_MARKED;
			  }*/
	      change = TRUE;
	      RegsetDup(BBL_REGS_LIVE_OUT(head),out);
	      continue;
	    }
	}

      if(!RegsetEquals(out,old_out))
	{
	  if (IsHellNode(head)) continue;
	  /*	  if (!BBL_FUN(head)) FATAL(("OOPS 1"));//continue;
	  	  if( FUN_IS_HELL(BBL_FUN(head)))
		  FATAL(("OOPS 2"));//continue;*/

	  BblMark(head);
	  if (CfgEdgeIsInterproc(i_edge))
	    {
	      if (BBL_FUN(head)) FUN_FLAGS(BBL_FUN(head))|=FF_IS_MARKED;
	    }
	  change = TRUE;
	  RegsetDup(BBL_REGS_LIVE_OUT(head),out);
	}
    }

  return change;

}


static void ComputeFunctionThroughRegs(t_cfg * cfg)
{
  /* iterators */

  t_bbl * i_bbl;
  t_function * i_fun;

  /* local variables */

  const t_architecture_description * description = cfg->sec->obj->description;

  t_bool change,big_change;
  t_regset in;

  /*
     we assume that bbl and fun stuff has been computed already
     here we just do the do fixpoint computation
   */

  InitSensitiveLivenessFixpoint(cfg);

  do
  {
    big_change = FALSE;
    CFG_FOREACH_FUN(cfg,i_fun)
    {
      if(!(FUN_FLAGS(i_fun)&FF_IS_MARKED))
	continue;

      FUN_FLAGS(i_fun)&=~FF_IS_MARKED;

      do
      {
	change = FALSE;

	FUN_FOREACH_BBL_R(i_fun,i_bbl)
	{
	  if(!BblIsMarked(i_bbl))
	    continue;

	  BblUnmark(i_bbl);

	  in = BblComputeLiveInFromLiveOut(i_bbl,1);

	  /* The rest is passing everything to the predecessors or function information */
	  if( BBL_IS_FIRST(i_bbl) && !FUN_IS_HELL(i_fun) && FunctionGetExitBlock(i_fun))
	    {	 
	      RegsetDup(FUN_REGS_THROUGH(i_fun),in);
	      RegsetSetUnion(FUN_REGS_THROUGH(i_fun),FUN_REGS_SAVED(i_fun));
	    }

	  change |= BblPropLivenessToPredButNoReturnEdge(i_bbl,in,description);

	  big_change |= change;
	  RegsetFree(in);
	}
      } while (change);
    }
  }
  while(big_change);

}


static void ComputeFunctionUsedRegs(t_cfg * cfg)
{
  /* iterators */

  t_bbl * i_bbl;
  t_function * i_fun;

  const t_architecture_description * description = cfg->sec->obj->description;

  /* local variables */

  t_bool change,big_change;
  t_regset in;

  /*
     we assume that bbl and fun stuff has been computed already
     here we just do the do fixpoint computation
   */

  InitSensitiveLivenessFixpoint(cfg);

  do
  {
    big_change = FALSE;
    CFG_FOREACH_FUN(cfg,i_fun)
    {
      if(!(FUN_FLAGS(i_fun)&FF_IS_MARKED))
	continue;

      FUN_FLAGS(i_fun)&=~FF_IS_MARKED;

      do
      {
	change = FALSE;

	FUN_FOREACH_BBL_R(i_fun,i_bbl)
	{
	  if(!BblIsMarked(i_bbl))
	    continue;

	  BblUnmark(i_bbl);

	  in = BblComputeLiveInFromLiveOut(i_bbl,2);

	  /* The rest is passing everything to the predecessors or function information */
	  if( BBL_IS_FIRST(i_bbl) && !FUN_IS_HELL(i_fun) )
	    {
	      /* Some registers are still used after they're saved on the stack,
	       * we have to include them in FUN_REGS_USED */
	      RegsetDup(FUN_REGS_USED(i_fun),in);
	      RegsetSetDiff(FUN_REGS_USED(i_fun),FUN_REGS_SAVED(i_fun));
	    }

	  change |= BblPropLivenessToPredButNoReturnEdge(i_bbl,in,description);
	  big_change |= change;
	  RegsetFree(in);
	}
      } while (change);
    }
  }
  while(big_change);
}

static void SensitiveLivenessFixpoint(t_cfg * cfg)
{
  /* iterators */

  t_bbl * i_bbl;
  t_function * i_fun;
#ifdef LIVENESS_OPTION_CONDITIONAL_CALL_OPT
  t_cfg_edge *i_edge;
#endif
  
  const t_architecture_description * description = cfg->sec->obj->description;

  /* local variables */

  t_bool change,big_change;
  t_regset in;
  /*
     we assume that bbl and fun stuff has been computed already
     here we just do the do fixpoint computation
   */

  InitSensitiveLivenessFixpoint(cfg);

  do
  {
    big_change = FALSE;
    CFG_FOREACH_FUN(cfg,i_fun)
    {
      if(!(FUN_FLAGS(i_fun)&FF_IS_MARKED))
	continue;

      FUN_FLAGS(i_fun)&=~FF_IS_MARKED;

      do
      {
	change = FALSE;

	FUN_FOREACH_BBL_R(i_fun,i_bbl)
	{
	  if(!BblIsMarked(i_bbl))
	    continue;

	  BblUnmark(i_bbl);

	  in = BblComputeLiveInFromLiveOut(i_bbl,0);

	  /* The rest is passing everything to the predecessors or function information */

	  change |= BblPropLivenessToPredIncludingReturnEdge(i_bbl,in,description);
	  big_change |= change;
	  RegsetFree(in);
	}
      } while (change);
    }
  }
  while(big_change);

#ifdef LIVENESS_OPTION_CONDITIONAL_CALL_OPT
  CFG_FOREACH_EDGE(cfg,i_edge)
    {
      if (EDGE_CAT(i_edge)==ET_CALL)
	if (BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)) && INS_IS_CONDITIONAL(BBL_INS_LAST(CFG_EDGE_HEAD(i_edge))))
	  {
	    /*	    DiabloPrint(stdout,"EXTRA FOR @B\n",CFG_EDGE_HEAD(i_edge));*/
	    RegsetSetUnion(BBL_REGS_LIVE_OUT(CFG_EDGE_HEAD(i_edge)),description->link_register);
	  }
    }
#endif
}


/* }}} */

void ComLiveness(t_cfg * cfg, liveness_level level)
{

  t_regset unused_regs = AllRegs();
  t_bbl * bbl;
  t_ins * ins;

  
  if(level==SENSITIVE)
    {
    ComLivenessSensitive(cfg);
  }
  else if(level==INSENSITIVE)
  {
    ComLivenessInsensitive(cfg);
  }
  else if(level==TRIVIAL)
  {
    ComLivenessTrivial(cfg);
  }
  else
  {
    FATAL(("Bad liveness analysis method\n"));
  } 

  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS(bbl,ins)
    {
      RegsetSetDiff(unused_regs,INS_REGS_USE(ins));
    }

  CFG_FOREACH_BBL(cfg,bbl)
    {
      RegsetSetDiff(BBL_REGS_LIVE_OUT(bbl),unused_regs);
    }

  /*  DiabloPrintArch(stdout,description,"saved: @A",unused_regs);*/

#ifdef IA64_SUPPORT
  Ia64UpdateLiveness(cfg);
#endif  
}

/* Dump Liveness Info {{{ */
/* This function dumps liveness info for all instructions in the object into the specified file */
void DumpLivenessInfo(t_string filename, t_object * obj)
{
  t_bbl * i_bbl;
  t_cfg * cfg = SECTION_CFG(obj->code[0]);
  FILE * f = fopen(filename,"w");

  if (!f) FATAL(("Could not open file %s for writing!",filename));

  /* compute accurate liveness information */
  ComputeSavedChangedRegisters(*(obj->code));
#ifdef IA64_SUPPORT
  Ia64UpdateLiveness(cfg);
#endif
  ComLiveness(cfg,SENSITIVE);

  /* now iterate over all blocks in the flowgraph and write out the liveness information */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_ins * i_ins;
    t_regset *live;
    t_uint32 count = 0;
    t_uint32 nins = BBL_NINS(i_bbl);

    live = (t_regset *) Malloc(sizeof(t_regset) * (nins+1));
    
    RegsetDup(live[nins],BBL_REGS_LIVE_OUT(i_bbl));
    
    BBL_FOREACH_INS_R(i_bbl,i_ins)
    {
      count++;
      RegsetDup(live[nins-count],live[nins-count+1]);
      RegsetSetDiff(live[nins-count],INS_REGS_DEF(i_ins));
      RegsetSetUnion(live[nins-count],INS_REGS_USE(i_ins));
    }

    /* now write the information to the file */
    fprintf(f,">>> %p\n",i_bbl);
    for (count = 0; count <= nins; count++)
    {
      PrintRegset(f,obj->description,live[count]);
    }
    Free(live);
  }
  fclose(f);
}

static void ComputeMeaningfulRegs(t_cfg * cfg)
{
#ifdef FORWARD_LIVENESS
  t_bbl * bbl;
  t_cfg_edge * edge;
  t_bool change;
  t_function * fun;

  const t_action_struct * actions = cfg->sec->obj->actions;

  const t_architecture_description * description = cfg->sec->obj->description;

  t_bool marked_interprocedural, marked_intraprocedural;
  t_ins * ins;

  if (global_options.rely_on_calling_conventions)
    {
      
      InitBblMarking();
      
      CFG_FOREACH_BBL(cfg,bbl)
	{
	  BblMark(bbl);
	  BBL_REGS_DEFINED_IN(bbl)=RegsetNew();
	  BBL_FOREACH_INS(bbl,ins)
	    RegsetSetUnion(BBL_REGS_DEFINED(bbl),INS_REGS_DEF(ins));
	}
      
      CFG_FOREACH_FUN(cfg,fun)
	{
	  FUN_FLAGS(fun)|=FF_IS_MARKED;
	}
      
      BBL_FOREACH_SUCC_EDGE(cfg->hell_node,edge)
	{
	  if (EDGE_CAT(edge)==ET_CALL)
	    {
	      RegsetDup(BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge)),description->callee_may_use);
	      RegsetSetUnion(BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge)),description->callee_saved);
	      RegsetSetUnion(BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge)),description->always_live);
	    }
	  else if (EDGE_CAT(edge)!=ET_RETURN)
	    {
	      BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge))=AllRegs();      
	    }
	}
      
      BBL_FOREACH_SUCC_EDGE(cfg->call_hell_node,edge)
	{
	  if (EDGE_CAT(edge)==ET_CALL)
	    {
	      RegsetDup(BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge)),description->callee_may_use);
	      RegsetSetUnion(BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge)),description->callee_saved);
	      RegsetSetUnion(BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge)),description->always_live);
	    }
	  else if (EDGE_CAT(edge)!=ET_RETURN)
	    {
	      BBL_REGS_DEFINED_IN(CFG_EDGE_TAIL(edge))=AllRegs();      
	    }
	}
      
      RegsetDup(BBL_REGS_DEFINED_IN(cfg->entries->entry_bbl),description->callee_may_use);
      RegsetSetUnion(BBL_REGS_DEFINED_IN(cfg->entries->entry_bbl),description->callee_saved);
      RegsetSetUnion(BBL_REGS_DEFINED_IN(cfg->entries->entry_bbl),description->always_live);
      
      do
	{
	  marked_interprocedural = 0;
	  
	  CFG_FOREACH_FUN(cfg,fun)
	    {
	      if(!(FUN_FLAGS(fun)&FF_IS_MARKED))
		continue;
	      
	      FUN_FLAGS(fun)&=~FF_IS_MARKED;
	      
	      do
		{
		  marked_intraprocedural = 0;
		  
		  FUN_FOREACH_BBL(fun,bbl)
		    {
		      if(!BblIsMarked(bbl))
			continue;
		      
		      BblUnmark(bbl);
		      
		      if (IsHellNode(bbl)) continue;
		      
		      BBL_FOREACH_SUCC_EDGE(bbl,edge)
			{
			  t_bbl * tail = CFG_EDGE_TAIL(edge);
			  t_regset old, new;
			  t_regset max;
			  
			  if (!tail) continue;
			  
			  if (IsHellNode(tail) && EDGE_CAT(edge)!=ET_CALL) continue;

			  RegsetDup(old,BBL_REGS_DEFINED_IN(tail));
			  RegsetDup(new,old);

			  RegsetDup(max,BBL_REGS_DEFINED_IN(bbl));
			  RegsetSetUnion(max,BBL_REGS_DEFINED(bbl));

			  RegsetSetIntersect(max,BBL_REGS_LIVE_OUT(bbl));

			  switch (EDGE_CAT(edge))
			    {
			    case ET_CALL:
			      {
				t_function * callee = BBL_FUN(tail);
				t_regset old2, new2;
				if (IsHellNode(tail))
				  {
				    t_bbl * return_node = CFG_EDGE_TAIL(edge->corr_edge);
				    t_regset prop;
				    RegsetDup(prop,description->callee_may_return);
				    RegsetSetUnion(prop,description->always_live);
				    RegsetSetUnion(prop,RegsetNewIntersect(description->callee_saved,max));
				
				    RegsetDup(old2,BBL_REGS_DEFINED_IN(return_node));
				    RegsetDup(new2,old2);
				    RegsetSetUnion(new2,prop);

				    if (!RegsetEquals(new2,old2))				  
				      {
					BblMark(return_node);
					RegsetDup(BBL_REGS_DEFINED_IN(return_node),new2);
					if (BBL_FUN(return_node) != BBL_FUN(bbl))
					  {
					    if (BBL_FUN(tail))
					      {
						FUN_FLAGS(BBL_FUN(tail))|=FF_IS_MARKED;
						marked_interprocedural=1;
					      }
					  }
					else
					  marked_intraprocedural=1;
					change = TRUE;
				      }
				  }
				else if (edge->corr_edge)
				  {
				    t_bbl * return_node;
				    t_regset prop;

				    return_node = CFG_EDGE_TAIL(edge->corr_edge);


				    RegsetSetIntersect(max,RegsetNewUnion(RegsetNewUnion(FUN_REGS_THROUGH(callee),FUN_REGS_USED(callee)),description->always_live));
				    if (actions->FunIsGlobal(callee) && actions->InsIsIndirectCall(BBL_INS_LAST(bbl)))
				      RegsetSetIntersect(max,RegsetNewUnion(RegsetNewUnion(description->callee_may_use,description->callee_saved),description->always_live));
				
				    if (actions->FunIsGlobal(callee) && actions->InsIsIndirectCall(BBL_INS_LAST(bbl)))
				      {
					RegsetDup(prop,RegsetNewIntersect(description->callee_saved,max));
				      }
				    else
				      {
					RegsetDup(prop,RegsetNewIntersect(FUN_REGS_THROUGH(callee),max));
				      }
				
				    RegsetDup(old2,BBL_REGS_DEFINED_IN(return_node));
				    RegsetDup(new2,old2);
				    RegsetSetUnion(new2,prop);

				    if (!RegsetEquals(new2,old2))				  
				      {
					BblMark(return_node);
					RegsetDup(BBL_REGS_DEFINED_IN(return_node),new2);
					if (BBL_FUN(return_node) != BBL_FUN(bbl))
					  {
					    if (BBL_FUN(tail))
					      {
						FUN_FLAGS(BBL_FUN(tail))|=FF_IS_MARKED;
						marked_interprocedural=1;
					      }
					  }
					else
					  marked_intraprocedural=1;
					change = TRUE;
				      }
				  }
				break;
			      }

			    case ET_RETURN:
			      {
				t_function * callee = BBL_FUN(bbl);
				t_regset def_caller_out;
				t_bbl * caller = CFG_EDGE_HEAD(edge->corr_edge);

				if (callee->name && PatternMatch(callee->name,"exit"))
				  continue;

				RegsetDup(def_caller_out,RegsetNewUnion(BBL_REGS_DEFINED(caller),BBL_REGS_DEFINED_IN(caller)));
				RegsetSetIntersect(def_caller_out,BBL_REGS_LIVE_OUT(caller));
			    
				if (actions->FunIsGlobal(callee) && actions->InsIsIndirectCall(BBL_INS_LAST(caller)))
				  RegsetSetIntersect(def_caller_out,description->callee_saved);
				else
				  RegsetSetIntersect(def_caller_out,FUN_REGS_THROUGH(callee));
			    
				if (actions->FunIsGlobal(callee) || actions->InsIsIndirectCall(BBL_INS_LAST(caller)))
				  {
				    RegsetSetIntersect(max,description->callee_may_return);
				  }

				RegsetSetUnion(max,def_caller_out);
				RegsetSetUnion(max,description->always_live);
			      }
			      break;
			    case ET_IPFALLTHRU:
			      if (BBL_FUN(bbl) && BBL_FUN(bbl)->name  && PatternMatch(BBL_FUN(bbl)->name,"exit"))
				continue;
			      break;
			    case ET_FUNLINK:
			      /*			    case ET_IPFUNLINK: LINK_REMOVAL*/
			      continue;
			    case ET_SWILINK:
			    case ET_IPSWILINK:
			      /* TODO: should verify that no edges are added following call_palls that are 
				 padding !!!
				 if (ALPHAINSIMM(BBL_INS_LAST(bbl))==0)
				 continue;*/
			      RegsetSetIntersect(max,description->callee_saved);
			      RegsetSetUnion(max,description->always_live);
			      RegsetSetUnion(max,description->callee_may_return);
			      RegsetSetUnion(max,description->callee_may_use);
			    }
		      
			  RegsetSetDiff(max,description->const_registers);
			  RegsetSetUnion(new,max);
		      
			  if (!RegsetEquals(new,old))
			    {
			      BblMark(tail);
			      RegsetDup(BBL_REGS_DEFINED_IN(tail),new);
			      if (CfgEdgeIsInterproc(edge))
				{
				  if (BBL_FUN(tail))
				    {
				      FUN_FLAGS(BBL_FUN(tail))|=FF_IS_MARKED;
				      marked_interprocedural=1;
				    }
				}
			      else
				marked_intraprocedural=1;
			      change = TRUE;
			    }
			}
		    }
		} while (marked_intraprocedural);
	    }
	}
      while(marked_interprocedural);
  

      CFG_FOREACH_BBL(cfg,bbl)
	{
	  t_regset live_out;
	  t_regset defined_out;
	  RegsetDup(live_out,BBL_REGS_LIVE_OUT(bbl));
	  RegsetDup(defined_out,BBL_REGS_DEFINED_IN(bbl));
	  RegsetSetUnion(defined_out,BBL_REGS_DEFINED(bbl));
	  RegsetSetIntersect(defined_out,live_out);
	  if (!RegsetIsEmpty(RegsetNewDiff(live_out,defined_out)))
	    {
	      /*	      DiabloPrintArch(stdout,description,"@B\nFROM @A\nTO @A\n",bbl,BBL_REGS_LIVE_OUT(bbl),defined_out);*/
	      RegsetDup(BBL_REGS_LIVE_OUT(bbl),defined_out);
	    }
	}
    }
#endif

#if 0
  int regsets[65];
  int i;
  t_reg reg;
  for (i=0;i<65;i++)
    regsets[i]=0;
      
  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (IsHellNode(bbl)) continue;
      if (!BBL_FUN(bbl)) continue;
      if (BBL_NINS(bbl)==0) continue;
      if (FUN_BBL_LAST(BBL_FUN(bbl))==bbl) continue;
      if (BBL_INS_FIRST(bbl) && INS_TYPE(BBL_INS_FIRST(bbl))==IT_DATA) continue;
      if ((GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl)) & 0xffff) == 0xffff) continue;
	  
      if (RegsetCountRegs(RegsetNewDiff(RegsetNewIntersect(BBL_REGS_LIVE_OUT(bbl),description->int_registers),description->const_registers))==0)
	{
	  /*printf("old: %x",GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl)));
	    DiabloPrintArch(stdout,description,"block with zero live registers @iB\n@A",bbl,BBL_REGS_LIVE_OUT(bbl) & description->int_registers);*/
	  continue;
	}
      regsets[RegsetCountRegs(RegsetNewDiff(RegsetNewIntersect(BBL_REGS_LIVE_OUT(bbl),description->int_registers),description->const_registers))]++;
    }
      
  printf("----------------------------------\n");
      
  for (i=0;i<RegsetCountRegs(description->int_registers);i++)
    {
      printf("REGS %10d\n",regsets[32-i-1]);
    }

  for (i=0;i<65;i++)
    regsets[i]=0;
      
  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (IsHellNode(bbl)) continue;
      if (!BBL_FUN(bbl)) continue;
      if (BBL_NINS(bbl)==0) continue;
      if (FUN_BBL_LAST(BBL_FUN(bbl))==bbl) continue;
      if (BBL_INS_FIRST(bbl) && INS_TYPE(BBL_INS_FIRST(bbl))==IT_DATA) continue;
      if ((GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl)) & 0xffff) == 0xffff) continue;
	  
      if (RegsetCountRegs(RegsetNewDiff(RegsetNewIntersect(BBL_REGS_LIVE_OUT(bbl),description->int_registers),description->const_registers))==0)
	{
	  /*printf("old: %x",GENERIC_EXTRACT_UINT32(BBL_OLD_ADDRESS(bbl)));
	    DiabloPrintArch(stdout,description,"block with zero live registers @iB\n@A",bbl,BBL_REGS_LIVE_OUT(bbl) & description->int_registers); */
	  continue;
	}
      REGSET_FOREACH_REG(RegsetNewDiff(description->int_registers,RegsetNewDiff(RegsetNewIntersect(BBL_REGS_LIVE_OUT(bbl),description->int_registers),description->const_registers)),reg)
	regsets[reg]++;
    }
      
  printf("----------------------------------\n");
      
  for (i=0;i<32;i++)
    {
      printf("HIST %10d\n",regsets[i]);
    }      
#endif

#if 0
  {
    t_bbl * bbl;
    CFG_FOREACH_BBL(cfg,bbl)
      DiabloPrintArch(stdout,description,"LIVE AFTER @iB\n @A\n",bbl,BBL_REGS_LIVE_OUT(bbl));
    exit(0);
  }
#endif

}

/*}}}*/

/* {{{ Dump dot files annotated with liveness information */
static t_bool _print_live_before = FALSE;
static t_bool _print_live_after  = FALSE;

static void __helper_annotate_liveness(t_bbl * bbl, FILE * fp)
{
  if (_print_live_before)
    DiabloPrintArch(fp,bbl->cfg->sec->obj->description,"live before: @A\\n",BblRegsLiveBefore(bbl));

  if (_print_live_after)
    DiabloPrintArch(fp,bbl->cfg->sec->obj->description,"live after: @A\\n",BblRegsLiveAfter(bbl));
}

void DumpCfgWithLiveness(t_cfg * cfg, t_string dotsdir, t_bool livebefore, t_bool liveafter, t_bool compute_fresh)
{
  t_function * fun;
  char filename[200];

  _print_live_before = livebefore;
  _print_live_after  = liveafter;

  if (compute_fresh)
  {
    ComLiveness(cfg,TRIVIAL);
    ComputeSavedChangedRegisters(cfg->sec);
    ComLiveness(cfg,SENSITIVE);
    ComputeSavedChangedRegisters(cfg->sec);
    ComLiveness(cfg,SENSITIVE);
  }

  DirectoryAssure(dotsdir,FALSE);
  CFG_FOREACH_FUN(cfg,fun)
  {
    if (FUN_NAME(fun))
      sprintf(filename,"%s/%s.dot",dotsdir,FUN_NAME(fun));
    else
#ifndef BIT64ADDRSUPPORT
      sprintf(filename,"%s/noname-0x%x.dot",dotsdir,BBL_OLD_ADDRESS(FUN_BBL_FIRST(fun)));
#else
      sprintf(filename,"%s/noname-0x%llx.dot",dotsdir,BBL_OLD_ADDRESS(FUN_BBL_FIRST(fun)));
#endif
    Export_Function_Annotate(fun,filename,__helper_annotate_liveness);
  }
}
/* }}} */

#define BBL_REGS_SAME_IN(bbl) (BBL_REGS_DEF_CERTAIN(bbl))
#define BBL_REGS_SAME(bbl) (BBL_REGS_DEF_PERHAPS(bbl))

t_regset ComputeSameOut(t_bbl * bbl)
{

  t_ins * ins;
  t_regset return_value;

  RegsetDup(return_value,BBL_REGS_SAME_IN(bbl));

  BBL_FOREACH_INS(bbl,ins)
    {
      if (INS_TYPE(ins)==IT_PSEUDO_LOAD || INS_TYPE(ins)==IT_PSEUDO_SAVE)
	{
	  RegsetSetUnion(return_value,INS_REGS_DEF(ins));
	  RegsetSetUnion(return_value,INS_REGS_USE(ins));
	}
      else
	RegsetSetDiff(return_value,INS_REGS_DEF(ins));
    }

  return return_value;
}

void ComputePseudoLoadsAndSaves(t_cfg * cfg)
{
  t_bbl * bbl;
  t_bool change, big_change;
  t_ins * ins;
  t_cfg_edge * edge;
  t_function * fun;

  const t_action_struct * actions = cfg->sec->obj->actions;
  
  const t_architecture_description * description = cfg->sec->obj->description;

  ComLiveness(cfg,INSENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);

  InitNodeMarking();

  CFG_FOREACH_FUN(cfg,fun)
    FUN_FLAGS(fun)|=FF_IS_MARKED;

  CFG_FOREACH_BBL(cfg,bbl)
    {
      t_regset same = RegsetNew();
      t_regset tmp;
      if (BBL_PRED_FIRST(bbl)==NULL && BBL_SUCC_FIRST(bbl)==NULL) continue;
      if (!BBL_FUN(bbl)) continue;

      NodeMark(bbl);

      RegsetDup(BBL_REGS_SAME_IN(bbl),description->int_registers);

      BBL_FOREACH_INS(bbl,ins)
	{
	  if (INS_TYPE(ins)==IT_PSEUDO_LOAD || INS_TYPE(ins)==IT_PSEUDO_SAVE)
	    {
	      RegsetSetUnion(same,INS_REGS_DEF(ins));
	      RegsetSetUnion(same,INS_REGS_USE(ins));
	    }
	  else
	    RegsetSetDiff(same,INS_REGS_DEF(ins));
	}
      RegsetDup(BBL_REGS_DEFINED(bbl), tmp = BblRegsDef(bbl));
      RegsetDup(BBL_REGS_SAME(bbl),same);
      
      NodeMark(bbl);

    }
  
  do
    {
      big_change = FALSE;
      CFG_FOREACH_FUN(cfg,fun)
	{
	  if(!(FUN_FLAGS(fun)&FF_IS_MARKED))
	    continue;
	  
	  FUN_FLAGS(fun)&=~FF_IS_MARKED;
	  
	  do
	    {
	      change = FALSE;
	      
	      FUN_FOREACH_BBL(fun,bbl)
		{
		  t_regset same_out;

		  if (!NodeIsMarked(bbl)) continue;
		  
		  NodeUnmark(bbl);
		  
		  if (!BBL_FUN(bbl)) continue;
		  
		  RegsetDup(same_out,BBL_REGS_SAME_IN(bbl));
		  
		  RegsetSetDiff(same_out,BBL_REGS_DEFINED(bbl));
		  
		  RegsetSetUnion(same_out,BBL_REGS_SAME(bbl));

		  BBL_FOREACH_SUCC_EDGE(bbl,edge)
		    {
		      t_regset new_same_out;
		      
		      t_bbl * tail = CFG_EDGE_TAIL(edge);
		      
		      if (EDGE_CAT(edge)==ET_RETURN)
			{
			  if (global_options.rely_on_calling_conventions)
			    {
			      RegsetDup(new_same_out,RegsetNewUnion(same_out,RegsetNewInvers(FUN_REGS_CHANGED(BBL_FUN(CFG_EDGE_HEAD(edge))))));
			      if (actions->FunIsGlobal(fun) || FunIsCalledByHell(fun))
				RegsetSetIntersect(new_same_out,description->callee_may_change);
			    }
			  else RegsetDup(new_same_out,same_out);

			}

		      else if (EDGE_CAT(edge)==ET_FUNLINK /* LINK_REMOVAL || EDGE_CAT(edge)==ET_IPFUNLINK*/)
			{
			  t_cfg_edge * call;
			  t_function * callee;
			  
			  if (global_options.rely_on_calling_conventions)
			    {  
			      BBL_FOREACH_SUCC_EDGE(bbl,call)
				if (EDGE_CAT(call)==ET_CALL) break;
			      
			      if (!call) FATAL(("No function call!"));
			      
			      callee = BBL_FUN(CFG_EDGE_TAIL(call));
			      
			      RegsetDup(new_same_out,RegsetNewUnion(same_out,FUN_REGS_CHANGED(callee)));
			    }
			  else
			    continue;
			}
		      else if (EDGE_CAT(edge)==ET_CALL)
			{
			  if (global_options.rely_on_calling_conventions)
			    {
			      RegsetDup(new_same_out,same_out);
			      if (actions->FunIsGlobal(BBL_FUN(CFG_EDGE_TAIL(edge)))  || FunIsCalledByHell(BBL_FUN(CFG_EDGE_TAIL(edge))) )
				RegsetSetIntersect(new_same_out,description->callee_may_use);
			    }
			  else
			    {
			      RegsetDup(new_same_out,same_out);
			    }
			}
		      else
			{
			  RegsetDup(new_same_out,same_out);
			}
		      
		      if (!RegsetEquals((RegsetNewIntersect(BBL_REGS_SAME_IN(tail),new_same_out)),BBL_REGS_SAME_IN(tail)))
			{
			  change |= TRUE;
			  big_change |= change;
			  if (CfgEdgeIsInterproc(edge))
			    FUN_FLAGS(BBL_FUN(tail))|=FF_IS_MARKED;
			  NodeMark(tail);
			  RegsetSetIntersect(BBL_REGS_SAME_IN(tail),new_same_out);
			}
		    }
		}
	    } while (change);
	}
    } while (big_change);
}

void OptimizePseudoSavesLoads(t_cfg * cfg)
{

  t_bbl * bbl;
  t_ins * ins, * tmp_ins;

 redo:

  ComputePseudoLoadsAndSaves(cfg);

  /* {{{ optimize */
    /*static int teller = 0;*/

  CFG_FOREACH_BBL(cfg,bbl)
    {
      t_regset same = RegsetNew();

      RegsetDup(same,BBL_REGS_SAME_IN(bbl));
      
      BBL_FOREACH_INS_SAFE(bbl,ins,tmp_ins)
	{
	  if (INS_TYPE(ins)==IT_PSEUDO_LOAD || INS_TYPE(ins)==IT_PSEUDO_SAVE)
	    {
	      if ((RegsetIsSubset(same,INS_REGS_USE(ins)) && INS_TYPE(ins)==IT_PSEUDO_SAVE) || (RegsetIsSubset(same,INS_REGS_DEF(ins)) && INS_TYPE(ins)==IT_PSEUDO_LOAD) )
		{
		  /*if (teller++ <  global_options.debugcounter)*/
		  {
		    /*DiabloPrintArch(stdout,description,"SAME IN @A\nSAME @A\nBBL @ieB\n",BBL_REGS_SAME_IN(bbl),BBL_REGS_SAME(bbl),bbl);*/
		    /*DiabloPrint(stdout,"KILLING LOAD OR SAVE @I\n",ins);*/
		    InsKill(ins);
		  }
		}
	      else
		{
		  RegsetSetUnion(same,INS_REGS_DEF(ins));
		  RegsetSetUnion(same,INS_REGS_USE(ins));
		}
	    }
	  else
	    {
	      RegsetSetDiff(same,INS_REGS_DEF(ins));
	    }
	}
    } /* }}} */
  
  ComLiveness(cfg,INSENSITIVE);
  ComputeSavedChangedRegisters(cfg->sec);
  ComLiveness(cfg,SENSITIVE);
  OptKillUseless(cfg);

  if (global_options.rely_on_calling_conventions == TRUE)
    { 
      global_options.rely_on_calling_conventions = FALSE;
      goto redo;
    }
  
  ComLiveness(cfg,INSENSITIVE);
  OptKillUseless(cfg);

  /*
  CFG_FOREACH_BBL(cfg,bbl)
    {
      t_regset movable_regs = RegsetNew();
      t_bool hoising_may_be_useful = FALSE;

      BBL_FOREACH_INS_R(bbl,ins)
	{
	  if (INS_TYPE(ins)==IT_PSEUDO_LOAD)
	    RegsetSetUnion(movable_regs,INS_REGS_DEFINED(ins));
	  else if (INS_TYPE(ins)==IT_PSEUDO_SAVE)
	    RegsetSetUnion(movable_regs,INS_REGS_USED(ins));
	  else
	    {
	      RegsetSetDiff(movable_regs,INS_REGS_USED(ins));
	      RegsetSetDiff(movable_regs,INS_REGS_DEFINED(ins));
	    }
	}

      if (RegsetIsEmpty(movable_regs)) continue;

    }
  */

  {  
    static unsigned dist[20];
    int i;
    CFG_FOREACH_BBL(cfg,bbl)
      {
	int teller = 0;
	BBL_FOREACH_INS(bbl,ins)
	  {
	    if (INS_TYPE(ins) == IT_PSEUDO_LOAD || INS_TYPE(ins) == IT_PSEUDO_SAVE)
	      teller++;
	  }
	dist[teller]++;
      }
    for (i=0; i<20; i++)
      printf("dist[%d] = %u\n",i,dist[i]);
  }
}


/* vim: set shiftwidth=2 foldmethod=marker: */
