/*{{{
 * Copyright 2001,2002,2003:
 * 			Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Headers for flowgraph manipulation 
 *
 * Written by Ludo Van Put (lvanput@elis.ugent.be)
 *
 * Copyright (c) 2002 Ludo Van Put
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Ludo Van Put (lvanput@elis.ugent.be) 
 ************************************************************}}}*/

#ifndef MY_LOOP_TYPEDEFS
#define MY_LOOP_TYPEDEFS
typedef struct _t_loop t_loop;
typedef struct _t_loopelem t_loopelem;
typedef struct _t_loopref t_loopref;
typedef struct _t_loopfunref t_loopfunref;
typedef struct _t_loopiterator t_loopiterator;
#endif

#ifndef MY_LOOP_H
#define MY_LOOP_H
#include "diablo_bbl.h"
#include "diablo_edge.h"
#include "diablo_flowgraph.h"
#include "../datastructures/data_dll.h"
#include "../datastructures/data_ll.h"

#define HASH_TABLE_SIZE 0x10
#define HASHINDEX(bbl)	(((*((unsigned int*)(&bbl)))>>4) & (HASH_TABLE_SIZE-1))

struct _t_loop
{
  /* Structures to maintain a dll of loops */
  struct _t_loop * prev;
  struct _t_loop * next;
  /* The edge that was responsible for uniquely defining this loop. */
  data_ll_elem * backedges;
  /* A table of linked lists, hashed by a very simple function */
  struct _t_loopelem * elems[16];
  struct _t_loopfunref * functions;
  /* The number of elements in this loop */
  t_uint32 count;
  /* Is the loop infinite? */
  t_bool infinite;
  /* The flowgraph this loop is in */
  t_cfg * cfg;
  /* The loop in which this loop is nested */
  struct _t_loopref * parents;
  struct _t_loopref * childs;
};

struct _t_loopiterator
{
  struct _t_loopelem * iterator_elem;
  t_uint32 iterator_hash;
  struct _t_loopfunref * iterator_fun;
  t_bbl * iterator_bbl;
  t_loop * loop;
};

#define T_LOOP(loop)		((t_loop*)loop)
#define LOOP_NEXT(loop)		(T_LOOP(loop)->next)
#define LOOP_PREV(loop)		(T_LOOP(loop)->prev)
#define LOOP_BACKEDGES(loop)	(T_LOOP(loop)->backedges)
#define LOOP_HEADER(loop)	(CFG_EDGE_TAIL(LOOP_BACKEDGES(loop)->data))
#define LOOP_ELEMS(loop)	(T_LOOP(loop)->elems)
#define LOOP_COUNT(loop)	(T_LOOP(loop)->count)
#define LOOP_CFG(loop)		(T_LOOP(loop)->cfg)
#define LOOP_INF(loop)		(T_LOOP(loop)->infinite)
#define LOOP_PARENTS(loop)	(T_LOOP(loop)->parents)
#define LOOP_CHILDS(loop)	(T_LOOP(loop)->childs)
#define LOOP_FUNCTIONS(loop)	(T_LOOP(loop)->functions)

struct _t_loopelem
{
  /* Structure to make a linked list */
  ll ll;
  /* The bbl that is in the loop */
  t_bbl * bbl;
  /* A bool that indicates if this bbl has an edge leaving the loop */
  t_bool exit;
};

struct _t_loopfunref
{
  ll ll;
  t_function * fun;
};

struct _t_loopref
{
  ll ll;
  t_loop * loop;
};
#define T_LOOPELEM(elem)	((t_loopelem *)elem)
#define LOOPELEM_BBL(elem)	(T_LOOPELEM(elem)->bbl)
#define LOOPELEM_EXIT(elem)	(T_LOOPELEM(elem)->exit)
#define T_LOOPREF(elem)		((t_loopref *) elem)
#define LOOPREF_LOOP(elem)	(T_LOOPREF(elem)->loop)

t_loop * LoopNew(t_cfg * cfg, t_cfg_edge * backedge);
void LoopKill(t_loop**);
t_bool LoopContainsBbl(t_loop * loop, t_bbl * bbl);
void LoopBblCleanup(t_bbl * bbl);
t_int32 EdgeLoopCount(t_cfg_edge * edge);
t_bool LoopBblIsLoopExit(t_loop * loop, t_bbl * bbl);
void LoopFindNestedLoops(t_cfg * cfg);
t_bool BblIsLoopHeader(t_bbl * bbl);
t_int32 EdgeLoopEntries(t_cfg_edge * edge);
t_int32 EdgeLoopExits(t_cfg_edge * edge);
void LoopMergeLoopsWithSameHeader(t_cfg * cfg);
void CfgMarkFunsWithoutIncomingJumps(t_cfg * cfg);
t_loopiterator * LoopNewIterator(t_loop * loop);
t_bbl * LoopGetNextBbl(t_loopiterator * iterator);

void TestBblOnAddress(t_cfg * cfg);

#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
