/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains function definitions to parse the 
 *       linker-generated map (this is a map containing the
 *       link- order of the different subobjects) and load
 *       all the objects subobjects. 
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

#define PARSEBUFFER 1000 /* TODO*/

/* Includes {{{ */
#include "../config.h"
#include "../datastructures/data_string.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_file.h"
#include "../utils/util_status.h"
#include "../simple_types.h"
#include "../diablo_options.h"
#include "diablo_parsemap.h"
#include "diablo_print.h"
#include "diablo_object.h"
#include "diablo_section.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
/*}}}*/

/* Node GNU ld constructor: TODO rename {{{ */
/*! 
 * parse a line from an arm map-file, returns a t_map_node.  
 * Memory allocated, don't forget to free! 
 */

t_map_node * I86Node(t_string name, t_string mapped_name, t_address start, t_address size, t_string file)
{
   t_map_node * ret=Calloc(1,sizeof(t_map_node));
   ret->base_addr=start;
   ret->sec_size=size;
   if ((file[0]!='(')&&(PatternMatch("*(*)",file)))
      {
	 t_string dup=StringDup(file);
	 t_string lib=strtok(dup,"(");
	 t_string obj=strtok(NULL,")");
	 t_string Name=NULL;
	 VERIFY_FASSERT(obj,("Illegal objectname in map: %s",file));
	 VERIFY_FASSERT(lib,("Illegal objectname in map: %s",file));
	 Name=Collate3(lib,":",obj);
	 Free(dup);
	 ret->object=Name;
      }
   else
   {
     ret->object=StringDup(file);
   }
   ret->sec_name=StringDup(name);
   ret->mapped_sec_name=StringDup(mapped_name);
   ret->idx=0;
   VERBOSE(2,("Section name= %s, starts at %x, size=%x, from file %s",name,start,size,file)); 
   return ret;
}
/* }}} */
/* Node Alpha Linker Constructor: TODO rename {{{ */
#ifdef ALPHA_SUPPORT
/*! TODO: document */
t_map_node * AlphaMapParseLine(t_string line)
{
   static t_uint16 tel=0;
   t_map_node * new_node=NULL;
   t_string_array * array=StringDivider(line," \t\n",FALSE,FALSE);
   t_string joined=StringArrayJoiner(array,":");
   static t_string last=NULL;
   t_uint64 sec_start, sec_size;
   t_uint64 tmp;
   t_uint16 size=StringArrayNElems(array);

   VERIFY_FASSERT((size==4)||(size==3)||(size==6),("Alpha Map file parse error! Line= %s",line));
   if (strcmp(joined,"LINK:EDITOR:MEMORY:MAP")==0) return NULL;
   if (strcmp(joined,"output:input:virtual")==0) return NULL;
   if (strcmp(joined,"section:section:address:size")==0) return NULL;

   if (size==6)
   {
      if ((strcmp(array->strings[3],"exception")==0) && (strcmp(array->strings[4],"gp")==0) && (strcmp(array->strings[5],"info")==0))
      {
	 WARNING(("Exception gp info... Not supported... Maybe this object has multiple gp's? Or is it dynamically linked?"));
	 return NULL;
      }
      else
      {
	 FATAL(("Expected exception gp info: %s",line));
      }
      
   }
   if (size==3)
   {
      if (last) if (strcmp(last,array->strings[0])==0) FATAL(("Parse error in %s",joined));
      last=StringDup(array->strings[0]);
      sec_start=strtol(array->strings[1],NULL,16); /* WARNING! On Linux, use strtoll */
      sec_size=strtol(array->strings[2],NULL,16);
      return NULL;
   }
   else
   {
      new_node = (t_map_node*) Malloc(sizeof(t_map_node));

      new_node->obj=NULL;
      new_node->flags = 0x0;	/* always initialise the flags field to zero! */
      new_node->idx=tel++;
      if (!last) FATAL(("Parse error in %s",joined));
      if (strcmp(last,array->strings[0])!=0) FATAL(("Parse error in %s",joined));
      /* Field 1: The name of the original section */
      new_node->sec_name=StringDup(array->strings[0]);
      new_node->mapped_sec_name=StringDup(last);
      if ((strcmp(new_node->sec_name,".text")==0) || (strcmp(new_node->sec_name,".init")==0) || (strcmp(new_node->sec_name,".fini")==0))
      {
	 new_node->type=Code;
      }
      else if ((strcmp(new_node->sec_name,".data")==0) || (strcmp(new_node->sec_name,".xdata")==0) 
	    || (strcmp(new_node->sec_name,".rdata")==0) ||(strcmp(new_node->sec_name,".rconst")==0) 
	    || (strcmp(new_node->sec_name,".pdata")==0) ||(strcmp(new_node->sec_name,".lita")==0)
	    || (strcmp(new_node->sec_name,".sdata")==0))
      {
	 new_node->type=Data;
      }
      else if ((strcmp(new_node->sec_name,".bss")==0)||(strcmp(new_node->sec_name,".sbss")==0))
      {
	 new_node->type=Zero;
      }
      else FATAL(("Unknown section %s",last));

      /* Field 2: The start of the section */
      tmp =  strtol(array->strings[1], NULL, 16);
      GenericSet_uint64(new_node->base_addr,tmp);
      if (tmp<0x140000000LL)
      {
	 new_node->attr=RO;
      }
      else
      {
	 new_node->attr=RW;
      }
      /* Field 3: The size of the section */
      tmp =  strtol(array->strings[2], NULL, 16);
      GenericSet_uint64(new_node->sec_size,tmp);
      /* Field 4: The object name */
      if ((array->strings[3][0]!='(')&&(PatternMatch("*(*)",array->strings[3])))
      {
	 t_string dup=StringDup(array->strings[3]);
	 t_string lib=strtok(dup,"(");
	 t_string obj=strtok(NULL,")");
	 t_string Name=NULL;
	 VERIFY_FASSERT(obj,("Illegal objectname in map: %s",array->strings[3]));
	 VERIFY_FASSERT(lib,("Illegal objectname in map: %s",array->strings[3]));
	 Name=Collate3(lib,":",obj);
	 Free(dup);
	 new_node->object=Name;
      }
      else
      {
	 new_node->object=StringDup(array->strings[3]);
      }
      return new_node;

   }
   /* exit(0); */ /* Should be impossible ! */
}
#endif
/*}}}*/
/* Node Destructor {{{ */
/*! 
 * \param node The node to free 
 *
 * \param param an unused parameter (neccesary for avlt_walk)  
 * 
 * releases allocated memory, this function is called by t_map_destroy */

void MapNodeFree( void * node)
{
  t_map_node * tmp;

  tmp = (t_map_node*) node;
  Free(tmp->sec_name);
  if (tmp->mapped_sec_name) Free(tmp->mapped_sec_name);
  Free(tmp->object);
  Free(tmp);
}
/*}}}*/
/* Node Compare {{{ */
/*!
 * \brief 
 *
 * \param constsection_a
 * \param  constsection_b
 * \param param
 *
 * \return int 
*/

static int MapNodeCompare(const void * section_a, const void * section_b)
{
  t_map_node * node_a, * node_b;

  node_a = (t_map_node *) section_a;
  node_b = (t_map_node *) section_b;
  if (GENERIC_LT(node_a->base_addr, node_b->base_addr)) return -1;
  else if (GENERIC_GT(node_a->base_addr, node_b->base_addr)) return 1;
  return 0;
}

/*!
 * \brief 
 *
 * \param constto_find
 * \param  constsection
 * \param param
 *
 * \return int 
 * 
 * compare function in order to retrieve a section given an address 
 */

static int MapNodeCompareAddress(const void * to_find, const void * section)
{
  t_address * item_to_find;
  t_map_node * node;

  item_to_find = (t_address *) to_find;
  node         = (t_map_node *) section;

  if (GENERIC_LT(*item_to_find, node->base_addr)) return -1;
  if (GENERIC_GE(*item_to_find, GENERIC_ADD(node->base_addr, node->sec_size))) return 1;
  return 0;
}
/*}}}*/
/* Node Print {{{ */
/*! 
 * \param void_node The node to print 
 * \param param an unused parameter (neccesary for avlt_walk)  
 *
 *
 * prints out a section from the map */

void MapNodePrint(void * void_node)
{
  t_map_node * node=(t_map_node* ) void_node;
  char type[5], attr[3];
  if (node == NULL) return;
  switch (node->type)
  {
    case Code : strcpy(type, "Code"); break;
    case Data : strcpy(type, "Data"); break;
    case Dbug : strcpy(type, "Dbug"); break;
    case Zero : strcpy(type, "Zero"); break;
    case TypeOther : strcpy(type, "Othe"); break;
    default   : WARNING(("Illegal type-field in map-tree"));
  }
  switch (node->attr)
  {
    case RO : strcpy(attr, "RO"); break;
    case RW : strcpy(attr, "RW"); break;
    case OT : strcpy(attr, "OT"); break;
    default : WARNING(("Illegal attr-field in map-tree"));
  }
  DiabloPrint(stdout,"@G @G %s  %s  %d  %s  %s\n",node->base_addr,node->sec_size,
          type,
          attr,
          node->idx,
          node->sec_name,
          node->object );
}
/*}}}*/
/* Node Verify: TODO remove side effect {{{ */

/*!  
 *
 * \param void_node The node to verify 
 * \param param an unused parameter (neccesary for avlt_walk)  
 * 
 * verifies if the information in the node and the loaded objectfile matches */
t_object * param;
static void MapNodeMoveToParent(void * void_node)
{
   t_map_node * node=void_node;
   t_object * obj=param;
   
   if (node->obj)
   {
      t_section * sec;

      sec= SectionGetFromObjectByName(node->obj,node->sec_name);

      if ( (strcmp(node->sec_name,".opd")) && (strcmp(node->sec_name,".IA_64.unwind")) 
#ifdef POWERPC_SUPPORT
	    && (strcmp(node->sec_name,".bss")) && (strcmp(node->sec_name,".sbss")) 
#endif
	    && (strcmp(node->sec_name,".common")) && (strcmp(node->sec_name,".scommon")) &&  (strcmp(node->sec_name,"COMMON")))
      {
	 if (!sec) FATAL(("Section %s:%s  is in map, but not in file!",node->object,node->sec_name));
	 
	 if (!(GENERIC_EQ(SECTION_OLD_SIZE(sec),node->sec_size)))
	 {
	    SECTION_OLD_SIZE(sec) = node->sec_size;
	    WARNING(("Section %s:%s section size differs in map",node->object,node->sec_name));
	 }

	 
	 if (!SECTION_IS_MAPPED(sec))
	 {
	    MAP_SECTION(sec,node->mapped_sec_name,SectionGetFromObjectByName(obj,node->mapped_sec_name));
	    SectionSetOldAddressSuper(sec,node->base_addr); 
	    SectionSetCAddressSuper(sec,node->base_addr); 
	 }
      }
   }
}
/*}}}*/
/* Node Load: TODO docs {{{*/
/*! 
 *  TODO 
 */

t_objects_ll * param2;

static void MapNodeLoad(void * node)
{
   t_objects_ll * objll=param2;
   /* TODO: Check type.... So, we need objecttype.... */

   if (strcmp(((t_map_node *) node)->object, "anon$$obj.o")==0)
   {
   }
   else
   if (strcmp(((t_map_node *) node)->object, "(common)")==0)
   {
   }
   else if (strcmp(((t_map_node *) node)->object ,"(scommon)")==0)
   {
   }
   else if (strcmp(((t_map_node *) node)->object ,"Linker")==0)
   {
   }
   else if (strcmp(((t_map_node *) node)->object ,"exception_gp_info")==0)
   {
   }
   else
   {
      ((t_map_node *) node)->obj=ObjectGet(((t_map_node *) node)->object,objll->parent);
   }
}
/*}}}*/


/* Map Constructor {{{ */
/*! Map Constructor */
t_map * MapNew(void)
{
  return (t_map *) ThreadedAvlTreeNew(0,MapNodeCompare);
}
/*}}} */
/* Map GNU ld constructor: TODO rename  {{{ */

extern FILE * ParseMapGCCin;
void ParseMapGCCBuildMap(t_object * tobj);

/*! TODO: document */
void GCCMapRead(t_object * obj, t_string infile)
{
  t_string filename = NULL;

  filename = Collate(infile,MAP_FILE_EXTENSION); 

  ParseMapGCCin = fopen((const char*) filename, "r");

  if (!ParseMapGCCin) FATAL(("Map file not found. Compile the binary with gcc -Wl,-Map,<output_name>.map -o <output_name> <obj1.o> <obj...>"));
  ParseMapGCCBuildMap(obj); 
  
  fclose(ParseMapGCCin);
  Free(filename);
  VERBOSE(1,("Parsing done"));
}
/*}}}*/
/* Map Alpha linker Constructor: TODO rename {{{ */
#ifdef ALPHA_SUPPORT

/*! 
 * \param obj    the object for which we want to read the map, 
 * \param infile a string containing the name of the object 
 * 
 * his function parses the file `infile.map', and stores it into 
 * obj->map. If no map if found obj->map is NULL, and a warning is printed.
 */

void AlphaMapParse(t_object * obj, t_string infile)
{
  FILE * map_fp;
  t_map * map=obj->map;
  t_string filename, parse_line = NULL;
  t_map_node * parsed_node;

  filename = (t_string) Malloc(strlen(infile)+strlen(MAP_FILE_EXTENSION)+1);
  strcpy((char *) filename, (const char *) infile);
  strcat((char *) filename, (const char *) MAP_FILE_EXTENSION);

  map_fp = fopen((const char*) filename, "r");
  if (map_fp == NULL) {
     errno=0; 
     Free(filename); 
     obj->map=NULL;
     WARNING(("Map not found!"));
     return;
  }
  else
  {
     fseek(map_fp,0,SEEK_END);
     VERIFY_FASSERT(ftell(map_fp),("Map file %s is an empty file! Remove, or add a correct map file")); 
     fseek(map_fp,0,SEEK_SET);
  }

  while((parse_line = GetLine(map_fp)) != NULL) /* while EOF not reached */
  {
    if (PatternMatch("*LINK*EDITOR*MEMORY*MAP*",parse_line)) break;
	Free(parse_line);   /* very important, avoids leaks */
  }

  if (!parse_line) FATAL(("Alpha map file corrupted!"));
  Free(parse_line);
  
  while((parse_line = GetLine(map_fp)) != NULL) /* while EOF not reached */
  {
    if (*parse_line == '\n') 
    {
       Free(parse_line);   /* very important, avoids leaks */
       continue;
    }

    /* Parse one line */
    parsed_node = AlphaMapParseLine(parse_line);
    if (parsed_node)
    {
       if (parsed_node->type!=Dbug)
	  MapInsertNode(map, parsed_node);    /* Insert in the avl-tree */ 
       else
	  MapNodeFree(parsed_node); /* If this is a line, with debugging 
						 info, we delete the node, because we
						 do not need this information */
    }
    Free(parse_line);   /* very important, avoids leaks */
  };

  fclose(map_fp);
  Free(filename);
  VERBOSE(1,("Parsing done"));
}
#endif

/* }}} */
/* Map Destructor {{{ */ 
void MapFree(t_map * map)
{
  ThreadedAvlTreeDestroy((t_threaded_avl_tree *) map, MapNodeFree);
}
/* }}} */
/* Map Print {{{ */
/*! prints out entire map-tree */
void MapPrint(t_map * map)
{
  ThreadedAvlTreeWalk((t_threaded_avl_tree *) map, MapNodePrint);
}
/* }}} */
/* Map Verify {{{ */
/*! TODO: document */
void MapMoveObjectsToParent(t_object * obj, t_map * map)
{
  param=obj;
  ThreadedAvlTreeWalk((t_threaded_avl_tree *) map, MapNodeMoveToParent);
}
/*}}} */
/* Map LoadObjects {{{ */

/*! TODO: document */
t_objects_ll MapLoadObjects(t_map * map, t_object * parent_object)
{
   t_objects_ll pass={NULL, NULL, parent_object};
   param2=&pass;
   ThreadedAvlTreeWalk((t_threaded_avl_tree *) map, MapNodeLoad);
   return pass;
}
/*}}}*/
/* Map InsertNode {{{ */
/*! inserts a node in the map-tree */
void MapInsertNode(t_map * map, t_map_node * node)
{
  ThreadedAvlTreeInsert((t_threaded_avl_tree *) map, (void *) node);
}
/* }}} */
/* Map GetSection {{{ */

/*! returns the node in which the input-adress is contained. Interacts directly
 * on the t_threaded_avl_tree. */

t_map_node * MapGetSection(t_map * map, t_address address)
{
  t_map_node * node = NULL;

  ((t_threaded_avl_tree *)map)->cmp = MapNodeCompareAddress;
  node = (t_map_node*)(ThreadedAvlTreeFind((t_threaded_avl_tree *) map, (const void *) &address));
  ((t_threaded_avl_tree *)map)->cmp = MapNodeCompare;
  return node;
}
/* }}} */
/* vim: set shiftwidth=2 foldmethod=marker: */
