/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//**********************************************************
 * FILE: diablo_print.h
 *
 * DESC: Headers for printing functions 
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) 
 ***********************************************************/


#ifndef DIABLO_PRINT_H
#define DIABLO_PRINT_H
#include "diablo_regset.h"
#include "diablo_object.h"
#include "diablo_bbl.h"
#include "../datastructures/data_string.h"
#include <stdarg.h>

void PrintRegset(FILE * f,const t_architecture_description *desc,t_regset regs);
void BblPrint(t_bbl * bbl, t_uint32 flags) ;
void ProcStatePrint(FILE *f, const t_architecture_description *desc, t_procstate * state);
void ProcStatePrintOpt(FILE *f, const t_architecture_description * desc, t_procstate * state, t_regset regs, t_uint8 options);
void FunctionPrint(t_function * fun, t_uint32 flags) ;
void DiabloPrint(FILE * f, char * desc_string, ...);
void DiabloPrintArch(FILE * f, const t_architecture_description * desc, char * desc_string, ...);
void LoopPrint(t_loop * loop, t_uint32 flags);
t_string vsDiabloPrint(char * desc_string,  va_list ap);
t_string sDiabloPrint(char * desc_string, ...);
void DiabloPrintStdout(char * desc_string, ...);
void TraceBblAtAddress(t_cfg * cfg, t_uint32 address, t_string place);

#define PRINT_PROC_STATE      1
#define PRINT_INSTRUCTIONS    2
#define PRINT_CORRESPONDING   4
#define PRINT_SUCCESSOR       8
#define PRINT_PREDECESSOR    16
#define PRINT_LIVENESS       32
#define PRINT_NEW_ADDRESS    64
#define PRINT_LIVENESS_END  128
#define PRINT_DOMINATES    256
#define PRINT_IDOMINATED_BY 512
#define PRINT_PDOMINATES    1024
#define PRINT_PDOMINATED_BY 2048
#define PRINT_DOM_INFO	(PRINT_DOMINATES|PRINT_IDOMINATED_BY|PRINT_PDOMINATES|PRINT_PDOMINATED_BY)

#define REG_FILTER            1
#define ONLY_VALUES           2

#endif




