/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * All functions to manipulate execution counts and other profiles
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.ugent.be) }}} 
 *****************************************************************************/
/* Includes {{{ */
#include <string.h>
#include "../config.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../datastructures/data_string.h"
#include "../datastructures/data_stack.h"
#include "../datastructures/data_ll.h"
#include "../diablo_options.h"
#include "diablo_flowgraph.h"
#include "diablo_function.h"
#include "diablo_bbl.h"
#include "diablo_edge.h"
#include "diablo_ins.h"
#include "diablo_print.h"
#include "diablo_dominator.h"
#include "diablo_reloc.h"
#include "diablo_iterators.h"
#include "diablo_programexit.h"
#include "diablo_profile.h"
#include <math.h>
/*}}}*/
/* CfgNew {{{*/ 

void CfgEstimateEdgeCounts(t_cfg * cfg)
{
  t_cfg_edge * edge;
  t_bool flag, innerflag;
  t_int32 count;
  t_function * ifun;
  t_bbl * bbl;

  CFG_FOREACH_EDGE(cfg,edge)
    {
      t_bbl * head = CFG_EDGE_HEAD(edge);
      t_bbl * tail = CFG_EDGE_TAIL(edge);
      EDGE_EXEC_COUNT(edge)=-1;

      if (BBL_EXEC_COUNT(head)==0 || BBL_EXEC_COUNT(tail)==0)
	{
	  EDGE_EXEC_COUNT(edge)=0;
	}
      else if (BBL_SUCC_FIRST(head)==BBL_SUCC_LAST(head))
	{
	  EDGE_EXEC_COUNT(edge)=BBL_EXEC_COUNT(head);
	}
      else if (BBL_PRED_FIRST(tail)==BBL_PRED_LAST(tail))
	{
	  EDGE_EXEC_COUNT(edge)=BBL_EXEC_COUNT(tail);
	}
      else if (EDGE_CAT(edge)==ET_FUNLINK || EDGE_CAT(edge)==ET_IPFUNLINK)
	{
	  EDGE_EXEC_COUNT(edge)=0;
	}
    }

  
  do
    {
      flag = FALSE;
      
      CFG_FOREACH_EDGE(cfg,edge)
        {
	  t_bbl *  hd;
	  t_bbl * tl;
	  t_cfg_edge* jedg = NULL;

	  if (EDGE_EXEC_COUNT(edge)>=0) 
	    continue;
	  
	  /*
	    see whether we can compute edge count by
	    looking at other outgoing edges
	  */
               
	  hd = CFG_EDGE_HEAD(edge);
	  count = BBL_EXEC_COUNT(hd);
            
	  BBL_FOREACH_SUCC_EDGE(hd,jedg)
            {
	      if( jedg == edge ) continue;
	      if( EDGE_EXEC_COUNT(jedg)<0)
		break;
	      else
		count -= EDGE_EXEC_COUNT(jedg);
            }
            
	  if( !jedg )
            {
	      if( count < 0 )
		count = 0;
	      EDGE_EXEC_COUNT(edge) = count;
	      flag = TRUE;
	      continue;
            }

	  /*
	    see whether we can compute edge count by
	    looking at other outgoing edges
	  */
            
	  tl = CFG_EDGE_TAIL(edge);
	  count = BBL_EXEC_COUNT(tl);

	  BBL_FOREACH_PRED_EDGE(tl,jedg)
            {
	      if( jedg == edge )
		continue;
	      if( EDGE_EXEC_COUNT(jedg)<0)
		break;
	      else
		count -= EDGE_EXEC_COUNT(jedg);
            }

	  if( !jedg )
            {
	      if( count < 0 )
		count = 0;
	      EDGE_EXEC_COUNT(edge) = count;
	      flag = TRUE;
	      continue;
            }
        }
    }
  while( flag );




  CFG_FOREACH_EDGE(cfg,edge)
    {
      if (EDGE_EXEC_COUNT(edge)>=0)
	{
	  EDGE_EXEC_COUNT_MAX(edge) = EDGE_EXEC_COUNT(edge);
	  EDGE_EXEC_COUNT_MIN(edge) = EDGE_EXEC_COUNT(edge);
	}
      else 
	{
	  t_int32 tl_cnt,hd_cnt;
	  hd_cnt = BBL_EXEC_COUNT(CFG_EDGE_HEAD(edge));
	  if (hd_cnt<0) hd_cnt=0;
	  tl_cnt = BBL_EXEC_COUNT(CFG_EDGE_TAIL(edge));
	  if (tl_cnt<0) tl_cnt=0;
	  EDGE_EXEC_COUNT_MAX(edge) = (hd_cnt<tl_cnt)?hd_cnt:tl_cnt;
	  EDGE_EXEC_COUNT_MIN(edge) = 0;
	}
    }
#if 0
  CFG_FOREACH_BBL(cfg,bbl)
    if (BBL_EXEC_COUNT(bbl)>0 && BblEndWithConditionalBranch(bbl))
      {
	t_cfg_edge * pred;
	t_bbl * bbl_pred1, *bbl_pred2;
	
	t_bbl * succ1 = CFG_EDGE_TAIL(TakenPath(bbl));
	t_bbl * succ2 = CFG_EDGE_TAIL(FallThroughPath(bbl));
	t_int32 count1=0,count2=0;
	t_int32 countbbl_pred1,countbbl_pred2,countsucc1,countsucc2,min11,min12,min21,min22;
	if (EDGE_EXEC_COUNT(TakenPath(bbl))>0) continue;
	if (succ1 == bbl || succ2 == bbl) continue;
	  
	BBL_FOREACH_PRED_EDGE(succ1,pred)
	  {
	    if (EDGE_EXEC_COUNT(pred)>=0) continue;
	    if (CFG_EDGE_HEAD(pred)==bbl) continue;
	    bbl_pred1 = CFG_EDGE_HEAD(pred);
	    count1++;
	  }
	  
	BBL_FOREACH_PRED_EDGE(succ2,pred)
	  {
	    if (EDGE_EXEC_COUNT(pred)>=0) continue;
	    if (CFG_EDGE_HEAD(pred)==bbl) continue;
	    bbl_pred2 = CFG_EDGE_HEAD(pred);
	    count2++;
	  }
	  
	if (count1 != 1 || count2 != 1) continue;
	  
	if (bbl_pred1 != bbl_pred2) continue;
	if (!BblEndWithConditionalBranch(bbl_pred2)) continue;
	  
	bbl_pred1 = bbl;

	countbbl_pred1=BBL_EXEC_COUNT(bbl_pred1);
	countbbl_pred2=BBL_EXEC_COUNT(bbl_pred2);
	countsucc1=BBL_EXEC_COUNT(succ1);
	countsucc2=BBL_EXEC_COUNT(succ2);

	BBL_FOREACH_PRED_EDGE(succ1,pred)
	  {
	    if (EDGE_EXEC_COUNT(pred)<0) continue;
	    countsucc1-=EDGE_EXEC_COUNT(pred);
	  }

	BBL_FOREACH_PRED_EDGE(succ2,pred)
	  {
	    if (EDGE_EXEC_COUNT(pred)<0) continue;
	    countsucc2-=EDGE_EXEC_COUNT(pred);
	  }
	  
	min11 = countbbl_pred1<countsucc1?countbbl_pred1:countsucc1;
	min12 = countbbl_pred1<countsucc2?countbbl_pred1:countsucc2;
	min21 = countbbl_pred2<countsucc1?countbbl_pred2:countsucc1;
	min22 = countbbl_pred2<countsucc1?countbbl_pred2:countsucc2;

	EDGE_EXEC_COUNT_MAX(TakenPath(bbl_pred1))   = (countbbl_pred1 - min12 + min11 + countsucc1 - min21 + min11)/4;
	EDGE_EXEC_COUNT_MIN(TakenPath(bbl_pred1))   = (countbbl_pred1 - min12 + min11 + countsucc1 - min21 + min11)/4;
	EDGE_EXEC_COUNT(TakenPath(bbl_pred1)) = (countbbl_pred1 - min12 + min11 + countsucc1 - min21 + min11)/4;
      }
#endif
  CFG_FOREACH_FUN(cfg,ifun)
    {
      do
	{
	  flag = FALSE;
	  
	  FUN_FOREACH_BBL(ifun,bbl)
	    {
	      t_int32 min,max;
	      t_int32 tmp_min,tmp_max;
		

	      do
		{
		  innerflag = FALSE;
		    
		  min = max = BBL_EXEC_COUNT(bbl);
		    
		  BBL_FOREACH_SUCC_EDGE(bbl,edge)
		    {
		      max -= EDGE_EXEC_COUNT_MAX(edge);
		      min -= EDGE_EXEC_COUNT_MIN(edge);
		    }
		    
		  BBL_FOREACH_SUCC_EDGE(bbl,edge)
		    {
		      if (CfgEdgeIsInterproc(edge)) continue;
		      if (EDGE_EXEC_COUNT(edge)>=0) continue;
			
		      tmp_min = max + EDGE_EXEC_COUNT_MAX(edge);
		      tmp_max = min + EDGE_EXEC_COUNT_MIN(edge);
			
		      if (tmp_min>EDGE_EXEC_COUNT_MIN(edge) && tmp_min<EDGE_EXEC_COUNT_MAX(edge))
			{
			  EDGE_EXEC_COUNT_MIN(edge)=tmp_min;
			  innerflag = flag = TRUE;
			}
		      if (tmp_max<EDGE_EXEC_COUNT_MAX(edge) && tmp_max>EDGE_EXEC_COUNT_MIN(edge))
			{
			  EDGE_EXEC_COUNT_MAX(edge)=tmp_max;
			  innerflag = flag = TRUE;
			}
		    }
		}
	      while (innerflag);
		
	      do
		{
		  innerflag = FALSE;
		    
		  min = max = BBL_EXEC_COUNT(bbl);
		    
		  BBL_FOREACH_PRED_EDGE(bbl,edge)
		    {
		      max -= EDGE_EXEC_COUNT_MAX(edge);
		      min -= EDGE_EXEC_COUNT_MIN(edge);
		    }
		    
		  BBL_FOREACH_PRED_EDGE(bbl,edge)
		    {
		      if (CfgEdgeIsInterproc(edge)) continue;
		      if (EDGE_EXEC_COUNT(edge)>=0) continue;
		      tmp_min = max + EDGE_EXEC_COUNT_MAX(edge);
		      tmp_max = min + EDGE_EXEC_COUNT_MIN(edge);
		      if (tmp_min>EDGE_EXEC_COUNT_MIN(edge) && tmp_min<EDGE_EXEC_COUNT_MAX(edge))
			{
			  EDGE_EXEC_COUNT_MIN(edge)=tmp_min;
			  innerflag = flag = TRUE;
			}
		      if (tmp_max<EDGE_EXEC_COUNT_MAX(edge) && tmp_max>EDGE_EXEC_COUNT_MIN(edge))
			{
			  EDGE_EXEC_COUNT_MAX(edge)=tmp_max;
			  innerflag = flag = TRUE;
			}
		    }
		}
	      while (innerflag);
	    }
	}
      while (flag);
	
      FUN_FOREACH_BBL(ifun,bbl)
	{
	  BBL_FOREACH_SUCC_EDGE(bbl,edge)
	    {
	      if (CfgEdgeIsInterproc(edge)) continue;
	      if (EDGE_EXEC_COUNT(edge)>=0) continue;
	      EDGE_EXEC_COUNT(edge)=(EDGE_EXEC_COUNT_MAX(edge)+EDGE_EXEC_COUNT_MIN(edge))/2;
	    }
	}
    }

}

void CfgReadInsExecutionCounts(t_cfg * cfg, t_string name)
{
  FILE* fp = fopen(name,"r");

  t_address address;
  t_int32 count;
  t_uint32 value;
  t_bbl * bbl;
  t_ins * ins;

  CFG_FOREACH_BBL(cfg,bbl)
    {
      BBL_FOREACH_INS(bbl,ins)
	INS_EXEC_COUNT(ins)=BBL_EXEC_COUNT(bbl);
    }
  
  if (fp)
    {
      while(!feof(fp)) 
	{
	  if (fscanf(fp,"%x %d\n",&value,&count) && count>0)
	    {
#ifdef BIT64ADDRSUPPORT
	      GenericSet_uint64(address,value);
#else
	      GenericSet_uint32(address,value);
#endif
	      CFG_FOREACH_BBL(cfg,bbl)
		{
		  t_bool exit = FALSE;
		  BBL_FOREACH_INS(bbl,ins)
		    {
		      if (GENERIC_EQ(address,INS_OLD_ADDRESS(ins)))
			{	
			  INS_EXEC_COUNT(ins) = count;
			  exit=TRUE;
			  break;
			}
		    }
		  if (exit) break;
		}
	    }
	}
    }
  else
    {
      FATAL(("oeps, no execution count file with name %s",name));
    }

}

void CfgReadBlockExecutionCounts(t_cfg * cfg, t_string name)
{
  FILE* fp = fopen(name,"r");

  t_address address;
  t_int32 count;
  t_uint32 value;
  t_bbl * bbl;
   
  CFG_FOREACH_BBL(cfg,bbl)
    bbl->execution_count = -1;

  if (fp)
    {
      while(!feof(fp)) 
	{
	  if (fscanf(fp,"%x %d\n",&value,&count) && count>0)
	    {
#ifdef BIT64ADDRSUPPORT
	      GenericSet_uint64(address,value);
#else
	      GenericSet_uint32(address,value);
#endif
	      CFG_FOREACH_BBL(cfg,bbl)
		{
		  if (GENERIC_EQ(address,BBL_OLD_ADDRESS(bbl)))
		    {	
		      bbl->execution_count = count;
		      break;
		    }
		}
	    }
	}

      CFG_FOREACH_BBL(cfg,bbl)
	{
	  if (IsHellNode(bbl)) continue;
	  if (!BBL_FUN(bbl)) continue;
	  if (BBL_IS_LAST(bbl)) continue;
	  if (BBL_EXEC_COUNT(bbl)>=0) continue;
	  BBL_EXEC_COUNT(bbl)=0;
	}
	
	
      CfgEstimateEdgeCounts(cfg);

    }
  else
    {
      FATAL(("oeps, no execution count file with name %s",name));
    }

}

static int cmp_count(t_bbl ** a, t_bbl ** b)
{
  t_bbl * bbla = *a;
  t_bbl * bblb = *b;

  if (BBL_EXEC_COUNT(bbla)<BBL_EXEC_COUNT(bblb))
    return 1;
  if (BBL_EXEC_COUNT(bbla)>BBL_EXEC_COUNT(bblb))
    return -1;
  return 0;

}

t_int32 CfgComputeWeight(t_cfg * cfg)
{
  t_bbl * bbl;
  t_int32 total_weight = 0;

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BBL_EXEC_COUNT(bbl)>0)
	{
	  total_weight += BBL_EXEC_COUNT(bbl)*BBL_NINS(bbl);
	}
    }
  return total_weight;

}

void CfgComputeHotBblThreshold(t_cfg * cfg, double weight_threshold)
{
  t_bbl * bbl;
  t_int32 total_count = 0;
  t_int32 total_weight = 0;
  t_int32 hot_weight = 0;
  t_int32 n_bbls = 0;
  t_bbl ** array;
  int i;

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BBL_EXEC_COUNT(bbl)>0)
	{
	  total_count += BBL_EXEC_COUNT(bbl);
	  total_weight += BBL_EXEC_COUNT(bbl)*BBL_NINS(bbl);
	  n_bbls++;
	}
    }

  array = Malloc(sizeof(t_bbl *)*n_bbls);

  n_bbls = 0;

  CFG_FOREACH_BBL(cfg,bbl)
    {
      if (BBL_EXEC_COUNT(bbl)>0)
	{
	  array[n_bbls++]=bbl;
	}
    }

  qsort(array, n_bbls, sizeof(t_bbl*), (int (*) (const void *,const void *)) cmp_count);

  for (i=0;i<n_bbls;i++)
    {
      bbl = array[i];
      hot_weight += BBL_EXEC_COUNT(bbl)*BBL_NINS(bbl);
      if (((double)hot_weight)>((double)total_weight)*weight_threshold)
	{
	  CFG_HOT_THRESHOLD_COUNT(cfg)=BBL_EXEC_COUNT(bbl);
	  break;
	}
    }
  /*
  for (i=0;i<n_bbls;i++)
    {
      bbl = array[i];
      if (BBL_EXEC_COUNT(bbl)>=CFG_HOT_THRESHOLD_COUNT(cfg))
	DiabloPrint(stdout,"HOT @B\n",array[i]);
    }

    printf("THRESHOLD: %d\n",CFG_HOT_THRESHOLD_COUNT(cfg));
  */
  Free(array);
}

t_bool FunIsHot(t_function * fun)
{
  t_cfg * cfg = fun->cfg;
  t_int32 threshold = CFG_HOT_THRESHOLD_COUNT(cfg);
  t_bbl * bbl;
  if (!global_options.blockprofilefile)
    return FALSE;
  FUN_FOREACH_BBL(fun,bbl)
    {
      if (BBL_EXEC_COUNT(bbl)>= threshold)
	return TRUE;
    }
  return FALSE;
}

t_bool BblIsHot(t_bbl * bbl)
{
  t_cfg * cfg;
  t_int32 threshold;

  if (!global_options.blockprofilefile)
    return FALSE;

  if (!BBL_FUN(bbl))
    return FALSE;
  cfg = bbl->fun->cfg;
  threshold = CFG_HOT_THRESHOLD_COUNT(cfg);
  if (BBL_EXEC_COUNT(bbl) >= threshold)
    return TRUE;
  return FALSE;
}

t_bool BblIsAlmostHot(t_bbl * bbl)
{
  t_cfg * cfg;
  t_int32 threshold;
  if (!global_options.blockprofilefile)
    return FALSE;
  if (!BBL_FUN(bbl))
    return FALSE;
  cfg = bbl->fun->cfg;
  threshold = CFG_HOT_THRESHOLD_COUNT(cfg);
  if (BBL_EXEC_COUNT(bbl) >= threshold/2)
    return TRUE;
  return FALSE;
}

t_bool EdgeIsHot(t_cfg_edge * edge)
{
  t_cfg * cfg = CFG_EDGE_TAIL(edge)->fun->cfg;
  t_int32 threshold = CFG_HOT_THRESHOLD_COUNT(cfg);
  if (!global_options.blockprofilefile)
    return FALSE;
  if (!BblIsHot(CFG_EDGE_HEAD(edge)) && !BblIsHot(CFG_EDGE_TAIL(edge)))
    return FALSE;
  if ((double)EDGE_EXEC_COUNT(edge) >= ((double)threshold)*0.90)
    return TRUE;
  return FALSE;
}

#if 1
double ORDER(t_int32 arg) 
{
  if (arg<=0) 
    {
      return 0.0;
    }
  return log10(((double)arg));
}
#else
double ORDER(t_int32 arg)
{
  double in = (double) arg;
  double out = 0;
  while (in>0)
    {
      out+=1;
      in/=10;
    }
  return out;
}
#endif
