/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

/* Includes {{{ */
#include "diablo_function.h"
#include "diablo_bbl.h"
#include "diablo_programexit.h"
#include "diablo_iterators.h"
#include "../datastructures/data_ll.h"
#include "../utils/util_status.h"
/* }}} */


void CfgCreateExitBlockList(t_cfg * cfg)
{
  t_function * fun;
  t_bbl * bbl;
  t_ins * ins;
  t_cfg_exits * exit_block;
  t_tristate is_exit;

  ANALYSIS_START("Detecting program exits");
  CFG_FOREACH_FUN(cfg, fun)
    FUN_FOREACH_BBL(fun,bbl)
      BBL_FOREACH_INS(bbl,ins)
      {
	if(INS_TYPE(ins) == IT_SWI)
	{
	  is_exit = cfg->sec->obj->actions->IsSyscallExit(ins);
	  if (is_exit != NO)
	  {
	    VERBOSE(1,("Adding @B as a exit block [%s]! @I",bbl,is_exit == YES ? "certain" : "perhaps", ins));
	    exit_block = (t_cfg_exits*) Malloc(sizeof(t_cfg_exits));
	    exit_block->exit_bbl = bbl;
	    exit_block->exit_ins = ins;
	    exit_block->certain = (is_exit == YES);
	    exit_block->next=cfg->exitblocks;
	    exit_block->tmp=NULL;
	    cfg->exitblocks=exit_block;
	    break;
	  }
	}
      }
  ANALYSIS_SIMPLE_END("Detecting program exits");
}

void CfgFreeExitBlockList(t_cfg * cfg)
{
  t_cfg_exits * to_free; 
  while(cfg->exitblocks)
  {
    to_free = cfg->exitblocks;
    cfg->exitblocks = cfg->exitblocks->next;
    Free(to_free);
  }
}
/* vim: set shiftwidth=2 foldmethod=marker: */
