/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//**********************************************************
 * FILE: diablo_gcc_maps.l
 *
 * DESC:  This file contains the definitions needed to build 
 * the global symbol table and handle all kinds of symbols
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) 
 ***********************************************************/

/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup SYMBOLS Symbols */
/*@{ */
#ifndef MY_SYMBOL_TYPEDEFS
#define MY_SYMBOL_TYPEDEFS
/*! Symbols in diablo, stored in t_symboltable */
typedef struct _t_symbol t_symbol;
/*! Symboltable */
typedef struct _t_symboltable t_symboltable;
/*! Symboltable Hash Elements */
typedef struct _t_symbol_he t_symbol_he;

#endif /*  MY_SYMBOL_TYPEDEFS */


#ifndef MY_SYMBOL_H
#define MY_SYMBOL_H
#include "diablo_reloc.h"
#include "../datastructures/data_hash.h"
#include "diablo_object.h"
/* Types */
#define FUNCTION_SYMBOL 1
#define GLOBAL_SYMBOL 2
#define UNDEFINED_SYMBOL 3
#define LOCAL_SYMBOL 4
#define CODE_IDENT_SYMBOL 5
#define DATA_IDENT_SYMBOL 6
#define ORIGINAL_SECTION_SYMBOL 7
#define ABS_SYMBOL 8
#define WEAK_UNDEFINED_SYMBOL 9
#define DIRECT_BRANCH_SYMBOL 10
#define WEAK_SYMBOL 11 
#define END_SYMBOL 13 
#define SECTION_END_SYMBOL 14
#define REMOVED_SYMBOL 15 
#define PLT_SYMBOL 16 

/* to types */
#define TO_SECTION 1 
#define TO_SUB_SECTION 2
#define TO_COMMON 3
#define TO_UNDEFINED 4
#define TO_NONE 5
#define TO_INSTRUCTION 6
#define TO_BBL 7
#define TO_REMOVED 8


#define ADDRESS_IS_CODE 1
#define ADDRESS_IS_DATA 2
#define ADDRESS_IS_SMALL_CODE 3

#define T_SYMBOL(x)   ((t_symbol *) x)
/*! A symbol */
struct _t_symbol
{
   t_uint32 to_type;
   t_uint32 type;
   t_uint32 relocrefcount;
   t_relocatable * base;
   t_address offset_from_start;
   t_address addend;
   t_address size;	/* size of a common symbol */
   t_address current_address;
   t_string name;
   t_uint32 prolog;
   t_bool local;
   struct _t_symbol * prev;
   struct _t_symbol * next;
   struct _t_symbol * mapped;
};

/*! The hash element stored in the Symbol(Hash)Table */
struct _t_symbol_he
{
  data_hash_node node;
  t_symbol * sym;
};

/*! A symbol table */
struct _t_symboltable
{
   t_object * obj;
   t_symbol * first; 
   t_symbol * last;
   t_uint32 nsyms;
   data_hash_table * hash_table; 
}; 



t_symboltable * SymbolTableNew(t_object *);

/*! \defgroup SYMBOL_ALLOC Constructors and destructors */
/*@{*/

t_symbol * SymbolTableAddSymbolToSection(t_symboltable * symbol_table,  t_uint32 type,  t_relocatable * sec,  t_address offset, t_string name, t_bool local);
t_symbol * SymbolTableAddSymbolToCommon(t_symboltable * symbol_table,  t_uint32 type, t_address align, t_address size, t_string name, t_bool local);
t_symbol * SymbolTableAddSymbolToSubSection(t_symboltable * symbol_table,  t_uint32 type,  t_relocatable * sec,  t_address offset, t_string name,t_bool local);
t_symbol * SymbolTableGetSymbolByName(t_symboltable * st, t_string name);
t_symbol * SymbolTableAddWeakUndefinedSymbol(t_symboltable * symbol_table,   t_string name);
t_symbol * SymbolTableAddAbsoluteSymbol(t_symboltable * symbol_table,  t_address offset, t_string name,t_bool local);
t_symbol *SymbolTableAddSymbol(t_symboltable * symbol_table, t_uint32 to_type, t_uint32 type,  t_relocatable * sec,  t_address offset, t_string name,t_bool local);
t_symbol * SymbolTableAddPltSymbol(t_symboltable * symbol_table,   t_string name, t_address pltaddress, t_address size);
/*@}*/



void SymbolTableFree(t_symboltable * st);
t_symbol * SymbolTableAddUndefinedSymbol(t_symboltable * symbol_table,   t_string name);
t_symbol * SymbolTableGetFirstSymbolWithName(t_symboltable * st, t_string name);
t_uint32 SymbolTableGetDataType(t_object * obj, t_address adr);
t_uint32 SymbolTableGetCodeType(t_object * obj, t_address addr);
t_symbol * SymbolTableGetSymbolByAddress(t_symboltable * st, t_address addr);
void SymbolPrintTable(t_symboltable * st);
void SymbolPrint(t_symbol * sym);
#endif
/*@}*/
/*@}*/
