/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be)
 ***********************************************************/

#ifndef _UD_CHAIN_H_
#define _UD_CHAIN_H_

#include <stdio.h>
#include "diablo_ins.h"
#include "diablo_regset.h"
#include "../datastructures/data_ll.h"

/*! A use-def chain */
struct _t_ud_chain {
  ll node;
  t_bbl * def;
  t_uint32 dups;
};

/*! An array of ud_chains */
struct _t_ud_chain_list {
  t_uint32 num_regs;
  t_ud_chain ** reg;
};

#ifndef UD_CHAIN_TYPEDEFS
#define UD_CHAIN_TYPEDEFS
typedef struct _t_ud_chain t_ud_chain;
typedef struct _t_ud_chain_list t_ud_chain_list;
#endif

t_ud_chain_list * UdChainListNew(t_uint32 regcount);
void UdChainListFree(t_ud_chain_list * udcl);
t_ud_chain_list * UdChainListClone(t_ud_chain_list * udcl);
void UdChainListDup(t_ud_chain_list * dest, t_ud_chain_list * src);
t_bool UdChainListJoin(t_ud_chain_list * dest, t_ud_chain_list * src, t_regset regs);
t_regset UdChainListGetTopRegs(t_ud_chain_list * udcl);
void UdChainListPrint(FILE * out, t_ud_chain_list * udcl);

t_ud_chain * UdChainNew(t_bbl * def);
void UdChainFree(t_ud_chain * udc);
void UdChainAppend(t_ud_chain ** head, t_ud_chain * insert);
t_ud_chain * UdChainDup(t_ud_chain * udc);
t_bool UdChainJoin(t_ud_chain ** dest, t_ud_chain * src);
void UdChainPrint(FILE * out, t_ud_chain * udcl);

#endif
