/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)
 ***********************************************************/


#ifndef MY_ACTION_TYPEDEFS
#define MY_ACTION_TYPEDEFS
typedef struct  _t_action_struct t_action_struct;
typedef struct  _t_fileformat_action_struct t_fileformat_action_struct;

/*! The things you can do with an object (all analyses and optimizations should be listed here. 
 * This is an extra step to check if the object is in the right state to perform a certain transformation) */ 
typedef enum _t_action
{
/*  0 */ GROUP_INFO                =  0,  /* NO ARGS */
/*  1 */ READ_MAP                  =  1,  /* ARG OBJECT, FILENAME (can be NULL) */
/*  2 */ DISASSEMBLE               =  2,  /* ARG OBJECT */
/*  3 */ DISASSEMBLE_SECTION       =  3,  /* ARG OBJECT, SECTION */
/*  4 */ FLOWGRAPH                 =  4,  /* ARG OBJECT */
/*  5 */ FLOWGRAPH_SECTION         =  5,  /* ARG OBJECT, SECTION */ 
/*  6 */ MAKE_ADDRESS_PRODUCERS    =  6,  /* ARG OBJECT, SECTION */
/*  7 */ DEFLOWGRAPH               =  7,  /* ARG OBJECT */
/*  8 */ DEFLOWGRAPH_SECTION       =  8,  /* ARG OBJECT, SECTION */ 
/*  9 */ ASSEMBLE                  =  9,  /* ARG OBJECT */
/* 10 */ ASSEMBLE_SECTION          = 10,  /* ARG OBJECT, SECTION */
/* 11 */ WRITE                     = 11,  /* ARG OBJECT, name */
/* 12 */ PRINT_INSTRUCTION         = 12,  /* ARG INS, t_string return */
/* 13 */ NOOP_INSTRUCTION          = 13,  /* ARG INS, t_string return */
/* 14 */ REGNAME                   = 14,  /* ARG REG */ 
/* 15 */ INS_EMUL                  = 15,
/* 16 */ EDGE_PROP                 = 16,
/* 17 */ INS_UNCONDITIONALIZE      = 17,
/* 18 */ INS_CONSTANT_OPTIMIZE     = 18,
/* 19 */ CONVERT_DATA_TO_BLOCKS    = 19,  /* ARG OBJECT */
/* 20 */ CONVERT_SECTION_TO_BLOCKS = 20   /* ARG OBJECT ,SECTION */
} t_action;

#endif /* MY_OBJECT_TYPEDEF */


#ifndef MY_ACTION_H
#define MY_ACTION_H
#include "../datastructures/data_string.h"
#include "diablo_section.h"
#include "diablo_ins.h"
#include "diablo_procstate.h"
#include "diablo_layout_script.h"
#include "diablo_code_layout.h"

/*! A struct describing the action that can be taken on an object */

struct  _t_action_struct
{
/*  0 */  void     (*GroupInfo)(void);
/*  1 */  void     (*ReadMap)(t_object *, t_string);
/*  2 */  void     (*Disassemble)(t_object *);
/*  3 */  void     (*DisassembleSec)(t_object *,t_section *);
/*  4 */  void     (*Flowgraph)(t_object *);
/*  5 */  void     (*FlowgraphSec)(t_object *,t_section *);
/*  6 */  void     (*MakeAddressProducers)(t_section *);
/*  7 */  void     (*DeFlowgraph)(t_object *);
/*  8 */  void     (*DeFlowgraphSec)(t_object *,t_section *);
/*  9 */  void     (*Assemble)(t_object *);
/* 10 */  void     (*AssembleSec)(t_object *,t_section *);
/* 11 */  void     (*Write)(t_object *, t_string);
/* 12 */  void     (*Print)(t_ins *, t_string);
/* 13 */  void     (*MakeNoop)(t_ins *);
/* 14 */  t_string (*RegName)(t_reg);
/* 15 */  void     (*InstructionEmulator)(t_ins *, t_procstate *);

/*! edge_propagator holds the procedure that sets the propagation procedures
 * for all edges, corresponding to the condition under which the branches can
 * be "executed" */

/* 16 */  void     (*EdgePropagator)(t_cfg_edge *,t_ins *);

/*! unconditionalizer holds a pointer to the procedure that turns conditional
 * instructions into unconditional ones */

/* 17 */  t_bool   (*InstructionUnconditionalizer)(t_ins *);

/*! ins_constant_optimizer points to the procedure that can actually optimize
 * instructions bases in the proc_state before the instruction is executed.*/

/* 18 */  t_bool   (*InstructionConstantOptimizer)(t_ins *,t_procstate *,t_procstate *);
/* 19 */  void     (*ObjectConvertToDataBlocks)(t_object *);
/* 20 */  void     (*ConvertToBlocksSec)(t_object *, t_section *);
/* 21 */  t_bool   (*CompareInstructions)(t_ins *, t_ins *);
/* 22 */  t_bool   (*IsCopy)(t_ins *, t_reg *, t_reg *);
/* 23 */  t_bool   (*InsHasSideEffect)(t_ins *);


/* 24 */  t_uint32 (*BblCmpExtReg)(t_bbl *,t_bbl *,t_uint32 *);
/* 25 */  t_uint32 (*BblAbstract)(t_bbl *,t_bbl *,t_uint32 *,t_uint32 *);
/* 26 */  void     (*FunComputeStackSavedRegisters)(t_function*);
/* 27 */  void     (*CopyInstructionPropagator)(t_ins *, t_equations);
/* 28 */  t_tristate (*IsSyscallExit)(t_ins *);
/* 29 */  t_bool   (*IsControlflow)(t_ins *);
/* 30 */  t_bool (*InsIsIndirectCall)(t_ins *);
/* 31 */  t_bool (*FunIsGlobal)(t_function *);
/* 32 */  void (*InsCleanup)(t_ins *);
/* 33 */  void (*InsDupDynamic)(t_ins *, t_ins *);
#ifdef STEGO
/* 34 */  void (*InsFindAlternatives)(t_ins * ins, t_regset dead);
/* 35 */  t_int8 (*InsCompare)(t_ins * ins1, t_ins * ins2, t_uint8 mode);
/* 36 */  void (*MorphInstructionTo)(t_ins * from, t_ins * to);
/* 37 */  void (*MarkLeadersThatCanBeGuessed)(t_chain_holder * ch);
#endif
};

struct _t_fileformat_action_struct {
  t_uint64 (*GetSizeofHeaders)(t_object * obj, t_layout_rule * layoutscript);
};

void ObjectAction(t_action action,...);
void ActionPrint(int start_char, const t_action_struct * action);
#endif
