/******************************************************************************
 * Copyright 2001,2002,2003: {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

/* Includes {{{ */
#include "../simple_types.h"
#include "diablo_addresses.h"
#include "../utils/util_error.h"
/* }}} */

t_address AlignAddress(t_address align, t_address start)
{
  t_uint32 alignment = GENERIC_EXTRACT_UINT32(align);
  t_uint32 diff;
  
  if (alignment)
    if ((diff = (GENERIC_EXTRACT_UINT32(start) % alignment)))
      return GENERIC_ADD_UINT32(start, alignment-diff);

  return start;
}

#ifdef GENERICADDRSUPPORT
t_address INVALID_GENERIC={ INVALID, {LL(-1)}};
t_address NULL_GENERIC={ EITHER, {0}};

t_address real_GENERIC_ADD(t_string file, t_uint32 line, t_address a, t_address b)
{
   t_address ret;
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid LHS in ADD at line %d in %s",line,file));
   VERIFY_FASSERT((b.type!=INVALID), ("Invalid RHS in ADD at line %d in %s",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses at line %d in %s %d %d",line,file, a.type, b.type));

   if (a.type!=EITHER) ret.type=a.type; 
   else ret.type=b.type;
   if ((a.type==BIT32)||(b.type==BIT32))
      ret.types.addr32=a.types.addr32+b.types.addr32;
   else ret.types.addr64=a.types.addr64+b.types.addr64;
   return ret;
}

t_address real_GENERIC_AND(t_string file, t_uint32 line, t_address a, t_address b)
{
   t_address ret;
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid LHS in AND at line %d in %s",line,file));
   VERIFY_FASSERT((b.type!=INVALID), ("Invalid RHS in AND at line %d in %s",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses at line %d in %s",line,file));

   if (a.type!=EITHER) ret.type=a.type; 
   else ret.type=b.type;
   if ((a.type==BIT32)||(b.type==BIT32))
      ret.types.addr32=a.types.addr32&b.types.addr32;
   else ret.types.addr64=a.types.addr64&b.types.addr64;
   return ret;
}

t_address real_GENERIC_OR(t_string file, t_uint32 line, t_address a, t_address b)
{
   t_address ret;
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid LHS in OR at line %d in %s",line,file));
   VERIFY_FASSERT((b.type!=INVALID), ("Invalid RHS in OR at line %d in %s",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses at line %d in %s",line,file));

   if (a.type!=EITHER) ret.type=a.type; 
   else ret.type=b.type;
   if ((a.type==BIT32)||(b.type==BIT32))
      ret.types.addr32=a.types.addr32|b.types.addr32;
   else ret.types.addr64=a.types.addr64|b.types.addr64;
   return ret;
}

t_address real_GENERIC_XOR(t_string file, t_uint32 line, t_address a, t_address b)
{
   t_address ret;
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid LHS in XOR at line %d in %s",line,file));
   VERIFY_FASSERT((b.type!=INVALID), ("Invalid RHS in XOR at line %d in %s",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses at line %d in %s",line,file));

   if (a.type!=EITHER) ret.type=a.type; 
   else ret.type=b.type;
   if ((a.type==BIT32)||(b.type==BIT32))
      ret.types.addr32=a.types.addr32^b.types.addr32;
   else ret.types.addr64=a.types.addr64^b.types.addr64;
   return ret;
}
t_address real_GENERIC_NOT(t_string file, t_uint32 line, t_address a)
{
   t_address ret;
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid LHS in NOT at line %d in %s",line,file));

   ret.type=a.type; 
   if (a.type==BIT32)
      ret.types.addr32=~a.types.addr32;
   else ret.types.addr64=~a.types.addr64;
   return ret;
}

t_address real_GENERIC_SHL(t_string file, t_uint32 line, t_address a, t_address b)
{
   t_address ret;
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid LHS in AND at line %d in %s",line,file));
   VERIFY_FASSERT((b.type!=INVALID), ("Invalid RHS in AND at line %d in %s",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses at line %d in %s",line,file));

   if (a.type!=EITHER) ret.type=a.type; 
   else ret.type=b.type;
   if ((a.type==BIT32)||(b.type==BIT32))
      ret.types.addr32=a.types.addr32 << b.types.addr32;
   else ret.types.addr64=a.types.addr64 << b.types.addr64;
   return ret;
}

t_address real_GENERIC_SHR(t_string file, t_uint32 line, t_address a, t_address b)
{
   t_address ret;
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid LHS in AND at line %d in %s",line,file));
   VERIFY_FASSERT((b.type!=INVALID), ("Invalid RHS in AND at line %d in %s",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses at line %d in %s",line,file));

   if (a.type!=EITHER) ret.type=a.type; 
   else ret.type=b.type;
   if ((a.type==BIT32)||(b.type==BIT32))
      ret.types.addr32=a.types.addr32 >> b.types.addr32;
   else ret.types.addr64=a.types.addr64 >> b.types.addr64;
   return ret;
}

t_address real_GENERIC_ADD_INT32(t_string file, t_uint32 line, t_address a, t_int32 b)
{
   t_address ret;
   if (a.type!=EITHER)
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid in GENERIC_ADD_UINT32 at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32+b;
   else ret.types.addr64=a.types.addr64+b;
   return ret;
}

t_address real_GENERIC_ADD_UINT64(t_string file, t_uint32 line, t_address a, t_uint64 b)
{
   t_address ret;
   if (a.type!=EITHER)
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid in GENERIC_ADD_UINT64 at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32+b;
   else ret.types.addr64=a.types.addr64+b;
   return ret;
}

t_address real_GENERIC_ADD_UINT32(t_string file, t_uint32 line, t_address a, t_uint32 b)
{
   t_address ret;
   if (a.type!=EITHER)
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid in GENERIC_ADD_UINT32 at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32+b;
   else ret.types.addr64=a.types.addr64+b;
   return ret;
}

t_address real_GENERIC_NAND_UINT32(t_string file, t_uint32 line, t_address a, t_uint32 b)
{
   t_address ret;
   if (a.type!=EITHER)
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid in GENERIC_NAND_UINT32 at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32&(~((t_uint32) (b)));
   else ret.types.addr64=a.types.addr64&(~((t_uint64) (b)));
   return ret;
}

t_address real_GENERIC_ADD_INT64(t_string file, t_uint32 line, t_address a, t_int64 b)
{
   t_address ret;
   if (a.type!=EITHER)
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid in GENERIC_ADD_UINT32 at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32+b;
   else ret.types.addr64=a.types.addr64+b;
   return ret;
}

t_address real_GENERIC_SUB_UINT32(t_string file, t_uint32 line, t_address a, t_uint32 b)
{
   t_address ret;
   if (a.type!=EITHER)
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid in GENERIC_SUB_UINT32 at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32-b;
   else ret.types.addr64=a.types.addr64-b;
   return ret;
}

t_address real_GENERIC_SUB_INT32(t_string file, t_uint32 line, t_address a, t_int32 b)
{
   t_address ret;
   if (a.type!=EITHER)
   VERIFY_FASSERT((a.type!=INVALID), ("Invalid in GENERIC_SUB_INT32 at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32-b;
   else ret.types.addr64=a.types.addr64-b;
   return ret;
}


t_address real_GENERIC_SUB(t_string file, t_uint32 line,t_address a, t_address b)
{
   t_address ret;
   VERIFY_FASSERT(a.type!=INVALID, ("Invalid LHS in generic SUB called at line %d in %s",line,file));
   VERIFY_FASSERT(b.type!=INVALID, ("Invalid RHS in generic SUB called at line %d in %s",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Sub'ed different size addresses at line %d in %s",line,file));
  
   ret.type=a.type;
   if (a.type==BIT32)
      ret.types.addr32=a.types.addr32-b.types.addr32;
   else ret.types.addr64=a.types.addr64-b.types.addr64;
   return ret;
}


t_bool real_GENERIC_GT(t_string file, t_uint32 line, t_address a, t_address b)
{
   VERIFY_FASSERT((a.type!=INVALID)&&(b.type!=INVALID), ("Invalid in ADD (called at line %d in %s)",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses at line %d in %s",line,file));
   if (a.type==BIT32)
      return a.types.addr32>b.types.addr32;
   else 
      return a.types.addr64>b.types.addr64;
}

t_bool real_GENERIC_GE(t_string file, t_uint32 line,t_address a, t_address b)
{
   VERIFY_FASSERT((a.type!=INVALID)&&(b.type!=INVALID), ("Invalid in GE (called at line %d in %s)",line,file));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("Added different size addresses"));
   if (a.type==BIT32)
      return a.types.addr32>=b.types.addr32;
   else 
      return a.types.addr64>=b.types.addr64;
}

t_bool real_GENERIC_LT(char * file, t_uint32 line, t_address a, t_address b)
{
   VERIFY_FASSERT((a.type!=INVALID)&&(b.type!=INVALID), ("Invalid in LEQ called in %s,%d",file,line));
   VERIFY_FASSERT((a.type==b.type)||(a.type==EITHER)||(b.type==EITHER), ("Added different size addresses at line %d in %s",line,file));
   if ((a.type==BIT32)||(b.type==BIT32))
     {
       return a.types.addr32<b.types.addr32;
     }
   else
     {
       return a.types.addr64<b.types.addr64;
     }
}

t_bool real_GENERIC_LE(char * file, t_uint32 line, t_address a, t_address b)
{
   VERIFY_FASSERT((a.type!=INVALID)&&(b.type!=INVALID), ("Invalid in ADD"));
   VERIFY_FASSERT((a.type==b.type)||(a.type==EITHER)||(b.type==EITHER), ("Added different size addresses"));
   if ((a.type==BIT32)||(b.type==BIT32))
     {
       return a.types.addr32<=b.types.addr32;
     }
   else
     {
       return a.types.addr64<=b.types.addr64;
     }
}


t_bool real_GENERIC_EQ(char * file, int line, t_address a, t_address b)
{
   VERIFY_FASSERT((a.type!=INVALID)&&(b.type!=INVALID), ("Invalid in GENERIC_EQ ( @G, @G ) called at %s:%d",a,b,file,line));
   if ((a.type!=EITHER)&&(b.type!=EITHER))
   VERIFY_FASSERT((a.type==b.type), ("GENERIC_EQ at line %d in %s has different type operands",line,file));
   if (a.type==BIT32)
      return a.types.addr32==b.types.addr32;
   else
      return a.types.addr64==b.types.addr64;
}

void * real_GENERIC_ADD_DISPL(char * file, int line,void * in,t_address addr)
{
   if (addr.type==BIT32)
   {
      return ((void *) (((char *) in) +  addr.types.addr32));
   }
   else
   {
      return ((void *) (((char *) in) +  addr.types.addr64));
   }
}
t_address real_GENERIC_MUL_UINT32(char * file, int line, t_address address, t_uint32 div)
{
   t_address ret;
   ret.type=address.type;
   if (address.type==BIT32)
   {
      ret.types.addr32=address.types.addr32*div;
   }
   else if (address.type==BIT64)
   {
      ret.types.addr64=address.types.addr64*((t_uint64) div);
   }
   else if (address.type==INVALID)
   {

      FATAL(("Division with invalid type!"));
   }
   else if (address.type==EITHER)
   {
      FATAL(("IMPLEMENT"));
   }
   else
   {
      FATAL(("Address is corrupt (called at file %s, line %d) ! %d",file,line,address.type));
   }
   return ret;
}

t_address real_GENERIC_DIV_UINT32(char * file, int line, t_address address, t_uint32 div)
{
   t_address ret;
   ret.type=address.type;
   if (address.type==BIT32)
   {
      ret.types.addr32=address.types.addr32/div;
   }
   else if (address.type==BIT64)
   {
      ret.types.addr64=address.types.addr64/((t_uint64) div);
   }
   else if (address.type==INVALID)
   {

      FATAL(("Division with invalid type!"));
   }
   else if (address.type==EITHER)
   {
      FATAL(("IMPLEMENT"));
   }
   else
   {
      FATAL(("Address is corrupt (called at file %s, line %d) ! %d",file,line,address.type));
   }
   return ret;
}

t_uint32 real_GENERIC_EXTRACT_UINT32(char * file, int line, t_address addr)
{
   if (addr.type==BIT32)
   {
      return addr.types.addr32;
   } 
   if (addr.types.addr64>0xffffffffull) FATAL(("Overflow at line %d in file %s\n",line,file));
   return (t_uint32) addr.types.addr64;
}

t_uint64 real_GENERIC_EXTRACT_UINT64(char * file, int line, t_address addr)
{
   if (addr.type==BIT32)
   {
      return (t_uint64) addr.types.addr32;
   } 
   return addr.types.addr64;
}

t_int32 real_GENERIC_EXTRACT_INT32(char * file, int line, t_address addr)
{
   if (addr.type==BIT32)
   {
      return (t_int32) addr.types.addr32;
   } 
   if (addr.types.addr64>0xffffffffull) 
   {
     if ((addr.types.addr64 & 0xffffffff80000000ull) != 0xffffffff80000000ull)
       FATAL(("Overflow at line %d in file %s\n",line,file));
     else
       /* negative number */
       return (t_int32) ((t_int64) addr.types.addr64);
   }
   return (t_int32) ((t_uint32) addr.types.addr64);
}

t_bool real_GENERICISNULL(t_address x)
{
   if (x.type==BIT32) return (x.types.addr32)==0;
   return (x.types.addr64)==0;
}


t_address real_GENERIC_SIGN_EXTEND(char * file, t_uint32 line, t_address x, t_uint32 signbit)
{
  if (x.type == BIT32)
  {
    x.types.addr32 = (x.types.addr32 ^ (1 << signbit)) - (1 << signbit);
  }
  else if (x.type == BIT64)
  {
    x.types.addr64 = (x.types.addr64 ^ (1 << signbit)) - (1 << signbit);
  }
  else if (x.type == EITHER)
  {
    FATAL(("Implement"));
  }
  else
  {
    FATAL(("address is corrupt"));
  }
  
  return x;
}

#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
