/******************************************************************************
 * Copyright 2001,2002,2003: {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/
#define NO_CONSTPROP_OVER_HELL
#define CONST_PROP_CONDITIONAL_BRANCHES_WITH_SIDE_EFFECT_OPT
/* Includes {{{ */
#include <stdlib.h>
#include "diablo_procstate.h"
#include "diablo_reloc.h"
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_function.h"
#include "diablo_print.h"
#include "diablo_constprop.h"
#include "diablo_iterators.h"
#include "diablo_liveness.h"
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../diablo_options.h"
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_constant_propagation.h"
#ifdef DIETER
#include "../arch/arm/arm_constantoptimizations.h"
#endif
#endif
/* }}} */
t_uint32 global_optimization_phase=0;

t_procstate * temp_procstate;
t_procstate * temp_procstate2;
/* Defines for static functions {{{ */
void FunctionPropagateConstants(t_object* obj, t_function * function, t_section *sec, t_cfg_edge * edge_in, t_bool during_fixpoint_calculations, t_bool propagate_over_all_outgoing_edges);
void BblPropagateConstants(t_object* obj, t_ConstantPropagationInsEmul propagator, t_bbl* bbl, t_cfg_edge * edge_out, t_bool during_fixpoint_calculations, t_bool propagate_over_all_outgoing_edges);
static void FunctionCIPropagateConstants(t_object* obj, t_function * function, t_section *sec,  t_bool during_fixpoint_calculations);
void BblCIPropagateConstants(t_object* obj, t_ConstantPropagationInsEmul propagator, t_bbl* bbl, t_bool during_fixpoint_calculations);
static void CfgRemoveNotExecutedIns(t_cfg *fg, t_reloc_table * table);
static void CfgApplyConstantInformation(t_object * obj, t_section * sec);
static void FlowgraphRemoveUnreachableEdges(t_cfg * cfg);
static void FlowgraphRemoveUnreachableBlocks(t_cfg *cfg);
t_bool FunctionUnmarkBbl(t_function * fun, t_bbl** bbl);
static void CfgUnconditionalizeIns(t_cfg * cfg);
static void FunctionDecrementUnexecedInsBy(t_function * fun, t_uint32 diff);
static void DumpCPInfo(t_object * obj,FILE * f);
/*}}}*/

/* Utility functions {{{ */

/* a first group of utility procedures is used to created linked lists
   of procedures, incoming edges into procedures and basic blocks in
   procedures that need re-evaluation 
*/
/*
  static void PrintMarkedFuns(t_cfg* cfg)
  {
  t_function * fun;
  t_cfg_edge * edge;

  printf("MARKED FUNS list by running over whole cfg\n");

  CFG_FOREACH_FUN(cfg,fun)
  if (FUN_NEXT_MARKED_IN_CFG(fun))
  {
  BblPrint(fun->bbl_first,0);
  BBL_FOREACH_PRED_EDGE(fun->bbl_first,edge)
  {
  CfgEdgePrint(edge,0);
  if (EDGE_NEXT_MARKED_INTO_FUN(edge))
  {
  printf("MARKED FUNS marked edge\n");
  }
  else
  {
  printf("MARKED FUNS unmarked edge\n");
  }
  }
  }
  
  printf("MARKED FUNS list by running over whole cfg\n");

  fun = FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg));

  while (fun && fun != CFG_MARKED_FUNS(cfg))
  {
  BblPrint(fun->bbl_first,0);
  fun = FUN_NEXT_MARKED_IN_CFG(fun);
  }
  }
*/

t_bool CfgUnmarkFun(t_cfg* cfg, t_function** fun)
{

  *fun = FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg));

  if (*fun==CFG_MARKED_FUNS(cfg))
    return FALSE;

  FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg)) = FUN_NEXT_MARKED_IN_CFG(*fun);
  FUN_NEXT_MARKED_IN_CFG(*fun) = NULL;

  return TRUE;
}

void CfgUnmarkAllFun(t_cfg * cfg)
{
  t_function * fun;
  if (!(CFG_MARKED_FUNS(cfg)))
    {
      CFG_MARKED_FUNS(cfg) = T_FUN(Malloc(sizeof(t_function)));
      FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg))=CFG_MARKED_FUNS(cfg);
    }
  else
    while (CfgUnmarkFun(cfg,&fun)) ;
}



void CfgMarkFun(t_cfg * cfg, t_function* fun)
{
  if (!(FUN_NEXT_MARKED_IN_CFG(fun)))
    {
      FUN_NEXT_MARKED_IN_CFG(fun) = FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg));
      FUN_NEXT_MARKED_IN_CFG(CFG_MARKED_FUNS(cfg)) = fun;
    }
}

t_bool FunctionUnmarkBbl(t_function * fun, t_bbl** bbl)
{
  *bbl = BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun));

  if (*bbl==FUN_MARKED_BBLS(fun))
    return FALSE;

  BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun)) = BBL_NEXT_MARKED_IN_FUN(*bbl);
  BBL_NEXT_MARKED_IN_FUN(*bbl) = NULL;

  return TRUE;
}

void FunctionUnmarkAllBbls(t_function * fun)
{
  t_bbl * bbl;
  if (!(FUN_MARKED_BBLS(fun)))
    {
      FUN_MARKED_BBLS(fun) = T_BBL(Malloc(sizeof(t_bbl)));
      BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(fun))=FUN_MARKED_BBLS(fun);
    }
  else
    while (FunctionUnmarkBbl(fun,&bbl)) ;

  /* make sure there is no garbage in the BBL_NEXT_MARKED_IN_FUN fields */
  FUN_FOREACH_BBL(fun,bbl)
    BBL_NEXT_MARKED_IN_FUN(bbl) = NULL;
}




void FunctionMarkBbl(t_function * fun, t_bbl * bbl)
{
  if (!(BBL_NEXT_MARKED_IN_FUN(bbl)))
    {
      BBL_NEXT_MARKED_IN_FUN(bbl) = BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(BBL_FUN(bbl)));
      BBL_NEXT_MARKED_IN_FUN(FUN_MARKED_BBLS(BBL_FUN(bbl))) = bbl;
    }
}

t_bool FunctionUnmarkEdge(t_function * fun,t_cfg_edge** edge)
{
  *edge = EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun));

  if (*edge==FUN_MARKED_EDGES(fun))
    return FALSE;

  EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun)) = EDGE_NEXT_MARKED_INTO_FUN(*edge);
  EDGE_NEXT_MARKED_INTO_FUN(*edge) = NULL;

  return TRUE;
}

void FunctionUnmarkAllEdges(t_function * fun)
{
  t_cfg_edge * edge;
  if (!(FUN_MARKED_EDGES(fun)))
    {
      FUN_MARKED_EDGES(fun) = T_CFG_EDGE(Malloc(sizeof(t_cfg_edge)));
      EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun))=FUN_MARKED_EDGES(fun);
    }
  else
    while (FunctionUnmarkEdge(fun,&edge)) ;
}

void FunctionFreeMarkedSpace(t_function * fun)
{
  if (FUN_MARKED_EDGES(fun))
    {
      Free(FUN_MARKED_EDGES(fun)); 
      FUN_MARKED_EDGES(fun)=NULL;
    }
  if (FUN_MARKED_BBLS(fun))
    {
      Free(FUN_MARKED_BBLS(fun));
      FUN_MARKED_BBLS(fun) = NULL;
    }
}

void FunctionMarkEdge(t_function * fun, t_cfg_edge * edge)
{
  if (!(EDGE_NEXT_MARKED_INTO_FUN(edge)))
    {
      EDGE_NEXT_MARKED_INTO_FUN(edge) = EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun));
      EDGE_NEXT_MARKED_INTO_FUN(FUN_MARKED_EDGES(fun)) = edge;
    }
}

void FunctionMarkAllTrueIncomingEdges(t_function * fun)
{
  t_bbl * bbl;
  t_cfg_edge * edge;

  FUN_FOREACH_BBL(fun,bbl)
    {
      BBL_FOREACH_PRED_EDGE(bbl,edge)
	{
	  if (CfgEdgeIsForwardInterproc(edge))
	    {
	      if (EdgeIsMarked(edge))
		{
		  FunctionMarkEdge(fun,edge);
		}
	    }
	}
    }
}

/*}}}*/
#ifdef ARM_SUPPORT
void UdLookForExtraConstants(t_section *sec,const char * fname);
#endif
/* CSConstantPropagation {{{ */
/*! 
  \param obj
  \param sec 
  \param outputfilename
   
  Now we come to the actual constant propagation implementation.  All
  information to apply a constant propagation should be found in the
  object obj. The constant propagation is performed for one section
  only: sec.

  if outputfilename is not NULL, we the results of constant propagation are dumped 
  into a file, and no optimizations are performed

  \pre initialize the obj properly
*/
   
void ConstantPropagation(t_cfg * cfg, char * outputfilename, t_bool after_ud_analysis)
{
  t_object * obj=CFG_OBJECT(cfg);
  t_section * sec=CFG_SECTION(cfg);

  /* iterators */
  t_function * i_fun;
  t_cfg_edge * i_edge;
  t_ins * i_ins;
  t_bbl * i_bbl;

  /* local variables */

  t_bbl * program_entry;
  t_function * entry_fun;

  temp_procstate = ProcStateNew();
  temp_procstate2 = ProcStateNew();

  ANALYSIS_START("Constant Propagation");
  /* All linked lists of to be evaluated items should be emptied. */

  
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	if (CfgEdgeIsFallThrough(i_edge))
	  {
	    CfgEdgeMoveToLastSucc(i_edge);
	    break;
	  }
    }

  InitEdgeMarking();

  CfgUnmarkAllFun(cfg);

  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionUnmarkAllEdges(i_fun);
      FunctionUnmarkAllBbls(i_fun);
    }

  /* All instructions are initially assumed not executed (not reachable), as
   * this constant propagation includes a reachable code detection. We also
   * assume that all (conditional) instructions, are always executed once they
   * are reachable.
   *
   * The IF_EXECED flag needs to be set during the instruction evaluation for
   * instructions that might be executed (are reachable and, for conditional
   * instructions, the condition flags might hold), the IF_ALWAYS_EXECED flag
   * needs to be unset for conditional instructions if their condition might
   * evaluate to false.
   *
   * As a result, after the partial evaluation, instructions in reachable
   * blocks with IF_EXECED not set are turned into NOOPS. Instructions in
   * unreachable blocks are removed turned into deadcOde. Conditional
   * instructions which have IF_EXECED and IF_ALWAYS_EXECUTED set are
   * unconditionalized.
   */

  ComLiveness(cfg,TRIVIAL);
  
  ComputeSavedChangedRegisters(cfg->sec);

  ComLiveness(cfg,SENSITIVE);

  ComputeSavedChangedRegisters(cfg->sec);

  CFG_FOREACH_FUN(cfg,i_fun)
    FUN_FOREACH_BBL(i_fun,i_bbl)
    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      INS_ATTRIB(i_ins) &= ~IF_EXECED;
      INS_ATTRIB(i_ins) |= IF_ALWAYS_EXECED;
    }

  BBL_REGS_LIVE_OUT(cfg->exit_call_hell_node) = AllRegs();
  BBL_REGS_LIVE_OUT(cfg->call_hell_node) = AllRegs();

  /* 
     No edge are initially assumed reachable, ProcStates are
     allocated, the architecture dependent evaluation of whether
     information needs to be propagated over an edge is initialized,
     and the registers to be propagated over edges is initialized.

     The latter for the moment is the same for all edges, as we do not
     have a saved-changed register analysis for procedures yet.
  */


  CFG_FOREACH_EDGE(cfg,i_edge)
    {
      EdgeUnmark(i_edge);
      if (CfgEdgeIsInterproc(i_edge))
	i_edge->procstate = ProcStateNew();
      CFG_OBJECT(cfg)->actions->EdgePropagator(i_edge,BBL_INS_LAST(CFG_EDGE_HEAD(i_edge)));
      RegsetDup(i_edge->prop_regs,obj->description->all_registers);

      if (EDGE_CAT(i_edge)==ET_RETURN)
	RegsetSetIntersect(i_edge->prop_regs,FUN_REGS_CHANGED(BBL_FUN(EDGE_HEAD(i_edge))));
   
      /* LINK_EDGE_REMOVAL 

      if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK)
      {
      t_cfg_edge * j_edge;
      BBL_FOREACH_SUCC_EDGE(EDGE_HEAD(i_edge),j_edge)
      if (EDGE_CAT(j_edge)==ET_CALL) break;

      if (j_edge)
      RegsetSetDiff(i_edge->prop_regs,FUN_REGS_CHANGED(BBL_FUN(EDGE_TAIL(j_edge))));
      else 
      FATAL(("Could not find corresponding call for @E", i_edge));
      }
      LINK_EDGE_REMOVAL */
      
      i_edge->n_unevaluated_ins = 0;
    }

  /*
    Constant propagation is started at the entry point of the
    program. This is not done through incoming edges, therefore the
    edge is null. As we want to propagate the outgoing information
    over outgoing edges, the last but one parameter is TRUE.  As we
    want to propagate over all backward outgoing edges, the LAST
    parameter is true as well.
  */

  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);
  FunctionPropagateConstants(obj,entry_fun,sec,NULL,TRUE,TRUE);

  BBL_FOREACH_PRED_EDGE(program_entry,i_edge)
    {
      ProcStateSetAllBot(i_edge->procstate,obj->description->all_registers);
      EdgeMark(i_edge);
    }
  
#ifdef NO_CONSTPROP_OVER_HELL
  FunctionPropagateConstants(obj,BBL_FUN(cfg->hell_node),sec,NULL,TRUE,TRUE);
  FunctionPropagateConstants(obj,BBL_FUN(cfg->call_hell_node),sec,NULL,TRUE,TRUE); /*BJORN*/
#endif
  /* 
     Actual interprocedural fixpoint stuff: as long as there are
     marked procedures, propagate information over marked (incoming)
     edges through them 
  */

  while (CfgUnmarkFun(cfg,&i_fun))
    {
      while (FunctionUnmarkEdge(i_fun,&i_edge))
	{
	  FunctionPropagateConstants(obj,i_fun, sec, i_edge,TRUE,FALSE);
	}
    }
  ANALYSIS_SIMPLE_END("Constant Propagation");

}

void OptUseConstantInformation(t_cfg * cfg, char * outputfilename, t_bool after_ud_analysis)
{
  t_object * obj=CFG_OBJECT(cfg);
  t_section * sec=CFG_SECTION(cfg);

  if (outputfilename != NULL)
    {
      /* dump the results of constantpropagation in the file */
      FILE * f = fopen(outputfilename,"w");

      VERIFY_FASSERT(f,("Could not open %s for writing!",outputfilename));
      DumpCPInfo(obj,f);
      fclose(f);
    }
  else if (!after_ud_analysis)
    {
      ANALYSIS_START("Constant propagation optimizations");
      /* 
	 Remove unreachable nodes from the graph 
      */

      FlowgraphRemoveUnreachableBlocks(cfg);

      /* 
	 The same for edges.
      */
      FlowgraphRemoveUnreachableEdges(cfg);

      /*
	Instructions that are executed and are always executed (conditions
	always hold) are unconditionalized.
      */

      CfgUnconditionalizeIns(cfg);


      /* 
	 Instructions that are never excecuted because of non holding
	 conditions are turned into NOOPS.
      */

      CfgRemoveNotExecutedIns(cfg,obj->reloc_table);

      /* 
	 All incoming edges will be propagated into the procedures, to
	 detect the constants inside the procedure and optimize the code
	 based on those constants.
      */

      InitDelayedInsKilling();
      CfgApplyConstantInformation(obj,sec);
      ApplyDelayedInsKilling();
      ANALYSIS_SIMPLE_END("Constant propagation optimizations");
    }
  else
    {
#ifdef ARM_SUPPORT
      UdLookForExtraConstants(sec,"./extraconst");
#endif
    }
}

void FreeConstantInformation(t_cfg * cfg)
{
  t_cfg_edge * i_edge;
  t_function * i_fun;
  /* Free the allocated space for interprocedural procstates.  */

  CFG_FOREACH_EDGE(cfg,i_edge)
    {
      if (CfgEdgeIsInterproc(i_edge))
	{
	  ProcStateFree(i_edge->procstate);
	  i_edge->procstate=NULL;
	}
    }
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionFreeMarkedSpace(i_fun);
    }

  ProcStateFree(temp_procstate);
  ProcStateFree(temp_procstate2);
  ProcStateRealFree();
  Free(CFG_MARKED_FUNS(cfg));
  CFG_MARKED_FUNS(cfg)=NULL;
}

/*}}}*/
/* CCIonstantPropagation {{{ */
void CIConstantPropagation(t_object * obj, 
			   t_section * sec, 
			   char * outputfilename,
			   t_bool after_ud_analysis)
{
  /* iterators */
  t_function * i_fun;
  t_cfg_edge * i_edge;
  t_ins * i_ins;
  t_bbl * i_bbl;


  /* local variables */

  t_bbl * program_entry;
  t_function * entry_fun;
  t_cfg *cfg = T_CFG(sec->data);


  /* All linked lists of to be evaluated items should be emptied. */

  InitEdgeMarking();

  CfgUnmarkAllFun(cfg);
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionUnmarkAllEdges(i_fun);
      FunctionUnmarkAllBbls(i_fun);
    }

  /* 
     All instructions are initially assumed not executed (not
     reachable), as this constant propagation includes a reachable
     code detection. We also assume that all (conditional)
     instructions, are always executed once they are reachable.

     The IF_EXECED flag needs to be set during the instruction
     evaluation for instructions that might be executed (are reachable
     and, for conditional instructions, the condition flags might
     hold), the IF_ALWAYS_EXECED flag needs to be unset for
     conditional instructions if their condition might evaluate to
     false.

     As a result, after the partial evaluation, instructions in
     reachable blocks with IF_EXECED not set are turned into
     NOOPS. Instructions in unreachable blocks are removed turned into
     deadcOde. Conditional instructions which have IF_EXECED and
     IF_ALWAYS_EXECUTED set are unconditionalized.

  */

  CFG_FOREACH_FUN(cfg,i_fun)
    FUN_FOREACH_BBL(i_fun,i_bbl)
    BBL_FOREACH_INS(i_bbl,i_ins)
    {
      INS_ATTRIB(i_ins) &= ~IF_EXECED;
      INS_ATTRIB(i_ins) |= IF_ALWAYS_EXECED;
    }
  
  /* 
     No edge are initially assumed reachable, ProcStates are
     allocated, the architecture dependent evaluation of whether
     information needs to be propagated over an edge is initialized,
     and the registers to be propagated over edges is initialized.

     The latter for the moment is the same for all edges, as we do not
     have a saved-changed register analyses for procedures yet.
  */
  
  
  CFG_FOREACH_EDGE(cfg,i_edge)
    {
      EdgeUnmark(i_edge);
      if (CfgEdgeIsInterproc(i_edge))
	i_edge->procstate = ProcStateNew();
      CFG_OBJECT(cfg)->actions->EdgePropagator(i_edge,CFG_EDGE_HEAD(i_edge)->ins_last);
      RegsetDup(i_edge->prop_regs,obj->description->all_registers);
      if (EDGE_CAT(i_edge)==ET_RETURN)
	RegsetSetIntersect(i_edge->prop_regs,FUN_REGS_CHANGED(BBL_FUN(EDGE_HEAD(i_edge))));

      /* LINK_EDGE_REMOVAL
	 if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK)
	 {
	 t_cfg_edge * j_edge;
	 BBL_FOREACH_SUCC_EDGE(EDGE_HEAD(i_edge),j_edge)
	 if (EDGE_CAT(j_edge)==ET_CALL) break;
	      
	 if (j_edge)
	 RegsetSetDiff(i_edge->prop_regs,FUN_REGS_CHANGED(BBL_FUN(EDGE_TAIL(j_edge))));
	 }
	 LINK_EDGE_REMOVAL */
      i_edge->n_unevaluated_ins = 0;
    }

  /*
    Constant propagation is started at the entry point of the
    program. This is not done through incoming edges, therefore the
    edge is null. As we want to propagate the outgoing information
    over outgoing edges, the last parameter is TRUE.
  */

  CFG_FOREACH_FUN(cfg,i_fun)
    {
      i_fun->procstate_in = ProcStateNew();
    }

  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);
  ProcStateSetAllBot(entry_fun->procstate_in,obj->description->all_registers);
  CfgMarkFun(cfg,entry_fun);

  /* 
     Actual interprocedural fixpoint stuff: as long as there are
     marked procedures, propagate information over marked (incoming)
     edges through them 
  */

  while (CfgUnmarkFun(cfg,&i_fun))
    {
      /* printf("CONSTANT PROPAGATION start fun %x\n",G_T_UINT32(BBL_BASE_ADDRESS(i_fun->bbl_first))); */

      FunctionCIPropagateConstants(obj,i_fun, sec, TRUE);
    }

  printf("END OF FIXPOINT\n");

  if (outputfilename != NULL)
    {
      /* dump the results of constantpropagation in the file */
      FILE * f = fopen(outputfilename,"w");

      VERIFY_FASSERT(f,("Could not open %s for writing!",outputfilename));
      DumpCPInfo(obj,f);
      fclose(f);
    }
  else if (!after_ud_analysis)
    {
      /* 
	 Remove unreachable nodes from the graph 
      */

      FlowgraphRemoveUnreachableBlocks(cfg);

      /* 
	 The same for edges.
      */
      FlowgraphRemoveUnreachableEdges(cfg);

      /*
	Instructions that are executed and are always executed (conditions
	always hold) are unconditionalized.
      */

      /*    CfgUnconditionalizeIns(cfg); */


      /* 
	 Instructions that are never excecuted because of non holding
	 conditions are turned into NOOPS.
      */


      /*    CfgRemoveNotExecutedIns(cfg,obj->reloc_table); */

      /* 
	 All incoming edges will be propagated into the procedures, to
	 detect the constants inside the procedure and optimize the code
	 based on those constants.
      */

      /*    DelayedRelocKilling(TRUE); */
      /*    CfgApplyConstantInformation(obj,sec); */
      /*    DelayedRelocKilling(FALSE); */
    }
  else
    {
      /* TODO: invullen! */
#ifdef ARM_SUPPORT
      UdLookForExtraConstants(sec,"./extraconst");
#endif
    }
  
  /*
    Free the allocated space for interprocedural procstates.
  */
  

  CFG_FOREACH_EDGE(cfg,i_edge)
    {
      if (CfgEdgeIsInterproc(i_edge))
	{
	  ProcStateFree(i_edge->procstate);
	  i_edge->procstate=NULL;
	}
    }
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionFreeMarkedSpace(i_fun);
    }
  Free(CFG_MARKED_FUNS(cfg));
  CFG_MARKED_FUNS(cfg) = NULL;

}
/*}}}*/

/* FunctionPropagateConstants {{{ */

/*! Propagate information through a procedure 
 *
 * \par obj
 * 
 * \par fun 
 * 
 * \par sec 
 * 
 * \par edge_in if NULL this propagation is not for an incoming edge, but for
 * the program entry point 
 *
 * \par during_fixpoint_calculations is false if no outgoing information has to
 * be propagated over interprocedural edges 
 */

void FunctionPropagateConstants(t_object * obj,
				t_function * fun, 
				t_section * sec, 
				t_cfg_edge * edge_in,
				t_bool during_fixpoint_calculations,
				t_bool propagate_over_all_outgoing_edges
				)
{
  /* iterators */

  t_cfg_edge * i_edge;
  t_bbl * i_bbl;

  /* local variables */

  t_cfg_edge * edge_out = NULL;
  t_cfg *cfg = T_CFG(sec->data);

  /* 
     Allocate init propagation inside fun 
  */

  FUN_FOREACH_BBL(fun,i_bbl)
    {
      BBL_PROCSTATE_IN(i_bbl)=ProcStateNew();
    }

  /* 
     propagate the necessary information into the procedure 
  */

  if (edge_in)
    {
      /* 
	 since edge_in is not NULL, we are not propagating for the entry procedure 
      */

      if (!during_fixpoint_calculations)
	{

	  /* 
	     this means we have to propagate information over all incoming edges into the procedure 
	  */

	  FUN_FOREACH_BBL(fun,i_bbl)
	    BBL_FOREACH_PRED_EDGE(i_bbl, i_edge) 
	    {
	      if (CfgEdgeIsForwardInterproc(i_edge))
		if (EdgeIsMarked(i_edge))
		  {
		    if (ProcStateJoinSimple(BBL_PROCSTATE_IN(i_bbl),i_edge->procstate,i_edge->prop_regs))
		      FunctionMarkBbl(BBL_FUN(i_bbl),i_bbl);
		    if (EDGE_CAT(i_edge) == ET_IPFUNLINK)
		      {
			/* this is a very special case: an interprocedural funlink
			 * is a forward interprocedural edge on which not all registers
			 * are propagated. Therefore, we also need to find the corresponding
			 * return edge and use the information on this edge as well */
			t_cfg_edge * ret_edge;

			FATAL(("IP FUNLINKS SHOULD NO LONGER EXIST"));

			BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(i_edge),ret_edge)
			  if (EDGE_CAT(ret_edge) == ET_RETURN)
			    break;
			VERIFY_FASSERT(ret_edge,("incoming ipfunlink without incoming return edge!"));

			if (ProcStateJoinSimple(BBL_PROCSTATE_IN(i_bbl),ret_edge->procstate,ret_edge->prop_regs))
			  FunctionMarkBbl(BBL_FUN(i_bbl),i_bbl);
		      }
		  }
	    }
	}
      else
	{
	  /*
	    Propagation will start at the tail of the edge, where
	    either the incoming information is used or everything is
	    set to bot (in case of hell nodes). This is not correct
	    once we will track addresses or relocs propagated to hell.
	  */
      
	  if (CFG_EDGE_TAIL(edge_in)==cfg->call_hell_node || CFG_EDGE_TAIL(edge_in)==cfg->hell_node) 
	    /*      if (CFG_EDGE_TAIL(edge_in)==cfg->hell_node) BJORN */
	    {
	      ProcStateSetAllBot(BBL_PROCSTATE_IN(CFG_EDGE_TAIL(edge_in)),RegsetNewInvers(obj->description->registers_prop_over_hell));	
	    }
      
      
	  ProcStateJoinSimple(BBL_PROCSTATE_IN(CFG_EDGE_TAIL(edge_in)),edge_in->procstate,edge_in->prop_regs);
      
	  if (EDGE_CAT(edge_in) == ET_IPFUNLINK)
	    {
	      /* for interprocedural funlinks we must also use the information
	       * from the corresponding return edge. this is because not all 
	       * registers are propagated over the funlink edge, and otherwise
	       * we wouldn't be able to propagate the information through the
	       * function at all. */
	      t_cfg_edge * ret_edge;

	      FATAL(("IP FUNLINKS SHOULD NO LONGER EXIST"));

	      BBL_FOREACH_PRED_EDGE(CFG_EDGE_TAIL(edge_in),ret_edge)
		if (EDGE_CAT(ret_edge) == ET_RETURN)
		  break;
	      VERIFY_FASSERT(ret_edge,("no return for interprocedural funlink"));
	  
	      ProcStateJoinSimple(BBL_PROCSTATE_IN(CFG_EDGE_TAIL(edge_in)),ret_edge->procstate,ret_edge->prop_regs);
	    }
      
	  FunctionMarkBbl(BBL_FUN(CFG_EDGE_TAIL(edge_in)),CFG_EDGE_TAIL(edge_in));
      
	  /*
	    For return or compensating outgoing edges, we only have to propagate over the corresponding edge.
	  */
      
	  edge_out = edge_in->corr_edge;
	}
    }
  else
    {
      /*
	We are working on the entry fun or on hell initially.
      */

      if (!during_fixpoint_calculations)
	{
	  /*
	    if there are more other incoming edges, propagate the information only if we are not in the fixpoint calculations
	  */

	  FUN_FOREACH_BBL(fun,i_bbl)
	    BBL_FOREACH_PRED_EDGE(i_bbl, i_edge) 
	    {
	      if (CfgEdgeIsForwardInterproc(i_edge))
		if (EdgeIsMarked(i_edge))
		  {
		    if (ProcStateJoinSimple(BBL_PROCSTATE_IN(i_bbl),i_edge->procstate,i_edge->prop_regs))
		      FunctionMarkBbl(BBL_FUN(i_bbl),i_bbl);
		  }
	    }
	}

      FunctionMarkBbl(fun,fun->bbl_first);

      if (IsHellNode(fun->bbl_first)) 
	/*    if (fun->bbl_first==fun->cfg->hell_node) BJORN */
	ProcStateSetAllBot(BBL_PROCSTATE_IN(fun->bbl_first),RegsetNewInvers(obj->description->registers_prop_over_hell));
      else
	ProcStateSetAllBot(BBL_PROCSTATE_IN(fun->bbl_first),AllRegs());
    }

  /* 
     do the actual fixpoint calculation inside the fun
  */

  while (FunctionUnmarkBbl(fun,&i_bbl))
    {
      BblPropagateConstants(obj,CFG_OBJECT(cfg)->actions->InstructionEmulator,i_bbl,edge_out,during_fixpoint_calculations,propagate_over_all_outgoing_edges);
    }

  /* 
     if the fixpoint calculation is still going on, count the number of unexecuted instructions and
     store it on the incoming edge that caused the propagation
  */
  if (during_fixpoint_calculations)
    {
      if (edge_in)
	{
	  t_uint32 unexeced = 0;
	  t_ins * i_ins;

	  /* count the number of unexecuted instructions */
	  FUN_FOREACH_BBL(fun,i_bbl)
	    {
	      BBL_FOREACH_INS(i_bbl,i_ins)
		{
		  if (!(INS_ATTRIB(i_ins) & IF_EXECED))
		    unexeced++;
		}
	    }
	  edge_in->n_unevaluated_ins = unexeced;
	}
    }

  /* 
     free the allocated memory
  */

  if (during_fixpoint_calculations)
    {
      FUN_FOREACH_BBL(fun,i_bbl)
	{
	  ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
	  BBL_PROCSTATE_IN(i_bbl) = NULL;
	}
    }
}
/*}}}*/
/* FunctionCIPropagateConstants {{{ */
static void FunctionCIPropagateConstants(t_object * obj,
					 t_function * fun, 
					 t_section *sec, 
					 t_bool during_fixpoint_calculations)
{
  /* iterators */

  t_cfg_edge * i_edge;
  t_bbl * i_bbl;

  /* local variables */

  t_cfg *cfg = T_CFG(sec->data);

  /* 
     Allocate init propagation inside fun 
  */

  /*printf("fun at ");BblPrint(fun->bbl_first,0);fflush(stdout);
    if (G_T_UINT32(BBL_BASE_ADDRESS(fun->bbl_first))==0xba6c)
    FunctionPrint(fun,PRINT_INSTRUCTIONS|PRINT_PREDECESSOR|PRINT_SUCCESSOR);fflush(stdout);
  */


  FUN_FOREACH_BBL(fun,i_bbl)
    {
      BBL_PROCSTATE_IN(i_bbl)=ProcStateNew();
    }

  /* 
     propagate the necessary information into the procedure 
  */

  FUN_FOREACH_BBL(fun,i_bbl)
    {
      if (!BBL_IS_FIRST(i_bbl))
	BBL_FOREACH_PRED_EDGE(i_bbl, i_edge) 
	{
	  if (CfgEdgeIsForwardInterproc(i_edge))
	    if (EdgeIsMarked(i_edge))
	      {
		if (ProcStateJoinSimple(BBL_PROCSTATE_IN(i_bbl),i_edge->procstate,i_edge->prop_regs))
		  FunctionMarkBbl(fun,i_bbl);
	      }
	}
    }

  FunctionMarkBbl(fun,fun->bbl_first);
  ProcStateDup(BBL_PROCSTATE_IN(fun->bbl_first),fun->procstate_in);

  /* 
     do the actual fixpoint calculation inside the fun
  */

  while (FunctionUnmarkBbl(fun,&i_bbl))
    {
      BblCIPropagateConstants(obj,CFG_OBJECT(cfg)->actions->InstructionEmulator,i_bbl,during_fixpoint_calculations);
    }

  /* 
     if the fixpoint calculation is still going on, count the number of unexecuted instructions and
     store it on the incoming edge that caused the propagation
  */
  if (during_fixpoint_calculations)
    {
      {
	t_uint32 unexeced = 0;
	t_ins * i_ins;

	/* count the number of unexecuted instructions */
	FUN_FOREACH_BBL(fun,i_bbl)
	  {
	    BBL_FOREACH_INS(i_bbl,i_ins)
	      {
		if (!(INS_ATTRIB(i_ins) & IF_EXECED))
		  unexeced++;
	      }
	  }
      }
    }

  /* 
     free the allocated memory
  */

  if (during_fixpoint_calculations)
    {
      FUN_FOREACH_BBL(fun,i_bbl)
	{
	  ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
	  BBL_PROCSTATE_IN(i_bbl) = NULL;

	}
    }
}
/*}}}*/

/*}}}*/
/*utility functions for PropOverEdge CI {{{ */

static void DoCIForwardInterproc(t_cfg_edge * edge, t_procstate * state_true);

static void DoCICall(t_cfg_edge * edge, t_procstate * state_true)
{
  t_cfg_edge * link_edge;

  FATAL(("replace call to DoCICall with call to DoCICall2"));

  link_edge=FindFunLinkForCall(edge);
  
  if (link_edge)
    {
      if (EDGE_CAT(link_edge)==ET_FUNLINK)
	{
	  if (ProcStateJoinSimple(BBL_PROCSTATE_IN(CFG_EDGE_TAIL(link_edge)),state_true,link_edge->prop_regs))
	    FunctionMarkBbl(BBL_FUN(CFG_EDGE_TAIL(link_edge)),CFG_EDGE_TAIL(link_edge));
	}
      else if (EDGE_CAT(link_edge)==ET_IPFUNLINK)
	{
	  FATAL(("IP FUNLINKS SHOULD NO LONGER EXIST"));
	  EdgeMark(link_edge);
	  if (ProcStateJoinSimple(link_edge->procstate,state_true,link_edge->prop_regs))
	    {
	      if (!BBL_IS_FIRST(CFG_EDGE_TAIL(link_edge)) || (ProcStateJoinSimple(CFG_EDGE_TAIL(link_edge)->fun->procstate_in,link_edge->procstate,link_edge->prop_regs)))
		CfgMarkFun(CFG_EDGE_CFG(link_edge),CFG_EDGE_TAIL(link_edge)->fun);
	    }
	  
	  /* now also propagate the corresponding edge of the funlink edge! */
	  DoCIForwardInterproc(link_edge,NULL);
	}
    }
}

static void DoCIForwardInterproc(t_cfg_edge * edge, t_procstate * state_true)
{
  t_cfg_edge * corr_edge = edge->corr_edge;

  if (corr_edge)
    {
      if (EdgeIsMarked(corr_edge))
	{
	  if (CFG_EDGE_TAIL(corr_edge)->fun==(CFG_EDGE_HEAD(edge)->fun) && CFG_EDGE_HEAD(edge)!=CFG_EDGE_TAIL(corr_edge))
	    {
	      if (ProcStateJoinSimple(BBL_PROCSTATE_IN(CFG_EDGE_TAIL(corr_edge)),corr_edge->procstate,corr_edge->prop_regs))
		{
		  FunctionMarkBbl(CFG_EDGE_TAIL(corr_edge)->fun,CFG_EDGE_TAIL(corr_edge));
		}
	    }
	}
    }

  if (EDGE_CAT(edge)==ET_CALL)
    {
      DoCICall(edge,state_true);
    }

}

/*}}} */

/* PropOverEdge {{{ */

/*! propagate a state over an edge 
 *
 * PRECONDITION: the edge_propagators must have
 * been initialized 
 *
 * POSTCONDITION: the state returned is not the original one:
 * it is adapted with information about the state in case the edge was not
 * followed: the returning state is the state if an edge is not taken 
 *
 * \return whether or not there is a need to propagate further (can it be that
 * the edge is not taken?) 
 *
 * \todo does not work for link edges if there need to
 * be propagated something over them */


static t_bool PropOverEdge(t_cfg_edge * edge, t_procstate ** state, t_bool do_it, t_bool during_fixpoint_calculations)
{
  t_procstate * state_true;
  t_procstate * state_false;

  edge->testcondition(edge,*state,&state_true,&state_false);
  
  if (state_true)
    {
      /* 
	 if the call to testcondition resulted in a state_true not
	 being null, the state pointed to by state_true can be
	 propagated over the edge */
      
      if (do_it)
	{
	  /* mark it as taken */
	  EdgeMark(edge);
	  
	  
	  if (CfgEdgeIsInterproc(edge))
	    {
	      /* 
		 if it is a true outgoing edge, the corresponding incoming edge must have its information propagated into the procedure (if it is marked that is).
	      */
	      
	      if (global_options.rely_on_calling_conventions)
		{
		  ProcStateSetAllBot(state_true,RegsetNewInvers(BBL_REGS_LIVE_OUT(CFG_EDGE_HEAD(edge))));
		  /*
		    if (IsHellNode(CFG_EDGE_HEAD(edge)))
		    {
		    DiabloPrintArch(stdout,CFG_EDGE_HEAD(edge)->fun->cfg->sec->obj->description,"CALLHELL @E @A\n",edge,BBL_REGS_LIVE_OUT(CFG_EDGE_HEAD(edge)));
		    if (CFG_EDGE_HEAD(edge)==CFG_EDGE_HEAD(edge)->fun->cfg->exit_call_hell_node) printf("YEP\n");
		    }*/
		  ProcStateSetAllBot(state_true,RegsetNewInvers(BblRegsLiveBefore(CFG_EDGE_TAIL(edge))));
		}

	      if (CfgEdgeIsForwardInterproc(edge))
		{

		  /* First, propagate what is known over corresponding edge */
		  t_cfg_edge * corr_edge = edge->corr_edge;
		  
		  if (corr_edge)
		    {
		      if (EdgeIsMarked(corr_edge))
			{
			  t_bbl * return_bbl = CFG_EDGE_TAIL(corr_edge);
			  if (return_bbl->fun==(CFG_EDGE_HEAD(edge)->fun) && CFG_EDGE_HEAD(edge)!=return_bbl)
			    {
			      if (ProcStateJoinSimple(BBL_PROCSTATE_IN(return_bbl),corr_edge->procstate,corr_edge->prop_regs))
				{
				  FunctionMarkBbl(CFG_EDGE_TAIL(corr_edge)->fun,CFG_EDGE_TAIL(corr_edge));
				}
			    }
			  
			  /* secondly, propagate over link_edge */
			  
			  if (EDGE_CAT(edge)==ET_CALL)
			    if (ProcStateJoinSimple(BBL_PROCSTATE_IN(return_bbl),state_true,RegsetNewInvers(corr_edge->prop_regs)))
			      FunctionMarkBbl(BBL_FUN(return_bbl),return_bbl);
			}
		    }
		}
	      
	      if (during_fixpoint_calculations)
		{
#if 0
		  if (edge->procstate==NULL) FATAL(("EDGE PROCSTATE == NULL!"));
#endif

		  if (ProcStateJoinSimple(edge->procstate,state_true,edge->prop_regs))
		    {
		      t_bbl* tail = CFG_EDGE_TAIL(edge);
		      if (!CfgEdgeIsForwardInterproc(edge))
			{
			  /* repropagation for all true incoming edges is necessary */
			  /*			  printf("marking a function and all its incoming edges!!!\n");
						  BblPrint(CFG_EDGE_TAIL(edge)->fun->bbl_first,0); */
			  CfgMarkFun(CFG_EDGE_CFG(edge),tail->fun);
			  FunctionMarkAllTrueIncomingEdges(tail->fun);
			}
		      else 
			{
			  /* only this edge will need to be repropagated */
			  CfgMarkFun(CFG_EDGE_CFG(edge),tail->fun);
			  FunctionMarkEdge(tail->fun,edge);
			}
		    }
		}
	    }
	  else 
	    {
	      t_bbl* tail = CFG_EDGE_TAIL(edge);

	      /* LINK_EDGE_REMOVAL
		 
	      if (global_options.rely_on_calling_conventions)
	      if (EDGE_CAT(edge)==ET_FUNLINK)
	      {
	      ProcStateSetAllBot(state_true,RegsetNewInvers(BBL_REGS_LIVE_OUT(CFG_EDGE_HEAD(edge))));
	      ProcStateSetAllBot(state_true,RegsetNewInvers(BblRegsLiveBefore(CFG_EDGE_TAIL(edge))));
	      }

	      LINK_EDGE_REMOVAL */

	      if (ProcStateJoinSimple(BBL_PROCSTATE_IN(tail),state_true,edge->prop_regs))
		{
		  FunctionMarkBbl(tail->fun,tail);
		}
	    }
	}
      if (state_false) 
	{
	  /* if there is a state to be propagated over the other
	     outgoing edges, the state over the true edge was freshly
	     allocated in the call to testcondition() and must be
	     free'ed here. The only exception is when both are the same */
	  if (state_false!=state_true)
	    {
	      ProcStateFree(state_true);
	    }
	  state_true = NULL;
	}
    }
  
  if (state_false)
    {
      /*
	The state to be sent over the next edge going out of a block
	is the false_state */
      *state=state_false;
      return TRUE;
    }
  else
    {
      /* 
	 if state_false is null after the call to testcondition(), it
	 was determined that no more states need to propagated over
	 outgoing edges of the basic block */
      
      return FALSE;
    }
}
/*}}}*/
/* PropOverEdge CI {{{ */

static t_bool CIPropOverEdge(t_cfg_edge * edge, t_procstate ** state, t_bool during_fixpoint_calculations)
{
  t_procstate * state_true;
  t_procstate * state_false;

  edge->testcondition(edge,*state,&state_true,&state_false);

  if (state_true)
    {
      /* 
	 if the call to testcondition resulted in a state_true not
	 being null, the state pointed to by state_true can be
	 propagated over the edge */

      /* mark it as taken */
      EdgeMark(edge);
     
      if (CfgEdgeIsInterproc(edge))
	{
	  /* 
	     if it is a true outgoing edge, the corresponding ingcoming edge must have its information propagated into the procedure (if it is marked that is).
	  */
	 
	  if (CfgEdgeIsForwardInterproc(edge))
	    DoCIForwardInterproc(edge,state_true);
	 
	  if (during_fixpoint_calculations)
	    {
	      if (edge->procstate==NULL) FATAL(("EDGE PROCSTATE == NULL!"));
	      if (ProcStateJoinSimple(edge->procstate,state_true,edge->prop_regs))
		{
		  if (!CfgEdgeIsForwardInterproc(edge))
		    {
		      /* repropagation for all true incoming edges is necessary */
		     
		      CfgMarkFun(CFG_EDGE_CFG(edge),CFG_EDGE_TAIL(edge)->fun);
		    }
		  else 
		    {
		      /* only this edge will need to be repropagated */
		      if (!BBL_IS_FIRST(CFG_EDGE_TAIL(edge)) || (ProcStateJoinSimple(CFG_EDGE_TAIL(edge)->fun->procstate_in,edge->procstate,edge->prop_regs)))
			CfgMarkFun(CFG_EDGE_CFG(edge),CFG_EDGE_TAIL(edge)->fun);
		    }
		}
	    }
	}
      else 
	{
	  if (BBL_PROCSTATE_IN(CFG_EDGE_TAIL(edge))==NULL) { BblPrint(CFG_EDGE_HEAD(edge),0xf);BblPrint(CFG_EDGE_TAIL(edge),0xf); FATAL(("BBL Procstate (not in fixpoint) @E @B=NULL... This probably means you didn't kill the dead edges",edge,CFG_EDGE_TAIL(edge))); }
	 
	 
	  if (ProcStateJoinSimple(BBL_PROCSTATE_IN(CFG_EDGE_TAIL(edge)),state_true,edge->prop_regs))
	    {
	      FunctionMarkBbl(CFG_EDGE_TAIL(edge)->fun,CFG_EDGE_TAIL(edge));
	    }
	}

      if (state_false)
	{
	  /* if there is a state to be propagated over the other
	     outgoing edges, the state over the true edge was freshly
	     allocated in the call to testcondition() and must be
	     free'ed here */
	 
	  if (state_false!=state_true)
	    ProcStateFree(state_true);
	  state_true = NULL;
	}
    }

  if (state_false)
    {
      /*
	The state to be sent over the next edge going out of a block
	is the false_state */
      *state=state_false;
      return TRUE;
    }
  else
    {
      /* 
	 if state_false is null after the call to testcondition(), it
	 was determined that no more states need to propagated over
	 outgoing edges of the basic block */

      return FALSE;
    }
}
/*}}}*/

/* Propagate BBL CS {{{ */

/* 
   Propagate a state over a bbl and to the successor edges

   TODO: link edges !!!

*/

void BblPropagateConstants(t_object* obj, t_ConstantPropagationInsEmul propagator,t_bbl * bbl, t_cfg_edge * edge_out, t_bool during_fixpoint_calculations, t_bool propagate_over_all_outgoing_edges)
{

  /* iterators */

  t_ins * i_ins;
  t_ins * last_ins;
  t_cfg_edge * i_edge;
#ifdef CONST_PROP_CONDITIONAL_BRANCHES_WITH_SIDE_EFFECT_OPT
  t_bool special_last_ins;
#endif
  /* local variables */

  t_bool prop_further = TRUE;

  t_regset allregs = obj->description->all_registers;

  if (!RegsetIsMutualExclusive(((t_procstate*)BBL_PROCSTATE_IN(bbl))->top,allregs))
    {
      return;
    }

  ProcStateDup(temp_procstate,BBL_PROCSTATE_IN(bbl));
  
#ifdef CONST_PROP_CONDITIONAL_BRANCHES_WITH_SIDE_EFFECT_OPT
  last_ins = BBL_INS_LAST(bbl);
  if (last_ins)
    special_last_ins = obj->actions->IsControlflow(last_ins) 
      && !RegsetIsEmpty(INS_REGS_DEF(last_ins)) 
      && INS_ATTRIB(last_ins) & IF_CONDITIONAL 
      && INS_TYPE(last_ins)!=IT_BRANCH;
  else 
    special_last_ins = FALSE;
#endif

  BBL_FOREACH_INS(bbl,i_ins) 
    {
#ifdef CONST_PROP_CONDITIONAL_BRANCHES_WITH_SIDE_EFFECT_OPT
      if (last_ins==i_ins && special_last_ins)
	{
	  ProcStateDup(temp_procstate2,temp_procstate);
	}
#endif
      if ((INS_ATTRIB(i_ins) & IF_FAST_CP_EVAL) && !RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(i_ins),temp_procstate->bot)))
	{
	  ProcStateSetAllBot(temp_procstate,INS_REGS_DEF(i_ins));
	  INS_ATTRIB(i_ins) |= IF_EXECED;
	  INS_ATTRIB(i_ins) &= ~IF_ALWAYS_EXECED;
	}
      else
	propagator(i_ins,temp_procstate);
    }


  /* first ordinary edges are looked at */
#ifdef CONST_PROP_CONDITIONAL_BRANCHES_WITH_SIDE_EFFECT_OPT
  if (!special_last_ins)
#endif
    {
      BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	{
	  if (EDGE_CAT(i_edge)==ET_FUNLINK) continue;
	  /*	  if (EDGE_CAT(i_edge)==ET_IPFUNLINK) FATAL(("IP FUNLINKS SHOULD NO LONGER EXIST"));*/

	  if (prop_further && temp_procstate) 
	    {
	      {
		prop_further = PropOverEdge(i_edge, &temp_procstate, (!CfgEdgeIsInterproc(i_edge) || CfgEdgeIsForwardInterproc(i_edge) || i_edge == edge_out || edge_out == NULL), during_fixpoint_calculations);
	      }
	    }
	}
    }
#ifdef CONST_PROP_CONDITIONAL_BRANCHES_WITH_SIDE_EFFECT_OPT
  else
    {

      /* special case for architectures like the arm: when an LDM
	 {r1,r2,r3,r4,...,pc} instruction is conditional, the old
	 values of the registers should be propagated over the
	 fallthrough edge */

      BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	{
	  if (prop_further && temp_procstate) 
	    {
	      if (EDGE_CAT(i_edge)==ET_FUNLINK) continue;
	      /*	      if (EDGE_CAT(i_edge)==ET_IPFUNLINK) FATAL(("IP FUNLINKS SHOULD NO LONGER EXIST"));*/
	      if (EDGE_CAT(i_edge)==ET_FALLTHROUGH)
		{
		  PropOverEdge(i_edge, &temp_procstate2, (!CfgEdgeIsInterproc(i_edge) || CfgEdgeIsForwardInterproc(i_edge) || i_edge == edge_out || edge_out == NULL), during_fixpoint_calculations);
		  prop_further = FALSE;
		}
	      else
		{
		  prop_further = PropOverEdge(i_edge, &temp_procstate, (!CfgEdgeIsInterproc(i_edge) || CfgEdgeIsForwardInterproc(i_edge) || i_edge == edge_out || edge_out == NULL), during_fixpoint_calculations);
		}
	    }
	}
    }
#endif
  /* then fall-through edges, as they are the edges that are taken in case a jump is not taken */
  /*  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
      {
      if (prop_further && temp_procstate)
      {
      tel1++;
      if (CfgEdgeIsFallThrough(i_edge))
      {
      tel2++;
      prop_further = PropOverEdge(i_edge, &temp_procstate, (!CfgEdgeIsInterproc(i_edge) || CfgEdgeIsForwardInterproc(i_edge) || i_edge == edge_out || edge_out == NULL), during_fixpoint_calculations);
      }
      }
      }
      if ((tel2%10000)==0)
      printf("%d/%d\n",tel2,tel1);
  */
}
/*}}}*/
/* Propagate BBL CI {{{ */
void BblCIPropagateConstants(t_object* obj, t_ConstantPropagationInsEmul propagator,t_bbl * bbl, t_bool during_fixpoint_calculations)
{

  /* iterators */

  t_ins * i_ins;
  t_cfg_edge * i_edge;
  t_procstate * procstate_out;
  /* local variables */

  t_bool prop_further;
  t_regset allregs;

  exit(-1);
  /* local variables */

  prop_further = TRUE;
  allregs = obj->description->all_registers;
  

  if (!RegsetIsMutualExclusive(((t_procstate*)BBL_PROCSTATE_IN(bbl))->top,allregs))
    {
    
      return;
    }
  
  procstate_out=ProcStateNewDup(BBL_PROCSTATE_IN(bbl));

  BBL_FOREACH_INS(bbl,i_ins) 
    {
      propagator(i_ins,procstate_out);
    }


  /* first ordinary edges are looked at */
  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
    {      
      if (prop_further && procstate_out)
	{
	  if (!CfgEdgeIsFallThrough(i_edge))
	    {
	      prop_further = CIPropOverEdge(i_edge, &procstate_out, during_fixpoint_calculations);
	    }
	}
    }

  /* then fall-through edges, as they are the edges that are taken in case a jump is not taken */
  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
    {
      if (prop_further && procstate_out)
	{
	  if (CfgEdgeIsFallThrough(i_edge))
	    {
	      prop_further = CIPropOverEdge(i_edge, &procstate_out, during_fixpoint_calculations);
	    }
	}
    }

  if (procstate_out)
    ProcStateFree(procstate_out);

}
/*}}}*/

/* Optimizations using CP information {{{*/
/* Remove unreachable edges {{{ */
static void FlowgraphRemoveUnreachableEdges(t_cfg * cfg)
{
  t_cfg_edge * i_edge, * i_safe;
  t_cfg_edge * corr_edge, * link_edge;

  int dead = 0;

  CFG_FOREACH_EDGE_SAFE(cfg,i_edge,i_safe)
    {
      if (CFG_EDGE_HEAD(i_edge) == cfg->unique_entry_node)
	{
	  continue;
	}

      if (!(EdgeIsMarked(i_edge)))
	{
	  if (EDGE_CAT(i_edge)!=ET_RETURN && EDGE_CAT(i_edge)!=ET_COMPENSATING &&  EDGE_CAT(i_edge)!=ET_FUNLINK  && EDGE_CAT(i_edge)!=ET_IPFUNLINK)
	    {
	      corr_edge = i_edge->corr_edge;

	      if (corr_edge)
		{
		  /* kill the corresponding edge */
		  if (corr_edge == i_safe) 
		    i_safe = T_CFG_EDGE(EDGE_NEXT(i_safe));
		  if (corr_edge->procstate)
		    ProcStateFree(corr_edge->procstate);
		  CfgEdgeKill(corr_edge);
		  dead++;
		}

	      if (EDGE_CAT(i_edge) == ET_CALL)
		{
		  /* kill the funlink as well */
		  link_edge = FindFunLinkForCall(i_edge);
		  if (link_edge)
		    {
		      corr_edge = link_edge->corr_edge;
		      if (corr_edge)
			{
			  /* kill corresponding edge for ipfunlink as well */
			  if (corr_edge == i_safe) 
			    i_safe = T_CFG_EDGE(EDGE_NEXT(i_safe));
			  if (corr_edge->procstate)
			    ProcStateFree(corr_edge->procstate);
			  CfgEdgeKill(corr_edge);
			  dead++;
			}
		      if (link_edge == i_safe) 
			i_safe = T_CFG_EDGE(EDGE_NEXT(i_safe));
		      if (link_edge->procstate)
			ProcStateFree(link_edge->procstate);
		      CfgEdgeKill(link_edge);
		      dead++;
		    }
		}
	
	      /* finally kill the edge itself */
	      if (i_edge->procstate) 
		ProcStateFree(i_edge->procstate);
	      CfgEdgeKill(i_edge);
	      dead++;
	    }
	}
    }
  VERBOSE(0,("%d unreachable edges removed",dead));
}
/*}}} */
/* Unconditionalize {{{ */
static void CfgUnconditionalizeIns(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_function * i_fun;
  t_ins * i_ins;

  int nr_uncond_ins = 0;

  CFG_FOREACH_FUN(cfg,i_fun)
    FUN_FOREACH_BBL(i_fun,i_bbl)
    BBL_FOREACH_INS(i_bbl,i_ins)
    if (INS_ATTRIB(i_ins) & IF_EXECED)
      if (INS_ATTRIB(i_ins) & IF_ALWAYS_EXECED)
	{
	  if (CFG_OBJECT(cfg)->actions->InstructionUnconditionalizer(i_ins))
	    {
	      nr_uncond_ins++;
	    }
	}

  VERBOSE(0,("%d instructions unconditionalized",nr_uncond_ins));

}
/*}}}*/
/* Remove unreachable blocks {{{ */
static void FlowgraphRemoveUnreachableBlocks(t_cfg *cfg)
{
  /* iterators */
  t_bbl * i_bbl;
  t_cfg_edge * i_edge;
  t_ins * i_ins, *i_tmp;

  /* local variables */
  int dead_ins = 0;
  int dead_blocks = 0;
  t_bool has_predecessor;
  int old_deadins = 0;

  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      if (IS_DATABBL(i_bbl)) continue;
      if (i_bbl == cfg->unique_entry_node) continue;

      BblUnmark(i_bbl);
      has_predecessor = FALSE;
      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  has_predecessor = TRUE;
	  if (EdgeIsMarked(i_edge))
	    BblMark(i_bbl);
	}

#ifdef ARM_SUPPORT
      if (BblIsMarked(i_bbl))
	{
	  RenameLocalAddressProducers(i_bbl);
	  continue;
	}
#endif
      if (has_predecessor)
	{
	  dead_blocks++;
	}

      old_deadins = dead_ins;

      BBL_FOREACH_INS_SAFE(i_bbl,i_ins,i_tmp)
	{
	  if (!(INS_ATTRIB(i_ins) & IF_EXECED))
	    {
	      if (has_predecessor)
		{
		  dead_ins++;
		}
	      InsKill(i_ins);
	    }
	}

      /* adjust n_unevaluated_ins for every incoming edge for the function containing this block */
      FunctionDecrementUnexecedInsBy(BBL_FUN(i_bbl),dead_ins-old_deadins);
    }
  VERBOSE(0,("%d unreachable instructions in %d unreachable blocks killed",dead_ins,dead_blocks));
}
/*}}}*/
/* RemoveNotExecutedIns {{{ */
static void CfgRemoveNotExecutedIns(t_cfg *cfg, t_reloc_table * table)
{

  /* iterators */

  t_bbl * i_bbl;
  t_cfg_edge * i_edge;
  t_ins * i_ins, * i_tmp;

  /* local variables */

  int dead_ins = 0;
  int old_deadins = 0;
  t_bool has_predecessor;

  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      if (IS_DATABBL(i_bbl)) continue;
      BblUnmark(i_bbl);
      has_predecessor = FALSE;
      BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	{
	  has_predecessor = TRUE;
	  if (EdgeIsMarked(i_edge))
	    BblMark(i_bbl);
	}
      if (!BblIsMarked(i_bbl)) continue;

      old_deadins = dead_ins;
      BBL_FOREACH_INS_SAFE(i_bbl,i_ins,i_tmp)
	{
	  if (!(INS_ATTRIB(i_ins) & IF_EXECED))
	    {
	      InsKill(i_ins);
	      if (has_predecessor)
		dead_ins++;
	    }
	}

      /* adjust n_unevaluated_ins for every incoming edge for the function containing this block */
      FunctionDecrementUnexecedInsBy(BBL_FUN(i_bbl),dead_ins-old_deadins);
    }
  VERBOSE(0,("%d unexecuted instructions killed",dead_ins));
}
/*}}} */
/*}}}*/
/* ApplyConstantInformation {{{ */
static void CfgApplyConstantInformation(t_object * obj, t_section * sec)
{
  /* iterators */
  
  t_function * i_fun;
  t_bbl * i_bbl;
  t_ins * i_ins, * i_tmp;
  
  /* local variables */
  
  t_bbl * program_entry;
  t_function * entry_fun;
  t_cfg * cfg;

  t_procstate * prev_state;
  t_procstate * next_state;
  
  t_regset allregs = obj->description->all_registers;
  
  cfg = T_CFG(sec->data);
  
  prev_state = ProcStateNew();
  next_state = ProcStateNew();
  
  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);
  
  /* first of all, if the instructions in the graph already contain regstates, remove them */
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  /*	  if (INS_USED_REGSTATE(i_ins)) Free(INS_USED_REGSTATE(i_ins));*/
	  /*	  if (INS_DEFINED_REGSTATE(i_ins)) Free(INS_DEFINED_REGSTATE(i_ins));*/
	  /*	  INS_USED_REGSTATE(i_ins) = INS_DEFINED_REGSTATE(i_ins) = NULL;*/
	}
    }
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionUnmarkAllBbls(i_fun);
    }
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FunctionPropagateConstants(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE,FALSE);
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      if (!RegsetIsMutualExclusive(((t_procstate*)BBL_PROCSTATE_IN(i_bbl))->top,allregs))
		continue;

	      ProcStateDup(next_state,BBL_PROCSTATE_IN(i_bbl));
	      ProcStateDup(prev_state,BBL_PROCSTATE_IN(i_bbl));
	      
	      /* safe needed because useless instruction can be killed in Optimizer */
	      
	      BBL_FOREACH_INS_SAFE(i_bbl,i_ins,i_tmp)
		{
		  if ((INS_ATTRIB(i_ins) & IF_FAST_CP_EVAL) && !RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(i_ins),temp_procstate->bot)))
		    {
		      ProcStateSetAllBot(next_state,INS_REGS_DEF(i_ins));
		    }
		  else
		    {
		      CFG_OBJECT(cfg)->actions->InstructionEmulator(i_ins,next_state);
		    }

		  if (CFG_OBJECT(cfg)->actions->InstructionConstantOptimizer(i_ins,prev_state,next_state))
		    {
		      /* save the states of the used and defined registers in the instruction */
		      /*INS_USED_REGSTATE(i_ins) = RegStateArrayCreateFromProcState(prev_state,INS_REGS_USE(i_ins), CFG_OBJECT(cfg)->description);*/
		      /*INS_DEFINED_REGSTATE(i_ins) = RegStateArrayCreateFromProcState(next_state,INS_REGS_DEF(i_ins), CFG_OBJECT(cfg)->description);*/
		    }
		  ProcStateDup(prev_state,next_state);
		}
	    }
	  
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
	      BBL_PROCSTATE_IN(i_bbl) = NULL;
	    }
	}
    }
  
  ProcStateFree(prev_state);
  ProcStateFree(next_state);
}

#ifdef DIETER
/* For ARM only */
void CfgCollectAddressProducerInformation(t_cfg * cfg)
{
  t_object * obj = CFG_OBJECT(cfg);
  t_section * sec = CFG_SECTION(cfg);

  /* iterators */
  
  t_function * i_fun;
  t_bbl * i_bbl;
  t_ins * i_ins;
  
  /* local variables */
  
  t_bbl * program_entry;
  t_function * entry_fun;

  t_procstate * prev_state;
  t_procstate * next_state;
  
  t_regset allregs = obj->description->all_registers;
  
  prev_state = ProcStateNew();
  next_state = ProcStateNew();
  
  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);
  
  /* first of all, if the instructions in the graph already contain regstates, remove them */
/*    CFG_FOREACH_BBL(cfg,i_bbl) */
/*      { */
/*        BBL_FOREACH_INS(i_bbl,i_ins) */
/*  	{ */
/*  	  if (INS_USED_REGSTATE(i_ins)) Free(INS_USED_REGSTATE(i_ins)); */
/*  	  if (INS_DEFINED_REGSTATE(i_ins)) Free(INS_DEFINED_REGSTATE(i_ins)); */
/*  	  INS_USED_REGSTATE(i_ins) = INS_DEFINED_REGSTATE(i_ins) = NULL; */
/*  	} */
/*      } */

  /* Remove hotnessinformation from previous optimization round. */
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  if(INS_REFERS_TO(i_ins))
	    {
	      INS_REFERS_TO(i_ins)->rel->hotness = 0;
	    }
	}
    }
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionUnmarkAllBbls(i_fun);
    }
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FunctionPropagateConstants(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE);
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      if (!RegsetIsMutualExclusive(((t_procstate*)BBL_PROCSTATE_IN(i_bbl))->top,allregs))
		continue;

	      ProcStateDup(next_state,BBL_PROCSTATE_IN(i_bbl));
	      ProcStateDup(prev_state,BBL_PROCSTATE_IN(i_bbl));
	      
	      BBL_FOREACH_INS(i_bbl,i_ins)
	      {

		if ((INS_ATTRIB(i_ins) & IF_FAST_CP_EVAL) && !RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(i_ins),temp_procstate->bot)))
		  {
		    ProcStateSetAllBot(next_state,INS_REGS_DEF(i_ins));
		  }
		else
		  {
		    CFG_OBJECT(cfg)->actions->InstructionEmulator(i_ins,next_state);
		  }

		/* Extract the info */
#ifdef ARM_SUPPORT
		ArmAddressAnalyser(i_ins, prev_state, next_state);
#endif
		ProcStateDup(prev_state,next_state);
	      }
	    }
	  
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
	      BBL_PROCSTATE_IN(i_bbl) = NULL;
	    }
	}
    }
  
  ProcStateFree(prev_state);
  ProcStateFree(next_state);
}

void FreeSectionConnections(t_section * sec)
{
  t_section_conn * next_conn, * conn;

  next_conn = sec->section_connections;
  while(next_conn)
    {
      conn = next_conn;
      next_conn = next_conn->next;
      SectionConnFree(conn);
    }
  
  sec->section_connections = NULL;
}

void FreeAddressProducerInformation(t_cfg * cfg)
{
  t_object * obj = CFG_OBJECT(cfg);
  t_uint32 i;
  
  printf("Freeing all section-connections.\n");
/*    for(i = 0 ; i < obj->ncodes; i++) */
/*      FreeSectionConnections(obj->code[i]); */
  
  for(i = 0 ; i < obj->ndatas; i++)
    FreeSectionConnections(obj->data[i]);

  for(i = 0 ; i < obj->nrodatas; i++)
    FreeSectionConnections(obj->rodata[i]);

  for(i = 0 ; i < obj->nbsss; i++)
    FreeSectionConnections(obj->bss[i]);

}

/* For ARM only */
void CfgOptimizeLoaders(t_cfg * cfg)
{
  t_object * obj = CFG_OBJECT(cfg);
  t_section * sec = CFG_SECTION(cfg);

  /* iterators */
  
  t_function * i_fun;
  t_bbl * i_bbl;
  t_ins * i_ins, * i_tmp;
  
  /* local variables */
  
  t_bbl * program_entry;
  t_function * entry_fun;

  t_procstate * prev_state;
  t_procstate * next_state;
  
  t_regset allregs = obj->description->all_registers;
  
  prev_state = ProcStateNew();
  next_state = ProcStateNew();
  
  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);
  
  /* first of all, if the instructions in the graph already contain regstates, remove them */
  /*    CFG_FOREACH_BBL(cfg,i_bbl) */
  /*      { */
  /*        BBL_FOREACH_INS(i_bbl,i_ins) */
  /*  	{ */
  /*  	  if (INS_USED_REGSTATE(i_ins)) Free(INS_USED_REGSTATE(i_ins)); */
  /*  	  if (INS_DEFINED_REGSTATE(i_ins)) Free(INS_DEFINED_REGSTATE(i_ins)); */
  /*  	  INS_USED_REGSTATE(i_ins) = INS_DEFINED_REGSTATE(i_ins) = NULL; */
  /*  	} */
  /*      } */
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionUnmarkAllBbls(i_fun);
    }
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FunctionPropagateConstants(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE,FALSE);
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      if (!RegsetIsMutualExclusive(((t_procstate*)BBL_PROCSTATE_IN(i_bbl))->top,allregs))
		continue;

	      ProcStateDup(next_state,BBL_PROCSTATE_IN(i_bbl));
	      ProcStateDup(prev_state,BBL_PROCSTATE_IN(i_bbl));

	      /* safe needed because useless instruction can be killed in Optimizer */
	      
	      BBL_FOREACH_INS_SAFE(i_bbl,i_ins,i_tmp)
	      {

		if ((INS_ATTRIB(i_ins) & IF_FAST_CP_EVAL) && !RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(i_ins),temp_procstate->bot)))
		  {
		    ProcStateSetAllBot(next_state,INS_REGS_DEF(i_ins));
		  }
		else
		  {
		    CFG_OBJECT(cfg)->actions->InstructionEmulator(i_ins,next_state);
		  }

		  /* Extract the info */
#ifdef ARM_SUPPORT
		ArmOptMemoryAccess(i_ins, prev_state, next_state);
#endif
		ProcStateDup(prev_state,next_state);
	      }
	    }
	  
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
	      BBL_PROCSTATE_IN(i_bbl) = NULL;
	    }
	}
    }
  
  ProcStateFree(prev_state);
  ProcStateFree(next_state);
}
#endif

/*}}} */


void CfgOptimizeAddressProducers(t_object * obj, t_section * sec)
{
  /* iterators */
  
  t_function * i_fun;
  t_bbl * i_bbl;
  t_ins * i_ins, * i_tmp;
  
  /* local variables */
  
  t_bbl * program_entry;
  t_function * entry_fun;
  t_cfg * cfg;

  t_procstate * prev_state;
  t_procstate * next_state;
  
  t_regset allregs = obj->description->all_registers;
  
  cfg = T_CFG(sec->data);
  
  prev_state = ProcStateNew();
  next_state = ProcStateNew();
  
  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);
  
  /* first of all, if the instructions in the graph already contain regstates, remove them */
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  /*	  if (INS_USED_REGSTATE(i_ins)) Free(INS_USED_REGSTATE(i_ins));*/
	  /*	  if (INS_DEFINED_REGSTATE(i_ins)) Free(INS_DEFINED_REGSTATE(i_ins));*/
	  /*	  INS_USED_REGSTATE(i_ins) = INS_DEFINED_REGSTATE(i_ins) = NULL;*/
	}
    }
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionUnmarkAllBbls(i_fun);
    }
  
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FunctionPropagateConstants(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE,FALSE);
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      if (!RegsetIsMutualExclusive(((t_procstate*)BBL_PROCSTATE_IN(i_bbl))->top,allregs))
		continue;

	      ProcStateDup(next_state,BBL_PROCSTATE_IN(i_bbl));
	      ProcStateDup(prev_state,BBL_PROCSTATE_IN(i_bbl));
	      
	      /* safe needed because useless instruction can be killed in Optimizer */
	      
	      BBL_FOREACH_INS_SAFE(i_bbl,i_ins,i_tmp)
		{
		  if ((INS_ATTRIB(i_ins) & IF_FAST_CP_EVAL) && !RegsetIsEmpty(RegsetNewIntersect(INS_REGS_USE(i_ins),temp_procstate->bot)))
		    ProcStateSetAllBot(next_state,INS_REGS_DEF(i_ins));
		  else
		    CFG_OBJECT(cfg)->actions->InstructionEmulator(i_ins,next_state);
#if 0		
		  CFG_OBJECT(cfg)->actions->ModelPotentialAddressOptimization(i_ins,prev_state,next_state);
#endif
		  ProcStateDup(prev_state,next_state);
		}
	    }
	  
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
	      BBL_PROCSTATE_IN(i_bbl) = NULL;
	    }
	}
    }

#if 0
  DetermineOrder(obj);
#endif

  ProcStateFree(prev_state);
  ProcStateFree(next_state);
}


/** Helper function that decrements adjusts the number of unexecuted
 * instructions in a function on all call sites
 * (FlowgraphRemoveUnreachableBlocks and CfgRemoveNotExecutedIns change this
 * number because they kill a number of these instructions */

static void FunctionDecrementUnexecedInsBy(t_function * fun, t_uint32 diff)
{
  t_bbl * i_bbl;
  t_cfg_edge * i_edge;

  if (diff == 0) return;

  FUN_FOREACH_BBL(fun,i_bbl)
    BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
    if (CfgEdgeIsForwardInterproc(i_edge))
      i_edge->n_unevaluated_ins -= diff;
}

/* Hieronder: tijdelijke functietjes voor Dominique, op zoek naar optimalisatiemogelijkheden adhv consprop en udanalyse {{{*/


#ifdef ARM_SUPPORT
#include "../arch/arm/arm_instructions.h"
t_string ArmRegisterName(t_reg reg);

static void ProcStateAfter(t_object * obj, t_ins * ins, t_procstate * out)
{
  t_bbl * bbl = INS_BBL(ins);
  t_ins * i_ins;

  ProcStateDup(out,BBL_PROCSTATE_IN(bbl));
  BBL_FOREACH_INS(bbl,i_ins)
    {
      obj->actions->InstructionEmulator(i_ins,out);
      if (i_ins == ins)
	break;
    }

  VERIFY_FASSERT(i_ins,("Instruction @I thinks it's in BBL @B, but isn't really there!",ins,bbl));
  
}
void UdLookForExtraConstants(t_section *sec,const char * fname)
{
  /* local variables */
  t_cfg * cfg = SECTION_CFG(sec);
  t_ud_graph * udg = CFG_UDGRAPH(cfg);
  t_object * obj = sec->obj;
  t_bbl * program_entry;
  t_function * entry_fun;
  FILE * ofile;
  t_regset allregs = obj->description->all_registers;


  /* iterators */
  t_function * i_fun;
  t_bbl * i_bbl;
  t_ins * i_ins;

  /* open output file */
  ofile = fopen(fname,"w");
  VERIFY_FASSERT(ofile,("Could not open file %s for writing.",fname));

  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);

  /* step 1: Propagate the constant information to each basic block */
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FunctionPropagateConstants(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE,FALSE);
	}
    }

  /* step 2a: remove unnecessary links from the ud-chain:
   * if a bbl has TOP at it's entry, it was never propagated through
   * so the code can never be executed -> the instructions here
   * shouldn't be considered as defining any registers */
  CFG_FOREACH_BBL(cfg,i_bbl)
    {
      if (IS_DATABBL(i_bbl)) continue; 
      if (RegsetIsMutualExclusive(BBL_PROCSTATE_IN(i_bbl)->top,allregs)) continue;

      /* Remove the links to and from each instruction of this bbl in the ud-graph */
      DiabloPrint(ofile,"Found TOP: bbl @B\n",i_bbl);
      BBL_FOREACH_INS(i_bbl,i_ins)
	{
	  t_edge * edge, * tmp;
	  NODE_FOREACH_SUCC_EDGE_SAFE(i_ins,edge,tmp)
	    GraphRemoveEdge(udg,edge);
	  NODE_FOREACH_PRED_EDGE_SAFE(i_ins,edge,tmp)
	    GraphRemoveEdge(udg,edge);
	}
    }
  
  /* step 2b: Look for interesting optimization opportunities */
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      t_reg r;
	      t_bool block_is_top = FALSE;
	      t_procstate * pstate = ProcStateNew();
	      ProcStateDup(pstate,BBL_PROCSTATE_IN(i_bbl));

	      REGSET_FOREACH_REG(allregs,r)
		{
		  if (r<18)
		    {
		      register_content dummy;
		      if (ProcStateGetReg(pstate,r, &dummy) == TOP)
			block_is_top = TRUE;
		    }
		  else
		    {
		      t_bool dummy;
		      if (ProcStateGetCond(pstate,r,&dummy) == TOP)
			block_is_top = TRUE;
		    }
		}

	      if (block_is_top) /* skip block if TOP at entry */
		continue;


	      BBL_FOREACH_INS(i_bbl,i_ins)
		{
		  t_reg r;
		  t_bool output_is_constant = TRUE;
		  t_bool input_is_constant = TRUE;

		  t_edge * i_edge;
		  t_procstate * pstmp = ProcStateNew();

		  int i = 0;
	  
		  /* check if output of instruction is constant */
		  obj->actions->InstructionEmulator(i_ins,pstate);
		  REGSET_FOREACH_REG(INS_REGS_DEF(i_ins),r)
		    {
	    
		      /* ARM-specifiek!!!!!!!!! */
		      if (r<18)
			{
			  register_content dummy;
			  if (ProcStateGetReg(pstate,r,&dummy) != VALUE)
			    output_is_constant = FALSE;
			} 
		      else
			{
			  t_bool dummy;
			  if (ProcStateGetCond(pstate,r,&dummy) != VALUE)
			    output_is_constant = FALSE;
			}
		    }

		  /* only non-constant outputs are potentially interesting */
		  if (output_is_constant)
		    continue;

		  /* now check if all incoming used registers are constant */
		  NODE_FOREACH_SUCC_EDGE(i_ins,i_edge)
		    {
		      i++;
		      ProcStateAfter(obj,T_INS(EDGE_TAIL(i_edge)),pstmp);

		      /* ARM-specifiek!!!!!!!!! */
		      if (EDGE_CAT(i_edge) < 18)
			{
			  register_content dummy;
			  if (ProcStateGetReg(pstmp,EDGE_CAT(i_edge), &dummy) != VALUE)
			    input_is_constant = FALSE;
			}
		      else
			{
			  t_bool dummy;
			  if (ProcStateGetCond(pstmp,EDGE_CAT(i_edge), &dummy) != VALUE)
			    input_is_constant = FALSE;
			}
		    }

		  /* no definers? skip it... */
		  if (i==0)
		    {
		      continue;
		    }

		  /* only constant inputs are potentially interesting */
		  if (!input_is_constant)
		    continue;
		  /* skip instructions that don't use any registers */
		  if (RegsetIsEmpty(INS_REGS_USE(i_ins)))
		    continue;
		  /* load instructions are not interesting: they mostly read from non-constant memory */
		  if (INS_TYPE(i_ins) == IT_LOAD)
		    continue;
	  
		  /* print out the information that was gathered */
		  DiabloPrint(ofile,"Instruction @I has non-constant output\n",i_ins);
		  DiabloPrint(ofile,"Defined regs are: ");
		  REGSET_FOREACH_REG(INS_REGS_DEF(i_ins),r)
		    DiabloPrint(ofile,"%s ",ArmRegisterName(r));
		  DiabloPrint(ofile,"\n");
	    
		  DiabloPrint(ofile,"Definers are:\n");
		  REGSET_FOREACH_REG(INS_REGS_USE(i_ins),r)
		    {
		      DiabloPrint(ofile,"For reg[%d]:\n",r);
		      NODE_FOREACH_SUCC_EDGE(i_ins,i_edge)
			{
			  if (EDGE_CAT(i_edge) != r) continue;
			  ProcStateAfter(obj,T_INS(EDGE_TAIL(i_edge)),pstmp);
			  /* ARM-specifiek!!!!!!!!!!!!!!!!!!! */
			  if (r<18)
			    {
			      register_content rc;
			      ProcStateGetReg(pstmp,r,&rc);
			      DiabloPrint(ofile,"\t[@G] @I\n",rc.i,T_INS(EDGE_TAIL(i_edge)));
			    }
			  else
			    {
			      t_bool val;
			      ProcStateGetCond(pstmp,r,&val);
			      DiabloPrint(ofile,"\t[%d] @I\n",val,T_INS(EDGE_TAIL(i_edge)));
			    }
			}
		    }
		  if ((ARMINSOPCODE(i_ins)==ARM_MOV) || (ARMINSOPCODE(i_ins)==ARM_MVN))
		    {
		      /* also print out all successors */
		      DiabloPrint(ofile,"Users of this information are:\n");
		      REGSET_FOREACH_REG(INS_REGS_DEF(i_ins),r)
			{
			  DiabloPrint(ofile,"-> reg[%d]:\n",r);
			  NODE_FOREACH_PRED_EDGE(i_ins,i_edge)
			    {
			      int defcount=0;
			      t_edge * i_edge2;
		
			      if (EDGE_CAT(i_edge)!=r) continue;
		
			      NODE_FOREACH_SUCC_EDGE(EDGE_HEAD(i_edge),i_edge2)
				{
				  if (EDGE_CAT(i_edge2) == r)
				    defcount++;
				}
			      DiabloPrint(ofile,"\t[%d definers] @I\n",defcount,EDGE_HEAD(i_edge));
			      if ((defcount == 1) && (ARMINSOPCODE(i_ins)==ARM_MOV))
				{
				  NODE_FOREACH_PRED_EDGE(EDGE_HEAD(i_edge),i_edge2)
				    {
				      int defcount2=0;
				      t_edge * i_edge3;
				      NODE_FOREACH_SUCC_EDGE(EDGE_HEAD(i_edge2),i_edge3)
					{
					  if (EDGE_CAT(i_edge3) == ARMINSREGA(EDGE_HEAD(i_edge2)))
					    defcount2++;
					}

				      DiabloPrint(ofile,"\t\t-> [%d definers] @I\n",defcount2,EDGE_HEAD(i_edge2));
				    }
				}
			    }
			}
		    }
		  DiabloPrint(ofile,"\n\n");

		  ProcStateFree(pstmp);
		}
	      ProcStateFree(pstate);
	    }
	}
    }

  /* step 3: Clean up the information from step 1 */
  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      ProcStateFree(BBL_PROCSTATE_IN(i_bbl));
	      BBL_PROCSTATE_IN(i_bbl) = NULL;
	    }
	}
    }

  fclose(ofile);

}
#endif
/*}}} */
/* Dump {{{ */
void DumpConstantPropagationInfo(t_string filename,t_object * obj)
{
  ConstantPropagation(obj->code[0]->data, filename, FALSE);
}

/* static helper function that actually dumps the constant propagation info in the file */
static void DumpCPInfo(t_object * obj,FILE * f)
{
  /* iterators */
  t_function * i_fun;
  t_bbl * i_bbl;
  t_ins * i_ins;

  /* local variables */
  t_bbl * program_entry;
  t_function * entry_fun;
  t_section * sec;
  t_cfg * cfg;

  t_procstate * state;

  sec = obj->code[0]; 
  cfg = T_CFG(sec->data);

  state = ProcStateNew();

  program_entry = cfg->entries->entry_bbl;
  entry_fun     = BBL_FUN(program_entry);

  CFG_FOREACH_FUN(cfg,i_fun)
    {
      FunctionUnmarkAllBbls(i_fun);
    }

  CFG_FOREACH_FUN(cfg,i_fun)
    {
      if (i_fun->bbl_first)
	{
	  FunctionPropagateConstants(obj,i_fun,sec,T_CFG_EDGE(i_fun!=entry_fun),FALSE,FALSE);

	  FUN_FOREACH_BBL(i_fun,i_bbl)
	    {
	      /* start new node */
	      fprintf(f,">>> %p\n",i_bbl);

	      ProcStateDup(state,BBL_PROCSTATE_IN(i_bbl));
	      ProcStatePrint(f, obj->description, state);

	      BBL_FOREACH_INS(i_bbl,i_ins)
		{
		  obj->actions->InstructionEmulator(i_ins,state);
		  ProcStatePrint(f,obj->description,state);
		}
	    }
	}
    }
  Free(state);
}
/*}}}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
