/*
 *{{{ Copyright 2001,2002,2003:
 *			Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *************************************************************}}}*/

#ifndef DIABLO_DESCRIPTION_TYPEDEFS 
#define DIABLO_DESCRIPTION_TYPEDEFS
typedef enum _t_address_type
{
  ADDRSIZE24,
  ADDRSIZE32,
  ADDRSIZE64,
  ADDRSIZENONE
} t_address_type;
typedef struct _t_arch_specific_optimization t_arch_specific_optimization;
typedef struct _t_architecture_description t_architecture_description;
#endif

#ifndef DIABLO_DESCRIPTION_H
#define DIABLO_DESCRIPTION_H
#include "../simple_types.h"
#include "diablo_regset.h"
#include "diablo_ins.h"
#include "diablo_flowgraph.h"
#ifdef INSTRUMENTATION
#include "../atom_internal.h"
#endif

/*! Struct that describes an architecture specific optimization. All
 * arch_specific_optimizations get only one parameter (a cfg). They can be
 * called by using ArchCall */
struct _t_arch_specific_optimization
{
  char * name;
  void (*callback)(t_cfg *);
}; 



/*! A description of the architecture */
struct _t_architecture_description
{
   t_uint32 minimal_encoded_instruction_size;
   t_uint32 maximum_encoded_instruction_size;
   t_uint32 encoded_instruction_mod_size; /* Normal this is equal to minimal */
   t_uint32 bundle_size;
   t_uint32 template_size;
   t_uint32 decoded_instruction_size;   
   t_address_type address_size;
   t_uint32 num_int_regs;
   t_uint32 num_float_regs;
   t_uint32 num_predicate_regs;
   t_uint32 num_branch_regs;
   t_uint32 num_special_regs;
   char ** register_names;
   t_regset all_registers;
   t_regset int_registers;
   t_regset flt_registers;
   t_regset cond_registers;
   t_regset callee_saved;
   t_regset callee_may_use;
   t_regset callee_may_change;
   t_regset callee_may_return;
   t_regset always_live;
   t_regset registers_prop_over_hell;
   t_regset const_registers;
   t_regset dead_over_call;	/* Registers that must not be propagated over call, swi and return edges */
   t_reg program_counter;   /* set to REG_NONE if there is no visible program counter */
   const t_arch_specific_optimization * opts;
#ifdef INSTRUMENTATION
   t_atom_actions * atom_actions;
#endif
   char ** unbehaved_funs;
  t_ins* (*addJumpInstruction)(t_bbl *);
  t_bool (*isProcedureCall)(t_ins *);
  t_regset link_register;
  void (*computeLiveRegsBeforeSwi)(t_regset *, t_ins *);
};
void ArchCall(char * name, t_cfg * cfg);
#define TotalRegs(x)	((x)->num_int_regs + (x)->num_float_regs + (x)->num_predicate_regs + (x)->num_branch_regs + (x)->num_special_regs)
#endif
/* vim: set shiftwidth=2 foldmethod=marker: */
