/**********************************************************
 * FILE: diablo_gcc_maps.l
 *
 * DESC: A simple lexer for GCC maps... 
 * Currently tested only for binutils 2.12.1
 * Very simple, but does the trick...
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) 
 ***********************************************************/

%{
#include "diablo_gcc_maps_parser.h"
#include <strings.h>
#include <string.h>
#include "../datastructures/data_string.h"
#include <stdio.h>
#define YY_NO_UNPUT
long line=1;
static int ParseMapGCCwrap(void) { return 1; }
static int open=0; 
%}
Name 		[0-9a-zA-Z_/\-\.()@]*

HexNumber	"0"[xX][0-9a-fA-F]+
OctNumber	"0"[0-7]*
DecNumber	[1-9][0-9]*

FltExponent   	[eE][-+]?[0-9]+
FltFract	([0-9]*"."[0-9]+)|([0-9]+".")
FltNumber	(({FltFract}{FltExponent}?)|([0-9]+{FltExponent}))[FfLl]?
Char            '(\\.|[^\\'\n])+'	
Regexp		[^ \n)(]*
%x PATTERN 
%x MEMBERS 
%x COMMONS 
%x MEMORY 
%x MAP
%x SKIP
%%
"Archive"[ ]*"member"[ ]*"included"[ ]*"because"[ ]*"of"[ ]*"file"[ ]*"("[ ]*"symbol"[ ]*")" { BEGIN(MEMBERS); }
"Memory"[ ]*"Configuration"[ ]* 		{ BEGIN(MEMORY); }
<MEMBERS>"Allocating"[ ]*"common"[ ]*"symbols" 	{ BEGIN(COMMONS); }
<MEMBERS>"Memory"[ ]*"Configuration"[ ]* 	{ BEGIN(MEMORY); }
<MEMBERS>.					{ }
<MEMBERS>\n					{ line++; }
<COMMONS>"Memory"[ ]*"Configuration"[ ]* 	{ BEGIN(MEMORY); }
<COMMONS>.					{ }
<COMMONS>\n					{ line++; }
<MEMORY>"Linker"[ ]*"script"[ ]*"and"[ ]*"memory"[ ]*"map"[ ]* { BEGIN(MAP); }
<MEMORY>.					{ }
<MEMORY>\n					{ line++; }
<MAP>^"OUTPUT".*				{ }
<MAP>^"LOAD"					{ return T_LOAD; }
<PATTERN>"("					{ open++; }
<PATTERN>")"					{ if (open==0) { BEGIN(MAP); return T_PATTERN;} else open--; }
<PATTERN>{Regexp}				{ }
<PATTERN>" "					{ } 
<MAP>^.*("*fill*"|"ALIGN"|"PROVIDE"|"CTORS"|"= ."|"LONG"|"START GROUP"|"END GROUP"|"__DYNAMIC"|". ="|"__global"|". +"|".="|"=."|"size before relaxing"|"EXCLUDE_FILE").*$ {}
<MAP>"*"[^(\n]*"("				{ open=0;  BEGIN(PATTERN);}
<SKIP>[\n]					{ BEGIN(MAP); }
<SKIP>.						{ }

<MAP>{HexNumber}				{ long long tmp; sscanf(ParseMapGCCtext,"0x%llx",&tmp);  ParseMapGCClval.ll.num1=tmp&((1LL<<32)-1); ParseMapGCClval.ll.num2=tmp>>31;  return T_HEX; }
<MAP>{OctNumber}				{ }
<MAP>{DecNumber}				{ }
<MAP>^{Name}					{ ParseMapGCClval.string=StringDup(ParseMapGCCtext); return T_LEADING_STRING; }
<MAP>{Name}					{ ParseMapGCClval.string=StringDup(ParseMapGCCtext); return T_STRING; }
<MAP>[ ]					{ }
<MAP>[\n]					{ line++; }
<MAP>"="					{ return T_IS; }
<MAP>","					{ }
<MAP>"&"					{ }
<MAP>"+"					{ }
<MAP>.						{printf("Unmatched %s\n",ParseMapGCCtext); exit(0);}
.						{ }
<MAP><<EOF>>					{ BEGIN(0);}
[\n]						{ line++; }
%%
