/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * A simple parser for GCC maps... 
 *  Currently tested only for binutils 2.12.1
 *  Very simple, but does the trick...
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) 
 ************************************************************/


%token T_LOAD 
%token T_STRING
%token T_HEX
%token T_PATTERN
%token T_LEADING_STRING
%token T_IS
%token T_PROVIDE

%union 
{
	long num;  /* Don't try to use long longs inhere .. Bison bug*/
	struct 
	{
		unsigned long num1;
		unsigned long num2;
	} ll;
	char * string;
}

%type <ll> T_HEX 
%type <string> T_STRING 
%type <string> T_LEADING_STRING


%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "diablo_object.h" 
#include "diablo_parsemap.h" 
static char * section_name=NULL;
static t_object * obj = NULL;
int ParseMapGCCerror();
int ParseMapGCClex();
%}
%%

start: program {if (section_name) Free(section_name);}
     ;

program: /* Empty */
       | program unary 
       ;

unary: load 
     | section
     ;


load: T_LOAD T_STRING     { Free($2); }
    ;

section: T_LEADING_STRING {if (section_name) Free(section_name); section_name=$1;} patterns		{}
       | T_LEADING_STRING {if (section_name) Free(section_name); section_name=$1;} T_HEX T_HEX patterns	{}
       ;

patterns: matched_pattern
	| T_PATTERN matched_pattern
	| patterns T_PATTERN matched_pattern
	;

matched_pattern: /* */
	       | matched_pattern matched {}
	       ;
	       
matched: T_STRING T_HEX T_HEX T_STRING  { 
       if ((strncmp($1,".gnu.warning",12)!=0) 
       && (strncmp($1,".got",4)!=0) 
       && (strncmp(section_name,".comment",8)!=0) 
       && (strncmp(section_name,".stab",5)!=0) 
       && (strncmp(section_name,".debug",6)!=0) 
       && (strncmp(section_name,".reginfo",8)!=0) 
       && (strncmp(section_name,".pdr",4)!=0) 
       ) { 
#ifdef GENERICADDRSUPPORT 
          t_address addr,size;
	  GenericSet_uint32(addr,$2.num1);
	  GenericSet_uint32(size,$3.num1);
	  MapInsertNode(obj->map, I86Node($1,section_name,addr,size,$4));
#else
#if BIT32ADDRSUPPORT
	  MapInsertNode(obj->map, I86Node($1,section_name,$2.num1,$3.num1,$4));
#else
	  MapInsertNode(obj->map, I86Node($1,section_name,((long long) ($2.num1))+((long long) ($2.num2)<<31LL),$3.num1,$4));
#endif
#endif
	  } /* Insert in the avl-tree */  
	  /* free the string values allocated in the lexer */
	  Free($1);
	  Free($4);
       }
       | matched T_HEX T_STRING { Free($3); }
       ;

%%

extern long line;
int ParseMapGCCerror()
{
   printf("Parse error! %ld\n",line);
   exit(0);
   return 0;
}


void ParseMapGCCBuildMap(t_object * tobj)
{
  line = 1;
  obj=tobj;
  obj->linker=GNU_LINKER;
  ParseMapGCCparse(); 
}
