/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/
/* Includes {{{ */
#include "diablo_instrument.h"
#include "diablo_object.h"
#include "diablo_reloc.h"
#include "diablo_iterators.h"
#include "diablo_print.h"
#include "../utils/util_error.h"
#include "../datastructures/data_string.h"
#include <string.h>
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_opcodes.h"
#include "../arch/arm/arm_registers.h"
#include "../arch/arm/arm_instructions.h"
#endif
/* }}} */

t_section *  InstrumentAddExtraSection(t_object * parent, t_section * parent_section, char type, char * objectname, char * sectionname, t_address total_size, t_address startaddress)
{
  t_object *newobject;
  t_section * newsection;

  newobject=ObjectGetFromCache(objectname,parent);
  if (!newobject) 
  {
    newobject=ObjectNewCached(objectname,parent);
    newobject->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);
    parent->mapped_last->next=newobject;
    newobject->next=NULL;
    parent->mapped_last=newobject;
    newobject->reloc_table=RelocTableNew(); 
    newobject->symbol_table=SymbolTableNew(newobject);
  }

  switch(type)
  {
    case CODE_SECTION:
      newsection=SectionNew(newobject,CODE_SECTION);
      T_RELOCATABLE(newsection)->type=RT_SUBSECTION;
      newsection->name=StringDup(sectionname);
      newobject->code=Realloc(newobject->code,(newobject->ncodes+1)*sizeof(t_section *));
      newobject->code[newobject->ncodes]=newsection;
      newobject->ncodes++;
      break;
    case RO_DATA_SECTION:
      newsection=SectionNew(newobject,RO_DATA_SECTION);
      T_RELOCATABLE(newsection)->type=RT_SUBSECTION;
      newsection->name=StringDup(sectionname);
      newobject->rodata=Realloc(newobject->rodata,(newobject->nrodatas+1)*sizeof(t_section *));
      newobject->rodata[newobject->nrodatas]=newsection;
      newobject->nrodatas++;
      break;
   case DATA_SECTION:
      newsection=SectionNew(newobject,DATA_SECTION);
      T_RELOCATABLE(newsection)->type=RT_SUBSECTION;
      newsection->name=StringDup(sectionname);
      newobject->data=Realloc(newobject->data,(newobject->ndatas+1)*sizeof(t_section *));
      newobject->data[newobject->ndatas]=newsection;
      newobject->ndatas++;
      break;
   case BSS_SECTION:
      newsection=SectionNew(newobject,BSS_SECTION);
      T_RELOCATABLE(newsection)->type=RT_SUBSECTION;
      newsection->name=StringDup(sectionname);
      newobject->bss=Realloc(newobject->bss,(newobject->nbsss+1)*sizeof(t_section *));
      newobject->bss[newobject->nbsss]=newsection;
      newobject->nbsss++;
      break;
    default:
      FATAL(("Unknown section type for vector: %c!",type));
      exit(0); /* keep the compiler happy */
  }

  newsection->data=Malloc(GENERIC_EXTRACT_UINT32(total_size));
  SECTION_OLD_SIZE(newsection)=total_size;
  SECTION_CSIZE(newsection)=total_size;
  SectionSetOldAddressSuper(newsection,startaddress);
  SectionSetCAddressSuper(newsection,startaddress);

 /* parent_section=SectionGetFromObjectByAddress(parent,SECTION_GET_CADDRESS(newsection));
*/  SECTION_ALIGN(newsection)=SECTION_ALIGN(parent_section);
  MAP_SECTION(newsection,parent_section->name, parent_section);

  HashInsert(newobject->section_table,(data_hash_node *) SectionHashNodeNew(newsection->name,newsection));

  return newsection;
}
/* vim: set shiftwidth=2 foldmethod=marker: */
