%{
/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the main object-file routines 
 * 
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/

#include <stdlib.h>
#include "../config.h"
#include "../datastructures/data_string.h"
#include "../utils/util_error.h"
#include "diablo_layout_script.h"
#include "diablo_layout_scripts_parser.h"

#define RNUMBER()	lparselval.longlong = StringToUint64(lparsetext,0); return NUMBER
#define RIDENT()	lparselval.string = StringDup(lparsetext); return IDENTIFIER
#define RTOKEN(tok)	return (tok)

int lparsewrap(void) { return 1; }

static int lineno=1;
static int nestingdepth=0;
static enum {section,subsection} prev_state;

%}

Identifier	[a-zA-Z_.][a-zA-Z0-9_.-]*
Decimal		[1-9][0-9]*
Octal		0[0-7]*
Hexadecimal	0(x|X)[0-9a-fA-F]+
LiteralTokens	\+|-|\*|\/|\&|\||\(|\)|;|~|:|=|\"

%s COMMENT
%s SECTION
%s SUBSECTION
%s SUBEXP
%option nounput
%%

<INITIAL>^SECTIONS	{ BEGIN(SECTION); RTOKEN(SECTIONS); }
<INITIAL>.		{ /* discard */ }
<INITIAL>\n		{ lineno++; /* printf("l: %3d\n",lineno); */ }

<SECTION>"/*"		{ BEGIN(COMMENT); prev_state = section; }
<SECTION,COMMENT,SUBSECTION,SUBEXP>\n	{ lineno++; }
<SECTION,SUBSECTION,SUBEXP>[[:blank:]]	{ /* discard */ }

<SECTION>ALIGN		{ RTOKEN(ALIGN); }
<SECTION>PROVIDE	{ RTOKEN(PROVIDE); }
<SECTION>\/DISCARD\/	{ RIDENT(); }
<SECTION>SEGMENT_START	{ RTOKEN(SEGMENT_START); }
<SECTION>SEGMENT_END	{ RTOKEN(SEGMENT_END); }
<SECTION>PUT_REMAINING_SECTIONS	{ RTOKEN(PUT_REMAINING_SECTIONS); }
<SECTION>{Identifier}	{ RIDENT(); }
<SECTION>{Octal}	{ RNUMBER(); }
<SECTION>{Decimal}	{ RNUMBER(); }
<SECTION>{Hexadecimal}	{ RNUMBER(); }
<SECTION>{LiteralTokens} { RTOKEN(lparsetext[0]); }
<SECTION>"{"		{ 
				nestingdepth++;
				if (nestingdepth == 2) BEGIN(SUBSECTION);
				RTOKEN('{');
			}
<SECTION>"}"		{
				nestingdepth--;
				if (nestingdepth == 0) BEGIN(INITIAL);
				RTOKEN('}');
			}
<SECTION>.		{ ECHO; printf(" is unknown token at line %d\n", lineno); exit(0); }

<SUBSECTION>\.[[:blank:]]*=[^;]*;	{ BEGIN(SUBEXP); yyless(0); }
<SUBSECTION>"/*"	{ BEGIN(COMMENT); prev_state = subsection; }
<SUBSECTION>"}"		{
				nestingdepth--;
				if (nestingdepth != 1)
					FATAL(("unbalanced braces at line %d",lineno));
				BEGIN(SECTION);
				RTOKEN('}');
			}
<SUBSECTION>"{"		{ FATAL(("Nesting too deep at line %d", lineno)); }
<SUBSECTION>.		{ /* discard */ }

<SUBEXP>;		{ BEGIN(SUBSECTION); RTOKEN(';'); }
<SUBEXP>ALIGN		{ RTOKEN(ALIGN); }
<SUBEXP>{Identifier}	{ RIDENT(); }
<SUBEXP>{Octal}		{ RNUMBER(); }
<SUBEXP>{Decimal}	{ RNUMBER(); }
<SUBEXP>{Hexadecimal}	{ RNUMBER(); }
<SUBEXP>{LiteralTokens} { RTOKEN(lparsetext[0]); }

<COMMENT>"*/"		{ 
				if (prev_state == section) 
					BEGIN(SECTION); 
				else if (prev_state == subsection)
					BEGIN(SUBSECTION);
				else
					FATAL(("Unknown return state from comment. line %d",lineno));
			}	
<COMMENT>.		{ /* discard */ }




%%

/* vim: set shiftwidth=2 foldmethod=marker: */
