/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)
 ***********************************************************/


#ifndef DIABLO_OBJECT_TABLES_TYPEDEFS
#define DIABLO_OBJECT_TABLES_TYPEDEFS
typedef enum 
{
   NOHANDLER,
   GROUPHANDLER,
   OBJECTHANDLER
} t_handler_type;

typedef struct _t_object_handler t_object_handler;

/*! The different objectfile formats */
typedef enum
{
#ifdef ELF_SUPPORT
   OBJECT_FILE_TYPE_ELF,
#endif
#ifdef AIF_SUPPORT
   OBJECT_FILE_TYPE_AIF,
#endif
#ifdef ALPHA_SUPPORT
   OBJECT_FILE_TYPE_ECOFF,
#endif
#ifdef PE_SUPPORT
   OBJECT_FILE_TYPE_DOS,
   OBJECT_FILE_TYPE_PE,
#endif
#ifdef JAVA_SUPPORT
   OBJECT_FILE_TYPE_JAVA,
#endif
#ifdef PALM_SUPPORT
   OBJECT_FILE_TYPE_PALM,
#endif
   T_UNKNOWN
} t_objecttype;

/*! Describes the size of a certain objectfile format */
typedef enum
{
   S_NOSIZE,  /* Depends on subtype */
   S_32BIT,   /* 32 bit risc */
   S_64BIT,   /* 64 bit risc */
   S_VARIABLE /* Cisc */
} t_instruction_size;

/*! Subtypes of objectfile formats */
typedef enum
{
   ST_UNKNOWN,
   OBJECT_FILE_SUBTYPE_ARMELFLSB,
   OBJECT_FILE_SUBTYPE_ARMELFMSB,
   OBJECT_FILE_SUBTYPE_SH3ELFLSB,
   OBJECT_FILE_SUBTYPE_SH3ELFMSB,
   OBJECT_FILE_SUBTYPE_MIPSELF,
   OBJECT_FILE_SUBTYPE_SPARCELF,
   OBJECT_FILE_SUBTYPE_I386ELF,
   OBJECT_FILE_SUBTYPE_IA64ELF,
   OBJECT_FILE_SUBTYPE_ALPHAELF,
   OBJECT_FILE_SUBTYPE_POWERPCELFLSB,
   OBJECT_FILE_SUBTYPE_POWERPCELFMSB
} t_objectsubtype;


#endif

#ifndef DIABLO_OBJECT_TABLES_H
#define DIABLO_OBJECT_TABLES_H

#include "diablo_object.h"
#include "diablo_action.h"
#include "diablo_description.h"
/*! How to handle an object */
struct _t_object_handler 
{
   /* Is this a handler for an object or a group of objects (eg. Elf) */
   t_handler_type htype;
   const struct _t_object_handler * subhandler;
   /* The type of the object */
   t_objecttype type;
   /* The subtype of the object */
   t_objectsubtype sub_type;
   /* How to identify an objectfile (see if this handler handles this type of objects) */
   t_bool (*ident)(FILE * fp);
   /* Read the objectfile (constructor for this type of objects) */
   void (*read)(FILE * fp, t_object *); 
   /* How to take an action on an object (see later) */
   void (*action)(t_action action,...);
   /* The real definitions for the object actions (see later) */
   const t_action_struct * acts;
   /* The callbacks for the fileformat-specific actions */
   const t_fileformat_action_struct * ff_acts;
   /* the (raw assembled) size of an instruction for this type of architecture */
   /* A string to identify this architecture / objectfileformat */
   const t_architecture_description * arch_desc;
   t_string name;
};
extern const t_object_handler handler_table[];

void ObjectKnownFileFormats(FILE * fp);

#endif
