/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bjorn De Sutter (brdsutte@elis.rug.ac.be) }}}
 *****************************************************************************/

/* Includes {{{ */
#include <stdlib.h>
#include "../utils/util_error.h"
#include "../utils/util_status.h"
#include "../utils/util_malloc.h"
#include "../simple_types.h"
#include "../diablo_options.h"
#include "diablo_ins.h"
#include "diablo_function.h"
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_liveness.h"
#include "diablo_constprop.h"
#include "diablo_object.h"
#include "diablo_regset.h"
#include "diablo_print.h"
#include "diablo_edge.h"
#include "diablo_opt_misc.h"
#include "diablo_iterators.h"
#ifdef ARM_SUPPORT
#include "../arch/arm/arm_instructions.h"
#include "../arch/arm/arm_opcodes.h"
#include "../arch/arm/arm_registers.h"
#include "../arch/arm/arm_utils.h"
#endif
#include "../utils/util_options.h"
#ifdef MIPS_SUPPORT
#include "../arch/mips/mips_opcodes.h"
#endif
#ifdef I386_SUPPORT
#include "../arch/i386/i386_instructions.h"
#include "../arch/i386/i386_registers.h"
#endif
/* }}} */

void OptKillNoops(t_cfg * cfg)
{
  t_bbl * bbl;
  t_ins * ins, *tmp;

  printf("enter\n");fflush(stdout);

  CFG_FOREACH_BBL(cfg,bbl)
    BBL_FOREACH_INS_SAFE(bbl,ins,tmp)
    {
      /*      if (RegsetIsEmpty(INS_REGS_DEF(ins)) && RegsetIsEmpty(INS_REGS_USE(ins)) && !InsHasSideEffect(ins))
	      InsKill(ins);*/
    }

  printf("exit\n");fflush(stdout);

}

t_bool OptKillUseless(t_cfg * cfg)
{
#ifdef ARM_SUPPORT
  t_regset condition_bits =RegsetNew();
#endif

  /* iterators */
  t_bbl * i_bbl;
  t_ins * i_ins, * i_tmp;

  /* local variables */

  t_regset live = RegsetNew();
  t_regset use  = RegsetNew();
  t_regset def  = RegsetNew();
  t_regset tmp  = RegsetNew();
  static int total_count = 0;
  t_uint32 kcount=0;

#ifdef ARM_SUPPORT
  RegsetAddReg(condition_bits,ARM_REG_Q_CONDITION);
  RegsetAddReg(condition_bits,ARM_REG_C_CONDITION);
  RegsetAddReg(condition_bits,ARM_REG_V_CONDITION);
  RegsetAddReg(condition_bits,ARM_REG_Z_CONDITION);
  RegsetAddReg(condition_bits,ARM_REG_N_CONDITION);
#endif

  ANALYSIS_START("Kill Useless Instructions");
  
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (!BBL_FUN(i_bbl)) continue;
    if( FUN_IS_HELL(BBL_FUN(i_bbl)) ) continue;

    RegsetDup(live,BBL_REGS_LIVE_OUT(i_bbl));

    BBL_FOREACH_INS_R_SAFE(i_bbl,i_ins,i_tmp)
    {
      RegsetDup(use,INS_REGS_USE(i_ins));
      RegsetDup(def,INS_REGS_DEF(i_ins));
      RegsetDup(tmp,def);
      RegsetSetIntersect(tmp,live);

#ifdef ARM_SUPPORT
      if (ARMINSOPCODE(i_ins)==ARM_LDM)
      {
	t_regset tmp2;
	RegsetDup(tmp2,def);
	RegsetSetDiff(tmp2,live);
	RegsetSubReg(tmp2,ARM_REG_R15);

	if (!RegsetIsEmpty(tmp))
	if ((live & ARMINSIMMEDIATE(i_ins))==0)
	{
	   if (!(ARMINSIMMEDIATE(i_ins) & (1<<ARM_REG_R15)))
	   {
	      if (ArmInsWriteBackHappens(i_ins))
	      {
		 int bits=0,tel;
		
		 for (tel=0; tel<32; tel++) if ((1<<tel) & ARMINSIMMEDIATE(i_ins)) bits++;
		 if (ARMINSFLAGS(i_ins) & FL_DIRUP)
		 {
		    ArmInstructionMakeAdd(i_ins,ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,4*bits,ARMINSCONDITION(i_ins));
		 }
		 else
		 {
		   /*
		    		  ArmInstructionMakeSub(j_ins,ARMINSREGB(i_ins),ARMINSREGB(i_ins),ARM_REG_NONE,4,ARMINSCONDITION(i_ins)); */
		 }
	      }
	   }
	}
	
      }
      if ((ARMINSFLAGS(i_ins) & FL_S)&&(ARMINSOPCODE(i_ins)!=ARM_CMP)&&(ARMINSOPCODE(i_ins)!=ARM_CMFE)&&(ARMINSOPCODE(i_ins)!=ARM_CMF)&&(ARMINSOPCODE(i_ins)!=ARM_CNF)&&(ARMINSOPCODE(i_ins)!=ARM_CNFE)&&(ARMINSOPCODE(i_ins)!=ARM_TST)&&(ARMINSOPCODE(i_ins)!=ARM_CMN))
      {
	 t_regset tmp2;
	 RegsetDup(tmp2,tmp);
	 RegsetSetIntersect(tmp2,condition_bits);
	 if(RegsetIsEmpty(tmp2))
	 {
	       ARMINSFLAGS(i_ins) &= ~FL_S;
	       INS_REGS_USE(i_ins) = ArmUsedRegisters(i_ins);
	       INS_REGS_DEF(i_ins) = ArmDefinedRegisters(i_ins);
	 }
      } 
#endif
      
      if (RegsetIsEmpty(tmp) && !InsHasSideEffect(i_ins))
      {
	/*	if (total_count < global_options.debugcounter)*/
	{
	  /*	  DiabloPrint(stdout,"-- Killing @I in @ieB\n",i_ins,INS_BBL(i_ins));*/
	  InsKill(i_ins);
	  total_count++;
	  kcount++;
	}
      }
      else
      {
	if (!INS_IS_CONDITIONAL(i_ins))
	  RegsetSetDiff(live,def);
	RegsetSetUnion(live,use);
      }
    }
  }

  RegsetFree(live);
  RegsetFree(use);
  RegsetFree(def);
  RegsetFree(tmp);

  ANALYSIS_END("Kill Useless Instructions",("%d instructions killed",kcount));
  return kcount > 0;
}

t_bool DefineIsLocal(t_ins * ins)
{
  t_regset defined;
  t_ins * i_ins;

  RegsetDup(defined, INS_REGS_DEF(ins));

  i_ins = INS_NEXT(ins);

  while (i_ins)
  {
    if (!INS_IS_CONDITIONAL(i_ins))
      RegsetSetDiff(defined,INS_REGS_DEF(i_ins));
    i_ins = INS_NEXT(i_ins);
  }

  return RegsetIsEmpty(defined);
}

/* merge blocks with only fallthrough out with successors with only fallthrough in */
void MergeBbls(t_cfg * cfg)
{
  t_cfg_edge * i_edge, * j_edge, * tmp;
  t_bbl * head;
  t_bbl * tail;
  t_ins * i_ins;
  int counter = 0;

  CFG_FOREACH_EDGE_SAFE(cfg,i_edge,tmp)
  {
    /* {{{ check for possibilities */
    if (EDGE_CAT(i_edge)!=ET_FALLTHROUGH) continue; 
    head = CFG_EDGE_HEAD(i_edge);
    tail = CFG_EDGE_TAIL(i_edge);
    if (NODE_SUCC_FIRST(head)!=NODE_SUCC_LAST(head)) continue;
    if (NODE_PRED_FIRST(tail)!=NODE_PRED_LAST(tail)) continue;
    if (!BBL_FUN(tail)) continue; /* Ignore blocks with no fun, they are not connected to the graph anyway */
    if (BBL_IS_LAST(tail)) continue;
    if (BBL_FUN(head)!=BBL_FUN(tail)) continue;
    /* }}} */

    /* {{{ do the merge */
    BBL_FOREACH_INS(tail,i_ins)
      INS_BBL(i_ins)=head;

    if (BBL_INS_FIRST(tail))
    {
      INS_PREV(BBL_INS_FIRST(tail))=BBL_INS_LAST(head);
      if (!BBL_INS_FIRST(head))
	BBL_INS_FIRST(head)=BBL_INS_FIRST(tail);
    }

    if (BBL_INS_LAST(head))
    {
      INS_NEXT(BBL_INS_LAST(head))=BBL_INS_FIRST(tail);
    }

    if (BBL_INS_LAST(tail))
      BBL_INS_LAST(head)=BBL_INS_LAST(tail);

    BBL_CSIZE(head) = GENERIC_ADD(BBL_CSIZE(head),BBL_CSIZE(tail));
    BBL_NINS(head)+=BBL_NINS(tail);
    BBL_NINS(tail)=0;
    BBL_INS_FIRST(tail)=BBL_INS_LAST(tail)=NULL;

    CfgEdgeKill(i_edge);

    BBL_FOREACH_SUCC_EDGE(tail,j_edge)
    {
      CFG_EDGE_HEAD(j_edge)=head;
    }

    NODE_SUCC_FIRST(head)=NODE_SUCC_FIRST(tail);
    NODE_SUCC_LAST(head)=NODE_SUCC_LAST(tail);

    NODE_SUCC_FIRST(tail)=NULL;
    NODE_SUCC_LAST(tail)=NULL;

    BblKill(tail);

    counter++;

    /* }}} */
  }

  printf("[bbl merging         ] %d merges\n",counter);
}

t_bool FtPath(t_bbl * i_bbl, t_bbl * j_bbl)
{
  t_bbl * it_bbl=i_bbl;
  t_cfg_edge * i_edge;
  while (1)
  {
    BBL_FOREACH_SUCC_EDGE(it_bbl,i_edge)
    {
      if ( EDGE_CAT(i_edge)==ET_FALLTHROUGH || EDGE_CAT(i_edge)==ET_IPFALLTHRU 
	  || EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK || 
	  EDGE_CAT(i_edge)==ET_SWILINK || EDGE_CAT(i_edge)==ET_IPSWILINK )
      {
	break;
      }

    }
    if (!i_edge) return FALSE;
    if (CFG_EDGE_TAIL(i_edge)==j_bbl) return TRUE;
    it_bbl=CFG_EDGE_TAIL(i_edge);
  }
}

void BranchForwarding(t_cfg * cfg)
{

  t_bbl * i_bbl;

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if (BBL_NINS(i_bbl)==1)
    {
      t_bool ok=TRUE;
      t_cfg_edge * i_edge=NULL;
      t_cfg_edge * j_edge=NULL;
      BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
      {
	switch	(EDGE_CAT(i_edge))
	{
	  case ET_FALLTHROUGH:
	  case ET_IPFALLTHRU:
	  
	  case ET_FUNLINK:
	  case ET_IPFUNLINK: 
	  
	  case ET_SWILINK:  
	  case ET_IPSWILINK:  
	  
	  case ET_UNKNOWN:   
	  case ET_IPUNKNOWN:   
	  
	  case ET_RETURN:   
	  case ET_CALL:   
	  
	  case ET_SWI: 
	  case ET_COMPENSATING:
	  
	  case ET_SWITCH:
	  case ET_IPSWITCH:
	    ok=FALSE;
	    break;
	    
	  
	  case ET_IPJUMP:
	    ok=FALSE; /* TODO */
	    break;
	  case ET_JUMP:
	    if (j_edge) FATAL(("Two jump edges?"));
	    if (!BBL_FUN(CFG_EDGE_TAIL(i_edge))) { ok=FALSE; break;}
	    if (CFG_EDGE_HEAD(i_edge)==CFG_EDGE_TAIL(i_edge)) { ok=FALSE; break;}
	    if (IsHellNode(CFG_EDGE_HEAD(i_edge))) { ok=FALSE; break;}
	    if (IsHellNode((CFG_EDGE_TAIL(i_edge)))) { ok=FALSE; break;}
	    if ((FunctionGetExitBlock(BBL_FUN(CFG_EDGE_TAIL(i_edge))))==CFG_EDGE_TAIL(i_edge)) { ok=FALSE; break;}

	    j_edge=i_edge;
	    break;

	  default:
	    FATAL(("Edge not handled!"));

	}
	if (!ok) break;
      }
      if (ok)
      {
	t_bool found=FALSE;
	do
	{
	  found=FALSE;
	  BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
	  {
	    switch (EDGE_CAT(i_edge))
	    {
	      case ET_FALLTHROUGH:
	      case ET_IPFALLTHRU:
		break;

	      case ET_FUNLINK:
	      case ET_IPFUNLINK: 
		break;

	      case ET_SWILINK:  
	      case ET_IPSWILINK:  
		break;

	      case ET_UNKNOWN:   
	      case ET_IPUNKNOWN:   
		break;

	      case ET_RETURN:   
		break;
	      case ET_CALL:   
		printf("Missing oppertunity!\n");

	      case ET_SWI: 
	      case ET_COMPENSATING:
		break;

	      case ET_SWITCH:
	      case ET_IPSWITCH:
		break;


	      case ET_IPJUMP:
		printf("Missing oppertunity!\n");
		break;
	      case ET_JUMP:
		if (CFG_EDGE_HEAD(i_edge)==CFG_EDGE_TAIL(i_edge)) break;
		if (IsHellNode(CFG_EDGE_HEAD(i_edge))) { break;}
		if (IsHellNode((CFG_EDGE_TAIL(i_edge)))) { break;}
		/*		DiabloPrint(stdout,"@ieB -> @ieB\n",CFG_EDGE_HEAD(i_edge), CFG_EDGE_TAIL(i_edge));*/
		EDGE_EXEC_COUNT(CfgEdgeNew(cfg,CFG_EDGE_HEAD(i_edge),CFG_EDGE_TAIL(j_edge),ET_JUMP))=EDGE_EXEC_COUNT(i_edge);
		CfgEdgeKill(i_edge);
		found=TRUE;
		break;

	      default:
		FATAL(("Edge not handled!"));

	    }
	    if (found)
	    {
	      /* Forward , break the loop*/
	      break;
	    }
	  }
	} while (found);
      }

    }
  }
}


/* turn jump edges into fallthrough edges */
void BranchElimination(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_cfg_edge * i_edge;

  static int nr_forwarded_branches = 0;

  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_bool flag_fallthrough = FALSE;
    t_bbl * j_bbl=NULL;
    t_cfg_edge * j_edge=NULL;
    t_uint32 count=0;
    
    if (IsHellNode(i_bbl)) continue;
    if (IS_DATABBL(i_bbl)) continue;
    
    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      if (
	  EDGE_CAT(i_edge)==ET_FALLTHROUGH || 
	  EDGE_CAT(i_edge)==ET_IPFALLTHRU || 
	  EDGE_CAT(i_edge)==ET_FUNLINK || 
	  EDGE_CAT(i_edge)==ET_IPFUNLINK || 
	  EDGE_CAT(i_edge)==ET_SWILINK || 
	  EDGE_CAT(i_edge)==ET_IPSWILINK || 
	  EDGE_CAT(i_edge)==ET_UNKNOWN ||  
	  EDGE_CAT(i_edge)==ET_RETURN ||  
	  EDGE_CAT(i_edge)==ET_CALL ||  
	  EDGE_CAT(i_edge)==ET_SWI ||
	  EDGE_CAT(i_edge)==ET_COMPENSATING ||  
	  EDGE_CAT(i_edge)==ET_SWITCH ||
	  EDGE_CAT(i_edge)==ET_IPSWITCH)
      {
	flag_fallthrough = TRUE;
	break;
      }
      else if (EDGE_CAT(i_edge)==ET_JUMP || EDGE_CAT(i_edge)==ET_IPJUMP)
      {
	if (j_bbl) { flag_fallthrough=TRUE; break;}
	j_bbl=CFG_EDGE_TAIL(i_edge);
	j_edge=i_edge;
      }
      else FATAL(("Implement @E\n",i_edge));
    }

    if (!j_bbl || flag_fallthrough || IsHellNode(j_bbl))
      continue;

    BBL_FOREACH_PRED_EDGE(i_bbl,i_edge)
    {
      t_cfg_edge * j_edge;
      BBL_FOREACH_SUCC_EDGE(CFG_EDGE_HEAD(i_edge),j_edge)
      {
	if (EDGE_CAT(j_edge)==ET_SWITCH)
	{
	  flag_fallthrough = TRUE;
	  break;
	}
      }
    }

    if (flag_fallthrough)
      continue;


    BBL_FOREACH_PRED_EDGE(j_bbl,i_edge)
    {
      if (
	  EDGE_CAT(i_edge)==ET_FALLTHROUGH || 
	  EDGE_CAT(i_edge)==ET_IPFALLTHRU || 
	  EDGE_CAT(i_edge)==ET_FUNLINK || 
	  EDGE_CAT(i_edge)==ET_IPFUNLINK || 
	  EDGE_CAT(i_edge)==ET_SWILINK || 
	  EDGE_CAT(i_edge)==ET_IPSWILINK || 
	  EDGE_CAT(i_edge)==ET_UNKNOWN ||  
	  EDGE_CAT(i_edge)==ET_SWITCH)
      {
	flag_fallthrough = TRUE;
	break;
      }
      else
      {
	count++;
      }
    }

    if (FtPath(j_bbl,i_bbl)) continue;

    if (flag_fallthrough)
      continue;

#if defined (ARM_SUPPORT)
    if (ARMINSOPCODE(BBL_INS_LAST(i_bbl))!=ARM_B) continue;
#elif defined (I386_SUPPORT)
    if (I386_INS_OPCODE(BBL_INS_LAST(i_bbl)) != I386_JMP) continue;
#else
    FATAL(("implement instruction check"));
#endif

    InsKill(BBL_INS_LAST(i_bbl));


    nr_forwarded_branches++;

    if (EDGE_CAT(j_edge)==ET_JUMP)
      EDGE_CAT(j_edge)=ET_FALLTHROUGH;
    else if(EDGE_CAT(j_edge)==ET_IPJUMP)
      EDGE_CAT(j_edge)=ET_IPFALLTHRU;
    else
      FATAL(("Oeps!"));
  }     
  printf("Branch forwarding : %d times!\n",nr_forwarded_branches);
}

#ifdef ARM_SUPPORT
void OptimizeStackLoadAndStores(t_cfg * cfg)
{
  t_bbl * i_bbl;
  t_ins *i_ins, *tmp;

  /*  printf("startdubbel\n");*/


  CFG_FOREACH_BBL(cfg,i_bbl)
    BBL_FOREACH_INS_SAFE(i_bbl,i_ins,tmp)
    {
      if (ARMINSOPCODE(i_ins)==ARM_STM && ARMINSREGB(i_ins)==ARM_REG_R13 && INS_NEXT(i_ins) && ARMINSOPCODE(INS_NEXT(i_ins))==ARM_STM && ARMINSREGB(INS_NEXT(i_ins))==ARM_REG_R13 && !(ARMINSIMMEDIATE(i_ins)&ARMINSIMMEDIATE(INS_NEXT(i_ins))))
      {
	int i;
	int num_regs = 0;

	for (i=0; i<16; i++)
	  if (ARMINSIMMEDIATE(i_ins) & (1 << i)) 
	    num_regs++;
	if (num_regs>=5)
	{
	  /*	  printf("dubbel ");*/
	  /*InsPrint(i_ins,0);
	  printf("dubbel ");
	  InsPrint(INS_NEXT(i_ins),0); */
	  INS_REGS_USE(i_ins)|=INS_REGS_USE(INS_NEXT(i_ins));
	  ARMINSIMMEDIATE(i_ins)|=ARMINSIMMEDIATE(INS_NEXT(i_ins));
	  INS_ATTRIB(i_ins)|=IF_CHANGED;
	  tmp = INS_NEXT(tmp);
	  InsKill(INS_NEXT(i_ins));
	}
      }

      if (ARMINSOPCODE(i_ins)==ARM_LDM && ARMINSREGB(i_ins)==ARM_REG_R13 && INS_NEXT(i_ins) && ARMINSOPCODE(INS_NEXT(i_ins))==ARM_LDM && ARMINSREGB(INS_NEXT(i_ins))==ARM_REG_R13 && !(ARMINSIMMEDIATE(i_ins)&ARMINSIMMEDIATE(INS_NEXT(i_ins))))
      {
	int i;
	int num_regs = 0;

	for (i=0; i<16; i++)
	  if (ARMINSIMMEDIATE(i_ins) & (1 << i)) 
	    num_regs++;
	if (num_regs>=5)
	{
	  /*	  printf("dubbel ");*/
	  /*InsPrint(i_ins,0); */
	  /*	  printf("dubbel ");*/
	  /*InsPrint(INS_NEXT(i_ins),0); */
	  INS_REGS_DEF(i_ins)|=INS_REGS_DEF(INS_NEXT(i_ins));
	  ARMINSIMMEDIATE(i_ins)|=ARMINSIMMEDIATE(INS_NEXT(i_ins));
	  INS_ATTRIB(i_ins)|=IF_CHANGED;
	  tmp = INS_NEXT(tmp);
	  InsKill(INS_NEXT(i_ins));
	}
      }
    }
}
#endif

void MarkAllSuccessorsOfEdge(t_cfg * cfg, t_cfg_edge * edge)
{
  t_bbl * bbl;
  t_cfg_edge * i_edge;
  t_bool change = TRUE;
  
  InitNodeMarking();

  /* walk upwards through caller chain */
  
  if (CfgEdgeIsForwardInterproc(edge) && edge->corr_edge)
    {
      /*      DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(edge->corr_edge));*/
      NodeMark(CFG_EDGE_TAIL(edge->corr_edge));
    }
  else
    {
      /*      DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(edge));*/
      NodeMark(CFG_EDGE_TAIL(edge));
    }

  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked(bbl))
	  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	  {
	    if (!CfgEdgeIsForwardInterproc(i_edge))
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge));
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge));*/
	      }
	    else if (i_edge->corr_edge)
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
	      }
	  }
    }
  
  NodeMark(CFG_EDGE_TAIL(edge));
  
  change = TRUE;

  /* walk downwards through callee-chain */
    
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked(bbl))
	  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	  {
	    if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK) continue;
	    if (!CfgEdgeIsBackwardInterproc(i_edge))
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge)) )
		  {
		    change = TRUE;
		    NodeMark(CFG_EDGE_TAIL(i_edge));
		    /*		    DiabloPrint(stdout,"marking 2 @B\n",CFG_EDGE_TAIL(i_edge));*/
		  }
		
		if (CfgEdgeIsForwardInterproc(i_edge) && i_edge->corr_edge)
		  if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		    {
		      change = TRUE;
		      NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		      /*		      DiabloPrint(stdout,"marking 2 @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
		    }
	      }
	  }
    }

  return;
}

void MarkAllSuccessorsOfEdgeHaltBbl(t_cfg * cfg, t_cfg_edge * edge, t_bbl * halt_bbl)
{
  t_bbl * bbl;
  t_cfg_edge * i_edge;
  t_bool change = TRUE;
  
  InitNodeMarking();

  /* walk upwards through caller chain */
  
  if (CfgEdgeIsForwardInterproc(edge) && edge->corr_edge)
    {
      /*      DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(edge->corr_edge));*/
      NodeMark(CFG_EDGE_TAIL(edge->corr_edge));
      if (BblIsMarked(halt_bbl)) return;
    }
  else
    {
      /*      DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(edge));*/
      NodeMark(CFG_EDGE_TAIL(edge));
      if (BblIsMarked(halt_bbl)) return;
    }

  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked(bbl))
	  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	  {
	    if (!CfgEdgeIsForwardInterproc(i_edge))
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge));
		if (BblIsMarked(halt_bbl)) return;
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge));*/
	      }
	    else if (i_edge->corr_edge)
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		if (BblIsMarked(halt_bbl)) return;
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
	      }
	  }
    }
  
  NodeMark(CFG_EDGE_TAIL(edge));
  if (BblIsMarked(halt_bbl)) return;

  change = TRUE;

  /* walk downwards through callee-chain */
    
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked(bbl))
	  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	  {
	    if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK) continue;
	    if (!CfgEdgeIsBackwardInterproc(i_edge))
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge)) )
		  {
		    change = TRUE;
		    NodeMark(CFG_EDGE_TAIL(i_edge));
		    if (BblIsMarked(halt_bbl)) return;
		    /*		    DiabloPrint(stdout,"marking 2 @B\n",CFG_EDGE_TAIL(i_edge));*/
		  }
		
		if (CfgEdgeIsForwardInterproc(i_edge) && i_edge->corr_edge)
		  if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		    {
		      change = TRUE;
		      NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		      if (BblIsMarked(halt_bbl)) return;
		      /*		      DiabloPrint(stdout,"marking 2 @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
		    }
	      }
	  }
    }

  return;
}

void MarkAllSuccessorsOfEdgeNotOverMarkedEdges(t_cfg * cfg, t_cfg_edge * edge)
{
  t_bbl * bbl;
  t_bool change = TRUE;
  t_cfg_edge * i_edge;
  
  InitNodeMarking();
  
  if (CfgEdgeIsForwardInterproc(edge) && edge->corr_edge)
    {
      if (!EdgeIsMarked(edge->corr_edge)) 
	NodeMark(CFG_EDGE_TAIL(edge->corr_edge));
    }
  else
    NodeMark(CFG_EDGE_TAIL(edge));
  
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked(bbl))
	  BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	  {
	    if (EdgeIsMarked(i_edge)) continue;
	    if (!CfgEdgeIsForwardInterproc(i_edge))
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge));
	      }
	    else if (i_edge->corr_edge && !EdgeIsMarked(i_edge->corr_edge))
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
	      }
	  }
    }

  NodeMark(CFG_EDGE_TAIL(edge));

  change = TRUE;
  
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	{
	  if (NodeIsMarked(bbl))
	    BBL_FOREACH_SUCC_EDGE(bbl,i_edge)
	    {
	      if (EdgeIsMarked(i_edge)) continue;
	      if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK) continue;
	      if (!CfgEdgeIsBackwardInterproc(i_edge))
		{
		  if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge)) )
		    change = TRUE;
		  NodeMark(CFG_EDGE_TAIL(i_edge));
		  
		  if (CfgEdgeIsForwardInterproc(i_edge) && i_edge->corr_edge)
		    if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		      {
			change = TRUE;
			NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		      }
		}
	    }
	}
    }

  return;
}

void MarkAllPredecessorOfEdge(t_cfg * cfg, t_cfg_edge * edge)
{
  t_cfg_edge * i_edge;
  t_bbl * bbl;
  t_bool change = TRUE;

  InitNodeMarking2();
  
  if (CfgEdgeIsBackwardInterproc(edge) && edge->corr_edge)
    NodeMark2(CFG_EDGE_HEAD(edge->corr_edge));
  else
    NodeMark2(CFG_EDGE_HEAD(edge));
  
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked2(bbl))
	  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	  {	
	    if (!CfgEdgeIsBackwardInterproc(i_edge))
	      {
		if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge)) )
		  change = TRUE;
		NodeMark2(CFG_EDGE_HEAD(i_edge));
	      }
	    else if (i_edge->corr_edge)
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
	      }
	  }
    }
  
  NodeMark2(CFG_EDGE_HEAD(edge));
  
  change = TRUE;
  
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked2(bbl))
	  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	  {
	    if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK) continue;
	    if (!CfgEdgeIsForwardInterproc(i_edge))
	      {
		if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge)) )
		  {
		    change = TRUE;
		    NodeMark2(CFG_EDGE_HEAD(i_edge));
		  }
		
		if (CfgEdgeIsBackwardInterproc(i_edge) && i_edge->corr_edge)
		  {
		    /*		    DiabloPrint(stdout,"@E\n",i_edge);fflush(stdout);fflush(stdout);*/
		    if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge->corr_edge)) )
		    {
		      change = TRUE;
		      NodeMark2(CFG_EDGE_HEAD(i_edge->corr_edge));			  
		    }
		  }
	      }
	  }
    }

  return;
}

void MarkAllPredecessorOfEdgeHatlBbl(t_cfg * cfg, t_cfg_edge * edge, t_bbl * halt_bbl)
{
  t_cfg_edge * i_edge;
  t_bbl * bbl;
  t_bool change = TRUE;

  InitNodeMarking2();
  
  if (CfgEdgeIsBackwardInterproc(edge) && edge->corr_edge)
    NodeMark2(CFG_EDGE_HEAD(edge->corr_edge));
  else
    NodeMark2(CFG_EDGE_HEAD(edge));
  
  if (BblIsMarked(halt_bbl)) return;

  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked2(bbl))
	  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	  {	
	    if (!CfgEdgeIsBackwardInterproc(i_edge))
	      {
		if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge)) )
		  change = TRUE;
		NodeMark2(CFG_EDGE_HEAD(i_edge));
		if (BblIsMarked(halt_bbl)) return;
	      }
	    else if (i_edge->corr_edge)
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		if (BblIsMarked(halt_bbl)) return;
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
	      }
	  }
    }
  
  NodeMark2(CFG_EDGE_HEAD(edge));

  if (BblIsMarked(halt_bbl)) return;

  change = TRUE;
  
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked2(bbl))
	  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	  {
	    if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK) continue;
	    if (!CfgEdgeIsForwardInterproc(i_edge))
	      {
		if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge)) )
		  {
		    change = TRUE;
		    NodeMark2(CFG_EDGE_HEAD(i_edge));
		    if (BblIsMarked(halt_bbl)) return;
		  }
		
		if (CfgEdgeIsBackwardInterproc(i_edge) && i_edge->corr_edge)
		  {
		    DiabloPrint(stdout,"@E\n",i_edge);fflush(stdout);fflush(stdout);
		    if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge->corr_edge)) )
		    {
		      change = TRUE;
		      NodeMark2(CFG_EDGE_HEAD(i_edge->corr_edge));
		      if (BblIsMarked(halt_bbl)) return;			  
		    }
		  }
	      }
	  }
    }

  return;
}

void MarkAllPredecessorOfEdgeNotOverMarkedEdges(t_cfg * cfg, t_cfg_edge * edge)
{
  t_cfg_edge * i_edge;
  t_bbl * bbl;
  t_bool change = TRUE;

  InitNodeMarking2();

  if (CfgEdgeIsBackwardInterproc(edge) && edge->corr_edge)
    {
      if (!EdgeIsMarked(edge->corr_edge))
	NodeMark2(CFG_EDGE_HEAD(edge->corr_edge));
    }
  else
    NodeMark2(CFG_EDGE_HEAD(edge));
  
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked2(bbl))
	  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	  {
	    if (EdgeIsMarked(i_edge)) continue;
	    if (!CfgEdgeIsBackwardInterproc(i_edge))
	      {
		if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge)) )
		  change = TRUE;
		NodeMark2(CFG_EDGE_HEAD(i_edge));
	      }
	    else if (i_edge->corr_edge && !EdgeIsMarked(i_edge->corr_edge))
	      {
		if (!NodeIsMarked(CFG_EDGE_TAIL(i_edge->corr_edge)) )
		  change = TRUE;
		NodeMark(CFG_EDGE_TAIL(i_edge->corr_edge));
		/*		DiabloPrint(stdout,"marking @B\n",CFG_EDGE_TAIL(i_edge->corr_edge));*/
	      }
	  }
    }
  
  NodeMark2(CFG_EDGE_HEAD(edge));

  change = TRUE;
  
  while (change)
    {
      change = FALSE;
      
      CFG_FOREACH_BBL(cfg,bbl)
	if (NodeIsMarked2(bbl))
	  BBL_FOREACH_PRED_EDGE(bbl,i_edge)
	  {
	    if (EdgeIsMarked(i_edge)) continue;
	    if (EDGE_CAT(i_edge)==ET_FUNLINK || EDGE_CAT(i_edge)==ET_IPFUNLINK) continue;
	    if (!CfgEdgeIsForwardInterproc(i_edge))
	      {
		if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge)) )
		  {
		    change = TRUE;
		    NodeMark2(CFG_EDGE_HEAD(i_edge));
		  }
		
		if (CfgEdgeIsBackwardInterproc(i_edge) && i_edge->corr_edge)
		  if (!NodeIsMarked2(CFG_EDGE_HEAD(i_edge->corr_edge)) )
		    {
		      change = TRUE;
		      NodeMark2(CFG_EDGE_HEAD(i_edge->corr_edge));			  
		    }
	      }
	  }
    }

  return;
}

typedef struct 
{
  t_cfg_edge * edge;
  int blocks_before;
  int blocks_after;
  int blocks_intersect;
  int blocks_union;
} t_trace_step;

int cmp_trace_step(const void * a, const void * b)
{
  t_trace_step *step1 = (t_trace_step*) a;
  t_trace_step *step2 = (t_trace_step*) b; 

  if (step1->blocks_before<step2->blocks_before)
    return -1;

  if (step1->blocks_before>step2->blocks_before)
    return 1;
  
  return 0;
}

void TryFindingNonLoopEdges(t_cfg * cfg)
{
  t_bbl * bbl;
  t_cfg_edge * edge;
  int i = 0,j;
  t_trace_step trace[100000];
  FILE * file = fopen("phases.dot","w");

  CFG_FOREACH_EDGE(cfg,edge)
    {

      if (BBL_PRED_FIRST(CFG_EDGE_HEAD(edge))==NULL) continue;

      if (EDGE_CAT(edge)==ET_FUNLINK || EDGE_CAT(edge)==ET_IPFUNLINK) continue;

      if (EDGE_CAT(edge)==ET_JUMP && BBL_IS_LAST(CFG_EDGE_TAIL(edge))) continue;


      MarkAllSuccessorsOfEdgeHaltBbl(cfg,edge,CFG_EDGE_HEAD(edge));

      if (!BblIsMarked(CFG_EDGE_HEAD(edge)))
	{
	  int x = 0;
	  int y = 0;
	  int z = 0;
	  
	  CFG_FOREACH_BBL(cfg,bbl)
	    {
	      if (NodeIsMarked(bbl))
		x+=BBL_NINS(bbl);
	    }
	  
	  /*	  MarkAllPredecessorOfEdge(cfg,edge);
	  
	  CFG_FOREACH_BBL(cfg,bbl)
	    {
	      if (NodeIsMarked2(bbl))
		{
		  if (NodeIsMarked(bbl))
		    z+=BBL_NINS(bbl);
		  y+=BBL_NINS(bbl);
		}
	    }
	  */

	  /*	  DiabloPrint(stdout,"FOUND PROGRESS EDGE @E\n",edge);*/

	  trace[i].edge=edge;
	  trace[i].blocks_after = x;
	  trace[i].blocks_before = y;
	  trace[i].blocks_intersect = z;
	  trace[i].blocks_union = x+y-z;
	  i++;
	  
	}
    }
  
  qsort(trace,i,sizeof(t_trace_step),cmp_trace_step);

  InitEdgeMarking();

  for (j=0;j<i;j++)
    {
      EdgeMark(trace[j].edge);
      if (EDGE_CAT(trace[j].edge)==ET_CALL)
	BBL_FOREACH_SUCC_EDGE(CFG_EDGE_HEAD(trace[j].edge),edge)
	  if (EDGE_CAT(edge)==ET_FUNLINK || EDGE_CAT(edge)==ET_IPFUNLINK)
	    EdgeMark(edge);
  
      DiabloPrint(stdout,
		  "BEFORE %d AFTER %d INTERSECT %d UNION %d PROGRESS EDGE @E\n",
		  trace[j].blocks_before,
		  trace[j].blocks_after,
		  trace[j].blocks_intersect,
		  trace[j].blocks_union,
		  trace[j].edge
		  );
    
    }
  
  DiabloPrint(file,"digraph phases {\n");

  for (j=0;j<i;j++)
    {

      int k = 0;

      /*      DiabloPrint(stdout,"BEFORE : %d AFTER %d NON-LOOP EDGE @E TOGETHER %d\n",trace[j].blocks_before,trace[j].blocks_after,trace[j].edge,trace[j].blocks_intersect);*/

      MarkAllSuccessorsOfEdgeNotOverMarkedEdges(cfg,trace[j].edge);

      for (k=0;k<i;k++)
	{
	  
	  if (k==j) continue;

	  if (NodeIsMarked(CFG_EDGE_HEAD(trace[k].edge)))
	    {
	      int w = 0;

	      /*	      MarkAllPredecessorOfEdgeNotOverMarkedEdges(cfg,trace[k].edge);
	      
	      	      CFG_FOREACH_BBL(cfg,bbl)
	      		if (NodeIsMarked2(bbl))
	      		  if (NodeIsMarked(bbl))
	      		    w+=BBL_NINS(bbl);
	      */
	      
	      DiabloPrint(
			  file, 
			  "\"@G->@G\" -> \"@G->@G\" [label=\"%d\"]\n",
			  BBL_CADDRESS(CFG_EDGE_HEAD(trace[j].edge)),BBL_CADDRESS(CFG_EDGE_TAIL(trace[j].edge)),
			  BBL_CADDRESS(CFG_EDGE_HEAD(trace[k].edge)),BBL_CADDRESS(CFG_EDGE_TAIL(trace[k].edge)),
			  w
			  );
	      
	      /*
	      DiabloPrint(
			  file, 
			  "\"@G->@G\" -> \"@G->@G\"\n",
			  BBL_CADDRESS(CFG_EDGE_HEAD(trace[j].edge)),BBL_CADDRESS(CFG_EDGE_TAIL(trace[j].edge)),
			  BBL_CADDRESS(CFG_EDGE_HEAD(trace[k].edge)),BBL_CADDRESS(CFG_EDGE_TAIL(trace[k].edge))
 			  );
	      
	      */
	      /*
	      if (BBL_IS_LAST(CFG_EDGE_TAIL(trace[j].edge)))
		{
		  if (BBL_IS_LAST(CFG_EDGE_TAIL(trace[k].edge)))
		    DiabloPrint(file, "\"@GX\" -> \"@GX\"\n",BBL_CADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(trace[j].edge)))),BBL_CADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(trace[k].edge)))));
		  else
		    DiabloPrint(file, "\"@GX\" -> \"@G\"\n",BBL_CADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(trace[j].edge)))),BBL_CADDRESS(CFG_EDGE_TAIL(trace[k].edge)));
		}
	      else
		if (BBL_IS_LAST(CFG_EDGE_TAIL(trace[k].edge)))
		  DiabloPrint(file, "\"@G\" -> \"@GX\"\n",BBL_CADDRESS(CFG_EDGE_TAIL(trace[j].edge)),BBL_CADDRESS(FUN_BBL_FIRST(BBL_FUN(CFG_EDGE_TAIL(trace[k].edge)))));
		else
		  DiabloPrint(file, "\"@G\" -> \"@G\"\n",BBL_CADDRESS(CFG_EDGE_TAIL(trace[j].edge)),BBL_CADDRESS(CFG_EDGE_TAIL(trace[k].edge)));
	      */
	    }
	  
	}

    }

  DiabloPrint(file,"}\n");
  fclose(file);
}

/* vim: set shiftwidth=2 foldmethod=marker: */
