/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains function definitions to parse the 
 *       linker-generated map (this is a map containing the
 *       link- order of the different subobjects) and load
 *       all the objects subobjects. 
 *
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup MAPS Maps */
/*@{ */
#ifndef DIABLO_PARSEMAP_TYPEDEFS
#define DIABLO_PARSEMAP_TYPEDEFS
#include "../datastructures/data_avlt.h"
/*! A map: maps are avl-trees */
typedef t_threaded_avl_tree * t_map;
/*! The type of a section in the map */
typedef enum {Code, Data, Dbug, Zero, TypeOther } t_map_section_type; 
/*! The attributes (Read Only, Read Write, ...) of a section in the map */
typedef enum {RO, RW, OT } t_map_section_attribute;
#endif

#ifndef DIABLO_PARSEMAP_H
#define DIABLO_PARSEMAP_H

#include <stdio.h>
#include "../simple_types.h"
#include "diablo_addresses.h"
#include "diablo_object.h" 


/*!A linked list of objects */
typedef struct 
{
   t_object * first;
   t_object * last;
   /* necessary for ObjectGet */
   t_object * parent;
} t_objects_ll;


#define MAP_FILE_EXTENSION   ".map"


/*! represents a section as found in the map-file */
typedef struct _t_map_node    
{
  t_threaded_avl_tree_node avltnode;
  t_address base_addr;
  t_address sec_size;
  t_string  object;
  t_string  sec_name;
  t_string  mapped_sec_name;
  t_map_section_type type;
  t_map_section_attribute attr;
  t_uint32 idx;
  t_object * obj;
  unsigned flags;
} t_map_node;


/* interface related functions */
/*! \defgroup MAP_ALLOC Constructors and destructors */
/*@{*/
void    GCCMapRead(t_object * obj, t_string infile);
void    AlphaMapParse(t_object * obj, t_string infile);
t_map * MapNew(void);
void    MapFree(t_map * map);
/*@}*/

/*! \defgroup MAP_NODE_ALLOC Constructors and destructors for map nodes */
/*@{*/
t_map_node * I86Node(t_string name, t_string mapped_name, t_address start, t_address size, t_string file);
t_map_node * AlphaMapParseLine(t_string line);
void         MapNodeFree(void * node);
/*@}*/

t_map_node * MapGetSection(t_map * map, t_address address);
void         MapPrint(t_map * map);
t_objects_ll MapLoadObjects(t_map * map,t_object * parent_object);
void         MapInsertNode(t_map * map, t_map_node * node);
void         MapVerify(t_object * obj, t_map * map);
void         MapMoveObjectsToParent(t_object * obj, t_map * map);

#endif
/*@}*/
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
