/***********************************************************
 *{{{ Copyright 2001,2002,2003:
 * 			Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * MAINTAINER: Bjorn De Sutter}}}
 ***********************************************************/


/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup REGSETS Register sets 
 *
 * Register sets are used in all kinds of analyses and optimisation: They are used to represent a set of registers */ 
/*@{ */
#ifndef REGSET_TYPEDEFS
#define REGSET_TYPEDEFS

#include "../simple_types.h"

#ifndef MAX_NR_REG_SUBSETS
#define MAX_NR_REG_SUBSETS 2
#endif
#ifndef REG_SUBSET_SIZE
#define REG_SUBSET_SIZE 16
#endif

#define REG_NONE (-1)

#define MAX_REG_ITERATOR (MAX_NR_REG_SUBSETS*REG_SUBSET_SIZE)
#if MAX_REG_ITERATOR > 64

#ifndef BIT64ADDRSUPPORT
typedef struct _t_regset t_regset;
typedef t_int16 t_reg;
#else
typedef struct _t_64_regset t_regset;
typedef t_int16 t_reg;
#endif

#elif MAX_REG_ITERATOR > 32
typedef t_uint64 t_regset;
typedef t_int16 t_reg;
#else
typedef t_uint32 t_regset;
typedef t_int16 t_reg;
#endif

#endif /* REGSET_TYPEDEFS*/

#ifndef REGSET_HEADER
#define REGSET_HEADER


/* Regsets > 64 {{{ */
#if MAX_REG_ITERATOR > 64

#ifndef BIT64ADDRSUPPORT
struct _t_regset
{
  t_uint16 max_reg;
  t_uint16 regset[MAX_REG_ITERATOR/16]; /* TODO: define REGS_IN_SUBSET for efficiency? */
};
#else
struct _t_64_regset
{
  t_uint16 max_reg;
  t_uint64 regset[MAX_REG_ITERATOR/64]; /* TODO: define REGS_IN_SUBSET for efficiency? */
};
#endif

extern   t_regset NullRegs;
//extern   t_regset ARM_ALL_BUT_PC_AND_COND;
t_regset AllRegs(void);
t_regset register_subset(t_uint16 index);

/* constructor and destructor */
#define  RegsetNew() NullRegs
t_regset RegsetNewInvers(t_regset x);
t_regset RegsetNewUnion(t_regset x, t_regset y);
t_regset RegsetNewIntersect(t_regset x, t_regset y);
t_regset RegsetNewDiff(t_regset x, t_regset y);
#define  RegsetNewDup(x) (x)

#define  RegsetFree(r) while(0)

/* regset manipulations with assignments */

#define RegsetSetAll(x) x=AllRegs()
#define RegsetSetEmpty(r) r=NullRegs
#define RegsetDup(x,y) (x)=(y)
#define RegsetSetInvers(x) real_RegsetSetInvers(&(x))
void real_RegsetSetInvers(t_regset* x); 
#define RegsetAddReg(r,i) real_RegsetAddReg(&(r),i)
void real_RegsetAddReg(t_regset* r, t_reg i);
/** \internal Should not be used, this is a remainder of an existing anomaly in ARM code*/
#define RegsetAddMultipleRegs(r,i) real_RegsetAddMultipleRegs(&(r),i)
void real_RegsetAddMultipleRegs(t_regset* r, t_reg i);
#define RegsetSetSingleton(r,i) real_RegsetSetSingleton(&(r),i)
void real_RegsetSetSingleton(t_regset* r, t_uint16 i);
#define RegsetSubReg(r,i) real_RegsetSubReg(&(r),i)
void real_RegsetSubReg(t_regset* r, t_uint16 i);
#define RegsetSetUnion(x,y) real_RegsetSetUnion(&(x),y)
void real_RegsetSetUnion(t_regset* x, t_regset y);
#define RegsetSetIntersect(x,y) real_RegsetSetIntersect(&(x),y)
void real_RegsetSetIntersect(t_regset* x, t_regset y);
#define RegsetSetDiff(x,y) real_RegsetSetDiff(&(x),y)
void real_RegsetSetDiff(t_regset* x, t_regset y);
#define RegsetSetMutualExclusive(x,y) real_RegsetSetMutualExclusive(&(x),y)
void real_RegsetSetMutualExclusive(t_regset* x, t_regset y);

t_uint32 RegsetCountRegs(t_regset r);

/* regset tests */

t_bool RegsetIsMutualExclusive(t_regset x, t_regset y); /*(!((x)&(y)))*/
t_bool RegsetIsEmpty(t_regset r); /*((r)==NullRegs)*/
t_bool RegsetEquals(t_regset x, t_regset y); /* ((x)==(y))*/
t_bool RegsetIn(t_regset x, t_reg y);    /*((x)&(1<<y))*/
t_bool RegsetIsSubset(t_regset x, t_regset y); /* (!((~(x))&y))*/

/* regset iterators */

t_regset RegsetNextSingleton(t_regset x);
#define REGSET_FOREACH_REG(r,i)    for (i=LL(0);i<MAX_REG_ITERATOR;i++) if (RegsetIn(r,i))
#define REGSET_FOREACH_REG_INVERSE(r,i) for (i=(MAX_REG_ITERATOR-1);i>=0 && i<MAX_REG_ITERATOR;i--) if (RegsetIn(r,i))
#define REGSET_FOREACH_SINGLETON_REGSET(r,i,j) for (i=LL(0),RegsetSetSingleton(j,0);i<MAX_REG_ITERATOR;i++,j=RegsetNextSingleton(j)) if (RegsetIn(r,i))
#define FOREACH_REG(i) for (i=LL(0);i<MAX_REG_ITERATOR;i++)
#define REGSUBSET_FOREACH_SINGLETON_REGSET3(s,r,i,j) for (i=s*REG_SUBSET_SIZE,RegsetSetSingleton(j,i);!RegsetIsEmpty(r);RegsetSetDiff(r,j),i++,RegsetSetSingleton(j,i)) if (RegsetIsSubset(r,j))
#define REGSUBSET_FOREACH_SINGLETON_REGSET4(s,r,i,j) for (i=s*REG_SUBSET_SIZE+(REG_SUBSET_SIZE-1),RegsetSetSingleton(j,i);!RegsetIsEmpty(r);RegsetSetDiff(r,j),i--,RegsetSetSingleton(j,i)) if (RegsetIsSubset(r,j))

/*}}}*/
/*Regsets < 64 {{{ */
#else  /* MAX_REG_ITERATOR <= 64 */

extern int RegTypeNameOffset[];
extern char* REGNAMES[];
extern t_regset register_subset[];

/* constants NullRegs and AllRegs should be defined with the correct data length */
#if MAX_REG_ITERATOR > 32
#define NullRegs LL(0x0)
#define AllRegs() ~(LL(0x0))
#else
#define NullRegs 0x0
#define AllRegs() ~0x0
#endif

#define register_subset(x) register_subset[x]

/*#define ARM_ALL_BUT_PC_AND_COND 0x7fff */
/* constructor and destructor */

/*! \defgroup REGSET_ALLOC Constructors and destructors */
/*@{*/
#define RegsetNew() NullRegs
#define RegsetNewInvers(x)  (~(x))
#define RegsetNewUnion(x,y) (x)|(y)
#define RegsetNewIntersect(x,y) (x)&(y)
#define RegsetNewDiff(x,y) (x)&(~(y))
#define RegsetNewDup(x) (x)

#define RegsetFree(r) while(0)
/*@}*/

/* regset manipulations with assignments */

#define RegsetSetAll(x) x=AllRegs()
#define RegsetSetEmpty(r) r=NullRegs
#define RegsetDup(x,y) (x)=(y)
#define RegsetSetInvers(x) (x)=(~(x))
#define RegsetAddReg(r,i)       (r)|=(LL(1)<<(i))
#define RegsetAddMultipleRegs(r,i) (r)|=(i)
#define RegsetSetSingleton(r,i) (r)=LL(1)<<(i)
#define RegsetSubReg(r,i)       (r)&=~(LL(1)<<(i))
#define RegsetSetUnion(x,y) (x)|=(y)
#define RegsetSetIntersect(x,y) (x)&=(y)
#define RegsetSetDiff(x,y) (x)&=(~(y))
#define RegsetSetMutualExclusive(x,y) (x)=((x)&~(y))|((~x)&(y))

/* regset tests */

#define RegsetIsMutualExclusive(x,y) (!((x)&(y)))
#define RegsetIsEmpty(r) ((r)==NullRegs)
#define RegsetEquals(x,y)  ((x)==(y))
#define RegsetIn(x,y)      ((x)& ( LL(1) << y )?TRUE:FALSE)
#define RegsetIsSubset(x,y) (!((~(x))&y))

/* regset iterators */

#define REGSET_FOREACH_REG(r,i)    for (i=LL(0);i<MAX_REG_ITERATOR;i++) if (RegsetIn(r,i))
#define REGSET_FOREACH_REG_INVERSE(r,i) for (i=(MAX_REG_ITERATOR-1);i>=0;i--) if (RegsetIn(r,i))

#define REGSET_FOREACH_SINGLETON_REGSET(r,i,j) for (i=LL(0),j=LL(1);i<MAX_REG_ITERATOR;i++,j<<=1) if (RegsetIn(r,i))

#define REGSUBSET_FOREACH_SINGLETON_REGSET(s,r,i,j) for (i=s*REG_SUBSET_SIZE,j=(LL(1))<<(s*REG_SUBSET_SIZE);i<(s+1)*REG_SUBSET_SIZE;i++,j<<=1) if (RegsetIn(r,i))
#define REGSUBSET_FOREACH_SINGLETON_REGSET2(s,r,i,j) for (i=s*REG_SUBSET_SIZE,j=(LL(1))<<(s*REG_SUBSET_SIZE);i<(s+1)*REG_SUBSET_SIZE;i++,j<<=1) if (RegsetIsSubset(r,j))
#define REGSUBSET_FOREACH_SINGLETON_REGSET3(s,r,i,j) for (i=s*REG_SUBSET_SIZE,j=(LL(1))<<(i);!RegsetIsEmpty(r);RegsetSetDiff(r,j),i++,j<<=1) if (RegsetIsSubset(r,j))
/*#define REGSUBSET_FOREACH_SINGLETON_REGSET4(s,r,i,j) for (i=s*REG_SUBSET_SIZE+(REG_SUBSET_SIZE-1),j=(LL(1))<<(i);!RegsetIsEmpty(r);RegsetSetDiff(r,j),i--,j>>=1) if (RegsetIsSubset(r,j)) */
#define REGSUBSET_FOREACH_SINGLETON_REGSET4(s,r,i,j) for (i=s*REG_SUBSET_SIZE+(REG_SUBSET_SIZE-1),RegsetSetSingleton(j,i);!RegsetIsEmpty(r);RegsetSetDiff(r,j),i--,RegsetSetSingleton(j,i)) if (RegsetIsSubset(r,j))

#define FOREACH_REG(i) for (i=LL(0);i<MAX_REG_ITERATOR;i++)

t_uint32 RegsetCountRegs(t_regset r);

#endif
#endif
/*}}}*/
/*@}*/
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
