/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * Diablo is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Diablo; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains the platform independent relocation handlers for
 * DIABLO. 
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/

/*#define VERBOSE_RELOCS   */
/*#define DEBUG_RELOCS   */
#define EDGE_REF_COUNTING



/* Includes {{{ */
#include <strings.h> 
#include "diablo_object.h" 
#include "diablo_print.h"
#include "diablo_reloc.h" 
#include "diablo_edge.h" 
#include "diablo_iterators.h" 
#include "../utils/util_malloc.h" 
#include "../utils/util_error.h"
/*}}}*/
/* Global variables and types {{{ */
/*! An element of the list of delayed relocs */
typedef struct _delayed_reloc_list_element { struct _delayed_reloc_list_element
  * next; t_reloc * reloc; t_reloc_table * table; } delayed_reloc_list_element;

/*  static t_bool delayed_reloc_killing = FALSE;*/

/*  static delayed_reloc_list_element * delayed_reloc_table = NULL; */
  /* }}} */

static void RelocUtilAddRefs(t_relocatable * from, t_relocatable * to, t_reloc * rel)
{
  t_reloc_ref * tmp;
  /* The refs */
  if (from)
  {
    tmp=RelocRefNew(rel);
    tmp->prev=NULL;
    tmp->next=RELOCATABLE_REFERS_TO(from);
    if (RELOCATABLE_REFERS_TO(from))
      RELOCATABLE_REFERS_TO(from)->prev=tmp;
    RELOCATABLE_REFERS_TO(from)=tmp;
    rel->from_ref=tmp;
  }
  /* the reged by */
  if (to)
  {
    tmp=RelocRefNew(rel);
    tmp->prev=NULL;
    tmp->next=RELOCATABLE_REFED_BY(to);
    if (RELOCATABLE_REFED_BY(to))
      RELOCATABLE_REFED_BY(to)->prev=tmp;
    RELOCATABLE_REFED_BY(to)=tmp;
    rel->to_ref=tmp;
  }
}

static void RelocUtilAddERefs(t_relocatable * to, t_reloc * rel)
{
  t_reloc_ref * tmp;
  /* the reged by */
  if (to)
  {
    tmp=RelocRefNew(rel);
    tmp->prev=NULL;
    tmp->next=RELOCATABLE_E_REFED_BY(to);
    if (RELOCATABLE_E_REFED_BY(to))
      RELOCATABLE_E_REFED_BY(to)->prev=tmp;
    RELOCATABLE_E_REFED_BY(to)=tmp;
    rel->e_to_ref=tmp;
  }
}
/* RelocTable Constructors/Destructors {{{*/
/*! Allocate a new relocation table */
t_reloc_table * RelocTableNew() { t_reloc_table *
  ret=Malloc(sizeof(t_reloc_table)); ret->first=ret->last=NULL; ret->nrelocs=0;
  return ret; }

  /*! Free a relocation table */
  void RelocTableFree(t_reloc_table * tofree) { t_reloc * next, *
    rel=tofree->first;

    while(rel) { 
      next=REL_NEXT(rel); 
      RelocTableRemoveReloc(tofree,rel);
      rel=next; 
    } 
    Free(tofree); 
  }
/*}}}*/
/* RelocTablePrint {{{ */
void RelocTablePrint(t_reloc_table  * table) 
{ 
  t_reloc * relptr; 

  printf("RELOC TABLE:\n"); 
  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) { DiabloPrint(stdout,"@R\n",relptr); } 
}
/* }}} */
/* RelocTableVerify {{{ */
void RelocTableVerify(t_object * obj, t_reloc_table  * table,t_bool hard) 
{
  t_reloc * relptr; 
  t_relocatable * to; 

  printf("VERIFY START for %s\n",obj->name);
  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
  { 
    t_reloc_ref * ref = RELOCATABLE_REFERS_TO(REL_FROM(relptr)), *prev;

    while (ref != NULL && ref->rel != relptr) { prev = ref; ref = ref->next; }

    if (!ref) FATAL(("VERIFY: Relocations corrupt! Could not find ref on from for reloc @R\n",relptr));


    if (!REL_IS_TO_RELOCATABLE(relptr)) { to=REL_TO_SYMBOL(relptr)->base; }
    else { to=REL_TO_RELOCATABLE(relptr); }

    if (!to) continue; ref = RELOCATABLE_REFED_BY(to);

    while (ref != NULL && ref->rel != relptr) { prev = ref; ref = ref->next; }

    if (!ref) FATAL(("VERIFY: Relocations corrupt! Could not find ref on to for reloc @R\n",relptr));
  }


  if (1) {
    t_object * subobj, *tmp;
    int tel;

    /* rodata {{{ */
    for (tel=0; tel<obj->nrodatas; tel++)
    {
      t_uint32 t=1;
      t_bool found=FALSE;
      t_relocatable * r;
      t_reloc_ref * ref=RELOCATABLE_REFED_BY(obj->rodata[tel]);
      DiabloPrint(stdout,"================REFS FROM of @T ==============\n",obj->rodata[tel]);

      while(ref)
      {
	DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	else r=REL_TO_SYMBOL(ref->rel)->base;
	if (r!=T_RELOCATABLE(obj->rodata[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}
	ref=ref->next;

      }

      t=1;
      DiabloPrint(stdout,"================REFS TO of @T ==============\n",obj->rodata[tel]);
      ref=RELOCATABLE_REFERS_TO(obj->rodata[tel]);
      while(ref)
      {
	DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	r=REL_FROM(ref->rel);

	if (r!=T_RELOCATABLE(obj->rodata[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}

	ref=ref->next;
      }
    }

    /* }}} */
    /* data {{{ */
    for (tel=0; tel<obj->ndatas; tel++)
    {
      t_uint32 t=1;
      t_bool found=FALSE;
      t_relocatable * r;
      t_reloc_ref * ref=RELOCATABLE_REFED_BY(obj->data[tel]);
      DiabloPrint(stdout,"================REFS FROM of @T ==============\n",obj->data[tel]);

      while(ref)
      {
	DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	else r=REL_TO_SYMBOL(ref->rel)->base;
	if (r!=T_RELOCATABLE(obj->data[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}
	ref=ref->next;

      }

      t=1;
      DiabloPrint(stdout,"================REFS TO of @T ==============\n",obj->data[tel]);
      ref=RELOCATABLE_REFERS_TO(obj->data[tel]);
      while(ref)
      {
	DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	r=REL_FROM(ref->rel);

	if (r!=T_RELOCATABLE(obj->data[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}

	ref=ref->next;
      }
    }

    /* }}} */
    /* bss {{{ */
    for (tel=0; tel<obj->nbsss; tel++)
    {
      t_uint32 t=1;
      t_bool found=FALSE;
      t_relocatable * r;
      t_reloc_ref * ref=RELOCATABLE_REFED_BY(obj->bss[tel]);
      DiabloPrint(stdout,"================REFS FROM of @T ==============\n",obj->bss[tel]);

      while(ref)
      {
	DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	else r=REL_TO_SYMBOL(ref->rel)->base;
	if (r!=T_RELOCATABLE(obj->bss[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}
	ref=ref->next;

      }

      t=1;
      DiabloPrint(stdout,"================REFS TO of @T ==============\n",obj->bss[tel]);
      ref=RELOCATABLE_REFERS_TO(obj->bss[tel]);
      while(ref)
      {
	DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	r=REL_FROM(ref->rel);

	if (r!=T_RELOCATABLE(obj->bss[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}

	ref=ref->next;
      }
    }

    /* }}} */
    /* code {{{ */
    for (tel=0; tel<obj->ncodes; tel++)
    {
      t_uint32 t=1;
      t_bool found=FALSE;
      t_relocatable * r;
      t_reloc_ref * ref=RELOCATABLE_REFED_BY(obj->code[tel]);
      t_reloc_ref * refi=RELOCATABLE_REFED_BY(obj->code[tel]);
      t_reloc_ref * ref2=RELOCATABLE_REFED_BY(obj->code[tel]);
      t_reloc_ref * ref3=RELOCATABLE_REFED_BY(obj->code[tel]);
      DiabloPrint(stdout,"================REFS FROM of @T ==============\n",obj->code[tel]);

      while (ref2)
      {
	refi=ref3;
	while(refi)
	{
	  if ((refi->rel==ref2->rel) && (refi!=ref2)) FATAL(("Duplicate ref!"));
	  refi=refi->next;
	}
	ref2=ref2->next;
      }
      while(ref)
      {
	DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	else r=REL_TO_SYMBOL(ref->rel)->base;
	if (r!=T_RELOCATABLE(obj->code[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}
	ref=ref->next;

      }

      t=1;
      DiabloPrint(stdout,"================REFS TO of @T ==============\n",obj->code[tel]);
      ref=RELOCATABLE_REFERS_TO(obj->code[tel]);
      while(ref)
      {
	DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	r=REL_FROM(ref->rel);

	if (r!=T_RELOCATABLE(obj->code[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	found=FALSE;

	if (hard)
	{
	  for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	  {
	    if (relptr==ref->rel)
	    {
	      found=TRUE;
	    }
	  }

	  if (!found) FATAL(("Reloc not found!"));
	}

	ref=ref->next;
      }
    }

    /* }}} */



    OBJECT_FOREACH_SUBOBJECT(obj,subobj,tmp)
    {

      printf("Verifying %s\n",subobj->name); 
      /* rodata {{{ */
      for (tel=0; tel<subobj->nrodatas; tel++)
      {
	t_uint32 t=1;
	t_bool found=FALSE;
	t_relocatable * r;
	t_reloc_ref * ref=RELOCATABLE_REFED_BY(subobj->rodata[tel]);

	t_reloc_ref * refi=RELOCATABLE_REFED_BY(subobj->rodata[tel]);
	t_reloc_ref * ref2=RELOCATABLE_REFED_BY(subobj->rodata[tel]);
	t_reloc_ref * ref3=RELOCATABLE_REFED_BY(subobj->rodata[tel]);

	while (ref2)
	{
	  refi=ref3;
	  while(refi)
	  {
	    if ((refi->rel==ref2->rel) && (refi!=ref2)) FATAL(("Duplicate ref!"));
	    refi=refi->next;
	  }
	  ref2=ref2->next;
	}
	DiabloPrint(stdout,"================REFS FROM of @T ==============\n",subobj->rodata[tel]);

	while(ref)
	{
	  DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	  if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	  else r=REL_TO_SYMBOL(ref->rel)->base;
	  if (r!=T_RELOCATABLE(subobj->rodata[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }
	  ref=ref->next;

	}

	t=1;
	DiabloPrint(stdout,"================REFS TO of @T ==============\n",subobj->rodata[tel]);
	ref=RELOCATABLE_REFERS_TO(subobj->rodata[tel]);
	while(ref)
	{
	  DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	  r=REL_FROM(ref->rel);

	  if (r!=T_RELOCATABLE(subobj->rodata[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }

	  ref=ref->next;
	}
      }

      /* }}} */
      /* data {{{ */
      for (tel=0; tel<subobj->ndatas; tel++)
      {
	t_uint32 t=1;
	t_bool found=FALSE;
	t_relocatable * r;
	t_reloc_ref * ref=RELOCATABLE_REFED_BY(subobj->data[tel]);
	DiabloPrint(stdout,"================REFS FROM of @T ==============\n",subobj->data[tel]);

	while(ref)
	{
	  DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	  if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	  else r=REL_TO_SYMBOL(ref->rel)->base;
	  if (r!=T_RELOCATABLE(subobj->data[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }
	  ref=ref->next;

	}

	t=1;
	DiabloPrint(stdout,"================REFS TO of @T ==============\n",subobj->data[tel]);
	ref=RELOCATABLE_REFERS_TO(subobj->data[tel]);
	while(ref)
	{
	  DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	  r=REL_FROM(ref->rel);

	  if (r!=T_RELOCATABLE(subobj->data[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }

	  ref=ref->next;
	}
      }

      /* }}} */
      /* bss {{{ */
      for (tel=0; tel<subobj->nbsss; tel++)
      {
	t_uint32 t=1;
	t_bool found=FALSE;
	t_relocatable * r;
	t_reloc_ref * ref=RELOCATABLE_REFED_BY(subobj->bss[tel]);
	DiabloPrint(stdout,"================REFS FROM of @T ==============\n",subobj->bss[tel]);

	while(ref)
	{
	  DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	  if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	  else r=REL_TO_SYMBOL(ref->rel)->base;
	  if (r!=T_RELOCATABLE(subobj->bss[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }
	  ref=ref->next;

	}

	t=1;
	DiabloPrint(stdout,"================REFS TO of @T ==============\n",subobj->bss[tel]);
	ref=RELOCATABLE_REFERS_TO(subobj->bss[tel]);
	while(ref)
	{
	  DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	  r=REL_FROM(ref->rel);

	  if (r!=T_RELOCATABLE(subobj->bss[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }

	  ref=ref->next;
	}
      }

      /* }}} */
      /* code {{{ */
      for (tel=0; tel<subobj->ncodes; tel++)
      {
	t_uint32 t=1;
	t_bool found=FALSE;
	t_relocatable * r;
	t_reloc_ref * ref=RELOCATABLE_REFED_BY(subobj->code[tel]);
	t_reloc_ref * refi=RELOCATABLE_REFERS_TO(subobj->code[tel]);
	t_reloc_ref * ref2=RELOCATABLE_REFERS_TO(subobj->code[tel]);
	t_reloc_ref * ref3=RELOCATABLE_REFERS_TO(subobj->code[tel]);

	while (ref2)
	{
	  refi=ref3;
	  while(refi)
	  {
	    if ((refi->rel==ref2->rel) && (refi!=ref2)) FATAL(("Duplicate ref!"));
	    refi=refi->next;
	  }
	  ref2=ref2->next;
	}
	DiabloPrint(stdout,"================REFS FROM of @T ==============\n",subobj->code[tel]);

	while(ref)
	{
	  DiabloPrint(stdout,"%d. @R\n",t++,ref->rel);
	  if (REL_IS_TO_RELOCATABLE(ref->rel)) r=REL_TO_RELOCATABLE(ref->rel);
	  else r=REL_TO_SYMBOL(ref->rel)->base;
	  if (r!=T_RELOCATABLE(subobj->code[tel])) FATAL(("VERIFY: Relocations corrupt!"));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }
	  ref=ref->next;

	}

	t=1;
	DiabloPrint(stdout,"================REFS TO of @T ==============\n",subobj->code[tel]);
	ref=RELOCATABLE_REFERS_TO(subobj->code[tel]);
	while(ref)
	{
	  DiabloPrint(stdout,"%d. USED REF: @R\n",t++,ref->rel);
	  r=REL_FROM(ref->rel);

	  if (r!=T_RELOCATABLE(subobj->code[tel])) FATAL(("VERIFY: Relocations corrupt! @R",ref->rel));

	  found=FALSE;

	  if (hard)
	  {
	    for (relptr=table->first; relptr!=NULL; relptr=REL_NEXT(relptr)) 
	    {
	      if (relptr==ref->rel)
	      {
		found=TRUE;
	      }
	    }

	    if (!found) FATAL(("Reloc not found!"));
	  }

	  ref=ref->next;
	}
      }

      /* }}} */
    }
  }
  printf("VERIFY END\n");

}
/* }}} */

#if 0
/* DelayedRelocKilling {{{ */
void DelayedRelocKilling(t_bool arg)
{

  if (delayed_reloc_killing == arg)
    return;

  delayed_reloc_killing = arg;

  if (delayed_reloc_killing==FALSE)
  {
    while (delayed_reloc_table)
    {
      delayed_reloc_list_element * tmp = delayed_reloc_table->next;
      RelocTableRemoveReloc(delayed_reloc_table->table,delayed_reloc_table->reloc);
      Free(delayed_reloc_table);
      delayed_reloc_table=tmp;
    }
  }
}
#endif
/*}}}*/

/* RelocTableAddRelocToSymbol {{{ */
t_reloc * RelocTableAddRelocToSymbol(
    t_reloc_table * table, 
    t_address addend, 
    t_relocatable * from, 
    t_address from_offset, 
    t_symbol * to, 
    t_bool hell, 
    t_cfg_edge * edge, 
    t_reloc * corresp, 
    t_symbol * e_sym, 
    t_string code, 
    t_bool first)
{
  t_reloc * rel=Malloc(sizeof(t_reloc));

  rel->flags = 0;
  rel->addend=addend;
  REL_FROM(rel)= from;
  rel->from_offset = from_offset;
  rel->to_symbol = TRUE;
  REL_TO_SYMBOL(rel) = to;
  to->relocrefcount++;
  rel->to_offset = GENERIC_SUB(from_offset,from_offset);
  rel->hell = hell;
  rel->corresponding = corresp;
  rel->e.sym = e_sym;
  rel->e_to_symbol = TRUE;
  rel->code = StringDup(code);
  rel->first = first;
  REL_NEXT(rel) = NULL;
  rel->label=NULL;
#ifdef DIETER
  rel->hotness = 0;
#endif

  if (table->first)
  {
    REL_NEXT(table->last)=rel;
    REL_PREV(rel)=table->last;
    table->last=rel;
  }
  else
  {
    table->first=table->last=rel;
    REL_NEXT(rel)=REL_PREV(rel)=NULL;
  }
  table->nrelocs++;
#ifdef VERBOSE_RELOCS 
  DiabloPrint(stdout,"Added symbol reloc %d: @R\n",table->nrelocs,rel);
#endif

  rel->edge = edge;
  if (edge)
  {
    if (edge->edge.category!=ET_CALL)
    { 
      FATAL(("Implement non calls! @R",rel));
    }
#if 0 
    rel->edge=CfgEdgeNewCall(edge->cfg, CFG_EDGE_HEAD(edge),  CFG_EDGE_TAIL(edge),CFG_EDGE_TAIL(edge->corr_edge), CFG_EDGE_HEAD(edge->corr_edge));
#else
    if (rel->edge->corr_edge) rel->edge->corr_edge->refcount++;
    edge->refcount++;
#endif
  }


  RelocUtilAddRefs(from,to->base,rel);
  if (e_sym) 
  {
    RelocUtilAddERefs(e_sym->base,rel);
  }

  return rel;
}
/* }}} */
/*RelocTableAddRelocToRelocatable {{{ */
t_reloc * RelocTableAddRelocToRelocatable(
    t_reloc_table * table, 
    t_address addend, 
    t_relocatable * from, 
    t_address from_offset, 
    t_relocatable * to, 
    t_address to_offset, 
    t_bool hell, 
    t_cfg_edge * edge, 
    t_reloc * corresp, 
    t_relocatable * sec, 
    t_string code, 
    t_bool first)
{
  t_reloc * rel=Malloc(sizeof(t_reloc));
  rel->flags = 0;


  rel->addend=addend;
  REL_FROM(rel)= from;
  rel->from_offset = from_offset;
  rel->to_symbol = FALSE;
  REL_TO_RELOCATABLE(rel) = to;
  rel->to_offset = to_offset;
  rel->hell = hell;
  rel->corresponding = corresp;
  rel->e.sec = sec;
  rel->e_to_symbol = FALSE;
  rel->code = StringDup(code);
  rel->first = first;
  REL_NEXT(rel) = NULL;
  rel->label=NULL;
#ifdef DIETER
  rel->hotness = 0;
#endif

  if (table->first)
  {
    REL_NEXT(table->last)=rel;
    REL_PREV(rel)=table->last;
    table->last=rel;
  }
  else
  {
    table->first=table->last=rel;
    REL_NEXT(rel)=REL_PREV(rel)=NULL;
  }
  table->nrelocs++;
#ifdef VERBOSE_RELOCS 
  DiabloPrint(stdout,"Added relocatable reloc %d: @R\n",table->nrelocs,rel);
#endif

  rel->edge = edge;
  if (edge)
  {
    if (edge->edge.category!=ET_CALL)
    { 
      FATAL(("Implement non calls! @R",rel));
    }
#if 0 
    rel->edge=CfgEdgeNewCall(edge->cfg, CFG_EDGE_HEAD(edge),  CFG_EDGE_TAIL(edge),CFG_EDGE_TAIL(edge->corr_edge), CFG_EDGE_HEAD(edge->corr_edge));
#else
    if (rel->edge->corr_edge) rel->edge->corr_edge->refcount++;
    edge->refcount++;
#endif
  }


  RelocUtilAddRefs(from,to,rel);
  if (sec) RelocUtilAddERefs(sec,rel);
  return rel;
}
/*}}}*/
/* RelocTableDupReloc {{{ */
t_reloc * RelocTableDupReloc(t_reloc_table * table, t_reloc *  orig)
{
  t_reloc * rel=Malloc(sizeof(t_reloc));
  t_relocatable * to, * from; 
  rel->addend = orig->addend;
  rel->from_offset=orig->from_offset;
  REL_FROM(rel)= REL_FROM(orig);
  rel->to_offset=orig->to_offset;
  rel->to_symbol = orig->to_symbol;
  REL_TO_VOID(rel)=REL_TO_VOID(orig);
  rel->hell = orig->hell;
  rel->corresponding = orig->corresponding;
#ifdef MIPS_SUPPORT
  if(orig->corresponding)
    rel->corresponding->corresponding=rel;
#endif
  rel->e.sec = orig->e.sec;
  rel->e_to_symbol = orig->e_to_symbol;
  rel->e_offset = orig->e_offset;
  REL_NEXT(rel)=NULL;
  rel->edge=NULL;
  rel->first = orig->first;
  rel->code = StringDup(orig->code);
  if (orig->label) rel->label = Collate("Copied",orig->label);
  else rel->label=NULL;
  rel->flags = orig->flags;

  if (table->first)
  {
    REL_NEXT(table->last)=rel;
    REL_PREV(rel)=table->last;
    table->last=rel;
  }
  else
  {
    table->first=table->last=rel;
    REL_NEXT(rel)=REL_PREV(rel)=NULL;
  }
  table->nrelocs++;


#ifdef VERBOSE_RELOCS 
  DiabloPrint(stdout,"Duplicated reloc @R\n",orig);
#endif

  if (orig->edge)
  {
    if (orig->edge->edge.category!=ET_CALL)
    {
      FATAL(("Implement non calls!"));
    }

#ifndef EDGE_REF_COUNTING
    rel->edge=CfgEdgeNewCall(orig->edge->cfg, CFG_EDGE_HEAD(orig->edge),  CFG_EDGE_TAIL(orig->edge), CFG_EDGE_TAIL(orig->edge->corr_edge), CFG_EDGE_HEAD(orig->edge->corr_edge));
#else
    rel->edge=orig->edge;
    rel->edge->refcount++;
    if (rel->edge->corr_edge) rel->edge->corr_edge->refcount++;
#endif
  }

  from=REL_FROM(rel);

  if (REL_IS_TO_RELOCATABLE(rel)) to=REL_TO_RELOCATABLE(rel);
  else to=REL_TO_SYMBOL(rel)->base;

  RelocUtilAddRefs(from,to,rel);
  if (rel->e.sec) RelocUtilAddERefs(rel->e.sec,rel);

  return rel;
}
/* }}} */

void RelocTableUnlinkReloc(t_reloc_table * table, t_reloc * reloc)
{
  if (table->first == reloc) table->first = REL_NEXT(reloc);
  if (table->last == reloc) table->last = REL_PREV(reloc);

  if (REL_PREV(reloc)) REL_NEXT(REL_PREV(reloc)) = REL_NEXT(reloc);
  if (REL_NEXT(reloc)) REL_PREV(REL_NEXT(reloc)) = REL_PREV(reloc);
}


void RelocTableLinkReloc(t_reloc_table * table, t_reloc * rel)
{
  if (table->first)
  {
    REL_NEXT(table->last)=rel;
    REL_PREV(rel)=table->last;
    table->last=rel;
  }
  else
  {
    table->first=table->last=rel;
    REL_NEXT(rel)=REL_PREV(rel)=NULL;
  }
}



/* RelocTableRemoveReloc {{{ */
void RelocTableRemoveReloc(t_reloc_table * table, t_reloc * reloc)
{
  static int reltel=0;
#ifdef DEBUG_RELOCS
  /* Debugging stuff */
  t_reloc * iter = table->first;
#endif

#if 0
  if (delayed_reloc_killing)
  {
    delayed_reloc_list_element * iter;
    delayed_reloc_list_element * tmp;
    iter=delayed_reloc_table;
    while(iter)
    {
      if ((iter->reloc==reloc) && (iter->table==table)) return;
      iter=iter->next;
    }

    tmp = (delayed_reloc_list_element*) Malloc(sizeof(delayed_reloc_list_element));
    tmp->next = delayed_reloc_table;
    tmp->reloc = reloc;
    tmp->table = table;
    delayed_reloc_table = tmp;

#ifdef VERBOSE_RELOCS 
    DiabloPrint(stdout,"Removed delayed reloc @R\n",reloc);
#endif
  }
  else
#endif
  {
    t_relocatable * relocatable;
#ifdef DEBUG_RELOCS
    /* Debugging stuff */
    while (iter && iter != reloc) iter=REL_NEXT(iter);
    if (!iter) FATAL(("Reloc table corrupt: while removing reloc @R, reloc not found in table",reloc));
#endif

    if (table->first == reloc) table->first = REL_NEXT(reloc);
    if (table->last == reloc) table->last = REL_PREV(reloc);

    if (REL_PREV(reloc)) REL_NEXT(REL_PREV(reloc)) = REL_NEXT(reloc);
    if (REL_NEXT(reloc)) REL_PREV(REL_NEXT(reloc)) = REL_PREV(reloc);

#ifdef VERBOSE_RELOCS 
    DiabloPrint(stdout,"Removed reloc @R\n",reloc);
#endif

    /* remove the refby to {{{ */
    if (!REL_IS_TO_RELOCATABLE(reloc)) 
    { 
      relocatable=REL_TO_SYMBOL(reloc)->base; 
      REL_TO_SYMBOL(reloc)->relocrefcount--; 
    }
    else 
      relocatable=REL_TO_RELOCATABLE(reloc);

    if (relocatable)
    {
      t_reloc_ref * refedby = relocatable->refedby, *prev = NULL;

      /* if the relocation is to a bbl and it has hell set to FALSE,
       * it probably is the corresponding reloc to a switch edge.
       * in that case we must find the edge and set its corresponding reloc field to NULL
       * in order to avoid dangling pointers */

#ifndef ALPHA_SUPPORT
      if (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(reloc)) == RT_BBL && !reloc->hell)
      {
	t_cfg_edge * edge;
	BBL_FOREACH_PRED_EDGE(REL_TO_RELOCATABLE(reloc),edge)
	  if (edge->rel == reloc)
	    {
	      edge->rel = NULL;
	    }
      }
#endif

      refedby=reloc->to_ref;
      prev=refedby->prev;

#ifdef VERBOSE_RELOCS 
      DiabloPrint(stdout,"Found! @R ...\n", refedby->rel);
#endif
      if (prev) prev->next = refedby->next;
      else relocatable->refedby = refedby->next;

      if (refedby->next) refedby->next->prev=prev;
      Free(refedby);
    }
    /* }}} */
    /* Remove refby from {{{ */
    relocatable=REL_FROM(reloc);
    if (relocatable)
    {
      t_reloc_ref * refto, * prev=NULL;
      refto=reloc->from_ref;
      prev=refto->prev;




      if (prev) prev->next = refto->next;
      else relocatable->refers = refto->next;
      if (refto->next) refto->next->prev=prev;
      Free(refto);
    }
    /* }}} */ 
    /* remove the extra refby to {{{ */
    if (reloc->e.sym)
    {
      if (!REL_E_IS_TO_RELOCATABLE(reloc)) 
      { 
	relocatable=REL_E_TO_SYMBOL(reloc)->base; 
      }
      else 
	relocatable=REL_E_TO_RELOCATABLE(reloc);


      if (relocatable)
      {
	t_reloc_ref * refedby = relocatable->e_refedby, *prev = NULL;
	refedby=reloc->e_to_ref;

	if (!refedby) FATAL(("Relocs corrupt for reloc @R!",reloc));
	prev=refedby->prev;

#ifdef VERBOSE_RELOCS 
	DiabloPrint(stdout,"Found! @R ...\n", refedby->rel);
#endif
	if (prev) prev->next = refedby->next;
	else relocatable->e_refedby = refedby->next;

	if (refedby->next) refedby->next->prev=prev;
	Free(refedby);
      }
    }
    /* }}} */
    /* removing the edge {{{  */
    if (reloc->edge) 
    {
#ifdef VERBOSE_RELOCS 
      DiabloPrint(stdout,"REMOVED EDGE @E\n",reloc->edge);
#endif
      if (reloc->edge->corr_edge)
      {
#ifdef VERBOSE_RELOCS 
	DiabloPrint(stdout,"REMOVED CORR EDGE @E\n",reloc->edge);
#endif
	CfgEdgeKill(reloc->edge->corr_edge);
      }
      CfgEdgeKill(reloc->edge); 
#ifdef VERBOSE_RELOCS 
      //printf("Number of successors: %d\n", hell_node->node.nsuccs);
#endif
      reltel++;
    }
    /* }}} */

    if (reloc->label)
      Free(reloc->label);
    if (reloc->code) 
      Free(reloc->code);
    bzero(reloc,sizeof(t_reloc));
    Free(reloc);
    table->nrelocs--;
  }
}
/*}}}*/
/* RelocGet {{{ */
/*! Get a relocation from the relocation table. 
 * \todo This code is really outdated... */
t_reloc * RelocGet(t_reloc_table * table,t_relocatable * section, t_address offset)
{
  t_reloc * rel=table->first;

  while(rel)
  {
    if (RELOCATABLE_TYPE(REL_FROM(rel))==RT_SUBSECTION)
    {
      if (REL_FROM(rel)==section)
      {
	if (GENERIC_EQ(offset,rel->from_offset))
	{
	  return rel;
	}
      }
    }
    rel=REL_NEXT(rel);
  }
  return NULL;
}
/* }}} */

void RelocSetFrom(t_reloc * rel, t_relocatable * new_from)
{
  t_reloc_ref * refto,* prev=NULL;
  refto=RELOCATABLE_REFERS_TO(REL_FROM(rel));

  if (REL_FROM(rel)==new_from) return;

  while (refto != NULL && refto->rel != rel)
  {
    prev = refto;
    refto = refto->next;
  }
  if (!refto)
  {
    FATAL(("Relocations corrupt! Could not find refedby for killed reloc @R\n",rel));
  }
  else 
  {
    if (prev) prev->next = refto->next;
    else RELOCATABLE_REFERS_TO(REL_FROM(rel)) = refto->next;

    if (refto->next) refto->next->prev=prev;
    Free(refto);
  }

  RelocUtilAddRefs(new_from,NULL,rel);
  REL_FROM(rel)=new_from;
}


void RelocRemoveToRefs(t_reloc * rel)
{
  t_relocatable * to=NULL;
  t_reloc_ref * refby;
  t_reloc_ref * prev=NULL;


  if (!REL_IS_TO_RELOCATABLE(rel))
  {
    to=REL_TO_SYMBOL(rel)->base;
  }
  else
  {
    to=REL_TO_RELOCATABLE(rel);
  }

  if (to)
  {
    refby=rel->to_ref;
    prev=rel->to_ref->prev;


    if (prev) prev->next = refby->next;
    else to->refedby = refby->next;

    if (refby->next) refby->next->prev=prev;
    Free(refby);
  }

}

void RelocSetEToSymbol(t_reloc * rel, t_symbol * new_to)
{
  t_reloc_ref * refby;
  t_reloc_ref * prev=NULL;
  t_relocatable * to;


  if (!REL_E_IS_TO_RELOCATABLE(rel)) 
  {
    to=REL_E_TO_SYMBOL(rel)->base;
  }
  else
  {
    FATAL(("NOT TO SYMBOL!"));
    exit(0); /* Keeps the compiler happy */
  }


  if ((to)&&(to!=new_to->base))
  {
    refby=RELOCATABLE_E_REFED_BY(to);

    while (refby != NULL && refby->rel != rel)
    {
      prev = refby;
      refby = refby->next;
    }
    if (!refby)
    {
      FATAL(("Relocations corrupt! Could not find extra refedby for killed reloc @R\n",rel));
    }
    else 
    {
      if (prev) prev->next = refby->next;
      else to->e_refedby = refby->next;

      if (refby->next) refby->next->prev=prev;
      Free(refby);
    }
  }

  /* the reged by */
  if ((new_to->base)&&(to!=new_to->base))
  {
    RelocUtilAddERefs(new_to->base,rel);
  }
  REL_E_TO_SYMBOL(rel)=new_to;
}

void RelocSetToSymbol(t_reloc * rel, t_symbol * new_to)
{
  t_reloc_ref * refby;
  t_reloc_ref * prev=NULL;
  t_relocatable * to;


  if (!REL_IS_TO_RELOCATABLE(rel)) 
  {
    to=REL_TO_SYMBOL(rel)->base;
  }
  else
  {
    FATAL(("NOT TO SYMBOL!"));
    exit(0); /* Keeps the compiler happy */
  }


  if ((to)&&(to!=new_to->base))
  {
    refby=RELOCATABLE_REFED_BY(to);

    while (refby != NULL && refby->rel != rel)
    {
      prev = refby;
      refby = refby->next;
    }
    if (!refby)
    {
      FATAL(("Relocations corrupt! Could not find refedby for killed reloc @R\n",rel));
    }
    else 
    {
      if (prev) prev->next = refby->next;
      else to->refedby = refby->next;

      if (refby->next) refby->next->prev=prev;
      Free(refby);
    }
  }

  /* the reged by */
  if ((new_to->base)&&(to!=new_to->base))
  {
    RelocUtilAddRefs(NULL, new_to->base,rel);
  }
  REL_TO_SYMBOL(rel)=new_to;
}


void RelocSetToSymbolBase(t_reloc * rel, t_relocatable * new_to)
{
  t_reloc_ref * tmp=NULL,  *tmpi;
  t_reloc_ref * refby;
  t_reloc_ref * prev=NULL;
  t_reloc_ref * next=NULL;
  t_symbol * search;

  t_relocatable * to;

  if (!REL_IS_TO_RELOCATABLE(rel)) 
  {
    search=REL_TO_SYMBOL(rel);
    to=REL_TO_SYMBOL(rel)->base;
  }
  else
  {
    FATAL(("NOT TO SYMBOL!"));
    exit(0); /* Keeps the compiler happy */
  }

  if (to)
  {
    refby=RELOCATABLE_REFED_BY(to);

    while (refby != NULL)
    {
      if ((REL_IS_TO_SYMBOL(refby->rel)) && (REL_TO_SYMBOL(refby->rel)==search))
      {
	if (prev)
	{
	  prev->next=refby->next;
	  if (refby->next) refby->next->prev=prev;
	}
	else
	{
	  RELOCATABLE_REFED_BY(to)=refby->next;
	  if (refby->next) refby->next->prev=NULL;
	}

	next=refby->next;	
	refby->next=tmp;
	if (tmp) tmp->prev=refby;
	tmp=refby;
      }
      else
      {
	next=refby->next;
	prev = refby;
      }
      refby = next ;
    }
  }

  tmpi=tmp;

  if (!tmp) FATAL(("Relocs corrupt"));

  prev=NULL;
  while(tmpi)
  {
    prev=tmpi;
    tmpi=tmpi->next;
    if (tmpi && tmpi->prev!=prev) FATAL(("CORRUPT"));
  }

  /* the reged by */
  if (new_to)
  {
    prev->next=RELOCATABLE_REFED_BY(new_to);
    if (RELOCATABLE_REFED_BY(new_to)) RELOCATABLE_REFED_BY(new_to)->prev=prev;
    RELOCATABLE_REFED_BY(new_to)=tmp;
    if (tmp) tmp->prev=NULL;
  }

  REL_TO_SYMBOL(rel)->base=new_to;
}


void RelocESetToRelocatable(t_reloc * rel, t_relocatable * new_to)
{
  t_reloc_ref * refby;
  t_reloc_ref * prev=NULL;
  t_relocatable * to;

  if (!REL_E_IS_TO_RELOCATABLE(rel)) 
  {
    REL_E_IS_TO_SYMBOL(rel) = FALSE;
    if (!REL_E_TO_SYMBOL(rel)) to=NULL;
    else to=REL_E_TO_SYMBOL(rel)->base;
  }
  else
  {
    to=REL_E_TO_RELOCATABLE(rel);
  }
  if ((to)&&(new_to!=to))
  {
    refby=RELOCATABLE_E_REFED_BY(to);

    while (refby != NULL && refby->rel != rel)
    {
      prev = refby;
      refby = refby->next;
    }
    if (!refby)
    {
      refby=RELOCATABLE_E_REFED_BY(to);
      while (refby != NULL)
	DiabloPrint(stdout,"refed by @R\n", refby->rel);

      FATAL(("Relocations corrupt! Could not find extra refedby (should be on @T) while setting the extra relocatable for reloc @R\n",rel->e.sec,rel));
    }
    else 
    {
      if (prev) prev->next = refby->next;
      else to->e_refedby = refby->next;

      if (refby->next) refby->next->prev=prev;
      Free(refby);
    }
  }

  /* the reged by */
  if ((new_to)&&(new_to!=to))
  {
    RelocUtilAddERefs(new_to,rel);
  }
  REL_E_TO_RELOCATABLE(rel)=new_to;
}


void RelocSetToRelocatable(t_reloc * rel, t_relocatable * new_to)
{
  t_reloc_ref * refby;
  t_reloc_ref * prev=NULL;
  t_relocatable * to;

  if (!REL_IS_TO_RELOCATABLE(rel)) 
  {
    REL_IS_TO_SYMBOL(rel) = FALSE;
    to=REL_TO_SYMBOL(rel)->base;
  }
  else
  {
    to=REL_TO_RELOCATABLE(rel);
  }

  if ((to)&&(new_to!=to))
  {
    refby=RELOCATABLE_REFED_BY(to);

    while (refby != NULL && refby->rel != rel)
    {
      prev = refby;
      refby = refby->next;
    }
    if (!refby)
    {
      refby=RELOCATABLE_REFED_BY(to);
      while (refby != NULL)
	DiabloPrint(stdout,"refed by @R\n", refby->rel);

      FATAL(("Relocations corrupt! Could not find refedby for killed reloc @R\n",rel));
    }
    else 
    {
      if (prev) prev->next = refby->next;
      else to->refedby = refby->next;

      if (refby->next) refby->next->prev=prev;
      Free(refby);
    }
  }

  /* the reged by */
  if ((new_to)&&(new_to!=to))
  {
    RelocUtilAddRefs(NULL, new_to,rel);
  }
  REL_TO_RELOCATABLE(rel)=new_to;
}


#ifdef DEBUG_MALLOC
t_reloc_ref * RealRelocRefNew(char * file, int lnno, t_reloc * rel)
#else
t_reloc_ref * RealRelocRefNew(t_reloc * rel)
#endif
{
#ifdef DEBUG_MALLOC
  t_reloc_ref * ret=RealMalloc(file,lnno,sizeof(t_reloc_ref));
#else
  t_reloc_ref * ret=Malloc(sizeof(t_reloc_ref));
#endif

  ret->rel=rel;
  ret->next=NULL;
  return ret;
}


/* vim: set shiftwidth=2 foldmethod=marker: */
