/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus }}}
 *****************************************************************************/

/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup RELOCS Relocations 
 *
 * 
 * */
/*@{ */
#ifndef DIABLO_RELOC_TYPEDEFS
#define DIABLO_RELOC_TYPEDEFS
typedef struct _t_reloc t_reloc;
typedef struct _t_reloc_table t_reloc_table;
#endif /* DIABLO_RELOC_TYPEDEFS */

#ifndef MY_RELOC_H
#define MY_RELOC_H
#include "../simple_types.h"
#include "diablo_relocatable.h"
#include "diablo_addresses.h"
#include "diablo_symbol.h"
#include "diablo_edge.h"

/* predefined stack machine programs */
#define CALC_REL	"SP-A+"
#define CALC_ABS	"SA+"

#define WRITE_32	"l*w\\s0000$"
#define WRITE_64	"L*W\\s0000$"

/* this is only interesting for IA-64*/
#define CALC_REL_ALIGN	"SP Ifffffffffffffff8 & -A+"

#define WRITE_IMM21	\
	"s0004 > I00000000001fffff &= " \
	"I00000000000fffff & s000d < " \
	"{ I0000000000100000 & s0010 <" \
	"f I000001ee00001fff & || F s0000$"

/*TODO: nog checken !!!*/	
#define WRITE_IMM22	\
	"I00000000003fffff &=== " \
	"I0000000000200000 & s000f <" \
	"{ I000000000000ff80 & s0014 <" \
	"{ I00000000001f0000 & s0006 <" \
	"{ I000000000000007f & s000d <" \
	"f I000001e000301fff & |||| F s0000$"	

#define WRITE_IMM64	\
	"===== " \
	"I8000000000000000 & s001b >" \
	"{ I000000000000ff80 & s0014 <" \
	"{ I00000000001f0000 & s0006 <" \
	"{ I0000000000200000 & " \
	"{ I000000000000007f & s000d <" \
	"{ I7fffffffffc00000 & s0016 >" \
	"{" \
	"f I000001e000101fff & ||||| F" \
	"F s0000$"
	
	
/* this one is only interesting for ARM:
 * TODO: put it somewhere in arch/arm */
#define WRITE_PC24	"= s0002 & % s0002 > i00ffffff &=l iff000000 &| SM?i10000000|ifeffffff& } s0017 < |: }* ! w\\l i00ffffff &-$"

/* SH3 */
#define WRITE_S16	"= sffff& k*v\\k#0f -$" 
#define WRITE_S8LO	"= s00ff& ksff00& |v\\ks00ff& #07 -$"
#define WRITE_S8HI	"= s00ff& s0008< ks00ff& |v\\ksff00& s0008> #07 -$"
#define WRITE_U8LO	"= s00ff& ksff00& |v\\ks00ff& -$"
#define WRITE_U8HI	"= s00ff& s0008< ks00ff& |v\\ksff00& s0008> -$"

/*! A relocation. Relocations describe the relation between data in one
 * relocatable and a referenced relocatable. There are two different
 * representations: a symbolical representation and the normal representation
 * used in diablo. The symbolic representation has a pointer to a symbol, which
 * holds a reference to a relocatable and is used during linking. The normal
 * representation simply holds a reference to a relocatable */

struct _t_reloc
{
  t_edge base_edge;
  t_string label;
  /** The addend from the relocation */
  t_address addend;
  /** The offset that belongs to the 'from' */
  t_address from_offset;
  /** Is the to pointing to a t_symbol (TRUE) or a t_relocatable (FALSE) */
  t_bool to_symbol;
  /** The offset that belongs to the 'to' */
  t_address to_offset;
  /** Does this relocation lead to a hell edge in the graph? */
  t_bool hell;
  /** The edge that was created because of this relocation */
  t_cfg_edge * edge;
  /** Another relocation, needed to correctly apply this relocation */
  struct _t_reloc * corresponding;
  /** An extra relocatable or symbol whose value can be needed during relocation */
  union
  {
    t_relocatable * sec;
    t_symbol * sym;
  } e;
  /** An extra offset that belongs to the extra relocatable 'sec' */
  t_address e_offset;
  /** Boolean that indicates what e, the extra field, is pointing to */
  t_bool e_to_symbol;
  /** The stack machine program that calculates the relocation */
  t_string code;
  /** Is this relocation the first or only one that comes from the 'from' */
  t_bool first;
  t_reloc_ref * from_ref;
  t_reloc_ref * to_ref;
  t_reloc_ref * e_from_ref;
  t_reloc_ref * e_to_ref;
  t_uint32      flags;

#ifdef DIETER
  t_uint32 hotness;
#endif
};

#define T_RELOC(x)			((t_reloc*)(x))
#define REL_TO_SYMBOL(x)		(T_SYMBOL(EDGE_TAIL(T_EDGE(x))))
#define REL_E_TO_SYMBOL(x)		(T_SYMBOL(x->e.sym))
#define REL_TO_VOID(x)		        ((void *) (EDGE_TAIL(T_EDGE(x))))
#define REL_TO_RELOCATABLE(x)		(T_RELOCATABLE(EDGE_TAIL(T_EDGE(x))))
#define REL_E_TO_RELOCATABLE(x)		(T_RELOCATABLE(x->e.sec))
#define REL_IS_TO_SYMBOL(x)		(T_RELOC(x)->to_symbol)
#define REL_E_IS_TO_SYMBOL(x)		(T_RELOC(x)->e_to_symbol)
#define REL_IS_TO_RELOCATABLE(x)	(!(T_RELOC(x)->to_symbol))
#define REL_E_IS_TO_RELOCATABLE(x)	(!(T_RELOC(x)->e_to_symbol))
#define REL_FROM(x)			(T_RELOCATABLE(EDGE_HEAD(T_EDGE(x))))
#define REL_NEXT(x)			(T_RELOC(EDGE_NEXT(T_EDGE(x))))
#define REL_PREV(x)			(T_RELOC(EDGE_PREV(T_EDGE(x))))
#define REL_FLAGS(x) 	                (T_RELOC(x)->flags)

#define REL_FLAG_MOVE_EXTRA_RELOCATABLE 0x1
#define REL_FLAG_USE_OLD_ADDRESSES      0x2

/*! A relocation table, keeping the number of t_reloc's included in nrelocs */
struct _t_reloc_table
{
   t_reloc * first;
   t_reloc * last;
   t_uint32 nrelocs;
};


t_reloc * RelocTableAddRelocToSymbol(t_reloc_table * table, t_address addend, t_relocatable * from, t_address from_offset, t_symbol * to, t_bool hell, t_cfg_edge * edge, t_reloc * corresp, t_symbol * e_sym, t_string code, t_bool first);
t_reloc * RelocTableAddRelocToRelocatable(t_reloc_table * table, t_address addend, t_relocatable * from, t_address from_offset, t_relocatable * to, t_address to_offset, t_bool hell, t_cfg_edge * edge, t_reloc * corresp, t_relocatable * sec, t_string code, t_bool first);


/*! \defgroup RELOC_ALLOC Constructors and destructors */
/*@{*/
t_reloc_table * RelocTableNew(void);
t_reloc * RelocTableDupReloc(t_reloc_table * table, t_reloc *  orig);
void RelocTableRemoveReloc(t_reloc_table * table, t_reloc * reloc);
/*@}*/

t_reloc * RelocGet(t_reloc_table * table,t_relocatable * section, t_address offset);
void RelocTableFree(t_reloc_table * tofree);

void RelocSetToRelocatable(t_reloc * rel, t_relocatable * new_to);
void RelocESetToRelocatable(t_reloc * rel, t_relocatable * new_to);
void RelocSetFrom(t_reloc * rel, t_relocatable * new_from);
void RelocSetToSymbolBase(t_reloc * rel, t_relocatable * new_to);
void RelocSetToSymbol(t_reloc * rel, t_symbol * new_to);
void RelocSetEToSymbol(t_reloc * rel, t_symbol * new_to);
void RelocTableUnlinkReloc(t_reloc_table * table, t_reloc * reloc);
void RelocTableLinkReloc(t_reloc_table * table, t_reloc * rel);
void RelocRemoveToRefs(t_reloc * rel);

void DelayedRelocKilling(t_bool arg);

void RelocPrint(t_reloc * reloc);

void RelocTablePrint(t_reloc_table * table);

#ifdef DEBUG_MALLOC
#define RelocRefNew(rel) RealRelocRefNew(__FILE__,__LINE__,rel)
t_reloc_ref * RealRelocRefNew(char * file, int lnno, t_reloc * rel);
#else
#define RelocRefNew(rel) RealRelocRefNew(rel)
t_reloc_ref * RealRelocRefNew(t_reloc * rel);
#endif



#endif
/*@}*/
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
