/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 *
 * Bruno De Bus, Bjorn De Sutter, Dominique Chanet, Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Written by Bruno De Bus (bdebus@elis.rug.ac.be)
 *
 * Copyright (c) 2002 Bruno De Bus
 *
 * THIS FILE IS PART OF DIABLO (KERNEL)
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be) }}} 
 *****************************************************************************/
/*!\addtogroup DIABLO_DATATYPES */
/*@{ */
/*! \defgroup RELOCATABLES Relocatables 
 *
 * 
 * */
/*@{ */
#ifndef DIABLO_RELOCATABLE_TYPEDEFS
#define DIABLO_RELOCATABLE_TYPEDEFS
typedef struct _t_relocatable t_relocatable;
typedef struct _t_reloc_ref t_reloc_ref;
#endif /* DIABLO_RELOC_TYPEDEFS */

#ifndef DIABLO_RELOCATABLE_H
#define DIABLO_RELOCATABLE_H
#include "../simple_types.h"
#include "diablo_addresses.h"
#include "../datastructures/data_graph.h"

/* t_relocatable types */
typedef enum {
  RT_NONE = 0,
  RT_SEGMENT,
  RT_SECTION,
  RT_SUBSECTION,
  RT_BBL,
  RT_INS
} relocatable_type;

/*! A structure describing all structs that can be relocated. All basic blocks,
 * functions, instructions, sections and subsections are instantions of this
 * type */

struct _t_relocatable
{
   t_node node;
   /* t_uint32 relocatable; */ /* Bit 0: Structure can be relocated or not, Bit 1: This is an instruction or not.... */
   relocatable_type type;
   t_address caddress;
   t_address csize;
   t_address old_address;
   t_address old_size;
   t_address min_address;
   t_address min_size;
   t_address alignment;
   t_reloc_ref * refers;
   t_reloc_ref * refedby;
   t_reloc_ref * e_refedby;
};

#define T_RELOCATABLE(relocatable) 	((t_relocatable *) relocatable)
#define RELOCATABLE_TYPE(x)		(T_RELOCATABLE(x)->type)
#define RELOCATABLE_REFERS_TO(x)	(T_RELOCATABLE(x)->refers)
#define RELOCATABLE_REFED_BY(x)	        (T_RELOCATABLE(x)->refedby)
#define RELOCATABLE_E_REFED_BY(x)	(T_RELOCATABLE(x)->e_refedby)
/* do not include this earlier, otherwise it creates dependency cycles */
#include "diablo_reloc.h"
/*! Struct to build a list of relocation references, as used in t_ins and t_bbl */
struct _t_reloc_ref
{
   struct _t_reloc_ref * next;
   struct _t_reloc_ref * prev;
   t_reloc * rel;
};

#endif
/*@}*/
/*@}*/
/* vim: set shiftwidth=2 foldmethod=marker: */
