/******************************************************************************
 * Copyright 2001,2002,2003: {{{ 
 * 
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * MAINTAINER: Bruno De Bus (bdebus@elis.rug.ac.be)  }}}
 *****************************************************************************/
/* Includes {{{ */
#include "../utils/util_malloc.h"
#include "../utils/util_error.h"
#include "diablo_edge.h"
#include "diablo_bbl.h"
#include "diablo_reloc.h"
#include "diablo_symbol.h"
#include "diablo_ins.h"
#include "diablo_flowgraph.h"
#include "diablo_print.h"
#include "diablo_linkers.h"
#include "diablo_iterators.h"
#include "../utils/util_status.h"
#include "../datastructures/data_graph.h"
#include <string.h>
/*}}}*/

/*!
 * \param obj the parent object for which we want to migrate the relocations 
 * \param code the code section
 * 
 * Migrates the relocations to the instruction representation
 *
 * \todo This should probably be RelocsMigrateToInstructions(cfg) 
 */ 

void RelocsMigrateToInstructions(t_object * obj,t_section * code)
/*{{{*/
{
  t_reloc * reloc;
  t_reloc * tmp;
  t_symbol * sym;
  t_address generic;
  t_ins * ins_s;

  /* migrate all symbols as well, as they can be handy later on for
   * reference purposes (e.g. for instrumentation) */
  SYMBOLTABLE_FOREACH_SYMBOL(obj->sub_symbol_table,sym)
  {

#ifndef KERNEL_OPT
    if (sym->local)
      if (!sym->name ||
	  (strncmp(sym->name,"$compiler",strlen("$compiler")) &&
	   strncmp(sym->name,"$handwritten",strlen("$handwritten"))))
	continue;	/* skip local symbols */
#endif

    if (sym->base && RELOCATABLE_TYPE(sym->base) == RT_SUBSECTION)
    {
      if (SECTION_PARENT_SECTION(sym->base) == code)
      {
	generic = GENERIC_ADD(sym->base->caddress,sym->offset_from_start);

	if(GENERIC_LT(generic, SECTION_GET_CADDRESS(code)) || GENERIC_GE(generic, GENERIC_ADD(SECTION_GET_CADDRESS(code), SECTION_CSIZE(code))))
	  FATAL(("trouble in paradise: symbols %s (base name = %s, parent = %s) points out of section SYM:@G=@G+@G SEC:@G+@G=@G", sym->name, T_SECTION(sym->base)->name, code->name, generic,sym->base->caddress,sym->offset_from_start,SECTION_GET_CADDRESS(code), SECTION_CSIZE(code),GENERIC_ADD(SECTION_GET_CADDRESS(code), SECTION_CSIZE(code))));
	
	ins_s=SecGetInsByAddress(code,generic);
	
	sym->base = T_RELOCATABLE(ins_s);
	sym->to_type = TO_INSTRUCTION;
	sym->offset_from_start = GENERIC_SUB(generic,INS_CADDRESS(ins_s));
	sym->current_address = generic;
      }
    }
  }
  OBJECT_FOREACH_RELOC_SAFE(obj,reloc,tmp)
  {

    /* By this time all relocs should point to subsections, never to normal sections */ 

    if (RELOCATABLE_TYPE(REL_FROM(reloc))==RT_SECTION) FATAL(("Still have relocs to normal section (@R)!",reloc));
    if (RELOCATABLE_TYPE(REL_FROM(reloc))!=RT_SUBSECTION) FATAL(("Still have relocs to normal section (@R)!",reloc));
#if 1
    if (SECTION_PARENT_SECTION(T_SECTION(REL_FROM(reloc)))==code) 
    {
      generic=GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_FROM(reloc))),reloc->from_offset);
      ins_s=SecGetInsByAddress(SECTION_PARENT_SECTION(T_SECTION(REL_FROM(reloc))),generic);

      RelocSetFrom(reloc,T_RELOCATABLE(ins_s));
      reloc->from_offset=GENERIC_SUB(generic,INS_CADDRESS(ins_s));
    }
#endif

#if 1
    if ((REL_TO_RELOCATABLE(reloc)) && (RELOCATABLE_TYPE(REL_TO_RELOCATABLE(reloc))==RT_SUBSECTION))
    {
      if (SECTION_PARENT_SECTION(T_SECTION(REL_TO_RELOCATABLE(reloc)))==code)
      {
	t_ins * ins_s;
	t_address generic;

	/*DiabloPrint(stdout,"Doing @R\n",reloc);
	DiabloPrint(stdout,"Doing %s in %s at @G\n",T_SECTION(REL_TO_RELOCATABLE(reloc))->name,T_SECTION(REL_TO_RELOCATABLE(reloc))->obj->name, SECTION_GET_CADDRESS(T_SECTION(REL_TO_RELOCATABLE(reloc)))); */
	/* Finding the instruction is straight forward: we know what needs to be relocated */
	generic=GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_TO_RELOCATABLE(reloc))),reloc->to_offset);
	ins_s=SecGetInsByAddress(SECTION_PARENT_SECTION(T_SECTION(REL_TO_RELOCATABLE(reloc))),generic);

        RelocSetToRelocatable(reloc,T_RELOCATABLE(ins_s));
	reloc->to_offset=GENERIC_SUB(generic,INS_CADDRESS(ins_s));
      }
    }


    if ((REL_E_TO_RELOCATABLE(reloc)) && (RELOCATABLE_TYPE(REL_E_TO_RELOCATABLE(reloc))==RT_SUBSECTION))
    {
      if (SECTION_PARENT_SECTION(T_SECTION(REL_E_TO_RELOCATABLE(reloc)))==code)
      {
	t_ins * ins_s;
	t_address generic;

	/* Finding the instruction is straight forward: we know what needs to be relocated */
	generic=GENERIC_ADD(SECTION_GET_CADDRESS(T_SECTION(REL_E_TO_RELOCATABLE(reloc))),reloc->e_offset);
	ins_s=SecGetInsByAddress(SECTION_PARENT_SECTION(T_SECTION(REL_E_TO_RELOCATABLE(reloc))),generic);

        RelocESetToRelocatable(reloc,T_RELOCATABLE(ins_s));
	reloc->e_offset=GENERIC_SUB(generic,INS_CADDRESS(ins_s));
      }
    }

#endif
  }


}/*}}}*/

void RelocsMigrateToRelocatables(t_object * obj)
/* {{{ Migrate relocs from symbols to relocatables */
{
  t_reloc * rel;
  t_symbol * sym;
  t_relocatable * r;


  if (!obj->reloc_table) FATAL(("No relocs!"));

  OBJECT_FOREACH_RELOC_R(obj,rel)
  {
    /* if there is an extra symbol, make it now point to a relocatable */
    if (rel->e_to_symbol && rel->e.sym)
    {
      rel->e_offset = rel->e.sym->offset_from_start;
      RelocESetToRelocatable(rel,rel->e.sym->base);
    }
    else if(rel->e.sym)
    {
    }
    else rel->e_to_symbol=FALSE;

    if (!REL_IS_TO_SYMBOL(rel)) continue;
    sym = REL_TO_SYMBOL(rel);
    r = sym->base;
    rel->label=sym->name?StringDup(sym->name):StringDup("Noname");

    /* migrate to symbol */

    if (r) RelocSetToRelocatable(rel,r);

    rel->addend = GENERIC_ADD(rel->addend,sym->addend);
    rel->to_offset = sym->offset_from_start;

    switch (sym->type)
    {
      case END_SYMBOL:
	{
	}
	break;
      case SECTION_END_SYMBOL:
	{
	  t_string tmp;
	  t_uint32 tel;
	  int a=0, b=0;

	  for (tel=0; tel<strlen(rel->code); tel++)
	    if(rel->code[tel] == 'S') b++;

	  tmp = Malloc(strlen(rel->code)+1+(b*2));
	  for (tel=0; tel<=strlen(rel->code); tel++)
	  {
	    tmp[tel+a]=rel->code[tel];
	    if (rel->code[tel]=='S') 
	    {
	      tmp[tel+(++a)]='Z';
	      tmp[tel+(++a)]='+';
	    }
	  }

	  Free(rel->code);
	  rel->code=tmp;
	  rel->to_offset = NULL_GENERIC;
	}
	break;
    }

    /* patch/verify special cases {{{ */
    switch (sym->to_type)
    {
      case TO_SECTION:
	VERIFY_FASSERT(r->type == RT_SECTION,("Symbol @S must point to parent section",sym));
	break;

      case TO_SUB_SECTION:
	VERIFY_FASSERT(r->type == RT_SUBSECTION,("Symbol @S must point to subsection reloc @R",sym, rel));
	break;

      case TO_COMMON:
	FATAL(("@S still points to common: shouldn't have that here",sym));
	break;

      case TO_UNDEFINED:
	VERIFY_FASSERT(r == NULL, ("@S must have NULL base",sym));
	RelocSetToRelocatable(rel,T_RELOCATABLE(undef_section));
	break;

      case TO_NONE:
	if (sym->type!=WEAK_UNDEFINED_SYMBOL)
	{
	  VERIFY_FASSERT(sym->type == ABS_SYMBOL ,("@S is TO_NONE but not ABS_SYMBOL",sym));
	  VERIFY_FASSERT(r == NULL, ("@S must have NULL base",sym));
	  RelocSetToRelocatable(rel,T_RELOCATABLE(abs_section));
	}
	else
	{
	  if (!PatternMatch("*P*",rel->code))
	  {
	    RelocSetToRelocatable(rel,T_RELOCATABLE(abs_section));
	    rel->addend = NULL_GENERIC;
	  }
	  else if (SymbolTableGetDataType(obj, GENERIC_ADD(REL_FROM(rel)->caddress,rel->from_offset)) == ADDRESS_IS_DATA)
	  {
	    t_uint32 tel;
	    t_string orig=rel->code;
	    rel->code=StringDup(rel->code);
	    Free(orig);
	    for (tel=0; tel<strlen(rel->code); tel++)
	    {
	      if (rel->code[tel]=='S') rel->code[tel]='P';
	    }
	    if(obj->linker == ADS_LINKER && !strcmp(rel->code,"PP-A+PM+\\l*w\\s0000$"))
	    {
	      rel->code = Realloc(rel->code,strlen("PP-A+\\l*w\\s0000$")+1);
	      strcpy(rel->code,"PP-A+\\l*w\\s0000$");
	    }
	      
	    RelocSetToRelocatable(rel,T_RELOCATABLE(undef_section));
	  }
	  else
	  {
	    t_string tmp_s=Malloc(strlen(rel->code)+1);
	    t_uint32 tel;
	    int a=0;

	    for (tel=0; tel<strlen(rel->code)+1; tel++)
	    {
	      if (rel->code[tel]=='S') 
	      {
		/* ARM ADS linker has special handling of weak undefined symbols */
		if (obj->linker == ADS_LINKER)
		{
		  tmp_s=Realloc(tmp_s,strlen(rel->code)+a+1+strlen("s0004P+"));
		  strcpy(tmp_s+tel+a,"s0004P+");
		  a+=strlen("s0004P+")-1;
		}
		else if (obj->linker == GNU_LINKER)
		{
		  tmp_s=Realloc(tmp_s,strlen(rel->code)+a+1+strlen("s0000+"));
		  strcpy(tmp_s+tel+a,"s0000");
		  a+=strlen("s0000")-1;
		}
		else
		  FATAL(("Unknown linker type"));
	      }
	      else
	      {
		tmp_s[tel+a]=rel->code[tel];
	      }
	    }
	    Free(rel->code);
	    rel->code=tmp_s;
	    RelocSetToRelocatable(rel,T_RELOCATABLE(undef_section));
	  }
	}
	break;

      case TO_INSTRUCTION:
	FATAL(("Should not have instructions?"));
	break;

      case TO_BBL:
	FATAL(("Should not have bbls?"));
	break;

      default:
	/* keep the compiler happy */
	FATAL(("@S has unknown to_type",sym));
    }
    /* }}} */

       if (!strcmp(rel->label,"__stack")) DiabloPrint(stdout,"To offset: @G\n",rel->to_offset);
  }



}
/* }}} */



/* vim: set shiftwidth=2 foldmethod=marker: */
