/******************************************************************************
 * Copyright 2001,2002,2003:  {{{
 *
 * Bertrand Anckaert
 * Bruno De Bus
 * Bjorn De Sutter
 * Dominique Chanet
 * Ludo Van Put
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * DESC: This file contains functionality to create an executable containing 
 * information to enable a next run of diablo
 * 
 * MAINTAINER: Bertrand Anckaert (banckaer@elis.UGent.be) }}} 
 *****************************************************************************/

#include "diablo_save_state.h"
#include "../utils/util_error.h"
#include "diablo_object.h"
#include "diablo_symbol.h"
#include "diablo_iterators.h"
#include "diablo_section.h"
#include "diablo_relocatable.h"
#include "diablo_reloc.h"
#include "diablo_addresses.h"
#include "diablo_print.h"
#include "../utils/util_malloc.h"
#include "../datastructures/data_hash.h"
#include "../diablo_options.h"
#include <stdio.h>
#include <string.h>


/*formaat:
->blokinformatie: aantal_sectie (Sectie)*
  Sectie: naam aantal_blokinfo (Blokinfo)*
  Blokinfo: reloc_nr base_addr sec_size

->relocatieinformatie: aantal_relocatie (Relocatie)*
  Relocatie: reloc_nr_from reloc_nr_to addend from_offset to_offset hell reloc_nr_e e_offset code

->code\data: aantal_sectie (Sectie)*
  Sectie: naam aantal_offset (offset)*
  ---idee: sla adres op dat verschilt van type van vorige, begin met code.--- 
*/

/* we reserveren 0 voor de NULL pointer, 1 voor *ABS* en 2 voor *UNDEF**/
t_uint32 section_count=3;

/* Helper functions for hash table {{{ */
typedef struct _hash_section
{
  data_hash_node node;
  t_uint32 nr;
} hash_section;

void HashSectionFree(void * f, void * data)
{
  hash_section * tf=(hash_section *) f;
  Free(tf);
}

t_uint32 pointer_hash(data_hash_table * table, void * key)
{
  return (t_uint32)(POINTER_TO_INT(key)%table->tsize);
}

int pointer_cmp(void * a, void * b)
{
  if (a<b) return -1;
  else if(a==b) return 0;
  else return 1;
}
void ObjectHashElementFree(void * tf,void * data);
/* }}} */

/* CalculateCommentAddress {{{ */
t_address CalculateCommentAddress(t_object * obj)
{
  return 1024*1024*1024;
}
/* }}} */
/* RestoreStateObjectAddSection {{{ */
t_section * RestoreStateObjectAddSection(t_object * obj, char type, t_address address, t_address size, t_address align, t_string name)
{
  t_section *** dest=NULL;
  t_uint32 * ndest=NULL;

  switch(type)
  {
    case CODE_SECTION:
      dest=&obj->code;
      ndest=&obj->ncodes;
      break;
    case RO_DATA_SECTION:
      dest=&obj->rodata;
      ndest=&obj->nrodatas;
      break;
    case DATA_SECTION:
      dest=&obj->data;
      ndest=&obj->ndatas;
      break;
    case NOTE_SECTION:
      dest=&obj->note;
      ndest=&obj->nnotes;
      break;
    case BSS_SECTION:
      dest=&obj->bss;
      ndest=&obj->nbsss;
      break;
    default:
      FATAL(("Unsupported type %c in add!",type));
  }

  *dest=Realloc(*dest,sizeof(t_section *)*((*ndest)+1));

  (*dest)[(*ndest)]=SectionNew(obj,type);

  T_RELOCATABLE((*dest)[(*ndest)])->type=RT_SUBSECTION;

  /* Set the name of the new section */
  (*dest)[(*ndest)]->name=StringDup(name);
  /* Set the size of the new section */
  SECTION_OLD_SIZE((*dest)[(*ndest)])=size;
  SECTION_CSIZE((*dest)[(*ndest)])=size;
  /* Set the start address */
  SectionSetOldAddressSuper(((*dest)[(*ndest)]),address);
  SectionSetCAddressSuper(((*dest)[(*ndest)]),address);
  SECTION_ALIGN((*dest)[(*ndest)])=align;

  SET_SECTION_DATA((*dest)[(*ndest)],NULL);

  /*HashInsert(obj->section_table,(data_hash_node *) SectionHashNodeNew((*dest)[(*ndest)]->name,(*dest)[(*ndest)]));*/
  (*ndest)++;
  return (*dest)[(*ndest)-1];
}
/* End RestoreStateObjectAddSection }}} */
/* RestoreSectionsParent {{{ */
void * RestoreSectionsParent(t_section *** p_tabel, t_object * obj, void * wijzer)
{
  t_uint32 aantal_secties, i=0;
  t_string naam;
  t_uint32 * t_uint32_ptr;
  char * char_ptr;
  
  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal_secties = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  section_count+=aantal_secties;
  * p_tabel = Realloc((* p_tabel), section_count*sizeof(t_section *));

  while(i<aantal_secties)
  {
    char_ptr = (char *) wijzer;
    naam = char_ptr;
    (*p_tabel)[3+i]=(t_section *)SectionGetFromObjectByName(obj, naam);
    if((*p_tabel)[3+i]==NULL)
      FATAL(("section not found"));   
    char_ptr +=strlen(naam)+1;
    wijzer = char_ptr;
    i++;
  }
  return wijzer;
}
/* }}} */
/* RestoreSection {{{ */
void * RestoreSection(t_section ** tabel, t_object * obj, void * wijzer, t_object * parent_obj)
{
  t_string naam;
  char type;
  t_address address, size;
  relocatable_type reloc_type;
  t_uint32 parent_section;
  t_bool is_subsection, is_mapped;
  char * char_ptr;
  t_address * address_ptr;
  relocatable_type * reloc_type_ptr; 
  t_uint32 * t_uint32_ptr;
  t_bool * t_bool_ptr;
  
  char_ptr = (char *) wijzer;
  naam = char_ptr;
  char_ptr += (strlen(naam)+1);
  wijzer = char_ptr;
  
  char_ptr = (char *) wijzer;
  type = * char_ptr;
  char_ptr ++;
  wijzer = char_ptr;

  address_ptr = (t_address *) wijzer;
  address = *address_ptr;
  address_ptr++;
  wijzer = address_ptr;
  
  address_ptr = (t_address *) wijzer;
  size = *address_ptr;
  address_ptr ++;
  wijzer = address_ptr;

  reloc_type_ptr = (relocatable_type *) wijzer;
  reloc_type = *reloc_type_ptr;
  reloc_type_ptr++;
  wijzer=reloc_type_ptr;
  
  t_uint32_ptr = (t_uint32 *) wijzer;
  parent_section = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  t_bool_ptr = (t_bool *) wijzer;
  is_subsection = *t_bool_ptr;
  t_bool_ptr++;
  wijzer = t_bool_ptr;

  t_bool_ptr = (t_bool *) wijzer;
  is_mapped = *t_bool_ptr;
  t_bool_ptr++;
  wijzer = t_bool_ptr;
  
  tabel[section_count]=RestoreStateObjectAddSection(obj, type, address, size, 0, naam);
  tabel[section_count]->is_subsection = is_subsection;
  tabel[section_count]->parent_section=tabel[parent_section];
  
  MAP_SECTION(tabel[section_count],tabel[parent_section]->name,tabel[parent_section]);
  
  /*Add it to the map*/
/*  switch(type){
    case RO_DATA_SECTION:
      map_insert = CreateMapNode(address, size, obj->name, naam, Data, RO, obj);
      break;
    case DATA_SECTION:
      map_insert = CreateMapNode(address, size, obj->name, naam, Data, RW, obj);
      break;
    case NOTE_SECTION:
      map_insert = CreateMapNode(address, size, obj->name, naam, Data, RW, obj);
      break;
    case BSS_SECTION:
      map_insert = CreateMapNode(address, size, obj->name, naam, Data, RW, obj);
      break;
    case CODE_SECTION:
      map_insert = CreateMapNode(address, size, obj->name, naam, Code, RO, obj);
      break;
    default: FATAL(("hello"));
  }
  MapInsertNode(parent_obj->map, map_insert);
*/  
  if(tabel[section_count]->parent_section->data!=NULL)
  {
    tabel[section_count]->data = Malloc(size);
    memcpy(
	tabel[section_count]->data, 
	(void *)  GENERIC_ADD_DISPL(
	  tabel[section_count]->parent_section->data,
	  GENERIC_SUB(
	    tabel[section_count]->relocatable.caddress,
	    tabel[section_count]->parent_section->relocatable.caddress
	    )
	  ), 
	size);
  }
    
  section_count++;
  
  return wijzer;
}
/* }}} */
/* PatchMergedCodeSubSection {{{ */
t_section * PatchMergedCodeSubSection(t_object * parent)
{
  
  t_object * newobject;
  t_section * newsection;
  t_section * parent_section;


  newobject=ObjectNewCached(".merged_code_subobject",parent);
  newobject->section_table=HashTableNew(7,0,string_hash,(int (*)(void *, void *))  strcmp,SectionHashNodeFree);
  parent->mapped_last->next=newobject;
  newobject->next=NULL;
  parent->mapped_last=newobject;
  newobject->reloc_table=RelocTableNew(); 
  newobject->symbol_table=SymbolTableNew(newobject);
      
  newsection=SectionNew(newobject,CODE_SECTION);
  T_RELOCATABLE(newsection)->type=RT_SUBSECTION;
  newsection->name=StringDup(".merged_code_subsection");
  newobject->code=Realloc(newobject->code,(newobject->ncodes+1)*sizeof(t_section *));
  newobject->code[newobject->ncodes]=newsection;
  newobject->ncodes++;


  parent_section=SectionGetFromObjectByName(parent,".merged_code");
  SECTION_ALIGN(newsection)=SECTION_ALIGN(parent_section);
  MAP_SECTION(newsection,parent_section->name,parent_section);
  newsection->is_subsection = TRUE;

  newsection->data = Malloc(parent_section->relocatable.csize);
  memcpy(
      newsection->data, 
      parent_section->data,
      parent_section->relocatable.csize);
  SECTION_GET_CADDRESS(newsection)=SECTION_GET_CADDRESS(parent_section);

  HashInsert(newobject->section_table,(data_hash_node *) SectionHashNodeNew(newsection->name,newsection));

  return newsection;
}
/* }}} */
/* RestoreSubObject {{{ */
void * RestoreSubObject(t_section *** p_tabel, t_object * parent, void * wijzer)
{
  int aantal_secties, i=0;
  t_string naam;
  t_object * sub_obj;
  char * char_ptr;
  t_uint32 * t_uint32_ptr;
  
  char_ptr = (char *)wijzer;
  naam = char_ptr;
  char_ptr += strlen(naam)+1;
  wijzer = char_ptr;

  sub_obj = ObjectNewCached(naam,parent);
  ObjectAppendSubObject(parent, sub_obj);
  sub_obj->section_table=HashTableNew(7,0,string_hash, string_cmp, SectionHashNodeFree);
 
  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal_secties = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  *p_tabel = Realloc(*p_tabel, (section_count+aantal_secties)*sizeof(t_section *));
  
  while(i<aantal_secties){
    wijzer = RestoreSection(*p_tabel, sub_obj, wijzer, parent);
    i++;
  }
  
  return wijzer;
}
/* }}} */
/* RestoreBlockInfo {{{ */
void * RestoreBlockInfo(t_section *** tabel, t_object * obj, void * wijzer)
{  
  t_uint32 i=0;
  t_uint32 aantal_sub_obj;
  t_uint32 * t_uint32_ptr;
  
  wijzer = RestoreSectionsParent(tabel, obj, wijzer);  

  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal_sub_obj = *t_uint32_ptr;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;
  
  obj->subobject_cache = HashTableNew(103,0,string_hash,(my_cmpresult (*)(void *, void *))  strcmp,ObjectHashElementFree);
  
  while(i<aantal_sub_obj){
    wijzer = RestoreSubObject(tabel, obj, wijzer);
    i++;
  }
  
  return wijzer;
}
/* }}} */
/* RestoreCodeDataBounds {{{ */
void * RestoreCodeDataBounds(t_section ** tabel, t_object * obj, void * wijzer)
{
  t_uint32 i=0;
  t_address offset_from_start, current_address;
  t_symbol_he * top, * he;
  t_symbol * tmp;
  t_uint32 base_nr;
  t_uint32 to_type;
  t_uint32 aantal_symbols;
  t_address * t_address_ptr;
  t_uint32 * t_uint32_ptr;
  t_symbol * tmp_symbol;
  
  t_address_ptr = (t_address *) wijzer;
  current_address = *t_address_ptr;
  t_address_ptr++;
  wijzer = t_address_ptr;
  
  SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,current_address,"$first",TRUE);
  
  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal_symbols = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  while(i<aantal_symbols){
    t_address_ptr = (t_address *) wijzer;
    offset_from_start = *t_address_ptr;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,offset_from_start,"$switch",TRUE);
  
    i++;
  }
  
  i=0;
  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal_symbols = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;
  
  while(i<aantal_symbols){
    t_address_ptr = (t_address *) wijzer;
    offset_from_start = *t_address_ptr;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_address_ptr = (t_address *) wijzer;
    current_address = *t_address_ptr;
    t_address_ptr++;
    wijzer = t_address_ptr;

    t_uint32_ptr = (t_uint32 *) wijzer;
    base_nr = *t_uint32_ptr;
    t_uint32_ptr++;
    wijzer = t_uint32_ptr;
  
    t_uint32_ptr = (t_uint32 *) wijzer;
    to_type = *t_uint32_ptr;
    t_uint32_ptr++;
    wijzer = t_uint32_ptr;
    
    if(to_type == TO_SECTION){
      tmp_symbol = SymbolTableAddSymbolToSection(obj->sub_symbol_table,  LOCAL_SYMBOL, (t_relocatable *) tabel[base_nr],  current_address, "$a", TRUE);
      tmp_symbol->offset_from_start=offset_from_start;
    }
    else if(to_type==TO_SUB_SECTION)
    {
      tmp_symbol = SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,  LOCAL_SYMBOL,  (t_relocatable *)tabel[base_nr],  current_address, "$a", TRUE);
      tmp_symbol->offset_from_start=offset_from_start;
    }
    else FATAL(("unexpected"));
  
    i++;
  }

  i=0;
  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal_symbols = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  while(i<aantal_symbols){
    t_address_ptr = (t_address *) wijzer;
    offset_from_start = *t_address_ptr;
    t_address_ptr++;
    wijzer = t_address_ptr;

    t_address_ptr = (t_address *) wijzer;
    current_address = *t_address_ptr;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_uint32_ptr = (t_uint32 *) wijzer;
    base_nr = *t_uint32_ptr;
    t_uint32_ptr++;
    wijzer = t_uint32_ptr;

    t_uint32_ptr = (t_uint32 *) wijzer;
    to_type = *t_uint32_ptr;
    t_uint32_ptr++;
    wijzer = t_uint32_ptr;
    
    if(to_type == TO_SECTION)
    {
      tmp_symbol = SymbolTableAddSymbolToSection(obj->sub_symbol_table,  LOCAL_SYMBOL,  (t_relocatable *)tabel[base_nr],  current_address, "$d", TRUE);
      tmp_symbol->offset_from_start=offset_from_start;
    }
    else if(to_type==TO_SUB_SECTION)
    {
      tmp_symbol= SymbolTableAddSymbolToSubSection(obj->sub_symbol_table,  LOCAL_SYMBOL,  (t_relocatable *)tabel[base_nr],  current_address, "$d", TRUE);
      tmp_symbol->offset_from_start=offset_from_start;
    }
    else FATAL(("unexpected"));
  
    i++;
  }
  /* Sort the $switch symbols */
  top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table,"$switch");
  if (top)
  {
    while(top->node.equal)
    {
      for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
      {
	if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
	{
	  tmp=top->sym;
	  top->sym=he->sym;
	  he->sym=tmp;
	}
      }
      top=(t_symbol_he *) top->node.equal;
    }
  }
  return wijzer;
}
/* }}} */
/* RestoreCodeDataBounds2 {{{ */
void * RestoreCodeDataBounds2(t_section ** tabel, t_object * obj, void * wijzer)
{
  t_uint32 * t_uint32_ptr;
  t_address * t_address_ptr;
  t_uint32 aantal;
  t_uint32 i = 0;
  t_address address;
  t_symbol_he * top, * he;
  t_symbol * tmp;
  
  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,0,"$switch",TRUE);
  
    
  while(i<aantal){
    t_address_ptr = (t_address *) wijzer;
    address = *t_address_ptr;
    t_address_ptr++;
    wijzer = t_address_ptr;
    SymbolTableAddAbsoluteSymbol(obj->sub_symbol_table,address,"$switch",TRUE);
    i++;
  }
  /* Sort the $switch symbols */
  top=(t_symbol_he *) HashLookup(obj->sub_symbol_table->hash_table,"$switch");
  if (top)
  {
    while(top->node.equal)
    {
      for(he=(t_symbol_he *) top->node.equal;he!=NULL;he=(t_symbol_he *) he->node.equal)
      {
	if (GENERIC_LT(he->sym->current_address,top->sym->current_address))
	{
	  tmp=top->sym;
	  top->sym=he->sym;
	  he->sym=tmp;
	}
      }
      top=(t_symbol_he *) top->node.equal;
    }
  }
  return wijzer;
}
/* }}} */
/* RestoreRelocationInfoReloc {{{ */
void * RestoreRelocationInfoReloc(t_section ** tabel, t_object * obj, void * wijzer, t_section * patched_section)
{
  t_uint32 reloc_nr_from, reloc_nr_to, reloc_nr_e;
  t_address addend, from_offset, to_offset, e_offset;
  t_string code;
  t_bool hell, first;
  t_reloc * reloc;
  t_uint32 * t_uint32_ptr;
  t_address * t_address_ptr;
  t_bool * t_bool_ptr;
  char * char_ptr;
  t_relocatable * to, * from, * e;

  t_uint32_ptr = (t_uint32 *) wijzer;
  reloc_nr_from = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer =t_uint32_ptr;
  
  t_uint32_ptr = (t_uint32 *) wijzer;
  reloc_nr_to = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer =t_uint32_ptr;
  
  t_address_ptr = (t_address *) wijzer;
  addend = *t_address_ptr;
  t_address_ptr++;
  wijzer = t_address_ptr;
  
  t_address_ptr = (t_address *) wijzer;
  from_offset = *t_address_ptr;
  t_address_ptr++;
  wijzer = t_address_ptr;
  
  t_address_ptr = (t_address *) wijzer;
  to_offset = *t_address_ptr;
  t_address_ptr++;
  wijzer = t_address_ptr;

  t_bool_ptr = (t_bool *) wijzer;
  hell = *t_bool_ptr;
  t_bool_ptr ++;
  wijzer = t_bool_ptr;
  
  t_uint32_ptr = (t_uint32 *) wijzer;
  reloc_nr_e = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer =t_uint32_ptr;
  
  t_address_ptr = (t_address *) wijzer;
  e_offset = *t_address_ptr;
  t_address_ptr++;
  wijzer = t_address_ptr;

  char_ptr = (char *)wijzer;
  code = char_ptr;
  char_ptr += strlen(code)+1;
  wijzer = char_ptr;
  
  t_bool_ptr = (t_bool *) wijzer;
  first = *t_bool_ptr;
  t_bool_ptr ++;
  wijzer = t_bool_ptr;
  
  from = (t_relocatable *)tabel[reloc_nr_from];
  to = (t_relocatable *)tabel[reloc_nr_to];
  e = (t_relocatable *)tabel[reloc_nr_e];

  if(! strcmp(T_SECTION(from)->name,".merged_code"))
  {
    if(patched_section==NULL)
      FATAL(("check correctness"));
    from = (t_relocatable *)patched_section;
  } 
  
  if(! strcmp(T_SECTION(to)->name,".merged_code"))
  {
    if(patched_section==NULL)
      FATAL(("check correctness"));
    to = (t_relocatable *)patched_section;
  } 
  
  if(e && ! strcmp(T_SECTION(e)->name,".merged_code"))
  {
    if(patched_section==NULL)
      FATAL(("check correctness"));
    e = (t_relocatable *)patched_section;
  } 
  
  reloc=RelocTableAddRelocToRelocatable(obj->reloc_table, addend, from, from_offset, to, to_offset, hell, NULL, NULL, e, code, first);
  reloc->e_to_symbol = FALSE;
  reloc->to_symbol = FALSE;
  reloc->edge = NULL;

  return wijzer;
}
/* }}} */
/* RestoreRelocationInfo {{{ */
void * RestoreRelocationInfo(t_section ** tabel, t_object * obj, void * wijzer, t_section * patched_section)
{
  t_uint32 i=0;
  t_uint32 aantal_reloc;
  t_uint32 * t_uint32_ptr;
  
  t_uint32_ptr = (t_uint32 *) wijzer;
  aantal_reloc = *t_uint32_ptr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;
  
  while(i<aantal_reloc){
    wijzer = RestoreRelocationInfoReloc(tabel, obj, wijzer, patched_section);
    i++;
  }
  return wijzer;
}
/* }}} */
/* RestoreState {{{ */
t_object * RestoreState()
{
  t_object * obj;
  void * wijzer;
  int i=0;
  t_linker * t_linker_ptr;
  t_section * patched_section=NULL;
  t_uint32 phase;
  t_uint32 * t_uint32_ptr;
 
  t_section ** tabel=Malloc(3*sizeof(t_section*));
  tabel[0]=NULL;
  tabel[1]=abs_section;
  tabel[2]=undef_section;  
  /* Read the main object */
  obj = ObjectGet(global_options.objectfilename,NULL);

  if (obj->reloc_table && obj->reloc_table->first ) 
  {
    FATAL(("Strange, this object already has relocations... This is currently not implemented!"));
  }
  
  while(strcmp(obj->data[i]->name,".comment"))
  {
    i++;
    if(i==obj->ndatas) FATAL((".comment section not found"));
  }
  if(SECTION_OLD_SIZE(obj->data[i])==0){
  obj->sub_symbol_table = SymbolTableNew(obj);
  obj->map = MapNew();
    return obj;
  }
  /*    FATAL(("comment section empty"));*/
  wijzer=obj->data[i]->data;
  
  obj->sub_symbol_table = SymbolTableNew(obj);
  obj->map = MapNew();
 
  t_linker_ptr =  (t_linker *) wijzer;
  obj->linker= *t_linker_ptr;
  t_linker_ptr ++;
  wijzer = t_linker_ptr;

  wijzer = RestoreBlockInfo(&tabel, obj, wijzer);

  t_uint32_ptr = (t_uint32 *) wijzer;
  phase = * t_uint32_ptr;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;

  if(phase == SAVESTATEDISASSEMBLED){
    patched_section = PatchMergedCodeSubSection(obj);
    wijzer = RestoreRelocationInfo(tabel, obj, wijzer, patched_section);
  }  /* TODO: Bertrand: controleer of dit correct is. patched_section kan NULL zijn, klopt dat? */
  else wijzer =  RestoreRelocationInfo(tabel, obj, wijzer, NULL);

  if(phase==SAVESTATENOTDISASSEMBLED){
    wijzer = RestoreCodeDataBounds(tabel, obj, wijzer);
  }
  else if(phase ==SAVESTATEDISASSEMBLED)
    wijzer = RestoreCodeDataBounds2(tabel, obj, wijzer);
  else FATAL(("unknown phase"));
  

  Free(tabel);

  return obj;
}
/* }}} */
/* RemoveCommentSection {{{ */
void RemoveCommentSection(t_object * obj)
{
  t_uint32 i = 0;
  t_section * comment_sectie=NULL;
  
  while(i<obj->ndatas){
    if(!strcmp(obj->data[i]->name,".comment"))
    {
      comment_sectie=obj->data[i];
      break;
    }
    i++;
  }
  if (!comment_sectie) return;

  while (i+1<obj->ndatas){
    obj->data[i]=obj->data[i+1];
    i++;
  }
  obj->ndatas--;
  Free(comment_sectie);
  
}
/* }}} */

/* SaveSection {{{ */
void * SaveSection(data_hash_table * tabel, t_section * section, void * wijzer)
{
  hash_section * insert, * item;
  char * char_ptr;
  t_address * t_address_ptr;
  relocatable_type * reloc_type_ptr;
  t_uint32 * t_uint32_ptr;
  t_bool * t_bool_ptr;
   
  
  char_ptr = (char *) wijzer;
  strcpy(char_ptr,section->name);
  char_ptr += strlen(section->name)+1;
  wijzer = char_ptr;

  
  char_ptr = (char *) wijzer;
  *char_ptr=section->type;
  char_ptr++;
  wijzer = char_ptr;

  t_address_ptr = (t_address *)wijzer;
  *t_address_ptr=section->relocatable.caddress;
  t_address_ptr++;
  wijzer = t_address_ptr;

  t_address_ptr = (t_address *)wijzer;
  *t_address_ptr=section->relocatable.csize;
  t_address_ptr++;
  wijzer = t_address_ptr;

  reloc_type_ptr = (relocatable_type *) wijzer;
  *reloc_type_ptr=section->relocatable.type;
  reloc_type_ptr++;
  wijzer = reloc_type_ptr;

  item = HashLookup(tabel, section->parent_section);
  if(item==NULL)
    FATAL(("couldn't find parent section"));
  
  t_uint32_ptr = (t_uint32 *)wijzer;
  *t_uint32_ptr=((hash_section *)HashLookup(tabel, section->parent_section))->nr;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  t_bool_ptr = (t_bool *)wijzer;
  *t_bool_ptr=section->is_subsection;
  t_bool_ptr++;
  wijzer = t_bool_ptr;
  
  t_bool_ptr = (t_bool *)wijzer;
  *t_bool_ptr=section->is_mapped;
  t_bool_ptr++;
  wijzer = t_bool_ptr;
  
  insert = Malloc(sizeof(hash_section));
  insert->node.key = section;
  insert->nr=section_count++;
  HashInsert(tabel, insert);

  return wijzer;
}
/* }}} */
/* SaveSubObject {{{ */
void * SaveSubObject(data_hash_table * tabel, t_object * sub_obj, void * wijzer, int phase)
{
  int aantal_secties=0, i=0;
  void * hulp;
  char * char_ptr;
  t_uint32 * t_uint32_ptr;
  
  char_ptr =(char *)wijzer;
  strcpy(char_ptr,sub_obj->name);
  char_ptr += strlen(sub_obj->name)+1;
  wijzer = char_ptr;
  
  hulp = wijzer;
  t_uint32_ptr = wijzer;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  if(phase == SAVESTATENOTDISASSEMBLED)
    while(i<sub_obj->ncodes){
      wijzer = SaveSection(tabel, sub_obj->code[i], wijzer);
      i++;
      aantal_secties++;
    }
  else if (phase != SAVESTATEDISASSEMBLED)
    FATAL(("unknown phase"));
  
  i=0;
  while(i<sub_obj->ndatas){
    if(strcmp(sub_obj->data[i]->name,".comment"))
    {
      wijzer = SaveSection(tabel, sub_obj->data[i], wijzer);
      aantal_secties++;
    }
    i++;
  }
  
  i=0;
  while(i<sub_obj->nrodatas){
    wijzer = SaveSection(tabel, sub_obj->rodata[i], wijzer);
    i++;
    aantal_secties++;
  }
  
  i=0;
  while(i<sub_obj->nnotes){
    wijzer = SaveSection(tabel, sub_obj->note[i], wijzer);
    i++;
    aantal_secties++;
  }
  
  i=0;
  while(i<sub_obj->nbsss){
    wijzer = SaveSection(tabel, sub_obj->bss[i], wijzer);
    i++;
    aantal_secties++;
  }

  *((t_uint32 *)hulp)=aantal_secties;
  return wijzer;
}
/* }}} */
/* SaveSectionsParent {{{ */
void * SaveSectionsParent(data_hash_table * tabel, t_object * obj, void * wijzer)
{
  t_uint32 aantal_secties=0, i=0;
  void * hulp;
  hash_section * insert;
  t_uint32 * t_uint32_ptr;
  char * char_ptr;

  hulp = wijzer;
  t_uint32_ptr = wijzer;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;

  while(i<obj->ncodes){
    char_ptr = (char *) wijzer;
    strcpy(char_ptr,obj->code[i]->name);
    char_ptr += strlen(obj->code[i]->name)+1;
    wijzer = char_ptr;

    insert = Malloc(sizeof(hash_section));
    insert->node.key = obj->code[i];
    insert->nr=section_count++;
    HashInsert(tabel, insert);

    i++;
    aantal_secties++;
  }

  i=0;
  while(i<obj->ndatas){
    if(strcmp(obj->data[i]->name,".comment"))
    {
      char_ptr = (char *) wijzer;
      strcpy(char_ptr,obj->data[i]->name);
      char_ptr += strlen(obj->data[i]->name)+1;
      wijzer = char_ptr;

      insert = Malloc(sizeof(hash_section));
      insert->node.key = obj->data[i];
      insert->nr=section_count++;
      HashInsert(tabel, insert);

      aantal_secties++;
    }
    i++;
  }

  i=0;
  while(i<obj->nrodatas){
    char_ptr = (char *) wijzer;
    strcpy(char_ptr,obj->rodata[i]->name);
    char_ptr += strlen(obj->rodata[i]->name)+1;
    wijzer = char_ptr;

    insert = Malloc(sizeof(hash_section));
    insert->node.key = obj->rodata[i];
    insert->nr=section_count++;
    HashInsert(tabel, insert);

    i++;
    aantal_secties++;
  }

  i=0;
  while(i<obj->nnotes){
    char_ptr = (char *) wijzer;
    strcpy(char_ptr,obj->note[i]->name);
    char_ptr += strlen(obj->note[i]->name)+1;
    wijzer = char_ptr;

    insert = Malloc(sizeof(hash_section));
    insert->node.key = obj->note[i];
    insert->nr=section_count++;
    HashInsert(tabel, insert);

    i++;
    aantal_secties++;
  }

  i=0;
  while(i<obj->nbsss){
    char_ptr = (char *) wijzer;
    strcpy(char_ptr,obj->bss[i]->name);
    char_ptr += strlen(obj->bss[i]->name)+1;
    wijzer = char_ptr;
    
    insert = Malloc(sizeof(hash_section));
    insert->node.key = obj->bss[i];
    insert->nr=section_count++;
    HashInsert(tabel, insert);

    i++;
    aantal_secties++;
  }

  *((t_uint32 *)hulp)=aantal_secties;
  return wijzer;
}
/* }}} */
/* SaveBlockInfo {{{ */
void * SaveBlockInfo(data_hash_table * tabel, t_object * obj, void * wijzer, int phase)
{
  t_object * sub_obj, *tmp;
  int aantal_sub_obj=0;
  void * hulp;
  t_uint32 * t_uint32_ptr;
  
  wijzer = SaveSectionsParent(tabel, obj, wijzer);
  
  hulp = wijzer;
  t_uint32_ptr = wijzer;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;

  OBJECT_FOREACH_SUBOBJECT(obj,sub_obj,tmp)
  {
    wijzer = SaveSubObject(tabel, sub_obj, wijzer, phase);
    aantal_sub_obj++;
  }
  
  *((t_uint32 *)hulp)=aantal_sub_obj;
  return wijzer;
}
/* }}} */
/* SaveCodeDataBounds {{{ */
void * SaveCodeDataBounds(data_hash_table * tabel, t_object * obj, void * wijzer)
{
  void * hulp;
  int aantal_symbols=0;
  t_symbol_he * symbool=HashLookup(obj->sub_symbol_table->hash_table, "$first");
  t_address * t_address_ptr;
  t_uint32 * t_uint32_ptr;

  t_address_ptr = (t_address *)wijzer;
  if(symbool)
    *t_address_ptr=symbool->sym->current_address;
  else
    *t_address_ptr=0;
  t_address_ptr ++;
  wijzer = t_address_ptr;

  symbool=HashLookup(obj->sub_symbol_table->hash_table, "$switch");

  hulp=wijzer;
  t_uint32_ptr = (t_uint32 *) wijzer;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;

  while(symbool)
  {

    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=symbool->sym->offset_from_start;
    t_address_ptr ++;
    wijzer = t_address_ptr;
    
    aantal_symbols ++;
    symbool = (t_symbol_he *)symbool->node.equal;
  }

  *((t_uint32 *)hulp)=aantal_symbols;
  
  aantal_symbols=0;
  symbool=HashLookup(obj->sub_symbol_table->hash_table, "$a");

  hulp=wijzer;
  t_uint32_ptr = (t_uint32 *) wijzer;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;

  while(symbool)
  {
    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=symbool->sym->offset_from_start;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=symbool->sym->current_address;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_uint32_ptr = (t_uint32 *) wijzer;
    *t_uint32_ptr=((hash_section *)HashLookup(tabel, symbool->sym->base))->nr;
    t_uint32_ptr ++;
    wijzer = t_uint32_ptr;
    
    t_uint32_ptr = (t_uint32 *) wijzer;
    *t_uint32_ptr=symbool->sym->to_type;
    t_uint32_ptr ++;
    wijzer = t_uint32_ptr;
    
    aantal_symbols ++;
    symbool = (t_symbol_he *)symbool->node.equal;
  }
  *((t_uint32 *)hulp)=aantal_symbols;
  
  aantal_symbols=0;
  symbool=HashLookup(obj->sub_symbol_table->hash_table, "$d");

  hulp=wijzer;
  t_uint32_ptr = (t_uint32 *) wijzer;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;

  while(symbool)
  {
    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=symbool->sym->offset_from_start;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=symbool->sym->current_address;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_uint32_ptr = (t_uint32 *) wijzer;
    *t_uint32_ptr=((hash_section *)HashLookup(tabel, symbool->sym->base))->nr;
    t_uint32_ptr ++;
    wijzer = t_uint32_ptr;
    
    t_uint32_ptr = (t_uint32 *) wijzer;
    *t_uint32_ptr=symbool->sym->to_type;
    t_uint32_ptr ++;
    wijzer = t_uint32_ptr;
    
    aantal_symbols ++;
    symbool = (t_symbol_he *)symbool->node.equal;
  }

  *((t_uint32 *)hulp)=aantal_symbols;
  
  return wijzer;
}
/* }}} */
/* SaveCodeDataBounds2 {{{ */
void * SaveCodeDataBounds2(data_hash_table * tabel, t_object * obj, void * wijzer)
{
  void * hulp;
  t_uint32 aantal=0;
  
  t_ins * ins;
  t_address * t_address_ptr;
  t_uint32 * t_uint32_ptr;

  hulp=wijzer;
  t_uint32_ptr = (t_uint32 *) wijzer;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;
  
    if(INS_IS_DATA(obj->code[0]->data))
    {
      ins = obj->code[0]->data;

      aantal++;

      t_address_ptr = (t_address *)wijzer;
      *t_address_ptr=INS_CADDRESS(ins);
      t_address_ptr ++;
      wijzer = t_address_ptr;
      
      /*FATAL(("code starts with data, not taken into account."));*/
    }
    for (ins = obj->code[0]->data; INS_NEXT(ins); ins = INS_NEXT(ins))
    {
      if(INS_IS_DATA(ins) != INS_IS_DATA(INS_NEXT(ins)))
      {
	aantal++;
	
	t_address_ptr = (t_address *)wijzer;
	*t_address_ptr=INS_CADDRESS(INS_NEXT(ins));
	t_address_ptr ++;
	wijzer = t_address_ptr;
	
      }
    }
  
  *((t_uint32 *)hulp)=aantal;

  return wijzer;
}
/* }}} */
/* SaveRelocationInfo {{{ */
void * SaveRelocationInfo(data_hash_table * tabel, t_object * obj, void * wijzer)
{
  t_reloc * relptr;
  void * hulp;
  int i=0;
  t_address from_offset, to_offset;
  t_section * from_section, * to_section;
  t_uint32 * t_uint32_ptr;
  t_address * t_address_ptr;
  t_bool * t_bool_ptr;
  char * char_ptr;
  hash_section * item;
  
  hulp = wijzer;
  t_uint32_ptr = (t_uint32 *) wijzer;
  t_uint32_ptr++;
  wijzer = t_uint32_ptr;

  for (relptr=obj->reloc_table->first; relptr!=NULL; relptr=REL_NEXT(relptr))
  {
    
    from_offset = relptr->from_offset;
    to_offset = relptr->to_offset;
    
    if(REL_FROM(&(relptr->base_edge))->type == RT_INS)
    {
      from_section = INS_SECTION(REL_FROM(&(relptr->base_edge)));
      if(from_section ==NULL) FATAL(("hello"));
      
      if(INS_SECTION(REL_FROM(&(relptr->base_edge)))!=NULL)
      {
        from_offset = GENERIC_ADD(GENERIC_SUB(((t_ins *)REL_FROM(&(relptr->base_edge)))->relocatable.caddress,INS_SECTION(REL_FROM(&(relptr->base_edge)))->relocatable.caddress),from_offset);
      }
      else{
	FATAL(("hello"));
	exit(0);
      }
    }
    else if(REL_FROM(&(relptr->base_edge))->type == RT_SECTION||REL_FROM(&(relptr->base_edge))->type == RT_SUBSECTION)
    {
      from_section = (t_section *)REL_FROM(&(relptr->base_edge));
    }
    else 
    {
      DiabloPrint(stdout,"@R\n",relptr); 
      FATAL(("hello"));
      exit(0);
    }
    
    if(REL_TO_RELOCATABLE(&(relptr->base_edge))->type == RT_INS)
    {
      to_section = INS_SECTION(REL_TO_RELOCATABLE(&(relptr->base_edge)));
      if(to_section ==NULL) FATAL(("hello"));
      
      if(INS_SECTION(REL_TO_RELOCATABLE(&(relptr->base_edge)))!=NULL)
      {
        to_offset = GENERIC_ADD(GENERIC_SUB(((t_ins *)REL_TO_RELOCATABLE(&(relptr->base_edge)))->relocatable.caddress,INS_SECTION(REL_TO_RELOCATABLE(&(relptr->base_edge)))->relocatable.caddress),to_offset);
      }
      else{
	FATAL(("hello"));
	exit(0);
      }
    }
    else if(REL_TO_RELOCATABLE(&(relptr->base_edge))->type == RT_SECTION||REL_TO_RELOCATABLE(&(relptr->base_edge))->type == RT_SUBSECTION)
    {
      to_section = (t_section *)REL_TO_RELOCATABLE(&(relptr->base_edge));
    }
    else 
    {
      FATAL(("hello"));
      exit(0);
    }
    
    
    t_uint32_ptr = (t_uint32*) wijzer;
    if(!strcmp(from_section->name,"*ABS*"))
      *t_uint32_ptr=1;
    else if(!strcmp(from_section->name,"*UNDEF*"))
      *t_uint32_ptr=2;
    else {
      item = HashLookup(tabel, from_section);
      if(item==NULL)
	FATAL(("Relocatable not found\n"));
      *t_uint32_ptr=item->nr;
    }
    t_uint32_ptr++;
    wijzer = t_uint32_ptr;


    t_uint32_ptr = (t_uint32*) wijzer;
    if(!strcmp(to_section->name,"*ABS*"))
      *t_uint32_ptr=1;
    else if(!strcmp(to_section->name,"*UNDEF*"))
      *t_uint32_ptr=2;
    else {
      item = HashLookup(tabel, to_section);
      if(item==NULL)
	FATAL(("Relocatable not found\n"));
      *t_uint32_ptr=item->nr;
    }
    t_uint32_ptr++;
    wijzer = t_uint32_ptr;
     
    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=relptr->addend;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=from_offset;
    t_address_ptr++;
    wijzer = t_address_ptr;


    if(relptr->to_symbol==TRUE)
      FATAL(("to_symbol should be false"));

    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=to_offset;
    t_address_ptr++;
    wijzer = t_address_ptr;
    
    t_bool_ptr = (t_bool *)wijzer;
    *t_bool_ptr=relptr->hell;
    t_bool_ptr++;
    wijzer = t_bool_ptr;

    if(relptr->e_to_symbol==TRUE) 
      FATAL(("e_to_symbol should not be TRUE"));
    
    if(relptr->e.sec==NULL)
      *((t_uint32 *)wijzer)=0;
    else
    {
      if(((t_section*)relptr->e.sec)->name=="*ABS*")
	*((t_uint32 *)wijzer)=1;
      else if(((t_section*)relptr->e.sec)->name=="*UNDEF*")
	*((t_uint32 *)wijzer)=2;
      else
      {	
	item = HashLookup(tabel, relptr->e.sec);
	if(item==NULL)
	  FATAL(("Relocatable not found\n"));
	*((t_uint32 *)wijzer)=item->nr;
      }
    }
    t_uint32_ptr = (t_uint32 *) wijzer;
    t_uint32_ptr++;
    wijzer = t_uint32_ptr;

    
    t_address_ptr = (t_address *)wijzer;
    *t_address_ptr=relptr->e_offset;
    t_address_ptr ++;
    wijzer = t_address_ptr;

    char_ptr = (char *) wijzer;
    strcpy(char_ptr ,relptr->code);
    char_ptr += strlen(relptr->code)+1;
    wijzer = char_ptr;

    t_bool_ptr = (t_bool *)wijzer;
    *t_bool_ptr=relptr->first;
    t_bool_ptr++;
    wijzer = t_bool_ptr;
    
    
    i++;
  }
  *((t_uint32 *)hulp)=i;
  return wijzer;
}
/* }}} */
/* SaveState {{{ */
void SaveState(t_object * obj, int phase)
{
#ifndef INSTRUMENTATION
  data_hash_table * tabel = HashTableNew(100, 0, pointer_hash, pointer_cmp, HashSectionFree);
  void * wijzer; 
  t_section * sectie;
  t_linker * t_linker_ptr;
  t_uint32 * t_uint32_ptr;
  char * end, * begin;
  section_count=3;

  
  /* Aanmaken .comment sectie*/
  if(!global_options.restoreState)
  {
    sectie = SectionNew(obj, DATA_SECTION);
    T_RELOCATABLE(sectie)->type=RT_SECTION;  
    SECTION_NAME(sectie)=StringDup(".comment");
    obj->data = Realloc(obj->data,(obj->ndatas+1)*sizeof(t_section *));
    obj->data[obj->ndatas]=sectie;
    obj->ndatas++;
  }
  else 
  {
    if(phase==SAVESTATENOTDISASSEMBLED)
      return;
    sectie = SectionGetFromObjectByName(obj, ".comment");
  }

  /*volgende drie lijnen te verschonen FIXME!*/
  SECTION_CSIZE(sectie)=2048*2048;
  sectie->data = Malloc(2048*2048);

  sectie->relocatable.caddress=CalculateCommentAddress(obj);

  
  /*Save Linker*/
  wijzer=sectie->data;

  t_linker_ptr = (t_linker *)wijzer;
  *t_linker_ptr=obj->linker;
  t_linker_ptr ++;
  wijzer = t_linker_ptr;
  
  /*Save Block Information*/
  wijzer = SaveBlockInfo(tabel,obj,wijzer,phase); 
 
  t_uint32_ptr = (t_uint32*) wijzer;
  *t_uint32_ptr = phase;
  t_uint32_ptr ++;
  wijzer = t_uint32_ptr;

  /*Save Relocation Information*/
  wijzer = SaveRelocationInfo(tabel, obj, wijzer);

  /*Save Code\Data Boundaries*/
  
  if(phase == SAVESTATEDISASSEMBLED)
    wijzer = SaveCodeDataBounds2(tabel, obj, wijzer);
  else if(phase == SAVESTATENOTDISASSEMBLED)
    wijzer = SaveCodeDataBounds(tabel, obj, wijzer);
  else FATAL(("unknown phase"));

  /*Teruggeven teveel aan geheugen*/
  end = (char*)wijzer;
  begin = (char*)sectie->data;
  sectie->data = Realloc(sectie->data, end - begin);
  SECTION_CSIZE(sectie)=end-begin;

  HashFree(tabel);
#endif
}
/* }}} */

/*
t_map_node * CreateMapNode(t_address base_addr, t_address sec_size, t_string object, t_string sec_name, t_map_section_type type, t_map_section_attribute attr, t_object * obj)
{
  t_map_node * ret = Malloc(sizeof(t_map_node));
  ret->base_addr = base_addr;
  ret->sec_size = sec_size;
  ret->object = StringDup(object);
  ret->sec_name = StringDup(sec_name);
  ret->type = type;
  ret->attr = attr;
  ret->obj = obj;
  ret->idx=0;
  ret->mapped_sec_name = NULL;

  return ret;
}*/

/* vim: set shiftwidth=2 foldmethod=marker: */
