/* {{{ Includes */
#include "diablo_schedule.h"
#include "diablo_function.h"
#include "diablo_iterators.h"
#include "diablo_bbl.h"
#include "diablo_print.h"
#include "../utils/util_error.h"
#include "../utils/util_status.h"

#ifdef ARM_SUPPORT
#include "../arch/arm/arm_registers.h"
#endif

#ifdef I386_SUPPORT
#include "../arch/i386/i386_description.h"
#include "../arch/i386/i386_instructions.h"
#endif
/* }}} */

void SchedulerUpdateAbsoluteTime(t_scheduler_dagnode * node);

/* temporary selection function, this should become some kind of arch-specific callback *//*{{{*/
t_scheduler_dagnode * SelectionFunction(t_scheduler_dag * dag, t_scheduler_list * list, t_uint32 cycle, t_bool has_pc_use)
{
#ifdef ARM_SUPPORT
  t_node * node;
  t_node * ret = NULL/* , *alter = NULL */;
  t_edge * i_edge;
  t_int32 weight = 0/* , slack = list->cycles */;
  t_int32 maxweight = -8000;
  t_int32 nedges = 0, max_nedges = 0, pred_edges=0, min_pred_edges = 0;
  t_int32 min_deadline = list->cycles;
  t_bool constrained = FALSE;
  int i;
  t_uint32 deadlines[2048] = {0};
  t_bool must_follow_deadline = FALSE;

  printf("Selecting for cycle %d\n",cycle);
  SchedulerListPrint(list);
#if 1
  /* When the block contains instructions that use the pc, we should do some bookkeeping to see if we are going to
   * create a valid schedule */
  if(has_pc_use)
  {
    GRAPH_FOREACH_NODE(dag,node)
    {
      if(DAGNODE_FLAGS(node) & SCHED_SCHEDULED) continue;

      /* Add 1 to all array elements from deadline to last */
      for (i=DAGNODE_DEADLINE(node)-cycle;i<list->cycles-cycle;i++)
	deadlines[i]++;
    }
    
    /*deadlines[list->cycles-cycle-1] = 0; */
    DiabloPrint(stdout,"@B\n",list->bbl);
    for(i = 0; i < list->cycles-cycle; i++)
      printf("Deadlines[%d] = %d\n",i,deadlines[i]);

    printf("<------------------------------------->\n");
    /* When a array element has a count higher than its position in the array, we MUST follow first deadline or else
     * a valid schedule will not be created */
    for (i=0;i<list->cycles-cycle;i++)
      if (deadlines[i]>i)
      { 
	printf("TRUE %d\n",i);
	must_follow_deadline = TRUE;
	break;
      }
  }
#endif

  GRAPH_FOREACH_NODE(dag,node)
  {
    if(DAGNODE_FLAGS(node) & SCHED_SCHEDULED) continue;
/*    if(DAGNODE_START(node) == DAGNODE_DEADLINE(node)) */

    if(DAGNODE_DEPENDENCYCOUNT(node) == 0)
      /*if(DAGNODE_START(node) <= cycle && DAGNODE_DEADLINE(node) >= cycle)*/
    {
      weight = 0;
      nedges = 0;
      pred_edges = 0;

      if(ret) DiabloPrint(stdout,"Current candidate = @I\n",DAGNODE_INS(ret));
      if(DAGNODE_DEADLINE(node) == cycle)
      {
	if(DAGNODE_INS(node))
	DiabloPrint(stdout,"Must take @I because deadline reached\n",DAGNODE_INS(node));
	return T_DAGNODE(node);
      }
      NODE_FOREACH_SUCC_EDGE(node,i_edge)
      {
	weight += (DAGEDGE_DEPENDENCY_TYPE(i_edge) == DEP_RAW)?(INS_TYPE(DAGNODE_INS(node))==IT_LOAD?4:3):(DAGEDGE_DEPENDENCY_TYPE(i_edge) == DEP_RAR)?0:1;/*DAGEDGE_MINWEIGHT(i_edge);*/
	if(DAGEDGE_DEPENDENCY_TYPE(i_edge) != DEP_RAR) nedges++;
      }

      NODE_FOREACH_PRED_EDGE(node,i_edge)
      {
	if(DAGEDGE_DEPENDENCY_TYPE(i_edge) == DEP_RAW && DAGNODE_START(EDGE_HEAD(i_edge)) == (cycle -1))
	{
	  if (INS_TYPE(DAGNODE_INS(EDGE_HEAD(i_edge)))==IT_LOAD)
	  {
	    printf("ok1\n");
	    weight -=3;
	  }
	  printf("ok2\n");
	  weight -=6;
	}
	else
	  weight -= (DAGEDGE_DEPENDENCY_TYPE(i_edge) == DEP_RAR)?0:1;
	if(DAGEDGE_DEPENDENCY_TYPE(i_edge) != DEP_RAR) pred_edges++;
      }
      /*slack = DAGNODE_DEADLINE(node) - DAGNODE_START(node);*/

      if(DAGNODE_DEADLINE(node) != DE_INFINITY)
      {
	if (must_follow_deadline)
	{
	  constrained = TRUE;
	  if(DAGNODE_DEADLINE(node) < min_deadline)
	  {
	    ret = node;
	    min_deadline = DAGNODE_DEADLINE(node);
	    maxweight = weight;
	    max_nedges = nedges;
	    min_pred_edges = pred_edges;
	  }
	  else if(DAGNODE_DEADLINE(node) == min_deadline)
	  {
	    if(pred_edges < min_pred_edges)
	    {
	      ret = node;
	      min_deadline = DAGNODE_DEADLINE(node);
	      maxweight = weight;
	      max_nedges = nedges;
	      min_pred_edges = pred_edges;
	    }
	  }
	}
	else
	{
	  constrained = FALSE;
	  weight-=DAGNODE_DEADLINE(node);
	}
      }
      if(!constrained)
      {
	/*NODE_FOREACH_SUCC_EDGE(node,i_edge)
	  {
	  weight += (DAGEDGE_DEPENDENCY_TYPE(i_edge) == DEP_RAW)?3:(DAGEDGE_DEPENDENCY_TYPE(i_edge) == DEP_RAR)?0:1;//DAGEDGE_MINWEIGHT(i_edge);
	  if(DAGEDGE_DEPENDENCY_TYPE(i_edge) != DEP_RAR) nedges++;
	  }*/
	if(weight > maxweight)
	{
	  ret = node;
	  maxweight = weight;
	  max_nedges = nedges;
	  min_pred_edges = pred_edges;
	}
	else if (weight == maxweight && nedges > max_nedges)
	{
	  if(!(nedges == max_nedges && pred_edges > min_pred_edges))
	  {
	    ret = node;
	    max_nedges = nedges;
	    min_pred_edges = pred_edges;
	  }
	}
      }
    }
  }
  return T_DAGNODE(ret);
#endif
  return NULL;
}/*}}}*/

void SchedulerCfgListSchedule(t_cfg * cfg)/*{{{*/
{
  t_function * fun;
  t_bbl * bbl;

  CFG_FOREACH_FUN(cfg,fun)
    FUN_FOREACH_BBL(fun,bbl)
    if(bbl->nins)
      SchedulerBblListSchedule(bbl);
}/*}}}*/


void SchedulerApplySchedule(t_scheduler_list * list);
void SchedulerBblListSchedule(t_bbl * bbl)/*{{{*/
{
  t_uint32 i;
  t_scheduler_dag * dag = SchedulerDagNew();
  t_scheduler_list * list = SchedulerListNew();
/*  static t_uint32 teller = 0;*/

  /*t_scheduler_listitem * item;*/
  t_ins * i_ins;

  list->bbl = bbl;

  SchedulerDagBuildForBbl(bbl,dag,NullRegs);

  SchedulerDagPrint(dag,bbl, FALSE);

  SchedulerGenerateListFromDag(list,dag,SelectionFunction,bbl);

  if(!list->schedule[0] || list->schedule[0]->node->ins)
    FATAL(("Should be a start node here!"));

  i = 1;

/*  if(global_options.debugcounter > teller)*/
    BBL_FOREACH_INS(bbl,i_ins)
    {
      if(list->schedule[i] && list->schedule[i]->node->ins != i_ins)
      {
	/*SchedulerListPrint(list);*/
	DiabloPrint(stdout,"Before: @iB\n",bbl);
	SchedulerApplySchedule(list);
	DiabloPrint(stdout,"After: @iB\n",bbl);
/*        teller++;*/
	break;
      }
      i++;
    }

  SchedulerDagFree(dag);
  SchedulerListFree(list);
}/*}}}*/



void SchedulerDagBuildForBbl(t_bbl * bbl, t_scheduler_dag * dag, t_regset exclude_set)/*{{{*/
{
  t_ins* ins, *iter_ins;
  t_node * node;
  t_regset currUsed, tmpUsed;
  t_regset currDefs, tmpDefs;
#ifdef ARM_SUPPORT
  t_reg pc = BBL_OBJECT(bbl)->description->program_counter;
#endif
  t_scheduler_dagedge * ret_edge;
  t_bool store_treated;

#ifdef I386_SUPPORT
  t_regset live = BBL_REGS_LIVE_OUT(bbl);
#endif
  /* create the dummy nodes */
  DAG_SOURCE(dag) = SchedulerDagNodeNew(dag,NULL);
  DAG_SINK(dag) = SchedulerDagNodeNew(dag,NULL);
  
  BBL_FOREACH_INS(bbl,ins)
  {
    SchedulerDagNodeNew(dag,ins);
  }

  
  BBL_FOREACH_INS_R(bbl,ins)
  {
    store_treated = FALSE;
    
    tmpUsed = currUsed = INS_REGS_USE(ins);
    tmpDefs = currDefs = INS_REGS_DEF(ins);
#ifdef I386_SUPPORT
    RegsetSetIntersect(tmpDefs, live);

    if(!I386InsIsConditional(ins))     
      RegsetSetDiff(live,currDefs);
    RegsetSetUnion(live,currUsed);
#endif


    /* Make sure that branches stay at the end of the bbl */
    /* TODO: Add edges for memory dependencies 
     */
    if(INS_OBJECT(ins)->actions->IsControlflow(ins))
    {
      iter_ins = INS_PREV(ins);
      while(iter_ins)
      {
	ret_edge = SchedulerDagEdgeNew(dag,iter_ins,ins,DEP_RAR);
	/* Need an arch call to determine the weights */
	DAGEDGE_MINWEIGHT(ret_edge) = 0;
	DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
	iter_ins = INS_PREV(iter_ins);
      }
      /*ret_edge = SchedulerDagEdgeNewToSink(dag,ins,DEP_RAR);
      DAGEDGE_MINWEIGHT(ret_edge) = 1;
      DAGEDGE_MAXWEIGHT(ret_edge) = 1;*/
    }    
#ifdef ARM_SUPPORT /*{{{*/
    /* Make sure that instructions that use the pc never get moved!!!*/
    if(RegsetIn(tmpUsed,pc))
    {
      ret_edge = SchedulerDagEdgeNewFromSource(dag,ins,DEP_RAR);
      DAGEDGE_MINWEIGHT(ret_edge) = DAGEDGE_MAXWEIGHT(ret_edge) = 1;
      
      iter_ins = INS_PREV(ins);
      while(iter_ins)
      {
	DAGEDGE_MINWEIGHT(ret_edge)++;
	iter_ins = INS_PREV(iter_ins);
      }
      DAGEDGE_MAXWEIGHT(ret_edge) = DAGEDGE_MINWEIGHT(ret_edge);

      ret_edge = SchedulerDagEdgeNewToSink(dag,ins,DEP_RAR);
      DAGEDGE_MINWEIGHT(ret_edge) = DAGEDGE_MAXWEIGHT(ret_edge) = 1;

      iter_ins = INS_NEXT(ins);
      while(iter_ins)
      {
	DAGEDGE_MINWEIGHT(ret_edge)++;
	iter_ins = INS_NEXT(iter_ins);
      }
      DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
    }
#endif /*}}}*/

    iter_ins = INS_PREV(ins);
    RegsetSetDiff(tmpUsed,exclude_set);
    while(iter_ins)
    {
      if(!RegsetIsMutualExclusive(tmpUsed,INS_REGS_DEF(iter_ins)))
      {
	/* Insert a RAW dependency, remove the defined regs from tmpSet, because we only have to
	 * insert an edge once for a RAW*/
	ret_edge = SchedulerDagEdgeNew(dag,iter_ins,ins,DEP_RAW);
	/* Need an arch call to determine the weights */
	DAGEDGE_MINWEIGHT(ret_edge) = 1;
	DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
	RegsetSetDiff(tmpUsed,INS_REGS_DEF(iter_ins));
      }
      iter_ins = INS_PREV(iter_ins);
      if(RegsetIsEmpty(tmpUsed)) break;
    }

    iter_ins = INS_PREV(ins);
    RegsetSetDiff(tmpDefs,exclude_set);
    while(iter_ins)
    {
      if(!RegsetIsMutualExclusive(tmpDefs,INS_REGS_DEF(iter_ins)))
      {
	ret_edge = SchedulerDagEdgeNew(dag,iter_ins,ins,DEP_WAW);
	/* Need an arch call to determine the weights */
	DAGEDGE_MINWEIGHT(ret_edge) = 1;
	DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
#ifndef I386_SUPPORT
	RegsetSetDiff(tmpDefs,INS_REGS_DEF(iter_ins));
#endif
      }
      if(!RegsetIsMutualExclusive(currDefs,INS_REGS_USE(iter_ins)))
      {
	ret_edge = SchedulerDagEdgeNew(dag,iter_ins,ins,DEP_WAR);
	/* Need an arch call to determine the weights */
	DAGEDGE_MINWEIGHT(ret_edge) = 1;
	DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
      }

      iter_ins = INS_PREV(iter_ins);
    }
    if(INS_TYPE(ins) == IT_STORE || INS_TYPE(ins) == IT_FLT_STORE || INS_TYPE(ins) == IT_STORE_MULTIPLE
#ifdef I386_SUPPORT
	|| I386InsIsStore(ins)
#endif
	)
    {
      iter_ins = INS_PREV(ins);
      while(iter_ins)
      {
	if(!store_treated && (INS_TYPE(iter_ins) == IT_STORE || INS_TYPE(iter_ins) == IT_FLT_STORE || INS_TYPE(iter_ins) == IT_STORE_MULTIPLE
#ifdef I386_SUPPORT
	|| I386InsIsStore(iter_ins)
#endif
	      ))
	{
	  ret_edge = SchedulerDagEdgeNew(dag,iter_ins,ins,DEP_WAW);
	  DAGEDGE_MINWEIGHT(ret_edge) = 1;
	  DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
	  store_treated = TRUE;
	}
	if(INS_TYPE(iter_ins) == IT_LOAD || INS_TYPE(iter_ins) == IT_FLT_LOAD || INS_TYPE(iter_ins) == IT_LOAD_MULTIPLE
#ifdef I386_SUPPORT
	|| I386InsIsLoad(iter_ins)
#endif
	    )
	{
	  ret_edge = SchedulerDagEdgeNew(dag,iter_ins,ins,DEP_WAR);
	  DAGEDGE_MINWEIGHT(ret_edge) = 1;
	  DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
	}
        iter_ins = INS_PREV(iter_ins);
      }
    }
    if(INS_TYPE(ins) == IT_LOAD || INS_TYPE(ins) == IT_FLT_LOAD || INS_TYPE(ins) == IT_LOAD_MULTIPLE
#ifdef I386_SUPPORT
	|| I386InsIsLoad(ins)
#endif
	)
    {
      iter_ins = INS_PREV(ins);
      while(iter_ins)
      {
	if(INS_TYPE(iter_ins) == IT_STORE || INS_TYPE(iter_ins) == IT_FLT_STORE || INS_TYPE(iter_ins) == IT_STORE_MULTIPLE
#ifdef I386_SUPPORT
	|| I386InsIsStore(iter_ins)
#endif
	    )
	{
	  ret_edge = SchedulerDagEdgeNew(dag,iter_ins,ins,DEP_RAW);
	  DAGEDGE_MINWEIGHT(ret_edge) = 1;
	  DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
	  break;
	}
        iter_ins = INS_PREV(iter_ins);
      }
    }
  }

  /* For node with no successor or predecessor, add edge to the source and sink */
  GRAPH_FOREACH_NODE(dag,node)
  {
    if(!NODE_SUCC_FIRST(node) && T_DAGNODE(node) != DAG_SINK(dag))
    {
      ret_edge = SchedulerDagEdgeNewToSink(dag,DAGNODE_INS(node),DEP_RAR);

      if (T_DAGNODE(node) != DAG_SOURCE(dag))
	{	
	  DAGEDGE_MINWEIGHT(ret_edge) = 1;
	  DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
	}
      else
	{
	  DAGEDGE_MINWEIGHT(ret_edge) = BBL_NINS(bbl)+1;
	  DAGEDGE_MAXWEIGHT(ret_edge) = BBL_NINS(bbl)+1;
	}
    }
    if(!NODE_PRED_FIRST(node) && T_DAGNODE(node) != DAG_SOURCE(dag))
    {
      ret_edge = SchedulerDagEdgeNewFromSource(dag,DAGNODE_INS(node),DEP_RAR);
      DAGEDGE_MINWEIGHT(ret_edge) = 1;
      DAGEDGE_MAXWEIGHT(ret_edge) = DE_INFINITY;
    }
  }
  
  GRAPH_FOREACH_NODE(dag,node)
    SchedulerDagNodeCalcHeight(T_DAGNODE(node));
    
}/*}}}*/

t_uint32 SchedulerGenerateListFromDag(t_scheduler_list * list, t_scheduler_dag * dag, SchedulerSelectionFunction SelectFunc/* een soort van selectiefunctie meegeven? */, t_bbl * bbl)/*{{{*/
{
#ifdef ARM_SUPPORT
  t_node * node;
  t_scheduler_dagnode * node_to_schedule;
  t_edge * i_edge;
  t_uint32 curr_cycle = 0;
  t_uint32 nnodes = 0;
  t_int32 last_pc_use = BBL_NINS(list->bbl);
  t_bool instructions_ready = FALSE;
  t_ins * ins;

  BBL_FOREACH_INS_R(bbl,ins)
  {
    if(RegsetIn(INS_REGS_USE(ins), ARM_REG_R15)) break;
    last_pc_use--;
  }
  DiabloPrint(stdout,"Last pc use in @iB at cycle %d\n",list->bbl,last_pc_use);
  

  DAGNODE_FLAGS(DAG_SOURCE(dag)) |= SCHED_SCHEDULED;
  NODE_FOREACH_SUCC_EDGE(DAG_SOURCE(dag),i_edge)
    DAGNODE_DEPENDENCYCOUNT(EDGE_TAIL(i_edge))--;

  instructions_ready = SchedulerDagUpdateIntervalsAndReadyList(dag,DAG_SOURCE(dag),curr_cycle);

  GRAPH_FOREACH_NODE(dag,node)
    nnodes++;

  /* Alloc nodes to create list */
  list->schedule = Calloc(nnodes,sizeof(t_scheduler_listitem * ));
  list->cycles = nnodes;

  
  SchedulerListInsertNode(list,DAG_SOURCE(dag));

  curr_cycle = 1;

  while(instructions_ready)
  {
    instructions_ready = FALSE;
    
    /* select the next node to schedule in current cycle */
    /* If it is possible we could select the next instruction to schedule using a callback SelectFunc? */
    node_to_schedule = SelectFunc(dag,list,curr_cycle,(curr_cycle<=last_pc_use));

    if(!node_to_schedule) break;
    else
    {
      t_uint32 i, j = 0;

      /* Here we should arch-dependend determine if the selected node can be scheduled */
      for(i=DAGNODE_START(node_to_schedule); i<=DAGNODE_DEADLINE(node_to_schedule); i++)
      {
	if(!list->schedule[i]) {j=i;break;}
      }
      if(j)
      {
	DAGNODE_START(node_to_schedule) = j;
	DAGNODE_DEADLINE(node_to_schedule) = j;
      }
      else
      {
	VERBOSE(0,("@iB",bbl));
	SchedulerListPrint(list);
        SchedulerDagPrint(dag,bbl,curr_cycle);
	FATAL(("Can not schedule node @I! Start = %d, deadline = %d",DAGNODE_INS(node_to_schedule),DAGNODE_START(node_to_schedule),DAGNODE_DEADLINE(node_to_schedule)));
      }

      /* Put it in the schedule and do some bookkeeping */
      SchedulerListInsertNode(list,node_to_schedule);
      DAGNODE_FLAGS(node_to_schedule) |= SCHED_SCHEDULED;
      NODE_FOREACH_SUCC_EDGE(node_to_schedule,i_edge)
	DAGNODE_DEPENDENCYCOUNT(EDGE_TAIL(i_edge))--;
    }

    /* Look for more ready instructions */
    /* TODO : implement a deadline-update function that returns TRUE if there are ready instructions left */
    instructions_ready = SchedulerDagUpdateIntervalsAndReadyList(dag,node_to_schedule,DAGNODE_START(node_to_schedule));
    curr_cycle++;
  }
  return curr_cycle;
#endif
  return 0;
}/*}}}*/

/* Create a new dag for scheduling a bbl */ /*{{{*/
t_scheduler_dag * SchedulerDagNew(void)
{
  t_scheduler_dag * ret = Calloc(1,sizeof(t_scheduler_dag));
#ifdef STEGO
  ret->nr_scheduled=0;
#endif
  GRAPH_NODE_SIZE(ret) = sizeof(t_scheduler_dagnode);
  GRAPH_EDGE_SIZE(ret) = sizeof(t_scheduler_dagedge);

  return ret;
}/*}}}*/

void SchedulerDagFree(t_scheduler_dag * dag)/*{{{*/
{
  t_node * node, * tmp_node;
  t_edge * edge, * tmp_edge;

  GRAPH_FOREACH_NODE_SAFE(dag,node,tmp_node)
    Free(node);
  GRAPH_FOREACH_EDGE_SAFE(dag,edge,tmp_edge)
    Free(edge);
  
  Free(dag);
}/*}}}*/

t_scheduler_list * SchedulerListNew(void)/*{{{*/
{
  t_scheduler_list * list = (t_scheduler_list *) Calloc(1,sizeof(t_scheduler_list));
  return list; 
}/*}}}*/

void SchedulerListFree(t_scheduler_list * list)/*{{{*/
{
  t_scheduler_listitem * free;
  t_uint32 i;
 
  /*DEBUGSchedulerListPrint(list,NULL);*/
  for(i = 0; i < list->cycles; i++)
  {
    free = list->schedule[i];
    while(free)
    {
      list->schedule[i] = list->schedule[i]->next;
      Free(free);
      free = list->schedule[i];
    }
  }

  Free(list->schedule);
  Free(list);

}/*}}}*/

/* Add a new node to the dag containing instruction ins */ /*{{{*/
t_scheduler_dagnode * SchedulerDagNodeNew(t_scheduler_dag * dag, t_ins *ins)
{
  t_scheduler_dagnode * ret = T_DAGNODE(GraphNewNode(T_GRAPH(dag),0));

  DAGNODE_INS(ret) = ins;
  DAGNODE_START(ret) = 0;
  DAGNODE_DEADLINE(ret) = DE_INFINITY;
  DAGNODE_FLAGS(ret) = 0;
  DAGNODE_DEPENDENCYCOUNT(ret) = 0;
  
  return ret;
}/*}}}*/

/* Add a new edge to the dag, from head to tail, with dependency given as arguments */ /*{{{*/
t_scheduler_dagedge * SchedulerDagEdgeNew(t_scheduler_dag * dag, t_ins * head, t_ins * tail, t_scheduler_dependency dependency)
{
  t_scheduler_dagedge * ret;
  t_node * node;
  t_node * head_node = NULL, * tail_node = NULL;
  t_edge * i_edge;

  /* First check if the node are already in the dag */
  GRAPH_FOREACH_NODE(dag,node)
  {
    if(DAGNODE_INS(node) == head) head_node = node;
    if(DAGNODE_INS(node) == tail) tail_node = node;

    if(tail_node && head_node) break;
  }
  if(!head_node && !tail_node) FATAL(("Head or tail not yet in dag!"));
  
  NODE_FOREACH_SUCC_EDGE(head_node,i_edge)
    if(EDGE_TAIL(i_edge) == tail_node)
    {
      if(DAGEDGE_DEPENDENCY_TYPE(i_edge) > dependency) DAGEDGE_DEPENDENCY_TYPE(i_edge) = dependency;
      return T_DAGEDGE(i_edge);
    }

  /* Create the new edge */
  ret = T_DAGEDGE(GraphNewEdge(T_GRAPH(dag),head_node,tail_node,dependency));
  DAGNODE_DEPENDENCYCOUNT(tail_node)++;

  return ret;
}/*}}}*/

/* Add a new edge to the dag, from the source to tail, with dependency given as arguments */ /*{{{*/
t_scheduler_dagedge * SchedulerDagEdgeNewFromSource(t_scheduler_dag * dag, t_ins * tail, t_scheduler_dependency dependency)
{
  t_scheduler_dagedge * ret;
  t_node * node;
  t_node * tail_node = NULL;
  t_edge * i_edge;

  GRAPH_FOREACH_NODE(dag,node)
  {
    if(DAGNODE_INS(node) == tail)
    {
      tail_node = node;
      break;
    }
  }
  if(!tail_node) FATAL(("Tail not yet in dag!"));
  if(!DAG_SOURCE(dag)) FATAL(("No source in dag yet!"));

  NODE_FOREACH_SUCC_EDGE(DAG_SOURCE(dag),i_edge)
    if(EDGE_TAIL(i_edge) == tail_node)
    {
      if(DAGEDGE_DEPENDENCY_TYPE(i_edge) > dependency) DAGEDGE_DEPENDENCY_TYPE(i_edge) = dependency;
      return T_DAGEDGE(i_edge);
    }

  /* Create the new edge */
  ret = T_DAGEDGE(GraphNewEdge(T_GRAPH(dag),T_NODE(DAG_SOURCE(dag)),tail_node,dependency));
  DAGNODE_DEPENDENCYCOUNT(tail_node)++;

  return ret;
}/*}}}*/

/* Add a new edge to the dag, from the head to the sink, with dependency given as arguments */ /*{{{*/
t_scheduler_dagedge * SchedulerDagEdgeNewToSink(t_scheduler_dag * dag, t_ins * head, t_scheduler_dependency dependency)
{
  t_scheduler_dagedge * ret;
  t_node * node;
  t_node * head_node = NULL;
  t_edge * i_edge;

  GRAPH_FOREACH_NODE(dag,node)
  {
    if(DAGNODE_INS(node) == head)
    {
      head_node = node;
      break;
    }
  }
  if(!head_node) FATAL(("Tail not yet in dag!"));
  if(!DAG_SINK(dag)) FATAL(("No source in dag yet!"));

  NODE_FOREACH_SUCC_EDGE(head_node,i_edge)
    if(T_DAGNODE(EDGE_TAIL(i_edge)) == DAG_SINK(dag))
    {
      if(DAGEDGE_DEPENDENCY_TYPE(i_edge) > dependency) DAGEDGE_DEPENDENCY_TYPE(i_edge) = dependency;
      return T_DAGEDGE(i_edge);
    }

  /* Create the new edge */
  ret = T_DAGEDGE(GraphNewEdge(T_GRAPH(dag),head_node,T_NODE(DAG_SINK(dag)),dependency));
  DAGNODE_DEPENDENCYCOUNT(DAG_SINK(dag))++;

  return ret;
}/*}}}*/


void SchedulerListInsertNode(t_scheduler_list * list, t_scheduler_dagnode * node_to_schedule)/*{{{*/
{
  t_scheduler_listitem * listitem = (t_scheduler_listitem*) Malloc(sizeof(t_scheduler_listitem));
  listitem->next = NULL;
  listitem->node = node_to_schedule;

  list->nins++;

  /* Insert in the linked list at cycle 'DAGNODE_START(node_to_schedule)' */
  if(list->schedule[DAGNODE_START(node_to_schedule)]) FATAL(("Pretty strange!"));
  listitem->next = list->schedule[DAGNODE_START(node_to_schedule)];
  list->schedule[DAGNODE_START(node_to_schedule)] = listitem;
  
  /*if(list->schedule[DAGNODE_START(node_to_schedule)])
  {
    list->last->next = listitem;
    list->last = listitem;
  }
  else
  {
    list->first = list->last = listitem;
  }*/

}/*}}}*/

  
/* Update the absolute timing of each node and return true when there is a node ready for scheduling {{{*/
t_bool SchedulerDagUpdateIntervalsAndReadyList(t_scheduler_dag *dag, t_scheduler_dagnode * node, t_uint32 curr_cycle)
{
  t_node * i_node;
  if(node)
  {
    DAGNODE_START(node) = curr_cycle;
    DAGNODE_DEADLINE(node) = curr_cycle;

    SchedulerUpdateAbsoluteTime(node);
  }

  GRAPH_FOREACH_NODE(dag,i_node)
    if(DAGNODE_DEPENDENCYCOUNT(i_node) == 0 && !(DAGNODE_FLAGS(i_node) & SCHED_SCHEDULED))
      return TRUE;

  return FALSE;

}/* }}} */

/* Update absolute timing, when a node gets scheduled {{{ */
void SchedulerUpdateAbsoluteTime(t_scheduler_dagnode * node)
{
  t_edge * edge;
  t_bool changes;
  t_uint32 min,max;
  if(DAGNODE_START(node) > DAGNODE_DEADLINE(node))
  {
    if(node->ins)
    FATAL(("Start greater than deadline! @I",node->ins));
    else
      FATAL(("Start comes after deadline"));
  }
  /*if(node->ins) DiabloPrint(stdout,"Doing node @I, %d, %d\n",node->ins,node->start,node->deadline);
  else DiabloPrint(stdout,"Doing node S/E, %d, %d\n",node->start,node->deadline);*/
  NODE_FOREACH_SUCC_EDGE(node,edge)
  {
    changes = FALSE;
    max = (DAGNODE_DEADLINE(node) == DE_INFINITY || DAGEDGE_MAXWEIGHT(edge) == DE_INFINITY)? DE_INFINITY:DAGNODE_DEADLINE(node) + DAGEDGE_MAXWEIGHT(edge);
    min = (DAGNODE_START(node) == DE_INFINITY || DAGEDGE_MINWEIGHT(edge) == DE_INFINITY)? DE_INFINITY:DAGNODE_START(node) + DAGEDGE_MINWEIGHT(edge);
    if(DAGNODE_START(EDGE_TAIL(edge)) < min)
    {
      DAGNODE_START(EDGE_TAIL(edge)) = min;
      changes = TRUE;
    }
    if(DAGNODE_DEADLINE(EDGE_TAIL(edge)) > max)
    {
      DAGNODE_DEADLINE(EDGE_TAIL(edge)) = max;
      changes = TRUE;
    }
    if(changes) SchedulerUpdateAbsoluteTime(T_DAGNODE(EDGE_TAIL(edge)));
  }
  NODE_FOREACH_PRED_EDGE(node,edge)
  {
    changes = FALSE;
    max = (DAGNODE_DEADLINE(node) < DAGEDGE_MINWEIGHT(edge))? 0: (DAGNODE_DEADLINE(node) == DE_INFINITY) ? DE_INFINITY: DAGNODE_DEADLINE(node) - DAGEDGE_MINWEIGHT(edge);
    min = (DAGNODE_START(node) < DAGEDGE_MAXWEIGHT(edge))? 0: DAGNODE_START(node) - DAGEDGE_MAXWEIGHT(edge);
    if(DAGNODE_START(EDGE_HEAD(edge)) < min)
    {
      DAGNODE_START(EDGE_HEAD(edge)) = min;
      changes = TRUE;
    }
    if(DAGNODE_DEADLINE(EDGE_HEAD(edge)) > max)
    {
      DAGNODE_DEADLINE(EDGE_HEAD(edge)) = max;
      changes = TRUE;
    }
    if(changes) SchedulerUpdateAbsoluteTime(T_DAGNODE(EDGE_HEAD(edge)));
  }
  /*if(node->ins) DiabloPrint(stdout,"DONE node @I, %d, %d\n",node->ins,node->start,node->deadline);*/
  /*else DiabloPrint(stdout,"DONE node S/E, %d, %d\n",node->start,node->deadline);*/
}/*}}}*/

/* Write out of dag to dots files */ /*{{{*/
void SchedulerDagPrint(t_scheduler_dag * dag, t_bbl * bbl, t_uint32 troubles)
{
  char * filename, buffer[512], labelbuf[10];
  FILE * of;
  t_node * node;
  t_edge * edge;
  t_string color = NULL, style = NULL;

  DirectoryAssure("./dots-dag",FALSE);
  if(!troubles)
    filename = sDiabloPrint("./dots-dag/@G-%s.dot",BBL_INS_FIRST(bbl)?INS_OLD_ADDRESS(BBL_INS_FIRST(bbl)):0x0,BBL_FUN(bbl)?(FUN_NAME(BBL_FUN(bbl))?FUN_NAME(BBL_FUN(bbl)):"no-name"):"no-name");
  else
    filename = sDiabloPrint("./dots-dag/@G-%d.dot",BBL_INS_FIRST(bbl)?INS_OLD_ADDRESS(BBL_INS_FIRST(bbl)):0x0,troubles);
  of = fopen(filename,"w");
  VERIFY_FASSERT(of,("Could not open %s for writing!",filename));

  DiabloPrint(of, "digraph \"@G\" {\n\tnode [shape=box]\n",BBL_OLD_ADDRESS(bbl));
  GRAPH_FOREACH_NODE(dag,node)
  {
    if(DAGNODE_INS(node))
    {
      BBL_OBJECT(bbl)->actions->Print(DAGNODE_INS(node),buffer);
      DiabloPrint(of,"\t\"%p\" [label=\"@G:%s, %d,%d\"];\n",node,INS_OLD_ADDRESS(DAGNODE_INS(node)),buffer, DAGNODE_START(node), DAGNODE_DEADLINE(node));
    }
    else if(T_DAGNODE(node) == DAG_SOURCE(dag))
      DiabloPrint(of,"\t\"%p\" [label=\"Source, %d,%d\"];\n",node, DAGNODE_START(node), DAGNODE_DEADLINE(node));
    else if(T_DAGNODE(node) == DAG_SINK(dag))
      DiabloPrint(of,"\t\"%p\" [label=\"Sink, %d,%d\"];\n",node, DAGNODE_START(node), DAGNODE_DEADLINE(node));


    NODE_FOREACH_SUCC_EDGE(node,edge)
    {
      /*
      t_node * current_node = EDGE_TAIL(edge);
      t_node * iter_node;
      */
      t_edge * i_edge,* j_edge;
      t_bool dontprint = FALSE;
      NODE_FOREACH_SUCC_EDGE(node,i_edge)
      {
	if(edge == i_edge) continue;
	NODE_FOREACH_SUCC_EDGE(EDGE_TAIL(i_edge),j_edge)
	{
	  if(EDGE_TAIL(j_edge) == EDGE_TAIL(edge))
	  {
	    dontprint = TRUE;
	    break;
	  }
	}
	if(dontprint) break;
      }

      if(!dontprint)
      {
	(DAGEDGE_MAXWEIGHT(edge) == DE_INFINITY)?sprintf(labelbuf,"INF"):sprintf(labelbuf,"%d",DAGEDGE_MAXWEIGHT(edge));
	switch(DAGEDGE_DEPENDENCY_TYPE(edge))
	{
	  case DEP_RAR:
	    color = "gray";
	    style = "dashed";
	    break;
	  case DEP_RAW:
	    color = "blue";
	    style = "solid";
	    break;
	  case DEP_WAR:
	    color = "green";
	    style = "solid";
	    break;
	  case DEP_WAW:
	    color = "red";
	    style = "solid";
	    break;
	  default:
	    color = "yellow";
	    style = "dashed";
	    break;
	}

	fprintf(of,"\t\"%p\" -> \"%p\" [label=\"%d,%s\",style=%s,color=%s];\n",node,EDGE_TAIL(edge),DAGEDGE_MINWEIGHT(edge),labelbuf,style,color);
      }
    }
  }
  fprintf(of, "}\n");

  Free(filename);
  fclose(of);

}/*}}}*/

/* Print a schedule {{{*/
void SchedulerListPrint(t_scheduler_list * list)
{
  t_uint32 i;
  t_bbl * bbl = list->bbl;
  t_scheduler_listitem * item=NULL;
  if(bbl)
    DiabloPrint(stdout,"Schedule of bbl @B in %s has %d instructions\n",bbl,bbl->fun->name?bbl->fun->name:"Unknown",list->nins-1);
  for(i = 1; i < list->cycles -1; i++)
  {
    DiabloPrint(stdout,"Cycle %d:",i);
    if(list->schedule[i]) item = list->schedule[i];
    while(item)
    {
      if(DAGNODE_INS(item->node))
	DiabloPrint(stdout,"@I\n",DAGNODE_INS(item->node));
      item = item->next;
    }
    DiabloPrint(stdout,"\n");
  }
  DiabloPrint(stdout,">--------------------------------------------<\n");
}/* }}} */

/* Calculate the height in the dag for a node {{{ */
t_uint32 SchedulerDagNodeCalcHeight(t_scheduler_dagnode * node)
{
  t_edge * edge;

  if(DAGNODE_FLAGS(node) & SCHED_HEIGHT)
    return DAGNODE_HEIGHT(node);

  NODE_FOREACH_SUCC_EDGE(node, edge)
  {
    t_uint32 inc = DAGEDGE_MINWEIGHT(edge);

    DAGNODE_HEIGHT(node) = MAX(DAGNODE_HEIGHT(node), SchedulerDagNodeCalcHeight(T_DAGNODE(EDGE_TAIL(edge)))+inc);
  }

  DAGNODE_FLAGS(node) |= SCHED_HEIGHT;
  return DAGNODE_HEIGHT(node);
}/* }}} */

void SchedulerApplySchedule(t_scheduler_list * list)
{
  t_uint32 i;
  t_ins *i_ins;
  t_bool differs;
  t_bbl * bbl = list->bbl;
  

  if(list->schedule[0] && !list->schedule[0]->node->ins)
  {
    do
    {
      differs = FALSE;
      i_ins = BBL_INS_FIRST(bbl);
      for(i = 1; i < list->cycles-1; i++)
      {
	if(!list->schedule[i])
	  FATAL(("Implement adding nops to schedule!"));
	else if(list->schedule[i]->node->ins != i_ins)
	{
	  if(i == 1)
	    BblMoveInstructionBefore(list->schedule[i]->node->ins, BBL_INS_FIRST(bbl));
	  else
	    BblMoveInstructionAfter(list->schedule[i]->node->ins,list->schedule[i-1]->node->ins);
	  differs = TRUE;
	  break;
	}
	else
	  i_ins = INS_NEXT(i_ins);
      }
    } while(differs);

  }
  else
    FATAL(("List is not a valid schedule, source is not present"));
}
/* vim: set shiftwidth=2 foldmethod=marker: */
