
#ifndef _SCHEDULER_TYPEDEFS
#define _SCHEDULER_TYPEDEFS
typedef struct _t_scheduler_dagedge t_scheduler_dagedge;
typedef struct _t_scheduler_dagnode t_scheduler_dagnode;
typedef struct _t_scheduler_dag t_scheduler_dag;

typedef struct _t_scheduler_list t_scheduler_list;
typedef struct _t_scheduler_listitem t_scheduler_listitem;

typedef enum _t_scheduler_dependency
{
  DEP_RAW = 0,
  DEP_WAW,
  DEP_ORDERED,
  DEP_WAR,
  DEP_WEAKLY_ORDERED,
  DEP_RAR
} t_scheduler_dependency;


#endif

#ifndef _SCHEDULER_H
#define _SCHEDULER_H
#include "../datastructures/data_graph.h"
#include "../simple_types.h"
#include "diablo_flowgraph.h"

typedef struct _t_scheduler_dagnode*  (*SchedulerSelectionFunction)(t_scheduler_dag * dag, t_scheduler_list * list, t_uint32 cycle, t_bool has_pc_use);
/*! \defgroup SCHEDULING Scheduler datastructures */
/*@{*/
 

struct _t_scheduler_dag
{
  t_graph graph;
  /* Maybe add some specific fields here? */
  t_scheduler_dagnode * source;
  t_scheduler_dagnode * sink;
 
#ifdef STEGO
  t_scheduler_dagnode * first;
  t_scheduler_dagnode * last;
  t_uint32 nr_scheduled;
#endif
};

struct _t_scheduler_dagnode
{
  t_node node;
  t_ins * ins;
  t_uint32 dependency_count; /* The node of predecessors that are unscheduled */
  t_uint32 start;	/* This node must be scheduled at first at time 'start' */
  t_uint32 deadline;	/* This node must be scheduled at last at time 'deadline' */
  t_uint32 height;	/* The height of the node in the dag */

  t_uint32 flags;

#ifdef STEGO
  t_scheduler_dagnode * corr;
  t_scheduler_dagnode * next;
  t_bool is_sink;
  t_bool is_source;
#endif
  /* We should add the resources here, in an arch independend manner */
};

struct _t_scheduler_dagedge
{
  t_edge edge;
  /* According to Allan V. et al, a maximum weight is usefull for eg. arch. with delayed branches (MIPS in the diablo case) */
  t_uint32 min_weight;
  t_uint32 max_weight;
};

#define T_DAG(x)	((t_scheduler_dag*)x)
#define DAG_SOURCE(x)	(T_DAG(x)->source)
#define DAG_SINK(x)	(T_DAG(x)->sink)

#define T_DAGNODE(x)	((t_scheduler_dagnode*)x)
#define DAGNODE_INS(x)		T_DAGNODE(x)->ins
#define DAGNODE_START(x)	T_DAGNODE(x)->start
#define DAGNODE_DEADLINE(x)	T_DAGNODE(x)->deadline
#define DAGNODE_FLAGS(x)	T_DAGNODE(x)->flags
#define DAGNODE_DEPENDENCYCOUNT(x) T_DAGNODE(x)->dependency_count
#define DAGNODE_HEIGHT(x)	T_DAGNODE(x)->height

#define SCHED_SCHEDULED 0x1	/* this node is scheduled and should not be considered anymore for scheduling */
#define SCHED_HEIGHT	0x2	/* we have already calculated the height of this node */
#define SCHED_SOURCE_DUMMY	0x4	/* A start dummy node, the unique source of the dag */
#define SCHED_SINK_DUMMY	0x8	/* A sink dummy node, the unique sink of the dag */

#define DAGNODE_IS_DUMMY(x)	(T_DAGNODE(x)->flags & (SCHED_SOURCE_DUMMY || SCHED_SINK_DUMMY))

#define T_DAGEDGE(x)	((t_scheduler_dagedge*)x)
#define DAGEDGE_MINWEIGHT(x)	T_DAGEDGE(x)->min_weight
#define DAGEDGE_MAXWEIGHT(x)	T_DAGEDGE(x)->max_weight
#define DAGEDGE_DEPENDENCY_TYPE(edge)	((t_scheduler_dependency)EDGE_CAT(edge))

#define DE_INFINITY	(0xfffffff0)	/*< An infinite weight for scheduler_dagedges */

struct _t_scheduler_listitem
{
  t_scheduler_dagnode * node;
  struct _t_scheduler_listitem *next;
};

struct _t_scheduler_list
{
  t_uint32 nins;
  t_uint32 cycles;
  t_bbl * bbl;

  t_scheduler_listitem ** schedule;/* first;*/
/*  t_scheduler_listitem * last;*/
};

t_scheduler_dag * SchedulerDagNew(void);
void SchedulerDagFree(t_scheduler_dag * dag);

t_scheduler_list * SchedulerListNew(void);
void SchedulerListFree(t_scheduler_list * dag);

t_scheduler_dagnode * SchedulerDagNodeNew(t_scheduler_dag * dag, t_ins *ins);
t_scheduler_dagedge * SchedulerDagEdgeNew(t_scheduler_dag * dag, t_ins * head, t_ins * tail, t_scheduler_dependency dependency);
t_scheduler_dagedge * SchedulerDagEdgeNewToSink(t_scheduler_dag * dag, t_ins * head, t_scheduler_dependency dependency);
t_scheduler_dagedge * SchedulerDagEdgeNewFromSource(t_scheduler_dag * dag, t_ins * tail, t_scheduler_dependency dependency);
/*void SchedulerDagSetEntries(t_scheduler_dag * dag);*/
t_uint32 SchedulerDagNodeCalcHeight(t_scheduler_dagnode * node);

void SchedulerDagBuildForBbl(t_bbl * bbl, t_scheduler_dag * dag, t_regset exclude_set);
t_uint32 SchedulerGenerateListFromDag(t_scheduler_list * schedule, t_scheduler_dag * dag, SchedulerSelectionFunction func/* een soort van selectiefunctie meegeven? */, t_bbl * bbl);

void SchedulerCfgListSchedule(t_cfg * cfg);
void SchedulerBblListSchedule(t_bbl * bbl);

void SchedulerDagPrint(t_scheduler_dag * dag, t_bbl * bbl, t_uint32 troubles);
void SchedulerListPrint(t_scheduler_list * list);

void SchedulerListInsertNode(t_scheduler_list * list, t_scheduler_dagnode * node_to_schedule);

t_bool SchedulerDagUpdateIntervalsAndReadyList(t_scheduler_dag * dag, t_scheduler_dagnode * node, t_uint32 curr_cycle);
/*@}*/
#endif
