/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef MY_TYPES_TYPEDEFS
#define MY_TYPES_TYPEDEFS

typedef unsigned int t_uint32;
typedef unsigned short t_uint16;
typedef unsigned char t_uint8;
#ifdef VISUALSTUDIO
typedef unsigned __int64 t_uint64;
typedef __int64 t_int64;
#define LL(x)   ((t_int64) x)
#define ULL(x)  ((t_uint64) x)
#define bzero(x,y)   memset(x,y,sizeof(char))
#else
#define LL(x)    (x##LL)
#define ULL(x)    (x##ULL)
typedef unsigned long long t_uint64;
typedef long long  t_int64;
#endif
typedef int t_int32;
typedef short t_int16;
typedef char t_int8;
typedef struct _t_float{
  long double value;
  t_uint32 flags;
} t_float;

typedef struct _t_uint128 t_uint128;

struct _t_uint128
{
    t_uint64 lower;
    t_uint64 upper;
};



#ifdef POINTERSIZE64
#define POINTER_TO_INT(x) ((t_uint64) x)
#else
#define POINTER_TO_INT(x) ((t_uint32) x)
#endif




#define FP_VALUE(fp) ((fp).value)
#define FP_FLAGS(fp) ((fp).flags)

#define FP_EQ(fp1, fp2) \
  (FP_VALUE(fp1)==FP_VALUE(fp2) && FP_FLAGS(fp1) == FP_FLAGS(fp2))


#ifdef FALSE 
#undef FALSE
#endif
#ifdef TRUE 
#undef TRUE 
#endif
#ifdef PERHAPS
#undef PERHAPS 
#endif
#ifdef YES
#undef YES 
#endif
#ifdef NO
#undef NO 
#endif


typedef enum {NO = 0,YES = 1,PERHAPS = 2} t_tristate;
typedef enum {FALSE = 0,TRUE = 1} t_bool;
typedef int my_cmpresult; 
typedef int Filepos;

typedef union{
    t_int64 quadword;
    struct{ t_int32 d1; t_int32 d2;                     }doubleword;
    struct{ t_int16 w1; t_int16 w2; t_int16 w3; t_int16 w4; }word;
    struct{ t_int8 b1;  t_int8 b2;  t_int8 b3;  t_int8 b4;
            t_int8 b5;  t_int8 b6;  t_int8 b7;  t_int8 b8;  }byte;
    struct{ t_uint8 b1; t_uint8 b2; t_uint8 b3; t_uint8 b4;
            t_uint8 b5; t_uint8 b6; t_uint8 b7; t_uint8 b8; }ubyte;
} type_quadword;


#ifdef HAVE_SYS_PARAM
#include <sys/param.h>
#else
#define MIN(x,y) (((x)<(y))?(x):(y))
#define MAX(x,y) (((x)>(y))?(x):(y))
#endif
#endif
