linker="ARM GNU LD"
fileformats="ELF"
architectures="ARM"

/* versions with lots of underscores */
BssStartWithSbss {
   action  { ADD_SYMBOL("__bss_start__") }
   trigger { UNDEFINED_SYMBOL("__bss_start__")   && SECTION_EXISTS(".sbss") }
   symbol  { START_OF_SECTION(".sbss") } 
}


BssStartWithBss {
   action  { ADD_SYMBOL("__bss_start__") }
   trigger { UNDEFINED_SYMBOL("__bss_start__")   && SECTION_EXISTS(".bss") }
   symbol  { START_OF_SECTION(".bss") } 
}

BssEndWithSbss {
   action  { ADD_SYMBOL("__bss_end__") }
   trigger { UNDEFINED_SYMBOL("__bss_end__")   && SECTION_EXISTS(".sbss") }
   symbol  { END_OF_SECTION(".sbss") } 
}


BssEndWithBss {
   action  { ADD_SYMBOL("__bss_end__") }
   trigger { UNDEFINED_SYMBOL("__bss_end__")   && SECTION_EXISTS(".bss") }
   symbol  { END_OF_SECTION(".bss") } 
}

/* version with less underscores (gcc 3.2.2 & binutils 2.13.2.1) */
BssStartWithSbssNoTrailing {
   action  { ADD_SYMBOL("__bss_start") }
   trigger { UNDEFINED_SYMBOL("__bss_start")   && SECTION_EXISTS(".sbss") }
   symbol  { START_OF_SECTION(".sbss") } 
}


BssStartWithBssNoTrailing  {
   action  { ADD_SYMBOL("__bss_start") }
   trigger { UNDEFINED_SYMBOL("__bss_start")   && SECTION_EXISTS(".bss") }
   symbol  { START_OF_SECTION(".bss") } 
}

BssEndWithSbssNoTrailing  {
   action  { ADD_SYMBOL("__bss_end") }
   trigger { UNDEFINED_SYMBOL("__bss_end")   && SECTION_EXISTS(".sbss") }
   symbol  { END_OF_SECTION(".sbss") } 
}


BssEndWithBssNoTrailing  {
   action  { ADD_SYMBOL("__bss_end") }
   trigger { UNDEFINED_SYMBOL("__bss_end")   && SECTION_EXISTS(".bss") }
   symbol  { END_OF_SECTION(".bss") } 
}

/***********************************************************************/

EhFrameBegin {
   action  { ADD_SYMBOL("__EH_FRAME_BEGIN__") }
   trigger { UNDEFINED_SYMBOL("__EH_FRAME_BEGIN__")   && SECTION_EXISTS(".eh_frame") }
   symbol  { START_OF_SECTION(".eh_frame") } 
}

End {
   action  { ADD_SYMBOL("end") }
   trigger { UNDEFINED_SYMBOL("end")    }
   symbol  { END_OF_SECTION(".bss") } 
}


CtorsVector {
   action {  VECTORIZE(".ctors") }
   trigger { SUBSECTION_EXISTS("*",".ctors") }
}

End {
   action  { ADD_SYMBOL("_CTOR_END__") }
   trigger { UNDEFINED_SYMBOL("__CTOR_END__") && SUBSECTION_EXISTS("Linker","VECTOR___.ctors") }
   symbol  { END_OF_SUBSECTION("Linker","VECTOR___.ctors") }
}


/* Some special sections:

   .preinit_array
   .init_array
   .fini_array

   We first vectorize them and add start and end symbols, with fallbacks if a
   section does not exists
*/

PreInitVector {
   action {  VECTORIZE(".preinit_array") }
   trigger { SUBSECTION_EXISTS("*",".preinit_array") }
}

InitVector {
   action {  VECTORIZE(".init_array") }
   trigger { SUBSECTION_EXISTS("*",".init_array") }
}

FiniVector {
   action {  VECTORIZE(".fini_array") }
   trigger { SUBSECTION_EXISTS("*",".fini_array") }
}

/* ========= Start of preinit ========= */

/* Have preinit */

PreInitStart {
 action  { ADD_SYMBOL("__preinit_array_start") }
 trigger { UNDEFINED_SYMBOL("__preinit_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.preinit_array") }
 symbol  { START_OF_SUBSECTION("Linker","VECTOR___.preinit_array") }
}

/* Start of preinit, no preinit, have init */

PreInitStartFALLBACK {
   action  { ADD_SYMBOL("__preinit_array_start") }
   trigger { UNDEFINED_SYMBOL("__preinit_array_start") } 
   symbol  { ABS(0) }
}

/* ========= End of preinit ========= */

/* Have preinit */

PreInitEnd {
 action  { ADD_SYMBOL("__preinit_array_end") }
 trigger { UNDEFINED_SYMBOL("__preinit_array_end") && SUBSECTION_EXISTS("Linker","VECTOR___.preinit_array") }
 symbol  { END_OF_SUBSECTION("Linker","VECTOR___.preinit_array") }
}

/* Have init, no preinit */

PreInitEndFALLBACK {
 action  { ADD_SYMBOL("__preinit_array_end") }
 trigger { UNDEFINED_SYMBOL("__preinit_array_end") }
 symbol  { ABS(0) }
}

/* ========= Start of init ========= */

/* Have init */

InitStart {
   action  { ADD_SYMBOL("__init_array_start") }
   trigger { UNDEFINED_SYMBOL("__init_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.init_array") }
   symbol  { START_OF_SUBSECTION("Linker","VECTOR___.init_array") }
}

InitStartFALLBACK {
   action  { ADD_SYMBOL("__init_array_start") }
   trigger { UNDEFINED_SYMBOL("__init_array_start") }
   symbol  { ABS(0) }
}

/* ========= End of init ========= */

/* Have init */

InitEnd {
   action  { ADD_SYMBOL("__init_array_end") }
   trigger { UNDEFINED_SYMBOL("__init_array_end") && SUBSECTION_EXISTS("Linker","VECTOR___.init_end") }
   symbol  { END_OF_SUBSECTION("Linker","VECTOR___.init_array") }
}

InitEndFALLBACK {
 action  { ADD_SYMBOL("__init_array_end") }
 trigger { UNDEFINED_SYMBOL("__init_array_end") }
 symbol  { ABS(0) }
}


/* ========= Start of fini ========= */

/* Have fini */

FiniStart {
   action  { ADD_SYMBOL("__fini_array_start") }
   trigger { UNDEFINED_SYMBOL("__fini_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.fini_array") }
   symbol  { START_OF_SUBSECTION("Linker","VECTOR___.fini_array") }
}

/* Start of finit, no fini => data */

FiniStartFALLBACK {
 action  { ADD_SYMBOL("__fini_array_start") }
 trigger { UNDEFINED_SYMBOL("__fini_array_start") }
 symbol  { ABS(0) }
}

/* ========= End of fini ========= */

/* Have fini */

FiniEnd {
   action  { ADD_SYMBOL("__fini_array_end") }
   trigger { UNDEFINED_SYMBOL("__fini_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.fini_array") }
   symbol  { END_OF_SUBSECTION("Linker","VECTOR___.fini_array") }
}

/* Start of finit, no fini => data */

FiniEndFALLBACK {
 action  { ADD_SYMBOL("__fini_array_end") }
 trigger { UNDEFINED_SYMBOL("__fini_array_end") }
 symbol  { ABS(0) }
}


/* ===================================*/

JcrVector {
  action { VECTORIZE(".jcr") }
  trigger { SUBSECTION_EXISTS("*",".jcr") }
}

JcrLIST {
  action  { ADD_SYMBOL("__JCR_LIST__") }
  trigger { SUBSECTION_EXISTS("Linker","VECTOR___.jcr") }
  symbol  { START_OF_SUBSECTION("Linker","VECTOR___.jcr") }
}

CtorVector {
  action { VECTORIZE(".ctors") }
  trigger { SUBSECTION_EXISTS("*",".ctors") }
}

DtorVector {
  action { VECTORIZE(".dtors") }
  trigger { SUBSECTION_EXISTS("*",".dtors") }
}

GotTable {
   action  { ADD_SYMBOL("_GLOBAL_OFFSET_TABLE_") }
   trigger { UNDEFINED_SYMBOL("_GLOBAL_OFFSET_TABLE_")  && SECTION_EXISTS(".got") }
   symbol  { START_OF_SECTION(".got") }
}


AtExitVecor {
  action { VECTORIZE("__libc_atexit") }
  trigger { SUBSECTION_EXISTS("*","__libc_atexit") }
  
}

StartOfAtExit {
   action  { ADD_SYMBOL("__start___libc_atexit") }
   trigger { SECTION_EXISTS("__libc_atexit") }
   symbol  { START_OF_SUBSECTION("Linker","VECTOR_____libc_atexit") }
}

StopOfAtExit {
   action  { ADD_SYMBOL("__stop___libc_atexit") }
   trigger { SECTION_EXISTS("__libc_atexit") }
   symbol  { END_OF_SUBSECTION("Linker","VECTOR_____libc_atexit") }
}

/* embedded layout script */
SECTIONS
{
  /* Read-only sections, merged into text segment: */
  . = 0x8000;

  SEGMENT_START("Code");
  
  . = 0x8000 + SIZEOF_HEADERS;

  .note.ABI-tag : {} =0
  PUT_REMAINING_SECTIONS(NOTE_SECTION);

  /* after passing through diablo, all code sections will be merged */
  .merged_code : {} =0

  .init           :
  {
  } =0
  .plt            : { *(.plt) }
  .text           :
  {
  } =0
  .fini           :
  {
  } =0
  PUT_REMAINING_SECTIONS(CODE_SECTION);

  .rodata         : { *(.rodata .rodata.* .gnu.linkonce.r.*) }
  .rodata1        : { *(.rodata1) }
  PUT_REMAINING_SECTIONS(RO_DATA_SECTION);

  SEGMENT_END("Code");

  SEGMENT_START("Data");
  
  .sdata2         : { *(.sdata2 .sdata2.* .gnu.linkonce.s2.*) }
  .sbss2          : { *(.sbss2 .sbss2.* .gnu.linkonce.sb2.*) }
  .eh_frame_hdr : { *(.eh_frame_hdr) }
  /* Adjust the address for the data segment.  We want to adjust up to
     the same address within the page on the next page up.  */
  /*. = ALIGN(256) + (. & (256 - 1)) ; */
  /* Ensure the __preinit_array_start label is properly aligned.  We
     could instead move the label definition inside the section, but
     the linker would then create the section even if it turns out to
     be empty, which isn't pretty.  */
  . = ALIGN(32 / 8);
  .preinit_array     : { *(.preinit_array) }
  .init_array     : { *(.init_array) }
  .fini_array     : { *(.fini_array) }
  .data           :
  {
  }
  .data1          : { *(.data1) }
  .tdata	  : { *(.tdata .tdata.* .gnu.linkonce.td.*) }
  .tbss		  : { *(.tbss .tbss.* .gnu.linkonce.tb.*) *(.tcommon) }
  .eh_frame       : { KEEP (*(.eh_frame)) }
  .gcc_except_table   : { *(.gcc_except_table) }
  .dynamic        : { *(.dynamic) }
  .ctors          :
  {
  }
  .dtors          :
  {
  }
  .jcr            : { KEEP (*(.jcr)) }
  .got            : { *(.got.plt) *(.got) }
  .sdata          :
  {
  }
  PUT_REMAINING_SECTIONS(DATA_SECTION);
  _edata = .;
  __bss_start = .;
  __bss_start__ = .;
  .sbss           :
  {
  }
  .bss            :
  {
   . = ALIGN(32 / 8);
  }
  PUT_REMAINING_SECTIONS(BSS_SECTION);
  
  
  . = ALIGN(32 / 8);
  _end = .;
  _bss_end__ = . ; __bss_end__ = . ; __end__ = . ;
  SEGMENT_END("Data");
  
  .stack         0x80000 :
  {
    _stack = .;
    *(.stack)
  } 
}


