linker="I386 GNU LD"
fileformats="ELF"
architectures="I386"

/* versions with lots of underscores */
BssStartWithSbss {
   action  { ADD_SYMBOL("__bss_start__") }
   trigger { UNDEFINED_SYMBOL("__bss_start__")   && SECTION_EXISTS(".sbss") }
   symbol  { START_OF_SECTION(".sbss") } 
}


BssStartWithBss {
   action  { ADD_SYMBOL("__bss_start__") }
   trigger { UNDEFINED_SYMBOL("__bss_start__")   && SECTION_EXISTS(".bss") }
   symbol  { START_OF_SECTION(".bss") } 
}

BssEndWithSbss {
   action  { ADD_SYMBOL("__bss_end__") }
   trigger { UNDEFINED_SYMBOL("__bss_end__")   && SECTION_EXISTS(".sbss") }
   symbol  { END_OF_SECTION(".sbss") } 
}


BssEndWithBss {
   action  { ADD_SYMBOL("__bss_end__") }
   trigger { UNDEFINED_SYMBOL("__bss_end__")   && SECTION_EXISTS(".bss") }
   symbol  { END_OF_SECTION(".bss") } 
}

/* version with less underscores (gcc 3.2.2 & binutils 2.13.2.1) */
BssStartWithSbssNoTrailing {
   action  { ADD_SYMBOL("__bss_start") }
   trigger { UNDEFINED_SYMBOL("__bss_start")   && SECTION_EXISTS(".sbss") }
   symbol  { START_OF_SECTION(".sbss") } 
}


BssStartWithBssNoTrailing  {
   action  { ADD_SYMBOL("__bss_start") }
   trigger { UNDEFINED_SYMBOL("__bss_start")   && SECTION_EXISTS(".bss") }
   symbol  { START_OF_SECTION(".bss") } 
}

BssEndWithSbssNoTrailing  {
   action  { ADD_SYMBOL("__bss_end") }
   trigger { UNDEFINED_SYMBOL("__bss_end")   && SECTION_EXISTS(".sbss") }
   symbol  { END_OF_SECTION(".sbss") } 
}


BssEndWithBssNoTrailing  {
   action  { ADD_SYMBOL("__bss_end") }
   trigger { UNDEFINED_SYMBOL("__bss_end")   && SECTION_EXISTS(".bss") }
   symbol  { END_OF_SECTION(".bss") } 
}

/***********************************************************************/

EhFrameBegin {
   action  { ADD_SYMBOL("__EH_FRAME_BEGIN__") }
   trigger { UNDEFINED_SYMBOL("__EH_FRAME_BEGIN__")   && SECTION_EXISTS(".eh_frame") }
   symbol  { START_OF_SECTION(".eh_frame") } 
}

End {
   action  { ADD_SYMBOL("end") }
   trigger { UNDEFINED_SYMBOL("end")    }
   symbol  { ABS(0) } 
}

_EndWithBss {
   action  { ADD_SYMBOL("_end") }
   trigger { UNDEFINED_SYMBOL("_end") && SECTION_EXISTS(".bss")   }
   symbol  { END_OF_SECTION(".bss") } 
}

_EndWithSbss {
   action  { ADD_SYMBOL("_end") }
   trigger { UNDEFINED_SYMBOL("_end") && SECTION_EXISTS(".sbss")  }
   symbol  { END_OF_SECTION(".sbss") } 
}

GotTable {
   action  { ADD_SYMBOL("_GLOBAL_OFFSET_TABLE_") }
   trigger { UNDEFINED_SYMBOL("_GLOBAL_OFFSET_TABLE_")  && SECTION_EXISTS(".got") }
   symbol  { START_OF_SECTION(".got") }
}

GotTableReservedEntries {
    action { ADD_SUBSECTION("Linker",".got",DATA,4) }
    trigger { SECTION_EXISTS(".got") }
    section { CONST32(0),CONST32(0),CONST32(0) }
    address { SYMBOL("_GLOBAL_OFFSET_TABLE_") }
}

/* Some special sections:

   .preinit_array
   .init_array
   .fini_array

   We first vectorize them and add start and end symbols, with fallbacks if a
   section does not exists
*/

PreInitVector {
   action {  VECTORIZE(".preinit_array") }
   trigger { SUBSECTION_EXISTS("*",".preinit_array") }
}

InitVector {
   action {  VECTORIZE(".init_array") }
   trigger { SUBSECTION_EXISTS("*",".init_array") }
}

FiniVector {
   action {  VECTORIZE(".fini_array") }
   trigger { SUBSECTION_EXISTS("*",".fini_array") }
}

/* ========= Start of preinit ========= */

/* Have preinit */

PreInitStart {
 action  { ADD_SYMBOL("__preinit_array_start") }
 trigger { UNDEFINED_SYMBOL("__preinit_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.preinit_array") }
 symbol  { START_OF_SUBSECTION("Linker","VECTOR___.preinit_array") }
}

/* Start of preinit, no preinit, have init */

PreInitStartFALLBACK {
   action  { ADD_SYMBOL("__preinit_array_start") }
   trigger { UNDEFINED_SYMBOL("__preinit_array_start") } 
   symbol  { ABS(0) }
}

/* ========= End of preinit ========= */

/* Have preinit */

PreInitEnd {
 action  { ADD_SYMBOL("__preinit_array_end") }
 trigger { UNDEFINED_SYMBOL("__preinit_array_end") && SUBSECTION_EXISTS("Linker","VECTOR___.preinit_array") }
 symbol  { END_OF_SUBSECTION("Linker","VECTOR___.preinit_array") }
}

/* Have init, no preinit */

PreInitEndFALLBACK {
 action  { ADD_SYMBOL("__preinit_array_end") }
 trigger { UNDEFINED_SYMBOL("__preinit_array_end") }
 symbol  { ABS(0) }
}

/* ========= Start of init ========= */

/* Have init */

InitStart {
   action  { ADD_SYMBOL("__init_array_start") }
   trigger { UNDEFINED_SYMBOL("__init_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.init_array") }
   symbol  { START_OF_SUBSECTION("Linker","VECTOR___.init_array") }
}

InitStartFALLBACK {
   action  { ADD_SYMBOL("__init_array_start") }
   trigger { UNDEFINED_SYMBOL("__init_array_start") }
   symbol  { ABS(0) }
}

/* ========= End of init ========= */

/* Have init */

InitEnd {
   action  { ADD_SYMBOL("__init_array_end") }
   trigger { UNDEFINED_SYMBOL("__init_array_end") && SUBSECTION_EXISTS("Linker","VECTOR___.init_end") }
   symbol  { END_OF_SUBSECTION("Linker","VECTOR___.init_array") }
}

InitEndFALLBACK {
 action  { ADD_SYMBOL("__init_array_end") }
 trigger { UNDEFINED_SYMBOL("__init_array_end") }
 symbol  { ABS(0) }
}


/* ========= Start of fini ========= */

/* Have fini */

FiniStart {
   action  { ADD_SYMBOL("__fini_array_start") }
   trigger { UNDEFINED_SYMBOL("__fini_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.fini_array") }
   symbol  { START_OF_SUBSECTION("Linker","VECTOR___.fini_array") }
}

/* Start of finit, no fini => data */

FiniStartFALLBACK {
 action  { ADD_SYMBOL("__fini_array_start") }
 trigger { UNDEFINED_SYMBOL("__fini_array_start") }
 symbol  { ABS(0) }
}

/* ========= End of fini ========= */

/* Have fini */

FiniEnd {
   action  { ADD_SYMBOL("__fini_array_end") }
   trigger { UNDEFINED_SYMBOL("__fini_array_start") && SUBSECTION_EXISTS("Linker","VECTOR___.fini_array") }
   symbol  { END_OF_SUBSECTION("Linker","VECTOR___.fini_array") }
}

/* Start of finit, no fini => data */

FiniEndFALLBACK {
 action  { ADD_SYMBOL("__fini_array_end") }
 trigger { UNDEFINED_SYMBOL("__fini_array_end") }
 symbol  { ABS(0) }
}


/* ===================================*/

JcrVector {
  action { VECTORIZE(".jcr") }
  trigger { SUBSECTION_EXISTS("*",".jcr") }
}

JcrLIST {
   action  { ADD_SYMBOL("__JCR_LIST__") }
   trigger { SECTION_EXISTS(".jcr") }
   symbol  { START_OF_SECTION(".jcr") } 
}

CtorVector {
  action { VECTORIZE(".ctors") }
  trigger { SUBSECTION_EXISTS("*",".ctors") }
}

DtorVector {
  action { VECTORIZE(".dtors") }
  trigger { SUBSECTION_EXISTS("*",".dtors") }
}

AtExitVecor {
  action { VECTORIZE("__libc_atexit") }
  trigger { SUBSECTION_EXISTS("*","__libc_atexit") }
  
}

StartOfAtExit {
   action  { ADD_SYMBOL("__start___libc_atexit") }
   trigger { SECTION_EXISTS("__libc_atexit") }
   symbol  { START_OF_SUBSECTION("Linker","VECTOR_____libc_atexit") }
}

StopOfAtExit {
   action  { ADD_SYMBOL("__stop___libc_atexit") }
   trigger { SECTION_EXISTS("__libc_atexit") }
   symbol  { END_OF_SUBSECTION("Linker","VECTOR_____libc_atexit") }
}

/* slightly adapted gcc linker script serves as our layout script */
SECTIONS
{
  /* Read-only sections, merged into text segment: */
  . = 0x08048000 + SIZEOF_HEADERS;

  .note.ABI-tag   : {} =0
  PUT_REMAINING_SECTIONS(NOTE_SECTION);

  .interp         : { *(.interp) }
  .hash           : { *(.hash) }
  .dynsym         : { *(.dynsym) }
  .dynstr         : { *(.dynstr) }
  .gnu.version    : { *(.gnu.version) }
  .gnu.version_d  : { *(.gnu.version_d) }
  .gnu.version_r  : { *(.gnu.version_r) }
  .rel.dyn        :
    {
      *(.rel.init)
      *(.rel.text .rel.text.* .rel.gnu.linkonce.t.*)
      *(.rel.fini)
      *(.rel.rodata .rel.rodata.* .rel.gnu.linkonce.r.*)
      *(.rel.data .rel.data.* .rel.gnu.linkonce.d.*)
      *(.rel.tdata .rel.tdata.* .rel.gnu.linkonce.td.*)
      *(.rel.tbss .rel.tbss.* .rel.gnu.linkonce.tb.*)
      *(.rel.ctors)
      *(.rel.dtors)
      *(.rel.got)
      *(.rel.bss .rel.bss.* .rel.gnu.linkonce.b.*)
    }
  .rela.dyn       :
    {
      *(.rela.init)
      *(.rela.text .rela.text.* .rela.gnu.linkonce.t.*)
      *(.rela.fini)
      *(.rela.rodata .rela.rodata.* .rela.gnu.linkonce.r.*)
      *(.rela.data .rela.data.* .rela.gnu.linkonce.d.*)
      *(.rela.tdata .rela.tdata.* .rela.gnu.linkonce.td.*)
      *(.rela.tbss .rela.tbss.* .rela.gnu.linkonce.tb.*)
      *(.rela.ctors)
      *(.rela.dtors)
      *(.rela.got)
      *(.rela.bss .rela.bss.* .rela.gnu.linkonce.b.*)
    }
  .rel.plt        : { *(.rel.plt) }
  .rela.plt       : { *(.rela.plt) }
  .init           :
  {
    KEEP (*(.init))
  } =0x90909090
  .plt            : { *(.plt) }
  .merged_code	  : {} =0x90909090
  .text           :
  {
    *(.text .stub .text.* .gnu.linkonce.t.*)
    /* .gnu.warning sections are handled specially by elf32.em.  */
    *(.gnu.warning)
  } =0x90909090
  .fini           :
  {
    KEEP (*(.fini))
  } =0x90909090
  PUT_REMAINING_SECTIONS(CODE_SECTION);
  PROVIDE (__etext = .);
  PROVIDE (_etext = .);
  PROVIDE (etext = .);
  .rodata         : { *(.rodata .rodata.* .gnu.linkonce.r.*) }
  .rodata1        : { *(.rodata1) }
  PUT_REMAINING_SECTIONS(RO_DATA_SECTION);
  .eh_frame_hdr : { *(.eh_frame_hdr) }
  /* Adjust the address for the data segment.  We want to adjust up to
     the same address within the page on the next page up.  */
  . = ALIGN(0x1000) + (. & 0xfff);

  /* Ensure the __preinit_array_start label is properly aligned.  We
     could instead move the label definition inside the section, but
     the linker would then create the section even if it turns out to
     be empty, which isn't pretty.  */
  . = ALIGN(32 / 8);
  PROVIDE (__preinit_array_start = .);
  .preinit_array     : { *(.preinit_array) }
  PROVIDE (__preinit_array_end = .);
  PROVIDE (__init_array_start = .);
  .init_array     : { *(.init_array) }
  PROVIDE (__init_array_end = .);
  PROVIDE (__fini_array_start = .);
  .fini_array     : { *(.fini_array) }
  PROVIDE (__fini_array_end = .);
  .data           :
  {
    *(.data .data.* .gnu.linkonce.d.*)
    SORT(CONSTRUCTORS)
  }
  .data1          : { *(.data1) }
  .tdata	  : { *(.tdata .tdata.* .gnu.linkonce.td.*) }
  .tbss		  : { *(.tbss .tbss.* .gnu.linkonce.tb.*) *(.tcommon) }
  .eh_frame       : { KEEP (*(.eh_frame)) }
  .gcc_except_table   : { *(.gcc_except_table) }
  .dynamic        : { *(.dynamic) }
  .ctors          :
  {
    /* gcc uses crtbegin.o to find the start of
       the constructors, so we make sure it is
       first.  Because this is a wildcard, it
       doesn't matter if the user does not
       actually link against crtbegin.o; the
       linker won't look for a file to match a
       wildcard.  The wildcard also means that it
       doesn't matter which directory crtbegin.o
       is in.  */
    KEEP (*crtbegin.o(.ctors))
    /* We don't want to include the .ctor section from
       from the crtend.o file until after the sorted ctors.
       The .ctor section from the crtend file contains the
       end of ctors marker and it must be last */
    KEEP (*(EXCLUDE_FILE (*crtend.o ) .ctors))
    KEEP (*(SORT(.ctors.*)))
    KEEP (*(.ctors))
  }
  .dtors          :
  {
    KEEP (*crtbegin.o(.dtors))
    KEEP (*(EXCLUDE_FILE (*crtend.o ) .dtors))
    KEEP (*(SORT(.dtors.*)))
    KEEP (*(.dtors))
  }
  .jcr            : { KEEP (*(.jcr)) }
  .got            : { *(.got.plt) *(.got) }
  PUT_REMAINING_SECTIONS(DATA_SECTION);
  _edata = .;
  PROVIDE (edata = .);
  __bss_start = .;
  .bss            :
  {
   *(.dynbss)
   *(.bss .bss.* .gnu.linkonce.b.*)
   *(COMMON)
   /* Align here to ensure that the .bss section occupies space up to
      _end.  Align after .bss to ensure correct alignment even if the
      .bss section disappears because there are no input sections.  */
   . = ALIGN(32 / 8);
  }
  . = ALIGN(32 / 8);
  PUT_REMAINING_SECTIONS(BSS_SECTION);
  _end = .;
  PROVIDE (end = .);

  .comment        : { *(.comment) }

}
