linker="I386 GNU LD"
fileformats="ELF"
architectures="I386"

/* linker script for the linux kernel for i386 */

BssStartWithSbss {
   action  { ADD_SYMBOL("__bss_start") }
   trigger { UNDEFINED_SYMBOL("__bss_start")   && SECTION_EXISTS(".sbss") }
   symbol  { START_OF_SECTION(".sbss") } 
}

BssStartWithBss {
   action  { ADD_SYMBOL("__bss_start") }
   trigger { UNDEFINED_SYMBOL("__bss_start")   && SECTION_EXISTS(".bss") }
   symbol  { START_OF_SECTION(".bss") } 
}

BssEndWithSbss {
   action  { ADD_SYMBOL("__bss_end") }
   trigger { UNDEFINED_SYMBOL("__bss_end")   && SECTION_EXISTS(".sbss") }
   symbol  { END_OF_SECTION(".sbss") } 
}

BssEndWithBss {
   action  { ADD_SYMBOL("__bss_end") }
   trigger { UNDEFINED_SYMBOL("__bss_end")   && SECTION_EXISTS(".bss") }
   symbol  { END_OF_SECTION(".bss") } 
}

_EndWithBss {
   action  { ADD_SYMBOL("_end") }
   trigger { UNDEFINED_SYMBOL("_end") && SECTION_EXISTS(".bss")   }
   symbol  { END_OF_SECTION(".bss") } 
}

_EndWithSbss {
   action  { ADD_SYMBOL("_end") }
   trigger { UNDEFINED_SYMBOL("_end") && SECTION_EXISTS(".sbss")  }
   symbol  { END_OF_SECTION(".sbss") } 
}

/* ========= kernel specific stuff (from arch/i386/vmlinux.lds in kernel sources) ========== */

_text {
  action  { ADD_SYMBOL("_text") }
  trigger { UNDEFINED_SYMBOL("_text") }
  symbol  { START_OF_SECTION(".text") }
}

_etext {
  action  { ADD_SYMBOL("_etext") }
  trigger { UNDEFINED_SYMBOL("_etext") }
  symbol  { END_OF_SECTION(".text") }
}

_edata {
  action  { ADD_SYMBOL("_edata") }
  trigger { UNDEFINED_SYMBOL("_edata") }
  symbol  { END_OF_SECTION(".data") }
}

__start___ksymtab {
  action  { ADD_SYMBOL("__start___ksymtab") }
  trigger { UNDEFINED_SYMBOL("__start___ksymtab") }
  symbol  { START_OF_SECTION("__ksymtab") }
}

__stop___ksymtab {
  action  { ADD_SYMBOL("__stop___ksymtab") }
  trigger { UNDEFINED_SYMBOL("__stop___ksymtab") }
  symbol  { END_OF_SECTION("__ksymtab") }
}

__init_begin { 
  action  { ADD_SYMBOL("__init_begin") }
  trigger { UNDEFINED_SYMBOL("__init_begin") }
  symbol  { START_OF_SECTION(".text.init") }
}

/* this is a kludge, because the linker script language is
 * not expressive enough to express the fact that symbols should
 * sometimes be aligned */
__init_end { 
  action  { ADD_SYMBOL("__init_end") }
  trigger { UNDEFINED_SYMBOL("__init_end") }
  symbol  { START_OF_SECTION(".data.page_aligned") }
}

ExtableVectorize {
  action  { VECTORIZE("__ex_table") }
  trigger { SECTION_EXISTS("__ex_table") }
}

__start___ex_table {
  action  { ADD_SYMBOL("__start___ex_table") }
  trigger { UNDEFINED_SYMBOL("__start___ex_table") && SUBSECTION_EXISTS("Linker","VECTOR_____ex_table") }
  symbol  { START_OF_SUBSECTION("Linker","VECTOR_____ex_table") }
}

__stop___ex_table {
  action  { ADD_SYMBOL("__stop___ex_table") }
  trigger { UNDEFINED_SYMBOL("__stop___ex_table") && SUBSECTION_EXISTS("Linker","VECTOR_____ex_table") }
  symbol  { END_OF_SUBSECTION("Linker","VECTOR_____ex_table") }
}

SetupVectorize {
  action  { VECTORIZE(".setup.init") }
  trigger { SECTION_EXISTS(".setup.init") }
}

__setup_start {
  action  { ADD_SYMBOL("__setup_start") }
  trigger { UNDEFINED_SYMBOL("__setup_start") && SUBSECTION_EXISTS("Linker","VECTOR___.setup.init") }
  symbol  { START_OF_SUBSECTION("Linker","VECTOR___.setup.init") }
}

__setup_end {
  action  { ADD_SYMBOL("__setup_end") }
  trigger { UNDEFINED_SYMBOL("__setup_end") && SUBSECTION_EXISTS("Linker","VECTOR___.setup.init") }
  symbol  { END_OF_SUBSECTION("Linker","VECTOR___.setup.init") }
}

InitcallVectorize {
  action  { VECTORIZE(".initcall.init") }
  trigger { SECTION_EXISTS(".initcall.init") }
}

__initcall_start {
  action  { ADD_SYMBOL("__initcall_start") }
  trigger { UNDEFINED_SYMBOL("__initcall_start") && SUBSECTION_EXISTS("Linker","VECTOR___.initcall.init") }
  symbol  { START_OF_SUBSECTION("Linker","VECTOR___.initcall.init") }
}

__initcall_end {
  action  { ADD_SYMBOL("__initcall_end") }
  trigger { UNDEFINED_SYMBOL("__initcall_end") && SUBSECTION_EXISTS("Linker","VECTOR___.initcall.init") }
  symbol  { END_OF_SUBSECTION("Linker","VECTOR___.initcall.init") }
}

SECTIONS
{
  . = 0xC0000000 + 0x100000;
  _text = .;			/* Text and read-only data */
  .text : {
	*(.text)
	*(.fixup)
	*(.gnu.warning)
	} = 0x90909090

  _etext = .;			/* End of text section */

  .rodata : { *(.rodata) *(.rodata.*) }
  .kstrtab : { *(.kstrtab) }

  . = ALIGN(16);		/* Exception table */
  __start___ex_table = .;
  __ex_table : { *(__ex_table) }
  __stop___ex_table = .;

  __start___ksymtab = .;	/* Kernel symbol table */
  __ksymtab : { *(__ksymtab) }
  __stop___ksymtab = .;

  .data : {			/* Data */
	*(.data)
	CONSTRUCTORS
	}

  _edata = .;			/* End of data section */

  . = ALIGN(8192);		/* init_task */
  .data.init_task : { *(.data.init_task) }

  . = ALIGN(4096);		/* Init code and data */
  __init_begin = .;
  .text.init : { *(.text.init) } = 0x90909090
  .data.init : { *(.data.init) }
  . = ALIGN(16);
  __setup_start = .;
  .setup.init : { *(.setup.init) }
  __setup_end = .;
  __initcall_start = .;
  .initcall.init : { *(.initcall.init) }
  __initcall_end = .;
  . = ALIGN(4096);
  __init_end = .;

  . = ALIGN(4096);
  .data.page_aligned : { *(.data.idt) }

  . = ALIGN(32);
  .data.cacheline_aligned : { *(.data.cacheline_aligned) }

  __bss_start = .;		/* BSS */
  .bss : {
	*(.bss)
	}
  _end = . ;

  /* Sections to be discarded */
  /DISCARD/ : {
	*(.text.exit)
	*(.data.exit)
	*(.exitcall.exit)
	}

  /* Stabs debugging sections.  */
  .stab 0 : { *(.stab) }
  .stabstr 0 : { *(.stabstr) }
  .stab.excl 0 : { *(.stab.excl) }
  .stab.exclstr 0 : { *(.stab.exclstr) }
  .stab.index 0 : { *(.stab.index) }
  .stab.indexstr 0 : { *(.stab.indexstr) }
  .comment 0 : { *(.comment) }
}
