/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* $Header: /home/cvs/cvsroot/diablo/utils/util_clientserver.c,v 1.2 2004/01/08 14:41:36 bdebus Exp $ */
#include "../config.h"
#ifdef HAVE_NETWORKING
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include "util_error.h"
#include "util_malloc.h"
#include "util_clientserver.h"
#include "util_timer.h"

#ifndef MSG_NOSIGNAL
void sig_pipe_ignore(int i)
{
}
#endif

connection_descriptor * NewConnectionDescriptor()
{
    connection_descriptor * ret;
    ret=(connection_descriptor *) calloc(1,sizeof(connection_descriptor));
    return ret;
}

connection_descriptor * CreateClient(char * host, int port)
{
    connection_descriptor * ret=NewConnectionDescriptor();
    int connect_ret=0;
    ret->hp = gethostbyname(host);
    VERIFY_FASSERT ((ret->hp != 0),("gethostbyname %s failed",host)); 

    ret->sa_in.sin_family = AF_INET;
    ret->sa_in.sin_addr.s_addr = ((struct in_addr *)(ret->hp->h_addr))->s_addr;
    ret->sa_in.sin_port = htons(port);
    ret->socket_descriptor = socket(AF_INET, SOCK_STREAM, 0);
    VERIFY_FASSERT ((ret->socket_descriptor  != -1), ("Socket failed"));

    connect_ret=connect(ret->socket_descriptor,(struct sockaddr *)  &(ret->sa_in), sizeof(struct sockaddr_in));

    VERIFY_FASSERT ((connect_ret  != -1), ("Connect failed"));
    
    return ret;
}


connection_descriptor * CreateServer(int port, int max_con)
{
    int fun_ret;
    int on=1;
    connection_descriptor * ret=NewConnectionDescriptor();
    ret->socket_descriptor = socket(AF_INET, SOCK_STREAM, 0);
    VERIFY_FASSERT ((ret->socket_descriptor != -1),("Socket failed"));

    ret->sa_in.sin_family = AF_INET;
    ret->sa_in.sin_addr.s_addr = INADDR_ANY;
    ret->sa_in.sin_port = htons(port);

    fun_ret=setsockopt(ret->socket_descriptor, SOL_SOCKET, SO_REUSEADDR, (const char *) &on, sizeof(on));
    VERIFY_FASSERT ((fun_ret == 0), ("Set socket options!"));

    fun_ret=bind(ret->socket_descriptor, (struct sockaddr *) &(ret->sa_in), sizeof(struct sockaddr_in));
    VERIFY_FASSERT ((fun_ret!=-1), ("Bind failed"));

    ret->type=SERVER;
    ret->max_connections=max_con;

    return ret;
}
void WaitOnceForConnection(connection_descriptor * in, void (*CallBack)(connection_descriptor *))
{
    int fun_ret=0;
    connection_descriptor * ret=NewConnectionDescriptor();
    ret->type=ACCEPTED_SERVER;
    fun_ret=listen(in->socket_descriptor, in->max_connections);
    VERIFY_FASSERT ((fun_ret != -1),  ("Listen failed"));
    do
    {
    ret->socket_descriptor = accept(in->socket_descriptor, (struct sockaddr *)  &(ret->sa_in), &(ret->addrlen));
    sleep(1);
    printf("Retrying accept\n");
    }
    while(ret->socket_descriptor==-1);
    VERIFY_FASSERT ( (ret->socket_descriptor !=-1), ("Accept failed"));
    CallBack(ret);
}
void WaitForConnections(connection_descriptor * in, void (*CallBack)(connection_descriptor *))
{
    while(1) WaitOnceForConnection(in,CallBack);
}

void close_connection_time(void * data)
{
    connection_descriptor * in=(connection_descriptor *) data;
    ConnectionEnd(in);
}

int ConnectionGet(connection_descriptor * in,char * data, t_uint32 max_len, long to)
{
    t_uint32 * size=(t_uint32 *) Malloc(sizeof(t_uint32));
    t_uint32 realsize;
    int cumsize=0;
    int locsize=0;
    timevals * timeout=NULL;
    if (to) timeout=TimerAdd(to,0,close_connection_time,(void *) in);
    while (cumsize<sizeof(t_uint32))
    {
	if ((locsize=recv(in->socket_descriptor, size+cumsize, sizeof(t_uint32)-cumsize, 0)))
	{
	   if (locsize<0)
	   {
	      if (to) 
	      { 
		 DeleteTimer(timeout); 
		 if (ErrorGetErrno()==EINTR) /* If the system call is  interrupted we replay it! */
		 { 
		    timeout=TimerAdd(to,0,close_connection_time,(void *) in);
		    ErrorClearErrno();
		    continue; 
		 } 
		 else 
		 { 
		    ErrorClearErrno();
		    Free(size); 
		    return -1;
		 } 
	      }
	      FATAL(("Recv failed!"));
	   }
	   cumsize+=locsize;
	}
    }

    if (to) DeleteTimer(timeout);
    realsize=ntohl(*size);
    VERIFY_FASSERT ( (realsize+1<=max_len) , ("Received size (%d = %x) is bigger then max_size (%d)",realsize,realsize,max_len));
    cumsize=0;
    locsize=0;
    if (to) timeout=TimerAdd(to,0,close_connection_time,in);
    while (cumsize<realsize)
    {
	if ((locsize=recv(in->socket_descriptor, data+cumsize, (realsize)-cumsize, 0)))
	{
	   if (locsize<0)
	   {
	       if (to) { 
		  DeleteTimer(timeout); 
		  if (ErrorGetErrno()==EINTR)
		  { 
		     timeout=TimerAdd(to,0,close_connection_time,(void *) in);
		     ErrorClearErrno();
		     continue; 
		  } 
		  else 
		  { 
		     ErrorClearErrno();
		     Free(size); 
		     return -1;
		  } 
	       }
	       FATAL(("Recv failed!"));
	    
	   }
	     cumsize+=locsize;
	}
    }
    data[realsize]=0;
    if (to) DeleteTimer(timeout);
    Free(size);
    return 0;
}

int ConnectionSend(connection_descriptor * in,char * data,t_uint32 size)
{
    int locsize;
    t_uint32 realsize=htonl(size);
   
#ifndef MSG_NOSIGNAL
    signal(SIGPIPE,sig_pipe_ignore); /* TODO: We should use sigaction, and restore the normal 
				   action for SIGPIPE after the send request */
    locsize=send(in->socket_descriptor, &realsize, sizeof(t_uint32), 0 );
#else
    locsize=send(in->socket_descriptor, &realsize, sizeof(t_uint32), MSG_NOSIGNAL);
#endif
    if (locsize < 0) return -1;
    VERIFY_FASSERT(locsize == sizeof(t_uint32), (("Implement non-blocking IO!")));

#ifndef MSG_NOSIGNAL
    locsize=send(in->socket_descriptor, data, size, 0 );
#else
    locsize=send(in->socket_descriptor, data, size, MSG_NOSIGNAL);
#endif
    if (locsize < 0) return -1;
    VERIFY_FASSERT(locsize == size, (("Implement non-blocking IO!")));
    return 0;
}

void ConnectionEnd(connection_descriptor * in)
{
	close(in->socket_descriptor);
}
#endif
