/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "../config.h"
#ifdef COLORS
#include <stdio.h>
#include <stdarg.h>
#ifdef HAVE_UNISTD
#include <unistd.h>
#endif
#include "util_color.h"
#include "util_error.h"
#include "util_term.h"

void ColorPrint(effect b,color fg, color bg,  char * arg, ...)
{
   long store;
   char color=0;
   va_list val;
   va_start(val,arg);

   store=ErrorGetErrno(); /* isatty can set errno */
#ifdef HAVE_ISATTY
#ifdef HAVE_FILENO
   if (isatty(fileno(stdout))) color=1; 
#else 
   color=1;
#endif
#else
   color=1;
#endif
   ErrorSetErrno(store);

   if (color) printf("\033[%d;3%d;4%dm", b, fg,bg); /* If we are outputting to a tty, 
   					    we want colors else we just print.... */
   vfprintf(stdout,arg,val);
   if (color) printf("\033[m"); /* Restore normal terminal colors */
   fflush(stdout); /* Push control output chars */
}

int ColorFPrint(effect e,color fg, color bg, int curspos, int indent,FILE * fp,  char * arg, ...)
{
   int ret=0;
   static char string[1000];
   long store;
   char color=0;
   va_list val;
   va_start(val,arg);
 
   store=ErrorGetErrno(); /* isatty can set errno */
#ifdef HAVE_ISATTY
#ifdef HAVE_FILENO
   if (isatty(fileno(fp))) color=1;
#else
   color=1;
#endif
#else
   color=1;
#endif

   ErrorSetErrno(store);

   if (color) fprintf(fp,"\033[%d;3%d;4%dm", e, fg,bg); /* If we are outputting to a tty, 
   					    we want colors else we just print.... */
   vsprintf(string,arg,val);
   ret=TermPrintClean(fp,string,curspos,indent,0);
   if (color) fprintf(fp,"\033[m"); /* Restore normal terminal colors */
   fflush(fp); /* Push control output chars */
   return ret;
}
#else
#warning Including color, but no color support
#endif
