/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* $Header: /home/cvs/cvsroot/diablo/utils/util_error.h,v 1.1.1.1 2003/10/10 12:12:23 bdebus Exp $ */
#ifndef MY_ERROR_H
#define MY_ERROR_H
#include <stdio.h>
#include <errno.h> 
#include "../simple_types.h" 
#include "util_options.h"

#define E_WARNING 1
#define E_FATAL   2
#define E_USAGE   4

#define PARANOIDLEVEL 100
#define VERIFYLEVEL 1

#ifndef DEBUGLEVEL
#define DEBUGLEVEL PARANOIDLEVEL
#endif

extern int errcurs;
void ErrorInit(char * fname, FILE * err, void (*usage)(void));
int PreError(t_uint32 type, char * file, t_uint32 lineno);
void Error(char * errormsg, ...);
void PostError(t_uint32 type);
void PrintExecName(FILE * fp);
void ErrorClearErrno(void);
t_int32 ErrorGetErrno(void);
void ErrorSetErrno(t_int32 in);
void ErrorFini(void);

#define WARNING(arg)	do { errcurs=PreError(E_WARNING, __FILE__, __LINE__); if (global_options.verbose) Error arg ; PostError(E_WARNING); } while(0)
#define FATAL(arg)	do { errcurs=PreError(E_FATAL, __FILE__, __LINE__); Error arg ; PostError(E_FATAL); } while(0)
#define FATAL_USAGE(arg)	do { PreError(E_FATAL|E_USAGE, __FILE__, __LINE__); Error arg ; PostError(E_FATAL|E_USAGE); } while(0)

#define PARANOID_FASSERT(cond,mesg) do {if (DEBUGLEVEL>=PARANOIDLEVEL) { if (!(cond)) FATAL(mesg); } } while(0)
#define VERIFY_FASSERT(cond,mesg) do {if (DEBUGLEVEL>=VERIFYLEVEL) { if (!(cond)) FATAL(mesg); } } while(0)
#endif
