/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *//* $Header: /home/cvs/cvsroot/diablo/utils/util_processes.c,v 1.2 2003/12/03 09:59:16 bdebus Exp $ */
#include "../config.h"
#ifdef HAVE_IPC
#ifdef HAVE_FILENO
/* WARNING: if you use this file on Tru64 Unix, you should use the sys5 library, or everything will fail! */
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/stropts.h>
#include <termio.h>
#include <stdarg.h>
#include <unistd.h>
#include "util_processes.h"
#include "util_error.h"

/* Don't forget to close with a NULL arg! */

void CreateProcPipe(int * retpid,FILE ** read_pipe, FILE ** write_pipe, char ** argv, char * exec, ...)
{
    /*struct termio io; */
    int filedes[2]; /* pipe */
    int file_rd=fileno(stdin);  /* filedescriptor = 0 */
    int file_wr=fileno(stdout); /* filedescriptor = 1 */
    int into_proc, used_byproc_in, outof_proc, used_byproc_out;
    va_list val;
    int ret;


    ret=pipe(filedes);                  /* Create process in */
    into_proc = filedes[1];
    used_byproc_in = filedes[0];

    ret=pipe(filedes);                  /* Create process out */
    outof_proc = filedes[0];
    used_byproc_out = filedes[1];

    if (!((*retpid)=fork()))
    {
        char * tmp;
	int nargs=0;
	/* child */
	/* stdin replaced by pipe output */
#ifdef ALPHA
	if (!isastream(used_byproc_out)) FATAL(("Seems like we are using a non-STREAMS-based pipe function\nThis happens on Tru64 Unix if you do not link with the sys5 library!")); 
#endif


	/* close unused */

	close(outof_proc);
	close(into_proc);

	ret=dup2(used_byproc_in,file_rd);
	/* stdout replaced by pipe input */
	dup2(used_byproc_out,file_wr);

#ifdef ALPHA
	/* Provide terminal functionality on pipe.... Sometimes this is neccesary */
        ret=ioctl(used_byproc_out,I_PUSH,"ptem");
	if (ret<0) FATAL(("Pseudo Terminal EMulation (output) failed...."));
        ret=ioctl(file_wr,I_PUSH,"ldterm");
	if (ret<0) FATAL(("Line Discipline module (output) failed"));
	ret=ioctl(file_rd,I_PUSH,"ptem");
	if (ret<0) FATAL(("Pseudo Terminal EMulation (input) failed...."));
        ret=ioctl(file_rd,I_PUSH,"ldterm");
	if (ret<0) FATAL(("Line Discipline module (input) failed"));
#endif
	
	setbuf(stdin,NULL);
	setbuf(stdout,NULL);
	if (!argv)
	{
	   argv=(char **) realloc(argv,sizeof(char *) * (nargs+1));
	   argv[nargs]=exec;
	   nargs++;
	   va_start(val,exec);
	   while ((tmp=va_arg(val, char *)))
	   {
	      argv=(char **) realloc(argv,sizeof(char *) * (nargs+1));
	    argv[nargs]=tmp;
	    nargs++;
	   }
	   argv=(char **) realloc(argv,sizeof(char *) * (nargs+1));
	   argv[nargs]=NULL;
	   nargs++;
	}

	/* Run the program */
	execvp(exec,argv);
	FATAL(("Execl %s failed!",exec));
    }
    else
    {



	/* close unused */

	close(used_byproc_in);
	close(used_byproc_out);


	if (read_pipe)
	{
	    *read_pipe=fdopen(outof_proc,"r");
	    setbuf(*read_pipe,NULL);
	    VERIFY_FASSERT (((*read_pipe)!=NULL) ,("Failed to open FILE descriptor for %s",exec));
	}

	if (write_pipe)
	{
	    *write_pipe=fdopen(into_proc,"w");
	    setbuf(*write_pipe,NULL);
	    VERIFY_FASSERT (((*write_pipe)!=NULL) ,("Failed to open FILE descriptor for %s",exec));
	}

    }

}


void MakeDaemon()
{
	int ret=0;
	if (fork()) exit(0); /* Kill the parent */
	ret=setsid(); /* Set session id (create new session, without a controling terminal....) */
	VERIFY_FASSERT(ret>0,("Could not become the session leader!"));
}
#endif
#endif
