/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h> 
#include <stdarg.h>
#include "util_status.h"
#include "util_term.h"
#include "util_malloc.h"
#include "../datastructures/data_string.h"
#include "../kernel/diablo_print.h"

typedef struct _status_struct
{
   t_string message;
   struct _status_struct * prev;
} status_struct;

status_struct * status_head=NULL;
t_uint32 status_depth=0;

void StatusPush(t_string message)
{
   status_struct * tmp=Malloc(sizeof(status_struct));
   tmp->prev=status_head;
   tmp->message=message;
   status_head=tmp;
   status_depth++;
}

void StatusPop()
{
    status_struct * tmp= status_head;
    status_head=status_head->prev;
    Free(tmp);
    status_depth--;
}

void verbose(char * s,...)
{
   t_string buffer;
   va_list val;

   va_start(val,s);
   buffer=vsDiabloPrint(s,val);
   TermPrintClean(stdout,buffer,0,3*status_depth,FINAL);
   Free(buffer);
}
