/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef STATUS_H
#define STATUS_H
#include "util_options.h"
#include "../diablo_options.h"
#include "../kernel/diablo_print.h"
extern t_uint32 last_warning;
extern t_uint32 warnings; 
extern t_uint32 status_depth;
void verbose(char * s,...);
void StatusPush(t_string message);
void StatusPop();
#define VERBOSE(x,y)   do { if (x<=global_options.verbose) verbose y; } while(0)         
#ifndef COLORS
#define ANALYSIS_START(name) do { fprintf(stdout,"Starting %s\n",name); last_warning=warnings; } while(0)
#define ANALYSIS_SIMPLE_END(name)   fprintf(stdout,"Done %s\n",name);
#define ANALYSIS_END(name,args)   fprintf(stdout,"Done %s\n",name); DiabloPrintStdout args;
#else
#include "util_color.h"
#define ANALYSIS_START(name)  do { t_uint32 tel; for(tel=0; tel<status_depth; tel++) fprintf(stdout,"   "); StatusPush(name);   fprintf(stdout,"Starting "); ColorFPrint(normal,cyan,black,0,0,stdout,"%s\n",name); last_warning=warnings;  } while(0)
#define ANALYSIS_SIMPLE_END(name)   do { t_uint32 tel; StatusPop();  for(tel=0; tel<status_depth; tel++) fprintf(stdout,"   "); fprintf(stdout,"Finished "); if (warnings!=last_warning) { ColorFPrint(normal,yellow,black,0,0,stdout,"PASSED "); printf("%d warnings\n", warnings-last_warning); last_warning=warnings; } else { ColorFPrint(normal,green,black,0,0,stdout,"OK\n"); } } while(0)
#define ANALYSIS_END(name,args)   do { t_uint32 tel; StatusPop();  for(tel=0; tel<status_depth; tel++) fprintf(stdout,"   ");  fprintf(stdout,"Finished ("); DiabloPrintStdout args; printf(") "); if (warnings!=last_warning) { ColorFPrint(normal,yellow,black,0,0,stdout,"PASSED "); printf("%d warnings\n", warnings-last_warning); last_warning=warnings; } else { ColorFPrint(normal,green,black,0,0,stdout,"OK\n"); } } while(0)
#endif
#endif
