/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/* $Header: /home/cvs/cvsroot/diablo/utils/util_timer.c,v 1.2 2004/01/08 14:41:36 bdebus Exp $ */


#ifdef __STRICT_ANSI__ 
/* This define enables the correct signal handling functions when compiling -ansi! */ 
#define _POSIX_SOURCE 

#endif

#include "../config.h"
#ifdef HAVE_TIME
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include "util_malloc.h"
#include "util_error.h"
#include "util_timer.h"

timevals * time_head=NULL;

void timer_handler(int i) 
{
    sigset_t mask_set;
    sigset_t old_set;
    struct itimerval itim; 
    timevals * curr=NULL;
    sigfillset(&mask_set);
    sigprocmask(SIG_SETMASK, &mask_set, &old_set); /* Thou shall not disturb me */
    if (time_head!=NULL)
    {
    	curr=time_head;
	if (time_head->cb) time_head->cb(time_head->data);
	time_head=DLLNext(time_head);
    }
    signal(SIGALRM,timer_handler);

    if (time_head)
    {
	itim.it_interval.tv_sec=0;
	itim.it_interval.tv_usec=0;
	itim.it_value.tv_usec=time_head->usecs%1000000;
	itim.it_value.tv_sec=time_head->usecs/1000000;
	setitimer(ITIMER_REAL,&itim, 0);
    }
    if (curr)
    DLLNext(curr)=DLLPrev(curr)=NULL;     
    sigprocmask(SIG_SETMASK, &old_set, NULL); /* Bug me if you want */
}

void DeleteTimer(timevals * delete_me)
{
     DLLUnlink(delete_me,(void *) &time_head,NULL);
     Free(delete_me);
}

timevals * TimerAdd(int usecs,int mode,timer_callback cb,void * data)
{
    long cumul;
    struct itimerval itim; 
    timevals * curr=(timevals *) Malloc(sizeof(timevals));
    timevals * it, * it2;
    sigset_t mask_set;
    sigset_t old_set;
    sigfillset(&mask_set);
    sigprocmask(SIG_SETMASK, &mask_set, &old_set); /* Thou shall not disturb me */
    curr->usecs=usecs;
    curr->cb=cb;
    curr->data=data;
    if (!time_head)
    {
    	DLLAppend(NULL,curr,(void *) &time_head,NULL);
	itim.it_interval.tv_sec=0;
	itim.it_interval.tv_usec=0;
	itim.it_value.tv_usec=usecs%1000000;
	itim.it_value.tv_sec=usecs/1000000;
	setitimer(ITIMER_REAL,&itim, 0);
    }
    else 
    {
	getitimer (ITIMER_REAL,&itim);
	cumul=itim.it_value.tv_sec*1000000+itim.it_value.tv_usec;
	if (cumul>usecs)
	{
	    time_head->usecs=cumul-usecs;
	    DLLAppend(NULL,curr,(void *) &time_head,NULL);
	    itim.it_interval.tv_sec=0;
	    itim.it_interval.tv_usec=0;
	    itim.it_value.tv_usec=usecs%1000000;
	    itim.it_value.tv_sec=usecs/1000000;
	    setitimer(ITIMER_REAL,&itim, 0);
	}
	else
	{
	    it2=time_head;
	    it=time_head;
	    while((it=DLLNext(it)))
	    {
		cumul+=it->usecs;
		if (cumul>usecs)
		{
		    DLLAppend(it2,curr,(void *)&time_head,NULL);
		    curr->usecs-=(cumul-it->usecs);
		    break;
		}
		it2=it;
	    }
	    if (!it)
	    {
		DLLAppend(it2,curr,(void *) &time_head,NULL);
		curr->usecs-=cumul;
	    }
	}
    }
    sigprocmask(SIG_SETMASK, &old_set, NULL); /* Bug me if you want */
    return curr;
}

void TimerInit()
{
    signal(SIGALRM,timer_handler);
}

/*
void print_data(void * data)
{
	long i=(long) data;
	printf("%ld\n",i);
}

main()
{
    signal(SIGALRM,timer_handler);
    TimerAdd(500000,0,print_data,(void *) 1);
    TimerAdd(300000,0,print_data,(void *) 2);
    TimerAdd(700000,0,print_data,(void *) 3);
    while(1)
    {
    }
}
*/
#endif
