/*
 * Copyright 2001,2002: Bruno De Bus
 *                      Bjorn De Sutter
 *                      Dominique Chanet
 *                      Ludo Van Put
 *
 *
 * This file is part of Diablo (Diablo is a better link-time optimizer)
 *
 * Diablo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Diablo is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Diablo; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include "ia64_scheduler.h"

#include "../../kernel/diablo_edge.h"
#include "../../utils/util_malloc.h"
#include "../../utils/util_error.h"
#include "ia64_opcodes.h"
#include "ia64_registers.h"
#include "ia64_print.h"
#include "ia64_utils.h"
#include "../../kernel/diablo_print.h"
#include "../../kernel/diablo_bbl.h"
#include "../../kernel/diablo_ins.h"
#include "../../kernel/diablo_regset.h"
#include "../../kernel/diablo_iterators.h"
#include "../../diablo_options.h"

#include "ia64_statusbar.h"

static void InitializePossibleTemplatesList(t_bool *possibleTemplates)
{
  t_uint8 i;
  
  for(i=0; i<16; i++)
    if(i == 0x03 || i == 0x0A || i == 0x0D || i == 0x0F)
      possibleTemplates[i] = 0;
    else
      possibleTemplates[i] = 1;
}



static void Ia64InsertInstruction(t_ins * ins, t_ins * new_ins)
{
  t_ins * next = INS_NEXT(ins);
  
  if(next == NULL)
  {
    INS_NEXT(ins) = new_ins;
    INS_PREV(new_ins) = ins;
  }
  else
  {
    INS_NEXT(ins) = new_ins;
    INS_NEXT(new_ins) = next;

    INS_PREV(next) = new_ins;
    INS_PREV(new_ins) = ins;
        
  }
  
}

static t_ins * Ia64InsertNops(t_section *code, t_bbl * bbl, t_ins * lastFromPrevBundle, 
                              t_bool *mustBeNop, t_uint8 templateNr)
{
  t_uint8 slot;
  t_ins * curr;
  t_ins * nop;
  
  //static int c=0;
  curr = lastFromPrevBundle;

//  printf("slot0:%d, slot1: %d,slot2: %d\n",templateField[templateNr][0],templateField[templateNr][1],templateField[templateNr][2]);

  /* debugging
  if(curr && INS_CADDRESS(INS_NEXT(curr)) == 0x4000000000001240ULL)
    *(int *)0 = 0;
  */
  
  for(slot = 0; slot < 3; slot++)
  {
    if(!mustBeNop[slot])
    {
      if(slot == 0 && curr == NULL)
        curr = bbl->ins_first;
      else
        //if(slot != 2)
          curr = INS_NEXT(curr);
      continue;
    }
    
    if(slot == 0)
    { /*\TODO ook nog de dingen (zoals BBL-address, is het een BBL-leader, ...) in generic juist zetten !!!*/
#if 0
      if((c++ & 0x3ff) == 0)
        printf("CREATE %d\n",c);
#endif      
      nop = Ia64InstructionCreateNoop(code, templateField[templateNr][slot]);
    }
    else
    {
#if 0
      if((c++ & 0x3ff) == 0)
        printf("CREATE %d\n",c);
#endif      
      nop = Ia64InstructionCreateNoop(code, templateField[templateNr][slot]);
    }

    INS_BBL(nop) = bbl;   
    BBL_CADDRESS(nop)=BBL_CADDRESS(bbl);
    //BBL_OLD_ADDRESS(nop)=(curr==NULL?BBL_OLD_ADDRESS(bbl->ins_first):BBL_OLD_ADDRESS(curr));
	    
    
    bbl->nins++;
    
    if(slot == 0 && curr == NULL)
    {
      if(bbl->ins_first)
      {
	INS_PREV(bbl->ins_first) = nop;
	INS_NEXT(nop) = bbl->ins_first;

	/* is nop; mag zeker in dezelfde instr groep zitten */
	IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(bbl->ins_first);
      }
      else
      {
	IA64INSINSTRGROUP(nop) = GetCurrentInstrGroupAndInc();
	bbl->ins_last = nop;
      }
      curr = bbl->ins_first = nop;
    }
    else
      if(slot == 2 && INS_NEXT(curr) == NULL)
      {
        INS_NEXT(bbl->ins_last) = nop;
        INS_PREV(nop) = bbl->ins_last;

        /* is nop; mag zeker in dezelfde instr groep zitten */
        IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(bbl->ins_last);

        curr = bbl->ins_last = nop;
      }
      else
      {
        /* is nop; mag zeker in dezelfde instr groep zitten :wel de juiste kiezen... */
	if(slot == 0 && INS_NEXT(curr))
	  IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(INS_NEXT(curr)); //deze van dezelfde bundel nemen
	else
        IA64INSINSTRGROUP(nop) = IA64INSINSTRGROUP(curr);
        
        Ia64InsertInstruction(curr, nop);
	
	if(bbl->ins_last == curr)
	  bbl->ins_last=nop;
	
	curr = nop;
      }
        
//    if(slot != 2)
//      curr = INS_NEXT(curr);
        
  }
  
  return curr;
}

static t_bool StopMatch(t_uint8 slot, t_ins * ins, t_uint8 templateNr)
{
//  t_ins * next_ins=INS_NEXT(ins);
  t_bool mustBeStopped;
  
  /* je hebt hiervoor toch altijd de twee opties */
  if(slot == 2)
    return 1;
    
  mustBeStopped = 
    INS_NEXT(ins) == NULL || (IA64INSINSTRGROUP(ins) != IA64INSINSTRGROUP(INS_NEXT(ins)));
    
  if(mustBeStopped)
  {
    if((slot == 1 && templateNr == 0x02) || (slot == 0 && templateNr == 0x0A))
      return 1;
    else
      return 0;
  }
      
  return 1;
}

static t_bool InstrTypeMatch(t_ia64_instr_type instr_type, t_ia64_unit unit)
{
  switch(instr_type)
  {
    case IA64_TYPE_A:
      return unit == IA64_UNIT_I || unit == IA64_UNIT_M;
    case IA64_TYPE_I:
      return unit == IA64_UNIT_I;
    case IA64_TYPE_M:
      return unit == IA64_UNIT_M;
    case IA64_TYPE_B:
      return unit == IA64_UNIT_B;
    case IA64_TYPE_F:
      return unit == IA64_UNIT_F;
    case IA64_TYPE_DUMMY:
      return unit == IA64_UNIT_L;
    case IA64_TYPE_X:
      return unit == IA64_UNIT_X;
    case IA64_TYPE_DATA:
      return 1;
    default:
//      printf("%d\n",instr_type);
      FATAL(("Wrong instruction match !!!"));
      /* Keep the compiler happy*/
      return 0;
  }
  
}

static void Ia64BBLBundeling(t_section *code, t_bbl * bbl)
{
  t_ins * ins;
  t_ins * lastFromPrevBundle = NULL;
  t_bool possibleTemplates[16];
  t_bool mustBeNop[3];
  t_uint8 slot = 0, i;
  t_bool isMatch;
  t_uint32 fullNOPBundle = 0;
 
  /* standaard moet alles nopkes zijn... */
  mustBeNop[0] = mustBeNop[1] = mustBeNop[2] = 1;
  
  InitializePossibleTemplatesList(possibleTemplates);
/*
  {
    t_ins *tmp_ins = bbl->ins_first;

    while(tmp_ins)
    {
      InsPrint(tmp_ins, 0);
      tmp_ins = INS_NEXT(tmp_ins);
    }

*/
  
  for(ins=bbl->ins_first; ins!=NULL;)  
  {
    isMatch = 0;

    /* first check if there is a dummy without a X-type instruction: this is possible if the X-type instruction was removed*/
    if(IA64INSTYPE(ins) == IA64_TYPE_DUMMY && (INS_NEXT(ins) == NULL || IA64INSTYPE(INS_NEXT(ins)) != IA64_TYPE_DUMMY  ))
    {
      t_ins * next = INS_NEXT(ins);
      InsKill(ins);
      ins = next; 
    }

    if(ins != NULL)
    {

      for(i = 0; i<16; i++)
      {
	if(possibleTemplates[i] && InstrTypeMatch(IA64INSTYPE(ins), templateField[i*2][slot]) 
	    && StopMatch(slot, ins, i*2))
	{
	  isMatch = 1;
	  break;
	}
      }   

      /*overrule...*/

      if(slot != 2 && (IA64OPCODEINFO(ins).flags & IA64_OPCODE_SLOT2))
	isMatch = 0;

      if(isMatch)
      { 
	for(i = 0; i<16; i++)
	  if(!(possibleTemplates[i] && InstrTypeMatch(IA64INSTYPE(ins), templateField[i*2][slot])
		&& StopMatch(slot, ins, i*2)))
	    possibleTemplates[i] = 0;
	ins=INS_NEXT(ins);

	/* we hebben een instructie die erin past => moet geen nop meer in */
	mustBeNop[slot] = 0;
      }

    }

    slot++;
    /* ook kijken of we aan het einde van de instructies gekomen zijn !!!*/
    if(slot == 3 || ins == NULL)
    {
      for(i = 0; i<16; i++)
	if(possibleTemplates[i])
	  break;

      if(i == 16)
	FATAL(("Something wrong with bundeling !!!"));

      if(mustBeNop[0] && mustBeNop[1] && mustBeNop[2])
      {
	printf("Warning: bundle full of nops !\n");
	printf("Current instruction: ");
	//InsPrint(ins, 0);
	
	fullNOPBundle++;
	if(fullNOPBundle >= 128)
	{
	  printf("Something wrong with bundeling: encountered a bundle full of nops ?!!\n");
	  if(ins == NULL)
	    printf("Current instruction was NULL [usually meaning end of BBL]\n");
	  else
	  {
	    printf("Current instruction was:\n");
	    //InsPrint(ins, 0);
	  }
	  FATAL(("Please Correct !"));
	}

      }
      else
	fullNOPBundle = 0;

      //      printf("inserting nops\n");  
      lastFromPrevBundle = Ia64InsertNops(code, bbl, lastFromPrevBundle, mustBeNop, i*2);


      mustBeNop[0] = mustBeNop[1] = mustBeNop[2] = 1;
      slot = 0; 

      InitializePossibleTemplatesList(possibleTemplates);
    }
  }
}






/*---------------------------------------------------------------------------------*/

#define IA64_MAXCOST 0x7fffffff


#define IA64_IS_NOP(ins) (IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_NOP)


/*statische variabele, zodat die niet altijd moet meegesleurd worden*/
t_section *__static_scheduler_code_section__;

static t_bool __gedetailleerde_latency_ = FALSE;

static t_bool __has_shift_ = FALSE;


#define Ia64IsNop(ins) (IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_NOP)



t_bool Ia64IsNopOrDummy(t_ins *ins)
{
  return IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_NOP
    || IA64INSTYPE(ins) == IA64_TYPE_DUMMY;
}





static t_uint32 Ia64RemoveNops(t_bbl* bbl)
{
  t_ins * ins;
  t_uint32 cnt = 0;

  BBL_FOREACH_INS(bbl, ins)
  {
    if(IA64_IS_NOP(ins))
    {
      InsKill(T_INS(ins));
      cnt++;
    }
  }
  
  //printf("removed %d NOPs in BBL @ %llx: current # instructions = %d\n", cnt, BBL_OLD_ADDRESS(bbl), bbl->nins);

  return cnt;

}

static t_uint32 Ia64RemoveDummies(t_bbl* bbl)
{
  t_ins * ins;
  t_uint32 cnt = 0;

  BBL_FOREACH_INS(bbl, ins)
  {
    if(IA64INSTYPE(ins) == IA64_TYPE_DUMMY)
    {
      InsKill(T_INS(ins));
      cnt++;
    }
  }
  
  //printf("removed %d NOPs in BBL @ %llx: current # instructions = %d\n", cnt, BBL_OLD_ADDRESS(bbl), bbl->nins);

  return cnt;

}








typedef struct _t_ia64_ins_list_item
{
  t_ins *ins;
  struct _t_ia64_ins_list_item *next;
} t_ia64_ins_list_item; 


t_ia64_ins_list_item *Ia64InsListItemAlloc(t_ins *ins)
{
  t_ia64_ins_list_item *item = (t_ia64_ins_list_item *)Calloc(1, sizeof(t_ia64_ins_list_item));

  item->ins = ins;
  
  return item;
}

void Ia64InsListItemFree(t_ia64_ins_list_item *item)
{
  Free(item);
}


typedef struct _t_ia64_ins_list
{
  t_uint32 nins;
  
  t_ia64_ins_list_item *first;
  t_ia64_ins_list_item *last;
} t_ia64_ins_list; 


#define IA64INSLIST_FOREACH_ITEM(list,item) for(item=list->first; item!=NULL; item=item->next)
#define IA64INSLIST_FOREACH_ITEM_SAFE(list,item,tmp)   for(item=list->first, tmp=item?item->next:NULL; item!=NULL; item=tmp, tmp=item?item->next:NULL)


void Ia64InsListAppendIns(t_ia64_ins_list *insList, t_ins *ins)
{
  insList->nins++;

  if(insList->last == NULL)
  {
    insList->last = Ia64InsListItemAlloc(ins);
    insList->first = insList->last;
  }
  else
  {
    insList->last->next = Ia64InsListItemAlloc(ins);
    insList->last = insList->last->next;
  }

}

t_ia64_ins_list *Ia64InsListAlloc(t_bbl *bbl)
{
  t_ia64_ins_list *insList = (t_ia64_ins_list *)Calloc(1, sizeof(t_ia64_ins_list));
  
  t_ins *ins;

  BBL_FOREACH_INS(bbl, ins)
  {
    if(!Ia64IsNopOrDummy(ins))
      Ia64InsListAppendIns(insList, ins);
  }
  
  return insList;
}

void Ia64InsListFree(t_ia64_ins_list *insList)
{
  t_ia64_ins_list_item *item, *next;

  IA64INSLIST_FOREACH_ITEM_SAFE(insList, item, next)
    Ia64InsListItemFree(item);

  Free(insList);
  
}


t_ia64_ins_list *Ia64InsListCopy(t_ia64_ins_list *insList)
{
  t_ia64_ins_list *newList = (t_ia64_ins_list *)Calloc(1, sizeof(t_ia64_ins_list));
  
  t_ia64_ins_list_item *item;

  IA64INSLIST_FOREACH_ITEM(insList, item)
    Ia64InsListAppendIns(newList, item->ins);

  return newList;
  
}

t_ins *Ia64InsListDelete(t_ia64_ins_list *insList, t_uint32 index)
{
  t_ia64_ins_list_item *item, *next;
  t_uint32 i = 1;
  t_ins *ins = NULL;

  item = insList->first;
  next = insList->first->next;

  insList->nins--;

  if(index == 0)
  {
    ins = item->ins;
    Ia64InsListItemFree(item);
    insList->first = next;

    return ins;
  }
  
  i = 1;
  while(i != index)
  {
    item = next;
    next = next->next;
    i++;
  }

  ins = next->ins;
  item->next = next->next;
 
  if(insList->last == next)
    insList->last = item;
  
  Ia64InsListItemFree(next);

  return ins;
}


void Ia64InsListAppendInsBeforeBranch(t_ia64_ins_list *insList, t_ins *ins)
{
  if(IA64INSTYPE(insList->last->ins) != IA64_TYPE_B)
    Ia64InsListAppendIns(insList, ins);
  else
  {
    /*invoegen op voorlaatste pos*/
    t_ia64_ins_list_item *item, *next;

    if(insList->first == insList->last)
    {
      item = Ia64InsListItemAlloc(ins);
      item->next = insList->first;
      insList->first = item;
   
      insList->nins++;
   
    }
    else
    {

      item = insList->first;
      next = insList->first->next;

      while(next != insList->last)
      {
	item = next;
	next = next->next;
      }

      item->next = Ia64InsListItemAlloc(ins);
      item->next->next = next;

      insList->nins++;
    }

  }
}


void Ia64InsListPrependIns(t_ia64_ins_list *insList, t_ins *ins)
{
  t_ia64_ins_list_item *item;
  
  insList->nins++;

  item = Ia64InsListItemAlloc(ins);
  
  if(insList->first == NULL)
    insList->first = insList->last = item;
  else
  {
    item->next = insList->first;
    insList->first = item;
  }
}


void Ia64InsListPrint(t_ia64_ins_list *insList)
{
  t_ia64_ins_list_item *item;
 
  FATAL(("Implement"));
/*  IA64INSLIST_FOREACH_ITEM(insList, item)
    InsPrint(item->ins, 0);
*/
    }













struct _t_ia64DAGItem;

typedef enum _t_ia64DAGDependencyType
{
  IA64_RAW = 0, IA64_WAW, IA64_ORDERED, IA64_WAR, IA64_WEAKLY_ORDERED, IA64_RAR
} t_ia64DAGDependencyType;

typedef struct _t_ia64DAGDependencyItem
{
  t_ia64DAGDependencyType type;
  t_uint8 gewicht;
  struct _t_ia64DAGItem *item;
} t_ia64DAGDependencyItem;


t_ia64DAGDependencyItem *Ia64DAGDependencyItemAlloc(t_ia64DAGDependencyType type, t_uint8 gewicht, 
    struct _t_ia64DAGItem *item)
{
  t_ia64DAGDependencyItem *depItem = (t_ia64DAGDependencyItem *)Calloc(1, sizeof(t_ia64DAGDependencyItem));

  depItem->type = type;
  depItem->gewicht = gewicht;
  depItem->item = item;
  
  return depItem;
}

void Ia64DAGDependencyItemFree(t_ia64DAGDependencyItem *depItem)
{
  Free(depItem);
}


/*deze structuur bevat 1 item van de Directed Acyclic Resource Graph, die de dependencies
* tussen instructies voorstelt...*/
typedef struct _t_ia64DAGItem
{
  t_ins * ins;                   /*ptr naar instructie*/
  t_uint32 start;                /*bevat de starttijd bij het schedulen van deze instructie*/
  t_uint32 deadline;             /*bevat deadline voor deze instructie*/
  t_uint32 special_count;        /*speciaal tellertje, voor het bepalen van de ready-set*/
  t_uint32 dependency_count;     /*bevat het aantal instructies waarvan het nog afhankelijk is 
                                   en die nog niet gescheduled zijn*/
  t_uint32 flag;                 /*vlagje voor bepaalde algo's*/
  t_bool hasSpecialNext;         /*cache die bijhoudt of er ook andere dan RAWs in zitten */
  t_uint32 nNext;                /*bevat de grootte van de next-tabel*/
  
  t_ia64DAGDependencyItem **next;/*bevat lijst van pointers naar de instructies die afhankelijk zijn van deze*/
  
  
} t_ia64DAGItem;



t_ia64DAGItem * Ia64DAGItemAlloc(t_ins *ins)
{
  t_ia64DAGItem *item = (t_ia64DAGItem *)Calloc(1, sizeof(t_ia64DAGItem));

  item->ins = ins;
  item->start = item->deadline = item->special_count = item->dependency_count = 0;
  item->hasSpecialNext = FALSE;
  item->flag = 0;

  item->nNext = 0;
  item->next = NULL;

  return item;
}


void Ia64DAGItemFree(t_ia64DAGItem * item)
{
  if(item->next != NULL)
  {
    t_uint32 i = 0;

    for(i = 0; i<item->nNext; i++)
      Ia64DAGDependencyItemFree(item->next[i]);
    
    Free(item->next);
  }
  
  Free(item);
}




/*scheduling environment*/
typedef struct _t_ia64_sched_env
{
  t_ia64DAGItem **Ia64DAGList;
  t_uint32 nIa64DAGList;
  t_ia64DAGItem *Ia64DAGNopList[6]; /*M, I, X, B, F*/

} t_ia64_sched_env;




/*
static t_ia64DAGItem **Ia64DAGList = NULL;
static t_uint32 nIa64DAGList = 0;
*/



/*deze slaat dummies en nops over*/

void Ia64DAGListAlloc(t_ia64_sched_env *env, t_ia64_ins_list *insList, t_uint32 startIndex, t_uint32 maxSize)
{
  t_ins *ins;
  t_uint32 cnt = 0, index = 0;
  t_ia64_ins_list_item *item;

  env->nIa64DAGList = maxSize; //bbl->nins;
  
  env->Ia64DAGList = (t_ia64DAGItem **)Calloc(env->nIa64DAGList, sizeof(t_ia64DAGItem*));

  IA64INSLIST_FOREACH_ITEM(insList, item)
//  BBL_FOREACH_INS(bbl, ins)
  {
    ins = item->ins;
    
    if(Ia64IsNopOrDummy(ins))
      continue;
    
    if(index++ < startIndex)
      continue;

    
    
    if(cnt == env->nIa64DAGList)
      FATAL(("Error in Ia64DAGListAlloc: there are more instructions than bbl->nins !"));
    
    env->Ia64DAGList[cnt] = Ia64DAGItemAlloc(ins);
    
    cnt++;

    if(maxSize == cnt)
      break;
  }

  env->nIa64DAGList = cnt;
  
/*  
  if(cnt < env->nIa64DAGList)
    FATAL(("Error in Ia64DAGListAlloc: there are less instructions than bbl->nins !"));
*/    
}


void Ia64DAGListFree(t_ia64_sched_env *env)
{
  t_uint32 i = 0;

  for(i = 0; i<env->nIa64DAGList; i++)
    Ia64DAGItemFree(env->Ia64DAGList[i]);
  
  Free(env->Ia64DAGList);

  env->Ia64DAGList = NULL;
  env->nIa64DAGList = 0;
}







/*M, I, X, B, F*/
/*
static t_ia64DAGItem *Ia64DAGNopList[6];
*/

void Ia64DAGNopListAlloc(t_ia64_sched_env *env)
{
  env->Ia64DAGNopList[IA64_UNIT_I] = 
    Ia64DAGItemAlloc(Ia64InstructionCreateNoop(__static_scheduler_code_section__, IA64_UNIT_I));

  env->Ia64DAGNopList[IA64_UNIT_M] = 
    Ia64DAGItemAlloc(Ia64InstructionCreateNoop(__static_scheduler_code_section__, IA64_UNIT_M));

  env->Ia64DAGNopList[IA64_UNIT_B] = 
    Ia64DAGItemAlloc(Ia64InstructionCreateNoop(__static_scheduler_code_section__, IA64_UNIT_B));

  env->Ia64DAGNopList[IA64_UNIT_F] = 
    Ia64DAGItemAlloc(Ia64InstructionCreateNoop(__static_scheduler_code_section__, IA64_UNIT_F));

  env->Ia64DAGNopList[IA64_UNIT_X] = 
    Ia64DAGItemAlloc(Ia64InstructionCreateNoop(__static_scheduler_code_section__, IA64_UNIT_X));

  env->Ia64DAGNopList[IA64_UNIT_L] = 
    Ia64DAGItemAlloc(Ia64InstructionCreateNoop(__static_scheduler_code_section__, IA64_UNIT_L));

}

#if 0
void Ia64DAGNopListAddNop(t_ia64DAGItem * nop)
{
  nIa64DAGNopList++;

  Ia64DAGNopList = (t_ia64DAGItem **)Realloc(Ia64DAGNopList, nIa64DAGNopList * sizeof(t_ia64DAGItem*));
  Ia64DAGNopList[nIa64DAGNopList - 1] = nop;
}

#endif

void Ia64DAGNopListFree(t_ia64_sched_env *env)
{
  t_uint32 i = 0;

  for(i = 0; i<6; i++)
  {
    InsFree(env->Ia64DAGNopList[i]->ins);
    Ia64DAGItemFree(env->Ia64DAGNopList[i]);
  }

}

/* shallow create nop*/
t_ia64DAGItem *Ia64DAGNopListCreateNop(t_ia64_sched_env *env, t_ia64_unit unit)
{
  return env->Ia64DAGNopList[unit];
}

/*now creates a real nop */
t_ia64DAGItem *Ia64DAGNopListDuplicateNop(t_ia64_sched_env *env, t_ia64_unit unit)
{
  return Ia64DAGItemAlloc(Ia64InstructionCreateNoop(__static_scheduler_code_section__, unit));
}










t_ia64_sched_env *Ia64SchedEnvAlloc()
{
  t_ia64_sched_env *env = (t_ia64_sched_env *)Calloc(1, sizeof(t_ia64_sched_env));

  env->Ia64DAGList = NULL;
  env->nIa64DAGList = 0;
  
  /*evt noplist ook op null initialiseren*/
  
  return env;
}


void Ia64SchedEnvFree(t_ia64_sched_env *env)
{
  Free(env);
}














t_int32 Ia64DependencyAlreadyExist(t_ia64DAGItem *src, t_ia64DAGItem *dest)
{
  t_int32 i = 0;
  
  for(i = 0; i<src->nNext; i++)
    if(src->next[i]->item == dest)
      return i;
  
  return -1;
}


t_bool Ia64TypeIsStrongerThan(t_ia64DAGDependencyType type1, t_ia64DAGDependencyType type2)
{
  return type1 < type2;
}


t_uint8 Ia64CalculateLatency(t_ins *src, t_ins *dest, t_ia64DAGDependencyType type)
{
  t_uint8 standaard;

  if((type != IA64_WAW && type != IA64_WAR  && type != IA64_RAW && 
	(IA64INSTYPE(src) == IA64INSTYPE(dest))) 
      /*|| IA64INSTYPE(src) == IA64_TYPE_B*/ 
      || (IA64INSTYPE(dest) == IA64_TYPE_B && IA64OPCODEINFO(src).instr_opcode != IA64_INSTR_FCMP)
    )
    standaard = 0;
  else  
    standaard = 1;

  
  

  if(!__gedetailleerde_latency_)
    return standaard;
  else
  {
    
    if(type == IA64_RAW)
    {
      if(IA64OPCODEINFO(src).instr_opcode == IA64_INSTR_LD)
      return 2; //minstens latency van 2...

      if(IA64OPCODEINFO(src).instr_opcode == IA64_INSTR_SHR || IA64OPCODEINFO(src).instr_opcode == IA64_INSTR_SHL)
	return 3;

      if(IA64OPCODEINFO(dest).instr_opcode == IA64_INSTR_SHR || IA64OPCODEINFO(dest).instr_opcode == IA64_INSTR_SHL)
	return 3;
      
    }
    
    
    return standaard;
  }
}


void Ia64DAGListAddDependency(t_ia64DAGItem * srcItem, t_ia64DAGItem * destItem,
   t_ia64DAGDependencyType type, t_bool hasSpecialNext)
{
  t_int32 index = Ia64DependencyAlreadyExist(srcItem, destItem);
  t_uint8 gewicht = 1;

  /*patch WAWs with mem-store: they are weaker than a full WAW*/
  /*TODO: also for cmp.and... and cmp.or...*/

  /*CHANGED*/
  if(type == IA64_WAW)
  {
/*    
    if(IA64OPCODEINFO(srcItem->ins).instr_opcode == IA64_INSTR_ST && IA64OPCODEINFO(destItem->ins).instr_opcode == IA64_INSTR_ST)
    {
      type = IA64_ORDERED;
      hasSpecialNext = TRUE;
    }
*/

    /*CHANGED !!!*/
    if((IA64OPCODEINFO(srcItem->ins).instr_opcode == IA64_INSTR_CMP && 
	  IA64OPCODEINFO(destItem->ins).instr_opcode == IA64_INSTR_CMP) 
	||
	(IA64OPCODEINFO(srcItem->ins).instr_opcode == IA64_INSTR_CMP4 && 
	  IA64OPCODEINFO(destItem->ins).instr_opcode == IA64_INSTR_CMP4)
	)
    {
      t_string desc1 = IA64OPCODEINFO(srcItem->ins).desc;
      t_string desc2 = IA64OPCODEINFO(destItem->ins).desc;
      if(((strstr(desc1, "or") != NULL || strstr(desc1, "and") != NULL )) && ((strstr(desc2, "or") != NULL || strstr(desc2, "and") != NULL )))
      {
	type = IA64_ORDERED;
	hasSpecialNext = TRUE;
      }
    }


  }
  
  

  

  
  gewicht = Ia64CalculateLatency(srcItem->ins, destItem->ins, type);
  
  if(index == -1)
  {
    srcItem->hasSpecialNext |= hasSpecialNext;
    
    srcItem->nNext++;
    srcItem->next = 
      (t_ia64DAGDependencyItem **)Realloc(srcItem->next, sizeof(t_ia64DAGDependencyItem *) * srcItem->nNext);
    srcItem->next[srcItem->nNext - 1] = Ia64DAGDependencyItemAlloc(type, gewicht, destItem);

    destItem->dependency_count ++;
  }
  else
  {
    if(Ia64TypeIsStrongerThan(type, srcItem->next[index]->type))
    {
      srcItem->hasSpecialNext |= hasSpecialNext;

      srcItem->next[index]->type = type; 
    }
  }

}



/*! voorlopig ENKEL gebruiken voor als ge zeker zijt dat ze in verschillende bbl zulle komen !!!*/
t_bool Ia64InstructionsHaveDependency(t_ins *src, t_ins *dest)
{
  t_regset currUsed = Ia64UsedRegisters(dest);
  t_regset currDef = Ia64DefinedRegisters(dest);
  t_regset prevDef = Ia64DefinedRegisters(src);
  t_regset prevUsed = Ia64UsedRegisters(src);

  t_int32 regnr = 0;

  /*TODO PATCH: geen dependencies vanuit branch*/
  if(IA64INSTYPE(dest) == IA64_TYPE_B && !Ia64IsModOrLoopBranch(dest))
    return FALSE;


  /*voor alle registers diedoor deze instructies gebruikt worden, kijken of er geen voorgaande
  * instructies zijn die deze definiren (RAW-dependencies)*/
  REGSET_FOREACH_REG(currUsed, regnr)
  {
    /*kleine verbetering: p0, r0, f0 en f1 kunnen nooit overschreven worden 
    * => kunnen geen afhankelijkheid veroorzaken*/
    if(regnr == IA64_P0 || regnr == IA64_R0 || regnr == IA64_F0 || regnr == IA64_F1)
      continue;

    /*CHANGED*/
    if(regnr == IA64_CFM)
      continue;

    /*wordt dit register hierin gedefinieerd ?*/
    if(RegsetIn(prevDef, regnr))
    {
#if 0      
      /*CHANGED !!!!!!!*/
      if(regnr == IA64_AR_FPSR || regnr == IA64_PSR || regnr == IA64_RESOURCE_MEM)
	return FALSE;
      else
#endif	
	return TRUE;
    }

  }

  /*voor alle registers die door deze instructies gedefinieerd worden, kijken of er geen voorgaande
  * instructies zijn die deze definiren of gebruiker: deze instructie moet dan immers na deze 
  * voorgaande komen te staan !*/
  REGSET_FOREACH_REG(currDef, regnr)
  {

    /*kleine verbetering: p0, r0, f0 en f1 kunnen nooit overschreven worden 
    * => kunnen geen afhankelijkheid veroorzaken*/
    if(regnr == IA64_P0 || regnr == IA64_R0 || regnr == IA64_F0 || regnr == IA64_F1)
      continue;

    /*CHANGED*/
    if(regnr == IA64_CFM)
      continue;

    /*wordt dit register hierin gedefinieerd ?*/
    if(RegsetIn(prevDef, regnr))
    {
      /*CHANGED*/
//      if(regnr == IA64_AR_FPSR || regnr == IA64_PSR || regnr == IA64_RESOURCE_MEM)
//	return FALSE;
//      else
	return TRUE;

    }

    /*wordt dit register hierin gebruikt ?*/
    if(RegsetIn(prevUsed, regnr))
    {
      /*TODO: verbeteren*/
      return TRUE;
    }

  }

  return FALSE;  

}


/*
* TODO: 
* - tbit, cmp, tnat en cmp4 hebben een paar specialkes: er wordt niet altijd geschreven
* - mf, sync, srlz: moeten barriere vormen !!!
* - ALLE instructies die met registers GR, FR en PR werken moeten ook CFM "Used" maken !
* 
*/

void Ia64DAGListBuild(t_ia64_sched_env *env)
{
  t_int32 i;

  /*voor alles behalve de eerste instructie (deze heeft immers geen voorgangers)*/
  for(i = env->nIa64DAGList - 1; i > 0; i--)
  {
    t_ia64DAGItem * currItem = env->Ia64DAGList[i];
    t_regset currUsed = Ia64UsedRegisters(currItem->ins);
    t_regset currDef = Ia64DefinedRegisters(currItem->ins);
    t_int32 regnr = 0;


    /*TODO PATCH: geen dependencies vanuit branch*/
    if(IA64INSTYPE(currItem->ins) == IA64_TYPE_B)
      continue;
    

    /*voor alle registers diedoor deze instructies gebruikt worden, kijken of er geen voorgaande
    * instructies zijn die deze definiren (RAW-dependencies)*/
    REGSET_FOREACH_REG(currUsed, regnr)
    {
      t_int32 j = i - 1; /*m.o.: i>0*/

      /*kleine verbetering: p0, r0, f0 en f1 kunnen nooit overschreven worden 
      * => kunnen geen afhankelijkheid veroorzaken*/
      if(regnr == IA64_P0 || regnr == IA64_R0 || regnr == IA64_F0 || regnr == IA64_F1)
	continue;

      /*CHANGED*/
      if(regnr == IA64_CFM)
	continue;
      
      for(; j >= 0; j--) /*nu wel evt tot aan de laatste instructie gaan*/
      {
	t_ia64DAGItem * prevItem = env->Ia64DAGList[j];
	t_regset prevDef = Ia64DefinedRegisters(prevItem->ins);

	/*wordt dit register hierin gedefinieerd ?*/
	if(RegsetIn(prevDef, regnr))
	{
	  /*CHANGED !!!!!!!*/
	  /*if(regnr == IA64_AR_FPSR || regnr == IA64_PSR || regnr == IA64_RESOURCE_MEM)
	    Ia64DAGListAddDependency(prevItem, currItem, IA64_ORDERED, TRUE);
	  else*/
	    Ia64DAGListAddDependency(prevItem, currItem, IA64_RAW, FALSE);
	  
	  break; /*je mag ook stoppen: je moet enkel de eerste, directe afhankelijkheid vinden*/
	}
      }

    }
    
    /*voor alle registers die door deze instructies gedefinieerd worden, kijken of er geen voorgaande
    * instructies zijn die deze definiren of gebruiker: deze instructie moet dan immers na deze 
    * voorgaande komen te staan !*/
    REGSET_FOREACH_REG(currDef, regnr)
    {
      t_int32 j = i - 1; /*m.o.: i>0*/

      /*kleine verbetering: p0, r0, f0 en f1 kunnen nooit overschreven worden 
      * => kunnen geen afhankelijkheid veroorzaken*/
      if(regnr == IA64_P0 || regnr == IA64_R0 || regnr == IA64_F0 || regnr == IA64_F1)
  	continue;
      
      /*CHANGED*/
      if(regnr == IA64_CFM)
	continue;
      
      for(; j >= 0; j--) /*nu wel evt tot aan de laatste instructie gaan*/
      {
	t_ia64DAGItem * prevItem = env->Ia64DAGList[j];
	t_regset prevDef = Ia64DefinedRegisters(prevItem->ins);
	t_regset prevUsed = Ia64UsedRegisters(prevItem->ins);


	/*wordt dit register hierin gedefinieerd ?*/
	if(RegsetIn(prevDef, regnr))
	{
	  /*CHANGED*/
	  if(regnr == IA64_AR_FPSR || regnr == IA64_PSR || regnr == IA64_RESOURCE_MEM)
	    Ia64DAGListAddDependency(prevItem, currItem, IA64_ORDERED, TRUE);
	  else
	    Ia64DAGListAddDependency(prevItem, currItem, IA64_WAW, FALSE);
	  
	  break; /*je mag ook stoppen: je moet enkel de eerste, directe afhankelijkheid vinden*/
	}

	/*wordt dit register hierin gebruikt ?*/
	if(RegsetIn(prevUsed, regnr))
	{
	  /*CHANGED !!!!!!!*/
#if 0	  
	  if(/*regnr == IA64_AR_FPSR || regnr == IA64_PSR ||*/ regnr == IA64_RESOURCE_MEM)
	    Ia64DAGListAddDependency(prevItem, currItem, IA64_ORDERED, TRUE);
	  else
#endif	    
  	  Ia64DAGListAddDependency(prevItem, currItem, IA64_WAR, FALSE);
	 
	  /*hier mag je nog niet stoppen ! alle voorgaande instructie die dit register gebruiken
	  * blijven afhankelijk, zolang er geen definitie tussen staat */
	}

      }

    }
    
    
  }
  

}




void Ia64DAGListOrderInstruction(t_ia64_sched_env *env, 
    t_ia64_instr_opcode opcode, t_ia64DAGDependencyType type)
{
  t_int32 i;

  /*voor alles behalve de eerste instructie (deze heeft immers geen voorgangers)*/
  for(i = 0; i < env->nIa64DAGList;i++)
  {
    t_ia64DAGItem * currItem = env->Ia64DAGList[i];
    
    if(IA64OPCODEINFO(currItem->ins).instr_opcode == opcode)
    {
      t_int32 j = 0;

      for(j = 0; j < i; j++) 
      {
	Ia64DAGListAddDependency(env->Ia64DAGList[j], currItem, type, TRUE);
      }

      for(j = i + 1; j < env->nIa64DAGList; j++)
      {
	Ia64DAGListAddDependency(currItem, env->Ia64DAGList[j], type, TRUE);
      }
      
      
    }
  }  
}

#define IA64ISSTACKEDREG(reg) ((reg) >= IA64_R32 && (reg) <= IA64_R127)

t_bool Ia64InsHasStackedRegs(t_ins *ins)
{
  if(IA64ISSTACKEDREG(IA64INSIN1(ins)))
    return TRUE;

  if(IA64ISSTACKEDREG(IA64INSIN2(ins)))
    return TRUE;

  if(IA64ISSTACKEDREG(IA64INSIN3(ins)))
    return TRUE;

  if(IA64ISSTACKEDREG(IA64INSOUT1(ins)))
    return TRUE;

  if(IA64ISSTACKEDREG(IA64INSOUT2(ins)))
    return TRUE;

  return FALSE;
}


/*CHANGED !!!*/
void Ia64DAGListOrderAlloc(t_ia64_sched_env *env)
{
  t_int32 i;

  /*voor alles behalve de eerste instructie (deze heeft immers geen voorgangers)*/
  for(i = 0; i < env->nIa64DAGList;i++)
  {
    t_ia64DAGItem * currItem = env->Ia64DAGList[i];
    
    if(IA64OPCODEINFO(currItem->ins).instr_opcode == IA64_INSTR_ALLOC)
    {
      t_int32 j = 0;

      for(j = 0; j < i; j++) 
      {
	if(Ia64InsHasStackedRegs(env->Ia64DAGList[j]->ins))
	  Ia64DAGListAddDependency(env->Ia64DAGList[j], currItem, IA64_ORDERED, TRUE);
      }

      for(j = i + 1; j < env->nIa64DAGList; j++)
      {
	if(Ia64InsHasStackedRegs(env->Ia64DAGList[j]->ins))
	  Ia64DAGListAddDependency(currItem, env->Ia64DAGList[j], IA64_ORDERED, TRUE);
      }
      
      
    }
  }  
}


void Ia64DAGListOrderSpecialLoads(t_ia64_sched_env *env)
{
  t_int32 i;

  /*voor alles behalve de eerste instructie (deze heeft immers geen voorgangers)*/
  for(i = 0; i < env->nIa64DAGList;i++)
  {
    t_ia64DAGItem * currItem = env->Ia64DAGList[i];

   
    if((IA64OPCODEINFO(currItem->ins).instr_opcode == IA64_INSTR_LD ||  
	IA64OPCODEINFO(currItem->ins).instr_opcode == IA64_INSTR_LDF ||
	IA64OPCODEINFO(currItem->ins).instr_opcode == IA64_INSTR_LDFP)
	&& 
       (strchr(IA64OPCODEINFO(currItem->ins).desc, '.')) != NULL)
    {
      t_int32 j = 0;

      for(j = 0; j < i; j++) 
      {
	Ia64DAGListAddDependency(env->Ia64DAGList[j], currItem, IA64_ORDERED, TRUE);
      }

      for(j = i + 1; j < env->nIa64DAGList; j++)
      {
	Ia64DAGListAddDependency(currItem, env->Ia64DAGList[j], IA64_ORDERED, TRUE);
      }
      
      
    }
  }  
}

void Ia64DAGListOrderBranches(t_ia64_sched_env *env)
{
  t_int32 i;

  /*voor alles behalve de eerste instructie (deze heeft immers geen voorgangers)*/
  for(i = 0; i < env->nIa64DAGList;i++)
  {
    t_ia64DAGItem * currItem = env->Ia64DAGList[i];

   
    if(IA64INSTYPE(currItem->ins) == IA64_TYPE_B)
    {
      t_int32 j = 0;

      for(j = 0; j < i; j++) 
      {
	if(IA64OPCODEINFO(env->Ia64DAGList[j]->ins).instr_opcode == IA64_INSTR_FCMP)
	{
//	  FATAL(("hello"));
	  Ia64DAGListAddDependency(env->Ia64DAGList[j], currItem, IA64_WAR, FALSE);
	}
	else
	  Ia64DAGListAddDependency(env->Ia64DAGList[j], currItem, IA64_ORDERED, TRUE);
      }

      for(j = i + 1; j < env->nIa64DAGList; j++)
      {
	Ia64DAGListAddDependency(currItem, env->Ia64DAGList[j], IA64_ORDERED, TRUE);
      }
      
      
    }
  }  
}




/* voegt o.a. ordening voor bepaalde instructies uit*/
void Ia64DAGListPostBuild(t_ia64_sched_env *env)
{
  Ia64DAGListOrderAlloc(env);
  //Ia64DAGListOrderInstruction(env, IA64_INSTR_ALLOC, IA64_ORDERED);
  Ia64DAGListOrderInstruction(env, IA64_INSTR_MF, IA64_ORDERED);
  Ia64DAGListOrderInstruction(env, IA64_INSTR_FLUSHRS, IA64_RAW);
  Ia64DAGListOrderInstruction(env, IA64_INSTR_BREAK, IA64_ORDERED);
  Ia64DAGListOrderSpecialLoads(env);
  Ia64DAGListOrderBranches(env);
}





void Ia64DAGListPrintItem(FILE *out, t_ia64DAGItem * item)
{
  t_int32 i = 0;
  static char buffer[512];
  t_string color = NULL, style=NULL;
  
  Ia64InstructionPrint(item->ins, buffer);
  
  fprintf(out, "\t\"%p\" [shape=record,label=\"<f0> (0x%llx)\\n%s |<f1> %d \"]\n", 
      item, INS_CADDRESS(item->ins), buffer, item->deadline);
   
  //DiabloPrint(out, "@I\"]\n", item->ins);
  
  for(i = 0; i<item->nNext; i++)
  {
    switch(item->next[i]->type)
    {
      case IA64_RAR:
//	color = "black";
	color = "gray";
	style = "dotted";
	break;
      case IA64_RAW:
	color = "blue";
	color = "black";
	style = "solid";
	break;
      case IA64_WAR:
	color = "yellow";
	color = "black";
	style = "dotted";
	break;
      case IA64_WAW:
	color = "red";
	color = "black";
	style = "dashed";
	break;
      default:
	color = "gray";
	style = "solid";
	break;
    }
    
//    fprintf(out, "\t\"%p\" -> \"%p\" [style=solid,color=%s];\n", item, item->next[i]->item, color);
    fprintf(out, "\t\"%p\" -> \"%p\" [label=\"%d\",style=%s,color=%s];\n", item, item->next[i]->item, item->next[i]->gewicht, style, color);
  }
}

void Ia64DAGListPrint(t_ia64_sched_env *env, t_bbl *bbl)
{
  char filename[512];
  t_int32 i = 0;
  FILE * out;
  sprintf(filename, "./dots/Ia64DAG_BBL@0x%llx.dot", BBL_OLD_ADDRESS(bbl));
  out = fopen(filename,"w");
  VERIFY_FASSERT(out,("Could not open %s for writing!",filename));
  fprintf(out, "digraph \"%llx\" {\n\tnode [shape=box]\n",BBL_OLD_ADDRESS(bbl));
  
  for(i = 0; i<env->nIa64DAGList; i++)
    Ia64DAGListPrintItem(out, env->Ia64DAGList[i]);
  
  fprintf(out, "}\n");
  fclose(out);

}


t_uint32 Ia64GetEdgeWeight(t_ia64DAGItem * item, t_uint32 nextIndex)
{
//  t_ia64DAGItem * next = item->next[nextIndex]->item;

  return item->next[nextIndex]->gewicht;
    
/*  
  if((
	item->next[nextIndex]->type != IA64_WAW && item->next[nextIndex]->type != IA64_WAR &&
	item->next[nextIndex]->type != IA64_RAW && IA64INSTYPE(item->ins) == IA64INSTYPE(next->ins)) || 
	IA64INSTYPE(item->ins) == IA64_TYPE_B || IA64INSTYPE(next->ins) == IA64_TYPE_B
      )
    return 0;
  else  
  
    return 1;
*/  
}


t_int32 Ia64DAGListItemCalcHeight(t_ia64DAGItem * item)
{
  t_int32 i = 0;
  
  if(item->flag == TRUE)
    return item->deadline;
  
  for(i = 0; i<item->nNext; i++)
  {
//    item->deadline = MAX(item->deadline, Ia64DAGListItemCalcHeight(item->next[i]->item) + 1);
    
    t_uint32 inc = Ia64GetEdgeWeight(item, i);
/*
    if(item->next[i]->type != IA64_RAW && 
	(IA64INSTYPE(item->ins) == IA64INSTYPE(item->next[i]->item->ins) || 
	 (IA64INSTYPE(item->ins) == IA64_TYPE_B) || IA64INSTYPE(item->next[i]->item->ins) == IA64_TYPE_B) )
      inc = 0;
*/      
    item->deadline = MAX(item->deadline, Ia64DAGListItemCalcHeight(item->next[i]->item) + inc);
    
  }
  
  item->flag = TRUE;

  return item->deadline;
}



void Ia64DAGListCalcDeadlines(t_ia64_sched_env *env)
{
  t_int32 i = 0;
  t_int32 maxHeight = 0;
  
  for(i = 0; i<env->nIa64DAGList; i++)
    maxHeight = MAX(maxHeight, Ia64DAGListItemCalcHeight(env->Ia64DAGList[i]));
  
  /*now invert the height in order to obtain the deadlines of each instruction*/
  for(i = 0; i<env->nIa64DAGList; i++)
    env->Ia64DAGList[i]->deadline = maxHeight - env->Ia64DAGList[i]->deadline;

  /*reset marking flag*/
  for(i = 0; i<env->nIa64DAGList; i++)
    env->Ia64DAGList[i]->flag = FALSE;
  
//#if 0
  /*TODO verbeter PATCH*/
  for(i = 0; i<env->nIa64DAGList-1; i++)
  {
    /*alloc altijd zo ver mogelijk vooraan plaatsen...*/
    /*vreemd genoeg staat deze niet altijd in het begin van een BBL*/
    if(IA64OPCODEINFO(env->Ia64DAGList[i]->ins).instr_opcode == IA64_INSTR_ALLOC)
    {
      /*TODO needs more work*/
      if(env->Ia64DAGList[i]->dependency_count == 0)
	env->Ia64DAGList[i]->deadline = 0;
    }
    
    if(IA64INSTYPE(env->Ia64DAGList[i]->ins) == IA64_TYPE_DUMMY)
    {
FATAL(("Implement"));
      //InsPrint(env->Ia64DAGList[i+1]->ins, 0);
      env->Ia64DAGList[i]->deadline = env->Ia64DAGList[i+1]->deadline; //sync dummy with X
    }
  }
//#endif  
}



t_bool Ia64InsIsShift(t_ins *ins)
{
  if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_SHL || 
	IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_SHR)
    return TRUE;
  else
    return FALSE;
}




void Ia64CheckHasShift(t_ia64_sched_env *env)
{
  __has_shift_ = FALSE;
  if(__gedetailleerde_latency_)
  {
    t_uint32 i = 0;

    for(i = 0; i < env->nIa64DAGList; i++)
    {
      if(Ia64InsIsShift(env->Ia64DAGList[i]->ins))
      {
	__has_shift_ = TRUE;
	return;
      }
    }
  }
}











typedef struct _t_ia64_set
{
  t_uint32 size;
  t_bool *set;
} t_ia64_set;


t_ia64_set *Ia64SetAlloc(t_uint32 size)
{
  t_ia64_set *set = (t_ia64_set *)Calloc(1, sizeof(t_ia64_set));
  set->size = size;
  set->set = (t_bool *)Calloc(size, sizeof(t_bool));

  return set;
}

void Ia64SetFree(t_ia64_set *set)
{
  Free(set->set);
  Free(set);
}


t_bool Ia64IsSetEmpty(t_ia64_set *set)
{
  t_int32 i = 0;
  
  for(i = 0; i<set->size; i++)
    if(set->set[i])
      return FALSE;

  return TRUE;
}

t_ia64_set *Ia64SetCopy(t_ia64_set *src)
{
  t_ia64_set *dest = Ia64SetAlloc(src->size);
  t_uint32 i = 0;

  for(i = 0; i < src->size; i++)
    dest->set[i] = src->set[i];

  return dest;
  
}

t_ia64_set *Ia64SetDiff(t_ia64_set *s1, t_ia64_set *s2)
{
  t_ia64_set *dest;
  t_uint32 i = 0;
  
  if(s1->size != s2->size)
    FATAL(("Ia64SetDiff: s1->size(%d) <> s2->size(%d)", s1->size, s2->size));

  dest = Ia64SetAlloc(s1->size);
  
  for(i = 0; i < s1->size; i++)
    dest->set[i] = (s1->set[i]) & (~(s2->set[i]));

  return dest;
  
}

t_ia64_set *Ia64SetUnion(t_ia64_set *s1, t_ia64_set *s2)
{
  t_ia64_set *dest;
  t_uint32 i = 0;
  
  if(s1->size != s2->size)
    FATAL(("Ia64SetUnion: s1->size(%d) <> s2->size(%d)", s1->size, s2->size));

  dest = Ia64SetAlloc(s1->size);
  
  for(i = 0; i < s1->size; i++)
    dest->set[i] = (s1->set[i]) | (s2->set[i]);

  return dest;
  
}

t_uint32 Ia64SetCount(t_ia64_set *s)
{
  t_uint32 i, cnt = 0;

  for(i = 0; i < s->size; i++)
    cnt += s->set[i];


  return cnt;
}


t_ia64_set *Ia64SetDelete(t_ia64_set *src, t_uint32 index)
{
  t_ia64_set *dest = Ia64SetAlloc(src->size - 1);
  t_uint32 i = 0, j = 0;

  for(i = 0; i < src->size; i++)
  {
    if(i != index)
    {
      dest->set[j] = src->set[i];
      j++;
    }
  }

  return dest;
  
}

t_int32 Ia64SetFirstInSet(t_ia64_set *s)
{
  t_uint32 i = 0;

  while(i < s->size && !s->set[i])
    i++;

  return i;
}

t_int32 Ia64SetLastInSet(t_ia64_set *s)
{
  t_int32 i = s->size - 1;

  while(i >= 0 && !s->set[i])
    i--;

  return i;
}


void Ia64DAGListDecDependencyCount(t_ia64_sched_env *env, t_ia64_set *p)
{
  t_uint32 i = 0;  
  
  for(i = 0; i < p->size; i++)
  {
    t_uint32 j = 0;
    
    if(!p->set[i])
      continue;
    
    for(j = 0; j < env->Ia64DAGList[i]->nNext; j++)
        env->Ia64DAGList[i]->next[j]->item->dependency_count--;
  }

}

void Ia64DAGListIncDependencyCount(t_ia64_sched_env *env, t_ia64_set *p)
{
  t_uint32 i = 0;  
  
  for(i = 0; i < p->size; i++)
  {
    t_uint32 j = 0;
    
    if(!p->set[i])
      continue;
    
    for(j = 0; j < env->Ia64DAGList[i]->nNext; j++)
        env->Ia64DAGList[i]->next[j]->item->dependency_count++;
  }

}

void Ia64DAGListMarkAsScheduled(t_ia64_sched_env *env, t_ia64_set *p)
{
  t_uint32 i = 0;  
  
  for(i = 0; i < p->size; i++)
  {
    if(p->set[i])
      env->Ia64DAGList[i]->flag = TRUE;
  }
}

void Ia64DAGListUnmarkAsScheduled(t_ia64_sched_env *env, t_ia64_set *p)
{
  t_uint32 i = 0;  
  
  for(i = 0; i < p->size; i++)
  {
    if(p->set[i])
      env->Ia64DAGList[i]->flag = FALSE;
  }
}





typedef enum _t_ia64_template
{
  IA64_MII = 0, IA64_MII_, IA64_MI_I, IA64_MI_I_, IA64_MLX, IA64_MLX_, 
  IA64_TEMPLATEFREE1, IA64_TEMPLATEFREE2, 
  IA64_MMI, IA64_MMI_, IA64_M_MI, IA64_M_MI_, IA64_MFI, IA64_MFI_,
  IA64_MMF, IA64_MMF_, IA64_MIB, IA64_MIB_, IA64_MBB, IA64_MBB_,
  IA64_TEMPLATEFREE3, IA64_TEMPLATEFREE4,
  IA64_BBB, IA64_BBB_, IA64_MMB, IA64_MMB_, 
  IA64_TEMPLATEFREE5, IA64_TEMPLATEFREE6,
  IA64_MFB, IA64_MFB_,
  IA64_TEMPLATEFREE7, IA64_TEMPLATEFREE8,
  IA64_TEMPLATEINVALID
} t_ia64_template;


t_ia64_unit Ia64TemplateGetUnit(t_ia64_template template, t_uint8 slot)
{
  return templateField[template][slot];
}


t_bool Ia64CanTypeInSlot(t_ia64_template template, t_ia64_instr_type type, t_uint8 slot)
{
  switch(Ia64TemplateGetUnit(template, slot))
  {
    case IA64_UNIT_I:
      return type == IA64_TYPE_I || type == IA64_TYPE_A;
    case IA64_UNIT_M:
      return type == IA64_TYPE_M || type == IA64_TYPE_A;
    case IA64_UNIT_B:
      return type == IA64_TYPE_B;
    case IA64_UNIT_F:
      return type == IA64_TYPE_F;
    case IA64_UNIT_L:
      return type == IA64_TYPE_DUMMY;
    case IA64_UNIT_X:
      return type == IA64_TYPE_X;

    
    default:
      return FALSE;
  }
}



  
typedef struct _t_ia64_schedule_bundle
{
  t_ia64_template template;
  t_ia64DAGItem *slot[3];
  struct _t_ia64_schedule_bundle *next;
} t_ia64_schedule_bundle;




t_ia64_schedule_bundle *Ia64ScheduleBundleAlloc()
{
  t_ia64_schedule_bundle *bundle = (t_ia64_schedule_bundle *)Calloc(1, sizeof(t_ia64_schedule_bundle));

  bundle->template = IA64_TEMPLATEINVALID;
  bundle->slot[0] = bundle->slot[1] = bundle->slot[2] = NULL;
  bundle->next = NULL;

  return bundle;  
}

void Ia64ScheduleBundleFree(t_ia64_schedule_bundle *bundle)
{
  Free(bundle);
}


typedef struct _t_ia64_scheduling
{
  t_int32 cost;
  t_uint8 slot; 
  t_ia64_schedule_bundle *first;
  t_ia64_schedule_bundle *last;
  
} t_ia64_scheduling;



t_ia64_scheduling *Ia64SchedulingAlloc()
{
  t_ia64_scheduling *scheduling = (t_ia64_scheduling *)Calloc(1, sizeof(t_ia64_scheduling));

  scheduling->cost = 0;
  scheduling->slot = 0;
  scheduling->first = NULL;
  scheduling->last = NULL;
  
  return scheduling;
}

void Ia64SchedulingFree(t_ia64_scheduling *scheduling)
{
  t_ia64_schedule_bundle *tmp = scheduling->first;
  while(tmp)
  {
    t_ia64_schedule_bundle *next = tmp->next;
    Free(tmp);
    tmp = next;
  }
  
  Free(scheduling);
}

void Ia64DAGListFreeFromScheduling(t_ia64_scheduling *scheduling)
{
  t_ia64_schedule_bundle *tmp = scheduling->first;
  while(tmp)
  {
    Ia64DAGItemFree(tmp->slot[0]);
    Ia64DAGItemFree(tmp->slot[1]);
    Ia64DAGItemFree(tmp->slot[2]);
    
    tmp = tmp->next;
  }
  
  
}


void Ia64DAGListFreeFromSchedulingExceptNops(t_ia64_scheduling *scheduling)
{
  t_ia64_schedule_bundle *tmp = scheduling->first;
  while(tmp)
  {
    if(!Ia64IsNop(tmp->slot[0]->ins))
      Ia64DAGItemFree(tmp->slot[0]);
    if(!Ia64IsNop(tmp->slot[1]->ins))
      Ia64DAGItemFree(tmp->slot[1]);
    if(!Ia64IsNop(tmp->slot[2]->ins))
      Ia64DAGItemFree(tmp->slot[2]);
    
    tmp = tmp->next;
  }
  
  
}


/*opgelet: de lijsten in s1 en s2 wordt verwijderd uit hun lijst ! Dit om shallow copy te kunnen doen*/
t_ia64_scheduling *Ia64SchedulingAdd(t_ia64_scheduling *s1, t_ia64_scheduling *s2)
{
  t_ia64_scheduling *dest = Ia64SchedulingAlloc();

  if(s2->first == NULL)
  {
    dest->cost = s1->cost;
    dest->slot = s1->slot;
    dest->first = s1->first;
    dest->last = s1->last;
  }
  else
  {
    
    if(s1->first == NULL)
    {
      dest->cost = s2->cost;
      dest->slot = s2->slot;
      dest->first = s2->first;
      dest->last = s2->last;      
    }
    else
    
    {

      dest->slot = s2->slot;
      dest->first = s1->first;
      dest->last = s2->last;

      if(s1->slot > 0)
      {
	if(s2->first == s2->last)
	  dest->last = s1->last;

	dest->cost = s1->cost + s2->cost - 3;

	s1->last->template = s2->first->template;
	s1->last->slot[2] = s2->first->slot[2];

	if(s1->slot == 1)
	  s1->last->slot[1] = s2->first->slot[1];

	s1->last->next = s2->first->next;

	Ia64ScheduleBundleFree(s2->first);
      }
      else
      {
	dest->cost = s1->cost + s2->cost;
	s1->last->next = s2->first;
      }
    }
  }

  s1->cost = 0;
  s1->slot = 0;
  s1->first = NULL;
  s1->last = NULL;

  s2->cost = 0;
  s2->slot = 0;
  s2->first = NULL;
  s2->last = NULL;

  return dest;
}

/*opgelet: de lijsten in src wordt verwijderd uit hun lijst ! Dit om shallow copy te kunnen doen*/
t_ia64_scheduling *Ia64SchedulingMove(t_ia64_scheduling *src)
{
  t_ia64_scheduling *dest = Ia64SchedulingAlloc();

  dest->cost = src->cost;
  dest->slot = src->slot;
  dest->first = src->first;
  dest->last= src->last;

  src->cost = 0;
  src->slot = 0;
  src->first = NULL;
  src->last = NULL;  

  return dest;
}

void Ia64SchedulingReserveTemplates(t_ia64_scheduling *scheduling, t_ia64_template template, t_uint32 count)
{
  t_uint32 i = 0;

  for(i = 0; i<count; i++)
  {
    t_ia64_schedule_bundle *bundle = Ia64ScheduleBundleAlloc();
    
    bundle->template = template;
    
    if(scheduling->last)
      scheduling->last->next = bundle;
    else
      scheduling->first = bundle;
    
    scheduling->last = bundle;
    
  }
}


typedef struct _t_ia64_ins_item
{
  t_ia64DAGItem *item;
  struct _t_ia64_ins_item *next;
} t_ia64_ins_item;



t_ia64_ins_item *Ia64InsItemAlloc(t_ia64DAGItem *item)
{
  t_ia64_ins_item *insItem = (t_ia64_ins_item *)Calloc(1, sizeof(t_ia64_ins_item));

  insItem->item = item;
  insItem->next = NULL;

  return insItem;
}

void Ia64InsItemFree(t_ia64_ins_item *insItem)
{
  Free(insItem);
}

#define AANTAL_INS_TYPES 8

typedef struct _t_ia64_ins_types
{
  t_uint32 counts[AANTAL_INS_TYPES];
  t_ia64_ins_item *typeItems[AANTAL_INS_TYPES]; /*cfr t_ia64_instr_type*/
} t_ia64_ins_types;


t_ia64_ins_types *Ia64InsTypesAlloc()
{
  t_ia64_ins_types *insTypes = (t_ia64_ins_types *)Calloc(1, sizeof(t_ia64_ins_types));
  t_uint32 i;

  for(i = 0; i<AANTAL_INS_TYPES; i++)
  {
    insTypes->counts[i] = 0;
    insTypes->typeItems[i] = NULL;
  }

  return insTypes;
}

void Ia64InsTypesFree(t_ia64_ins_types *insTypes)
{
  t_uint32 i;

  for(i = 0; i<AANTAL_INS_TYPES; i++)
  {
    t_ia64_ins_item *tmp = insTypes->typeItems[i];
    
    while(tmp)
    {
      t_ia64_ins_item *next = tmp->next;
      Ia64InsItemFree(tmp);
      tmp = next;
    }
  }
  
  
  Free(insTypes);
}


t_bool Ia64InsTypesIsEmpty(t_ia64_ins_types *insTypes)
{
  t_uint32 i;

  for(i = 0; i<AANTAL_INS_TYPES; i++)
    if(insTypes->counts[i] != 0)
      return FALSE;

  return TRUE;

}

void Ia64InsTypesInsertItem(t_ia64_ins_types *insTypes, t_ia64DAGItem *item)
{
  t_ia64_instr_type type;
  t_ia64_ins_item *next, *newItem;

  type = IA64INSTYPE(item->ins);

  insTypes->counts[type]++;

  newItem = Ia64InsItemAlloc(item);

  /*TODO: dit zou sneller kunnen...*/
  next = insTypes->typeItems[type];
  if(next == NULL)
    insTypes->typeItems[type] = newItem;
  else
  {
    while(next->next)
      next = next->next;

    next->next = newItem;
  }

}


/* insType <-> resourceType: is enkel van belang bij A-type: insType is dan A, en resourceType
*  kan dan M of I zijn (maar ook A), als A enkel van de M of I mag eten */
void Ia64SchedulingFillBundles(t_ia64_scheduling *scheduling,  
    t_ia64_ins_types *insTypes, t_ia64_instr_type insType,  t_ia64_instr_type resourceType, t_uint32 count)
{
  t_uint32 i = 0;
  t_uint8 slot = 0;
  t_ia64_schedule_bundle *currBundle = scheduling->first, *lastBundle = NULL;

  while(i<count)
  {
    if(!currBundle)
    {
      FATAL(("Error not enough templates reserved !!!"));
    }

    if(currBundle->slot[slot] == NULL && Ia64CanTypeInSlot(currBundle->template, resourceType, slot))
    {
      t_ia64_ins_item *next = insTypes->typeItems[insType]->next;

      insTypes->counts[insType]--;
      currBundle->slot[slot] = insTypes->typeItems[insType]->item;
//      currBundle->slot[slot]->position = i * 3 + slot;

      if(IA64OPCODEINFO(currBundle->slot[slot]->ins).instr_opcode == IA64_INSTR_ALLOC)
      {
	if(slot == 0)
	{
	  if(lastBundle)
	    if((lastBundle->template&1) == 0)
	      lastBundle->template++;
	}
	else
	{
	  t_ia64_template template = currBundle->template;

	  if(template != IA64_M_MI && template != IA64_M_MI_)
	  {
	    if(template == IA64_MMI)
	      currBundle->template = IA64_M_MI;
	    else
	    {
	      if(template == IA64_MMI_)
		currBundle->template = IA64_M_MI_;
	      else
	      {
		/*MMF, MMB, ... => we draaien beide om */

		t_ia64DAGItem * item = currBundle->slot[0];
		currBundle->slot[0] = currBundle->slot[1];
		currBundle->slot[1] = item;

		printf("WARNING: moved alloc in bundle: see source-file for details !\n");
	      }
	    }
	  }
	}

      }

      Free(insTypes->typeItems[insType]);
      insTypes->typeItems[insType] = next;

      i++;
    }
    
    slot ++;
    
    if(slot == 3)
    {
      slot = 0;
      lastBundle = currBundle;
      currBundle = currBundle->next;
    }
    
  }
}

void Ia64SchedulingFillBundlesWithNops(t_ia64_sched_env *env, t_ia64_scheduling *scheduling, 
    t_ia64_instr_type type, t_int32 maxCount)
{
  t_ia64_schedule_bundle *ptr = scheduling->first;

  t_ia64_unit unit1, unit2;

  t_int32 count = 0;

  switch(type)
  {
    case IA64_TYPE_A:
      unit1 = IA64_UNIT_I;
      unit2 = IA64_UNIT_M;
      break;
    case IA64_TYPE_I:
      unit1 = unit2 = IA64_UNIT_I;
      break;
    case IA64_TYPE_M:
      unit1 = unit2 = IA64_UNIT_M;
      break;
    case IA64_TYPE_B:
      unit1 = unit2 = IA64_UNIT_B;
      break;
    case IA64_TYPE_F:
      unit1 = unit2 = IA64_UNIT_F;
      break;
    case IA64_TYPE_DUMMY:
      unit1 = unit2 = IA64_UNIT_L;
      break;
    case IA64_TYPE_X:
      unit1 = unit2 = IA64_UNIT_X;
      break;
      
    default:
      FATAL(("wrong type in Ia64SchedulingFillBundlesWithNops"));
      return;
  }
  

  while(ptr)
  {
    t_uint8 slot = 0;

    for(slot = 0; slot < 3; slot++)
    {
      t_ia64_unit unit = Ia64TemplateGetUnit(ptr->template, slot);

      if(count == maxCount)
	return;
      
      if(ptr->slot[slot] == NULL && (unit == unit1 || unit == unit2))
      {
	count ++;
	if(type != IA64_TYPE_DUMMY)
          scheduling->cost++;
        ptr->slot[slot] = Ia64DAGNopListCreateNop(env, unit);
      }
    }
    
    ptr = ptr->next;
  }
}



void Ia64AlgoDec(t_ia64_sched_env *env, t_ia64_scheduling *scheduling, t_ia64_ins_types *typeCnt, 
        t_ia64_instr_type type, t_int32 num)
{
  if(num <= 0)
    return;
  
  if(num < typeCnt->counts[type])
  {
    Ia64SchedulingFillBundles(scheduling, typeCnt, type, type, num); 
    return;
  }

  num -= typeCnt->counts[type];
  Ia64SchedulingFillBundles(scheduling, typeCnt, type, type, typeCnt->counts[type]);

  if(type == IA64_TYPE_M || type == IA64_TYPE_I)
  {
    if(num < typeCnt->counts[IA64_TYPE_A])
    {
      Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_A, type, num); 
      return;
    }

    num -= typeCnt->counts[IA64_TYPE_A];
    Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_A, type, typeCnt->counts[IA64_TYPE_A]);
  }

  /*nu al de rest van dat type met nops vullen !!!*/
  Ia64SchedulingFillBundlesWithNops(env, scheduling, type, -1);
}

#if 0
t_bool Ia64AlgoIsValidSchedule(t_ia64_scheduling *scheduling)
{
  t_ia64_schedule_bundle *ptr = scheduling->first;

  while(ptr)
  {
    t_uint8 slot = 0;
    
    for(slot = 0; slot < 3; slot++)
    {
      t_ia64DAGItem *item = ptr->slot[slot];

      if(item->hasSpecialNext)
      {
	t_uint32 i = 0;

	for(i = 0; i < item->nNext; i++)
	{
	  switch(item->next[i]->type)
	  {
	    /* geen WAR: mag door elkaar in instruction-group*/
	    case IA64_WAW:
	    case IA64_ORDERED:
	      {
		if(item->next[i]->item->position < item->position)
		  return FALSE;
	        break;
	      }
	    default:
	      break;
	  }
	}
      }
    }
    
    ptr = ptr->next;
  }

  return TRUE;
}

#endif


t_bool Ia64AlgoMayRemoveStop(t_ia64_template template, t_ia64DAGItem *src, t_ia64DAGItem *dest, t_ia64_scheduling *prev)
{
  if(template == IA64_M_MI)
  {
    t_uint32 i = 0;
    
    if(dest == NULL)
      return TRUE;

    /*viz: je mag stop voor alloc niet verwijderen !*/
    if(IA64OPCODEINFO(src->ins).instr_opcode == IA64_INSTR_ALLOC || IA64OPCODEINFO(dest->ins).instr_opcode == IA64_INSTR_ALLOC)
      return FALSE;

    for(i = 0; i < src->nNext; i++)
    {
      if(src->next[i]->item == dest)
      {
	if(src->next[i]->type == IA64_WAR || src->next[i]->type == IA64_ORDERED)
	{
	  /*nu, als prev != NULL => wel stop toevoegen als er een prev->last->prev is...*/
	  if(prev && prev->first != prev->last)
	  {
	    t_ia64_schedule_bundle *next = prev->first;

	    while(next->next != prev->last)
	      next = next->next;

	    if((next->template & 1) == 0)
	      next->template++;
	  }
	  
	  return TRUE;
	}
	else
	{
	  /*kijken of zo niet om RESOURCE_MEM vechten...RAW of WAW*/
	  return FALSE;
	}
      }
    }
    
    /*zou niet mogen voorkomen*/
    return FALSE;
  }
  else
    return FALSE;
}




t_ia64_scheduling * Ia64AlgoFindTemplates(t_ia64_sched_env *env, t_ia64_scheduling *prev, 
    t_ia64_ins_types *typeCnt, t_uint8 slot)
{
  t_ia64_scheduling *scheduling = Ia64SchedulingAlloc();
  t_int32 count = 0, gap = 0, i = 0;
  t_ia64_instr_type big, sml;
  t_bool hasBType, dontWaitForB;

  hasBType = typeCnt->counts[IA64_TYPE_B] > 0;
  dontWaitForB = typeCnt->counts[IA64_TYPE_B] != 1;


  /*CHANGED*/
  if(Ia64InsTypesIsEmpty(typeCnt) && !__has_shift_)
  {
    scheduling->slot = slot;
    scheduling->cost = 0;
    return scheduling;
  }

  
  if(slot == 1)
  {
    Ia64SchedulingReserveTemplates(scheduling, IA64_M_MI, 1);
    Ia64SchedulingFillBundlesWithNops(env, scheduling, IA64_TYPE_M, 1);
    
    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 1);   
    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_I, 1); 
  }

  if(slot == 2)
  {
    Ia64SchedulingReserveTemplates(scheduling, IA64_MI_I, 1);
    Ia64SchedulingFillBundlesWithNops(env, scheduling, IA64_TYPE_M, 1);
    Ia64SchedulingFillBundlesWithNops(env, scheduling, IA64_TYPE_I, 1);
      
    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_I, 1);
  }

  /*CHANGED*/
  /*kijken of de typeCnt niet ledig is => maw lege cyclus*/
  if(Ia64InsTypesIsEmpty(typeCnt))
  {
    if(slot == 0)
    {
      Ia64SchedulingReserveTemplates(scheduling, IA64_M_MI, 1);
    }
  }

  
  
  count = typeCnt->counts[IA64_TYPE_X];
  Ia64SchedulingReserveTemplates(scheduling, IA64_MLX, count);
  Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_X, IA64_TYPE_X, count);
  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, count);
  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_DUMMY, count);
  //Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_DUMMY, IA64_TYPE_DUMMY, count);


  if(typeCnt->counts[IA64_TYPE_I] >= typeCnt->counts[IA64_TYPE_M])
  {
    count = typeCnt->counts[IA64_TYPE_F];
    Ia64SchedulingReserveTemplates(scheduling, IA64_MFI, count);
    Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_F, IA64_TYPE_F, count);

    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_I, count);
    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, count);
  }
  else
  {
    t_int32 extraM = (t_int32)typeCnt->counts[IA64_TYPE_M] - (t_int32)typeCnt->counts[IA64_TYPE_I];
    
    count = typeCnt->counts[IA64_TYPE_F];
    
//    if(2*extraM >= count)
    if(extraM >= 2*count)
    {
      Ia64SchedulingReserveTemplates(scheduling, IA64_MMF, count);
      Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_F, IA64_TYPE_F, count);

      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 2*count);
    }
    else
    {
      Ia64SchedulingReserveTemplates(scheduling, IA64_MMF, extraM/2);
      Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_F, IA64_TYPE_F, extraM/2);

      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 2*(extraM/2));
      
      /* Ia64SchedulingFillBundles vermindert ook typeCnt->counts[IA64_TYPE_F] */
      count = typeCnt->counts[IA64_TYPE_F];
      
      Ia64SchedulingReserveTemplates(scheduling, IA64_MFI, count);
      Ia64SchedulingFillBundles(scheduling, typeCnt, IA64_TYPE_F, IA64_TYPE_F, count);

      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, count);
      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_I, count);
      
      
    }
  }

  
  if(typeCnt->counts[IA64_TYPE_I] > typeCnt->counts[IA64_TYPE_M])
  {
    big = IA64_TYPE_I;
    sml = IA64_TYPE_M;
  }
  else
  {
    big = IA64_TYPE_M;
    sml = IA64_TYPE_I;
  }
  
  gap = ((2*((t_int32)typeCnt->counts[sml]) - (t_int32)typeCnt->counts[big]) + 2*((t_int32)dontWaitForB))/3;

  for(i = 0; i < gap; i++)
  {
    Ia64SchedulingReserveTemplates(scheduling, IA64_MMI, 1);
    Ia64SchedulingReserveTemplates(scheduling, IA64_MII, 1);
  }

  if(gap > 0)
  {
    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 3*gap);
    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_I, 3*gap);
  }

  if(big == IA64_TYPE_M)
    Ia64SchedulingReserveTemplates(scheduling, IA64_MMI, (typeCnt->counts[big]+dontWaitForB)/2);
  else
    Ia64SchedulingReserveTemplates(scheduling, IA64_MII, (typeCnt->counts[big]+dontWaitForB)/2);

  
  Ia64AlgoDec(env, scheduling, typeCnt, sml, (typeCnt->counts[big]+dontWaitForB)/2);
  Ia64AlgoDec(env, scheduling, typeCnt, big, 2*((typeCnt->counts[big]+dontWaitForB)/2));

  if(typeCnt->counts[IA64_TYPE_A] == 1)
    Ia64SchedulingReserveTemplates(scheduling, IA64_MMI, (typeCnt->counts[IA64_TYPE_A]+2*dontWaitForB)/3);
  else
    Ia64SchedulingReserveTemplates(scheduling, IA64_MII, (typeCnt->counts[IA64_TYPE_A]+2*dontWaitForB)/3);
  
  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_A, 3*((typeCnt->counts[IA64_TYPE_A]+2*dontWaitForB)/3));

  /*en nu nog evt B:
  * TODO: volgens mij kan dit beter versmolten worden met evt A, M, F of I's !!!*/

  if(hasBType)
  {
    if(dontWaitForB)
    {
      Ia64SchedulingReserveTemplates(scheduling, IA64_BBB, (typeCnt->counts[IA64_TYPE_B]+2)/3);
      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, typeCnt->counts[IA64_TYPE_B]);
    }
    else
    {
      /*Nu kunnen we nog M's, I's en A's over hebben */
      while(typeCnt->counts[IA64_TYPE_M] + typeCnt->counts[IA64_TYPE_I] + typeCnt->counts[IA64_TYPE_A] > 2)
      {
#if 0	
	printf("hallo %d %d %d ", typeCnt->counts[IA64_TYPE_M], 
	    typeCnt->counts[IA64_TYPE_I], typeCnt->counts[IA64_TYPE_A]);
#endif	    
	if(big == IA64_TYPE_M)
	  Ia64SchedulingReserveTemplates(scheduling, IA64_MMI, 1);
	else
	  Ia64SchedulingReserveTemplates(scheduling, IA64_MII, 1);

 
	Ia64AlgoDec(env, scheduling, typeCnt, sml, 1);
	Ia64AlgoDec(env, scheduling, typeCnt, big, 2);
      }

      if(typeCnt->counts[IA64_TYPE_M] > 1 || typeCnt->counts[IA64_TYPE_I] > 1)
	FATAL(("Error in Ia64AlgoFindTemplates: counts[IA64_TYPE_M] > 1 || counts[IA64_TYPE_I] > 1"));

      if(typeCnt->counts[IA64_TYPE_A] == 2 || 
	  (typeCnt->counts[IA64_TYPE_A] == 1 && 
	   (typeCnt->counts[IA64_TYPE_I] == 1 || typeCnt->counts[IA64_TYPE_M] == 1)))
      {
	Ia64SchedulingReserveTemplates(scheduling, IA64_MIB, 1);
	Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);

	Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, typeCnt->counts[IA64_TYPE_M]);
	Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_I, typeCnt->counts[IA64_TYPE_I]);
	Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_A, typeCnt->counts[IA64_TYPE_A]);
      }
      else
	if (typeCnt->counts[IA64_TYPE_I] == 1)
	{
	  Ia64SchedulingReserveTemplates(scheduling, IA64_MIB, 1);
	  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);

	  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 1);
	  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_I, 1);
	}
	else
	{
	  t_bool todo = typeCnt->counts[IA64_TYPE_M] + typeCnt->counts[IA64_TYPE_A] > 0;

	  /*CHANGED*/

	  if(!todo && scheduling->last && (scheduling->last->slot[2] == NULL ||
	       IA64OPCODEINFO(scheduling->last->slot[2]->ins).instr_opcode == IA64_INSTR_NOP))
	  {
	    switch(scheduling->last->template)
	    {
	      /*kijken of we geen patch kunnen doorvoeren: als laatste == MMI en I=nop*/
	      case IA64_MMI:
		scheduling->last->template = IA64_MMB;
		if(scheduling->last->slot[2] != NULL)
		  scheduling->cost--;
		scheduling->last->slot[2] = NULL;
		Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);

		break;
		
	      /*kijken of we geen patch kunnen doorvoeren: als laatste == MFI en I=nop*/
	      case IA64_MFI:
		scheduling->last->template = IA64_MFB;
		if(scheduling->last->slot[2] != NULL)
		  scheduling->cost--;
		scheduling->last->slot[2] = NULL;
		Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);
		break;

	      /*kijken of we geen patch kunnen doorvoeren: als laatste == MII en I=nop*/
	      case IA64_MII:
		scheduling->last->template = IA64_MIB;
		if(scheduling->last->slot[2] != NULL)
		  scheduling->cost--;
		scheduling->last->slot[2] = NULL;
		Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);
		break;

	      /*kijken of we geen patch kunnen doorvoeren: als laatste == MI_I en I=nop*/
		/*als we in dit geval terechtkomen, dan is die stop eigenlijk overbodig !*/
		/*eigenlijk wel enkel als de te plaatsen B effectief een BR is !*/
	      case IA64_MI_I:
		/*evt checken of wel degelijk een BR-instructie is...*/
		scheduling->last->template = IA64_MIB;
		if(scheduling->last->slot[2] != NULL)
		  scheduling->cost--;
		scheduling->last->slot[2] = NULL;
		Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);
		break;

	      /*kijken of we geen patch kunnen doorvoeren: als laatste == M_MI en I=nop*/
	      /*dit mag wel enkel als we de stop mogen opheffen (dit mag bij WAW en WAR-dependency)*/
	      case IA64_M_MI:
		if((prev && scheduling->first == scheduling->last && 
		      Ia64AlgoMayRemoveStop(IA64_M_MI, prev->last->slot[0], scheduling->last->slot[1], prev)) 
		    || (scheduling->first != scheduling->last && 
		      Ia64AlgoMayRemoveStop(IA64_M_MI, scheduling->last->slot[0], scheduling->last->slot[1], NULL))) 
		  /*we kijken evt daarvoor naar de voorganger !*/
		{
		  scheduling->last->template = IA64_MMB;
		  if(scheduling->last->slot[2] != NULL)
		    scheduling->cost--;
		  scheduling->last->slot[2] = NULL;
		  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);
		  break;
		}
		else
		{
		  Ia64SchedulingReserveTemplates(scheduling, IA64_MFB, 1);
		  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);
		  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_F, 1);

		  if(typeCnt->counts[IA64_TYPE_M] + typeCnt->counts[IA64_TYPE_A] == 0)
		    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 1);
		  else
		  {
		    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, typeCnt->counts[IA64_TYPE_M]);
		    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_A, typeCnt->counts[IA64_TYPE_A]);
		  }

		  break;

		}
		
	      default:

		Ia64SchedulingReserveTemplates(scheduling, IA64_MFB, 1);
		Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);
		Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_F, 1);

		if(typeCnt->counts[IA64_TYPE_M] + typeCnt->counts[IA64_TYPE_A] == 0)
		  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 1);
		else
		{
		  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, typeCnt->counts[IA64_TYPE_M]);
		  Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_A, typeCnt->counts[IA64_TYPE_A]);
		}

		break;

	    }
	  }
	  else
	  {

	    Ia64SchedulingReserveTemplates(scheduling, IA64_MFB, 1);
	    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_B, 1);
	    Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_F, 1);

	    if(typeCnt->counts[IA64_TYPE_M] + typeCnt->counts[IA64_TYPE_A] == 0)
	      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, 1);
	    else
	    {
	      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_M, typeCnt->counts[IA64_TYPE_M]);
	      Ia64AlgoDec(env, scheduling, typeCnt, IA64_TYPE_A, typeCnt->counts[IA64_TYPE_A]);
	    }
	  }

#if 0
	  /*kijken of we geen patch kunnen doorvoeren: als laatste == MMI en I=nop*/
	  if(!todo && scheduling->last && 
	      scheduling->last->template == IA64_MMI && 
	      (scheduling->last->slot[2] == NULL ||
	       IA64OPCODEINFO(scheduling->last->slot[2]->ins).instr_opcode == IA64_INSTR_NOP))
	  {
	    scheduling->last->template = IA64_MMB;
	    scheduling->last->slot[2] = NULL;
	    Ia64AlgoDec(scheduling, typeCnt, IA64_TYPE_B, 1);
	  }
	  else
	  {
	    /*CHANGED*/
	    /*kijken of we geen patch kunnen doorvoeren: als laatste == MFI en I=nop*/
	    if(!todo && scheduling->last && 
		scheduling->last->template == IA64_MFI && 
		(scheduling->last->slot[2] == NULL ||
		 IA64OPCODEINFO(scheduling->last->slot[2]->ins).instr_opcode == IA64_INSTR_NOP))
	    {
	      scheduling->last->template = IA64_MFB;
	      scheduling->last->slot[2] = NULL;
	      Ia64AlgoDec(scheduling, typeCnt, IA64_TYPE_B, 1);
	    }
	    else
	    {



	      Ia64SchedulingReserveTemplates(scheduling, IA64_MFB, 1);
	      Ia64AlgoDec(scheduling, typeCnt, IA64_TYPE_B, 1);
	      Ia64AlgoDec(scheduling, typeCnt, IA64_TYPE_F, 1);

	      if(typeCnt->counts[IA64_TYPE_M] + typeCnt->counts[IA64_TYPE_A] == 0)
		Ia64AlgoDec(scheduling, typeCnt, IA64_TYPE_M, 1);
	      else
	      {
		Ia64AlgoDec(scheduling, typeCnt, IA64_TYPE_M, typeCnt->counts[IA64_TYPE_M]);
		Ia64AlgoDec(scheduling, typeCnt, IA64_TYPE_A, typeCnt->counts[IA64_TYPE_A]);
	      }
	    }
	  }

#endif	  
	}

    }
  }
   
  /*slot toevoegen*/
  
  /* bij laatste bundle: stop toevoegen */
  if(scheduling->last && (scheduling->last->template == IA64_MMI || scheduling->last->template == IA64_MII))
  {
    if(scheduling->last->template == IA64_MII)
    {
      if(IA64OPCODEINFO(scheduling->last->slot[2]->ins).instr_opcode == IA64_INSTR_NOP)
      {
	/*CHANGED*/
	/*evt MI_I => M_MI als eerste I = nop*/
	if(IA64OPCODEINFO(scheduling->last->slot[1]->ins).instr_opcode == IA64_INSTR_NOP)
	{
	  scheduling->last->template = IA64_M_MI;
	  scheduling->slot = 1;
	  scheduling->last->slot[1] = NULL;
	  
	}
	else
	{
	  scheduling->last->template = IA64_MI_I;
	  scheduling->slot = 2;
	}
      }
      else
      {
	scheduling->last->template++; /*viz: voegt stop toe op het einde*/
	scheduling->slot = 0;
      }
    } 
    else
    {
      if(IA64OPCODEINFO(scheduling->last->slot[1]->ins).instr_opcode == IA64_INSTR_NOP && 
	  IA64OPCODEINFO(scheduling->last->slot[2]->ins).instr_opcode == IA64_INSTR_NOP)
      {
	scheduling->last->template = IA64_M_MI;
	scheduling->slot = 1;
      }
      else
      {
	scheduling->last->template++; /*viz: voegt stop toe op het einde*/
	scheduling->slot = 0;
      }
      
    }


  }
  else
  {
    scheduling->last->template++; /*viz: voegt stop toe op het einde*/
    scheduling->slot = 0;
  }
  
  /* nu voor de veiligheid controleren of we wel alles opgebruikt hebben*/
  if(!Ia64InsTypesIsEmpty(typeCnt))
  {
     *(int*)0 =0; 

    FATAL(("Error in Ia64AlgoFindTemplates: not all types have been scheduled !"));
  }

  /* make sure all NULL slots are filled with nops */
  
  Ia64SchedulingFillBundlesWithNops(env, scheduling, IA64_TYPE_A, -1);
  Ia64SchedulingFillBundlesWithNops(env, scheduling, IA64_TYPE_B, -1);
  Ia64SchedulingFillBundlesWithNops(env, scheduling, IA64_TYPE_F, -1);

#if 0  
  /*Yet another small PATCH: alloc must be first in instr group*/
  if(slot > 0)
  {
    if(scheduling->first->next != NULL &&
      IA64OPCODEINFO(scheduling->first->next->slot[0]->ins).instr_opcode == IA64_INSTR_ALLOC)
    {
      if((scheduling->first->template&1) == 0)
	scheduling->first->template++;
    }
  }
  else
  {
    if(IA64OPCODEINFO(scheduling->first->slot[1]->ins).instr_opcode == IA64_INSTR_ALLOC)
    {
      t_ia64_template template = scheduling->first->template;
      
      if(template != IA64_M_MI && template != IA64_M_MI_)
      {
	if(template == IA64_MMI)
	  scheduling->first->template = IA64_M_MI;
	else
	{
	  if(template == IA64_MMI_)
	    scheduling->first->template = IA64_M_MI_;
	  else
	  {
	    /*MMF, MMB, ... => we draaien beide om */

	    t_ia64DAGItem * item = scheduling->first->slot[0];
	    scheduling->first->slot[0] = scheduling->first->slot[1];
	    scheduling->first->slot[1] = item;
	    
	    printf("WARNING: moved alloc in bundle: see source-file for details !\n");
	  }
	}
      }
	
      
    }
    
  }
#endif

/*  
  if(!Ia64AlgoIsValidSchedule(scheduling))
    scheduling->cost = IA64_MAXCOST;
*/  
  
  return scheduling;
}



t_ia64_set *Ia64AlgoReady(t_ia64_sched_env *env, t_ia64_set *u, t_uint32 clk)
{
  t_ia64_set *r = Ia64SetAlloc(u->size);
  t_uint32 i = 0;

  for(i = 0; i<env->nIa64DAGList; i++)
  {
    t_ia64DAGItem *item = env->Ia64DAGList[i];
    t_uint32 j = 0;

    if(!item->flag && item->dependency_count - item->special_count == 0)
    {
      for(j = 0; j < item->nNext; j++)
      {
	if(Ia64GetEdgeWeight(item, j) == 0)
	  item->next[j]->item->special_count ++;
      }    

      r->set[i] = TRUE;
    }
    
    item->special_count = 0;
  }
  
  return r;
}



t_bool Ia64HasDependencyWithBranch(t_ins *ins, t_ins *branch)
{
  if(IA64INSPRED(branch) != IA64_P0)
  {
    if(IA64INSOUT1(ins) == IA64INSPRED(branch) || IA64INSOUT2(ins) == IA64INSPRED(branch))
      return TRUE;
  }

  if(IA64_IS_BR(IA64INSOUT1(branch)))
  {
    if(IA64INSOUT1(ins) == IA64INSOUT1(branch) || IA64INSOUT2(ins) == IA64INSOUT1(branch))
      return TRUE;    

    if(IA64INSIN1(ins) == IA64INSOUT1(branch) || IA64INSIN2(ins) == IA64INSOUT1(branch) || IA64INSIN3(ins) == IA64INSOUT1(branch))
      return TRUE;    
    
  }

  if(IA64_IS_BR(IA64INSIN1(branch)))
  {
    if(IA64INSOUT1(ins) == IA64INSIN1(branch) || IA64INSOUT2(ins) == IA64INSIN1(branch))
      return TRUE;    
  }

  
  return FALSE;
}



t_ia64_set *Ia64AlgoReverseReadySkipBranchAndShift(t_ia64_sched_env *env, t_ia64_set *u, t_uint32 clk)
{
  t_ia64_set *r = Ia64SetAlloc(u->size);
  t_int32 i = 0;

  for(i = env->nIa64DAGList - 1; i>=0; i--)
  {
    t_ia64DAGItem *item = env->Ia64DAGList[i];
    t_uint32 j = 0;

    item->special_count = 0;
/*
    if(Ia64InsIsShift(item->ins))
      continue;
*/    
    if(IA64INSTYPE(item->ins) == IA64_TYPE_B)
    {
      item->special_count = 1;
      continue;
    }

    r->set[i] = TRUE;
    
    for(j = 0; j < item->nNext; j++)
    {
      t_ia64DAGItem *next = item->next[j]->item;

      if((IA64INSTYPE(next->ins) == IA64_TYPE_B && Ia64HasDependencyWithBranch(item->ins, next->ins)))
      {
        r->set[i] = FALSE;
	break;

      }
      
      if(!next->flag && (next->special_count == 0 || Ia64GetEdgeWeight(item, j) != 0 ))
      {
	r->set[i] = FALSE;
	break;
      }
    }

    if(r->set[i])
      item->special_count ++;

  }
  
  return r;
}


t_ia64_set *Ia64AlgoNotYetCritical(t_ia64_sched_env *env, t_ia64_set *r, t_uint32 clk)
{
  t_ia64_set *s = Ia64SetAlloc(r->size);
  t_uint32 i = 0;

  for(i = 0; i<env->nIa64DAGList; i++)
  {
    if(r->set[i] && env->Ia64DAGList[i]->deadline != clk)
      s->set[i] = TRUE;
  }

  
  
  return s;
}

t_ia64_ins_types *Ia64AlgoCounts(t_ia64_sched_env *env, t_ia64_set *p)
{
  t_uint32 i = 0;
  t_ia64_ins_types *insTypes = Ia64InsTypesAlloc();

  for(i = 0; i<p->size; i++)
  {
    if(!p->set[i])
      continue;
    
    Ia64InsTypesInsertItem(insTypes, env->Ia64DAGList[i]);
  }
  
  return insTypes;
}



static t_uint32 __bestCost__ = IA64_MAXCOST;

t_uint32 Ia64AlgoGetBestCostSoFar()
{
  return __bestCost__;
}

void Ia64AlgoTestAndSetBestCostSoFar(t_uint32 cost)
{
  __bestCost__ = MIN(__bestCost__, cost);
}


//TODO
t_bool Ia64AlgoPrune(t_uint32 prevCost, t_int32 cCost, t_int32 uCost, t_ia64_set *s, t_ia64_set *c)
{
/*  
  if(prevCost + cCost + uCost >= Ia64AlgoGetBestCostSoFar())
    return TRUE;
*/  
  return FALSE;
}



/* TODO: volgens dat random shiften zoals beschreven in dat artikel */
void Ia64SetNext(t_ia64_set *s)
{
  t_uint32 i = 0;
  t_bool carry = TRUE;

  while(carry)
  {
    if(s->set[i])
      s->set[i] = 0;
    else
    {
      s->set[i] = 1;
      carry = FALSE;
    }
    i++;
  }
}

t_ia64_set *Ia64SetApplyMask(t_ia64_set *src, t_ia64_set *mask)
{
  t_ia64_set *dest = Ia64SetAlloc(src->size);
  t_uint32 i = 0, j = 0;


  for(i = 0; i < src->size; i++)
    if(src->set[i])
      if(mask->set[j++])
	dest->set[i] = TRUE;
      
  
  return dest;
}


static t_uint32 __start_stage__ = 0;
static t_uint32 __stop_stage__ = 2;

#define IA64_ALGO_MAX_DEPTH_EAGER_LAZY 8
#define IA64_ALGO_MAX_DEPTH_FULL 2
#define IA64_ALGO_FULL_MAX_COUNT 8


t_ia64_scheduling * Ia64AlgoScheduleRecursive(t_ia64_sched_env *env, t_ia64_scheduling *prev, t_uint32 clk, t_ia64_set *u, t_uint8 slot, 
    t_uint32 prevCost, t_uint32 choiceDepth)
{
  t_ia64_scheduling *bestScheduling;
  
  t_ia64_set *r, *s, *m, *mask = NULL;
  t_uint32 stage = 0;
  t_bool done = FALSE;
  t_uint32 start_stage = 0, stop_stage = 2;
  t_bool fullSearch = FALSE;
  t_uint32 sCount = 0, step = 1, maxCount = 0;
  
  bestScheduling = Ia64SchedulingAlloc();
  bestScheduling->cost = IA64_MAXCOST;


  r = Ia64AlgoReady(env, u, clk);
  s = Ia64AlgoNotYetCritical(env, r, clk);
  m = Ia64SetDiff(r, s);

  sCount = Ia64SetCount(s);

  if(clk < IA64_ALGO_FULL_MAX_COUNT)
    maxCount = IA64_ALGO_FULL_MAX_COUNT - clk;

  fullSearch = IA64_ALGO_MAX_DEPTH_FULL != -1 
    && sCount > 0 
    && maxCount > 0;
//    && choiceDepth <= IA64_ALGO_MAX_DEPTH_FULL;

  if(fullSearch)
  {
    mask = Ia64SetAlloc(sCount);

    start_stage = 0;
    stop_stage = (1<<sCount);

//    step = (stop_stage + IA64_ALGO_FULL_MAX_COUNT) / IA64_ALGO_FULL_MAX_COUNT;
    step = (stop_stage + maxCount) / maxCount;
  }
  else
  {
    if(choiceDepth > IA64_ALGO_MAX_DEPTH_EAGER_LAZY)
    //  start_stage = 1;
      stop_stage = 1;

  }
  
  
  for(stage = start_stage; stage < stop_stage && !done; stage +=step)
  {
    t_ia64_set *c = NULL, *p, *u_min_p;
    t_ia64_ins_types *curTypes, *leftTypes;
    t_ia64_scheduling *cScheduling = NULL, *uScheduling = NULL, *rScheduling, *costScheduling;
    t_bool hasBType = FALSE;

    if(fullSearch)
    {
      t_uint32 stepi = 0;

      if(stage + step >=stop_stage)
	c = Ia64SetCopy(s);
      else      
      {
	c = Ia64SetApplyMask(s, mask);
	
	while(stepi++ != step)
	  Ia64SetNext(mask);
      }
    }
    else
    {

      if(stage == 1)
	c = Ia64SetCopy(s); 
      if(stage == 0)
	c = Ia64SetAlloc(u->size); //empty set

    }
    
    p = Ia64SetUnion(c, m);

    Ia64DAGListMarkAsScheduled(env, p);

    curTypes = Ia64AlgoCounts(env, p);

    hasBType = curTypes->counts[IA64_TYPE_B] > 0;

    cScheduling = Ia64AlgoFindTemplates(env, prev, curTypes, slot);    


    u_min_p = Ia64SetDiff(u, p);

    if(!Ia64IsSetEmpty(u_min_p))
    {
      /*als er nog zijn: eerst kijken of er geen B-type instructie */
      if(hasBType)
      {
	cScheduling->cost = IA64_MAXCOST; /* zal zeker geen goeie oplossing zijn */
	costScheduling = Ia64SchedulingMove(cScheduling);
      }
      else
      {

	leftTypes = Ia64AlgoCounts(env, u_min_p);
#if 0
	uScheduling = Ia64AlgoFindTemplates(leftTypes, cScheduling->slot);
	/* TODO: als cSched->slot > 0: kost - 3 doen !*/
	if(cScheduling->slot > 0)
	  uScheduling->cost -= 3;
#endif
	if(TRUE/*!Ia64AlgoPrune(prevCost, cScheduling->cost, uScheduling->cost, s, c)*/)
	{
	  Ia64DAGListDecDependencyCount(env, p);

	  rScheduling = Ia64AlgoScheduleRecursive(env, cScheduling->first?cScheduling:NULL, clk + 1, u_min_p, cScheduling->slot, 
	      prevCost + cScheduling->cost, choiceDepth + 1/* (1-Ia64IsSetEmpty(s))*/  );

	  Ia64DAGListIncDependencyCount(env, p);
/*
	  if(rScheduling->cost == IA64_MAXCOST)
	    costScheduling = Ia64SchedulingMove(rScheduling);
	  else
*/	  
  	    costScheduling = Ia64SchedulingAdd(cScheduling, rScheduling);

	  Ia64SchedulingFree(rScheduling);

	}
	else
	{
	  costScheduling = Ia64SchedulingMove(cScheduling);
	}

	Ia64InsTypesFree(leftTypes);
      }
    }
    else
    {
      costScheduling = Ia64SchedulingMove(cScheduling);
    }

    if(bestScheduling->cost > costScheduling->cost - (slot > 0?3:0))
    {
      Ia64SchedulingFree(bestScheduling);
      bestScheduling = Ia64SchedulingMove(costScheduling);
      
      //Ia64AlgoTestAndSetBestCostSoFar(bestScheduling->cost + prevCost);
    }
#if 0
    if(uScheduling && bestScheduling->cost == uScheduling->cost + cScheduling->cost/* hier niet +cShed->cost ??*/)
      done = TRUE;
#endif
    if(uScheduling)
    {
      Ia64SchedulingFree(uScheduling);
      uScheduling = NULL;
    }

    Ia64SchedulingFree(costScheduling);

    Ia64DAGListUnmarkAsScheduled(env, p);

    Ia64SchedulingFree(cScheduling);


    Ia64InsTypesFree(curTypes);


    Ia64SetFree(u_min_p);
    Ia64SetFree(p);
    Ia64SetFree(c);
  }

  if(mask != NULL)
    Ia64SetFree(mask);
  
  Ia64SetFree(r);
  Ia64SetFree(s);
  Ia64SetFree(m);

  return bestScheduling;
}


void Ia64ApplySchedulingToBbl(t_ia64_sched_env *env, t_ia64_scheduling *scheduling, t_bbl *bbl)
{
  t_ia64_schedule_bundle *bundle = scheduling->first;

  /* start from scratch */
  bbl->nins = 0;
  bbl->ins_first = bbl->ins_last = NULL;
  
  while(bundle)
  {
    t_uint32 slot = 0;

    for(slot = 0; slot < 3; slot++)
    {
    
      if(IA64OPCODEINFO(bundle->slot[slot]->ins).instr_opcode == IA64_INSTR_NOP ||
	  IA64INSTYPE(bundle->slot[slot]->ins) == IA64_TYPE_DUMMY)
	bundle->slot[slot] = Ia64DAGNopListDuplicateNop(env, Ia64TemplateGetUnit(bundle->template, slot));
    
      if(IA64INSTYPE(bundle->slot[slot]->ins) == IA64_TYPE_A)
      {
	IA64INSATTR(bundle->slot[slot]->ins) &= ~IA64_INSTR_FLAG_USE_M_FOR_A;
	IA64INSATTR(bundle->slot[slot]->ins) &= ~IA64_INSTR_FLAG_USE_I_FOR_A;

	if(Ia64TemplateGetUnit(bundle->template, slot) == IA64_UNIT_I)
	  IA64INSATTR(bundle->slot[slot]->ins) |= IA64_INSTR_FLAG_USE_I_FOR_A;
	else
	  if(Ia64TemplateGetUnit(bundle->template, slot) == IA64_UNIT_M)
	    IA64INSATTR(bundle->slot[slot]->ins) |= IA64_INSTR_FLAG_USE_M_FOR_A;
	  else FATAL(("A must be mapped onto I or M !"));
      }
      
      INS_PREV(bundle->slot[slot]->ins) = INS_NEXT(bundle->slot[slot]->ins) = NULL;
      InsAppendToBbl(bundle->slot[slot]->ins, bbl);
    }

    /*CHANGED !!!!*/
    if(bundle == scheduling->last)
    {
      if(bundle->template == IA64_M_MI && Ia64IsNop(bundle->slot[1]->ins) && Ia64IsNop(bundle->slot[2]->ins))
	bundle->template = IA64_MMI;

      if(bundle->template == IA64_MI_I && Ia64IsNop(bundle->slot[2]->ins))
	bundle->template = IA64_MII;
      
    }
    

    /* now format the stops... */
    if(bundle->template == IA64_M_MI || bundle->template == IA64_M_MI_)
      IA64INSINSTRGROUP(bundle->slot[0]->ins) = GetCurrentInstrGroupAndInc();
    else
      IA64INSINSTRGROUP(bundle->slot[0]->ins) = GetCurrentInstrGroup();

    if(bundle->template == IA64_MI_I || bundle->template == IA64_MI_I_)
      IA64INSINSTRGROUP(bundle->slot[1]->ins) = GetCurrentInstrGroupAndInc();
    else
      IA64INSINSTRGROUP(bundle->slot[1]->ins) = GetCurrentInstrGroup();

    /*CHANGED !!!!*/
    if(bundle == scheduling->last || (bundle->template & 1) == 1) /*oneven ? => stop*/
      IA64INSINSTRGROUP(bundle->slot[2]->ins) = GetCurrentInstrGroupAndInc();
    else
      IA64INSINSTRGROUP(bundle->slot[2]->ins) = GetCurrentInstrGroup();

    
    
    bundle = bundle->next;
  }
}


void Ia64BblPrint(t_bbl *bbl)
{
  t_ins *ins;
  
  BblPrint(bbl, 0);

  ins = bbl->ins_first;
  while(ins)
  {
    //InsPrint(ins, 0);
    ins = INS_NEXT(ins);
  }
  
}


void Ia64PatchTemplatesForAllocs(t_ia64_scheduling *scheduling)
{
  if(IA64OPCODEINFO(scheduling->first->slot[0]->ins).instr_opcode == IA64_INSTR_ALLOC)
  {
    if(scheduling->first->template == IA64_M_MI)
      scheduling->first->template = IA64_MMI;
    else if(scheduling->first->template == IA64_M_MI_)
      scheduling->first->template = IA64_MMI_;
  }
}


/*-----------------------------------------------------------------------*/

t_ia64DAGItem *Ia64AlgoGetHighestPriority(t_ia64_sched_env *env, t_ia64_set *r, t_uint32 *index)
{
  t_ia64DAGItem *item = NULL;
  t_uint32 i = 0;

  for(i = 0; i<env->nIa64DAGList; i++)
  {
    if(r->set[i])
    {
      if(item == NULL)
      {
	*index = i;
	item = env->Ia64DAGList[i];
      }
      else
      {
	if(item->deadline > env->Ia64DAGList[i]->deadline)
	{
	  *index = i;
	  item = env->Ia64DAGList[i];
	}
      }
    }
  }

  
  
  return item;
}


void Ia64ExpandInsTypes(t_ia64_ins_types ***insTypes, t_uint32 newSize)
{
  *insTypes = (t_ia64_ins_types **)Realloc(*insTypes, sizeof(t_ia64_ins_types *) * newSize);
  (*insTypes)[newSize - 1] = Ia64InsTypesAlloc();
}

t_uint32 Ia64CountInstructionsInCycle(t_ia64_ins_types *insTypes)
{
  t_uint32 i = 0, count = 0;

  for(i = 0; i < AANTAL_INS_TYPES; i++)
    count += insTypes->counts[i];
  
  return count;
}


t_bool Ia64ResourceConstraintsSatisfied(t_ia64_ins_types *insTypes, t_ia64_instr_type type)
{
#define __AANTAL_M 2
#define __AANTAL_I 2  
  
  t_int32 totalA = insTypes->counts[IA64_TYPE_A];
  t_int32 totalM = insTypes->counts[IA64_TYPE_M];
  t_int32 totalI = insTypes->counts[IA64_TYPE_I] + insTypes->counts[IA64_TYPE_X];
  t_int32 totalAforM = MAX(0, totalA - (__AANTAL_M - totalI));
  t_int32 totalAforI = MAX(0, totalA - (__AANTAL_I - totalM));

  //return TRUE;
  
  switch(type)
  {
    case IA64_TYPE_A:
      return totalA + totalM + totalI + 1 <= __AANTAL_I + __AANTAL_M;
    case IA64_TYPE_M:
      return totalM + 1 + totalAforM <= __AANTAL_M;
    case IA64_TYPE_I:
    case IA64_TYPE_X:
      return totalI + 1 + totalAforI <= __AANTAL_I;
    case IA64_TYPE_F:
      return insTypes->counts[IA64_TYPE_F] + 1 <= 2;
    case IA64_TYPE_B:
      return insTypes->counts[IA64_TYPE_B] + 1 <= 3;
      
    
    case IA64_TYPE_DUMMY:
    default:
      FATAL(("Invalid instruction type in Ia64ResourceConstraintsSatisfied"));
      return FALSE;
  }
  
  return FALSE;
}


t_bool Ia64AlgoWaitForB(t_ia64_sched_env *env, t_ia64_set *ready, t_ia64_set *todo)
{
  t_uint32 i = 0;
  t_bool hasB = FALSE;
  t_bool hasAM = FALSE;

  if(Ia64SetCount(todo) != 2)
    return FALSE;
  
  if(Ia64SetCount(ready) != Ia64SetCount(todo))
    return FALSE;

  for(i = 0; i < env->nIa64DAGList; i++)
  {
    if(!ready->set[i])
      continue;
    
    if(IA64INSTYPE(env->Ia64DAGList[i]->ins) == IA64_TYPE_B)
      hasB = TRUE;
    if(IA64INSTYPE(env->Ia64DAGList[i]->ins) == IA64_TYPE_A || IA64INSTYPE(env->Ia64DAGList[i]->ins) == IA64_TYPE_M)
      hasAM = TRUE;
  }

  if(!hasB || !hasAM)
    return FALSE;

  return TRUE;  
}


t_ia64_scheduling *Ia64ListScheduling(t_ia64_sched_env *env)
{
  t_ia64_set *todo, *ready;
  /*t_uint32 globalCycle = 0;*/

  t_uint32 nInsTypes = 0, i = 0;
  t_ia64_ins_types **insTypes = NULL;

  t_ia64_scheduling *scheduling = NULL;

  t_uint8 slot = 0;

  /*t_bool hasShiftInstruction = FALSE;*/

  

  Ia64ExpandInsTypes(&insTypes, ++nInsTypes);

  todo = Ia64SetAlloc(env->nIa64DAGList);
  for(i = 0; i < todo->size; i++)
    todo->set[i] = TRUE;

  
  
  ready = Ia64AlgoReady(env, todo, 0);

  while(!Ia64IsSetEmpty(todo))
  {
    t_uint32 index = 0;
    t_ia64DAGItem *item = Ia64AlgoGetHighestPriority(env, ready, &index);
    t_uint32 localCycle = item->start;
    t_bool OK = TRUE;
/*
    if(!hasShiftInstruction && (IA64OPCODEINFO(item->ins).instr_opcode == IA64_INSTR_SHL || IA64OPCODEINFO(item->ins).instr_opcode == IA64_INSTR_SHR))
      hasShiftInstruction = TRUE;
*/
    /*CHANGED*/
    /*patch, zodanig dat B meer geintegreerd kan worden*/
/*    
    if(Ia64AlgoWaitForB(ready, todo))
    {
      printf("********* Wait for B\n");
      localCycle = nInsTypes;
    }
*/    
    while(localCycle >= nInsTypes)
    {
      Ia64ExpandInsTypes(&insTypes, ++nInsTypes);
    }


    do
    {
      //      printf("Hallo %d %d\n", index, localCycle);
      //#if 0

      while(Ia64CountInstructionsInCycle(insTypes[localCycle]) >= 6)
      {
	//	printf("InstrCount >= 6\n");

	localCycle++; 
	if(localCycle == nInsTypes)
	  Ia64ExpandInsTypes(&insTypes, ++nInsTypes);
      }

      if(IA64INSTYPE(item->ins) == IA64_TYPE_F || IA64INSTYPE(item->ins) == IA64_TYPE_B)
      {
	while(insTypes[localCycle]->counts[IA64_TYPE_F] == 2 ||  /*CHANGED && to ||*/
	    insTypes[localCycle]->counts[IA64_TYPE_B] == 3)
	{
	  localCycle++; 
	  if(localCycle == nInsTypes)
	    Ia64ExpandInsTypes(&insTypes, ++nInsTypes);
	}
      }
      //#endif

      /*CHANGED*/
      if(Ia64ResourceConstraintsSatisfied(insTypes[localCycle], IA64INSTYPE(item->ins)))
      {

	Ia64InsTypesInsertItem(insTypes[localCycle], item);

	todo->set[index] = FALSE;
	OK = TRUE;
      }
      else
      {
	//	printf("constraints not satisfied\n");

	localCycle++; 
	if(localCycle == nInsTypes)
	  Ia64ExpandInsTypes(&insTypes, ++nInsTypes);


	OK = FALSE;
      }

    } while (!OK);

    /* Now add newly ready instructions */

    /*mark as scheduled */
    env->Ia64DAGList[index]->flag = TRUE;

    for(i = 0; i < env->Ia64DAGList[index]->nNext; i++)
    {
      t_ia64DAGItem *next = env->Ia64DAGList[index]->next[i]->item;
      t_uint32 inc = Ia64GetEdgeWeight(item, i);
            
      next->dependency_count--;

      next->start = MAX(next->start, localCycle + inc);
    }

    Ia64SetFree(ready);

    ready = Ia64AlgoReady(env, todo, 0);

  }

  
  for(i = 0; i < nInsTypes; i++)
  {
    t_ia64_scheduling *sched;
    
    /*als een cyclus leeg zou zijn, sla die dan over, in plaats van een leeg ding te krijgen...*/
/*    
    if(Ia64InsTypesIsEmpty(insTypes[i]) && !hasShiftInstruction)
      continue;
*/    
    sched = Ia64AlgoFindTemplates(env, scheduling, insTypes[i], slot);
/*
    if(IA64INSTYPE(sched->last->slot[2]->ins) == IA64_TYPE_B && i+1 < nInsTypes)
      FATAL(("error"));
*/    
    if(scheduling == NULL)
      scheduling = Ia64SchedulingMove(sched);
    else
      scheduling = Ia64SchedulingAdd(scheduling, sched);

    slot = scheduling->slot;
    
    Ia64SchedulingFree(sched);
    
    Ia64InsTypesFree(insTypes[i]);
  }

  Free(insTypes);
  
  
  Ia64SetFree(todo);
  Ia64SetFree(ready);

  return scheduling;
}





//#define IA64_RESPECT_LATENCIES_FOR_NOP_SCHEDULER


#define IA64_PRINT_DEBUG /*TRUE*/ FALSE
#define IA64_PRINT_BBL_NINS_THRESHOLD 500


#define MAX_BBL_SCHEDULING_SIZE 20
#define BBL_SPLIT_THRESHOLD  30


/*
#define MAX_BBL_SCHEDULING_SIZE 25
#define BBL_SPLIT_THRESHOLD  40
*/
/*
#define MAX_BBL_SCHEDULING_SIZE 30
#define BBL_SPLIT_THRESHOLD  50
*/

//#if 0

static t_uint32 minimumNops = 0;
static t_uint32 verwijderbareNops = 0;

static t_uint32 __bblCount = 0;
static t_uint32 __hotCount = 0;



t_uint32 Ia64CountNopsAndDummiesInBbl(t_bbl *bbl)
{
  t_ins *ins;
  t_uint32 cnt = 0;

  BBL_FOREACH_INS(bbl, ins)
  {
    if(Ia64IsNopOrDummy(ins))
      cnt++;
  }

  return cnt;
}



t_ia64_scheduling *Ia64PerformScheduling(t_ia64_sched_env *env, t_ia64_ins_list *insList, t_bbl *bbl)
{
  t_int32 i;
  t_ia64_scheduling *totalScheduling = NULL, *scheduling = NULL;
  t_ia64_set *u;
  /*t_bool finished = TRUE;*/
  t_uint32 splitSize = 0, splitCount = 0, startIndex = 0;

  t_uint32 realNins = insList->nins;


  
  if(global_options.nopscheduler)
  {
    if(realNins >= BBL_SPLIT_THRESHOLD)
    {
      if(realNins >= 2*MAX_BBL_SCHEDULING_SIZE)
	splitCount = (realNins + MAX_BBL_SCHEDULING_SIZE)/MAX_BBL_SCHEDULING_SIZE;
      else
	splitCount = 2;

      splitSize = (realNins + splitCount - 1)/splitCount;
    }
    else
      splitSize = realNins;
  }

  if(global_options.listscheduler)
    splitSize = realNins;

  
  do
  {

    Ia64DAGListAlloc(env, insList, startIndex, splitSize);

    startIndex += splitSize;

    Ia64DAGListBuild(env);
    Ia64DAGListPostBuild(env);
    Ia64DAGListCalcDeadlines(env);
    /*print dots of DAG - werkt fantastisch goed !*/
/*    
    if(bbl)
      Ia64DAGListPrint(env, bbl);
*/

    u = Ia64SetAlloc(env->nIa64DAGList);
    for(i = 0; i < u->size; i++)
      u->set[i] = TRUE;

  if(global_options.listscheduler)  
  {
    __gedetailleerde_latency_ = TRUE;
    Ia64CheckHasShift(env);
    scheduling = Ia64ListScheduling(env);
  }
  
  if(global_options.nopscheduler)  
  {

#ifdef IA64_RESPECT_LATENCIES_FOR_NOP_SCHEDULER
    __gedetailleerde_latency_ = TRUE;
#else    
    __gedetailleerde_latency_ = FALSE;
#endif    
    Ia64CheckHasShift(env);
    scheduling = Ia64AlgoScheduleRecursive(env, totalScheduling, 0, u, totalScheduling == NULL?0: totalScheduling->slot, 0, 0);
  }    

    if(totalScheduling == NULL)
      totalScheduling = Ia64SchedulingMove(scheduling);
    else
    {
      t_ia64_scheduling *tmpSched = totalScheduling;
      totalScheduling = Ia64SchedulingAdd(tmpSched, scheduling);
      Ia64SchedulingFree(tmpSched);
    }


    Ia64SchedulingFree(scheduling); 
    Ia64SetFree(u);

  } while(startIndex < realNins);

  return totalScheduling;
  
}



void Ia64ScheduleBbl(t_section* code, t_bbl* bbl)
{
  t_uint32 nops;
  t_ia64_scheduling *totalScheduling = NULL;
  t_bool printFlag = FALSE;

  t_uint32 nopsAndDummies = 0;
  t_uint32 realNins = 0;

  t_ia64_sched_env *env;

  t_ia64_ins_list *insList;

  if(!global_options.listscheduler && !global_options.nopscheduler)
  {
    Ia64BBLBundeling(code, bbl);
    return;
  }

  if(global_options.listscheduler && global_options.nopscheduler)
    FATAL(("Both options cannot be activated !!!"));

#if 0
  
  if(BBL_PROFILE_INFO(bbl) >= 1000000)
  {
    __hotCount++;
    __bblCount++;
    
    printf("hots: %d\n", __hotCount);
  }
  
  if(__bblCount >= global_options.debugcounter)
  //if(++__bblCount == 22 || __bblCount == 23)
  {
    
    if(BBL_PROFILE_INFO(bbl) >= 1000000 && __bblCount >= global_options.debugcounter && __bblCount <= global_options.debugcounter + 20)  
    {
      printf("*********** <%lld>\n", BBL_PROFILE_INFO(bbl));
      Ia64BblPrint(bbl);
    }

    //printf("%d\n", global_options.debugcounter);
    Ia64BBLBundeling(code, bbl);
    return;
  }

  
#endif

  
  
//  __static_scheduler_code_section__ = code;

  /*Data: need not be scheduled*/
  if(bbl->ins_first && IA64INSTYPE(bbl->ins_first) == IA64_TYPE_DATA)
    return;

  if(IA64_PRINT_DEBUG && bbl->nins >= IA64_PRINT_BBL_NINS_THRESHOLD)
    printFlag = TRUE;


  if(printFlag)
    Ia64BblPrint(bbl);

  nopsAndDummies = Ia64CountNopsAndDummiesInBbl(bbl);

  realNins = bbl->nins - nopsAndDummies;
  
#if 0  
  /* verwijder eerst alle resterende nops ...*/
  nops = Ia64RemoveNops(bbl);
  Ia64RemoveDummies(bbl);
#endif
  

  if(realNins == 0)  
  {
    Ia64RemoveNops(bbl);
    Ia64RemoveDummies(bbl);
    
    if(bbl->ins_first != 0 || bbl->ins_last != 0)
      FATAL(("Error in Ia64ScheduleBbl: bbl->nins = 0, but bbl->ins_first != 0 || bbl->ins_last != 0 !"));
    return;
  }


  insList = Ia64InsListAlloc(bbl);

  env = Ia64SchedEnvAlloc();
  Ia64DAGNopListAlloc(env);
  

  totalScheduling = Ia64PerformScheduling(env, insList, bbl);
    


  /*ze hier pas effectief verwijderen*/
  nops = Ia64RemoveNops(bbl);
  Ia64RemoveDummies(bbl);

  Ia64ApplySchedulingToBbl(env, totalScheduling, bbl);


  /*DEBUGGING*/
  if(IA64_PRINT_DEBUG && nops < totalScheduling->cost)
  {
    t_uint32 align = (totalScheduling->cost - nops)%3;
    nops += align;
    verwijderbareNops+=totalScheduling->cost - nops;


    if(totalScheduling->cost - nops > 0)
    {
      printf("Removed NOPs(%d) <-> Final NOPs(%d) @ 0x%llx\n", nops, totalScheduling->cost, BBL_OLD_ADDRESS(bbl));
      printf("Totaal verwijderbare nops: %d\n", verwijderbareNops);
      Ia64BblPrint(bbl);
      printf("\n");
    }

  }  

  /*DEBUGGING*/
  if(printFlag)
  {
    Ia64BblPrint(bbl);
    printf("\n");
  }


  Ia64DAGListFreeFromScheduling(totalScheduling);
  Ia64SchedulingFree(totalScheduling);
  Ia64DAGNopListFree(env);
  Free(env->Ia64DAGList);
  Ia64SchedEnvFree(env);

  Ia64InsListFree(insList);

}














/* ---------------------------------------------------------------------------- */

/*Global scheduling code */

/*kan dit eig niet met de oorspronkelijke?*/
t_ins * Ia64InsDup(t_ins * ins)
{
//    FATAL(("Use orig?"));

    t_ins * ret;
    t_section * sec;
    t_object * obj;
    t_reloc_ref * refs=RELOCATABLE_REFERS_TO(ins);

    sec=InsGetSection(ins);
    obj=sec->obj;
/*
    if (sec->type!=FLOWGRAPHING_CODE_SECTION && sec->type!=FLOWGRAPHED_CODE_SECTION)
    {
      FATAL(("Cannot duplicate instructions in fase %c",sec->type));
    }
*/
    ret=InsNewForSec(sec);
    memcpy(ret,ins,obj->description->decoded_instruction_size);
    INS_OLD_ADDRESS(ret)= NULL_GENERIC;
    INS_CADDRESS(ret)= NULL_GENERIC;
    INS_OLD_SIZE(ret)= NULL_GENERIC;
    INS_CSIZE(ret)=INS_CSIZE(ins);
    T_RELOCATABLE(ret)->refedby = NULL;
    INS_PREV(ret)=INS_NEXT(ret)=NULL;

    /* all node structures (cf ud-graph) must be set to NULL.
     * the copied instruction is _never_ part of the ud-graph
     * until it is inserted in the graph */
    NODE_NEXT(ret) = NODE_PREV(ret) = NULL;
    NODE_SUCC_FIRST(ret) = NODE_SUCC_LAST(ret) = NULL;
    NODE_PRED_FIRST(ret) = NODE_PRED_LAST(ret) = NULL;

    RELOCATABLE_REFERS_TO(ret)=NULL;
    RELOCATABLE_REFED_BY(ret)=NULL;

    INS_COPY(ins)=ret;

    while (refs)
  {
    /* Duplicate the reference */
    t_reloc * copy=RelocTableDupReloc(obj->reloc_table,refs->rel);

    VERIFY_FASSERT(RELOCATABLE_TYPE(REL_FROM(copy))==RT_INS,("Illegal reference with instruction @I: @R!",ins,copy));
    VERIFY_FASSERT(REL_TO_RELOCATABLE(copy),("@R has no to (code= %s)",copy,copy->code));

    RelocSetFrom(copy,T_RELOCATABLE(ret));
    refs=refs->next;
  }

  if(INS_OBJECT(ins)->actions->InsDupDynamic)
    INS_OBJECT(ins)->actions->InsDupDynamic(ret,ins);
  return ret;
    
#if 0
   t_ins * ret;
   t_section * sec;
   t_object * obj;
   t_reloc_ref * refs=INS_REFERS_TO(ins);
   
   if (INS_ATTRIB(ins) & IF_IN_BBL)
     {
       sec=ins->bbl->cfg->sec;
     }
   else
     {
       sec=ins->sec;
     }

   obj=sec->obj;
/*   
   if (sec->type!=FLOWGRAPHING_CODE_SECTION && sec->type!=FLOWGRAPHED_CODE_SECTION)
   {
     FATAL(("Cannot duplicate instructions in fase %c",sec->type));
   }
*/   
   ret=InsNewForSec(sec);
   memcpy(ret,ins,obj->description->decoded_instruction_size);
   INS_CADDRESS(ret)= NULL_GENERIC;
   INS_OLD_ADDRESS(ret) = NULL_GENERIC;
   T_RELOCATABLE(ret)->refedby = NULL;
   INS_PREV(ret)=INS_NEXT(ret)=NULL;
   INS_USED_REGSTATE(ret) = INS_DEFINED_REGSTATE(ret) = NULL;

   /* all node structures (cf ud-graph) must be set to NULL.
    * the copied instruction is _never_ part of the ud-graph
    * until it is inserted in the graph */
   NODE_NEXT(ret) = NODE_PREV(ret) = NULL;
   NODE_SUCC_FIRST(ret) = NODE_SUCC_LAST(ret) = NULL;
   NODE_PRED_FIRST(ret) = NODE_PRED_LAST(ret) = NULL;
   
   INS_REFERS_TO(ret)=NULL;
   INS_IS_REFED_BY(ret)=NULL;
   
   ins->copy=ret;
   
   while (refs)
     {
       /* Duplicate the reference */
       t_reloc_ref * ref=RelocRefNew(RelocTableDupReloc(obj->reloc_table,refs->rel));
       ref->next=INS_REFERS_TO(ret);
       
       INS_REFERS_TO(ret)=ref;
       
       if (RELOCTABLE_TYPE(REL_TO_RELOCATABLE(ref->rel))==RT_INS)
	 {
	   REL_FROM(ref->rel)=T_RELOCATABLE(ret);
	 }
       else
	 {
	   FATAL(("Illegal reference with instruction!"));
	 }
       
       if ((Rref->rel->to->to_type == TO_BBL) || (ref->rel->to->to_type == TO_INSTRUCTION))
       {
	 t_reloc_ref * ref2 = RelocRefNew(ref->rel);
	 ref2->next = INS_IS_REFED_BY(ref->rel->to->base);
	 INS_IS_REFED_BY(ref->rel->to->base) = ref2;
	 //DiabloPrint(stdout,"Adding refedby to bbl @B for reloc @R\n",ref->rel->to->base,ref->rel);
       }
       
       refs=refs->next;
     }


   /* Refby is not set and shouldn't be, because the original relocation keeps pointing at the original instruction! */
   
   return ret;
#endif
}




typedef enum _t_ia64_motion
{
  IA64_MOTION_UP, IA64_MOTION_DOWN
} t_ia64_motion;


typedef enum _t_ia64_reduction_type
{
  IA64_REDUCE_CYCLES, IA64_REDUCE_NOPS
} t_ia64_reduction_type;






static t_uint64 __maxBblProfileCount = 0;


void Ia64CalcMaxBblProfileCount(t_cfg *cfg)
{
  t_bbl *bbl;
  
  CFG_FOREACH_BBL(cfg, bbl)
  {
    __maxBblProfileCount = MAX(__maxBblProfileCount, BBL_PROFILE_INFO(bbl));
  }
}



t_bool Ia64InsIsPredicatable(t_ins *ins)
{
  if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_ALLOC || 
      IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_M25 || 
      IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B2 ||
      IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B6 ||
      IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B7 ||
      IA64OPCODEINFO(ins).instr_format == IA64_INSTR_FORMAT_B8)
    return FALSE;
  else
    return TRUE;
}



t_bool Ia64BblIsPredicatable(t_bbl *bbl, t_bool ignoreFinalBranch)
{
  t_ins *ins;

  BBL_FOREACH_INS(bbl, ins)
  {
    if(ignoreFinalBranch && ins == bbl->ins_last && IA64INSTYPE(ins) == IA64_TYPE_B)
      continue;

    if(!Ia64InsIsPredicatable(ins))
      return FALSE;
  }

  return TRUE;
}



t_ia64_sched_env *Ia64BuildDAGEnv(t_ia64_ins_list *insList)
{
  t_ia64_sched_env *env = NULL;

  if(insList->nins == 0)
    return NULL;


  env = Ia64SchedEnvAlloc();

  Ia64DAGNopListAlloc(env);

//#if 0  
  Ia64DAGListAlloc(env, insList, 0, insList->nins);


  Ia64DAGListBuild(env);
  Ia64DAGListPostBuild(env);
  Ia64DAGListCalcDeadlines(env);
  /*print dots of DAG - werkt fantastisch goed !*/
  //Ia64DAGListPrint(bbl);
//#endif
  return env;
}

void Ia64DestroyDAGEnv(t_ia64_sched_env *env)
{
  Ia64DAGListFree(env);
  Free(env->Ia64DAGList);
  Ia64SchedEnvFree(env);
}


#if 0
void Ia64CheckPredicatesAvailable(t_cfg *cfg)
{
  t_bbl *i_bbl;
  
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_ins *ins;

    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(IA64INSTYPE(ins) == IA64_TYPE_DATA)
	continue;
      
      if(IA64INSPRED(ins) >= IA64_P1 && IA64INSPRED(ins) <= IA64_P5)
      {
	InsPrint(ins, 0);
	FATAL(("Predicate %d is apparently in use", IA64INSPRED(ins) - IA64_PR));
      }
    }
  }

  
}
#endif

#define IA64_IS_FIXED_PR(reg) ((reg) >= IA64_P0 && (reg) <= IA64_P15)

t_ia64_reg Ia64GetFirstFreePredicateInFunction(t_function *fun)
{
  t_bbl *i_bbl;
  static t_ia64_reg freePreds[16];
  t_uint32 i = 0;

  for(i = 0; i < 16; i++)
    freePreds[i] = TRUE;
  
  
  FUN_FOREACH_BBL(fun,i_bbl)
  {
    t_ins *ins;

    BBL_FOREACH_INS(i_bbl, ins)
    {
      if(IA64INSTYPE(ins) == IA64_TYPE_DATA)
	continue;

      freePreds[IA64INSPRED(ins) - IA64_PR] = FALSE;
      
      if(IA64_IS_FIXED_PR(IA64INSOUT1(ins)))
	freePreds[IA64INSOUT1(ins) - IA64_PR] = FALSE;
      
      if(IA64_IS_FIXED_PR(IA64INSOUT2(ins)))
	freePreds[IA64INSOUT2(ins) - IA64_PR] = FALSE;
    }
  }  

  for(i = 6; i < 16; i++)
    if(freePreds[i])
      return IA64_PR + i;
  
  
  return IA64_NONE;
}

t_bool Ia64InsIsCompare(t_ins *ins)
{
  if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_CMP || IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_CMP4 ||
      IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_TBIT || IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_TNAT)
    return TRUE;
  else
    return FALSE;
}


t_ins *Ia64SearchForCompare(t_bbl *bbl)
{
  t_ins *ins;
  
  BBL_FOREACH_INS_R(bbl, ins)
  {
    if(Ia64InsIsCompare(ins))
      return ins;
  }

  return NULL;
}

t_bool Ia64IsBranchOrReturnOrCall(t_ins *ins)
{
  return Ia64IsBranch(ins) || Ia64IsReturn(ins) || Ia64IsCall(ins);
}

t_bool Ia64InstructionsAllHaveP0(t_bbl *bbl)
{
  t_ins *ins;

  BBL_FOREACH_INS(bbl, ins)
  {
    if(IA64INSPRED(ins) != IA64_P0)
      return FALSE;
  }

  return TRUE;
  
}

void Ia64AppendBblWithPredicate(t_bbl *destBbl, t_bbl *srcBbl, t_ia64_reg pred)
{
  t_ins *ins, *tmp;
  
  BBL_FOREACH_INS_SAFE(srcBbl, ins, tmp)
  {
    IA64INSPRED(ins) = pred;

    INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
    INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);

    INS_PREV(ins) = INS_NEXT(ins) = NULL;
    InsAppendToBbl(ins, destBbl);
    srcBbl->nins--;
    
  }

  srcBbl->ins_first = srcBbl->ins_last = NULL;
  
}


void Ia64CreateHyperblocks(t_section *code, t_cfg *cfg)
{
  t_function *i_fun;

  t_uint32 possibleCount = 0, directCount = 0;
  
  CFG_FOREACH_FUN(cfg, i_fun)
  {
    t_bbl *i_bbl, *tmp;
    t_ia64_reg freePred;
    t_bool restartFunction = FALSE;

    freePred = Ia64GetFirstFreePredicateInFunction(i_fun);

/*    
    if(freePred == IA64_NONE)
      continue;
*/
    do
    {
      restartFunction = FALSE;

      FUN_FOREACH_BBL_SAFE(i_fun, i_bbl, tmp)
      {
	static t_bbl *nextBbl[2];
	static t_cfg_edge *nextEdge[2];
	t_uint32 cnt = 0;
	t_cfg_edge *i_edge, *tmp_edge;
	t_ins *cmp;

	if(restartFunction)
	  break;

	if(i_bbl->ins_last == NULL)
	  continue;

	if(!Ia64IsBranch(i_bbl->ins_last) || Ia64IsCall(i_bbl->ins_last))
	  continue;

	if(IA64INSPRED(i_bbl->ins_last) == IA64_P0)
	  continue;

	BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
	{
	  if(cnt == 2)
	  {
	    cnt = 1000;
	    break;
	  }

	  nextEdge[cnt] = i_edge;
	  nextBbl[cnt] = CFG_EDGE_TAIL(i_edge);

	  cnt++;
	}

	if(cnt != 2)
	  continue;

	/*geen loopkes*/
	if(i_bbl == nextBbl[0] || i_bbl == nextBbl[1])
	  continue;

	/*geen lege opvolgerkes*/
	if(BBL_NINS(nextBbl[0]) == 0 || BBL_NINS(nextBbl[1]) == 0)
	  continue;

	/* nu checken of de opvolgers wel maar 1 inkomende pijl heeft*/

	/* evt wel speciaal geval voorzien: 
	*
	*        BBBBBBB
	*         /  |
	*        /   |
	*     BBBBBB |
	*       \    |
	*        \   |
	*        BBBBBB
	* 
	* */

	if(NODE_PRED_FIRST(nextBbl[0]) != NODE_PRED_LAST(nextBbl[0]))
	  continue;

	if(NODE_PRED_FIRST(nextBbl[1]) != NODE_PRED_LAST(nextBbl[1]))
	  continue;

	if(CfgEdgeIsInterproc(nextEdge[0]) || CfgEdgeIsInterproc(nextEdge[1]))
	  continue;
	

	if(!Ia64BblIsPredicatable(nextBbl[0], FALSE) || !Ia64BblIsPredicatable(nextBbl[1], FALSE))
	  continue;

	/*now search for the compare instruction...*/
	/*      
		cmp = INS_PREV(i_bbl->ins_last);

		if(IA64OPCODEINFO(cmp).instr_opcode != IA64_INSTR_CMP && IA64OPCODEINFO(cmp).instr_opcode != IA64_INSTR_CMP4)
		printf("WARNING: didn't find cmp[4] instruction\n");
	 */

	cmp = Ia64SearchForCompare(i_bbl);

	if(cmp == NULL)
	  continue;

	if(IA64INSPRED(i_bbl->ins_last) != IA64INSOUT1(cmp) && IA64INSPRED(i_bbl->ins_last) != IA64INSOUT2(cmp))
	  continue;

	if(freePred == IA64_NONE && (IA64INSOUT1(cmp) == IA64_P0 || IA64INSOUT2(cmp) == IA64_P0))
	  continue;


	if(Ia64IsBranchOrReturnOrCall(nextBbl[0]->ins_last) && Ia64IsBranchOrReturnOrCall(nextBbl[1]->ins_last))
	  continue;

	if(!Ia64IsBranchOrReturnOrCall(nextBbl[0]->ins_last) && !Ia64IsBranchOrReturnOrCall(nextBbl[1]->ins_last))
	  continue;


	if(EDGE_CAT(nextEdge[0]) == ET_FALLTHROUGH && Ia64IsBranchOrReturnOrCall(nextBbl[0]->ins_last))
	  continue;
	if(EDGE_CAT(nextEdge[1]) == ET_FALLTHROUGH && Ia64IsBranchOrReturnOrCall(nextBbl[1]->ins_last))
	  continue;

	/*maak index 0 de fallthrough*/
	if(EDGE_CAT(nextEdge[0]) != ET_FALLTHROUGH)
	{
	  t_cfg_edge *tmpEdge = nextEdge[0];
	  t_bbl *tmpBbl = nextBbl[0];

	  nextEdge[0] = nextEdge[1];
	  nextBbl[0] = nextBbl[1];

	  nextEdge[1] = tmpEdge;
	  nextBbl[1] = tmpBbl;

	}

	/*mag geen call zijn => voor de funlink*/
	if(Ia64IsCall(nextBbl[1]->ins_last))
	  continue;
	

	/* missing opportunity*/
	if(!Ia64InstructionsAllHaveP0(nextBbl[0]) || !Ia64InstructionsAllHaveP0(nextBbl[1]))
	  continue;

	if(BBL_NINS(nextBbl[0]) >= 9 && BBL_NINS(nextBbl[1]) >= 9)
	  continue;


	if(BBL_NINS(nextBbl[0]) <= 3 || BBL_NINS(nextBbl[1]) <= 3)
	{
	  t_ia64_reg jumpPred = IA64INSPRED(i_bbl->ins_last);
	  t_ia64_reg defaultPred;
	  
	  //printf("BblProfileCounts: %d %d \n", BBL_PROFILE_INFO(nextBbl[0]), BBL_PROFILE_INFO(nextBbl[1]));
	  
	  if(possibleCount >= global_options.debugcounter)
	    continue;
	  
	  if(IA64INSOUT1(cmp) == IA64_P0)
	  {
	    IA64INSOUT1(cmp) = freePred;
	    defaultPred = freePred;

	    INS_REGS_USE(cmp)=Ia64UsedRegisters(cmp);
	    INS_REGS_DEF(cmp)=Ia64DefinedRegisters(cmp);

	    freePred = Ia64GetFirstFreePredicateInFunction(i_fun);
	  }
	  else
	    if (IA64INSOUT2(cmp) == IA64_P0)
	    {
	      IA64INSOUT2(cmp) = freePred;
	      defaultPred = freePred;

	      INS_REGS_USE(cmp)=Ia64UsedRegisters(cmp);
	      INS_REGS_DEF(cmp)=Ia64DefinedRegisters(cmp);

	      freePred = Ia64GetFirstFreePredicateInFunction(i_fun);
	    }
	    else
	    {
	      if(IA64INSOUT1(cmp) == jumpPred)
	      	defaultPred = IA64INSOUT2(cmp);
	      else
		defaultPred = IA64INSOUT1(cmp);
	    }


	  /*kill the branch*/
	  InsKill(i_bbl->ins_last);

	  
//	  CfgEdgeKill(nextEdge[0]);
//	  CfgEdgeKill(nextEdge[1]);

	  /* shift these outgoing edges to the collective bbl*/	  
	  BBL_FOREACH_SUCC_EDGE_SAFE(nextBbl[0], i_edge, tmp_edge)
	  {
	    CfgEdgeNew(cfg, i_bbl, T_BBL(EDGE_TAIL(i_edge)), EDGE_CAT(i_edge));
	    CfgEdgeKill(i_edge);
	  }

	  BBL_FOREACH_PRED_EDGE_SAFE(nextBbl[0], i_edge, tmp_edge)
	    CfgEdgeKill(i_edge);
	  
	  
	  BBL_FOREACH_SUCC_EDGE_SAFE(nextBbl[1], i_edge, tmp_edge)
	  {
	    CfgEdgeNew(cfg, i_bbl, T_BBL(EDGE_TAIL(i_edge)), EDGE_CAT(i_edge));
	    CfgEdgeKill(i_edge);
	  }
	  
	  BBL_FOREACH_PRED_EDGE_SAFE(nextBbl[1], i_edge, tmp_edge)
	    CfgEdgeKill(i_edge);
	  
	  Ia64BblPrint(i_bbl);
	  printf(">******\n\n");

	  Ia64AppendBblWithPredicate(i_bbl, nextBbl[0], defaultPred);

	  Ia64BblPrint(i_bbl);
	  printf(">>******\n\n");

	  
	  Ia64AppendBblWithPredicate(i_bbl, nextBbl[1], jumpPred);

	  Ia64BblPrint(i_bbl);
	  printf(">>>******\n\n");


//	  BblKill(nextBbl[0]);
//	  BblKill(nextBbl[1]);
	  
	  /*rescheduling new BBL */
	  Ia64ScheduleBbl(code, i_bbl);

	  Ia64BblPrint(i_bbl);
	  printf("******\n\n");

          directCount++;
	  possibleCount++;

	  restartFunction = TRUE;
	  break;
	  
	}


	possibleCount++;
      }
      
    } while(restartFunction);

  }  
  printf("Statistics: %d %d\n", directCount, possibleCount);
  
}



#if 0
void Ia64MergingTwins(t_cfg *cfg)
{
  t_bbl *i_bbl, *tmp;

  t_uint32 mergeCount = 0;
  
  CFG_FOREACH_BBL_SAFE(cfg,i_bbl, tmp)
  {
    t_uint32 cnt = 0;
    t_cfg_edge *i_edge;
    t_bbl *tail;

    if(BBL_NINS(i_bbl) == 0) 
      continue;
    
    BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
    {
      tail = CFG_EDGE_TAIL(i_edge);
      if(BBL_NINS(tail) == 0) 
	cnt = 1000;
      cnt++;
    }

    if(cnt > 1)
      continue;

    cnt = 0;

    BBL_FOREACH_PRED_EDGE(tail,i_edge)
      cnt++;

    if(cnt > 1)
      continue;

    /*those bbls should be merged !*/

    mergeCount++;
/*
    Ia64BblPrint(i_bbl);
    printf("-----\n");
    Ia64BblPrint(tail);
    printf("\n\n");
*/

    if(Ia64IsBranch(i_bbl->ins_last) && Ia64IsBranch(tail->ins_last))
    {
      t_ins *ins;
      
      /*kill the edge*/
      CfgEdgeKill(i_edge);

      

      

      BblKill(tail);
    }
    else
    {
      
    }
    
  }  

  printf("%d BBLs could be merged\n", mergeCount);
}

#endif




t_ins *Ia64FirstInsInSet(t_ia64_set *s, t_ia64_ins_list *insListNext)
{
  t_uint32 i = 0;
  t_ia64_ins_list_item *item;

  item = insListNext->first;

  while (i < s->size && !s->set[i])
  {
    /*evt range check !!*/
    
    item = item->next;
    i++;
  }

  if(i == s->size)
    return NULL;
  
  return item->ins;
  
}


t_ins *Ia64GetInsInList(t_ia64_ins_list *insList, t_uint32 index)
{
  t_uint32 i = 0;
  t_ia64_ins_list_item *item;

  item = insList->first;

  while (i < index)
  {
    /*evt range check !!*/
    
    item = item->next;
    i++;
  }

  return item->ins;
  
}



t_bool Ia64HasTwin(t_bbl *i_bbl, t_bbl **nextBbl, t_cfg_edge **nextEdge)
{
  t_uint32 cnt = 0;
  t_cfg_edge *i_edge, *tmp_edge;
  t_bool hasBranch;

  if(i_bbl->ins_last == NULL)
    return FALSE;

  hasBranch = IA64INSTYPE(i_bbl->ins_last) == IA64_TYPE_B;


  if(hasBranch)
  {

    if(IA64INSPRED(i_bbl->ins_last) != IA64_P0)
      return FALSE;  

    if(!Ia64IsBranch(i_bbl->ins_last) || Ia64IsCall(i_bbl->ins_last))
      return FALSE;
  }
  
  /*Data: need not be scheduled*/
  if(IA64INSTYPE(i_bbl->ins_first) == IA64_TYPE_DATA)
    return FALSE;


  BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
  {
    if(cnt == 1)
    {
      cnt = 1000;
      break;
    }

    nextEdge[cnt] = i_edge;
    nextBbl[cnt] = CFG_EDGE_TAIL(i_edge);

    cnt++;
  }

  if(cnt != 1)
    return FALSE;

  /*geen loopkes*/
  if(i_bbl == nextBbl[0])
    return FALSE;

  /*geen lege opvolgerkes*/
  if(BBL_NINS(nextBbl[0]) == 0)
    return FALSE;

  /*Data: need not be scheduled*/
  if(IA64INSTYPE(nextBbl[0]) == IA64_TYPE_DATA)
    return FALSE;
  

  /* nu checken of de opvolgers wel maar 1 inkomende pijl heeft*/

  if(NODE_PRED_FIRST(nextBbl[0]) != NODE_PRED_LAST(nextBbl[0]))
    return FALSE;

  if(CfgEdgeIsInterproc(nextEdge[0]))
    return FALSE;

  return TRUE;

  
}

t_bool Ia64HasTriangle(t_bbl *i_bbl, t_bbl **nextBbl, t_cfg_edge **nextEdge)
{
  t_uint32 cnt = 0;
  t_cfg_edge *i_edge, *tmp_edge;

  if(i_bbl->ins_last == NULL)
    return FALSE;

  if(!Ia64IsBranch(i_bbl->ins_last) || Ia64IsCall(i_bbl->ins_last))
    return FALSE;

  if(IA64INSPRED(i_bbl->ins_last) == IA64_P0)
    return FALSE;


  /*Data: need not be scheduled*/
  if(IA64INSTYPE(i_bbl->ins_last) == IA64_TYPE_DATA)
    return FALSE;



  BBL_FOREACH_SUCC_EDGE(i_bbl,i_edge)
  {
    if(cnt == 2)
    {
      cnt = 1000;
      break;
    }

    nextEdge[cnt] = i_edge;
    nextBbl[cnt] = CFG_EDGE_TAIL(i_edge);

    cnt++;
  }

  if(cnt != 2)
    return FALSE;

  /*geen loopkes*/
  if(i_bbl == nextBbl[0] || i_bbl == nextBbl[1])
    return FALSE;

  /*geen lege opvolgerkes*/
  if(BBL_NINS(nextBbl[0]) == 0 || BBL_NINS(nextBbl[1]) == 0)
    return FALSE;


  /*Data: need not be scheduled*/
  if(IA64INSTYPE(nextBbl[0]) == IA64_TYPE_DATA)
    return FALSE;


  /*Data: need not be scheduled*/
  if(IA64INSTYPE(nextBbl[1]) == IA64_TYPE_DATA)
    return FALSE;



  /* nu checken of de opvolgers wel maar 1 inkomende pijl heeft*/

  if(NODE_PRED_FIRST(nextBbl[0]) != NODE_PRED_LAST(nextBbl[0]))
    return FALSE;

  if(NODE_PRED_FIRST(nextBbl[1]) != NODE_PRED_LAST(nextBbl[1]))
    return FALSE;

  if(CfgEdgeIsInterproc(nextEdge[0]) || CfgEdgeIsInterproc(nextEdge[1]))
    return FALSE;

  /*nu nog sorteren: sprongpad in [0], falltrough in [1] */
  if(EDGE_CAT(nextEdge[0]) == ET_FALLTHROUGH)
  {
    t_cfg_edge *tmpEdge = nextEdge[0];
    t_bbl *tmpBbl = nextBbl[0];

    nextEdge[0] = nextEdge[1];
    nextBbl[0] = nextBbl[1];

    nextEdge[1] = tmpEdge;
    nextBbl[1] = tmpBbl;

  }



  return TRUE;

}




t_uint32 Ia64CalcNopsInBbl(t_bbl *bbl)
{
  t_ins *ins;
  t_uint32 nops = 0;

  BBL_FOREACH_INS(bbl, ins)
  {
    if(IA64OPCODEINFO(ins).instr_opcode == IA64_INSTR_NOP)
      nops++;
  }

  return nops;
}




t_uint32 Ia64CalcNopsInScheduling(t_ia64_scheduling *scheduling)
{
  t_uint32 nops = 0;
  t_ia64_schedule_bundle *bundle;

  bundle = scheduling->first;

  while(bundle)
  {
    if(Ia64IsNop(bundle->slot[0]->ins))
      nops++;
    if(Ia64IsNop(bundle->slot[1]->ins))
      nops++;
    if(Ia64IsNop(bundle->slot[2]->ins))
      nops++;
    
    bundle = bundle->next;
  }
  
  return nops;
}


#if 0

t_bool Ia64ReduceNopsIfTwin(t_section *code, t_cfg *cfg, t_function *i_fun, t_bbl *i_bbl, t_uint32 *twinCount)
{
  static t_bbl *nextBbl;
  static t_cfg_edge *nextEdge;

  t_bbl *firstBbl = i_bbl;


  if(!Ia64HasTwin(firstBbl, &nextBbl, &nextEdge))
    return FALSE;


  if(Ia64CalcNopsInBbl(firstBbl) == 0 && Ia64CalcNopsInBbl(nextBbl) == 0)
    return FALSE;


  /*er valt nie veel te verbeteren...*/
  if(BBL_NINS(firstBbl) <= 3 || BBL_NINS(nextBbl) <= 3)
    return FALSE;


  (*twinCount)++;

  return TRUE;
}


t_bool Ia64ReduceNopsIfTriangle(t_section *code, t_cfg *cfg, t_function *i_fun, t_bbl *i_bbl, t_uint32 *triangleCount)
{
  static t_bbl *nextBbl[2];
  static t_cfg_edge *nextEdge[2];

  t_bbl *firstBbl = i_bbl;


  if(!Ia64HasTriangle(firstBbl, nextBbl, nextEdge))
    return FALSE;

  if(Ia64CalcNopsInBbl(firstBbl) == 0 && Ia64CalcNopsInBbl(nextBbl[0]) == 0 && Ia64CalcNopsInBbl(nextBbl[1]) == 0)
    return FALSE;

  /*er valt nie veel te verbeteren...*/
  if(BBL_NINS(firstBbl) <= 3 || BBL_NINS(nextBbl[0]) <= 3 || BBL_NINS(nextBbl[1]) <= 3)
    return FALSE;



  (*triangleCount)++;


  return TRUE;
}







void Ia64ReduceNopsByExchangingInstructions(t_section *code, t_cfg *cfg)
{
  t_function *i_fun;

  t_uint32 twinCount = 0, triangleCount = 0;
  
  CFG_FOREACH_FUN(cfg, i_fun)
  {
    t_bbl *i_bbl, *tmp;
    t_ia64_reg freePred;

    freePred = Ia64GetFirstFreePredicateInFunction(i_fun);    

    FUN_FOREACH_BBL(i_fun, i_bbl)
    {
      if(Ia64ReduceNopsIfTwin(code, cfg, i_fun, i_bbl, &twinCount))
	continue;

      if(Ia64ReduceNopsIfTriangle(code, cfg, i_fun, i_bbl, &triangleCount))
	continue;
      
    }
    
  }
  
  printf("Statistics: twin: %d; triangle: %d\n", twinCount, triangleCount);
  
}



#endif















t_uint32 Ia64CalcStopsInBbl(t_bbl *bbl)
{
  t_ins *ins, *next;
  t_uint32 stops = 1;

  BBL_FOREACH_INS_SAFE(bbl, ins, next)
  {
    if(next && IA64INSINSTRGROUP(ins) != IA64INSINSTRGROUP(next) )
      stops++;
  }

  return stops;
}


t_uint32 Ia64CalcStopsInScheduling(t_ia64_scheduling *scheduling)
{
  t_uint32 stops = 0;
  t_ia64_schedule_bundle *bundle;

  bundle = scheduling->first;

  while(bundle)
  {
    
    if(bundle == scheduling->last || (bundle->template & 1) != 0)
      stops++;

    switch(bundle->template)
    {
      case IA64_MI_I:
      case IA64_MI_I_:
      case IA64_M_MI:
      case IA64_M_MI_:
	stops++;
	break;
      default:
	break;
    }
    
    bundle = bundle->next;
  }
  
  if(stops == 0)
    return 1;
  else
    return stops;
}


typedef struct _t_ia64_pred_cmp
{
//  t_ins *cmp;
  
  t_ia64_reg predTrue;
  t_ia64_reg predFalse;
  
} t_ia64_pred_cmp;



typedef struct _t_ia64_motion_stats
{
  t_uint32 motionCount;

  t_uint32 motionStaticUpCount;
  t_uint32 motionStaticDownCount;
  
  t_uint64 motionDynamicUpCount;
  t_uint64 motionDynamicDownCount;
  
} t_ia64_motion_stats;


t_ia64_motion_stats *Ia64MotionStatsAlloc()
{
  return (t_ia64_motion_stats *)Calloc(1, sizeof(t_ia64_motion_stats));
}

void Ia64MotionStatsFree(t_ia64_motion_stats *stats)
{
  Free(stats);
}


typedef struct _t_ia64_region_bblcounts
{
  t_uint64 bblCountFirst;
  t_uint64 bblCountNext[2];
  
} t_ia64_region_bblcounts;


t_uint64 Ia64CalcResourceCost(t_ia64_reduction_type reductionType, t_ia64_region_bblcounts *bblCounts, 
    t_uint64 nResourceFirst, t_uint64 nResourceNext1, t_uint64 nResourceNext2)
{
  if(reductionType == IA64_REDUCE_CYCLES)
  {
    return bblCounts->bblCountFirst * nResourceFirst + 
      bblCounts->bblCountNext[0] * nResourceNext1 + 
      bblCounts->bblCountNext[1] * nResourceNext2;
  }
  else
    return nResourceFirst + nResourceNext1 + nResourceNext2;
}


typedef struct _t_ia64_region_data
{
  t_ia64_sched_env *envFirst;
  t_ia64_sched_env *envNext[2];
    
  t_ia64_ins_list *insListFirst;
  t_ia64_ins_list *insListNext[2];

  t_ia64_scheduling *schedulingFirst;
  t_ia64_scheduling *schedulingNext[2];

  t_uint32 nResourceFirst;
  t_uint32 nResourceNext[2];

  t_ia64_set *uFirst;
  t_ia64_set *uNext[2];

  t_uint64 cost;

  t_uint8 nNext;
  
} t_ia64_region_data;


t_ia64_region_data *Ia64RegionDataAlloc()
{
  return (t_ia64_region_data *)Calloc(1, sizeof(t_ia64_region_data));
}


void Ia64RegionDataFree(t_ia64_region_data *data)
{
  t_uint8 i = 0;

  if(data->schedulingFirst)
    Ia64SchedulingFree(data->schedulingFirst);

  if(data->uFirst)
    Ia64SetFree(data->uFirst);

  Ia64DAGListFree(data->envFirst);
  Ia64DAGNopListFree(data->envFirst);
  Ia64SchedEnvFree(data->envFirst);
  Ia64InsListFree(data->insListFirst);

  for(i = 0; i < data->nNext; i++)
  {
    if(data->schedulingNext[i])
      Ia64SchedulingFree(data->schedulingNext[i]);
  
    if(data->uNext[i])
      Ia64SetFree(data->uNext[i]);

    Ia64DAGListFree(data->envNext[i]);
    Ia64DAGNopListFree(data->envNext[i]);
    Ia64SchedEnvFree(data->envNext[i]);
    Ia64InsListFree(data->insListNext[i]);
  }

  Free(data);
}



t_ia64_region_data *Ia64RegionDataAllocAndSchedule(t_ia64_reduction_type reductionType, 
    t_ia64_ins_list *insListFirst, t_ia64_ins_list *insListNext1, t_ia64_ins_list *insListNext2)
{
  t_ia64_region_data *tmpData;
  t_uint8 i = 0;
  
  tmpData = Ia64RegionDataAlloc();

  tmpData->nNext = insListNext2 == NULL? 1 : 2;
  
  tmpData->insListFirst = Ia64InsListCopy(insListFirst);
  tmpData->insListNext[0] = Ia64InsListCopy(insListNext1);
  if(insListNext2)
    tmpData->insListNext[1] = Ia64InsListCopy(insListNext2);

  tmpData->envFirst = Ia64SchedEnvAlloc();
  Ia64DAGNopListAlloc(tmpData->envFirst);
  tmpData->schedulingFirst = Ia64PerformScheduling(tmpData->envFirst, tmpData->insListFirst, NULL);

  if(reductionType == IA64_REDUCE_CYCLES)
    tmpData->nResourceFirst = Ia64CalcStopsInScheduling(tmpData->schedulingFirst);
  else
    tmpData->nResourceFirst = Ia64CalcNopsInScheduling(tmpData->schedulingFirst);

  
  for(i = 0; i < tmpData->nNext; i++)
  {
    tmpData->envNext[i] = Ia64SchedEnvAlloc();
    Ia64DAGNopListAlloc(tmpData->envNext[i]);
    tmpData->schedulingNext[i] = Ia64PerformScheduling(tmpData->envNext[i], tmpData->insListNext[i], NULL);

    if(reductionType == IA64_REDUCE_CYCLES)
      tmpData->nResourceNext[i] = Ia64CalcStopsInScheduling(tmpData->schedulingNext[i]);
    else
      tmpData->nResourceNext[i] = Ia64CalcNopsInScheduling(tmpData->schedulingNext[i]);
  }

  return tmpData;
}

/*TODO: dit uitbreiden ?*/
t_bool Ia64ReductionCanImprove(t_ia64_reduction_type reductionType, t_ia64_motion direction, 
    t_uint8 nNext, t_uint32 nResourceFirst, t_uint32 nResourceNext1, t_uint32 nResourceNext2, 
    t_uint32 ninsFirst, t_uint32 ninsNext1, t_uint32 ninsNext2
/*    t_ia64_region_data *regionData*/
    )
{
  if(reductionType == IA64_REDUCE_CYCLES)
  {
    if(direction == IA64_MOTION_UP)
    {
      if(nResourceNext1 == 1)
	return FALSE;

      if(nNext == 2 && nResourceNext2 == 1)
	return FALSE;
    }
    else
    {
      if(nResourceFirst == 1)
	return FALSE;
    }
  }
  else
  {
    if(nResourceFirst == 0 && nResourceNext1 == 0 && nResourceNext2 == 0) 
      return FALSE;

    if(nNext == 1)
    {
      if(nResourceFirst + nResourceNext1 + nResourceNext2 < 3)
	return FALSE;

      if(direction == IA64_MOTION_UP)
      {
	if(nNext == 1)
	{
	  if(ninsNext1 <= 3)
	    return FALSE;
	}
	else
	{
	  if(ninsNext1 <= 3 && ninsNext2 <= 3)
	    return FALSE;
	}
      }
      else
      {
	if(ninsFirst <= 3)
	  return FALSE;
      }
    }
  }

  return TRUE;  
}


#define IA64CODEMOTIONMAXDEPTH_UP 4
#define IA64CODEMOTIONMAXDEPTH_DOWN 2


/*evt verfijnen: als laatste instructie geen verloop wijzigt: evt toch doen => evt geen enkele instructie meer over in BBL !!!*/
/*dan wel alle return-condities checken: vb ook: if(regionData->insListNext->nins <= 3)*/

/*retourneert de best gevonden configuratie*/
t_ia64_region_data *Ia64ReduceWithTwinsRec(t_ia64_reduction_type reductionType, t_ia64_motion direction, /*t_ia64_set *u, */
    t_ia64_region_data *regionData, t_uint32 depth, t_uint64 bestSolutionCost, t_ia64_region_bblcounts *bblCounts, t_ia64_pred_cmp *predCmp)
{
  t_ia64_region_data *bestData = NULL, *newData, *nextData;

  t_ia64_set *u;
  t_ia64_set *r, *s;

  t_uint8 stage = 0;
  t_ins *dupIns = NULL;

  t_uint8 choiceNext = 0;
  t_bool predChanged = FALSE;

  if(direction == IA64_MOTION_UP && depth >= IA64CODEMOTIONMAXDEPTH_UP)
    return NULL;

  
  if(direction == IA64_MOTION_DOWN && depth >= IA64CODEMOTIONMAXDEPTH_DOWN)
    return NULL;
/*
  if(__static_scheduler_code_section__ == NULL)
  {
    *(int *)0 = 0;
  }
*/  
  
  /*je kan dit al niet verminderen !!!*/
  if(!Ia64ReductionCanImprove(reductionType, direction, regionData->nNext, regionData->nResourceFirst, regionData->nResourceNext[0],
	regionData->nResourceNext[1], regionData->insListFirst->nins, regionData->insListNext[0]->nins, 
	regionData->nNext == 2?regionData->insListNext[1]->nins:0))
    return NULL;

  if(direction == IA64_MOTION_UP && regionData->nNext == 2)
    choiceNext = depth % 2;
  
  
  if(direction == IA64_MOTION_UP)
    u = regionData->uNext[choiceNext];
  else
    u = regionData->uFirst;

  if(Ia64IsSetEmpty(u))
    return NULL;

  if(direction == IA64_MOTION_UP)
    r = Ia64AlgoReady(regionData->envNext[choiceNext], u, 0);
  else
    r = Ia64AlgoReverseReadySkipBranchAndShift(regionData->envFirst, u, 0);
  
  /*er zijn er geen ready ...*/
  if(direction == IA64_MOTION_UP)
  {
    if(Ia64IsSetEmpty(r) || Ia64SetCount(u) <= 1) /*evt verfijnen: zie commentaar hierboven */
      return NULL;
  }
  else
  {
    if(Ia64IsSetEmpty(r) || Ia64SetCount(u) <= 1)
      return NULL;
  }

  if(direction == IA64_MOTION_UP)
    s = Ia64AlgoNotYetCritical(regionData->envNext[choiceNext], r, depth);  
  else
    s = Ia64SetCopy(r);//Ia64AlgoReverseNotYetCritical(regionData->envFirst, r, depth);

  for(stage = 0; stage < 2; stage++)
  {
    t_int32 index;
    t_ia64_set *newU, *m;
    t_ins *ins = NULL;

    predChanged = FALSE;

    m = Ia64SetDiff(r, s);

    /*eerst kijken of er geen kritische zijn...*/
    if (direction == IA64_MOTION_UP)
    {
      index = Ia64SetFirstInSet(m);

      if(index == m->size)
	index = Ia64SetFirstInSet(r);

      if(index == u->size)
        FATAL(("something wrong"));
    }
    else
    {
      index = Ia64SetLastInSet(m);

      if(index == -1)
	index = Ia64SetLastInSet(r);

      if(index == -1)
        FATAL(("something wrong"));
    }

    Ia64SetFree(m);

    nextData = NULL;

    newData = Ia64RegionDataAlloc();
    newData->insListFirst = Ia64InsListCopy(regionData->insListFirst);
    newData->insListNext[0] = Ia64InsListCopy(regionData->insListNext[0]);
    if(regionData->nNext == 2)
      newData->insListNext[1] = Ia64InsListCopy(regionData->insListNext[1]);
    
    newData->nNext = regionData->nNext;
    
//#if 0    
    /*eerst effen kijken of de instructie wel prediceerbaar is*/
    if(stage == 0 && direction == IA64_MOTION_UP && regionData->nNext == 2)
    {
      t_ins *tmpIns = Ia64GetInsInList(regionData->insListNext[choiceNext], index);

      if(!Ia64InsIsPredicatable(tmpIns) || IA64INSPRED(tmpIns) != IA64_P0 || Ia64InsIsCompare(tmpIns))
        stage = 1; /*we nemen hem niet...*/
      else
      {
	predChanged = TRUE;

	if(choiceNext == 0)
	  IA64INSPRED(tmpIns) = predCmp->predTrue;
	else
	  IA64INSPRED(tmpIns) = predCmp->predFalse;

        INS_REGS_USE(tmpIns)=Ia64UsedRegisters(tmpIns);
	INS_REGS_DEF(tmpIns)=Ia64DefinedRegisters(tmpIns);	
      }
    }
//#endif

    /*nu twee mogelijkheden: ofwel neem je hem, ofwel neem je hem niet*/
    if(stage == 0)
    { /* we nemen hem */

      t_ia64_region_data *tmpData;

      newU = Ia64SetDelete(u, index);

      if(direction == IA64_MOTION_UP)
      {
        ins = Ia64InsListDelete(newData->insListNext[choiceNext], index);
	Ia64InsListAppendInsBeforeBranch(newData->insListFirst, ins);
      }
      else
      {
	ins = Ia64InsListDelete(newData->insListFirst, index);
	Ia64InsListPrependIns(newData->insListNext[0], ins);
	/*evt dupliceren en ook aan de andere toevoegen !*/
	if(regionData->nNext == 2)
	{
	  //InsPrint(ins, 0);
	  dupIns = Ia64InsDup(ins);
	  Ia64InsListPrependIns(newData->insListNext[1], dupIns);
	}
      }

      
      newData->envFirst = Ia64BuildDAGEnv(newData->insListFirst);
      newData->envNext[0] = Ia64BuildDAGEnv(newData->insListNext[0]);
      newData->envNext[1] = regionData->nNext == 2? Ia64BuildDAGEnv(newData->insListNext[1]): NULL;
      
      tmpData = Ia64RegionDataAllocAndSchedule(reductionType, newData->insListFirst, newData->insListNext[0], newData->insListNext[1]);

      /*nu dus kijken of dit niet tot een betere oplossing leidt*/
      {
	/*TODO: functie van maken: bij cycles en triangle immers rekening houden met bbl_count*/
	t_uint64 total_new = Ia64CalcResourceCost(reductionType, bblCounts, 
	    tmpData->nResourceFirst, tmpData->nResourceNext[0], tmpData->nResourceNext[1]);
	  //tmpData->nResourceFirst + tmpData->nResourceNext[0] + tmpData->nResourceNext[1];

	if(total_new < bestSolutionCost)
	{ /*YES!*/
	  printf("Total %lld -> %lld\n", bestSolutionCost, total_new);
	  printf("cost: %d %d %d\n", tmpData->nResourceFirst, tmpData->nResourceNext[0], tmpData->nResourceNext[1]);
	  fflush(stdout);
	  
	  bestData = newData;

	  bestData->cost = total_new;
	}
      }

      newData->nResourceFirst = tmpData->nResourceFirst;
      newData->nResourceNext[0] = tmpData->nResourceNext[0];
      newData->nResourceNext[1] = tmpData->nResourceNext[1];
      
      Ia64RegionDataFree(tmpData);

    }
    else
    {
      newU = Ia64SetCopy(u);
      newU->set[index] = 0;

      newData->envFirst = Ia64BuildDAGEnv(newData->insListFirst);
      newData->envNext[0] = Ia64BuildDAGEnv(newData->insListNext[0]);
      if(newData->nNext == 2)
	newData->envNext[1] = Ia64BuildDAGEnv(newData->insListNext[1]);

      newData->nResourceFirst = regionData->nResourceFirst;
      newData->nResourceNext[0] = regionData->nResourceNext[0];
      newData->nResourceNext[1] = regionData->nResourceNext[1];
    }
    

    if(direction == IA64_MOTION_UP)
    {
      newData->uNext[choiceNext] = newU;
      if(regionData->nNext == 2)
        newData->uNext[1 - choiceNext] = Ia64SetCopy(regionData->uNext[1 - choiceNext]);
    }
    else
      newData->uFirst = newU;

    
    
    /*nu dus kijken of dit niet tot een betere oplossing leidt*/
    
    nextData = Ia64ReduceWithTwinsRec(reductionType, direction, /*newU,*/ newData, depth + 1, 
	bestData?bestData->cost:bestSolutionCost, bblCounts, predCmp);
    
    if(nextData != NULL)
    {
      t_uint64 total_new = Ia64CalcResourceCost(reductionType, bblCounts, 
	    nextData->nResourceFirst, nextData->nResourceNext[0], nextData->nResourceNext[1]);
	//nextData->nResourceFirst + nextData->nResourceNext[0] + nextData->nResourceNext[1];
      
      if(total_new < bestSolutionCost)
      { /*YES!*/
        printf("Total %lld -> %lld\n", bestSolutionCost, total_new);
	printf("cost: %d %d %d\n", nextData->nResourceFirst, nextData->nResourceNext[0], nextData->nResourceNext[1]);
        fflush(stdout);
	
	
	bestData = nextData;
	bestData->cost = total_new;
      }
      
    }

    if(stage == 0 && (bestData == NULL || (bestData != nextData && bestData != newData)))
      if(predChanged)
      {
	IA64INSPRED(ins) = IA64_P0;
	
        INS_REGS_USE(ins)=Ia64UsedRegisters(ins);
	INS_REGS_DEF(ins)=Ia64DefinedRegisters(ins);	
      }
      

    /*geen opl of opl van deze stage => oplossing van vorige stage is ongeldig => evt dupIns en */
    if(stage == 1 && (bestData == NULL || bestData == nextData || bestData == newData))
      if(dupIns)
      {
	  InsFree(dupIns);
	  dupIns = NULL;
      }

    if(bestData != newData)
      Ia64RegionDataFree(newData);
    
    if(nextData && bestData != nextData)
      Ia64RegionDataFree(nextData);

    /*Ia64SetFree(newU);*/
  }

  Ia64SetFree(r);
  Ia64SetFree(s);
  
  return bestData;
}


void Ia64ReduceWithTwins(t_ia64_reduction_type reductionType, t_ia64_motion direction, 
    t_bbl *firstBbl, t_bbl *nextBbl1, t_bbl *nextBbl2, t_ia64_motion_stats *stats, t_ia64_pred_cmp *predCmp)
{
  t_int32 i;
  t_ia64_region_data *input;
  t_ia64_region_data *result;
  t_ia64_region_data *tmpData;
  t_ia64_set *u;
  t_ia64_ins_list *insListNext[2];
  t_ia64_ins_list *insListFirst;
/*  t_uint32 nopsFirst = 0, nopsNext = 0;*/
  t_uint32 nResourceFirst = 0, nResourceNext1 = 0, nResourceNext2 = 0;
  
  t_uint64 total_orig = 0;
  
  t_ia64_region_bblcounts bblCounts;

  /*je kan dit al niet verminderen !!!*/
  if(reductionType == IA64_REDUCE_CYCLES)
  {
    nResourceFirst = Ia64CalcStopsInBbl(firstBbl);
    nResourceNext1 = Ia64CalcStopsInBbl(nextBbl1);
    nResourceNext2 = nextBbl2?Ia64CalcStopsInBbl(nextBbl2):0;
  }
  else
  {
    nResourceFirst = Ia64CalcNopsInBbl(firstBbl);
    nResourceNext1 = Ia64CalcNopsInBbl(nextBbl1);
    nResourceNext2 = nextBbl2?Ia64CalcNopsInBbl(nextBbl2):0;
  }

  if(!Ia64ReductionCanImprove(reductionType, direction, nextBbl2?2:1, nResourceFirst, nResourceNext1, nResourceNext2, 
      BBL_NINS(firstBbl), BBL_NINS(nextBbl1), nextBbl2?BBL_NINS(nextBbl2):0 ))
    return;


  bblCounts.bblCountFirst = BBL_PROFILE_INFO(firstBbl);
  bblCounts.bblCountNext[0] = BBL_PROFILE_INFO(nextBbl1);
  bblCounts.bblCountNext[1] = nextBbl2?BBL_PROFILE_INFO(nextBbl2):0;
  
  insListFirst = Ia64InsListAlloc(firstBbl);
  insListNext[0] = Ia64InsListAlloc(nextBbl1);
  insListNext[1] = nextBbl2?Ia64InsListAlloc(nextBbl2):NULL;


  input = Ia64RegionDataAlloc();
  input->insListFirst = insListFirst;
  input->insListNext[0] = insListNext[0];
  input->insListNext[1] = insListNext[1];
  input->envFirst = Ia64BuildDAGEnv(insListFirst);
  input->envNext[0] = Ia64BuildDAGEnv(insListNext[0]);
  input->envNext[1] = nextBbl2?Ia64BuildDAGEnv(insListNext[1]):NULL;
  input->nNext = nextBbl2?2:1;

  if(reductionType == IA64_REDUCE_CYCLES)
  {
    input->nResourceFirst = Ia64CalcStopsInBbl(firstBbl);
    input->nResourceNext[0] = Ia64CalcStopsInBbl(nextBbl1);
    input->nResourceNext[1] = nextBbl2?Ia64CalcStopsInBbl(nextBbl2):0;
  }
  else
  {
    input->nResourceFirst = Ia64CalcNopsInBbl(firstBbl);
    input->nResourceNext[0] = Ia64CalcNopsInBbl(nextBbl1);
    input->nResourceNext[1] = nextBbl2?Ia64CalcNopsInBbl(nextBbl2):0;
  }

  input->cost = Ia64CalcResourceCost(reductionType, &bblCounts, 
	    input->nResourceFirst, input->nResourceNext[0], input->nResourceNext[1]);

  if(direction == IA64_MOTION_UP)
  {
    if(nextBbl2)
    {
      u = Ia64SetAlloc(insListNext[1]->nins);
      input->uNext[1] = u;

      for(i = 0; i < u->size; i++)
	u->set[i] = TRUE;
    }
    
    u = Ia64SetAlloc(insListNext[0]->nins);
    input->uNext[0] = u;
    
    for(i = 0; i < u->size; i++)
      u->set[i] = TRUE;

  }
  else
  {
    u = Ia64SetAlloc(insListFirst->nins);
    input->uFirst = u;

    for(i = 0; i < u->size; i++)
      u->set[i] = TRUE;
  }
  
  
  if(reductionType == IA64_REDUCE_CYCLES)
    printf("total dynamical input stops: %lld\n", input->cost);
  else
    printf("total input NOPs: %lld\n", input->cost);
  
  result = Ia64ReduceWithTwinsRec(reductionType, direction, /*u,*/ input, 0, input->cost, &bblCounts, predCmp);
  
  if(result != NULL)
  {
    t_uint32 total_orig = input->nResourceFirst + input->nResourceNext[0] + input->nResourceNext[1];
    t_uint32 total_new = result->nResourceFirst + result->nResourceNext[0] + result->nResourceNext[1];

    result->cost = Ia64CalcResourceCost(reductionType, &bblCounts, 
	    result->nResourceFirst, result->nResourceNext[0], result->nResourceNext[1]);


    if(direction == IA64_MOTION_UP)
    {
      stats->motionStaticUpCount += total_orig - total_new;
      
      //if(reductionType == IA64_REDUCE_CYCLES)
        stats->motionDynamicUpCount += input->cost - result->cost;//BBL_PROFILE_INFO(firstBbl)*(total_orig - total_new);
    }
    else
    {
      stats->motionStaticDownCount += total_orig - total_new;

      //if(reductionType == IA64_REDUCE_CYCLES)
        stats->motionDynamicDownCount += input->cost - result->cost;//BBL_PROFILE_INFO(firstBbl)*(total_orig - total_new);
    }

    if(reductionType == IA64_REDUCE_CYCLES)    
      printf("YES: found a better solution for a region: %lld cycles in stead of %lld cycles!\n", 
	  result->cost, input->cost);
    else
      printf("YES: found a better solution for a region: %lld NOPs in stead of %lld NOPs!\n", 
	  result->cost, input->cost);
    
    printf("Previous configuration:\n");
    Ia64BblPrint(firstBbl);
    Ia64BblPrint(nextBbl1);
    if(nextBbl2)
      Ia64BblPrint(nextBbl2);
    printf("\n");

    tmpData = Ia64RegionDataAllocAndSchedule(reductionType, result->insListFirst, result->insListNext[0], result->insListNext[1]);


    /*apply to bbl e.d.*/
    Ia64ApplySchedulingToBbl(tmpData->envFirst, tmpData->schedulingFirst, firstBbl);
    Ia64ApplySchedulingToBbl(tmpData->envNext[0], tmpData->schedulingNext[0], nextBbl1);
    if(nextBbl2)
      Ia64ApplySchedulingToBbl(tmpData->envNext[1], tmpData->schedulingNext[1], nextBbl2);

    printf("New configuration:\n");
    Ia64BblPrint(firstBbl);
    Ia64BblPrint(nextBbl1);
    if(nextBbl2)
      Ia64BblPrint(nextBbl2);
    printf("\n");
    fflush(stdout);

    
    Ia64RegionDataFree(tmpData);
    Ia64RegionDataFree(result);
  }

  Ia64RegionDataFree(input);
  
}



t_bool Ia64ReduceCyclesIfTwin(t_section *code, t_cfg *cfg, t_function *i_fun, t_bbl *i_bbl, t_ia64_motion_stats *stats)
{
  static t_bbl *nextBbl;
  static t_cfg_edge *nextEdge;

  t_bbl *firstBbl = i_bbl;


  if(!Ia64HasTwin(firstBbl, &nextBbl, &nextEdge))
    return FALSE;

  if(BBL_PROFILE_INFO(firstBbl) == 0 && BBL_PROFILE_INFO(nextBbl) == 0)
    return FALSE;

  if(Ia64CalcStopsInBbl(firstBbl) == 1 && Ia64CalcStopsInBbl(nextBbl) == 1)
    return FALSE;

  printf("First in twin: <bblcount = %lld>\n", BBL_PROFILE_INFO(firstBbl));
  Ia64BblPrint(firstBbl);
  printf("Next in twin: <bblcount = %lld>\n", BBL_PROFILE_INFO(nextBbl));
  Ia64BblPrint(nextBbl);
  printf("\n");

  stats->motionCount++;

  printf("Trying upward code motion...\n");
  Ia64ReduceWithTwins(IA64_REDUCE_CYCLES, IA64_MOTION_UP, firstBbl, nextBbl, NULL, stats, NULL);

  
  printf("Trying downward code motion...\n");
  Ia64ReduceWithTwins(IA64_REDUCE_CYCLES, IA64_MOTION_DOWN, firstBbl, nextBbl, NULL, stats, NULL);
  

  return TRUE;
}


t_bool Ia64ReduceCyclesIfTriangle(t_section *code, t_cfg *cfg, t_function *i_fun, t_bbl *i_bbl, t_ia64_motion_stats *stats)
{
  static t_bbl *nextBbl[2];
  static t_cfg_edge *nextEdge[2];

  t_bbl *firstBbl = i_bbl;


  if(!Ia64HasTriangle(firstBbl, nextBbl, nextEdge))
    return FALSE;

  

  if(BBL_PROFILE_INFO(firstBbl) == 0 && BBL_PROFILE_INFO(nextBbl[0]) == 0 && BBL_PROFILE_INFO(nextBbl[1]) == 0)
    return FALSE;

  if(Ia64CalcStopsInBbl(firstBbl) == 1 && Ia64CalcStopsInBbl(nextBbl[0]) == 1 && Ia64CalcStopsInBbl(nextBbl[1]) == 1)
    return FALSE;
  

  printf("First in triangle: <bblcount = %lld>\n", BBL_PROFILE_INFO(firstBbl));
  Ia64BblPrint(firstBbl);
  printf("Next-TRUE in triangle: <bblcount = %lld>\n", BBL_PROFILE_INFO(nextBbl[0]));
  Ia64BblPrint(nextBbl[0]);
  printf("Next-FALSE in triangle: <bblcount = %lld>\n", BBL_PROFILE_INFO(nextBbl[1]));
  Ia64BblPrint(nextBbl[1]);
  printf("\n");

  stats->motionCount++;

//#if 0  
  printf("Trying upward code motion...\n");
  
  {
    t_ia64_pred_cmp predCmp;
    t_ins *cmp;
    
    cmp = Ia64SearchForCompare(firstBbl);

    if(cmp && (IA64INSPRED(firstBbl->ins_last) != IA64INSOUT1(cmp) && IA64INSPRED(firstBbl->ins_last) != IA64INSOUT2(cmp)))
      cmp = NULL;

    /*TODO: dit wegwerken !!!*/
    /*merk op: ook aanpassen bij nops !!!*/
    if(cmp && (IA64INSOUT1(cmp) == IA64_P0 || IA64INSOUT1(cmp) == IA64_P0))
      cmp = NULL;
    
    if(cmp)
    {
      predCmp.predTrue = IA64INSOUT1(cmp);
      predCmp.predFalse = IA64INSOUT2(cmp);

      if(predCmp.predTrue != IA64INSPRED(firstBbl->ins_last))
      {
	/*just omgekeerd dan...*/
	predCmp.predTrue = IA64INSOUT2(cmp);
	predCmp.predFalse = IA64INSOUT1(cmp);
      }
      
      /*cmp zoeken, predicaat wijzigen(als niet succesvol => evt resetten)...*/
      Ia64ReduceWithTwins(IA64_REDUCE_CYCLES, IA64_MOTION_UP, firstBbl, nextBbl[0], nextBbl[1], stats, &predCmp);
    }
  }
//#endif
  
  printf("Trying downward code motion...\n");
  Ia64ReduceWithTwins(IA64_REDUCE_CYCLES, IA64_MOTION_DOWN, firstBbl, nextBbl[0], nextBbl[1], stats, NULL);
  
  return TRUE;
}


void Ia64ReduceCyclesByExchangingInstructions(t_section *code, t_cfg *cfg)
{
  t_function *i_fun;

  t_ia64_motion_stats *twinStats = Ia64MotionStatsAlloc();
  t_ia64_motion_stats *triangleStats = Ia64MotionStatsAlloc();
  
  CFG_FOREACH_FUN(cfg, i_fun)
  {
    t_bbl *i_bbl, *tmp;
    t_ia64_reg freePred;

    freePred = Ia64GetFirstFreePredicateInFunction(i_fun);    

    FUN_FOREACH_BBL(i_fun, i_bbl)
    {
      if(Ia64ReduceCyclesIfTwin(code, cfg, i_fun, i_bbl, twinStats))
	continue;

      if(Ia64ReduceCyclesIfTriangle(code, cfg, i_fun, i_bbl, triangleStats))
	continue;
    
      /*nog een funnel-geval voorzien (trechter)*/  
    }
    
  }


  printf("\n");
  printf("--    Code Motion Statistics    --\n");
  printf("---      Reducing Cycles       ---\n");
  printf("Twins:                    %8d\n", twinStats->motionCount);
  printf("  Static by UP motion:    %8d\n", twinStats->motionStaticUpCount);
  printf("  Static by DOWN motion:  %8d\n", twinStats->motionStaticDownCount);
  printf("  Dynamic by UP motion:   %8lld\n", twinStats->motionDynamicUpCount);
  printf("  Dynamic by DOWN motion: %8lld\n", twinStats->motionDynamicDownCount);
  printf("Triangles:                %8d\n", triangleStats->motionCount);
  printf("  Static by UP motion:    %8d\n", triangleStats->motionStaticUpCount);
  printf("  Static by DOWN motion:  %8d\n", triangleStats->motionStaticDownCount);
  printf("  Dynamic by UP motion:   %8lld\n", triangleStats->motionDynamicUpCount);
  printf("  Dynamic by DOWN motion: %8lld\n", triangleStats->motionDynamicDownCount);
  printf("--        END statistics        --\n");
  printf("\n");
  

  Ia64MotionStatsFree(twinStats);
  Ia64MotionStatsFree(triangleStats);
  
  //printf("Statistics: twin: %d; triangle: %d\n", twinCount, triangleCount);
}






t_bool Ia64ReduceNopsIfTwin(t_section *code, t_cfg *cfg, t_function *i_fun, t_bbl *i_bbl, t_ia64_motion_stats *stats)
{
  static t_bbl *nextBbl;
  static t_cfg_edge *nextEdge;

  t_bbl *firstBbl = i_bbl;

  t_uint32 nopsFirst = 0, nopsNext = 0;


  if(!Ia64HasTwin(firstBbl, &nextBbl, &nextEdge))
    return FALSE;

  nopsFirst = Ia64CalcNopsInBbl(firstBbl);
  nopsNext = Ia64CalcNopsInBbl(nextBbl);

  if(nopsFirst == 0 && nopsNext == 0)
    return FALSE;
  
  if(nopsFirst + nopsNext < 3)
    return FALSE;

  if(BBL_NINS(firstBbl) <= 3 && BBL_NINS(nextBbl) <= 3)
    return FALSE;

  printf("First in twin: <bblcount = %lld>\n", BBL_PROFILE_INFO(firstBbl));
  Ia64BblPrint(firstBbl);
  printf("Next in twin: <bblcount = %lld>\n", BBL_PROFILE_INFO(nextBbl));
  Ia64BblPrint(nextBbl);
  printf("\n");

  stats->motionCount++;

  printf("Trying upward code motion...\n");
  Ia64ReduceWithTwins(IA64_REDUCE_NOPS, IA64_MOTION_UP, firstBbl, nextBbl, NULL, stats, NULL);

  
  printf("Trying downward code motion...\n");
  Ia64ReduceWithTwins(IA64_REDUCE_NOPS, IA64_MOTION_DOWN, firstBbl, nextBbl, NULL, stats, NULL);
  

  return TRUE;
}




t_bool Ia64ReduceNopsIfTriangle(t_section *code, t_cfg *cfg, t_function *i_fun, t_bbl *i_bbl, t_ia64_motion_stats *stats)
{
  static t_bbl *nextBbl[2];
  static t_cfg_edge *nextEdge[2];

  t_bbl *firstBbl = i_bbl;

  t_uint32 nopsFirst = 0, nopsNext1 = 0, nopsNext2 = 0;
  

  if(!Ia64HasTriangle(firstBbl, nextBbl, nextEdge))
    return FALSE;

  
  nopsFirst = Ia64CalcNopsInBbl(firstBbl);
  nopsNext1 = Ia64CalcNopsInBbl(nextBbl[0]);
  nopsNext2 = Ia64CalcNopsInBbl(nextBbl[1]);

  if(nopsFirst == 0 && nopsNext1 == 0 && nopsNext2 == 0)
    return FALSE;
  

#if 0  
  if(nopsFirst + nopsNext1 + nopsNext2 < 3)
    return FALSE;

  if(BBL_NINS(firstBbl) <= 3 && BBL_NINS(nextBbl[0]) <= 3 && BBL_NINS(nextBbl[1]) <= 3)
    return FALSE;
#endif

  printf("First in triangle: <bblcount = %lld>\n", BBL_PROFILE_INFO(firstBbl));
  Ia64BblPrint(firstBbl);
  printf("Next-TRUE in triangle: <bblcount = %lld>\n", BBL_PROFILE_INFO(nextBbl[0]));
  Ia64BblPrint(nextBbl[0]);
  printf("Next-FALSE in triangle: <bblcount = %lld>\n", BBL_PROFILE_INFO(nextBbl[1]));
  Ia64BblPrint(nextBbl[1]);
  printf("\n");

  stats->motionCount++;

//#if 0  
  printf("Trying upward code motion...\n");
  
  {
    t_ia64_pred_cmp predCmp;
    t_ins *cmp;
    
    cmp = Ia64SearchForCompare(firstBbl);

    if(cmp && (IA64INSPRED(firstBbl->ins_last) != IA64INSOUT1(cmp) && IA64INSPRED(firstBbl->ins_last) != IA64INSOUT2(cmp)))
      cmp = NULL;

    /*TODO: dit wegwerken !!!*/
    /*merk op: ook aanpassen bij nops !!!*/
    if(cmp && (IA64INSOUT1(cmp) == IA64_P0 || IA64INSOUT1(cmp) == IA64_P0))
      cmp = NULL;
    
    if(cmp)
    {
      predCmp.predTrue = IA64INSOUT1(cmp);
      predCmp.predFalse = IA64INSOUT2(cmp);

      if(predCmp.predTrue != IA64INSPRED(firstBbl->ins_last))
      {
	/*just omgekeerd dan...*/
	predCmp.predTrue = IA64INSOUT2(cmp);
	predCmp.predFalse = IA64INSOUT1(cmp);
      }
      
      /*cmp zoeken, predicaat wijzigen(als niet succesvol => evt resetten)...*/
      Ia64ReduceWithTwins(IA64_REDUCE_NOPS, IA64_MOTION_UP, firstBbl, nextBbl[0], nextBbl[1], stats, &predCmp);
    }
  }
//#endif
  
  printf("Trying downward code motion...\n");
  Ia64ReduceWithTwins(IA64_REDUCE_NOPS, IA64_MOTION_DOWN, firstBbl, nextBbl[0], nextBbl[1], stats, NULL);
  
  return TRUE;
}








void Ia64ReduceNopsByExchangingInstructions(t_section *code, t_cfg *cfg)
{
  t_function *i_fun;

  t_ia64_motion_stats *twinStats = Ia64MotionStatsAlloc();
  t_ia64_motion_stats *triangleStats = Ia64MotionStatsAlloc();
  
  t_uint32 aantalFun = 0, teller = 0;

  CFG_FOREACH_FUN(cfg, i_fun)
    aantalFun++;
  
  
  CFG_FOREACH_FUN(cfg, i_fun)
  {
    t_bbl *i_bbl, *tmp;
    t_ia64_reg freePred;

    freePred = Ia64GetFirstFreePredicateInFunction(i_fun); 

    FUN_FOREACH_BBL(i_fun, i_bbl)
    {
      if(Ia64ReduceNopsIfTwin(code, cfg, i_fun, i_bbl, twinStats))
	continue;

      if(Ia64ReduceNopsIfTriangle(code, cfg, i_fun, i_bbl, triangleStats))
	continue;
    
      /*nog een funnel-geval voorzien (trechter)*/  
    }

    teller++;
    printf("Message: %d/%d (%d%%) functions completed!\n", teller, aantalFun, 100*teller/aantalFun);
    
  }


  printf("\n");
  printf("--    Code Motion Statistics    --\n");
  printf("---       Reducing NOPs        ---\n");
  printf("Twins:                    %8d\n", twinStats->motionCount);
  printf("  Static by UP motion:    %8d\n", twinStats->motionStaticUpCount);
  printf("  Static by DOWN motion:  %8d\n", twinStats->motionStaticDownCount);
  printf("  Dynamic by UP motion:   %8lld\n", twinStats->motionDynamicUpCount);
  printf("  Dynamic by DOWN motion: %8lld\n", twinStats->motionDynamicDownCount);
  printf("Triangles:                %8d\n", triangleStats->motionCount);
  printf("  Static by UP motion:    %8d\n", triangleStats->motionStaticUpCount);
  printf("  Static by DOWN motion:  %8d\n", triangleStats->motionStaticDownCount);
  printf("  Dynamic by UP motion:   %8lld\n", triangleStats->motionDynamicUpCount);
  printf("  Dynamic by DOWN motion: %8lld\n", triangleStats->motionDynamicDownCount);
  printf("--        END statistics        --\n");
  printf("\n");
  

  Ia64MotionStatsFree(twinStats);
  Ia64MotionStatsFree(triangleStats);
  
  //printf("Statistics: twin: %d; triangle: %d\n", twinCount, triangleCount);
}






void Ia64StandardScheduling(t_section *code, t_cfg *cfg)
{
  t_uint64 bblCount = 0;

  t_bbl *i_bbl;
  
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if(BBL_NINS(i_bbl) != 0)
      bblCount ++;
  }

  Ia64StatusBarInit(bblCount, 75);
  Ia64StatusBarStart();

  __static_scheduler_code_section__ = code;

  bblCount = 0;

  /* nu nog eerst het inbundelen van de basic blocks uitvoeren */
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    if(BBL_NINS(i_bbl) != 0)
      bblCount++;

    Ia64ScheduleBbl(code, i_bbl);
    Ia64StatusBarUpdate(bblCount);
  }

  Ia64StatusBarStop();

}


void Ia64GlobalScheduling(t_section *code, t_cfg *cfg)
{
  t_uint8 stage = 1;

  __static_scheduler_code_section__ = code;

  
  if(global_options.useprofile)
    Ia64CalcMaxBblProfileCount(cfg);
  
  printf("\n");
  printf("Initializing global scheduler\n");
  printf("\n");

  

  
  printf("Entering Stage %d: Applying local scheduler to all BBLs\n", stage);
  Ia64StandardScheduling(code, cfg);
  printf("Completed Stage %d\n", stage);
  printf("\n");
  stage++;
  
#if 0
  if(global_options.useprofile)
  {

    printf("Entering Stage %d: Detecting and converting to hyperblocks\n", stage);
    Ia64CreateHyperblocks(code, cfg);
    printf("Completed Stage %d\n", stage);
    printf("\n");
    stage++;
  }
#endif  

  if(global_options.nopscheduler)
  {

    printf("Entering Stage %d: Reducing nops by exchanging instructions between BBLs\n", stage);
    Ia64ReduceNopsByExchangingInstructions(code, cfg);
    printf("Completed Stage %d\n", stage);
    printf("\n");
    stage++;
  }
  

  if(global_options.useprofile && global_options.listscheduler)
  {

    printf("Entering Stage %d: Reducing the number of cycles by exchanging instructions between BBLs\n", stage);
    Ia64ReduceCyclesByExchangingInstructions(code, cfg);
    printf("Completed Stage %d\n", stage);
    printf("\n");
    stage++;
  }
  
  
  
#if 0  
  printf("Entering Stage %d: Merging twins!\n", stage);
  Ia64MergingTwins(cfg);
  printf("Completed Stage %d\n", stage);
  printf("\n");
  stage++;
#endif  
  
  printf("Terminating global scheduler\n");
  printf("\n");
  
}


t_bool Ia64InsBetweenBblHaveDependency(t_bbl *first, t_bbl *next)
{
  t_ins *ins1, *ins2;

  t_uint32 instr_group1 = IA64INSINSTRGROUP(first->ins_last);
  t_uint32 instr_group2 = IA64INSINSTRGROUP(next->ins_first);

  BBL_FOREACH_INS_R(first, ins1)
  {
    if(IA64INSINSTRGROUP(ins1) != instr_group1)
      break;
    
    BBL_FOREACH_INS(next, ins2)
    {
      if(IA64INSINSTRGROUP(ins2) != instr_group2)
        break;

      if(Ia64InstructionsHaveDependency(ins1, ins2))
	return TRUE;
    }
  }

  return FALSE;
}



void Ia64EliminateFinalStop(t_section* code, t_cfg *cfg)
{
  t_bbl *i_bbl;
  t_uint32 cnt;

  printf("Eliminating final stops...\n");
  
  CFG_FOREACH_BBL(cfg,i_bbl)
  {
    t_bbl *nextBbl = NULL;
    t_int32 disp = 0;
    t_ins *ins;
    
    if(BBL_NINS(i_bbl) == 0)
      continue;

    /*Data: need not be scheduled*/
    if(i_bbl->ins_first && IA64INSTYPE(i_bbl->ins_first) == IA64_TYPE_DATA)
      continue;

    if(!i_bbl->ins_last)
      continue;

    /*slechts 1 uitgaande pijl*/
    if(NODE_SUCC_FIRST(i_bbl) != NODE_SUCC_LAST(i_bbl))
      continue;
    
    
    /*TODO dit evt nog verfijnen*/    
    if(/*Ia64IsBranchOrCallOrReturn(i_bbl->ins_last) || */ IA64INSTYPE(i_bbl->ins_last) == IA64_TYPE_B || Ia64IsBreak(i_bbl->ins_last))
      continue;

    nextBbl = CFG_EDGE_TAIL(T_CFG_EDGE(NODE_SUCC_FIRST(i_bbl)));

    if(BBL_NINS(nextBbl) == 0)
      continue;

    /*Data: need not be scheduled*/
    if(nextBbl->ins_first && IA64INSTYPE(nextBbl->ins_first) == IA64_TYPE_DATA)
      continue;

    if(!nextBbl->ins_first)
      continue;

    if(IA64OPCODEINFO(nextBbl->ins_first).instr_opcode == IA64_INSTR_ALLOC)
      continue;


    /*evt nog proberen afzetten*/
    if(IA64INSTYPE(nextBbl->ins_last) != IA64_TYPE_B)
      continue;
    

    if(Ia64InsBetweenBblHaveDependency(i_bbl, nextBbl))
      continue;
    

    disp = (t_uint32)IA64INSINSTRGROUP(i_bbl->ins_last) - (t_uint32)IA64INSINSTRGROUP(nextBbl->ins_first);

    if(disp != -1)
      continue;
      
/*
    if(cnt >= global_options.debugcounter)
      continue;
*/
    BBL_FOREACH_INS(nextBbl, ins)
//      IA64INSINSTRGROUP(ins)--;
      IA64INSINSTRGROUP(ins) = (t_int32)IA64INSINSTRGROUP(ins) + disp;

    if(IA64INSINSTRGROUP(i_bbl->ins_last) != IA64INSINSTRGROUP(nextBbl->ins_first))
      FATAL(("huh ?"));

    cnt++;
/*
    Ia64BblPrint(i_bbl);
    Ia64BblPrint(nextBbl);
    printf("\n");
*/    
  }  

  printf("Removed %d stops.\n", cnt);
}



void Ia64PatchEmptyRefedBbls(t_section* code, t_cfg *cfg)
{
  t_bbl *bbl;
  CFG_FOREACH_BBL(cfg, bbl)
  {
    if(BBL_NINS(bbl) == 0 && RELOCATABLE_REFED_BY(bbl))
    {
      t_ins * ins;
      t_uint64 instrgroup = IA64INSINSTRGROUP(BBL_INS_FIRST(BBL_NEXT(bbl)));

      ins = Ia64InstructionCreateNoop(code, IA64_UNIT_M);
      IA64INSINSTRGROUP(ins) = instrgroup;
      InsAppendToBbl(ins, bbl);
      
      ins = Ia64InstructionCreateNoop(code, IA64_UNIT_F);
      IA64INSINSTRGROUP(ins) = instrgroup;
      InsAppendToBbl(ins, bbl);
      
      ins = Ia64InstructionCreateNoop(code, IA64_UNIT_B);
      IA64INSINSTRGROUP(ins) = instrgroup;
      InsAppendToBbl(ins, bbl);
    }
  }
}




void Ia64ScheduleCfg(t_section* code, t_cfg *cfg)
{
  printf("Starting scheduling flowgraph\n");

  if(global_options.globalscheduler)
  {
    if(!global_options.nopscheduler && !global_options.listscheduler)
      FATAL(("Global scheduler only works with nopscheduler of listscheduler"));

#if 0
    /*dit zou evt kunnen worden uitgebreid...*/
    if(!global_options.useprofile)
      FATAL(("Global scheduler currently only works with profile-information"));
#endif    
    
    Ia64GlobalScheduling(code, cfg);
  }
  else
    Ia64StandardScheduling(code, cfg);

  /*CHANGED !!!!!!!*/
  if(global_options.nopscheduler || global_options.listscheduler)
  {
    Ia64PatchEmptyRefedBbls(code, cfg);
    Ia64EliminateFinalStop(code, cfg);
  }

  printf("Finished scheduling flowgraph\n");
}






/* vim: set shiftwidth=2: */
